-- This is a Musubi input file for benchmarking the implementation.
-- Musubi input files are Lua scripts (http://www.lua.org)
--------------------------------------------------------------------------------

-- The following three settings are the most relevant ones for
-- this benchmarking.

-- This states the number of elements to use.
-- There will be 8^ref_level elements in the simulation.
-- The following statement will get this setting from the
-- environment variable $MUS_LEVEL.
ref_level = tonumber(os.getenv('MUS_LEVEL'))

-- Number of iterations to do. This controls, how long the simulation will run.
-- If you have a higher level, you typically want less iterations, to keep
-- compute time bounded. There needs to be at least one iteration.
-- Getting this setting from the environment variable $MUS_ITER here.
-- Note, that you could also limit the execution time by setting the clock
-- to a given number of seconds in the sim_control.time_control setting
-- below, but this will introduce additional communication.
iterations = tonumber(os.getenv('MUS_ITER'))

-- How many outputs do you want to write during this benchmark?
-- In every output the complete data will be written, that would be
-- required to restart the simulation, or visualize the complete
-- result. If it is not set, or set to 0, no output will be generated.
noutputs = os.getenv('MUS_OUTPUTS')

-- The io_buffer_size specifies the amount of memory reserved for blocks of
-- temporary memory allocated for copies of data in MB.
-- (If you want to measure IO performance, higher settings here, might improve
--  the IO time.)
io_buffer_size = 8 -- default is 8 MB

-- Communication pattern to use in parallel executions
-- There 4 different communications implemented, that can be selected
-- by the commpattern setting, usually we observe best scalability with
-- isend_irecv_overlap, but with little differences to the other
-- patterns.
commpattern = 'isend_irecv' -- default is simple isend-irecv exchange
--          = 'isend_irecv_overlap' allow overlapping of isends and irecvs
--          = 'typed_isend_irecv' use a MPI datatype to describe the data
--          = 'gathered_type' minimized number of blocks

-- With the simulation name you can put a label on your simulation run, to have
-- some concise reference. The name is arbitrary and will be provided in the
-- timing.res output.
simulation_name = 'bench_L'..ref_level..'-'..commpattern

-- Print memory upon finalize?
-- This is obtained from the Linux pseudo file /proc/self/status.
-- It is only printed by the first process in a prallel run, and only provides
-- data on this first process.
printRuntimeInfo = true --default is true

-- Settings for the primary logging object provided by treelm.
-- (Root outputs to stdout)
-- The main control here is the level, which tells the logger how verbose
-- the output should be. The higher the log level, the more verbose your
-- output will be. Default is 1.
-- Use 0 to deactivate log messages in the primary logger.
-- Further settings are real_form to define the format of real numbers
-- (default = 'EN12.3') and int_form to define the format of integer
-- numbers (default = 'I0')
logging = { level = 1 }

-- simulation control settings
sim_control = { 
  -- Actual time definitions
  -- Time definitions are provided in terms of:
  -- * sim:   simulation time, this is the time measurement for the
  --          transient phenomenon that is simulated.
  -- * iter:  the number of iterations that were done
  -- * clock: the real time that has passed since the beginning of
  --          the simulation in seconds.
  -- All three measures might be set, whatever is encountered first,
  -- will trigger the setting:
  -- {sim = 1.23, iter = 123, clock=12.3}
  time_control = { 
    max = { iter = iterations }, -- maximum number of iterations to reach in the finest level
    interval = { iter = iterations } -- interval to perform checks total density
  },
  abort_criteria = { 
    stop_file = '', -- provide some filename, if it exists, simultion will stop
                    -- (can be used to stop the run, by 'touch'ing the given file)
  }
}  

--------------------------------------------------------------------------------
-- MESH DEFINITION
-- Defining a predefined mesh:
--    Available are fully periodic meshes:
--    - 'cube': a full cube, defined by origin, length and refinementLevel.
--    - 'slice': a plane out of the full cube.
--    - 'line': a line out of the full cube, if you use element_count instead
--              of refinementLevel for this mesh, an arbitrary number of
--              elements can be used for the mesh. In this case the length
--              refers to the overall length for the given number of elements
--              not the cube extent.

-- Here we define a fully periodic cube:
mesh = {
        predefined = 'cube',           -- use the predefined full cube
        origin = { -1.0, -1.0, -1.0 }, -- origin is the corner of the cube
        length = 2.0,                  -- length of the cube
        refinementLevel = ref_level    -- refinement level to resolve the cube
       }
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
-- RESTART SETTINGS (only active, if outputs requested).
if noutputs ~= nil then
  if tonumber(noutputs) > 0 then
    res_interval = math.floor(iterations/tonumber(noutputs))
    restart = {
               -- Header of restart, to start simulation from.
               -- example:
               -- read = './restart/sample_lastHeader.lua',
    
               -- Prefix for the files, that are to be written during the
               -- simulation. If this ends with a path separator, the restart files
               -- will be written into the specified directory and that directory
               -- has to exist!
               write = 'restart_', -- 'restart/' for a directory
    
               -- The points in time at which restart files are to be written is
               -- controlled by the time_control definition.
               -- Time definitions are provided in terms of:
               -- * sim:   simulation time, this is the time measurement for the
               --          transient phenomenon that is simulated.
               -- * iter:  the number of iterations that were done
               -- * clock: the real time that has passed since the beginning of
               --          the simulation in seconds.
               -- All three measures might be set, whatever is encountered first,
               -- will trigger the setting:
               -- {sim = 1.23, iter = 123, clock=12.3}
               -- If instead of a table a single number is provided, this is
               -- interpreted as the setting for sim.
               -- All not-given times are set to never.
               time_control = {
                                -- Starting point after which restart files should
                                -- be written.
                                -- Setting iter to 0 here, results in restart files
                                -- being written from the initial condition onwards.
                                min = {iter = 0},
    
                                -- The maximal point in time, up to which, restarts
                                -- should be written.
                                -- Note, that if this is not defined at all, it will
                                -- be set to never, resulting in doing restarts for
                                -- all times, including a final restart after
                                -- reaching the termination of the time loop.
                                max      = {iter = iterations},
    
                                -- Frequency at which restart files are to be
                                -- written between min and max.
                                interval = {iter = res_interval},
    
                                -- Controlling if a restart file should be written
                                -- or not involves communication, if a clock setting
                                -- is used.
                                -- This might have a negative impact on the
                                -- performance, if it is done every iteration and
                                -- a single iteration is too short.
                                -- By setting check_iter the interval at which
                                -- these checks are done, can be increased, and
                                -- thus performance impacts reduced.
                                -- CAUTION: be aware that increasing this to values
                                --          larger than 1, decreases the accuracy
                                --          of the points in time at which restarts
                                --          are written!
                                -- default:
                                -- check_iter = 1
                              }
              }
    --------------------------------------------------------------------------------
  end
end


-- Actual simulation parameters, should not be very relevant for benchmarking.
-- Global parameters
rho0 = 1.
cs2  = 1./3.

-- Define the interpolation method to use between different refinement levels
-- available are: average, eqneq, quadratic, quadraticCompact
interpolation_method = 'average' -- default: average

control_routine = 'multischeme'

scaling = 'acoustic' -- or diffusive

comm_reduced = true   -- communicate reduced set of links?

init_allElems = false -- besides fluids, initialize ghost and halo elements?
                      -- only use this feature for debugging!

init_allTimeLayers = true -- initialize all time layers

-- scheme model for single fluid simulation
scheme = {
  identify = {
      label = 'global',  -- scheme name 
      kind = 'lbm',      -- simulation type of this scheme ( lbm, lbm_incomp, lbm_ps, ...)
      relaxation = 'bgk', -- relaxation type (bgk, mrt, ...)
      layout = 'd3q19'},

  -- Define the Lattice Boltzmann fluid properties
  fluid = { omega = 1.7, rho0 = rho0 },

  -- Define initial conditions per variable.
  -- They might be constants, predefined functions (described with a table), or
  -- lua functions.
  initial_condition = { 
       pressure =  {
         predefined='gausspulse', 
         center={0.0, 0.0, 0.0}, 
         halfwidth=0.5,
         amplitude=1.20, 
         background=1.000},
       velocityX = 0.0,
       velocityY = 0.0,
       velocityZ = 0.0
  }
}
