?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Boundary condition treatment routines for fluid simulation
!!
!! A detailed description on the implementation details are given in
!! \ref boundary_implementation
!!
module mus_bc_fluid_experimental_module

  ! include treelm modules
  use env_module,               only: rk, long_k, labelLen
  use tem_param_module,         only: div1_3, cs2, cs2inv, qDirName
  use tem_spacetime_fun_module, only: tem_spacetime_for
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_var_system_module,    only: tem_var_system_type
  use tem_subTree_type_module,  only: tem_subTree_type
  use tem_debug_module,         only: dbgUnit
  use tem_logging_module,       only: logUnit
  use tem_ini_condition_module, only: tem_ini_condition_type
  use tem_math_module,          only: invert_matrix
  use tem_property_module,      only: prp_missingNeigh

  ! include musubi modules
  use mus_bc_header_module,     only: boundary_type, glob_boundary_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type 
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_pdf_module,           only: pdf_global_type
  use mus_variable_module,      only: mus_derVarPos_type
  use mus_param_module,         only: mus_param_type
  use mus_physics_module,       only: mus_physics_type
  use mus_mixture_module,       only: mus_mixture_type
  use mus_solSpecHelpers_module, only: getFieldVariable, getFieldPrefixes

  implicit none

  private

  public :: outlet_extrapolation
  public :: outlet_extrapolation_microscopic
  public :: outlet_bouzidi
  public :: test_bcNeighs
  public :: moments_inflow, moments_outflow
  public :: spc_moments_outflow
  public :: spc_bb_wall
  public :: spc_bb_vel_test

contains

!******************************************************************************!
  !> Higher order no-slip wall linear interpolation bounce back boundary 
  !! conditions.
  !! \author Manuel Hasert, Jiaxing Qi
  !!
  !! An interpolation is performed, which can exactly set the wall to a position
  !! between the fluid and the solid node. The geometric information is stored
  !! in the qVal, which gives the proportion of the wall position by the real
  !! position between fluid and solid node. \n
  !! After the paper [1]  M. Bouzidi, M. Firdaouss, and P. Lallemand,
  !! “Momentum transfer of a Boltzmann-lattice fluid with boundaries,”
  !! Physics of Fluids, vol. 13, no. 11, pp. 3452–3459, Nov. 2001 \n
  !! Equations 5 (linear interpolation). quadratic is not there yet.
  !!
  !  subroutine wall_bouzidi
?? copy :: bc_routineHeader( spc_ubb_bouzidi )
    !---------------------------------------------------------------------------
    ! fIn:        incoming link of boundary element
    ! fOut:       outgoing link of boundary element
    ! fOutNeigh:  outgoing link of neighbor of boundary element
    real(kind=rk) :: fIn, fOut, fOutNeigh
    integer :: iElem, iDir, invDir
    real(kind=rk) :: qVal
    !---------------------------------------------------------------------------

    ! Loop over all boundary elements in this level
    do iElem = 1, globBC%nElems( iLevel )
      ! Loop over all links except central one.
      do iDir = 1, layout%stencil(1)%QQN
        ! bitmask points inverse to boundary direction incoming Dir (iDir)
        ! Whereas qVal is stored in outgoing Dir ( invDir )
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then

          ! Only treat the required directions
          invDir = layout%stencil(1)%cxDirInv( iDir )
          qVal = globBC%elems(iLevel)%qVal( invDir, iElem )

          ! --------------------------------------------------------------------
          ! JQ: for PULL, should update f_outgoing( invDir ), as in kernel it
          !               will bounce back (stream)
          !     for PUSH, should update f_incoming( iDir   ), as in kernel it
          !               will stay (relax), iDir = neigh( invDir )
          !               so we should use SAVE( invDir )
          if( qVal >= 0.5_rk ) then

            fIn  = state(                                                      &
& ?SAVE?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
            fOut = state(                                                      &
& ?SAVE?( varPos(invDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)
            state(                                                             &
& ?SAVE?( varPos(invDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) = &
              &  (1._rk - 0.5_rk/qVal) * fIn + (0.5_rk/qVal) * fOut

          ! --------------------------------------------------------------------
          ! JQ: for PULL, should update f_outgoing( invDir ), as in kernel it
          !               will bounce back (stream)
          !     for PUSH, should update f_incoming( iDir   ), as in kernel it
          !               will stay (relax), iDir = neigh( invDir )
          !     so we should use SAVE( invDir ), i.e. update post-collision value
          else ! qVal < 0.5

            fOutNeigh  = state(                                                &
& ?FETCH?( varPos(invDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ), currT)
            fOut       = state(                                                &
& ?SAVE?( varPos(invDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ), currT)
            state(                                                             &
& ?SAVE?( varPos(invDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ), currT) = &
              &   (1._rk - 2._rk*qVal) * fOutNeigh + (2._rk*qVal) * fOut

          endif ! qVal > 0.5 or qVal < 0.5
        endif ! BitMask
      enddo ! iDir
    enddo ! iElem

  end subroutine spc_ubb_bouzidi
!******************************************************************************!


!******************************************************************************!
  !> species bounce back velocity boundary 
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_bb_vel_test', 
  !!    velocityX = 0.1,
  !!    velocityY = 0.0,
  !!    velocityZ = 0.0
  !!     }
  !! }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_bb_vel_test )
    !------------------------------------------------------------------------ 
!    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
!    real(kind=rk) :: fEq
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fEq( layout%stencil(1)%QQ, varSys%nVars )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: rho_spc(1), eqVel(3)
    integer :: iELem, iDir, iFieldLoc, nFields, pos, invPos
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    integer :: iNeigh, tmpDir, elemPos, fluidNeighbor
    real(kind=rk) :: mass_dens(varSys%nVars), vel(3, varSys%nVars)
    real(kind=rk) :: input_loc(4*varSys%nVars)
    real(kind=rk) :: fTmp_allNeigh( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_allNeighPre( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_Next( layout%stencil(1)%QQ*varSys%nVars )
    !real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ )
    real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ, varSys%nVars )
    real(kind=rk) :: fEqNext( layout%stencil(1)%QQ )
    real(kind=rk) :: rhoTmp(1), uxTmp(3), ucx
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel))
    !------------------------------------------------------------------------ 
!!write(dbgUnit(1),*) 'Boundary label ', trim(me%label)
!!write(dbgUnit(1),*) 'iField ', iField

    ! Get velocity which might be defined to vary in space and time
     ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    !write(*,*) 'phy Vel', ux, uy, uz
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel


    nFields = varSys%nVars
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = 1

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

!!write(dbgUnit(1),*) 'iElem ', iElem
      fTmp_all = 0.0_rk
      ! Calculate the density of current element
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  

      do iDir = 1,layout%stencil(1)%QQ
        fTmp(iDir) = bcBuffer( & 
          &?IDX?(varpos(iDir),globBC%elems(iLevel)%elemBuffer(iElem),nScalars))
      end do

!!write(dbgUnit(1),*) 'fTmp_all ', fTmp_all
      fTmp_Next = 0.0_rk 
      do iFieldLoc = 1, nFields
        call varSys%derive( derVarPosAll(iFieldLoc)%equilFromState )%do(    &
          &                 input = fTmp_all,                               &
          &                 bound = bound,                                  &
          &                 tree = tree,                                    &
          &                 subTree = subTree,                              &
          &                 requVar = varSys%Variable%val(                  &
          &                           derVarPosAll(iFieldLoc)%equilFromState ),&
          &                 globSys = varSys, stencil = layout%stencil(1),  &
          &                 res = fEq(:, iFieldLoc),                        &
          &                 conf = varSys%solSpec%conf )

        !write(*,*) 'iField ', iFieldLoc, 'local fEq ', fEq
        do iDir = 1,layout%stencil(1)%QQ
          if( globBC%elems(iLevel)%bitmask( iDir, iElem ) .or. &
            & iDir == layout%stencil(1)%restPosition ) then
            pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
            invPos = varSys%Variable%val(iFieldLoc)%varPos( &
              & layout%stencil(1)%cxDirInv(iDir) )
            !write(*,*) 'iDir ', iDir, pos, invPos  
            fTmp_Next(pos) = ( fTmp_all( invPos )                      &
              &            + mixture%omega_diff * 0.5_rk *             &
              &              fEq( layout%stencil(1)%cxDirInv(iDir), iFieldLoc ) ) &
              &            / (1.0_rk+mixture%omega_diff*0.5_rk)
          endif
        enddo
      enddo  

!!write(dbgUnit(1),*) 'local fTmp_Next ', fTmp_Next

      ! current element position in state array
      elemPos = globBC%elems(iLevel)%elem(iElem)
!!write(dbgUnit(1),*) 'elemPos', elemPos, 'treeID ', glob%levelDesc(iLevel)%total(elemPos)

      ! Transform g to f in each neighbor node
      do iNeigh = 1,layout%stencil(1)%QQN
        ! fluid neighbor in inverse direction of iNeigh (x-c_i)
        fluidNeighbor =  &
  &?ElemIDX?(neigh( ?NGPOS?( varPos( iNeigh ), elemPos ,nScalars )),nScalars)

        !if (elemPos /= fluidNeighbor) then
        if( .not. globBC%elems(iLevel)%bitmask( iNeigh, iElem ) ) then
          if (elemPos == fluidNeighbor) then
            fTmp_allNeigh =  fTmp_all 
          else  
            ! get pdf of neighbor in iNeigh
            do iFieldLoc = 1, nFields
              do iDir = 1,layout%stencil(1)%QQ
                pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
                fTmp_allNeigh(pos) =  & 
                  & state( ?SAVE?( pos, fluidNeighbor, nScalars, neigh ), currT)
              end do    
            end do !iFieldLoc
          endif

!!write(dbgUnit(1),*) 'iDir ', iNeigh, 'fluidNeighbor ', fluidNeighbor
!!write(dbgUnit(1),*) 'neighID ', glob%levelDesc(iLevel)%total(fluidNeighbor)
!!write(dbgUnit(1),*) 'fTmp_allNeigh ', fTmp_allNeigh

          ! compute equilibrium of neighbor
          fEqNeigh = 0.0_rk
          do iFieldLoc = 1, nFields
            call varSys%derive( derVarPosAll(iFieldLoc)%equilFromState )%do(    &
              &                 input = fTmp_allNeigh,                          &
              &                 bound = bound,                                  &
              &                 tree = tree,                                    &
              &                 subTree = subTree,                              &
              &                 requVar = varSys%Variable%val(                  &
              &                           derVarPosAll(iFieldLoc)%equilFromState ),&
              &                 globSys = varSys, stencil = layout%stencil(1),  &
              &                 res = fEqNeigh(:, iFieldLoc),                   &
              &                 conf = varSys%solSpec%conf )
          !write(*,*) 'iFieldLoc ', iFieldLoc, 'fEqNeigh ', fEqNeigh(:, iFieldLoc)
          !write(*,*) 'iFieldLoc ', iFieldLoc, 'sum(fEqNeigh) ', sum(fEqNeigh(:, iFieldLoc))
          end do  
  
          do iFieldLoc = 1, nFields
            pos = varSys%Variable%val(iFieldLoc)%varPos(iNeigh)
            fTmp_Next(pos) = ( fTmp_allNeigh(pos) &
              &               + mixture%omega_diff*0.5_rk*fEqNeigh(iNeigh, iFieldLoc) ) &
              &               / (1.0_rk+mixture%omega_diff*0.5_rk)
          !fTmp_Next(varSys%Variable%val(iFieldLoc)%varPos(tmpDir)) = &
          !  & ( fTmp_allNeigh(varSys%Variable%val(iFieldLoc)%varPos(iNeigh)) &
          !  &               + mixture%omega_diff*0.5_rk*fEqNeigh(iNeigh, iFieldLoc) ) &
          !  &               / (1.0_rk+mixture%omega_diff*0.5_rk)
          end do  
        endif
      end do !iNeigh
!write(dbgUnit(1),*) 'fTmp_Next ', fTmp_Next

      mass_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fTmp_Next(pos)
        enddo 
      enddo  
!write(dbgUnit(1),*) 'mass_dens ', mass_dens
!write(dbgUnit(1),*) 'vel ', ux(iElem), uy(iElem), uz(iElem)

      ! update untransformed f with velocity bounce back
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem ) ) then
!write(dbgUnit(1),*) 'iDir', iDir
          ucx = layout%stencil(1)%cxDir(1,iDir) * ux(iElem)   &
            & + layout%stencil(1)%cxDir(2,iDir) * uy(iElem)    &
            & + layout%stencil(1)%cxDir(3,iDir) * uz(iElem)   
          do iFieldLoc = 1, nFields
            pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
            fTmp_Next(pos) = fTmp_Next(pos) &
              & + layout%weight(iDir) * 2._rk * cs2inv * mass_Dens(iFieldLoc) &
              & * ucx 
          enddo  
        endif  
      enddo
!write(dbgUnit(1),*) 'after ubb fTmp_Next ', fTmp_Next

      mass_dens = 0.0_rk
      vel = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fTmp_Next(pos)
          vel(:, iFieldLoc)  = vel(:, iFieldLoc) &
            & + real(layout%stencil(1)%cxDir( :, iDir), kind=rk) * fTmp_Next(pos)
        enddo 
        !vel(:, iFieldLoc)  = (/ ux(iElem), uy(iElem), uz(iElem) /)
      enddo  
      do iFieldLoc = 1, nFields
        vel(:, iFieldLoc) = vel(:,iFieldLoc)/mass_dens(iFieldLoc)
      end do  

      input_loc = 0.0_rk
      do iFieldLoc = 1, nFields
        input_loc( iFieldLoc ) = mass_dens(iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 1 ) = vel(1, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 2 ) = vel(2, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 3 ) = vel(3, iFieldLoc)
      end do  
!write(dbgUnit(1),*) 'after ubb mass_dens ', mass_dens
!write(dbgUnit(1),*) 'after ubb vel ', vel 

      ! Calculate the equilibrium distribution
      call varSys%derive( derVarPos%equilFromMacro )%do(                         &
        &                input = input_loc,                                      &
        &                bound = bound,                                          &
        &                tree = tree,                                            &
        &                subTree = subTree,                                      &
        &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
        &                globSys = varSys,                                       &
        &                stencil = layout%stencil(1),                            &
        &                conf = varSys%solSpec%conf,                             & 
        &                res = fEqNext )

!write(dbgUnit(1),*) 'fEqNext ', fEqNext  

      ! density
      call varSys%derive( derVarPos%density )%do(        &
        &                 input = fTmp_all,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%density ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = rho_spc,                             &
        &                 conf = varSys%solSpec%conf )

      ! equilibrium velocity
      call varSys%derive( derVarPos%equilvelFromState )%do(        &
        &                 input = fTmp_Next,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%equilvelFromState ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = eqVel,                             &
        &                 conf = varSys%solSpec%conf )


      ! set boundary 
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
&?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
              !! simple Bounce Back
              !& = fTmp(layout%stencil(1)%cxDirInv( iDir )) 

              ! using f in next time step 
              & = fTmp_Next(varPos(iDir)) &
              & + mixture%omega_diff*0.5_rk*( fTmp_Next(varPos(iDir)) &
              & - fEqNext(iDir) )

              ! Bounce back with correction term for g with fEq in next time step 
              !& = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              !& + mixture%omega_diff*0.5_rk*(fEq(layout%stencil(1)%cxDirInv( iDir ), iField)&
              !& - fEqNext(iDir) )

              !& = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              !& - layout%weight( iDir )&
              !& * mixture%omega_diff*cs2inv*rho_spc(1)                        &
              !&    * ( layout%stencil(1)%cxDir( 1, iDir )*eqVel(1)            &
              !&    +   layout%stencil(1)%cxDir( 2, iDir )*eqVel(2)            &
              !&    +   layout%stencil(1)%cxDir( 3, iDir )*eqVel(3))

        endif
      enddo
      endif  
    enddo

!KM!    ! debug output
!KM!    if (iField == nFields) then
!KM!    do iElem = 1, globBC%nElems( iLevel )
!KM!      !ux = 0.0_rk
!KM!      do iFieldLoc = 1, nFields
!KM!        do iDir = 1,layout%stencil(1)%QQ
!KM!          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
!KM!          fTmp_all(pos) = state(                                                  &
!KM!&?FETCH?( pos, globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
!KM!        & currT)
!KM!        end do
!KM!      end do  
!KM!      bound(1)=iElem
!KM!      bound(2)=iElem
!KM!      write(dbgUnit(1),*) 'iElem ', iElem
!KM!      write(dbgUnit(1),*) 'fTmp_all ', fTmp_all
!KM!      ! density
!KM!      do iFieldLoc = 1, nFields
!KM!        call varSys%derive( derVarPosAll(iFieldLoc)%density )%do(        &
!KM!          &                 input = fTmp_all,                               &
!KM!          &                 bound = bound,                                  &
!KM!          &                 tree = tree,                                    &
!KM!          &                 subTree = subTree,                              &
!KM!          &                 requVar = varSys%Variable%val(                  &
!KM!          &                           derVarPosAll(iFieldLoc)%density ), &
!KM!          &                 globSys = varSys, stencil = layout%stencil(1),  &
!KM!          &                 res = rhoTmp,                             &
!KM!          &                 conf = varSys%solSpec%conf )
!KM!
!KM!        call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
!KM!          &                 input   = fTmp_all,                                  &
!KM!          &                 bound   = bound,                                     &
!KM!          &                 tree    = tree,                                      &
!KM!          &                 subTree = subTree,                                   &
!KM!          &                 requVar = varSys%Variable%val(&
!KM!          &                           derVarPosAll(iFieldLoc)%velocity),   &
!KM!          &                 globSys = varSys,                                    &
!KM!          &                 stencil = layout%stencil(1),                         &
!KM!          &                 conf    = varSys%solSpec%conf,                       &
!KM!          &                 res     = uxTmp )
!KM!        write(dbgUnit(1),*) 'iField ', iFieldLoc, 'density ', rhoTmp
!KM!        write(dbgUnit(1),*) 'iField ', iFieldLoc, 'velocity ', uxTmp
!KM!      end do  
!KM!    end do
!KM!    endif
  end subroutine spc_bb_vel_test
!******************************************************************************!

!******************************************************************************!
  !> species bounce back wall boundary 
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_bb_wall', 
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_bb_wall )
    !------------------------------------------------------------------------ 
!    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
!    real(kind=rk) :: fEq
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fEq( layout%stencil(1)%QQ, varSys%nVars )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: rho_spc(1), eqVel(3)
    integer :: iELem, iDir, iFieldLoc, nFields, pos, invPos
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    integer :: iNeigh, tmpDir, elemPos, fluidNeighbor
    real(kind=rk) :: mass_dens(varSys%nVars), vel(3, varSys%nVars)
    real(kind=rk) :: input_loc(4*varSys%nVars)
    real(kind=rk) :: fTmp_allNeigh( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_allNeighPre( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_Next( layout%stencil(1)%QQ*varSys%nVars )
    !real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ )
    real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ, varSys%nVars )
    real(kind=rk) :: fEqNext( layout%stencil(1)%QQ )
    real(kind=rk) :: rhoTmp(1), uxTmp(3)
    !------------------------------------------------------------------------ 
!write(dbgUnit(1),*) 'Boundary label ', trim(me%label)
!write(dbgUnit(1),*) 'iField ', iField
    nFields = varSys%nVars
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = 1

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

!write(dbgUnit(1),*) 'iElem ', iElem
      fTmp_all = 0.0_rk
      ! Calculate the density of current element
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  

      do iDir = 1,layout%stencil(1)%QQ
        fTmp(iDir) = bcBuffer( & 
          &?IDX?(varpos(iDir),globBC%elems(iLevel)%elemBuffer(iElem),nScalars))
      end do

!write(dbgUnit(1),*) 'fTmp_all ', fTmp_all
      fTmp_Next = 0.0_rk 
      do iFieldLoc = 1, nFields
        call varSys%derive( derVarPosAll(iFieldLoc)%equilFromState )%do(    &
          &                 input = fTmp_all,                               &
          &                 bound = bound,                                  &
          &                 tree = tree,                                    &
          &                 subTree = subTree,                              &
          &                 requVar = varSys%Variable%val(                  &
          &                           derVarPosAll(iFieldLoc)%equilFromState ),&
          &                 globSys = varSys, stencil = layout%stencil(1),  &
          &                 res = fEq(:, iFieldLoc),                        &
          &                 conf = varSys%solSpec%conf )
!write(dbgUnit(1),*) 'iFieldLoc ', iFieldLoc, 'fEqLocal ', fEq(:, iFieldLoc)

        !write(*,*) 'iField ', iFieldLoc, 'local fEq ', fEq
        do iDir = 1,layout%stencil(1)%QQ
          if( globBC%elems(iLevel)%bitmask( iDir, iElem ) .or. &
            & iDir == layout%stencil(1)%restPosition ) then
            pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
            invPos = varSys%Variable%val(iFieldLoc)%varPos( &
              & layout%stencil(1)%cxDirInv(iDir) )
            !write(*,*) 'iDir ', iDir, pos, invPos  
            fTmp_Next(pos) = ( fTmp_all( invPos )                      &
              &            + mixture%omega_diff * 0.5_rk *             &
              &              fEq( layout%stencil(1)%cxDirInv(iDir), iFieldLoc ) ) &
              &            / (1.0_rk+mixture%omega_diff*0.5_rk)
          endif
        enddo
      enddo  

!write(dbgUnit(1),*) 'local fTmp_Next ', fTmp_Next

      ! current element position in state array
      elemPos = globBC%elems(iLevel)%elem(iElem)
!write(dbgUnit(1),*) 'elemPos', elemPos, 'treeID ', glob%levelDesc(iLevel)%total(elemPos)

      ! Transform g to f in each neighbor node
      do iNeigh = 1,layout%stencil(1)%QQN
        ! fluid neighbor in inverse direction of iNeigh (x-c_i)
        fluidNeighbor =  &
  &?ElemIDX?(neigh( ?NGPOS?( varPos( iNeigh ), elemPos ,nScalars )),nScalars)

        !if (elemPos /= fluidNeighbor) then
        if( .not. globBC%elems(iLevel)%bitmask( iNeigh, iElem ) ) then
          if (elemPos == fluidNeighbor) then
            fTmp_allNeigh =  fTmp_all 
          else  
            ! get pdf of neighbor in iNeigh
            do iFieldLoc = 1, nFields
              do iDir = 1,layout%stencil(1)%QQ
                pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
                fTmp_allNeigh(pos) =  & 
                  & state( ?SAVE?( pos, fluidNeighbor, nScalars, neigh ), currT)
              end do    
            end do !iFieldLoc
          endif

!write(dbgUnit(1),*) 'iDir ', iNeigh, 'fluidNeighbor ', fluidNeighbor
!write(dbgUnit(1),*) 'neighID ', glob%levelDesc(iLevel)%total(fluidNeighbor)
!write(dbgUnit(1),*) 'fTmp_allNeigh ', fTmp_allNeigh

          ! compute equilibrium of neighbor
          fEqNeigh = 0.0_rk
          do iFieldLoc = 1, nFields
            call varSys%derive( derVarPosAll(iFieldLoc)%equilFromState )%do(    &
              &                 input = fTmp_allNeigh,                          &
              &                 bound = bound,                                  &
              &                 tree = tree,                                    &
              &                 subTree = subTree,                              &
              &                 requVar = varSys%Variable%val(                  &
              &                           derVarPosAll(iFieldLoc)%equilFromState ),&
              &                 globSys = varSys, stencil = layout%stencil(1),  &
              &                 res = fEqNeigh(:, iFieldLoc),                   &
              &                 conf = varSys%solSpec%conf )
!write(dbgUnit(1),*) 'iFieldLoc ', iFieldLoc, 'fEqNeigh ', fEqNeigh(:, iFieldLoc)
!write(dbgUnit(1),*) 'iFieldLoc ', iFieldLoc, 'sum(fEqNeigh) ', sum(fEqNeigh(:, iFieldLoc))
          end do  
  
          do iFieldLoc = 1, nFields
            pos = varSys%Variable%val(iFieldLoc)%varPos(iNeigh)
            !fTmp_Next(pos) = fTmp_allNeigh(pos) 
            fTmp_Next(pos) = ( fTmp_allNeigh(pos) &
              &               + mixture%omega_diff*0.5_rk*fEqNeigh(iNeigh, iFieldLoc) ) &
              &               / (1.0_rk+mixture%omega_diff*0.5_rk)
          !fTmp_Next(varSys%Variable%val(iFieldLoc)%varPos(tmpDir)) = &
          !  & ( fTmp_allNeigh(varSys%Variable%val(iFieldLoc)%varPos(iNeigh)) &
          !  &               + mixture%omega_diff*0.5_rk*fEqNeigh(iNeigh, iFieldLoc) ) &
          !  &               / (1.0_rk+mixture%omega_diff*0.5_rk)
          end do  
        endif
      end do !iNeigh
!write(dbgUnit(1),*) 'fTmp_Next ', fTmp_Next

      mass_dens = 0.0_rk
      vel = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fTmp_Next(pos)
          vel(:, iFieldLoc)  = vel(:, iFieldLoc) &
            & + real(layout%stencil(1)%cxDir( :, iDir), kind=rk) * fTmp_Next(pos)
        enddo 
      enddo  
      do iFieldLoc = 1, nFields
        vel(:, iFieldLoc) = vel(:,iFieldLoc)/mass_dens(iFieldLoc)
      end do  

      input_loc = 0.0_rk
      do iFieldLoc = 1, nFields
        input_loc( iFieldLoc ) = mass_dens(iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 1 ) = vel(1, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 2 ) = vel(2, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 3 ) = vel(3, iFieldLoc)
      end do  
!write(dbgUnit(1),*) 'mass_dens ', mass_dens
!write(dbgUnit(1),*) 'vel ', vel 

      ! Calculate the equilibrium distribution
      call varSys%derive( derVarPos%equilFromMacro )%do(                         &
        &                input = input_loc,                                      &
        &                bound = bound,                                          &
        &                tree = tree,                                            &
        &                subTree = subTree,                                      &
        &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
        &                globSys = varSys,                                       &
        &                stencil = layout%stencil(1),                            &
        &                conf = varSys%solSpec%conf,                             & 
        &                res = fEqNext )

!write(dbgUnit(1),*) 'fEqNext ', fEqNext  

      ! density
      call varSys%derive( derVarPos%density )%do(        &
        &                 input = fTmp_all,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%density ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = rho_spc,                             &
        &                 conf = varSys%solSpec%conf )

      ! equilibrium velocity
      call varSys%derive( derVarPos%equilvelFromState )%do(        &
        &                 input = fTmp_Next,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%equilvelFromState ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = eqVel,                             &
        &                 conf = varSys%solSpec%conf )


      ! set boundary 
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
&?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
              !! simple Bounce Back
              !& = fTmp(layout%stencil(1)%cxDirInv( iDir )) 

              ! using f in next time step 
              & = fTmp_Next(varPos(iDir)) &
              & + mixture%omega_diff*0.5_rk*( fTmp_Next(varPos(iDir)) &
              & - fEqNext(iDir) )

              ! Bounce back with correction term for g with fEq in next time step 
              !& = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              !& + mixture%omega_diff*0.5_rk*(fEq(layout%stencil(1)%cxDirInv( iDir ), iField)&
              !& - fEqNext(iDir) )

              !& = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              !& + mixture%omega_diff*0.5_rk*(fEq(layout%stencil(1)%cxDirInv( iDir ), iField)&
              !& - fEq(iDir, iField) )

              !& = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              !& - layout%weight( iDir )&
              !& * mixture%omega_diff*cs2inv*rho_spc(1)                        &
              !&    * ( layout%stencil(1)%cxDir( 1, iDir )*eqVel(1)            &
              !&    +   layout%stencil(1)%cxDir( 2, iDir )*eqVel(2)            &
              !&    +   layout%stencil(1)%cxDir( 3, iDir )*eqVel(3))

        endif
      enddo
      endif  
    enddo

    ! debug output
!KM!    if (iField == nFields) then
!KM!    do iElem = 1, globBC%nElems( iLevel )
!KM!      !ux = 0.0_rk
!KM!      do iFieldLoc = 1, nFields
!KM!        do iDir = 1,layout%stencil(1)%QQ
!KM!          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
!KM!          fTmp_all(pos) = state(                                                  &
!KM!&?FETCH?( pos, globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
!KM!        & currT)
!KM!        end do
!KM!      end do  
!KM!      bound(1)=iElem
!KM!      bound(2)=iElem
!KM!!write(dbgUnit(1),*) 'iElem ', iElem
!KM!!write(dbgUnit(1),*) 'fTmp_all ', fTmp_all
!KM!      ! density
!KM!      do iFieldLoc = 1, nFields
!KM!        call varSys%derive( derVarPosAll(iFieldLoc)%density )%do(        &
!KM!          &                 input = fTmp_all,                               &
!KM!          &                 bound = bound,                                  &
!KM!          &                 tree = tree,                                    &
!KM!          &                 subTree = subTree,                              &
!KM!          &                 requVar = varSys%Variable%val(                  &
!KM!          &                           derVarPosAll(iFieldLoc)%density ), &
!KM!          &                 globSys = varSys, stencil = layout%stencil(1),  &
!KM!          &                 res = rhoTmp,                             &
!KM!          &                 conf = varSys%solSpec%conf )
!KM!
!KM!        call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
!KM!          &                 input   = fTmp_all,                                  &
!KM!          &                 bound   = bound,                                     &
!KM!          &                 tree    = tree,                                      &
!KM!          &                 subTree = subTree,                                   &
!KM!          &                 requVar = varSys%Variable%val(&
!KM!          &                           derVarPosAll(iFieldLoc)%velocity),   &
!KM!          &                 globSys = varSys,                                    &
!KM!          &                 stencil = layout%stencil(1),                         &
!KM!          &                 conf    = varSys%solSpec%conf,                       &
!KM!          &                 res     = uxTmp )
!KM!!write(dbgUnit(1),*) 'iField ', iFieldLoc, 'density ', rhoTmp
!KM!!write(dbgUnit(1),*) 'iField ', iFieldLoc, 'velocity ', uxTmp
!KM!      end do  
!KM!    end do
!KM!    endif
  end subroutine spc_bb_wall
!******************************************************************************!


!******************************************************************************!
  !> species bounce back wall boundary 
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_bb_wall', 
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_bb_wall3 )
    !------------------------------------------------------------------------ 
!    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
!    real(kind=rk) :: fEq
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fEq( layout%stencil(1)%QQ, varSys%nVars )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: rho_spc(1), eqVel(3)
    integer :: iELem, iDir, iFieldLoc, nFields, pos, invPos
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    integer :: iNeigh, tmpDir, elemPos, fluidNeighbor
    real(kind=rk) :: mass_dens(varSys%nVars), vel(3, varSys%nVars)
    real(kind=rk) :: input_loc(4*varSys%nVars)
    real(kind=rk) :: fTmp_allNeigh( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_allNeighPre( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_Next( layout%stencil(1)%QQ*varSys%nVars )
    !real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ )
    real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ, varSys%nVars )
    real(kind=rk) :: fEqNext( layout%stencil(1)%QQ )
    real(kind=rk) :: rhoTmp(1), uxTmp(3)
    !------------------------------------------------------------------------ 
    write(dbgUnit(1),*) 'Boundary label ', trim(me%label)
    write(dbgUnit(1),*) 'iField ', iField
    nFields = varSys%nVars
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = 1

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      write(dbgUnit(1),*) 'iElem ', iElem
      fTmp_all = 0.0_rk
      ! Calculate the density of current element
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  

      do iDir = 1,layout%stencil(1)%QQ
        fTmp(iDir) = bcBuffer( & 
          &?IDX?(varpos(iDir),globBC%elems(iLevel)%elemBuffer(iElem),nScalars))
      end do

      write(dbgUnit(1),*) 'fTmp_all ', fTmp_all
      fTmp_Next = 0.0_rk 
      do iFieldLoc = 1, nFields
        call varSys%derive( derVarPosAll(iFieldLoc)%equilFromState )%do(    &
          &                 input = fTmp_all,                               &
          &                 bound = bound,                                  &
          &                 tree = tree,                                    &
          &                 subTree = subTree,                              &
          &                 requVar = varSys%Variable%val(                  &
          &                           derVarPosAll(iFieldLoc)%equilFromState ),&
          &                 globSys = varSys, stencil = layout%stencil(1),  &
          &                 res = fEq(:, iFieldLoc),                        &
          &                 conf = varSys%solSpec%conf )

        !write(*,*) 'iField ', iFieldLoc, 'local fEq ', fEq
        do iDir = 1,layout%stencil(1)%QQ
          if( globBC%elems(iLevel)%bitmask( iDir, iElem ) .or. &
            & iDir == layout%stencil(1)%restPosition ) then
            pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
            invPos = varSys%Variable%val(iFieldLoc)%varPos( &
              & layout%stencil(1)%cxDirInv(iDir) )
            !write(*,*) 'iDir ', iDir, pos, invPos  
            fTmp_Next(pos) = ( fTmp_all( invPos )                      &
              &            + mixture%omega_diff * 0.5_rk *             &
              &              fEq( layout%stencil(1)%cxDirInv(iDir), iFieldLoc ) ) &
              &            / (1.0_rk+mixture%omega_diff*0.5_rk)
          endif
        enddo
      enddo  

      write(dbgUnit(1),*) 'local fTmp_Next ', fTmp_Next

      ! current element position in state array
      elemPos = globBC%elems(iLevel)%elem(iElem)
      write(dbgUnit(1),*) 'elemPos', elemPos, 'treeID ', glob%levelDesc(iLevel)%total(elemPos)

      ! Transform g to f in each neighbor node
      do iNeigh = 1,layout%stencil(1)%QQN
        ! fluid neighbor in inverse direction of iNeigh (x-c_i)
        fluidNeighbor =  &
  &?ElemIDX?(neigh( ?NGPOS?( varPos( iNeigh ), elemPos ,nScalars )),nScalars)

        !if (elemPos /= fluidNeighbor) then
        if( .not. globBC%elems(iLevel)%bitmask( iNeigh, iElem ) ) then
          if (elemPos == fluidNeighbor) then
            fTmp_allNeigh =  fTmp_all 
          else  
            ! get pdf of neighbor in iNeigh
            do iFieldLoc = 1, nFields
              do iDir = 1,layout%stencil(1)%QQ
                pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
                fTmp_allNeigh(pos) =  & 
                  & state( ?SAVE?( pos, fluidNeighbor, nScalars, neigh ), currT)
              end do    
            end do !iFieldLoc
          endif

          write(dbgUnit(1),*) 'iDir ', iNeigh, 'fluidNeighbor ', fluidNeighbor
          write(dbgUnit(1),*) 'neighID ', glob%levelDesc(iLevel)%total(fluidNeighbor)
          write(dbgUnit(1),*) 'fTmp_allNeigh ', fTmp_allNeigh

          ! compute equilibrium of neighbor
          fEqNeigh = 0.0_rk
          do iFieldLoc = 1, nFields
            call varSys%derive( derVarPosAll(iFieldLoc)%equilFromState )%do(    &
              &                 input = fTmp_allNeigh,                          &
              &                 bound = bound,                                  &
              &                 tree = tree,                                    &
              &                 subTree = subTree,                              &
              &                 requVar = varSys%Variable%val(                  &
              &                           derVarPosAll(iFieldLoc)%equilFromState ),&
              &                 globSys = varSys, stencil = layout%stencil(1),  &
              &                 res = fEqNeigh(:, iFieldLoc),                   &
              &                 conf = varSys%solSpec%conf )
          write(dbgUnit(1),*) 'iFieldLoc ', iFieldLoc, 'fEqNeigh ', fEqNeigh(:, iFieldLoc)
          write(dbgUnit(1),*) 'iFieldLoc ', iFieldLoc, 'sum(fEqNeigh) ', sum(fEqNeigh(:, iFieldLoc))
          end do  
  
          do iFieldLoc = 1, nFields
            pos = varSys%Variable%val(iFieldLoc)%varPos(iNeigh)
            fTmp_Next(pos) = fTmp_allNeigh(pos) 
            !fTmp_Next(pos) = ( fTmp_allNeigh(pos) &
            !  &               + mixture%omega_diff*0.5_rk*fEqNeigh(iNeigh, iFieldLoc) ) &
            !  &               / (1.0_rk+mixture%omega_diff*0.5_rk)
          !fTmp_Next(varSys%Variable%val(iFieldLoc)%varPos(tmpDir)) = &
          !  & ( fTmp_allNeigh(varSys%Variable%val(iFieldLoc)%varPos(iNeigh)) &
          !  &               + mixture%omega_diff*0.5_rk*fEqNeigh(iNeigh, iFieldLoc) ) &
          !  &               / (1.0_rk+mixture%omega_diff*0.5_rk)
          end do  
        endif
      end do !iNeigh
      write(dbgUnit(1),*) 'fTmp_Next ', fTmp_Next

      mass_dens = 0.0_rk
      vel = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fTmp_Next(pos)
          vel(:, iFieldLoc)  = vel(:, iFieldLoc) &
            & + real(layout%stencil(1)%cxDir( :, iDir), kind=rk) * fTmp_Next(pos)
        enddo 
      enddo  
      do iFieldLoc = 1, nFields
        vel(:, iFieldLoc) = vel(:,iFieldLoc)/mass_dens(iFieldLoc)
      end do  

      input_loc = 0.0_rk
      do iFieldLoc = 1, nFields
        input_loc( iFieldLoc ) = mass_dens(iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 1 ) = vel(1, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 2 ) = vel(2, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 3 ) = vel(3, iFieldLoc)
      end do  
      write(dbgUnit(1),*) 'mass_dens ', mass_dens
      write(dbgUnit(1),*) 'vel ', vel 

      ! Calculate the equilibrium distribution
      call varSys%derive( derVarPos%equilFromMacro )%do(                         &
        &                input = input_loc,                                      &
        &                bound = bound,                                          &
        &                tree = tree,                                            &
        &                subTree = subTree,                                      &
        &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
        &                globSys = varSys,                                       &
        &                stencil = layout%stencil(1),                            &
        &                conf = varSys%solSpec%conf,                             & 
        &                res = fEqNext )

      write(dbgUnit(1),*) 'fEqNext ', fEqNext  

      ! density
      call varSys%derive( derVarPos%density )%do(        &
        &                 input = fTmp_all,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%density ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = rho_spc,                             &
        &                 conf = varSys%solSpec%conf )

      ! equilibrium velocity
      call varSys%derive( derVarPos%equilvelFromState )%do(        &
        &                 input = fTmp_Next,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%equilvelFromState ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = eqVel,                             &
        &                 conf = varSys%solSpec%conf )


      ! set boundary 
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
&?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
              !! simple Bounce Back
              !& = fTmp(layout%stencil(1)%cxDirInv( iDir )) 

              ! using f in next time step 
              !& = fTmp_Next(varPos(iDir)) &
              !& + mixture%omega_diff*0.5_rk*( fTmp_Next(varPos(iDir)) &
              !& - fEqNext(iDir) )

              ! Bounce back with correction term for g with fEq in next time step 
              & = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              & + mixture%omega_diff*0.5_rk*(fEq(layout%stencil(1)%cxDirInv( iDir ), iField)&
              & - fEqNext(iDir) )

              !& = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              !& + mixture%omega_diff*0.5_rk*(fEq(layout%stencil(1)%cxDirInv( iDir ), iField)&
              !& - fEq(iDir, iField) )

              !& = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              !& - layout%weight( iDir )&
              !& * mixture%omega_diff*cs2inv*rho_spc(1)                        &
              !&    * ( layout%stencil(1)%cxDir( 1, iDir )*eqVel(1)            &
              !&    +   layout%stencil(1)%cxDir( 2, iDir )*eqVel(2)            &
              !&    +   layout%stencil(1)%cxDir( 3, iDir )*eqVel(3))

        endif
      enddo
      endif  
    enddo

    ! debug output
    if (iField == nFields) then
    do iElem = 1, globBC%nElems( iLevel )
      !ux = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all(pos) = state(                                                  &
&?FETCH?( pos, globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
        & currT)
        end do
      end do  
      bound(1)=iElem
      bound(2)=iElem
      write(dbgUnit(1),*) 'iElem ', iElem
      write(dbgUnit(1),*) 'fTmp_all ', fTmp_all
      ! density
      do iFieldLoc = 1, nFields
        call varSys%derive( derVarPosAll(iFieldLoc)%density )%do(        &
          &                 input = fTmp_all,                               &
          &                 bound = bound,                                  &
          &                 tree = tree,                                    &
          &                 subTree = subTree,                              &
          &                 requVar = varSys%Variable%val(                  &
          &                           derVarPosAll(iFieldLoc)%density ), &
          &                 globSys = varSys, stencil = layout%stencil(1),  &
          &                 res = rhoTmp,                             &
          &                 conf = varSys%solSpec%conf )

        call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
          &                 input   = fTmp_all,                                  &
          &                 bound   = bound,                                     &
          &                 tree    = tree,                                      &
          &                 subTree = subTree,                                   &
          &                 requVar = varSys%Variable%val(&
          &                           derVarPosAll(iFieldLoc)%velocity),   &
          &                 globSys = varSys,                                    &
          &                 stencil = layout%stencil(1),                         &
          &                 conf    = varSys%solSpec%conf,                       &
          &                 res     = uxTmp )
        write(dbgUnit(1),*) 'iField ', iFieldLoc, 'density ', rhoTmp
        write(dbgUnit(1),*) 'iField ', iFieldLoc, 'velocity ', uxTmp
      end do  
    end do
    endif
  end subroutine spc_bb_wall3
!******************************************************************************!

!******************************************************************************!
?? copy :: bc_routineHeader( spc_bb_wall2 )
    !------------------------------------------------------------------------ 
!    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
!    real(kind=rk) :: fEq
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fEq( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: rho_spc(1), eqVel(3)
    integer :: iELem, iDir, iFieldLoc, nFields, pos
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    integer :: iDir_2, elemPos, fluidNeighbor
    real(kind=rk) :: mass_dens(varSys%nVars), vel(3, varSys%nVars)
    real(kind=rk) :: input_loc(4*varSys%nVars)
    real(kind=rk) :: fTmp_allNeigh( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_allNeighPre( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fTmp_Next( layout%stencil(1)%QQ*varSys%nVars )
    !real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ )
    real(kind=rk) :: fEqNeigh( layout%stencil(1)%QQ, varSys%nVars )
    real(kind=rk) :: fEqNext( layout%stencil(1)%QQ )
    real(kind=rk) :: rhoTmp(1), uxTmp(3)
    !------------------------------------------------------------------------ 
    write(*,*) 'Boundary label ', trim(me%label)
    write(*,*) 'iField ', iField
    nFields = varSys%nVars

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      fTmp_all = 0.0_rk
      ! Calculate the density of current element
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  

      do iDir = 1,layout%stencil(1)%QQ
        fTmp(iDir) = bcBuffer( & 
          &?IDX?(varpos(iDir),globBC%elems(iLevel)%elemBuffer(iElem),nScalars))
      end do

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      ! density
      call varSys%derive( derVarPos%density )%do(        &
        &                 input = fTmp_all,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%density ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = rho_spc,                             &
        &                 conf = varSys%solSpec%conf )

      ! equilibrium velocity
      call varSys%derive( derVarPos%equilvelFromState )%do(        &
        &                 input = fTmp_all,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%equilvelFromState ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = eqVel,                             &
        &                 conf = varSys%solSpec%conf )

      call varSys%derive( derVarPos%equilFromState )%do(        &
        &                 input = fTmp_all,                               &
        &                 bound = bound,                                  &
        &                 tree = tree,                                    &
        &                 subTree = subTree,                              &
        &                 requVar = varSys%Variable%val(                  &
        &                           derVarPos%equilFromState ), &
        &                 globSys = varSys, stencil = layout%stencil(1),  &
        &                 res = fEq,                             &
        &                 conf = varSys%solSpec%conf )

      write(*,*) 'iElem', iElem
      write(*,*) 'density ', rho_spc
      !write(*,*) 'eqVel ', eqVel
      elemPos = globBC%elems(iLevel)%elem(iElem)
      !write(*,*) 'currID ',glob%levelDesc(iLevel)%total(elemPos) 

      fTmp_Next = 0.0_rk
      do iDir_2 = 1,layout%stencil(1)%QQ
        fluidNeighbor =  &
  !&?ElemIDX?(neigh( ?NGPOS?( varPos( layout%stencil(1)%cxDirInv( iDir_2 ) ), elemPos ,nScalars )),nScalars)
  &?ElemIDX?(neigh( ?NGPOS?( varPos( iDir_2 ), elemPos ,nScalars )),nScalars)
        write(*,*) 'iDir ', iDir_2, 'fluidNeighbor ', fluidNeighbor
        !write(*,*) 'neighID ',glob%levelDesc(iLevel)%total(fluidNeighbor) 
  
        do iFieldLoc = 1, nFields
          do iDir = 1,layout%stencil(1)%QQ
            pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
            !fTmp_allNeigh(pos) =  & 
            !  & state( ?SAVE?( pos, fluidNeighbor, nScalars, neigh ), currT)
            if (elemPos == fluidNeighbor) then
              fTmp_allNeigh(pos) =  & 
                & bcBuffer( ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
            else    
              fTmp_allNeigh(pos) =  & 
                & state( ?SAVE?( pos, fluidNeighbor, nScalars, neigh ), currT)
            endif

            fTmp_allNeighPre(pos) =  & 
              & state( ?FETCH?( pos, fluidNeighbor, nScalars, neigh ), currT)
          end do    
        end do !iFieldLoc

        write(*,*) 'fTmp_allNeigh ', fTmp_allNeigh

        !mass_dens = 0.0_rk
        !vel = 0.0_rk
        !do iFieldLoc = 1, nFields
        !  do iDir = 1,layout%stencil(1)%QQ
        !    pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
        !    mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fTmp_allNeigh(pos)
        !    vel(:, iFieldLoc)  = vel(:, iFieldLoc) &
        !      & + real(layout%stencil(1)%cxDir( :, iDir), kind=rk) * fTmp_allNeigh(pos)
        !  enddo 
        !enddo  
        !do iFieldLoc = 1, nFields
        !  vel(:, iFieldLoc) = vel(:,iFieldLoc)/mass_dens(iFieldLoc)
        !end do  
        !write(*,*) 'neigh mass_dens ', mass_dens
        !write(*,*) 'neigh vel ', vel 

        fEqNeigh = 0.0_rk
        do iFieldLoc = 1, nFields
          call varSys%derive( derVarPosAll(iFieldLoc)%equilFromState )%do(    &
            &                 input = fTmp_allNeigh,                          &
            &                 bound = bound,                                  &
            &                 tree = tree,                                    &
            &                 subTree = subTree,                              &
            &                 requVar = varSys%Variable%val(                  &
            &                           derVarPosAll(iFieldLoc)%equilFromState ),&
            &                 globSys = varSys, stencil = layout%stencil(1),  &
            &                 res = fEqNeigh(:, iFieldLoc),                   &
            &                 conf = varSys%solSpec%conf )
          write(*,*) 'iFieldLoc ', iFieldLoc, 'fEqNeigh ', fEqNeigh(:, iFieldLoc)
          write(*,*) 'iFieldLoc ', iFieldLoc, 'sum(fEqNeigh) ', sum(fEqNeigh(:, iFieldLoc))
        end do  
        stop

        do iFieldLoc = 1, nFields
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir_2)
          !fTmp_Next(pos) = ( fTmp_allNeigh(pos))
          !fTmp_Next(pos) = fEqNeigh(iDir_2)
          !fTmp_Next(pos) = ( fTmp_allNeigh(pos)+fEqNeigh(iDir_2))*0.5_rk
          fTmp_Next(pos) = ( fTmp_allNeigh(pos) &
            &               + mixture%omega_diff*0.5_rk*fEqNeigh(iDir_2, iFieldLoc) ) &
            &               / (1.0_rk+mixture%omega_diff*0.5_rk)
          !write(*,*) 'iFieldLoc ', iFieldLoc, 'fTmp_Next ', fTmp_Next(pos)

          !mass_dens = 0.0_rk
          !vel = 0.0_rk
          !do iDir = 1,layout%stencil(1)%QQ
          !  mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fEqNeigh(iDir)
          !  vel(:, iFieldLoc)  = vel(:, iFieldLoc) &
          !    & + real(layout%stencil(1)%cxDir( :, iDir), kind=rk) * fEqNeigh(iDir)
          !enddo 
          !vel(:, iFieldLoc) = vel(:,iFieldLoc)/mass_dens(iFieldLoc)
          !write(*,*) 'neigh Equil mass_dens ', mass_dens(iFieldLoc)
          !write(*,*) 'neigh Equil vel ', vel(:, iFieldLoc) 

        end do ! iFieldLoc   
      end do ! iDir_2 
      write(*,*) 'fTmp_Next ', fTmp_Next

      mass_dens = 0.0_rk
      vel = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fTmp_Next(pos)
          vel(:, iFieldLoc)  = vel(:, iFieldLoc) &
            & + real(layout%stencil(1)%cxDir( :, iDir), kind=rk) * fTmp_Next(pos)
        enddo 
      enddo  
      do iFieldLoc = 1, nFields
        vel(:, iFieldLoc) = vel(:,iFieldLoc)/mass_dens(iFieldLoc)
      end do  

      input_loc = 0.0_rk
      do iFieldLoc = 1, nFields
        input_loc( iFieldLoc ) = mass_dens(iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 1 ) = vel(1, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 2 ) = vel(2, iFieldLoc)
        input_loc( nFields + (iFieldLoc-1)*3 + 3 ) = vel(3, iFieldLoc)
      end do  
      write(*,*) 'mass_dens ', mass_dens
      write(*,*) 'vel ', vel 

      ! Calculate the equilibrium distribution
      call varSys%derive( derVarPos%equilFromMacro )%do(                         &
        &                input = input_loc,                                      &
        &                bound = bound,                                          &
        &                tree = tree,                                            &
        &                subTree = subTree,                                      &
        &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
        &                globSys = varSys,                                       &
        &                stencil = layout%stencil(1),                            &
        &                conf = varSys%solSpec%conf,                             & 
        &                res = fEqNext )

      write(*,*) 'fEqNext ', fEqNext  


      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
&?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
              & = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              & + mixture%omega_diff*0.5_rk*(fEq(layout%stencil(1)%cxDirInv( iDir ))&
              & - fEqNext(iDir) )
              !& + mixture%omega_diff*0.5_rk*(fEq(layout%stencil(1)%cxDirInv( iDir ))& 
              !& - fEq( iDir ) )
              !& - layout%weight( iDir )&
              !& * mixture%omega_diff*cs2inv*rho_spc(1)                        &
              !&    * ( layout%stencil(1)%cxDir( 1, iDir )*eqVel(1)            &
              !&    +   layout%stencil(1)%cxDir( 2, iDir )*eqVel(2)            &
              !&    +   layout%stencil(1)%cxDir( 3, iDir )*eqVel(3))
        endif
      enddo
      endif  
    enddo

    if (iField == nFields) then
    do iElem = 1, globBC%nElems( iLevel )
      !ux = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all(pos) = state(                                                  &
&?FETCH?( pos, globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
        & currT)
        end do
      end do  
      bound(1)=iElem
      bound(2)=iElem
      write(*,*) 'iElem ', iElem
      write(*,*) 'fTmp_all ', fTmp_all
      ! density
      do iFieldLoc = 1, nFields
        call varSys%derive( derVarPosAll(iFieldLoc)%density )%do(        &
          &                 input = fTmp_all,                               &
          &                 bound = bound,                                  &
          &                 tree = tree,                                    &
          &                 subTree = subTree,                              &
          &                 requVar = varSys%Variable%val(                  &
          &                           derVarPosAll(iFieldLoc)%density ), &
          &                 globSys = varSys, stencil = layout%stencil(1),  &
          &                 res = rhoTmp,                             &
          &                 conf = varSys%solSpec%conf )

        call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
          &                 input   = fTmp_all,                                  &
          &                 bound   = bound,                                     &
          &                 tree    = tree,                                      &
          &                 subTree = subTree,                                   &
          &                 requVar = varSys%Variable%val(&
          &                           derVarPosAll(iFieldLoc)%velocity),   &
          &                 globSys = varSys,                                    &
          &                 stencil = layout%stencil(1),                         &
          &                 conf    = varSys%solSpec%conf,                       &
          &                 res     = uxTmp )
        write(*,*) 'iField ', iFieldLoc, 'density ', rhoTmp
        write(*,*) 'iField ', iFieldLoc, 'velocity ', uxTmp
      end do  
    end do
    endif
!
!  stop

  end subroutine spc_bb_wall2
!******************************************************************************!

!******************************************************************************!
  !> molar flux boundary condition like moments velocity bc type
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moments_outflow', 
  !!    moleflux = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moments_outflow )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: moleFluxX(globBC%nElems(iLevel)),                         &
      &              moleFluxY(globBC%nElems(iLevel)),                         &
      &              moleFluxZ(globBC%nElems(iLevel))
    integer :: iELem, iDir, iFieldLoc, nFields, pos, tmpDir
    real(kind=rk) :: rho, press, dens_correction
    integer :: nLinks, iLink, elemPos, neighPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp(:)
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: toMoments( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: vel(3,varSys%nVars), spc_vel(3), vel_inG(3)
    real(kind=rk) :: mom(3,varSys%nVars), spc_mom(3), mom_inG(3)
    real(kind=rk) :: mass_dens(varSys%nVars), num_dens(varSys%nVars)
    real(kind=rk) :: spc_num_dens(1), totMassDens, velAvg(3), velNew(3)
    real(kind=rk) :: moleFrac(varSys%nVars)
    integer :: nField
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk) :: resi_coeff( varSys%nVars, varSys%nVars )
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !------------------------------------------------------------------------ 
    !write(*,*) 'Boundary label ', trim(me%label)

    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    end do    
    
!    write(*,*) 'iField ', iField
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
!      write(*,*) 'iElem ', iElem  

      fTmp = 0.0_rk
      rho = 0
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )
      neighPos = me%elems(iLevel)%neigh(1,iElem)

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do 


      ! local state vector to compute density and velocity of other species
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            !& state(?FETCH?( pos, elemPos, nScalars, neigh ), currT) 
            & state(?SAVE?( pos, neighPos, nScalars, neigh ), currT) 
        enddo
      enddo

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      do iFieldLoc = 1, nFields
        call varSys%derive( derVarPosAll(iFieldLoc)%numDensity )%do(        &
          &                 input = fTmp_all,                               &
          &                 bound = bound,                                  &
          &                 tree = tree,                                    &
          &                 subTree = subTree,                              &
          &                 requVar = varSys%Variable%val(                  &
          &                           derVarPosAll(iFieldLoc)%numDensity ), &
          &                 globSys = varSys, stencil = layout%stencil(1),  &
          &                 res = spc_num_dens,                             &
          &                 conf = varSys%solSpec%conf )
        num_dens(iFieldLoc) = spc_num_dens(1)

        call varSys%derive( derVarPosAll(iFieldLoc)%momentum )%do(              &
          &                       input = fTmp_all,                             &
          &                       bound = bound,                                &
          &                       tree = tree,                                  &
          &                       subTree = subTree,                            &
          &                       requVar = varSys%Variable%val(                &
          &                                 derVarPosAll(iFieldLoc)%momentum ), &
          &                       globSys = varSys, stencil = layout%stencil(1),&
          &                       res = spc_mom,                                &
          &                       conf = varSys%solSpec%conf )

        mom(:, iFieldLoc) = spc_mom
      enddo  

      ! mass density
      mass_dens = num_dens * molWeight
      totMassDens = sum(mass_dens)

      rho = mass_dens(iField)
!      write(*,*) 'density ', rho

      ! mole fraction
      moleFrac = num_dens/sum(num_dens)

!      do iFieldLoc = 1, nFields
!        write(*,*) iFieldLoc, 'vel ', vel(:, iFieldLoc), &
!          & 'dens ', mass_dens(iFieldLoc), 'moleFrac ', moleFrac(iFieldLoc)
!      enddo  
      
      ! compute velocity of transformed variable g. Eq. which is normally
      ! used to compute actual velocity by solving LSE
      mom_inG = mom(:, iField)
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG + ( mixture%omega_diff * 0.5_rk                    &
            &               * resi_coeff(iField, iFieldLoc) * phi(iField)    &
            &               * moleFrac(iFieldLoc) / mixture%paramB )         &
            &             * mom(:, iField)
      end do
      !set non-diagonal part
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG - ( mixture%omega_diff * 0.5_rk                    &
          &                 * resi_coeff(iField, iFieldLoc) * phi(iFieldLoc) &
          &                 * moleFrac(iField) / mixture%paramB )            &
          &               * mom(:,iFieldLoc)
      end do

      vel_inG = mom_inG/rho
!      write(*,*) 'transformed momentum ', mom_inG

      ! mass averaged mixture velocity
      velAvg(1) = sum(mom(1,:)) / totMassDens
      velAvg(2) = sum(mom(2,:)) / totMassDens
      velAvg(3) = sum(mom(3,:)) / totMassDens

!      write(*,*) 'velAvg ', velAvg

      press = cs2*rho*fieldProp%species%molWeigRatio
      !write(*,*) 'rho ', rho, 'press', press

      toMoments = matmul(layout%moment%toMoments%A, fTmp)
      !write(*,*) 'fTmp ', fTmp
      !write(*,*) 'toMoments ', toMoments

      ! moments for liquid mixture
      moments = (/ rho, mom_inG(1), mom_inG(2), press+rho*velAvg(1)**2,    &
        & press+rho*velAvg(2)**2, rho*velAvg(1)*velAvg(2), cs2*mom_inG(2), &
        & cs2*mom_inG(1), cs2*(press+rho*velAvg(1)**2+rho*velAvg(2)**2) /)

      !write(*,*) 'Moments ', moments
      !write(*,*) 'knownMomPos ', me%elems(iLevel)%moments(iElem)%knownMom_pos

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          & - toMoments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) )
      end do
      !write(*,*) 'rhs ', rhs

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )
      !write(*,*) 'unknown fTmp ', unKnown_fTmp

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

      endif
    end do

  end subroutine spc_moments_outflow
!******************************************************************************!


!******************************************************************************!
  !> Moment based velocity boundary condition from Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! \author Kannan Masilamani
  !!
?? copy :: bc_routineHeader( moments_inflow )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rho, rho0, vel(3), dens_correction
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel)), pressB(globBC%nElems(iLevel))
    integer :: iELem, iDir, tmpDir
    integer :: nLinks, iLink, elemPos, neighPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp( : )
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    integer :: cornerNeighDir
    real(kind=rk) :: fTmp_2( layout%stencil(1)%QQ )
    real(kind=rk) :: rhoTmp, uxTmp(3)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------
    !write(*,*) 'Boundary label ', trim(me%label)
    ! Get velocity which might be defined to vary in space and time
    ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    !write(*,*) 'phy Vel', ux, uy, uz
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel

    pressB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,     &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = sim_time,                              &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%general%solver%conf(1),         &
      &                       dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    pressB = pressB/physics%fac( iLevel )%press

    rho0 = fieldProp%fluid%rho0

    do iElem = 1, globBC%nElems( iLevel )

      write(*,*) 'iElem ', iElem
      write(*,*) 'velocity ', ux(iElem), uy(iElem), uz(iElem)
      vel = (/ ux(iElem), uy(iElem), uz(iElem) /)

      fTmp = 0.0_rk
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
!           & bcBuffer( ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
!        write(*,*) 'fTmp0 ',  fTmp(iDir)
      end do 

      rho = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          tmpDir = layout%stencil(1)%cxDirInv(iDir)
          if( globBC%elems(iLevel)%bitmask( tmpDir, iElem )) then
            tmpDir = layout%stencil(1)%cxDirInv(globBC%elems(iLevel)%normalInd(iElem))
          endif
        else
          tmpDir = iDir
        endif
        rho  = rho + fTmp(tmpDir)
      end do

write(*,*) 'dens before correction ', rho

      ! add small correction term to density which comes from momentum in the
      ! flow direction while substituing unknowns pdfs terms obtained
      ! from moments BC derivation to compute density.
      ! With this correction term density is recovered correctly at this node
      dens_correction = dot_product(vel, globBC%elems(iLevel)%normal(:, iElem))
      ! for compressible model, rho_new = rho_old + rho_new*dens_correction
      ! thus, rho_new = rho_old/(1-dens_correction)
      ! correction term for in-compressible mode
      rho = rho + rho0*dens_correction
write(*,*) 'dens after correction ', rho, rho*cs2
write(*,*) 'dens from press', pressB(iElem)*cs2inv, pressB(iElem)

      rho = pressB(iElem) * cs2inv + rho0

      !!! add small correction term to density which comes from momentum in the
      !!! flow direction while substituing unknowns pdfs terms obtained
      !!! from moments BC derivation to compute density.
      !!! With this correction term density is recovered correctly at this node
      !!dens_correction = dot_product(vel, globBC%elems(iLevel)%normal(:, iElem))
      !!! correction term for Multispecies
      !!! rho_new = rho_old/(1.0 - dens_correction)
      !!rho = rho + rho0*dens_correction

      !write(*,*) 'fTmp ', fTmp

      ! moments for compressible model
      !moments = (/ rho, rho*ux(iElem), rho*uy(iElem), press+rho*ux(iElem)**2, &
      !  & press+rho*uy(iElem)**2, rho*ux(iElem)*uy(iElem), cs2*rho*uy(iElem), &
      !  & cs2*rho*ux(iElem), cs2*(press+rho*ux(iElem)**2+rho*uy(iElem)**2) /)
      ! moments for in-compressible model
      moments = (/ rho, rho0*vel(1), rho0*vel(2), pressB(iElem)+rho0*vel(1)**2,&
        & pressB(iElem)+rho0*vel(2)**2, rho0*vel(1)*vel(2), cs2*rho0*vel(2),   &
        & cs2*rho0*vel(1), cs2*(pressB(iElem)+rho0*vel(1)**2+rho0*vel(2)**2) /)
      !write(*,*) 'Moments ', moments
      !write(*,*) 'knownMomPos ', me%elems(iLevel)%moments(iElem)%knownMom_pos

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do
     !write(*,*) 'rhs ', rhs

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )
      !write(*,*) 'unknown fTmp ', unKnown_fTmp

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

!      do iDir = 1,layout%stencil(1)%QQ
!        write(*,*) 'idir ', iDir, 'pdf ',state(                                &
!&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
!        & currT)
!      end do  

    end do

    do iElem = 1, globBC%nElems( iLevel )
      !ux = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        fTmp_2(iDir) = state(                                                  &
&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
        & currT)
      end do  
      rhoTmp = sum(fTmp_2)
      bound(1)=iElem
      bound(2)=iElem
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                 input   = fTmp_2,                                    &
        &                 bound   = bound,                                     &
        &                 tree    = tree,                                      &
        &                 subTree = subTree,                                   &
        &                 requVar = varSys%Variable%val(derVarPos%velocity),   &
        &                 globSys = varSys,                                    &
        &                 stencil = layout%stencil(1),                         &
        &                 conf    = varSys%solSpec%conf,                       &
        &                 res     = uxTmp )
      write(*,*) 'iElem ', iElem
      write(*,*) 'density ', rhoTmp, 'press ', cs2*rhoTmp, cs2*rhoTmp*physics%fac( iLevel )%press
      write(*,*) 'velocity ', uxTmp
!      write(*,*) 'velocity ', uxTmp*rhoTmp
    end do
!
!  stop

  end subroutine moments_inflow
!******************************************************************************!


!******************************************************************************!
  !> Moment based velocity boundary condition from Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! \author Kannan Masilamani
  !!
?? copy :: bc_routineHeader( moments_outflow )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rho, rho0, press, vel(3)
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel)), pressB(globBC%nElems(iLevel))
    integer :: iELem, iDir
    integer :: nLinks, iLink, elemPos, neighPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp( : )
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    integer :: cornerNeighDir
    ! Vel on frst neighbor element
    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel))
    ! density on first neighbor element
    real(kind=rk) :: rho_1(globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !real(kind=rk) :: fTmp_neigh( layout%stencil(1)%QQ )
    !real(kind=rk) :: rhoTmp, uxTmp(3)
    !---------------------------------------------------------------------------
    !write(*,*) 'Boundary label ', trim(me%label)

    rho0 = fieldProp%fluid%rho0

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! velocity is taken from precollision of current
    call varSys%derive( derVarPos%velocity )%do(                               &
        &                  input   = me%elems(iLevel)%neighBufferPre(1,:),     &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = uxB_1 )

    ! density is taken from precollision of current
    call varSys%derive( derVarPos%density )%do(                               &
        &                  input   = me%elems(iLevel)%neighBufferPre(1,:),     &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%density ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = rho_1 )

    do iElem = 1, globBC%nElems( iLevel )

      !write(*,*) 'iElem ', iElem
      rho = rho_1(iElem)
      vel = (/ uxB_1((iElem-1)*3+1), uxB_1((iElem-1)*3+2), uxB_1((iElem-1)*3+3) /)
      !write(*,*) 'dens', rho, 'velocity ', vel 

      fTmp = 0.0_rk
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
!           & bcBuffer( ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
!        write(*,*) 'fTmp0 ',  fTmp(iDir)
      end do 

      press = rho * cs2
      
      !write(*,*) 'fTmp ', fTmp

      ! moments for compressible model
      !moments = (/ rho, rho*ux(iElem), rho*uy(iElem), press+rho*ux(iElem)**2, &
      !  & press+rho*uy(iElem)**2, rho*ux(iElem)*uy(iElem), cs2*rho*uy(iElem), &
      !  & cs2*rho*ux(iElem), cs2*(press+rho*ux(iElem)**2+rho*uy(iElem)**2) /)
      ! moments for in-compressible model
      moments = (/ rho, rho0*vel(1), rho0*vel(2), press+rho0*vel(1)**2, &
        & press+rho0*vel(2)**2, rho0*vel(1)*vel(2), cs2*rho0*vel(2), &
        & cs2*rho0*vel(1), cs2*(press+rho0*vel(1)**2+rho0*vel(2)**2) /)
      !write(*,*) 'Moments ', moments
      !write(*,*) 'knownMomPos ', me%elems(iLevel)%moments(iElem)%knownMom_pos

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do
      !write(*,*) 'rhs ', rhs

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )
      !write(*,*) 'unknown fTmp ', unKnown_fTmp

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

!      do iDir = 1,layout%stencil(1)%QQ
!        write(*,*) 'idir ', iDir, 'pdf ',state(                                &
!&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
!        & currT)
!      end do  

    end do

!!    do iElem = 1, globBC%nElems( iLevel )
!!      !ux = 0.0_rk
!!      do iDir = 1,layout%stencil(1)%QQ
!!        fTmp_2(iDir) = state(                                                  &
!!&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
!!        & currT)
!!      end do  
!!      rhoTmp = sum(fTmp_2)
!!      bound(1)=iElem
!!      bound(2)=iElem
!!      call varSys%derive( derVarPos%velocity )%do(                             &
!!        &                 input   = fTmp_2,                                    &
!!        &                 bound   = bound,                                     &
!!        &                 tree    = tree,                                      &
!!        &                 subTree = subTree,                                   &
!!        &                 requVar = varSys%Variable%val(derVarPos%velocity),   &
!!        &                 globSys = varSys,                                    &
!!        &                 stencil = layout%stencil(1),                         &
!!        &                 conf    = varSys%solSpec%conf,                       &
!!        &                 res     = uxTmp )
!!      write(*,*) 'iElem ', iElem
!!      write(*,*) 'density ', rhoTmp, cs2*rhoTmp
!!!      write(*,*) 'velocity ', uxTmp
!!!      write(*,*) 'velocity ', uxTmp*rhoTmp
!!    end do
!
!  stop

  end subroutine moments_outflow
!******************************************************************************!


!******************************************************************************!
  !> Outlet boundary condition with experimental higher order macroscopic 
  !! quantity extrapolation
  !! \author Manuel Hasert
  !!
  !!  with higher order macroscopic value extrapolation
  !! This is only experimental and follows no paper reference.
  !! We use an extrapolation of the macroscopic values and then do a anti bounce
  !! back with the extrapolated values
  !!
  ! subroutine outlet_extrapolation
?? copy :: bc_routineHeader( outlet_extrapolation )
    !---------------------------------------------------------------------------
    real(kind=rk) :: fEqPlus, fPlus, omega
    real(kind=rk) :: uxB(3), uxB_1(3), uxB_2(3), uxB_3(3), uxB_4(3)
    real(kind=rk) :: rhoB, rhoB_1, rhoB_2, rhoB_3, rhoB_4
    real(kind=rk) :: fTmp_1( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_2( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_3( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_4( layout%stencil(1)%QQ )
    real(kind=rk) :: usq!, rho_temp(1) 
    integer :: iELem, iDir
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------
    write(logUnit(1),*)'MH: Warning: I didnt adapt these BCs to the new '//    &
      &                'bitmasks! Wrong for sure'

    ! transient pressure value
!    rho_temp = tem_transient_for( me%bc_states%pressure%transient, real(sim_time%iTime,kind=rk) )
    omega = fieldProp%fluid%omLvl( iLevel )

    do iElem = 1, globBC%nElems( iLevel )
      ! Get local and neighbor velocity
      rhoB_1 = 0.0_rk
      rhoB_2 = 0.0_rk
      rhoB_3 = 0.0_rk
      rhoB_4 = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        fTmp_1( iDir ) = bcBuffer( &
          & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ) )
        fTmp_2 = me%elems( iLevel )%neighBufferPre( 1, &
          & ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        if( me%order > 1) &
          & fTmp_3 = me%elems( iLevel )%neighBufferPre( 2, &
          & ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )

        if( me%order > 2) &
          & fTmp_4 = me%elems( iLevel )%neighBufferPre( 3, &
          & ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
      enddo
     
      rhoB_1 = sum(fTmp_1)
      rhoB_2 = sum(fTmp_2)
      if( me%order > 1)  rhoB_3 = sum(fTmp_3)
      if( me%order > 2)  rhoB_4 = sum(fTmp_4)
     
      bound(1)=iElem
      bound(2)=iElem
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                 input   = fTmp_1,                                    &
        &                 bound   = bound,                                     &
        &                 tree    = tree,                                      &
        &                 subTree = subTree,                                   &
        &                 requVar = varSys%Variable%val(derVarPos%velocity),   &
        &                 globSys = varSys,                                    &
        &                 stencil = layout%stencil(1),                         &
        &                 conf    = varSys%solSpec%conf,                       &
        &                 res     = uxB_1 )

      call varSys%derive( derVarPos%velocity )%do(                             &
        &                 input   = fTmp_2,                                    &
        &                 bound   = bound,                                     &
        &                 tree    = tree,                                      &
        &                 subTree = subTree,                                   &
        &                 requVar = varSys%Variable%val(derVarPos%velocity),   &
        &                 globSys = varSys,                                    &
        &                 stencil = layout%stencil(1),                         &
        &                 conf    = varSys%solSpec%conf,                       &
        &                 res     = uxB_2 )

     if( me%order > 1)                                                      &
       &call varSys%derive( derVarPos%velocity )%do(                           &
          &                 input   = fTmp_3,                                  &
          &                 bound   = bound,                                   &
          &                 tree    = tree,                                    &
          &                 subTree = subTree,                                 &
          &                 requVar = varSys%Variable%val(derVarPos%velocity), &
          &                 globSys = varSys,                                  &
          &                 stencil = layout%stencil(1),                       &
          &                 conf    = varSys%solSpec%conf,                     &
          &                 res     = uxB_3 )

     if( me%order > 2)                                                      &
       &call varSys%derive( derVarPos%velocity )%do(                           &
          &                 input   = fTmp_4,                                  &
          &                 bound   = bound,                                   &
          &                 tree    = tree,                                    &
          &                 subTree = subTree,                                 &
          &                 requVar = varSys%Variable%val(derVarPos%velocity), &
          &                 globSys = varSys,                                  &
          &                 stencil = layout%stencil(1),                       &
          &                 conf    = varSys%solSpec%conf,                     &
          &                 res     = uxB_4 )

      ! Extrapolate velocities on the boundary node
      ! Note here, that the position is q=0.5
      ! Other interpolation formula required for different q
      select case( me%order )
      case( 1 ) ! Linear
        uxB  = 1.5_rk*uxB_1 - 0.5_rk*uxB_2
        rhoB = 1.5_rk*rhoB_1 - 0.5_rk*rhoB_2
      case( 2 ) ! Quadratic
        uxB  = 1.875_rk*uxB_1 - 1.25_rk*uxB_2 + 0.375_rk*uxB_3
        rhoB = 1.875_rk*rhoB_1 - 1.25_rk*rhoB_2 + 0.375_rk*rhoB_3
      case( 3 ) ! Cubic
        uxB  = 2.1875_rk*uxB_1 - 2.1875_rk*uxB_2 + 1.3125_rk*uxB_3             &
          &  - 0.3125_rk*uxB_4
        rhoB = 2.1875_rk*rhoB_1 - 2.1875_rk*rhoB_2 + 1.3125_rk*rhoB_3          &
          &  - 0.3125_rk*rhoB_4
      end select

      ! Outlet pressure specified in configuration file
      ! rho = me%bc_states%pressure
      ! transient pressure  time spatial pressure
!      rhoB = rho_temp(1) * globBC%elems( iLevel )%spatial_term( 1, iElem )

      usq = uxB(1)*uxB(1) + uxB(2)*uxB(2) + uxB(3)*uxB(3) 

      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Calc the correction term and the Dirichlet (Eq) term
          fEqPlus = layout%weight( iDir ) * rhoB                               &
            &     + 4.5_rk*layout%weight( iDir )*fieldProp%fluid%rho0*((       &
            &       layout%stencil(1)%cxDir( 1, iDir )*uxB(1)                  &
            &     + layout%stencil(1)%cxDir( 2, iDir )*uxB(2)                  &
            &     + layout%stencil(1)%cxDir( 3, iDir )*uxB(3))**2              &
            &     - div1_3*usq)
          fPlus = 0.5_rk*(                                                     &
  state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT) &
+ state( ?SAVE?( varPos( layout%stencil(1)%cxDirInv(iDir )), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ) &
        , currT))

 ! Now assign the values
  state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
 ! antibounceback term
 - state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)   &
          ! Dirichlet pressure term
          & + 2._rk*fEqPlus                                                                     &
          ! Correction term
          & + (2._rk - omega )*( fPlus - fEqPlus )
        endif
      enddo
    enddo
 
  end subroutine outlet_extrapolation
!******************************************************************************!


!******************************************************************************!
  !> Outlet boundary conditions with zero-f-gradient
  !! \author Manuel Hasert
  !!
  !! Simple microscopic extrapolation outlet boundaries with zero-f-gradient.
  !! This follows equation (18) in S. Chen, D. Martinez, and R. Mei, 
  !!  "On boundary conditions in lattice Boltzmann methods," 
  !!  Physics of Fluids, vol. 8, no. 9, p. 2527, 1996.
  !! 
  !! MH: These are not working yet!!
  ! subroutine outlet_extrapolation_microscopic
?? copy :: bc_routineHeader( outlet_extrapolation_microscopic )
    !--------------------------------------------------------------------------- 
    ! defining local variables
    integer :: iELem, iDir, fluidNeighbor, fluidNeighbor2
    !--------------------------------------------------------------------------- 
    write(logUnit(1),*)'MH: Warning: I didnt adapt these BCs to the new '//    &
      &                'bitmasks! Wrong for sure'

      select case( me%order) 
      case( 0 ) ! Zero order extrapolation
      do iElem = 1, globBC%nElems( iLevel )
        do iDir = 1,layout%stencil(1)%QQ
          if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
            ! Only treat the required directions
            ! f_i(xB) = f_i(xf)
            ! But we have already bounced back intrinsically, so we take the first f_i = f_i'
            ! and also have to trace the f_i', which was actually already pushed further with PUSH
state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)  =       &
state( ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ) &
           , currT)
          endif
        enddo
      enddo
      case( 1 ) ! Linear extrapolation
      do iElem = 1, globBC%nElems( iLevel )
        do iDir = 1,layout%stencil(1)%QQ
          if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
            ! Only treat the required directions
            ! f_i(xB) = 2*f_i(xf) - f_i(xf+1)
            ! But we have already bounced back intrinsically, so we take the first f_i = f_i'
            ! and also have to trace the f_i', which was actually already pushed further with PUSH
            fluidNeighbor =  &
?ElemIDX?(neigh( ?NGPOS?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), globBC%elems(iLevel)%elem(iElem),nScalars )),nScalars)
    state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)  =       &
    2._rk * state( &
    ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  &
 - state( ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), fluidNeighbor, nScalars, neigh ), currT)  
          endif
        enddo
      enddo
      case( 2 ) ! Quadratic extrapolation
        do iElem = 1, globBC%nElems( iLevel )
          do iDir = 1,layout%stencil(1)%QQ
            if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
              ! Only treat the required directions
              ! f_i(xB) = 3*f_i(xf) - 3*f_i(xf+1)+ 3*f_i(xf+2)
              ! But we have already bounced back intrinsically, so we take the first f_i = f_i'
              ! and also have to trace the f_i', which was actually already pushed further with PUSH
              fluidNeighbor =  &
?ElemIDX?(neigh( ?NGPOS?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), globBC%elems(iLevel)%elem(iElem),nScalars )),nScalars)
              fluidNeighbor2 = &
?ElemIDX?(neigh( ?NGPOS?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), fluidNeighbor,nScalars )),nScalars)
         state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  =       &
         3._rk * state( &
         ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh )&
         , currT)  &
         - 3._rk * state( &
         ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), fluidNeighbor, nScalars, neigh ), currT)  &
        + state( ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir ) ), fluidNeighbor2, nScalars, neigh ),&
        currT)  
            endif
          enddo
        enddo
      end select
 
    end subroutine outlet_extrapolation_microscopic
!******************************************************************************!


!******************************************************************************!
  !> Outlet Pressure Bounce Back boundary condition 
  !! These boundary conditions are taken from the paper:  
  !! M. Bouzidi et al, “ Momentum  transfer  of  a 
  !!  Boltzmann-lattice  fluid  with  boundaries,” PHYSICS OF
  !! FLUIDS, vol. 13, no. 11, 2001.
  !!
  !>\todo: these BC are implemented like in the DC-code, they need to be tested
  !!      and checked
  ! subroutine outlet_bouzidi
?? copy :: bc_routineHeader( outlet_bouzidi )
    !---------------------------------------------------------------------------
    real(kind=rk) ux(3), usq, rho(globBC%nElems(iLevel))!, rho_temp
    integer :: iELem, iDir
    !---------------------------------------------------------------------------
    write(logUnit(1),*)'MH: Warning: I didnt adapt these BCs to the new '//    &
      &                'bitmasks! Wrong for sure'

    ! transient pressure value
    rho = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,        &
      &                      treeIds = glob%levelDesc( iLevel )%total( globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                      tree    = tree,                                   &
      &                      time    = sim_time,                               &
      &                      n       = globBC%nElems(iLevel),                  &
      &                      conf    = params%general%solver%conf(1))

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rho = rho/physics%fac( iLevel )%press * cs2inv + fieldProp%fluid%rho0

    do iElem = 1, globBC%nElems( iLevel )
      ! Calculate local pressure and velocity moments
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)
      enddo 
      ux  = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        ux(:)  = ux(:) + layout%stencil(1)%cxDir( :, iDir)                     &
          &    * state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)
      enddo 

      ! get the actual velocity out of the moments
      ux = ux/rho
      ! Outlet density specified in configuration file
      ! rho = me%density 
      ! transient density  time spatial density
!      rho = rho_temp * me%density%spatial_val( iLevel )%val( iElem )

      usq = ux(1)*ux(1) + ux(2)*ux(2) + ux(3)*ux(3) 

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        !KM:\todo @Manuel: Please check whether it is SAVE or FETCH
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)  =   &
            &       layout%weight(iDir)*rho(iElem)*(2._rk+9._rk*( (            &
            &                    layout%stencil(1)%cxDir( 1, iDir )*ux(1)      &
            &             +      layout%stencil(1)%cxDir( 2, iDir )*ux(2)      &
            &             +      layout%stencil(1)%cxDir( 3, iDir )*ux(3))**2  &
            &                 - div1_3*usq))                                   &
          & - state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)
        endif
      enddo
    enddo
  end subroutine outlet_bouzidi
!******************************************************************************!


!******************************************************************************!
  !> Test routine for boundary neighbors
  !!
  ! subroutine test_bcNeighs
?? copy :: bc_routineHeader( test_bcNeighs )
    !---------------------------------------------------------------------------
    real(kind=rk) :: fTmp_1( layout%stencil(1)%QQ )
    ! need to compute precollision velocity of neighbor element
    real(kind=rk) :: rho_loc(1)  ! Temporary density for transient 
    real(kind=rk) :: uxB(3)      ! Velocity on boundary element
    integer :: iDir              ! Direction counter
    integer :: iElem, iNeigh     ! Element counter
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 

    ! Run over all elements in list for this boundary
    write(dbgUnit(1),*) '  iElem     elemPos       Neigh      density '//      &
      &                     '    velocity'
    do iElem = 1, globBC%nElems( iLevel )
      do iNeigh = 1, me%nNeighs
        do iDir = 1,layout%stencil(1)%QQ
          fTmp_1(iDir) = state( ?IDX?( varPos(iDir), me%elems( iLevel )%neigh(iNeigh,iElem), nScalars, neigh ), currT) 
        enddo
        bound(1)=iElem
        bound(2)=iElem
        call varSys%derive( derVarPos%density )%do(                            &
          &                input   = fTmp_1,                                   &
          &                bound   = bound,                                    &
          &                tree    = tree,                                     &
          &                subTree = subTree,                                  &
          &                requVar = varSys%Variable%val( derVarPos%velocity ),&
          &                globSys = varSys, stencil = layout%stencil(1),      &
          &                conf    = varSys%solSpec%conf,                      &
          &                res     = rho_loc )

        call varSys%derive( derVarPos%velocity )%do(                           &
          &                input   = fTmp_1,                                   &
          &                bound   = bound,                                    &
          &                tree    = tree,                                     &
          &                subTree = subTree,                                  &
          &                requVar = varSys%Variable%val( derVarPos%velocity ),&
          &                globSys = varSys, stencil = layout%stencil(1),      &
          &                conf    = varSys%solSpec%conf,                      &
          &                res     = uxB )

        write(dbgUnit(1),*) iElem, glob%levelDesc( iLevel )%total(             &
          &                        me%elems( iLevel )%neigh( iNeigh,iElem )),  &
          &                        iNeigh, rho_loc, ',',                       &
          &                        uxB(1), ',', uxB(2), ',', uxB(3)
      enddo
    enddo

  end subroutine test_bcNeighs
!******************************************************************************!


end module mus_bc_fluid_experimental_module
!******************************************************************************!
