!******************************************************************************!
!> Boundary condition treatment routines for fluid simulation
!! Details on implementation can be found in \ref tem_bc_module 
!!
!! \author Manuel Hasert, Kannan Masilamani
?? include 'header/lbm_macros.inc'
module mus_bc_fluid_module

  ! include treelm modules
  use env_module,               only: rk
  use tem_param_module,         only: div1_3, csInv, cs, cs2, cs2inv
  use tem_spacetime_fun_module, only: tem_spacetime_for
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_var_system_module,    only: tem_var_system_type
  use tem_property_module,      only: prp_missingNeigh, prp_hasQVal
  use tem_subTree_type_module,  only: tem_subTree_type
  use tem_ini_condition_module, only: tem_ini_condition_type
  use tem_geometry_module,      only: tem_BaryOfId
  use tem_logging_module,       only: logUnit

  ! include musubi modules
  use mus_bc_header_module,     only: boundary_type, glob_boundary_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_pdf_module,           only: pdf_global_type
  use mus_variable_module,      only: mus_derVarPos_type
  use mus_param_module,         only: mus_param_type
  use mus_physics_module,       only: mus_physics_type
  use mus_mixture_module,       only: mus_mixture_type

  implicit none

  private

  public :: inlet_ubb
  public :: inlet_eq
  public :: inlet_mfr
  public :: inlet_mfr_eq

  public :: outlet_pab
  public :: outlet_eq
  public :: outlet_dnt
  public :: outlet_expol
  public :: outlet_zero_prsgrd

  public :: outlet_nrbc
  public :: outlet_nrbc_eq

  ! moments based BC
  public :: moments_vel
  public :: moments_press
  public :: moments_wall

  public :: moments_vel_incomp
  public :: moments_press_incomp

contains
!******************************************************************************!
  !> Inlet Velocity Equilibrium type boundary conditions
  !!
  !! The incoming densities are set to the equilibrium distribution with 
  !! macroscopic values:
  !! - density is extrapolated (0 or 1st order) from the flow domain
  !!   + 0-order: so we simply use the density of the boundary element itself
  !! - velocity is used as defined in the configuration file
  !!
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'inlet_eq', 
  !!    velocityX = 0.06, 
  !!    velocityY = 0.0, 
  !!    velocityZ = 0.0 }
  !!\endverbatim
  !!
  ! subroutine inlet_eq
?? copy :: bc_routineHeader( inlet_eq )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: rho
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel))
    integer :: iELem, iDir
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 

    ! Get velocity which might be defined to vary in space and time
    ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel

    input_loc = 0.0_rk
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + bcBuffer(                                                  &
   & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
      end do
      input_loc( iElem ) = rho
    enddo
    do iElem = 1, globBC%nElems(iLevel)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1 ) = ux(iElem)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2 ) = uy(iElem)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3 ) = uz(iElem)
    end do
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input   = input_loc,                                    &
      &                bound   = bound,                                        &
      &                tree    = tree,                                         &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf    = varSys%solSpec%conf,                          &
      &                res     = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
      endif  
    enddo

  end subroutine inlet_eq
!******************************************************************************!

!******************************************************************************!
  !> Inlet Velocity Equilibrium type boundary conditions with mass flow rate
  !! as input
  !!
  !! The incoming densities are set to the equilibrium distribution with 
  !! macroscopic values:
  !! - density is extrapolated (0 or 1st order) from the flow domain
  !!   + 0-order: so we simply use the density of the boundary element itself
  !! - velocity is used as defined in the configuration file
  !!
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'inlet_eq', 
  !!    massflowrate = 0.1 }
  !!\endverbatim
  !!
?? copy :: bc_routineHeader( inlet_mfr_eq )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: massFlowRate(globBC%nElems(iLevel))
    real(kind=rk) :: rho, velocity(3), massFlowRateToVel
    integer :: iELem, iDir, iLvl
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    reaL(kind=rk) :: area
    !--------------------------------------------------------------------------- 

    !Multiply transient velocity with spatial velocity. Spatial term includes
    !maxvalue
    massFlowRate = tem_spacetime_for(                       &
      &  me      = me%bc_states%massFlowRate%spacetime,     &
      &  treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                            elems(iLevel)%elem(:) ), &
      &  tree    = tree,                                    &
      &  time    = sim_time,                                &
      &  n       = globBC%nElems(iLevel),                   &
      &  conf    = params%general%solver%conf(1),           &
      &  dt_phy  = physics%dt )

    ! convertion factor to convert mass flow rate to velocity
    ! vel_lattice = massFlowRate/(density*areaOfCrossSection)
    ! area of cross section = sum(nElems(iLevel)*dx(iLevel)**2)
    area = 0.0_rk
    do iLvl = tree%global%minLevel, tree%global%maxLevel
      area = area + globBC%nElems(iLvl) * physics%dxLvl(iLvl)**2
    end do
    massFlowRateToVel = 1.0_rk / ( physics%rho0 * area )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
!write(*,*) 'velocity_P ', massFlowRateToVel, 1.0/massFlowRateToVel
    massFlowRateToVel = massFlowRateToVel / physics%fac( iLevel )%vel
!write(*,*) 'velocity_L ', massFlowRateToVel 

    input_loc = 0.0_rk
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + bcBuffer(                                                  &
& ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
      end do
      input_loc( iElem ) = rho
    enddo
    do iElem = 1, globBC%nElems(iLevel)
      ! compute velocity from massFlowRate
      velocity = massFlowRate(iElem) * massFlowRateToVel                       &
        & * layout%stencil(1)%cxDir(:, globBC%elems(iLevel)%normalInd(iElem))
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1 ) = velocity(1)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2 ) = velocity(2)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3 ) = velocity(3)
    end do
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input   = input_loc,                                    &
      &                bound   = bound,                                        &
      &                tree    = tree,                                         &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf    = varSys%solSpec%conf,                          &
      &                res     = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain,
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
      endif
    enddo

  end subroutine inlet_mfr_eq
!******************************************************************************!


!******************************************************************************!
  !> Outlet Pressure Equilibrium type boundary conditions
  !!
  !! The incoming densities are set to the equilibrium distribution with 
  !! macroscopic values used from:
  !! * Velocity: Linear extrapolation of velocity of precollision of current 
  !!   time step
  !! * Pressure as defined in the configuration file
  !!
  !! Usage:\n
  !! \verbatim
  !!  boundary_condition = {
  !!    { label = 'outlet',
  !!       kind = 'outlet_eq',
  !!       pressure = 1.0,
  !!    }
  !!  }
  !! \endverbatim
  !!
  !! This is described in the paper:\n
  !! S. Izquierdo, P. Martinez-Lera, and N. Fueyo,\n
  !! Analysis of open boundary effects in unsteady lattice Boltzmann simulations
  !! Comput. Math. Applications, vol. 58, no. 5, pp. 914–921, Sep. 2009.\n
  !!
  !! The velocity is extrapolated (1st order) from the flow domain
  !! See equation (6)
  !!
  !subroutine outlet_eq
?? copy :: bc_routineHeader( outlet_eq )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    ! need to compute precollision velocity of neighbor element
    ! Density on boundary element
    real(kind=rk) :: rhoB( globBC%nElems(iLevel))
    ! Velocity on boundary element
    real(kind=rk) :: uxB(3*globBC%nElems(iLevel))
    ! Vel on frst neighbor element
    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel))
    ! Vel on scnd neighbor element
    real(kind=rk) :: uxB_2(3*globBC%nElems(iLevel))
    integer :: iDir              ! Direction counter
    integer :: iElem             ! Element counter
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 

    ! get density value from the definition in the lua file
    rhoB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,       &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = sim_time,                              &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%general%solver%conf(1),         &
      &                       dt_phy  = physics%dt )
    ! convert physical pressure to LB density
    rhoB =   rhoB / physics%fac( iLevel )%press * cs2inv &
      &    + fieldProp%fluid%rho0

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    !KM: velocity is taken from precollision of current
    call varSys%derive( derVarPos%velocity )%do(                               &
        &                  input   = me%elems(iLevel)%neighBufferPre(1,:),     &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = uxB_1 )

    call varSys%derive( derVarPos%velocity )%do(                               &
        &                  input   = me%elems(iLevel)%neighBufferPre(2,:),     &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = uxB_2 )

    ! prepare density and velocity for equilibrium calculation
    input_loc = 0.0_rk
    input_loc( 1:globBC%nElems(iLevel) ) = rhoB
    ! do iElem = 1, globBC%nElems(iLevel)
    !   input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1) = uxB((iElem-1)*3+1)
    !   input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2) = uxB((iElem-1)*3+2)
    !   input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3) = uxB((iElem-1)*3+3)
    ! end do
    ! Linear extrapolation
    ! Note, that the position is q=0.5
    input_loc( globBC%nElems(iLevel)+1 : globBC%nElems(iLevel)*4 ) &
      &                                           = 1.5_rk*uxB_1 - 0.5_rk*uxB_2

    ! \todo: Passing the complete treeID list here is false, only the elements
    !        for this boundary have to be passed
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
        &            input   = input_loc,                                      &
        &            bound   = bound,                                          &
        &            tree    = tree,                                           &
        &            subTree = subTree,                                        &
        &            requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
        &            globSys = varSys,                                         &
        &            stencil = layout%stencil(1),                              &
        &            conf    = varSys%solSpec%conf,                            &
        &            res     = fEq )


    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems( iLevel )
      if( .not. btest( glob%levelDesc(iLevel)%property(                        &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh ))then
        do iDir = 1,layout%stencil(1)%QQ
          ! the bitmask points into the incoming direction, which must
          ! be updated. As the kernel reads this value with FETCH,
          ! we store it to the FETCH position.
          ! This results for
          ! - PUSH in storing the density to where the bitmask
          !   points, i.e. incoming
          ! - PULL storing the density pointing outwards, but as
          !   this density is being bounced back before collision in
          !   the kernel, this is the correct position
          if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
            ! Now assign the values
state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT) = &
            & fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
          endif ! bitMask
        enddo ! iDir
      endif ! not prp_missingNeigh
    enddo ! iElem

  end subroutine outlet_eq
!******************************************************************************!


!******************************************************************************!
!> Outlet Pressure do-nothing boundary
!! Here: Linear extrapolation of velocity of precollistion of current time step
!! is used
!!
!! This is taken from the paper: 
!! M. Junk and Z. Yang, 
!! Asymptotic Analysis of Lattice Boltzmann Outflow Treatments
!! Commun. Comput. Phys., pp. 1–11, 2011.
!! KM: This boundary is stable only for omega <=1.925. 
!! This boundary condition prescribes density = 1.0 at outlet as default
!!
!! Algorithm used in this boundary condition:
!! fEq(1,u) and fEq(rho,u) are computed using macroscopic values from current element
!! In fNeq, pre-collision of current time step is used
!! for non-normal direction, extrapolated velocity from two neighbor element 
!! at current time step is used
!!
!  subroutine outlet_dnt
?? copy :: bc_routineHeader( outlet_dnt )
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fEq0( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fNEq( layout%stencil(1)%QQ )
    real(kind=rk) :: visc    ! LB viscosity
    real(kind=rk) :: rho     ! Temporary density for transient
    real(kind=rk) :: uxB(3)  ! Velocity on boundary element
    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel)) ! Vel on first neigh element
    real(kind=rk) :: uxB_2(3*globBC%nElems(iLevel)) ! Vel on scond neigh element
    real(kind=rk) :: uxB_3(3*globBC%nElems(iLevel)) ! Vel on third neigh element
    real(kind=rk) :: omega   ! Relaxation parameter
    integer :: iDir          ! Direction counter
    integer :: iElem         ! Element counter
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel))
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------- 

    omega = fieldProp%fluid%omLvl( iLevel )
    visc  = fieldProp%fluid%kine_viscosityLB

    input_loc = 0.0_rk
    do iElem = 1, globBC%nElems(iLevel)
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(             &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        rho = rho + fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
      enddo
      input_loc( iElem ) = rho
    enddo

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )

    ! Get local and neighbor velocity
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = fTmp,                                     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec%conf,                      &
      &                    res     = uxB_1 )

    !KM: velocity is taken from precollision of current
    ! time step for extrapolation
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = me%elems(iLevel)%neighBufferPre(1,:),     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec%conf,                      &
      &                    res     = uxB_2 )

    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = me%elems(iLevel)%neighBufferPre(2,:),     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec%conf,                      &
      &                    res     = uxB_3 )

    do iElem = 1, globBC%nElems(iLevel)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1) = uxB_1((iElem-1)*3+1)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2) = uxB_1((iElem-1)*3+2)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3) = uxB_1((iElem-1)*3+3)
    end do
    ! \todo: Passing the complete treeID list here is false, only the elements
    !        for this boundary have to be passed
    ! Calculate the equilibrium distribution
    ! KM: post collision macroscopic values are used to compute fEq
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec%conf,                            &
      &              res     = fEq )

    ! Calculate the reference equilibrium distribution
    input_loc( 1:globBC%nElems(iLevel) ) = fieldProp%fluid%rho0

    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec%conf,                            &
      &              res     = fEq0 )


    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems(iLevel)

      ! Extrapolate velocities on the boundary node
      ! !! Note here, that the position is q=0.5
      ! Other interpolation formula required for different q
      ! Zero order extrapolation
!      uxB = uxB_1((iElem-1)*3+1 : iElem*3)
      ! MH: Linear extrapolation
      !     Compare with 
      uxB = 1.5_rk*uxB_2((iElem-1)*3+1 : iElem*3) &
        & - 0.5_rk*uxB_3((iElem-1)*3+1 : iElem*3)

      do iDir = 1,layout%stencil(1)%QQ
        fNEq(iDir) =  fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) &
          &          - fEq(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
      end do

      ! Treat all directions, but actually we only want to treat 
      ! non-normal directions
      ! Equation (3.2b)
      do iDir = 1,layout%stencil(1)%QQ
        ! the bitmask points into the incoming direction, which must
        ! be updated. As the kernel reads this value with FETCH,
        ! we store it to the FETCH position. 
        ! This results for 
        ! - PUSH in storing the density to where the bitmask 
        !   points, i.e. incoming
        ! - PULL storing the density pointing outwards, but as 
        !   this density is being bounced back before collision in 
        !   the kernel, this is the correct position
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
!          fluidNeighbor = &
! &?ElemIDX?(neigh( ?NGPOS?( varPos( iDir ), globBC%elems(iLevel)%elem(iElem),nScalars )),nScalars)
          ! update the incoming velocity direction 
          state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
         ! This bc is stable only using FETCH with inverse direction of bitmask
         ! For PULL this means, get the outgoing one from upstream neighbor
         ! For PUSH this means, get the outgoing one from local element
         ! KM: Using FETCH gives stable result
         ! \todo test for different testcase for stability
            & state(                                                           &
&?FETCH?(varPos(layout%stencil(1)%cxDirInv(iDir)),globBC%elems(iLevel)%elem(iElem),nScalars,neigh),currT) &
!    &    fTmp(?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ))  &
            & + 6._rk*layout%weight( iDir )                                    &
            !extrapolated velocity is used here to remove fluctuations
            & * ( layout%stencil(1)%cxDir( 1, iDir )*uxB(1)                    &
            & +   layout%stencil(1)%cxDir( 2, iDir )*uxB(2)                    &
            & +   layout%stencil(1)%cxDir( 3, iDir )*uxB(3))
        endif
      enddo

      ! then overwrite the normal direction with special treatment
      ! Equation (3.2a)
      iDir = globBC%elems( iLevel )%normalInd( iElem )
state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
      ! KM:we use fEq2 since this equilibrium has to be computed from
      ! reference density
        & fEq0(?IDX?(iDir, iElem, layout%stencil(1)%QQ))                       &
        & - (visc*omega -1._rk)*fNeq( layout%stencil(1)%cxDirInv( iDir ))
    enddo

  end subroutine outlet_dnt
!******************************************************************************!


!******************************************************************************!
!> Outlet Pressure extrapolation boundary
!!
!! This is taken from the paper:
!! M. Junk and Z. Yang,
!! Asymptotic Analysis of Lattice Boltzmann Outflow Treatments
!! Commun. Comput. Phys., pp. 1–11, 2011.
!!
!! This boundary condition prescribes density = 1.0 at outlet by forcing the
!! equlibirium part of \f$ f_i \f$ in normal direction (equation 3.13a):
!! \f$
!!  f_i(n+1,j_0) = \left[ F^{eq}_{i}(1,u) + f^{neq}_i \right] (n,j_0)
!! \f$
!! For non-normal direction, \f$ f_i \f$ is extrapolated from neighrbors by
!! (equation 3.13b):
!! \f$
!!  f_i(n+1,j_0) = 2f_i(n+1,j_0-n) - f_i(n+1,j_0-2n)
!! \f$
!!
!  subroutine outlet_expol
?? copy :: bc_routineHeader( outlet_expol )
    !---------------------------------------------------------------------------
    real(kind=rk), allocatable :: fTmp(:)
    real(kind=rk) :: fTmp_1
    real(kind=rk) :: fTmp_2
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fEq0( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: fNEq( layout%stencil(1)%QQ )
    real(kind=rk) :: uxB( 3 * globBC%nElems(iLevel) ) ! Velocity of BC elements
    integer :: iDir, iElem, QQ
    ! density and velocity of BC elements, used for equilibirium calculation
    real(kind=rk) :: input_loc( 4 * globBC%nElems(iLevel) )
    real(kind=rk) :: rhoB( globBC%nElems(iLevel) )
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    QQ = layout%stencil(1)%QQ

    ! get transient pressure from me%bc_states%pressure
    rhoB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,       &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = sim_time,                              &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%general%solver%conf(1),         &
      &                       dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoB = rhoB / physics%fac( iLevel )%press * cs2inv + fieldProp%fluid%rho0

    ! initialize array which stores vel and den for eq calculation
    input_loc = 0._rk

    ! transfer bcBuffer to fTmp
    allocate( fTmp( nScalars * globBC%nElems( iLevel ) ) )
    do iElem = 1, globBC%nElems( iLevel )
      do iDir = 1, QQ
        fTmp( ?IDX?(iDir, iElem, nScalars) ) = &
          & bcBuffer( ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
      end do
    end do

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )

    ! call derive function pointer to calculate density
    call varSys%derive( derVarPos%density )%do(                      &
      &          input   = fTmp,                                     &
      &          bound   = bound,                                    &
      &          tree    = tree,                                     &
      &          subTree = subTree,                                  &
      &          requVar = varSys%Variable%val( derVarPos%density ), &
      &          globSys = varSys,                                   &
      &          stencil = layout%stencil(1),                        &
      &          conf    = varSys%solSpec%conf,                      &
      &          res     = input_loc( 1:globBC%nElems(iLevel) )      )

    ! Get local velocity
    uxB = 0._rk

    ! KM: velocity is taken from postcollision of current
    call varSys%derive( derVarPos%velocity )%do(                     &
      &          input   = fTmp,                                     &
      &          bound   = bound,                                    &
      &          tree    = tree,                                     &
      &          subTree = subTree,                                  &
      &          requVar = varSys%Variable%val( derVarPos%velocity ),&
      &          globSys = varSys,                                   &
      &          stencil = layout%stencil(1),                        &
      &          conf    = varSys%solSpec%conf,                      &
      &          res     = uxB )

    do iElem = 1, globBC%nElems(iLevel)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1) = uxB((iElem-1)*3+1)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2) = uxB((iElem-1)*3+2)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3) = uxB((iElem-1)*3+3)
    end do

    ! Calculate the equilibrium distribution
    ! KM: we use post collision macroscopic values and pdf to 
    ! compute equilibrium and fNeq 
    ! Calculate local equilibrium (use local rho and uxB )
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec%conf,                            &
      &              res     = fEq )

    !MH: using not rho0 but rhoB as the inlet pressure
    input_loc( 1:globBC%nElems(iLevel) ) = rhoB(1:globBC%nELems(iLevel))

    ! Calculate reference equilibrium (use rhoB and uxB )
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &              input   = input_loc,                                      &
      &              bound   = bound,                                          &
      &              tree    = tree,                                           &
      &              subTree = subTree,                                        &
      &              requVar = varSys%Variable%val( derVarPos%equilFromMacro ),&
      &              globSys = varSys,                                         &
      &              stencil = layout%stencil(1),                              &
      &              conf    = varSys%solSpec%conf,                            &
      &              res     = fEq0 )

    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest( glob%levelDesc(iLevel)%property(                        &
        &       globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh ) ) then
        ! Treat ALL directions,
        ! but NORMAL direction will be overwriten later by Equation (3.2b)
        do iDir = 1, QQ
          ! the bitmask points into the incoming direction, which must
          ! be updated. As the kernel reads this value with FETCH,
          ! we store it to the FETCH position.
          ! This results for
          ! - PUSH in storing the density to where the bitmask
          !   points, i.e. incoming
          ! - PULL storing the density pointing outwards, but as
          !   this density is being bounced back before collision in
          !   the kernel, this is the correct position
          if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
            ! KM: pre-collision of current time step must be used for
            ! extrapolation
            fTmp_1 = me%elems(iLevel)%neighBufferPre(1, ?IDX?(iDir, iElem, QQ))
            fTmp_2 = me%elems(iLevel)%neighBufferPre(2, ?IDX?(iDir, iElem, QQ))

            ! if any of the neighbor has property qVal then fall back to simple
            ! equilibrium boundary condition
            if( btest( glob%levelDesc(iLevel)%property(                        &
              &        me%elems(iLevel)%neigh(1, iElem)),  prp_hasQVal )       &
              & .or. btest( glob%levelDesc(iLevel)%property(                   &
              &        me%elems(iLevel)%neigh(2, iElem)),  prp_hasQVal ) ) then
              ! update the incoming velocity direction
              state(                                                           &
&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
              & currT)  =                                                      &
              & fEq0(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
            else ! extrapolate pdf from neighbor
              state(                                                           &
&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
              & currT)  =                                                      &
              ! linear extrapolation for wall at half-position
              ! equation 3.13b
              ! \todo: according to equation 3.13b, parameters should be 2 and 1
              !        but better results can be obtained with 1.5 and 0.5
              &  1.5_rk*fTmp_1 - 0.5_rk*fTmp_2
              ! &  2.0_rk*fTmp_1 - 1.0_rk*fTmp_2
            end if ! globBC%elems(iLevel)%bitmask( iDir, iElem )
          end if ! not prp_missingNeigh
        end do ! iDir

        ! non-equilibrium part for normal direction
        ! \todo: fNeq should be converted from post to pre
        do iDir = 1,layout%stencil(1)%QQ
          fNEq(iDir) =  fTmp(?IDX?(iDir, iElem, nScalars))&
            &         - fEq(?IDX?(iDir, iElem, QQ))
        end do !iDir

        ! then overwrite the normal direction with special treatment
        ! Equation (3.13a)
        iDir = globBC%elems( iLevel )%normalInd( iElem )
        state(                                                                 &
&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
        & currT)  = &
        ! KM:we use fEq0 since this equilibrium has been computed from
        ! defined density at boundary
        & fEq0(?IDX?(iDir, iElem, QQ)) + fNeq( layout%stencil(1)%cxDirInv(iDir))
      end if
    end do ! iElem

    deallocate( fTmp )

  end subroutine outlet_expol
!******************************************************************************!


!******************************************************************************!
!> Inlet Velocity Bounce Back boundary condition
!!
!! This is taken from the paper:
!! [1] S. Izquierdo and N. Fueyo, “Characteristic  non-reflecting
!! boundary  conditions  for  open  boundaries  in  lattice
!!  Boltzmann  methods,” Physical Review E, vol. 78, no. 46707, 2008
!! The incoming densities \f$ \bar{f_{alpha}} \f$ are reconstructed from the
!! already reflected densities, enforcing the given velocity me\%ubb\%velocity.
!! [2] M. Junk and Z. Yang. One-point boundary condition for the lattice Boltzmann method,
!! Physical Review E, vol 72, issu 8, year 2005.
!!
!! This boundary condition has error of 1st order pressure and 2nd order velocity
!! only if wall is located exactly at q=1/2 else the accuracy of both pressure
!! and velocity reduce by order 1.
!!
!  subroutine inlet_ubb
?? copy :: bc_routineHeader( inlet_ubb)
    !---------------------------------------------------------------------------
    real(kind=rk) :: eqPlus, rho0
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: ux(globBC%nElems(iLevel))
    real(kind=rk) :: uy(globBC%nElems(iLevel))
    real(kind=rk) :: uz(globBC%nElems(iLevel))
    integer :: iELem, iDir
    !---------------------------------------------------------------------------

    rho0 = 1.0_rk

    ! Multiply transient velocity with spatial velocity. Spatial term includes
    ! maxvalue
    ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
!write(*,*) 'velocity_P ', ux(1),uy(1),uz(1) 
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel
!write(*,*) 'velocity_L ', ux(1),uy(1),uz(1) 
!stop

    ! element loop
    do iElem = 1, globBC%nElems( iLevel )
      if( .not. btest( glob%levelDesc(iLevel)%property(                        &
        &              globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh))then

        !store state in temp array
        !KM:produce no checker board effect if we use pre-collision values of 
        !current time step
        do iDir = 1,layout%stencil(1)%QQ
          !KM: Fetch gives precollision value at current time step and it produce
          ! no fluctuation at inlet.
          !SAVE use post collision values and produce fluctuation. 
          !\todo [1] and [2] use post collision values so we need to validate this
          !KM: It was found by Kartick with his testcase, SAVE is more stable then FETCH
          !fTmp( iDir ) = state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)
          fTmp( iDir ) = bcBuffer(                                               &
        & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo

        do iDir = 1,layout%stencil(1)%QQ
          ! Write the values
          if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
            ! According to the bounce back rule, 
            ! Subtract the equlibrium if its computed from alpha-(inverse of
            ! bitmask) or add the equlibrium if its computed in the direction of
            ! bitmask.
            ! In the Izquiredo paper, this is rho0, the reference density
            eqPlus =    layout%weight(iDir) * 6._rk * rho0                   &
              &       * (   layout%stencil(1)%cxDirRK(1,iDir) * ux(iElem)    &
              &           + layout%stencil(1)%cxDirRK(2,iDir) * uy(iElem)    &
              &           + layout%stencil(1)%cxDirRK(3,iDir) * uz(iElem)    )

            ! We need to get post-collision pdf in direction
            ! alpha-, which is the inverse direction of bitmask
            ! For PULL this means, get the outgoing one, as this is the one which will be bounced back
            ! For PUSH this means, get the already bounced back pdf back, so take the incoming
state(?FETCH?(varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh), currT) = &
              & fTmp( layout%stencil(1)%cxDirInv( iDir ) ) + eqPlus
          endif
        enddo ! iDir
      endif
    enddo ! iElem

  end subroutine inlet_ubb
!******************************************************************************!


  !******************************************************************************!
  !> Inlet Velocity Bounce Back boundary condition with mass flow rate as input
  !!
  !! This is taken from the paper: 
  !! [1] S. Izquierdo and N. Fueyo, “Characteristic  non-reflecting 
  !! boundary  conditions  for  open  boundaries  in  lattice 
  !!  Boltzmann  methods,” Physical Review E, vol. 78, no. 46707, 2008
  !! The incoming densities \f$ \bar{f_{alpha}} \f$ are reconstructed from the
  !! already reflected densities, enforcing the given velocity me\%ubb\%velocity.
  !! [2] M. Junk and Z. Yang. One-point boundary condition for the 
  !! lattice Boltzmann method,
  !! Physical Review E, vol 72, issu 8, year 2005.
  !!
  !! This boundary condition has error of 1st order pressure and 2nd order velocity
  !! only if wall is located exactly at q=1/2 else the accuracy of both pressure
  !! and velocity reduce by order 1.
  !!  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'inlet_mfr', 
  !!    massflowrate = 0.1 }
  !!\endverbatim
?? copy :: bc_routineHeader( inlet_mfr)
    !--------------------------------------------------------------------------- 
    real(kind=rk) :: eqPlus, rho0
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: massFlowRate(globBC%nElems(iLevel))
    real(kind=rk) :: velocity(3), massFlowRateToVel
    integer :: iELem, iDir, iLvl
    reaL(kind=rk) :: area
    !--------------------------------------------------------------------------- 

    rho0 = 1._rk

    ! Multiply transient velocity with spatial velocity. Spatial term includes
    ! maxvalue
    massFlowRate = tem_spacetime_for(                       &
      &  me      = me%bc_states%massFlowRate%spacetime,     &
      &  treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                            elems(iLevel)%elem(:) ), &
      &  tree    = tree,                                    &
      &  time    = sim_time,                                &
      &  n       = globBC%nElems(iLevel),                   &
      &  conf    = params%general%solver%conf(1),           &
      &  dt_phy  = physics%dt )

    ! convertion factor to convert mass flow rate to velocity
    ! vel_lattice = massFlowRate/(density*nElems*dx^2)
    ! area of cross section = sum(nElems(iLevel)*dx(iLevel)**2)
    area = 0.0_rk
    do iLvl = tree%global%minLevel, tree%global%maxLevel
      area = area + globBC%nElems_totalLevel(iLvl) * physics%dxLvl(iLvl)**2
    end do  

    massFlowRateToVel = 1.0_rk / ( physics%rho0 * area )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
!write(*,*) 'velocity_P ', massFlowRateToVel, 1.0/massFlowRateToVel
    massFlowRateToVel = massFlowRateToVel / physics%fac( iLevel )%vel
!write(*,*) 'velocity_L ', massFlowRateToVel 

    !element loop
    do iElem = 1, globBC%nElems( iLevel )
      if( .not. btest( glob%levelDesc(iLevel)%property(                        &
        &              globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh))then

       ! compute velocity from massFlowRate
        velocity = massFlowRate(iElem) * massFlowRateToVel                     &
          & * layout%stencil(1)%cxDir(:, globBC%elems(iLevel)%normalInd(iElem))

        !store state in temp array
        !KM:produce no checker board effect if we use pre-collision values of 
        !current time step
        do iDir = 1,layout%stencil(1)%QQ
          !KM: Fetch gives precollision value at current time step and it produce
          ! no fluctuation at inlet.
          !SAVE use post collision values and produce fluctuation. 
          !\todo [1] and [2] use post collision values so we need to validate this
          !KM: It was found by Kartick with his testcase, SAVE is more stable then FETCH
          !fTmp( iDir ) = state( ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)
          fTmp( iDir ) = bcBuffer(                                               &
        & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo 

        do iDir = 1,layout%stencil(1)%QQ
          ! Write the values
          if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
            ! Accourding to the bounce back rule, 
            ! Subtract the equlibrium if its computed from alpha-(inverse of
            ! bitmask) or add the equlibrium if its computed in the direction of
            ! bitmask.
            ! In the Izquiredo paper, this is rho0, the reference density
            eqPlus = layout%weight( iDir ) * 6._rk * rho0            &
              &    * ( layout%stencil(1)%cxDir( 1, iDir )*velocity(1)            &
              &    +   layout%stencil(1)%cxDir( 2, iDir )*velocity(2)            &
              &    +   layout%stencil(1)%cxDir( 3, iDir )*velocity(3))
 state(?FETCH?(varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh), currT) = &
            ! We need to get post-collision pdf in direction
            ! alpha-, which is the inverse direction of bitmask
            ! For PULL this means, get the outgoing one, as this is the one which will be bounced back
            ! For PUSH this means, get the already bounced back pdf back, so take the incoming
              & fTmp( layout%stencil(1)%cxDirInv( iDir ) ) + eqPlus
            ! MH: Please check this density!
          endif
        enddo
      endif

    enddo
  end subroutine inlet_mfr
!******************************************************************************!


!******************************************************************************!
!> Outlet boundary conditions with zero pressure gradient.
!! \author Kannan Masilamani
!! 
!! These boundary conditions use the neighbor cells density and velocity in the
!! equilibrium function. It is not necessary to specify density at boundary in
!! the lua configuration file \f$ f = f^{eq}(\rho_{b-1},u_{b-1}) \f$
!!
!! MH: Kannan, please update them
!!
!  subroutine outlet_zero_prsgrd
?? copy :: bc_routineHeader( outlet_zero_prsgrd)
    !---------------------------------------------------------------------------
    ! defining local variables
    real(kind=rk) :: rho, ux(3*globBC%nElems(iLevel)), usq, rho0
    integer :: iELem, iDir
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    rho0 = 1._rk

    do iElem = 1, globBC%nElems(iLevel)
      !Calculate local density and velocity moments
      do iDir = 1,layout%stencil(1)%QQ
        fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(             &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
      enddo
    enddo
    ! Get local velocity
    !KM: velocity is taken from postcollision of current
    ux  = 0._rk
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = fTmp,                                     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec%conf,                      & 
      &                    res     = ux )

    do iElem = 1, globBC%nElems(iLevel)
      !Calculate local density and velocity moments
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        rho = rho + fTmp(?IDX?(iDir, iElem, layout%stencil(1)%QQ))
      enddo

      usq = 0.0_rk
      usq = ux((iElem-1)*3+1)*ux((iElem-1)*3+1)                                &
        & + ux((iElem-1)*3+2)*ux((iElem-1)*3+2)                                &
        & + ux((iElem-1)*3+3)*ux((iElem-1)*3+3) 

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)  =   &
            &  layout%weight( iDir )*( 2._rk*rho + 9._rk*rho0* ( ( &
            &            layout%stencil(1)%cxDir( 1, iDir )*ux((iElem-1)*3+1)  &
            &         +  layout%stencil(1)%cxDir( 2, iDir )*ux((iElem-1)*3+2)  &
            &         +  layout%stencil(1)%cxDir( 3, iDir )*ux((iElem-1)*3+3)  &
            &         )**2                                                     &
            &         -  div1_3*usq))                                          &
          ! We need to get post-collision pdf in direction
          ! alpha-, which is the inverse direction of bitmask
          ! For PULL this means, get the outgoing one, as this is the one which will be bounced back
          ! For PUSH this means, get the already bounced back pdf back, so take the incoming
            &         - fTmp(                                                  &
        & ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ) )
        endif
      enddo
    enddo

  end subroutine outlet_zero_prsgrd
!******************************************************************************!


!******************************************************************************!
!> Outlet Pressure Bounce Back boundary condition 
!! \author Manuel Hasert
!!
!! An anti-bounce back is performed for the outgoing links, so that the 
!! incoming densities are set with the defined reference density while 
!! maintaining the correct non-equilibrium part.
!! The velocity is extrapolated from two neighboring fluid elements (boundary element
!! itself + one fluid neighbor). The position of the wall is hence between the boundary 
!! element and the (non-existing) solid element
!! Based on the Bouzidi Boundary conditions, the current implementation is 
!! taken from the paper: 
!! S. Izquierdo and N. Fueyo, “Characteristic  non-reflecing  boundary
!!  conditions  for  open  boundaries  in  lattice  Boltzmann
!!  methods,” Physical Review E, vol. 78, no. 46707, 2008.
!!
!! MH: I deactivated the correction term as it involves information from the
!! old time step. In parallel, this can not be provided.
!!
!subroutine outlet_pab
?? copy :: bc_routineHeader( outlet_pab )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rhoB(globBC%nElems(iLevel)) ! Density on boundary element
    real(kind=rk) :: rhoF    ! density on fluid element 
    real(kind=rk) :: uxF(3)  ! Velocity on fluid element
    real(kind=rk) :: uxB(3)  ! Velocity on boundary element
    ! Velocity on first neighbor element
    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel) )
    ! Velocity on second neighbor element
    real(kind=rk) :: uxB_2(3*globBC%nElems(iLevel) )
    real(kind=rk) :: usq, usqF        ! Squared velocity magnitude 
    real(kind=rk) :: fPlusFluid       ! f Plus for anti-bounceback correction
    ! Equilibrium Plus for anti-bounceback correction
    real(kind=rk) :: fEqPlus
    ! Equlibrium Plus from the old time step
    real(kind=rk) :: fEqPlusFluid
    real(kind=rk) :: rho0
    real(kind=rk) :: fTmpPost( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: omega       ! Relaxation parameter
    integer :: iDir              ! Direction counter
    integer :: iElem             ! Element counter
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    rho0 = 1._rk

    ! get transient density value with reference value from me%bc_states%pressure
    rhoB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,       &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = sim_time,                              &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%general%solver%conf(1),         &
      &                       dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoB = rhoB / physics%fac( iLevel )%press * cs2inv + fieldProp%fluid%rho0

    omega = fieldProp%fluid%omLvl( iLevel )

    do iElem = 1, globBC%nElems( iLevel )
      do iDir = 1,layout%stencil(1)%QQ
        fTmpPost(?IDX?(iDir, iElem, layout%stencil(1)%QQ)) = bcBuffer(         &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
      enddo
    enddo

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Get local and neighbor velocity
    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = fTmpPost,                                 &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec%conf,                      &
      &                    res     = uxB_1 )

    call varSys%derive( derVarPos%velocity )%do(                               &
      &                    input   = me%elems(iLevel)%neighBufferPre(1,:),     &
      &                    bound   = bound,                                    &
      &                    tree    = tree,                                     &
      &                    subTree = subTree,                                  &
      &                    requVar = varSys%Variable%val( derVarPos%velocity ),&
      &                    globSys = varSys,                                   &
      &                    stencil = layout%stencil(1),                        &
      &                    conf    = varSys%solSpec%conf,                      &
      &                    res     = uxB_2 )

    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems( iLevel )
      !KM: Deactivated correction term
      rhoF = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        ! previous time step precollision to compute only density and velocity  
!        fTmpPre(iDir) = state(                                                 &
!          &  ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh ),&
!          &  lastT) 
        rhoF = rhoF + fTmpPost( ?IDX?(iDir, iElem, layout%stencil(1)%QQ) )
      enddo

      ! Extrapolate velocities on the boundary node
      ! !! Note here, that the position is q=0.5
      ! Other interpolation formula required for different q
      uxB =   1.5_rk * uxB_1((iElem-1)*3+1 : iElem*3)  &
        &   - 0.5_rk * uxB_2((iElem-1)*3+1 : iElem*3)
      uxF = uxB_1((iElem-1)*3+1 : iElem*3)

      usq = uxB(1)*uxB(1) + uxB(2)*uxB(2) + uxB(3)*uxB(3)
      usqF = uxF(1)*uxF(1) + uxF(2)*uxF(2) + uxF(3)*uxF(3)

      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Calc the correction term and the Dirichlet (Eq) term
          fEqPlus = layout%weight( iDir )*rhoB(iElem)                          &
            &     + 4.5_rk*layout%weight( iDir )*rho0*((       &
            &       layout%stencil(1)%cxDir( 1, layout%stencil(1)%             &
            &                                          cxDirInv(iDir ))*uxB(1) &
            &     + layout%stencil(1)%cxDir( 2, layout%stencil(1)%             &
            &                                          cxDirInv(iDir ))*uxB(2) &
            &     + layout%stencil(1)%cxDir( 3, layout%stencil(1)%             &
            &                                          cxDirInv(iDir ))*uxB(3) &
            &     )**2                                                         &
            &     - div1_3*usq)
          fEqPlusFluid = layout%weight( iDir )*rhoF                            &
            &          + 4.5_rk*layout%weight( iDir )*rho0*((  &
            &            layout%stencil(1)%cxDir( 1, layout%stencil(1)%        &
            &                                          cxDirInv(iDir ))*uxF(1) &
            &          + layout%stencil(1)%cxDir( 2, layout%stencil(1)%        &
            &                                          cxDirInv(iDir ))*uxF(2) &
            &          + layout%stencil(1)%cxDir( 3, layout%stencil(1)%        &
            &                                          cxDirInv(iDir ))*uxF(3) &
            &          )**2                                                    &
            &          - div1_3*usqF)

fPlusFluid = 0.5_rk*( &
    &     fTmpPost( ?IDX?(iDir, iElem, layout%stencil(1)%QQ) ) &
    &   + fTmpPost( ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ) ) )

          ! Now assign the values
          ! Actually, we should take fPlusLAst and fEqPlusLast.
          ! But this introduces errors in parallel
          state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
          ! antibounceback term
!    &   - state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)    &
          ! We need to get post-collision pdf in direction
          ! alpha-, which is the inverse direction of bitmask
          ! For PULL this means, get the outgoing one, as this is the one which will be bounced back
          ! For PUSH this means, get the already bounced back pdf back, so take the incoming
    &   - fTmpPost( ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ) ) &
    &   + 2._rk*fEqPlus     &                          ! Dirichlet pressure term
    &   + (2._rk - omega )*( fPlusFluid - fEqPlusFluid ) ! Correction term
        endif
      enddo
    enddo

  end subroutine outlet_pab
!******************************************************************************!


!******************************************************************************!
  !> Characteristic-based non-reflective open boundary conditions
  !! \todo: add explaination of these steps
  !!
  !! These boundary conditions are taken from the paper:
  !!   S. Izquierdo and N. Fueyo,
  !!   "Characteristic nonreﬂecting boundary
  !!   conditions for open boundaries in lattice Boltzmann methods,"
  !!   Physical Review E, vol. 78, no. 46707, 2008.
  !!
  !! > Note: unstable behavior for high omega relaxation parameters.
  !! >       It is recommended to use a sponge layer with a
  !! >       \ref mus_aux_module::mus_setspatialomega "spatial omega"
  !! >       where the omega is decreased towards the boundaries to increase
  !! >       stability
  !!
  !! \author Manuel Hasert
  !!
  !  subroutine outlet_nrbc
?? copy :: bc_routineHeader( outlet_nrbc )
    !---------------------------------------------------------------------------
    real(kind=rk) :: fEqPlus, fPlus
    real(kind=rk) :: usq, rhoTmp, rhoTmp_2
    real(kind=rk) :: rhoRef(globBC%nElems( iLevel ))
    real(kind=rk) :: Ma, swap, sigma, kappa, K_mod
    real(kind=rk) :: L1, L2, L3, L4, L5
    real(kind=rk) :: dudx, dvdx, dwdx, drhodx
    real(kind=rk) :: rhoN1Prev, uN1Prev(3), rhoN2Prev, uN2Prev(3)
    real(kind=rk) :: rhoLodi, uLodi(3), csMod
    real(kind=rk) :: fTmpPost( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmpPost_2( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0ref
    integer :: iELem, iDir
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    rho0ref = 1.0_rk
    ! get physical pressure from lua and convert it to LB density
    rhoRef = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,     &
      &                         treeIDs = glob%levelDesc( iLevel )%total(      &
      &                                         globBC%elems(iLevel)%elem(:)), &
      &                         tree    = tree,                                &
      &                         time    = sim_time,                            &
      &                         n       = globBC%nElems(iLevel),               &
      &                         conf    = params%general%solver%conf(1),       &
      &                         dt_phy  = physics%dt )
    rhoRef = rhoRef / physics%fac(iLevel)%press * cs2inv + fieldProp%fluid%rho0

    sigma = me%nrbc%sigma
    kappa = me%nrbc%kappa  ! =1 for D2Q9, =1.66666??? for D3Q19
    csMod = me%nrbc%cs_mod

    do iElem = 1, globBC%nElems( iLevel )

      ! my density and velocity
      rhoLodi = me%elems( iLevel )%lodi( 1,   1, iElem )
      uLodi   = me%elems( iLevel )%lodi( 2:4, 1, iElem )
      ! density and velocity of my first neighbor from last iteration
      rhoN1Prev = me%elems( iLevel )%lodi( 1,   2, iElem )
      uN1Prev   = me%elems( iLevel )%lodi( 2:4, 2, iElem )
      ! density and velocity of my secnd neighbor from last iteration
      rhoN2Prev = me%elems( iLevel )%lodi( 1,   3, iElem )
      uN2Prev   = me%elems( iLevel )%lodi( 2:4, 3, iElem )

      ! Use a swapping variable to treat +x and -x directions
      swap = 1._rk
      ! x-direction
      if( globBC%elems( iLevel )%normal( 1, iElem ) /= 0 ) then 
        swap = 1._rk*(-real( globBC%elems( iLevel )%normal( 1, iElem ),kind=rk))

        ! LODI Treatment
        Ma   = abs(uLodi(1))*csInv ! X direction
        ! Equation (17) for the linear relaxation model constant
        ! lodi_length= distance outlet-inlet
        K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)

        ! calculate spatial derivatives according to (15) which will be used to
        ! compute the characteristics
        ! This is a quadratic extrapolation
        drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
        dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
        dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
        dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))

        ! wave amplitudes
        ! Incoming wave L1 is approached with a linear relaxation model (16)
        L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
        L2 = (uLodi(1))*dvdx  ! equation (9) second row
        L3 = (uLodi(1))*dwdx  ! equation (9) second row adapted for z-direction
        ! equation (9) third row with EOS equation (3) used
        L4 = (uLodi(1))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
        ! equation (9) fourth row with EOS equation (3) used
        ! MH: Here rhoLodi should be replaced with rho0
        L5 = ( uLodi(1) + swap * csMod )                                       &
          &* ( csMod * csMod * drhodx / kappa + rho0ref * swap * csMod * dudx )

        ! LODI Equations (18)
        rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
        ! Eq. (18b) for x-velocity
        uLodi(1) = uLodi(1) - ((L5-L1))/(rho0Ref*swap*csMod)*0.5d0
        uLodi(2) = uLodi(2) - L2  ! equation (18c) for y-direction
        uLodi(3) = uLodi(3) - L3  ! equation (18c) for z-direction
      else
        ! Treat y-direction
        if( globBC%elems( iLevel )%normal( 2, iElem ) /= 0 ) then 
          swap = 1._rk*(-real( globBC%elems( iLevel )%normal(2,iElem),kind=rk))

          ! LODI Treatment
          Ma   = abs(uLodi(2))*csInv ! X direction
          ! lodi_length= distance outlet-inlet
          K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)

          ! calculate derivatives
          ! This is a quadratic extrapolation
          drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
          dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
          dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
          dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))

          ! wave amplitudes
          L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
          L2 = ( uLodi(2) ) * dudx
          L3 = ( uLodi(2) ) * dwdx
          L4 = ( uLodi(2) ) * csMod * csMod * drhodx * ( 1.d0 - 1.d0 / kappa )
          L5 = ( uLodi(2) + swap * csMod )                                     &
            &* ( csMod * csMod * drhodx / kappa                                &
            &+   rho0ref * swap * csMod * dvdx )

          ! LODI Equations
          rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
          uLodi(1) = uLodi(1) - L2
          uLodi(2) = uLodi(2) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          uLodi(3) = uLodi(3) - L3
        else
          ! Treat z-direction
          if( globBC%elems( iLevel )%normal( 3, iElem ) /= 0 ) then 
            swap = 1._rk*(-real( globBC%elems(iLevel)%normal(3,iElem),kind=rk ))

            ! LODI Treatment
            Ma   = abs(uLodi(3))*csInv ! X direction
            !lodi_length= distance outlet-inlet
            K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)

            ! calculate derivatives
           ! This is a quadratic extrapolation
            drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
            dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
            dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
            dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))

            ! wave amplitudes
            L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
            L2 = ( uLodi(3) ) * dudx
            L3 = ( uLodi(3) ) * dvdx
            L4 = ( uLodi(3) ) * csMod * csMod * drhodx * ( 1.d0 - 1.d0 / kappa )
            L5 = ( uLodi(3) +swap * csMod )                                    &
              &* ( csMod * csMod * drhodx / kappa                              &
              &+   rho0ref * swap * csMod * dwdx )

            ! LODI Equations
            rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
            uLodi(1) = uLodi(1) - L2 
            uLodi(2) = uLodi(2) - L3
            uLodi(3) = uLodi(3) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          endif
        endif
      endif

      ! Now store the previous variables for next iteration
      me%elems( iLevel )%lodi( 1,   1, iElem ) = rhoLodi
      me%elems( iLevel )%lodi( 2:4, 1, iElem ) = uLodi
      ! density and velocity of my first neighbor
      ! MH: Velocity and Density should be calculated with access as SAVE
      ! For pull, this is fine however, as IDX and SAVE coincide
      rhoTmp = 0.0_rk
      rhoTmp_2 = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
      ! MH: These should be calculated with a SAVE access!
        fTmpPost( iDir ) = bcBuffer(                                           &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        fTmpPost_2(iDir) = me%elems( iLevel )%neighBufferPost( 1,              &
          &                  ?IDX?( iDir, iElem, layout%stencil(1)%QQ ))
        rhoTmp = rhoTmp + fTmpPost(iDir)
        rhoTmp_2 = rhoTmp_2 + fTmpPost_2(iDir)
      enddo

      ! density
      me%elems( iLevel )%lodi( 1, 2, iElem ) = rhoTmp
      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1

      ! velocity
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost,                                 &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,2,iElem) )

      ! density and velocity of my secnd neighbor
      me%elems( iLevel )%lodi( 1, 3, iElem )  = rhoTmp_2
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost_2,                               &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,3,iElem) )

      usq = uLodi(1)*uLodi(1) + uLodi(2)*uLodi(2) + uLodi(3)*uLodi(3)
      ! Now do the anti-bounce back
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! MH: This calculation must be the same as for the outlet_pab,
          !     except that we are using the Lodi values
          ! Calc the correction term and the Dirichlet (Eq) term
          ! MH: It makes no difference to use either iDir or inv(iDir)
          fEqPlus = layout%weight(iDir)*rhoLodi                                &
            &     + 4.5_rk*layout%weight(iDir)*rho0Ref*((         &
            &       layout%stencil(1)%cxDir( 1, layout%stencil(1)%             &
            &                                        cxDirInv(iDir ))*uLodi(1) &
            &     + layout%stencil(1)%cxDir( 2, layout%stencil(1)%             &
            &                                        cxDirInv(iDir ))*uLodi(2) &
            &     + layout%stencil(1)%cxDir( 3, layout%stencil(1)%             &
            &                                        cxDirInv(iDir ))*uLodi(3) &
            &     )**2 - div1_3*usq)

          fPlus = 0.5_rk*(                                                     &
&  state( ?FETCH?( varPos( iDir ), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT) &
&+ state( ?FETCH?( varPos( layout%stencil(1)%cxDirInv( iDir )), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ) &
&           , currT))

          ! Now assign the values
state( ?FETCH?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)  =     &
          ! antibounceback term
  & - state( ?FETCH?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)  &
  & + 2._rk*fEqPlus !     &                                      ! Dirichlet pressure term
       ! MH: Correciton term increases error!
!       + (2._rk - omega )*( fPlus - fEqPlus )                      ! Correction term
        endif
      enddo
    enddo

  end subroutine outlet_nrbc
!******************************************************************************!


!******************************************************************************!
  !> Characteristic-based non-reflective open boundary conditions 
  !!
  !! These boundary conditions are taken from the paper:  
  !!   S. Izquierdo and N. Fueyo, 
  !!   "Characteristic nonreﬂecting boundary 
  !!   conditions for open boundaries in lattice Boltzmann methods," 
  !!   Physical Review E, vol. 78, no. 46707, 2008.
  !! 
  !! > Note: unstable behavior for high omega relaxation parameters.
  !! >       It is recommended to use a sponge layer with a 
  !! >       \ref mus_aux_module::mus_setspatialomega "spatial omega" 
  !! >       where the omega is decreased towards the boundaries to increase 
  !! >       stability
  !!
  !! \author Manuel Hasert
  !!
  !  subroutine outlet_nrbc_eq
?? copy :: bc_routineHeader( outlet_nrbc_eq )
    !---------------------------------------------------------------------------
    real(kind=rk) :: usq, rhoTmp, rhoTmp_2
    real(kind=rk) :: rhoRef(globBC%nElems( iLevel ))
    real(kind=rk) :: Ma, swap, sigma, kappa, K_mod
    real(kind=rk) :: L1, L2, L3, L4, L5
    real(kind=rk) :: dudx, dvdx, dwdx, drhodx
    real(kind=rk) :: rhoN1Prev, uN1Prev(3), rhoN2Prev, uN2Prev(3)
    real(kind=rk) :: rhoLodi, uLodi(3), csMod
    real(kind=rk) :: fEq( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmpPost( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmpPost_2( layout%stencil(1)%QQ )
    real(kind=rk) :: input_loc(4), rho0ref
    integer :: iELem, iDir
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    rho0ref = 1.0_rk
    rhoRef = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,     &
      &                         treeIDs = glob%levelDesc( iLevel )%total(      &
      &                                      globBC%elems(iLevel)%elem(:)),    &
      &                         tree    = tree,                                &
      &                         time    = sim_time,                            &
      &                         n       = globBC%nElems(iLevel),               &
      &                         conf    = params%general%solver%conf(1),       &
      &                         dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    rhoRef = rhoRef/physics%fac( iLevel )%press * cs2inv + fieldProp%fluid%rho0

    sigma = me%nrbc%sigma
    kappa = me%nrbc%kappa  ! =1 for D2Q9, D3Q19
    csMod = me%nrbc%cs_mod

    do iElem = 1, globBC%nElems( iLevel )
      rhoLodi = me%elems( iLevel )%lodi( 1,   1, iElem )
      uLodi   = me%elems( iLevel )%lodi( 2:4, 1, iElem )
      ! density and velocity of my first neighbor from last iteration
      rhoN1Prev = me%elems( iLevel )%lodi( 1,   2, iElem )
      uN1Prev   = me%elems( iLevel )%lodi( 2:4, 2, iElem )
      ! density and velocity of my secnd neighbor from last iteration
      rhoN2Prev = me%elems( iLevel )%lodi( 1,   3, iElem )
      uN2Prev   = me%elems( iLevel )%lodi( 2:4, 3, iElem )

      ! Use a swapping variable to treat +x and -x directions
      swap = 1._rk
      ! x-direction
      if( globBC%elems( iLevel )%normal( 1, iElem ) /= 0 ) then 
        ! Use a swapping variable to treat +x and -x directions
        swap = 1._rk*(-real( globBC%elems( iLevel )%normal( 1, iElem ),kind=rk))

        ! LODI Treatment
        Ma   = abs(uLodi(1))*csInv ! X direction
        ! Equation (17) for the linear relaxation model constant
        ! lodi_length= distance outlet-inlet
        K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)

        ! calculate spatial derivatives according to (15) which will be used to
        ! compute the characteristics
        ! This is a quadratic extrapolation
        drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
        dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
        dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
        dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))

        ! wave amplitudes
        ! Incoming wave L1 is approached with a linear relaxation model (16)
        L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
        L2 = (uLodi(1))*dvdx  ! equation (9) second row
        L3 = (uLodi(1))*dwdx  ! equation (9) second row adapted for z-direction
        ! equation (9) third row with EOS equation (3) used
        L4 = (uLodi(1))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
        ! equation (9) fourth row with EOS equation (3) used
        ! MH: Here rhoLodi should be replaced with rho0
        L5 = ( uLodi(1) + swap * csMod )                                       &
          &* ( csMod * csMod * drhodx / kappa                                  &
          &+   rho0ref * swap * csMod * dudx )

        ! LODI Equations (18)
        rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
        ! Eq. (18b) for x-velocity
        uLodi(1) = uLodi(1) - ((L5-L1))/(rho0Ref*swap*csMod)*0.5d0
        uLodi(2) = uLodi(2) - L2  ! equation (18c) for y-direction
        uLodi(3) = uLodi(3) - L3  ! equation (18c) for z-direction
      else
        ! Treat y-direction
        if( globBC%elems( iLevel )%normal( 2, iElem ) /= 0 ) then 
          swap = 1._rk*(-real( globBC%elems( iLevel )%normal(2,iElem),kind=rk ))

          ! LODI Treatment
          Ma   = abs(uLodi(2))*csInv ! X direction
          ! lodi_length= distance outlet-inlet
          K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)

          ! calculate derivatives
          ! This is a quadratic extrapolation
          drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
          dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
          dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
          dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))

          ! wave amplitudes
          L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
          L2 = (uLodi(2))*dudx
          L3 = (uLodi(2))*dwdx
          L4 = (uLodi(2))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
          L5 = ( uLodi(2) + swap * csMod )                                     &
            &* ( csMod * csMod * drhodx / kappa                                &
            &+   rho0ref * swap * csMod * dvdx )

          ! LODI Equations
          rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
          uLodi(1) = uLodi(1) - L2 
          uLodi(2) = uLodi(2) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          uLodi(3) = uLodi(3) - L3 
        else
          ! Treat z-direction
          if( globBC%elems( iLevel )%normal( 3, iElem ) /= 0 ) then 
            swap = 1._rk*(-real(globBC%elems( iLevel )%normal(3,iElem),kind=rk))

            ! LODI Treatment
            Ma   = abs(uLodi(3))*csInv ! X direction
            !lodi_length= distance outlet-inlet
            K_mod = swap*sigma*(1.d0 - Ma*Ma)*cs/(me%nrbc%lodi_length)

            ! calculate derivatives
           ! This is a quadratic extrapolation
            drhodx =  swap*div1_3*(8.d0*rhoLodi  - 9.d0*rhoN1Prev  + rhoN2Prev)
            dudx   =  swap*div1_3*(8.d0*uLodi(1) - 9.d0*uN1Prev(1) + uN2Prev(1))
            dvdx   =  swap*div1_3*(8.d0*uLodi(2) - 9.d0*uN1Prev(2) + uN2Prev(2))
            dwdx   =  swap*div1_3*(8.d0*uLodi(3) - 9.d0*uN1Prev(3) + uN2Prev(3))

            ! wave amplitudes
            L1 = K_mod*csMod*csMod*(rhoLodi-rhoRef(iElem))
            L2 = (uLodi(3))*dudx
            L3 = (uLodi(3))*dvdx
            L4 = (uLodi(3))*csMod*csMod*drhodx*(1.d0-1.d0/kappa)
            L5 = ( uLodi(3) + swap * csMod )                                   &
              &* ( csMod * csMod * drhodx / kappa                              &
              &+   rho0ref * swap * csMod * dwdx )

            ! LODI Equations
            rhoLodi  = rhoLodi  - (L4+0.5d0*(L5+L1))/(csMod*csMod)
            uLodi(1) = uLodi(1) - L2
            uLodi(2) = uLodi(2) - L3
            uLodi(3) = uLodi(3) - ((L5-L1))/(rho0ref*swap*csMod)*0.5d0
          endif
        endif
      endif

      ! Now store the previous variables for next iteration
      me%elems( iLevel )%lodi( 1,   1, iElem ) = rhoLodi
      me%elems( iLevel )%lodi( 2:4, 1, iElem ) = uLodi 
      ! density and velocity of my first neighbor
      ! MH: Velocity and Density should be calculated with access as SAVE
      ! For pull, this is fine however, as IDX and SAVE coincide
      rhoTmp = 0.0_rk
      rhoTmp_2 = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
      ! MH: These should be calculated with a SAVE access!
        fTmpPost( iDir ) = bcBuffer(                                           &
      & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        fTmpPost_2(iDir) = me%elems( iLevel )%neighBufferPost( 1,              &
          &                  ?IDX?( iDir, iElem, layout%stencil(1)%QQ ))
        rhoTmp = rhoTmp + fTmpPost(iDir)
        rhoTmp_2 = rhoTmp_2 + fTmpPost_2(iDir)
      enddo
      !density
      me%elems( iLevel )%lodi( 1,   2, iElem ) = rhoTmp
      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      !velocity
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost,                                 &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,2,iElem) )

      ! density and velocity of my secnd neighbor
      me%elems( iLevel )%lodi( 1,   3, iElem )  = rhoTmp_2
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpPost_2,                               &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = me%elems(iLevel)%lodi(2:4,3,iElem) )

      ! Calculate the equilibrium distribution
      input_loc = (/ rhoLodi, uLodi(1), uLodi(2), uLodi(3) /)
      call varSys%derive( derVarPos%equilFromMacro )%do(                       &
        &              input   = fTmpPost_2,                                   &
        &              bound   = bound,                                        &
        &              tree    = tree,                                         &
        &              subTree = subTree,                                      &
        &              requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
        &              globSys = varSys,                                       &
        &              stencil = layout%stencil(1),                            &
        &              conf    = varSys%solSpec%conf,                          &
        &              res     = fEq )

      usq = uLodi(1)*uLodi(1) + uLodi(2)*uLodi(2) + uLodi(3)*uLodi(3) 
      ! Now do the anti-bounce back
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Now assign the values
 state( ?FETCH?( varPos( iDir ), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT)  = &
          fEq( iDir )
        endif
      enddo
    enddo

  end subroutine outlet_nrbc_eq
!******************************************************************************!


!******************************************************************************!
  !> Moment based open boundary condition from Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! for weakly compressible LBM model
  !! \author Kannan Masilamani
  !!
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'moments_press', 
  !!    pressure = 0.0 }
  !!\endverbatim
?? copy :: bc_routineHeader( moments_press )
    !---------------------------------------------------------------------------
    real(kind=rk) :: pressB( globBC%nElems(iLevel) ), rho, vel(3)
    integer :: iELem, iDir
    integer :: nLinks, iLink, elemPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp(:)
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !---------------------------------------------------------------------------
!    write(*,*) 'Boundary label ', trim(me%label)
    ! get transient density value with reference value from
    ! me%bc_states%pressure
    pressB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,     &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = sim_time,                              &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%general%solver%conf(1),         &
      &                       dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    pressB = pressB/physics%fac( iLevel )%press

    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    do iElem = 1, globBC%nElems( iLevel )
      fTmp = 0.0_rk
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(unKnown_fTmp(nLinks))
      allocate(rhs(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink + 1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state( ?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), currT )
        end if
      end do

      ! density from defined pressure
      rho = cs2inv*pressB(iElem) + fieldProp%fluid%rho0

      !local velocity
      vel = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        vel(:)  = vel(:) &
&+bcBuffer( ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))&
          & * layout%stencil(1)%cxDir(:, iDir)
      end do
      vel = vel/rho

      ! moments for in-compressible model
      first_moments = rho*vel !momX, momY, momZ 
      second_moments = (/ pressB(iElem) + rho*vel(1)*vel(1), & !momXX
        &                 pressB(iElem) + rho*vel(2)*vel(2), & !momYY
        &                 pressB(iElem) + rho*vel(3)*vel(3), & !momZZ
        &                 rho*vel(1)*vel(2),                 & !momXY
        &                 rho*vel(2)*vel(3),                 & !momYZ
        &                 rho*vel(3)*vel(1) /)                 !momXZ
      third_moments = (/ cs2*rho*vel(2), & !momXXY
        &                cs2*rho*vel(3), & !momXXZ
        &                cs2*rho*vel(1), & !momYYX
        &                cs2*rho*vel(3), & !momYYZ
        &                cs2*rho*vel(1), & !momZZX
        &                cs2*rho*vel(2) /) !momZZY
      !fourth_moments = cs2*pressB(iElem) !momXXYY, momYYZZ, momZZXX 
      fourth_moments = (/ &
        & cs2*(pressB(iElem)+rho*( vel(1)**2 + vel(2)**2 - 0.5_rk*vel(3)**2 )), &!momXXYY
        & cs2*(pressB(iElem)+rho*(-0.5_rk*vel(1)**2 + vel(2)**2 + vel(3)**2 )), &!momYYZZ
        & cs2*(pressB(iElem)+rho*( vel(1)**2 - 0.5_rk*vel(2)**2 + vel(3)**2 ))  &!momZZXX
        & /) 

      ! moments for pressure-wall corner
      ! KM: for pressure-velocity, velocity at boundary node must be defined
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      ! compute rhs of lse to compute unknown pdfs
      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do

      ! unknown pdfs
      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )

      do iLink = 1, nLinks
        state(                                                                 &
          & ?FETCH?( varPos(missing_links(iLink)), elemPos, nScalars, neigh ), &
          & currT)  = unKnown_fTmp(iLink)
      end do

     deallocate(missing_links)
     deallocate(unKnown_fTmp)
     deallocate(rhs)
   end do

  end subroutine moments_press
!******************************************************************************!



!******************************************************************************!
  !> Moment based open boundary condition from Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! for incompressible LBM model
  !! \author Kannan Masilamani
  !!
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'moments_press', 
  !!    pressure = 0.0 }
  !!\endverbatim
?? copy :: bc_routineHeader( moments_press_incomp )
    !---------------------------------------------------------------------------
    real(kind=rk) :: pressB( globBC%nElems(iLevel) ), rho0, rho, vel(3)
    integer :: iELem, iDir
    integer :: nLinks, iLink, elemPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp(:)
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !---------------------------------------------------------------------------
!    write(*,*) 'Boundary label ', trim(me%label)
    rho0 = fieldProp%fluid%rho0

    ! get transient density value with reference value from
    ! me%bc_states%pressure
    pressB = tem_spacetime_for( me      = me%bc_states%pressure%spacetime,     &
      &                       treeIDs = glob%levelDesc( iLevel )%total(globBC% &
      &                                               elems(iLevel)%elem(:) ), &
      &                       tree    = tree,                                  &
      &                       time    = sim_time,                              &
      &                       n       = globBC%nElems(iLevel),                 &
      &                       conf    = params%general%solver%conf(1),         &
      &                       dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    pressB = pressB/physics%fac( iLevel )%press

    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    do iElem = 1, globBC%nElems( iLevel )
      fTmp = 0.0_rk
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(unKnown_fTmp(nLinks))
      allocate(rhs(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink + 1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            & state( ?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), currT )
        end if
      end do

      !local velocity
      vel = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        vel(:)  = vel(:) &
&+bcBuffer( ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))&
          & * layout%stencil(1)%cxDir(:, iDir)
      end do
      vel = vel/rho0

      ! density from defined pressure
      rho = cs2inv*pressB(iElem) + rho0

      ! moments for in-compressible model
      !first_moments = (/ 0.0_rk, 0.0_rk, 0.0_rk /) !momX, momY, momZ 
      first_moments = rho0*vel !momX, momY, momZ 
      second_moments = (/ pressB(iElem) + rho0*vel(1)*vel(1), & !momXX
        &                 pressB(iElem) + rho0*vel(2)*vel(2), & !momYY
        &                 pressB(iElem) + rho0*vel(3)*vel(3), & !momZZ
        &                 rho0*vel(1)*vel(2),                 & !momXY
        &                 rho0*vel(2)*vel(3),                 & !momYZ
        &                 rho0*vel(3)*vel(1) /)                 !momXZ
      third_moments = (/ cs2*rho0*vel(2), & !momXXY
        &                cs2*rho0*vel(3), & !momXXZ
        &                cs2*rho0*vel(1), & !momYYX
        &                cs2*rho0*vel(3), & !momYYZ
        &                cs2*rho0*vel(1), & !momZZX
        &                cs2*rho0*vel(2) /) !momZZY
      !fourth_moments = cs2*pressB(iElem) !momXXYY, momYYZZ, momZZXX 
      fourth_moments = (/ &
        & cs2*(pressB(iElem)+rho0*( vel(1)**2 + vel(2)**2 - 0.5_rk*vel(3)**2 )), &!momXXYY
        & cs2*(pressB(iElem)+rho0*(-0.5_rk*vel(1)**2 + vel(2)**2 + vel(3)**2 )), &!momYYZZ
        & cs2*(pressB(iElem)+rho0*( vel(1)**2 - 0.5_rk*vel(2)**2 + vel(3)**2 ))  &!momZZXX
        & /) 

      ! moments for pressure-wall corner
      ! KM: for pressure-velocity, velocity at boundary node must be defined
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      ! compute rhs of lse to compute unknown pdfs
      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do

      ! unknown pdfs
      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )

      do iLink = 1, nLinks
        state(                                                                 &
          & ?FETCH?( varPos(missing_links(iLink)), elemPos, nScalars, neigh ), &
          & currT)  = unKnown_fTmp(iLink)
      end do

     deallocate(missing_links)
     deallocate(unKnown_fTmp)
     deallocate(rhs)
   end do

  end subroutine moments_press_incomp
!******************************************************************************!


!******************************************************************************!
  !> Moment based wall boundary condition from Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! \author Kannan Masilamani
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'moments_wall'} 
  !!\endverbatim
  !!
?? copy :: bc_routineHeader( moments_wall )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rhoB, ux(3), press
    integer :: iELem, iDir, tmpDir
    integer :: nLinks, iLink, elemPos, neighPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp( : )
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !---------------------------------------------------------------------------

    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    do iElem = 1, globBC%nElems( iLevel )

      fTmp = 0.0_rk
      rhoB = 0
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do  

      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          tmpDir = layout%stencil(1)%cxDirInv(iDir)
          if( globBC%elems(iLevel)%bitmask( tmpDir, iElem )) then
            tmpDir = layout%stencil(1)%cxDirInv(globBC%elems(iLevel)%normalInd(iElem))
          endif
        else
          tmpDir = iDir
        endif
        rhoB  = rhoB + fTmp(tmpDir)
      end do
      press = cs2 * rhoB

      ! moments for in-compressible model
      first_moments = (/ 0.0_rk, 0.0_rk, 0.0_rk /) !momX, momY, momZ 
      second_moments = (/ press, & !momXX
        &                 press, & !momYY
        &                 press, & !momZZ
        &                 0.0_rk,& !momXY
        &                 0.0_rk,& !momYZ
        &                 0.0_rk /)!momXZ
      third_moments = 0.0_rk
      fourth_moments = cs2*press !momXXYY, momYYZZ, momZZXX 

      ! moments for pressure-wall corner
      moments = 0.0_rk
      moments(1) = rhoB
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

!     write(*,*) 'Moments ', moments

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do
     !write(*,*) 'rhs ', rhs

     unKnown_fTmp = matmul( &
       & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )

     !write(*,*) 'unknown fTmp ', unKnown_fTmp

     do iLink = 1, nLinks
       state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
        & currT)  = unKnown_fTmp(iLink)
     end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

    end do

  end subroutine moments_wall
!******************************************************************************!


!******************************************************************************!
  !> Moment based velocity boundary condition for weakly compressible LBM model
  !! Based on Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! \author Kannan Masilamani
  !!  Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'moments_vel'
  !!    velocityX = 0.06, 
  !!    velocityY = 0.0, 
  !!    velocityZ = 0.0 }
  !!\endverbatim
 
?? copy :: bc_routineHeader( moments_vel )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rho, press, dens_correction, vel(3)
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel))
    integer :: iELem, iDir, tmpDir
    integer :: nLinks, iLink, elemPos, neighPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp( : )
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !---------------------------------------------------------------------------
    !write(*,*) 'Boundary label ', trim(me%label)
    ! Get velocity which might be defined to vary in space and time
    ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    !write(*,*) 'phy Vel', ux, uy, uz
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel

    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    do iElem = 1, globBC%nElems( iLevel )

      vel = (/ ux(iElem), uy(iElem), uz(iElem) /)

      fTmp = 0.0_rk
      rho = 0
      dens_correction = 0.0_rk
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do 

      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          tmpDir = layout%stencil(1)%cxDirInv(iDir)
          if( globBC%elems(iLevel)%bitmask( tmpDir, iElem )) then
            tmpDir = layout%stencil(1)%cxDirInv(globBC%elems(iLevel)%normalInd(iElem))
          endif
        else
          tmpDir = iDir
        endif
        rho  = rho + fTmp(tmpDir)
      end do

      ! add small correction term to density which comes from momentum in the
      ! flow direction while substituing unknowns pdfs terms obtained
      ! from moments BC derivation to compute density.
      ! With this correction term density is recovered correctly at this node
      dens_correction = dot_product(vel, globBC%elems(iLevel)%normal(:, iElem))
      ! for compressible model, rho_new = rho_old + rho_new*dens_correction
      ! thus, rho_new = rho_old/(1-dens_correction)
      rho = rho/(1.0_rk-dens_correction)

      press = cs2 * rho

      ! moments for in-compressible model
      first_moments = vel !momX, momY, momZ 
      second_moments = (/ press + rho*vel(1)*vel(1), & !momXX
        &                 press + rho*vel(2)*vel(2), & !momYY
        &                 press + rho*vel(3)*vel(3), & !momZZ
        &                 rho*vel(1)*vel(2),                 & !momXY
        &                 rho*vel(2)*vel(3),                 & !momYZ
        &                 rho*vel(3)*vel(1) /)                 !momXZ
      third_moments = (/ cs2*rho*vel(2), & !momXXY
        &                cs2*rho*vel(3), & !momXXZ
        &                cs2*rho*vel(1), & !momYYX
        &                cs2*rho*vel(3), & !momYYZ
        &                cs2*rho*vel(1), & !momZZX
        &                cs2*rho*vel(2) /) !momZZY
      !fourth_moments = cs2*pressB(iElem) !momXXYY, momYYZZ, momZZXX 
      fourth_moments = (/ &
        & cs2*(press+rho*( vel(1)**2 + vel(2)**2 - 0.5_rk*vel(3)**2 )), &!momXXYY
        & cs2*(press+rho*(-0.5_rk*vel(1)**2 + vel(2)**2 + vel(3)**2 )), &!momYYZZ
        & cs2*(press+rho*( vel(1)**2 - 0.5_rk*vel(2)**2 + vel(3)**2 ))  &!momZZXX
        & /) 

      ! moments for pressure-wall corner
      ! KM: for pressure-velocity, velocity at boundary node must be defined
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

    end do

  end subroutine moments_vel
!******************************************************************************!


!******************************************************************************!
  !> Moment based velocity boundary condition for incompressible LBM model 
  !! Based on Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! \author Kannan Masilamani
  !!  Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'inlet', 
  !!    kind = 'moments_vel'
  !!    velocityX = 0.06, 
  !!    velocityY = 0.0, 
  !!    velocityZ = 0.0 }
  !!\endverbatim
?? copy :: bc_routineHeader( moments_vel_incomp )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rho, rho0, press, dens_correction, vel(3)
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel))
    integer :: iELem, iDir, tmpDir
    integer :: nLinks, iLink, elemPos, neighPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp( : )
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !---------------------------------------------------------------------------
    !write(*,*) 'Boundary label ', trim(me%label)
    ! Get velocity which might be defined to vary in space and time
    ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    !write(*,*) 'phy Vel', ux, uy, uz
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel

    rho0 = fieldProp%fluid%rho0

    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    do iElem = 1, globBC%nElems( iLevel )

      vel = (/ ux(iElem), uy(iElem), uz(iElem) /)

      fTmp = 0.0_rk
      rho = 0
      dens_correction = 0.0_rk
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do 

      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          tmpDir = layout%stencil(1)%cxDirInv(iDir)
          if( globBC%elems(iLevel)%bitmask( tmpDir, iElem )) then
            tmpDir = layout%stencil(1)%cxDirInv(globBC%elems(iLevel)%normalInd(iElem))
          endif
        else
          tmpDir = iDir
        endif
        rho  = rho + fTmp(tmpDir)
      end do

      ! add small correction term to density which comes from momentum in the
      ! flow direction while substituing unknowns pdfs terms obtained
      ! from moments BC derivation to compute density.
      ! With this correction term density is recovered correctly at this node
      dens_correction = dot_product(vel, globBC%elems(iLevel)%normal(:, iElem))
      ! for compressible model, rho_new = rho_old + rho_new*dens_correction
      ! thus, rho_new = rho_old/(1-dens_correction)
      ! correction term for in-compressible mode
      rho = rho + rho0*dens_correction

      press = cs2 * rho

      ! moments for in-compressible model
      first_moments = vel !momX, momY, momZ 
      second_moments = (/ press + rho0*vel(1)*vel(1), & !momXX
        &                 press + rho0*vel(2)*vel(2), & !momYY
        &                 press + rho0*vel(3)*vel(3), & !momZZ
        &                 rho0*vel(1)*vel(2),                 & !momXY
        &                 rho0*vel(2)*vel(3),                 & !momYZ
        &                 rho0*vel(3)*vel(1) /)                 !momXZ
      third_moments = (/ cs2*rho0*vel(2), & !momXXY
        &                cs2*rho0*vel(3), & !momXXZ
        &                cs2*rho0*vel(1), & !momYYX
        &                cs2*rho0*vel(3), & !momYYZ
        &                cs2*rho0*vel(1), & !momZZX
        &                cs2*rho0*vel(2) /) !momZZY
      !fourth_moments = cs2*pressB(iElem) !momXXYY, momYYZZ, momZZXX 
      fourth_moments = (/ &
        & cs2*(press+rho0*( vel(1)**2 + vel(2)**2 - 0.5_rk*vel(3)**2 )), &!momXXYY
        & cs2*(press+rho0*(-0.5_rk*vel(1)**2 + vel(2)**2 + vel(3)**2 )), &!momYYZZ
        & cs2*(press+rho0*( vel(1)**2 - 0.5_rk*vel(2)**2 + vel(3)**2 ))  &!momZZXX
        & /)

      ! moments for pressure-wall corner
      ! KM: for pressure-velocity, velocity at boundary node must be defined
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

    end do

  end subroutine moments_vel_incomp
!******************************************************************************!

end module mus_bc_fluid_module
!******************************************************************************!
