!******************************************************************************!
!> This module contains general boundary routines 
!!
!! like initializing boundary, setting boundary at every time step for each 
!! field and other general boundary routines
!!
!! \author Manuel Hasert
?? include 'header/lbm_macros.inc'
module mus_bc_general_module

  ! include treelm modules
  use env_module,              only: rk, globalMaxLevels, long_k, labelLen
  use tem_param_module,        only: cs, qDirName
  use treelmesh_module,        only: treelmesh_type
  use tem_aux_module,          only: tem_abort
  use tem_var_system_module,   only: tem_var_system_type
  use tem_bc_module,           only: tem_bc_state_type
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_topology_module,     only: tem_levelOF
  use tem_subTree_type_module, only: tem_subTree_type
  use tem_debug_module,        only: dbgUnit
  use tem_logging_module,      only: logUnit
  use tem_math_module,         only: invert_matrix
  use tem_geometry_module,     only: tem_PosOfID, tem_determine_discreteVector

  ! include musubi modules
  use mus_field_module,                 only: mus_field_type
  use mus_field_prop_module,            only: mus_field_prop_type
  use mus_scheme_layout_module,         only: mus_scheme_layout_type
  use mus_scheme_header_module,         only: mus_scheme_header_type
  use mus_pdf_module,                   only: pdf_data_type, pdf_global_type,  &
    &                                         pdf_globInfo_type 
  use mus_bc_header_module,             only: boundary_type, glob_boundary_type
  use mus_bc_fluid_module,              only: inlet_ubb, inlet_eq, outlet_nrbc,&
    &                                         outlet_nrbc_eq, outlet_dnt,      &
    &                                         outlet_expol, outlet_eq,         &
    &                                         outlet_pab, outlet_zero_prsgrd,  &
    &                                         inlet_mfr, inlet_mfr_eq,         &
    &                                         moments_press, moments_wall,     &
    &                                         moments_vel, moments_vel_incomp, &
    &                                         moments_press_incomp
  use mus_bc_fluid_wall_module,         only: wall_bouzidi, slip_wall,         &
    &                                         wall_multiReflection
  use mus_bc_fluid_experimental_module, only: outlet_bouzidi, test_bcNeighs,   &
    &                                         outlet_extrapolation,            &
    &                                         outlet_extrapolation_microscopic,&
    &                                         moments_inflow, moments_outflow, &
    &                                         spc_moments_outflow,             &
    &                                         spc_bb_wall, spc_bb_vel_test
  use mus_bc_species_module,            only: spc_outlet_zero_prsgrd,          &
    &                                         spc_moleFrac, spc_moleFlux,      &
    &                                         spc_moleFrac_wtdf,               &
    &                                         spc_moleDiff_Flux, spc_inlet_eq, &
    &                                         spc_outlet_eq, spc_outlet_vel,   &
    &                                         spc_outlet_expol,                &
    &                                         spc_moleFlux_eq, spc_vel_bb,  &
    &                                         spc_blackbox_mem_ion,            &
    &                                         spc_blackbox_mem_solvent,        &
    &                                         spc_moments_moleFrac,            &
    &                                         spc_moments_moleFlux,            &
    &                                         spc_moments_wall,                &
    &                                         spc_moments_vel,                 &
    &                                         spc_moments_blackbox_mem_ion
  use mus_variable_module,              only: mus_derVarPos_type
  use mus_param_module,                 only: mus_param_type
  use mus_physics_module,               only: mus_physics_type
  use mus_mixture_module,               only: mus_mixture_type

  implicit none

  private

  public :: set_boundary
  public :: init_boundary

contains


!******************************************************************************!
  !> Call the functions associated with each boundary condition
  !! 
  !! Loop over each field and Run over all the boundary conditions for 
  !! current iLevel and call the function pointer. 
  !! The function pointer was before assigned in init_boundary
  !! This routine is being called from the 
  !! \ref mus_control_module::do_recursive_multischeme "main control routine"
  !! before the compute (advection_relaxation) kernel call
  subroutine set_boundary( field, pdf, glob, tree, iLevel, pdfGlobal,          &
    &                      params, layout, varSys, derVarPos, globBC,          &
    &                      mixture, physics )
    !--------------------------------------------------------------------------- 
    !> fluid parameters and properties
    type( mus_field_type ), intent(inout) :: field(:)
    !> state and neighbor arrays for all iLevels
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !>gloabl type contains iLevel descriptor
    type( pdf_global_type ), intent(in) :: glob
    !> global treelm mesh 
    type( treelmesh_type ), intent(in) ::tree
    !> the iLevel on which this boundary was invoked
    integer, intent(in) :: iLevel
    !> global information 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal 
    !> global parameters
    type(mus_param_type),intent(in) :: params   
    !> stencil layout information
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !> scheme global boundary type
    type( glob_boundary_type ), intent(in) :: globBC(:)
    !> scheme global boundary type
    type( mus_physics_type ), intent(in) :: physics
    !> mixture info
    type(mus_mixture_type), intent(in) :: mixture
    !--------------------------------------------------------------------------- 
    integer :: iField, nFields
    integer :: iBnd   ! counter for the boundary condition
    integer :: nStateVars, iVar, iComp, iPos
    integer, allocatable :: stateVarPos(:)
    !--------------------------------------------------------------------------- 

    nFields = size( field )

    ! get the correct varpos for each field, only works for ONE field!
    iPos = 1
    allocate( stateVarPos( varSys%nScalars / nFields ) )
    nStateVars = varSys%nVars / nFields
    do iVar = 1, nStateVars
      do iComp = 1, varSys%Variable%val( iVar )%nComponents
        stateVarPos( iPos ) = varSys%variable%val( iVar )%varPos( iComp )
        iPos = iPos + 1
      end do
    end do

    if ( pdfGlobal%nBCs > 0 ) then
      call fill_neighBuffer(                                                 &
        &            currState = pdf( iLevel )%state( :, pdf(iLevel)%nNow ), &
        &            bcBuffer  = pdf( iLevel )%bcBuffer,                     &
        &            neigh     = pdf( iLevel )%neigh,                        &
        &            glob      = glob,                                       &
        &            globBC    = globBC,                                     &
        &            pdfGlobal = pdfGlobal,                                  &
        &            field     = field,                                      &
        &            varSys    = varSys,                                     &
        &            layout    = layout,                                     &
        &            iLevel    = iLevel )

      ! loop over fields
      do iField = 1, nFields

        ! Treat all boundary conditions
        do iBnd = 1, pdfGlobal%nBCs
          write(dbgUnit(10),*) 'setting bnd'// trim( globBC( iBnd )%label )
          if( associated( field(iField)%bc( iBnd )%fnct ))   then
            call field(iField)%bc( iBnd )%fnct(                              &
              &      state     = pdf( iLevel )%state( :, : ),                &
              ! we want to write into the current time step
              ! the field pointers have been inverted, so the newest
              ! information is stored in nNow
              ! So we want to proivde the information for the coming
              ! kernel update, which reads the info from nNow with FETCH
              &      currT        = pdf(iLevel)%nNow,                        &
              &      lastT        = pdf(iLevel)%nNext,                       &
              &      bcBuffer     = pdf( iLevel )%bcBuffer,                  &
              &      globBC       = globBC( iBnd ),                          &
              &      glob         = glob,                                    &
              &      tree         = tree,                                    &
              &      iLevel       = iLevel,                                  &
              &      sim_time     = params%general%simControl%now,           &
              &      params       = params,                                  &
              &      physics      = physics,                                 &
              &      mixture      = mixture,                                 &
              &      IC_fields    = field(:)%ic,                             &
              &      iField       = iField,                                  &
              &      neigh        = pdf( iLevel )%neigh,                     &
              &      layout       = layout,                                  &
              &      fieldProp    = field(iField)%fieldProp,                 &
              &      varSys       = varSys,                                  &
              &      derVarPos    = derVarPos(iField),                       &
              &      derVarPosAll = derVarPos(:),                            &
              ! each field may has multiple state variables
              &      varPos       = varSys%Variable%val(iField)%varPos,      &
              ! &      varPos       = stateVarPos,                             &
              &      nScalars     = varSys%nScalars )
          end if
        end do  ! iBnd
      end do  ! iField
    end if

    deallocate( stateVarPos )

  end subroutine set_boundary
!******************************************************************************!


!******************************************************************************!
  ! > This routine call init_boundary to initialize boundary for each field 
  !!
  subroutine init_boundary( field, pdf, tree, glob, pdfGlobal, layout,         &
    &                       schemeHeader, varSys, derVarPos, globBC, bc_prop )
    !---------------------------------------------------------------------------
    !> fluid parameters and properties
    type( mus_field_type ), intent(inout) :: field(:)
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global treelm mesh
    type( treelmesh_type ), intent(in) ::tree
    !> global info type
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> scheme layout type
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> scheme header info
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !> global pdf type
    type( pdf_global_type ), intent(in) :: glob
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !> scheme global boundary type
    type( glob_boundary_type ), intent(inout) :: globBC(:)
    !> boundary property type
    type( tem_bc_prop_type ) :: bc_prop
    !---------------------------------------------------------------------------
    integer :: iField
    !---------------------------------------------------------------------------

    if ( pdfGlobal%nBCs > 0 ) then
      do iField = 1, size(field)
        call init_boundary_single(                                             &
          &                 bc           = field(iField)%bc,                   &
          &                 pdf          = pdf,                                &
          &                 tree         = tree,                               &
          &                 fieldProp    = field(iField)%fieldProp,            &
          &                 glob         = glob,                               &
          &                 pdfGlobal    = pdfGlobal,                          &
          &                 layout       = layout,                             &
          &                 schemeHeader = schemeHeader,                       &
          &                 varPos       = varSys%Variable%val(iField)%varPos, &
          &                 nScalars     = varSys%nScalars,                    &
          &                 varSys       = varSys,                             &
          &                 dervarPos    = derVarPos(iField),                  &
          &                 globBC       = globBC,                             &
          &                 bc_prop      = bc_prop )
      end do
    end if

  end subroutine init_boundary
!******************************************************************************!


!******************************************************************************!
  !> This subroutine sets the right boundary conditions for the different
  !! boundaries.
  !!
  subroutine init_boundary_single(bc, pdf, tree, fieldProp, glob, pdfGlobal,   &
    &                             layout, schemeHeader, varPos, nScalars,      &
    &                             varSys, derVarPos, globBC, bc_prop )
    !--------------------------------------------------------------------------- 
    !> global array boundary type
    type( boundary_type ) :: bc(:)
    !> contains global state vector
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global treelm mesh
    type( treelmesh_type ), intent(in) ::tree
    !> fluid parameters
    type( mus_field_prop_type ), intent(in) :: fieldProp
    !> global info type
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> scheme layout type
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> scheme header info
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !> global pdf type
    type( pdf_global_type ), intent(in) :: glob
    !> varPos of current field variable
    integer, intent(in) :: varPos(:)
    !> number of scalars in global system
    integer, intent(in) :: nScalars
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos
    !> scheme global boundary type
    type( glob_boundary_type ), intent(inout) :: globBC(:)
    !> boundary property type
    type( tem_bc_prop_type ) :: bc_prop
    !--------------------------------------------------------------------------- 
    integer :: iBnd
    !--------------------------------------------------------------------------- 

    !> The boundary conditions were read in the order of config.lua
    !! the order in the boundary condition description file however 
    !! might be different.
    !! We have to match the labels.
    write(logUnit(1),'(A,I0,A)') ' Setting up ', pdfGlobal%nBCs, ' boundary conditions'

    do iBnd = 1, pdfGlobal%nBCs
      if( trim(bc( iBnd )%label ) /= trim( globBC( iBnd )%label )) then
        write(logUnit(1),*) 'Error: The boundary with label from the mesh:'    &
          &              //trim( globBC(iBnd)%label ) 
        write(logUnit(1),*) 'does not match the one specified in the lua file:'&
          &              //trim(bc( iBnd  )%label)
        write(logUnit(1),*) 'Stopping ...' 
        call tem_abort()
      end if

      write(logUnit(1),*) 'Initializing BC with label: '//trim(bc(iBnd)%label)

      select case (trim(bc( iBnd )%BC_kind))
      case('wall')
      case('spc_bb_wall')
        bc( iBnd )%fnct => spc_bb_wall
      case('spc_bb_vel_test')
        bc( iBnd )%fnct => spc_bb_vel_test
      case('wall_linearInterpolation')
        bc( iBnd )%fnct => wall_bouzidi
        ! Here we have to allocate and set the q-values if it is not
        ! provided by seeder and set the qVal from musubi.lua
        if( .not. globBC(iBnd)%hasQVal .and.                                   &
          & .not. globBC(iBnd)%qValInitialized ) then
          call init_qVals( bc        = bc( iBnd ),                             &
            &              tree      = tree,                                   &
            &              bc_prop   = bc_prop,                                &
            &              pdfGlobal = pdfGlobal,                              &
            &              layout    = layout,                                 &
            &              globBC    = globBC(iBnd),                           &
            &              iBnd      = int(iBnd, kind=long_k) )
        endif
      case('wall_multireflection','wall_mr')
        bc( iBnd )%fnct => wall_multiReflection
        ! Here we have to allocate and set the q-values
        !call init_qVals( bc = bc( iBnd ), tree = tree, bc_prop = bc_prop,      &
        !  &              pdfGlobal = pdfGlobal, layout = layout,               &
        !  &              globBC = globBC(iBnd), iBnd = int(iBnd, kind=long_k) )
      case('slip_wall')
        bc( iBnd )%fnct => slip_wall
      case('inlet_ubb')
        bc( iBnd )%fnct => inlet_ubb
      case('inlet_eq')
        bc( iBnd )%fnct => inlet_eq
      case('inlet_mfr')
        bc( iBnd )%fnct => inlet_mfr
      case('inlet_mfr_eq')
        bc( iBnd )%fnct => inlet_mfr_eq
      case('outlet_nrbc','outlet_nrbc_eq')
        select case (trim(bc( iBnd  )%BC_kind))
          case( 'outlet_nrbc' )
            bc( iBnd )%fnct => outlet_nrbc
          case( 'outlet_nrbc_eq' )
            bc( iBnd )%fnct => outlet_nrbc_eq
        end select
        call init_nrbc( bc        = bc(iBnd),                                  &
          &             pdf       = pdf,                                       &
          &             fieldProp = fieldProp,                                 &
          &             layout    = layout,                                    &
          &             pdfGlobal = pdfGlobal,                                 &
          &             varPos    = varPos,                                    &
          &             nScalars  = nScalars,                                  &
          &             varSys    = varSys,                                    &
          &             derVarPos = derVarPos,                                 &
          &             globBC    = globBC(iBnd),                              &
          &             tree      = tree ) 

      case('outlet_zfg','outlet_extrapolation_microscopic')
        bc( iBnd )%fnct => outlet_extrapolation_microscopic
        write(logUnit(1),*) 'Microscopic extrapolation. Zero Gradient for pdfs'

      case('outlet_dnt')
        bc( iBnd )%fnct => outlet_dnt
      case('outlet_expol')
        bc( iBnd )%fnct => outlet_expol
      case('outlet_eq')
        bc( iBnd )%fnct => outlet_eq
      case('outlet_pab')
        bc( iBnd )%fnct => outlet_pab 
      case('outlet_bouzidi')
        bc( iBnd )%fnct => outlet_bouzidi 
      case('outlet_zero_prsgrd')
        bc( iBnd )%fnct => outlet_zero_prsgrd 
      ! passive scalar boundary conditions
      case('test')
        bc( iBnd )%fnct => test_bcNeighs
      ! multispecies boundary conditions
      case('spc_outlet_zero_prsgrd')
        bc( iBnd )%fnct => spc_outlet_zero_prsgrd 
      case('spc_inlet_eq')
        bc( iBnd )%fnct => spc_inlet_eq
      case('spc_vel_bb')
        bc( iBnd )%fnct => spc_vel_bb
      case('spc_outlet_eq')
        bc( iBnd )%fnct => spc_outlet_eq
      case('spc_outlet_expol')
        bc( iBnd )%fnct => spc_outlet_expol
      case('spc_outlet_vel')
        bc( iBnd )%fnct => spc_outlet_vel
      case('spc_molefrac')
        bc( iBnd )%fnct => spc_moleFrac
      case('spc_molefrac_wtdf')
        bc( iBnd )%fnct => spc_moleFrac_wtdf
      case('spc_moleflux')
        bc( iBnd )%fnct => spc_moleFlux
      case('spc_moleflux_eq')
        bc( iBnd )%fnct => spc_moleFlux_eq
      case('spc_molediff_flux')
        bc( iBnd )%fnct => spc_moleDiff_Flux
      case( 'spc_blackbox_mem_ion')  
        bc( iBnd )%fnct => spc_blackbox_mem_ion
      case( 'spc_blackbox_mem_solvent')  
        bc( iBnd )%fnct => spc_blackbox_mem_solvent
      case( 'moments_press', 'moments_wall', 'moments_vel', 'moments_inflow',   &
        &   'moments_outflow',                                                  &
        &   'spc_moments_molefrac', 'spc_moments_moleflux', 'spc_moments_wall', &
        &   'spc_moments_vel', 'spc_moments_outflow',                           &
        &   'spc_moments_blackbox_mem_ion')  
        select case (trim(bc( iBnd  )%BC_kind))
        case( 'moments_press')
          select case (trim(schemeHeader%kind))
          case ('lbm')
            bc( iBnd )%fnct => moments_press
          case ('lbm_incomp')
            bc( iBnd )%fnct => moments_press_incomp
          case default
            write(logUnit(1),*) 'Chosen boundary kind '                        &
              &                 //trim(bc(iBnd)%BC_kind)//' is not supported'  &
              &                //'scheme kind'//trim(schemeHeader%kind)
            call tem_abort()  
          end select
        case( 'moments_wall')  
          bc( iBnd )%fnct => moments_wall
        case( 'moments_vel')  
          select case (trim(schemeHeader%kind))
          case ('lbm')
            bc( iBnd )%fnct => moments_vel
          case('lbm_incomp')
            bc( iBnd )%fnct => moments_vel_incomp
          case default
            write(logUnit(1),*) 'Chosen boundary kind '                        &
              &                 //trim(bc(iBnd)%BC_kind)//' is not supported'  &
              &                //'scheme kind'//trim(schemeHeader%kind)
            call tem_abort()  
          end select
        case( 'moments_inflow')  
          bc( iBnd )%fnct => moments_inflow
        case( 'moments_outflow') 
          bc( iBnd )%fnct => moments_outflow
        case( 'spc_moments_molefrac')  
          bc( iBnd )%fnct => spc_moments_moleFrac
        case( 'spc_moments_moleflux')  
          bc( iBnd )%fnct => spc_moments_moleFlux
        case( 'spc_moments_outflow')  
          bc( iBnd )%fnct => spc_moments_outflow
        case( 'spc_moments_wall')  
          bc( iBnd )%fnct => spc_moments_wall
        case( 'spc_moments_vel')  
          bc( iBnd )%fnct => spc_moments_vel
        case( 'spc_moments_blackbox_mem_ion')  
          bc( iBnd )%fnct => spc_moments_blackbox_mem_ion
        end select
        call init_momentsBC( bc        = bc(iBnd),     &
          &                  glob      = glob,         &
          &                  pdfGlobal = pdfGlobal,    &
          &                  layout    = layout,       &
          &                  globBC    = globBC(iBnd), &
          &                  tree      = tree,         &
          &                  pdf       = pdf,          &
          &                  varPos    = varPos,       &
          &                  nScalars = nScalars ) 
      case default 
        write(logUnit(1),*)'BC type '//trim(bc( iBnd  )%bc_kind)// &
          & 'for label'//trim(bc( iBnd  )%label)//'not found'
        call tem_abort()
      end select
    end do

  end subroutine init_boundary_single
!******************************************************************************!

  !****************************************************************************!
  !> Initialize the values required for the moments BC
  subroutine init_momentsBC( bc, glob, pdfGlobal, layout, globBC, tree, pdf,   &
    &                        varPos, nScalars )
    !---------------------------------------------------------------------------
    !> global array boundary type
    type( boundary_type ), intent(inout) :: bc
    !> global pdf type
    type( pdf_global_type ), intent(in) :: glob
    !> global information 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal 
    type( mus_scheme_layout_type), intent(in) :: layout
    !> scheme global boundary type
    type( glob_boundary_type ), intent(inout) :: globBC
    !> global treelm mesh
    type( treelmesh_type ), intent(in) ::tree
    !> contains global state vector
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> varPos of current field variable
    integer, intent(in) :: varPos(:)
    !> number of scalars in global system
    integer, intent(in) :: nScalars
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, iDir, iDims
    integer :: d2q9_xNormal(3), d2q9_yNormal(3)
    integer :: d3q19_xNormal(5), d3q19_yNormal(5), d3q19_zNormal(5)
    integer, allocatable, dimension(:) :: xNormal_mom, yNormal_mom,            &
      &                                   zNormal_mom,                         &
      &                                   xyNormal_mom, yzNormal_mom,          &
      &                                   xzNormal_mom,                        &
      &                                   xyzNormal_mom
    integer, allocatable, dimension(:,:) :: xNorm_links, yNorm_links,          &
      &                                     zNorm_links, xyNorm_links,         &
      &                                     yzNorm_links, xzNorm_links,        &
      &                                     xyzNorm_links
    character(len=labelLen) :: normalIndex  
    real(kind=rk) :: normal(3)
    integer :: nLinks, iLink, normal_nLinks, edge_nLinks, corner_nLinks
    integer, allocatable :: missing_links(:)
    integer :: edge_normal(3)
    real(kind=rk), allocatable :: unKnown_Mat(:,:)
    integer :: idPos, Neighbor, elemPos, normalIndTreelm
    logical :: bitmask( layout%stencil(1)%QQ )
    integer(kind=long_k), allocatable :: corner_elems(:)
    integer(kind=long_k) :: treeID
    logical :: corner_node, update_allMoments
    integer :: iCorner
    !---------------------------------------------------------------------------
!    write(dbgUnit(1),*) 'Boundary label ', trim(bc%label)

    ! \todo KM: move this generic info to tem_stencil_module or tem_param_module
    ! known moments positions in moments array
    select case (trim(bc%BC_kind))
    case('moments_press', 'moments_outflow', 'spc_moments_molefrac')
      d2q9_xNormal = (/ 1, 5, 3/) !m0, mY, mYY
      d2q9_yNormal = (/ 1, 4, 2/) !m0, mX, mXX
      d3q19_xNormal = (/ 1, 3, 4, 6, 7/) !m0, mY, mZ, mYY, mZZ
      d3q19_yNormal = (/ 1, 2, 4, 5, 7/) !m0, mX, mZ, mXX, mZZ
      d3q19_zNormal = (/ 1, 2, 3, 5, 6/) !m0, mX, mY, mXX, mYY
    case('moments_wall','moments_vel','moments_inflow', &
      &  'spc_moments_moleflux','spc_moments_wall','spc_moments_vel', &
      & 'spc_moments_outflow', 'spc_moments_blackbox_mem_ion' )
      ! ordering is based on such that 1st knownMoments pos is
      ! moment in normal direction and rest follows
      d2q9_xNormal = (/ 2, 3, 5/) !mX, mY, mYY
      d2q9_yNormal = (/ 3, 2, 4/) !mX, mY, mXX
      d3q19_xNormal = (/ 2, 3, 4, 6, 7/) !mX, mY, mZ, mYY, mZZ
      d3q19_yNormal = (/ 3, 2, 4, 5, 7/) !mX, mY, mZ, mXX, mZZ
      d3q19_zNormal = (/ 4, 2, 3, 5, 6/) !mX, mY, mZ, mXX, mYY
    case default
      write(logUnit(1),*)'This boundary kind is not supported'
      call tem_abort()
    end select

    select case (trim(layout%stencil(1)%label))
    case('d2q9')
      normal_nLinks = 3
      ! in 2d edge and corner or same
      edge_nLinks = 5
      corner_nLinks = 5
      !axis-normal
      allocate(xNormal_mom(normal_nLinks))
      allocate(yNormal_mom(normal_nLinks))
      !edge normal
      allocate(xyNormal_mom(edge_nLinks))
      xNormal_mom = d2q9_xNormal 
      yNormal_mom = d2q9_yNormal 
      xyNormal_mom = (/ 1, 2, 3, 4, 5/) !m0, mX, mY, mXX, mYY

      ! normal links
      allocate(xNorm_links(normal_nLinks,2))
      xNorm_links = reshape((/ 1, 5, 6,    & !x-
        &                      3, 7, 8 /), & !x+
        &                  (/normal_nLinks,2/))

      allocate(yNorm_links(normal_nLinks,2))
      yNorm_links = reshape((/ 2, 5, 7,    & !y-
        &                      4, 6, 8 /), & !y+
        &                  (/normal_nLinks,2/))
        
      allocate(xyNorm_links(edge_nLinks,4))
      xyNorm_links = reshape((/ 1, 2, 5, 6, 7,   & !x-,y-
        &                       1, 4, 5, 6, 8,   & !x-.y+
        &                       2, 3, 5, 7, 8,   & !x+,y-
        &                       3, 4, 6, 7, 8 /),& !x+,y+
        &                   (/edge_nLinks,4/))
      allocate(zNorm_links(normal_nLinks,2))
      zNorm_links = 0
      allocate(yzNorm_links(edge_nLinks,4))
      yzNorm_links = 0
      allocate(xzNorm_links(edge_nLinks,4))
      xzNorm_links = 0
      allocate(xyzNorm_links(corner_nLinks,8))
      xyzNorm_links = 0
    case('d3q19')
      normal_nLinks = 5
      edge_nLinks = 9
      corner_nLinks = 12
      !axis-normal
      allocate(xNormal_mom(normal_nLinks))
      allocate(yNormal_mom(normal_nLinks))
      allocate(zNormal_mom(normal_nLinks))
      !edge normal
      allocate(xyNormal_mom(edge_nLinks))
      allocate(yzNormal_mom(edge_nLinks))
      allocate(xzNormal_mom(edge_nLinks))
      !corner normal
      allocate(xyzNormal_mom(corner_nLinks))
      xNormal_mom = d3q19_xNormal 
      yNormal_mom = d3q19_yNormal 
      zNormal_mom = d3q19_zNormal 
      ! m0, mX, mY, mZ, mXX, mYY, mZZ, mYZ, mZZX
      xyNormal_mom = (/ 1, 2, 3, 4, 5, 6, 7, 9, 15/)
      ! m0, mX, mY, mZ, mXX, mYY, mZZ, mXZ, mXXY
      yzNormal_mom = (/ 1, 2, 3, 4, 5, 6, 7, 10, 11/)
      ! m0, mX, mY, mZ, mXX, mYY, mZZ, mXY, mYYZ
      xzNormal_mom = (/ 1, 2, 3, 4, 5, 6, 7, 8, 14/)
      ! m0, mX, mY, mZ, mXX, mYY, mZZ, mXY, mYZ, mXXY, mYYX, mZZY
      xyzNormal_mom = (/ 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 13, 16/)

      ! normal links
      allocate(xNorm_links(normal_nLinks,2))
      xNorm_links = reshape((/ 1, 11, 13, 15, 16,    & !x- 
        &                      4, 12, 14, 17, 18 /), & !x+
        &                  (/normal_nLinks,2/))

      allocate(yNorm_links(normal_nLinks,2))
      yNorm_links = reshape((/ 2,  7,  8, 15, 17,    & !y-
        &                      5,  9, 10, 16, 18 /), & !y+
        &                  (/normal_nLinks,2/) ) 

      allocate(zNorm_links(normal_nLinks,2))
      zNorm_links = reshape((/ 3,  7,  9, 11, 12,    & !z-
        &                      6,  8, 10, 13, 14 /), & !z+
        &                  (/normal_nLinks,2/))

      allocate(xyNorm_links(edge_nLinks,4))
      xyNorm_links = reshape((/ 1,  2,  7,  8, 11, 13, 15, 16, 17,   & !x-,y-
        &                       1,  5,  9, 10, 11, 13, 15, 16, 18,   & !x-,y+
        &                       2,  4,  7,  8, 12, 14, 15, 17, 18,   & !x+,y-
        &                       4,  5,  9, 10, 12, 14, 16, 17, 18 /),& !x+,y+
        &                   (/edge_nLinks,4/))
        
      allocate(yzNorm_links(edge_nLinks,4))
      yzNorm_links = reshape((/ 2,  3,  7,  8,  9, 11, 12, 15, 17,   & !y-,z-
        &                       2,  6,  7,  8, 10, 13, 14, 15, 17,   & !y-,z+
        &                       3,  5,  7,  9, 10, 11, 12, 16, 18,   & !y+,z-
        &                       5,  6,  8,  9, 10, 13, 14, 16, 18 /),& !y+,z+
        &                   (/edge_nLinks,4/))
        
      allocate(xzNorm_links(edge_nLinks,4))
      xzNorm_links = reshape((/ 1,  3,  7,  9, 11, 12, 13, 15, 16,   & !x-,z-
        &                       1,  6,  8, 10, 11, 13, 14, 15, 16,   & !x-,z+
        &                       3,  4,  7,  9, 11, 12, 14, 17, 18,   & !x+,z-
        &                       4,  6,  8, 10, 12, 13, 14, 17, 18 /),& !x+,z+
        &                   (/edge_nLinks,4/))
        
      allocate(xyzNorm_links(corner_nLinks,8))
      xyzNorm_links = reshape((/ &
        &        1,  2,  3,  7,  8,  9, 11, 12, 13, 15, 16, 17,   & !x-,y-,z-
        &        1,  2,  6,  7,  8, 10, 11, 13, 14, 15, 16, 17,   & !x-,y-,z+
        &        1,  3,  5,  7,  9, 10, 11, 12, 13, 15, 16, 18,   & !x-,y+,z-
        &        1,  5,  6,  8,  9, 10, 11, 13, 14, 15, 16, 18,   & !x-,y+,z+
        &        2,  3,  4,  7,  8,  9, 11, 12, 14, 15, 17, 18,   & !x+,y-,z-
        &        2,  4,  6,  7,  8, 10, 12, 13, 14, 15, 17, 18,   & !x+,y-,z+
        &        3,  4,  5,  7,  9, 10, 11, 12, 14, 16, 17, 18,   & !x+,y+,z-
        &        4,  5,  6,  8,  9, 10, 12, 13, 14, 16, 17, 18 /),& !x+,y+,z+
        &                    (/corner_nLinks,8/))
    case default
      write(logUnit(1),*) trim(layout%stencil(1)%label)//&
        &' layout is not supported for this boundary'
      call tem_abort()
    end select


    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      allocate( bc%elems( iLevel )%moments( globBC%nElems( iLevel )))

      ! list of corner elements treeIDs
      allocate(corner_elems(globBC%cornerBC%nElems(iLevel)))
      corner_elems = glob%levelDesc(iLevel)%total( &
        & globBC%cornerBC%elems(iLevel)%elem(:))

      iCorner = 0
      do iElem = 1, globBC%nElems( iLevel )
        ! when moments combination are not defined properly 
        ! update all moments
        update_allMoments = .false.
        ! if corrent node is corner node i.e intersected by multiple boundaries
        corner_node = .false.

!write(dbgUnit(1),*) 'iElem ', iElem
        treeID = glob%levelDesc( iLevel )%total(  &
          &      globBC%elems(iLevel)%elem(iElem)) 
        if ( any(corner_elems .eq. treeID) ) corner_node = .true.
!write(dbgUnit(1),*) 'corner_node ', corner_node

        ! update bitmask, normal and normalInd with cornerBC information
        ! if corrent node is corner node
        if (corner_node) then
          iCorner = iCorner + 1
          globBC%elems(iLevel)%bitmask(:, iElem) = &
            & globBC%cornerBC%elems(iLevel)%bitmask(:, iCorner)
          globBC%elems(iLevel)%normal(:, iElem) =  &
            & globBC%cornerBC%elems(iLevel)%normal(:, iCorner)
          globBC%elems(iLevel)%normalInd(iElem) =  &
            & globBC%cornerBC%elems(iLevel)%normalInd(iCorner)
        end if ! corner node  

        ! element position in state array and treeID list
        elemPos = globBC%elems(iLevel)%elem(iElem)
        ! normal of this corrent node
        normal = globBC%elems(iLevel)%normal(:,iElem)
        bitmask = globBC%elems(iLevel)%bitmask(:, iElem)

        nLinks = count(globBC%elems(iLevel)%bitmask(:, iElem))
        bc%elems(iLevel)%moments(iElem)%nUnKnownPdfs = nLinks

        allocate(bc%elems(iLevel)%moments(iElem)%knownMom_pos(nLinks))
        allocate(missing_links(nLinks))
        iLink = 0
        do iDir = 1,layout%stencil(1)%QQ
          if (globBC%elems(iLevel)%bitmask(iDir, iElem)) then
            iLink = iLink+1
            missing_links(iLink) = iDir
          endif
        end do

        normalIndex = 'default'

        ! axis normals
        if (nLinks == normal_nLinks) then
          do iDir = 1, 2
            if ( all(missing_links == xNorm_links(:, iDir)) ) &
              & normalIndex = 'xNormal'
            if ( all(missing_links == yNorm_links(:, iDir)) ) &
              & normalIndex = 'yNormal'
            if ( all(missing_links == zNorm_links(:, iDir)) ) &
              & normalIndex = 'zNormal'
          end do
        else if (nLinks == edge_nLinks) then
        ! edge direction
          do iDir = 1, 4 
            if ( all(missing_links == xyNorm_links(:, iDir)) ) &
              & normalIndex = 'xyNormal'

            if ( all(missing_links == yzNorm_links(:, iDir)) ) &
              & normalIndex = 'yzNormal'

            if ( all(missing_links == xzNorm_links(:, iDir)) ) &
              & normalIndex = 'xzNormal'
          end do
        else if (nLinks == corner_nLinks) then
        ! corner direction
          do iDir = 1, 8 
            if ( all(missing_links == xyzNorm_links(:,iDir)) ) &
              & normalIndex = 'xyzNormal'
          end do    
        end if

!write(dbgUnit(1),*) 'normal ', globBC%elems(iLevel)%normal(:, iElem)
!write(dbgUnit(1),*) 'normalIndex ', normalIndex
!write(dbgUnit(1),*) 'missing_links ', missing_links
        select case(trim(normalIndex))
        case ('xNormal')
          bc%elems(iLevel)%moments(iElem)%knownMom_pos = xNormal_mom
        case ('yNormal')
          bc%elems(iLevel)%moments(iElem)%knownMom_pos = yNormal_mom
        case ('zNormal')
          bc%elems(iLevel)%moments(iElem)%knownMom_pos = zNormal_mom
        case ('xyNormal')
          bc%elems(iLevel)%moments(iElem)%knownMom_pos = xyNormal_mom
        case ('yzNormal')
          bc%elems(iLevel)%moments(iElem)%knownMom_pos = yzNormal_mom
        case ('xzNormal')
          bc%elems(iLevel)%moments(iElem)%knownMom_pos = xzNormal_mom
        case ('xyzNormal')
          bc%elems(iLevel)%moments(iElem)%knownMom_pos = xyzNormal_mom
        case default
          ! undefined normal so update all links for this boundary
          globBC%elems(iLevel)%bitmask(:, iElem) = .true.
          nLinks = layout%stencil(1)%QQ
          deallocate(bc%elems(iLevel)%moments(iElem)%knownMom_pos)
          allocate(bc%elems(iLevel)%moments(iElem)%knownMom_pos(nLinks))
          do iLink = 1,nLinks
            bc%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) = iLink
          enddo  

          deallocate(missing_links)
          allocate(missing_links(nLinks))
          iLink = 0
          do iDir = 1,layout%stencil(1)%QQ
            if (globBC%elems(iLevel)%bitmask(iDir, iElem)) then
              iLink = iLink+1
              missing_links(iLink) = iDir
            endif
          end do  
        end select

        bc%elems(iLevel)%moments(iElem)%nUnKnownPdfs = nLinks
        allocate(bc%elems(iLevel)%moments(iElem)%                  &
          &                       unKnownPdfs_MatInv(nLinks,nLinks))
        allocate(unKnown_Mat(nLinks,nLinks))

        do iLink = 1, nLinks
          unKnown_Mat(:,iLink) = layout%moment%toMoments%A( &
            & bc%elems(iLevel)%moments(iElem)%knownMom_pos, & 
            & missing_links(iLink) )
        end do  

        bc%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv &
          & = invert_matrix(unKnown_Mat)

        deallocate(unKnown_Mat)
        deallocate(missing_links)

      end do ! iElem  
    end do ! level  

  end subroutine init_momentsBC
  !****************************************************************************!


!******************************************************************************!
  !> JQ: assign qVal to corresponding BC and level-wise
  !! todo: first check if qVal is given by files,
  !! if not, try to receive qVal from lua file.
  subroutine init_qVals( bc, tree, bc_prop, pdfGlobal, layout, globBC, iBnd )
    !---------------------------------------------------------------------------
    type( boundary_type ) :: bc !< global array boundary type
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)        :: tree
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)      :: bc_prop
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)     :: pdfGlobal
    type( mus_scheme_layout_type), intent(in) :: layout
    type( glob_boundary_type ), intent(inout) :: globBC
    integer(kind=long_k), intent(in)          :: iBnd
    !---------------------------------------------------------------------------
    integer, allocatable :: nBnds(:) ! number of bc elements in each level
    integer(kind=long_k) :: bID
    integer :: iElem, iLevel, level, iDir, iQQQ, iElem_qVal
    logical :: found
    !---------------------------------------------------------------------------
    globBC%qValInitialized = .true.

    allocate( nBnds( globalMaxLevels ))
    nBnds = 0
    ! element pointer to qVal array in bc_prop
    iElem_qVal = 0

    ! 2. allocate qVal array level-wise
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      allocate( globBC%elems( iLevel )%qVal(                                   &
        &       layout%stencil(1)%QQN, globBC%nElems( iLevel )))
      globBC%elems( iLevel )%qVal = -1._rk
    end do

    ! loop over all boundary elements
    do iElem = 1, tree%property(1)%nElems

      found = .false.

      ! count elements that has prp_hasQVal property
      iElem_qVal = iElem_qval + 1

      ! run over each direction in the stencil
      do iDir = 1, layout%stencil(1)%QQN
        ! Stencil order of directions might differ from the treelm numbering
        ! iDir is musubi numbering
        ! iQQQ is treelm numbering.
        iQQQ = layout%stencil(1)%map( iDir )
        ! Get the boundary ID as stored in the mesh
        bID = bc_prop%boundary_ID( iQQQ, iElem )

        if( bID == iBnd ) then
          level = tem_LevelOf( tree%treeID( tree%property(1)%ElemID( iElem )))

          if( .not. found ) then
            nBnds( level ) = nBnds( level ) + 1
            found = .true.
          end if

          globBC%elems( level )%qVal( iDir, nBnds(level))  = bc%qVal
        end if
      end do ! iDir
    end do ! iElem

  end subroutine init_qVals
!******************************************************************************!


!******************************************************************************!
  !> Initialize the values required for the characteristic boundary conditions
  !!
  !! Simply calculate the macroscopic values from the current pdf distributions
  !! in the boundary element and its neighbor elements and store it to the 
  !! %nrbc%lodi field
  !! NOTE:   
  !! To be consistent, the same calculation procedure of the LODI values need to
  !! be performed as in the 
  !! \ref mus_bc_fluid_module::outlet_nrbc "NRBC routine" itself
  !!
  subroutine init_nrbc( bc, pdf, fieldProp, layout, pdfGlobal, varPos,         &
    &                   nScalars, varSys, derVarPos, globBC, tree )
    !---------------------------------------------------------------------------
    !> global array boundary type
    type( boundary_type ) :: bc
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> fluid parameters
    type( mus_field_prop_type ), intent(in) :: fieldProp
    type( mus_scheme_layout_type), intent(in) :: layout
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal 
    !> varPos of current field variable
    integer, intent(in) :: varPos(:)
    !> number of scalars in global system
    integer, intent(in) :: nScalars
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> position of derived quantities in varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos
    !> scheme global boundary type
    type( glob_boundary_type ), intent(in) :: globBC
    !> global treelm mesh
    type( treelmesh_type ), intent(in) ::tree
    !---------------------------------------------------------------------------
    integer :: iLevel, iElem, iDir
    ! Neighbor element variables
    integer :: iNeigh, neighPos
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    if( bc%nrbc%lodi_length == 0._rk) &
        bc%nrbc%lodi_length = 1.0_rk
    write(logUnit(1),*) 'Initializing NRBC with parameters:' 
    write(logUnit(1),*) '    kappa:         ',bc%nrbc%kappa
    write(logUnit(1),*) '    sigma:         ',bc%nrbc%sigma
    write(logUnit(1),*) '    Char. length:  ',bc%nrbc%lodi_length
    bc%nrbc%cs_mod = cs*sqrt( bc%nrbc%kappa )

    ! Assign the initial values to the lodi variables
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      allocate( bc%elems( iLevel )%lodi( 4, 3, globBC%nElems( iLevel )))
      ! reset 
      do iElem = 1, globBC%nElems( iLevel )
        ! adopt initial density from the fluid cells
        bc%elems( iLevel )%lodi( :, 1, iElem) = 0._rk
        do iDir = 1,layout%stencil(1)%QQ
          fTmp(iDir) = pdf(iLevel)%state(                                      &
& ?SAVE?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars,pdf(iLevel)%neigh ), &
& pdf( iLevel )%nNext ) 

          bc%elems( iLevel )%lodi( 1, 1, iElem ) =                             &
            &               bc%elems( iLevel )%lodi( 1, 1, iElem ) + fTmp(iDir) 
        end do    
          
        ! initialize the lower and upper bound
        bound(1) = 1
        bound(2) = 1
        ! adapt initial velocities from the fluid cells
        call varSys%derive( derVarPos%velocity )%do(                           &
          &                input   = fTmp,                                     &
          &                bound   = bound,                                    &
          &                tree    = tree,                                     &
          &                subTree = subTree,                                  &
          &                requVar = varSys%Variable%val( derVarPos%velocity ),&
          &                globSys = varSys,                                   &
          &                stencil = layout%stencil(1),                        &
          &                conf    = varSys%solSpec%conf,                      &
          &                res     = bc%elems(iLevel)%lodi(2:4,1,iElem) )

        do iNeigh = 1,2
          ! Get the values for the neighbors as well
          neighPos = bc%elems( iLevel )%neigh( iNeigh, iElem )
          bc%elems( iLevel )%lodi( :, 1+iNeigh, iElem) = 0._rk
          if( neighPos > 0 ) then
            do iDir = 1,layout%stencil(1)%QQ
              fTmp(iDir) = pdf(iLevel)%state(                                  &
& ?SAVE?( varPos(iDir), neighPos, nScalars, pdf(iLevel)%neigh ), pdf( iLevel )%nNext ) 
              bc%elems( iLevel )%lodi( 1, 1+iNeigh, iElem ) =                  &
                &    bc%elems( iLevel )%lodi( 1, 1+iNeigh, iElem ) + fTmp(iDir)
            end do

            ! velocities
            call varSys%derive( derVarPos%velocity )%do(                       &
              &            input   = fTmp,                                     &
              &            bound   = bound,                                    &
              &            tree    = tree,                                     &
              &            subTree = subTree,                                  &
              &            requVar = varSys%Variable%val( derVarPos%velocity ),&
              &            globSys = varSys,                                   &
              &            stencil = layout%stencil(1),                        &
              &            conf    = varSys%solSpec%conf,                      &
              &            res     = bc%elems(iLevel)%lodi(2:4,1+iNeigh,iElem))

          end if
        end do
      end do
    end do

  end subroutine init_nrbc
!******************************************************************************!


!******************************************************************************!
  !> This routine will fill BC buffer with state values of boundary element
  !! of all fields from current time step
  !!
  !! \todo KM: Please give reasons why the access patterns are used
  !! Access to state array
  !! ---------------------
  !! * bcBuffer:        SAVE 
  !! * NeighBufferPre: FETCH
  !! * NeighBufferPost: SAVE 
  !!
  subroutine fill_neighBuffer( currState, bcBuffer, neigh, glob, globBC,       &
    &                          pdfGlobal, field, varSys, layout, iLevel )
    !--------------------------------------------------------------------------- 
    !> Current state vector of iLevel  
    real(kind=rk),intent(in) :: currState(:) 
    !> state values of boundary elements of all fields of iLevel
    real(kind=rk),intent(inout) :: bcBuffer(:) 
    !> connectivity array corresponding to state vector
    integer,intent(in) :: neigh(:)       
    !>gloabl type contains iLevel descriptor
    type( pdf_global_type ), intent(in) :: glob
    !> scheme global boundary type
    type( glob_boundary_type ), intent(in) :: globBC(:)
    !> global information 
    type( pdf_globInfo_type), intent(in) :: pdfGlobal 
    !> fluid parameters and properties
    type( mus_field_type ), intent(inout) :: field(:)
    !> scheme variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> stencil layout information
    type( mus_scheme_layout_type ), intent(in) ::layout
    !> the iLevel on which this boundary was invoked
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    !!SZ: \todo: this is only valid in case each field only owns one state
    !!           variable!!!!!
    integer :: iField
    integer :: iElem, iDir, iBnd, iNeigh
    integer :: pos, elemPos
    !---------------------------------------------------------------------------

    do iField = 1, size( field )
      do iElem = 1, glob%levelDesc( iLevel )%bc_elemBuffer%nVals 
        elemPos =  glob%levelDesc( iLevel )%bc_elemBuffer%val( iElem )
        do iDir = 1, layout%stencil(1)%QQ
          pos = varSys%Variable%val( iField )%varPos( iDir )
          bcBuffer( ?IDX?( pos, iElem, varSys%nScalars ) ) = currState(        &
& ?SAVE?( pos, elemPos, varSys%nScalars, neigh ) )
!& ?FETCH?( pos, elemPos, varSys%nScalars, neigh ) )
          ! MH: Why do we use SAVE access here?
          ! KM: Store final result of compute kernel 
        end do ! iDir
      end do ! iElem
    end do ! iField

    do iField = 1, size( field )
      !fill boundary neighbor pre- and post-collision state values 
      do iBnd = 1, pdfGlobal%nBCs
        do iNeigh = 1, field(iField)%bc( iBnd )%nNeighs
          do iElem = 1, globBC( iBnd )%nElems( iLevel )
            elemPos =  field(iField)%bc( iBnd )%elems( iLevel )%               &
              &                                          neigh( iNeigh, iElem )
            do iDir = 1, layout%stencil(1)%QQ
              pos = varSys%Variable%val( iField )%varPos( iDir )
              !pre-collision neigh
              field(iField)%bc( iBnd )%elems( iLevel )%neighBufferPre(         &
& iNeigh, ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) ) = currState(            & 
& ?FETCH?( pos, elemPos, varSys%nScalars, neigh ) )
                !post-collision neigh  
                field(iField)%bc( iBnd )%elems( iLevel )%neighBufferPost(      &          
& iNeigh, ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) ) = currState(            & 
& ?SAVE?( pos, elemPos, varSys%nScalars, neigh ) )
            end do ! iDir
          end do ! iElem
        end do ! iNeigh
      end do ! iBnd
    end do ! iField

  end subroutine fill_neighBuffer
!******************************************************************************!


end module mus_bc_general_module
!******************************************************************************!
