!******************************************************************************!
!> This module contains boundary type definitions, different boundary treatment
!! types, abtract interface and routine to load boundary table from config file
!!
!! The boundary conditions are stored in boundary type definitions for each 
!! boundary which is defined in the main lua file. 
!! \verbatim
!! boundary_condition = {  }
!! \endverbatim
!!
!! The definitions there must equal the boundary definition within the mesh on 
!! disk, in terms of the amount of boundary conditions and the labels.
!! A detailed description on the implementation details are given in 
!! \ref tem_bc_module "boundary condition implementation"
!!
!! This module will be used as a header to use boundary definitions in other
!! modules.
!!
module mus_bc_header_module

  ! include treelm modules
  use env_module,               only: rk, globalMaxLevels, LabelLen
  use tem_aux_module,           only: tem_abort
  use tem_bc_module,            only: tem_bc_state_type, tem_load_bc_state
  use tem_bc_header_module,     only: tem_load_bc_header, tem_bc_header_type
  use tem_bc_prop_module,       only: tem_bc_prop_type
  use tem_tools_module,         only: tem_horizontalSpacer
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_var_system_module,    only: tem_var_system_type
  use tem_ini_condition_module, only: tem_ini_condition_type
  use tem_logging_module,       only: logUnit
  use tem_stlb_io_module,       only: tem_dump_stlb

  ! include aotus modules
  use aotus_module,     only: flu_State, aoterr_Fatal, aoterr_NonExistent,     &
    &                         aoterr_WrongType, aot_top_get_val, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_pdf_module,           only: pdf_global_type
  use mus_variable_module,      only: mus_derVarPos_type
  use mus_param_module,         only: mus_param_type
  use mus_physics_module,       only: mus_physics_type
  use mus_mixture_module,       only: mus_mixture_type

  implicit none

  private

  public :: mus_bc_corner_type
  public :: boundary_type
  public :: glob_boundary_type
  public :: mus_load_bc
  public :: mus_setBC_stencilPos

  !> information needed for moments based boundary condition
  type bc_moments_type
    integer :: nUnKnownPdfs
    integer, allocatable :: knownMom_pos(:)
    real(kind=rk), allocatable :: unKnownPdfs_MatInv(:,:)
  end type bc_moments_type

  !> Level wise boundary elements information
  type bc_elems_type
    !> list of Tree-ID positions, size: nElems
    integer, allocatable :: elem(:)
    !> Position of this boundary in the bc_elemBuffer
    integer, allocatable :: elemBuffer(:)
    !> Normal vector for each boundary element pointing into the domain.
    !! size: 3, nElems
    integer, allocatable :: normal(:,:)
    !> which is the index in the stencil corresponding to the normal vector?
    integer, allocatable :: normalInd(:)
    !> bit mask for each node holding directions which have to be updated
    logical, allocatable :: bitmask(:,:)
    !> The q-Values for the exact wall positions
    real(kind=rk), allocatable :: qVal(:,:)
  end type bc_elems_type

  !> contains information about corner boundary nodes
  type bcCornerElems_type
    !> list of Tree-ID positions, size: nElems
    integer, allocatable :: elem(:)
    !> bit mask for each corner node holding directions which have to be updated
    logical, allocatable :: bitmask(:,:)
    !> Position of this boundary in the bc_elemBuffer
    integer, allocatable :: elemBuffer(:)
    !> Normal vector for each boundary element pointing into the domain.
    !! size: 3, nElems
    integer, allocatable :: normal(:,:)
    !> which is the index in the stencil corresponding to the normal vector?
    integer, allocatable :: normalInd(:)
  end type bcCornerElems_type

  !> contains boundary ids of corner node
  type bcCornerID_type
    !> list of boundary IDs
    integer, allocatable :: bcID(:)
  end type bcCornerID_type

  !> contains information needed to treat corner nodes or nodes intersected
  !! by more than one boundary type
  type mus_bc_corner_type
    !> number of local boundary elements per level
    integer :: nElems( globalMaxLevels )
    !> boundary neighbor in different refinement level
    type( bcCornerElems_type ) :: elems( globalMaxLevels )
    !> boundary id in each direction for each corner node
    !! size layout%stencil%QQ, nElems
    integer, allocatable :: bcid(:,:)
    type( bcCornerID_type ), allocatable :: cornerIDs(:)
  end type mus_bc_corner_type

  !> variable definition for non-reflective type of boundary conditions\n
  !! These boundary condition is taken from the paper:
  !! \verbatim
  !!   S. Izquierdo and N. Fueyo,
  !!   "Characteristic nonreﬂecting boundary
  !!   conditions for open boundaries in lattice Boltzmann methods,"
  !!   Physical Review E, vol. 78, no. 46707, 2008.
  !! \endverbatim
  type bc_outlet_nrbc_type

    !> specific heat ratio
    !! appears in eq 3: cs = sqrt( kappa * R * T )
    real(kind=rk) :: kappa

    !> from eq 17: k = sigma * ( 1 - Ma^2 ) * cs / L
    real(kind=rk) :: K_mod

    !> used in eq 17
    real(kind=rk) :: cs_mod

    !> used in eq 17
    real(kind=rk) :: sigma

    !> length between inlet and outlet, represented by L in the paper
    real(kind=rk) :: lodi_length

  end type bc_outlet_nrbc_type

  !> Contains variables for black-box membrane boundary condition
  type bc_blackbox_membrane_type
    !> type of membrance (AEM or CEM)
    character(len=labelLen) :: label
    !> transference number defining transfer property
    !! of ionic species on the membrance
    !! KM \todo check transNr constrain
    !!          sumOfIonicSpecies(transNr) = 1
    real(kind=rk) :: transNr
    !> osmotic solvent transport coefficient
    real(kind=rk) :: solvTrans_coeff
    !> direction of flux
    integer :: flux_dir(3)
  end type bc_blackbox_membrane_type

  !> Contains slip factor for slip wall boundary
  type bc_slip_wall_type
    real(kind=rk) :: fac !< slip factor
  end type bc_slip_wall_type

  !> contains all possible bc state variables needed for boundary
  type bc_states_type
    type(tem_bc_state_type) :: velocity(3) !< velocity at boundary node
    !> mass flow rate at boundary node for inlet_mfr (inlet mass flow rate)
    !! boundary condition
    type(tem_bc_state_type) :: massFlowRate 
    type(tem_bc_state_type) :: pressure !< pressure at boundary node
    !> mole fraction for species boundary
    type(tem_bc_state_type) :: moleFrac
    !> molar flux for species (n_i*u_i)
    type(tem_bc_state_type) :: moleFlux(3)
    !> molar diffusion flux for species
    type(tem_bc_state_type) :: moleDiff_Flux(3)
  end type bc_states_type

  !> Information about boundary elements
  !!
  !! This derived type encapsulates the definition of a boundary
  !! with all necessary parameters and data fields
  type glob_boundary_type
    !> name of this boundary
    character(len=LabelLen) :: label
    !> elements list of this boundary in different refinement level
    type( bc_elems_type ) :: elems( globalMaxLevels )
    !> list corner boundary elements
    type( mus_bc_corner_type ) :: cornerBC
    !> number of local boundary elements per level
    integer :: nElems( globalMaxLevels )
    !> number of total boundary elements per level
    integer :: nElems_totalLevel( globalMaxLevels )
    !> number of local boundary elements
    integer :: nElems_local = 0
    !> number of total boundary elements
    integer :: nElems_total = 0
    !> whether this boundary requires qVal
    logical :: hasQVal = .false.
    !> has qVal initialized with default qVal = 0.5
    logical :: qValInitialized = .false.
    !> all bc states types of all fields of this boundary
    type( bc_states_type ), allocatable :: bc_states(:)
  end type glob_boundary_type

  !> Level wise information about neighbor of boundary elements
  type bc_neigh_type
    !> Possible neighbor locations in the local tree-ID vector,
    !! size: nNeighs, nElems
    integer, allocatable :: neigh(:,:)
    !> Pre-collision state values of this boundary neighbor,
    !! size: nNeighs, nElems*stencil%QQ
    real(kind=rk), allocatable :: neighBufferPre(:,:)
    !> Post-coliision state values of this boundary neighbor,
    !! size: nNeighs, nElems*stencil%QQ
    real(kind=rk), allocatable :: neighBufferPost(:,:)
    !> LODI array for the NRBC
    real(kind=rk), allocatable :: lodi(:,:,:)
    !> defined moments position in moments matrix and
    !! inverse of unknown pdf matrix
    type( bc_moments_type ), allocatable :: moments(:)
  end type bc_neigh_type

  !> Boundary treatment data type.
  !! For the same boundary elements, each field may have a different treatment.
  !! It includes function pointers and neighbor elements required by this
  !! boundary treatment
  type boundary_type
    character(len=LabelLen) :: BC_kind !< kind of this boundary
    character(len=LabelLen) :: label  !< name of this boundary
    !> number of neighbors fluid element needed for extrapolation to
    !! get data at boundary element
    integer :: nNeighs =  0
    !> Level wise neighbor information of boundary elements
    !! Each field may have a different handlement on the same boundary
    !! Thus requires a different neighborhood
    type( bc_neigh_type ) :: elems( globalMaxLevels )
    !> function pointer to boundary routine to call
    procedure( boundaryRoutine ),pointer :: fnct => null()
    !> musubi bc state variables
    type( bc_states_type ) :: BC_states
    !> order of boundary which determines number of neighbor need for this boundary
    integer :: order
    !> standard q-value for higher order boundaries without any definition
    !! in seeder
    real(kind=rk) :: qVal
    !> Non-reflective boundary type
    type( bc_outlet_nrbc_type ) :: nrbc
    !> slip wall boundary
    type( bc_slip_wall_type ) :: slip

    !> black box membrane boundary
    type( bc_blackbox_membrane_type ) :: blackbox_mem
    !> position of the corresponding stencil in the stencil array of the schemes
    integer :: stencilPos
  end type boundary_type

  !> Interface definition for the boundary condition subroutines
  abstract interface
    ! This interface definition must match all the boundary routine interfaces.
    ! these are all defined via the macro bc_routineHeader( bc_name )
    subroutine boundaryRoutine( me, state, currT, lastT, bcBuffer, globBC,     &
      & glob, tree, iLevel, sim_time, neigh, layout, fieldProp, varPos, nScalars,&
      & varSys, derVarPos, params, physics, iField, mixture, IC_fields,        &
      & derVarPosAll )
      import :: boundary_type, rk, mus_field_prop_type, mus_scheme_layout_type,&
        & tem_time_type, pdf_global_type, treelmesh_type, mus_derVarPos_type,  &
        & tem_var_system_type, glob_boundary_type, mus_param_type,             &
        & mus_physics_type, mus_mixture_type, tem_ini_condition_type
      !> field boundary type
      class( boundary_type ) :: me
      real(kind=rk), intent(inout) :: state(:,:)  !< Current state vector
      !> current time step
      integer,intent(in) :: currT
      !> last time step
      integer,intent(in) :: lastT
      !> state values of boundary elements of all fields of iLevel
      real(kind=rk),intent(in) :: bcBuffer(:)
      !> gloabl type contains iLevel descriptor
      type( pdf_global_type ), intent(in) :: glob
      !> global treelm mesh 
      type( treelmesh_type ), intent(in) ::tree
      !> level which invokes boundary
      integer,intent(in) :: iLevel
      !> global time information
      type( tem_time_type ), intent(in)  :: sim_time
      !> global parameters
      type(mus_param_type),intent(in) :: params   
      integer,intent(in)  :: neigh(:)  !< connectivity array
      !> scheme layout
      type( mus_scheme_layout_type ),intent(in) :: layout
      !> Fluid property
      type( mus_field_prop_type ), intent(in)  :: fieldProp
      !> pointer to field variable in the state vector
      integer, intent(in) :: varPos(:)
      !> number of Scalars in the scheme var system
      integer, intent(in) :: nScalars
      !> scheme variable system
      type( tem_var_system_type ), intent(in) :: varSys
      !> position of derived quantities in varsys
      type( mus_derVarPos_type ), intent(in) :: derVarPos
      !> position of derived quantities in varsys
      type( mus_derVarPos_type ), intent(in) :: derVarPosAll(:)
      !> scheme global boundary type
      type( glob_boundary_type ), intent(in) :: globBC
      !> scheme global boundary type
      type( mus_physics_type ), intent(in) :: physics
      !> current field
      integer, intent(in) :: iField
      !> mixture info
      type(mus_mixture_type), intent(in) :: mixture
      !> InitialConditions of all fields
      type(tem_ini_condition_type), intent(in) :: IC_fields(:)
    end subroutine boundaryRoutine
  end interface

contains

!******************************************************************************!
  !> Read in the boundary conditions from the LUA parameter file.\n
  !!
  subroutine mus_load_bc( me, boundary, conf, path, parent )
    !---------------------------------------------------------------------------
    !> field boundary type 
    type( boundary_type ), allocatable   :: me(:)
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in) :: boundary
    !> lua flu state
    type( flu_state )               :: conf
    !> lua path
    type(aot_path_type), intent(in) :: path
    !> bc parent handle
    integer, intent(in), optional   :: parent
    !---------------------------------------------------------------------------
    integer :: bc_handle
    integer :: sub_handle
    integer :: iError
    integer :: vError(3), errFatal(3)
    type(aot_path_type) :: bc_path ! boundary_condition path
    type(aot_path_type) :: bc_sub_path ! boundary sub table path
    ! loop variable
    integer :: iBC
    ! boundary header type
    type( tem_bc_header_type ) :: bc_header
    !---------------------------------------------------------------------------
    errFatal = aoterr_Fatal

    ! if fluid informations in scheme table parentHandle /= 0
    ! load boundary_condition table and match label specified in config lua with 
    ! boundary label loaded from mesh (information in tem_bc_prop_type)
    if( present(parent) ) then 
      call tem_load_bc_header( me           = bc_header,                       &
        &                      conf         = conf,                            &
        &                      parentHandle = parent,                          &
        &                      BC_prop      = boundary )
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent,                                   &
        &                  thandle = bc_handle,                                &
        &                  key     = 'boundary_condition' )
    else
      call tem_load_bc_header( me      = bc_header,                            &
        &                      conf    = conf,                                 &
        &                      BC_prop = boundary )
      call aot_table_open( L       = conf,                                     &
        &                  thandle = bc_handle,                                &
        &                  key     = 'boundary_condition' )
    end if

    call tem_horizontalSpacer(fUnit = logUnit(1))
    write(logUnit(1),*) ' Loading boundary informations for nBCs',             &
      &                 bc_header%nBCs 

    ! set bc_path to boundary_condition table in lua file
    bc_path = path
    call aot_path_addNode( me       = bc_path,                                 &
      &                    NodeType ='table',                                  &
      &                    key = 'boundary_condition' )

    ! load tables inside the boundary table for bc kind and their state
    ! variables boundaries defined in lua are match with bc label read from
    ! mesh using bc_header
    allocate( me( bc_header%nBCs ))

    do iBC = 1, bc_header%nBCs
        call aot_table_open( L       = conf,                                   &
          &                  parent  = bc_handle,                              &
          &                  thandle = sub_handle,                             &
          &                  pos     = iBC )
        if ( sub_handle /= 0 ) then
          ! set bc_sub_path to boundary_condition sub_table in lua file
          bc_sub_path = bc_path
          call aot_path_addNode( me       = bc_sub_path,                       &
            &                    NodeType = 'table',                           &
            &                    pos      = iBC )

          ! matching the boundary label from mesh to boundary label read for lua 
          ! file bc_header%BC_ID(iBC) returns the position of boundary label iBC 
          ! in the mesh. 
          me( bc_header%BC_ID(iBC) )%label   = bc_header%label(iBC)
          me( bc_header%BC_ID(iBC) )%BC_kind = bc_header%BC_kind(iBC)

          write(logUnit(1),*) '   Boundary',iBC
          select case( trim( bc_header%BC_kind(iBC) ))
            case( 'wall' )
              write(logUnit(1),*) '    Simple bounce back treatment for '//    &
                &  'no-slip wall '// trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'spc_bb_wall' )
              write(logUnit(1),*) '    Species bounce back treatment for '//    &
                &  'no-slip wall '// trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 1

            case( 'moments_wall', 'spc_moments_wall' )
              write(logUnit(1),*) '    Moments wall treatment for '//    &
                &  'no-slip wall '// trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'spc_moments_outflow' )
              write(logUnit(1),*) '    Spc moments outflow treatment for '//    &
                &  trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 1

            case( 'wall_multiReflection','wall_mr')
              bc_header%BC_kind(iBC) = 'wall_multiReflection'
              me( bc_header%BC_ID(iBC) )%BC_kind = bc_header%BC_kind(iBC)
              write(logUnit(1),*) '    Wall multi reflection for no-slip wall '&
                &             //trim(me( bc_header%BC_ID(iBC) )%label)
              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )%qVal,     &
                &               ErrCode = iError,                              &
                &               key     = 'qVal',                              &
                &               default = 0.5_rk )
              me( bc_header%BC_ID(iBC) )%nNeighs = 2

            case( 'wall_linearInterpolation',                                  &
              &   'wall_higher_order',                                         &
              &   'wall_linearInterp' )
              bc_header%BC_kind(iBC) = 'wall_linearInterpolation'
              me( bc_header%BC_ID(iBC) )%BC_kind = bc_header%BC_kind(iBC)
              write(logUnit(1),*) '    Wall higher order with linear'   //     &
                &             ' interpolation treatment for no-slip wall '//   &
                &            trim(me( bc_header%BC_ID(iBC) )%label)
              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )%qVal,     &
                &               ErrCode = iError,                              &
                &               key     = 'qVal',                              &
                &               default = 0.5_rk )
              ! No normal neighbor is required
              ! Interpolation neighbors are directly identified in the boundary
              ! routine using element index macro
              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'slip_wall' )
              write(logUnit(1),*) '   '//trim( bc_header%BC_kind(iBC) ) //     &
                &    ' treatment for '//trim(me( bc_header%BC_ID(iBC) )%label)
              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )%slip%fac, &
                &               ErrCode = iError,                              &
                &               key     = 'fac',                               &
                &               default = 1.0_rk )
              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'inlet_ubb', 'inlet_eq', 'spc_inlet_eq', 'spc_outlet_vel',   &
              &   'spc_vel_bb', 'moments_vel', 'moments_inflow',            &
              &   'spc_moments_vel', 'spc_bb_vel_test')
              write(logUnit(1),*) '   '//trim( bc_header%BC_kind(iBC) ) //     &
                &    ' treatment for '//trim(me( bc_header%BC_ID(iBC) )%label)

              call tem_load_bc_state(                                          &
                & bc         = me(bc_header%BC_ID(iBC))%BC_states%velocity(1), &
                & state_name = 'velocityX',                                    &
                & conf       = conf,                                           &
                & bc_handle  = sub_handle,                                     &
                & path       = bc_sub_path )

              call tem_load_bc_state(                                          &
                & bc         = me(bc_header%BC_ID(iBC))%BC_states%velocity(2), &
                & state_name = 'velocityY',                                    &
                & conf       = conf,                                           &
                & bc_handle  = sub_handle,                                     &
                & path       = bc_sub_path )

              call tem_load_bc_state(                                          &
                & bc         = me(bc_header%BC_ID(iBC))%BC_states%velocity(3), &
                & state_name = 'velocityZ',                                    &
                & conf       = conf,                                           &
                & bc_handle  = sub_handle,                                     &
                & path       = bc_sub_path )
              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )%order,    &
                &               ErrCode = iError,                              &
                &               key     = 'order',                             &
                &               default = 1 )

              me(bc_header%BC_ID(iBC))%nNeighs = me(bc_header%BC_ID(iBC))%order

              ! load molefraction in addition for spc_inlet_eq
              select case( trim( bc_header%BC_kind(iBC) ))
                case( 'spc_inlet_eq' )
                  call tem_load_bc_state(                                      &
                    &      bc = me( bc_header%BC_ID(iBC) )%BC_states%moleFrac, &
                    &      state_name = 'molefraction', conf = conf,           &
                    &      bc_handle = sub_handle, path = bc_sub_path)
                case( 'moments_inflow' )
                  call tem_load_bc_state(                                      &
                    &    bc         = me( bc_header%BC_ID(iBC) )%              &
                    &                                      BC_states%pressure, &
                    &    state_name = 'pressure',                              &
                    &    conf       = conf,                                    &
                    &    bc_handle  = sub_handle,                              &
                    &    path       = bc_sub_path )
              end select

            ! inlet mass flow rate boundary condition
            case( 'inlet_mfr', 'inlet_mfr_eq' )
              write(logUnit(1),*) '   '//trim( bc_header%BC_kind(iBC) ) //     &
                &    ' treatment for '//trim(me( bc_header%BC_ID(iBC) )%label)

              call tem_load_bc_state(                                          &
                & bc         = me( bc_header%BC_ID(iBC) )%                     &
                &                                       BC_states%massFlowRate,&
                & state_name = 'massflowrate',                                 &
                & conf       = conf,                                           &
                & bc_handle  = sub_handle,                                     &
                & path       = bc_sub_path )

            case( 'outlet_zfg',                                                &
              &   'outlet_extrapolation',                                      &
              &   'outlet_extrapolation_microscopic' )
              write(logUnit(1),*) '    Outlet extrapolation for kind '//       &
                &    trim( me( bc_header%BC_ID(iBC) )%BC_kind )//' with label '&
                &      //trim(me( bc_header%BC_ID(iBC) )%label )

              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )%order,    &
                &               ErrCode = iError,                              &
                &               key     = 'order',                             &
                &               default = 1 )
              me( bc_header%BC_ID(iBC) )%nNeighs                               &
                & = me( bc_header%BC_ID(iBC) )%order

            case( 'outlet_eq', 'outlet_pab', 'outlet_bouzidi','spc_outlet_eq', &
              &   'outlet_expol', 'moments_press')
              select case( trim( bc_header%BC_kind(iBC) ))
              case( 'outlet_eq' )
                write(logUnit(1),*) '    Outlet pressure equilibrium type for '&
                  &             //trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 2

              case( 'outlet_pab' )

                write(logUnit(1),*) '    Outlet pressure anti bounce back for '&
                  &            //trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 1

              case( 'outlet_bouzidi' )
                write(logUnit(1),*) '    Outlet bouzidi boundary condition for'&
                  &             //' '//trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 1

              case( 'spc_outlet_eq' )
                write(logUnit(1),*) '    Species outlet pressure '//           &
                  &                 'equilibrium type for '//                  &
                  &                  trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 2
              case( 'outlet_expol' )
                write(logUnit(1),*) '    Outlet extrapolation type for '//     &
                  &                      trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 2
              case( 'moments_press' )
                write(logUnit(1),*) '    Moments based pressure boundary '//   &
                  &                 'type for '//                              &
                  &                 trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 0
              end select

              call tem_load_bc_state(                                          &
                &    bc         = me( bc_header%BC_ID(iBC) )%                  &
                &                                          BC_states%pressure, &
                &    state_name = 'pressure',                                  &
                &    conf       = conf,                                        &
                &    bc_handle  = sub_handle,                                  &
                &    path       = bc_sub_path )

              if( .not.                                                        &
                & me( bc_header%BC_ID(iBC) )%BC_states%pressure%isDefined ) then
                write(logUnit(1),*)' Error: No pressure was set for boundary ' &
                  &    //'condition ' //trim(me( bc_header%BC_ID(iBC) )%label)
                call tem_abort()
              endif

            case( 'outlet_nrbc', 'outlet_nrbc_eq' )
              select case( trim( bc_header%BC_kind(iBC) ))
              case( 'outlet_nrbc' )
                write(logUnit(1),*) '    Characteristic based non-reflective ' &
                  &           //'boundary condition '                     //   &
                  &             trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 2

              case( 'outlet_nrbc_eq' )
                write(logUnit(1),*) '    Eq-type Characteristic based '   //   &
                  &             'non-reflective boundary condition '      //   &
                  &             trim(me( bc_header%BC_ID(iBC) )%label)
                me( bc_header%BC_ID(iBC) )%nNeighs = 2
              end select 

              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )%          &
                &                                                  nrbc%kappa, &
                &               ErrCode = iError,                              &
                &               key     = 'kappa',                             &
                &               default = 1.0_rk )
              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )%          &
                &                                                  nrbc%sigma, &
                &               ErrCode = iError,                              &
                &               key     = 'sigma',                             &
                &               default = 0.0001_rk )
              call aot_get_val(                                                &
                &         L       = conf,                                      &
                &         thandle = sub_handle,                                &
                &         val     = me( bc_header%BC_ID(iBC) )%                &
                &                                            nrbc%lodi_length, &
                &         ErrCode = iError,                                    &
                &         key     = 'L',                                       &
                &         default = 100._rk )
              write(logUnit(1),*)' kappa:    ',                                &
                &                         me( bc_header%BC_ID(iBC) )%nrbc%kappa 
              write(logUnit(1),*)' sigma:    ',                                &
                &                         me( bc_header%BC_ID(iBC) )%nrbc%sigma 
              write(logUnit(1),*)' lodi L:   ',                                &
                &                   me( bc_header%BC_ID(iBC) )%nrbc%lodi_length 

              call tem_load_bc_state(                                          &
                &    bc         = me( bc_header%BC_ID(iBC) )%                  &
                &                                          BC_states%pressure, &
                &    state_name = 'pressure',                                  &
                &    conf       = conf,                                        &
                &    bc_handle  = sub_handle,                                  &
                &    path = bc_sub_path )

            case( 'outlet_zero_prsgrd', 'spc_outlet_zero_prsgrd' )
              write(logUnit(1),*) '    Outlet zero pressure gradient for '  // &
                &                      trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'outlet_dnt' )
              write(logUnit(1),*) '    Outlet do-nothing type for '         // &
                &                      trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 2

            case( 'moments_outflow' )
              write(logUnit(1),*) '    Outlet moments type for '         // &
                &                      trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 2

            case( 'spc_outlet_expol' )
              write(logUnit(1),*) '    Species outlet extrapolation type for ' &
                &                 // trim(me( bc_header%BC_ID(iBC) )%label)
              me( bc_header%BC_ID(iBC) )%nNeighs = 2
              call tem_load_bc_state(                                          &
                &           bc = me( bc_header%BC_ID(iBC) )%BC_states%pressure,&
                &           state_name = 'pressure', conf = conf,              &
                &           bc_handle = sub_handle, path = bc_sub_path)

            case( 'spc_molefrac', 'spc_molefrac_wtdf', 'spc_moments_molefrac' )
              write(logUnit(1),*) '    Species '//trim(bc_header%BC_kind(iBC)) &
                & //'  treatment for '//trim(me( bc_header%BC_ID(iBC) )%label) 
              call tem_load_bc_state(                                          &
                &           bc = me( bc_header%BC_ID(iBC) )%BC_states%moleFrac,&
                &           state_name = 'molefraction', conf = conf,          &
                &           bc_handle = sub_handle, path = bc_sub_path)
              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'spc_moleflux', 'spc_moleflux_eq', 'spc_moments_moleflux' )
              write(logUnit(1),*) '    Species molar flux treatment for '//    &
                &        trim(me( bc_header%BC_ID(iBC) )%label)
              call tem_load_bc_state(                                          &
                &          bc = me( bc_header%BC_ID(iBC) )%                    &
                &                                        BC_states%moleFlux(1),&
                &          state_name = 'molefluxX', conf = conf,              &
                &          bc_handle = sub_handle, path = bc_sub_path)

              call tem_load_bc_state(                                          &
                &          bc = me( bc_header%BC_ID(iBC) )%                    &
                &                                        BC_states%moleFlux(2),&
                &          state_name = 'molefluxY', conf = conf,              &
                &          bc_handle = sub_handle, path = bc_sub_path)

              call tem_load_bc_state(                                          &
                &          bc = me( bc_header%BC_ID(iBC) )%                    &
                &                                        BC_states%moleFlux(3),&
                &          state_name = 'molefluxZ', conf = conf,              &
                &          bc_handle = sub_handle, path = bc_sub_path) 

              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'spc_blackbox_mem_ion', 'spc_blackbox_mem_solvent',          &
              &   'spc_moments_blackbox_mem_ion' )
              write(logUnit(1),*) '    Species black box membrance '//         &
                & 'treatment for '//trim(me( bc_header%BC_ID(iBC) )%label) 

              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )           &
                &                         %blackbox_mem%transNr,               &
                &               ErrCode = iError,                              &
                &               key     = 'transference_number' )
              if( btest( iError, aoterr_Fatal ))then
                write(logUnit(1),*)'FATAL Error occured in definition of '//   &
                  &            'black box membrance boundary condition'
                write(logUnit(1),*)'while retrieving transference number'
                write(logUnit(1),*)'STOPPING'
                call tem_abort()
              end if

              call aot_get_val( L       = conf,                                &
                &               thandle = sub_handle,                          &
                &               val     = me( bc_header%BC_ID(iBC) )           &
                &                         %blackbox_mem%flux_dir,              &
                &               ErrCode = vError,                              &
                &               key     = 'flux_dir' )
              if( any(btest( vError, errFatal)) )then
                write(logUnit(1),*)'FATAL Error occured in definition of '//   &
                  &            'black box membrance boundary condition'
                write(logUnit(1),*)'while retrieving flux dir with 3 entries'
                write(logUnit(1),*)'STOPPING'
                call tem_abort()
              end if

              select case( trim( bc_header%BC_kind(iBC) ))
                case( 'spc_blackbox_mem_solvent' )
                  call aot_get_val( L       = conf,                            &
                    &               thandle = sub_handle,                      &
                    &               val     = me( bc_header%BC_ID(iBC) )       &
                    &                         %blackbox_mem%solvTrans_coeff,   &
                    &               ErrCode = iError,                          &
                    &               key     = 'transference_number' )
                  if( btest( iError, aoterr_Fatal ))then
                    write(logUnit(1),*)'FATAL Error occured in definition of ' &
                      &            //'black box '//                            &
                      &            'membrance solvent boundary condition'
                    write(logUnit(1),*)'while retrieving solv_trans_coeff'
                    write(logUnit(1),*)'STOPPING'
                    call tem_abort()
                  end if
              end select
            case( 'spc_molediff_flux' )
              write(logUnit(1),*) '    Species molar diffusion flux '//        &
                & 'treatment for '//trim(me( bc_header%BC_ID(iBC) )%label) 
              call tem_load_bc_state(                                          &
                &           bc = me( bc_header%BC_ID(iBC) )%BC_states%moleFrac,&
                &           state_name = 'molefraction', conf = conf,          &
                &           bc_handle = sub_handle, path = bc_sub_path)
              call tem_load_bc_state(                                          &
                &     bc = me( bc_header%BC_ID(iBC) )%BC_states%               &
                &                                             moleDiff_Flux(1),&
                &     state_name = 'molediff_fluxX', conf = conf,              &
                &     bc_handle = sub_handle, path = bc_sub_path)

              call tem_load_bc_state(                                          &
                &     bc = me( bc_header%BC_ID(iBC) )%BC_states%               &
                &                                             moleDiff_Flux(2),&
                &     state_name = 'molediff_fluxY', conf = conf,              &
                &     bc_handle = sub_handle, path = bc_sub_path)

              call tem_load_bc_state(                                          &
                &     bc = me( bc_header%BC_ID(iBC) )%BC_states%               &
                &                                             moleDiff_Flux(3),&
                &     state_name = 'molediff_fluxZ', conf = conf,              &
                &     bc_handle = sub_handle, path = bc_sub_path) 

              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'flekkoy_outlet' )
              write(logUnit(1),*) '   Outlet BCs for the flekkoy '//           &
                &                 'diffusion model'
              me( bc_header%BC_ID(iBC) )%nNeighs = 2

            case( 'flekkoy_inlet' )
              write(logUnit(1),*) '   Inlet BCs for the flekkoy '//            & 
                &                 'diffusion model'
              me( bc_header%BC_ID(iBC) )%nNeighs = 0

            case( 'test' )
              write(logUnit(1),*) '   WARNING!! ' 
              write(logUnit(1),*) '   Test boundary conditions for testing ' //&
                &             'neighbor info exchange' 
              me( bc_header%BC_ID(iBC) )%nNeighs = 20

            case default
              write(logUnit(1),*) 'Unknown Boundary condition kind!'
              write(logUnit(1),*) 'Stopping ...' 
              call tem_abort()
          end select
        write(logUnit(1),*)'      number of neighbors ',                     &
          &                             me( bc_header%BC_ID( iBC ))%nNeighs 
      end if  ! sub_handle/=0
      call aot_table_close( L=conf, thandle=sub_handle )
    end do ! iBC
    call tem_horizontalSpacer(fUnit = logUnit(1))

  end subroutine mus_load_bc
!******************************************************************************!


!******************************************************************************!
  !> Set the correct position of the stencils in the stencil array in the scheme
  !!
  subroutine mus_setBC_stencilPos( me, iField, nBCs )
    !---------------------------------------------------------------------------
    !> global boundary type 
    type( boundary_type ) :: me(nBCs)
    !> current field
    integer :: iField
    !> number of boundary conditions
    integer :: nBCs
    !---------------------------------------------------------------------------
    integer :: iBC
    !---------------------------------------------------------------------------

    do iBC = 1, nBCs
      me( iBC )%stencilPos = ( iField - 1 ) * nBCs + iBC + 1
    end do

  end subroutine mus_setBC_stencilPos
!******************************************************************************!


end module mus_bc_header_module
!******************************************************************************!
