?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> This module provides the definition and methods for MSLiquid
!! bgk advection relaxation scheme.
!! \author Kannan Masilamani
module mus_MSLiquid_module
  use env_module,               only: rk, long_k, eps
  use tem_var_system_module,    only: tem_var_system_type
  use tem_param_module,         only: div1_3, div1_36, div1_8, div3_4h,        &
    &                                 div1_4, div3_8, div3_2, div9_16, div3_16,&
    &                                 cs2inv, div1_6, cs4inv,                  &
    &                                 t2cs2inv, t2cs4inv, div1_18
  use treelmesh_module,         only: treelmesh_type
  use tem_math_module,          only: invert_matrix
  use tem_isNaN_module,         only: tem_isNaN
  use tem_debug_module,         only: dbgUnit
  use tem_spacetime_fun_module, only: tem_spacetime_for

  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_param_module,         only: mus_param_type
  use mus_mixture_module,       only: mus_mixture_type
  use mus_eNRTL_module,         only: mus_calc_thermFactor,                    &
    &                                 mus_calc_MS_DiffMatrix

  implicit none

  private

  public :: mrt_advRel_MSLiquid
  public :: mrt_advRel_d3q19f3_MSLiquid
  public :: bgk_advRel_MSLiquid
  public :: bgk_forcing_advRel_MSLiquid
  public :: bgk_advRel_d3q19f3_MSLiquid
  public :: externalForce_MSLiquid
  public :: externalForce_MSLiquid_WTF
  public :: bgk_advRel_MSLiquid_WTDF
  public :: bgk_advRel_MSLiquid_WTDF_2Spc

  !=======================================================================
  ! D3Q19 flow model
  !=======================================================================
  !> Definition of the discrete velocity set

  integer, parameter :: QQ  = 19   !> number of pdf directions
  integer, parameter :: QQN = 18   !> number of neighbors

  integer,parameter :: q__W     = 1     !< west             x-
  integer,parameter :: q__S     = 2     !< south            y-
  integer,parameter :: q__B     = 3     !< bottom           z-
  integer,parameter :: q__E     = 4     !< east             x+
  integer,parameter :: q__N     = 5     !< north            y+
  integer,parameter :: q__T     = 6     !< top              z+
  integer,parameter :: q_BS     = 7     !<                  z-,y-
  integer,parameter :: q_TS     = 8     !<                  z+,y-
  integer,parameter :: q_BN     = 9     !<                  z-,y+
  integer,parameter :: q_TN     = 10    !<                  z+,y+
  integer,parameter :: q_BW     = 11    !<                  x-,z-
  integer,parameter :: q_BE     = 12    !<                  x+,z-
  integer,parameter :: q_TW     = 13    !<                  x-,z+
  integer,parameter :: q_TE     = 14    !<                  x+,z+
  integer,parameter :: q_SW     = 15    !<                  y-,x-
  integer,parameter :: q_NW     = 16    !<                  y+,x-
  integer,parameter :: q_SE     = 17    !<                  y-,x+
  integer,parameter :: q_NE     = 18    !<                  y+,x+
  integer,parameter :: q__0     = 19    !< rest density is last

  integer, parameter :: QQF3  = 57   !> number of pdf directions for 3 species

  integer,dimension(QQ,3), parameter :: varPos = &
   & reshape(& 
   &  [q__w, q__s, q__b, q__e, q__n, q__t, q_bs, q_ts, q_bn, &
   &   q_tn, q_bw, q_be, q_tw, q_te, q_sw, q_nw, q_se, q_ne, q__0, & 
   &   q__w+19, q__s+19, q__b+19, q__e+19, q__n+19, q__t+19, &
   &   q_bs+19, q_ts+19, q_bn+19, q_tn+19, q_bw+19, q_be+19, &
   &   q_tw+19, q_te+19, q_sw+19, q_nw+19, q_se+19, q_ne+19, q__0+19,& 
   &   q__w+38, q__s+38, q__b+38, q__e+38, q__n+38, q__t+38, &
   &   q_bs+38, q_ts+38, q_bn+38, q_tn+38, q_bw+38, q_be+38, &
   &   q_tw+38, q_te+38, q_sw+38, q_nw+38, q_se+38, q_ne+38, q__0+38] &
   &  ,[QQ,3])
contains

!******************************************************************************
  !> Optimized Advection relaxation routine for the MSLiquid BGK model 
  !! for d3q19 layout with three species.
  !!
  !! This routine contains the implementation of semi-implicit lattice boltzmann
  !! equation using variable transformation based on the paper
  !! "Multi-species Lattice Boltzmann Model and Practical Examples. Short Course 
  !! material Pietro Asinari PhD." \n
  !! Refer page: \ref multispecies "Multispecies" for more information
  !! In the variable tranformation steps, we can skip the step 1 and step 3
  !! and evaluate only step 2 based on tranformed variable g
  !! only prerequisite is to compute feq which depends on original f not on g.
  !! feq is depend on density and velocity. Where density can be computed 
  !! directly from g and velocity computed from linear system of equation 
  !! given in the reference page \ref multispecies "Multispecies.
  !! KM: This is an non-optimized kernel
?? copy :: compute_routineHeader( bgk_advRel_d3q19f3_MSLiquid )
    !------------------------------------------------------------------------ 
    !local variables
    ! temporary local pdf values
    real(kind=rk) :: pdfTmp( layout%stencil(1)%QQ, 3 ) 
    integer :: iElem
    real(kind=rk), dimension(3) :: rsigma, uxsigma, rsigmaInv,                 &
      & uysigma, uzsigma, qxsigma, qysigma, qzsigma, gqxsigma, gqysigma,       &
      & gqzsigma
    real(kind=rk), dimension(3) :: uxstar, uystar, uzstar
    real(kind=rk), dimension(3) :: molWeight, molWeight_inv, phi,              &
      &                            num_dens, moleFrac
    real(kind=rk) :: usqr(3), totNum_dens_inv, totMass_densInv
    real(kind=rk) :: velAvg(3)
    real(kind=rk) :: velQuadTerm_x(3), velQuadTerm_y(3), velQuadTerm_z(3)
    real(kind=rk) :: omega, omegadiv2, omega_fac, omega_o
    real(kind=rk), dimension(3,3) :: A, mbb
    real(kind=rk) :: zfac, yfac, a12_11, a13_11, &
      & a23fac, a21_11, a31_11, zcoeff
    real(kind=rk) :: theta_eq, theta_eq_spc, paramB_inv, a11_inv, a22fac_inv
    real(kind=rk) :: B_12, B_13, B_23
    real(kind=rk) :: Bratio_12, Bratio_13, Bratio_23
    real(kind=rk) :: mbbEq_12, mbbEq_13, mbbEq_21, mbbEq_23, mbbEq_31, mbbEq_32
    real(kind=rk) :: chi12, chi13, chi21, chi23, chi31, chi32
    real(kind=rk),dimension(3) :: omegaRho, omegaRho_d, omegaRho_dm, &
      & omegaRho_o, omegaRho_om, sum1_1, sum2_1, sum3_1,                       &
      & sum4_1, sum5_1, sum6_1, sum7_1, sum8_1, sum9_1, sum1_2, sum2_2, sum3_2,&
      & sum4_2, sum5_2, sum6_2, sum7_2, sum8_2, sum9_2
    !--------------------------------------------------------------------------- 
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    molWeight_inv = 1.0_rk/molWeight
    !molecular weight ratios
    phi = fieldProp(:)%species%molWeigRatio

    !omega
    omega = scheme(iScheme)%mixture%relaxLvl(Level)%omega_diff
    omega_fac = (1.0_rk/(1.0_rk/omega + 0.5_rk))
    omegadiv2 = omega*0.5_rk
    omega_o = 1.0_rk - omega_fac

    !free parameter B
    paramB_inv = 1.0_rk/scheme(iScheme)%mixture%paramB

    !equilibrium theta
    theta_eq = scheme(iScheme)%mixture%theta_eq
    theta_eq_spc = 1.0_rk - theta_eq

    !resistivities
    !resistivity coeff are symmetric
    !B_12 = B_21, B_13=B31, B23=B32
    B_12 = fieldProp( 1 )%species%resi_coeff(2)
    B_13 = fieldProp( 1 )%species%resi_coeff(3)
    B_23 = fieldProp( 2 )%species%resi_coeff(3)

    Bratio_12 = B_12*paramB_inv
    Bratio_13 = B_13*paramB_inv
    Bratio_23 = B_23*paramB_inv
!write(dbgUnit,*) 'phi ', phi
    mbbEq_12 = Bratio_12*phi(1)
    mbbEq_13 = Bratio_13*phi(1)

    mbbEq_21 = Bratio_12*phi(2)
    mbbEq_23 = Bratio_23*phi(2)

    mbbEq_31 = Bratio_13*phi(3)
    mbbEq_32 = Bratio_23*phi(3)

    ! factor need to build matrixA to solve LSE
    ! dummy diagonal entries 
    mbb(1,1) = -1.0_rk
    mbb(1,2) = mbbEq_21*omegadiv2
    mbb(1,3) = mbbEq_31*omegadiv2

    mbb(2,1) = mbbEq_12*omegadiv2
    mbb(2,2) = -1.0_rk
    mbb(2,3) = mbbEq_32*omegadiv2

    mbb(3,1) = mbbEq_13*omegadiv2
    mbb(3,2) = mbbEq_23*omegadiv2
    mbb(3,3) = -1.0_rk

    nodeloop: do iElem = nStart, nEnd

      !species 1
      pdfTmp( Q__W,1 ) = instate(?FETCH?( varPos(Q__W,1), iElem, QQF3, neigh ))
      pdfTmp( Q__S,1 ) = instate(?FETCH?( varPos(Q__S,1), iElem, QQF3, neigh ))
      pdfTmp( Q__B,1 ) = instate(?FETCH?( varPos(Q__B,1), iElem, QQF3, neigh ))
      pdfTmp( Q__E,1 ) = instate(?FETCH?( varPos(Q__E,1), iElem, QQF3, neigh ))
      pdfTmp( Q__N,1 ) = instate(?FETCH?( varPos(Q__N,1), iElem, QQF3, neigh ))
      pdfTmp( Q__T,1 ) = instate(?FETCH?( varPos(Q__T,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BS,1 ) = instate(?FETCH?( varPos(Q_BS,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TS,1 ) = instate(?FETCH?( varPos(Q_TS,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BN,1 ) = instate(?FETCH?( varPos(Q_BN,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TN,1 ) = instate(?FETCH?( varPos(Q_TN,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BW,1 ) = instate(?FETCH?( varPos(Q_BW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BE,1 ) = instate(?FETCH?( varPos(Q_BE,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TW,1 ) = instate(?FETCH?( varPos(Q_TW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TE,1 ) = instate(?FETCH?( varPos(Q_TE,1), iElem, QQF3, neigh ))
      pdfTmp( Q_SW,1 ) = instate(?FETCH?( varPos(Q_SW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_NW,1 ) = instate(?FETCH?( varPos(Q_NW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_SE,1 ) = instate(?FETCH?( varPos(Q_SE,1), iElem, QQF3, neigh ))
      pdfTmp( Q_NE,1 ) = instate(?FETCH?( varPos(Q_NE,1), iElem, QQF3, neigh ))
      pdfTmp( Q__0,1 ) = instate(?FETCH?( varPos(Q__0,1), iElem, QQF3, neigh ))
      !species 2
      pdfTmp( Q__W,2 ) = instate(?FETCH?( varPos(Q__W,2), iElem, QQF3, neigh ))
      pdfTmp( Q__S,2 ) = instate(?FETCH?( varPos(Q__S,2), iElem, QQF3, neigh ))
      pdfTmp( Q__B,2 ) = instate(?FETCH?( varPos(Q__B,2), iElem, QQF3, neigh ))
      pdfTmp( Q__E,2 ) = instate(?FETCH?( varPos(Q__E,2), iElem, QQF3, neigh ))
      pdfTmp( Q__N,2 ) = instate(?FETCH?( varPos(Q__N,2), iElem, QQF3, neigh ))
      pdfTmp( Q__T,2 ) = instate(?FETCH?( varPos(Q__T,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BS,2 ) = instate(?FETCH?( varPos(Q_BS,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TS,2 ) = instate(?FETCH?( varPos(Q_TS,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BN,2 ) = instate(?FETCH?( varPos(Q_BN,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TN,2 ) = instate(?FETCH?( varPos(Q_TN,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BW,2 ) = instate(?FETCH?( varPos(Q_BW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BE,2 ) = instate(?FETCH?( varPos(Q_BE,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TW,2 ) = instate(?FETCH?( varPos(Q_TW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TE,2 ) = instate(?FETCH?( varPos(Q_TE,2), iElem, QQF3, neigh ))
      pdfTmp( Q_SW,2 ) = instate(?FETCH?( varPos(Q_SW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_NW,2 ) = instate(?FETCH?( varPos(Q_NW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_SE,2 ) = instate(?FETCH?( varPos(Q_SE,2), iElem, QQF3, neigh ))
      pdfTmp( Q_NE,2 ) = instate(?FETCH?( varPos(Q_NE,2), iElem, QQF3, neigh ))
      pdfTmp( Q__0,2 ) = instate(?FETCH?( varPos(Q__0,2), iElem, QQF3, neigh ))
      !species 3
      pdfTmp( Q__W,3 ) = instate(?FETCH?( varPos(Q__W,3), iElem, QQF3, neigh ))
      pdfTmp( Q__S,3 ) = instate(?FETCH?( varPos(Q__S,3), iElem, QQF3, neigh ))
      pdfTmp( Q__B,3 ) = instate(?FETCH?( varPos(Q__B,3), iElem, QQF3, neigh ))
      pdfTmp( Q__E,3 ) = instate(?FETCH?( varPos(Q__E,3), iElem, QQF3, neigh ))
      pdfTmp( Q__N,3 ) = instate(?FETCH?( varPos(Q__N,3), iElem, QQF3, neigh ))
      pdfTmp( Q__T,3 ) = instate(?FETCH?( varPos(Q__T,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BS,3 ) = instate(?FETCH?( varPos(Q_BS,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TS,3 ) = instate(?FETCH?( varPos(Q_TS,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BN,3 ) = instate(?FETCH?( varPos(Q_BN,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TN,3 ) = instate(?FETCH?( varPos(Q_TN,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BW,3 ) = instate(?FETCH?( varPos(Q_BW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BE,3 ) = instate(?FETCH?( varPos(Q_BE,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TW,3 ) = instate(?FETCH?( varPos(Q_TW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TE,3 ) = instate(?FETCH?( varPos(Q_TE,3), iElem, QQF3, neigh ))
      pdfTmp( Q_SW,3 ) = instate(?FETCH?( varPos(Q_SW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_NW,3 ) = instate(?FETCH?( varPos(Q_NW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_SE,3 ) = instate(?FETCH?( varPos(Q_SE,3), iElem, QQF3, neigh ))
      pdfTmp( Q_NE,3 ) = instate(?FETCH?( varPos(Q_NE,3), iElem, QQF3, neigh ))
      pdfTmp( Q__0,3 ) = instate(?FETCH?( varPos(Q__0,3), iElem, QQF3, neigh ))

      !density 1
      rsigma(1) = pdfTmp(Q_NE,1) + pdfTmp(Q__N,1) + pdfTmp(Q_NW,1)             &
        &       + pdfTmp(Q__W,1) + pdfTmp(Q_SW,1) + pdfTmp(Q__S,1)             &
        &       + pdfTmp(Q_SE,1) + pdfTmp(Q__E,1) + pdfTmp(Q__T,1)             &
        &       + pdfTmp(Q_TE,1) + pdfTmp(Q_TN,1) + pdfTmp(Q_TW,1)             &
        &       + pdfTmp(Q_TS,1) + pdfTmp(Q__B,1) + pdfTmp(Q_BE,1)             &
        &       + pdfTmp(Q_BN,1) + pdfTmp(Q_BW,1) + pdfTmp(Q_BS,1)             &
        &       + pdfTmp(Q__0,1)

      !density 2
      rsigma(2) = pdfTmp(Q_NE,2) + pdfTmp(Q__N,2) + pdfTmp(Q_NW,2)             &
        &       + pdfTmp(Q__W,2) + pdfTmp(Q_SW,2) + pdfTmp(Q__S,2)             &
        &       + pdfTmp(Q_SE,2) + pdfTmp(Q__E,2) + pdfTmp(Q__T,2)             &
        &       + pdfTmp(Q_TE,2) + pdfTmp(Q_TN,2) + pdfTmp(Q_TW,2)             &
        &       + pdfTmp(Q_TS,2) + pdfTmp(Q__B,2) + pdfTmp(Q_BE,2)             &
        &       + pdfTmp(Q_BN,2) + pdfTmp(Q_BW,2) + pdfTmp(Q_BS,2)             &
        &       + pdfTmp(Q__0,2)

      !density 3
      rsigma(3) = pdfTmp(Q_NE,3) + pdfTmp(Q__N,3) + pdfTmp(Q_NW,3)             &
        &       + pdfTmp(Q__W,3) + pdfTmp(Q_SW,3) + pdfTmp(Q__S,3)             &
        &       + pdfTmp(Q_SE,3) + pdfTmp(Q__E,3) + pdfTmp(Q__T,3)             &
        &       + pdfTmp(Q_TE,3) + pdfTmp(Q_TN,3) + pdfTmp(Q_TW,3)             &
        &       + pdfTmp(Q_TS,3) + pdfTmp(Q__B,3) + pdfTmp(Q_BE,3)             &
        &       + pdfTmp(Q_BN,3) + pdfTmp(Q_BW,3) + pdfTmp(Q_BS,3)             &
        &       + pdfTmp(Q__0,3)

 
      !local x,y,z - momentum of species 1
      gqxsigma(1) = pdfTmp(Q_NE,1) + pdfTmp(Q_SE,1) + pdfTmp(Q__E,1)           &
        &         + pdfTmp(Q_TE,1) + pdfTmp(Q_BE,1)                            &
        &         - pdfTmp(Q_NW,1) - pdfTmp(Q__W,1) - pdfTmp(Q_SW,1)           &
        &         - pdfTmp(Q_TW,1) - pdfTmp(Q_BW,1)
                         
      gqysigma(1) = pdfTmp(Q_NE,1) + pdfTmp(Q__N,1) + pdfTmp(Q_NW,1)           &
        &         + pdfTmp(Q_TN,1) + pdfTmp(Q_BN,1)                            &
        &         - pdfTmp(Q_SW,1) - pdfTmp(Q__S,1) - pdfTmp(Q_SE,1)           &
        &         - pdfTmp(Q_TS,1) - pdfTmp(Q_BS,1)
                         
      gqzsigma(1) = pdfTmp(Q__T,1) + pdfTmp(Q_TE,1) + pdfTmp(Q_TN,1)           &
        &         + pdfTmp(Q_TW,1) + pdfTmp(Q_TS,1)                            & 
        &         - pdfTmp(Q__B,1) - pdfTmp(Q_BE,1) - pdfTmp(Q_BN,1)           &
        &         - pdfTmp(Q_BW,1) - pdfTmp(Q_BS,1)

      !local x,y,z - momentum of species 2
      gqxsigma(2) = pdfTmp(Q_NE,2) + pdfTmp(Q_SE,2) + pdfTmp(Q__E,2)           &
        &         + pdfTmp(Q_TE,2) + pdfTmp(Q_BE,2)                            &
        &         - pdfTmp(Q_NW,2) - pdfTmp(Q__W,2) - pdfTmp(Q_SW,2)           &
        &         - pdfTmp(Q_TW,2) - pdfTmp(Q_BW,2)
                         
      gqysigma(2) = pdfTmp(Q_NE,2) + pdfTmp(Q__N,2) + pdfTmp(Q_NW,2)           &
        &         + pdfTmp(Q_TN,2) + pdfTmp(Q_BN,2)                            &
        &         - pdfTmp(Q_SW,2) - pdfTmp(Q__S,2) - pdfTmp(Q_SE,2)           &
        &         - pdfTmp(Q_TS,2) - pdfTmp(Q_BS,2)
                         
      gqzsigma(2) = pdfTmp(Q__T,2) + pdfTmp(Q_TE,2) + pdfTmp(Q_TN,2)           &
        &         + pdfTmp(Q_TW,2) + pdfTmp(Q_TS,2)                            & 
        &         - pdfTmp(Q__B,2) - pdfTmp(Q_BE,2) - pdfTmp(Q_BN,2)           &
        &         - pdfTmp(Q_BW,2) - pdfTmp(Q_BS,2)

      !local x,y,z - momentum of species 3
      gqxsigma(3) = pdfTmp(Q_NE,3) + pdfTmp(Q_SE,3) + pdfTmp(Q__E,3)           &
        &         + pdfTmp(Q_TE,3) + pdfTmp(Q_BE,3)                            &
        &         - pdfTmp(Q_NW,3) - pdfTmp(Q__W,3) - pdfTmp(Q_SW,3)           &
        &         - pdfTmp(Q_TW,3) - pdfTmp(Q_BW,3)
                         
      gqysigma(3) = pdfTmp(Q_NE,3) + pdfTmp(Q__N,3) + pdfTmp(Q_NW,3)           &
        &         + pdfTmp(Q_TN,3) + pdfTmp(Q_BN,3)                            &
        &         - pdfTmp(Q_SW,3) - pdfTmp(Q__S,3) - pdfTmp(Q_SE,3)           &
        &         - pdfTmp(Q_TS,3) - pdfTmp(Q_BS,3)
                         
      gqzsigma(3) = pdfTmp(Q__T,3) + pdfTmp(Q_TE,3) + pdfTmp(Q_TN,3)           &
        &         + pdfTmp(Q_TW,3) + pdfTmp(Q_TS,3)                            & 
        &         - pdfTmp(Q__B,3) - pdfTmp(Q_BE,3) - pdfTmp(Q_BN,3)           &
        &         - pdfTmp(Q_BW,3) - pdfTmp(Q_BS,3)

!write(dbgUnit,*) 'First moments 1', gqxsigma(1), gqysigma(1), gqzsigma(1)
!write(dbgUnit,*) 'First moments 2', gqxsigma(2), gqysigma(2), gqzsigma(2)
!write(dbgUnit,*) 'First moments 3', gqxsigma(3), gqysigma(3), gqzsigma(3)

      !number density
      num_dens(:) = rsigma(:)*molWeight_inv(:)
!write(dbgUnit,*) 'Number density ', num_dens(:)      

      !total number density
      totNum_dens_inv = 1.0_rk/(num_dens(1)+num_dens(2)+num_dens(3))

      !mole fraction
      moleFrac(:) = num_dens(:)*totNum_dens_inv
!write(dbgUnit,*) 'moleFrac ', moleFrac(:)      

      !matrix A of the velocity lse
      A(1,1) = 1.0_rk + moleFrac(2)*mbb(2,1) + moleFrac(3)*mbb(3,1) 
      A(1,2) = - moleFrac(1)*mbb(1,2)
      A(1,3) = - moleFrac(1)*mbb(1,3)

      A(2,1) = - moleFrac(2)*mbb(2,1)
      A(2,2) = 1.0_rk + moleFrac(1)*mbb(1,2) + moleFrac(3)*mbb(3,2)
      A(2,3) = - moleFrac(2)*mbb(2,3)

      A(3,1) = - moleFrac(3)*mbb(3,1)
      A(3,2) = - moleFrac(3)*mbb(3,2)
      A(3,3) = 1.0_rk + moleFrac(1)*mbb(1,3) + moleFrac(2)*mbb(2,3)
!write(dbgUnit,*) 'matriA ', A

      a11_inv = 1.0_rk/A(1,1)
      a13_11 = A(1,3)*a11_inv
      a12_11 = A(1,2)*a11_inv
      a31_11 = A(3,1)*a11_inv
      a21_11 = A(2,1)*a11_inv
      a23fac = A(2,3) - A(2,1)*a13_11
      a22fac_inv = 1.0_rk/(A(2,2) - A(2,1)*a12_11)
  
      zfac = (A(3,1)*a12_11 - A(3,2)) * a22fac_inv
      zcoeff = 1._rk / (a23fac*zfac - A(3,1)*a13_11 + A(3,3))
!      !x-momentum
!      !species 3
      yfac = gqxsigma(2) - gqxsigma(1)*a21_11

      qxsigma(3) = (gqxsigma(3) - a31_11*gqxsigma(1) + yfac*zfac)*zcoeff
!      write(*,*)'z',qxsigma(3)

      !species 2
      qxsigma(2) = (yfac - a23fac*qxsigma(3))*a22fac_inv
      !species 1
      qxsigma(1) = (gqxsigma(1) - A(1,2)*qxsigma(2) - A(1,3)*qxsigma(3))*a11_inv
!      write(*,*) 'qxsigma',qxsigma
      !y-momentum
      yfac = gqysigma(2) - gqysigma(1)*a21_11
      qysigma(3) = (gqysigma(3) - a31_11*gqysigma(1) + yfac*zfac)*zcoeff

      qysigma(2) = (yfac - a23fac*qysigma(3))*a22fac_inv

      qysigma(1) = (gqysigma(1) - A(1,2)*qysigma(2) - A(1,3)*qysigma(3))*a11_inv
!      write(*,*) 'qysigma',qysigma
      !z-momentum
      yfac = gqzsigma(2) - gqzsigma(1)*a21_11
      qzsigma(3) = (gqzsigma(3) - a31_11*gqzsigma(1) + yfac*zfac)*zcoeff
      
      qzsigma(2) = (yfac - a23fac*qzsigma(3))*a22fac_inv

      qzsigma(1) = (gqzsigma(1) - A(1,2)*qzsigma(2) - A(1,3)*qzsigma(3))*a11_inv
!      write(*,*) 'qzsigma',qzsigma
!      stop

      rsigmaInv(1) = 1.0_rk/rsigma(1)
      rsigmaInv(2) = 1.0_rk/rsigma(2)
      rsigmaInv(3) = 1.0_rk/rsigma(3)

      uxsigma(1) = qxsigma(1)*rsigmaInv(1)
      uxsigma(2) = qxsigma(2)*rsigmaInv(2)
      uxsigma(3) = qxsigma(3)*rsigmaInv(3)

      uysigma(1) = qysigma(1)*rsigmaInv(1)
      uysigma(2) = qysigma(2)*rsigmaInv(2)
      uysigma(3) = qysigma(3)*rsigmaInv(3)

      uzsigma(1) = qzsigma(1)*rsigmaInv(1)
      uzsigma(2) = qzsigma(2)*rsigmaInv(2)
      uzsigma(3) = qzsigma(3)*rsigmaInv(3)

!write(dbgUnit,*) 'velocity 1', uxsigma(1), uysigma(1), uzsigma(1)
!write(dbgUnit,*) 'velocity 2', uxsigma(2), uysigma(2), uzsigma(2)
!write(dbgUnit,*) 'velocity 3', uxsigma(3), uysigma(3), uzsigma(3)

      !compute equilibrium velocity to compute feq
      chi12 = mbbEq_12 * moleFrac(2)
      chi13 = mbbEq_13 * moleFrac(3)
      chi21 = mbbEq_21 * moleFrac(1) 
      chi23 = mbbEq_23 * moleFrac(3)
      chi31 = mbbEq_31 * moleFrac(1)
      chi32 = mbbEq_32 * moleFrac(2)
      !ux
      uxstar(1) =  uxsigma(1) + chi12*(uxsigma(2)-uxsigma(1))                  &
                             &+ chi13*(uxsigma(3)-uxsigma(1))

      uxstar(2) =  uxsigma(2) + chi21*(uxsigma(1)-uxsigma(2))                  &
                             &+ chi23*(uxsigma(3)-uxsigma(2))

      uxstar(3) =  uxsigma(3) + chi31*(uxsigma(1)-uxsigma(3))                  &
                             &+ chi32*(uxsigma(2)-uxsigma(3))
      !uy
      uystar(1) =  uysigma(1) + chi12*(uysigma(2)-uysigma(1))                  &
                             &+ chi13*(uysigma(3)-uysigma(1))

      uystar(2) =  uysigma(2) + chi21*(uysigma(1)-uysigma(2))                  &
                             &+ chi23*(uysigma(3)-uysigma(2))

      uystar(3) =  uysigma(3) + chi31*(uysigma(1)-uysigma(3))                  &
                             &+ chi32*(uysigma(2)-uysigma(3))
      !uz
      uzstar(1) =  uzsigma(1) + chi12*(uzsigma(2)-uzsigma(1))                  &
                             &+ chi13*(uzsigma(3)-uzsigma(1))

      uzstar(2) =  uzsigma(2) + chi21*(uzsigma(1)-uzsigma(2))                  &
                             &+ chi23*(uzsigma(3)-uzsigma(2))

      uzstar(3) =  uzsigma(3) + chi31*(uzsigma(1)-uzsigma(3))                  &
                             &+ chi32*(uzsigma(2)-uzsigma(3))

!write(dbgUnit,*) 'eqVel 1', uxstar(1), uystar(1), uzstar(1)
!write(dbgUnit,*) 'eqVel 2', uxstar(2), uystar(2), uzstar(2)
!write(dbgUnit,*) 'eqVel 3', uxstar(3), uystar(3), uzstar(3)

      ! total mass density
      totMass_densInv = 1.0_rk/(rsigma(1)+rsigma(2)+rsigma(3))

      ! mass averaged mixture velocity three components
      velAvg(1) = ( rsigma(1) * uxsigma(1) &
        &         + rsigma(2) * uxsigma(2) &
        &         + rsigma(3) * uxsigma(3) ) * totMass_densInv * theta_eq

      velAvg(2) = ( rsigma(1) * uysigma(1) &
        &         + rsigma(2) * uysigma(2) &
        &         + rsigma(3) * uysigma(3) ) * totMass_densInv * theta_eq

      velAvg(3) = ( rsigma(1) * uzsigma(1) &
        &         + rsigma(2) * uzsigma(2) &
        &         + rsigma(3) * uzsigma(3) ) * totMass_densInv * theta_eq
      
      ! velocity in quadratic term of equilibrium
      velQuadTerm_x(1) = velAvg(1) + theta_eq_spc*uxstar(1)
      velQuadTerm_y(1) = velAvg(2) + theta_eq_spc*uystar(1)
      velQuadTerm_z(1) = velAvg(3) + theta_eq_spc*uzstar(1)

      !compute equilibrium and do collision
      usqr(1) = (velQuadTerm_x(1) * velQuadTerm_x(1)                           &
        &     + velQuadTerm_y(1) * velQuadTerm_y(1)                            &
        &     + velQuadTerm_z(1) * velQuadTerm_z(1)) * t2cs2inv

      omegaRho(:) = rsigma(:)*omega_fac
      omegaRho_d(:) = omegaRho(:) * div1_4

      ! species 1
      ! equilibrium at rest
      outstate( ?SAVE?( varPos(Q__0,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q__0,1) +                                       &
              & div1_3 * omegaRho(1) * (( 3._rk - 2._rk * phi(1)) - usqr(1) )
!write(dbgUnit,*) 'omegaRho ', omegaRho(1), rsigma(1), omega_fac, omega_o 
!write(dbgUnit,*) 'pdfTmp ', pdfTmp(Q__0,1), omega_o*pdfTmp(Q__0,1)              
!write(dbgUnit,*) 'eq ', div1_3 * rsigma(1) * (( 3._rk - 2._rk * phi(1)) - usqr(1) )              
!write(dbgUnit,*) 'state ', outstate( ?SAVE?( varPos(Q__0,1), iElem, QQF3, neigh ) )
!stop

      omegaRho_dm(1) = omegaRho(1) * (phi(1) - usqr(1)) * div1_18

      !directional velocity factor 
      sum1_1(1) = omegaRho_d(1) * uxstar(1) * div3_4h
      sum1_2(1) = omegaRho_d(1) * velQuadTerm_x(1)**2.0 + omegaRho_dm(1)

      outstate( ?SAVE?( varPos(Q__W,1), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__W,1) - sum1_1(1) + sum1_2(1) 
      outstate( ?SAVE?( varPos(Q__E,1), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__E,1) + sum1_1(1) + sum1_2(1)

      sum2_1(1) = omegaRho_d(1) * uystar(1) * div3_4h
      sum2_2(1) = omegaRho_d(1) * velQuadTerm_y(1)**2.0 + omegaRho_dm(1)

      outstate( ?SAVE?( varPos(Q__S,1), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__S,1) - sum2_1(1) + sum2_2(1)
      outstate( ?SAVE?( varPos(Q__N,1), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__N,1) + sum2_1(1) + sum2_2(1)

      sum3_1(1) = omegaRho_d(1) * uzstar(1) * div3_4h
      sum3_2(1) = omegaRho_d(1) * velQuadTerm_z(1)**2.0 + omegaRho_dm(1)

      outstate( ?SAVE?( varPos(Q__B,1), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__B,1) - sum3_1(1) + sum3_2(1)
      outstate( ?SAVE?( varPos(Q__T,1), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__T,1) + sum3_1(1) + sum3_2(1)

      !top north / bottom south        
      omegaRho_o(1) = omegaRho_d(1) * 0.5_rk
      omegaRho_om(1) = omegaRho_dm(1) * 0.5_rk
      sum4_1(1) = omegaRho_o(1) * ( uystar(1) + uzstar(1) ) *div3_4h
      sum4_2(1) = omegaRho_o(1) * ( velQuadTerm_y(1) + velQuadTerm_z(1) )**2.0 &
        &       + omegaRho_om(1)

      outstate( ?SAVE?( varPos(Q_BS,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BS,1) - sum4_1(1) + sum4_2(1) 
      outstate( ?SAVE?( varPos(Q_TN,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TN,1) + sum4_1(1) + sum4_2(1) 

      !top south / bottom north
      sum5_1(1) = omegaRho_o(1) * ( -uystar(1) + uzstar(1) ) *div3_4h
      sum5_2(1) = omegaRho_o(1) * ( -velQuadTerm_y(1) + velQuadTerm_z(1) )**2.0&
        &       + omegaRho_om(1)

      outstate( ?SAVE?( varPos(Q_BN,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BN,1) - sum5_1(1) + sum5_2(1) 
      outstate( ?SAVE?( varPos(Q_TS,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TS,1) + sum5_1(1) + sum5_2(1)

      !top east / bottom west
      sum6_1(1) = omegaRho_o(1) * ( uxstar(1) + uzstar(1) ) *div3_4h
      sum6_2(1) = omegaRho_o(1) * ( velQuadTerm_x(1) + velQuadTerm_z(1) )**2.0 &
        &       + omegaRho_om(1)

      outstate( ?SAVE?( varPos(Q_BW,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BW,1) - sum6_1(1) + sum6_2(1) 
      outstate( ?SAVE?( varPos(Q_TE,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TE,1) + sum6_1(1) + sum6_2(1) 
 
      !top west / bottom east
      sum7_1(1) = omegaRho_o(1) * ( -uxstar(1) + uzstar(1) ) *div3_4h
      sum7_2(1) = omegaRho_o(1) * ( -velQuadTerm_x(1) + velQuadTerm_z(1) )**2.0&
        &       + omegaRho_om(1)

      outstate( ?SAVE?( varPos(Q_BE,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BE,1) - sum7_1(1) + sum7_2(1) 
      outstate( ?SAVE?( varPos(Q_TW,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TW,1) + sum7_1(1) + sum7_2(1) 

      !north east / south west
      sum8_1(1) = omegaRho_o(1) * ( uxstar(1) + uystar(1) ) *div3_4h
      sum8_2(1) = omegaRho_o(1) * ( velQuadTerm_x(1) + velQuadTerm_y(1) )**2.0 &
        &       + omegaRho_om(1)

      outstate( ?SAVE?( varPos(Q_SW,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_SW,1) - sum8_1(1) + sum8_2(1) 
      outstate( ?SAVE?( varPos(Q_NE,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_NE,1) + sum8_1(1) + sum8_2(1) 

      !north west / south east
      sum9_1(1) = omegaRho_o(1) * ( -uxstar(1) + uystar(1) ) *div3_4h
      sum9_2(1) = omegaRho_o(1) * ( -velQuadTerm_x(1) + velQuadTerm_y(1) )**2.0&
        &       + omegaRho_om(1)

      outstate( ?SAVE?( varPos(Q_SE,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_SE,1) - sum9_1(1) + sum9_2(1)
      outstate( ?SAVE?( varPos(Q_NW,1), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_NW,1) + sum9_1(1) + sum9_2(1)

      ! species 2
      ! velocity in quadratic term of equilibrium
      velQuadTerm_x(2) = velAvg(1) + theta_eq_spc*uxstar(2)
      velQuadTerm_y(2) = velAvg(2) + theta_eq_spc*uystar(2)
      velQuadTerm_z(2) = velAvg(3) + theta_eq_spc*uzstar(2)

      usqr(2) = (velQuadTerm_x(2) * velQuadTerm_x(2)                           &
        &     + velQuadTerm_y(2) * velQuadTerm_y(2)                            &
        &     + velQuadTerm_z(2) * velQuadTerm_z(2)) * t2cs2inv

      outstate( ?SAVE?( varPos(Q__0,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q__0,2) + & 
              & div1_3 * omegaRho(2) * (( 3._rk - 2._rk * phi(2)) - usqr(2) )

      omegaRho_dm(2) = omegaRho(2) * (phi(2) - usqr(2)) * div1_18
      !directional velocity factor 
      sum1_1(2) = omegaRho_d(2) * uxstar(2) * div3_4h
      sum1_2(2) = omegaRho_d(2) * velQuadTerm_x(2)**2.0 + omegaRho_dm(2)

      outstate( ?SAVE?( varPos(Q__W,2), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__W,2) - sum1_1(2) + sum1_2(2)
      outstate( ?SAVE?( varPos(Q__E,2), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__E,2) + sum1_1(2) + sum1_2(2)

      sum2_1(2) = omegaRho_d(2) * uystar(2) * div3_4h
      sum2_2(2) = omegaRho_d(2) * velQuadTerm_y(2)**2.0 + omegaRho_dm(2)

      outstate( ?SAVE?( varPos(Q__S,2), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__S,2) - sum2_1(2) + sum2_2(2)
      outstate( ?SAVE?( varPos(Q__N,2), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__N,2) + sum2_1(2) + sum2_2(2)

      sum3_1(2) = omegaRho_d(2) * uzstar(2) * div3_4h
      sum3_2(2) = omegaRho_d(2) * velQuadTerm_z(2)**2.0 + omegaRho_dm(2)

      outstate( ?SAVE?( varPos(Q__B,2), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__B,2) - sum3_1(2) + sum3_2(2)
      outstate( ?SAVE?( varPos(Q__T,2), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__T,2) + sum3_1(2) + sum3_2(2)

      !top north / bottom south        
      omegaRho_o(2) = omegaRho_d(2) * 0.5_rk
      omegaRho_om(2) = omegaRho_dm(2) * 0.5_rk
      sum4_1(2) = omegaRho_o(2) * ( uystar(2) + uzstar(2) ) *div3_4h
      sum4_2(2) = omegaRho_o(2) * ( velQuadTerm_y(2) + velQuadTerm_z(2) )**2.0 &
        &       + omegaRho_om(2)

      outstate( ?SAVE?( varPos(Q_BS,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BS,2) - sum4_1(2) + sum4_2(2) 
      outstate( ?SAVE?( varPos(Q_TN,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TN,2) + sum4_1(2) + sum4_2(2) 

      !top south / bottom north
      sum5_1(2) = omegaRho_o(2) * ( -uystar(2) + uzstar(2) ) *div3_4h
      sum5_2(2) = omegaRho_o(2) * ( -velQuadTerm_y(2) + velQuadTerm_z(2) )**2.0&
        &       + omegaRho_om(2)

      outstate( ?SAVE?( varPos(Q_BN,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BN,2) - sum5_1(2) + sum5_2(2) 
      outstate( ?SAVE?( varPos(Q_TS,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TS,2) + sum5_1(2) + sum5_2(2)

      !top east / bottom west
      sum6_1(2) = omegaRho_o(2) * ( uxstar(2) + uzstar(2) ) *div3_4h
      sum6_2(2) = omegaRho_o(2) * ( velQuadTerm_x(2) + velQuadTerm_z(2) )**2.0 &
        &       + omegaRho_om(2)

      outstate( ?SAVE?( varPos(Q_BW,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BW,2) - sum6_1(2) + sum6_2(2) 
      outstate( ?SAVE?( varPos(Q_TE,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TE,2) + sum6_1(2) + sum6_2(2) 

      !top west / bottom east
      sum7_1(2) = omegaRho_o(2) * ( -uxstar(2) + uzstar(2) ) *div3_4h
      sum7_2(2) = omegaRho_o(2) * ( -velQuadTerm_x(2) + velQuadTerm_z(2) )**2.0&
        &       + omegaRho_om(2)

      outstate( ?SAVE?( varPos(Q_BE,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BE,2) - sum7_1(2) + sum7_2(2) 
      outstate( ?SAVE?( varPos(Q_TW,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TW,2) + sum7_1(2) + sum7_2(2) 

      !north east / south west
      sum8_1(2) = omegaRho_o(2) * ( uxstar(2) + uystar(2) ) *div3_4h
      sum8_2(2) = omegaRho_o(2) * ( velQuadTerm_x(2) + velQuadTerm_y(2) )**2.0 &
        &       + omegaRho_om(2)

      outstate( ?SAVE?( varPos(Q_SW,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_SW,2) - sum8_1(2) + sum8_2(2) 
      outstate( ?SAVE?( varPos(Q_NE,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_NE,2) + sum8_1(2) + sum8_2(2) 

      !north west / south east
      sum9_1(2) = omegaRho_o(2) * ( -uxstar(2) + uystar(2) ) *div3_4h
      sum9_2(2) = omegaRho_o(2) * ( -velQuadTerm_x(2) + velQuadTerm_y(2) )**2.0&
        &       + omegaRho_om(2)

      outstate( ?SAVE?( varPos(Q_SE,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_SE,2) - sum9_1(2) + sum9_2(2)
      outstate( ?SAVE?( varPos(Q_NW,2), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_NW,2) + sum9_1(2) + sum9_2(2)

      ! species 3
      ! velocity in quadratic term of equilibrium
      velQuadTerm_x(3) = velAvg(1) + theta_eq_spc*uxstar(3)
      velQuadTerm_y(3) = velAvg(2) + theta_eq_spc*uystar(3)
      velQuadTerm_z(3) = velAvg(3) + theta_eq_spc*uzstar(3)

      usqr(3) = (velQuadTerm_x(3) * velQuadTerm_x(3)                           &
        &     + velQuadTerm_y(3) * velQuadTerm_y(3)                            &
        &     + velQuadTerm_z(3) * velQuadTerm_z(3)) * t2cs2inv

      outstate( ?SAVE?( varPos(Q__0,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q__0,3) + & 
              & div1_3 * omegaRho(3) * (( 3._rk - 2._rk * phi(3)) - usqr(3) )

      omegaRho_dm(3) = omegaRho(3) * (phi(3) - usqr(3)) * div1_18
      !directional velocity factor 
      sum1_1(3) = omegaRho_d(3) * uxstar(3) * div3_4h
      sum1_2(3) = omegaRho_d(3) * velQuadTerm_x(3)**2.0 + omegaRho_dm(3)

      outstate( ?SAVE?( varPos(Q__W,3), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__W,3) - sum1_1(3) + sum1_2(3)
      outstate( ?SAVE?( varPos(Q__E,3), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__E,3) + sum1_1(3) + sum1_2(3)

      sum2_1(3) = omegaRho_d(3) * uystar(3) * div3_4h
      sum2_2(3) = omegaRho_d(3) * velQuadTerm_y(3)**2.0 + omegaRho_dm(3)

      outstate( ?SAVE?( varPos(Q__S,3), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__S,3) - sum2_1(3) + sum2_2(3)
      outstate( ?SAVE?( varPos(Q__N,3), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__N,3) + sum2_1(3) + sum2_2(3)

      sum3_1(3) = omegaRho_d(3) * uzstar(3) * div3_4h
      sum3_2(3) = omegaRho_d(3) * velQuadTerm_z(3)**2.0 + omegaRho_dm(3)

      outstate( ?SAVE?( varPos(Q__B,3), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__B,3) - sum3_1(3) + sum3_2(3)
      outstate( ?SAVE?( varPos(Q__T,3), iElem, QQF3, neigh ) )                 &
        & = omega_o * pdfTmp(Q__T,3) + sum3_1(3) + sum3_2(3)

      !top north / bottom south        
      omegaRho_o(3) = omegaRho_d(3) * 0.5_rk
      omegaRho_om(3) = omegaRho_dm(3) * 0.5_rk
      sum4_1(3) = omegaRho_o(3) * ( uystar(3) + uzstar(3) ) *div3_4h
      sum4_2(3) = omegaRho_o(3) * ( velQuadTerm_y(3) + velQuadTerm_z(3) )**2.0 &
        &       + omegaRho_om(3)

      outstate( ?SAVE?( varPos(Q_BS,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BS,3) - sum4_1(3) + sum4_2(3) 
      outstate( ?SAVE?( varPos(Q_TN,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TN,3) + sum4_1(3) + sum4_2(3) 

      !top south / bottom north
      sum5_1(3) = omegaRho_o(3) * ( -uystar(3) + uzstar(3) ) *div3_4h
      sum5_2(3) = omegaRho_o(3) * ( -velQuadTerm_y(3) + velQuadTerm_z(3) )**2.0&
        &       + omegaRho_om(3)

      outstate( ?SAVE?( varPos(Q_BN,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BN,3) - sum5_1(3) + sum5_2(3) 
      outstate( ?SAVE?( varPos(Q_TS,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TS,3) + sum5_1(3) + sum5_2(3)

      !top east / bottom west
      sum6_1(3) = omegaRho_o(3) * ( uxstar(3) + uzstar(3) ) *div3_4h
      sum6_2(3) = omegaRho_o(3) * ( velQuadTerm_x(3) + velQuadTerm_z(3) )**2.0 &
        &       + omegaRho_om(3)

      outstate( ?SAVE?( varPos(Q_BW,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BW,3) - sum6_1(3) + sum6_2(3) 
      outstate( ?SAVE?( varPos(Q_TE,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TE,3) + sum6_1(3) + sum6_2(3) 

      !top west / bottom east
      sum7_1(3) = omegaRho_o(3) * ( -uxstar(3) + uzstar(3) ) *div3_4h
      sum7_2(3) = omegaRho_o(3) * ( -velQuadTerm_x(3) + velQuadTerm_z(3) )**2.0&
        &       + omegaRho_om(3)

      outstate( ?SAVE?( varPos(Q_BE,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_BE,3) - sum7_1(3) + sum7_2(3) 
      outstate( ?SAVE?( varPos(Q_TW,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_TW,3) + sum7_1(3) + sum7_2(3) 
              
      !north east / south west
      sum8_1(3) = omegaRho_o(3) * ( uxstar(3) + uystar(3) ) *div3_4h
      sum8_2(3) = omegaRho_o(3) * ( velQuadTerm_x(3) + velQuadTerm_y(3) )**2.0 &
        &       + omegaRho_om(3)

      outstate( ?SAVE?( varPos(Q_SW,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_SW,3) - sum8_1(3) + sum8_2(3) 
      outstate( ?SAVE?( varPos(Q_NE,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_NE,3) + sum8_1(3) + sum8_2(3) 

      !north west / south east
      sum9_1(3) = omegaRho_o(3) * ( -uxstar(3) + uystar(3) ) *div3_4h
      sum9_2(3) = omegaRho_o(3) * ( -velQuadTerm_x(3) + velQuadTerm_y(3) )**2.0&
        &       + omegaRho_om(3)

      outstate( ?SAVE?( varPos(Q_SE,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_SE,3) - sum9_1(3) + sum9_2(3)
      outstate( ?SAVE?( varPos(Q_NW,3), iElem, QQF3, neigh ) ) =               &
              & omega_o*pdfTmp(Q_NW,3) + sum9_1(3) + sum9_2(3)

!write(dbgUnit,*) 'outpdf1', outstate(varPos(:,1))
!write(dbgUnit,*) 'outpdf2', outstate(varPos(:,2))
!write(dbgUnit,*) 'outpdf3', outstate(varPos(:,3))
!stop
    end do nodeloop

  end subroutine bgk_advRel_d3q19f3_MSLiquid


!******************************************************************************
  !> Unoptimized Advection relaxation routine for the multispecies BGK model 
  !!
  !! This routine contains the implementation based on the paper
  !! "A Lattice Boltzmann Scheme for liquid mixtures - Part II: Discretization  
  !! and Numerics, Jens Zudrop, Sabine Roller, Pietro Asinari. "\n
?? copy :: compute_routineHeader( bgk_advRel_MSLiquid )
    !------------------------------------------------------------------------ 
    !local variables
    real(kind=rk) :: pdfTmp( varSys%nScalars )  !< temporary local pdf values
    integer :: iElem, iField, iField_2, nFields, iDir
    integer :: vPos( layout%stencil(1)%QQ )
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: mass_dens, num_dens
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: moleFrac
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: first_moments
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: velocity
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: eqVel
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: molWeight, phi 
    real(kind=rk) :: totNum_dens, usqr, ucx, feq
    real(kind=rk) :: velAvg(3), velNew(3), ucxQuadTerm, theta_eq, totMassDens
    real(kind=rk) :: omega, omega_new, paramB
    real(kind=rk), &
      & dimension(scheme(iScheme)%nFields,scheme(iScheme)%nFields) :: matA,    &
      & invA, resi_coeff
    !--------------------------------------------------------------------------- 
    nFields = scheme(iScheme)%nFields
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    !molecular weight ratios
    phi = fieldProp(:)%species%molWeigRatio
    !phi = real(fieldProp(:)%species%molWeigRatio)
    !resistivities
    do iField = 1, nFields
      resi_coeff(iField,:) = fieldProp( iField )%species%resi_coeff
    enddo

    omega = scheme(iScheme)%mixture%relaxLvl(Level)%omega_diff
!    write(*,*) 'omega ', omega
    omega_new = (1.0_rk/(1.0_rk/omega + 0.5_rk))
!    write(*,*) 'omega_new ', omega_new

    paramB = scheme(iScheme)%mixture%paramB
    !equilibrium theta
    theta_eq = scheme(iScheme)%mixture%theta_eq

!    write(*,*) 'paramB ', paramB

!    write(dbgUnit,*) 'B 1', resi_coeff(1,:)
!    write(dbgUnit,*) 'B 2', resi_coeff(2,:)
!    write(dbgUnit,*) 'B 3', resi_coeff(3,:)
!write(dbgUnit(1),*) 'phi ', phi
!write(dbgUnit(1),*) 'weights ', layout%weight

    ! KM: debugging multispecies
    !outstate = instate

!write(dbgUnit(1),*)'time ', params%general%simControl%now%iter
    nodeloop: do iElem = nStart, nEnd
!write(dbgUnit(1),*) 'iElem', iElem

      mass_dens = 0._rk
      first_moments = 0.0_rk
      velocity = 0.0_rk
      eqVel = 0.0_rk
      pdfTmp = 0._rk

      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        ! compute field density and first moments
        do iDir = 1, layout%stencil(1)%QQ
          !store all field pdf in single element to local array pdfTmp
          pdfTmp( vPos(iDir) ) =                                               &
            & instate( ?FETCH?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
          !field density
          mass_dens( iField ) = mass_dens( iField ) + pdfTmp( vPos(iDir) )

          !field momentum (rho*u)
          first_moments( 1, iField ) = first_moments( 1, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 1, iDir )

          first_moments( 2, iField ) = first_moments( 2, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 2, iDir )

          first_moments( 3, iField ) = first_moments( 3, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 3, iDir )
!write(dbgunit,*) 's',s,'iDir', iDir, layout%stencil(1)%cxDir(3,iDir), pdfTmp( vPos(iDir) ), gqzsigma(s)
        enddo
      enddo
!write(dbgUnit(1),*) 'mass_dens ', mass_dens
!write(dbgUnit(1),*) 'first_moments ', first_moments

      !total mass density
      totmassDens = sum(mass_dens)

      ! solve linear system of equation for actual moments
      ! number density of all species
      num_dens(:) = mass_dens(:)/molWeight(:)

      !total number density
      totNum_dens = sum(num_dens(:))

      !mole fraction
      moleFrac(:) =  num_dens(:)/totNum_dens

      matA = 0.0_rk
      !build up matrix to solver LSE for actual velocity
      do iField = 1, nFields
        !set diagonal part 
        matA(iField, iField) = 1.0_rk
        do iField_2 = 1, nFields
          matA(iField, iField) = matA(iField, iField) + omega * 0.5_rk         &
            &                  * resi_coeff(iField, iField_2) * phi(iField)    &
            &                  * moleFrac(iField_2) / paramB
        end do
        !set non-diagonal part
        do iField_2 = 1, nFields
          matA(iField, iField_2) = matA(iField, iField_2) - omega * 0.5_rk     &
            &                    * resi_coeff(iField, iField_2) * phi(iField_2)&
            &                    * moleFrac(iField) / paramB
        end do
      end do

      ! invert matrix
      invA = invert_matrix( matA )

      !actual velocity of all species
      !KM: TGV testcase
      velocity(1, :) = matmul( invA, first_moments(1,:) ) / mass_dens(:)
      velocity(2, :) = matmul( invA, first_moments(2,:) ) / mass_dens(:)
      velocity(3, :) = matmul( invA, first_moments(3,:) ) / mass_dens(:)
!write(dbgUnit(1),*) 'velocity ', velocity

      ! compute equilibrim velocity
      do iField = 1, nFields
        eqVel( :, iField ) = velocity( :, iField )
        do iField_2 = 1, nFields
          eqVel( :, iField ) = eqVel( :, iField )                              &
            &                + resi_coeff( iField, iField_2 ) * phi(iField)    &
            &                * moleFrac(iField_2)                              &
            &                * (velocity(:, iField_2) - velocity(:,iField))    &
            &                / paramB 
        end do  
      end do
!write(dbgUnit(1),*) 'eqVel ', eqVel

      !compute mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens, velocity(1,:) )/totmassDens
      velAvg(2) = dot_product( mass_dens, velocity(2,:) )/totmassDens
      velAvg(3) = dot_product( mass_dens, velocity(3,:) )/totmassDens

      !compute equilibrium and do collision
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        !KM: TGV testcase
        velNew = theta_eq*velAvg + (1.0_rk - theta_eq)*eqVel(:, iField)

        !usqr = eqVel(1, iField) * eqVel(1, iField) &
        !  &  + eqVel(2, iField) * eqVel(2, iField) &
        !  &  + eqVel(3, iField) * eqVel(3, iField) 
        usqr = dot_product( velNew, velNew ) 
        usqr = usqr * t2cs2inv
!write(dbgUnit(1),*) 'omega_new ', omega_new
        do iDir = 1, layout%stencil(1)%QQ
          ucx = dble( layout%stencil(1)%cxDir( 1, iDir ))*eqVel(1, iField)     &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*eqVel(2, iField)     &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*eqVel(3, iField)

          ucxQuadTerm = dot_product(                                           &
            &           dble(layout%stencil(1)%cxDir(:, iDir)), velNew )

          !feq = layout%weight(iDir) * mass_dens(iField) * ( phi(iField)        &
          !  & + ucx * cs2inv + ucx * ucx * t2cs4inv - usqr )
          feq = layout%weight(iDir) * mass_dens(iField) * ( phi(iField)        &
            & + ucx * cs2inv + ucxQuadTerm * ucxQuadTerm * t2cs4inv - usqr )
          
          if ( iDir == layout%stencil(1)%restPosition ) then      
          ! equilibrium at rest
            select case( trim(scheme(iScheme)%header%layout) )
            case('d2q9')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 9._rk - 5._rk * phi(iField) )/4._rk - usqr )
            case('d3q19')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 3._rk - 2._rk * phi(iField) ) - usqr )
            end select
          end if  
!write(dbgUnit(1),*) 'iDir ', iDir, 'pdf ', pdfTmp(vPos(iDir)), ' fEq ', feq
!write(dbgUnit(1),*) 'iDir ', iDir, (feq-pdfTmp(vPos(iDir))), omega_new*(feq-pdfTmp(vPos(iDir)))
          outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) = &
            & pdfTmp( vPos(iDir) ) + omega_new * ( feq - pdfTmp( vPos(iDir) ) ) 

!write(dbgUnit(1),*) 'finalPdf ', outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
!          if (tem_isnan( outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ))) then
!            write(*,*) 'found nan', iElem, iField
!            stop
!          endif
!          if (outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) < 0.0_rk) then
!            write(*,*)'Kernel: time ', params%general%simControl%now%iter
!            write(*,*) 'found negative', iElem, iField, &
!              &outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
!            stop
!          endif
        enddo !iDir
          
!      write(*,*) 'outpdf',s, outstate(vPos(:)) 
      enddo !iField
!    stop
    enddo nodeloop

  end subroutine bgk_advRel_MSLiquid
!******************************************************************************

!******************************************************************************
  !> Unoptimized Advection relaxation routine for the multispecies BGK model
  !! with external forcing term
  !!
  !! This routine contains the implementation based on the paper
  !! "A Lattice Boltzmann Scheme for liquid mixtures - Part II: Discretization  
  !! and Numerics, Jens Zudrop, Sabine Roller, Pietro Asinari. "\n
?? copy :: compute_routineHeader( bgk_forcing_advRel_MSLiquid )
    !------------------------------------------------------------------------ 
    !local variables
    real(kind=rk) :: pdfTmp( varSys%nScalars )  !< temporary local pdf values
    integer :: iElem, iField, iField_2, nFields, iDir
    integer :: vPos( layout%stencil(1)%QQ )
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: mass_dens, num_dens
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: moleFrac, massFrac
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: first_moments
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: velocity
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: eqVel
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: molWeight, phi, &
      & chargeNr
    real(kind=rk) :: totNum_dens, totMass_dens, usqr, ucx, feq
    real(kind=rk) :: omega, omega_new, paramB, minMolWeight
    real(kind=rk), &
      & dimension(scheme(iScheme)%nFields,scheme(iScheme)%nFields) :: matA, &
      & invA, resi_coeff
    type(mus_mixture_type) :: mixture
    real(kind=rk), dimension(3, scheme(iScheme)%nFields ) :: mixForce,       &
      & diffForce
    real(kind=rk) :: totMixForce(3), diff_forceTerm, mix_forceTerm
    !real(kind=rk), dimension(nEnd-nStart):: forceState
    real(kind=rk) :: velAvg(3), velNew(3), ucxQuadTerm, theta_eq
    !------------------------------------------------------------------------ 
    ! initialize forceState
    !forceState = -1.0_rk

    nFields = scheme(iScheme)%nFields
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    minMolWeight = minval(molWeight)
    !specific charge of ionic species
    chargeNr = fieldProp(:)%species%chargeNr
    !molecular weight ratios
    phi = fieldProp(:)%species%molWeigRatio
    !resistivities
    do iField = 1, nFields
      resi_coeff(iField,:) = fieldProp( iField )%species%resi_coeff
    enddo
    
    mixture = scheme(iScheme)%mixture
    omega = mixture%relaxLvl(Level)%omega_diff
    omega_new = (1.0_rk/(1.0_rk/omega + 0.5_rk))

    paramB = mixture%paramB
    !equilibrium theta
    theta_eq = mixture%theta_eq

    ! loop over elements
    nodeloop: do iElem = nStart, nEnd

      mass_dens = 0._rk
      first_moments = 0.0_rk
      velocity = 0.0_rk
      eqVel = 0.0_rk
      pdfTmp = 0._rk

      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        ! compute field density and first moments
        do iDir = 1, layout%stencil(1)%QQ
          !store all field pdf in single element to local array pdfTmp
          pdfTmp( vPos(iDir) ) =                                               &
            & instate( ?FETCH?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
          !field density
          mass_dens( iField ) = mass_dens( iField ) + pdfTmp( vPos(iDir) )

          !field momentum (rho*u)
          first_moments( 1, iField ) = first_moments( 1, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 1, iDir )

          first_moments( 2, iField ) = first_moments( 2, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 2, iDir )

          first_moments( 3, iField ) = first_moments( 3, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 3, iDir )
        enddo
      enddo

      !total mass density
      totmass_Dens = sum(mass_dens)

      !mass fraction
      massFrac(:) = mass_dens(:)/totMass_dens

      ! solve linear system of equation for actual moments
      ! number density of all species
      num_dens(:) = mass_dens(:)/molWeight(:)

      !total number density
      totNum_dens = sum(num_dens(:))

      !mole fraction
      moleFrac(:) =  num_dens(:)/totNum_dens

      ! compute external forcing term
      ! d^m_k = w_m*c_m*( n0*min_a(m_a)*F_k + (1/cs2)*(rho_k*g + n_k*z_k*F*E) )

      ! forcing term in the mixture for momentum equation
      ! mixForce = 1/cs2 * rho_k * g + n_k * z_k * F * E
      mixforce = 0.0_rk
      do iField = 1, nFields
        mixforce(:,iField) = cs2inv *                                          &
          &                ( mass_dens(iField) * mixture%gravityField(:)       &
          &                + num_dens(iField) * chargeNr(iField)               &
          &                * mixture%faradayLB * mixture%electricField(:) )
      end do
 
      ! diffusive forcing term in the Maxwell-Stefan equations
      ! F_k = 1/(nRT) rho_k (g - z_k Faraday/m_k E) 
      !     - y_k/(nRT) sum_l(rho_l (g - z_l Faraday/m_l E)
 
      ! diffusive force on each species
      ! diffForce = rho_k (g - z_k Faraday/m_k E)/
      diffForce = 0.0_rk
      do iField = 1, nFields
        diffForce(:, iField) = mass_dens(iField) * ( mixture%gravityField(:)    &
          &                  + chargeNr(iField) * mixture%faradayLB             &
          &                  * mixture%electricField(:) / molWeight(iField) )
      end do

      ! total mixture force in diffusive forcing term
      ! totMixForce = sum_l(rho_l (g - z_l Faraday/m_l E)
      totMixForce = 0.0_rk
      do iField = 1, nFields
        totMixForce(:) = totMixForce(:) + diffForce(:, iField)
      end do
 
      ! update diffForce with fraction of totMixForce and 
      ! constant term 1/(nRT) * n0 * min_a(m_a)
      do iField = 1, nFields
        !diffForce(:, iField) = mixture%moleDens0LB * minMolWeight              &
        diffForce(:, iField) = totNum_dens * minMolWeight                      &
          &                  * ( diffForce(:, iField)                          &
          &                  - massFrac(iField)*totMixForce(:) )               &
          &                  / (totNum_dens *mixture%gasConst_R_LB             &
          &                  * mixture%temp0LB )
      end do
      
      
      matA = 0.0_rk
      !build up matrix to solver LSE for actual velocity
      do iField = 1, nFields
        !set diagonal part 
        matA(iField, iField) = 1.0_rk
        do iField_2 = 1, nFields
          matA(iField, iField) = matA(iField, iField) + omega * 0.5_rk         &
            &                  * resi_coeff(iField, iField_2) * phi(iField)    &
            &                  * moleFrac(iField_2) / paramB
        end do
        !set non-diagonal part
        do iField_2 = 1, nFields
          matA(iField, iField_2) = matA(iField, iField_2) - omega * 0.5_rk     &
            &                    * resi_coeff(iField, iField_2) * phi(iField_2)&
            &                    * moleFrac(iField) / paramB
        end do
      end do

      ! invert matrix
      invA = invert_matrix( matA )

      ! actual velocity of all species
      velocity(1, :) = matmul( invA, first_moments(1,:) ) / mass_dens(:)
      velocity(2, :) = matmul( invA, first_moments(2,:) ) / mass_dens(:)
      velocity(3, :) = matmul( invA, first_moments(3,:) ) / mass_dens(:)

      ! compute equilibrim velocity
      do iField = 1, nFields
        eqVel( :, iField ) = velocity( :, iField )
        do iField_2 = 1, nFields
          eqVel( :, iField ) = eqVel( :, iField )                              &
            &                + resi_coeff( iField, iField_2 ) * phi(iField)    &
            &                * moleFrac(iField_2)                              &
            &                * (velocity(:, iField_2) - velocity(:,iField))    &
            &                / paramB 
        end do  
      end do

      !compute mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens, velocity(1,:) )/totmass_Dens
      velAvg(2) = dot_product( mass_dens, velocity(2,:) )/totmass_Dens
      velAvg(3) = dot_product( mass_dens, velocity(3,:) )/totmass_Dens

      !compute equilibrium and do collision
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        !KM: TGV testcase
        velNew = theta_eq*velAvg + (1.0_rk - theta_eq)*eqVel(:, iField)
        !usqr = eqVel(1, iField) * eqVel(1, iField) &
        !  &  + eqVel(2, iField) * eqVel(2, iField) &
        !  &  + eqVel(3, iField) * eqVel(3, iField) 
        usqr = dot_product( velNew, velNew ) 
        usqr = usqr * t2cs2inv
        do iDir = 1, layout%stencil(1)%QQ
          !forcing terms
          diff_forceTerm =  &
            &   dble( layout%stencil(1)%cxDir( 1, iDir ))*diffForce(1, iField) &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*diffForce(2, iField) &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*diffForce(3, iField)
          mix_forceTerm =  &
            &   dble( layout%stencil(1)%cxDir( 1, iDir ))*mixForce(1, iField)  &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*mixForce(2, iField)  &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*mixForce(3, iField)

          ucx = dble( layout%stencil(1)%cxDir( 1, iDir ))*eqVel(1, iField)     &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*eqVel(2, iField)     &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*eqVel(3, iField)

          ucxQuadTerm = dble( layout%stencil(1)%cxDir( 1, iDir ))*velNew(1)    &
            &         + dble( layout%stencil(1)%cxDir( 2, iDir ))*velNew(2)    &
            &         + dble( layout%stencil(1)%cxDir( 3, iDir ))*velNew(3)

          feq = layout%weight(iDir) * mass_dens(iField) * ( phi(iField)        &
            & + ucx * cs2inv + ucxQuadTerm * ucxQuadTerm * t2cs4inv - usqr )
          
          if ( iDir == layout%stencil(1)%restPosition ) then      
          ! equilibrium at rest
            select case( trim(scheme(iScheme)%header%layout) )
            case('d2q9')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 9._rk - 5._rk * phi(iField) )/4._rk - usqr )
            case('d3q19')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 3._rk - 2._rk * phi(iField) ) - usqr )
            end select
          end if  

          outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) = &
            & pdfTmp( vPos(iDir) ) + omega_new * ( feq - pdfTmp( vPos(iDir) ) ) &
            ! forcing term
            & + layout%weight( iDir ) * ( diff_forceTerm + mix_forceTerm )
        enddo !iDir
          
      enddo !iField
 
    enddo nodeloop

  end subroutine bgk_forcing_advRel_MSLiquid
!******************************************************************************

!******************************************************************************
  !> Unoptimized Advection relaxation routine for the multispecies BGK model 
  !!
  !! This routine contains the implementation based on the paper
  !! "A Lattice Boltzmann Scheme for liquid mixtures - Part II: Discretization  
  !! and Numerics, Jens Zudrop, Sabine Roller, Pietro Asinari. "\n
  !! MRT paper
  !! "Lattice Boltzmann liquid mixture modeling for electrodialytic engineering
  !!  applications - J. Zudrop, K. Masilamani, S. Roller and P. Asinari"\n
?? copy :: compute_routineHeader( mrt_advRel_MSLiquid )
    !------------------------------------------------------------------------ 
    !local variables
    real(kind=rk) :: pdfTmp( varSys%nScalars )  !< temporary local pdf values
    integer :: iElem, iField, iField_2, nFields, iDir
    integer :: vPos( layout%stencil(1)%QQ )
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: mass_dens, num_dens
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: moleFrac
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: first_moments
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: velocity
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: eqVel
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: molWeight, phi 
    real(kind=rk) :: totNum_dens, usqr, ucx, feq
    real(kind=rk) :: omega, omega_new, paramB, theta_eq
    real(kind=rk) :: velAvg(3), velNew(3), ucxQuadTerm, totMassDens
    real(kind=rk) :: omegaMoments(layout%stencil(1)%QQ,layout%stencil(1)%QQ)
    real(kind=rk) :: fneq(layout%stencil(1)%QQ), fneq_om(layout%stencil(1)%QQ)
    real(kind=rk), &
      & dimension(scheme(iScheme)%nFields,scheme(iScheme)%nFields) :: matA, &
      & invA, resi_coeff
    !------------------------------------------------------------------------ 
    nFields = scheme(iScheme)%nFields
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    !molecular weight ratios
    phi = fieldProp(:)%species%molWeigRatio
    !resistivities
    do iField = 1, nFields
      resi_coeff(iField,:) = fieldProp( iField )%species%resi_coeff
    enddo

    omega = scheme(iScheme)%mixture%relaxLvl(level)%omega_diff
!    write(*,*) 'omega ', omega
    omega_new = (1.0_rk/(1.0_rk/omega + 0.5_rk))
!    write(*,*) 'omega_new ', omega_new

    paramB = scheme(iScheme)%mixture%paramB
    !equilibrium theta
    theta_eq = scheme(iScheme)%mixture%theta_eq
!    write(*,*) 'paramB ', paramB

!    write(dbgUnit,*) 'B 1', resi_coeff(1,:)
!    write(dbgUnit,*) 'B 2', resi_coeff(2,:)
!    write(dbgUnit,*) 'B 3', resi_coeff(3,:)

    ! KM: debugging multispecies
    !outstate = instate

!write(dbgUnit(1),*)'time ', params%general%simControl%now%iter
    nodeloop: do iElem = nStart, nEnd
!write(dbgUnit(1),*) 'iElem', iElem

      mass_dens = 0._rk
      first_moments = 0.0_rk
      velocity = 0.0_rk
      eqVel = 0.0_rk
      pdfTmp = 0._rk

      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        ! compute field density and first moments
        do iDir = 1, layout%stencil(1)%QQ
          !store all field pdf in single element to local array pdfTmp
          pdfTmp( vPos(iDir) ) =                                               &
            & instate( ?FETCH?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
          !field density
          mass_dens( iField ) = mass_dens( iField ) + pdfTmp( vPos(iDir) )

          !field momentum (rho*u)
          first_moments( 1, iField ) = first_moments( 1, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 1, iDir )

          first_moments( 2, iField ) = first_moments( 2, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 2, iDir )

          first_moments( 3, iField ) = first_moments( 3, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 3, iDir )
!          write(dbgunit,*) 's',s,'iDir', iDir, layout%stencil(1)%cxDir(3,iDir), pdfTmp( vPos(iDir) ), gqzsigma(s)
        enddo
      enddo

      !total mass density
      totmassDens = sum(mass_dens)

      ! solve linear system of equation for actual moments
      ! number density of all species
      num_dens(:) = mass_dens(:)/molWeight(:)

      !total number density
      totNum_dens = sum(num_dens(:))

      !mole fraction
      moleFrac(:) =  num_dens(:)/totNum_dens

      matA = 0.0_rk
      !build up matrix to solver LSE for actual velocity
      do iField = 1, nFields
        !set diagonal part 
        matA(iField, iField) = 1.0_rk
        do iField_2 = 1, nFields
          matA(iField, iField) = matA(iField, iField) + omega * 0.5_rk         &
            &                  * resi_coeff(iField, iField_2) * phi(iField)    &
            &                  * moleFrac(iField_2) / paramB
        end do
        !set non-diagonal part
        do iField_2 = 1, nFields
          matA(iField, iField_2) = matA(iField, iField_2) - omega * 0.5_rk     &
            &                    * resi_coeff(iField, iField_2) * phi(iField_2)&
            &                    * moleFrac(iField) / paramB
        end do
      end do

      ! invert matrix
      invA = invert_matrix( matA )

      ! actual velocity of all species
      velocity(1, :) = matmul( invA, first_moments(1,:) ) / mass_dens(:)
      velocity(2, :) = matmul( invA, first_moments(2,:) ) / mass_dens(:)
      velocity(3, :) = matmul( invA, first_moments(3,:) ) / mass_dens(:)

      ! compute equilibrim velocity
      do iField = 1, nFields
        eqVel( :, iField ) = velocity( :, iField )
        do iField_2 = 1, nFields
          eqVel( :, iField ) = eqVel( :, iField )                              &
            &                + resi_coeff( iField, iField_2 ) * phi(iField)    &
            &                * moleFrac(iField_2)                              &
            &                * (velocity(:, iField_2) - velocity(:,iField))    &
            &                / paramB 
        end do  
      end do

      !compute mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens, velocity(1,:) )/totmassDens
      velAvg(2) = dot_product( mass_dens, velocity(2,:) )/totmassDens
      velAvg(3) = dot_product( mass_dens, velocity(3,:) )/totmassDens

      !compute equilibrium and do collision
      do iField = 1, nFields
        !omega moments = (M^-1 RelaxMat M)*(I+(M^-1 RelMat M)/2)^-1
        omegaMoments = fieldProp( iField )%species%mrt( level )%omegaMoments
        vPos = varSys%Variable%val( iField )%varPos

        !KM: TGV testcase
        velNew = theta_eq*velAvg + (1.0_rk - theta_eq)*eqVel(:, iField)

        !usqr = eqVel(1, iField) * eqVel(1, iField) &
        !  &  + eqVel(2, iField) * eqVel(2, iField) &
        !  &  + eqVel(3, iField) * eqVel(3, iField) 
        usqr = dot_product( velNew, velNew ) 
        usqr = usqr * t2cs2inv

        do iDir = 1, layout%stencil(1)%QQ
          ucx = dble( layout%stencil(1)%cxDir( 1, iDir ))*eqVel(1, iField)     &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*eqVel(2, iField)     &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*eqVel(3, iField)

          ucxQuadTerm = dot_product(                                           &
            &           dble(layout%stencil(1)%cxDir(:, iDir)), velNew )

          feq = layout%weight(iDir) * mass_dens(iField) * ( phi(iField)        &
            & + ucx * cs2inv + ucxQuadTerm * ucxQuadTerm * t2cs4inv - usqr )
          
          if ( iDir == layout%stencil(1)%restPosition ) then      
          ! equilibrium at rest
            select case( trim(scheme(iScheme)%header%layout) )
            case('d2q9')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 9._rk - 5._rk * phi(iField) )/4._rk - usqr )
            case('d3q19')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 3._rk - 2._rk * phi(iField) ) - usqr )
            end select
          end if  
          fneq(iDir) = ( feq - pdfTmp( vPos(iDir) ) )
        enddo !iDir
!write(dbgUnit(1),*) 'omegaMoments ', omegaMoments
!write(dbgUnit(1),*) 'before fnEq ', fnEq
        fneq_om = matmul( omegaMoments, fneq )  
!do iDir = 1, layout%stencil(1)%QQ
!  write(dbgUnit(1),*) 'iDir ', iDir, fnEq(iDir), fnEq_om(iDir)
!end do  

        do iDir = 1, layout%stencil(1)%QQ
          outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) = &
            & pdfTmp( vPos(iDir) ) + fneq_om( iDir )
!          if (tem_isnan( outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ))) then
!            write(*,*) 'found nan', iElem, iField
!            stop
!          endif
        enddo !iDir
          
!      write(*,*) 'outpdf',s, outstate(vPos(:)) 
      enddo !iField
!    stop
    enddo nodeloop

  end subroutine mrt_advRel_MSLiquid
!******************************************************************************


!******************************************************************************
  !> Otimized Advection relaxation routine for the multispecies mrt model 
  !! for d3q19 with 3 species
  !!
  !! This routine contains the implementation based on the paper
  !! "A Lattice Boltzmann Scheme for liquid mixtures - Part II: Discretization  
  !! and Numerics, Jens Zudrop, Sabine Roller, Pietro Asinari. "\n
  !! MRT paper
  !! "Lattice Boltzmann liquid mixture modeling for electrodialytic engineering
  !!  applications - J. Zudrop, K. Masilamani, S. Roller and P. Asinari"\n
?? copy :: compute_routineHeader( mrt_advRel_d3q19f3_MSLiquid )
    !------------------------------------------------------------------------ 
    !local variables
    ! temporary local pdf values
    real(kind=rk) :: pdfTmp( layout%stencil(1)%QQ, 3 ) 
    integer :: iElem!, nFields
    real(kind=rk), dimension(3) :: rsigma, uxsigma,                            &
      & uysigma, uzsigma, qxsigma, qysigma, qzsigma, gqxsigma, gqysigma,       &
      & gqzsigma
    real(kind=rk), dimension(3) :: uxstar, uystar, uzstar
    real(kind=rk), dimension(3) :: molWeight, molWeight_inv, phi,              &
      &                            num_dens, moleFrac
    real(kind=rk) :: usqr, totNum_dens_inv, totMass_dens
    real(kind=rk) :: velAvg_x, velAvg_y, velAvg_z
    real(kind=rk) :: velQuadTerm_x, velQuadTerm_y, velQuadTerm_z
    real(kind=rk) :: omega, omegadiv2, omega_fac, omega_o
    real(kind=rk), dimension(3,3) :: A, mbb
    real(kind=rk) :: zfac, yfac, a12_11, a13_11, &
      & a23fac, a21_11, a31_11, zcoeff
    real(kind=rk) :: theta_eq, theta_eq_spc, paramB_inv, a11_inv, a22fac_inv
    real(kind=rk) :: B_12, B_13, B_23
    real(kind=rk) :: Bratio_12, Bratio_13, Bratio_23
    real(kind=rk) :: mbbEq_12, mbbEq_13, mbbEq_21, mbbEq_23, mbbEq_31, mbbEq_32
    real(kind=rk) :: chi12, chi13, chi21, chi23, chi31, chi32
    real(kind=rk) :: rho_d, rho_dm, rho_o, rho_om, sum1_1, sum2_1, sum3_1,     &
      & sum4_1, sum5_1, sum6_1, sum7_1, sum8_1, sum9_1, sum1_2, sum2_2, sum3_2,&
      & sum4_2, sum5_2, sum6_2, sum7_2, sum8_2, sum9_2
    real(kind=rk) :: omegaMoments(layout%stencil(1)%QQ,layout%stencil(1)%QQ,3)
    real(kind=rk) :: fneq(layout%stencil(1)%QQ), fneq_om(layout%stencil(1)%QQ)
    integer :: iField
    !--------------------------------------------------------------------------- 
    !nFields = scheme(iScheme)%nFields
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    molWeight_inv = 1.0_rk/molWeight
    !molecular weight ratios
    phi = fieldProp(:)%species%molWeigRatio

    !omega moments = (M^-1 RelaxMat M)*(I+(M^-1 RelMat M)/2)^-1
    omegaMoments(:,:,1) = fieldProp(1)%species%mrt( level )%omegaMoments
    omegaMoments(:,:,2) = fieldProp(2)%species%mrt( level )%omegaMoments
    omegaMoments(:,:,3) = fieldProp(3)%species%mrt( level )%omegaMoments

    !omega
    omega = scheme(iScheme)%mixture%relaxLvl(Level)%omega_diff
    omega_fac = (1.0_rk/(1.0_rk/omega + 0.5_rk))
    omegadiv2 = omega*0.5_rk
    omega_o = 1.0_rk - omega_fac

    !free parameter B
    paramB_inv = 1.0_rk/scheme(iScheme)%mixture%paramB

    !equilibrium theta
    theta_eq = scheme(iScheme)%mixture%theta_eq
    theta_eq_spc = 1.0_rk - theta_eq

    !resistivities
    !resistivity coeff are symmetric
    !B_12 = B_21, B_13=B31, B23=B32
    B_12 = fieldProp( 1 )%species%resi_coeff(2)
    B_13 = fieldProp( 1 )%species%resi_coeff(3)
    B_23 = fieldProp( 2 )%species%resi_coeff(3)

    Bratio_12 = B_12*paramB_inv
    Bratio_13 = B_13*paramB_inv
    Bratio_23 = B_23*paramB_inv
    mbbEq_12 = Bratio_12*phi(1)
    mbbEq_13 = Bratio_13*phi(1)

    mbbEq_21 = Bratio_12*phi(2)
    mbbEq_23 = Bratio_23*phi(2)

    mbbEq_31 = Bratio_13*phi(3)
    mbbEq_32 = Bratio_23*phi(3)

    ! factor need to build matrixA to solve LSE
    ! dummy diagonal entries 
    mbb(1,1) = -1.0_rk
    mbb(1,2) = mbbEq_21*omegadiv2
    mbb(1,3) = mbbEq_31*omegadiv2

    mbb(2,1) = mbbEq_12*omegadiv2
    mbb(2,2) = -1.0_rk
    mbb(2,3) = mbbEq_32*omegadiv2

    mbb(3,1) = mbbEq_13*omegadiv2
    mbb(3,2) = mbbEq_23*omegadiv2
    mbb(3,3) = -1.0_rk

    nodeloop: do iElem = nStart, nEnd

      !species 1
      pdfTmp( Q__W,1 ) = instate(?FETCH?( varPos(Q__W,1), iElem, QQF3, neigh ))
      pdfTmp( Q__S,1 ) = instate(?FETCH?( varPos(Q__S,1), iElem, QQF3, neigh ))
      pdfTmp( Q__B,1 ) = instate(?FETCH?( varPos(Q__B,1), iElem, QQF3, neigh ))
      pdfTmp( Q__E,1 ) = instate(?FETCH?( varPos(Q__E,1), iElem, QQF3, neigh ))
      pdfTmp( Q__N,1 ) = instate(?FETCH?( varPos(Q__N,1), iElem, QQF3, neigh ))
      pdfTmp( Q__T,1 ) = instate(?FETCH?( varPos(Q__T,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BS,1 ) = instate(?FETCH?( varPos(Q_BS,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TS,1 ) = instate(?FETCH?( varPos(Q_TS,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BN,1 ) = instate(?FETCH?( varPos(Q_BN,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TN,1 ) = instate(?FETCH?( varPos(Q_TN,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BW,1 ) = instate(?FETCH?( varPos(Q_BW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_BE,1 ) = instate(?FETCH?( varPos(Q_BE,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TW,1 ) = instate(?FETCH?( varPos(Q_TW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_TE,1 ) = instate(?FETCH?( varPos(Q_TE,1), iElem, QQF3, neigh ))
      pdfTmp( Q_SW,1 ) = instate(?FETCH?( varPos(Q_SW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_NW,1 ) = instate(?FETCH?( varPos(Q_NW,1), iElem, QQF3, neigh ))
      pdfTmp( Q_SE,1 ) = instate(?FETCH?( varPos(Q_SE,1), iElem, QQF3, neigh ))
      pdfTmp( Q_NE,1 ) = instate(?FETCH?( varPos(Q_NE,1), iElem, QQF3, neigh ))
      pdfTmp( Q__0,1 ) = instate(?FETCH?( varPos(Q__0,1), iElem, QQF3, neigh ))
      !species 2
      pdfTmp( Q__W,2 ) = instate(?FETCH?( varPos(Q__W,2), iElem, QQF3, neigh ))
      pdfTmp( Q__S,2 ) = instate(?FETCH?( varPos(Q__S,2), iElem, QQF3, neigh ))
      pdfTmp( Q__B,2 ) = instate(?FETCH?( varPos(Q__B,2), iElem, QQF3, neigh ))
      pdfTmp( Q__E,2 ) = instate(?FETCH?( varPos(Q__E,2), iElem, QQF3, neigh ))
      pdfTmp( Q__N,2 ) = instate(?FETCH?( varPos(Q__N,2), iElem, QQF3, neigh ))
      pdfTmp( Q__T,2 ) = instate(?FETCH?( varPos(Q__T,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BS,2 ) = instate(?FETCH?( varPos(Q_BS,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TS,2 ) = instate(?FETCH?( varPos(Q_TS,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BN,2 ) = instate(?FETCH?( varPos(Q_BN,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TN,2 ) = instate(?FETCH?( varPos(Q_TN,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BW,2 ) = instate(?FETCH?( varPos(Q_BW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_BE,2 ) = instate(?FETCH?( varPos(Q_BE,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TW,2 ) = instate(?FETCH?( varPos(Q_TW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_TE,2 ) = instate(?FETCH?( varPos(Q_TE,2), iElem, QQF3, neigh ))
      pdfTmp( Q_SW,2 ) = instate(?FETCH?( varPos(Q_SW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_NW,2 ) = instate(?FETCH?( varPos(Q_NW,2), iElem, QQF3, neigh ))
      pdfTmp( Q_SE,2 ) = instate(?FETCH?( varPos(Q_SE,2), iElem, QQF3, neigh ))
      pdfTmp( Q_NE,2 ) = instate(?FETCH?( varPos(Q_NE,2), iElem, QQF3, neigh ))
      pdfTmp( Q__0,2 ) = instate(?FETCH?( varPos(Q__0,2), iElem, QQF3, neigh ))
      !species 3
      pdfTmp( Q__W,3 ) = instate(?FETCH?( varPos(Q__W,3), iElem, QQF3, neigh ))
      pdfTmp( Q__S,3 ) = instate(?FETCH?( varPos(Q__S,3), iElem, QQF3, neigh ))
      pdfTmp( Q__B,3 ) = instate(?FETCH?( varPos(Q__B,3), iElem, QQF3, neigh ))
      pdfTmp( Q__E,3 ) = instate(?FETCH?( varPos(Q__E,3), iElem, QQF3, neigh ))
      pdfTmp( Q__N,3 ) = instate(?FETCH?( varPos(Q__N,3), iElem, QQF3, neigh ))
      pdfTmp( Q__T,3 ) = instate(?FETCH?( varPos(Q__T,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BS,3 ) = instate(?FETCH?( varPos(Q_BS,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TS,3 ) = instate(?FETCH?( varPos(Q_TS,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BN,3 ) = instate(?FETCH?( varPos(Q_BN,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TN,3 ) = instate(?FETCH?( varPos(Q_TN,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BW,3 ) = instate(?FETCH?( varPos(Q_BW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_BE,3 ) = instate(?FETCH?( varPos(Q_BE,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TW,3 ) = instate(?FETCH?( varPos(Q_TW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_TE,3 ) = instate(?FETCH?( varPos(Q_TE,3), iElem, QQF3, neigh ))
      pdfTmp( Q_SW,3 ) = instate(?FETCH?( varPos(Q_SW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_NW,3 ) = instate(?FETCH?( varPos(Q_NW,3), iElem, QQF3, neigh ))
      pdfTmp( Q_SE,3 ) = instate(?FETCH?( varPos(Q_SE,3), iElem, QQF3, neigh ))
      pdfTmp( Q_NE,3 ) = instate(?FETCH?( varPos(Q_NE,3), iElem, QQF3, neigh ))
      pdfTmp( Q__0,3 ) = instate(?FETCH?( varPos(Q__0,3), iElem, QQF3, neigh ))

      !density 1
      rsigma(1) = pdfTmp(Q_NE,1) + pdfTmp(Q__N,1) + pdfTmp(Q_NW,1)             &
        &       + pdfTmp(Q__W,1) + pdfTmp(Q_SW,1) + pdfTmp(Q__S,1)             &
        &       + pdfTmp(Q_SE,1) + pdfTmp(Q__E,1) + pdfTmp(Q__T,1)             &
        &       + pdfTmp(Q_TE,1) + pdfTmp(Q_TN,1) + pdfTmp(Q_TW,1)             &
        &       + pdfTmp(Q_TS,1) + pdfTmp(Q__B,1) + pdfTmp(Q_BE,1)             &
        &       + pdfTmp(Q_BN,1) + pdfTmp(Q_BW,1) + pdfTmp(Q_BS,1)             &
        &       + pdfTmp(Q__0,1)

      !density 2
      rsigma(2) = pdfTmp(Q_NE,2) + pdfTmp(Q__N,2) + pdfTmp(Q_NW,2)             &
        &       + pdfTmp(Q__W,2) + pdfTmp(Q_SW,2) + pdfTmp(Q__S,2)             &
        &       + pdfTmp(Q_SE,2) + pdfTmp(Q__E,2) + pdfTmp(Q__T,2)             &
        &       + pdfTmp(Q_TE,2) + pdfTmp(Q_TN,2) + pdfTmp(Q_TW,2)             &
        &       + pdfTmp(Q_TS,2) + pdfTmp(Q__B,2) + pdfTmp(Q_BE,2)             &
        &       + pdfTmp(Q_BN,2) + pdfTmp(Q_BW,2) + pdfTmp(Q_BS,2)             &
        &       + pdfTmp(Q__0,2)

      !density 3
      rsigma(3) = pdfTmp(Q_NE,3) + pdfTmp(Q__N,3) + pdfTmp(Q_NW,3)             &
        &       + pdfTmp(Q__W,3) + pdfTmp(Q_SW,3) + pdfTmp(Q__S,3)             &
        &       + pdfTmp(Q_SE,3) + pdfTmp(Q__E,3) + pdfTmp(Q__T,3)             &
        &       + pdfTmp(Q_TE,3) + pdfTmp(Q_TN,3) + pdfTmp(Q_TW,3)             &
        &       + pdfTmp(Q_TS,3) + pdfTmp(Q__B,3) + pdfTmp(Q_BE,3)             &
        &       + pdfTmp(Q_BN,3) + pdfTmp(Q_BW,3) + pdfTmp(Q_BS,3)             &
        &       + pdfTmp(Q__0,3)

 
      !local x,y,z - momentum of species 1
      gqxsigma(1) = pdfTmp(Q_NE,1) + pdfTmp(Q_SE,1) + pdfTmp(Q__E,1)           &
        &         + pdfTmp(Q_TE,1) + pdfTmp(Q_BE,1)                            &
        &         - pdfTmp(Q_NW,1) - pdfTmp(Q__W,1) - pdfTmp(Q_SW,1)           &
        &         - pdfTmp(Q_TW,1) - pdfTmp(Q_BW,1)
                         
      gqysigma(1) = pdfTmp(Q_NE,1) + pdfTmp(Q__N,1) + pdfTmp(Q_NW,1)           &
        &         + pdfTmp(Q_TN,1) + pdfTmp(Q_BN,1)                            &
        &         - pdfTmp(Q_SW,1) - pdfTmp(Q__S,1) - pdfTmp(Q_SE,1)           &
        &         - pdfTmp(Q_TS,1) - pdfTmp(Q_BS,1)
                         
      gqzsigma(1) = pdfTmp(Q__T,1) + pdfTmp(Q_TE,1) + pdfTmp(Q_TN,1)           &
        &         + pdfTmp(Q_TW,1) + pdfTmp(Q_TS,1)                            & 
        &         - pdfTmp(Q__B,1) - pdfTmp(Q_BE,1) - pdfTmp(Q_BN,1)           &
        &         - pdfTmp(Q_BW,1) - pdfTmp(Q_BS,1)

      !local x,y,z - momentum of species 2
      gqxsigma(2) = pdfTmp(Q_NE,2) + pdfTmp(Q_SE,2) + pdfTmp(Q__E,2)           &
        &         + pdfTmp(Q_TE,2) + pdfTmp(Q_BE,2)                            &
        &         - pdfTmp(Q_NW,2) - pdfTmp(Q__W,2) - pdfTmp(Q_SW,2)           &
        &         - pdfTmp(Q_TW,2) - pdfTmp(Q_BW,2)
                         
      gqysigma(2) = pdfTmp(Q_NE,2) + pdfTmp(Q__N,2) + pdfTmp(Q_NW,2)           &
        &         + pdfTmp(Q_TN,2) + pdfTmp(Q_BN,2)                            &
        &         - pdfTmp(Q_SW,2) - pdfTmp(Q__S,2) - pdfTmp(Q_SE,2)           &
        &         - pdfTmp(Q_TS,2) - pdfTmp(Q_BS,2)
                         
      gqzsigma(2) = pdfTmp(Q__T,2) + pdfTmp(Q_TE,2) + pdfTmp(Q_TN,2)           &
        &         + pdfTmp(Q_TW,2) + pdfTmp(Q_TS,2)                            & 
        &         - pdfTmp(Q__B,2) - pdfTmp(Q_BE,2) - pdfTmp(Q_BN,2)           &
        &         - pdfTmp(Q_BW,2) - pdfTmp(Q_BS,2)

      !local x,y,z - momentum of species 3
      gqxsigma(3) = pdfTmp(Q_NE,3) + pdfTmp(Q_SE,3) + pdfTmp(Q__E,3)           &
        &         + pdfTmp(Q_TE,3) + pdfTmp(Q_BE,3)                            &
        &         - pdfTmp(Q_NW,3) - pdfTmp(Q__W,3) - pdfTmp(Q_SW,3)           &
        &         - pdfTmp(Q_TW,3) - pdfTmp(Q_BW,3)
                         
      gqysigma(3) = pdfTmp(Q_NE,3) + pdfTmp(Q__N,3) + pdfTmp(Q_NW,3)           &
        &         + pdfTmp(Q_TN,3) + pdfTmp(Q_BN,3)                            &
        &         - pdfTmp(Q_SW,3) - pdfTmp(Q__S,3) - pdfTmp(Q_SE,3)           &
        &         - pdfTmp(Q_TS,3) - pdfTmp(Q_BS,3)
                         
      gqzsigma(3) = pdfTmp(Q__T,3) + pdfTmp(Q_TE,3) + pdfTmp(Q_TN,3)           &
        &         + pdfTmp(Q_TW,3) + pdfTmp(Q_TS,3)                            & 
        &         - pdfTmp(Q__B,3) - pdfTmp(Q_BE,3) - pdfTmp(Q_BN,3)           &
        &         - pdfTmp(Q_BW,3) - pdfTmp(Q_BS,3)


      !number density
      num_dens(:) = rsigma(:)*molWeight_inv(:)

      !total number density
      totNum_dens_inv = 1.0_rk/sum(num_dens(:))

      !mole fraction
      moleFrac(:) = num_dens(:)*totNum_dens_inv

      !matrix A of the velocity lse
      A(1,1) = 1.0_rk + moleFrac(2)*mbb(2,1) + moleFrac(3)*mbb(3,1) 
      A(1,2) = - moleFrac(1)*mbb(1,2)
      A(1,3) = - moleFrac(1)*mbb(1,3)

      A(2,1) = - moleFrac(2)*mbb(2,1)
      A(2,2) = 1.0_rk + moleFrac(1)*mbb(1,2) + moleFrac(3)*mbb(3,2)
      A(2,3) = - moleFrac(2)*mbb(2,3)

      A(3,1) = - moleFrac(3)*mbb(3,1)
      A(3,2) = - moleFrac(3)*mbb(3,2)
      A(3,3) = 1.0_rk + moleFrac(1)*mbb(1,3) + moleFrac(2)*mbb(2,3)

      a11_inv = 1.0_rk/A(1,1)
      a13_11 = A(1,3)*a11_inv
      a12_11 = A(1,2)*a11_inv
      a31_11 = A(3,1)*a11_inv
      a21_11 = A(2,1)*a11_inv
      a23fac = A(2,3) - A(2,1)*a13_11
      a22fac_inv = 1.0_rk/(A(2,2) - A(2,1)*a12_11)
  
      zfac = (A(3,1)*a12_11 - A(3,2)) * a22fac_inv
      zcoeff = 1._rk / (a23fac*zfac - A(3,1)*a13_11 + A(3,3))
      !x-momentum
      !species 3
      yfac = gqxsigma(2) - gqxsigma(1)*a21_11

      qxsigma(3) = (gqxsigma(3) - a31_11*gqxsigma(1) + yfac*zfac)*zcoeff

      !species 2
      qxsigma(2) = (yfac - a23fac*qxsigma(3))*a22fac_inv
      !species 1
      qxsigma(1) = (gqxsigma(1) - A(1,2)*qxsigma(2) - A(1,3)*qxsigma(3))*a11_inv
      !y-momentum
      yfac = gqysigma(2) - gqysigma(1)*a21_11
      qysigma(3) = (gqysigma(3) - a31_11*gqysigma(1) + yfac*zfac)*zcoeff

      qysigma(2) = (yfac - a23fac*qysigma(3))*a22fac_inv

      qysigma(1) = (gqysigma(1) - A(1,2)*qysigma(2) - A(1,3)*qysigma(3))*a11_inv
      !z-momentum
      yfac = gqzsigma(2) - gqzsigma(1)*a21_11
      qzsigma(3) = (gqzsigma(3) - a31_11*gqzsigma(1) + yfac*zfac)*zcoeff
      
      qzsigma(2) = (yfac - a23fac*qzsigma(3))*a22fac_inv

      qzsigma(1) = (gqzsigma(1) - A(1,2)*qzsigma(2) - A(1,3)*qzsigma(3))*a11_inv

      ! species velocity
      uxsigma = qxsigma/rsigma
      uysigma = qysigma/rsigma
      uzsigma = qzsigma/rsigma

      !compute equilibrium velocity to compute feq
      chi12 = mbbEq_12 * moleFrac(2)
      chi13 = mbbEq_13 * moleFrac(3)
      chi21 = mbbEq_21 * moleFrac(1) 
      chi23 = mbbEq_23 * moleFrac(3)
      chi31 = mbbEq_31 * moleFrac(1)
      chi32 = mbbEq_32 * moleFrac(2)
      !ux
      uxstar(1) =  uxsigma(1) + chi12*(uxsigma(2)-uxsigma(1))                  &
                             &+ chi13*(uxsigma(3)-uxsigma(1))

      uxstar(2) =  uxsigma(2) + chi21*(uxsigma(1)-uxsigma(2))                  &
                             &+ chi23*(uxsigma(3)-uxsigma(2))

      uxstar(3) =  uxsigma(3) + chi31*(uxsigma(1)-uxsigma(3))                  &
                             &+ chi32*(uxsigma(2)-uxsigma(3))
      !uy
      uystar(1) =  uysigma(1) + chi12*(uysigma(2)-uysigma(1))                  &
                             &+ chi13*(uysigma(3)-uysigma(1))

      uystar(2) =  uysigma(2) + chi21*(uysigma(1)-uysigma(2))                  &
                             &+ chi23*(uysigma(3)-uysigma(2))

      uystar(3) =  uysigma(3) + chi31*(uysigma(1)-uysigma(3))                  &
                             &+ chi32*(uysigma(2)-uysigma(3))
      !uz
      uzstar(1) =  uzsigma(1) + chi12*(uzsigma(2)-uzsigma(1))                  &
                             &+ chi13*(uzsigma(3)-uzsigma(1))

      uzstar(2) =  uzsigma(2) + chi21*(uzsigma(1)-uzsigma(2))                  &
                             &+ chi23*(uzsigma(3)-uzsigma(2))

      uzstar(3) =  uzsigma(3) + chi31*(uzsigma(1)-uzsigma(3))                  &
                             &+ chi32*(uzsigma(2)-uzsigma(3))

      ! total mass density
      totMass_dens = sum(rsigma)

      ! mass averaged mixture velocity
      velAvg_x = dot_product( rsigma, uxsigma )/totMass_dens
      velAvg_y = dot_product( rsigma, uysigma )/totMass_dens
      velAvg_z = dot_product( rsigma, uzsigma )/totMass_dens

      !compute equilibrium and do collision
      do iField = 1, 3!nFields
        ! \todo KM: Bug remove omega factor from equilibrium
        rho_d = rsigma(iField) * div1_4

        velQuadTerm_x = theta_eq*velAvg_x + theta_eq_spc*uxstar(iField)
        velQuadTerm_y = theta_eq*velAvg_y + theta_eq_spc*uystar(iField)
        velQuadTerm_z = theta_eq*velAvg_z + theta_eq_spc*uzstar(iField)

        usqr = (velQuadTerm_x * velQuadTerm_x                                  &
          &  + velQuadTerm_y * velQuadTerm_y                                   &
          &  + velQuadTerm_z * velQuadTerm_z) * t2cs2inv
        
        ! at rest
        fneq(Q__0) = div1_3 * rsigma(iField) * (( 3._rk - 2._rk * phi(iField)) &
          &        - usqr ) - pdfTmp(Q__0, iField)

        rho_dm = rsigma(iField) * (phi(iField) - usqr) * div1_18
        !directional velocity factor 
        sum1_1 = rho_d * uxstar(iField) * div3_4h
        sum1_2 = rho_d * velQuadTerm_x**2.0 + rho_dm
        fneq(Q__W) = -sum1_1 + sum1_2 - pdfTmp(Q__W,iField)
        fneq(Q__E) = sum1_1 + sum1_2 - pdfTmp(Q__E,iField)

        sum2_1 = rho_d * uystar(iField) * div3_4h
        sum2_2 = rho_d * velQuadTerm_y**2.0 + rho_dm
        fneq(Q__S) = -sum2_1 + sum2_2 - pdfTmp(Q__S,iField)
        fneq(Q__N) = sum2_1 + sum2_2 - pdfTmp(Q__N,iField)

        sum3_1 = rho_d * uzstar(iField) * div3_4h
        sum3_2 = rho_d * velQuadTerm_z**2.0 + rho_dm
        fneq(Q__B) = -sum3_1 + sum3_2 - pdfTmp(Q__B,iField)
        fneq(Q__T) = sum3_1 + sum3_2 - pdfTmp(Q__T,iField)

        rho_o = rho_d * 0.5_rk
        rho_om = rho_dm * 0.5_rk
        !top north / bottom south        
        sum4_1 = rho_o * ( uystar(iField) + uzstar(iField) ) *div3_4h
        sum4_2 = rho_o * ( velQuadTerm_y + velQuadTerm_z )**2.0 &
        &       + rho_om
        fneq(Q_BS) = -sum4_1 + sum4_2 - pdfTmp(Q_BS,iField)
        fneq(Q_TN) = sum4_1 + sum4_2 - pdfTmp(Q_TN,iField)

        !top south / bottom north
        sum5_1 = rho_o * ( -uystar(iField) + uzstar(iField) ) *div3_4h
        sum5_2 = rho_o * ( -velQuadTerm_y + velQuadTerm_z )**2.0&
          &       + rho_om
        fneq(Q_BN) = -sum5_1 + sum5_2 - pdfTmp(Q_BN,iField)
        fneq(Q_TS) = sum5_1 + sum5_2 - pdfTmp(Q_TS,iField)
        
        !top east / bottom west
        sum6_1 = rho_o * ( uxstar(iField) + uzstar(iField) ) *div3_4h
        sum6_2 = rho_o * ( velQuadTerm_x + velQuadTerm_z )**2.0 &
          &       + rho_om
        fneq(Q_BW) = -sum6_1 + sum6_2 - pdfTmp(Q_BW,iField)
        fneq(Q_TE) = sum6_1 + sum6_2 - pdfTmp(Q_TE,iField)

        !top west / bottom east
        sum7_1 = rho_o * ( -uxstar(iField) + uzstar(iField) ) *div3_4h
        sum7_2 = rho_o * ( -velQuadTerm_x + velQuadTerm_z )**2.0&
          &       + rho_om
        fneq(Q_BE) = -sum7_1 + sum7_2 - pdfTmp(Q_BE,iField)
        fneq(Q_TW) = sum7_1 + sum7_2 - pdfTmp(Q_TW,iField)

        !north east / south west
        sum8_1 = rho_o * ( uxstar(iField) + uystar(iField) ) *div3_4h
        sum8_2 = rho_o * ( velQuadTerm_x + velQuadTerm_y )**2.0 &
          &       + rho_om
        fneq(Q_SW) = -sum8_1 + sum8_2 - pdfTmp(Q_SW,iField)
        fneq(Q_NE) = sum8_1 + sum8_2 - pdfTmp(Q_NE,iField)

        !north west / south east
        sum9_1 = rho_o * ( -uxstar(iField) + uystar(iField) ) *div3_4h
        sum9_2 = rho_o * ( -velQuadTerm_x + velQuadTerm_y )**2.0&
          &       + rho_om
        fneq(Q_SE) = -sum9_1 + sum9_2 - pdfTmp(Q_SE,iField)
        fneq(Q_NW) = sum9_1 + sum9_2 - pdfTmp(Q_NW,iField)

        ! omega * fNonEq
        fneq_om = matmul( omegaMoments(:,:, iField), fneq )  

        !outstate
        outstate( ?SAVE?( varPos(Q__0,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q__0, iField ) + fneq_om( Q__0 )

        outstate( ?SAVE?( varPos(Q__W,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q__W, iField ) + fneq_om( Q__W )
        outstate( ?SAVE?( varPos(Q__E,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q__E, iField ) + fneq_om( Q__E )
        outstate( ?SAVE?( varPos(Q__S,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q__S, iField ) + fneq_om( Q__S )
        outstate( ?SAVE?( varPos(Q__N,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q__N, iField ) + fneq_om( Q__N )
        outstate( ?SAVE?( varPos(Q__B,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q__B, iField ) + fneq_om( Q__B )
        outstate( ?SAVE?( varPos(Q__T,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q__T, iField ) + fneq_om( Q__T )
        outstate( ?SAVE?( varPos(Q_BS,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_BS, iField ) + fneq_om( Q_BS )
        outstate( ?SAVE?( varPos(Q_TN,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_TN, iField ) + fneq_om( Q_TN )
        outstate( ?SAVE?( varPos(Q_BN,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_BN, iField ) + fneq_om( Q_BN )
        outstate( ?SAVE?( varPos(Q_TS,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_TS, iField ) + fneq_om( Q_TS )
        outstate( ?SAVE?( varPos(Q_BW,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_BW, iField ) + fneq_om( Q_BW )
        outstate( ?SAVE?( varPos(Q_TE,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_TE, iField ) + fneq_om( Q_TE )
        outstate( ?SAVE?( varPos(Q_BE,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_BE, iField ) + fneq_om( Q_BE )
        outstate( ?SAVE?( varPos(Q_TW,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_TW, iField ) + fneq_om( Q_TW )
        outstate( ?SAVE?( varPos(Q_SW,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_SW, iField ) + fneq_om( Q_SW )
        outstate( ?SAVE?( varPos(Q_NE,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_NE, iField ) + fneq_om( Q_NE )
        outstate( ?SAVE?( varPos(Q_SE,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_SE, iField ) + fneq_om( Q_SE )
        outstate( ?SAVE?( varPos(Q_NW,iField), iElem, QQF3, neigh ) ) =             &
          & pdfTmp( Q_NW, iField ) + fneq_om( Q_NW )

      enddo !iField
    enddo nodeloop

  end subroutine mrt_advRel_d3q19f3_MSLiquid
!******************************************************************************!

!******************************************************************************!
  !> This routine update currState with external force fields computed from 
  !! previous state
  subroutine externalForce_MSLiquid( iScheme, fieldProp, inState, outState,  &
    &                                neigh, nStart, nEnd, level, layout,     &
    &                                pdfGlobal, params, scheme, varSys, tree )
    !------------------------------------------------------------------------ 
    integer, intent(in) :: iScheme !< current scheme
    type(mus_field_prop_type), intent(in) :: fieldProp(:)   !< Global fluid 
    real(kind=rk), dimension(:), intent(in)  :: inState   !< input pdf vector
    real(kind=rk), dimension(:), intent(inout) :: outState  !< output pdf vector
    integer, dimension(:), intent(in) :: neigh   !< connectivity vector 
    integer, intent(in) :: nStart               !< start positions in pdf vector 
    integer, intent(in) :: nEnd                 !< end positions in pdf vector 
    integer,intent(in) :: level                 !< current level
    type(mus_scheme_layout_type), intent(in) :: layout
    type(pdf_globInfo_type), intent(in) :: pdfGlobal !< global info type
    type(mus_param_type),intent(in) :: params   !< global parameters
    type( mus_scheme_type ), intent(in) :: scheme(:) !< array of all schemes
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    type( treelmesh_type ), intent(in) :: tree !< array of all schemes
    !------------------------------------------------------------------------ 
    real(kind=rk), dimension(nEnd-nStart+1, 3) :: electricField, gravityField
    real(kind=rk), dimension(3, scheme(iScheme)%nFields ) :: mixForce,       &
      & diffForce, gravityTerm, electricTerm
    real(kind=rk) :: totMixForce(3), diff_forceTerm, mix_forceTerm
    integer :: vPos( layout%stencil(1)%QQ )
    integer :: iElem, iField, iDir, nFields, nElems
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: mass_dens, num_dens
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: massFrac
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: molWeight, chargeNr
    type(mus_mixture_type) :: mixture
    real(kind=rk) :: totNum_dens, totMass_dens, minMolWeight
    real(kind=rk) :: electricPhyFac, gravityPhyFac
    !--------------------------------------------------------------------------!
    nElems = nEnd - nStart + 1
    nFields = scheme(iScheme)%nFields
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    minMolWeight = minval(molWeight)
    !specific charge of ionic species
    chargeNr = fieldProp(:)%species%chargeNr

    mixture = scheme(iScheme)%mixture

    ! unit N/C
    electricPhyFac = params%physics%coulomb0                          &
      &           / params%physics%fac(level)%force         

    ! unit m/s^2
    gravityPhyFac = params%physics%dtLvl(level)**2                    &
      &           / params%physics%dxLvl(level)

    ! Get Electric field which might be defined to vary in space and time
    electricField(:,1) = tem_spacetime_for(                                    &
      & me      = scheme(iScheme)%ext_force%electricField(1),                  &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),    &
      & tree    = tree,                                                        &
      & time    = params%general%simControl%now,                               &
      & n       = nElems,                                                      &
      & conf    = params%general%solver%conf(1) )
  
    electricField(:,2) = tem_spacetime_for(                                    &
      & me      = scheme(iScheme)%ext_force%electricField(2),                  &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),    &
      & tree    = tree,                                                        &
      & time    = params%general%simControl%now,                               &
      & n       = nElems,                                                      &
      & conf    = params%general%solver%conf(1) )
  
    electricField(:,3) = tem_spacetime_for(                                    &
      & me      = scheme(iScheme)%ext_force%electricField(3),                  &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),    &
      & tree    = tree,                                                        &
      & time    = params%general%simControl%now,                               &
      & n       = nElems,                                                      &
      & conf    = params%general%solver%conf(1) )
    
    ! convert physical to lattice
    electricField = electricField * electricPhyFac

    ! Get Electric field which might be defined to vary in space and time
    gravityField(:,1) = tem_spacetime_for(                                     &
      & me      = scheme(iScheme)%ext_force%gravityField(1),                   &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),    &
      & tree    = tree,                                                        &
      & time    = params%general%simControl%now,                               &
      & n       = nElems,                                                      &
      & conf    = params%general%solver%conf(1) )
  
    gravityField(:,2) = tem_spacetime_for(                                     &
      & me      = scheme(iScheme)%ext_force%gravityField(2),                   &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),    &
      & tree    = tree,                                                        &
      & time    = params%general%simControl%now,                               &
      & n       = nElems,                                                      &
      & conf    = params%general%solver%conf(1) )
  
    gravityField(:,3) = tem_spacetime_for(                                     &
      & me      = scheme(iScheme)%ext_force%gravityField(3),                   &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),    &
      & tree    = tree,                                                        &
      & time    = params%general%simControl%now,                               &
      & n       = nElems,                                                      &
      & conf    = params%general%solver%conf(1) )

    ! convert physical to lattice
    gravityField = gravityField * gravityPhyFac  

    nodeloop: do iElem = nStart, nEnd
      mass_dens = 0._rk
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        ! compute field density and first moments
        do iDir = 1, layout%stencil(1)%QQ
          !field density
          mass_dens( iField ) = mass_dens( iField )                           &
            & + instate( ?FETCH?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
        enddo
      enddo
      !total mass density
      totmass_Dens = sum(mass_dens)

      !mass fraction
      massFrac(:) = mass_dens(:)/totMass_dens

      ! solve linear system of equation for actual moments
      ! number density of all species
      num_dens(:) = mass_dens(:)/molWeight(:)

      !total number density
      totNum_dens = sum(num_dens(:))

      ! compute external forcing term
      ! d^m_k = w_m*c_m*( n0*min_a(m_a)*F_k + (1/cs2)*(rho_k*g + n_k*z_k*F*E) )

      gravityTerm = 0.0_rk
      electricTerm = 0.0_rk
      do iField = 1, nFields
        gravityTerm(:, iField) = mass_dens(iField) * gravityField(iElem,:)
        electricTerm(:, iField) = num_dens(iField) * chargeNr(iField)          &
          &                     * mixture%faradayLB * electricField(iElem,:)
      end do

      ! forcing term in the mixture for momentum equation
      ! mixForce = (1/cs2)*(rho_k*g + n_k*z_k*F*E)
      mixforce = 0.0_rk
      do iField = 1, nFields
        mixforce(:,iField) = cs2inv                                          &
          &             * (gravityTerm(:, iField) + electricTerm(:, iField))
      end do
 
      ! diffusive forcing term in the Maxwell-Stefan equations
      ! F_k = 1/(nRT) rho_k (g - z_k Faraday/m_k E) 
      !     - y_k/(nRT) sum_l(rho_l (g - z_l Faraday/m_l E)
 
      ! diffusive force on each species
      ! diffForce = rho_k (g - z_k Faraday/m_k E)/
      diffForce = 0.0_rk
      do iField = 1, nFields
        diffForce(:, iField) = gravityTerm(:, iField)                        &
          &                  + electricTerm(:, iField)
      end do

      ! total mixture force in diffusive forcing term
      ! totMixForce = sum_l(rho_l (g - z_l Faraday/m_l E)
      totMixForce = 0.0_rk
      do iField = 1, nFields
        totMixForce(:) = totMixForce(:) + diffForce(:, iField)
      end do

      ! update diffForce with fraction of totMixForce and 
      ! constant term 1/(nRT) * n0 * min_a(m_a)
      ! \todo KM: test moleDens0LB = totNum_dens below
      do iField = 1, nFields
        diffForce(:, iField) = totNum_dens * minMolWeight                      &
          &                  * ( diffForce(:, iField)                          &
          &                  - massFrac(iField)*totMixForce(:) )               &
          &                  / (totNum_dens *mixture%gasConst_R_LB             &
          &                  * mixture%temp0LB )
      end do

      ! forcing term
      ! d^m_k = min_a(m_a)*weight_m*cxDir_m * n * F_k
      !       + weight_m*cxDir_m*cs2inv*(rho_k*g - n_k*z_k Faraday E)
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        do iDir = 1, layout%stencil(1)%QQ
          diff_forceTerm =  &
            &   dble( layout%stencil(1)%cxDir( 1, iDir ))*diffForce(1, iField) &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*diffForce(2, iField) &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*diffForce(3, iField)
          mix_forceTerm =  &
            &   dble( layout%stencil(1)%cxDir( 1, iDir ))*mixForce(1, iField)  &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*mixForce(2, iField)  &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*mixForce(3, iField)

          outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) =    &
           & outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) )   &
           & + layout%weight( iDir ) * ( diff_forceTerm + mix_forceTerm )
        end do
      end do
    end do nodeloop

  end subroutine externalForce_MSLiquid  
!******************************************************************************

!******************************************************************************!
  !> This routine update currState with external force fields computed from 
  !! previous state
  subroutine externalForce_MSLiquid_WTF( iScheme, fieldProp, inState, outState,&
    &                                neigh, nStart, nEnd, level, layout,       &
    &                                pdfGlobal, params, scheme, varSys, tree )
    !------------------------------------------------------------------------ 
    integer, intent(in) :: iScheme !< current scheme
    type(mus_field_prop_type), intent(in) :: fieldProp(:)   !< Global fluid 
    real(kind=rk), dimension(:), intent(in)  :: inState   !< input pdf vector
    real(kind=rk), dimension(:), intent(inout) :: outState  !< output pdf vector
    integer, dimension(:), intent(in) :: neigh   !< connectivity vector 
    integer, intent(in) :: nStart               !< start positions in pdf vector 
    integer, intent(in) :: nEnd                 !< end positions in pdf vector 
    integer,intent(in) :: level                 !< current level
    type(mus_scheme_layout_type), intent(in) :: layout
    type(pdf_globInfo_type), intent(in) :: pdfGlobal !< global info type
    type(mus_param_type),intent(in) :: params   !< global parameters
    type( mus_scheme_type ), intent(in) :: scheme(:) !< array of all schemes
    !> variable system definition
    type(tem_var_system_type), intent(in) :: varSys
    type( treelmesh_type ), intent(in) :: tree !< array of all schemes
    !------------------------------------------------------------------------ 
    real(kind=rk), dimension(nEnd-nStart+1, 3) :: electricField, gravityField
    real(kind=rk), dimension(3, scheme(iScheme)%nFields ) :: mixForce,       &
      & diffForce, gravityTerm, electricTerm, diffForce_WTF
    real(kind=rk) :: totMixForce(3), diff_forceTerm, mix_forceTerm
    integer :: vPos( layout%stencil(1)%QQ )
    integer :: iElem, iField, iDir, nFields, nElems, iField_2
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: mass_dens, num_dens
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: massFrac, moleFrac
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: molWeight, chargeNr
    type(mus_mixture_type) :: mixture
    real(kind=rk) :: totNum_dens, totMass_dens, minMolWeight
    real(kind=rk) :: electricPhyFac, gravityPhyFac
    real(kind=rk), &
      & dimension(scheme(iScheme)%nFields,scheme(iScheme)%nFields) ::          &
      & thermodynamic_fac, inv_thermodyn_fac
    real(kind=rk) :: temp, press
    !--------------------------------------------------------------------------!
    nElems = nEnd - nStart + 1
    nFields = scheme(iScheme)%nFields
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    minMolWeight = minval(molWeight)
    !specific charge of ionic species
    chargeNr = fieldProp(:)%species%chargeNr

    mixture = scheme(iScheme)%mixture

    ! temperature
    temp = mixture%temp0
    ! atmospheric pressure
    press = mixture%atm_press

    ! initialize thermodynamic factor 
    ! compute activity_coeff from matthias c-code for each element
    thermodynamic_fac = 0.0_rk
    do iField = 1, nFields
      thermodynamic_fac(iField, iField) = 1.0_rk
    end do

    ! unit N/C
    electricPhyFac = params%physics%coulomb0 / params%physics%fac(level)%force

    ! unit m/s^2
    gravityPhyFac = params%physics%dtLvl(level)**2 / params%physics%dxLvl(level)

    ! Get Electric field which might be defined to vary in space and time
    electricField(:,1) = tem_spacetime_for(                                  &
      & me      = scheme(iScheme)%ext_force%electricField(1),                &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),  &
      & tree    = tree,                                                      &
      & time    = params%general%simControl%now,                             &
      & n       = nElems,                                                    &
      & conf    = params%general%solver%conf(1) )
  
    electricField(:,2) = tem_spacetime_for(                                  &
      & me      = scheme(iScheme)%ext_force%electricField(2),                &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),  &
      & tree    = tree,                                                      &
      & time    = params%general%simControl%now,                             &
      & n       = nElems,                                                    &
      & conf    = params%general%solver%conf(1) )
  
    electricField(:,3) = tem_spacetime_for(                                  &
      & me      = scheme(iScheme)%ext_force%electricField(3),                &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),  &
      & tree    = tree,                                                      &
      & time    = params%general%simControl%now,                             &
      & n       = nElems,                                                    &
      & conf    = params%general%solver%conf(1) )
    
    ! convert physical to lattice
    electricField = electricField * electricPhyFac

    ! Get Electric field which might be defined to vary in space and time
    gravityField(:,1) = tem_spacetime_for(                                   &
      & me      = scheme(iScheme)%ext_force%gravityField(1),                 &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),  &
      & tree    = tree,                                                      &
      & time    = params%general%simControl%now,                             &
      & n       = nElems,                                                    &
      & conf    = params%general%solver%conf(1) )
  
    gravityField(:,2) = tem_spacetime_for(                                   &
      & me      = scheme(iScheme)%ext_force%gravityField(2),                 &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),  &
      & tree    = tree,                                                      &
      & time    = params%general%simControl%now,                             &
      & n       = nElems,                                                    &
      & conf    = params%general%solver%conf(1) )
  
    gravityField(:,3) = tem_spacetime_for(                                   &
      & me      = scheme(iScheme)%ext_force%gravityField(3),                 &
      & treeIDs = scheme(iScheme)%glob%levelDesc(level)%total(nStart:nEnd),  &
      & tree    = tree,                                                      &
      & time    = params%general%simControl%now,                             &
      & n       = nElems,                                                    &
      & conf    = params%general%solver%conf(1) )

    ! convert physical to lattice
    gravityField = gravityField * gravityPhyFac  

    nodeloop: do iElem = nStart, nEnd
      mass_dens = 0._rk
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        ! compute field density and first moments
        do iDir = 1, layout%stencil(1)%QQ
          !field density
          mass_dens( iField ) = mass_dens( iField )                           &
            & + instate( ?FETCH?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
        enddo
      enddo
      !total mass density
      totmass_Dens = sum(mass_dens)

      !mass fraction
      massFrac(:) = mass_dens(:)/totMass_dens

      ! solve linear system of equation for actual moments
      ! number density of all species
      num_dens(:) = mass_dens(:)/molWeight(:)

      !total number density
      totNum_dens = sum(num_dens(:))

      ! mole fraction
      moleFrac(:) =  num_dens(:)/totNum_dens

      ! compute external forcing term
      ! d^m_k = w_m*c_m*( n0*min_a(m_a)*F_k + (1/cs2)*(rho_k*g + n_k*z_k*F*E) )

      gravityTerm = 0.0_rk
      electricTerm = 0.0_rk
      do iField = 1, nFields
        gravityTerm(:, iField) = mass_dens(iField) * gravityField(iElem,:)
        electricTerm(:, iField) = num_dens(iField) * chargeNr(iField)          &
          &                     * mixture%faradayLB * electricField(iElem,:)
      end do

      ! forcing term in the mixture for momentum equation
      ! mixForce = (1/cs2)*(rho_k*g + n_k*z_k*F*E)
      mixforce = 0.0_rk
      do iField = 1, nFields
        mixforce(:,iField) = cs2inv                                          &
          &             * (gravityTerm(:, iField) + electricTerm(:, iField))
      end do
 
      ! diffusive forcing term in the Maxwell-Stefan equations
      ! F_k = 1/(nRT) rho_k (g - z_k Faraday/m_k E) 
      !     - y_k/(nRT) sum_l(rho_l (g - z_l Faraday/m_l E)
 
      ! diffusive force on each species
      ! diffForce = rho_k (g - z_k Faraday/m_k E)
      diffForce = 0.0_rk
      do iField = 1, nFields
        diffForce(:, iField) = gravityTerm(:, iField)                        &
          &                  + electricTerm(:, iField)
      end do

      ! total mixture force in diffusive forcing term
      ! totMixForce = sum_l(rho_l (g - z_l Faraday/m_l E)
      totMixForce = 0.0_rk
      do iField = 1, nFields
        totMixForce(:) = totMixForce(:) + diffForce(:, iField)
      end do
 
      ! update diffForce, F_k with fraction of totMixForce
      ! totalNumDens term in F_k cancels with totalNumDens term in 
      ! external driving force equation d_k but it is still used in calculation
      do iField = 1, nFields
        diffForce(:, iField) = ( diffForce(:, iField)                          &
          &                  - massFrac(iField)*totMixForce(:) )               &
          &                  / (totNum_dens * mixture%gasConst_R_LB            &
          &                  * mixture%temp0LB )
      end do

      ! Thermodynamic factor from C++ code
      call mus_calc_thermFactor( nFields, temp, press, moleFrac,               &
        &                        thermodynamic_fac )
      inv_thermodyn_fac = invert_matrix( thermodynamic_fac )

      ! multiply diffForce with thermFacInv * min_a(m_a)
      ! diffusive driving force term:
      !  min_a(m_a) * c_m* w_m * sum_l(themFacInv_k_l * n * F_l)
      diffForce_WTF = 0.0_rk
      do iField = 1, nFields
        do iField_2 = 1, nFields
          diffForce_WTF(:, iField ) = diffForce_WTF(:, iField)                 &
            &                       + inv_thermodyn_fac(iField, iField_2)      &
            &                       * totNum_dens * diffForce(:, iField_2)
        end do
        diffForce_WTF(:, iField ) = minMolWeight * diffForce_WTF(:, iField)
      end do

      ! forcing term
      ! d^m_k = min_a(m_a)*weight_m*cxDir_m*sum_l(themFacInv_k_l * n * F_l) 
      !       + weight_m*cxDir_m*cs2inv*(rho_k*g - n_k*z_k Faraday E)
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        do iDir = 1, layout%stencil(1)%QQ
          diff_forceTerm =  &
            &   dble( layout%stencil(1)%cxDir( 1, iDir ))                          &
            &   * diffForce_WTF(1, iField) &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))                          &
            &   * diffForce_WTF(2, iField) &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))                          &
            &   * diffForce_WTF(3, iField)
          mix_forceTerm =  &
            &   dble( layout%stencil(1)%cxDir( 1, iDir ))*mixForce(1, iField)  &
            & + dble( layout%stencil(1)%cxDir( 2, iDir ))*mixForce(2, iField)  &
            & + dble( layout%stencil(1)%cxDir( 3, iDir ))*mixForce(3, iField)

          outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) =    &
           & outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) )   &
           & + layout%weight( iDir ) * ( diff_forceTerm + mix_forceTerm )
        end do
      end do
    end do nodeloop

  end subroutine externalForce_MSLiquid_WTF 
!******************************************************************************


!******************************************************************************
  !> Unoptimized Advection relaxation routine for the multispecies BGK model 
  !! with thermodynamic factors in Maxwell-Stefan formulation
  !!
  !! This routine contains the implementation based on the paper
  !! "A Lattice Boltzmann Scheme for liquid mixtures - Part II: Discretization  
  !! and Numerics, Jens Zudrop, Sabine Roller, Pietro Asinari. "\n
?? copy :: compute_routineHeader( bgk_advRel_MSLiquid_WTDF )
    !------------------------------------------------------------------------ 
    !local variables
    real(kind=rk) :: pdfTmp( varSys%nScalars )  !< temporary local pdf values
    integer :: iElem, iField, iField_2, iField_3, nFields, iDir
    integer :: vPos( layout%stencil(1)%QQ )
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: mass_dens, num_dens
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: moleFrac
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: first_moments
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: velocity
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: eqVel
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: molWeight, phi 
    real(kind=rk) :: totNum_dens, usqr, ucx, feq
    real(kind=rk) :: velAvg(3), velQuad(3), ucxQuadTerm, theta_eq, totMassDens
    real(kind=rk) :: omega, omega_new, paramB
    real(kind=rk), &
      & dimension(scheme(iScheme)%nFields,scheme(iScheme)%nFields) :: matA,    &
      & invA, resi_coeff, activity_coeff, thermodynamic_fac, inv_thermodyn_fac,&
      & diff_coeff
    real(kind=rk) :: temp, press, moleDens0, phy_moleDens_fac
    !--------------------------------------------------------------------------- 
    nFields = scheme(iScheme)%nFields

    ! constant mixture number density
    moleDens0 = scheme(iScheme)%mixture%moleDens0

    !KM \todo check moleDens for multilevel
    phy_moleDens_fac = params%physics%moleDens0
   
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    !molecular weight ratios
    phi = fieldProp(:)%species%molWeigRatio
    !resistivities
    do iField = 1, nFields
      resi_coeff(iField,:) = fieldProp( iField )%species%resi_coeff
    enddo

    omega = scheme(iScheme)%mixture%relaxLvl(Level)%omega_diff
!    write(*,*) 'omega ', omega
    omega_new = (1.0_rk/(1.0_rk/omega + 0.5_rk))
!    write(*,*) 'omega_new ', omega_new

    paramB = scheme(iScheme)%mixture%paramB
    !equilibrium theta
    theta_eq = scheme(iScheme)%mixture%theta_eq

    ! temperature
    temp = scheme(iScheme)%mixture%temp0
    ! atmospheric pressure
    press = scheme(iSCheme)%mixture%atm_press

!write(*,*) 'resi_coeff LB', real(resi_coeff)
!write(*,*) 'temp ', temp, ' press ', press
    ! initialize thermodynamic factor and activity coeff
    ! compute activity_coeff from matthias c-code for each element
    activity_coeff = 0.0_rk
    thermodynamic_fac = 0.0_rk
    do iField = 1, nFields
      thermodynamic_fac(iField, iField) = 1.0_rk
    end do
!    write(*,*) 'paramB ', paramB

!    write(dbgUnit,*) 'B 1', resi_coeff(1,:)
!    write(dbgUnit,*) 'B 2', resi_coeff(2,:)
!    write(dbgUnit,*) 'B 3', resi_coeff(3,:)

    ! KM: debugging multispecies
    !outstate = instate

!write(dbgUnit(1),*)'time ', params%general%simControl%now%iter
    nodeloop: do iElem = nStart, nEnd
!write(dbgUnit(1),*) 'iElem', iElem
!write(*,*) 'iElem', iElem

      mass_dens = 0._rk
      first_moments = 0.0_rk
      velocity = 0.0_rk
      eqVel = 0.0_rk
      pdfTmp = 0._rk

      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        ! compute field density and first moments
        do iDir = 1, layout%stencil(1)%QQ
          !store all field pdf in single element to local array pdfTmp
          pdfTmp( vPos(iDir) ) =                                               &
            & instate( ?FETCH?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
          !field density
          mass_dens( iField ) = mass_dens( iField ) + pdfTmp( vPos(iDir) )

          !field momentum (rho*u)
          first_moments( 1, iField ) = first_moments( 1, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 1, iDir )

          first_moments( 2, iField ) = first_moments( 2, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 2, iDir )

          first_moments( 3, iField ) = first_moments( 3, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 3, iDir )
!write(dbgunit,*) 's',s,'iDir', iDir, layout%stencil(1)%cxDir(3,iDir), pdfTmp( vPos(iDir) ), gqzsigma(s)
        enddo
      enddo
!write(*,*) 'first velocity '      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' vel ', first_moments(:, iField)/mass_dens(iField)
!enddo

      !total mass density
      totmassDens = sum(mass_dens)

      ! solve linear system of equation for actual moments
      ! number density of all species
      num_dens(:) = mass_dens(:)/molWeight(:)

      !total number density
      totNum_dens = sum(num_dens(:))

      !mole fraction
      moleFrac(:) =  num_dens(:)/totNum_dens
!write(*,*) 'molefraction ', moleFrac      

      ! MS-Diff coeff matrix from C++ code
      call mus_calc_MS_DiffMatrix( nFields, temp, press,                       &
        &                        num_dens*phy_moleDens_fac, diff_coeff )
!write(*,*) 'num_dens ', num_dens, scheme(iScheme)%mixture%moleDens0LB,moleFrac*moleDens0 
!write(*,*) 'physical ', totNum_dens/scheme(iScheme)%mixture%moleDens0LB, totNum_dens
!write(*,*) 'physical ', totNum_dens*phy_moleDens_fac
!write(*,*) 'num_dens ', num_dens*phy_moleDens_fac
!write(*,*) 'lattice ', scheme(iScheme)%mixture%moleDens0LB
!write(*,*) 'lattice ', scheme(iScheme)%mixture%moleDens0LB*phy_moleDens_fac
!write(*,*) 'diff_coeff', real(diff_coeff) 
!stop
      ! Convert to lattice unit
      resi_coeff = params%physics%fac(level)%diffusivity/diff_coeff
!write(*,*) 'resi_coeff LB', real(resi_coeff) 
      
!stop

      ! Thermodynamic factor from C++ code
      call mus_calc_thermFactor( nFields, temp, press, moleFrac,               &
        &                        thermodynamic_fac )

      inv_thermodyn_fac = invert_matrix( thermodynamic_fac )
!write(*,*) 'Therm_factors '
!do iField = 1, nFields
!write(*,*) 'iField ', iField, ' factors ', thermodynamic_fac(iField,:)
!end do
!write(*,*) 'inv Therm_factors '
!do iField = 1, nFields
!write(*,*) 'iField ', iField, ' factors ', inv_thermodyn_fac(iField,:)
!end do
!write(*,*) 'Identity', matmul(thermodynamic_fac, inv_thermodyn_fac)

      matA = 0.0_rk
      !build up matrix to solver LSE for actual velocity
      do iField = 1, nFields
        !set diagonal part 
        matA(iField, iField) = 1.0_rk
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            matA(iField, iField_2) = matA(iField, iField_2) + omega * 0.5_rk   &
              &                    * inv_thermodyn_fac(iField, iField_2)       &
              &                    * resi_coeff(iField_2, iField_3)            &
              &                    * phi(iField_2) * moleFrac(iField_3)        &
              &                    / paramB
          end do
        end do
        !set non-diagonal part
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            matA(iField, iField_3) = matA(iField, iField_3) - omega * 0.5_rk   &
              &                    * inv_thermodyn_fac(iField, iField_2)       &
              &                    * resi_coeff(iField_2, iField_3)            &
              &                    * phi(iField_3) * moleFrac(iField_2)        &
              &                    / paramB
          end do
        end do
      end do
!write(*,*) 'matrixA'      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' 2_matA ', matA(iField, :)
!end do
!      matA = 0.0_rk
!      !build up matrix to solver LSE for actual velocity
!      do iField = 1, nFields
!        !set diagonal part 
!        matA(iField, iField) = 1.0_rk
!        do iField_2 = 1, nFields
!          matA(iField, iField) = matA(iField, iField) + omega * 0.5_rk         &
!            &                  * resi_coeff(iField, iField_2) * phi(iField)    &
!            &                  * moleFrac(iField_2) / paramB
!        end do
!        !set non-diagonal part
!        do iField_2 = 1, nFields
!          matA(iField, iField_2) = matA(iField, iField_2) - omega * 0.5_rk     &
!            &                    * resi_coeff(iField, iField_2) * phi(iField_2)&
!            &                    * moleFrac(iField) / paramB
!        end do
!      end do

      ! invert matrix
      invA = invert_matrix( matA )
!write(*,*) 'INV matrixA'      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' inv ', invA(iField, :)
!end do

      !actual velocity of all species
      !KM: TGV testcase
      velocity(1, :) = matmul( invA, first_moments(1,:) ) / mass_dens(:)
      velocity(2, :) = matmul( invA, first_moments(2,:) ) / mass_dens(:)
      velocity(3, :) = matmul( invA, first_moments(3,:) ) / mass_dens(:)
!write(*,*) 'density ', mass_dens      
!write(*,*) 'velocity '      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' vel ', velocity(:, iField)
!enddo
!stop

!write(*,*) 'test first_moments'
!do iField = 1, nFields
!  velocity(:,iField) = velocity(:,iField)*mass_dens(iField)
!end do
!write(*,*) 'x ', matmul(matA, velocity(1,:))/mass_dens(:)
!write(*,*) 'y ', matmul(matA, velocity(2,:))/mass_dens(:)
!write(*,*) 'z ', matmul(matA, velocity(3,:))/mass_dens(:)

      ! compute equilibrim velocity
!      do iField = 1, nFields
!        eqVel( :, iField ) = mass_dens(iField)*velocity( :, iField )
!        do iField_2 = 1, nFields
!          eqVel( :, iField ) = eqVel( :, iField )                              &
!            &                + mass_dens(iField)                               &
!            &                * resi_coeff( iField, iField_2 ) * phi(iField)    &
!            &                * moleFrac(iField_2)                              &
!            &                * (velocity(:, iField_2) - velocity(:,iField))    &
!            &                / paramB 
!        end do  
!      end do


      do iField = 1, nFields
        eqVel( :, iField ) = mass_dens(iField)*velocity( :, iField )
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            eqVel( :, iField ) = eqVel( :, iField )                              &
              &                + inv_thermodyn_fac(iField, iField_2)             &
              &                * mass_dens(iField_2)                             &
              &                * resi_coeff( iField_2, iField_3 ) * phi(iField_2)&
              &                * moleFrac(iField_3)                              &
              &                * (velocity(:, iField_3) - velocity(:,iField_2))  &
              &                / paramB 
          end do    
        end do  
      end do

!write(*,*) 'eq_velocity '      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' eqvel ', eqVel(:, iField)
!enddo
!stop

      !compute mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens, velocity(1,:) )/totmassDens
      velAvg(2) = dot_product( mass_dens, velocity(2,:) )/totmassDens
      velAvg(3) = dot_product( mass_dens, velocity(3,:) )/totmassDens

      !compute equilibrium and do collision
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        !KM: TGV testcase
        velQuad = theta_eq*velAvg                                              &
          &     + (1.0_rk - theta_eq) * eqVel(:, iField) / mass_dens(iField)

        usqr = dot_product( velQuad, velQuad ) * t2cs2inv
!write(dbgUnit(1),*) 'omega_new ', omega_new
        do iDir = 1, layout%stencil(1)%QQ
          ucx = dot_product(                                                   &
            & dble(layout%stencil(1)%cxDir(:, iDir)), eqVel(:, iField) )

          ucxQuadTerm = dot_product(                                           &
            & dble(layout%stencil(1)%cxDir(:, iDir)), velQuad )

          !feq = layout%weight(iDir) * mass_dens(iField) * ( phi(iField)        &
          !  & + ucx * cs2inv + ucx * ucx * t2cs4inv - usqr )
          feq = layout%weight(iDir) * ( mass_dens(iField) * ( phi(iField)        &
            & + ucxQuadTerm * ucxQuadTerm * t2cs4inv - usqr ) + ucx * cs2inv )
          
          if ( iDir == layout%stencil(1)%restPosition ) then      
          ! equilibrium at rest
            select case( trim(scheme(iScheme)%header%layout) )
            case('d2q9')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 9._rk - 5._rk * phi(iField) )/4._rk - usqr )
            case('d3q19')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 3._rk - 2._rk * phi(iField) ) - usqr )
            end select
          end if  
!write(dbgUnit(1),*) 'iDir ', iDir, (feq-pdfTmp(vPos(iDir))), omega_new*(feq-pdfTmp(vPos(iDir)))
          outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) = &
            & pdfTmp( vPos(iDir) ) + omega_new * ( feq - pdfTmp( vPos(iDir) ) ) 
          if (tem_isnan( outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ))) then
            write(*,*) 'found nan', iElem, iField
            stop
          endif
!          if (outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) < 0.0_rk) then
!            write(*,*)'Kernel: time ', params%general%simControl%now%iter
!            write(*,*) 'found negative', iElem, iField, &
!              &outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
!            stop
!          endif
        enddo !iDir
          
!      write(*,*) 'outpdf',s, outstate(vPos(:)) 
      enddo !iField
!    stop
    enddo nodeloop
!stop
  end subroutine bgk_advRel_MSLiquid_WTDF
!******************************************************************************

!******************************************************************************
  !> Unoptimized Advection relaxation routine for the multispecies BGK model 
  !! with thermodynamic factors in Maxwell-Stefan formulation
  !!
  !! This routine contains the implementation based on the paper
  !! "A Lattice Boltzmann Scheme for liquid mixtures - Part II: Discretization  
  !! and Numerics, Jens Zudrop, Sabine Roller, Pietro Asinari. "\n
?? copy :: compute_routineHeader( bgk_advRel_MSLiquid_WTDF_2Spc )
    !------------------------------------------------------------------------ 
    !local variables
    real(kind=rk) :: pdfTmp( varSys%nScalars )  !< temporary local pdf values
    integer :: iElem, iField, iField_2, iField_3, nFields, iDir
    integer :: vPos( layout%stencil(1)%QQ )
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: mass_dens, num_dens
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: moleFrac
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: first_moments
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: velocity
    real(kind=rk), dimension(3, scheme(iScheme)%nFields) :: eqVel
    real(kind=rk), dimension(scheme(iScheme)%nFields) :: molWeight, phi 
    real(kind=rk) :: totNum_dens, usqr, ucx, feq
    real(kind=rk) :: velAvg(3), velQuad(3), ucxQuadTerm, theta_eq, totMassDens
    real(kind=rk) :: omega, omega_new, paramB
    real(kind=rk), &
      & dimension(scheme(iScheme)%nFields,scheme(iScheme)%nFields) :: matA,    &
      & invA, resi_coeff, thermodynamic_fac, inv_thermodyn_fac,&
      & diff_coeff
    real(kind=rk), dimension(3,3) :: tdf_3Spc, diff_coeff_3Spc
    real(kind=rk) :: num_dens_3Spc(3), moleFrac_3Spc(3)
    real(kind=rk) :: temp, press, moleDens0, phy_moleDens_fac
    !--------------------------------------------------------------------------- 
    nFields = scheme(iScheme)%nFields

    ! constant mixture number density
    moleDens0 = scheme(iScheme)%mixture%moleDens0

    !KM \todo check moleDens for multilevel
    phy_moleDens_fac = params%physics%moleDens0
   
    !molecular weights
    molWeight = fieldProp(:)%species%molweight
    !molecular weight ratios
    phi = fieldProp(:)%species%molWeigRatio
    !resistivities
    do iField = 1, nFields
      resi_coeff(iField,:) = fieldProp( iField )%species%resi_coeff
    enddo

    omega = scheme(iScheme)%mixture%relaxLvl(Level)%omega_diff
!    write(*,*) 'omega ', omega
    omega_new = (1.0_rk/(1.0_rk/omega + 0.5_rk))
!    write(*,*) 'omega_new ', omega_new

    paramB = scheme(iScheme)%mixture%paramB
    !equilibrium theta
    theta_eq = scheme(iScheme)%mixture%theta_eq

    ! temperature
    temp = scheme(iScheme)%mixture%temp0
    ! atmospheric pressure
    press = scheme(iSCheme)%mixture%atm_press

!write(*,*) 'resi_coeff LB', real(resi_coeff)
!write(*,*) 'temp ', temp, ' press ', press
    ! initialize thermodynamic factor 
    thermodynamic_fac = 0.0_rk
    do iField = 1, nFields
      thermodynamic_fac(iField, iField) = 1.0_rk
      diff_coeff(iField, iField) = 1.0_rk
    end do
!    write(*,*) 'paramB ', paramB

!    write(dbgUnit,*) 'B 1', resi_coeff(1,:)
!    write(dbgUnit,*) 'B 2', resi_coeff(2,:)
!    write(dbgUnit,*) 'B 3', resi_coeff(3,:)

    ! KM: debugging multispecies
    !outstate = instate

!write(dbgUnit(1),*)'time ', params%general%simControl%now%iter
    nodeloop: do iElem = nStart, nEnd
!write(dbgUnit(1),*) 'iElem', iElem
!write(*,*) 'iElem', iElem

      mass_dens = 0._rk
      first_moments = 0.0_rk
      velocity = 0.0_rk
      eqVel = 0.0_rk
      pdfTmp = 0._rk

      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        ! compute field density and first moments
        do iDir = 1, layout%stencil(1)%QQ
          !store all field pdf in single element to local array pdfTmp
          pdfTmp( vPos(iDir) ) =                                               &
            & instate( ?FETCH?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
          !field density
          mass_dens( iField ) = mass_dens( iField ) + pdfTmp( vPos(iDir) )

          !field momentum (rho*u)
          first_moments( 1, iField ) = first_moments( 1, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 1, iDir )

          first_moments( 2, iField ) = first_moments( 2, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 2, iDir )

          first_moments( 3, iField ) = first_moments( 3, iField )              &
            & + pdfTmp( vPos(iDir) ) * layout%stencil(1)%cxDir( 3, iDir )
!write(dbgunit,*) 's',s,'iDir', iDir, layout%stencil(1)%cxDir(3,iDir), pdfTmp( vPos(iDir) ), gqzsigma(s)
        enddo
      enddo
!write(*,*) 'first velocity '      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' vel ', first_moments(:, iField)/mass_dens(iField)
!enddo

      !total mass density
      totmassDens = sum(mass_dens)

      ! solve linear system of equation for actual moments
      ! number density of all species
      num_dens(:) = mass_dens(:)/molWeight(:)

      !total number density
      totNum_dens = sum(num_dens(:))

      !mole fraction
      moleFrac(:) =  num_dens(:)/totNum_dens
!write(*,*) 'molefraction ', moleFrac      

      num_dens_3Spc(1) = num_dens(1)
      num_dens_3Spc(2) = num_dens(2)
      num_dens_3Spc(3) = num_dens(2)
      moleFrac_3Spc(1) = moleFrac(1)
      moleFrac_3Spc(2) = moleFrac(2)
      moleFrac_3Spc(3) = moleFrac(2)
!write(*,*) 'moleFrac_3Spc ', moleFrac_3Spc
      ! MS-Diff coeff matrix from C++ code
      call mus_calc_MS_DiffMatrix( 3, temp, press,                            & 
        &                        num_dens_3Spc*phy_moleDens_fac,              &
        &                        diff_coeff_3Spc )
!write(*,*) 'diff_coeff 3Spc ', diff_coeff_3Spc
      diff_coeff(1,2) = (diff_coeff_3Spc(1,2)*diff_coeff_3Spc(1,3))            &
        &             / (diff_coeff_3Spc(1,2)+diff_coeff_3Spc(1,3))
      diff_coeff(2,1) = diff_coeff(1,2)
!write(*,*) 'diff_coeff ', diff_coeff 
      ! Convert to lattice unit
      resi_coeff = params%physics%fac(level)%diffusivity/diff_coeff
!write(*,*) 'resi_coeff LB', real(resi_coeff) 
!stop
      call mus_calc_thermFactor( 3, temp, press, moleFrac_3Spc, tdf_3Spc )

      thermodynamic_fac(1,1) = tdf_3Spc(1,1)               
      thermodynamic_fac(1,2) = tdf_3Spc(1,2)               
      thermodynamic_fac(2,1) = tdf_3Spc(2,1)               
      thermodynamic_fac(2,2) = tdf_3Spc(2,2)               
!write(*,*) 'thermodynamic_fac ', thermodynamic_fac
!stop

      inv_thermodyn_fac = invert_matrix( thermodynamic_fac )
!write(*,*) 'Therm_factors '
!do iField = 1, nFields
!write(*,*) 'iField ', iField, ' factors ', thermodynamic_fac(iField,:)
!end do
!write(*,*) 'inv Therm_factors '
!do iField = 1, nFields
!write(*,*) 'iField ', iField, ' factors ', inv_thermodyn_fac(iField,:)
!end do
!write(*,*) 'Identity', matmul(thermodynamic_fac, inv_thermodyn_fac)

      matA = 0.0_rk
      !build up matrix to solver LSE for actual velocity
      do iField = 1, nFields
        !set diagonal part 
        matA(iField, iField) = 1.0_rk
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            matA(iField, iField_2) = matA(iField, iField_2) + omega * 0.5_rk   &
              &                    * inv_thermodyn_fac(iField, iField_2)       &
              &                    * resi_coeff(iField_2, iField_3)            &
              &                    * phi(iField_2) * moleFrac(iField_3)        &
              &                    / paramB
          end do
        end do
        !set non-diagonal part
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            matA(iField, iField_3) = matA(iField, iField_3) - omega * 0.5_rk   &
              &                    * inv_thermodyn_fac(iField, iField_2)       &
              &                    * resi_coeff(iField_2, iField_3)            &
              &                    * phi(iField_3) * moleFrac(iField_2)        &
              &                    / paramB
          end do
        end do
      end do
!write(*,*) 'matrixA'      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' 2_matA ', matA(iField, :)
!end do
!      matA = 0.0_rk
!      !build up matrix to solver LSE for actual velocity
!      do iField = 1, nFields
!        !set diagonal part 
!        matA(iField, iField) = 1.0_rk
!        do iField_2 = 1, nFields
!          matA(iField, iField) = matA(iField, iField) + omega * 0.5_rk         &
!            &                  * resi_coeff(iField, iField_2) * phi(iField)    &
!            &                  * moleFrac(iField_2) / paramB
!        end do
!        !set non-diagonal part
!        do iField_2 = 1, nFields
!          matA(iField, iField_2) = matA(iField, iField_2) - omega * 0.5_rk     &
!            &                    * resi_coeff(iField, iField_2) * phi(iField_2)&
!            &                    * moleFrac(iField) / paramB
!        end do
!      end do

      ! invert matrix
      invA = invert_matrix( matA )
!write(*,*) 'INV matrixA'      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' inv ', invA(iField, :)
!end do

      !actual velocity of all species
      !KM: TGV testcase
      velocity(1, :) = matmul( invA, first_moments(1,:) ) / mass_dens(:)
      velocity(2, :) = matmul( invA, first_moments(2,:) ) / mass_dens(:)
      velocity(3, :) = matmul( invA, first_moments(3,:) ) / mass_dens(:)
!write(*,*) 'density ', mass_dens      
!write(*,*) 'velocity '      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' vel ', velocity(:, iField)
!enddo
!stop

!write(*,*) 'test first_moments'
!do iField = 1, nFields
!  velocity(:,iField) = velocity(:,iField)*mass_dens(iField)
!end do
!write(*,*) 'x ', matmul(matA, velocity(1,:))/mass_dens(:)
!write(*,*) 'y ', matmul(matA, velocity(2,:))/mass_dens(:)
!write(*,*) 'z ', matmul(matA, velocity(3,:))/mass_dens(:)

      ! compute equilibrim velocity
!      do iField = 1, nFields
!        eqVel( :, iField ) = mass_dens(iField)*velocity( :, iField )
!        do iField_2 = 1, nFields
!          eqVel( :, iField ) = eqVel( :, iField )                              &
!            &                + mass_dens(iField)                               &
!            &                * resi_coeff( iField, iField_2 ) * phi(iField)    &
!            &                * moleFrac(iField_2)                              &
!            &                * (velocity(:, iField_2) - velocity(:,iField))    &
!            &                / paramB 
!        end do  
!      end do


      do iField = 1, nFields
        eqVel( :, iField ) = mass_dens(iField)*velocity( :, iField )
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            eqVel( :, iField ) = eqVel( :, iField )                              &
              &                + inv_thermodyn_fac(iField, iField_2)             &
              &                * mass_dens(iField_2)                             &
              &                * resi_coeff( iField_2, iField_3 ) * phi(iField_2)&
              &                * moleFrac(iField_3)                              &
              &                * (velocity(:, iField_3) - velocity(:,iField_2))  &
              &                / paramB 
          end do    
        end do  
      end do

!write(*,*) 'eq_velocity '      
!do iField = 1, nFields
!  write(*,*) 'iField ', iField, ' eqvel ', eqVel(:, iField)
!enddo
!stop

      !compute mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens, velocity(1,:) )/totmassDens
      velAvg(2) = dot_product( mass_dens, velocity(2,:) )/totmassDens
      velAvg(3) = dot_product( mass_dens, velocity(3,:) )/totmassDens

      !compute equilibrium and do collision
      do iField = 1, nFields
        vPos = varSys%Variable%val( iField )%varPos
        !KM: TGV testcase
        velQuad = theta_eq*velAvg                                              &
          &     + (1.0_rk - theta_eq) * eqVel(:, iField) / mass_dens(iField)

        usqr = dot_product( velQuad, velQuad ) * t2cs2inv
!write(dbgUnit(1),*) 'omega_new ', omega_new
        do iDir = 1, layout%stencil(1)%QQ
          ucx = dot_product(                                                   &
            & dble(layout%stencil(1)%cxDir(:, iDir)), eqVel(:, iField) )

          ucxQuadTerm = dot_product(                                           &
            & dble(layout%stencil(1)%cxDir(:, iDir)), velQuad )

          !feq = layout%weight(iDir) * mass_dens(iField) * ( phi(iField)        &
          !  & + ucx * cs2inv + ucx * ucx * t2cs4inv - usqr )
          feq = layout%weight(iDir) * ( mass_dens(iField) * ( phi(iField)        &
            & + ucxQuadTerm * ucxQuadTerm * t2cs4inv - usqr ) + ucx * cs2inv )
          
          if ( iDir == layout%stencil(1)%restPosition ) then      
          ! equilibrium at rest
            select case( trim(scheme(iScheme)%header%layout) )
            case('d2q9')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 9._rk - 5._rk * phi(iField) )/4._rk - usqr )
            case('d3q19')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 3._rk - 2._rk * phi(iField) ) - usqr )
            end select
          end if  
!write(dbgUnit(1),*) 'iDir ', iDir, (feq-pdfTmp(vPos(iDir))), omega_new*(feq-pdfTmp(vPos(iDir)))
          outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) = &
            & pdfTmp( vPos(iDir) ) + omega_new * ( feq - pdfTmp( vPos(iDir) ) ) 
!          if (tem_isnan( outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ))) then
!            write(*,*) 'found nan', iElem, iField
!            stop
!          endif
!          if (outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) ) < 0.0_rk) then
!            write(*,*)'Kernel: time ', params%general%simControl%now%iter
!            write(*,*) 'found negative', iElem, iField, &
!              &outstate( ?SAVE?( vPos(iDir), iElem, varSys%nScalars, neigh ) )
!            stop
!          endif
        enddo !iDir
          
!      write(*,*) 'outpdf',s, outstate(vPos(:)) 
      enddo !iField
!    stop
    enddo nodeloop
!stop
  end subroutine bgk_advRel_MSLiquid_WTDF_2Spc
!******************************************************************************

end module mus_MSLiquid_module

