! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Routines and parameter definitions for the standard D2Q9 model
!! 
!! \author Manuel Hasert, Kannan Masilamani
?? include 'header/lbm_macros.inc'
module mus_d2q9_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_var_system_module, only: tem_var_system_type
  use tem_geometry_module,   only: tem_baryOfId
  use tem_param_module,      only: cs2inv, t2cs2inv,t2cs4inv
  use treelmesh_module,      only: treelmesh_type

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_geom_module,          only: mus_geom_type 
  use mus_param_module,         only: mus_param_type

  implicit none
  private
  public :: bgk_advRel_d2q9_explicit_fixedVel
  public :: bgk_advRel_d2q9_explicit_incomp_fixedVel
  public :: bgk_advRel_d2q9_explicitOmegaSpatial
  public :: mrt_advRel_d2q9
  public :: mrt_advRel_d2q9_explicitEq

  !=============================================================================
  ! D2Q9 flow model
  !=============================================================================
  !> Definition of the discrete velocity set
  integer, parameter :: QQ  = 9   !> number of pdf directions
  integer, parameter :: QQN = 8   !> number of neighbors

  integer,parameter :: q__W     = 1     !< west             x-
  integer,parameter :: q__S     = 2     !< south            y-
  integer,parameter :: q__E     = 3     !< east             x+
  integer,parameter :: q__N     = 4     !< north            y+
  integer,parameter :: q_SW     = 5     !<                  y-,x-
  integer,parameter :: q_NW     = 6     !<                  y+,x-
  integer,parameter :: q_SE     = 7     !<                  y-,x+
  integer,parameter :: q_NE     = 8     !<                  y+,x+
  integer,parameter :: q__0     = 9     !< rest density is last

   ! MUSUBI D2Q9 MRT
   real(kind=rk), dimension(9,9),parameter  :: MMtrD2Q9 = &
   reshape((/ &
     1._rk,     1._rk,     1._rk,     1._rk,     1._rk,     1._rk,     1._rk,     1._rk,     1._rk,&
    -1._rk,    -1._rk,    -1._rk,    -1._rk,     2._rk,     2._rk,     2._rk,     2._rk,    -4._rk,&
    -2._rk,    -2._rk,    -2._rk,    -2._rk,     1._rk,     1._rk,     1._rk,     1._rk,     4._rk,&
    -1._rk,     0._rk,     1._rk,     0._rk,    -1._rk,    -1._rk,     1._rk,     1._rk,     0._rk,&
     2._rk,     0._rk,    -2._rk,     0._rk,    -1._rk,    -1._rk,     1._rk,     1._rk,     0._rk,&
     0._rk,    -1._rk,     0._rk,     1._rk,    -1._rk,     1._rk,    -1._rk,     1._rk,     0._rk,&
     0._rk,     2._rk,     0._rk,    -2._rk,    -1._rk,     1._rk,    -1._rk,     1._rk,     0._rk,&
     1._rk,    -1._rk,     1._rk,    -1._rk,     0._rk,     0._rk,     0._rk,     0._rk,     0._rk,&
     0._rk,     0._rk,     0._rk,     0._rk,     1._rk,    -1._rk,    -1._rk,     1._rk,     0._rk &
             /),(/9,9/))


    real(kind=rk), dimension(9,9),parameter  :: MMIvD2Q9 =                &
            reshape((/                                                    &
1._rk/9._rk, -1._rk/36._rk, -1._rk/18._rk,-1._rk/6._rk,                   & 
      1._rk/6._rk,     0._rk,     0._rk,  0.25_rk,    0._rk,              & ! W
 1._rk/9._rk, -1._rk/36._rk, -1._rk/18._rk, 0._rk,                        & 
      0._rk,    -1._rk/6._rk,  1._rk/6._rk, -0.25_rk, 0._rk,              & ! S
 1._rk/9._rk, -1._rk/36._rk, -1._rk/18._rk, 1._rk/6._rk,                  & 
     -1._rk/6._rk,    0._rk,     0._rk,  0.25_rk,    0._rk,               & ! E
 1._rk/9._rk, -1._rk/36._rk, -1._rk/18._rk, 0._rk   ,                     & 
      0._rk,     1._rk/6._rk, -1._rk/6._rk, -0.25_rk,    0._rk,           & ! N
 1._rk/9._rk,  1._rk/18._rk,  1._rk/36._rk,-1._rk/6._rk,                  &
     -1._rk/12._rk,-1._rk/6._rk, -1._rk/12._rk,    0._rk, 0.25_rk,        & ! SW
 1._rk/9._rk,  1._rk/18._rk,  1._rk/36._rk,-1._rk/6._rk,                  &
     -1._rk/12._rk, 1._rk/6._rk,  1._rk/12._rk,    0._rk,-0.25_rk,        & ! NW
 1._rk/9._rk,  1._rk/18._rk,  1._rk/36._rk, 1._rk/6._rk,                  &
      1._rk/12._rk,-1._rk/6._rk, -1._rk/12._rk,    0._rk, -0.25_rk,       & ! SE
 1._rk/9._rk,  1._rk/18._rk,  1._rk/36._rk, 1._rk/6._rk,                  &
      1._rk/12._rk, 1._rk/6._rk,  1._rk/12._rk,    0._rk, 0.25_rk,        & ! NE
 1._rk/9._rk, -1._rk/9._rk,   1._rk/9._rk,  0._rk,                        &
      0._rk,    0._rk,     0._rk,     0._rk,    0._rk                     & ! 0
                      /),(/9,9/))
 

contains

!******************************************************************************!
  !> Advection relaxation routine for the D2Q9 model with BGK for 
  !! spatially varying relaxation parameter omega
  !! Definition in the configuration file according to 
  !! \ref mus_aux_module::mus_setspatialomega "spatial omega"
  !!
?? copy :: compute_routineHeader( bgk_advRel_d2q9_explicitOmegaSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem,iDir                       ! voxel element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)  ! temporary local pdf values
    real(kind=rk) equil(layout%stencil(1)%QQ)   ! temporary equilibrium values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) usq, ucx     ! square velocity
    real(kind=rk) omega, omSrc ! relaxation parameter
    !---------------------------------------------------------------------------

    ! Read the reference relaxation parameter for the current level
    omSrc   = fieldProp(1)%fluid%rampOmegas( level )

!$omp do schedule(static)
    nodeloop: do iElem=nStart,nEnd       
      omega   = fieldProp(1)%fluid%viscSpatial%stored( level )%val(iElem)*omSrc
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW) 
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE) 
                         
      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y

      do iDir = 1, layout%stencil(1)%QQ
        ! velocity times lattice unit velocity
        ucx = dble( layout%stencil(1)%cxDir( 1, iDir ) ) * u_x                 &
          & + dble( layout%stencil(1)%cxDir( 2, iDir ) ) * u_y 
    
        ! calculate equilibrium density
        equil( iDir ) = layout%weight( iDir ) * rho * ( 1._rk + ucx*cs2inv     &
          &           + ucx*ucx*cs2inv*cs2inv*0.5_rk                           &
          &           - usq*cs2inv*0.5_rk )
    
        outState(?SAVE?(iDir, iElem, QQ, neigh)) =                             &
          &                pdfTmp(iDir) - omega* (pdfTmp(iDir) - equil( iDir ))
      enddo
    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d2q9_explicitOmegaSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D2Q9 model with BGK.
  !! 
  !! Streaming is defined in the header/lbm_macros.inc in STREAM_PUSH = .false.
  !! This routine ONLY contains pure computation
  !! and should be implemented in a very efficient way 
  !! \f$ f_\alpha(x_i+e_{\alpha,i},t+1) = 
  !! f_\alpha(x_i,t) - \omega(f_\alpha(x_i,t)-f^{eq}_{\alpha}(x_i,t)) \f$  
  !!
?? copy :: compute_routineHeader( bgk_advRel_d2q9_explicit_fixedVel )
    !---------------------------------------------------------------------------
    integer :: iElem,iDir                       ! voxel element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)  ! temporary local pdf values
    real(kind=rk) equil(layout%stencil(1)%QQ)   ! temporary equilibrium values
    real(kind=rk) rho     ! local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) usq, ucx     ! square velocity
    real(kind=rk) omega ! relaxation parameter
    integer, parameter :: iField = 1
    !---------------------------------------------------------------------------

    ! Read the relaxation paraemter omega for the current level
    omega   = fieldProp(1)%fluid%rampOmegas( level )

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__0)

      ! read velocity from IC field
      u_x = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(1)%       &
        &                                          stored( level )%val( iElem )
      u_y = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(2)%       &
        &                                          stored( level )%val( iElem )

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y

      do iDir = 1, layout%stencil(1)%QQ
        ! velocity times lattice unit velocity
        ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk ) * u_x        &
          & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk ) * u_y 
    
        ! calculate equilibrium density
        equil( iDir ) = layout%weight( iDir ) * rho * ( 1._rk + ucx*cs2inv     &
          &           + ucx*ucx*cs2inv*cs2inv*0.5_rk                           &
          &           - usq*cs2inv*0.5_rk )
    
        outState(?SAVE?(iDir, iElem, QQ, neigh)) =                             &
          &                pdfTmp(iDir) - omega* (pdfTmp(iDir) - equil( iDir ))
      enddo
    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d2q9_explicit_fixedVel
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D2Q9 model with BGK 
  !! for the incompressible lbm model.
  !! 
  !! Streaming is defined in the header/lbm_macros.inc in STREAM_PUSH = .false.
  !! This routine ONLY contains pure computation
  !! and should be implemented in a very efficient way 
  !! \f$ f_\alpha(x_i+e_{\alpha,i},t+1) = 
  !! f_\alpha(x_i,t) - \omega(f_\alpha(x_i,t)-f^{eq}_{\alpha}(x_i,t)) \f$  
  !!
?? copy :: compute_routineHeader( bgk_advRel_d2q9_explicit_incomp_fixedVel )
    !---------------------------------------------------------------------------
    integer :: iElem,iDir                       ! voxel element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)  ! temporary local pdf values
    real(kind=rk) equil(layout%stencil(1)%QQ)   ! temporary equilibrium values
    real(kind=rk) rho      ! local density
    real(kind=rk) rho0     ! referece density
    real(kind=rk) inv_rho0 ! inverse local density
    real(kind=rk) u_x      ! local x-velocity
    real(kind=rk) u_y      ! local y-velocity
    real(kind=rk) usq, ucx ! square velocity
    real(kind=rk) omega ! relaxation parameters
    integer, parameter :: iField = 1
    !---------------------------------------------------------------------------

    ! some global variables
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    rho0 = fieldProp(1)%fluid%rho0
    ! inverse local density
    inv_rho0 = 1._rk / rho0

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__0)

      ! read velocity from IC field
      u_x = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(1)%       &
        &                                          stored( level )%val( iElem )
      u_y = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(2)%       &
        &                                          stored( level )%val( iElem )

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y
      !MH: for debugging, not correct so far.
      !equil = getEquilibrium( rho, (/u_x,u_y,0._rk/), layout )

      do iDir = 1, layout%stencil(1)%QQ
        ! velocity times lattice unit velocity
        ucx = dble( layout%stencil(1)%cxDir( 1, iDir ) ) * u_x                 &
          & + dble( layout%stencil(1)%cxDir( 2, iDir ) ) * u_y 
    
        ! calculate equilibrium density
        equil( iDir ) = layout%weight( iDir ) * ( rho + rho0*(ucx*cs2inv       &
          &           + ucx*ucx*cs2inv*cs2inv*0.5_rk                           &
          &           - usq*cs2inv*0.5_rk ) )
    
      !enddo
      !do iDir = 1, QQ
        outState(?SAVE?(iDir, iElem, QQ, neigh)) =                             &
          &                pdfTmp(iDir) - omega* (pdfTmp(iDir) - equil( iDir ))
      enddo
    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d2q9_explicit_incomp_fixedVel
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D2Q9 model with MRT 
  !! the equilibrium moments are taken from the pdf equilibriums
  !! if the given eq moments from the Lallemand 2000 MRT paper is used
  !! then there is a deviation
?? copy :: compute_routineHeader( mrt_advRel_d2q9 )
    !---------------------------------------------------------------------------
    integer :: iElem,iDir,ll                       ! voxel element counter
    real(kind=rk) ::  pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: mout(layout%stencil(1)%QQ)    ! equilibrium moments
    real(kind=rk) :: meq(layout%stencil(1)%QQ)     ! equilibrium moments
    real(kind=rk) :: mneq(layout%stencil(1)%QQ)    ! non-equilibrium moments
    real(kind=rk) :: s_mrt(layout%stencil(1)%QQ)   ! relaxation parameters
    real(kind=rk) :: fTmpOut(layout%stencil(1)%QQ) ! intermediate pdf value
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: drho    ! density fluctuation
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    !---------------------------------------------------------------------------
    ! some global variables
    s_mrt(:) = fieldProp(1)%fluid%mrt( level )%s_mrt(:)
    
    nodeloop: do iElem=nStart,nEnd       
      !> First load all local values into temp array
      pdfTmp(  Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp(  Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp(  Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp(  Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp(  Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp(  Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp(  Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp(  Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp(  Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW) 
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE) 
      ! If the velocity field should be fixed, get the velocity from somewhere 
      ! else!
      
      ! Incompressible
      !drho = (rho - fieldProp(1)%fluid%rho0)
      drho = rho
      meq(1) =  rho ! rho
      meq(2) =  -2._rk*drho + 3._rk*(u_x*u_x + u_y*u_y) ! e
      meq(3) =  drho - 3._rk*(u_x*u_x + u_y*u_y) ! eps
      meq(4) =  u_x ! jx
      meq(5) = -u_x ! qx
      meq(6) =  u_y ! jy
      meq(7) = -u_y ! qy
      meq(8) =  u_x*u_x - u_y*u_y ! pxx
      meq(9) =  u_x*u_y ! pxy
      mneq(:) = 0.0_rk
      do iDir = 1, layout%stencil(1)%QQ
         ! Berechnung der  nicht-gg-Werte im Momentenraum
         do ll=1, layout%stencil(1)%QQ
            mneq( iDir ) = mneq( iDir ) + pdfTmp( ll )*MMtrD2Q9( ll, iDir )
         end do
      end do !m ok
      mout(:) = s_mrt( :)*(mneq( :) - meq( :))

      fTmpOut(:)  = 0.0_rk
      do iDir = 1, layout%stencil(1)%QQ
        ! back transformation 
        ! calculation of non-equilibrium moments        
        do ll=1,layout%stencil(1)%QQ
           fTmpOut(iDir)  = fTmpOut(iDir)                                      &
             &            + MMIvD2Q9(ll, iDir)*s_mrt( ll)*( mneq(ll)-meq(ll) ) 
        end do
      end do

      outState(?SAVE?( Q__0, iElem, QQ, neigh))=pdfTmp(Q__0 ) - fTmpOut(Q__0)
      outState(?SAVE?( Q__E, iElem, QQ, neigh))=pdfTmp(Q__E ) - fTmpOut(Q__E)
      outState(?SAVE?( Q__N, iElem, QQ, neigh))=pdfTmp(Q__N ) - fTmpOut(Q__N)
      outState(?SAVE?( Q__W, iElem, QQ, neigh))=pdfTmp(Q__W ) - fTmpOut(Q__W)
      outState(?SAVE?( Q__S, iElem, QQ, neigh))=pdfTmp(Q__S ) - fTmpOut(Q__S)
      outState(?SAVE?( Q_NE, iElem, QQ, neigh))=pdfTmp(Q_NE ) - fTmpOut(Q_NE)
      outState(?SAVE?( Q_NW, iElem, QQ, neigh))=pdfTmp(Q_NW ) - fTmpOut(Q_NW)
      outState(?SAVE?( Q_SW, iElem, QQ, neigh))=pdfTmp(Q_SW ) - fTmpOut(Q_SW)
      outState(?SAVE?( Q_SE, iElem, QQ, neigh))=pdfTmp(Q_SE ) - fTmpOut(Q_SE)
    enddo nodeloop

  end subroutine mrt_advRel_d2q9
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D2Q9 model with MRT 
  !! the equilibrium moments are taken from the pdf equilibriums
  !! if the given eq moments from the Lallemand 2000 MRT paper is used
  !! then there is a deviation
  !! WARNING: Somehow, when the relaxation parameters for the collision 
  !! invariants( density and momentum ) are set to zero, the code crashes.
  !! This should be valid, however, as the invariants do not change during a
  !! collision. 
?? copy :: compute_routineHeader( mrt_advRel_d2q9_explicitEq_mlb )
    !--------------------------------------------------------------------------- 
    integer :: iElem,iDir,ll                      ! voxel element counter
    real(kind=rk) ::  pdfTmp(layout%stencil(1)%QQ)! temporary local pdf values
    real(kind=rk) :: equil(layout%stencil(1)%QQ)  ! temporary equilibrium values
    real(kind=rk) :: mout(layout%stencil(1)%QQ)   ! equilibrium moments
    real(kind=rk) :: meq(layout%stencil(1)%QQ)    ! equilibrium moments
    real(kind=rk) :: mneq(layout%stencil(1)%QQ)   ! non-equilibrium moments
    real(kind=rk) :: s_mrt(layout%stencil(1)%QQ)  ! relaxation parameters
    real(kind=rk) :: fTmpOut(layout%stencil(1)%QQ)! intermediate pdf value
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: usq, ucx
    !--------------------------------------------------------------------------- 
    ! set all relaxation parameters to omega -> bgk
    ! Something is going wrong with the direct equilibrium values.
    s_mrt(:) = fieldProp(1)%fluid%mrt( level )%s_mrt(:)
    
    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      pdfTmp(  Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp(  Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp(  Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp(  Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp(  Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp(  Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp(  Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp(  Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp(  Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
!      pdfTmp( : ) = (/ 0.11111111191331_rk,  &
!         0.113324411115526_rk, &
!         0.11144571110917_rk,  &
!         0.1234551111115526_rk, &
!         0.024777777807804_rk, &
!         0.028977777807804_rk, &
!         0.025377777777683_rk, &
!         0.026777777777683_rk, &
!         0.452144444454952_rk/) 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW) 
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE) 

      ! square of velocity
      usq = u_x*u_x + u_y*u_y 
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx =   real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk) * u_x       &
            & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk) * u_y

        ! calculate equilibrium density
        equil( iDir ) = layout%weight( iDir ) * rho * ( 1.0d0 + ucx*cs2inv     &
          &           + ucx*ucx*t2cs4inv - usq*t2cs2inv )

      enddo

      do iDir = 1, layout%stencil(1)%QQ
         meq( iDir ) = 0.0_rk
         ! Berechnung der  nicht-gg-Werte im Momentenraum
         do ll=1, layout%stencil(1)%QQ
            meq( iDir ) = meq( iDir ) + equil( ll )*MMtrD2Q9( ll, iDir )
         end do
      end do

      mneq(:) = 0.0_rk
      do iDir = 1, layout%stencil(1)%QQ
         ! Berechnung der  nicht-gg-Werte im Momentenraum
         do ll=1, layout%stencil(1)%QQ
            mneq( iDir ) = mneq( iDir )                                        &
              &          + pdfTmp( ll )*real(MMtrD2Q9( ll, iDir ), kind = rk )
         end do
      end do !m ok
      mout(:) = s_mrt( :)*(mneq( :) - meq( :))
      fTmpOut(:)  = 0.0_rk
      do iDir = 1, layout%stencil(1)%QQ
        ! back transformation 
        ! calculation of non-equilibrium moments        
        do ll=1,layout%stencil(1)%QQ
           fTmpOut(iDir)  = fTmpOut(iDir)                                      &
             &            + MMIvD2Q9(ll, iDir)*s_mrt( ll)*( mneq(ll)-meq(ll) ) 
        end do
      end do
      outState(?SAVE?( Q__0, iElem, QQ, neigh))=pdfTmp(Q__0 ) - fTmpOut(Q__0)
      outState(?SAVE?( Q__E, iElem, QQ, neigh))=pdfTmp(Q__E ) - fTmpOut(Q__E)
      outState(?SAVE?( Q__N, iElem, QQ, neigh))=pdfTmp(Q__N ) - fTmpOut(Q__N)
      outState(?SAVE?( Q__W, iElem, QQ, neigh))=pdfTmp(Q__W ) - fTmpOut(Q__W)
      outState(?SAVE?( Q__S, iElem, QQ, neigh))=pdfTmp(Q__S ) - fTmpOut(Q__S)
      outState(?SAVE?( Q_NE, iElem, QQ, neigh))=pdfTmp(Q_NE ) - fTmpOut(Q_NE)
      outState(?SAVE?( Q_NW, iElem, QQ, neigh))=pdfTmp(Q_NW ) - fTmpOut(Q_NW)
      outState(?SAVE?( Q_SW, iElem, QQ, neigh))=pdfTmp(Q_SW ) - fTmpOut(Q_SW)
      outState(?SAVE?( Q_SE, iElem, QQ, neigh))=pdfTmp(Q_SE ) - fTmpOut(Q_SE)
    enddo nodeloop

  end subroutine mrt_advRel_d2q9_explicitEq_mlb
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D2Q9 model with MRT 
  !! the equilibrium moments are taken from the pdf equilibriums
  !! if the given eq moments from the Lallemand 2000 MRT paper is used
  !! then there is a deviation
  !! This is the one from glbe.m  
  !! http://www2.fiu.edu/~sanwa001/Site/LB%20Code.html
  !! 
  !! WARNING: Somehow, when the relaxation parameters for the collision 
  !! invariants( density and momentum ) are set to zero, the code crashes.
  !! This should be valid, however, as the invariants do not change during a
  !! collision. 
  !! Also, the MRT is less stable than bgk, which also points to some bug.
?? copy :: compute_routineHeader( mrt_advRel_d2q9_explicitEq )
    !---------------------------------------------------------------------------
    integer :: iElem,iDir,ll                       ! voxel element counter
    real(kind=rk) :: pdfTmp(QQ)  ! temporary local pdf values
    real(kind=rk) :: mpdf(QQ)    ! equilibrium moments
    real(kind=rk) :: mout(QQ)    ! equilibrium moments
    real(kind=rk) :: meq(QQ)     ! equilibrium moments
    real(kind=rk) :: s_mrt(QQ)   ! relaxation parameters
    real(kind=rk) :: fTmpOut(QQ) ! intermediate pdf value
    real(kind=rk) :: rho      ! local density
    real(kind=rk) :: u_x, j_x ! local x-velocity
    real(kind=rk) :: u_y, j_y ! local y-velocity
    !---------------------------------------------------------------------------

    s_mrt(:) = fieldProp(1)%fluid%mrt( level )%s_mrt(:)

    nodeloop: do iElem = nStart, nEnd
      ! First load all local values into temp array
      pdfTmp(  Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp(  Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp(  Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp(  Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp(  Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp(  Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp(  Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp(  Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp(  Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW) &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S) &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__0)

      ! local x-, y-velocity
      j_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E) &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)

      j_y = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW) &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S) - pdfTmp(Q_SE)

      u_x = j_x / rho
      u_y = j_y / rho

      meq(1) =  rho
      meq(2) =  -2._rk * rho + 3._rk * (u_x*u_x + u_y*u_y) ! e
      meq(3) =  rho - 3._rk*(u_x*u_x + u_y*u_y) ! eps
      meq(4) =  u_x ! jx
      meq(5) = -u_x ! qx
      meq(6) =  u_y ! jy
      meq(7) = -u_y ! qy
      meq(8) =  u_x*u_x - u_y*u_y ! pxx
      meq(9) =  u_x*u_y ! pxy

      mpdf(:) = 0.0_rk
      do iDir = 1, QQ
         do ll = 1, QQ
            mpdf(iDir) = mpdf(iDir) + pdfTmp(ll) * MMtrD2Q9(ll,iDir)
         end do
      end do !m ok
      mout(:) = mpdf(:) - s_mrt(:) * ( mpdf(:) - meq(:) )

      fTmpOut(:)  = 0.0_rk
      do iDir = 1, QQ
        ! back transformation
        do ll = 1, QQ
           fTmpOut(iDir) = fTmpOut(iDir) + MMIvD2Q9(ll,iDir)*mout(ll)
        end do
      end do

      outState(?SAVE?( Q__0, iElem, QQ, neigh))=fTmpOut(Q__0)
      outState(?SAVE?( Q__E, iElem, QQ, neigh))=fTmpOut(Q__E)
      outState(?SAVE?( Q__N, iElem, QQ, neigh))=fTmpOut(Q__N)
      outState(?SAVE?( Q__W, iElem, QQ, neigh))=fTmpOut(Q__W)
      outState(?SAVE?( Q__S, iElem, QQ, neigh))=fTmpOut(Q__S)
      outState(?SAVE?( Q_NE, iElem, QQ, neigh))=fTmpOut(Q_NE)
      outState(?SAVE?( Q_NW, iElem, QQ, neigh))=fTmpOut(Q_NW)
      outState(?SAVE?( Q_SW, iElem, QQ, neigh))=fTmpOut(Q_SW)
      outState(?SAVE?( Q_SE, iElem, QQ, neigh))=fTmpOut(Q_SE)
    enddo nodeloop

  end subroutine mrt_advRel_d2q9_explicitEq
!******************************************************************************!


end module mus_d2q9_module
!******************************************************************************!
