!******************************************************************************!
!> This module provides the MUSUBI specific functions for calculating
!! macroscopic quantities from the state variables for incompressible LBM
!! models.\n
!! Notice that only those quantities that related to density should have a
!! formula which differs from normal LBM model.
!!
!! The depending common interface between MUSUBI and ATELES is defined in the
!! tem_derived_module. The functionality for accessing a variable from the state
!! and evaluating a lua function are also provided in the tem_derived module.
!!
!! \author Kannan Masilamani, Jiaxing Qi
!!
?? include 'header/lbm_macros.inc'
?? include 'treelm/source/deriveMacros.inc'
module mus_derQuanIncomp_module

  ! include treelm modules
  use tem_param_module,        only: div1_2, div1_3, div1_54, div1_9, div3_4,  &
    &                                sqrt3, cs2inv, cs2
  use env_module,              only: rk, long_k, globalMaxLevels
  use tem_var_system_module,   only: tem_var_system_type
  use tem_variable_module,     only: tem_variable_type
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_topology_module,     only: tem_levelOf
  use tem_time_module,         only: tem_time_type
  use treelmesh_module,        only: treelmesh_type
  use tem_subTree_type_module, only: tem_subTree_type, tem_treeIDfrom_subTree

  ! include musubi modules
  use mus_solSpecHelpers_module, only: getWeights, getFieldVariable
  use mus_derivedQuantities_module2, only: secondMom

  ! include aotus modules
  use aotus_module, only: flu_State

  implicit none

  private

  public :: derivePressureIncomp_FromState
  public :: deriveVelocityIncomp_FromState
  public :: deriveEquilIncomp_FromState
  public :: deriveNonEquilIncomp_FromState
  public :: deriveEquilIncomp_FromMacro
  public :: deriveKineEnerIncomp_FromState
  public :: deriveStrainRateIncomp_FromState
  public :: deriveStrainRateIncompnNwtn_FromState

contains


!******************************************************************************!
!       Subroutines with common interface for the function pointers            !
!******************************************************************************!

!******************************************************************************!
  !> Calculate the pressure of a given set of elements (sum up all links).
  !!
  !! Pressure calculation according to the isentropic equation of state for 
  !! the LBM \f$ p = \rho c_s^2 \f$
  !! with the calculation of density as in deriveDensity_FromState
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(derivePressureIncomp_FromState)
    !---------------------------------------------------------------------------
    !local variable
    integer :: iDep, dep, iElem, buf_start, buf_end, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens, rho0
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'pressure',                           &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      dens = 0.0_rk
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )

        ! set the start and end position in the state array
        buf_start = (iElem - 1)*globSys%nScalars      &
          &   + globSys%variable%val( dep )%varPos( 1 )

        buf_end   = (iElem - 1)*globSys%nScalars          &
          &   + globSys%variable%val( dep )%varPos(       &
          &       globSys%variable%val( dep )%nComponents )

        ! derive the density and store it in local_res
        ! derive the density and store it in the input (for more than
        ! 1 species (iDep) this will be the mixture density)
        dens = dens + sum( input( buf_start : buf_end ))

      end do  ! iDep
      local_res( iElem ) = ( dens - rho0 ) * cs2
    end do  ! iElem

  end subroutine derivePressureIncomp_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the velocity of a given element number with the given input
  !!        vector (sum up all values) for incompressible model
  !! \f$ \vec{u} = \frac{\sum(f_i c_i)}{\rho_0} \f$
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveVelocityIncomp_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem, nElems, iDep, iDir, pos, dep, iComp
    real(kind=rk) :: rho0, vel( requVar%nComponents )
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'velMag',                             &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    do iElem = 1, nElems

      ! initialize the local velocity for every element
      vel = 0._rk
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )

        do iDir = 1, globSys%variable%val( dep )%nComponents
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( dep )%varPos( iDir )
          do iComp = 1,requVar%nComponents
            vel(iComp) = vel(iComp) + input( (iElem - 1) *                     &
              &                              globSys%nScalars + pos )          &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do ! iComp
        end do ! iDir
      end do ! iDep

      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &    = vel / rho0

    end do ! iElem

  end subroutine deriveVelocityIncomp_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! state vector for incompressible model
  !! \f$ \rho0 = 1.0 \f$
  !! \f$ f_eq = w_\alpha*(\rho + \rho0( (u . c_i) + (u . c_i)^2 - u^2 )  \f$)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveEquilIncomp_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens !< temporary array for density
    real(kind=rk) :: vel(3) !< temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: posBgn, posEnd
    real(kind=rk) :: ucx, usq
    real(kind=rk) :: rho0
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'equilFromState',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents

    ! Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk

      ! set the begin and end position in the state array to compute density
      posBgn = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( 1 )
      posEnd = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( posBgn:posEnd ) )

      ! Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / rho0

      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)

      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)

        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) =                             &
          &   weights( iDir ) * ( dens + rho0 * (                              &
          &                              ucx * cs2inv                          &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk ) )
      end do ! iDir
    end do ! iElem

  end subroutine deriveEquilIncomp_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! vector.
  !!
  !! input array contains density and velocity in the following order
  !! |dens_1 | dens_2 | ... | dens_nElems | u_x_1 | u_y_1 | u_z_1 | ....
  !! .. | u_x_nElems | u_y_nElems | u_z_nElems |
  !! \f$ \rho0 = 1.0 \f$
  !! \f$ f_eq = w_\alpha*(\rho + \rho0( (u . c_i) + (u . c_i)^2 - u^2 )  \f$)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveEquilIncomp_FromMacro)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: ucx, usq
    real(kind=rk) :: rho0
    integer :: iDir, iElem, posVel, offVel, nElems
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'equilFromState',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    offVel = nElems   !< offset of velocity components in input array

    do iElem = 1, nElems
      ! square of velocity
      posVel = (iElem-1) * 3 + offVel ! position of this element's vel in input array
      usq =   input( posVel+1 ) * input( posVel+1 )                            &
        &   + input( posVel+2 ) * input( posVel+2 )                            &
        &   + input( posVel+3 ) * input( posVel+3 )
      
      do iDir = 1, stencil%QQ
      
        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk)*input( posVel+1 )    &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*input( posVel+2 )    &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*input( posVel+3 )
      
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * ( input( iElem ) + rho0 * ( ucx*cs2inv                   &
          &                                    + ucx*ucx*cs2inv*cs2inv*0.5_rk  &
          &                                    - usq*cs2inv*0.5_rk ) )
      
      enddo ! iDir
    end do ! iElem

  end subroutine deriveEquilIncomp_FromMacro
!******************************************************************************!


!******************************************************************************!
  !> Calculate the Non-Equlibrium of given elements with given input state
  !! array.
  !!
?? copy :: derive_headtxt(deriveNonEquilIncomp_FromState)
    !---------------------------------------------------------------------------
    !> local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens !< temporary array for density
    real(kind=rk) :: vel(3) !< temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: posBgn, posEnd
    real(kind=rk) :: ucx, usq
    real(kind=rk) :: rho0
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'equilFromState',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents
    ! Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk

      ! set the begin and end position in the state array to compute density
      posBgn =    (iElem-1) * globSys%nScalars            &
        &       + globSys%variable%val( dep )%varPos( 1 )
      posEnd =    (iElem-1) * globSys%nScalars            &
        &       + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( posBgn:posEnd ) )

      ! Calculate velocity
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / rho0

      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)

      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)

        ! calculate Non-Equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) =                             &
          &   input( (iElem-1)*stencil%QQ + iDir )                             &
          & - weights( iDir ) * ( dens + rho0 * (                              &
          &                              ucx * cs2inv                          &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk ) )
      end do ! iDir
    end do ! iElem

  end subroutine deriveNonEquilIncomp_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the strain rate ( or rate of strain, or rate of deformation)
  !! \author Jiaxing Qi
  !!
  !! The formula is:
  !! \f[
  !!  \tau_{\alpha \beta}=
  !!    -\frac{3\omega}{2{\rho}_0} \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta}
  !! \f]
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibrium pdf.\n
  !! For more information, please refer to: equation 45 in\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.\n
  !!
  !! For multi-level mesh, Omega on finer level needs to be adjusted in order to
  !! get the correct shearstress calculation.\n
  !! First, we defines c as the dx ratio between finer and coarse level.\n
  !! \f$ c={ \Delta dx }_{ c }/{ \Delta dx }_{ f } \f$
  !! Then the viscosity on the different levels must satisfy:\n
  !! \f$ \frac { { \nu  }_{ f } }{ { \nu  }_{ c } } =c \f$
  !! This constrain leads to a relationship of omega on different levels:\n
  !! \f$ {\omega}_f = \frac {1}{ {\lambda}(\frac{1}{{\omega}_c}-0.5)+0.5 } \f$
  !! For more information, please refer to:\n
  !! Manuel H, Harald K, Joerg B, Sabine R. Aeroacoustic validation of the
  !! lattice boltzmann method on non-uniform grids. ECCOMAS 2012
  !!
?? copy :: derive_headtxt(deriveStrainRateIncomp_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: tau(requVar%nComponents)
    real(kind=rk), allocatable :: nonEq(:)
    integer       :: iElem, iLevel, buf_start, buf_end, nElems
    integer       :: nonEq_pos, pdf_pos, pdf_nComps
    real(kind=rk) :: omega( globalMaxLevels), iOmega, rho0
    integer(kind=long_k), allocatable :: treeID(:)
    integer :: localBound(2)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( nonEq( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    ! Get all omegas from all levels
    omega(:) = getFieldVariable( conf = globSys%solSpec%conf,                  &
      &                       varLabel = trim(requVar%label) ,                 &
      &                       varName = 'shearstress',                         &
      &                       fieldVar = 'omLvl',                              &
      &                       fieldProp = 'fluid', nVals = globalMaxLevels )

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'velMag',                             &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! position of pdf in glob system
    pdf_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents
    ! position of nonEquilibrium in glob system
    nonEq_pos = globSys%variable%val( requVar%mapping )%varDep( 2 )

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem
      nonEq = 0._rk

      ! set the begin and end position in the input state array
      buf_start = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! Calculate nonEquilibrium density
      ! for PULL, this is post-collision Non-Eq
      ! for PUSH, this is  pre-collision Non-Eq
      call globSys%derive( nonEq_pos )%do(                         &
        &                    input   = input(buf_start:buf_end),   &
        &                    bound   = localBound,                 &
        &                    tree    = tree,                       &
        &                    subTree = subTree,                    &
        &                    requVar = globSys%variable%val(nonEq_pos) ,&
        &                    globSys = globSys,                    &
        &                    stencil = stencil,                    &
        &                    conf    = conf,                       &
        &                    res     = nonEq  )

      ! get my current level
      iLevel = tem_levelOf( treeID( iElem ) )
      ! get the correct omega value
      iOmega = omega( iLevel )

      ! Shear stress requires pre-collision NonEq
      ! Convert post-coll to pre-coll by a factor of 1/(1-omega), if use PULL
?? if( .not. STREAM_PUSH ) then
      if( iOmega /= 1._rk ) then
        nonEq = nonEq / ( 1._rk - iOmega )
      else
        nonEq = 0._rk
      endif
?? endif

      ! compute shear stress
      tau(:) = secondMom( stencil%cxcx(:,:), nonEq(:), stencil%QQ )

      local_res( (iElem-1)*6+1 : iElem*6 ) = -1.5_rk * iOmega / rho0 * tau(:)
    enddo ! iElem

  end subroutine deriveStrainRateIncomp_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the strain rate ( or rate of strain, or rate of deformation)
  !! for nonNewtonian model.
  !! \author Jiaxing Qi
  !!
  !! The following needs to be updated!
  !! The formula is:
  !! \f[
  !!  \tau_{\alpha \beta}=
  !!    -\frac{3\omega}{2{\rho}_0} \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta}
  !! \f]
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibrium pdf.\n
  !! For more information, please refer to: equation 45 in\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.\n
  !!
  !! For multi-level mesh, Omega on finer level needs to be adjusted in order to
  !! get the correct shearstress calculation.\n
  !! First, we defines c as the dx ratio between finer and coarse level.\n
  !! \f$ c={ \Delta dx }_{ c }/{ \Delta dx }_{ f } \f$
  !! Then the viscosity on the different levels must satisfy:\n
  !! \f$ \frac { { \nu  }_{ f } }{ { \nu  }_{ c } } =c \f$
  !! This constrain leads to a relationship of omega on different levels:\n
  !! \f$ {\omega}_f = \frac {1}{ {\lambda}(\frac{1}{{\omega}_c}-0.5)+0.5 } \f$
  !! For more information, please refer to:\n
  !! Manuel H, Harald K, Joerg B, Sabine R. Aeroacoustic validation of the
  !! lattice boltzmann method on non-uniform grids. ECCOMAS 2012
  !!
?? copy :: derive_headtxt(deriveStrainRateIncompnNwtn_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: tau(requVar%nComponents)
    real(kind=rk), allocatable :: nonEq(:)
    integer :: iElem, iLevel, buf_start, buf_end, nElems
    integer :: nonEq_pos, nonEq_nComps, pdf_pos, pdf_nComps
    integer :: omega_pos, omega_nComps
    real(kind=rk) :: omega, rho0
    integer(kind=long_k), allocatable :: treeID(:)
    integer :: localBound(2)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if


    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    !get reference density
    rho0 = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'velMag',                             &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! position of pdf in glob system
    pdf_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents
    ! omega position in glob system
    omega_pos = globSys%variable%val( requVar%mapping )%varDep( 2 )
    omega_nComps = globSys%variable%val( omega_pos )%nComponents
    ! position of nonEquilibrium in glob system
    nonEq_pos = globSys%variable%val( requVar%mapping )%varDep( 3 )
    nonEq_nComps = globSys%variable%val( nonEq_pos )%nComponents

    allocate( nonEq( nonEq_nComps ))

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem
      nonEq = 0._rk

      ! set the begin and end position in the input state array
      buf_start = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! Calculate nonEquilibrium density
      ! for PULL, this is post-collision Non-Eq
      ! for PUSH, this is  pre-collision Non-Eq
      call globSys%derive( nonEq_pos )%do(                         &
        &                    input   = input(buf_start:buf_end),   &
        &                    bound   = localBound,                 &
        &                    tree    = tree,                       &
        &                    subTree = subTree,                    &
        &                    requVar = globSys%variable%val(nonEq_pos) ,&
        &                    globSys = globSys,                    &
        &                    stencil = stencil,                    &
        &                    conf    = conf,                       &
        &                    res     = nonEq  )

      ! get omega from state vector
      omega = input( (iElem-1) * globSys%nScalars + &
        &               globSys%variable%val(omega_pos)%varPos(omega_nComps) )


      ! Shear stress requires pre-collision NonEq
      ! Convert post-coll to pre-coll by a factor of 1/(1-omega), if use PULL
?? if( .not. STREAM_PUSH ) then
      if( omega /= 1._rk ) then
        nonEq = nonEq / ( 1._rk - omega )
      else
        nonEq = 0._rk
      endif
?? endif

      ! compute shear stress
      tau(:) = secondMom( stencil%cxcx(:,:), nonEq(:), stencil%QQ )

      local_res( (iElem-1)*6+1 : iElem*6 ) = -1.5_rk * omega / rho0 * tau(:)
    enddo ! iElem

  end subroutine deriveStrainRateIncompnNwtn_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the kinetic energy as
  !! \f[
  !!  E = \frac{1}{2} \rho ( u_x^2 + u_y^2 + u_z^2 )
  !! \f]
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveKineEnerIncomp_FromState)
    !---------------------------------------------------------------------------
    !> local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens ! temporary array for density
    real(kind=rk) :: vel(3) ! temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: buf_start, buf_end
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    !get reference density
    dens = getFieldVariable( conf      = globSys%solSpec%conf,                 &
      &                      varLabel  = trim(requVar%label),                  &
      &                      varName   = 'kinetic_energy',                     &
      &                      fieldVar  = 'rho0',                               &
      &                      fieldProp = 'fluid' )

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents
    ! Calculation of density and velocity
    do iElem = 1, nElems
      vel = 0._rk

      ! set the begin and end position in the state array to compute density
      buf_start = (iElem-1) * globSys%nScalars                                 &
        &   + globSys%variable%val( dep )%varPos( 1 )
      buf_end = (iElem-1) * globSys%nScalars                                   &
        &   + globSys%variable%val( dep )%varPos( nComp )

      ! Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / dens

      local_res( iElem ) = 0.5_rk*dens*(                                       &
        &       vel(1)*vel(1) + vel(2)*vel(2) + vel(3)*vel(3) )
    end do ! iElem

  end subroutine deriveKineEnerIncomp_FromState
!******************************************************************************!


end module mus_derQuanIncomp_module
!******************************************************************************!
