!******************************************************************************!
!> This module provides the MUSUBI specific functions for calculating
!! macroscopic quantities from the state variables multispecies.
!!
!! The depending common interface between MUSUBI and ATELES is defined in the
!! tem_derived_module. The functionality for accessing a variable from the state
!! and evaluating a lua function are also provided in the tem_derived module.
!!
!! \author Kannan Masilmani, Simon Zimny
!!
?? include 'treelm/source/deriveMacros.inc'
module mus_derQuanMSGas_module

  ! include treelm modules
  use env_module,                only: rk, long_k, PathLen, labelLen
  use tem_var_system_module,     only: tem_var_system_type
  use tem_variable_module,       only: tem_variable_type
  use tem_stencil_module,        only: tem_stencilHeader_type
  use tem_param_module,          only: cs2inv, t2cs4inv
  use tem_time_module,           only: tem_time_type
  use treelmesh_module,          only: treelmesh_type
  use tem_subTree_type_module,   only: tem_subTree_type, tem_treeIDfrom_subTree
  use mus_solSpecHelpers_module, only: getWeights, getNFields, getFieldVariable

  ! include aotus modules
  use aotus_module, only: flu_State

  implicit none

  private

  ! functions for multispecies
  public :: deriveMolfracMSGas_FromState
  public :: deriveVelocityMSGas_FromState
  public :: deriveVelMagMSGas_FromState
  public :: deriveEquilMSGas_FromState
  public :: deriveEquilMSGas_FromMacro
  public :: deriveEquilMSGas_2d_FromState
  public :: deriveEquilMSGas_2d_FromMacro

contains


!******************************************************************************!
!       Subroutines with common interface for the function pointers            !
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity of a given element number with the given input
  !!        vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveVelocityMSGas_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: iDep, iDir, pos, dep, iComp
    integer :: buf_start, buf_end, nElems
    ! array to store the intermediate results for vel and dens
    real(kind=rk) :: dens, vel( requVar%nComponents )
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !\todo solve system of equation for multispecies velocity
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      vel = 0._rk
      dens = 0._rk ! initialize the local density variable
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
      
        ! \todo:SZ: is this really optimal or better calculate dens in the
        !           direction loop from the links?????
      
        ! set the start and end position in the input array to compute the
        ! density
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos( 1 )
      
        buf_end   = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos(                            &
          &                            globSys%variable%val( dep )%nComponents )
      
        ! derive the density and store it in the input (for more than
        ! 1 species (iDep) this will be the mixture density)
        dens = dens + sum( input( buf_start : buf_end ))
      
        do iDir = 1, globSys%variable%val( dep )%nComponents
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( dep )%varPos( iDir )
          do iComp = 1,requVar%nComponents
            vel(iComp) = vel(iComp) + input( (iElem - 1)                       &
              &                              * globSys%nScalars + pos )        &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !icomp
        end do !idir
      end do !idep
      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &     = vel/dens
    end do

  end subroutine deriveVelocityMSGas_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the velocity magnitude of a given element number with the given 
  !! input vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveVelMagMSGas_FromState)
    !---------------------------------------------------------------------------
    ! local variables
    ! array to store the intermediate results for vel and dens
    integer :: iElem
    integer :: iDep, iDir, pos, dep
    integer :: buf_start, buf_end, nElems
    real(kind=rk) :: dens, vel( 3 )
    ! array to store the intermediate results for vel and dens
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      ! initialize the local velocity for every element
      vel = 0._rk
      dens = 0._rk ! initialize the local density variable
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
      
        ! \todo:SZ: is this really optimal or better calculate dens in the
        !           direction loop from the links?????
      
        ! set the start and end position in the input array to compute the
        ! density
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos( 1 )
      
        buf_end   = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos(                            &
          &                            globSys%variable%val( dep )%nComponents )
      
        ! derive the density and store it in the input (for more than
        ! 1 species (iDep) this will be the mixture density)
        dens = dens + sum( input( buf_start : buf_end ))
      
        do iDir = 1, globSys%variable%val( dep )%nComponents
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( dep )%varPos( iDir )
          vel = vel + input( (iElem - 1) * globSys%nScalars + pos )            &
            & * real(stencil%cxDir( :, iDir ), kind=rk)
        end do !idir
      end do ! idep
      vel = vel/dens
      ! copy the values to the right position in the inout array input
      local_res(iElem) = sqrt(vel(1)*vel(1)+vel(2)*vel(2)+vel(3)*vel(3))
    end do

  end subroutine deriveVelMagMSGas_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! state vector.
  !!
  !! The equilibrium distribution function is:\n
  !! \f$ f^{eq}_rest = \rho w_i( (3-2 mRatio_iField ) + 3(cx . u) 
  !!            + 9/2 (cx . u )^2 - 3/2 u^2 ) \f$
  !! \f$ f^{eq}_i = \rho w_i( mRatio_iField + 3(cx . u)
  !!            + 9/2 (cx . u )^2 - 3/2 u^2 ) \f$
  !! where \f$w_i\f$ is the weight in each direction,\n
  !! \f$\rho\f$ is the macroscopic value of density,\n
  !! \f$c_s\f$ is the speed of sound,\n
  !! \f$\vec c_i\f$ is the lattice unit velocity in each direction,\n
  !! \f$\vec u\f$ is the macroscopic value of velocity.
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveEquilMSGas_FromState)
    !---------------------------------------------------------------------------
    ! local variables
    ! array to store the intermediate results for vel and dens
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    ! molecular weight ratio
    real(kind=rk) :: m_ratio
    real(kind=rk) :: dens ! temporary array for density
    real(kind=rk) :: vel(3) ! temporary array for velocity
    integer :: iDir, iElem, dep, nComp
    integer :: posBgn, posEnd, nElems
    real(kind=rk) :: ucx, usq
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------
  
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))
    
    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get molecular weight ratio
    m_ratio = getFieldVariable( conf = globSys%solSpec%conf,                   &
      &                         varLabel = trim(requVar%label),                &
      &                         varName  = 'equilFromState',                   &
      &                         fieldVar = 'molWeigRatio',                     &
      &                         fieldProp = 'species' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )
  
    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents
    !> Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk
      
      ! set the begin and end position in the state array to compute density
      posBgn = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( 1 )
      posEnd = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( posBgn:posEnd ) )
      
      !> Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / dens
      
      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)
      
      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)
      
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &                   * dens * ( m_ratio  + ucx * cs2inv               &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk )
      end do ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*stencil%QQ + stencil%restPosition ) =               &
        &                     weights( stencil%restPosition ) * dens           &
        &                   * ( (3._rk - 2._rk*m_ratio)                        &
        &                   - usq*cs2inv*0.5_rk )
    end do ! iElem

  end subroutine deriveEquilMSGas_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! vector.
  !!
  !! input array contains density and velocity in the following order
  !! |dens_1 | dens_2 | ... | dens_nElems | u_x_1 | u_y_1 | u_z_1 | ....
  !! .. | u_x_nElems | u_y_nElems | u_z_nElems |
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveEquilMSGas_FromMacro)
    !---------------------------------------------------------------------------
    ! local variables
    ! array to store the intermediate results for vel and dens
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    ! molecular weight ratio
    real(kind=rk) :: m_ratio
    real(kind=rk) :: ucx, usq
    integer :: iDir, iElem, posVel, offVel, nElems
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------
  
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))
  
    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get molecular weight ratio
    m_ratio = getFieldVariable( conf = globSys%solSpec%conf,                   &
      &                         varLabel = trim(requVar%label),                &
      &                         varName  = 'equilFromMacro',                   &
      &                         fieldVar = 'molWeigRatio',                     &
      &                         fieldProp = 'species' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )
  
    offVel = nElems   ! offset of velocity components in input array
  
    do iElem = 1, nElems
      ! position of this element's vel in input array
      posVel = (iElem-1) * 3 + offVel
      ! square of velocity
      usq =   input( posVel+1 ) * input( posVel+1 )                            &
        &   + input( posVel+2 ) * input( posVel+2 )                            &
        &   + input( posVel+3 ) * input( posVel+3 )
      
      do iDir = 1, stencil%QQ
      
        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk)*input( posVel+1 )    &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*input( posVel+2 )    &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*input( posVel+3 )
      
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * input( iElem ) * ( m_ratio  + ucx * cs2inv               &
          &                          + ucx*ucx*cs2inv*cs2inv*0.5_rk            &
          &                          - usq*cs2inv*0.5_rk )
      
      enddo ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*stencil%QQ + stencil%restPosition ) =               &
        &                     weights( stencil%restPosition ) * input( iElem ) &
        &                   * ( (3._rk - 2._rk*m_ratio)                        &
        &                   - usq*cs2inv*0.5_rk )
    end do ! iElem

  end subroutine deriveEquilMSGas_FromMacro
!******************************************************************************!

  !> Calculate the equlibrium of a given element number with the given input
  !! state vector for 2D layout.
  !!
  !! This function is used only for single field not for mixtures
  !! Never evaluate the equilibrium for mixtures
  !! The equilibrium distribution function is:\n
  !! \f$ f^{eq}_rest = \rho w_i( (9-5 mRatio_iField )/4 + 3(cx . u)
  !!            + 9/2 (cx . u )^2 - 3/2 u^2 ) \f$
  !! \f$ f^{eq}_i = \rho w_i( mRatio_iField + 3(cx . u)
  !!            + 9/2 (cx . u )^2 - 3/2 u^2 ) \f$
  !! where \f$w_i\f$ is the weight in each direction,\n
  !! \f$\rho\f$ is the macroscopic value of density,\n
  !! \f$c_s\f$ is the speed of sound,\n
  !! \f$\vec c_i\f$ is the lattice unit velocity in each direction,\n
  !! \f$\vec u\f$ is the macroscopic value of velocity.
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveEquilMSGas_2d_FromState)
    !---------------------------------------------------------------------------
    ! local variables
    ! array to store the intermediate results for vel and dens
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    ! molecular weight ratio
    real(kind=rk) :: m_ratio
    real(kind=rk) :: dens ! temporary array for density
    real(kind=rk) :: vel(3) ! temporary array for velocity
    integer :: iDir, iElem, dep, nComp
    integer :: posBgn, posEnd, nElems
    real(kind=rk) :: ucx, usq
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------
  
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get molecular weight ratio
    m_ratio = getFieldVariable( conf = globSys%solSpec%conf,                   &
      &                         varLabel = trim(requVar%label),                &
      &                         varName  = 'equilFromState',                   &
      &                         fieldVar = 'molWeigRatio',                     &
      &                         fieldProp = 'species' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )
  
    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents
    !> Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk
      
      ! set the begin and end position in the state array to compute density
      posBgn = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( 1 )
      posEnd = (iElem-1) * globSys%nScalars                                    &
        &   + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( posBgn:posEnd ) )
      
      !> Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / dens
      
      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)
      
      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)
      
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &                   * dens * ( m_ratio  + ucx * cs2inv               &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk )
      end do ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*stencil%QQ + stencil%restPosition ) =               &
        &                     weights( stencil%restPosition ) * dens           &
        &                   * ( (9._rk - 5._rk*m_ratio)/4.0_rk                 &
        &                   - usq*cs2inv*0.5_rk )
    end do ! iElem

  end subroutine deriveEquilMSGas_2d_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! state vector for 2D layout.
  !!
  !! input array contains density and velocity in the following order
  !! |dens_1 | dens_2 | ... | dens_nElems | u_x_1 | u_y_1 | u_z_1 | ....
  !! .. | u_x_nElems | u_y_nElems | u_z_nElems |
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveEquilMSGas_2d_FromMacro)
    !---------------------------------------------------------------------------
    !> array to store the intermediate results for vel and dens
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    ! molecular weight ratio
    real(kind=rk) :: m_ratio
    real(kind=rk) :: ucx, usq
    integer :: iDir, iElem, posVel, offVel, nElems
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------
  
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))
  
    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    !get molecular weight ratio
    m_ratio = getFieldVariable( conf = globSys%solSpec%conf,                   &
      &                         varLabel = trim(requVar%label),                &
      &                         varName  = 'equilFromMacro',                   &
      &                         fieldVar = 'molWeigRatio',                     &
      &                         fieldProp = 'species' )

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    offVel = nElems   ! offset of velocity components in input array

    do iElem = 1, nElems
      ! position of this element's vel in input array
      posVel = (iElem-1) * 3 + offVel
      ! square of velocity
      usq =   input( posVel+1 ) * input( posVel+1 )                            &
        &   + input( posVel+2 ) * input( posVel+2 )                            &
        &   + input( posVel+3 ) * input( posVel+3 )
      
      do iDir = 1, stencil%QQ
      
        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk)*input( posVel+1 )    &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*input( posVel+2 )    &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*input( posVel+3 )
      
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * input( iElem ) * ( m_ratio  + ucx * cs2inv               &
          &                          + ucx*ucx*cs2inv*cs2inv*0.5_rk            &
          &                          - usq*cs2inv*0.5_rk )
      
      enddo ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*stencil%QQ + stencil%restPosition ) =               &
        &                     weights( stencil%restPosition ) * input( iElem ) &
        &                   * ( (9._rk - 5._rk*m_ratio)/4.0_rk                 &
        &                   - usq*cs2inv*0.5_rk )
    end do ! iElem
  
  end subroutine deriveEquilMSGas_2d_FromMacro
!******************************************************************************!


!******************************************************************************!
  !> Calculate the molefraction of a given element number with the given state
  !!        vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveMolfracMSGas_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk), allocatable :: m_ratio(:), psigma(:)
    real(kind=rk) :: ptol, molfrac
    integer :: iDep, iElem, iField, nFields, buf_start, buf_end, nElems
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! \todo: SZ: get molecular weight ratio from sol_specstring?? or only nFields??
    nFields = globSys%nVars!getNFields( globSys%solSpec%conf )

    allocate(m_ratio(nFields))
    do iField = 1, nFields
      m_ratio(iField) = getFieldVariable( conf = globSys%solSpec%conf,         &
        &                varLabel = trim(globSys%variable%val(iField)%label),  &
        &                varName  = 'pdf',                                     &
        &                fieldVar = 'molWeigRatio',                            &
        &                fieldProp = 'species' )
    end do

    !partial pressure
    allocate(psigma(nFields))
    do iElem = 1, nElems
      psigma = 0.0_rk
      molfrac = 0.0_rk
      do iField = 1, nFields
        ! set the start and end position in the state array
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( iField )%varPos( 1 )
      
        buf_end   = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( iField )%varPos(                         &
          &                         globSys%variable%val( iField )%nComponents )
      
        !partialPressur, p_1=rho_1*m_ratio/3.0_rk
        psigma(iField) = sum(input( buf_start : buf_end )) * m_ratio(iField)   &
          &            / 3.0_rk
      end do
      !total partial pressure
      ptol = sum(psigma)
      
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        molfrac = molfrac + psigma( globSys%variable%val(                      &
          &                              requVar%mapping )%varDep(iDep) ) / ptol
      enddo
      local_res( iElem ) = molfrac
    end do

  end subroutine deriveMolfracMSGas_FromState
!******************************************************************************!


end module mus_derQuanMSGas_module
!******************************************************************************!
