!******************************************************************************!
!> This module provides the MUSUBI specific functions for calculating
!! macroscopic quantities from the state variables for multispecies liquid model.
!!
!! The depending common interface between MUSUBI and ATELES is defined in the
!! tem_derived_module. The functionality for accessing a variable from the state
!! and evaluating a lua function are also provided in the tem_derived module.
!!
!! \author Kannan Masilamani
!!
?? include 'treelm/source/deriveMacros.inc'
module mus_derQuanMSLiquid_module
  ! include treelm modules
  use env_module,              only: rk, long_k, globalMaxLevels, labelLen
  use tem_param_module,        only: div1_2, div1_3, div1_54, div1_9, div3_4,  &
    &                                sqrt3, cs2inv, cs2, t2cs2inv, t2cs4inv,   &
    &                                cs4inv
  use tem_aux_module,          only: tem_abort
  use tem_var_system_module,   only: tem_var_system_type
  use tem_variable_module,     only: tem_variable_type
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_logging_module,      only: logUnit
  use tem_topology_module,     only: tem_levelOf
  use tem_time_module,         only: tem_time_type
  use treelmesh_module,        only: treelmesh_type
  use tem_math_module,         only: invert_matrix
  use tem_spatial_module,      only: tem_spatial_for
  use tem_subTree_type_module, only: tem_subTree_type, tem_treeIDfrom_subTree
  use tem_debug_module,        only: dbgUnit

  use mus_solSpecHelpers_module, only: getWeights, getFieldVariable,           &
    &                                  getVariable_FromTable, getFieldPrefixes
  use mus_eNRTL_module,          only: mus_calc_thermFactor,                   &
    &                                  mus_calc_MS_DiffMatrix                 

  ! include aotus modules
  use aotus_module, only: flu_State

  implicit none

  private

  public :: deriveNumDensityMSLiquid_FromState
  public :: derivePressureMSLiquid_FromState
  public :: deriveKinePressMSLiquid_FromState
  public :: deriveMoleFracMSLiquid_FromState
  public :: deriveMoleFluxMSLiquid_FromState
  public :: deriveVelocityMSLiquid_FromState
  public :: deriveMomentumMSLiquid_FromState
  public :: deriveVelMagMSLiquid_FromState
  public :: deriveEquilMSLiquid_FromState
  public :: deriveEquilMSLiquid_2d_FromState
  !KM : equilibrium from macro is same as gas. remove this
  public :: deriveEquilMSLiquid_FromMacro
  public :: deriveEquilMSLiquid_2d_FromMacro
  public :: deriveEquilVelMSLiquid_FromMacro
  public :: deriveEquilVelMSLiquid_FromState

  ! routines which uses thermodynamic factor and variable diffusivities
!  public :: deriveVelocityWTDF_MSLiquid_FromState
!  public :: deriveEquilWTDF_MSLiquid_FromState
!  public :: deriveEquilWTDF_MSLiquid_2d_FromState
!  public :: deriveEquilWTDF_MSLiquid_FromMacro
!  public :: deriveEquilWTDF_MSLiquid_2d_FromMacro

!  public :: deriveEquilVelFromMacro
!  public :: deriveMomentumFromMacroFromLSE

contains

!******************************************************************************!
!       Subroutines with common interface for the function pointers            !
!******************************************************************************!

!******************************************************************************!
  !> Calculate the number density of a given element for single species or mixture
  !! with the given input state vector for liquid mixture.
  !! Number density = density/molecular weight
  !! mixture number density = sum(number_density)
?? copy :: derive_headtxt(deriveNumDensityMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: molWeigh( globSys%nVars ) !, phi(globSys%nVars)
    !mass density of nSpecies
    real(kind=rk) :: mass_dens
!    real(kind=rk) :: mass_dens( globSys%nVars ), dens
    !number density of nSpecies 
    real(kind=rk) :: num_dens
!    real(kind=rk) :: moleDens0, tot_massDens, kinepress
    integer :: iElem, iDep, dep, nFields, iField, nElems
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars
    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeigh(iField) = getFieldVariable( conf = globSys%solSpec%conf,        &
        &           varLabel = trim(fieldPrefix(iField))//'numDensity',        &
        &           varName = 'numDensity',                                    &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      !phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
      !  &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
      !  &           varName = 'velocity',                                      &
      !  &           fieldVar = 'molWeigRatio',                                 &
      !  &           fieldProp = 'species' )                                
    enddo
    !moleDens0 = getVariable_FromTable( conf = globSys%solSpec%conf,               &
    !  &                            varLabel = 'moleDens0',                        &
    !  &                            table = 'mixture')

    ! loop over elements
    do iElem = 1, nElems
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
!      tot_massDens = 0.0_rk
!      do iField = 1, nFields
!        buf_start = (iElem - 1)*globSys%nScalars                               &
!          &       + globSys%variable%val( iField )%varPos(1)
!        buf_end = (iElem-1)*globSys%nScalars                                   &
!          &     + globSys%variable%val( iField )%varPos(                       &
!          &               globSys%variable%val( iField )%nComponents )
!        mass_dens(iField) = sum( input( buf_start : buf_end ) )
!      end do
!
!      tot_massDens = sum(mass_dens)
!      kinePress = cs2 * ( dot_product(phi, mass_dens)  &
!        & - minval(molWeigh)*moleDens0 ) / tot_massDens
!
!      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
!        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
!        dens = mass_dens(dep) - cs2inv * mass_dens(dep) * kinePress/phi(dep)
!        num_dens = num_dens + dens / molWeigh(dep)
!      end do
      
      !> vardep > 1 fir mixture
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( dep )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( dep )%varPos(                       &
          &               globSys%variable%val( dep )%nComponents )
        
        ! mass density of species
        mass_dens = sum( input( buf_start : buf_end ))
        ! number density of species   
        num_dens = num_dens + mass_dens / molWeigh( dep )  
      end do !iDep      
      local_res(iElem) = num_dens
    end do  !iElem

  end subroutine deriveNumDensityMSLiquid_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate mixture kinematic pressure. 
  !! This routine requires initial mole fraction of all species which must
  !! be provided through add_variable table in scheme and set depvar in tracking
  !! for kinematic pressure to the newly added variable.
  !! This will be used to evaluate spatial function for initial mole fraction
  !! from initial mole fraction initial mixture molar density will be computed.
  !! Formula to compute kinematic pressure
  !! \f$ p = c^2_s (\sum_k \rho_k \phi_k - min_l (m_l) n_0)/\rho_0 \f$
  !! here, \f$ \rho_k \f$ - species density, \\
  !! \f$ \phi_k \f$ - species molecular weight ratio, \\
  !! \f$ n_0 \f$ - mixture number density,\\
  !! \f$ \rho_0 \f$ - reference density.
  !! Example:
  !! In scheme table,
  !! \verbatim
  !! add_variable = { name = "initialMoleFrac", ncomponents=1,')
  !!   initialMoleFrac = { { kind="combined", spatial=luaFunc}
  !!                      { kind="combined", spatial=luaFunc2}}
  !! }
  !! \endverbatim
  !! In tracking,
  !! \verbatim
  !! variable = {{"kinematicpressure", 1, dep = { "initialMoleFrac"}}}
  !! \endverbatim
  !! \todo use constant total number density
?? copy :: derive_headtxt(deriveKinePressMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: molWeigh( globSys%nVars )
    real(kind=rk) :: rho0, moleDens0, min_molWeight
    integer :: iElem, nFields, iField, press_pos, dens_pos, nElems
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    real(kind=rk), allocatable :: tot_massDens(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeigh(iField) = getFieldVariable( conf = globSys%solSpec%conf,        &
        &           varLabel = trim(fieldPrefix(iField))//'kinematicPressure', &
        &           varName = 'kinematicPressure',                             &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
    enddo
    min_molWeight = minval(molWeigh)

    moleDens0 = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'moleDens0',                     &
      &                            table = 'mixture')
    ! KM:\todo use tot_massDens of each element for rho0
    rho0 = getVariable_FromTable( conf = globSys%solSpec%conf,                 &
      &                            varLabel = 'rho0',                          &
      &                            table = 'mixture')

    ! point to pressure derive quantities which return \rho*cs2
    press_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    if( size( globSys%variable%val( press_pos )%varDep) /= nFields) then
      write(logUnit(1),*)' ERROR: Kinematic pressure can be derived only '//   &
        &                'for mixture'
      call tem_abort()
    endif
    ! mixture pressure
    call globSys%derive( press_pos )%do( input = input,                        &
      &                                  bound = bound,                        &
      &                                  tree = tree,                          &
      &                                  subTree = subTree,                    & 
      &                                  requVar = globSys%variable%val(press_pos),&
      &                                  globSys = globSys,                    &
      &                                  stencil = stencil,                    &
      &                                  res = local_res,                      &
      &                                  conf = conf )      
    
    ! point to density derive quantities which return sum(\rho)
    dens_pos = globSys%variable%val( requVar%mapping )%varDep( 2 )
    ! mixture density
    allocate(tot_massDens(nElems))
    call globSys%derive( dens_pos )%do( input = input,                        &
      &                                  bound = bound,                        &
      &                                  tree = tree,                          &
      &                                  subTree = subTree,                    & 
      &                                  requVar = globSys%variable%val(dens_pos),&
      &                                  globSys = globSys,                    &
      &                                  stencil = stencil,                    &
      &                                  res = tot_massDens,                   &
      &                                  conf = conf )      

    ! loop over elements
    do iElem = 1, nElems
      !mixture kinematic pressure
      ! \f$ p = c^2_s (\sum_k \rho_k \phi_k - min_l (m_l) n_0)/\rho_0 \f$
      !write(*,*) 'kinePress ', local_res(iElem), min_molWeight * moleDens0 * cs2, &
      !  & rho0, tot_massDens(iElem)
      local_res(iElem) = ( local_res(iElem) &
        & - min_molWeight * moleDens0 * cs2 ) / tot_massDens(iElem)

    end do  

  end subroutine deriveKinePressMSLiquid_FromState
!******************************************************************************!



!******************************************************************************!
  !> Calculate species and total pressure both for gas and liquid model 
  !! Note: This routine can also be used to compute species pressure
  !! In case of gas mixture, it is partial pressure where as in
  !! liquid mixture this is not valid. 
  !! Formula to ressure
  !! \f$ p = c^2_s (\sum_k \rho_k \phi_k ) \f$
  !! here, \f$ \rho_k \f$ - species density, \\
  !! \f$ \phi_k \f$ - species molecular weight ratio, \\
?? copy :: derive_headtxt(derivePressureMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: phi(globSys%nVars) !, molWeigh( globSys%nVars )
    !mass density of Species
    real(kind=rk) :: mass_dens
    ! press return
    real(kind=rk) :: press
    real(kind=rk) :: dens
    integer :: iElem, iDep, dep, nFields, iField, nElems
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars
    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'Pressure',          &
        &           varName = 'Pressure',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
    enddo

    ! loop over elements
    do iElem = 1, nElems
      mass_dens = 0.0_rk
      dens = 0.0_rk
      ! return single species number density or mixture number density
      ! depends on size of varDep
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( dep)%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( dep )%varPos(                       &
          &               globSys%variable%val( dep )%nComponents )
        
        ! mass density of species
        dens = sum( input( buf_start : buf_end ))
        mass_dens = mass_dens + dens * phi(dep)
        !write(*,*) dep, 'dens ', dens, dens*phi(dep)
      end do
      
      !single species pressure or mixture pressure
      !write(*,*) 'press ', mass_dens
      press = mass_dens * cs2
   
      !pressure
      local_res(iElem) = press
    end do  

  end subroutine derivePressureMSLiquid_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity of a given element for single species or mixture
  !! with the given input state vector for liquid mixture.
  !! Need to solve the system of equations to compute this velocity since
  !! first order moments gives only moments of transformed pdfs. Hence
  !! to obtain the first order moments of actual pdfs we need to solve
  !! system of equation of size = nSpecies for each velocity components
  !! \todo Add equation and reference
?? copy :: derive_headtxt(deriveVelocityMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: dens, vel( requVar%nComponents )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    !mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    real(kind=rk) :: totNum_densInv
    !first moments of nSpecies
    real(kind=rk) :: first_moments( requVar%nComponents, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( requVar%nComponents, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv
    integer :: iElem, iDep, iDir, pos, dep, iComp, nFields, iField, nElems
    integer :: iFieldDia, iFieldNonDia
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))

        ! number density of species   
        num_dens(iField) = mass_dens(iField) * molWeighInv(iField)

        ! all fields have some nComponents
        do iDir = 1, stencil%QQ 
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          do iComp = 1, requVar%nComponents
            first_moments( iComp, iField ) = first_moments( iComp, iField )    &
              &          + input( (iElem - 1) * globSys%nScalars + pos )       &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !iComp    
        end do !iDir
      end do !iField

      ! total number density Inv
      totNum_densInv = 1.0_rk/sum(num_dens(:))

      ! molefraction
      moleFraction(:) = num_dens(:)*totNum_densInv

      ! build up the equation system for momentum
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do
        
      invA = invert_matrix( matrixA )

      ! momentum of all species
      momentum = 0.0_rk
      do iComp = 1, requVar%nComponents
        momentum( iComp, : ) = matmul( invA, first_moments( iComp, : ) )
      end do

      !momentum = deriveMomentumFromMacroFromLSE( moleFraction = moleFraction,  &
      !  &        first_moments = first_moments,                                &
      !  &        nFields = nFields, phi = phi, resi_coeff = resi_coeff,        &
      !  &        omega_fac = omega_fac, paramBInv = paramBInv,               &
      !  &        nComp = requVar%nComponents )

      ! find required velocity   
      dens = 0.0_rk
      vel = 0.0_rk
      ! varDep >1 only for mixture velocity
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
             
        dens = dens + mass_dens( dep )    
        vel(:) = vel(:) + momentum( :, dep )
      end do !idep

      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = vel/dens
    end do

  end subroutine deriveVelocityMSLiquid_FromState  
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity of a given element for single species or mixture
  !! with the given input state vector for liquid mixture with thermodynamic
  !! factor and variable diffusivity coeffcients.
  !! Need to solve the system of equations to compute this velocity since
  !! first order moments gives only moments of transformed pdfs. Hence
  !! to obtain the first order moments of actual pdfs we need to solve
  !! system of equation of size = nSpecies for each velocity components
  !! \todo Add equation and reference
?? copy :: derive_headtxt(deriveVelocityWTDF_MSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: dens, vel( requVar%nComponents )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    !mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    real(kind=rk) :: totNum_densInv
    !first moments of nSpecies
    real(kind=rk) :: first_moments( requVar%nComponents, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( requVar%nComponents, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv
    integer :: iElem, iDep, iDir, pos, dep, iComp, nFields, iField, nElems
    integer :: iFieldDia, iFieldNonDia
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))

        ! number density of species   
        num_dens(iField) = mass_dens(iField) * molWeighInv(iField)

        ! all fields have some nComponents
        do iDir = 1, stencil%QQ 
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          do iComp = 1, requVar%nComponents
            first_moments( iComp, iField ) = first_moments( iComp, iField )    &
              &          + input( (iElem - 1) * globSys%nScalars + pos )       &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !iComp    
        end do !iDir
      end do !iField

      ! total number density Inv
      totNum_densInv = 1.0_rk/sum(num_dens(:))

      ! molefraction
      moleFraction(:) = num_dens(:)*totNum_densInv

      ! build up the equation system for momentum
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do
        
      invA = invert_matrix( matrixA )

      ! momentum of all species
      momentum = 0.0_rk
      do iComp = 1, requVar%nComponents
        momentum( iComp, : ) = matmul( invA, first_moments( iComp, : ) )
      end do

      !momentum = deriveMomentumFromMacroFromLSE( moleFraction = moleFraction,  &
      !  &        first_moments = first_moments,                                &
      !  &        nFields = nFields, phi = phi, resi_coeff = resi_coeff,        &
      !  &        omega_fac = omega_fac, paramBInv = paramBInv,               &
      !  &        nComp = requVar%nComponents )

      ! find required velocity   
      dens = 0.0_rk
      vel = 0.0_rk
      ! varDep >1 only for mixture velocity
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
             
        dens = dens + mass_dens( dep )    
        vel(:) = vel(:) + momentum( :, dep )
      end do !idep

      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = vel/dens
    end do

  end subroutine deriveVelocityWTDF_MSLiquid_FromState  
!******************************************************************************!


!******************************************************************************!
  !> velocity magnitude
!******************************************************************************!
?? copy :: derive_headtxt(deriveVelMagMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: dens, vel( requVar%nComponents )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    !mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    real(kind=rk) :: totNum_densInv
    !first moments of nSpecies
    real(kind=rk) :: first_moments( requVar%nComponents, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( requVar%nComponents, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv
    integer :: iElem, iDep, iDir, pos, dep, iComp, nFields, iField, nElems
    integer :: ifieldDia, ifieldNonDia
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'velMag',            &
        &           varName = 'velMag',                                        &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'velMag',            &
        &           varName = 'velMag',                                        &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'velMag',            &
        &           varName = 'velMag',                                        &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))

        ! number density of species
        num_dens(iField) = mass_dens(iField) * molWeighInv(iField)

        ! all fields have some nComponents
        do iDir = 1, stencil%QQ 
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          do iComp = 1, requVar%nComponents
            first_moments( iComp, iField ) = first_moments( iComp, iField )    &
              &          + input( (iElem - 1) * globSys%nScalars + pos )       &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !iComp    
        end do !iDir
      end do !iField

      ! total number density Inv
      totNum_densInv = 1.0_rk/sum(num_dens(:))

      ! molefraction
      moleFraction(:) = num_dens(:)*totNum_densInv
       
      ! derive momentum by solving LSE 
      !momentum = deriveMomentumFromMacroFromLSE( moleFraction = moleFraction,  &
      !  &        first_moments = first_moments,                                &
      !  &        nFields = nFields, phi = phi, resi_coeff = resi_coeff,        &
      !  &        omega_fac = omega_fac, paramBInv = paramBInv,                 &
      !  &        nComp = requVar%nComponents )
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do
      invA = invert_matrix( matrixA )

      ! momentum of all species
      momentum = 0.0_rk
      do iComp = 1, requVar%nComponents
        momentum( iComp, : ) = matmul( invA, first_moments( iComp, : ) )
      end do

      ! find required velocity   
      dens = 0.0_rk
      vel = 0.0_rk
      ! varDep >1 only for mixture velocity
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
             
        dens = dens + mass_dens( dep )    
        vel(:) = vel(:) + momentum( :, dep )
      end do !idep
      ! compute velocity from momentum
      vel(:) = vel(:) / dens
      
      ! copy the values to the right position in the inout array input
      local_res(iElem) = sqrt(vel(1)*vel(1)+vel(2)*vel(2)+vel(3)*vel(3))
    end do

  end subroutine deriveVelMagMSLiquid_FromState  
!******************************************************************************!


!******************************************************************************!
  !> Calculate the momentum of a given element for single species or mixture
  !! with the given input state vector for liquid mixture.
  !! Need to solve the system of equations to compute this momentum since
  !! first order moments gives only moments of transformed pdfs. Hence
  !! to obtain the first order moments of actual pdfs we need to solve
  !! system of equation of size = nSpecies for each velocity components
  !! \todo Add equation and reference
?? copy :: derive_headtxt(deriveMomentumMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: mom( requVar%nComponents )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    !mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    real(kind=rk) :: totNum_densInv
    !first moments of nSpecies
    real(kind=rk) :: first_moments( requVar%nComponents, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( requVar%nComponents, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv
    integer :: iElem, iDep, iDir, pos, dep, iComp, nFields, iField, nElems
    integer :: iFieldDia, iFieldNonDia
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))

        ! number density of species   
        num_dens(iField) = mass_dens(iField) * molWeighInv(iField)

        ! all fields have some nComponents
        do iDir = 1, stencil%QQ 
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          do iComp = 1, requVar%nComponents
            first_moments( iComp, iField ) = first_moments( iComp, iField )    &
              &          + input( (iElem - 1) * globSys%nScalars + pos )       &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !iComp    
        end do !iDir
      end do !iField

      ! total number density Inv
      totNum_densInv = 1.0_rk/sum(num_dens(:))

      ! molefraction
      moleFraction(:) = num_dens(:)*totNum_densInv

      ! build up the equation system for momentum
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do
        
      invA = invert_matrix( matrixA )

      ! momentum of all species
      momentum = 0.0_rk
      do iComp = 1, requVar%nComponents
        momentum( iComp, : ) = matmul( invA, first_moments( iComp, : ) )
      end do

      !momentum = deriveMomentumFromMacroFromLSE( moleFraction = moleFraction,  &
      !  &        first_moments = first_moments,                                &
      !  &        nFields = nFields, phi = phi, resi_coeff = resi_coeff,        &
      !  &        omega_fac = omega_fac, paramBInv = paramBInv,               &
      !  &        nComp = requVar%nComponents )

      ! find required momentum 
      mom = 0.0_rk
      ! varDep >1 only for mixture velocity
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
             
        mom(:) = mom(:) + momentum( :, dep )
      end do !idep

      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = mom
    end do

  end subroutine deriveMomentumMSLiquid_FromState  
!******************************************************************************!


!******************************************************************************!
  !> Equilibrium velocity from density and velocity
  !!
  !! input array contains density and velocity in the following order
  !! |dens_1_sp1 | dens_1_sp2 | dens_2_sp1 | dens_2_sp2| ... | dens_nElems_sp1 | 
  !! |dens_nElems_sp2 | u_x_1_sp1 | u_y_1_sp1 | u_z_1_sp1 | 
  !! !u_x_1_sp2 | u_y_1_sp2 | u_z_1_sp2 |....
  !! .. | u_x_nElems_sp1 | u_y_nElems_sp1 | u_z_nElems_sp1 | 
  !! |u_x_nElems_sp2 | u_y_nElems_sp2 | u_z_nElems_sp2 |
?? copy :: derive_headtxt(deriveEquilVelMSLiquid_FromMacro)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: vel( requVar%nComponents, globSys%nVars )
    real(kind=rk) :: eqVel( requVar%nComponents )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    ! total number density
    real(kind=rk) :: totNum_densInv
    ! mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: paramBInv
    integer :: iElem, dep, nFields, iField, nElems
    integer :: offVel, posDens, posVel
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'equilvelocity',     &
        &           varName = 'equilvelocity',                                 &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'equilvelocity',     &
        &           varName = 'equilvelocity',                                 &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'equilvelocity',     &
        &           varName = 'equilvelocity',                                 &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    ! get the position of the depending variable in the global varSys
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )

    !\todo use required nComp for velocity 
    offVel = nElems*nFields
    ! loop over elements
    do iElem = 1, nElems
      posDens = (iElem-1)*nFields
      !num_dens for all fields
      num_dens(:) = input( posDens + 1 : posDens + nFields )*molWeighInv(:)
      totNum_densInv = 1.0_rk/sum( num_dens(:) )
      ! mole fraction
      moleFraction(:) = num_dens(:) * totNum_densInv
      ! velocity nComp = 3
      posVel = offVel + (iElem-1)*nFields*3
      do iField=1,nFields
        vel(:, iField) = input( (posVel + (iField-1)*3 + 1) : (posVel + (iField-1)*3 + 3) )
      end do  

      eqVel(:) = vel(:, dep)
      do iField = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(dep, iField)*phi(dep)*molefraction(iField) &
          &      * ( vel(:, iField) - vel(:, dep) ) * paramBInv
      end do

      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = eqVel

!write(*,*) 'In derive EqVelMacro:'
!write(*,*) 'moleFraction ', moleFraction(dep)
!write(*,*) 'mass_dens ', num_dens(dep)
!write(*,*) 'vel ', vel(:, dep)
!write(*,*) 'eqVel ', eqVel(:)
 
    end do 
  end subroutine deriveEquilVelMSLiquid_FromMacro
!******************************************************************************!


!******************************************************************************!
  !> Equlibrium velocity from macro
  function deriveEquilVelFromMacro( iField, moleFraction, velocity, nFields,   &
    &                               paramBInv, phi, resi_coeff ) result( eqVel )
    !---------------------------------------------------------------------------
    !> current field
    integer, intent(in) :: iField
    !> mole fraction of all species
    real(kind=rk), intent(in) :: moleFraction(nFields)
    !> velocity of all species
    real(kind=rk), intent(in) :: velocity(3,nFields)
    !> number of species
    integer, intent(in) :: nFields
    !> free parameter B
    real(kind=rk), intent(in) :: paramBInv
    !> molecular weight ratio
    real(kind=rk), intent(in) :: phi(nFields)
    !> resistivity coefficients
    real(kind=rk), intent(in) :: resi_coeff(nFields)
    !> return equilibrium velocity
    real(kind=rk) :: eqVel(3)
    !---------------------------------------------------------------------------
    integer :: ifld
    !---------------------------------------------------------------------------
    eqVel(:) = velocity(:,iField)
    do ifld = 1, nFields
      eqVel(:) = eqVel(:) + resi_coeff(ifld)*phi(iField)*molefraction(ifld) &
        &      * ( velocity(:,ifld) - velocity(:,iField) ) * paramBInv
    end do

  end function deriveEquilVelFromMacro
!******************************************************************************!

!******************************************************************************!
  !> derive original velocity of species by solving system of equations of
  !! nSpecies
  function deriveMomentumFromMacroFromLSE( moleFraction, first_Moments, nFields,&
    &                                      phi, resi_coeff, omega_fac,          &
    &                                      paramBInv, nComp ) result(momentum)
    !---------------------------------------------------------------------------
    !> number of species
    integer, intent(in) :: nFields
    !> number of velocity components
    integer, intent(in) :: nComp
    !> molefraction  of all species
    real(kind=rk), intent(in) :: moleFraction(nFields)
    !> velocity of all species
    real(kind=rk), intent(in) :: First_Moments(nComp, nFields)
    !> free parameter B
    real(kind=rk), intent(in) :: paramBInv
    !> molecular weight ratio of iField
    real(kind=rk), intent(in) :: phi(nFields)
    !> resistivity coefficients
    real(kind=rk), intent(in) :: resi_coeff(nFields, nFields)
    !> relaxation parameter
    real(kind=rk), intent(in) :: omega_fac
    !> return actual momentum
    real(kind=rk) :: momentum(nComp, nFields)
    !---------------------------------------------------------------------------
    integer :: ifield, ifieldDia, ifieldNonDia, icomp
    real(kind=rk) :: matrixA( nFields, nFields ), invA( nFields, nFields )
    !---------------------------------------------------------------------------
    ! build up the equation system for momentum
    matrixA = 0.0_rk
    do iField = 1, nFields
      ! set diagonal part
      matrixA( iField, iField ) = 1.0_rk
      do iFieldDia = 1, nFields
        matrixA( iField, iField ) = matrixA( iField, iField )                &
          & + omega_fac * resi_coeff( iField, iFieldDia )          &
          & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
      end do
      ! set nonDiagonal
      do iFieldNonDia = 1,nFields
        matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )     &
          & - omega_fac * resi_coeff( iField, iFieldNonDia )        &
          & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
      end do
    end do
      
    invA = invert_matrix( matrixA )

    ! momentum of all species
    momentum = 0.0_rk
    do iComp = 1, nComp
      momentum( iComp, : ) = matmul( invA, first_Moments( iComp, : ) )
    end do

!    !velocity of all species
!    do iField = 1,nFields
!      vel(:, iField) = vel(:, iField)/density(iField)
!    end do

  end function deriveMomentumFromMacroFromLSE
!******************************************************************************!

!******************************************************************************!
  !> Equilibrium velocity from state
  !! Must compute species velocity of all species by solving LSE
  !!
  !! input array contains pdfs in the following order
  !! |pdf_1_sp1 | pdf_1_sp2 | pdf_2_sp1 | pdf_2_sp2| ... | pdf_nElems_sp1 | 
  !! |pdf_nElems_sp2 | 
?? copy :: derive_headtxt(deriveEquilVelMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: vel( requVar%nComponents, globSys%nVars )
    real(kind=rk) :: eqvel( requVar%nComponents )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    ! total number density
    real(kind=rk) :: totNum_densInv
    ! mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    !first moments of nSpecies
    real(kind=rk) :: first_moments( requVar%nComponents, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( requVar%nComponents, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv
    integer :: iElem, iDir, pos, dep, iComp, nFields, iField, nElems
    integer :: ifieldDia, ifieldNonDia, nComp
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    ! get the position of the depending variable in the global varSys
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))

        ! number density of species   
        num_dens( iField ) = mass_dens( iField ) * molWeighInv( iField )  

        ! all fields have some nComponents
        do iDir = 1, stencil%QQ
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          do iComp = 1, requVar%nComponents
            first_moments( iComp, iField ) = first_moments( iComp, iField )    &
              &          + input( (iElem - 1) * globSys%nScalars + pos )       &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !iComp    
        end do !iDir
      end do !iField

      ! total number density
      totNum_densInv = 1.0_rk/sum( num_dens(:) )
      ! mole fraction
      moleFraction(:) = num_dens(:) * totNum_densInv

      ! build up the equation system for momentum
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do
        
      invA = invert_matrix( matrixA )

      ! momentum of all species
      do iComp = 1, requVar%nComponents
        momentum( iComp, : ) = matmul( invA, first_moments( iComp, : ) )
      end do
       
      !velocity of all species
      do iField = 1,nFields
        vel( :, iField) = momentum( :, iField) / mass_dens(iField)
      end do

!      momentum = deriveMomentumFromMacroFromLSE( mass_dens, first_moments, &
!        &        nFields, phi, resi_coeff, omega_diff, molWeigh, paramB,            &
!        &        requVar%nComponents )

      eqVel(:) = vel(:, dep)
      do iField = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(dep,iField)*phi(dep)*molefraction(iField) &
          &      * ( vel(:, iField) - vel(:, dep) ) * paramBInv
      end do

     ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = eqVel
!write(*,*) 'DEREQVEL eqVel ', eqVel(:)
 
    end do 
  end subroutine deriveEquilVelMSLiquid_FromState
!******************************************************************************!


!******************************************************************************!
  !> Equilibrium from state
?? copy :: derive_headtxt(deriveEquilMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end, nElems
    real(kind=rk) :: vel( 3, globSys%nVars )
    real(kind=rk) :: eqvel( 3 )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    ! mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    real(kind=rk) :: totMass_densInv, totNum_densInv
    real(kind=rk) :: first_moments( 3, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( 3, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ),  &
      &              phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv, theta_eq, theta_eq_spc
    integer :: iElem, iDir, pos, dep, iComp, nFields, iField, nComp
    real(kind=rk) :: ucx, usq, ucxQuadTerm, velAvg(3), velQuadTerm(3)
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    integer :: ifieldDia, ifieldNonDia
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromState',    &
        &           varName = 'equilFromState',                                &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromState',    &
        &           varName = 'equilFromState',                                &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromState',    &
        &           varName = 'equilFromState',                                &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    theta_eq = getVariable_FromTable( conf = globSys%solSpec%conf,             &
      &                            varLabel = 'theta_eq',                      &
      &                            table = 'mixture')
    theta_eq_spc = 1.0_rk - theta_eq  

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    ! get the position of the depending variable in the global varSys
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))

        ! number density of species
        num_dens(iField) = mass_dens(iField) * molWeighInv(iField)

        ! all fields have some nComponents
        do iDir = 1, nComp
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          first_moments( :, iField ) = first_moments( :, iField )    &
            &          + input( (iElem - 1) * globSys%nScalars + pos )       &
            & * real(stencil%cxDir( :, iDir ), kind=rk)
        end do !iDir
      end do !iField

      ! total number density Inv
      totNum_densInv = 1.0_rk/sum(num_dens(:))

      ! molefraction
      moleFraction(:) = num_dens(:)*totNum_densInv

!write(*,*) 'DEREq:mass_dens ', mass_dens(dep)
      ! derive momentum by solving LSE 
      !momentum = deriveMomentumFromMacroFromLSE( moleFraction = moleFraction,  &
      !  &        first_moments = first_moments,                                &
      !  &        nFields = nFields, phi = phi, resi_coeff = resi_coeff,        &
      !  &        omega_fac = omega_fac, paramBInv = paramBInv, nComp = 3 )
      ! build up the equation system for momentum
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do

      invA = invert_matrix( matrixA )

      ! momentum of all species
      momentum = 0.0_rk
      do iComp = 1, 3
        momentum( iComp, : ) = matmul( invA, first_Moments( iComp, : ) )
      end do

      !velocity of all species
      do iField = 1,nFields
        vel( :, iField) = momentum( :, iField) / mass_dens(iField)
      end do

!write(*,*) 'DEREq: vel ', vel(:, iField)
      eqVel(:) = vel(:, dep)
      do iField = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(dep, iField)*phi(dep)*molefraction(iField) &
          &      * ( vel(:, iField) - vel(:, dep) ) * paramBInv
      end do
!write(*,*) 'DEREq:eqVel ', eqVel
!stop
      totMass_densInv = 1.0_rk/sum(mass_dens)
      ! mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens(:), vel(1, :) )*totMass_densInv
      velAvg(2) = dot_product( mass_dens(:), vel(2, :) )*totMass_densInv
      velAvg(3) = dot_product( mass_dens(:), vel(3, :) )*totMass_densInv

      ! velocity in quadratic term of equilibrium
      velQuadTerm(:) = theta_eq*velAvg(:) + theta_eq_spc*eqVel(:)
      
      ! Calculate the square of velocity
      !usq = eqvel(1) * eqvel(1) + eqvel(2) * eqvel(2) + eqvel(3) * eqvel(3)
      ! square of velocity
      usq = ( velQuadTerm(1)*velQuadTerm(1)                                    &
          + velQuadTerm(2)*velQuadTerm(2)                                      &
          + velQuadTerm(3)*velQuadTerm(3) )*t2cs2inv

      do iDir = 1, nComp
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * eqvel(1)           &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * eqvel(2)           &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * eqvel(3)

        ucxQuadTerm =    &
          &      real( stencil%cxDir( 1, iDir ), kind=rk)*velQuadTerm( 1 )     &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*velQuadTerm( 2 )     &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*velQuadTerm( 3 )

        ! calculate equilibrium density
        local_res( (iElem-1)*nComp + iDir ) = weights( iDir )                  &
          &                   * mass_dens(dep) * ( phi(dep)  + ucx * cs2inv    &
          &                            + ucxQuadTerm * ucxQuadTerm * t2cs4inv  &
          &                            - usq )
      end do ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*nComp + stencil%restPosition ) =                    &
        &                     weights( stencil%restPosition ) * mass_dens(dep) &
        &                   * ( (3._rk - 2._rk*phi(dep)) - usq )
    end do !iElem

   end subroutine deriveEquilMSLiquid_FromState
!******************************************************************************!

!******************************************************************************!
  !> Equlibrium from macro
  !!
  !! input array contains density and velocity in the following order
  !! |dens_1_sp1 | dens_1_sp2 | dens_2_sp1 | dens_2_sp2| ... | dens_nElems_sp1 | 
  !! |dens_nElems_sp2 | u_x_1_sp1 | u_y_1_sp1 | u_z_1_sp1 | 
  !! !u_x_1_sp2 | u_y_1_sp2 | u_z_1_sp2 |....
  !! .. | u_x_nElems_sp1 | u_y_nElems_sp1 | u_z_nElems_sp1 | 
  !! |u_x_nElems_sp2 | u_y_nElems_sp2 | u_z_nElems_sp2 |
  !!
  !! u_x, u_y, u_z are velocity of non-transformed pdf obtained
  !! by solving LSE
?? copy :: derive_headtxt(deriveEquilMSLiquid_FromMacro)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: ucx, usq, ucxQuadTerm, velAvg(3), velQuadTerm(3)
    integer :: iDir, iElem, iField, dep
    integer :: nElems, nFields, posDens, posVel, offVel
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    real(kind=rk) :: vel( 3, globSys%nVars ), eqvel( 3 )
    real(kind=rk) :: mass_dens( globSys%nVars ), num_dens( globSys%nVars )
    real(kind=rk) :: moleFraction(globSys%nVars)
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    real(kind=rk) :: theta_eq, theta_eq_spc, paramBInv
    real(kind=rk) :: totMass_densInv, totNum_densInv
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !real(kind=rk), dimension(globSys%nVars, globSys%nVars) :: &
    !  & thermodynamic_fac, inv_thermodyn_fac
    !integer :: iField_2, iField_3  
    !---------------------------------------------------------------------------
 
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))
  
    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars
   
    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )
   
    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromMacro',    &
        &           varName = 'equilFromMacro',                                &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      !get molecular weight ratio
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromMacro',    &
        &           varName = 'equilFromMacro',                                &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromMacro',    &
        &           varName = 'equilFromMacro',                                &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    theta_eq = getVariable_FromTable( conf = globSys%solSpec%conf,             &
      &                            varLabel = 'theta_eq',                      &
      &                            table = 'mixture')
    theta_eq_spc = 1.0_rk - theta_eq  

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )
  
    ! get the position of the depending variable in the global varSys
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )

    offVel = nElems*nFields   ! offset of velocity components in input array
    ! loop over elements
    do iElem = 1, nElems
      posDens = (iElem-1)*nFields
      ! mass density of all fields
      mass_dens(:) = input( posDens + 1 : posDens + nFields )
      ! total mass density
      totMass_DensInv = 1.0_rk/sum(mass_dens)
      ! number density of all fields
      num_dens(:) = mass_dens(:)*molWeighInv(:)
      ! total number density
      totNum_DensInv = 1.0_rk/sum(num_dens)
      ! mole fraction
      moleFraction(:) = num_dens(:) * totNum_densInv

      !call mus_calc_thermFactor( nFields, 298.15_rk, 1.01325e5_rk , moleFraction,&
      !  &                        thermodynamic_fac )

      !inv_thermodyn_fac = invert_matrix( thermodynamic_fac )

      ! position of this element's vel in input array
      posVel = offVel + (iElem-1)*nFields*3
      do iField=1,nFields
        vel(:, iField) = input( (posVel + (iField-1)*3 + 1) : &
          &                     (posVel + (iField-1)*3 + 3) )
      end do  

!write(*,*) dep, ' iElem ', iElem, ' vel ', vel(:, dep)
     ! equilibrium velocity 
      eqVel(:) = vel(:, dep)
!!  write(*,*) 'eqVel ', eqVel
      do iField = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(dep, iField)*phi(dep)*molefraction(iField) &
          &      * ( vel(:, iField) - vel(:, dep) ) * paramBInv
!!  write(*,*) iField, ' eqVel ', eqVel
      end do
      eqVel(:) = mass_dens(dep)*eqVel(:)
!!write(*,*) 'iElem ', iElem, ' eqvel ', eqVel(:)

!      eqVel(:) = mass_dens(dep)*vel( :, dep )
!      do iField_2 = 1, nFields
!        do iField_3 = 1, nFields
!          eqVel( :) = eqVel( : )                                               &
!            &                + inv_thermodyn_fac(dep, iField_2)                &
!            &                * mass_dens(iField_2)                             &
!            &                * resi_coeff( iField_2, iField_3 ) * phi(iField_2)&
!            &                * moleFraction(iField_3)                          &
!            &                * (vel(:, iField_3) - vel(:,iField_2))  &
!            &                * paramBInv 
!        end do    
!      end do  
!write(*,*) dep, 'iElem ', iElem, ' eqvel ', eqVel(:)/mass_dens(dep)

      ! mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens(:), vel(1, :) )*totMass_densInv
      velAvg(2) = dot_product( mass_dens(:), vel(2, :) )*totMass_densInv
      velAvg(3) = dot_product( mass_dens(:), vel(3, :) )*totMass_densInv

      ! velocity in quadratic term of equilibrium
      velQuadTerm(:) = theta_eq*velAvg(:) + theta_eq_spc*eqVel(:)/mass_dens(dep)
      
      ! square of velocity
      usq = ( velQuadTerm(1)*velQuadTerm(1)                                    &
          + velQuadTerm(2)*velQuadTerm(2)                                      &
          + velQuadTerm(3)*velQuadTerm(3) )*t2cs2inv

      do iDir = 1, stencil%QQ
      
        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk)*eqVel( 1 )           &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*eqVel( 2 )           &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*eqVel( 3 )
        ucxQuadTerm =    &
          &      real( stencil%cxDir( 1, iDir ), kind=rk)*velQuadTerm( 1 )     &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*velQuadTerm( 2 )     &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*velQuadTerm( 3 )

        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * ( mass_dens(dep) * ( phi(dep)                            &
          &                          + ucxQuadTerm*ucxQuadTerm*t2cs4inv        &
          &                         - usq ) + ucx * cs2inv )
      
      enddo ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*stencil%QQ + stencil%restPosition ) =               &
        &                     weights( stencil%restPosition ) * mass_dens(dep) &
        &                   * ( (3._rk - 2._rk*phi(dep)) - usq )
    end do ! iElem

  end subroutine deriveEquilMSLiquid_FromMacro
!******************************************************************************!

!******************************************************************************!
  !> Equilibrium from state 2d
?? copy :: derive_headtxt(deriveEquilMSLiquid_2d_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end, nElems
    real(kind=rk) :: vel( 3, globSys%nVars )
    real(kind=rk) :: eqvel( 3 )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    ! mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    real(kind=rk) :: totMass_densInv, totNum_densInv
    !first moments of nSpecies
    real(kind=rk) :: first_moments( 3, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( 3, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv, theta_eq, theta_eq_spc
    integer :: iElem, iDir, pos, dep, iComp, nFields, iField, nComp
    real(kind=rk) :: ucx, usq, ucxQuadTerm, velAvg(3), velQuadTerm(3)
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    integer :: ifieldDia, ifieldNonDia
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromState',    &
        &           varName = 'equilFromState',                                &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromState',    &
        &           varName = 'equilFromState',                                &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromState',    &
        &           varName = 'equilFromState',                                &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    theta_eq = getVariable_FromTable( conf = globSys%solSpec%conf,             &
      &                            varLabel = 'theta_eq',                      &
      &                            table = 'mixture')
    theta_eq_spc = 1.0_rk - theta_eq  

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    ! get the position of the depending variable in the global varSys
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))

        ! number density of species
        num_dens(iField) = mass_dens(iField) * molWeighInv(iField)

        ! all fields have some nComponents
        do iDir = 1, nComp
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          first_moments( :, iField ) = first_moments( :, iField )    &
            &      + input( (iElem - 1) * globSys%nScalars + pos )       &
            & * real(stencil%cxDir( :, iDir ), kind=rk)
        end do !iDir
      end do !iField

      ! total number density Inv
      totNum_densInv = 1.0_rk/sum(num_dens(:))

      ! molefraction
      moleFraction(:) = num_dens(:)*totNum_densInv

      ! derive momentum by solving LSE 
      !momentum = deriveMomentumFromMacroFromLSE( moleFraction = moleFraction,  &
      !  &        first_moments = first_moments,                                &
      !  &        nFields = nFields, phi = phi, resi_coeff = resi_coeff,        &
      !  &        omega_fac = omega_fac, paramBInv = paramBInv, nComp = 3 )
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do

      invA = invert_matrix( matrixA )

      ! momentum of all species
      momentum = 0.0_rk
      do iComp = 1, 3
        momentum( iComp, : ) = matmul( invA, first_Moments( iComp, : ) )
      end do

      ! find required equlibrium velocity   
      do iField = 1,nFields
        vel( :, iField) = momentum( :, iField) / mass_dens(iField)
      end do

      !write(*,*) 'derived density ', mass_dens
      !write(*,*) 'derived vel ', vel

      ! equilibrium velocity
      eqVel(:) = vel(:, dep)
      do iField = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(dep, iField)*phi(dep)*molefraction(iField) &
          &      * ( vel(:, iField) - vel(:, dep) ) * paramBInv
      end do

!write(dbgUnit(1),*) 'EquilFromState: dep ', dep
!write(dbgUnit(1),*) 'EquilFromState: dens ', mass_dens
!write(dbgUnit(1),*) 'EquilFromState: moleFrac ', molefraction
!write(dbgUnit(1),*) 'EquilFromState: resi_coeff ', resi_coeff
!write(dbgUnit(1),*) 'EquilFromState: phi ', phi(dep)
!write(dbgUnit(1),*) 'EquilFromState: paramBInv ', paramBinv
!write(dbgUnit(1),*) 'EquilFromState: vel ', Vel
!write(dbgUnit(1),*) 'EquilFromState: eqVel ', eqVel
!write(dbgUnit(1),*) 'EquilFromState: weights ', weights
!write(dbgUnit(1),*) 'EquilFromState: pdf ', weights*phi(dep)*mass_dens(dep)

      totMass_densInv = 1.0_rk/sum(mass_dens)
      ! mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens(:), vel(1, :) )*totMass_densInv
      velAvg(2) = dot_product( mass_dens(:), vel(2, :) )*totMass_densInv
      velAvg(3) = dot_product( mass_dens(:), vel(3, :) )*totMass_densInv

      ! velocity in quadratic term of equilibrium
      velQuadTerm(:) = theta_eq*velAvg(:) + theta_eq_spc*eqVel(:)
      
      ! square of velocity
      usq = ( velQuadTerm(1)*velQuadTerm(1)                                    &
          + velQuadTerm(2)*velQuadTerm(2)                                      &
          + velQuadTerm(3)*velQuadTerm(3) )*t2cs2inv
 
      ! Calculate the square of velocity
      !usq = eqvel(1) * eqvel(1) + eqvel(2) * eqvel(2) + eqvel(3) * eqvel(3)

      do iDir = 1, nComp
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * eqvel(1)           &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * eqvel(2)           &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * eqvel(3)

        ucxQuadTerm =    &
          &      real( stencil%cxDir( 1, iDir ), kind=rk)*velQuadTerm( 1 )     &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*velQuadTerm( 2 )     &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*velQuadTerm( 3 )

        ! calculate equilibrium density
        local_res( (iElem-1)*nComp + iDir ) = weights( iDir )                  &
          &                   * mass_dens(dep) * ( phi(dep)  + ucx * cs2inv    &
          &                          + ucxQuadTerm*ucxQuadTerm*t2cs4inv - usq )
      end do ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*nComp + stencil%restPosition ) =                    &
        &                     weights( stencil%restPosition ) * mass_dens(dep) &
        &                   * ( (9._rk - 5._rk*phi(dep))/4.0_rk - usq )
    end do !iElem
   end subroutine deriveEquilMSLiquid_2d_FromState
!******************************************************************************!

!******************************************************************************!
  !> Equlibrium from macro 2d
  !!
  !! input array contains density and velocity in the following order
  !! |dens_1_sp1 | dens_1_sp2 | dens_2_sp1 | dens_2_sp2| ... | dens_nElems_sp1 | 
  !! |dens_nElems_sp2 | u_x_1_sp1 | u_y_1_sp1 | u_z_1_sp1 | 
  !! !u_x_1_sp2 | u_y_1_sp2 | u_z_1_sp2 |....
  !! .. | u_x_nElems_sp1 | u_y_nElems_sp1 | u_z_nElems_sp1 | 
  !! |u_x_nElems_sp2 | u_y_nElems_sp2 | u_z_nElems_sp2 |
  !!
  !! u_x, u_y, u_z are velocity of non-transformed pdf obtained
  !! by solving LSE
?? copy :: derive_headtxt(deriveEquilMSLiquid_2d_FromMacro)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: ucx, usq, ucxQuadTerm, velAvg(3), velQuadTerm(3)
    integer :: iDir, iElem, iField, dep
    integer :: nElems, nFields, posDens, posVel, offVel
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    real(kind=rk) :: vel( 3, globSys%nVars ), eqvel( 3 )
    real(kind=rk) :: mass_dens( globSys%nVars ), num_dens( globSys%nVars )
    real(kind=rk) :: moleFraction(globSys%nVars)
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    real(kind=rk) :: theta_eq, theta_eq_spc, paramBInv
    real(kind=rk) :: totMass_densInv, totNum_densInv
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !---------------------------------------------------------------------------
 
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))
  
    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )
   
    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromMacro',    &
        &           varName = 'equilFromMacro',                                &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      !get molecular weight ratio
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromMacro',    &
        &           varName = 'equilFromMacro',                                &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'equilFromMacro',    &
        &           varName = 'equilFromMacro',                                &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    theta_eq = getVariable_FromTable( conf = globSys%solSpec%conf,             &
      &                            varLabel = 'theta_eq',                      &
      &                            table = 'mixture')
    theta_eq_spc = 1.0_rk - theta_eq  

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )
  
    ! get the position of the depending variable in the global varSys
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )

    offVel = nElems*nFields   ! offset of velocity components in input array
    ! loop over elements
    do iElem = 1, nElems
      posDens = (iElem-1)*nFields
      ! mass density of all fields
      mass_dens(:) = input( posDens + 1 : posDens + nFields )
      ! total mass density
      totMass_DensInv = 1.0_rk/sum(mass_dens)
      ! number density of all fields
      num_dens(:) = mass_dens(:)*molWeighInv(:)
      ! total number density
      totNum_DensInv = 1.0_rk/sum(num_dens)
      ! mole fraction
      moleFraction(:) = num_dens(:) * totNum_densInv

      ! position of this element's vel in input array
      posVel = offVel + (iElem-1)*nFields*3
      do iField=1,nFields
        vel(:, iField) = input( (posVel + (iField-1)*3 + 1) : &
          &                     (posVel + (iField-1)*3 + 3) )
      end do  

      ! equilibrium velocity 
      eqVel(:) = vel(:, dep)
      do iField = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(dep, iField)*phi(dep)*molefraction(iField) &
          &      * ( vel(:, iField) - vel(:, dep) ) * paramBInv
        !write(*,*) 'EquilFromMacro: iField ', iField, 'eqVel ', eqVel
      end do
!write(dbgUnit(1),*) 'EquilFromMacro: dep ', dep
!write(dbgUnit(1),*) 'EquilFromMacro: dens ', mass_dens
!write(dbgUnit(1),*) 'EquilFromMacro: moleFrac ', molefraction
!write(dbgUnit(1),*) 'EquilFromMacro: resi_coeff ', resi_coeff
!write(dbgUnit(1),*) 'EquilFromMacro: phi ', phi(dep)
!write(dbgUnit(1),*) 'EquilFromMacro: paramBInv ', paramBinv
!write(dbgUnit(1),*) 'EquilFromMacro: vel ', Vel
!write(dbgUnit(1),*) 'EquilFromMacro: eqVel ', eqVel

      ! mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens(:), vel(1, :) )*totMass_densInv
      velAvg(2) = dot_product( mass_dens(:), vel(2, :) )*totMass_densInv
      velAvg(3) = dot_product( mass_dens(:), vel(3, :) )*totMass_densInv

      ! velocity in quadratic term of equilibrium
      velQuadTerm(:) = theta_eq*velAvg(:) + theta_eq_spc*eqVel(:)
      
      ! square of velocity
      usq = ( velQuadTerm(1)*velQuadTerm(1)                                    &
          + velQuadTerm(2)*velQuadTerm(2)                                      &
          + velQuadTerm(3)*velQuadTerm(3) )*t2cs2inv
      
      do iDir = 1, stencil%QQ
      
        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk)*eqVel( 1 )           &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*eqVel( 2 )           &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*eqVel( 3 )
        ucxQuadTerm =    &
          &      real( stencil%cxDir( 1, iDir ), kind=rk)*velQuadTerm( 1 )     &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*velQuadTerm( 2 )     &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*velQuadTerm( 3 )
     
        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * mass_dens(dep) * ( phi(dep)  + ucx * cs2inv              &
          &                          + ucxQuadTerm*ucxQuadTerm*t2cs4inv - usq )
      
      enddo ! iDir
      !set equilibrium at rest
      local_res( (iElem-1)*stencil%QQ + stencil%restPosition ) =               &
        &                     weights( stencil%restPosition ) * mass_dens(dep) &
        &                   * ( (9._rk - 5._rk*phi(dep))/4.0_rk - usq )
    end do ! iElem

   end subroutine deriveEquilMSLiquid_2d_FromMacro
!******************************************************************************!

!******************************************************************************!
  !> mole fraction
?? copy :: derive_headtxt(deriveMoleFracMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: molWeigh( globSys%nVars )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density of nSpecies 
    real(kind=rk) :: num_dens( globSys%nVars )
    ! mole fraction
    real(kind=rk) :: moleFraction( globSys%nVars )
    ! mole fraction single species or mixture
    real(kind=rk) :: molFrac_loc
    ! total number density
    real(kind=rk) :: totNum_dens
    integer :: iElem, iDep, dep, nFields, iField, nElems
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeigh(iField) = getFieldVariable( conf = globSys%solSpec%conf,        &
        &           varLabel = trim(fieldPrefix(iField))//'molfrac',           &
        &           varName = 'molfrac',                                       &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
    enddo

    ! loop over elements
    do iElem = 1, nElems
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = mass_dens( iField )                              &
          &                 + sum( input( buf_start : buf_end ))
        ! number density of species   
        num_dens( iField ) = mass_dens( iField ) / molWeigh ( iField )  
      end do
      
      ! total number density
      totNum_dens = sum( num_dens(:) )
      ! mole fraction
      moleFraction(:) = num_dens(:) / totNum_dens
 
      molFrac_loc = 0.0_rk
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
        molFrac_loc = molFrac_loc + moleFraction(dep)
      end do

      local_res(iElem) = molFrac_loc
    end do  

  end subroutine deriveMoleFracMSLiquid_FromState
!******************************************************************************!

!******************************************************************************!
  !> mole flux = numDens_i*velocity_i
?? copy :: derive_headtxt(deriveMoleFluxMSLiquid_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer :: buf_start, buf_end
    real(kind=rk) :: dens, ndens, vel( requVar%nComponents )
    !mass density of nSpecies
    real(kind=rk) :: mass_dens( globSys%nVars )
    !number density
    real(kind=rk) :: num_dens( globSys%nVars )
    real(kind=rk) :: moleFraction(globSys%nVars)
    real(kind=rk) :: totNum_densInv
    !first moments of nSpecies
    real(kind=rk) :: first_moments( requVar%nComponents, globSys%nVars )
    !momentum from linear system of equation
    real(kind=rk) :: momentum( requVar%nComponents, globSys%nVars )
    !parameters from solver specific conf
    !field specific info from field table
    real(kind=rk) :: molWeighInv( globSys%nVars ), phi(globSys%nVars)
    real(kind=rk) :: resi_coeff( globSys%nVars, globSys%nVars )
    character(len=labelLen) :: fieldPrefix( globSys%nVars )
    !mixture info
    real(kind=rk) :: omega_diff, omega_fac, paramBInv
    integer :: iElem, iDep, iDir, pos, dep, iComp, nFields, iField, nElems
    integer :: ifieldDia, ifieldNonDia
    real(kind=rk) :: matrixA( globSys%nVars, globSys%nVars )
    real(kind=rk) :: invA( globSys%nVars, globSys%nVars )
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    nFields = globSys%nVars

    fieldPrefix = getFieldPrefixes( conf = globSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iField = 1, nFields
      molWeighInv(iField) = getFieldVariable( conf = globSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      molWeighInv(iField) = 1.0_rk/molWeighInv(iField)  
      phi(iField) = getFieldVariable( conf = globSys%solSpec%conf,             &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iField, :) = getFieldVariable( conf = globSys%solSpec%conf,   &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo

    omega_diff = getVariable_FromTable( conf = globSys%solSpec%conf,           &
      &                            varLabel = 'omega_diff',                    &
      &                            table = 'mixture')
    omega_fac = omega_diff * 0.5_rk

    paramBInv = getVariable_FromTable( conf = globSys%solSpec%conf,            &
      &                            varLabel = 'paramB',                        &
      &                            table = 'mixture')
    paramBInv = 1.0_rk/paramBInv  

    ! loop over elements
    do iElem = 1, nElems
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iField = 1, nFields
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &       + globSys%variable%val( iField )%varPos(1)
        buf_end = (iElem-1)*globSys%nScalars                                   &
          &     + globSys%variable%val( iField )%varPos(                       &
          &               globSys%variable%val( iField )%nComponents )
        
        ! mass density of species
        mass_dens( iField ) = sum( input( buf_start : buf_end ))
        ! number density of species
        num_dens(iField) = mass_dens(iField) * molWeighInv(iField)

        ! all fields have some nComponents
        do iDir = 1, globSys%variable%val( iField )%nComponents
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( iField )%varPos( iDir )
          ! number of required velocity components
          do iComp = 1, requVar%nComponents
            first_moments( iComp, iField ) = first_moments( iComp, iField )    &
              &          + input( (iElem - 1) * globSys%nScalars + pos )       &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !iComp    
        end do !iDir
      end do !iField

      ! total number density Inv
      totNum_densInv = 1.0_rk/sum(num_dens(:))

      ! molefraction
      moleFraction(:) = num_dens(:)*totNum_densInv
      ! momentum 
      !momentum = deriveMomentumFromMacroFromLSE( moleFraction = moleFraction,  &
      !  &        first_moments = first_moments,                                &
      !  &        nFields = nFields, phi = phi, resi_coeff = resi_coeff,        &
      !  &        omega_fac = omega_fac, paramBInv = paramBInv,                 &
      !  &        nComp = requVar%nComponents )
      matrixA = 0.0_rk
      do iField = 1, nFields
        ! set diagonal part
        matrixA( iField, iField ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iField, iField ) = matrixA( iField, iField )                &
            & + omega_fac * resi_coeff( iField, iFieldDia )          &
            & * phi( iField ) * moleFraction( iFieldDia ) * paramBInv
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iField, iFieldNonDia ) = matrixA( iField, iFieldNonDia )    &
            & - omega_fac * resi_coeff( iField, iFieldNonDia )       &
            & * phi( iFieldNonDia ) * moleFraction( iField ) * paramBInv
        end do
      end do

      invA = invert_matrix( matrixA )

      ! momentum of all species
      momentum = 0.0_rk
      do iComp = 1, requVar%nComponents
        momentum( iComp, : ) = matmul( invA, first_Moments( iComp, : ) )
      end do

      ! find required velocity   
      dens = 0.0_rk
      vel = 0.0_rk
      ndens = 0.0_rk
      ! varDep >1 only for mixture velocity
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )
             
        dens = dens + mass_dens( dep )    
        ! number density of species   
        ndens = ndens + num_dens( dep ) 

        vel(:) = vel(:) + momentum( :, dep )
      end do !idep

      vel = vel/dens

      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = ndens * vel
    end do

  end subroutine deriveMoleFluxMSLiquid_FromState
!******************************************************************************!

end module mus_derQuanMSLiquid_module

