!******************************************************************************!
!> This module provides the MUSUBI specific functions for calculating
!! macroscopic quantities from the state variables in physical units.
!!
!! The depending common interface between MUSUBI and ATELES is defined in the
!! tem_derived_module. The functionality for accessing a variable from the state
!! and evaluating a lua function are also provided in the tem_derived module.
!!
!! \author Kannan Masilamani
!!
?? include 'treelm/source/deriveMacros.inc'
module mus_derQuanPhysics_module

  ! include treelm modules
  use tem_param_module,        only: div1_2, div1_3, div1_54, div1_9, div3_4,  &
    &                                sqrt3, cs2inv, cs2
  use env_module,              only: rk, long_k
  use tem_var_system_module,   only: tem_var_system_type
  use tem_variable_module,     only: tem_variable_type
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_topology_module,     only: tem_levelOf
  use tem_time_module,         only: tem_time_type
  use treelmesh_module,        only: treelmesh_type
  use tem_aux_module,          only: tem_abort
  use tem_spatial_module,      only: tem_spatial_for
  use tem_subTree_type_module, only: tem_subTree_type, tem_treeIDfrom_subTree

  ! include musubi modules
  use mus_solSpecHelpers_module, only: getConversionFac

  ! include aotus modules
  use aotus_module, only: flu_State

  implicit none

  private

  public :: deriveDensityPhy_FromState
  public :: derivePressurePhy_FromState
  public :: deriveKinePressPhy_FromState
  public :: deriveVelocityPhy_FromState
  public :: deriveMomentumPhy_FromState
  public :: deriveWSSPhy_FromState
  public :: deriveShearStressPhy_FromState
  public :: deriveShearMagPhy_FromState
  public :: deriveVelMagPhy_FromState
  public :: deriveKineEnerPhy_FromState
  public :: deriveTempPhy_FromState
  public :: deriveNumDensityPhy_FromState
  public :: deriveMoleFluxPhy_FromState
  public :: deriveStrainRatePhy_FromState
  public :: deriveShearRatePhy

contains


!******************************************************************************!
!       Subroutines with common interface for the function pointers            !
!******************************************************************************!

!******************************************************************************!
  !> Calculate the density of a given set of elements (sum up all links).
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveDensityPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer :: dens_pos
    integer, allocatable :: level(:)
    integer :: iElem, nElems
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, facName = 'density', &
              &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position of pressure in glob system
    dens_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( dens_pos )%do( input = input,                        &
      &                                 bound = bound,                        &
      &                                 tree = tree,                          &
      &                                 subTree = subTree,                    &
      &                                 requVar = globSys%variable%val(dens_pos),&
      &                                 globSys = globSys,                    &
      &                                 stencil = stencil,                    &
      &                                 res = local_res,                      &
      &                                 conf = conf )

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    level(1:nElems) = tem_levelOf( treeID, nElems)

    do iElem = 1,nElems
      local_res(iElem) = local_res(iElem)*convFac( level( iElem ))
    end do

  end subroutine deriveDensityPhy_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the density of a given set of elements (sum up all links).
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveNumDensityPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac_moleDens(tree%global%minLevel:tree%global%maxLevel)
    integer :: iElem, nElems, numdens_pos
    integer, allocatable :: level(:)
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(level(nElems))

    convFac_moleDens(:) = getConversionFac( conf = globSys%solSpec%conf,       &
      &               facName = 'moleDens',                                    &
      &               nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position of pressure in glob system
    numdens_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( numdens_pos )%do( input = input,                      &
      &                                  bound = bound,                        &
      &                                  tree = tree,                          &
      &                                  subTree = subTree,                    & 
      &                                  requVar = globSys%variable%val(numdens_pos),&
      &                                  globSys = globSys,                    &
      &                                  stencil = stencil,                    &
      &                                  res = local_res,                      &
      &                                  conf = conf )

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    level(1:nElems) = tem_levelOf( treeID, nElems) 

    !number density physical [mol/m^3]
    do iElem = 1,nElems
      local_res(iElem) = local_res(iElem) * convFac_moleDens(level( iElem ))
    end do

  end subroutine deriveNumDensityPhy_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the pressure of a given set of elements (sum up all links).
  !!
  !! Pressure calculation according to the isentropic equation of state for 
  !! the LBM \f$ p = \rho c_s^2 \f$
  !! with the calculation of density as in deriveDensityPhy_FromState
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(derivePressurePhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, press_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, facName = 'press',&
              &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position of pressure in glob system
    press_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( press_pos )%do( input = input,                        &
      &                                  bound = bound,                        &
      &                                  tree = tree,                          &
      &                                  subTree = subTree,                    & 
      &                                  requVar = globSys%variable%val(press_pos),&
      &                                  globSys = globSys,                    &
      &                                  stencil = stencil,                    &
      &                                  res = local_res,                      &
      &                                  conf = conf )

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    level(1:nElems) = tem_levelOf( treeID, nElems) 
  
    do iElem = 1,nElems
      local_res(iElem) = local_res(iElem)*convFac( level( iElem ))
    end do

  end subroutine derivePressurePhy_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the kinematic mixture pressure in physical 
  !! of a given set of elements (sum up all links).
  !!
  !! This routine requires initial mole fraction of all species which must
  !! be provided through add_variable table in scheme and set depvar in tracking
  !! for kinematic pressure to the newly added variable.
  !! This will be used to evaluate spatial function for initial mole fraction
  !! from initial mole fraction initial mixture molar density will be computed.
  !! Formula to compute kinematic pressure
  !! \f$ p = c^2_s (\sum_k \rho_k \phi_k - min_l (m_l) n_0)/\rho_0 \f$
  !! here, \f$ \rho_k \f$ - species density, \\
  !! \f$ \phi_k \f$ - species molecular weight ratio, \\
  !! \f$ n_0 \f$ - mixture number density,\\
  !! \f$ \rho_0 \f$ - reference density.
  !! Example:
  !! In scheme table,
  !! \verbatim
  !! add_variable = { name = "initialMolFrac", ncomponents=1,')
  !!   initialMolFrac = { { kind="combined", spatial=luaFunc}
  !!                      { kind="combined", spatial=luaFunc2}}
  !! }
  !! \endverbatim
  !! In tracking,
  !! \verbatim
  !! variable = {{"kinematicpressure_phy", 1, dep = { "initialMolFrac"}}}
  !! \endverbatim
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveKinePressPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac_len(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convFac_time(tree%global%minLevel:tree%global%maxLevel)
    integer :: iElem, nElems, press_pos
    integer, allocatable :: level(:)
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(level(nElems))

    convFac_len(:) = getConversionFac( conf = globSys%solSpec%conf, &
      &              facName = 'length', &
      &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    convFac_time(:) = getConversionFac( conf = globSys%solSpec%conf, &
      &              facName = 'time', &
      &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! point to pressure derive quantities which return \rho*cs2
    press_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    ! mixture pressure
    call globSys%derive( press_pos )%do( input = input,                        &
      &                                  bound = bound,                        &
      &                                  tree = tree,                          &
      &                                  subTree = subTree,                    & 
      &                                  requVar = globSys%variable%val(press_pos),&
      &                                  globSys = globSys,                    &
      &                                  stencil = stencil,                    &
      &                                  res = local_res,                      &
      &                                  conf = conf,                          &
      &                                  bary = bary,                          &
      &                                  time = time )      

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    level(1:nElems) = tem_levelOf( treeID, nElems) 
  
    !kinematic pressure (p/rho0)[m^2/s^2]
    do iElem = 1,nElems
      local_res(iElem) = local_res(iElem) &
        & * (convFac_len( level( iElem )) / convFac_time( level( iElem ) ) )**2
    end do

  end subroutine deriveKinePressPhy_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity of a given element number with the given input
  !!        vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveVelocityPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, vel_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))
    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, facName = 'vel' ,&
              &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position of vel in glob system
    vel_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive(vel_pos)%do( input = input,                            &
      &                              bound = bound,                            &
      &                              tree = tree,                              &
      &                              subTree = subTree,                        &
      &                              requVar = globSys%variable%val(vel_pos),  &
      &                              globSys = globSys,                        &
      &                              stencil = stencil,                        &
      &                              res = local_res,                          &
      &                              conf = conf )
  
    level(1:nElems) = tem_levelOf( treeID, nElems ) 
  
    do iElem = 1,nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &   *convFac( level( iElem ) )
    end do


  end subroutine deriveVelocityPhy_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity magnitude of a given element number with the given
  !! input vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveVelMagPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, velMag_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, facName = 'vel' ,&
              &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position in glob system
    velMag_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( velMag_pos )%do( input = input,                       &
      &                                 bound = bound,                         &
      &                                 tree = tree,                           &
      &                                 subTree = subTree,                     & 
      &                                 requVar = globSys%variable%val(velMag_pos),&
      &                                 globSys = globSys,                     &
      &                                 stencil = stencil,                     &
      &                                 res = local_res,                       &
      &                                 conf = conf )
  
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    level(1:nElems) = tem_levelOf( treeID, nElems )
  
    do iElem = 1,nElems
      local_res(iElem) = local_res(iElem)*convFac( level( iElem ))
    end do

  end subroutine deriveVelMagPhy_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the momentum of a given element number with the given input
  !!        vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveMomentumPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, mom_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))
    convFac(:) = getConversionFac(                           &
      & conf    = globSys%solSpec%conf,                      &
      & facName = 'flux' ,                                   &
      & nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position of momentum in glob system
    mom_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive(mom_pos)%do( input = input,                            &
      &                              bound = bound,                            &
      &                              tree = tree,                              &
      &                              subTree = subTree,                        &
      &                              requVar = globSys%variable%val(mom_pos),  &
      &                              globSys = globSys,                        &
      &                              stencil = stencil,                        &
      &                              res = local_res,                          &
      &                              conf = conf )
  
    level(1:nElems) = tem_levelOf( treeID, nElems ) 
  
    do iElem = 1,nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &   *convFac( level( iElem ) )
    end do


  end subroutine deriveMomentumPhy_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the velocity magnitude of a given element number with the given
  !! input vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveTempPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, temp_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, facName = 'temp' ,             &
              &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position in glob system
    temp_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( temp_pos )%do( input = input,                         &
      &                                 bound = bound,                         &
      &                                 tree = tree,                           &
      &                                 subTree = subTree,                     & 
      &                                 requVar = globSys%variable%val(temp_pos),&
      &                                 globSys = globSys,                     &
      &                                 stencil = stencil,                     &
      &                                 res = local_res,                       &
      &                                 conf = conf )
  
    level(1:nElems) = tem_levelOf( treeID, nElems ) 
  
    do iElem = 1,nElems
      local_res(iElem) = local_res(iElem)*convFac( level( iElem ))
    end do


  end subroutine deriveTempPhy_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the kinetic energy in physical units
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveKineEnerPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, kinec_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, facName = 'energy' ,             &
              &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position in glob system
    kinec_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( kinec_pos )%do( input = input,                        &
      &                                  bound = bound,                        &
      &                                  tree = tree,                          &
      &                                  subTree = subTree,                    & 
      &                                  requVar = globSys%variable%val(kinec_pos),&
      &                                  globSys = globSys,                    &
      &                                  stencil = stencil,                    &
      &                                  res = local_res,                      &
      &                                  conf = conf )

    level(1:nElems) = tem_levelOf( treeID, nElems )
  
    do iElem = 1,nElems
      local_res(iElem) = local_res(iElem)*convFac( level( iElem ))
    end do


  end subroutine deriveKineEnerPhy_FromState
!******************************************************************************!



!******************************************************************************!
  !> Calculate the wall shear stress (WSS) of a given element with the given input
  !! \author Jiaxing Qi
  !!
  !! The wall shear stress is shear stress exerted on the wall. Since it is well
  !! known that shear stress gets its maximum value on the wall, here we can
  !! directly calculate the principle stress of the stress tensor instead of 
  !! multiplying the stress tensor by normal vector of the plane.
  !! The principle stresses are just the eigenvalues of the stress tensor. To
  !! get those eigenvalues, we need solve the characteristic equation:\n
  !! \f$ {\tau}^3 + a_{2}{\tau}^2 + a_1{\tau} + a_0 = 0 \f$
  !! where \f$ a_2 = \tau_{ii} \f$
  !! \f$ a_1 = (\tau_{ii}\tau_{jj} - \tau_{ij}\tau_{ji}) / 2 \f$
  !! \f$ a_0 = det(\tau_{ij}) \f$. Here Einstein notation is used.
?? copy :: derive_headtxt(deriveWSSPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, wss_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, &
      &                            facName = 'press' ,          &
      &           nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position in glob system
    wss_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( wss_pos )%do( &
      &             input = input,                          &
      &             bound = bound,                          &
      &             tree = tree,                            &
      &             subTree = subTree,                      &
      &             requVar = globSys%variable%val(wss_pos),&
      &             globSys = globSys,                      &
      &             stencil = stencil,                      &
      &             res = local_res,                        &
      &             conf = conf )

    level(1:nElems) = tem_levelOf( treeID, nElems)

    do iElem = 1,nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &   *convFac( level( iElem ) )
    end do

  end subroutine deriveWSSPhy_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the strain rate (or called rate of strain)
  !! \author Jiaxing Qi
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
?? copy :: derive_headtxt(deriveStrainRatePhy_FromState)
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:) ! levels of elements
    integer :: iElem, nElems, strain_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, &
      &                            facName = 'strainRate' ,     &
      &       nLevels = tree%global%maxLevel-tree%global%minLevel+1 )

    ! position of shear strain in glob system
    strain_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    ! calculate shear strain in LB unit
    call globSys%derive( strain_pos )%do(                     &
      &             input = input,                            &
      &             bound = bound,                            &
      &             tree = tree,                              &
      &             subTree = subTree,                        &
      &             requVar = globSys%variable%val(strain_pos),&
      &             globSys = globSys,                        &
      &             stencil = stencil,                        &
      &             res = local_res,                          &
      &             conf = conf )

    ! get the level of elements
    level(1:nElems) = tem_levelOf( treeID, nElems)

    ! convert to physical units
    do iElem = 1,nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &   * convFac( level( iElem ) )
    end do

  end subroutine deriveStrainRatePhy_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the shear rate
  !! \author Jiaxing Qi
  !!
?? copy :: derive_headtxt(deriveShearRatePhy)
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:) ! levels of elements
    integer :: iElem, nElems, shear_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, &
      &                            facName = 'strainRate' ,     &
      &       nLevels = tree%global%maxLevel-tree%global%minLevel+1 )

    ! position of shear strain in glob system
    shear_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    ! calculate shear strain in LB unit
    call globSys%derive( shear_pos )%do(                     &
      &             input = input,                            &
      &             bound = bound,                            &
      &             tree = tree,                              &
      &             subTree = subTree,                        &
      &             requVar = globSys%variable%val(shear_pos),&
      &             globSys = globSys,                        &
      &             stencil = stencil,                        &
      &             res = local_res,                          &
      &             conf = conf )

    ! get the level of elements
    level(1:nElems) = tem_levelOf( treeID, nElems)

    ! convert to physical units
    do iElem = 1, nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &   * convFac( level( iElem ) )
    end do

  end subroutine deriveShearRatePhy
!******************************************************************************!

!******************************************************************************!
  !> Calculate the viscous shear stress (exclude pressure) (no mixtures)
  !! \author Jiaxing Qi
  !!
  !! The formula is:\n
  !! \f$ \tau_{\alpha \beta}=
  !! -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta} \f$\n
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.
  !!
?? copy :: derive_headtxt(deriveShearStressPhy_FromState)
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, shear_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, &
      &                            facName = 'press' ,          &
      &       nLevels = tree%global%maxLevel-tree%global%minLevel+1 )

    ! position of shear stress in glob system
    shear_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    ! calculate shear stress in LB unit
    call globSys%derive( shear_pos )%do(                      &
      &             input = input,                            &
      &             bound = bound,                            &
      &             tree = tree,                              &
      &             subTree = subTree,                        &
      &             requVar = globSys%variable%val(shear_pos),&
      &             globSys = globSys,                        &
      &             stencil = stencil,                        &
      &             res = local_res,                          &
      &             conf = conf )

    ! get the level of elements
    level(1:nElems) = tem_levelOf( treeID, nElems)

    ! convert to physical units
    do iElem = 1,nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &   * convFac( level( iElem ) )
    end do

  end subroutine deriveShearStressPhy_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the viscous shear stress magnitude
  !! \author Manuel Hasert
  !!
  !! Compute von Mises stress yield
?? copy :: derive_headtxt(deriveShearMagPhy_FromState)
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac(tree%global%minLevel:tree%global%maxLevel)
    integer, allocatable :: level(:)
    integer :: iElem, nElems, shearMag_pos
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    allocate(level(nElems))

    convFac(:) = getConversionFac( conf = globSys%solSpec%conf, facName = 'press' ,&
              &   nLevels = tree%global%maxLevel-tree%global%minLevel+1)

    ! position in glob system
    shearMag_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( shearMag_pos )%do(               &
      &             input = input,                        &
      &             bound = bound,                        &
      &             tree = tree,                          &
      &             subTree = subTree,                    &
      &             requVar = globSys%variable%val(shearMag_pos),&
      &             globSys = globSys,                    &
      &             stencil = stencil,                    &
      &             res = local_res,                      &
      &             conf = conf )

    level(1:nElems) = tem_levelOf( treeID, nElems)

    do iElem = 1,nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &   *convFac( level( iElem ) )
    end do

  end subroutine deriveShearMagPhy_FromState
!******************************************************************************!

!******************************************************************************!
!                                     MULTISPECIES                             !
!******************************************************************************!
!******************************************************************************!
  !> Calculate the density of a given set of elements (sum up all links).
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
?? copy :: derive_headtxt(deriveMoleFluxPhy_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: convFac_moleFlux(tree%global%minLevel:tree%global%maxLevel)
    integer :: iElem, nElems, moleFlux_pos
    integer, allocatable :: level(:)
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(level(nElems))

    convFac_moleFlux(:) = getConversionFac( conf = globSys%solSpec%conf,       &
      &               facName = 'moleFlux',                                    &
      &               nLevels = tree%global%maxLevel-tree%global%minLevel+1)

!write(*,*) 'moleFlux in derivePhy ',convFac_moleFlux
!stop
    ! position of pressure in glob system
    moleFlux_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )

    call globSys%derive( moleFlux_pos )%do( input = input,                     &
      &                                  bound = bound,                        &
      &                                  tree = tree,                          &
      &                                  subTree = subTree,                    & 
      &                                  requVar = globSys%variable%val(moleFlux_pos),&
      &                                  globSys = globSys,                    &
      &                                  stencil = stencil,                    &
      &                                  res = local_res,                      &
      &                                  conf = conf,                          &
      &                                  bary = bary,                          &
      &                                  time = time )      

    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    level(1:nElems) = tem_levelOf( treeID, nElems) 

    !number density physical [mol/m^3]
    do iElem = 1,nElems
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        & = local_res( (iElem-1)*requVar%nComponents +                         &
        &                                        1 : iElem*requVar%nComponents)&
        & * convFac_moleFlux(level( iElem ))
    end do

  end subroutine deriveMoleFluxPhy_FromState
!******************************************************************************!


end module mus_derQuanPhysics_module
!******************************************************************************!
