!******************************************************************************!
!> This module provides the MUSUBI specific functions for calculating
!! macroscopic quantities from the state variables.
!!
!! The depending common interface between MUSUBI and ATELES is defined in the
!! tem_derived_module. The functionality for accessing a variable from the state
!! and evaluating a lua function are also provided in the tem_derived module.
!!
!! \author Simon Zimny
!!
?? include 'header/lbm_macros.inc'
?? include 'treelm/source/deriveMacros.inc'
module mus_derQuan_module

  ! include treelm modules
  use tem_param_module,        only: div1_2, div1_3, div1_54, div1_9, div3_4,  &
    &                                sqrt3, cs2inv, cs2
  use env_module,              only: rk, long_k, globalMaxLevels
  use tem_var_system_module,   only: tem_var_system_type
  use tem_variable_module,     only: tem_variable_type
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_topology_module,     only: tem_levelOf
  use tem_time_module,         only: tem_time_type
  use treelmesh_module,        only: treelmesh_type
  use tem_subTree_type_module, only: tem_subTree_type, tem_treeIDfrom_subTree
  use tem_logging_module,      only: logUnit

  ! include musubi modules
  use mus_solSpecHelpers_module,     only: getWeights, getFieldVariable
  use mus_derivedQuantities_module2, only: secondMom, getShearRate

  ! include aotus modules
  use aotus_module, only: flu_State

  implicit none

  private

  public :: derivePressure_FromState
  public :: deriveDensity_FromState
  public :: deriveVelocity_FromState
  public :: deriveMomentum_FromState
  public :: deriveEquil_FromState
  public :: deriveNonEquil_FromState
  public :: deriveEquil_FromMacro
  ! public :: deriveNonEquil_FromMacro
  public :: deriveEquil_ps_FromMacro
  public :: deriveWSS_FromState
  public :: deriveShearStress_FromState
  public :: deriveShearMag_FromState
  public :: deriveVelMag_FromState
  public :: deriveKineEner_FromState
  public :: deriveTemp_FromState
  public :: deriveStrainRate_FromState
  public :: deriveStrainRatenNwtn_FromState
  public :: deriveShearStressnNwtn_FromState
  public :: deriveShearRate

contains


!******************************************************************************!
!       Subroutines with common interface for the function pointers            !
!******************************************************************************!

!******************************************************************************!
  !> Calculate the density of a given set of elements (sum up all links).
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveDensity_FromState)
    !---------------------------------------------------------------------------
    !local variable
    integer :: iDep, dep, iElem, buf_start, buf_end, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      dens = 0.0_rk
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )

        ! set the start and end position in the state array
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos( 1 )

        buf_end   = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos(                            &
          &                            globSys%variable%val( dep )%nComponents )

        ! derive the density and store it in local_res
        ! derive the density and store it in the input (for more than
        ! 1 species (iDep) this will be the mixture density)
        dens = dens + sum( input( buf_start : buf_end ))

      end do  ! iDep
      local_res( iElem ) = dens
    end do  ! iElem

  end subroutine deriveDensity_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the temperature of a given set of elements (sum up all links).
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveTemp_FromState)
    !---------------------------------------------------------------------------
    !local variable
    integer :: iElem, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    local_res( 1:nElems ) = div1_3

  end subroutine deriveTemp_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the pressure of a given set of elements (sum up all links).
  !!
  !! Pressure calculation according to the isentropic equation of state for 
  !! the LBM \f$ p = \rho c_s^2 \f$
  !! with the calculation of density as in deriveDensity_FromState
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(derivePressure_FromState)
    !---------------------------------------------------------------------------
    !local variable
    integer :: iDep, dep, iElem, buf_start, buf_end, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      dens = 0.0_rk
      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )

        ! set the start and end position in the state array
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos( 1 )

        buf_end   = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos(                            &
          &                            globSys%variable%val( dep )%nComponents )

        ! derive the density and store it in local_res
        ! derive the density and store it in the input (for more than
        ! 1 species (iDep) this will be the mixture density)
        dens = dens + sum( input( buf_start : buf_end ))

      end do  ! iDep
      local_res( iElem ) = dens*cs2
    end do  ! iElem

  end subroutine derivePressure_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity of a given element number with the given input
  !!        vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveVelocity_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem, nElems
    integer :: iDep, iDir, pos, dep, iComp
    integer :: buf_start, buf_end
    real(kind=rk) :: dens, vel( requVar%nComponents )
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      ! initialize the local velocity and density for every element
      vel = 0._rk
      dens = 0._rk

      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )

        ! \todo:SZ: is this really optimal or better calculate dens in the
        !           direction loop from the links?????

        ! set the start and end position in the input array to compute the
        ! density
        buf_start = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos( 1 )

        buf_end   = (iElem - 1)*globSys%nScalars                               &
          &   + globSys%variable%val( dep )%varPos(                            &
          &                            globSys%variable%val( dep )%nComponents )

        ! derive the density and store it in the input (for more than
        ! 1 species (iDep) this will be the mixture density)
        dens = dens + sum( input( buf_start : buf_end ))

        do iDir = 1, globSys%variable%val( dep )%nComponents
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( dep )%varPos( iDir )
          do iComp = 1, requVar%nComponents
            vel(iComp) = vel(iComp) + input( (iElem - 1) * globSys%nScalars    &
              &                    + pos )                                     &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !icomp
        end do !idir
      end do !idep
      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = vel/dens
    end do

  end subroutine deriveVelocity_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the velocity magnitude of a given element number with the given
  !! input vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveVelMag_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem, vel_pos, nElems, localBound(2)
    integer :: pdf_pos, pdf_nComps, buf_start, buf_end
    real(kind=rk) :: vel( 3 ) ! velocity results
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! position of vel in glob system
    vel_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    ! position of pdf in glob variable system
    pdf_pos = globSys%variable%val( vel_pos )%varDep(1)
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents

    do iElem = 1, nElems
      localBound(1) = iElem
      localBound(2) = iElem

      ! get pdf position in input array
      buf_start = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! First calculate velocity vector, by calling its function pointer
      ! save results to vel array
      call globSys%derive( vel_pos )%do(                                &
        &                      input   = input(buf_start:buf_end),      &
        &                      bound   = localBound,                    &
        &                      tree    = tree,                          &
        &                      subTree = subTree,                       &
        &                      requVar = globSys%variable%val(vel_pos), &
        &                      globSys = globSys,                       &
        &                      stencil = stencil,                       &
        &                      conf    = conf,                          &
        &                      res     = vel )

      ! Calculate the square root of velocity vector
      local_res( iElem ) = sqrt( vel(1)*vel(1) + vel(2)*vel(2) + vel(3)*vel(3) )
    enddo ! iElem

  end subroutine deriveVelMag_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the momentum of a given element number with the given input
  !!        vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveMomentum_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem, nElems
    integer :: iDep, iDir, pos, dep, iComp
    real(kind=rk) :: mom( requVar%nComponents )
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      ! initialize the local velocity and density for every element
      mom = 0._rk

      do iDep = 1, size( globSys%variable%val( requVar%mapping )%varDep )
        ! get the position of the depending variable in the global varSys
        dep = globSys%variable%val( requVar%mapping )%varDep( iDep )

        ! \todo:SZ: is this really optimal or better calculate dens in the
        !           direction loop from the links?????

        do iDir = 1, globSys%variable%val( dep )%nComponents
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = globSys%variable%val( dep )%varPos( iDir )
          do iComp = 1, requVar%nComponents
            mom(iComp) = mom(iComp) + input( (iElem - 1) * globSys%nScalars    &
              &                    + pos )                                     &
              & * real(stencil%cxDir( requVar%sub_index(iComp), iDir ), kind=rk)
          end do !icomp
        end do !idir
      end do !idep
      ! copy the values to the right position in the inout array input
      local_res( (iElem-1)*requVar%nComponents + 1 : iElem*requVar%nComponents)&
        &         = mom
    end do

  end subroutine deriveMomentum_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equlibrium of given elements with the given input state
  !! array.
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveEquil_FromState)
    !---------------------------------------------------------------------------
    !> local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens ! temporary array for density
    real(kind=rk) :: vel(3) ! temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: buf_start, buf_end
    real(kind=rk) :: ucx, usq
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents
    !> Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk

      ! set the begin and end position in the state array to compute density
      buf_start = (iElem-1) * globSys%nScalars                                 &
        &   + globSys%variable%val( dep )%varPos( 1 )
      buf_end = (iElem-1) * globSys%nScalars                                   &
        &   + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( buf_start:buf_end ) )

      !> Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / dens

      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)

      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)

        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) =                             &
          &   weights( iDir ) * dens * ( 1.d0                                  &
          &                            + ucx * cs2inv                          &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk )
      end do ! iDir
    end do ! iElem

  end subroutine deriveEquil_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! vector.
  !!
  !! input array contains density and velocity in the following order:
  !! |dens_1 | dens_2 | ... | dens_nElems | u_x_1 | u_y_1 | u_z_1 | ....
  !! .. | u_x_nElems | u_y_nElems | u_z_nElems |
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveEquil_FromMacro)
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk) :: ucx, usq
    integer :: iDir, iElem, posVel, offVel, nElems
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    offVel = nElems   ! offset of velocity components in input array

    do iElem = 1, nElems
      ! position of this element's vel in input array
      posVel = (iElem-1) * 3 + offVel
      ! square of velocity
      usq = sum( input(posVel+1:posVel+3) * input(posVel+1:posVel+3) )

      do iDir = 1, stencil%QQ

        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir(1,iDir),rk) * input( posVel+1 )    &
          &    + real( stencil%cxDir(2,iDir),rk) * input( posVel+2 )    &
          &    + real( stencil%cxDir(3,iDir),rk) * input( posVel+3 )

        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * input( iElem ) * ( 1._rk + ucx*cs2inv                    &
          &                                    + ucx*ucx*cs2inv*cs2inv*0.5_rk  &
          &                                    - usq*cs2inv*0.5_rk )

      enddo ! iDir
    end do ! iElem

  end subroutine deriveEquil_FromMacro
!******************************************************************************!

!******************************************************************************!
  !> Calculate the equlibrium of a given element number with the given input
  !! vector.
  !!
  !! input array contains density and velocity in the following order:
  !! |dens_1 | dens_2 | ... | dens_nElems | u_x_1 | u_y_1 | u_z_1 | ....
  !! .. | u_x_nElems | u_y_nElems | u_z_nElems |
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveEquil_ps_FromMacro)
    !---------------------------------------------------------------------------
    ! local variables
    real(kind=rk) :: ucx
    integer :: iDir, iElem, posVel, offVel, nElems
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    offVel = nElems   ! offset of velocity components in input array

    do iElem = 1, nElems
      ! position of this element's vel in input array
      posVel = (iElem-1) * 3 + offVel

      do iDir = 1, stencil%QQ

        ! velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk)*input( posVel+1 )    &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk)*input( posVel+2 )    &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk)*input( posVel+3 )

        ! calculate equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = weights( iDir )             &
          &         * input( iElem ) * ( 1._rk + ucx*cs2inv )

      enddo ! iDir
    end do ! iElem

  end subroutine deriveEquil_ps_FromMacro
!******************************************************************************!


!******************************************************************************!
  !> Calculate the Non-Equlibrium of given elements with given input state 
  !! array.
  !!
?? copy :: derive_headtxt(deriveNonEquil_FromState)
    !---------------------------------------------------------------------------
    !> local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens ! temporary array for density
    real(kind=rk) :: vel(3) ! temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: buf_start, buf_end
    real(kind=rk) :: ucx, usq, fEq, fNeq
    real(kind=rk), allocatable :: weights(:) ! weights read from spec_string
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( weights( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the weights from the solver specific character
    weights = getWeights( globSys%solSpec%conf, stencil )

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents

    !> Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk

      ! set the begin and end position in the state array to compute density
      buf_start = (iElem-1) * globSys%nScalars                                 &
        &   + globSys%variable%val( dep )%varPos( 1 )
      buf_end = (iElem-1) * globSys%nScalars                                   &
        &   + globSys%variable%val( dep )%varPos( nComp )

      dens = sum( input( buf_start:buf_end ) )

      ! Calculate velocity
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem - 1) * globSys%nScalars +                    &
          & globSys%variable%val(dep)%varPos( iDir ) )                         &
          &   * real(stencil%cxDir( 1:3, iDir ), kind=rk)
      end do
      vel = vel / dens

      ! Calculate the square of velocity
      usq = vel(1) * vel(1) + vel(2) * vel(2) + vel(3) * vel(3)

      do iDir = 1, stencil%QQ
        ! Velocity times lattice unit velocity
        ucx =    real( stencil%cxDir( 1, iDir ), kind=rk) * vel(1)             &
          &    + real( stencil%cxDir( 2, iDir ), kind=rk) * vel(2)             &
          &    + real( stencil%cxDir( 3, iDir ), kind=rk) * vel(3)
        fEq =  weights( iDir ) * dens * ( 1._rk                                &
          &                            + ucx * cs2inv                          &
          &                            + ucx * ucx * cs2inv * cs2inv * 0.5_rk  &
          &                            - usq * cs2inv * 0.5_rk )
        ! calculate Non-Equilibrium density
        local_res( (iElem-1)*stencil%QQ + iDir ) = &
          &                         input( (iElem-1)*stencil%QQ + iDir )  - fEq
      end do ! iDir
    end do ! iElem

  end subroutine deriveNonEquil_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the strain rate ( or rate of strain, or rate of deformation)
  !! \author Jiaxing Qi
  !!
  !! The equation is:
  !! \f[
  !!  \tau_{\alpha \beta}=
  !!    -\frac{3\omega}{2\rho} \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta}
  !! \f]
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibrium pdf.\n
  !! For more information, please refer to: equation 45 in\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.\n
  !!
  !! For multi-level mesh, Omega on finer level needs to be adjusted in order to
  !! get the correct shearstress calculation.\n
  !! First, we defines c as the dx ratio between finer and coarse level.\n
  !! \f$ c={ \Delta dx }_{ c }/{ \Delta dx }_{ f } \f$
  !! Then the viscosity on the different levels must satisfy:\n
  !! \f$ \frac { { \nu  }_{ f } }{ { \nu  }_{ c } } =c \f$
  !! This constrain leads to a relationship of omega on different levels:\n
  !! \f$ {\omega}_f = \frac {1}{ {\lambda}(\frac{1}{{\omega}_c}-0.5)+0.5 } \f$
  !! For more information, please refer to:\n
  !! Manuel H, Harald K, Joerg B, Sabine R. Aeroacoustic validation of the
  !! lattice boltzmann method on non-uniform grids. ECCOMAS 2012
  !!
?? copy :: derive_headtxt(deriveStrainRate_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: tau(requVar%nComponents), dens
    real(kind=rk), allocatable           :: nonEq(:) 
    integer       :: iElem, iLevel, buf_start, buf_end, nElems
    integer       :: nonEq_pos, pdf_pos, pdf_nComps
    real(kind=rk) :: omega( globalMaxLevels), iOmega
    integer(kind=long_k), allocatable    :: treeID(:)
    integer :: localBound(2)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( nonEq( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    ! Get all omegas from all levels
    omega(:) = getFieldVariable( conf   = globSys%solSpec%conf,                &
      &                       varLabel  = trim(requVar%label) ,                &
      &                       varName   = 'shearstress',                       &
      &                       fieldVar  = 'omLvl',                             &
      &                       fieldProp = 'fluid', nVals = globalMaxLevels )

    ! position of pdf in glob system
    pdf_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents
    ! position of nonEquilibrium in glob system
    nonEq_pos = globSys%variable%val( requVar%mapping )%varDep( 2 )

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem
      nonEq = 0._rk

      ! set the begin and end position in the input state array
      buf_start = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! calculate density
      dens = 0._rk
      dens = sum( input( buf_start : buf_end ))

      ! Calculate nonEquilibrium density
      ! for PULL, this is post-collision Non-Eq
      ! for PUSH, this is  pre-collision Non-Eq
      call globSys%derive( nonEq_pos )%do(                         &
        &                    input   = input(buf_start:buf_end),   &
        &                    bound   = localBound,                 &
        &                    tree    = tree,                       &
        &                    subTree = subTree,                    &
        &                    requVar = globSys%variable%val(nonEq_pos) ,&
        &                    globSys = globSys,                    &
        &                    stencil = stencil,                    &
        &                    conf    = conf,                       &
        &                    res     = nonEq  )

      ! get my current level
      iLevel = tem_levelOf( treeID( iElem ) )
      ! get the correct omega value
      iOmega = omega( iLevel )

      ! Shear stress requires pre-collision NonEq
      ! Convert post-coll to pre-coll by a factor of 1/(1-omega), if use PULL
?? if( .not. STREAM_PUSH ) then
      if( iOmega /= 1._rk ) then
        nonEq = nonEq / ( 1._rk - iOmega )
      else
        nonEq = 0._rk
      endif
?? endif

      ! compute shear stress
      tau(:) = secondMom( stencil%cxcx(:,:), nonEq(:), stencil%QQ )

      local_res( (iElem-1)*6+1 : iElem*6 ) = -1.5_rk * iOmega / dens * tau(:)
    enddo ! iElem

  end subroutine deriveStrainRate_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the strain rate ( or rate of strain, or rate of deformation)
  !! \author Jiaxing Qi
  !!
  !! The equation is:
  !! \f[
  !!  \tau_{\alpha \beta}=
  !!    -\frac{3\omega}{2\rho} \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta}
  !! \f]
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibrium pdf.\n
  !! For more information, please refer to: equation 45 in\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.\n
  !!
  !! For multi-level mesh, Omega on finer level needs to be adjusted in order to
  !! get the correct shearstress calculation.\n
  !! First, we defines c as the dx ratio between finer and coarse level.\n
  !! \f$ c={ \Delta dx }_{ c }/{ \Delta dx }_{ f } \f$
  !! Then the viscosity on the different levels must satisfy:\n
  !! \f$ \frac { { \nu  }_{ f } }{ { \nu  }_{ c } } =c \f$
  !! This constrain leads to a relationship of omega on different levels:\n
  !! \f$ {\omega}_f = \frac {1}{ {\lambda}(\frac{1}{{\omega}_c}-0.5)+0.5 } \f$
  !! For more information, please refer to:\n
  !! Manuel H, Harald K, Joerg B, Sabine R. Aeroacoustic validation of the
  !! lattice boltzmann method on non-uniform grids. ECCOMAS 2012
  !!
?? copy :: derive_headtxt(deriveStrainRatenNwtn_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: tau(requVar%nComponents), dens, omega
    real(kind=rk), allocatable :: nonEq(:)
    integer :: iElem, buf_start, buf_end, nElems
    integer :: nonEq_pos, nonEq_nComps, pdf_pos, pdf_nComps
    integer :: omega_pos, omega_nComps
    integer(kind=long_k), allocatable :: treeID(:)
    integer :: localBound(2)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! position of pdf in glob system
    pdf_pos       = globSys%variable%val( requVar%mapping )%varDep( 1 )
    pdf_nComps    = globSys%variable%val( pdf_pos )%nComponents
    ! omega position in glob system
    omega_pos     = globSys%variable%val( requVar%mapping )%varDep( 2 )
    omega_nComps  = globSys%variable%val( omega_pos )%nComponents
    ! nonEquilibrium position in glob system
    nonEq_pos     = globSys%variable%val( requVar%mapping )%varDep( 3 )
    nonEq_nComps  = globSys%variable%val( nonEq_pos )%nComponents
    allocate( nonEq( nonEq_nComps ))

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1:2) = iElem
      nonEq = 0._rk

      ! set the begin and end position in the input state array
      buf_start = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                &
        &   + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! calculate density
      dens = sum( input( buf_start : buf_end ))

      ! Calculate nonEquilibrium density
      ! for PULL, this is post-collision Non-Eq
      ! for PUSH, this is  pre-collision Non-Eq
      call globSys%derive( nonEq_pos )%do(                         &
        &                    input   = input(buf_start:buf_end),   &
        &                    bound   = localBound,                 &
        &                    tree    = tree,                       &
        &                    subTree = subTree,                    &
        &                    requVar = globSys%variable%val(nonEq_pos) ,&
        &                    globSys = globSys,                    &
        &                    stencil = stencil,                    &
        &                    conf    = conf,                       &
        &                    res     = nonEq  )

      ! get omega from state vector
      omega = input( (iElem-1) * globSys%nScalars + &
        &               globSys%variable%val(omega_pos)%varPos(omega_nComps) )

      ! Shear stress requires pre-collision NonEq
      ! Convert post-coll to pre-coll by a factor of 1/(1-omega), if use PULL
?? if( .not. STREAM_PUSH ) then
      if( omega /= 1._rk ) then
        nonEq = nonEq / ( 1._rk - omega )
      else
        nonEq = 0._rk
      endif
?? endif

      ! compute shear stress
      tau(:) = secondMom( stencil%cxcx(:,:), nonEq(:), stencil%QQ )

      local_res( (iElem-1)*6+1 : iElem*6 ) = -1.5_rk * omega / dens * tau(:)
    enddo ! iElem

    deallocate( nonEq )

  end subroutine deriveStrainRatenNwtn_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the shear rate
  !! \author Jiaxing Qi
  !!
  !! The Shear Rate is defined as
  !! \f[
  !!  \dot{\gamma} = \sqrt{ 2D_{II} }
  !! \f]
  !! where \f$ D_{II} \f$ is the second invariant of the strain rate tensor and
  !! defined as
  !! \f[
  !!    D_{II} = \sum^{l}_{\alpha,\beta=l} S_{\alpha\beta} S_{\alpha\beta}
  !! \f]
  !! where \f$ S_{\alpha\beta} \f$ is the strain rate tensor.
  !!
?? copy :: derive_headtxt(deriveShearRate)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk), allocatable :: strain(:)
    integer       :: iElem, nElems
    integer       :: strain_pos, nComps
    integer(kind=long_k), allocatable :: treeID(:)
    integer :: localBound(2)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! position of strain rate in glob system
    strain_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComps = globSys%variable%val( strain_pos )%nComponents

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate( strain(nComps*nElems) )

    ! Calculate strain rate
    call globSys%derive( strain_pos )%do(                               &
      &                    input   = input(:),                          &
      &                    bound   = bound,                             &
      &                    tree    = tree,                              &
      &                    subTree = subTree,                           &
      &                    requVar = globSys%variable%val(strain_pos) , &
      &                    globSys = globSys,                           &
      &                    stencil = stencil,                           &
      &                    conf    = conf,                              &
      &                    res     = strain )


    do iElem = 1, nElems
      ! calculate shear rate from strain rate
local_res( iElem ) = getShearRate( strain( 1+(iElem-1)*nComps : nComps*iElem) )
    enddo ! iElem

    deallocate( strain )

  end subroutine deriveShearRate
!******************************************************************************!

!******************************************************************************!
  !> Calculate the deviatoric shear stress for Newtonian fluid
  !! (exclude pressure) (no mixtures).\n
  !! Shear Stress depends on variable: nonEquilibirium
  !! \author Jiaxing Qi
  !!
  !! The formula is
  !! \f[
  !!  \tau_{\alpha \beta}=
  !!    -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta}
  !! \f]
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.\n
  !! Thus this variable is dependent on another variable: nonEquilibirium.
  !!
  !! For multi-level mesh, Omega on finer level needs to be adjusted in order to
  !! get the correct shearstress calculation.\n
  !! First, we defines c as the dx ratio between finer and coarse level.\n
  !! \f$ c={ \Delta dx }_{ c }/{ \Delta dx }_{ f } \f$
  !! Then the viscosity on the different levels must satisfy:\n
  !! \f$ \frac { { \nu  }_{ f } }{ { \nu  }_{ c } } =c \f$
  !! This constrain leads to a relationship of omega on different levels:\n
  !! \f$ {\omega}_f = \frac {1}{ {\lambda}(\frac{1}{{\omega}_c}-0.5)+0.5 } \f$
  !! For more information, please refer to:\n
  !! Manuel H, Harald K, Joerg B, Sabine R. Aeroacoustic validation of the
  !! lattice boltzmann method on non-uniform grids. ECCOMAS 2012
  !!
?? copy :: derive_headtxt(deriveShearStress_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: tau(requVar%nComponents)
    real(kind=rk), allocatable :: nonEq(:) 
    integer       :: iLevel, iElem, nElems, nonEq_pos, pdf_pos, pdf_nComps
    real(kind=rk) :: omega( globalMaxLevels), iOmega
    integer(kind=long_k), allocatable :: treeID(:)
    integer :: localBound(2), buf_start, buf_end
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    allocate( nonEq( stencil%QQ ))

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    ! Get all omegas from all levels
    omega(:) = getFieldVariable( conf = globSys%solSpec%conf,                  &
      &                       varLabel = trim(requVar%label) ,                 &
      &                       varName = 'shearstress',                         &
      &                       fieldVar = 'omLvl',                              &
      &                       fieldProp = 'fluid', nVals = globalMaxLevels )

    ! position of nonEq in glob variable system
    nonEq_pos = globSys%variable%val( requVar%mapping )%varDep(1)
    ! position of pdf in glob variable system
    pdf_pos = globSys%variable%val( nonEq_pos )%varDep(1)
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem

      ! get pdf position in input array
      buf_start = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! Calculate NonEquilibrium density
      ! for PULL, this is post-collision Non-Eq
      ! for PUSH, this is  pre-collision Non-Eq
      nonEq = 0._rk
      call globSys%derive( nonEq_pos )%do(                         &
        &                    input   = input(buf_start:buf_end),   &
        &                    bound   = localBound,                 &
        &                    tree    = tree,                       &
        &                    subTree = subTree,                    &
        &                    requVar = globSys%variable%val( nonEq_pos ),  &
        &                    globSys = globSys,                    &
        &                    stencil = stencil,                    &
        &                    conf    = conf,                       &
        &                    res     = nonEq  )

      ! get my current level
      iLevel = tem_levelOf( treeID( iElem ) )
      ! get the correct omega value
      iOmega = omega( iLevel )

      ! Shear stress requires pre-collision NonEq
      ! Convert post-coll to pre-coll by a factor of 1/(1-omega), if use PULL
?? if( .not. STREAM_PUSH ) then
      if( iOmega /= 1._rk ) then
        nonEq = nonEq / ( 1._rk - iOmega )
      else
        nonEq = 0._rk
      endif
?? endif

      ! compute shear stress
      tau(:) = secondMom( stencil%cxcx(:,:), nonEq(:), stencil%QQ )

      local_res( (iElem-1)*6+1 : iElem*6 ) = (iOmega * 0.5_rk - 1._rk) * tau(:)
    enddo ! iElem

  end subroutine deriveShearStress_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the deviatoric shear stress for Newtonian fluid
  !! (exclude pressure) (no mixtures).\n
  !! Shear Stress depends on variable: nonEquilibirium
  !! \author Jiaxing Qi
  !!
  !! The formula is
  !! \f[
  !!  \tau_{\alpha \beta}=
  !!    -(1-\frac{\omega}{2}) \sum_{i} f^{neq}_{i} c_{i\alpha} c_{i\beta}
  !! \f]
  !! where \f$ \tau_{\alpha \beta}\f$ is the stress
  !! in the \f$\beta\f$-direction on a face normal to the \f$\alpha\f$-axis,\n
  !! \f$ f^{neq}_i = f_i - f^{eq}_i\f$ is the non-equilibirium density.\n
  !! For more information, please refer to:\n
  !! Krueger T, Varnik F, Raabe D. Shear stress in lattice Boltzmann
  !! simulations. Physical Review E. 2009;79(4):1-14.\n
  !! Thus this variable is dependent on another variable: nonEquilibirium.
  !!
  !! For multi-level mesh, Omega on finer level needs to be adjusted in order to
  !! get the correct shearstress calculation.\n
  !! First, we defines c as the dx ratio between finer and coarse level.\n
  !! \f$ c={ \Delta dx }_{ c }/{ \Delta dx }_{ f } \f$
  !! Then the viscosity on the different levels must satisfy:\n
  !! \f$ \frac { { \nu  }_{ f } }{ { \nu  }_{ c } } =c \f$
  !! This constrain leads to a relationship of omega on different levels:\n
  !! \f$ {\omega}_f = \frac {1}{ {\lambda}(\frac{1}{{\omega}_c}-0.5)+0.5 } \f$
  !! For more information, please refer to:\n
  !! Manuel H, Harald K, Joerg B, Sabine R. Aeroacoustic validation of the
  !! lattice boltzmann method on non-uniform grids. ECCOMAS 2012
  !!
?? copy :: derive_headtxt(deriveShearStressnNwtn_FromState)
    !---------------------------------------------------------------------------
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: tau(requVar%nComponents), omega
    real(kind=rk), allocatable :: nonEq(:)
    integer :: nonEq_pos, nonEq_nComps, pdf_pos, pdf_nComps
    integer :: omega_pos, omega_nComps
    integer(kind=long_k), allocatable :: treeID(:)
    integer :: localBound(2), buf_start, buf_end, iLevel, iElem, nElems
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! position of nonEq in glob variable system
    nonEq_pos = globSys%variable%val( requVar%mapping )%varDep(1)
    nonEq_nComps = globSys%variable%val( nonEq_pos )%nComponents
    ! position of pdf in glob variable system
    pdf_pos = globSys%variable%val( nonEq_pos )%varDep(1)
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents
    ! omega position in glob system
    omega_pos = globSys%variable%val( requVar%mapping )%varDep( 2 )
    omega_nComps = globSys%variable%val( omega_pos )%nComponents
    allocate( nonEq( nonEq_nComps ))

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1:2) = iElem

      ! get pdf position in input array
      buf_start = (iElem - 1)*globSys%nScalars                       &
        &       + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                       &
        &       + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! Calculate NonEquilibrium density
      ! for PULL, this is post-collision Non-Eq
      ! for PUSH, this is  pre-collision Non-Eq
      nonEq = 0._rk
      call globSys%derive( nonEq_pos )%do(                         &
        &                    input   = input(buf_start:buf_end),   &
        &                    bound   = localBound,                 &
        &                    tree    = tree,                       &
        &                    subTree = subTree,                    &
        &                    requVar =globSys%variable%val( nonEq_pos ),  &
        &                    globSys = globSys,                    &
        &                    stencil = stencil,                    &
        &                    conf    = conf,                       &
        &                    res     = nonEq  )

      ! get omega from state vector
      omega = input( (iElem-1) * globSys%nScalars + &
        &               globSys%variable%val(omega_pos)%varPos(omega_nComps) )

      ! Shear stress requires pre-collision NonEq
      ! Convert post-coll to pre-coll by a factor of 1/(1-omega), if use PULL
?? if( .not. STREAM_PUSH ) then
      if( omega /= 1._rk ) then
        nonEq = nonEq / ( 1._rk - omega )
      else
        nonEq = 0._rk
      endif
?? endif

      ! compute shear stress
      tau(:) = secondMom( stencil%cxcx(:,:), nonEq(:), stencil%QQ )

      local_res( (iElem-1)*6+1 : iElem*6 ) = (omega * 0.5_rk - 1._rk) * tau(:)
    enddo ! iElem

    deallocate( nonEq )

  end subroutine deriveShearStressnNwtn_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the wall shear stress (WSS) of a given element with the given
  !! input
  !! \author Jiaxing Qi
  !!
  !! The wall shear stress is shear stress exerted on the wall. Since it is well
  !! known that shear stress gets its maximum value on the wall, here we can
  !! directly calculate the principle stress of the stress tensor instead of 
  !! multiplying the stress tensor by normal vector of the plane.
  !! The principle stresses are just the eigenvalues of the stress tensor. To
  !! get those eigenvalues, we need solve the characteristic equation:\n
  !! \f$ {\tau}^3 + a_{2}{\tau}^2 + a_1{\tau} + a_0 = 0 \f$
  !! where \f$ a_2 = \tau_{ii} \f$
  !! \f$ a_1 = (\tau_{ii}\tau_{jj} - \tau_{ij}\tau_{ji}) / 2 \f$
  !! \f$ a_0 = det(\tau_{ij}) \f$. Here Einstein notation is used.
  !!
?? copy :: derive_headtxt(deriveWSS_FromState)
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iElem, shear_pos, nElems, localBound(2)
    integer :: nonEq_pos, pdf_pos, pdf_nComps, buf_start, buf_end
    real(kind=rk) :: tau(6) ! shear stress components
    real(kind=rk) :: a0, a1, a2
    real(kind=rk) :: q_stress, q_stress_cube, q_term, r_stress, d_stress
    real(kind=rk) :: theta_stress, cos_term, shear_1, shear_2, shear_3
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! wall shear stress is depend on shear stress. get its position in globSys
    shear_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    ! position of pdf in glob variable system
    nonEq_pos = globSys%variable%val( shear_pos )%varDep(1)
    pdf_pos = globSys%variable%val( nonEq_pos )%varDep(1)
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents

    do iElem = 1, nElems
      ! set the bounds to be iElem call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem

      ! get pdf position in input array
      buf_start = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! first compute shear stress, save it to array tau
      ! the shear tensor has the following order: x, y, z, xy, xz, yz
      tau = 0._rk
      call globSys%derive( shear_pos )%do(                                     &
        &                      input   = input(buf_start:buf_end),             &
        &                      bound   = localBound,                           &
        &                      tree    = tree,                                 &
        &                      subTree = subTree,                              &
        &                      requVar = globSys%variable%val(shear_pos),      &
        &                      globSys = globSys,                              &
        &                      stencil = stencil,                              &
        &                      conf    = conf,                                 &
        &                      res     = tau )

      ! tau_x + tau_y + tau_z
      a2 = -1._rk * ( tau(1) + tau(2) + tau(3) )

      ! tau_x * tau_y + tau_y * tau_z + tau_x * tau_z - tau_xy^2 - tau_xz^2 - tau_yz^2
      a1 =    tau(1) * tau(2) + tau(2) * tau(3) + tau(1) * tau(3)              &
        &   - tau(4) * tau(4) - tau(5) * tau(5) - tau(6) * tau(6)

      a0 = -1._rk * (   tau(1) * tau(2) * tau(3)                               &
        &             + 2._rk * tau(4) * tau(5) * tau(6)                       &
        &             - tau(1) * tau(6) * tau(6)                               &
        &             - tau(2) * tau(5) * tau(5)                               &
        &             - tau(3) * tau(4) * tau(4) )

      q_stress = div1_9 * ( 3._rk * a1 - a2 * a2 )
      if ( q_stress < 0._rk ) then
        q_term = 2._rk * sqrt( -q_stress )
      else
        q_term = 0._rk
      end if
      r_stress = div1_54 * ( 9._rk*a2*a1 - 27._rk*a0 - 2._rk*a2*a2*a2 )
      q_stress_cube = q_stress * q_stress * q_stress
      d_stress = q_stress_cube + r_stress * r_stress

      if ( d_stress < 0._rk ) then
        ! the solutions are real and unequal
        theta_stress = div1_3 * acos( r_stress / sqrt( -q_stress_cube ))

        cos_term = div3_4 * q_term * cos( theta_stress )

        ! three values of maximum shear stress
        shear_1 = abs( sqrt3 * div1_2 * q_term * sin( theta_stress ))

        shear_2 = abs( cos_term - shear_1*div1_2 )

        shear_3 = abs( cos_term + shear_1*div1_2 )

        ! choose the max shear as Wall Shear Stress(wss)
        local_res(iElem) = max(shear_1, shear_2, shear_3)
      else
        local_res(iElem) = 0._rk
      end if
    end do ! iElem

  end subroutine deriveWSS_FromState
!******************************************************************************!

!******************************************************************************!
  !> Calculate the shear stress magnitude of a given element number with the
  !! given
  !! input vector (sum up all values)
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveShearMag_FromState)
    !---------------------------------------------------------------------------
    integer :: iElem, tau_pos, nElems, localBound(2)
    integer :: pdf_pos, pdf_nComps, buf_start, buf_end
    real(kind=rk) :: tau(6) ! shear stress result
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! position of shear stress in glob system
    tau_pos = globSys%variable%val( requVar%mapping )%varDep( 1 )
    ! position of pdf in glob variable system
    pdf_pos = globSys%variable%val( tau_pos )%varDep(1)
    pdf_nComps = globSys%variable%val( pdf_pos )%nComponents

    do iElem = 1, nElems
      ! set the bounds to be 1 call is done elementwise
      localBound(1) = iElem
      localBound(2) = iElem

      ! get pdf position in input array
      buf_start = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( 1 )
      buf_end   = (iElem - 1)*globSys%nScalars                                 &
        &       + globSys%variable%val( pdf_pos )%varPos( pdf_nComps )

      ! First calculate velocity vector, by calling its function pointer
      ! save results to tau array
      call globSys%derive( tau_pos )%do(                                  &
        &                      input   = input(buf_start:buf_end),        &
        &                      bound   = localBound,                      &
        &                      tree    = tree,                            &
        &                      subTree = subTree,                         &
        &                      requVar = globSys%variable%val(tau_pos),   &
        &                      globSys = globSys,                         &
        &                      stencil = stencil,                         &
        &                      conf    = conf,                            &
        &                      res     = tau )

      ! Von mises-criterion (see wikipedia article on von Mises yield criterion)
      ! wss = (tau_xx-tau_yy)**2 + (tau_yy-tau_zz)**2 + (tau_xx-tau_zz)**2  &
      ! &   + 6._rk*( tau_xy*tau_xy + tau_yz*tau_yz + tau_xz*tau_xz)
      ! wss = sqrt(0.5_rk*wss )
      local_res( iElem ) = sqrt( 0.5_rk*(                                      &
        & (tau(1)-tau(2))**2 + (tau(2)-tau(3))**2 + (tau(1)-tau(3))**2         &
        &   + 6._rk*( tau(4)*tau(4) + tau(5)*tau(5) + tau(6)*tau(6))           &
        &   ))
    enddo ! iElem

  end subroutine deriveShearMag_FromState
!******************************************************************************!


!******************************************************************************!
  !> Calculate the kinetic energy as
  !!  E = 1/2 rho ( ux^2 + uy^2 +uz^2 )
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_module#derive "derive".
  !!
?? copy :: derive_headtxt(deriveKineEner_FromState)
    !---------------------------------------------------------------------------
    !> local variables
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dens ! temporary array for density
    real(kind=rk) :: vel(3) ! temporary array for velocity
    integer :: iDir, iElem, dep, nComp, nElems
    integer :: buf_start, buf_end
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the position of the depending variable in the global varSys
    ! it is always for only one field.
    dep = globSys%variable%val( requVar%mapping )%varDep( 1 )
    nComp = globSys%variable%val(dep)%nComponents

    ! Calculation of density and velocity
    do iElem = 1, nElems
      dens = 0._rk
      vel = 0._rk

      ! set the begin and end position in the state array to compute density
      buf_start =   (iElem-1) * globSys%nScalars              &
        &         + globSys%variable%val( dep )%varPos( 1 )
      buf_end   =   (iElem-1) * globSys%nScalars              &
        &         + globSys%variable%val( dep )%varPos( nComp )
      dens = sum( input( buf_start:buf_end ) )

      ! Velocity calculation
      do iDir = 1, stencil%QQ
        vel = vel + input( (iElem-1) * globSys%nScalars + &
          & globSys%variable%val(dep)%varPos( iDir ) )    &
          &   * real(stencil%cxDir( 1:3, iDir ), rk)
      end do
      vel = vel / dens

      local_res( iElem ) = sum( vel(:)*vel(:) ) * 0.5_rk * dens
    end do ! iElem

  end subroutine deriveKineEner_FromState
!******************************************************************************!

end module mus_derQuan_module
!******************************************************************************!
