! See copyright notice in the COPYRIGHT file.
?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> This module provides functionality to perform geometry change  
!! depending on the flow properties. Fluidify/Solidify, proximity, dependent 
!! scheme, flow variables, thresholds and operators are supplied in the lua
!! file. Checks are performed against those parameters and geometry is 
!! accordingly changed (an element solidified or fluidified)

!! \author Kartik Jain, Jiaxing Qi
!!
!> Geometry Increment module
!! 
module mus_geomIncr_module

  ! include treelm modules
  use env_module,              only: rk, long_k, io_buffer_size, LabelLen
  use tem_element_module,      only: eT_GhostFromCoarserElem,                  &
    &                                eT_ghostFromFinerElem
  use tem_topology_module,     only: tem_LevelOf
  use tem_property_module,     only: prp_fluidify
  use tem_comm_module,         only: tem_commPattern_type
  use tem_timeControl_module,  only: tem_timeControl_check 
  use tem_condition_module,    only: comparator
  use tem_depend_module,       only: tem_init_depend
  use tem_subTree_type_module, only: tem_subTree_type
  use tem_logging_module,      only: logUnit
  use tem_general_module,      only: tem_general_type

  ! include musubi modules
  use mus_pdf_module,         only: pdf_globInfo_type, pdf_global_type
  use mus_geom_module,        only: mus_geom_type
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_buffer_module,      only: mus_trans_serialize
  use mus_param_module,       only: mus_param_type
  use mus_time_module,        only: mus_time_modulo, mus_timeControl_homogenize
  use mus_physics_module,     only: mus_physics_type


  implicit none

  private

  public :: mus_geomIncr, proximity
  public :: mus_init_geomIncr
  public :: prepare_target_elem_list
  public :: update_connectivity
  public :: fillMinePropFromFiner, fillFinerPropFromMe

contains

!******************************************************************************!
  !> This subroutine initializes the geometry increment.
  !!
  subroutine mus_init_geomIncr( geometry, scheme, physics, params)
    !---------------------------------------------------------------------------
    !> geometry type
    type( mus_geom_type ), intent(inout)     :: geometry
    !> scheme information
    type( mus_scheme_type ), intent(inout)   :: scheme(:)
    !> global physics info
    type( mus_physics_type ), intent(in) :: physics
    !> global parameters
    type(mus_param_type),intent(in)       :: params
    !---------------------------------------------------------------------------
    integer :: iGeom
    character(len=labelLen) :: label_res
    !---------------------------------------------------------------------------
    do iGeom = 1, size( geometry%geomIncr ) 
      ! Initialize the timer
      call mus_timeControl_homogenize(                                         &
        &         me = geometry%geomIncr( iGeom )%timeControl,                 &
        &         dt = physics%dtLvl( geometry%tree%global%maxLevel ),         &
        &         reqInt = params%reqInterval )

      ! Initialize the depends which were loaded before
      label_res = 'geomIncr'
      call tem_init_depend( medepend    = geometry%geomIncr( iGeom )%depend,   &
        &                   varSys      = scheme(:)%varSys,                    &
        &                   resSyslabel = label_res,                           &
        &                   label_list  = scheme(:)%header%label,              &
        &                   iScheme     = 0 )

    end do

  end subroutine mus_init_geomIncr
!******************************************************************************!


!******************************************************************************!
  !> This subroutine checks for various conditions defined in the geomIncr table
  !! within the lua file, calculates the requested macroscpoic variables and 
  !! then compares them against the specified threshold. 
  !! Accordingly then solidification or fluidification of elements is performed.
  !!
  subroutine mus_geomIncr( geometry, scheme, pdfGlobal, commPattern, params )
    !---------------------------------------------------------------------------
    !>
    type( mus_geom_type )                    :: geometry
    !>
    type( mus_scheme_type ), intent(inout)   :: scheme(:)
    !>
    type( pdf_globInfo_type ), intent(inout) :: pdfGlobal
    !>
    type( tem_commPattern_type ), intent(in) :: commPattern
    !> global parameters
    type(mus_param_type),intent(in) :: params
    !---------------------------------------------------------------------------
    integer :: pot_nElems ! potential elements for gi temporarily
    integer :: nElems_gi  ! number of elements which participate in geomincr
    integer :: nElems ! Total number of elements in the tree
    integer :: nVars, nDepends
    ! treeID list of element to be solidify
    integer(kind=long_k), allocatable :: treeIDlist(:)
    ! levelPointer list of element to be solidify
    integer, allocatable :: LPlist(:), LPtreeList(:)
    ! List of target iElems
    integer, allocatable :: target_ielem_list(:)
    real(kind=rk), allocatable :: chunk(:)
    integer :: iElem, iVar, iGeom, iDepend, iChunk
    integer :: iScheme, nComp, chunkSize, nChunks
    integer :: nChunkElems, iLevel, iSchm
    integer :: elemoff ! starting index of iElem for different chunks
    real(kind=rk), allocatable :: local_res(:)
    integer :: bound(2)
    type(tem_subTree_type) :: subTree
    integer :: res_size
    logical :: triggered
    !---------------------------------------------------------------------------

    ! The number of elements in the tree
    nElems = geometry%tree%nElems

    allocate( treeIDlist( size(geometry%tree%treeID)) )
    allocate( LPlist(     size(geometry%tree%treeID)) )
    allocate( LPtreeList( size(geometry%tree%treeID)) )
    allocate( target_ielem_list(size(geometry%tree%treeID)) )

    allocate(chunk(io_buffer_size))

    ! \todo: SZ: initialize a subTree here using the global tree parameters
    allocate( subTree%map2Global(nElems) )
    allocate( subTree%treeID(nElems) )
    subTree%useGlobalMesh = .true.

    !\todo: if the mesh changes set the flag in geometry%tree%global to true

    ! Loop over the number of geomIncrs defined in the Lua file
    do iGeom = 1, size( geometry%geomIncr )
      call tem_timeControl_check(                                              &
        &               me     = geometry%geomIncr(iGeom)%timeControl,         &
        &               now    = params%general%simControl%now,                &
        &               comm   = params%general%proc%comm,                     &
        &               triggered = triggered )

      ! time_modulo is to do only when one complete cyle is done with multilevel
      ! when read from restart trigger is active at 1st time step
      if( triggered ) then
        
        ! Initialize the total number of target elements to zero
        nElems_gi = 0
        nDepends = size( geometry%geomIncr(iGeom)%depend )
        ! Loop over the depend tables inside geometry increase table
        do iDepend = 1, nDepends 
          !> grab the index for scheme from depend type
          iScheme = geometry%geomIncr(iGeom)%depend(iDepend)%useScheme
     
          ! If proximity flag is active, call the proximity routine to generate
          ! a list of treeIDs which pass the proximity check
          if ( geometry%geomIncr(iGeom)%proximity .and. iDepend == 1 ) then
            call proximity( treeID       = geometry%tree%treeID,               &
              &             levelPointer = pdfGlobal%levelPointer,             &
              &             scheme       = scheme(iScheme),                    &
              &             pot_tid      = treeIDlist(:),                      &
              &             pot_nElems   = pot_nElems,                         &
              &             pot_lp       = LPlist(:),                          &
              &             pot_lp_tree  = LPtreeList(:) )
          ! If proximity is not activated, then we perform the checks for all 
          ! the elements, i.e. all elements in the tree are potential elements
          else
            pot_nElems = nElems
            LPlist     = pdfglobal%levelPointer
            treeIDlist = geometry%tree%treeID
            do iElem = 1, nElems 
              LPtreeList(iElem) = iElem
            end do
          end if

          nVars = geometry%geomIncr(iGeom)%depend(iDepend)%resSys%nVars
          chunkSize = io_buffer_size / scheme(iScheme)%varSys%nScalars
          nChunks = ceiling( real(pot_nElems, kind=rk) /                       &
            &                real(chunkSize, kind=rk) )

          do iChunk = 1, nChunks
            elemOff = ( (iChunk-1)*chunkSize )

            !\todo: SZ: passing treeID  = treeIDlist(1:pot_nElems) seems to be
            !           wrong here
            !           in case treeIDList is larger than chunkSize multiple
            !           chunks are used but the same treeIDs are passed!
            !KJ: passing treeIDlist(1:pot_nElems) is absolutely right. The
            ! treeIDlist basically contains only pot_nElems. Even if we pass
            ! it without index it is fine but its better to use this indexing
            ! to avoid Memory thrashing by several compilers

            ! Extract the state values for the treeIDs into the array chunk
            ! nChunkElems is the number of elements in this chunk which is 
            ! returned by the trans_serialize routine
            call mus_trans_serialize( pdf         = scheme(iScheme)%pdf,       &
              &                       globSys     = scheme(iScheme)%varSys,    &
              &                       chunk       = chunk,                     &
              &                       iChunk      = iChunk,                    &
              &                       chunkSize   = chunkSize,                 &
              &                       nChunkElems = nChunkElems,               &
              &                       treeID      = treeIDlist(1:pot_nElems),  &
              &                       elemPos     = LPlist(1:pot_nELems) )
            
            ! Compute the element lower and upper bound for the current chunk
            bound(1) = (iChunk-1)*chunkSize+1
            bound(2) = (iChunk-1)*chunkSize+nChunkElems

            ! \todo: SZ: initialize the treeID array in the subTree

            ! Loop over the number of variables inside each depend table
            do iVar = 1, nVars
              ! get the no. of components for each variable system
              nComp = geometry%geomIncr(iGeom)%depend(iDepend)%resSys          &
                &     %variable%val(iVar)%nComponents

              ! allocate local array where the results are placed
              res_size = nComp*nChunkElems
              allocate( local_res(res_size) )

              ! Use the function pointers to calculate variables identified 
              ! from Lua results are placed in array local_res
              !\todo: SZ: passing the complete treeID list of the global tree
              !           is wrong
              !KJ: treeID list was not a part of derived quantities function
              !pointers when I implemented this routine and it looks like its
              ! not a part of it again and thus the call below is correct.
              !\todo: subTREE is a mandatory parameter in derived quantities
              !which should not be the case. I think it should be OPTIONAL
              call scheme( iScheme )%varSys%derive(                            &
                &          geometry%geomIncr(iGeom)%depend(iDepend)%resSys%    &
                &          variable%val(iVar)%mapping )                        &
                &          %do( input   = chunk,                               &
                &               bound   = bound,                               &
                &               subTree = subTree,                             &
                &               tree    = geometry%tree,                       &
                &               requVar = geometry%geomIncr(iGeom)             &
                &                         %depend(iDepend)%resSys%Variable     &
                &                         %val(iVar),                          &
                &               globSys = scheme( iScheme )%varSys,            &
                &               stencil = scheme(iScheme)%layout%stencil(1),   &
                &               res     = local_res(1:res_size),               &
                &               conf    = scheme(iScheme)%varSys%solSpec%conf )

              ! Prepare the list of target elements and set the property bits 
              ! of these elements
              call prepare_target_elem_list(                                   &
                &                       iVar              = iVar,              &
                &                       nVars             = nVars,             &
                &                       nComp             = nComp,             &
                &                       iGeom             = iGeom,             &
                &                       iDepend           = iDepend,           &
                &                       elemoff           = elemoff,           &
                &                       LPList            = LPList,            &
                &                       nChunkElems       = nChunkElems,       &
                &                       LPtreeList        = LPtreeList,        &
                &                       treeIDlist        = treeIDlist,        &
                &                       scheme            = scheme,            &
                &                       res               = local_res,         &
                &                       iScheme           = iScheme,           &
                &                       target_ielem_list = target_ielem_list, &
                &                       nElems_gi         = nElems_gi,         &
                &                       geometry          = geometry )

              ! \todo: SZ: Update the treeID list according to the 
              !            prepare_target_elem_list call

              ! Deallocate the array containing computed results
              deallocate( local_res )
            end do ! iVar
          end do ! iChunk
          
          if(nElems_gi > 0) then
            write(logUnit(1),*)'Elements found for geometry increase:', nElems_gi
          end if

          do iSchm = 1, size( scheme )
            do iLevel = geometry%tree%global%minLevel, geometry%tree%global%maxLevel
              ! Communicate the property bits of halo elements for all schemes
              call commPattern%exchange_long(                                  &
                &  send = scheme( iSchm )%glob%levelDesc( iLevel )%sendbuffer, &
                &  recv = scheme( iSchm )%glob%levelDesc( iLevel )%recvbuffer, &
                &  state = scheme( iSchm )%glob%levelDesc( iLevel )%property,  &
                &  message_flag = iLevel,                                      &
                &  proc = params%general%proc )
              
              ! Invoke the update connectivity routine which updates the 
              ! neighbor array of target elements and the neighbor array of 
              ! their neighbors. This has to be done for all schemes and not
              ! only for that scheme on which GI is currently dependent upon.
              ! This will ensure change in flow properties for all schemes
              call update_connectivity( scheme(iSchm), pdfGlobal, iLevel)
            end do
          end do ! Updation of all schemes

        end do ! iDepend
      end if ! Timing check
    end do ! iGeom
    ! Deallocate the chunk array
    deallocate( chunk )
    deallocate( treeIDlist )
    deallocate( LPlist )
    deallocate( target_ielem_list )
  end subroutine mus_geomIncr
!******************************************************************************!


!******************************************************************************!
  !> A subroutine which checks the proximity condition and generates a new 
  !! treeID list of potential target elements which can then be checked against
  !! various thresholds as defined in the lua configuration file.
  !!
  !! Proximity condition means that an element can be solidified only if it has 
  !! at least one wall as a neighboring element. In biological clotting models 
  !! it ensures that no isolated clot is produced in the fluid domain.
  !!
  subroutine proximity( treeID, levelPointer, scheme, pot_tid, pot_nELems,     &
    &                   pot_lp, pot_lp_tree )
    !---------------------------------------------------------------------------
    !> scheme information
    type(mus_scheme_type), intent(inout)  :: scheme
    !> elements treeIDlist
    integer(kind=long_k), intent(in)      :: treeID(:)
    !> array of levelPointers
    integer, intent(in)                   :: levelPointer(:)
    !> list of potential treeIDs to convert
    integer(kind=long_k), intent(inout)   :: pot_tid(:)  
    !> number of potential elements found
    integer, intent(inout) :: pot_nElems
    !> level pointer of potential elements
    integer, intent(inout) :: pot_lp(:)
    !> position of potential elements in the treeID list
    integer, intent(inout) :: pot_lp_tree(:)
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iElem, iLevel, iField, iDir
    integer :: neighPos, elemPos 
    integer :: QQN, QQ, neighbor
    !---------------------------------------------------------------------------

    pot_nElems = 0
    QQN = scheme%layout%stencil(1)%QQN
    QQ  = scheme%layout%stencil(1)%QQ

    do iElem = 1, size( treeID )
      do iField = 1, scheme%nFields
        iLevel = tem_levelOf( treeID( iElem ))
        elemPos = levelPointer( iElem )
        do iDir = 1, QQN
          ! Get the neighbor 
          neighbor = scheme%pdf( iLevel )%neigh( ?NGPOS?(iDir, elemPos, QQ ))
          ! position in total list
          neighPos = ?ElemIDX?( neighbor, QQ)

          !> Acc the proximity condition, as soon as its found that any of the 
          !! neighbors of current element is solid, that element is stored in 
          !! the new list of potential elements. Loop is terminated for this 
          !! element and check is performed for other element.
          if (neighPos == elemPos) then
            pot_nElems = pot_nElems + 1
            pot_tid(pot_nElems) = treeID(iElem)
            pot_lp(pot_nElems) = elemPos
            pot_lp_tree(pot_nElems) = iElem
            exit
          end if
          enddo ! iDir
      enddo ! iField
    enddo ! iElem

  end subroutine proximity
!******************************************************************************!


!******************************************************************************!
  !> This routine compares the macroscopic quantity obtained for an element
  !! against specified threshold and populates the list of target elements 
  !! which participate in geometry change. In addition it sets the property
  !! bit(s) of target elements to fluidify in case of solidification and clears
  !! this bit in case of fluidification
  !!
  subroutine prepare_target_elem_list( iVar, nVars, nComp, iGeom, iDepend,     &
    &                                  elemoff, nChunkElems, LPList, scheme,   &
    &                                  LPtreeList, treeIDlist, iScheme, res,   &
    &                                  nElems_gi, target_ielem_list, geometry )
    !---------------------------------------------------------------------------
    !> current variable
    integer, intent(in)    :: iVar
    !> Total number of variables
    integer, intent(in)    :: nVarS
    !> total number of components
    integer, intent(in)    :: nComp
    !> index to current geom incr
    integer, intent(in)    :: iGeom
    !> index to current depend table
    integer, intent(in)    :: iDepend
    !> index of the depending scheme
    integer, intent(in)    :: iScheme
    !> offset of elements in chunk
    integer, intent(in)    :: elemoff
    !> number of elements in chunk
    integer, intent(in)    :: nChunkElems
    !> list of level pointers
    integer, intent(in)    :: LPList(:)
    !> number of final elements
    integer, intent(inout) :: nElems_gi
    ! todo: SZ: describe these variables
    !>
    integer, intent(in)    :: LPtreeList(:) 
    !>
    integer, intent(inout) :: target_ielem_list(:) 
    !>
    integer(kind=long_k), intent(in)       :: treeIDlist(:) 
    !>
    type( mus_scheme_type ), intent(inout) :: scheme(:)
    !>
    type( mus_geom_type ), intent(inout)   :: geometry
    !>
    real(kind=rk), intent(in)              :: res(:)
    !---------------------------------------------------------------------------
    ! local variables
    integer                :: local_elem
    integer                :: iElem, iLevel, iSchm, elemPos
    integer(kind=long_k)   :: treeID, elemProp, nDepends
    logical                :: comp, isSolid
    character(len=2)       :: oper 
    real(kind=rk)          :: threshold 
    !---------------------------------------------------------------------------
    ! Get the operator and threshold from lua file
    oper = geometry%geomIncr(iGeom)%depend(iDepend)%cond(iVar)%operation
    threshold = geometry%geomIncr(iGeom)%depend(iDepend)%cond(iVar)%threshold
    nDepends = size( geometry%geomIncr(iGeom)%depend )

    ! Loop over all elements in chunk for the first variable in var system
    ! For solidify( solidify = true ), we check only fluid element( isSolid = 
    ! false ) and for fluidify( solidify = false), we check only solid element
    ! ( isSolid = true ) by using XOR operator( neqv ), we can achieve our 
    ! purpose in a universal way
    if ( iVar == 1 .and. iDepend == 1) then
      local_elem = 0
      do iElem = elemoff+1, elemoff+nChunkElems
        local_elem = local_elem + 1
        elemPos = LPlist( iElem ) ! elem position in total list
        treeID = treeIDlist( iElem )
        iLevel = tem_LevelOf( treeID )
        elemProp = scheme( iScheme )%glob%levelDesc( iLevel )                  &
          &                %property( elemPos )
        isSolid = btest( elemProp, prp_fluidify )
        if( isSolid .neqv. geometry%geomIncr(iGeom)%solidify ) then
          ! Velocity has 3 components, so the index of res should be adjusted
          comp = comparator(res((local_elem-1)*nComp+1), oper, threshold )
          if ( comp ) then
            nElems_gi = nElems_gi + 1
            target_ielem_list( nElems_gi ) = iElem
            ! Update the property bit if nVars is 1 and this is the
            ! only depend table, otherwise check for other
            ! variables and then update the bit
            if (nVars == iVar .and. nDepends == iDepend) then
              ! Update the property bits for ALL SCHEMES
              do iSchm = 1, size( scheme )
                call mus_setSolidProp( property = scheme(iSchm)%glob           &
                  &                    %levelDesc(iLevel)%property(elemPos),   &
                  &                    solidify = geometry%geomIncr(iGeom)     &
                  &                    %solidify )
              end do ! scheme loop
              ! and the ones in the global tree
              call mus_setSolidProp( property = geometry%tree%ElemPropertyBits &
                &                                 (LPtreeList(iElem)),         &
                &                    solidify = geometry%geomIncr(iGeom)       &
                &                                 %solidify )
              ! and update the flag meshChange because something changed
              ! geometry%tree%global%meshChange = .true.

            end if ! nVars == 1
          end if ! comparison
        end if ! is my target
      end do ! iElem
      ! For subsequent variables, loop over the tree ID list which was found 
      ! for the first variable. Doing this we ensure the AND condition among 
      ! different variables within a depend table and as the number of target 
      ! elements is hardly 5% of the total geometry, it improves computational
      ! efficiency of the code.
      else ! iVar != 1 or iDepend != 1
        if(nElems_gi > 0) then
          local_elem = 1
          do iElem = 1, nElems_gi
            elemPos = LPlist( target_ielem_list( iElem ))
            treeID = treeIDlist( target_ielem_list( iElem ) )
            iLevel = tem_LevelOf( treeID )
            elemProp = scheme( iScheme )%glob%levelDesc( iLevel )              &
              &        %property( elemPos )
            isSolid = btest( elemProp, prp_fluidify )
            if( isSolid .neqv. geometry%geomIncr(iGeom)%solidify ) then
              comp = comparator(res( (target_ielem_list(iElem)-1)*nComp+1),    &
              &                 oper, threshold )
              ! Update the local tree ID list
              if (comp) then
                target_ielem_list(local_elem) = target_ielem_list(iElem)
                ! Update the property bit if this is the final variable and 
                ! the depend table is the last one, otherwise check for other 
                ! variables (and/or depends) and then update the bit
                if (nVars == iVar .and. nDepends == iDepend) then
                  ! Update the property bits for ALL SCHEMES
                  do iSchm = 1, size( scheme )
                    call mus_setSolidProp(                                     &
                      &       property = scheme(iSchm)%glob%levelDesc(iLevel)% &
                      &                    property(elemPos),                  &
                      &       solidify = geometry%geomIncr(iGeom)%solidify )
                  end do ! scheme loop
                  ! and the ones in the global tree
                  call mus_setSolidProp(                                       &
                    &   property = geometry%tree%ElemPropertyBits              &
                    &                (LPtreeList(iElem)),                      &
                    &   solidify = geometry%geomIncr(iGeom)%solidify )
                  ! and update the flag meshChange because something changed
                  ! geometry%tree%global%meshChange = .true.

                end if ! nVars == 1
                local_elem = local_elem + 1
              end if ! comparison
            end if ! is my target
          end do ! iElem
          ! number of target elements is updated
          nElems_gi = local_elem - 1
        end if ! nElems_gi > 0
      end if ! iVar == 1

  end subroutine prepare_target_elem_list

!******************************************************************************!


!******************************************************************************!
  !> Construct the propagation stencil for each element
  !!
  !! the bounce back rules have to be applied here
  !!
  subroutine update_connectivity( scheme, pdfGlobal, iLevel)
    !-------------------------------------------------------------------------- 
    !> global pdf information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> scheme information
    type( mus_scheme_type ), intent(inout) :: scheme
    !> current level
    integer, intent(in)     :: iLevel
    !-------------------------------------------------------------------------- 
    ! defining local variables
    integer :: iElem  ! element counter for current element in treeID list
    integer :: iField ! field counter
    integer :: idx_idir ! current direction in global varsys
    integer :: iDir ! counter for current direction in iElem
    integer :: invDir ! inverse direction for iDir
    ! result: from which direction to take pdf from (inverse for bounceback) 
    integer :: GetFromDir
    ! result: from which element to take the pdf from (local for bounceback) 
    integer :: GetFromPos
    integer :: neighPos      ! position of current neighElem in treeID list
    integer(kind=long_k)  :: neighProp, elemProp
    !---------------------------------------------------------------------------

    do iElem = 1, scheme%pdf( iLevel )%nElems_local
      do iField = 1, scheme%nFields
        elemProp = scheme%glob%levelDesc( iLevel )%property( iElem )

        ! Loop over every link
        do iDir  = 1, scheme%layout%stencil(1)%QQ 
          if ( iDir == scheme%layout%stencil(1)%QQ  ) then
            ! Rest density takes from itself, iElem and iDir
            GetFromPos = iElem
            GetFromDir = scheme%varSys%Variable%val(iField)%varPos( iDir )
          else ! iDir not rest density
            ! For push we need to find different neighbors than for pull
            ! -> ngDir
            neighPos = scheme%glob%levelDesc( iLevel )%neigh(1)%               &
              &        nghElems( ?NgDir?( iDir, scheme%layout%stencil(1) ),    &
              &                  iElem )

            neighProp = 0_long_k
            ! get neighbor's property
            if( neighPos > 0 )                                                 &
              & neighProp = scheme%glob%levelDesc( iLevel )%property( neighPos )

            if( (neighPos <= 0) .or. (btest(neighProp, prp_fluidify))        &
                &  .or. (btest(elemProp, prp_fluidify))) then
              ! 1. Fluid element has boundary in current link direction
              ! 2. neighbor element has fluidify property (still solid now)
              ! Get inverse direction of LOCAL element
              invDir = scheme%layout%stencil(1)%cxDirInv( iDir )
              GetFromDir = scheme%varSys%Variable%val(iField)%varPos(invDir)
              GetFromPos = iElem
              ! I set the source to the target link for halo cells 
              ! with no neighbor in that direction 
            else
              ! neighElem = scheme%glob%levelDesc( iLevel )%total( neighPos )
              GetFromDir = scheme%varSys%Variable%val(iField)%varPos(iDir)
              GetFromPos = neighPos
            endif ! has no boundary -> regular treatment
          endif ! not the 0 velocity
 
          idx_idir = scheme%varSys%Variable%val(iField)%varPos(iDir)
          scheme%pdf( iLevel )%neigh(                                          &
            &         ?NGPOS?( idx_idir, iElem, scheme%varSys%nScalars )) =    &
            &         ?IDX?( GetFromDir, GetFromPos, scheme%varSys%nScalars )
          ! Set all the pdf links of target element to an infinitesimally 
          ! small value. This will prevent density from becoming NaN if pdfs 
          ! are set to zero and besides changed elements will have a very 
          ! small density which will help in visualization and differentiate
          ! them from other fluid elements
          if (btest(elemProp, prp_fluidify) ) then
            scheme%pdf(iLevel)%state(                                          &
              &        ?IDX?(idx_idir, iElem, scheme%varSys%nScalars),:) =     &
              &        0.000000001_rk
          end if

        enddo ! neighLoop
      enddo ! fieldLoop
    enddo ! elemLoop

  end subroutine update_connectivity
!******************************************************************************!


!******************************************************************************!
  !> Fill the Properties of GhostFromFiner elements on my level
  !!
  subroutine fillMinePropFromFiner( glob, level )
    !---------------------------------------------------------------------------
    !> state properties
    type( pdf_global_type ), intent(inout) :: glob
    !> current level 
    integer,intent(in) :: level
    !---------------------------------------------------------------------------
    integer :: srcElem     ! treeId of current source element
    integer :: tgtLvl      ! level of target elements
    integer :: tgtElem     ! treeId of current source element
    integer :: iElem       ! current target element (for outer loop)
    integer :: nSrcElem, iSrcElem   ! number of source elements
    integer(kind=long_k) :: srcProp
    logical :: setSolid   ! whether to set target ghost element to be solid
    integer :: srcLvl
    !---------------------------------------------------------------------------
 
    write(logUnit(1),*) '*** Get into fillMinePropFromFiner ***'
    tgtLvl = level      ! fill ghost of current level
    srcLvl = level + 1

    ! Treat every coarse target element:
    do iElem = 1, glob%levelDesc( tgtLvl )%nElems_ghostFromFiner
      ! number of source elements
      nSrcElem = glob%levelDesc( tgtLvl )%depFromFiner( iElem )%elem%nVals
      setSolid = .false.

      do iSrcElem = 1, nSrcElem
        ! get pos of source element in total list
        srcElem = glob%levelDesc( tgtLvl )%depFromFiner( iElem )%              &
          &                                                 elem%val( iSrcElem )
        ! get property of source element
        srcProp = glob%levelDesc( srcLvl )%property( srcElem )
        ! check whether the source elem is solid
        ! if one of the source elems is solid, set the target ghost to be solid
        !>\todo HK: Confused code!!!
        !!          The statement above states something different,
        !!          than the commented out original code!
        !!          XOR is first of all NOT STANDARD and supposed to be bitwise
        !!          that is, it should be applied to integers, not logicals!
        !!          Then, according to the comment, the code should actually
        !!          NOT to an exclusive or, which would be .neqv., but just
        !!          a simple logical or, otherwise setSolid will be true for
        !!          even numbers of srcElem with prp_fluidify.
        !!          Also it is confusing, that you are using prp_fluidify when
        !!          doing setSolid?
        !!          I changed this now to do the straight .or. operation.
        !!setSolid = xor( setSolid, btest( srcProp, prp_fluidify ))
        setSolid = ( setSolid .or. btest(srcProp, prp_fluidify) )
      end do

      ! if criterion is satisfied, set the target ghost to be solid(prp_fludify)
      if( setSolid ) then
        tgtElem = iElem + glob%levelDesc( tgtLvl )%                            &
          &                                   offset( 1, eT_ghostFromFinerElem )
        glob%levelDesc( tgtLvl )%property( tgtElem ) =                         &
          &  ibset( glob%levelDesc( tgtLvl )%property( tgtElem ), prp_fluidify )
      endif
    enddo
  end subroutine fillMinePropFromFiner
!******************************************************************************!


!******************************************************************************!
  !> Fill the properties of ghost elements on the next(finer) level
  !!
  subroutine fillFinerPropFromMe( glob, level )
    !---------------------------------------------------------------------------
    !> state properties
    type( pdf_global_type ), intent(inout) :: glob
    !> current level
    integer,intent(in) :: level
    !---------------------------------------------------------------------------
    integer :: srcLvl      ! level of source elements
    integer :: srcElem     ! treeId of current source element
    integer :: tgtLvl      ! level of target elements
    integer :: tgtElem     ! treeId of current source element
    integer :: iElem       ! current target element (for outer loop)
    integer :: nSrcElem, iSrcElem
    integer(kind=long_k) :: srcProp
    logical :: setSolid
    !---------------------------------------------------------------------------
    write(logUnit(1),*) '*** Get into fillFinerPropFromMe ***'
    srcLvl = level
    tgtLvl = level + 1

    ! Treat every fine target element:
    do iElem = 1, glob%levelDesc( tgtLvl )%nElems_ghostFromCoarser

      ! number of source elements
      nSrcElem = glob%levelDesc( tgtLvl )%depFromCoarser( iElem )%elem%nVals

      ! loop over source elements
      do iSrcElem = 1, nSrcElem
        ! get pos of source element in total list
        srcElem = glob%levelDesc( tgtLvl )%depFromCoarser( iElem )%            &
          &                                                 elem%val( iSrcElem )
        ! get property of source element
        srcProp = glob%levelDesc( srcLvl )%property( srcElem )
        ! check whether the source elem is solid
        ! if one of the source elems is solid, set the target ghost to be solid
        setSolid = btest( srcProp, prp_fluidify )
      enddo

      ! if criterion is satisfied, set the target ghost to be solid(prp_fludify)
      if( setSolid ) then
        tgtElem = iElem + glob%levelDesc( tgtLvl )%                            &
          &                                 offset( 1, eT_ghostFromCoarserElem )
        glob%levelDesc( tgtLvl )%property( tgtElem ) =                         &
          &  ibset( glob%levelDesc( tgtLvl )%property( tgtElem ), prp_fluidify )
      endif

    enddo ! ind%nVals

  end subroutine fillFinerPropFromMe
!******************************************************************************!


!******************************************************************************!
  !> wall shear stress gradient model for aneurysm initialization
  !!
  subroutine mus_wssg_model( geometry, scheme, pdfGlobal, general)
    !---------------------------------------------------------------------------
    type( mus_geom_type )                    :: geometry
    type( mus_scheme_type ), intent(inout)   :: scheme(:)
    type( pdf_globInfo_type ), intent(inout) :: pdfGlobal
    type(tem_general_type), intent(in) :: general
    !---------------------------------------------------------------------------
    ! Local variables
    integer :: iGeom, iDepend, iScheme, iLevel, iVar, iElem, iCond
    integer :: nVars, nComp, QQ
    integer :: neigh__w, neigh_sw, neigh_nw
    integer( kind=long_k ) :: elemProp
    integer( kind=long_k ) :: prop_neigh__w, prop_neigh_sw, prop_neigh_nw
    real( kind=rk ) :: state( 3 * scheme(1)%layout%stencil(1)%QQ )
    real( kind=rk ) :: thrh
    character(len=2) :: oper
    real( kind=rk ), allocatable :: res(:)
    logical :: comp(2)
    ! \todo: SZ: most of these parameters are currently not used, delete them??
    ! Local direction parameters
    integer,parameter :: incr_q__W     = 1     ! west             x-
!    integer,parameter :: incr_q__S     = 2     ! south            y-
!    integer,parameter :: incr_q__E     = 3     ! east             x+
!    integer,parameter :: incr_q__N     = 4     ! north            y+
    integer,parameter :: incr_q_SW     = 5     !                  y-,x-
    integer,parameter :: incr_q_NW     = 6     !                  y+,x-
!    integer,parameter :: incr_q_SE     = 7     !                  y-,x+
!    integer,parameter :: incr_q_NE     = 8     !                  y+,x+
!    integer,parameter :: incr_q__0     = 9     ! rest density is last
    logical :: triggered
    !---------------------------------------------------------------------------

    do iGeom = 1, size( geometry%geomIncr )
      call tem_timeControl_check(                                              &
        &               me     = geometry%geomIncr(iGeom)%timeControl,         &
        &               now    = general%simControl%now,                       &
        &               comm   = general%proc%comm,                            &
        &               triggered = triggered )

      if( triggered )                                                          &
        & write(logUnit(1),*) 'Initiating the Geometry Increment Routine'

      do iDepend = 1, size( geometry%geomIncr(iGeom)%depend )
        !> grab the index for scheme from depend type
        iScheme = geometry%geomIncr(iGeom)%depend(iDepend)%useScheme
        nVars = geometry%geomIncr(iGeom)%depend(iDepend)%resSys%nVars
        QQ = scheme( iScheme )%layout%stencil(1)%QQ

        do iVar = 1, nVars
          ! get the no. of components for each variable system
          nComp = geometry%geomIncr(iGeom)%depend(iDepend)%resSys%variable     &
            &     %val(iVar)%nComponents
          allocate( res( nComp*3 ))
  
          do iLevel = geometry%tree%global%minLevel, geometry%tree%global%maxLevel
            do iElem = 1, scheme( ischeme )%glob%leveldesc( ilevel )%nElems_fluid
              ! Get position of three neighbors on left side
              neigh__W = scheme( ischeme )%glob%leveldesc( ilevel )%           &
                &                          neigh(1)%nghElems( incr_q__W, iElem )
              neigh_SW = scheme( ischeme )%glob%leveldesc( ilevel )%           &
                &                          neigh(1)%nghElems( incr_q_SW, iElem )
              neigh_NW = scheme( ischeme )%glob%leveldesc( ilevel )%           &
                &                          neigh(1)%nghElems( incr_q_NW, iElem )

              ! get property of target element and neighbors
              elemProp = scheme( ischeme )%glob%leveldesc( ilevel )%           &
                &                                              property( iElem )
              prop_neigh__w = scheme( ischeme )%glob%leveldesc( ilevel )%      &
                &                                           property( neigh__w )
              prop_neigh_sw = scheme( ischeme )%glob%leveldesc( ilevel )%      &
                &                                           property( neigh_sw )
              prop_neigh_nw = scheme( ischeme )%glob%leveldesc( ilevel )%      &
                &                                           property( neigh_nw )

              ! check whether target element is solid and
              ! its neighbors are fluids
              if(         btest( elemProp,      prp_fluidify ) .and.           &
                &   .not. btest( prop_neigh__w, prp_fluidify ) .and.           &
                &   .not. btest( prop_neigh_sw, prp_fluidify ) .and.           &
                &   .not. btest( prop_neigh_nw, prp_fluidify ) ) then

               ! extract pdf to state array
               state(     1:  QQ) = scheme( iScheme )%pdf( iLevel )%           &
                 &                     state( (neigh__w-1)*QQ+1:neigh__w*QQ, 2 )
               state(  QQ+1:2*QQ) = scheme( iScheme )%pdf( iLevel )%           &
                 &                     state( (neigh_sw-1)*QQ+1:neigh_sw*QQ, 2 )
               state(2*QQ+1:3*QQ) = scheme( iScheme )%pdf( iLevel )%           &
                 &                     state( (neigh_nw-1)*QQ+1:neigh_nw*QQ, 2 )

                !\todo: SZ: passing the complete treeID list of the global tree
                !           is wrong
                ! call function pointer to get direved quantities
!                call scheme( iScheme )%varSys%derive(                          &
!                  &    geometry%geomIncr(iGeom)%depend(iDepend)%resSys%        &
!                  &    variable%val(iVar)%mapping )                            &
!                  &    %do(  input   = state,                                  &
!                  &          nElems  = 3,                                      &
!                  &          tree = geometry%tree,                             &
!                  &          requVar = geometry%geomIncr(iGeom)%depend(iDepend)&
!                  &                    %resSys%Variable%val(iVar),             &
!                  &          globSys = scheme( iScheme )%varSys,               &
!                  &          stencil = scheme(iScheme)%layout%stencil(1),      &
!                  &          treeID  = geometry%tree%treeID,                   &
!                  &          res = res,                                        &
!                  &          conf= scheme(iScheme)%varSys%solSpec%conf )

                ! get the wss gradient from sw and nw neighbors
                res(2) = abs( res(2) - res(3) ) * 0.5_rk

                ! compare wss and wssg with thresholds
                do iCond = 1, size( geometry%geomIncr(iGeom)%depend(iDepend)%cond )
                  oper = geometry%geomIncr(iGeom)%depend(iDepend)%             &
                    &                                    cond( iCond )%operation
                  thrh = geometry%geomIncr(iGeom)%depend(iDepend)%             &
                    &                                    cond( iCond )%threshold
                  comp( iCond ) = comparator( res( iCond ), oper, thrh )
                enddo

                ! if satisfy, update its property
                if( comp(1) .and. comp(2) ) then
                  call mus_setSolidProp(                                       &
                    &    property = scheme(iScheme)%glob%levelDesc(iLevel)%    &
                    &                                         property(iElem), &
                    &    solidify = geometry%geomIncr(iGeom)%solidify )
                endif
                ! and the ones of the global tree
                call mus_setSolidProp(                                         &
                  &          property = geometry%tree%ElemPropertyBits(iElem), &
                  &          solidify = geometry%geomIncr(iGeom)%solidify )

              endif ! check element property
            enddo ! iElem
          enddo ! iLevel
        enddo ! iVar
      enddo ! iScheme

    enddo ! iGeom

    ! Communicate the property bits of halo elements for all schemes
    do iScheme = 1, size( scheme )
      do iLevel = geometry%tree%global%minLevel, geometry%tree%global%maxLevel
        call general%commPattern%exchange_long(                                &
          &  send = scheme( iScheme )%glob%levelDesc( iLevel )%sendbuffer,     &
          &  recv = scheme( iScheme )%glob%levelDesc( iLevel )%recvbuffer,     &
          &  state = scheme( iScheme )%glob%levelDesc( iLevel )%property,      &
          &  message_flag = iLevel,                                            &
          &  proc = general%proc )
        
        ! Invoke the update connectivity routine which updates the neighbor
        ! array of target elements and the neighbor array of their neighbors
        ! This has to be done for all schemes and not only for that scheme on 
        ! which GI is currently dependent upon. This will ensure change in 
        ! flow properties for all schemes
        call update_connectivity( scheme(iScheme), pdfGlobal, iLevel )
      end do
    end do ! Updation of all schemes

  end subroutine mus_wssg_model
!******************************************************************************!


!******************************************************************************!
  !> This routine update the solid property of an element
  !! If solidify == true , set prp_fluidify
  !! If solidify == false, clean prp_fluidify
  subroutine mus_setSolidProp( property, solidify )
    !---------------------------------------------------------------------------
    integer( kind=long_k ) :: property
    logical :: solidify
    !---------------------------------------------------------------------------

    if( solidify ) then
      ! set solid property (prp_fluidify)
      property = ibset( property, prp_fluidify )
    else
      ! clean solid property (prp_fluidify)
      property = ibclr( property, prp_fluidify )
    endif

  end subroutine mus_setSolidProp
!******************************************************************************!


end module mus_geomIncr_module
!******************************************************************************!
