! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains the definition of geometry type

!! \author Kartik Jain

module mus_geom_module
  use mpi
!$ use omp_lib

  ! include treelm modules
  use treelmesh_module,        only: treelmesh_type
  use env_module,              only: rk, long_k, globalMaxLevels, rk_mpi,      &
    &                                PathLen, tem_create_EndianSuffix, newunit
  use tem_balance_module,      only: tem_balance_type, Bal_logUnit
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_aux_module,          only: tem_abort
  use tem_topology_module,     only: tem_levelOf
  use tem_construction_module, only: tem_levelDesc_type
  use tem_logging_module,      only: logUnit
  use tem_tools_module,        only: tem_horizontalSpacer 
  use tem_dyn_array_module,    only: dyn_intArray_type

  ! include musubi modules
  use mus_geomIncrHead_module, only: mus_geomIncrHead_type
  use mus_param_module,        only: mus_param_type
  use mus_pdf_module,          only: pdf_globInfo_type, pdf_global_type
  use mus_field_module,        only: mus_field_type

  implicit none

  private

  public mus_geom_type
  public mus_geom_setWeight
  public mus_geom_setLevelWeight
  public mus_geom_cleanWeight
  public mus_geom_dumpWeights

  !> Geometric information and definitions
  type mus_geom_type
    !> tree data type
    type(treelmesh_type)                      :: tree
    !> boundary information as stored on disk
    type( tem_bc_prop_type )                  :: boundary
    !> The header type containing all the geometry increase information
    type( mus_geomIncrHead_type), allocatable :: geomIncr(:)
    !> Logical to define whether geometry increase is active or not
    logical                                   :: dynamicGeom = .false.
    ! todo: SZ: description is missing
    real(kind=rk), allocatable                :: weight(:)
    integer(kind=long_k), allocatable         :: offset(:)
    real(kind=rk)                             :: levelWeight(globalMaxLevels)
  end type mus_geom_type

contains

!******************************************************************************!
  !> set the weight for each level ( initial load balancing )
  !!
  subroutine mus_geom_setLevelWeight( params, geometry, balance, levelWeight )
    !---------------------------------------------------------------------------
    !> parameters
    type( mus_param_type ), intent(inout)    :: params
    !> geometry information
    type( mus_geom_type ), intent(inout) :: geometry
    !> load balancing info
    type( tem_balance_type ), intent(in) :: balance
    !> weights for each level
    real(kind=rk), intent(out)           :: levelWeight(globalMaxLevels)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    ! Set level weight
    levelWeight = 1._rk
    if( trim( balance%balanceKind ) == 'none' .or. .not. balance%static) then
      write(logUnit(1),*)'Enforcing same number of elements on all processes.'
    else
      write(logUnit(1),*)'Initial load balancing with level weight: ',        &
        &             real(params%scaleFactor )
      do iLevel = 1, globalMaxLevels
        levelWeight( iLevel ) = real(params%scaleFactor,kind=rk)**( iLevel ) 
      end do
    end if
    call tem_horizontalSpacer(fUnit = logUnit(1))
  end subroutine mus_geom_setLevelWeight
!******************************************************************************!


!******************************************************************************!
  !> set weights for each element in the treeIDlist
  !!
  subroutine mus_geom_setWeight( params, geometry, balance, glob, pdfGlobal,   &
    &                            field)
    !---------------------------------------------------------------------------
    !> parameters
    type( mus_param_type ),       intent(inout)      :: params
    !> geometry information
    type( mus_geom_type ),    intent(inout)          :: geometry
    !> load balancing information
    type( tem_balance_type ), intent(in)             :: balance
    !> Glob type 
    type(pdf_global_type), intent(in), optional      :: glob
    !> PdfGlobal type
    type( pdf_globInfo_type ), intent(in), optional  :: pdfGlobal
    !> The field type
    type( mus_field_type ), intent(in), optional     :: field
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, refLevel, level
    integer :: elemPos
    real(kind=rk) :: totalWeight
    !---------------------------------------------------------------------------
    allocate( geometry%weight( geometry%tree%nElems ))
    refLevel = geometry%tree%global%minLevel

    write(logUnit(1),*)'Setting weights for load balancing...'
    select case( trim(balance%balanceKind ))
    case( 'simple' )
      write(logUnit(1),*)'  Using base weight:  ', real( params%scaleFactor )
      do iElem = 1, geometry%tree%nElems
        level = tem_levelOf( geometry%tree%treeID( iElem ))
        ! set weight to fac^( level - minlevel ) 
        !                 acoustic      diffusive
        !  ------------------------------------------------
        !  fac =         |   2              4
        !  ------------------------------------------------
        !  minLevel      |   1              1
        !  minLevel + 1  |   2              4
        !  minLevel + 2  |   4              16
        geometry%weight( iElem ) =                                             & 
          &              real( params%scaleFactor, kind=rk )**( level-refLevel )
      enddo

    ! Setting level wise weights for dynamic load balancing
    case('levelwise')
      write(logUnit(1),*)'Using levelwise weights'  
      do iElem = 1, geometry%tree%nElems
        level = tem_levelOf( geometry%tree%treeID( iElem ))
        geometry%weight( iElem ) = params%level_weight( level )
      enddo
    ! As a special case, we treat the boundary elements with additional
    ! weights. The boundary computations especially in the case of complex
    ! higher order boundaries is expensive and causes load imbalance. Weights
    ! which were computed before are just added to elements which are
    ! boundary. GLOB AND PDFGLOBAL HAVE TO BE PRESENT FOR THIS OPERATION 
    case('boundary')
      write(logUnit(1),*)'Using levelwise weights and boundary weights'
      ! First set the level weights for all elements
      do iElem = 1, geometry%tree%nElems
        level = tem_levelOf( geometry%tree%treeID( iElem ))
        geometry%weight( iElem ) = params%level_weight( level )
      enddo
      ! Set the additional weights for boundary elements
      if( present(glob) .and. present(pdfGlobal) ) then
        do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
          if( glob%levelDesc(iLevel)%bc_elemBuffer%nVals > 0 ) then
            do iElem = 1, glob%levelDesc(iLevel)%bc_elemBuffer%nVals 
              elemPos = glob%levelDesc( iLevel )%bc_elemBuffer%val( iElem )
              geometry%weight( elemPos ) = params%level_weight_bnd( iLevel )
            end do
          end if ! nBCs>0
        end do ! iLevel
      end if ! IF present

    case default
      write(logUnit(1),*) ' Given load balancing not found in Musubi'//        &
        &             trim(balance%balanceKind )
      call tem_abort
    end select
    call tem_horizontalSpacer(fUnit=Bal_logUnit(3))
    ! Calculate the totalweight of this partition
    totalWeight = 0._rk
    do iElem = 1, geometry%tree%nElems
      totalWeight = totalWeight + geometry%weight( iElem )
    end do
    write(Bal_logUnit(3),*)'Totalweight: ', totalWeight
    call tem_horizontalSpacer(fUnit=Bal_logUnit(3))

  end subroutine mus_geom_setWeight
!******************************************************************************!


!******************************************************************************!
  !> delete weight array
  !!
  subroutine mus_geom_cleanWeight( geometry )
    !---------------------------------------------------------------------------
    !> geometry information
    type( mus_geom_type ), intent(inout)  :: geometry
    !---------------------------------------------------------------------------
    if( allocated(geometry%weight) ) deallocate( geometry%weight)

  end subroutine mus_geom_cleanWeight
!******************************************************************************!


!******************************************************************************!
  !> Dump weights to a file.
  !!
  subroutine mus_geom_dumpWeights(tree, weight, filename)
    !---------------------------------------------------------------------------
    type(treelmesh_type), intent(in) :: tree
    real(kind=rk), intent(in) :: weight(:)
    character(len=*), optional :: filename
    !---------------------------------------------------------------------------
    integer :: globElems, elemOff, locElems
    integer :: ErrErr
    integer :: ftype
    integer :: iError
    integer :: funit
    integer :: iostatus( MPI_STATUS_SIZE )
    character(len=PathLen) :: WeightsFileName
    character(len=4) :: EndianSuffix
    character(len=100) :: IOError
    integer :: resultlen = 100
    !---------------------------------------------------------------------------

    ErrErr  = 0
    iError  = 0
    IOError = ''

    globElems = int(tree%global%nElems)
    elemOff   = int(tree%elemOffset)
    locElems  = int(tree%nElems)

    EndianSuffix = tem_create_EndianSuffix()

    if( present( filename ))then
      WeightsFileName = trim(filename)//EndianSuffix
    else
      ! define the filename to dump weights to
      WeightsFileName = trim(tree%global%dirname) // 'weights' // EndianSuffix
    end if

    ! MPI_TYPE_CREATE_SUBARRAY(ndims, array_of_sizes, array_of_subsizes,
    !                          array_of_starts, order, oldtype, newtype)
    call MPI_Type_create_subarray( 1, [ globElems ], [ locElems ], &
      &                            [ elemOff ], MPI_ORDER_FORTRAN, &
      &                            rk_mpi,                         &
      &                            ftype, iError)

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      write(logUnit(0),*) 'Error creating the subarray MPI WARNING: '//        &
        &                 trim(IOError)
      call tem_abort()
    end if

    ! commit the new contiguous type
    call MPI_Type_commit( ftype, iError )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      write(logUnit(0),*) 'Error commiting the subarray MPI WARNING: '//       &
        &                 trim(IOError)
      call tem_abort()
    end if

    ! get a new file unit
    funit = newunit()

    ! open the binary file for MPI I/O
    call MPI_File_open( tree%global%comm,                &
      &                 trim(WeightsFileName),           &
      &                 MPI_MODE_WRONLY+MPI_MODE_CREATE, &
      &                 MPI_INFO_NULL, funit,            &
      &                 iError                           )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      write(logUnit(0),*) 'Error opening the file MPI WARNING: '//trim(IOError)
      call tem_abort()
    end if

    ! set the view of each process on the file opened above
    call MPI_File_set_view( funit, 0_MPI_OFFSET_KIND, &
      &                     rk_mpi, ftype, "native",  &
      &                     MPI_INFO_NULL, iError     )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      write(logUnit(0),*) 'Error setting the file view MPI WARNING: '//        &
        &                 trim(IOError)
      call tem_abort()
    end if

    ! Dump the weights to disk.
    call mpi_file_write_all( funit, weight,    &
      &                      locElems, rk_mpi, &
      &                      iostatus, iError  )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      write(logUnit(0),*) 'Error writing the data MPI WARNING: '//trim(IOError)
      call tem_abort()
    end if

    ! free the mpi type
    call MPI_TYPE_FREE( ftype, iError )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      write(logUnit(0),*) 'Free Type MPI WARNING: '// trim(IOError) 
      call tem_abort()
    end if

    ! Close the weights file again.
    call MPI_File_close( funit, iError )

    if (iError .ne. MPI_SUCCESS) then
      call MPI_ERROR_STRING(iError, IOError, resultlen, ErrErr)
      write(logUnit(0),*) 'Error closing the file MPI WARNING: '//trim(IOError)
      call tem_abort()
    end if

  end subroutine mus_geom_dumpWeights
!******************************************************************************!


end module mus_geom_module
!******************************************************************************!
