!******************************************************************************!
!> This module contains routines which initiliaze advection relaxation and
!! flow field for multispecies lbm gas model and liquid model.
module mus_initMultispecies_module

  ! include treelm modules
  use env_module,         only: rk
  use tem_aux_module,     only: tem_abort
  use tem_logging_module, only: logUnit

  ! include musubi modules
  use mus_MSGas_module, only: bgk_advRel_MSGas,                                &
    &                         bgk_advRel_d3q19f3_MSGas
  use mus_MSLiquid_module, only: bgk_advRel_MSLiquid,                          &
    &                            bgk_advRel_MSLiquid_WTDF,                     &
    &                            bgk_advRel_MSLiquid_WTDF_2Spc,                &
    &                            bgk_advRel_d3q19f3_MSLiquid,                  &
    &                            bgk_forcing_advRel_MSLiquid,                  &
    &                            mrt_advRel_MSLiquid,                          &
    &                            mrt_advRel_d3q19f3_MSLiquid,                  &
    &                            externalForce_MSLiquid,                       &
    &                            externalForce_MSLiquid_WTF
  use mus_scheme_type_module, only: mus_scheme_type

  implicit none

  private

  public :: mus_init_advRel_multispecies_gas
  public :: mus_init_advRel_multispecies_liquid

contains

!******************************************************************************!
  !> Initialize the relaxation model for multispecies gas model
  subroutine mus_init_advRel_multispecies_gas( scheme )
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: scheme !< scheme type 
    !---------------------------------------------------------------------------
    write(logUnit(1),*) 'Choosing MULTI-SPECIES GAS relaxation model: '        &
      &                //trim( scheme%header%relaxation )
    select case( trim( scheme%header%relaxation ))
    case( 'bgk' )
      select case( trim(scheme%header%layout) )
      case( 'd3q19' )
        if(scheme%nFields==3 ) then
          write(logUnit(1),*) 'Choosen optimized d3q19f3 multispecies '//      &
            &                 'gas kernel'
          scheme%compute => bgk_advRel_d3q19f3_MSGas
        else
          write(logUnit(1),*) 'Choosen testing multispecies gas kernel'
          scheme%compute => bgk_advRel_MSGas
        end if  
      case( 'd2q9' )
        write(logUnit(1),*) 'Choosen testing multispecies gas kernel'
        scheme%compute => bgk_advRel_MSGas
      case default
        write(logUnit(1),*) 'The selected stencil is not supported at '//      &
          &                 'this moment: '//trim(scheme%header%layout) 
        call tem_abort()
      end select
    end select
  end subroutine mus_init_advRel_multispecies_gas  

!******************************************************************************!
  !> Initialize the relaxation model for multispecies gas model
  subroutine mus_init_advRel_multispecies_liquid( scheme )
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: scheme !< scheme type 
    !---------------------------------------------------------------------------
    write(logUnit(1),*) 'Choosing MULTI-SPECIES LIQUID relaxation model: '     &
      &                //trim( scheme%header%relaxation )
 
    ! only if external force is defined
    !write(*,*) 'active ', scheme%ext_force%active
    !stop
    if ( scheme%ext_force%active ) &
      & scheme%compute_force => externalForce_MSLiquid

    select case( trim( scheme%header%relaxation ))
    case( 'bgk' )
      select case( trim(scheme%header%layout) )
      case( 'd3q19' )
        if(scheme%nFields==3 ) then
          write(logUnit(1),*) 'Choosen optimized d3q19f3 multispecies '//      &
            &                 'liquid kernel'
          scheme%compute => bgk_advRel_d3q19f3_MSLiquid
        else
          scheme%compute => bgk_advRel_MSLiquid
        end if  
      case( 'd2q9' )
        scheme%compute => bgk_advRel_MSLiquid
      case default
        write(logUnit(1),*) 'The selected stencil is not supported at '//      &
          &                 'this moment: '//trim(scheme%header%layout) 
        call tem_abort()
      end select
    case( 'bgk_forcing' )
      select case( trim(scheme%header%layout) )
      case( 'd3q19' , 'd2q9')
        write(logUnit(1),*) 'Choosen unoptimized bgk forcing multispecies '//  &
          &                 'liquid kernel'
        scheme%compute => bgk_forcing_advRel_MSLiquid
      case default
        write(logUnit(1),*) 'The selected stencil is not supported at '//      &
          &                 'this moment: '//trim(scheme%header%layout) 
        call tem_abort()
      end select
    case( 'bgk_withthermodynfac' )
      ! overwrite external force function pointer if scheme is withthermodynfac
      if ( scheme%ext_force%active ) &
        & scheme%compute_force => externalForce_MSLiquid_WTF
      select case( trim(scheme%header%layout) )
      case( 'd3q19' , 'd2q9')
        if(scheme%nFields==2 ) then
          write(logUnit(1),*) 'Choosen 2-species with 3-species '//            &
            &                 'thermodynamic factor liquid kernel'
          scheme%compute => bgk_advRel_MSLiquid_WTDF_2Spc
        else  
          write(logUnit(1),*) 'Choosen unoptimized bgk with thermodynamic '//  &
            &             'factor multispecies liquid kernel'
          scheme%compute => bgk_advRel_MSLiquid_WTDF
        endif  
      case default
        write(logUnit(1),*) 'The selected stencil is not supported at '//      &
          &             'this moment: trim(scheme%header%layout)'
        call tem_abort()
      end select
    case( 'mrt' )
      select case( trim(scheme%header%layout) )
      case( 'd3q19' )
        if(scheme%nFields==3 ) then
          write(logUnit(1),*) 'Choosen optimized mrt d3q19f3 multispecies '//  &
            &                 'liquid kernel'
          scheme%compute => mrt_advRel_d3q19f3_MSLiquid
        else
          write(logUnit(1),*) 'Choosen unoptimized mrt multispecies liquid '// &
            &                 'kernel'
          scheme%compute => mrt_advRel_MSLiquid
        end if  
      case( 'd2q9' )
        write(logUnit(1),*) 'Choosen unoptimized mrt multispecies liquid kernel'
        scheme%compute => mrt_advRel_MSLiquid
      case default
        write(logUnit(1),*) 'The selected stencil is not supported at '//      &
          &                 'this moment: '//trim(scheme%header%layout) 
        call tem_abort()
      end select
    case default
      write(logUnit(1),*) 'The selected relaxation model is not supported: '   &
        &              //trim( scheme%header%relaxation )
      call tem_abort()
    end select
  end subroutine mus_init_advRel_multispecies_liquid

end module mus_initMultispecies_module

