! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Linear Interpolation of flow quantities between different grid levels
!!
!! \section intp_main Interpolation 
!! The routines defined here, fill up the ghost elements with valid data.
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with 
!! valid values.
!! Source elements are the fluid elements from other levels, from where to 
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source 
!! fluid elements on the source level.
!!
!! \ref tem_topology_module For a detailed description of the grid 
!!
!! \section workflow Workflow
!!
!! Each interpolation routine acts on a list of ghost elements.
!! This list contains pointers to the position in the total list.
!! For each of these ghost elements, the source elements are identified.
!! Before that, the sourceLevel is identified. However, the code is restricted
!! to work with a level jump of only one level, so the sourceLevel is:\n
!! for fillMyGhostsFromFiner: sourceLevel = targetLevel + 1
!! for fillFinerGhostsFromMe: sourceLevel = targetLevel - 1
!!
!! For an overview over implemented interpolation methods, see
!! \ref intp_methods "Interpolation methods"
!!
!!\author Manuel Hasert
!!
?? include 'header/lbm_macros.inc'
module mus_interpolate_linear_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_element_module,    only: eT_GhostFromCoarserElem,                    &
    &                              eT_ghostFromFinerElem
  use tem_grow_array_module, only: grw_intArray_type
  use tem_param_module,      only: cs2inv, cs2
  use tem_var_system_module, only: tem_var_system_type
  use tem_comm_env_module,   only: tem_comm_env_type
  ! use tem_debug_module,      only: dbgUnit

  ! include musubi modules
  use mus_field_prop_module,         only: mus_field_prop_type
  use mus_interpolate_header_module, only: interpolation_type,                 &
                                           determine_timeLayer
  use mus_pdf_module,                only: pdf_data_type, pdf_global_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_variable_module,           only: mus_derVarPos_type
  use mus_derivedQuantities_module2, only: convPrePost, getNEq_diffusive,      &
    &                                      getShearRateTensor_diffusive
  use mus_physics_module,            only: mus_physics_type
  use mus_moments_module,            only: get_moment

  implicit none

  private

  public :: fillMyGhostsFromFiner_linearDiffusive
  public :: fillFinerGhostsFromMe_linearDiffusive
  public :: fillMyGhostsFromFiner_linear
  public :: fillFinerGhostsFromMe_linear
  public :: fillMyGhostsFromFiner_linearIncomp
  public :: fillFinerGhostsFromMe_linearIncomp
  public :: fillMyGhostsFromFiner_linMoment2d
  public :: fillMyGhostsFromFiner_linMoment3d

  contains

!******************************************************************************!
  !> \ref intp_linear "Linear interpolation" of ghostFromFiner
  !!
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linearDiffusive )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iVal, jVal     ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,intp%nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetVel(3)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(intp%nMaxSources)
    real(kind=rk) :: targetRho  !< macroscopic density for all source elements
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on
    integer :: t_source ! which time layer to work on
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    !> Strain tensor of the source elements
    real(kind=rk) :: sourceS(3,3, intp%nMaxSources)
    !> Strain tensor of the target element
    real(kind=rk) :: targetS(3,3)
    real(kind=rk) :: rho0
    !---------------------------------------------------------------------------

    sourceLevel = level + 1
    targetLevel = level
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    rho0 = fieldProp( iVar )%fluid%rho0

! DEBUG output
! write(dbgUnit(1),*) '----- Enter fillMyGhostsFromFiner_linearDiffusive -----------'
! write(dbgUnit(1),*) 'source level: ', sourceLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( sourceLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( sourceLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( sourceLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( sourceLevel )%strainRate
! write(dbgUnit(1),*) 'target level: ', targetLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( targetLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( targetLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( targetLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( targetLevel )%strainRate

    ! Treat all coarse target elements
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( level )                             &
        &                      %offset( 1, eT_ghostFromFinerElem )
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceELems = glob%levelDesc( level )%depFromFiner( iElem )%elem%nVals

      ! Compute sourceRho and velocity for each field and interpolate pdf
      ! field wise
        sourceRho = 0._rk
        sourceVel = 0._rk
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems

          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )     &
            &                                       %elem%val( iSourceElem )

          ! Calculate Rho and Velocity
          do iDir = 1, layout%stencil(1)%QQ
             fTmp(iDir) = pdf(sourceLevel)%state(                              &
& ?SAVE?( varSys%variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
        & t_source )
             sourceRho(iSourceElem) = sourceRho(iSourceElem) + fTmp(iDir)
             ! momentum = velocity for diffusive scaling (incompressible)
             sourceVel(:,iSourceElem) = sourceVel(:,iSourceElem)               &
               &       + fTmp(iDir) * layout%stencil(1)%cxDirRK(:,iDir)
          enddo

          ! Calculate Strain rate
          sourceS(:,:,iSourceElem ) = getShearRateTensor_diffusive( &
            &                               subset = fTmp, layout = layout,   &
            &               omega = fieldProp( iVar )%fluid%omLvl( sourceLevel ))

          ! Convert Rho, Vel, Strain from LB to physical
          ! pressure is actually only the fluctuating part
          sourceRho(iSourceElem)   =   (sourceRho(iSourceElem) - rho0) * cs2  &
            &                        * physics%fac( sourceLevel )%press
          sourceVel(:,iSourceElem) =   sourceVel(:,iSourceElem)               &
            &                        * physics%fac( sourceLevel )%vel
          sourceS(:,:,iSourceElem) =   sourceS(:,:,iSourceElem)               &
            &                        * physics%fac( sourceLevel )%strainRate
          !-------------- done converting

! write(dbgUnit(1),"(A, F12.5, A, 3F12.5, A, 3F12.5)") &
!   &       'Source pressPhy: ', sourceRho(iSourceElem), &
!   &       '   vel: ', sourceVel(1,iSourceElem), sourceVel(2,iSourceElem), sourceVel(3,iSourceElem),&
!   &       '   S: ', sourceS(1,1,iSourceElem), sourceS(2,2,iSourceElem), sourceS(1,2,iSourceElem)
        end do ! iSourceElem

        !------------------ LINEAR INTERPOLATION -------------------
        ! This is all done in physical units
        ! Ok, now I have: Rho(pressure), Vel, Strain of all my neighbor elements.
        ! Interpolate by average
        targetRho    = sum(sourceRho(1:nSourceElems))   / real(nSourceElems,rk)
        targetVel(1) = sum(sourceVel(1,1:nSourceElems)) / real(nSourceElems,rk)
        targetVel(2) = sum(sourceVel(2,1:nSourceElems)) / real(nSourceElems,rk)
        targetVel(3) = sum(sourceVel(3,1:nSourceElems)) / real(nSourceElems,rk)
        ! Interpolate the shear stress
        targetS = 0._rk
        do jVal = 1, layout%stencil(1)%nDims
          do iVal = 1, layout%stencil(1)%nDims
          targetS(iVal,jVal) =                                               &
            & sum( sourceS(iVal,jVal,1:nSourceElems)) / real(nSourceElems,rk)
          end do
        end do
        !------------------ LINEAR INTERPOLATION -------------------

! write(dbgUnit(1),"(A, F12.5, A, 3F12.5, A, 3F12.5)") &
!   &       'Target pressPhy: ', targetRho,&
!   &       '   vel: ', targetVel(1), targetVel(2), targetVel(3),&
!   &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
! write(dbgUnit(1),"(A)") ''

        ! Convert target Rho (pressure), Vel, Strain from physical to LB
        targetRho    = targetRho * cs2inv / physics%fac( targetLevel )%press + rho0
        targetVel(:) = targetVel(:)       / physics%fac( targetLevel )%vel
        targetS(:,:) = targetS(:,:)       / physics%fac( targetLevel )%strainRate

        ! Calculate fEq
?? copy :: Eq_incomp( targetEq, targetRho, targetVel, layout, rho0 )

        ! Calculate fNeq
        targetNeq(:) = getnEq_diffusive(                                          &
          &               layout = layout,                                        &
          &               omega  = fieldProp( iVar )%fluid%omLvl( targetLevel ),  &
          &               Sxx    = targetS(:,:))

        ! Now write the resulting pdf in the current direction to the target
        ! Element position resulting pdf = f_eq + fac*f_neq
        do iDir = 1, layout%stencil(1)%QQ
          pdf( targetLevel )%state(                                            &
&?SAVE?( varSys%Variable%val(iVar)%varPos( iDir ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
           & t_target ) = targetEq( iDir ) + targetNeq( iDir )
        end do ! iDir

! DEBUG output
! Double check strain rate from pdf
! fTmp = targetEq + targetNeq
! targetRho = sum(fTmp) * cs2 * physics%fac( targetLevel )%press
! targetVel(1) = sum( fTmp(:) * real( layout%stencil(1)%cxDir(1,:), rk) )
! targetVel(2) = sum( fTmp(:) * real( layout%stencil(1)%cxDir(2,:), rk) )
! targetVel(3) = sum( fTmp(:) * real( layout%stencil(1)%cxDir(3,:), rk) )
! targetVel = targetVel * physics%fac( targetLevel )%vel
! targetS = getShearRateTensor_diffusive( ftmp, &
!   &                             fieldProp(1)%fluid%omLvl( targetLevel ), &
!   &                             layout ) * physics%fac( targetLevel )%strainRate
! write(dbgUnit(1),"(A, F12.5, A, 3F12.5, A, 3F12.5)") &
!   &       'Target pressPhy: ', targetRho,&
!   &       '   vel: ', targetVel(1), targetVel(2), targetVel(3),&
!   &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
! write(dbgUnit(1),"(A)") ''

    end do

! write(dbgUnit(1), *) '---- Leave fillMyGhostsFromFiner_linearDiffusive -----------'
! write(dbgUnit(1), *) ''

  end subroutine fillMyGhostsFromFiner_linearDiffusive
!******************************************************************************!


!******************************************************************************!
  !>  \ref intp_linear "Linear interpolation" of ghostFromCoarser
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_linearDiffusive )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iVal, jVal     ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,intp%nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetVel(3)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(intp%nMaxSources)
    real(kind=rk) :: targetRho  !< macroscopic density for all source elements
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on for the target elements
    integer :: t_source ! which time layer to work on for the source elements
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: weight( intp%nMaxSources )
    ! Shear tensor of the source elements
    real(kind=rk) :: sourceS(3,3, intp%nMaxSources)
    ! Shear tensor of the target element
    real(kind=rk) :: targetS(3,3)
    real(kind=rk) :: rho0
    !--------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    rho0 = fieldProp( iVar )%fluid%rho0

! DEBUG output
! write(dbgUnit(1),*) '----- Enter fillFinerGhostsFromMe_linearDiffusive -----------'
! write(dbgUnit(1),*) 'source level: ', sourceLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( sourceLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( sourceLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( sourceLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( sourceLevel )%strainRate
! write(dbgUnit(1),*) 'target level: ', targetLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( targetLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( targetLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( targetLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( targetLevel )%strainRate

    ! Treat all fine target elements
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromCoarserElem )
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )     &
        &                 %elem%nVals

      !Compute rho and velocity for each field and interpolate pdf field wise
        sourceRho = 0._rk
        sourceVel = 0._rk

        ! Now loop over all fine source elements for this target
        do iSourceElem = 1, nSourceElems
          ! Get the source element position
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                       %elem%val( iSourceElem )

          ! Read the pre-calculated weights (like in Average intp.)
          weight( iSourceElem ) = glob%levelDesc( targetLevel )                &
            &                     %depFromCoarser( iElem )%weight( iSourceElem )

          ! Calculate Rho and Vel
          do iDir = 1, layout%stencil(1)%QQ
             fTmp(iDir) = pdf(sourceLevel)%state(                              &
& ?SAVE?( varSys%variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
        & t_source )
             sourceRho(iSourceElem)   = sourceRho(iSourceElem) + fTmp(iDir)
             sourceVel(:,iSourceElem) = sourceVel(:,iSourceElem)           &
               & + fTmp(iDir) * real(layout%stencil(1)%cxDir(:,iDir), rk)
          enddo

          ! Calculate Strain rate
          sourceS(:,:,iSourceElem) = getShearRateTensor_diffusive( &
            &           subset = fTmp, layout = layout,   &
            &           omega = fieldProp( iVar )%fluid%omLvl( sourceLevel ))

          ! Convert Rho, Vel, Strain from LB to physical
          ! convert rho to pressure to physical
          sourceRho(iSourceElem)   =   (sourceRho(iSourceElem) - rho0 ) * cs2 &
            &                        * physics%fac( sourceLevel )%press
          sourceVel(:,iSourceElem) =   sourceVel(:,iSourceElem)             &
            &                        * physics%fac(sourceLevel)%vel
          sourceS(:,:,iSourceElem) =   sourceS(:,:,iSourceElem)             &
            &                        * physics%fac(sourceLevel)%strainRate

! DEBUG output: check rho, vel, strain of source elements
! write(dbgUnit(1),"(A, F12.5, A, 3F12.5, A, 3F12.5)") &
!   &       'Source pressPhy: ', sourceRho(iSourceElem), &
!   &       '   vel: ', sourceVel(1,iSourceElem), sourceVel(2,iSourceElem), sourceVel(3,iSourceElem),&
!   &       '   S: ', sourceS(1,1,iSourceElem), sourceS(2,2,iSourceElem), sourceS(1,2,iSourceElem)
        end do ! iSourceElem

        !------------------ LINEAR INTERPOLATION -------------------
        ! linearly interpolate Rho, Vel, and Strain
        targetRho    = sum( weight(1:nSourceElems)*sourceRho(1:nSourceElems)   )
        targetVel(1) = sum( weight(1:nSourceElems)*sourceVel(1,1:nSourceElems) )
        targetVel(2) = sum( weight(1:nSourceElems)*sourceVel(2,1:nSourceElems) )
        targetVel(3) = sum( weight(1:nSourceElems)*sourceVel(3,1:nSourceElems) )
        targetS = 0._rk
        do jVal = 1, layout%stencil(1)%nDims
          do iVal = 1, layout%stencil(1)%nDims
          targetS(iVal,jVal) =                                   &
            & sum( weight(1:nSourceElems) * sourceS(iVal,jVal,1:nSourceElems) )
          end do
        end do
        !------------------ LINEAR INTERPOLATION -------------------

! DEBUG output: check rho, vel, strain of target elements
! write(dbgUnit(1),"(A, F12.5, A, 3F12.5, A, 3F12.5)") &
!   &       'Target pressPhy: ', targetRho,&
!   &       '   vel: ', targetVel(1), targetVel(2), targetVel(3),&
!   &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
! write(dbgUnit(1),"(A)") ''

        !---------------------------------------------
        ! Convert from physical to LB
        targetRho    = targetRho * cs2inv / physics%fac( targetLevel )%press + rho0
        targetVel(:) = targetVel(:)       / physics%fac( targetLevel )%vel
        targetS(:,:) = targetS(:,:)       / physics%fac( targetLevel )%strainRate

        ! Calculate fEq and fNeq
        targetEq  = 0._rk
        targetNEq = 0._rk
?? copy :: Eq_incomp( targetEq, targetRho, targetVel, layout, rho0 )

        targetNeq(:) = getnEq_diffusive(                                       &
          &            layout = layout,                                        &
          &            omega  = fieldProp( iVar )%fluid%omLvl( targetLevel ),  &
          &            Sxx    = targetS(:,:))

        ! Now write pdf (fEq+fNeq) to the target element
        do iDir = 1, layout%stencil(1)%QQ
          pdf( targetLevel )%state(                                            &
&?SAVE?( varSys%Variable%val( iVar )%varPos( iDir ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ), &
            & t_target ) = targetEq( iDir ) + targetNeq( iDir )
        end do ! iDir

! DEBUG output: Double check strain rate from pdf
! fTmp = targetEq + targetNeq
! targetRho = sum(fTmp) * cs2 * physics%fac( targetLevel )%press
! targetVel(1) = sum( fTmp(:) * real( layout%stencil(1)%cxDir(1,:), rk) )
! targetVel(2) = sum( fTmp(:) * real( layout%stencil(1)%cxDir(2,:), rk) )
! targetVel(3) = sum( fTmp(:) * real( layout%stencil(1)%cxDir(3,:), rk) )
! targetVel = targetVel * physics%fac( targetLevel )%vel
! targetS = getShearRateTensor_diffusive( ftmp, &
!   &                             fieldProp(1)%fluid%omLvl( targetLevel ), &
!   &                             layout ) * physics%fac( targetLevel )%strainRate
! write(dbgUnit(1),"(A, F12.5, A, 3F12.5, A, 3F12.5)") &
!   &       'Target pressPhy: ', targetRho,&
!   &       '   vel: ', targetVel(1), targetVel(2), targetVel(3),&
!   &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
! write(dbgUnit(1),"(A)") ''

    end do ! indElem

! write(dbgUnit(1),*) '----- Leave fillFinerGhostsFromMe_linearDiffusive -----------'
  end subroutine fillFinerGhostsFromMe_linearDiffusive
!******************************************************************************!


!******************************************************************************!
  !> \ref intp_linear "Linear interpolation" of ghostFromFiner
  !! The interpolation procedure used in this routine is:\n
  !! 1. calculate rho(pressure), velocity, fNeq from source elements
  !! 2. interpolate rho and velocity to target element
  !! 3. calculate fEq of target element from rho and velocity
  !! 4. interpolation fNeq from source elements to target element
  !! 5. calculate pdf of target element = fEq + fac * fNeq
  !!
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linear )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,intp%nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetVel(3)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(intp%nMaxSources)
    real(kind=rk) :: targetRho  ! macroscopic density for all source elements
    real(kind=rk) :: fac        ! scaling factor
    ! temporary equilirbium part
    real(kind=rk) :: fEq( layout%stencil(1)%QQ )
    ! current equilirbium part
    ! real(kind=rk) :: sourceEq( layout%stencil(1)%QQ, intp%nMaxSources ) 
    ! current non-equilibrium part
    real(kind=rk) :: sourceNEq( layout%stencil(1)%QQ, intp%nMaxSources )
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: srcRho, srcVel(3)
    !---------------------------------------------------------------------------
    sourceLevel = level + 1
    targetLevel = level
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

! DEBUG output
! write(dbgUnit(1),*) '----- Enter fillMyGhostsFromFiner_linearAcoustic -----------'
! write(dbgUnit(1),*) 'source level: ', sourceLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( sourceLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( sourceLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( sourceLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( sourceLevel )%strainRate
! write(dbgUnit(1),*) 'target level: ', targetLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( targetLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( targetLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( targetLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( targetLevel )%strainRate

    ! Treat all coarse target elements
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element
      targetElem = iElem + glob%levelDesc( level )                             &
        &                      %offset( 1, eT_ghostFromFinerElem )
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceELems = glob%levelDesc( level )%depFromFiner( iElem )%elem%nVals

      ! Compute sourceRho and velocity for each field and interpolate pdf
      ! field wise
      do iVar = 1, varSys%nVars
        ! Relationship between shear stress on different levels 
        ! We know that f = f^neq + f^eq
        ! f^eq,f = f^eq,c
        ! and for the f^neq, the following equation holds
        ! tau_ij = (1-1/(2*tau))*sum_k( f^neq*( cx_(i,k)*cx_(j,k) 
        !         - 0.5*cx_(:,k)*cx_(:,k)*d_ij))
        ! for coarse and fine, the stress must be constant
        ! tau_ij^c = tau_ij^f
        ! it results for each term of the sum
        ! (1-1/(2*tau_c))*f^neq,c = (1-1/(2*tau_f))*f^neq,f
        ! So, for fine to coarse: f^neq,c = fac*f^neq,f
        ! We are working on post-collision values.
        ! The derivations are however all done for pre-collision values, which
        ! is the intrinsically correct way for the LBE.
        ! Additionally, we need to convert these with the convPrePost factor
        fac = ((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( sourceLevel ))   &
          &       *convPrePost(fieldProp( iVar )%fluid%omLvl( sourceLevel )))  &
          &    /((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( targetLevel )) &
          &       *convPrePost(fieldProp( iVar )%fluid%omLvl( targetLevel )))

        sourceRho = 0._rk
        sourceVel = 0._rk
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems

          ! Get the source element
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )     &
            &                                       %elem%val( iSourceElem )

          ! Calculate rho and vel for equil
          do iDir = 1, layout%stencil(1)%QQ
             fTmp(iDir) = pdf(sourceLevel)%state(                              &
& ?SAVE?( varSys%variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
        & t_source )
             sourceRho(iSourceElem)   = sourceRho(iSourceElem) + fTmp(iDir)
             sourceVel(:,iSourceElem) = sourceVel(:,iSourceElem)              &
               &       + fTmp(iDir) * real(layout%stencil(1)%cxDir(:,iDir), rk)
          end do
          sourceVel(:,iSourceElem) = sourceVel(:,iSourceElem) / sourceRho(iSourceElem)

          srcRho = sourceRho( iSourceElem )
          srcVel = sourceVel( :,iSourceElem )
?? copy :: Eq_comp( fEq, srcRho, srcVel, layout )
          ! sourceEq(:,iSourceElem)  = fEq(:)
          sourceNEq(:,iSourceElem) = fTmp(:) - fEq(:)

! DEBUG output: check rho, vel of source element
! write(dbgUnit(1),"(A, F12.5, A, 3F12.5)") &
!   &       'Source pressPhy: ', sourceRho(iSourceElem), &
!   &       '   vel: ', sourceVel(1,iSourceElem), sourceVel(2,iSourceElem), sourceVel(3,iSourceElem)

        end do  ! iSourceElem

        ! Ok, now I have: rho, vel, fNeq of all my neighbor elements.
        ! Interpolate rho, vel, and fNeq by average
        targetRho    = sum(sourceRho(1:nSourceElems))   / real(nSourceElems, rk)
        targetVel(1) = sum(sourceVel(1,1:nSourceElems)) / real(nSourceElems, rk)
        targetVel(2) = sum(sourceVel(2,1:nSourceElems)) / real(nSourceElems, rk)
        targetVel(3) = sum(sourceVel(3,1:nSourceElems)) / real(nSourceElems, rk)
        ! calculate eq and store in targetEq
?? copy :: Eq_comp( targetEq, targetRho, targetVel, layout )

        do iDir = 1,layout%stencil(1)%QQ
          targetNeq(iDir) =   sum( sourceNeq(iDir,1:nSourceElems)) &
            &               / real(nSourceElems, rk)
        end do

! DEBUG output: check rho, vel of target element
! write(dbgUnit(1),"(A, F12.5, A, 3F12.5)") &
!   &       'Target pressPhy: ', targetRho,&
!   &       '   vel: ', targetVel(1), targetVel(2), targetVel(3)
! write(dbgUnit(1),"(A)") ''

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        ! resulting pdf = f_eq + fac*f_neq
        do iDir = 1, layout%stencil(1)%QQ
          pdf( targetLevel )%state(                                            &
&?SAVE?( varSys%Variable%val(iVar)%varPos( iDir ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
           & t_target ) = targetEq( iDir ) + fac * targetNeq( iDir )
        enddo ! iDir

      enddo ! ivar loop
    enddo

  end subroutine fillMyGhostsFromFiner_linear
!******************************************************************************!


!******************************************************************************!
  !> \ref intp_linear "Linear interpolation" of ghostFromFiner
  !!
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linearIncomp )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,intp%nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetVel(3)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(intp%nMaxSources)
    real(kind=rk) :: targetRho  ! macroscopic density for all source elements
    real(kind=rk) :: fac(varSys%nVars)  ! scaling factor for fNeq
    ! current equilirbium part
    ! real(kind=rk) :: sourceEq( layout%stencil(1)%QQ, intp%nMaxSources )
    ! current non-equilibrium part
    real(kind=rk) :: sourceNEq( layout%stencil(1)%QQ, intp%nMaxSources )
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on?
    integer :: t_source ! which time layer to work on?
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ ), fEq( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0, srcRho, srcVel(3)
    !---------------------------------------------------------------------------

    sourceLevel = level + 1
    targetLevel = level
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

! write(*,"(A)") "Inside fill my ghost from Finer linear incomp routine"
! write(*,"(A,I0,A,I0)") "  source level: ", sourceLevel, ", read  buffer: ", t_source
! write(*,"(A,I0,A,I0)") "  target level: ", targetLevel, ", write buffer: ", t_target

    ! Conversion factor(fac) of fNeq on different levels
    ! We know that f = f^neq + f^eq
    ! f^eq,f = f^eq,c
    ! and for the f^neq, the following equation holds
    ! tau_ij = (1-1/(2*tau))*sum_k( f^neq*( cx_(i,k)*cx_(j,k)
    !         - 0.5*cx_(:,k)*cx_(:,k)*d_ij))
    ! for coarse and fine, the stress must be constant
    ! tau_ij^c = tau_ij^f
    ! it results for each term of the sum
    ! (1-1/(2*tau_c))*f^neq,c = (1-1/(2*tau_f))*f^neq,f
    ! So, for fine to coarse: f^neq,c = fac*f^neq,f
    do iVar = 1, varSys%nVars
      fac(iVar) = ((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl(sourceLevel)) &
        &     *convPrePost(fieldProp( iVar )%fluid%omLvl( sourceLevel )))      &
        &   /((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( targetLevel ))    &
        &     *convPrePost(fieldProp( iVar )%fluid%omLvl( targetLevel )))
    end do

    ! Treat all coarse target elements
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( level )                             &
        &                      %offset( 1, eT_ghostFromFinerElem )

      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceELems = glob%levelDesc( level )%depFromFiner( iElem )%elem%nVals

      ! Compute sourceRho and velocity for each field and interpolate pdf
      do iVar = 1, varSys%nVars
        rho0 = fieldProp( iVar )%fluid%rho0

        sourceRho = 0._rk
        sourceVel = 0._rk
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems

          ! Get the source element position
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )     &
            &                                       %elem%val( iSourceElem )

          ! Calculate rho and vel for equil
          do iDir = 1, layout%stencil(1)%QQ
             fTmp(iDir) = pdf(sourceLevel)%state(                              &
& ?SAVE?( varSys%variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
        & t_source )
             sourceRho(iSourceElem)   = sourceRho(iSourceElem) + fTmp(iDir)
             sourceVel(:,iSourceElem) = sourceVel(:,iSourceElem)           &
               &        + fTmp(iDir) * real(layout%stencil(1)%cxDir(:,iDir), rk)
          enddo

          srcRho = sourceRho( iSourceElem )
          srcVel = sourceVel( :, iSourceElem )
?? copy :: Eq_incomp( fEq, srcRho, srcVel, layout, rho0 )
          sourceNEq(:, iSourceElem ) = fTmp(:) - fEq(:)

        end do ! iSourceElem

        ! Ok, now I have: rho, vel, fNeq of all my neighbor elements.
        ! interpolate rho, vel, and fNeq by average
        targetRho    = sum(sourceRho(1:nSourceElems))   / real(nSourceElems, rk)
        targetVel(1) = sum(sourceVel(1,1:nSourceElems)) / real(nSourceElems, rk)
        targetVel(2) = sum(sourceVel(2,1:nSourceElems)) / real(nSourceElems, rk)
        targetVel(3) = sum(sourceVel(3,1:nSourceElems)) / real(nSourceElems, rk)
        do iDir = 1,layout%stencil(1)%QQ
          targetNeq(iDir) =   sum( sourceNeq(iDir,1:nSourceElems)) &
            &               / real(nSourceElems, rk)
        end do

?? copy :: Eq_incomp( targetEq, targetRho, targetVel, layout, rho0 )

        ! Now write the resulting pdf (= fEq + fac*fNeq) to the target element
        do iDir = 1, layout%stencil(1)%QQ
          pdf( targetLevel )%state(                                           &
&?SAVE?( varSys%Variable%val(iVar)%varPos( iDir ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
           & t_target ) = targetEq( iDir ) + targetNeq( iDir ) * fac( iVar )
        end do ! iDir
      end do ! ivar loop
    enddo

  end subroutine fillMyGhostsFromFiner_linearIncomp
!******************************************************************************!


!******************************************************************************!
  !>  \ref intp_linear "Linear interpolation" of ghostFromCoarser
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_linearIncomp )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iVal           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,intp%nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetVel(3)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(intp%nMaxSources)
    real(kind=rk) :: targetRho  ! macroscopic density for all source elements
    real(kind=rk) :: fac(varSys%nVars)       ! scaling factor
    ! current non-equilibrium part
    real(kind=rk) :: sourceNEq( layout%stencil(1)%QQ, intp%nMaxSources )
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on for the target elements
    integer :: t_source ! which time layer to work on for the source elements
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ ), fEq( layout%stencil(1)%QQ )
    real(kind=rk) :: weight( intp%nMaxSources ) 
    real(kind=rk) :: rho0
    ! Shear tensor of the target element
    real(kind=rk) :: srcVel(3), srcRho
    !--------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    do iVar = 1, varSys%nVars
      ! See explanation for the factor in fineToCoarse
      fac(iVar) = ((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl(sourceLevel)) &
        &     *convPrePost(fieldProp( iVar )%fluid%omLvl( sourceLevel )))      &
        &   /((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( targetLevel ))    &
        &     *convPrePost(fieldProp( iVar )%fluid%omLvl( targetLevel )))
    end do

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromCoarserElem )
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )     &
        &                 %elem%nVals

      ! accumulators for target equilibrium and non-equilibrium parts
      targetEq   = 0._rk
      targetNEq  = 0._rk

      !Compute rho and velocity for each field and interpolate pdf field wise
      do iVar = 1, varSys%nVars
        rho0 = fieldProp( iVar )%fluid%rho0
        sourceRho = 0._rk 
        sourceVel = 0._rk

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                       %elem%val( iSourceElem )

          ! Read the pre-calculated weights (like in Average intp.)
          weight( iSourceElem ) = glob%levelDesc( targetLevel )                &
            &                     %depFromCoarser( iElem )%weight( iSourceElem )
          ! Calculate rho and vel for equil
          do iDir = 1, layout%stencil(1)%QQ
             fTmp(iDir) = pdf(sourceLevel)%state(                              &
& ?SAVE?( varSys%variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
        & t_source )
             sourceRho(iSourceElem) = sourceRho(iSourceElem) + fTmp(iDir)
             sourceVel(:,iSourceElem ) = sourceVel( :, iSourceElem )           &
               & + fTmp( iDir)*real( layout%stencil(1)%cxDir(:,iDir), rk)
          enddo

          srcRho = sourceRho( iSourceElem )
          srcVel = sourceVel( :, iSourceElem )
?? copy :: Eq_incomp( fEq, srcRho, srcVel, layout, rho0 )
          ! sourceEq( :,iSourceElem) = fEq(:)
          sourceNEq(:,iSourceElem) = fTmp(:) - fEq(:)

        end do ! iSourceElem

        ! Ok, now I have: rho, vel, fNeq of all my neighbor elements.
        ! linearly interpolate rho, vel, and fNeq
        targetRho = sum( weight(1:nSourceElems)*sourceRho(1:nSourceElems) )
        do iDir = 1,3
          targetVel(iDir) =   sum( weight(1:nSourceElems)    &
            &               * sourceVel(iDir,1:nSourceElems))
        enddo
        do iDir = 1,layout%stencil(1)%QQ
          targetNeq(iDir) =   sum( weight(1:nSourceElems)    &
            &               * sourceNeq(iDir,1:nSourceElems) )
        end do

        ! calculate target Eq
?? copy :: Eq_incomp( targetEq, targetRho, targetVel, layout, rho0 )

        ! Now write the resulting pdf = f_eq + fac*f_neq to the target element
        do iDir = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state(                                              &
&?SAVE?( varSys%Variable%val( iVar )%varPos( iDir ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ), &
         & t_target ) = targetEq( iDir ) + fac(iVar)*targetNeq( iDir )
        end do ! iDir
      end do ! ivar loop
    end do

  end subroutine fillFinerGhostsFromMe_linearIncomp
!******************************************************************************!


!******************************************************************************!
  !>  \ref intp_linear "Linear interpolation" of ghostFromCoarser
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_linear )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,intp%nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetVel(3)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(intp%nMaxSources)
    real(kind=rk) :: targetRho  ! macroscopic density for all source elements
    real(kind=rk) :: fac(varSys%nVars)        ! scaling factor
    ! current equilirbium part
    ! real(kind=rk) :: sourceEq( layout%stencil(1)%QQ, intp%nMaxSources )
    ! temporary equilirbium part
    real(kind=rk) :: fEq( layout%stencil(1)%QQ )
    ! current non-equilibrium part
    real(kind=rk) :: sourceNEq( layout%stencil(1)%QQ, intp%nMaxSources )
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on for the target elements
    integer :: t_source ! which time layer to work on for the source elements
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: srcRho, srcVel(3)
    real(kind=rk) :: weight( intp%nMaxSources ) 
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

! write(dbgUnit(1),*) '----- Enter fillMyGhostsFromFiner_linearAcoustic -----------'
! write(dbgUnit(1),*) 'source level: ', sourceLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( sourceLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( sourceLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( sourceLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( sourceLevel )%strainRate
! write(dbgUnit(1),*) 'target level: ', targetLevel
! write(dbgUnit(1),*) '  omega: ', fieldProp(1)%fluid%omLvl( targetLevel )
! write(dbgUnit(1),*) '  pressure factor: ',physics%fac( targetLevel )%press
! write(dbgUnit(1),*) '  velocity factor: ',physics%fac( targetLevel )%vel
! write(dbgUnit(1),*) '  strain   factor: ',physics%fac( targetLevel )%strainRate

    do iVar = 1, varSys%nVars
      ! See explanation for the factor in fineToCoarse
      fac(iVar) = ((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl(sourceLevel)) &
        &     *convPrePost(fieldProp( iVar )%fluid%omLvl( sourceLevel )))      &
        &   /((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( targetLevel ))    &
        &     *convPrePost(fieldProp( iVar )%fluid%omLvl( targetLevel )))
    end do

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromCoarserElem )
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )     &
        &                 %elem%nVals

      ! accumulators for target equilibrium and non-equilibrium parts
      targetEq   = 0._rk
      targetNEq  = 0._rk

      !Compute rho and velocity for each field and interpolate pdf field wise
      do iVar = 1, varSys%nVars
        ! See explanation for the factor in fineToCoarse
        sourceRho = 0._rk
        sourceVel = 0._rk

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                       %elem%val( iSourceElem )

          ! Read the pre-calculated weights (like in Average intp.)
          weight( iSourceElem ) = glob%levelDesc( targetLevel )                &
            &                     %depFromCoarser( iElem )%weight( iSourceElem )

          ! Calculate rho and vel
          do iDir = 1, layout%stencil(1)%QQ
             fTmp(iDir) = pdf(sourceLevel)%state(                              &
& ?SAVE?( varSys%variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
        & t_source )
             sourceRho(iSourceElem) = sourceRho(iSourceElem) + fTmp(iDir)
             sourceVel(:,iSourceElem ) = sourceVel( :, iSourceElem )           &
               & + fTmp( iDir)*real( layout%stencil(1)%cxDir(:,iDir), rk)
          enddo
          sourceVel(:,iSourceElem) =   sourceVel(:,iSourceElem) &
            &                        / sourceRho(iSourceElem)

          srcRho = sourceRho( iSourceElem )
          srcVel = sourceVel( :,iSourceElem )
?? copy :: Eq_comp( fEq, srcRho, srcVel, layout )
          ! sourceEq(:,iSourceElem)  = fEq(:)
          sourceNEq(:,iSourceElem) = fTmp(:) - fEq(:)

! write(dbgUnit(1),"(A, F12.5, A, 3F12.5)") &
!   &       'Source pressPhy: ', sourceRho(iSourceElem), &
!   &       '   vel: ', sourceVel(1,iSourceElem), sourceVel(2,iSourceElem), sourceVel(3,iSourceElem)
        end do ! iSourceElem

        ! Ok, now I have: rho, vel, fNeq of all my neighbor elements.
        ! linearly interpolate rho, vel, and fNeq
        targetRho    = sum( weight(1:nSourceElems)*sourceRho(1:nSourceElems))
        do iDir = 1,3
          targetVel(iDir) = sum( weight(1:nSourceElems)                        &
            &               *sourceVel(iDir,1:nSourceElems))
        enddo
        do iDir = 1,layout%stencil(1)%QQ
          targetNeq(iDir) = sum( weight(1:nSourceElems)                        &
            &               *sourceNeq(iDir,1:nSourceElems) )
        end do

! write(dbgUnit(1),"(A, F12.5, A, 3F12.5)") &
!   &       'Target pressPhy: ', targetRho,&
!   &       '   vel: ', targetVel(1), targetVel(2), targetVel(3)
!   ! &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
! write(dbgUnit(1),"(A)") ''

        ! calculate fEq
?? copy :: Eq_comp( targetEq, targetRho, targetVel, layout )

        ! Now write the resulting pdf to the target element
        do iDir = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state(                                              &
&?SAVE?( varSys%Variable%val( iVar )%varPos( iDir ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ), &
         & t_target ) = targetEq( iDir ) + fac(iVar)*targetNeq( iDir )
        end do
      enddo ! ivar loop
    enddo ! element loop

  end subroutine fillFinerGhostsFromMe_linear
!******************************************************************************!


!******************************************************************************!
  !> Doing the linear interpolation just like the quadratic one by using 
  !! the moment representation
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linMoment3d )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal, iDir     ! value counter, iDir required by eq calculation
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! shear stress scaling factor
    real(kind=rk) :: fac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )  ! source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    real(kind=rk) :: targetRho, targetVel(3)
    !---------------------------------------------------------------------------

    sourceLevel = level + 1
    targetLevel = level
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromFiner%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromFiner%              &
        &                                                    val( iSourceElem )
      !\todo iVar loop
      iVar = 1
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo

      tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)
      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5 ) = tMom( 5 ) - cs2*tMom( 1) & !Pxx
                   - tMom( 2)*tMom( 2)/tMom( 1)
      tMom( 6 ) = tMom( 6 ) - cs2*tMom( 1) & !Pyy
                   - tMom( 3)*tMom( 3)/tMom( 1)
      tMom( 7 ) = tMom( 7 ) - cs2*tMom( 1) & !Pzz
                   - tMom( 4)*tMom( 4)/tMom( 1)
      tMom( 8 ) = tMom( 8 )                & !Pxy
                   - tMom( 2)*tMom( 3)/tMom( 1)
      tMom( 9 ) = tMom( 9 )                & !Pyz
                   - tMom( 3)*tMom( 4)/tMom( 1)
      tMom(10 ) = tMom(10 )                & !Pzx
                   - tMom( 2)*tMom( 4)/tMom( 1)

      ! Store all the moments to the current source element's position in
      ! the buffer
      glob%levelDesc( targetLevel )%intpBufFromFiner( :, iSourceElem ) = tMom
    enddo

    ! Treat all fine target elements:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromFiner
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      !test = .false.
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                    offset( 1, eT_ghostFromFinerElem )
      ! get the target coordinates 
      targetCoord = glob%levelDesc( targetLevel )%depFromFiner( iElem )%coord
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromFiner( iElem )%      &
        &                                                            elem%nVals

      !Compute rho and velocity for each field and interpolate pdf field wise
      do iVar = 1, varSys%nVars
        fac = ((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( sourceLevel ))   &
          &     *convPrePost(fieldProp( iVar )%fluid%omLvl( sourceLevel )))    &
          &   /((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( targetLevel ))  &
          &     *convPrePost(fieldProp( iVar )%fluid%omLvl( targetLevel )))

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element in the levelwise buffer
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
            &                                     elemBuffer%val( iSourceElem )
          ! and read its entries
          sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
            &                                 intpBufFromFiner( :, sourceElem )
        end do

        do iVal = 1, 10
          targetMom(iVal) = sum( sourceMom(iVal, 1:nSourceElems))              &
            &             / real( nSourceElems, kind=rk )
        enddo

        ! Rescale the shear stress moments
        targetMom(5:10) = targetMom(5:10)*fac
        ! and add back the pi(0) part to the tensor
        ! No need to rescale higher order moments as they were set to
        ! equilibrium
        targetMom( 5 ) = targetMom( 5 ) + cs2*targetMom( 1) & !Pxx
                     + targetMom( 2)*targetMom( 2)/targetMom( 1)
        targetMom( 6 ) = targetMom( 6 ) + cs2*targetMom( 1) & !Pyy
                     + targetMom( 3)*targetMom( 3)/targetMom( 1)
        targetMom( 7 ) = targetMom( 7 ) + cs2*targetMom( 1) & !Pzz
                     + targetMom( 4)*targetMom( 4)/targetMom( 1)
        targetMom( 8 ) = targetMom( 8 )                     & !Pxy
                     + targetMom( 2)*targetMom( 3)/targetMom( 1)
        targetMom( 9 ) = targetMom( 9 )                     & !Pyz
                     + targetMom( 3)*targetMom( 4)/targetMom( 1)
        targetMom(10 ) = targetMom(10 )                     & !Pzx
                     + targetMom( 2)*targetMom( 4)/targetMom( 1)

        ! And calculate the equilibrium distribution from the macroscopic
        ! quantities in tMom
        targetRho = targetMom(1)
        targetVel(1:3) = targetMom(2:4)/targetMom(1)
?? copy :: Eq_comp( targetEq, targetRho, targetVel, layout )

        ! Set the further higher moments to their equilibrium distribution
        targetMom(11 ) = get_moment( targetEq, layout, (/2,1,0/) ) !mxxy
        targetMom(12 ) = get_moment( targetEq, layout, (/2,0,1/) ) !mxxz
        targetMom(13 ) = get_moment( targetEq, layout, (/1,2,0/) ) !myyx
        targetMom(14 ) = get_moment( targetEq, layout, (/0,2,1/) ) !myyz
        targetMom(15 ) = get_moment( targetEq, layout, (/1,0,2/) ) !mzzx
        targetMom(16 ) = get_moment( targetEq, layout, (/0,1,2/) ) !mzzy
        targetMom(17 ) = get_moment( targetEq, layout, (/2,2,0/) ) !mxxyy
        targetMom(18 ) = get_moment( targetEq, layout, (/0,2,2/) ) !myyzz
        targetMom(19 ) = get_moment( targetEq, layout, (/2,0,2/) ) !mzzxx

        ! Transform back into pdf space
        targetPdf = matmul( layout%moment%toPdf%A, targetMom )

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state( &
         &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target  ) &
         & = targetPdf(iVal)
         enddo
      enddo ! ivar loop
    enddo

  end subroutine fillMyGhostsFromFiner_linMoment3d
!******************************************************************************!


!******************************************************************************!
  !> Doing the linear interpolation just like the quadratic one by using
  !! the moment representation
  !! moment 1:     rho      from source element
  !! moment 2,3:   velocity from source element
  !! moment 4,5,6: strain rate from source element and scaled
  !! moment 7,8,9: calculated from fEq of target element
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linMoment2d )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal, iDir     ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! shear stress scaling factor
    real(kind=rk) :: fac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )  ! source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    real(kind=rk) :: targetRho, targetVel(3)
   !--------------------------------------------------------------------------

    sourceLevel = level + 1
    targetLevel = level
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromFiner%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromFiner%              &
        &                                                    val( iSourceElem )
      !\todo iVar loop
      iVar = 1
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo

      tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)
      ! remove pi(0) parts from the second order moment
      tMom(4) = tMom(4) - tMom(2)*tMom(2)/tMom(1) - cs2*tMom(1)
      tMom(5) = tMom(5) - tMom(3)*tMom(3)/tMom(1) - cs2*tMom(1)
      tMom(6) = tMom(6) - tMom(2)*tMom(3)/tMom(1)
      ! Store all the moments to the current source element's position in the
      ! buffer
      glob%levelDesc( targetLevel )%intpBufFromFiner( :, iSourceElem ) = tMom
    enddo

    ! Treat all fine target elements:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromFiner
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      !test = .false.
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                    offset( 1, eT_ghostFromFinerElem )
      ! get the target coordinates 
      targetCoord = glob%levelDesc( targetLevel )%depFromFiner( iElem )%coord
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromFiner( iElem )%      &
        &                                                            elem%nVals

      !Compute rho and velocity for each field and interpolate pdf field wise
      do iVar = 1, varSys%nVars
        fac = ((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( sourceLevel ))   &
          &     *convPrePost(fieldProp( iVar )%fluid%omLvl( sourceLevel )))    &
          &   /((1._rk - 0.5_rk*fieldProp( iVar )%fluid%omLvl( targetLevel ))  &
          &     *convPrePost(fieldProp( iVar )%fluid%omLvl( targetLevel )))

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element in the levelwise buffer
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
            &                                     elemBuffer%val( iSourceElem )
          ! and read its entries
          sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
            &                                 intpBufFromFiner( :, sourceElem )
        end do

        do iVal = 1, 6
          targetMom(iVal) = sum( sourceMom(iVal, 1:nSourceElems))              &
           &              / real( nSourceElems, kind=rk )
        enddo

        ! Rescale the shear stress moments
        targetMom(4:6) = targetMom(4:6)*fac
        ! targetMom(4:6) = targetMom(4:6)/physics%fac( targetLevel )%press
        ! and add back the pi(0) part to the tensor
        targetMom(4) = targetMom(4) + cs2*targetMom(1) &
                     + targetMom(2)*targetMom(2)/targetMom(1)
        targetMom(5) = targetMom(5) + cs2*targetMom(1) &
                     + targetMom(3)*targetMom(3)/targetMom(1)
        targetMom(6) = targetMom(6)                    &
                     + targetMom(2)*targetMom(3)/targetMom(1)

        ! Get the equilibrium of the target distribution
        targetRho = targetMom(1)
        targetVel(1:2) = targetMom(2:3)/targetMom(1)
        targetVel(3)   = 0._rk
?? copy :: Eq_comp( targetEq, targetRho, targetVel, layout )

        ! Set the further higher moments to equilibrium
        targetMom( 7 ) = get_moment( targetEq, layout, (/2,1,0/) )
        targetMom( 8 ) = get_moment( targetEq, layout, (/1,2,0/) )
        targetMom( 9 ) = get_moment( targetEq, layout, (/2,2,0/) )

        ! Transform back into pdf space
        targetPdf = matmul( layout%moment%toPdf%A, targetMom )

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state( &
&?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target  ) &
         & = targetPdf(iVal)
         enddo
      enddo ! ivar loop
    enddo

  end subroutine fillMyGhostsFromFiner_linMoment2d
!******************************************************************************!

end module mus_interpolate_linear_module
!******************************************************************************!
