! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
?? include 'header/lbm_macros.inc'
!> \brief Interpolation of flow quantities between different grid levels
!!
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with 
!! valid values.
!! Source elements are the fluid elements from other levels, from where to 
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source 
!! fluid elements on the source level.
!!
!! \ref For a detailed description of the grid
!!
module mus_interpolate_quadratic_module

  ! include treelm modules
  use env_module,              only: rk
  use tem_aux_module,          only: tem_abort
  use tem_param_module,        only: cs2inv, cs2
  use tem_math_module,         only: invert_matrix, matrix_type
  use tem_element_module,      only: eT_GhostFromCoarserElem
  use tem_construction_module, only: depSource_type
  use tem_grow_array_module,   only: grw_intArray_type
  use tem_rotation_module,     only: tem_assign_rotatedNeighbors,    &
    &                                tem_assign_rotatedNeighbors_ref2d
  use tem_var_system_module,   only: tem_var_system_type
  use tem_debug_module,        only: dbgUnit
  use tem_logging_module,      only: logUnit

  ! include musubi modules
  use mus_field_prop_module,          only: mus_field_prop_type
  use mus_pdf_module,                 only: pdf_data_type, pdf_global_type
  use mus_scheme_layout_module,       only: mus_scheme_layout_type
  use mus_scheme_type_module,         only: mus_scheme_type
  use mus_interpolate_header_module,  only: interpolation_type,                &
    &                                       refinement_type,                   &
    &                                       determine_timeLayer
  use mus_variable_module,            only: mus_derVarPos_type
  use mus_derivedQuantities_module2,  only: getNEq_diffusive,     &
    &                                       getEquilibrium, &
    &                                       getEquilibriumIncomp, &
    &                                       getShearRateTensor_diffusive
  use mus_interpolate_compact_module, only:                                    &
    &                                     unitTest_interpolate_quadraticCompact
  use mus_moments_module,             only: get_moment
  use mus_physics_module,             only: mus_physics_type

  implicit none

  private

  public :: mus_init_intpMatrices
  public :: fillFinerGhostsFromMe_quadratic2dDiffusive
  public :: fillFinerGhostsFromMe_quadratic3dDiffusive
  public :: fillFinerGhostsFromMe_quadratic2d
  public :: fillFinerGhostsFromMe_quadratic2dIncomp
  public :: fillFinerGhostsFromMe_quadratic3d
  public :: fillFinerGhostsFromMe_quadratic3dIncomp
  public :: mus_interpolate_quadratic3d_leastSq

  integer, parameter :: QmomentumX = 1
  integer, parameter :: QmomentumY = 2
  integer, parameter :: QmomentumZ = 3
  integer, parameter :: QPxx       = 4
  integer, parameter :: QPyy       = 5
  integer, parameter :: QPzz       = 6
  integer, parameter :: QPxy       = 7
  integer, parameter :: QPyz       = 8
  integer, parameter :: QPzx       = 9
  integer, parameter :: c_x  = 1 
  integer, parameter :: c_y  = 2 
  integer, parameter :: c_z  = 3 

  interface mus_interpolate_quadratic3d_leastSq
    module procedure mus_interpolate_quadratic3d_leastSq_scal
    module procedure mus_interpolate_quadratic3d_leastSq_vect
  end interface mus_interpolate_quadratic3d_leastSq

 contains

!******************************************************************************!
  !> Initialize the matrix for finding the interpolation coefficients
  !!
  subroutine mus_init_intpMatrices( me, intp, layout, dep, glob, fieldProp,    &
    &                               compact, debug, targetLevel, sourceLevel )
    !---------------------------------------------------------------------------
    !> refinement type
    type( refinement_type), intent(out) :: me
    !> interpolation type
    type( interpolation_type), intent(in) :: intp
    !> scheme layout including the stencil
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> dependencies of the target ghost elements
    type( depSource_type ), intent(in) :: dep(:)
    !> fluid parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !> field and fluid properties
    type( mus_field_prop_type ), intent(in) :: fieldProp
    !> current interpolation with a compact inerpolation stencil?
    logical, intent(in) :: compact
    !> debug?
    logical, intent(in) :: debug
    !> level of the target elements to interpolate to
    integer, intent(in) :: targetLevel
    !> level of the source elements (=sample points) for the interpolation 
    integer, intent(in) :: sourceLevel
    !---------------------------------------------------------------------------
    real(kind=rk) :: omega, factor
    real(kind=rk) :: wxy
    real(kind=rk), allocatable :: AT(:,:)
    real(kind=rk), allocatable :: ATA(:,:)
    real(kind=rk), allocatable :: Ainv(:,:)
    ! row and column counter
    integer :: iRow, iCase, nCases, iElem
    ! Perform a test of this routine
    logical :: unitTest
    logical :: passed
    logical :: require_matrices
    !---------------------------------------------------------------------------
    require_matrices = .false.
    unitTest = .false.

    omega = fieldProp%fluid%omLvl( sourceLevel )
    if( unitTest ) omega = 1.8_rk

?? if( STREAM_PUSH ) then
    ! Pre collision values
    ! \sigma^{(1),LB} = -(1-\frac{\omega}{2})\sum_i c_{\alpha i } c_{\beta i} f_i^{neq}
    factor = 1._rk/omega
?? else
    ! For post-collision (PULL) interpolation: use 
    ! \f$
    ! \sigma^{(1),LB} = -\frac{(1-\frac{\omega}{2})}{(1-\omega)}
    ! \sum_i c_{\alpha i } c_{\beta i} f_i^{neq} \f$
    factor = (1._rk/omega-1._rk)
?? endif
    wxy =  -cs2*factor

    nCases = 8
    if( .not. compact ) then
      nCases = 1
    endif

    allocate( me%matrix( nCases ))
    allocate( me%matrixInv( nCases ))

    ! Choose an element with high order interpolation
    if( targetLevel > sourceLevel ) then
      if( glob%levelDesc( targetLevel )%intpFromCoarser_high%nVals > 0 ) then
        require_matrices = .true.
        iElem = glob%levelDesc( targetLevel )%intpFromCoarser_high%val(1)
      endif
    else
      if( glob%levelDesc( targetLevel )%intpFromFiner_high%nVals > 0 ) then
        require_matrices = .true.
        iElem = glob%levelDesc( targetLevel )%intpFromFiner_high%val(1)
      endif
    endif

    if( require_matrices ) then
      call mus_init_coeffMatrix( intp = intp,                                  &
        &                        dep = dep,                                    &
        &                        compact = compact,                            &
        &                        nSourceElems = dep( iElem )%elem%nVals,       &
        &                        firstElem = iElem,                            &
        &                        matrix = me%matrix(:),                        &
        &                        layout = layout )

      ! Allocate intermediate matrices
      allocate( AT(    me%matrix( 1 )%nEntries(2), me%matrix( 1 )%nEntries(1)))
      allocate( ATA(   me%matrix( 1 )%nEntries(2), me%matrix( 1 )%nEntries(2)))
      allocate( Ainv(  me%matrix( 1 )%nEntries(2), me%matrix( 1 )%nEntries(2)))

      ! Set the matrices for all eight cases which might appear
      do iCase = 1, nCases
        me%matrixInv( iCase )%nEntries(1:2) =                                  &
          &        (/ me%matrix( 1 )%nEntries(2), me%matrix( 1 )%nEntries(1) /)
        allocate( me%matrixInv( iCase )%A( me%matrixInv( iCase )%nEntries(1),  &
          &                                me%matrixInv( iCase )%nEntries(2)))
        me%matrixInv( iCase )%A = -1._rk
        AT   = -1._rk
        ATA  = -1._rk
        Ainv = -1._rk

        AT = transpose(me%matrix( iCase )%A)

        ! debug output
        write(dbgUnit(10),*) 'sourceLevel ', sourceLevel, &
          &                  ' targetLevel ',targetLevel
        write(dbgUnit(10),*) 'Case ', iCase
        write(dbgUnit(10),*) 'Original me%matrix'

        do iRow = 1, me%matrix( iCase )%nEntries(1)
          write(dbgUnit(10),*) me%matrix( iCase )%A( iRow, : )
        enddo
        ! Matrix multiplication AT*A
        ATA = matmul(AT, me%matrix( iCase )%A)

        ! Invert matrix (AT*A)^-1
        Ainv = invert_matrix( ATA ) !, size(ATA, 2))
        ! Calculate the resulting inverse me%matrix
        ! This me%matrix is then mutliplied with the rhs of the equation system
        ! to get the interpolation coefficients
        me%matrixInv( iCase )%A = matmul( Ainv, AT )
        ! debug output
        write(dbgUnit(10),*)'Matrix inverted *AT' 
        write(dbgUnit(10),*)'size ', size( me%matrixInv( iCase )%A,1 ),        &
          &                  size(me%matrixInv( iCase )%A,2 )
        write(dbgUnit(10),*) 'Original me%matrix'
        do iRow = 1, size( me%matrixInv( iCase )%A, 1)
          write(dbgUnit(10),*) me%matrixInv( iCase )%A( iRow, : )
        enddo
        if( unitTest ) then
          ! Test the matrices with reference data
          call unitTest_interpolate_quadraticCompact( me%matrixInv( iCase )%A, &
            &                                         iCase, passed )
          if( passed ) then
            write(logUnit(1),*) ' Testing: Interpolation matrix PASSED '//     &
              &                 'for case', iCase 
          else
            write(logUnit(1),*) ' Testing: Interpolation matrix FAILED '//     &
              &                 'for case', iCase 
            call tem_abort
          endif
        endif
      enddo ! fill up all cases 
    endif ! require matrices?

  end subroutine mus_init_intpMatrices
!******************************************************************************!


!******************************************************************************!
  !> This routine fill Finer Ghost from Coarser elements.
  !! Target: finer   element (L+1)
  !! Source: coarser element (L)
  !! 1. Calculate moments for all source elements and store into buffer
  !! 2. For each target, read moment of source elements from buffer.
  !! 3. Do quadratic interpolation by target coordinates.
?? copy :: interpolateRoutine_header(fillFinerGhostsFromMe_quadratic2dIncomp)
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! shear stress scaling factor
    real(kind=rk) :: fac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )  ! source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    real(kind=rk) :: rho0
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    iVar = 1
    fac = getNoneqFac( fieldProp( iVar )%fluid%omLvl( sourceLevel ), &
      &                fieldProp( iVar )%fluid%omLvl( targetLevel )  )

    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    rho0 = fieldProp( iVar )%fluid%rho0

! write(*,"(A)") "Inside fill finer ghost from my coarser quadratic 2D incomp routine"
! write(*,"(A,I0,A,I0)") "  source level: ", sourceLevel, ",  read buffer: ", t_source
! write(*,"(A,I0,A,I0)") "  target level: ", targetLevel, ", write buffer: ", t_target

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals

      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromCoarser%val( iSourceElem )

      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo

      tMom(:) = matmul( layout%moment%toMoments%A, sourcePdf)
      ! remove pi(0) parts from the second order moment
      tMom(4) = tMom(4) - tMom(2)*tMom(2) - cs2*tMom(1)
      tMom(5) = tMom(5) - tMom(3)*tMom(3) - cs2*tMom(1)
      tMom(6) = tMom(6) - tMom(2)*tMom(3)

      ! Store all the moments to the buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser( :, iSourceElem ) = tMom
    enddo

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )

      ! get the target coordinates
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems =                                                           &
        &      glob%levelDesc( targetLevel )%depFromCoarser( iElem )%elem%nVals

      !compute rho and velocity for each field and interpolate pdf field wise
      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element in the levelwise buffer
        sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )   &
          &              %elemBuffer%val( iSourceElem )
        ! and read its entries
        sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )           &
          &                          %intpBufFromCoarser( :, sourceElem )
      end do

      do iVal = 1, layout%stencil(1)%QQ
        targetMom(iVal) = mus_interpolate_quadratic2d_leastSq(               &
          &                 sourceMom(iVal, 1:nSourceElems),                 &
          &                 targetCoord(1:2),                                &
          &                 nSourceElems,                                    &
          &                 intp%fromCoarser( targetLevel )%matrixInv(1)%A )
      enddo

      ! Rescale the shear stress moments
      ! \todo: equation from asymptotic paper is used here.
      !        use fac or omega?
      targetMom(4:6) = targetMom(4:6)*fac
      ! and add back the pi(0) part to the tensor
      targetMom(4) = targetMom(4) + targetMom(2)*targetMom(2) + cs2*targetMom(1)
      targetMom(5) = targetMom(5) + targetMom(3)*targetMom(3) + cs2*targetMom(1)
      targetMom(6) = targetMom(6) + targetMom(2)*targetMom(3)

      ! Get the equilibrium of the target distribution
      targetEq = getEquilibriumIncomp( dens = targetMom(1),     &
        &          vel = (/ targetMom(2), targetMom(3), 0._rk /),&
        &          rho0 = rho0, layout = layout )

      ! Set the further higher moments to equilibrium
      targetMom( 7 ) = get_moment( targetEq, layout, (/2,1,0/) ) !mxxy
      targetMom( 8 ) = get_moment( targetEq, layout, (/1,2,0/) ) !mxyy
      targetMom( 9 ) = get_moment( targetEq, layout, (/2,2,0/) ) !mxxyy

      ! Transform back into pdf space
      targetPdf = matmul( layout%moment%toPdf%A, targetMom )

      ! Now write the resulting pdf into target element
      targetElem = iElem + glob%levelDesc( targetLevel )           &
        &                      %offset( 1, eT_ghostFromCoarserElem )

      do iVal = 1, layout%stencil(1)%QQ
      pdf( targetLevel )%state( &
       &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
       & t_target  ) &
       & = targetPdf(iVal)
      enddo

    end do ! indElem

  end subroutine fillFinerGhostsFromMe_quadratic2dIncomp
!******************************************************************************!


!******************************************************************************!
  !> This routine fill Finer from Coarser.
  !! Important steps includes:
  !! 1. Calculate moment of all source elements.
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic2d )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! shear stress scaling factor
    real(kind=rk) :: fac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )  ! source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    fac = getNoneqFac( fieldProp( iVar )%fluid%omLvl( sourceLevel ), &
      &                fieldProp( iVar )%fluid%omLvl( targetLevel )  )

    ! \todo: This section can become a seperate routine
    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromCoarser%            &
        &                                                    val( iSourceElem )
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo
      tMom(:) = matmul( layout%moment%toMoments%A, sourcePdf)
      ! remove pi(0) parts from the second order moment
      tMom(4) = tMom(4) - cs2*tMom(1) - tMom(2)*tMom(2)/tMom(1)
      tMom(5) = tMom(5) - cs2*tMom(1) - tMom(3)*tMom(3)/tMom(1)
      tMom(6) = tMom(6)               - tMom(2)*tMom(3)/tMom(1)
      ! Transfer momentum field to velocity field
      tMom(2:3) = tMom(2:3)/tMom(1)
      ! Store all the moments to the current source element's position in
      ! the buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser(:,iSourceElem) = tMom
    enddo

    ! Treat all fine target elements:
    !do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                  offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%    &
        &                                                            elem%nVals

      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element in the levelwise buffer
        sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
          &                                     elemBuffer%val( iSourceElem )
        ! and read its entries
        sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
          &                               intpBufFromCoarser( :, sourceElem )
      end do

      do iVal = 1, layout%stencil(1)%QQ
        targetMom(iVal) = mus_interpolate_quadratic2d_leastSq(               &
          &               sourceMom(iVal, 1:nSourceElems),                   &
          &               targetCoord(1:2),                                  &
          &               nSourceElems,                                      &
          &               intp%fromCoarser( targetLevel )%matrixInv(1)%A )
      enddo

      ! Rescale the shear stress moments
      targetMom(4:6) = targetMom(4:6)*fac
      ! Transfer back velocity field to momentum field
      targetMom(2:3) = targetMom(2:3)*targetMom(1)
      ! and add back the pi(0) part to the tensor
      targetMom(4) = targetMom(4) + cs2*targetMom(1) + targetMom(2)*targetMom(2)/targetMom(1)
      targetMom(5) = targetMom(5) + cs2*targetMom(1) + targetMom(3)*targetMom(3)/targetMom(1)
      targetMom(6) = targetMom(6)                    + targetMom(2)*targetMom(3)/targetMom(1)

      ! Get the equilibrium of the target distribution
      targetEq = getEquilibrium( targetMom( 1),         & ! density
        &             (/ targetMom( 2)/targetMom(1),    & ! velX
        &                targetMom( 3)/targetMom(1),    & ! velY
        &                0._rk/),                       & ! velZ
        &                layout )

      ! Set the further higher moments to equilibrium
      targetMom( 7 ) = get_moment( targetEq, layout, (/2,1,0/) )
      targetMom( 8 ) = get_moment( targetEq, layout, (/1,2,0/) )
      targetMom( 9 ) = get_moment( targetEq, layout, (/2,2,0/) )

      ! Transform back into pdf space
      targetPdf = matmul( layout%moment%toPdf%A, targetMom )

      ! Now write the resulting pdf in the current direction to the target
      ! Element position
      do iVal = 1, layout%stencil(1)%QQ
      pdf( targetLevel )%state( &
       &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
       & t_target  ) &
       & = targetPdf(iVal)
      end do
      ! enddo ! ivar loop
    end do ! indElem

  end subroutine fillFinerGhostsFromMe_quadratic2d
!******************************************************************************!


!******************************************************************************!
  !>
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3dDiffusive )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    integer :: iVar ! current field variable
    integer :: iDir ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    ! Shear tensor of the source elements
    real(kind=rk) :: sourceS(3,3)
    ! Shear tensor of the target element
    real(kind=rk) :: targetS(3,3)
    real(kind=rk) :: rho0
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    rho0 = fieldProp( iVar )%fluid%rho0

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromCoarser%            &
        &                                                    val( iSourceElem )
      tMom = 0._rk
      do iDir = 1, layout%stencil(1)%QQ
        sourcePdf( iDir )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
        tMom(1)   = tMom(1) + sourcePdf(iDir)
        tMom(2:4) =   tMom(2:4)           &
          &         + sourcePdf(iDir) * real(layout%stencil(1)%cxDir(1:3,iDir),rk)
      enddo

      !-----------------------------------------
      ! Get the strain rate tensor
      sourceS(:,:) = getShearRateTensor_diffusive(                              &
        &                   subset = sourcePdf, layout = layout,                &
        &                   omega = fieldProp( iVar )%fluid%omLvl( sourceLevel ))
      ! and copy to the moment vector
      tMom( 5) = sourceS(1,1)
      tMom( 6) = sourceS(2,2)
      tMom( 7) = sourceS(3,3)
      tMom( 8) = sourceS(1,2)
      tMom( 9) = sourceS(2,3)
      tMom(10) = sourceS(1,3)

      !-----------------------------------------
      ! Convert from LB to physical
      ! pressure is actually only the fluctuating part
      ! p = (rho-rho0)*cs2*convFactor with convFactor = rho0*dx^2/dt^2
      tMom(1)    = ( tMom(1) - rho0 ) * cs2 * physics%fac( sourceLevel )%press
      tMom(2:4)  = tMom(2:4)                * physics%fac( sourceLevel )%vel
      ! strain rate conversion
      tMom(5:10) = tMom(5:10)               * physics%fac( sourceLevel )%strainRate
      !-----------------------------------------

      ! Store all the moments to the current source element's position in the
      ! buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser( :, iSourceElem ) = tMom
    enddo

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                  offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%    &
        &                                                            elem%nVals

        sourceMom = 0._rk
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element in the levelwise buffer
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
            &                                     elemBuffer%val( iSourceElem )
          ! and read its entries
          sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
            &                               intpBufFromCoarser( :, sourceElem )
        end do

        ! interpolate the first ten entries:
        ! pressure(1), velocity(3), strain rate(6)
        do iVal = 1, 10
          targetMom(iVal) = mus_interpolate_quadratic3d_leastSq(               &
            &               sourceMom(iVal, 1:nSourceElems),                   &
            &               targetCoord(1:3),                                  &
            &               nSourceElems,                                      &
            &               intp%fromCoarser( targetLevel )%matrixInv(1)%A )
        enddo

        ! Copy the shear rate tensor entries to an explicit tensor for the
        ! neq-computation
        targetS = 0._rk
        targetS(1,1) = targetMom(5)
        targetS(2,2) = targetMom(6)
        targetS(3,3) = targetMom(7)
        targetS(1,2) = targetMom(8)
        targetS(2,1) = targetMom(8)
        targetS(2,3) = targetMom(9)
        targetS(3,2) = targetMom(9)
        targetS(1,3) = targetMom(10)
        targetS(3,1) = targetMom(10)

        !---------------------------------------------
        ! Convert from physical to LB
        ! Re-scale the pressure and velocity from physical units to LB units on
        ! the targetlevel
        targetMom(1)   = targetMom(1)   / physics%fac(targetLevel)%press * cs2inv + rho0
        targetMom(2:4) = targetMom(2:4) / physics%fac(targetLevel)%vel
        targetS(:,:)   = targetS(:,:)   / physics%fac(targetLevel)%strainRate
        !---------------------------------------------

        targetPdf = getEquilibriumIncomp( targetMom(1), targetMom(2:4), layout,&
          &            fieldProp( iVar )%fluid%rho0 )
        targetNeq(:) = getnEq_diffusive(                                       &
          &            layout = layout,                                        &
          &            omega  = fieldProp( iVar )%fluid%omLvl( targetLevel ),  &
          &            Sxx    = targetS(:,: ))

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
pdf( targetLevel )%state( &
  &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
  & t_target  )  = targetPdf(iVal)  + targetNeq(iVal)
        end do
    end do

  end subroutine fillFinerGhostsFromMe_quadratic3dDiffusive
!******************************************************************************!


!******************************************************************************!
  !>
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic2dDiffusive )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    integer :: iVar ! current field variable
    integer :: iDir ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    ! Shear tensor of the source elements
    real(kind=rk) :: sourceS(3,3) 
    ! Shear tensor of the target element
    real(kind=rk) :: targetS(3,3)
    real(kind=rk) :: rho0
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))

    iVar = 1
    rho0 = fieldProp( iVar )%fluid%rho0

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromCoarser%            &
        &                                                    val( iSourceElem )
      tMom = 0._rk
      do iDir = 1, layout%stencil(1)%QQ
        sourcePdf( iDir )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
        tMom(1) = tMom(1) + sourcePdf(iDir)
        tMom(2:3) = tMom(2:3)                                               &
          & + sourcePdf(iDir) * real(layout%stencil(1)%cxDir(1:2,iDir), rk)
      end do

      !-----------------------------------------
      ! Get the strain rate tensor
      sourceS(:,:) = getShearRateTensor_diffusive(                             &
        &                   subset = sourcePdf, layout = layout,                &
        &                   omega = fieldProp( iVar )%fluid%omLvl( sourceLevel ))
      ! and copy to the moment vector
      tMom(4) = sourceS(1,1)
      tMom(5) = sourceS(2,2)
      tMom(6) = sourceS(1,2)

      !-----------------------------------------
      ! Convert from LB to physical
      ! pressure is actually only the fluctuating part
      tMom(1)   = ( tMom(1) - rho0 ) * cs2 * physics%fac( sourceLevel )%press
      tMom(2:3) = tMom(2:3)     * physics%fac( sourceLevel )%vel
      tMom(4:6) = tMom(4:6)     * physics%fac( sourceLevel )%strainRate
      !-----------------------------------------

      ! Store all the moments to the current source element's position in
      ! the buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser( :, iSourceElem ) = tMom
    end do ! iSourceElem

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                  offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%    &
        &                                                            elem%nVals

        sourceMom = 0._rk
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element in the levelwise buffer
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
            &                                     elemBuffer%val( iSourceElem )
          ! and read its entries
          sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
            &                               intpBufFromCoarser( :, sourceElem )
        end do

        do iVal = 1, 6
          targetMom(iVal) = mus_interpolate_quadratic2d_leastSq(               &
            &               sourceMom(iVal, 1:nSourceElems),                   &
            &               targetCoord(1:2),                                  &
            &               nSourceElems,                                      &
            &               intp%fromCoarser( targetLevel )%matrixInv(1)%A )
        end do

        ! Copy the shear rate tensor entries to an explicit tensor for the
        ! neq-computation
        targetS = 0._rk
        targetS(1,1) = targetMom(4)
        targetS(2,2) = targetMom(5)
        targetS(1,2) = targetMom(6)
        targetS(2,1) = targetMom(6)

        !---------------------------------------------
        ! Convert from physical to LB
        ! Re-scale the pressure and velocity from physical units to LB units on
        ! the targetlevel
        targetMom(1)   = targetMom(1) * cs2inv / physics%fac(targetLevel)%press + rho0
        targetMom(2:3) = targetMom(2:3)        / physics%fac(targetLevel)%vel
        targetS(:,:)   = targetS(:,:)          / physics%fac(targetLevel)%strainRate
        !---------------------------------------------

        targetPdf = getEquilibriumIncomp( targetMom(1),                        &
          &                               [ targetMom(2), targetMom(3), 0._rk ], &
          &                               layout, rho0 )
        targetNeq(:) = getnEq_diffusive(                                       &
          &            layout = layout,                                        &
          &            omega  = fieldProp( iVar )%fluid%omLvl( targetLevel ),  &
          &            Sxx    = targetS(:,: ))

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state( &
         &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target  )  = targetPdf(iVal)  + targetNeq(iVal)
        end do
    end do

  end subroutine fillFinerGhostsFromMe_quadratic2dDiffusive
!******************************************************************************!


!******************************************************************************!
  !>  \ref intp_linear "Linear interpolation" of ghostFromCoarser
  !! This routine is not used by anywhere
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic2dDiffusiveLinStyle )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iVal, jVal     ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,intp%nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetVel(3)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(intp%nMaxSources)
    real(kind=rk) :: targetRho  ! macroscopic density for all source elements
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: targetNeq( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to work on for the target elements
    integer :: t_source ! which time layer to work on for the source elements
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: weight( intp%nMaxSources ) 
    ! Shear tensor of the source elements
    real(kind=rk) :: sourceS(3,3, intp%nMaxSources) 
    ! Shear tensor of the target element
    real(kind=rk) :: targetS(3,3), targetCoord(3)
    real(kind=rk) :: rho0
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    rho0 = fieldProp( iVar )%fluid%rho0

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )                       &
        &                      %offset( 1, eT_ghostFromCoarserElem )
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord
      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )     &
        &                 %elem%nVals

      ! accumulators for target equilibrium and non-equilibrium parts
      targetEq   = 0._rk
      targetNEq  = 0._rk

      !Compute rho and velocity for each field and interpolate pdf field wise
        sourceRho = 0._rk
        sourceVel = 0._rk

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                       %elem%val( iSourceElem )

          ! Read the pre-calculated weights (like in Average intp.)
          weight( iSourceElem ) = glob%levelDesc( targetLevel )                &
            &                     %depFromCoarser( iElem )%weight( iSourceElem )
          ! Calculate rho and vel for equil
          do iDir = 1, layout%stencil(1)%QQ
             fTmp(iDir) = pdf(sourceLevel)%state(                              &
& ?SAVE?( varSys%variable%val(iVar)%varPos(iDir), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh), &
        & t_source )
             sourceRho(iSourceElem) = sourceRho(iSourceElem) + fTmp(iDir)
             sourceVel(:,iSourceElem ) = sourceVel( :, iSourceElem )           &
               & + fTmp( iDir)*real( layout%stencil(1)%cxDir(:,iDir), kind=rk)
          enddo
          !-----------------------------------------
          ! Get the second velocity moments of the source element's pdf
          sourceS(:,:,iSourceElem ) = getShearRateTensor_diffusive( &
            &  subset = fTmp, layout = layout,   & 
            &  omega = fieldProp( iVar )%fluid%omLvl( sourceLevel ))
          !-----------------------------------------

          ! Convert from LB to physical
          ! pressure is actually only the fluctuating part
          sourceRho(iSourceElem)   = ( sourceRho(iSourceElem) - rho0 ) * cs2 &
            &                        * physics%fac( sourceLevel )%press
          sourceVel(:,iSourceElem) = sourceVel( :, iSourceElem)     &
            &                        * physics%fac( sourceLevel )%vel
          sourceS(:,:,iSourceElem) = sourceS(:,:,iSourceElem)              &
            &                        * physics%fac( sourceLevel )%strainRate
          !-----------------------------------------

        end do ! iSourceElem

        !------------------ LINEAR INTERPOLATION -------------------
        ! This is all done in physical units
        ! Ok, now I have: rho, vel, fNeq of all my neighbor elements.
        ! linearly interpolate rho, vel, and strain rate
        targetRho       = mus_interpolate_quadratic2d_leastSq(                 &
          &               sourceRho(1:nSourceElems),                           &
          &               targetCoord(1:2),                                    &
          &               nSourceElems,                                        &
          &               intp%fromCoarser( targetLevel )%matrixInv(1)%A )
        do iVal = 1,3
          targetVel(iVal) = mus_interpolate_quadratic2d_leastSq(               &
            &             sourceVel(iVal, 1:nSourceElems),                     &
            &             targetCoord(1:2),                                    &
            &             nSourceElems,                                        &
            &             intp%fromCoarser( targetLevel )%matrixInv(1)%A )
        end do
        ! Interpolate the shear stress
        targetS = 0._rk
        do jVal = 1, layout%stencil(1)%nDims
          do iVal = 1, layout%stencil(1)%nDims
          targetS(iVal, jVal) = mus_interpolate_quadratic2d_leastSq(           &
            &             sourceS(iVal,jVal, 1:nSourceElems),                  &
            &             targetCoord(1:2),                                    &
            &             nSourceElems,                                        &
            &             intp%fromCoarser( targetLevel )%matrixInv(1)%A )
          end do
        end do
        !------------------ LINEAR INTERPOLATION -------------------

        !---------------------------------------------
        ! Convert from physical to LB
        ! Re-scale the pressure and velocity from physical units to LB units on
        ! the targetlevel
        targetRho = targetRho/physics%fac( targetLevel )%press*cs2inv + rho0
        targetVel(:) = targetVel(:)/physics%fac( targetLevel )%vel
        targetS(:,:) = targetS(:,:)/physics%fac( targetLevel)%strainRate
        !---------------------------------------------

        targetEq = getEquilibriumIncomp( targetRho, targetVel, layout, rho0 )
        targetNeq(:) = getnEq_diffusive(                                       &
          &            layout = layout,                                        &
          &            omega  = fieldProp( iVar )%fluid%omLvl( targetLevel ),  &
          &            Sxx    = targetS(:,:) )

        ! Now write the resulting pdf in the current direction to the target
        ! resulting pdf = f_eq + f_neq
        do iVal = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state(                                              &
&?SAVE?( varSys%Variable%val( iVar )%varPos( iVal ), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ), &
         & t_target ) = targetEq( iVal ) + targetNeq( iVal )
        enddo
    enddo

  end subroutine fillFinerGhostsFromMe_quadratic2dDiffusiveLinStyle
!******************************************************************************!


!******************************************************************************!
  !>  
  !!  
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3dIncomp )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )  ! source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    real(kind=rk) :: rho0, fac
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    fac = getNoneqFac( fieldProp( iVar )%fluid%omLvl( sourceLevel ), &
      &                fieldProp( iVar )%fluid%omLvl( targetLevel )  )
    rho0 = fieldProp( iVar )%fluid%rho0

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromCoarser%            &
        &                                                    val( iSourceElem )
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo

      ! Calculate the raw moments form the pdf
      ! \todo: use density or pressure?
      tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)

      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5 ) = tMom( 5 ) - tMom(2)*tMom(2)*rho0 - cs2*tMom(1) !Pxx
      tMom( 6 ) = tMom( 6 ) - tMom(3)*tMom(3)*rho0 - cs2*tMom(1) !Pyy
      tMom( 7 ) = tMom( 7 ) - tMom(4)*tMom(4)*rho0 - cs2*tMom(1) !Pzz
      tMom( 8 ) = tMom( 8 ) - tMom(2)*tMom(3)*rho0               !Pxy
      tMom( 9 ) = tMom( 9 ) - tMom(3)*tMom(4)*rho0               !Pyz
      tMom(10 ) = tMom(10 ) - tMom(2)*tMom(4)*rho0               !Pxz

      ! Store all the moments to the current source element's position in the
      ! buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser( :, iSourceElem ) = tMom
    enddo

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                  offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord

      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%elem%nVals

      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element in the levelwise buffer
        sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
          &                                     elemBuffer%val( iSourceElem )
        ! and read its entries
        sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
          &                               intpBufFromCoarser( :, sourceElem )
      end do

      targetMom( 1:19) = mus_interpolate_quadratic3d_leastSq(                &
        &   srcMom =    sourceMom(1:19, 1:nSourceElems),                     &
        &   targetCoord =  targetCoord,                                      &
        &   nSourceElems =  nSourceElems,                                    &
        &   nVals =     19,                                                  &
        &   matrix =    intp%fromCoarser( targetLevel )%matrixInv(1)%A )

      ! Rescale the shear stress moments
      targetMom(5:10) = targetMom(5:10)*fac

      ! and add back the pi(0) part to the tensor
      ! No need to rescale higher order moments as they were set to
      ! equilibrium
      targetMom( 5 ) = targetMom( 5 ) + cs2*targetMom( 1) & !Pxx
                   + targetMom( 2)*targetMom( 2)*rho0
      targetMom( 6 ) = targetMom( 6 ) + cs2*targetMom( 1) & !Pyy
                   + targetMom( 3)*targetMom( 3)*rho0
      targetMom( 7 ) = targetMom( 7 ) + cs2*targetMom( 1) & !Pzz
                   + targetMom( 4)*targetMom( 4)*rho0
      targetMom( 8 ) = targetMom( 8 )                     & !Pxy
                   + targetMom( 2)*targetMom( 3)*rho0
      targetMom( 9 ) = targetMom( 9 )                     & !Pyz
                   + targetMom( 3)*targetMom( 4)*rho0
      targetMom(10 ) = targetMom(10 )                     & !Pzx
                   + targetMom( 2)*targetMom( 4)*rho0

      ! And calculate the equilibrium distribution from the macroscopic 
      ! quantities in tMom
      targetEq = getEquilibriumIncomp( targetMom( 1),    & ! density
        &             (/ targetMom( 2),                  & ! velX
        &                targetMom( 3),                  & ! velY
        &                targetMom( 4)/),                & ! velZ
        &                layout, rho0 )

      ! \todo: how to set higher moments?
      ! Set the further higher moments to their equilibrium distribution
      targetMom(11 ) = get_moment( targetEq, layout, (/2,1,0/) ) !mxxy
      targetMom(12 ) = get_moment( targetEq, layout, (/2,0,1/) ) !mxxz
      targetMom(13 ) = get_moment( targetEq, layout, (/1,2,0/) ) !myyx
      targetMom(14 ) = get_moment( targetEq, layout, (/0,2,1/) ) !myyz
      targetMom(15 ) = get_moment( targetEq, layout, (/1,0,2/) ) !mzzx
      targetMom(16 ) = get_moment( targetEq, layout, (/0,1,2/) ) !mzzy
      targetMom(17 ) = get_moment( targetEq, layout, (/2,2,0/) ) !mxxyy
      targetMom(18 ) = get_moment( targetEq, layout, (/0,2,2/) ) !myyzz
      targetMom(19 ) = get_moment( targetEq, layout, (/2,0,2/) ) !mzzxx

      ! transform the moments back to the pdfs
      targetPdf = matmul( layout%moment%toPdf%A, targetMom )

      ! Now write the resulting pdf in the current direction to the target
      ! Element position
      do iVal = 1, layout%stencil(1)%QQ
      pdf( targetLevel )%state( &
       &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
       & t_target  ) &
       & = targetPdf(iVal)
      end do
    end do ! indElem

  end subroutine fillFinerGhostsFromMe_quadratic3dIncomp
!******************************************************************************!


!******************************************************************************!
  !>  
  !!  
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3d )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! shear stress scaling factor
    real(kind=rk) :: fac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )  ! source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    fac = getNoneqFac( fieldProp( iVar )%fluid%omLvl( sourceLevel ), &
      &                fieldProp( iVar )%fluid%omLvl( targetLevel )  )

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromCoarser%            &
        &                                                    val( iSourceElem )
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo
      ! Calculate the raw moments form the pdf
      tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)

      ! Transfer momentum to velocity
      tMom(2:4) = tMom(2:4)/tMom(1)
      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5 ) = tMom( 5 ) - tMom(2)*tMom(2)*tMom(1) - cs2*tMom(1) !Pxx
      tMom( 6 ) = tMom( 6 ) - tMom(3)*tMom(3)*tMom(1) - cs2*tMom(1) !Pyy
      tMom( 7 ) = tMom( 7 ) - tMom(4)*tMom(4)*tMom(1) - cs2*tMom(1) !Pzz
      tMom( 8 ) = tMom( 8 ) - tMom(2)*tMom(3)*tMom(1)               !Pxy
      tMom( 9 ) = tMom( 9 ) - tMom(3)*tMom(4)*tMom(1)               !Pyz
      tMom(10 ) = tMom(10 ) - tMom(2)*tMom(4)*tMom(1)               !Pxz

      ! Store all the moments to the current source element's position in the
      ! buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser( :, iSourceElem ) = tMom
    enddo

    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                  offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates 
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord
!      targetCoord_trafo = abs( targetCoord -0.5_rk )

      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%    &
        &                                                            elem%nVals

      !Compute rho and velocity for each field and interpolate pdf field wise
      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element in the levelwise buffer
        sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
          &                                     elemBuffer%val( iSourceElem )
        ! and read its entries
        sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%          &
          &                               intpBufFromCoarser( :, sourceElem )
      end do
!      iCase = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%childNum

      targetMom( 1:19) = mus_interpolate_quadratic3d_leastSq(                &
        &   srcMom =    sourceMom(1:19, 1:nSourceElems),                     &
        &   targetCoord =  targetCoord,                                      &
        &   nSourceElems =  nSourceElems,                                    &
        &   nVals =     19,                                                  &
        &   matrix =    intp%fromCoarser( targetLevel )%matrixInv(1)%A )

      ! Transfer velocity to momentum
      targetMom(2:4) = targetMom(2:4)*targetMom(1)

      ! Rescale the shear stress moments
      targetMom(5:10) = targetMom(5:10)*fac

      ! and add back the pi(0) part to the tensor
      ! No need to rescale higher order moments as they were set to
      ! equilibrium
      targetMom( 5 ) = targetMom( 5 ) + cs2*targetMom( 1) & !Pxx
                   + targetMom( 2)*targetMom( 2)/targetMom( 1)
      targetMom( 6 ) = targetMom( 6 ) + cs2*targetMom( 1) & !Pyy
                   + targetMom( 3)*targetMom( 3)/targetMom( 1)
      targetMom( 7 ) = targetMom( 7 ) + cs2*targetMom( 1) & !Pzz
                   + targetMom( 4)*targetMom( 4)/targetMom( 1)
      targetMom( 8 ) = targetMom( 8 )                     & !Pxy
                   + targetMom( 2)*targetMom( 3)/targetMom( 1)
      targetMom( 9 ) = targetMom( 9 )                     & !Pyz
                   + targetMom( 3)*targetMom( 4)/targetMom( 1)
      targetMom(10 ) = targetMom(10 )                     & !Pzx
                   + targetMom( 2)*targetMom( 4)/targetMom( 1)

      ! And calculate the equilibrium distribution from the macroscopic 
      ! quantities in tMom
      targetEq = getEquilibrium( targetMom( 1),        & ! density
        &             (/ targetMom( 2)/targetMom(1),   & ! velX
        &                targetMom( 3)/targetMom(1),   & ! velY
        &                targetMom( 4)/targetMom(1)/), & ! velZ
        &                layout )

      ! Set the further higher moments to their equilibrium distribution
      targetMom(11 ) = get_moment( targetEq, layout, (/2,1,0/) ) !mxxy
      targetMom(12 ) = get_moment( targetEq, layout, (/2,0,1/) ) !mxxz
      targetMom(13 ) = get_moment( targetEq, layout, (/1,2,0/) ) !myyx
      targetMom(14 ) = get_moment( targetEq, layout, (/0,2,1/) ) !myyz
      targetMom(15 ) = get_moment( targetEq, layout, (/1,0,2/) ) !mzzx
      targetMom(16 ) = get_moment( targetEq, layout, (/0,1,2/) ) !mzzy
      targetMom(17 ) = get_moment( targetEq, layout, (/2,2,0/) ) !mxxyy
      targetMom(18 ) = get_moment( targetEq, layout, (/0,2,2/) ) !myyzz
      targetMom(19 ) = get_moment( targetEq, layout, (/2,0,2/) ) !mzzxx

      ! transform the moments back to the pdfs
      targetPdf = matmul( layout%moment%toPdf%A, targetMom )

      ! Now write the resulting pdf in the current direction to the target
      ! Element position
      do iVal = 1, layout%stencil(1)%QQ
      pdf( targetLevel )%state( &
       &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
       & t_target  ) &
       & = targetPdf(iVal)
      enddo
    end do ! indElem

  end subroutine fillFinerGhostsFromMe_quadratic3d
!******************************************************************************!


!******************************************************************************!
  !>  
  !!  
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3dUnBuffered )
    !---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal           ! value counter
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) ! pdf to reconstruct
    real(kind=rk) :: sourcePdf( layout%stencil(1)%QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%stencil(1)%QQ, intp%nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    ! shear stress scaling factor
    real(kind=rk) :: fac
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: t_source ! which time layer to read from
    real(kind=rk) :: targetEq( layout%stencil(1)%QQ )  !source equilirbium part
    real(kind=rk) :: tMom( layout%stencil(1)%QQ )  ! temp moment calculation
    !---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
    t_source = determine_timeLayer( intp%timeLayer, pdf( sourceLevel ))
    iVar = 1
    fac = getNoneqFac( fieldProp( iVar )%fluid%omLvl( sourceLevel ), &
      &                fieldProp( iVar )%fluid%omLvl( targetLevel )  )

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, glob%levelDesc( targetLevel )%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = glob%levelDesc( targetLevel )%sourceFromCoarser%            &
        &                                                    val( iSourceElem )
      do iVal = 1, layout%stencil(1)%QQ
        sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &      t_source  )
      enddo
      ! Calculate the raw moments form the pdf
      tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)

      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5 ) = tMom( 5 ) - tMom(2)*tMom(2)*tMom(1) - cs2*tMom(1) !Pxx
      tMom( 6 ) = tMom( 6 ) - tMom(3)*tMom(3)*tMom(1) - cs2*tMom(1) !Pyy
      tMom( 7 ) = tMom( 7 ) - tMom(4)*tMom(4)*tMom(1) - cs2*tMom(1) !Pzz
      tMom( 8 ) = tMom( 8 ) - tMom(2)*tMom(3)*tMom(1)               !Pxy
      tMom( 9 ) = tMom( 9 ) - tMom(3)*tMom(4)*tMom(1)               !Pyz
      tMom(10 ) = tMom(10 ) - tMom(2)*tMom(4)*tMom(1)               !Pxz

      ! Store all the moments to the current source element's position in the
      ! buffer
      glob%levelDesc( targetLevel )%intpBufFromCoarser( :, iSourceElem ) = tMom
    enddo
 
    ! Treat all fine target elements:
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      ! Read the target element treeId
      targetElem = iElem + glob%levelDesc( targetLevel )%                      &
        &                                  offset( 1, eT_ghostFromCoarserElem )
      ! get the target coordinates 
      targetCoord = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%coord

      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%    &
        &                                                            elem%nVals

      !Compute rho and velocity for each field and interpolate pdf field wise
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
          ! Get the source element in the levelwise buffer
       !   sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%elemBuffer%val( iSourceElem )
          ! and read its entries
       !   sourceMom(:, iSourceElem ) = glob%levelDesc( targetLevel )%intpBufFromCoarser( :, sourceElem )
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
            &                                           elem%val( iSourceElem )
          do iVal = 1, layout%stencil(1)%QQ
            sourcePdf( iVal )  = pdf( sourceLevel )%state( &
  &  ?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), sourceElem, varSys%nScalars, pdf(sourceLevel)%neigh ),&
  &          t_source  )
          enddo
          ! Calculate the raw moments form the pdf
          tMom( : ) = matmul( layout%moment%toMoments%A, sourcePdf)

          ! remove pi(0) parts from the second order moment
          ! that is ((rho u_\alpha * rho u_\beta) / rho)
          ! and additionally rho*cs2 from diagonal entries
          ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
          tMom( 5 ) = tMom( 5 ) - tMom(2)*tMom(2)*tMom(1) - cs2*tMom(1) !Pxx
          tMom( 6 ) = tMom( 6 ) - tMom(3)*tMom(3)*tMom(1) - cs2*tMom(1) !Pyy
          tMom( 7 ) = tMom( 7 ) - tMom(4)*tMom(4)*tMom(1) - cs2*tMom(1) !Pzz
          tMom( 8 ) = tMom( 8 ) - tMom(2)*tMom(3)*tMom(1)               !Pxy
          tMom( 9 ) = tMom( 9 ) - tMom(3)*tMom(4)*tMom(1)               !Pyz
          tMom(10 ) = tMom(10 ) - tMom(2)*tMom(4)*tMom(1)               !Pxz

          sourceMom(:, iSourceElem ) = tMom

          ! debug output
          write(dbgUnit(10),*) iSourceElem, ' srcMom ',                        &
            &                  sourceMom( :, iSourceElem)

        end do

        targetMom( 1:19) = mus_interpolate_quadratic3d_leastSq(                &
          &   srcMom =    sourceMom(1:19, 1:nSourceElems),                     &
          &   targetCoord =  targetCoord,                                      &
          &   nSourceElems =  nSourceElems,                                    &
          &   nVals =     19,                                                  &
          &   matrix =    intp%fromCoarser( targetLevel )%matrixInv(1)%A )

        ! Rescale the shear stress moments
        targetMom(5:10) = targetMom(5:10)*fac

        ! and add back the pi(0) part to the tensor
        ! No need to rescale higher order moments as they were set to
        ! equilibrium
        targetMom( 5 ) = targetMom( 5 ) + cs2*targetMom( 1) & !Pxx
                     + targetMom( 2)*targetMom( 2)/targetMom( 1)
        targetMom( 6 ) = targetMom( 6 ) + cs2*targetMom( 1) & !Pyy
                     + targetMom( 3)*targetMom( 3)/targetMom( 1)
        targetMom( 7 ) = targetMom( 7 ) + cs2*targetMom( 1) & !Pzz
                     + targetMom( 4)*targetMom( 4)/targetMom( 1)
        targetMom( 8 ) = targetMom( 8 )                     & !Pxy
                     + targetMom( 2)*targetMom( 3)/targetMom( 1)
        targetMom( 9 ) = targetMom( 9 )                     & !Pyz
                     + targetMom( 3)*targetMom( 4)/targetMom( 1)
        targetMom(10 ) = targetMom(10 )                     & !Pzx
                     + targetMom( 2)*targetMom( 4)/targetMom( 1)

        ! And calculate the equilibrium distribution from the macroscopic 
        ! quantities in tMom
        targetEq = getEquilibrium( targetMom( 1),        & ! density
          &             (/ targetMom( 2)/targetMom(1),   & ! velX
          &                targetMom( 3)/targetMom(1),   & ! velY
          &                targetMom( 4)/targetMom(1)/), & ! velZ
          &                layout )

        ! Set the further higher moments to their equilibrium distribution
        targetMom(11 ) = get_moment( targetEq, layout, (/2,1,0/) ) !mxxy
        targetMom(12 ) = get_moment( targetEq, layout, (/2,0,1/) ) !mxxz
        targetMom(13 ) = get_moment( targetEq, layout, (/1,2,0/) ) !myyx
        targetMom(14 ) = get_moment( targetEq, layout, (/0,2,1/) ) !myyz
        targetMom(15 ) = get_moment( targetEq, layout, (/1,0,2/) ) !mzzx
        targetMom(16 ) = get_moment( targetEq, layout, (/0,1,2/) ) !mzzy
        targetMom(17 ) = get_moment( targetEq, layout, (/2,2,0/) ) !mxxyy
        targetMom(18 ) = get_moment( targetEq, layout, (/0,2,2/) ) !myyzz
        targetMom(19 ) = get_moment( targetEq, layout, (/2,0,2/) ) !mzzxx

        ! transform the moments back to the pdfs
        targetPdf = matmul( layout%moment%toPdf%A, targetMom )

        ! Now write the resulting pdf in the current direction to the target
        ! Element position
        do iVal = 1, layout%stencil(1)%QQ
        pdf( targetLevel )%state( &
         &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
         & t_target  ) &
         & = targetPdf(iVal)
         enddo
    enddo

  end subroutine fillFinerGhostsFromMe_quadratic3dUnbuffered
!******************************************************************************!


!******************************************************************************!
  !> Initialize the matrix for finding the interpolation coefficients
  !!
  subroutine mus_init_coeffMatrix( intp, dep, compact, nSourceElems,           &
    &                              firstElem, matrix, layout )
    !---------------------------------------------------------------------------
    !> interpolation method info
    type( interpolation_type ), intent(in) :: intp
    !>
    integer, intent(in) :: nSourceElems
    !> dependencies of the target ghost elements
    type( depSource_type ), intent(in) :: dep(:)
    !> first element with the corresponding correct information in the
    !! dep(:) array
    integer, intent(in) :: firstElem
    !> current interpolation with a compact inerpolation stencil?
    logical, intent(in) :: compact
    !> resulting interpolation matrix
    type(matrix_type), intent(inout) :: matrix(:)
    !> Factor for the shear stress components
    type(mus_scheme_layout_type), intent(in) :: layout
    !---------------------------------------------------------------------------
    ! Temporary matrix to assemble (in a transposed fashion)
    real(kind=rk), allocatable :: AT(:,:)
    ! Coordinates of the source elements
    real(kind=rk), allocatable :: sourceCoord(:,:)
    integer  :: iElem, iCase, iTask
    ! Number of degerees of freedom for interpolation
    ! 2d:   6: a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y
    ! 3d:  10: a0 + a1*x + a2*y + a3*z + a4*x^2 + a5*y^2 + a6*z^2 
    !                                  a7*x*y + a8*y*z + a9*z*x 
    integer  :: nDOFs
    ! amount of tasks in the interpolation method
    integer  :: nTasks
    ! task definition for the interpolation method
    integer, allocatable :: task(:)
    !---------------------------------------------------------------------------
    allocate( sourceCOord(3, nSourceElems))
    write(logUnit(10),"(A)")    '   Initializing the interpolation matrix'
    write(logUnit(10),"(A,I0)") '   nSourceElems: ', nSourceElems

    select case( trim(layout%stencil(1)%label))
    case( 'd2q9' )
      !-------------------------- D2Q9 Compact ---------------------------------
      if( compact ) then
        nTasks = 5
        nDOFs = 12
        matrix( : )%nEntries(1) = nSourceElems*nTasks 
        matrix( : )%nEntries(2) = nDOFs
        allocate( AT( matrix(1)%nEntries(2), matrix(1)%nEntries(1) ))
        ! Allocate Matrices for all the childNum cases
        allocate( task( nTasks ))
        task(1:nTasks) = (/QmomentumX, QmomentumY, QPxx, QPyy, QPxy /)
        do iCase = 1,8
          allocate( matrix( iCase )%A( matrix(1)%nEntries(1),                  &
            &                          matrix(1)%nEntries(2) ))
          ! Assign source coordinates
          sourceCoord = real( tem_assign_rotatedNeighbors_ref2d( iCase ),      &
            &                                                          kind=rk)
          do iTask = 1, nTasks
            do iElem = 1, nSourceElems
              AT( :, iElem+(iTask-1)*nSourceElems )                            &
                &            = mus_init_coeffRow2d(                            &
                &                 sourceCoord = sourceCoord(:,iElem),          &
                &                 nEntries = nDOFs,                            &
                &                 iTask = task( iTask ))
            end do
          end do
          matrix( iCase )%A = transpose(AT)
        end do ! iCase
      else
      !-------------------------- D2Q9 Non-Compact------------------------------
        nDOFs = 6
        matrix( : )%nEntries(1) = nSourceElems
        matrix( : )%nEntries(2) = nDOFs
        allocate( AT( matrix(1)%nEntries(2), matrix(1)%nEntries(1) ))
        ! Allocate Matrices for all the childNum cases
        do iCase = 1,1
          allocate( matrix( iCase )%A( matrix(1)%nEntries(1),                  &
            &                          matrix(1)%nEntries(2) ))
          ! Assign source element's coordinates from the target element's
          ! parent's point of view
          ! the matrix rows 
          do iElem = 1, nSourceElems
            sourceCoord( :, iElem ) = dep(firstElem)%stencil( :, iElem )
            AT( :, iElem ) = mus_init_coeffRow2d(                              &
              &                 sourceCoord = sourceCoord(:,iElem),            &
              &                 nEntries = nDOFs,                              &
              &                 iTask = QmomentumX )
          end do
          matrix( iCase )%A = transpose(AT)
        end do
      endif
    case( 'd3q19' )
      !-------------------------- D3Q19 Compact------------------------------
      if( compact ) then
        !nSourceElems = 7
        nTasks = 9
        nDOFs = 30
        matrix( : )%nEntries(1) = nSourceElems*nTasks
        matrix( : )%nEntries(2) = nDOFs
        allocate( AT( matrix(1)%nEntries(2), matrix(1)%nEntries(1) ))
        ! Allocate Matrices for all the childNum cases
        allocate( task( nTasks ))
          task(:) = (/QmomentumX, QmomentumY, QmomentumZ,    &
            &         QPxx, QPyy, QPzz,                      &
            &         QPxy, QPyz, QPzx /)
        do iCase = 1,8
          allocate( matrix( iCase )%A( matrix(1)%nEntries(1),&
            &                          matrix(1)%nEntries(2) ))
          ! Assign source coordinates
          sourceCoord = real( tem_assign_rotatedNeighbors( iCase ), kind=rk)
          do iTask = 1, nTasks
            do iElem = 1, nSourceElems
              AT( :, iElem+(iTask-1)*nSourceElems )                            &
                &            = mus_init_coeffRow3d(                            &
                &                 sourceCoord = sourceCoord(:,iElem),          &
                &                 nEntries = nDOFs,                            &
                &                 iTask = task( iTask ))
            end do
          end do
          matrix( iCase )%A = transpose(AT)
        enddo
      else
        !-------------------------- D3Q19 non-compact---------------------------
        nDOFs = 10
        matrix( : )%nEntries(1) = nSourceElems
        matrix( : )%nEntries(2) = nDOFs
        allocate( AT( matrix(1)%nEntries(2), matrix(1)%nEntries(1) ))
        ! Allocate Matrices for all the childNum cases
        do iCase = 1,1
          allocate( matrix( iCase )%A( matrix(1)%nEntries(1),   &
            &                          matrix(1)%nEntries(2) )  )
          ! Assign the matrix rows
          do iElem = 1, nSourceElems
            sourceCoord( :, iElem ) = dep( firstElem )%stencil( :, iElem )
            AT( :, iElem ) = mus_init_coeffRow3d(                              &
              &                 sourceCoord = sourceCoord(:,iElem),            &
              &                 nEntries = nDOFs,                              &
              &                 iTask = QmomentumX )
          enddo
          matrix( iCase )%A = transpose(AT)
        enddo
      endif
    end select
    !--END--------------- select stencil  and interpolation scheme ------------


    write(logUnit(10),"(A)")    '     interpolation: '//trim( intp%label )
    write(logUnit(10),"(A,I0)") '     nSourceElems: ', nSourceElems
    write(logUnit(10),"(A,I0)") '     nDOFs: ', nDOFS

  end subroutine mus_init_coeffMatrix
!******************************************************************************!


!******************************************************************************!
  !> Build up the matrix rows for iTask
  !! 1 momentumX
  !! 2 momentumY
  !! 3 Pxx
  !! 4 Pyy
  !! 5 Pxy
  !! For the interpolation ansatz
  !!   f(x) = a0 + a1*x + a2*y + a3*x^2 + a4*y^2 + a5*x*y
  !!   g(x) = b0 + b1*x + b2*y + b3*x^2 + b4*y^2 + b5*x*y
  !!
  function mus_init_coeffRow2d( sourceCoord, nEntries, iTask ) result( row )
    !---------------------------------------------------------------------------
    !>
    real(kind=rk), intent(in) :: sourceCoord(3)
    !>
    integer, intent(in) :: nEntries
    !>
    integer, intent(in) :: iTask
    !>
    real(kind=rk) :: row(nEntries)
    !---------------------------------------------------------------------------

    row = 0._rk
    select case( iTask )
    case( QmomentumX )
      row( 1) = 1._rk                               ! a0
      row( 2) = sourceCoord(1)     ! a1
      row( 3) = sourceCoord(2)     ! a2
      row( 4) = sourceCoord(1)**2  ! a3
      row( 5) = sourceCoord(2)**2  ! a4
      row( 6) = sourceCoord(1)*sourceCoord(2)
    case( QmomentumY )
      row( 7) = 1._rk                               ! b0
      row( 8) = sourceCoord(1)     ! b1
      row( 9) = sourceCoord(2)     ! b2
      row(10) = sourceCoord(1)**2  ! b3
      row(11) = sourceCoord(2)**2  ! b4
      row(12) = sourceCoord(1)*sourceCoord(2)
    case( QPxx ) ! = 2*df/dx = 2*a1 + 4*a3*x + 2*a5*y
      row( 2) = 2._rk                ! 2*a1
      row( 4) = 4._rk*sourceCoord(1) ! 4*a3*x
      row( 6) = 2._rk*sourceCoord(2) ! 2*a5*y
    case( QPyy ) ! = 2*dg/dy = 2*b2 + 4*b4*y + 2*b5*x
      row( 9) = 2._rk                ! 2*b2
      row(11) = 4._rk*sourceCoord(2) ! 4*b4*y
      row(12) = 2._rk*sourceCoord(1) ! 2*b5*x
    case( QPxy ) ! = df/dy + dg/dx = a2 + 2*a4*y + a5*x + b1 + 2*b3*x + b5*y
      row( 3) = 1._rk                ! a2
      row( 5) = 2._rk*sourceCoord(2) ! 2*a4*y
      row( 6) = 1._rk*sourceCoord(1) ! a5*x
      row( 8) = 1._rk                ! b1  
      row(10) = 2._rk*sourceCoord(1) ! 2*b3*x
      row(12) = 1._rk*sourceCoord(2) ! b5*y
    case default
      write(logUnit(1),*) 'Undefined task', iTask
      call tem_abort
    end select

  end function mus_init_coeffRow2d
!******************************************************************************!


!******************************************************************************!
  !> Build up the matrix rows for iTask
  !! 1 momentumX   4 Pxx     7 Pxy 
  !! 2 momentumY   5 Pyy     8 Pyz
  !! 3 momentumY   6 Pzz     9 Pzx
  !! For the interpolation ansatz
  !!   f(x) = a0 + a1*x + a2*y + a3*z + a4*x^2 + a5*y^2 + a6*z^2 + a7*x*y 
  !!             + a8*y*z + a9*z*x
  !!   g(x) = b0 + b1*x + b2*y + b3*z + b4*x^2 + b5*y^2 + b6*z^2 + b7*x*y 
  !!             + b8*y*z + b9*z*x
  !!   h(x) = c0 + c1*x + c2*y + c3*z + c4*x^2 + c5*y^2 + c6*z^2 + c7*x*y 
  !!             + c8*y*z + c9*z*x
  !!
  function mus_init_coeffRow3d( sourceCoord, iTask, nEntries ) result( row )
    !---------------------------------------------------------------------------
    !>
    real(kind=rk), intent(in) :: sourceCoord(3)
    !>
    integer, intent(in) :: iTask
    !>
    integer, intent(in) :: nEntries
    !>
    real(kind=rk) :: row( nEntries )
    !---------------------------------------------------------------------------
    row = 0._rk
    select case( iTask )
    case( QmomentumX )       ! only set entries for a
      row( 1) = 1._rk                                ! a0
      row( 2) = sourceCoord(1)                       ! a1
      row( 3) = sourceCoord(2)                       ! a2    
      row( 4) = sourceCoord(3)                       ! a3    
      row( 5) = sourceCoord(1)**2                    ! a4     
      row( 6) = sourceCoord(2)**2                    ! a5     
      row( 7) = sourceCoord(3)**2                    ! a6     
      row( 8) = sourceCoord(1)*sourceCoord(2)        ! a7     
      row( 9) = sourceCoord(2)*sourceCoord(3)        ! a8     
      row(10) = sourceCoord(3)*sourceCoord(1)        ! a9     
    case( QmomentumY )      ! only set entries for b
      row(11) = 1._rk                                ! b0
      row(12) = sourceCoord(1)                       ! b1
      row(13) = sourceCoord(2)                       ! b2
      row(14) = sourceCoord(3)                       ! b3 
      row(15) = sourceCoord(1)**2                    ! b4 
      row(16) = sourceCoord(2)**2                    ! b5   
      row(17) = sourceCoord(3)**2                    ! b6   
      row(18) = sourceCoord(1)*sourceCoord(2)        ! b7   
      row(19) = sourceCoord(2)*sourceCoord(3)        ! b8   
      row(20) = sourceCoord(3)*sourceCoord(1)        ! b9   
    case( QmomentumZ )      ! only set entries for c
      row(21) = 1._rk                                ! c0
      row(22) = sourceCoord(1)                       ! c1     
      row(23) = sourceCoord(2)                       ! c2     
      row(24) = sourceCoord(3)                       ! c3     
      row(25) = sourceCoord(1)**2                    ! c4  
      row(26) = sourceCoord(2)**2                    ! c5  
      row(27) = sourceCoord(3)**2                    ! c6  
      row(28) = sourceCoord(1)*sourceCoord(2)        ! c7
      row(29) = sourceCoord(2)*sourceCoord(3)        ! c8
      row(30) = sourceCoord(3)*sourceCoord(1)        ! c9
    case( QPxx ) ! = 2*df/dx = 2*a1 + 4*a4*x + 2*a7*y +2*a9*z
      row( 2) = 2._rk                ! 2*a1
      row( 5) = 4._rk*sourceCoord(1) ! 4*a4*x
      row( 8) = 2._rk*sourceCoord(2) ! 2*a7*y
      row(10) = 2._rk*sourceCoord(3) ! 2*a9*z
    case( QPyy ) ! = 2*dg/dy = 2*b2 + 4*b5*y + 2*b7*x + 2*b8*z
      row(13) = 2._rk                ! 2*b2
      row(16) = 4._rk*sourceCoord(2) ! 4*b5*y
      row(18) = 2._rk*sourceCoord(1) ! 2*b7*x
      row(19) = 2._rk*sourceCoord(3) ! 2*b8*z
    case( QPzz ) ! = 2*dh/dz = 2*c3 + 4*c6*z + 2*c8*y + 2*c9*x
      row(24) = 2._rk                ! 2*c3
      row(27) = 4._rk*sourceCoord(3) ! 4*c6*z
      row(29) = 2._rk*sourceCoord(2) ! 2*c8*y
      row(30) = 2._rk*sourceCoord(1) ! 2*c9*x
    case( QPxy ) ! = df/dy                     + dg/dx 
                 ! = a2 + 2*a5*y + a7*x + a8*z + b1 + 2*b4*x + b7*y + b9*z
      row( 3) = 1._rk                ! a2
      row( 6) = 2._rk*sourceCoord(2) ! 2*a5*y
      row( 8) = sourceCoord(1)       ! a7*x
      row( 9) = sourceCoord(3)       ! a8*z
      row(12) = 1._rk                ! b1
      row(15) = 2._rk*sourceCoord(1) ! 2*b4*x
      row(18) = sourceCoord(2)       ! b7*y
      row(20) = sourceCoord(3)       ! b9*z

    case( QPyz ) ! = dg/dz                     + dh/dy 
                 ! = b3 + 2*b6*z + b8*y + b9*x + c2 + 2*c5*y + c7*x + c8*z
      row(14) = 1._rk                ! b3
      row(17) = 2._rk*sourceCoord(3) ! 2*b6*z
      row(19) = sourceCoord(2)       ! b8*y
      row(20) = sourceCoord(1)       ! b9*x
      row(23) = 1._rk                ! c2
      row(26) = 2._rk*sourceCoord(2) ! 2*c5*y
      row(28) = sourceCoord(1)       ! c7*x
      row(29) = sourceCoord(3)       ! c8*z

    case( QPzx ) ! = dh/dx                     + df/dz 
                 ! = c1 + 2*c4*x + c7*y + c9*z + a3 + 2*a6*z + a8*y + a9*x
      row(22) = 1._rk                ! c1
      row(25) = 2._rk*sourceCoord(1) ! 2*c4*x
      row(28) = sourceCoord(2)       ! c7*y
      row(30) = sourceCoord(3)       ! c9*z
      row( 4) = 1._rk                ! a3
      row( 7) = 2._rk*sourceCoord(3) ! 2*a6*z
      row( 9) = sourceCoord(2)       ! a8*y
      row(10) = sourceCoord(1)       ! a9*x
    end select

  end function mus_init_coeffRow3d
!******************************************************************************!


!******************************************************************************!
  !> Biquadratic interpolation for a single quantity phi
  !!
  pure function mus_interpolate_quadratic2d_leastSq( srcMom, targetCoord,      &
    &                                      nSourceElems, matrix ) result( phi )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: nSourceElems
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(nSourceElems)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(2)
    !> relaxation time for the non-eq moments
    real(kind=rk), intent(in) :: matrix(:,:)
    !> interpolated value
    real(kind=rk) :: phi
    !---------------------------------------------------------------------------
    real(kind=rk) :: a(6) ! Coefficients
    real(kind=rk) :: rhs(nSourceElems), lhs(6) ! Coefficients
    !---------------------------------------------------------------------------
    ! We extract momentum information completely on the view of the source
    ! coordinate system
    ! Set the right hand side of the equation system
    rhs(:)  = srcMom(:)
    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    lhs = matmul( matrix, rhs )
 
    a(1:6) = lhs(1:6)
    ! Evaluate the bubble function with the above calculated coefficients
    ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
    phi = a(1) + a(2)*targetCoord(c_x) + a(3)*targetCoord(c_y)                 &
      &           + a(4)*targetCoord(c_x)*targetCoord(c_x)                     &
      &           + a(5)*targetCoord(c_y)*targetCoord(c_y)                     &
      &           + a(6)*targetCoord(c_x)*targetCoord(c_y)

  end function mus_interpolate_quadratic2d_leastSq
!******************************************************************************!


!******************************************************************************!
  !> Triquadratic interpolation for a single quantity phi
  !!
  pure function mus_interpolate_quadratic3d_leastSq_scal( srcMom, targetCoord, &
    &                                     nSourceElems, matrix ) result( phi )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: nSourceElems
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(nSourceELems)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(3)
    !> relaxation time for the non-eq moments
    real(kind=rk), intent(in) :: matrix(:,:)
    !> interpolated value
    real(kind=rk) :: phi
    !---------------------------------------------------------------------------
    real(kind=rk) :: a(10) ! Coefficients
    !---------------------------------------------------------------------------
    ! We extract momentum information completely on the view of the source
    ! coordinate system
    ! Set the right hand side of the equation system
    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    a(:)= matmul( matrix, srcMom(:))

    ! Evaluate the bubble function with the above calculated coefficients
    ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
    phi = a(1) + a( 2)*targetCoord(c_x)                                        &
      &         + a( 3)*targetCoord(c_y)                                       &
      &         + a( 4)*targetCoord(c_z)                                       &
      &         + a( 5)*targetCoord(c_x)*targetCoord(c_x)                      &
      &         + a( 6)*targetCoord(c_y)*targetCoord(c_y)                      &
      &         + a( 7)*targetCoord(c_z)*targetCoord(c_z)                      &
      &         + a( 8)*targetCoord(c_x)*targetCoord(c_y)                      &
      &         + a( 9)*targetCoord(c_y)*targetCoord(c_z)                      &
      &         + a(10)*targetCoord(c_z)*targetCoord(c_x)

  end function mus_interpolate_quadratic3d_leastSq_scal
!******************************************************************************!


!******************************************************************************!
  !> Triquadratic interpolation for a single quantity phi
  pure function mus_interpolate_quadratic3d_leastSq_vect( srcMom, targetCoord, &
    &              nSourceElems, nVals, matrix ) result( phi )
    !---------------------------------------------------------------------------
    !>
    integer, intent(in) :: nSourceElems
    !>
    integer, intent(in) :: nVals
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(nVals, nSourceELems)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(3)
    !> relaxation time for the non-eq moments
    real(kind=rk), intent(in) :: matrix(:,:)
    !> interpolated value
    real(kind=rk) :: phi( nVals )
    !---------------------------------------------------------------------------
    real(kind=rk) :: a(10) ! Coefficients
    integer :: iVal
    !---------------------------------------------------------------------------
    ! We extract momentum information completely on the view of the source
    ! coordinate system
    ! Set the right hand side of the equation system
    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    do iVal = 1, nVals
    a(:)= matmul( matrix, srcMom(iVal, :))

    ! Evaluate the bubble function with the above calculated coefficients
    ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
    phi( iVal ) =   a( 1) &
      &           + a( 2)*targetCoord(c_x)                  &
      &           + a( 3)*targetCoord(c_y)                  &
      &           + a( 4)*targetCoord(c_z)                  &
      &           + a( 5)*targetCoord(c_x)*targetCoord(c_x) &
      &           + a( 6)*targetCoord(c_y)*targetCoord(c_y) &
      &           + a( 7)*targetCoord(c_z)*targetCoord(c_z) &
      &           + a( 8)*targetCoord(c_x)*targetCoord(c_y) &
      &           + a( 9)*targetCoord(c_y)*targetCoord(c_z) &
      &           + a(10)*targetCoord(c_z)*targetCoord(c_x)
    enddo

  end function mus_interpolate_quadratic3d_leastSq_vect
!******************************************************************************!

!******************************************************************************!
  !> Calculate the conversion factor for nonEq in difference levels
  pure function getNoneqFac( omegaS, omegaT ) result( fac )
    !---------------------------------------------------------------------------
    real(kind=rk), intent(in) :: omegaS  !< omega value on source level
    real(kind=rk), intent(in) :: omegaT  !< omage value on target level
    real(kind=rk) :: fac
    !---------------------------------------------------------------------------

??  if( STREAM_PUSH ) then
      fac = ( 1._rk - 0.5_rk * omegaS ) / ( 1._rk - 0.5_rk * omegaT )
??  else
      fac =   ( 1._rk - 0.5_rk * omegaS ) * ( 1._rk - omegaT ) &
        &   / ( 1._rk - omegaS ) / ( 1._rk - 0.5_rk * omegaT )
??  endif

  end function getNoneqFac
!******************************************************************************!


end module mus_interpolate_quadratic_module
!******************************************************************************!
