! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Interpolation scheme tools
!!
!! For an overview over implemented interpolation methods, see
!! \ref intp_methods "Interpolation methods"
!!
!!\author Manuel Hasert
!!
?? include 'header/lbm_macros.inc'
module mus_interpolate_tools_module

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length

  ! include treelm modules
  use env_module,              only: rk, labelLen, long_k, newUnit, pathLen
  use treelmesh_module,        only: treelmesh_type
  use tem_aux_module,          only: tem_abort
  use tem_construction_module, only: tem_levelDesc_type
  use tem_element_module,      only: eT_GhostFromCoarserElem,                  &
    &                                eT_ghostFromFinerElem
  use tem_geometry_module,     only: tem_baryOfId, tem_intp_bilinear,          &
    &                                tem_elemSize
  use tem_grow_array_module,   only: grw_intArray_type
  use tem_param_module,        only: cs2inv, cs2
  use tem_var_system_module,   only: tem_var_system_type
  use tem_comm_env_module,     only: tem_comm_env_type

  ! include musubi modules
  use mus_field_prop_module,         only: mus_field_prop_type 
  use mus_interpolate_header_module, only: interpolation_type,                 &
    &                                      determine_timeLayer
  use mus_pdf_module,                only: pdf_data_type, pdf_global_type,     &
    &                                      pdf_globInfo_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_moments_module,            only: init_transformation_matrix
  use mus_variable_module,           only: mus_derVarPos_type

  implicit none

  private

  public :: dump_intpLists        
  public :: debug_dependencies

  contains
!******************************************************************************!
  !> check the dependencies from Finer
  !!
  subroutine debug_dependencies( pdfGlobal, intp, layout, glob, tree, proc ) 
    !---------------------------------------------------------------------------
    !> global pdf information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> state properties
    type( pdf_global_type ), intent(in) :: glob
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> musubi mpi communicator environment
    type( tem_comm_env_type ), intent(in) :: proc
    !---------------------------------------------------------------------------
    integer :: nUnit
    character(len=17)  :: buffer
    !---------------------------------------------------------------------------
    nUnit = newunit()
    write(buffer,'(a4,i7.7)') 'debugDep.', proc%rank
    open(file = trim(buffer), unit= nUnit, recl=1024)

    write(nUnit,*) 
    write(nUnit,*) 'GhostFromFiner      '
    write(nUnit,*) 
    write(nUnit,*) '-----------------------------------------------------------'
    call dump_MyGhostsFromFiner( pdfGlobal, intp, layout, glob, nUnit, tree )
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*) 
    write(nUnit,*) 'GhostFromCoarser    '
    write(nUnit,*) 
    write(nUnit,*) '-----------------------------------------------------------'
    call dump_FinerGhostsFromMe( pdfGlobal, intp, layout, glob, nUnit, tree )
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*) 
    write(nUnit,*) 'GhostFromCoarserBuffer'
    write(nUnit,*) 
    write(nUnit,*) '-----------------------------------------------------------'
    call dump_FinerGhostsFromMeBuffer( pdfGlobal, intp, layout, glob, nUnit,   &
      &                                tree )
    write(nUnit,*) '-----------------------------------------------------------'

    close( nUnit )

  end subroutine debug_dependencies
!******************************************************************************!


!******************************************************************************!
  !> check the dependencies from Finer
  !!
  subroutine dump_intpLists( pdfGlobal, intp, layout, glob, tree, proc ) 
    !---------------------------------------------------------------------------
    !> global pdf information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> state properties
    type( pdf_global_type ), intent(in) :: glob
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> musubi mpi communicator environment
    type( tem_comm_env_type ), intent(in) :: proc
    !---------------------------------------------------------------------------
    integer :: nUnit, iLevel
    character(len=17)  :: buffer
    !---------------------------------------------------------------------------
    nUnit = newunit()
    write(buffer,'(a4,i7.7)') 'dumpIntpLists.', proc%rank
    open(file = trim(buffer), unit= nUnit, recl=1024)

    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      write(nUnit,*) 
      write(nUnit,*) 'GhostFromFiner      '
      write(nUnit,*) 
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'LOW ---fromFiner-----------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromFinerElem, iLevel = iLevel,      &
        &    pdfGlobal = pdfGlobal, intp = intp, layout = layout, glob = glob, &
        &    ind = glob%levelDesc( iLevel)%intpFromFiner_low, nUnit = nUnit,   &
        &    tree = tree )
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'HIGH --fromFiner-----------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromFinerElem, iLevel = iLevel,      &
        &    pdfGlobal = pdfGlobal, intp = intp, layout = layout, glob = glob, &
        &    ind = glob%levelDesc( iLevel)%intpFromFiner_high, nUnit = nUnit,  &
        &    tree = tree )
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 
      write(nUnit,*) 'GhostFromCoarser    '
      write(nUnit,*) 
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'LOW ---fromCoarser---------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromCoarserElem, iLevel = iLevel,    &
        &    pdfGlobal = pdfGlobal, intp = intp, layout = layout, glob = glob, &
        &    ind = glob%levelDesc( iLevel)%intpFromCoarser_low, nUnit = nUnit, &
        &    tree = tree )
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'HIGH --fromCoarser --------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromCoarserElem, iLevel = iLevel,    &
        &    pdfGlobal = pdfGlobal, intp = intp, layout = layout, glob = glob, &
        &    ind = glob%levelDesc( iLevel)%intpFromCoarser_high, nUnit = nUnit,&
        &    tree = tree )
      write(nUnit,*) '---------------------------------------------------------'
    end do
    close( nUnit )

  end subroutine dump_intpLists
!******************************************************************************!


!******************************************************************************!
  !> check the dependencies from Finer and write them out so we can compare 
  !!
  subroutine dump_intpList( iLevel, eType, pdfGlobal, intp, layout, glob,      &
    &                       ind, nUnit, tree ) 
    !---------------------------------------------------------------------------
    !> global pdf information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> state properties
    type( pdf_global_type ), intent(in) :: glob
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> indirectio list
    type( grw_intArray_type ), intent(in)  :: ind
    !>
    integer, intent(in) :: nUnit
    !>
    integer, intent(in) :: iLevel
    !>
    integer, intent(in) :: eType
    !---------------------------------------------------------------------------
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: targetElem     ! element counter for indirection list
    integer(kind=long_k) :: tID
    character(len=pathLen) :: buffer
    !---------------------------------------------------------------------------

    buffer = ''
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      targetElem = iElem + glob%levelDesc( iLevel )                            &
        &                      %offset( 1, eType )
      tID = glob%levelDesc( iLevel )%total( targetElem )
      write(buffer, '(2a, i9)') trim(buffer), ' ', tID
      if( mod( indElem, 32 ) == 0 .or. indElem == ind%nVals ) then
        write(nUnit, *) trim(buffer)
        buffer = ''
      end if
    enddo

  end subroutine dump_intpList
!******************************************************************************!


!******************************************************************************!
  !> check the dependencies from Finer and write them out so we can compare 
  !!
  subroutine dump_MyGhostsFromFiner( pdfGlobal, intp, layout, glob, nUnit, tree)
    !---------------------------------------------------------------------------
    !> global pdf information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> state properties
    type( pdf_global_type ), intent(in) :: glob
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> unit to write to
    integer, intent(in) :: nUnit
    !--------------------------------------------------------------------------
    integer :: ilevel         ! grid refinement level
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iElem          ! current target element (for outer loop)
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer(kind=long_k) :: tID(0:intp%nMaxSources )
    !--------------------------------------------------------------------------
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel - 1
      sourceLevel = ilevel + 1
      targetLevel = ilevel
     
      ! Treat all coarse target elements
      do iElem = 1, glob%levelDesc( targetLevel )%nElems_ghostFromFiner
        ! Read the target element treeId
        targetElem = iElem + glob%levelDesc( targetlevel )                     &
          &                      %offset( 1, eT_ghostFromFinerElem )
        ! Find out how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = glob%levelDesc( ilevel )%depFromFiner( iElem )%elem%nVals
        tID = 0
        tID( 0 ) = glob%levelDesc( targetLevel )%total( targetElem )
          
        write(nUnit, '(a,i9,3f10.5)') ' targetGhost  ', tID( 0 ),              &
          &                           tem_baryOfId( tree, tID(0) )
         
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
     
          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )     &
            &                                       %elem%val( iSourceElem )
          tID( iSourceElem ) = glob%levelDesc( sourceLevel )%total( sourceElem )
          
          call dump_elemDep( targetElem = tID(0),                              &
            &                sourceElem = tID( iSourceElem ),                  &
            &                nUnit = nUnit,                                    &
            &                levelDesc = glob%levelDesc,                       &
            &                tree = tree)
        end do  ! iSourceElem  
      enddo
    enddo

  end subroutine dump_MyGhostsFromFiner
!******************************************************************************!


!******************************************************************************!
  !> check the dependencies from Coarser
  !!
  subroutine dump_FinerGhostsFromMe( pdfGlobal, intp, layout, glob, nUnit, tree)
    !---------------------------------------------------------------------------
    !> global pdf information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> state properties
    type( pdf_global_type ), intent(in) :: glob
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> unit to write to
    integer, intent(in) :: nUnit
    !---------------------------------------------------------------------------
    integer :: ilevel         ! grid refinement level
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iElem          ! current target element (for outer loop)
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer(kind=long_k) :: tID(0:intp%nMaxSources )
    logical :: weights
    !---------------------------------------------------------------------------
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel - 1
      sourceLevel = ilevel
      targetLevel = ilevel + 1
     
      ! Treat all coarse target elements
      do iElem = 1, glob%levelDesc( targetLevel )%nElems_ghostFromCoarser
        ! Read the target element treeId
        targetElem = iElem + glob%levelDesc( targetlevel )                     &
          &                      %offset( 1, eT_ghostFromCoarserElem )
        ! Find out how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = glob%levelDesc( targetlevel )%depFromCoarser( iElem )%  &
          &                                                          elem%nVals
        tID = 0
        tID( 0 ) = glob%levelDesc( targetLevel )%total( targetElem )
        write(nUnit, '(a,i9,3f10.5)') ' targetGhost  ', tID( 0 ),              &
          &                                        tem_baryOfId( tree, tID(0) )
         
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
     
          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                       %elem%val( iSourceElem )
          tID( iSourceElem ) = glob%levelDesc( sourceLevel )%total( sourceElem )
        
          if( allocated( glob%levelDesc( targetLevel )                         &
              &       %depFromCoarser( iElem )%weight)) then
              weights = .true. 
          else
            weights = .false.
          end if
          if( weights ) then
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                levelDesc = glob%levelDesc,                     &
              &                tree = tree,                                    &
              &                weight = glob%levelDesc( targetLevel )          &
              &                   %depFromCoarser( iElem )%weight( iSourceElem))
          else
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                levelDesc = glob%levelDesc,                     &
              &                tree = tree )
          end if
  
        end do  ! iSourceElem  
      enddo
    enddo

  end subroutine dump_FinerGhostsFromMe
!******************************************************************************!


!******************************************************************************!
  !> check the dependencies from Coarser
  !!
  subroutine dump_FinerGhostsFromMeBuffer( pdfGlobal, intp, layout, glob,      &
    &                                      nUnit, tree ) 
    !---------------------------------------------------------------------------
    !> global pdf information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !> state properties
    type( pdf_global_type ), intent(in) :: glob
    !> the layout used
    type( mus_scheme_layout_type), intent(in)   :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> unit to write to
    integer, intent(in) :: nUnit
    !---------------------------------------------------------------------------
    integer :: ilevel         ! grid refinement level
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iElem          ! current target element (for outer loop)
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer(kind=long_k) :: tID(0:intp%nMaxSources )
    integer :: totalPos ! Position of the element in the total list
    logical :: weights
    !---------------------------------------------------------------------------
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel - 1
      sourceLevel = ilevel
      targetLevel = ilevel + 1
     
      ! Treat all coarse target elements
      do iElem = 1, glob%levelDesc( targetLevel )%nElems_ghostFromCoarser
        ! Read the target element treeId
        targetElem = iElem + glob%levelDesc( targetlevel )                     &
          &                      %offset( 1, eT_ghostFromCoarserElem )
        ! Find out how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = glob%levelDesc( targetlevel )%depFromCoarser( iElem )%  &
          &                                                          elem%nVals
        tID = 0
        tID( 0 ) = glob%levelDesc( targetLevel )%total( targetElem )
        write(nUnit, '(a,i9,3f10.5)') ' targetGhost  ', tID( 0 ),              &
          &                           tem_baryOfId( tree, tID(0) )
         
        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems
     
          ! Get the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                    %elemBuffer%val( iSourceElem )
          totalPos = glob%levelDesc( targetLevel )%sourceFromCoarser%          &
            &                                                 val( sourceElem )
          tID( iSourceElem ) = glob%levelDesc( sourceLevel )%total( totalPos )
        
          if( allocated( glob%levelDesc( targetLevel )                         &
              &       %depFromCoarser( iElem )%weight)) then
              weights = .true. 
          else
            weights = .false.
          end if
          if( weights ) then
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                levelDesc = glob%levelDesc,                     &
              &                tree = tree,                                    &
              &                weight = glob%levelDesc( targetLevel )          &
              &                   %depFromCoarser( iElem )%weight( iSourceElem))
          else
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                levelDesc = glob%levelDesc,                     &
              &                tree = tree )
          end if
  
        end do  ! iSourceElem  
      enddo
    enddo

  end subroutine dump_FinerGhostsFromMeBuffer
!******************************************************************************!


!******************************************************************************!
  !> dump dependencies for one element
  !!
  subroutine dump_elemDep( targetElem, sourceElem, nUnit, levelDesc, tree,     &
    &                      weight )
    !---------------------------------------------------------------------------
    !>
    integer(kind=long_k), intent(in) :: targetElem
    !>
    integer(kind=long_k), intent(in) :: sourceElem
    !>
    type( tem_levelDesc_type ), intent(in) :: levelDesc(:)
    !>
    integer, intent(in) :: nUnit
    !>
    type( treelmesh_type ), intent(in) :: tree
    !>
    real(kind=rk), optional :: weight
    !---------------------------------------------------------------------------
    real(kind=rk) :: xTarget(3), xSource(3)
    real(kind=rk) :: elemSize
    character(len=pathLen) :: buffer
    !---------------------------------------------------------------------------
    buffer = ''
    elemSize = tem_elemSize( tree, targetElem )
    xTarget = tem_baryOfId( tree, targetElem )
    xSource = tem_baryOfId( tree, sourceElem )
    write(buffer,'(a,i9, 6f10.5)') '  ', sourceElem,                           &
      &          ((xSource(1:3) - xTarget(1:3))/elemSize), xSource(1:3)
    if( present( weight )) then
      write(buffer, '(a,a,f8.5)') trim( buffer), '    weight: ', weight
    end if
    write(nUnit, '(a)') trim(buffer)

  end subroutine dump_elemDep
!******************************************************************************!

 
end module mus_interpolate_tools_module
!******************************************************************************!
