! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Interpolation scheme tools
!!
!! For an overview over implemented interpolation methods, see
!! \ref intp_methods "Interpolation methods"
!!
!!\author Manuel Hasert
!!
?? include 'header/lbm_macros.inc'
module mus_interpolate_verify_module

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val, close_config
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length

  ! include treelm modules
  use mpi
  use env_module,              only: rk, labelLen, long_k, newUnit, pathLen,   &
    &                                rk_mpi
  use treelmesh_module,        only: treelmesh_type
  use tem_aux_module,          only: tem_abort
  use tem_construction_module, only: tem_levelDesc_type
  use tem_element_module,      only: eT_GhostFromCoarserElem,                  &
    &                                eT_ghostFromFinerElem, eT_fluidElem
  use tem_geometry_module,     only: tem_baryOfId
  use tem_grow_array_module,   only: grw_intArray_type, init, append, destroy
  use tem_param_module,        only: cs2inv, cs2
  use tem_var_system_module,   only: tem_var_system_type
  use tem_subTree_type_module, only: tem_subTree_type, tem_subTree_from
  use tem_logging_module,      only: logUnit
  use tem_spatial_module,      only: tem_spatial_for

  ! include musubi modules
  use mus_config_module,             only: mus_open_config
  use mus_field_prop_module,         only: mus_field_prop_type 
  use mus_param_module,              only: mus_param_type
  use mus_interpolate_header_module, only: interpolation_type,                 &
                                           determine_timeLayer
  use mus_pdf_module,                only: pdf_data_type, pdf_global_type,     &
    &                                      pdf_globInfo_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_moments_module,            only: init_transformation_matrix,         &
    &                                      set_momentIndices
  use mus_variable_module,           only: mus_derVarPos_type

  implicit none

  private

  public :: mus_testInterpolation

 contains

!******************************************************************************!
  !> Call tests to determine the actual error from the interpolation routines on
  !! the ghost elements. Compare against the analytical solution, which is
  !! given in terms of the initial conditions.
  !! Call this routine after the initial values are set and the ghost elements
  !! have been filled once, but no computation was started 
  !! -> after fillHelperElements in the mus_aux_module
  !!
  subroutine mus_testInterpolation( scheme, tree, params,                      &
    &                       iLevel, intp, pdfGlobal, iTimeLayer  )
    !---------------------------------------------------------------------------
    !>
    type(mus_scheme_type)  :: scheme
    !> treelm tree
    type( treelmesh_type ), intent(in) :: tree
    !>
    type(mus_param_type)       :: params  
    !>
    type(interpolation_type) :: intp      
    !>
    type(pdf_globInfo_type) :: pdfGlobal
    !>Level counter variable
    integer, intent(in) :: iLevel
    !> which time layer to fill
    integer, intent(in) :: iTimeLayer
    !---------------------------------------------------------------------------
    integer :: iElem     ! element counter
    type( grw_intArray_type ) :: local_ind
    logical :: checkFluid, checkFF, checkFC
    !---------------------------------------------------------------------------
    checkFluid = intp%testFluids
    checkFF    = .true.
    checkFC    = .true.
    write(logUnit(1),*)' Checking error on level...', iLevel
    write(logUnit(1),*)' Fluid ...', iLevel
    call init( me = local_ind, length = scheme%pdf( iLevel )%nElems_fluid )
    if( checkFluid ) then
      do iElem = 1, scheme%pdf( iLevel )%nElems_fluid
        call append( me = local_ind, val = iElem )
      end do
      call mus_intp_error( pdf       = scheme%pdf,                             &
        &                  glob      = scheme%glob,                            &
        &                  scheme    = scheme,                                 &
        &                  params    = params,                                 &
        &                  tree      = tree,                                   &
        &                  fieldProp = scheme%field(:)%fieldProp,              &
        &                  level     = iLevel,                                 &
        &                  eType     = eT_fluidElem,                           &
        &                  intp      = intp,                                   &
        &                  ind       = local_ind,                              &
        &                  layout    = scheme%layout,                          &
        &                  varSys    = scheme%varSys,                          &
        &                  derVarPos = scheme%derVarPos  )  
      call destroy( me = local_ind )
    end if
    if( iLevel < pdfGLobal%maxLevel .and. checkFF ) then
      write(logUnit(1),*)' Checking fromFiner interpolation error...'
      call mus_intp_error( pdf   = scheme%pdf,                                 &
          &                glob      = scheme%glob,                            &
          &                scheme    = scheme,                                 &
          &                params    = params,                                 &
          &                tree      = tree,                                   &
          &                fieldProp = scheme%field(:)%fieldProp,              &
          &                level     = iLevel,                                 &
          &                eType     = eT_ghostFromFinerElem,                  &
          &                intp      = intp,                                   &
          &                ind       = scheme%glob%levelDesc(iLevel)%          &
          &                                                intpFromFiner_high, &
          &                layout    = scheme%layout,                          &
          &                varSys    = scheme%varSys,                          &
          &                derVarPos = scheme%derVarPos  )  
    end if
    if( iLevel > pdfGLobal%minLevel .and. checkFC ) then
      write(logUnit(1),*)' Checking fromCoarser interpolation error...'
      call mus_intp_error(                                                     &
        &      pdf       = scheme%pdf,                                         &
        &      glob      = scheme%glob,                                        &
        &      scheme    = scheme,                                             &
        &      params    = params,                                             &
        &      tree      = tree,                                               &
        &      fieldProp = scheme%field(:)%fieldProp,                          &
        &      level     = iLevel,                                             &
        &      eType     = eT_ghostFromCoarserElem,                            &
        &      intp      = intp,                                               &
        &      ind       = scheme%glob%levelDesc(iLevel)%intpFromCoarser_high, &
        &      layout    = scheme%layout,                                      &
        &      varSys    = scheme%varSys,                                      &
        &      derVarPos = scheme%derVarPos  )  
    end if
 
    
  end subroutine mus_testInterpolation
!******************************************************************************!


!******************************************************************************!
  !> Determine the numerical error of the interpolated quantities to the given 
  !! initial conditions
  !!
  subroutine mus_intp_error( pdf, glob, fieldProp, level, tree,                &
    &           params,scheme, layout, intp, ind, varSys, derVarPos, eType ) 
    !---------------------------------------------------------------------------
    !> global fluid parameters 
    type( pdf_data_type ), intent(inout) :: pdf(:)
    !> global parameters 
    type( pdf_global_type ), intent(inout) :: glob
    !>
    type(mus_param_type)       :: params  
    !> treelm tree
    type( treelmesh_type ), intent(in) :: tree
    !> the current scheme 
    type( mus_scheme_type ), intent(in)   :: scheme
    !> fluid parameters
    type( mus_field_prop_type ), intent(in) ::  fieldProp(:)
    !> level
    integer, intent(in)   :: level
    !> element type, eT_ghostFromFinerElem, eT_ghostFromCoarserElem
    integer, intent(in)   :: eType 
    !> the layout used
    type( mus_scheme_layout_type), intent(in) :: layout
    !> interpolation method info
    type( interpolation_type ), intent(in)   :: intp
    !> indirection list
    type( grw_intArray_type ), intent(in)   :: ind
    !> global variable system
    type( tem_var_system_type ), intent(in) :: varSys
    !> required variable system which maps to global varsys
    type( mus_derVarPos_type ), intent(in) :: derVarPos(:)
    !---------------------------------------------------------------------------
    integer :: targetLevel      ! level of target elements
    integer :: targetElem       ! treeId of current source element
    integer :: iVal, iPos       ! value counter
    integer :: iElem, nElems, globnElems    ! element counter for outer loop
    integer :: nStresses        ! number of stress variables
    integer :: indElem          ! element index in from the index list
    integer :: fUnit            ! unit for file write access
    real(kind=rk) :: targetPdf( layout%stencil(1)%QQ ) !< pdf to reconstruct
    real(kind=rk) :: targetMom( layout%stencil(1)%QQ )
    integer :: iVar ! current field variable
    integer :: t_target ! which time layer to write to
    integer :: iErr ! mpi error
    integer :: nChunkELems, nDims, iVelMin, iVelMax
    integer :: iStressMin, iStressMax, iPress
    integer(kind=long_k) :: targetID
    integer(kind=long_k), allocatable :: minErrorID(:), maxErrorID(:)
    real(kind=rk) :: targetVel(3), refVel(3), targetPress(1), refPress
    real(kind=rk) :: targetStress(6), targetStrain(6)
    ! temporary handle for the config file
    type(flu_state), allocatable :: conf(:)
    real(kind=rk), allocatable :: xc(:,:), ux(:), uy(:), uz(:), rho(:)
    real(kind=rk), allocatable :: Sxx(:,:)
    real(kind=rk), allocatable :: maxError(:), minError(:), error(:)
    real(kind=rk), allocatable :: errorNorm(:)
    real(kind=rk), allocatable :: globMaxError(:), globminError(:)
    real(kind=rk), allocatable :: globErrorNorm(:)
    character(len=labelLen) :: intpString 
    character(len=PathLen) :: buffer, filename, header
    logical :: file_exists, dumpElementError
    ! local subTree needed to use the function pointers
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    dumpElementError = intp%testEachElement
    ! Allocate even though no test might been done, for parallel runs
    ! Other processes might have elements to check
    allocate(Error(layout%stencil(1)%QQ))
    allocate(minError(layout%stencil(1)%QQ))
    allocate(maxError(layout%stencil(1)%QQ))
    allocate(minErrorID(layout%stencil(1)%QQ))
    allocate(maxErrorID(layout%stencil(1)%QQ))
    allocate(errorNorm(layout%stencil(1)%QQ))
    allocate(globerrorNorm(layout%stencil(1)%QQ))
    allocate(globminError(layout%stencil(1)%QQ))
    allocate(globmaxError(layout%stencil(1)%QQ))
    minError = huge(minError)
    maxError = tiny(minError)
    error = 0._rk
    errorNorm = 0._rk
    nElems = ind%nVals

    ! copy the global information of the global tree to the subTree
    subTree%global = tree%global

    ! Only compute errors if there are any elements to test.
    if( nElems  > 0 ) then
      if( eType == eT_ghostFromCoarserElem ) then
        intpString = 'intpFromCoarser'
      elseif( eType == eT_ghostFromFinerElem ) then
        intpString = 'intpFromFiner'
      else
        intpString = 'noIntp'
      endif
      nDims = layout%stencil(1)%nDims
      call set_momentIndices( nDims, iPress, iVelMin, iVelMax,                 &
        &                     iStressMin, iStressMax )
      nStresses = iStressMax - iStressMin + 1

      call mus_open_config( conf     = conf,                                   &
        &                   filename = params%general%solver%configFile,       &
        &                   proc     = params%general%proc )
      targetLevel = level
      t_target = determine_timeLayer( intp%timeLayer, pdf( targetLevel ))
      nChunkElems = 1
      allocate(xc(nChunkElems, 3))
      allocate(rho(nChunkElems))
      allocate(ux(nChunkElems))
      allocate(uy(nChunkElems))
      allocate(uz(nChunkElems))
      allocate(Sxx(6,nChunkElems))
      minErrorID = 0_long_k
      maxErrorID = 0_long_k
      errorNorm = 0._rk

      maxError = tiny(1._rk)
      minError = huge(1._rk)
      do indElem = 1, nElems
        iElem = ind%val( indElem )
        ! Read the target element treeId
        targetElem = iElem + glob%levelDesc( targetLevel )%offset( 1, eType )
        targetID   = glob%levelDesc( targetLevel )%total( targetElem )
        ! get the target coordinates 
        xc(1,:) = tem_baryOfId( tree, targetID )

        ! initialize the lower and upper bound
        bound(1)=indElem
        bound(2)=indElem
        call tem_subTree_from( me = subTree, treeID = (/targetID/))

        do iVar = 1, varSys%nVars
          do iVal = 1, layout%stencil(1)%QQ
            targetPdf(iVal) =  &
            pdf( targetLevel )%state( &
  &?SAVE?( varSys%Variable%val(iVar)%varPos(iVal), targetElem, varSys%nScalars, pdf(targetLevel)%neigh ),&
             & t_target  )
          enddo
          targetMom = matmul( layout%moment%toMoments%A, targetPdf)

          call varSys%derive( derVarPos(iVar)%pressure )%do(                   &
            &       input = targetPdf,                                         &
            &       bound = bound,                                             &
            &       tree = tree,                                               &
            &       subTree = subTree,                                         &
            &       requVar = varSys%Variable%val( derVarPos(iVar)%pressure ), &
            &       globSys = varSys,                                          &
            &       stencil = layout%stencil(1),                               &
            &       conf = varSys%solSpec%conf,                                & 
            &       res = targetPress )
          targetPress = targetPress * params%physics%fac( targetLevel )%press
          call varSys%derive( derVarPos(iVar)%velocity )%do(                   &
            &       input = targetPdf,                                         &
            &       bound = bound,                                             &
            &       tree = tree,                                               &
            &       subTree = subTree,                                         &
            &       requVar = varSys%Variable%val( derVarPos(iVar)%velocity ), &
            &       globSys = varSys,                                          &
            &       stencil = layout%stencil(1),                               &
            &       conf = varSys%solSpec%conf,                                &
            &       res = targetVel(:) )
          targetVel = targetVel * params%physics%fac( targetLevel )%vel
          call varSys%derive( derVarPos(iVar)%shearstress )%do(                &
            &    input = targetPdf,                                            &
            &    bound = bound,                                                &
            &    tree = tree,                                                  &
            &    subTree = subTree,                                            &
            &    requVar = varSys%Variable%val( derVarPos(iVar)%shearstress ), &
            &    globSys = varSys,                                             &
            &    stencil = layout%stencil(1),                                  &
            &    conf = varSys%solSpec%conf,                                   &
            &    res = targetStress(:) )
          targetStress(:) = targetStress(:)                                    &
            &             * params%physics%fac( targetLevel)%press
          !if( params%scaling == 'diffusive' ) then
          targetStrain = targetStress/params%physics%fac(targetLevel)%visc     &
            &          / params%physics%rho0
          !end if

          rho = tem_spatial_for(conf=conf(1),                                  &
            & me=scheme%field(iVar)%ic%ini_state(1),                           &
            & coord=xc, n=nChunkElems )
          ux =  tem_spatial_for(conf=conf(1),                                  &
            & me=scheme%field(iVar)%ic%ini_state(2),                           &
            & coord=xc, n=nChunkElems )
          uy =  tem_spatial_for(conf=conf(1),                                  &
            & me=scheme%field(iVar)%ic%ini_state(3),                           &
            & coord=xc, n=nChunkElems )
          uz =  tem_spatial_for(conf=conf(1),                                  &
            & me=scheme%field(iVar)%ic%ini_state(4),                           &
            & coord=xc, n=nChunkElems )
          ! Read in the shear rate tensor
          ! This corresponds to S = grad(u) + grad(u)^T
          Sxx = 0._rk
          Sxx(1,:) =  tem_spatial_for(conf=conf(1),                            &
            & me=scheme%field(iVar)%ic%ini_state(5),                           &
            & coord=xc, n=nChunkElems )
          Sxx(2,:) =  tem_spatial_for(conf=conf(1),                            &
            & me=scheme%field(iVar)%ic%ini_state(6),                           &
            & coord=xc, n=nChunkElems )
          Sxx(3,:) =  tem_spatial_for(conf=conf(1),                            &
            & me=scheme%field(iVar)%ic%ini_state(7),                           &
            & coord=xc, n=nChunkElems )
          Sxx(4,:) =  tem_spatial_for(conf=conf(1),                            &
            & me=scheme%field(iVar)%ic%ini_state(8),                           &
            & coord=xc, n=nChunkElems )
          Sxx(5,:) =  tem_spatial_for(conf=conf(1),                            &
            & me=scheme%field(iVar)%ic%ini_state(9),                           &
            & coord=xc, n=nChunkElems )
          Sxx(6,:) =  tem_spatial_for(conf=conf(1),                            &
            & me=scheme%field(iVar)%ic%ini_state(10),                          &
            & coord=xc, n=nChunkElems )
          if( scheme%layout%stencil(1)%nDims == 2 ) then
            Sxx(3,:) = Sxx(4,:)
            targetStrain(3) = targetStrain(4)
          end if

          refVel(:) = [ux(1), uy(1), uz(1)]
          refPress  = rho(1)
          ! Error computation
          error = 0._rk
          ! Absolute error
          error(1) = abs( (targetPress(1)-refPress) )
          error(iVelMin:iVelMax) = abs( (targetVel(1:nDims)-refVel(1:nDims)) )
          error(iStressMin:iStressMax ) = abs( (targetStrain(1:nStresses)      &
            &                                 - Sxx(1:nStresses,1)) )
          ! Relative error for pressure
          if( abs(refPress) > 0.000000001_rk ) then
            error(1) = error(1)/refPress
          end if

          ! Relative error for velocity
          do iVal = iVelMin, iVelMax
            if( abs(refVel(iVal-1)) > 0.000000001_rk ) then
              error( iVal ) = abs(error( iVal)/refVel(iVal-1 ))
            end if
          end do
          ! Relative error for shear stress
          do iVal = iStressMin, iStressMax
            iPos = iVal - iStressMin + 1
            if( abs(Sxx(iPos, 1)) > 0.000000001_rk ) then
              error( iVal ) = abs(error( iVal)/Sxx(iPos, 1 ))
            end if
          end do
          ! Determine min and max error and where it occurred
          do iVal = 1, size( error )
            errorNorm( iVal ) = errorNorm( iVal ) + error( iVal )**2
            if( abs(error( iVal ))> maxError( iVal)) then 
              maxError( iVal ) = abs(error( iVal ))
              maxErrorID( iVal ) = targetID
            end if
            if( abs(error( iVal)) < minError( iVal)) then 
              minError( iVal ) = abs(error( iVal ))
              minErrorID( iVal ) = targetID
            end if
          end do
          if( dumpElementError ) then
            write(logUnit(1),*) targetID 
            write(logUnit(1),*) 'ref  ',refPress, refVel(1:nDims),             &
              &             Sxx(1:nStresses, 1)
            write(logUnit(1),*) 'val  ',targetPress(1), targetVel(1:nDims),    &
              &             targetStrain(1:nStresses)
            write(logUnit(1),*) 'error ',error(1:iStressMax)
          end if
        enddo ! ivar loop
      enddo
      ! Get rid of the max and min values
      do iVal = 1, size(error)
        if( minError( iVal ) < 1.E-90_rk ) then
          minError( iVal ) = 0._rk
        end if
        if( maxError( iVal ) < 1.E-90_rk ) then
          maxError( iVal ) = 0._rk
        end if
      end do
    end if ! nVals > 0


    if( params%general%proc%comm_size > 1 ) then
      ! Communicate error to root process
      call mpi_reduce( errorNorm, globErrorNorm, size(errorNorm), rk_mpi,      &
        &              mpi_sum, 0, params%general%proc%comm, iErr )
      call mpi_reduce( maxError, globMaxError, size(maxError), rk_mpi,         &
        &              mpi_max, 0, params%general%proc%comm, iErr )
      call mpi_reduce( minError, globMinError, size(minError), rk_mpi,         &
        &              mpi_min, 0, params%general%proc%comm, iErr )
      call mpi_reduce( nElems,   globnELems,  1, mpi_integer,                  &
        &              mpi_sum, 0, params%general%proc%comm, iErr )
    else
      globErrorNorm = errorNorm
      globMaxError  = maxError  
      globMinError  = minError  
      globnElems    = nElems
    end if

    if( params%general%proc%rank == 0 ) then
      ! Build the L2 norm error = 1/n*(sqrt( sum (phi - phi_error)^2 )
      globErrorNorm = sqrt( globErrorNorm ) / real( globnElems, kind=rk )
      write(buffer,'(2i4)') tree%global%minLevel, tree%global%maxlevel
      write(buffer,'(3a)') trim( buffer ), '  ', trim( params%scaling )
      write(buffer,'(3a)') trim( buffer ), '  ', trim( intp%label )
      do iVal = 1, size(error)
        write(logUnit(1),*)trim(layout%moment%momLabel( iVal )),               &
          &                   ' E max ', globmaxError( iVal ),                 &
          &                   ' E min ', globminError( iVal ),                 &
          &                   ' norm ', globErrorNorm( iVal )
        write(buffer, '(2a, e15.5)') trim( buffer ), '  ',                     &
          &     maxError( iVal )
      end do
      do iVal = 1, size( error )
        write(buffer, '(a, e16.5)') trim( buffer ),                            &
          &     globErrorNorm( iVal )
      end do
      !call close_config(conf)
    
      write(filename,'(4a)') 'error_', trim(intpString), '.res'
      header = '' 
      write(header,'(a,a1)') trim(header), '#'
      write(header,'(a,a)') trim(header), 'Lmin Lmax'
      write(header,'(3a)') trim(header), '  scaling'
      write(header,'(3a)') trim(header), '  intp '
      do iVal = 1, size( error )
        write(header,'(4a)') trim(header), ' max',                             &
          &         trim(adjustl(layout%moment%momLabel( iVal )))
      end do
      do iVal = 1, size( error )
        write(header,'(4a)') trim(header), ' rms',                             &
          &         trim(adjustl(layout%moment%momLabel( iVal )))
      end do

      inquire(file=filename,exist=file_exists)
      fUnit = newunit()
      open(unit=fUnit,file=trim(filename),position='append' ) 
      if( .not. file_exists ) then
         write(fUnit,'(a)') trim(header)
      end if
      write( fUnit, *) trim( buffer )
      close( fUnit )
    end if
      
  end subroutine mus_intp_error
!******************************************************************************!


end module mus_interpolate_verify_module
!******************************************************************************!
