!******************************************************************************!
!> Auxiliary functionality
!!
!! \author Manuel Hasert
!!
module mus_aux_module

  ! include musubi modules
  use mus_comm_module,               only: mus_exchange
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_param_module,              only: mus_param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_tools_module,              only: perform_checks,                     &
    &                                      check_streaming_layout
  use mus_tracking_module,           only: mus_tracker, mus_track_convergence
  use mus_restart_module,            only: mus_readRestart, mus_writeRestart
  use mus_bc_general_module,         only: init_boundary
  use mus_interpolate_header_module, only: interpolation_type
  use mus_interpolate_module,        only: mus_init_interpolate
  use mus_interpolate_verify_module, only: mus_testInterpolation
  use mus_moments_module,            only: mus_init_moments
  use mus_scheme_module,             only: mus_init_sources
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_scheme_header_module,      only: mus_scheme_header_type
  use mus_geomIncr_module,           only: mus_geomIncr, mus_init_geomIncr
  use mus_geom_module,               only: mus_geom_setWeight,                 &
    &                                      mus_geom_dumpWeights
  use mus_interpolate_tools_module,  only: debug_dependencies, dump_intpLists
  use mus_time_module,               only: mus_time_modulo,                    &
    &                                      mus_time_homogenize,                &
    &                                      mus_timeControl_homogenize
  use mus_fluid_module,              only: mus_update_relaxation, &
    &                                      mus_alloc_mrt

  ! include treelm modules
  use env_module,               only: rk, fin_env, PathLen, pathSep
  use treelmesh_module,         only: treelmesh_type, unload_tem
  use tem_timer_module,         only: tem_addTimer
  use tem_aux_module,           only: tem_print_execInfo, utc_date_string
  use tem_param_module,         only: cs2inv, cs2, div1_2
  use tem_math_module,          only: invert_matrix
  use tem_restart_module,       only: tem_init_restart
  use tem_spatial_module,       only: tem_spatial_for
  use tem_tools_module,         only: tem_horizontalSpacer
  use tem_balance_module,       only: Bal_logUnit, tem_balance_initTracks
  use tem_balanceSparta_module, only: tem_balanceSparta
  use tem_tracking_module,      only: tem_init_tracker
  use tem_transient_module,     only: tem_transient_for
  use tem_timeControl_module,   only: tem_timeControl_check,                   &
    &                                 tem_timeControl_update,                  &
    &                                 tem_timeControl_dump
  use tem_simControl_module,    only: tem_simControl_syncUpdate
  use tem_time_module,          only: tem_time_dump
  use tem_debug_module,         only: main_debug
  use tem_solveHead_module,     only: tem_solveHead_type
  use tem_logging_module,       only: logUnit
  use tem_global_module,        only: tem_global_type
  use tem_bc_prop_module,       only: dump_tem_BC_prop, dump_tem_BC_qVal

  implicit none
  private

  public :: check_flow_status
  public :: mus_init_aux
  public :: mus_setParameters
  public :: mus_balance
  public :: mus_addExternalForceToState
  public :: mus_banner
  public :: fillHelperElementsCoarseToFine
  public :: fillHelperElementsFineToCoarse

contains

!******************************************************************************!
  !> This routine performs several tasks: geometry increment, time updating,
  !! tracking, density checking, restart
  subroutine check_flow_status( scheme, params, geometry, pdfGlobal)
    !--------------------------------------------------------------------------
    !> containers for the different schemes
    type(mus_scheme_type), intent(inout)   :: scheme(:)
    !> global flow quantity properties
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> geometry infomation
    type(mus_geom_type),intent(in)         :: geometry
    !> global parameters
    type(mus_param_type),intent(inout)         :: params
    !--------------------------------------------------------------------------
    integer :: iScheme !< scheme counter
    !--------------------------------------------------------------------------

    ! Call the geometry increment module which  performs solidification 
    ! or liquification based on certain criteria read from the Lua file
    if( geometry%dynamicGeom ) then
      call mus_geomIncr(geometry = geometry, scheme = scheme,                  &
        &               pdfGlobal = pdfGlobal,                                 &
        &               commPattern = params%general%commPattern,              &
        &               params = params )
    end if ! Time for geom incr

    ! perform density check for all schemes
    do iScheme = 1, pdfGlobal%nSchemes
      ! Perform run-time checks for total density, NaNs 
      call perform_checks( scheme(iScheme), pdfGlobal, params )

      ! check for convergence only if abortCriteria is set for steady_state
      if( params%general%simControl%abortCriteria%steady_state) then
        ! check for steady state convergence based on convergence criteria
        ! defined in tracking table
        call mus_track_convergence( pdf = scheme(iScheme)%pdf,                 &
            &                       track = scheme(iScheme)%track,             &
            &                       params = params,                           &
            &                       layout = scheme(iScheme)%layout,           &
            &                       globSys = scheme(iScheme)%varSys,          &
            &                       glob_tree = geometry%tree )

      endif
    end do ! iScheme

    ! in musubi, advance time step separtely in control_routine so no need
    ! to pass dt to syncUpdate
    ! This sync update: check for stop file and time 
    ! control interval trigger, communicate status bits and update timeControl
    call tem_simControl_syncUpdate( me      = params%general%simControl,     &
      &                             proc    = params%general%proc ) 

    ! perform tracking for all schemes
    do iScheme = 1, pdfGlobal%nSchemes
      ! tracking
      if( scheme( iScheme )%trackingControl%active ) then
        call mus_tracker( pdf = scheme(iScheme)%pdf,   &
          &           track = scheme(iScheme)%track,   &
          &           params = params,                 &
          &           layout = scheme(iScheme)%layout, &
          &           glob_tree = geometry%tree,       &
          &           globSys = scheme(iScheme)%varSys )
      end if
    end do ! iScheme

    if( params%general%restart%controller%writeRestart ) then
      ! check time control and update it if triggered
      call tem_timeControl_check(                                              &
        &               me     = params%general%restart%controller%timeControl,&
        &               now    = params%general%simControl%now,                &
        &               comm   = params%general%proc%comm,                     &
        &               triggered = params%restart_triggered )


      ! dump restart if restart timeControl is triggered 
      if( params%restart_triggered )  then

        write(logUnit(1),*) 'Writing restart at time...'
        call tem_time_dump( params%general%simControl%now, logUnit(1) )
        call mus_writeRestart( pdfGlobal = pdfGlobal,                          &
          &                    restart   = params%general%restart,             &
          &                    scheme    = scheme,                             &
          &                    tree      = geometry%tree,                      &
          &                    timing    = params%general%simControl%now,      &
          &                    timerData = params%general%timerData,           &
          &                    timerHandle = params%wRestartTimer              )
        write(logUnit(1),*) 'Done writing'
      end if
    end if ! write Restart

  end subroutine check_flow_status
!******************************************************************************!



!******************************************************************************!
  !> This routine performs the load balancing for multilevel simulations. The
  !! weights are calculated on the basis of levelwise run time, which are then
  !! fed to sparta for calculation of splitting positions. Restart files are 
  !! saved and the simulation is restarted with the newly distributed mesh
  subroutine mus_balance( scheme, pdfGlobal, params, geometry )
    !---------------------------------------------------------------------------
    !> The scheme type
    type(mus_scheme_type), allocatable, intent(inout)   :: scheme(:)
    !> global flow quantity properties
    type(pdf_globInfo_type), intent(inout) :: pdfGlobal
    !> global parameters
    type(mus_param_type), intent(inout)    :: params
    !> geometry infomation
    type(mus_geom_type),intent(inout)      :: geometry
    !---------------------------------------------------------------------------
    integer :: iLevel
    ! characters for logging output
    character(len=PathLen)    :: header
    character(len=PathLen)    :: output
    character(len=PathLen)    :: filename
    character(len=PathLen)    :: suffix
    integer :: pos
    ! tmp global type for keeping old tree information
    type( tem_global_type ) :: locGlobal
    !---------------------------------------------------------------------------

    if( .not. allocated(params%level_weight) ) &
      & allocate ( params%level_weight(pdfGlobal%minLevel:pdfGlobal%maxLevel) )
    if( .not. allocated(params%level_weight_bnd) ) &
      & allocate ( params%level_weight_bnd(pdfGlobal%minLevel:pdfGlobal%maxLevel) )

    ! Here weights are assigned depending on user request. In case of level wise
    ! weights, only levelwise weights are calculated whereas in case of
    ! boundary, the weights are calculated which are based on boundary time and
    ! then passed on to set weight routine for special treatment

    select case( trim(params%general%balance%balanceKind) )
    case( 'levelwise' )
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        if( scheme(1)%glob%levelDesc(iLevel)%nElems_fluid > 0) then
          params%level_weight(iLevel) = ( params%timer(iLevel)%value(params%   &
            &                 lw_tIntp(iLevel) )                               &
            &               + params%timer(iLevel)%value(params%               &
            &                 lw_tCompute(iLevel) )                            & 
            &               / scheme( 1 )%glob%levelDesc(iLevel)%nElems_fluid ) 
        end if
      end do
    case( 'boundary' )
      call tem_horizontalSpacer(fUnit=Bal_logUnit(3))
      header = ''
      if( any( scheme( 1 )%glob%levelDesc(:)%nElems_fluid > 0 ))then
        write(header, '(a,a7)')  trim(header), 'Level'
        write(header, '(a,a7)')  trim(header), 'rank'
        write(header, '(a,a20)') trim(header), 'treeID first'
        write(header, '(a,a20)') trim(header), 'treeID last'
        write(header, '(a,a20)') trim(header), 'Level Weight'
        write(header, '(a,a20)') trim(header), 'Time Level Weight'
        write(header, '(a,a20)') trim(header), 'nElems Level'
      end if
      if( any(scheme(1)%glob%levelDesc(:)%bc_elemBuffer%nVals > 0 )) then
        write(header, '(a,a20)') trim(header), 'BND Weight'
        write(header, '(a,a20)') trim(header), 'Time BND Weight'
        write(header, '(a,a20)') trim(header), 'nElems BND'
      end if
      write( Bal_logUnit(3),*) trim(header)

      ! In case of boundaries, level weights are calculated as above in addition
      ! to the specific boundary weights
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        output = ''
        if( scheme(1)%glob%levelDesc(iLevel)%nElems_fluid > 0) then
          params%level_weight(iLevel) = ( params%timer(iLevel)%value(params%   &
            &                 lw_tIntp(iLevel) )                               &
            &               + params%timer(iLevel)%value(params%               &
            &                 lw_tCompute(iLevel) ))                           & 
            &               / scheme( 1 )%glob%levelDesc(iLevel)%nElems_fluid
          write(output, '(a,i7)')     trim(output), iLevel
          write(output, '(a,i7)')     trim(output), geometry%tree%global%myPart+1
          write(output, '(a,i20)')    trim(output), geometry%tree%Part_First(  &
            &                                       geometry%tree%global%myPart+1)
          write(output, '(a,i20)')    trim(output), geometry%tree%Part_Last(  &
            &                                       geometry%tree%global%myPart+1)
          write(output, '(a,en20.2)') trim(output), params%level_weight(iLevel)
          write(output, '(a,en20.2)') trim(output),                            &
            &         ( params%timer(iLevel)%value(params%lw_tIntp(iLevel) )   &
            &         + params%timer(iLevel)%value(params%lw_tCompute(iLevel)))
          write(output, '(a,i20)')    trim(output), scheme( 1 )%               &
            &                   glob%levelDesc(iLevel)%nElems_fluid
        end if

        if( scheme(1)%glob%levelDesc(iLevel)%bc_elemBuffer%nVals > 0 ) then
            !Set weights according to BC time
            params%level_weight_bnd(iLevel) = params%level_weight(iLevel)      &
              &        + params%timer(iLevel)%value(params%lw_tSetBnd(iLevel)) &
              &        / scheme( 1 )%glob%levelDesc(iLevel)%bc_elemBuffer%nVals
          write(output, '(a,en20.2)') trim(output), params%level_weight_bnd(iLevel)
          write(output, '(a,en20.2)') trim(output),                            &
            &           params%timer(iLevel)%value(params%lw_tSetBnd(iLevel) )
          write(output, '(a,i20)')    trim(output), scheme( 1 )%               &
              &                       glob%levelDesc(iLevel)%bc_elemBuffer%nVals
        end if
        write( Bal_logUnit(3),*)    trim(output)
      end do
      call tem_horizontalSpacer(fUnit=Bal_logUnit(3))
    case default
      write(logUnit(1),*) ' Given load balancing not found in Musubi'//        &
        &             trim(params%general%balance%balanceKind )
    end select

    ! update the suffix
    params%general%balance%fileID = mod(params%general%balance%fileID,2)+1

    ! copy the current tree global info
    locGlobal = geometry%tree%global
    write(filename,'(a,i1,a)')                                                 &
      &           trim(params%general%balance%restart%controller%writePrefix), &
      &           params%general%balance%fileID,'_'

    pos = INDEX(trim(filename), pathSep, .true.)
    suffix = trim(filename(pos+1:))

    ! set the new filename and set the meshChange flag to true
    geometry%tree%global%dirname    = trim(filename)
    geometry%tree%global%meshChange = .true.

    ! Write a restart file
    write(logUnit(1),*) 'Writing restart for dynamic load balancing ...'
    call tem_time_dump( params%general%simControl%now, logUnit(1) )
    call mus_writeRestart( pdfGlobal = pdfGlobal,                              &
      &                    restart   = params%general%balance%restart,         &
      &                    scheme    = scheme,                                 &
      &                    tree      = geometry%tree,                          &
      &                    timing    = params%general%simControl%now,          &
      &                    timerData = params%general%timerData,               &
      &                    timerHandle = params%wRestartTimer,                 &
      &                    suffix    = suffix )
    write(logUnit(1),*) 'Done writing'

    ! Set the weights of each element according to levelwise weight obtained
    call mus_geom_setWeight ( params    = params,                              &
      &                       geometry  = geometry,                            &
      &                       balance   = params%general%balance,              &
      &                       glob      = scheme(1)%glob,                      &
      &                       pdfGlobal = pdfGlobal,                           &
      &                       field     = scheme(1)%field(1) )

    ! Call the sparta to get new splitting positions wrt new weights
    call tem_balanceSparta( weight = geometry%weight,                          &
      &                     myPart = params%general%proc%rank,                 &
      &                     nParts = params%general%proc%comm_size,            &
      &                     offset = geometry%offset,                          &
      &                     comm   = params%general%proc%comm )

    if( geometry%tree%global%nProperties > 0 )then
      write(logUnit(1),*) 'Writing boundary information for dynamic load '//   &
        &                 'balancing ...'
      ! Write the boundary information to disc in the same directory
      call dump_tem_BC_prop( me       = geometry%boundary,                     &
        &                    offset   = geometry%tree%Property(1)%Offset,      &
        &                    nElems   = geometry%tree%Property(1)%nElems,      &
        &                    basename = trim(filename)//'bnd',                 &
        &                    myPart   = params%general%proc%rank,              &
        &                    comm     = params%general%proc%comm )
      write(logUnit(1),*) 'Done writing'
    end if

    if( geometry%tree%global%nProperties > 1 )then
      write(logUnit(1),*) 'Writing qVal information for dynamic load '//       &
        &                 'balancing ...'
      ! dump the qVals to disc using the dynLB basename
      call dump_tem_BC_qVal( me       = geometry%boundary,                     &
        &                    offset   = geometry%tree%Property(2)%Offset,      &
        &                    nElems   = geometry%tree%Property(2)%nElems,      &
        &                    basename = trim(filename)//'qval',                &
        &                    mypart   = params%general%proc%rank,              &
        &                    comm     = params%general%proc%comm )
      write(logUnit(1),*) 'Done writing'
    end if

    ! copy back original tree global type
    geometry%tree%global = locGlobal

    ! Dump the weights using the current slot in the filename
    write(filename,'(a,i4.4,a)')                                               &
      &           trim(params%general%balance%restart%controller%writePrefix), &
      &           params%general%balance%iSlot,'_weights'
    call mus_geom_dumpWeights( tree     = geometry%tree,                       &
      &                        weight   = geometry%weight,                     &
      &                        filename = trim(filename) )

    ! Unload the tree
    call unload_tem( me = geometry%tree )

  end subroutine mus_balance
!******************************************************************************!

!******************************************************************************!
  !> Initialize musubi solverHead and print musubi banner to screen
  subroutine mus_banner( solver )
    !---------------------------------------------------------------------------
    !> solver definition
    type(tem_solveHead_type), intent(in) :: solver
    !---------------------------------------------------------------------------
    character(len=26) :: dat_string
    !---------------------------------------------------------------------------
    write(logUnit(1),*)""
    write(logUnit(1),*)" .___  ___.  __    __       _______. __    __  .______    __  "
    write(logUnit(1),*)" |   \/   | |  |  |  |     /       ||  |  |  | |   _  \  |  | "
    write(logUnit(1),*)" |  \  /  | |  |  |  |    |   (----`|  |  |  | |  |_)  | |  | "
    write(logUnit(1),*)" |  |\/|  | |  |  |  |     \   \    |  |  |  | |   _  <  |  | "
    write(logUnit(1),*)" |  |  |  | |  `--'  | .----)   |   |  `--'  | |  |_)  | |  | "
    write(logUnit(1),*)" |__|  |__|  \______/  |_______/     \______/  |______/  |_"  & 
      &              //trim(solver%version)
    write(logUnit(1),*)""
    write(logUnit(1),*)" (C) 2012 German Research School for Simulation Sciences"
    write(logUnit(1),*)""
    call tem_print_execInfo()
    write(logUnit(1),*)""
    dat_string = utc_date_string()
    write(logUnit(1),*)"Run at: "//trim(dat_string)
    write(logUnit(1),*)""

  end subroutine mus_banner
!******************************************************************************!


!******************************************************************************!
  !> Init auxiliary features such as interpolation boundaries, restart and 
  !! the tracker
  subroutine mus_init_aux( scheme, geometry, params, pdfGlobal, intp )
    !---------------------------------------------------------------------------
    !> containers for the different schemes
    type(mus_scheme_type), intent(inout)     :: scheme(:)
    !> geometry infomation
    type(mus_geom_type), intent(inout)       :: geometry
    !> global parameters
    type(mus_param_type),intent(inout)       :: params
    !> Interpolation information
    type(interpolation_type), intent(inout)  :: intp
    !> global flow quantity properties
    type(pdf_globInfo_type), intent(inout)   :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: iScheme, iField, iLevel, iTimeLayer, iTrack
    integer :: tLayerMin, tLayerMax
    !---------------------------------------------------------------------------

    ! verify correct settings for the streaming layout
    call check_streaming_layout( pdfGlobal )
    ! \todo dt is taken from the first scheme. There should be a global dt

    ! When the restart is read from separate restart table, we start the 
    ! simulation from the time step given in restart file. In the case,
    ! when restart is read from initial condition table, the simulation start
    ! time step is taken as the one defined in configuration file
    call mus_time_homogenize( me = params%general%simControl%now,              &
      &                       dt = params%physics%dtLvl(                       &
      &                            geometry%tree%global%maxLevel ),            &
      &              readRestart = params%general%restart%controller           &
      &                             %readRestart )
    call mus_timeControl_homogenize(                                           &
      &                       me = params%general%simControl%timeControl,      &
      &                       dt = params%physics%dtLvl(                       &
      &                            geometry%tree%global%maxLevel ),            &
      &                       reqInt = params%reqInterval )

    ! De-activate the check density for time step 1
    ! \todo KM: is this call necessary with new timeControl
    call tem_timeControl_update( me   = params%general%simControl%timeControl, &
      &                          now  = params%general%simControl%now,         &
      &                          comm = params%general%proc%comm )

    ! dynamic load balance time control homogenize
    if ( params%general%balance%dynamic ) then
      write(logUnit(1),*) 'Balance...'
      call mus_timeControl_homogenize(                                         &
        &     me = params%general%balance%timeControl,                         &
        &     dt = params%physics%dtLvl(                                       &
        &          geometry%tree%global%maxLevel ),                            &
        &     reqInt = params%reqInterval )
    endif

    if( ( params%general%restart%controller%writeRestart  .or.                 &
      &   params%general%restart%controller%readRestart ))then

      ! initialize the restart
      write(logUnit(1),*) 'Initializing restart...'
      call tem_init_restart( me      = params%general%restart,                 &
        &                    solver  = params%general%solver,                  &
        &                    nSys    = pdfGlobal%nSchemes,                     &
        &                    varSys  = scheme(:)%varSys,                       &
        &                    tree    = geometry%tree )
      call mus_timeControl_homogenize(                                         &
        &    me  = params%general%restart%controller%timeControl,              &
        &    dt  = params%physics%dtLvl( geometry%tree%global%maxLevel ),      &
        &                       reqInt = params%reqInterval )
    end if

    !\todo: no general read option available in restart read
    if(params%general%restart%controller%readRestart )then
      call mus_readRestart( pdfGlobal = pdfGlobal,                             &
        &                   restart   = params%general%restart,                &
        &                   scheme    = scheme,                                &
        &                   tree      = geometry%tree )
    end if !restart?

    ! allocate the tracking headers in the tem_balance_type with the number
    ! of schemes
    allocate( params%general%balance%trackHeaders( pdfGlobal%nSchemes ))

    ! scheme loop to initialize tracking, boundary, interpolation and output
    do iScheme = 1, pdfGlobal%nSchemes

      call mus_setSpatialOmega( scheme    = scheme( iScheme ),                 &
        &                       pdfGlobal = pdfGlobal,                         &
        &                       tree      = geometry%tree,                     &
        &                       params    = params )

      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel

        ! allocate fluid mrt array
        call mus_alloc_mrt( scheme( iScheme )%field(1)%fieldProp%fluid%mrt(iLevel)%s_mrt,&
          &                 scheme( iScheme )%layout%stencil(1)%QQ )
        ! set relaxation parameter for MRT
        call mus_setParameters( scheme( iScheme ), iLevel, params )

        ! add a levelwise timer for load balancing
        call tem_addTimer( me          = params%timer(iLevel),                 &
          &                timerHandle = params%lw_tIntp(iLevel),              &
          &                timerName   = 'tIntp' )

        call tem_addTimer( me          = params%timer(iLevel),                 &
          &                timerHandle = params%lw_tCompute(iLevel),           &
          &                timerName   = 'tCompute' )

        call tem_addTimer( me          = params%timer(iLevel),                 &
          &                timerHandle = params%lw_tSetBnd(iLevel),            &
          &                timerName   = 'tSetBnd' )
      end do

      ! initialize the balance tracking headers for a clean restart in case
      ! of dynamic load balancing
      call tem_balance_initTracks( me    = params%general%balance%             &
        &                                         trackHeaders(iScheme)%header,&
        &                          track = scheme(iScheme)%track )

      if( scheme( iScheme )%sourceActive ) then
        ! initialize the source terms for this field
        call mus_init_sources( me    = scheme(iScheme),                        &
          &                    tree  = geometry%tree )
      end if

      ! tracking objects
      write(logUnit(1),*) 'Initializing tracker...'

      do iTrack = 1, size( scheme( iScheme )%track )
        call mus_timeControl_homogenize(                                       &
          &        me = scheme( iScheme )%track( iTrack )%header%timeControl,  &
          &        dt = params%physics%dtLvl(geometry%tree%global%maxLevel ),  &
          &        reqInt = params%reqInterval )
      end do

      call tem_init_tracker( me           = scheme(iScheme)%track,             &
        &                    tCtrl        = scheme(iScheme)%trackingControl,   & 
        &                    levelPointer = pdfGlobal%levelPointer,            &
        &                    tree         = geometry%tree,                     &
        &                    proc         = params%general%proc,               &
        &                    solver       = params%general%solver,             &
        &                    varSys       = scheme( iScheme )%varSys)

      write(logUnit(1),*) 'Initializing interpolation...'

      if( pdfGlobal%minLevel /= pdfGlobal%maxlevel ) then
        ! initialize the interpolation
        call mus_init_interpolate(                                             &
          &                  intp      = intp,                                 &
          &                  glob      = scheme(iScheme)%glob,                 &
          &                  scheme    = scheme( iScheme ),                    &
          &                  fieldProp = scheme( iScheme )%field(:)%fieldProp, &
          &                  tree      = geometry%tree,                        &
          &                  layout    = scheme( iScheme )%layout,             &
          &                  pdfGlobal = pdfGlobal, params = params )
      end if
      if( main_debug%debugDependencies) then
          call debug_dependencies( pdfGlobal, intp, scheme( iScheme )%layout,  &
            &                      scheme( iScheme )%glob, geometry%tree,      &
            &                      params%general%proc )  
          call dump_intpLists( pdfGlobal, intp, scheme( iScheme )%layout,      &
            &                  scheme( iScheme )%glob, geometry%tree,          &
            &                  params%general%proc )
      end if

      ! Fill all elements (ghost, halo) with valid values from fluid elements
      ! for both time layers now and next
      tLayerMin = 1
      if( params%init_allTimeLayers ) then
        ! Initialize all time layers
        tLayerMax = 2
      else
        ! Initialize only nNext ( iTimeLayer = 1 )
        tLayerMax = 1
      end if

      do iTimeLayer = tLayerMax, tLayerMin, -1
        call fillHelperElementsFineToCoarse( scheme     = scheme(iScheme),     &
          &                                  geometry   = geometry,            &
          &                                  params     = params,              &
          &                                  iLevel     = pdfGlobal%minLevel,  &
          &                                  intp       = intp,                &
          &                                  pdfGlobal  = pdfGlobal,           &
          &                                  iTimeLayer = iTimeLayer  )
        call fillHelperElementsCoarseToFine( scheme     = scheme(iScheme),     &
          &                                  geometry   = geometry,            &
          &                                  params     = params,              &
          &                                  iLevel     = pdfGlobal%minLevel,  &
          &                                  intp       = intp,                &
          &                                  pdfGlobal  = pdfGlobal,           &
          &                                  iTimeLayer = iTimeLayer  )
      end do

      if( intp%testInterpolation ) then
        write(logUnit(1),*)'Testing  for the interpolation method.'
        do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
          call mus_testInterpolation( scheme     = scheme(iScheme),            &
            &                         tree       = geometry%tree,              &
            &                         params     = params,                     &
            &                         iLevel     = iLevel,                     &
            &                         intp       = intp,                       &
            &                         pdfGlobal  = pdfGlobal,                  &
            &                         iTimeLayer = iTimeLayer  )
        end do
      end if
      ! And set back the time layer to standard to interpolate the correct
      ! time layer in the interpolation module
      intp%timeLayer = 1


      if ( scheme(iScheme)%trackingControl%active ) then
        call mus_tracker( pdf       = scheme(iScheme)%pdf,                     &
          &               track     = scheme(iScheme)%track,                   &
          &               params    = params,                                  &
          &               layout    = scheme(iScheme)%layout,                  &
          &               glob_tree = geometry%tree,                           &
          &               globSys   = scheme(iScheme)%varSys )
      end if

      ! initialize each field boundary by looping over the field inside
      ! init_boundary_field
      write(logUnit(1),*) 'Initializing boundary conditions...'
      call init_boundary( field        = scheme(iScheme)%field,       &
        &                 pdf          = scheme(iScheme)%pdf,         &
        &                 tree         = geometry%tree,               &
        &                 glob         = scheme(iScheme)%glob,        &
        &                 pdfGlobal    = pdfGlobal,                   &
        &                 layout       = scheme(iScheme)%layout,      &
        &                 schemeHeader = scheme(iScheme)%header,      & 
        &                 varSys       = scheme(iScheme)%varSys,      &
        &                 derVarPos    = scheme( iScheme)%derVarPos,  &
        &                 globBC       = scheme( iScheme )%globBC,    &
        &                 bc_prop      = geometry%boundary            )
      write(logUnit(1),*) 'Done initializing boundary conditions.'

    end do !scheme loop


    if( geometry%dynamicGeom ) then
      write(logUnit(1),*) 'Initializing geomIncr ...'
      call mus_init_geomIncr( geometry  = geometry,                            &
        &                     scheme    = scheme,                              &
        &                     physics   = params%physics,                      &
        &                     params    = params )
      write(logUnit(1),*) 'Done initializing geomIncr.'
    end if ! Time for geom incr

    ! restart at initial time step. Useful to check whether
    ! intialization is done properly.
    if( params%general%restart%controller%writeRestart ) then
      ! check time control and update it if triggered
      call tem_timeControl_check(                                              &
        &               me     = params%general%restart%controller%timeControl,&
        &               now    = params%general%simControl%now,                &
        &               comm   = params%general%proc%comm,                     &
        &               triggered = params%restart_triggered )

      if ( params%restart_triggered ) then
        write(logUnit(1),*) ' Writing restart at:'
        call tem_time_dump( params%general%simControl%now, logUnit(1) )
        call mus_writeRestart( pdfGlobal = pdfGlobal,                          &
          &                    restart   = params%general%restart,             &
          &                    scheme    = scheme(:),                          &
          &                    tree      = geometry%tree,                      &
          &                    timing    = params%general%simControl%now,      &
          &                    timerData = params%general%timerData,           &
          &                    timerHandle = params%wRestartTimer              )
        write(logUnit(1),*) 'Done writing'
      end if
    end if

    write(logUnit(1),*) "Starting Musubi MAIN loop with time control:"
    call tem_timeControl_dump(params%general%simControl%timeControl,logUnit(1))

    call tem_horizontalSpacer( after = 1, fUnit = logUnit(1) )

  end subroutine mus_init_aux
!******************************************************************************!


!******************************************************************************!
  !> Recursively fill all the helper elements (i.e. ghost, halo) with valid
  !! information from the fluid elements.
  !! 
  !! This step is required before each run of the simulation. It would be
  !! possible to fill also the helper elements with the initial conditions.
  !! However, we are only able to fill the fluid elements with valid data
  !! (restart files have no information about the helper elements)
  !!
  recursive subroutine fillHelperElementsFineToCoarse( scheme, geometry,       &
    &                       params, iLevel, intp, pdfGlobal, iTimeLayer  )
    !---------------------------------------------------------------------------
    !> containers for the scheme
    type(mus_scheme_type) :: scheme
    !> geometry infomation
    type(mus_geom_type)   :: geometry
    !> global parameters
    type(mus_param_type)      :: params  
    !> Interpolation information
    type(interpolation_type) :: intp      
    !> global flow quantity properties
    type(pdf_globInfo_type) :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !> which time layer to fill
    integer, intent(in) :: iTimeLayer
    !---------------------------------------------------------------------------
    integer :: timeLayer ! which time layer to fill
    !---------------------------------------------------------------------------

    if( iTimeLayer == 1 ) then
      timeLayer = scheme%pdf( iLevel )%nNext
    else
      timeLayer = scheme%pdf( iLevel )%nNow 
    end if

    if( iLevel < pdfGlobal%maxLevel ) then
      call fillHelperElementsFineToCoarse( scheme, geometry, params,           &
        &                             iLevel+1, intp, pdfGlobal, iTimeLayer )  
      !\todo loop over field, pass varPos, nScalars
      intp%timeLayer = iTimeLayer
      call mus_exchange(                                                       &
        &     recv    = scheme%glob%levelDesc( iLevel )%recvbufferFromFiner,   &
        &     send    = scheme%glob%levelDesc( iLevel )%sendbufferFromFiner,   &
        &     state   = scheme%pdf( iLevel )%state( :, timeLayer ),            &
        &     pattern = params%general%commPattern,                            &
        &     level   = iLevel,                                                &
        &     proc    = params%general%proc )
      call intp%fillMineFromFiner_low(                                         &
        &      pdf       = scheme%pdf,                                         &
        &      glob      = scheme%glob,                                        &
        &      scheme    = scheme,                                             &
        &      fieldProp = scheme%field(:)%fieldProp,                          &
        &      level     = iLevel,                                             &
        &      layout    = scheme%layout,                                      &
        &      intp      = intp,                                               &
        &      ind       = scheme%glob%levelDesc( iLevel)%intpFromFiner_low,   &
        &      varSys    = scheme%varSys,                                      &
        &      physics   = params%physics,                                     &
        &      derVarPos = scheme%derVarPos )

      call intp%fillMineFromFiner(                                             &
        &      pdf       = scheme%pdf,                                         &
        &      glob      = scheme%glob,                                        &
        &      scheme    = scheme,                                             &
        &      fieldProp = scheme%field(:)%fieldProp,                          &
        &      level     = iLevel,                                             &
        &      layout    = scheme%layout,                                      &
        &      intp      = intp,                                               &
        &      ind       = scheme%glob%levelDesc(iLevel)%intpFromFiner_high,   &
        &      varSys    = scheme%varSys,                                      &
        &      physics   = params%physics,                                     &
        &      derVarPos = scheme%derVarPos  )
      call mus_exchange(                                                       &
        &     recv    = scheme%glob%levelDesc( iLevel )%recvbufferFromFiner,   &
        &     send    = scheme%glob%levelDesc( iLevel )%sendbufferFromFiner,   &
        &     state   = scheme%pdf( iLevel )%state( :, timeLayer ),            &
        &     pattern = params%general%commPattern,                            &
        &     level   = iLevel,                                                &
        &     proc    = params%general%proc )
    end if

    call mus_exchange( recv    = scheme%glob%levelDesc( iLevel )%recvbuffer,   &
      &                send    = scheme%glob%levelDesc( iLevel )%sendbuffer,   &
      &                state   = scheme%pdf( iLevel )%state( :, timeLayer ),   &
      &                pattern = params%general%commPattern,                   &
      &                level   = iLevel,                                       &
      &                proc    = params%general%proc )                                
                                                                              
  end subroutine fillHelperElementsFineToCoarse
!******************************************************************************!


!******************************************************************************!
  !> Recursively fill all the helper elements (i.e. ghost, halo) with valid
  !! information from the fluid elements.
  !! 
  !! This step is required before each run of the simulation. It would be
  !! possible to fill also the helper elements with the initial conditions.
  !! However, we are only able to fill the fluid elements with valid data 
  !! (restart files have no information about the helper elements)
  !!
  recursive subroutine fillHelperElementsCoarseToFine( scheme, geometry,       &
    &                       params, iLevel, intp, pdfGlobal, iTimeLayer  )
    !---------------------------------------------------------------------------
    !> containers for the scheme
    type(mus_scheme_type)     :: scheme
    !> geometry infomation
    type(mus_geom_type)       :: geometry
    !> global parameters
    type(mus_param_type)      :: params  
    !> Interpolation information
    type(interpolation_type)  :: intp      
    !> global flow quantity properties
    type(pdf_globInfo_type)   :: pdfGlobal
    !> Level counter variable
    integer, intent(in)       :: iLevel
    !> which time layer to fill
    integer, intent(in)       :: iTimeLayer
    !---------------------------------------------------------------------------
    integer :: timeLayer ! which time layer to fill
    !---------------------------------------------------------------------------

    if( iTimeLayer == 1 ) then
      timeLayer = scheme%pdf( iLevel )%nNext
    else
      timeLayer = scheme%pdf( iLevel )%nNow 
    end if

    if( iLevel > pdfGlobal%minLevel ) then
      call mus_exchange(                                                       &
        &     recv    = scheme%glob%levelDesc( iLevel )%recvbufferFromCoarser, &
        &     send    = scheme%glob%levelDesc( iLevel )%sendbufferFromCoarser, &
        &     state   = scheme%pdf( iLevel )%state( :, timeLayer ),            &
        &     pattern = params%general%commPattern,                            &
        &     level   = iLevel,                                                &
        &     proc    = params%general%proc )                                
    end if

    if( iLevel < pdfGlobal%maxLevel ) then
      intp%timeLayer = iTimeLayer
      call intp%fillFinerFromMe_low(                                           &
        &    pdf       = scheme%pdf,                                           &
        &    glob      = scheme%glob,                                          &
        &    scheme    = scheme,                                               &
        &    fieldProp = scheme%field(:)%fieldProp,                            &
        &    level     = iLevel,                                               &
        &    intp      = intp,                                                 &
        &    ind       = scheme%glob%levelDesc( iLevel+1)%intpFromCoarser_low, &
        &    layout    = scheme%layout,                                        &
        &    varSys    = scheme%varSys,                                        &
        &    physics   = params%physics,                                       &
        &    derVarPos = scheme%derVarPos )

      call intp%fillFinerFromMe(                                               &
        &   pdf       = scheme%pdf,                                            &
        &   glob      = scheme%glob,                                           &
        &   scheme    = scheme,                                                &
        &   fieldProp = scheme%field(:)%fieldProp,                             &
        &   level     = iLevel,                                                &
        &   intp      = intp,                                                  &
        &   ind       = scheme%glob%levelDesc( iLevel+1)%intpFromCoarser_high, &
        &   layout    = scheme%layout,                                         &
        &   varSys    = scheme%varSys,                                         &
        &   physics   = params%physics,                                        &
        &   derVarPos = scheme%derVarPos  )

      call fillHelperElementsCoarseToFine( scheme, geometry, params,           &
        &                               iLevel+1, intp, pdfGlobal, iTimeLayer )

    end if

  end subroutine fillHelperElementsCoarseToFine
!******************************************************************************!



!******************************************************************************!
  !> Set relaxation parameters for MRT
  !!
  subroutine mus_setParameters( scheme, iLevel, params )
    !---------------------------------------------------------------------------
    !> scheme type
    type(mus_scheme_type)               :: scheme
    !> level
    integer, intent(in)                 :: iLevel
    !> global parameters
    type(mus_param_type),intent(in)     :: params
    !---------------------------------------------------------------------------

    select case(trim(scheme%header%kind))
    case('lbm', 'lbm_incomp')
      ! \todo KM: must set omega at each time step only if transient is
      ! active
      ! call setParameters_lbm( scheme = scheme, iField = 1, &
      !   &                     iLevel = iLevel, &
      !   &                     params = params )
      call mus_update_relaxation( me = scheme%field(1)%fieldProp%fluid, &
        &                         general = params%general, &
        &                         relaxation = scheme%header%relaxation, &
        &                         layout = scheme%header%layout, &
        &                         level = iLevel )
    case('multi-species_gas','multi-species_liquid')
      call setParameters_multispecies( scheme = scheme, &
        &                              iLevel = iLevel, params = params )
    end select

  end subroutine mus_setParameters
!******************************************************************************!

!******************************************************************************!
  !> Set parameters for multispecies
  subroutine setParameters_multispecies( scheme, iLevel, params )
    !--------------------------------------------------------------------------
    type(mus_scheme_type)               :: scheme    !< Scheme type
    integer, intent(in)                 :: iLevel    !<  level
    !> global parameters
    type(mus_param_type),intent(in)     :: params
    !--------------------------------------------------------------------------
    real(kind=rk) :: omega_diff ! diffusive relaxation parameter
    real(kind=rk) :: omega_kine ! kinematic shear relaxation parameter
    real(kind=rk) :: omega_bulk ! bulk relaxation parameter
    real(kind=rk) :: fac ! ramping multiplication factor
    integer :: iField ! field counter
    real(kind=rk), dimension( scheme%layout%stencil(1)%QQ,                     &
      & scheme%layout%stencil(1)%QQ ) :: identity, tmpMatrix, tmpMatrixInv
    integer :: iDir
    !--------------------------------------------------------------------------
!write(dbgUnit,*) 'setting omega paramter for multi-species'
    ! get the transient multiplication factor
    fac = tem_transient_for( transient = scheme%mixture%omega_ramping,         &
      &                      time      = params%general%simControl%now,        &
      &                      conf      = params%general%solver%conf(1) )
    !! Relaxation parameter for each level
    !! kine_viscosity = dt*c^2*(1/omega - 0.5)
    !!    =>   omega = dt / (viscosity/c^2 + dt/2)
    !! omega for each level is stored at fluid%omLvl
    omega_diff  =  fac * scheme%mixture%relaxLvl( iLevel )%paramB * cs2
!write(dbgUnit,*) 'omega_diff ', omega_diff
    scheme%mixture%relaxLvl( iLevel )%omega_diff = omega_diff


    ! if relaxation is mrt set kinematic and bulk viscosity omega
    if( scheme%header%relaxation(1:3) == 'mrt' ) then
      do iField = 1, scheme%nFields
        if( .not. allocated( scheme%field( iField )%fieldProp%species          &
          &          %mrt( iLevel )%s_mrt))                                    &
          & allocate( scheme%field( iField )%fieldProp%species                 &
          &          %mrt( iLevel )%s_mrt( scheme%layout%stencil(1)%QQ,        &
          &                                scheme%layout%stencil(1)%QQ))

        if( .not. allocated( scheme%field( iField )%fieldProp%species          &
          &          %mrt( iLevel )%omegaMoments))                             &
          & allocate( scheme%field( iField )%fieldProp%species                 &
          &          %mrt( iLevel )%omegaMoments( scheme%layout%stencil(1)%QQ, &
          &                                       scheme%layout%stencil(1)%QQ))
        scheme%field( iField )%fieldProp%species%mrt( iLevel )                 &
          & %omegaMoments = 0.0_rk

        tmpMatrix    = 0.0_rk
        tmpMatrixInv = 0.0_rk
        Identity     = 0.0_rk
        do iDir = 1, scheme%layout%stencil(1)%QQ
          Identity(iDir, iDir) = 1.0_rk
        end do

        ! viscous omega
        omega_kine = scheme%mixture%relaxLvl( iLevel )%omega_kine * fac
        ! bulk omega
        omega_bulk = scheme%field(iField)%fieldProp%species%omBulkLvl(iLevel)  &
          &        * fac
!write(*,*) 'omega_diff ', omega_diff 
!write(*,*) 'omega_kine ', omega_kine        
!write(*,*) 'omega_bulk ', omega_bulk    
        ! Set the relaxation parameters (only the non-relaxing modes)
        ! \ref Multiple-relaxation-time lattice Boltzmann scheme for 
        !      homogeneous mixture flows with external force- Pietro Asinari
        ! \ref Jens ICCMES paper
        ! initialize all entries in relaxation matrix
        scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt(:,:) = 0.0_rk
        select case(trim(scheme%header%layout))
        case('d2q9' )
!write(*,*) 'omega_diaCenter ', (omega_bulk + omega_kine )/2.0_rk
!write(*,*) 'omega_nondia ', (omega_bulk - omega_kine )/2.0_rk
          ! diffusivity omegas
          do iDir=2,3
            scheme%field( iField )%fieldProp%species%mrt( iLevel )             &
              & %s_mrt( iDir, iDir ) = omega_diff
          end do    

          do iDir=4,5
            scheme%field( iField )%fieldProp%species%mrt( iLevel )             &
              & %s_mrt( iDir, iDir ) = ( omega_bulk + omega_kine ) * div1_2
          end do

          scheme%field( iField )%fieldProp%species%mrt( iLevel )               &
              & %s_mrt( 5, 4 ) = ( omega_bulk - omega_kine ) * div1_2

          scheme%field( iField )%fieldProp%species%mrt( iLevel )               &
              & %s_mrt( 4, 5 ) = ( omega_bulk - omega_kine ) * div1_2

          scheme%field( iField )%fieldProp%species%mrt( iLevel )               &
            & %s_mrt( 6, 6 ) = omega_kine

          do iDir=7,9
            scheme%field( iField )%fieldProp%species%mrt( iLevel )             &
              & %s_mrt( iDir, iDir ) = scheme%mixture%omega_hom
          end do

        case('d3q19')
!write(*,*) 'omega_diaCenter ', (omega_bulk + 2.0_rk * omega_kine )/3.0_rk
!write(*,*) 'omega_nondia ', (omega_bulk - omega_kine )/3.0_rk
          do iDir=2,4
            scheme%field( iField )%fieldProp%species%mrt( iLevel )             &
              & %s_mrt( iDir, iDir ) = omega_diff
          end do

          do iDir=5,7
            scheme%field( iField )%fieldProp%species%mrt( iLevel )             &
              & %s_mrt( iDir, iDir ) = ( omega_bulk + 2.0_rk * omega_kine ) / 3.0_rk
          end do

          scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt( 5, 6 ) &
            & = ( omega_bulk - omega_kine ) / 3.0_rk
          scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt( 5, 7 ) &
            & = ( omega_bulk - omega_kine ) / 3.0_rk

          scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt( 6, 5 ) &
            & = ( omega_bulk - omega_kine ) / 3.0_rk
          scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt( 6, 7 ) &
            & = ( omega_bulk - omega_kine ) / 3.0_rk

          scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt( 7, 5 ) &
            & = ( omega_bulk - omega_kine ) / 3.0_rk
          scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt( 7, 6 ) &
            & = ( omega_bulk - omega_kine ) / 3.0_rk

          do iDir=8,10 
            scheme%field( iField )%fieldProp%species%mrt( iLevel )             &
              & %s_mrt( iDir, iDir ) = omega_kine
          end do

          do iDir=11,19
            scheme%field( iField )%fieldProp%species%mrt( iLevel )             &
              & %s_mrt( iDir, iDir ) = scheme%mixture%omega_hom
          end do
        end select

        !omega moments = (M^-1 RelaxMat M)*(I+(M^-1 RelMat M)/2)^-1
        tmpMatrix = matmul( matmul( scheme%layout%moment%toPdf%A,              &
          & scheme%field( iField )%fieldProp%species%mrt( iLevel )%s_mrt ),    &
          & scheme%layout%moment%toMoments%A )

        tmpMatrixInv = invert_matrix( Identity + tmpMatrix * 0.5_rk )  

        scheme%field( iField )%fieldProp%species%mrt( iLevel )%omegaMoments =  &
          & matmul(tmpMatrix, tmpMatrixInv)

      end do ! iField
    end if ! mrt?
!stop
  end subroutine setParameters_multispecies
!******************************************************************************!


!******************************************************************************!
  !> Update omega by transient factor
  !! This routine is called every iteration to update omega
  !! \todo: this routine is replace by mus_update_relaxation in mus_fluid_module
  !!        It can be removed after some time.
  !! last update: Sep 4. 2014
  !!
  subroutine setParameters_lbm( scheme, iField, iLevel, params )
    !--------------------------------------------------------------------------
    type(mus_scheme_type)           :: scheme    !< Scheme type
    integer, intent(in)             :: iField    !< field counter
    integer, intent(in)             :: iLevel    !< level
    type(mus_param_type),intent(in) :: params    !< global parameters
    !--------------------------------------------------------------------------
    real(kind=rk) :: fac   ! ramping multiplication factor
    real(kind=rk) :: omega ! the shear relaxation parameter
    !--------------------------------------------------------------------------

    ! get the transient multiplication factor
    fac = tem_transient_for( transient = scheme%field( iField )%fieldprop      &
      &                                                  %fluid%omega_ramping, &
      &                      time = params%general%simControl%now,             &
      &                      conf = params%general%solver%conf(1) )

    ! calculate omega from viscostiy
    omega = 1._rk  /                                                    &
     &  ( cs2inv*scheme%field( iField )%fieldProp%fluid%kine_viscosityLB + 0.5_rk)
    scheme%field(iField )%fieldProp%fluid%omLvl( iLevel ) = omega * fac

    if( scheme%header%relaxation(1:3) == 'mrt' ) then
      if( .not. allocated(scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt))&
        & allocate( scheme%field(iField)%fieldProp%fluid              &
        &          %mrt( iLevel )%s_mrt( scheme%layout%stencil(1)%QQ ))

      select case ( trim(scheme%header%relaxation) )
      case ( 'mrt_bgk' )
        ! set all relaxation parameters to the same omega value
        scheme%field( iField )%fieldProp%fluid%mrt( iLevel )%s_mrt = omega * fac

      case ( 'mrt' )
        ! \todo From which paper these relaxation values are taken?
        select case(trim(scheme%header%layout))
        case('d2q9' )
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(:)   = 0._rk
          ! set by bulk viscosity
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(2)   =  &
            & fac*scheme%field(iField)%fieldProp%fluid%omLvl_bulk(iLevel)
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(3)   = 1.14_rk
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(5)   = 1.92_rk
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(7)   = 1.92_rk
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(8:9) = omega
        case('d3q19')
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(:)     = 1._rk
          ! set relaxation by bulk viscostiy
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(2)     =  &
            &       fac*scheme%field(iField)%fieldProp%fluid%omLvl_bulk(iLevel)
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(10)    = omega
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(12)    = omega
          scheme%field(iField)%fieldProp%fluid%mrt(iLevel)%s_mrt(14:16) = omega
        end select ! d2q9 or d3q19?
      end select ! mrt or mrt_bgk?
    end if ! mrt?

  end subroutine setParameters_lbm
!******************************************************************************!


!******************************************************************************!
  !> This routine sets the spatial value for the omega for 
  !! spatial omega definition
  !!
  !! \verbatim 
  !! -- Definition of the spatial function for how to vary omega 
  !! -- in the simulation domain
  !! function spatialFunction(x,y,z)
  !!    if(  x > xmax) then
  !!      res =w_max
  !!    elseif( x < xmin ) then
  !!      res =  w_min
  !!    else
  !!      slope = (w_max-w_min)/(cutoff_min-cutoff_max)
  !!      res =  r*slope + w_max - cutoff_min*slope
  !!    end
  !!    return res
  !! end
  !! 
  !! fluid = { omega_spatial = {predefined = 'lua_fun',
  !!                            omega_spatial = spatialFunction, store = true }
  !! identify = { relaxation = 'bgk_omega_spatial'} 
  !! \endverbatim
  !!
  subroutine mus_setSpatialOmega( scheme, tree, pdfGlobal, params )
    !--------------------------------------------------------------------------
    !> scheme to work on
    type( mus_scheme_type ), intent(inout) :: scheme
    !> global treelm mesh 
    type( treelmesh_type ), intent(in)     ::tree
    !> pdf global info type
    type( pdf_globInfo_type )              :: pdfGlobal
    !> global parameters
    type(mus_param_type),intent(in)        :: params
    !---------------------------------------------------------------------------
    integer :: iLevel, nElems, iField, iVel
    real(kind=rk) :: nu_LB
    !---------------------------------------------------------------------------

    select case(trim(scheme%header%kind))
    case('lbm', 'lbm_incomp')
      iField = 1
      if( scheme%field(iField)%fieldprop%fluid%active ) then
        if( scheme%field(iField)%fieldProp%fluid%viscSpatial%toStore ) then
          write(logUnit(1),*)' Storing spatial omega ...'
          ! loop over each refinement iLevels
          do iLevel  = pdfGlobal%minLevel, pdfGlobal%maxLevel
            nu_LB   = scheme%field(iField)%fieldprop%fluid%kine_viscosityLB
            nElems = scheme%pdf( iLevel )%nElems_solve
            allocate( scheme%field(iField)%fieldprop%fluid%viscSpatial%        &
              &                               stored( iLevel )%val( nElems ))  

            scheme%field(iField)%fieldprop%fluid%viscSpatial%                  &
              &                            stored( iLevel )%val( : ) = 0.0_rk  
            ! get spatial term for all elements in iLevel
            ! in terms of viscosity
            scheme%field(iField)%fieldprop%fluid%viscSpatial%                  &
              &                      stored( iLevel )%val = tem_spatial_for(   &
              & me      = scheme%field(iField)%fieldprop%fluid%viscSpatial,    &
              & treeIds = scheme%glob%levelDesc( iLevel )%total( 1:nElems ),   &
              & tree    = tree,                                                &
              & n       = nElems,                                              &
              & conf     = params%general%solver%conf(1) ) 
            ! and now translate the viscosity in physical units to lattice
            ! omega multiplication factor
            ! The sponge factor gives the multiplication factor of the
            ! physical viscosity
            ! The factor store here in viscSpatial is given as a
            ! multiplication to the original omega (which might be ramped as
            ! well in time -> multiply). However, the relationship between
            ! viscosity and omega is non-linear with nu = 1/3(1/omega - 1/2)
            ! omega_sponged = factor * omega
            !  -> factor = omega_sponged / omega
            ! omega = 1/(3*nu_phys*dt/dx^2 + 1/2 )
            ! or with lattice viscosity
            ! omega = 1/(3*nu_LB + 1/2)
            ! the factor for converting the physical viscosity to lattice
            ! viscosity is collected in q = dt/dx^2
            scheme%field(iField)%fieldprop%fluid%viscSpatial%                  &
              &                                       stored( iLevel )%val     &
              &  = (3._rk*nu_LB + 0.5_rk) / (3._rk*nu_LB                       &
              &  * scheme%field(iField)%fieldprop%fluid%viscSpatial%           &
              &                                       stored( iLevel )%val     &
              &  + 0.5_rk )
          end do
        endif
        do iVel = 1,3
          if( scheme%field(iField)%fieldProp%fluid%                            &
            &                               velSpatial(iVel)%toStore ) then
            write(logUnit(1),*)' Storing velocity field...'
            ! loop over each refinement iLevels
            do iLevel  = pdfGlobal%minLevel, pdfGlobal%maxLevel
              nElems = scheme%pdf( iLevel )%nElems_solve
              allocate( scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)  &
                &             %stored( iLevel )%val( nElems ))
              scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)            &
                &                 %stored( iLevel )%val( : ) = 0.0_rk
              ! get spatial term all elements in iLevel
              scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)            &
                &                  %stored( iLevel )%val = tem_spatial_for(    &
                & me      = scheme%field(iField)%ic%ini_state(1+iVel),         &
                & treeIds = scheme%glob%levelDesc(iLevel)%total( 1:nElems ),   &
                & tree    = tree,                                              &
                & n       = nElems,                                            &
                & conf    = params%general%solver%conf(1) ) 
              scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)%           &
                &                                     stored( iLevel )%val     &
                & = scheme%field(iField)%fieldprop%fluid%velSpatial(iVel)%     &
                &                                     stored( iLevel )%val     &
                & / params%physics%fac( iLevel )%vel
            end do
          endif
        end do
      endif
    end select

  end subroutine mus_setSpatialOmega
!******************************************************************************!

!******************************************************************************!
  !> Compute external force and add it to outstate
  !! \todo KM: Generalize sources and use source to add external forces to state
  subroutine mus_addExternalForceToState( iScheme, scheme, pdfGlobal, tree,    &
    &                                     iLevel, params )
    !--------------------------------------------------------------------------
    !> current scheme
    integer, intent(in)                    :: iScheme
    !> scheme to work on
    type( mus_scheme_type ), intent(inout) :: scheme(:)
    !> pdf global info type
    type( pdf_globInfo_type )              :: pdfGlobal
    !> global treelm mesh
    type( treelmesh_type ), intent(in)     ::tree
    !> global parameters
    type(mus_param_type),intent(in)        :: params
    !> current level
    integer, intent(in)                    :: iLevel
    !--------------------------------------------------------------------------

    !--------------------------------------------------------------------------
    if (associated( scheme( iScheme)%compute_force ) ) then
      call scheme( iScheme)%compute_force(                                     &
        &           iScheme   = iScheme,                                       &
        &           fieldProp = scheme( iScheme )%field(:)%fieldProp,          &
        &           inState   = scheme( iScheme )%pdf(iLevel)%state(:,         &
        &                               scheme( iScheme )%pdf(iLevel)%nNow),   &
        &           outState  = scheme( iScheme )%pdf(iLevel)%state(:,         &
        &                              scheme( iScheme )%pdf(iLevel)%nNext),   &
        &           neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),        &
        &           nStart    = 1,                                             &
        &           nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,    &
        &           level     = iLevel,                                        &
        &           layout    = scheme( iScheme )%layout,                      &
        &           pdfGlobal = pdfGlobal,                                     &
        &           params    = params,                                        &
        &           scheme    = scheme,                                        &
        &           tree      = tree,                                          &
        &           varSys    = scheme( iScheme )%varSys )
    endif

  end subroutine mus_addExternalForceToState
!******************************************************************************!

end module mus_aux_module
!******************************************************************************!
