! See copyright notice in the COPYRIGHT file.
!*****************************************************************************!
!> In this module, all parameter files are read in 
!! as lua script or a sample configuration is being loaded
!!
!! Possible Parameter configuration
!! - \ref General Parameters
!! - \ref mus_scheme_layout_module "Scheme Definitions"
!! - \ref tem_debug_module "Debug Parameters"
!!
module mus_config_module

  ! include musubi modules
  use mpi
  use mus_param_module,              only: mus_param_type, mus_load_param
  use mus_geom_module,               only: mus_geom_type, mus_geom_setWeight,  &
    &                                      mus_geom_cleanWeight,               &
    &                                      mus_geom_setLevelWeight
  use mus_interpolate_header_module, only: interpolation_type,                 &
    &                                      mus_load_interpolate
  use mus_pdf_module,                only: pdf_globInfo_type, &
    &                                      mus_globInfo_set_levels, &
    &                                      mus_globInfo_set_dxdt
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_scheme_module,             only: mus_load_schemes
  use mus_geomIncrHead_module,       only: mus_geomIncrHead_load
  use mus_physics_module,            only: mus_create_funcStr, mus_load_physics

  ! include treelm modules
  use env_module,               only: pathLen
  use treelmesh_module,         only: load_tem, unload_tem
  use tem_aux_module,           only: tem_abort, tem_open_distconf_array
  use tem_property_module,      only: prp_hasBnd, prp_hasQVal
  use tem_balance_module,       only: tem_balance_loadConfig
  use tem_balanceSparta_module, only: tem_balanceSparta
  use tem_restart_module,       only: tem_restart_readHeader
  use tem_bc_prop_module,       only: load_tem_BC_prop,                        &
    &                                 tem_empty_BC_prop, load_tem_bc_qVal
  use tem_restart_module,       only: tem_load_restart
  use tem_time_module,          only: tem_time_isDefined
  use tem_timer_module,         only: tem_addTimer, tem_startTimer,            &
    &                                 tem_stopTimer
  use tem_tools_module,         only: tem_horizontalSpacer
  use tem_geometry_module,      only: tem_setEffBoundingBox
  use tem_general_module,       only: tem_load_general
  use tem_logging_module,       only: logUnit
  use tem_comm_env_module,      only: tem_comm_env_type
  use tem_timeControl_module,   only: tem_timeControl_start_at_sim

  ! include aotus modules
  use aot_path_module, only: aot_path_type, assignment(=), aot_init_path
  use aotus_module,    only: flu_State, open_config_chunk

  implicit none

  private

  public :: mus_load_config
  public :: mus_open_config
  public :: mus_load_bc_data

contains

!******************************************************************************!
  !> Read in LUA parameter file 
  !! See http://www.lua.org for a reference on how to use
  !! Lua is a scripting language in itself which allows 
  !! more complex parameter files including comments 
  !! And load / create the mesh depending on the configuration
  subroutine mus_load_config( scheme, geometry, params, pdfGlobal, intp )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ),    allocatable   :: scheme(:)
    !> Treelmesh data
    type( mus_geom_type ),      intent(inout) :: geometry
    !> Global parameters
    type( mus_param_type ),     intent(inout) :: params
    !> global information
    type( pdf_globInfo_type ),  intent(out)   :: pdfGlobal
    !> interpolation method
    type( interpolation_type ), intent(out)   :: intp
    !---------------------------------------------------------------------------
    character(len=PathLen) :: filename
    integer                :: iError
    type( aot_path_type )  :: init_path
    integer                :: total_nElems
    !---------------------------------------------------------------------------

    call tem_startTimer( timerHandle = params%loadMeshTimer,    &
      &                  me          = params%general%timerData )

    ! check whether params%general%solver%configFile is defined. When loading 
    ! musubi from apes, musubi config file is provided via params
    if ( trim(params%general%solver%configFile) == '' ) then
      ! Get filename from command line argument
      call get_command_argument( 1,filename )
      if( trim(filename) == '' ) then
        filename = 'musubi.lua'
      end if

      params%general%solver%configFile = filename
    else
      filename = params%general%solver%configFile
    endif

    write(logUnit(1),*) "Loading configuration file: "//trim( filename )
    call tem_horizontalSpacer(fUnit = logUnit(1))

    ! open musubi config file and solver specific lua functions as chunk
    call mus_open_config( conf     = params%general%solver%conf,             &
      &                   filename = filename,                               &
      &                   proc     = params%general%proc )

    ! Set aotus initial path 
    call aot_init_path( me = init_path, Filename = filename )

    ! load general information
    call tem_load_general( me   = params%general,                            &
      &                    conf = params%general%solver%conf(1))

    ! load global musubi params
    call mus_load_param( params = params,                                    &
      &                  conf   = params%general%solver%conf(1))

    ! Get the load balancing information
    call tem_balance_loadConfig( conf      = params%general%solver%conf(1),  &
      &                          me        = params%general%balance,         &
      &                          rank      = params%general%proc%rank,       &
      &                          defaultActive = .false. )

    !-------------------------------------------------------------------------
    !  Load mesh                                                             !
    !-------------------------------------------------------------------------
    ! Set the initial weights
    call mus_geom_setLevelWeight( params      = params,                      &
      &                           geometry    = geometry,                    &
      &                           balance     = params%general%balance,      &
      &                           levelWeight = geometry%levelWeight )

    ! First check, if we are starting from a restart
    ! KJ: Now the restart is read from initial conditions, which are loaded
    ! much later than the config in the process flow
    call tem_load_restart( me             = params%general%restart,          &
      &                    conf           = params%general%solver%conf(1),   &
      &                    tree           = geometry%tree,                   &
      &                    timing         = params%general%simControl%now,   &
      &                    proc           = params%general%proc )

    if( .not. params%general%restart%controller%readRestart ) then
      write(logUnit(1),*)'No read restart given. Loading mesh file'
      ! load the tree from the mesh = '' definition in case no restartRead is 
      ! given in this case the mesh from the restart header is read (in 
      ! tem_load_restart)
      call load_tem( me          = geometry%tree,                            &
        &            conf        = params%general%solver%conf(1),            &
        &            myPart      = params%general%proc%rank,                 &
        &            levelWeight = geometry%levelWeight,                     &
        &            nParts      = params%general%proc%comm_size,            &
        &            comm        = params%general%proc%comm,                 &
        &            meshDir     = params%general%solver%meshFolder )
    else
      ! If there is a restart, the timings in the params type have to be
      ! updated to those read from the restart
      call tem_timeControl_start_at_sim(                                     &
        & me = params%general%simControl%timeControl,                        &
        & now = params%general%simControl%now )
    end if
    ! CODE BLOCK ABOUT RESETTING PARAMS TIME ETC

    ! calculate the effective origin and effective length
    ! KM :\todo get effective bounding box from seeder header
    call tem_setEffBoundingBox( geometry%tree )

    ! Do an initial load-balancing only when there is no dynamic load balancing
    if( params%general%balance%static ) then
      call tem_horizontalSpacer(fUnit = logUnit(1))
      write(logUnit(1),*)'Load balancing ...' 
      ! Get the total number of elements from all the processes
      call mpi_allreduce( geometry%tree%nElems, total_nElems, 1, mpi_integer,  &
        &                 mpi_sum, params%general%proc%comm, iError)
      write(logUnit(1),*)'  Number of elements before load balancing',         &
        &               geometry%tree%nElems
      ! Invoke the load balancing based on the geometry
      call mus_geom_setWeight( params   = params,                              &
        &                      geometry = geometry,                            &
        &                      balance  = params%general%balance )
      call tem_balanceSparta( weight = geometry%weight,                        &
        &                     myPart = params%general%proc%rank,               &
        &                     nParts = params%general%proc%comm_size,          &
        &                     offset = geometry%offset,                        &
        &                     comm   = params%general%proc%comm )
      ! unload tree ...
      call unload_tem( me = geometry%tree ) 
      ! ... and reload it with the offsets obtained from the balancing algorithm
      call load_tem( me     = geometry%tree,                                   &
        &            conf   = params%general%solver%conf(1),                   &
        &            myPart = params%general%proc%rank,                        &
        &            nParts = params%general%proc%comm_size,                   &
        &            comm   = params%general%proc%comm,                        &
        &            offset = geometry%offset,                                 &
        &            meshDir = params%general%solver%meshFolder )
      call mus_geom_cleanWeight( geometry )
      call mpi_allreduce( geometry%tree%nElems, total_nElems, 1, mpi_integer,  &
        &                 mpi_sum, params%general%proc%comm, iError)
      write(logUnit(1),*)'  Number of elements after load balancing',          &
        &               geometry%tree%nElems
      write(logUnit(1),*)'Finished load balancing.' 
      call tem_horizontalSpacer(fUnit = logUnit(1))
    endif
    !---------------------------------------------------------------------------
    ! Done loading mesh.                                                       !
    !---------------------------------------------------------------------------
    !----------------------Load the boundary conditions------------------------
    ! Boundary conditions are loaded even in the case of dynamic load balancing
    ! when the restart is done at requisite load balancing intervals
    call mus_load_bc_data( geometry, params, .true. )
 
    ! Initialize the LB time and space steps (for multilevel)
    ! \todo KM: move all variables from pdfGlobal to params
    call mus_globInfo_set_levels( pdfGlobal,geometry%tree%global%minLevel, &
      &                                     geometry%tree%global%maxLevel )
    call mus_globInfo_set_dxdt( pdfGlobal, params%scaleFactor )
    pdfGlobal%nBCs = geometry%boundary%nBCtypes
    ! Initialize requiredInterval for multilevel to determine one complete cycle
    params%reqInterval =                                                       &
      &              params%scaleFactor**(pdfGlobal%maxLevel-pdfGlobal%minLevel)

    ! load physics table for unit converstion
    call mus_load_physics( me      = params%physics,                &
      &                    conf    = params%general%solver%conf(1), &
      &                    tree    = geometry%tree,                 &
      &                    scaleFactor = params%scaleFactor         )

    ! The scheme information is loaded from the configuration file by invoking
    ! the function. THIS IS REQUIRED EVEN IN THE CASE OF DYNAMIC LOAD BALANCING
    ! DUE TO PRESENCE OF DERIVED VARIABLE LISTS BUILDING INSIDE THE LOAD SCHEME
    ! SINGLE ROUTINE
    call mus_load_schemes( me        = scheme,                                 &
      &                    boundary  = geometry%boundary,                      &
      &                    conf      = params%general%solver%conf(1),          &
      &                    params    = params,                                 &
      &                    pdfGlobal = pdfGlobal,                              &
      &                    tree      = geometry%tree,                          &
      &                    path      = init_path )

    if( geometry%tree%global%minLevel .ne. geometry%tree%global%maxLevel ) then
      ! get the interpolation method.
      call mus_load_interpolate( me        = intp,                             &
        &                        conf      = params%general%solver%conf(1),    &
        &                        pdfGlobal = pdfGlobal )
    endif
    call tem_horizontalSpacer(fUnit = logUnit(1))

    ! Load solidification/fluidification settings
    call mus_geomIncrHead_load( me          = geometry%geomIncr,             &
      &                         params      = params,                        &
      &                         conf        = params%general%solver%conf(1), &
      &                         dynamicGeom = geometry%dynamicGeom )

    call tem_stopTimer( timerHandle = params%loadMeshTimer,                  &
      &                 me          = params%general%timerData )

  end subroutine mus_load_config
!******************************************************************************!


!******************************************************************************!
  !> This routine invokes the treelm routines to load the boundary conditions
  subroutine mus_load_bc_data( geometry, params, useBase )
    !--------------------------------------------------------------------------
    type( mus_geom_type ),  intent(inout)  :: geometry !< Treelmesh data
    !> Global parameters
    type( mus_param_type ), intent(inout)  :: params
    !> use the basename??
    logical, intent(in) :: useBase
    !--------------------------------------------------------------------------
    integer :: iProp
    logical :: found_bnd
    character(len=pathLen) :: meshDir_base
    integer :: meshDir_basePos
    !-------------- boundary conditions ----------------------------------------
    ! load boundary information from mesh
    ! generate / load the mesh
    ! per default, load the mesh from the location specified in the lua file
    found_bnd = .false.
    ! Check if there is any boundary property
    do iProp = 1, geometry%tree%global%nProperties
      if( geometry%tree%global%property( iProp )%bitpos == prp_hasBnd ) then
        if( geometry%tree%global%property( iProp )%nElems > 0 ) then
          found_bnd = .true.
        end if
      end if
    end do
    if( useBase )then
      ! get the actual directory from the tree%global%dirname for loading
      ! mesh and qVals, since both do not have timestamp
      ! this is mainly used in a regular restart
      meshDir_basePos = index(geometry%tree%global%dirname,'/',.true.)
      meshDir_base = geometry%tree%global%dirname(1:meshDir_basePos)
    else
      ! use the complete tree%global%dirname for loading bc's and qVals,
      ! this is used when dynLB is active
      meshDir_base = trim(geometry%tree%global%dirname)
    end if

    do iProp = 1, geometry%tree%global%nProperties
      if( geometry%tree%global%property( iProp )%nElems > 0 ) then
        select case( geometry%tree%global%property( iProp )%bitpos )
        case( prp_hasBnd )
          ! Load boundary condition from disk
          write(logUnit(1),*)'Loading boundary data from directory "'          &
            &             //trim(meshDir_base)//'"'
          call load_tem_BC_prop(                                               &
            &            me       = geometry%boundary,                         &
            &            offset   = geometry%tree%Property(1)%Offset,          &
            &            nElems   = geometry%tree%Property(1)%nElems,          &
            &            basename = trim(meshDir_base)//'bnd',                 &
            &            mypart   = params%general%proc%rank,                  &
            &            comm     = params%general%proc%comm )
          write(logUnit(1),*)'Done, reading the boundary!'

        case( prp_hasQVal )
          ! Load qVal from disk
          ! prp_hasQVal is the 2nd property in mesh
          write(logUnit(1),*)'Loading qVal data from directory "'              &
            &             //trim(meshDir_base)//'"'
          call load_tem_BC_qVal(                                               &
            &           me       = geometry%boundary,                          &
            &           offset   = geometry%tree%Property(2)%Offset,           &
            &           nElems   = geometry%tree%Property(2)%nElems,           &
            &           basename = trim(meshDir_base)//'qval',                 &
            &           mypart   = params%general%proc%rank,                   &
            &           comm     = params%general%proc%comm )
          write(logUnit(1),*)'Done, reading the qVal!'
        end select ! property( iProp )%bitpos
      endif ! property( iProp )%nElems > 0
    enddo ! iProp
    if( .not. found_bnd ) then
      ! No boundary conditions available!
      call tem_empty_BC_prop( geometry%boundary )
    endif
    !-------------- end boundary conditions ------------------------------------
  end subroutine mus_load_bc_data
!******************************************************************************!


!******************************************************************************!
  !> This routine loads musubi specific lua function from string and musubi
  !! input configuration file
  subroutine mus_open_config( conf, filename, proc )
    !---------------------------------------------------------------------------
    !> lua state to be stored
    type(flu_State), allocatable :: conf(:)
    !> name of the config file to be opened
    character(len=*), intent(in) :: filename
    !> process description to use
    type(tem_comm_env_type), intent(in) :: proc
    !---------------------------------------------------------------------------
    character(len=2048) :: fun_str !contains function to be used in musubi 
    integer :: iThread
    !---------------------------------------------------------------------------

    ! allocate the array of lua states
    allocate( conf( proc%nThreads ))

    call mus_create_funcStr( fun_str = fun_str )

    !pre-loading musubi specific lua functions using same conf used to load
    !main musubi config file
    do iThread=1, proc%nThreads
      call open_config_chunk(L = conf(iThread), chunk = fun_str) 
    end do

!    call tem_open_distconf(L = conf, filename = trim(filename), proc = proc )
    ! Open the lua config
    call tem_open_distconf_array( L        = conf,                             &
      &                           filename = trim(filename),                   &
      &                           proc     = proc )

  end subroutine mus_open_config
!******************************************************************************!


end module mus_config_module
!******************************************************************************!

