! See copyright notice in the COPYRIGHT file.
!!******************************************************************************!
!> \ref mus_construct "Creation of the data structures"
!! from the information in the configuration and
!! from the mesh read from disk
!!
!! \author Manuel Hasert
!!
?? include 'header/lbm_macros.inc'
module mus_construction_module

  ! include treelm modules
  use mpi
  use env_module,              only: rk, long_k, globalMaxLevels, eps
  use treelmesh_module,        only: treelmesh_type
  use tem_comm_module,         only: tem_communication_type,                   &
    &                                tem_commpattern_type
  use tem_param_module,        only: childPosition, parentNeighbor, PI,        &
    &                                tem_rotationMatrix, qDirName
  use tem_topology_module,     only: tem_LevelOf, tem_ChildNumber,             &
    &                                tem_CoordOfId, tem_IdOfCoord,             &
    &                                tem_FirstIdAtLevel
  use tem_geometry_module,     only: tem_determine_discreteVector,             &
    &                                tem_ElemSizeLevel
  use tem_tools_module,        only: tem_horizontalSpacer
  use tem_grow_array_module,   only: grw_intArray_type, init, append, destroy, &
    &                                empty
  use tem_dyn_array_module,    only: dyn_intArray_type, init, append, expand,  &
    &                                destroy, empty, PositionOfVal
  use tem_property_module,     only: tem_property_type, prp_solid,             &
    &                                prp_missingNeigh, prp_hasQVal,            &
    &                                prp_sendHalo
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_construction_module, only: tem_init_elemLevels, tem_find_allElements,&
    &                                tem_build_verticalDependencies,           &
    &                                tem_build_horizontalDependencies,         &
    &                                tem_eTypeOfTotalPos,                      &
    &                                tem_cleanup_arrays, tem_levelDesc_type,   &
    &                                tem_treeIDinTotal, tem_elemList_dump,     &
    &                                tem_levelNeighbor_type,                   &
    &                                tem_dumpTreeIDlists, tem_dumpPropBits,    &
    &                                tem_debugDependencies
  use tem_element_module,      only: eT_haloElem, eT_ghostFromCoarserElem,     &
    &                                eT_ghostFromFinerElem
  use tem_timer_module,        only: tem_startTimer, tem_stopTimer,            &
    &                                tem_getTimerVal
  use tem_stencil_module,      only: tem_stencilHeader_type,                   &
    &                                tem_stencil_map_ToTreelmDef,              &
    &                                tem_treelmDef_map_toStencil
  use tem_debug_module,        only: main_debug, dbgUnit
  use tem_reduction_module,    only: tem_GetRealBoundingCube
  use tem_aux_module,          only: tem_abort
  use tem_logging_module,      only: logUnit
  use tem_subTree_module,      only: tem_write_debugMesh
  use tem_comm_env_module,     only: tem_comm_env_type

  ! include musubi modules
  use mus_bc_header_module,          only: boundary_type, glob_boundary_type
  use mus_param_module,              only: mus_param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_interpolate_header_module, only: interpolation_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_pdf_module,                only: pdf_globInfo_type, pdf_data_type
  use mus_statistics_module,         only: mus_statistics_type,                &
    &                                      mus_calc_commAmount
  use mus_debug_module,              only: debug_normals, &
    &                                      debug_dependenciesFromFiner,        &
    &                                      debug_dependenciesFromCoarser,      &
    &                                      debug_connectivity
  use mus_comm_module,               only: mus_init_longBuffers
  use mus_field_module,              only: mus_field_type

  implicit none

  private

  public :: mus_construct
  public :: mus_construct_cleanup

  !> Contains bitmask to reduce halo elements communciation links.
  !! This type is used only in init_levelBuffers
  type halo_commBitmask_type
    !> Bitmask of which direction to send,
    !! only directions of fluid neighbor need to be update
    !! Size: nDofs, nElems
    logical, allocatable :: bitmask(:,:)
  end type halo_commBitmask_type

contains

!******************************************************************************!
  !> Initialize Musubi data strucutres based on data provided by Treelm
  !!
  !! Load the mesh and boundary conditions for this process from disk.
  !! Get the level-wise treeID lists and create the required ghost and halo
  !! elements.
  !!
  !! This is achieved by a two-folded identification of elements.
  !!
  !! -# the theoretically required elements are collected based on
  !! \ref mus_scheme_layout_module::mus_scheme_layout_type "stencil information"
  !! The \ref tem_construction_module::tem_findneighbor "find neighbor routine"
  !! performs this task for compute fluid elements.
  !! For boundaries which require information from neighbor elements, these
  !! required \ref tem_topology_module "treeIDs" are collected into the
  !! \ref mus_bc_header_module
  !! "boundary element type"
  !! -# All required elements are created in the
  !! \ref tem_construction_module::tem_createleveldescriptor
  !! "Level Descriptor creation routine"
  !!
  !! \section mus_construction_additionalTasks Additional Tasks
  !! - receive \ref tem_construction_module::tem_buildhorizontaldependencies
  !! "horizontal"
  !! (within a level for the element updates)
  !! - and \ref tem_construction::tem_buildverticaldependencies "vertical"
  !! dependencies (between levels for ghost-interpolations).
  !! - The main state vector and the neighbor lists on which the kernel then
  !!    acts is created
  !! - The MPI buffers are created.
  !! - For each \ref mus_scheme_module "Scheme", the
  !! \ref tem_construction_module::level_descriptor_type "Level Descriptor"
  !! is created
  !!
  !! \section mus_construction_result Result
  !!
  !! After this routine, all data structures for starting the main loop of the
  !! solver are allocated and ready.
  !!
  subroutine mus_construct( scheme, geometry, params, pdfGlobal, intp )
    !---------------------------------------------------------------------------
    !> global information
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal
    !> run-time Parameters
    type( mus_param_type ), intent(inout)     :: params
    !> geometric information
    type( mus_geom_type ), intent(inout)      :: geometry
    !> scheme information including fluid, boundary and flow information
    type( mus_scheme_type ), intent(inout)    :: scheme(:)
    !> interpolation information
    type( interpolation_type ), intent(inout) :: intp
    !---------------------------------------------------------------------------
    integer :: iLevel      ! counter for level
    integer :: iStencil    ! lists with different stencils
    integer :: iBC         ! counter for different boundary conditions
    integer :: iScheme     ! counter for different schemes
    integer :: iField      ! counter for different fields
    logical :: requireAll_loc ! are all links required
    integer :: minLevel, maxLevel
    !---------------------------------------------------------------------------

    minLevel = pdfGlobal%minLevel
    maxLevel = pdfGlobal%maxLevel

    write(dbgUnit(6),*) 'before construction'

    call tem_startTimer( timerHandle = params%initLvlDTimer,    &
      &                  me          = params%general%timerData )

    ! Output basic mesh infos to the screen
    call dump_meshHeader(tree = geometry%tree)

    ! For all the schemes, set up the data structures for the kernel
    ! \todo: SZ: update this subroutine for multiple fields
    do iScheme = 1, pdfGlobal%nSchemes

      allocate( scheme(iScheme)%pdf( globalMaxLevels ))

      ! check nBC in mesh and lua, for each filed
      do iField = 1, scheme( iScheme )%nFields
        if( geometry%boundary%nBCtypes /= &
          &                        size(scheme(iScheme)%field(iField)%bc) ) then
          write(logUnit(1),*) 'The number of boundary conditions specified'//  &
            & ' in '//trim( params%general%solver%configFile )//' for field '//&
            & trim( scheme( iScheme )%field( iField )%label )// ' in Scheme '//&
            & trim( scheme( iScheme )%header%label )//' does not correspond'// &
            & ' to the ones specified in the mesh'
          write(logUnit(1),*) 'Mesh File, nBC : ', geometry%boundary%nBCtypes
          write(logUnit(1),*) 'Lua File, nBC  : ', &
            &                 size(scheme(iScheme)%field(iField)%bc)
          call tem_abort()
        end if
      end do ! iField

      write(logUnit(1),*) 'Starting to initialize the geometry'
      call tem_horizontalSpacer(fUnit = logUnit(1))

      write(logUnit(1),*) 'Map stencil to treelm definition and vice versa ...'
      ! map the stencil offsets to the treelm definitions
      ! required to map boundary ID direction defined by treeLM directions
      ! from fluid stencil directions
      call tem_stencil_map_toTreelmDef(                                        &
        &                            me = scheme( iScheme )%layout%stencil(1))
      ! map the stencil offsets to the treelm definitions
      call tem_treelmDef_map_toStencil(                                        &
        &                            me = scheme( iScheme )%layout%stencil(1))

      ! Stencil(1) stands for fluid elements
      ! intialize stencil for all fluid elements
      scheme( iScheme )%layout%stencil(1)%nElems = geometry%tree%nElems

      ! 1. determine the number of elements for each mesh boundary
      ! 2. copy the information to globbc
      ! 3. determine the normal vector and bitmasks for the individual
      !    boundaries
      ! 4. allocate the BC lists in globbc
      ! THIS ROUTINE DOES NOT MODIFY THE BC LISTS IN THE FIELDS!!!
      call build_BClists( globBC    = scheme( iScheme )%globBC,                &
        &                 tree      = geometry%tree,                           &
        &                 boundary  = geometry%boundary,                       &
        &                 property  = geometry%tree%Property,                  &
        &                 pdfGlobal = pdfGlobal,                               &
        &                 layout    = scheme( iScheme )%layout,                &
        &                 field     = scheme( iScheme )%field(:),              &
        &                 proc      = params%general%proc )

      ! Treat the boundary lists: in addition to the normal stencil treatment
      ! one has to find all those elements along the element list in the stencil
      ! type, which are boundary elements and give them the special
      ! boundary treatment

      ! 1. define the stencil for the individual boundaries (in the fields)
      !    by setting useAll, requireNeighNeigh, QQ, QQN, nElems, cxDir
      ! 2. fill stencil%elem with the level wise element positions and
      !    stencil%elemLvl with the level wise positions in the primal mesh
      do iField = 1, scheme( iScheme )%nFields
        do iBC = 1, pdfGlobal%nBCs
          call mus_build_BCStencils(                                           &
            &      globbc  = scheme( iScheme )%globbc( iBC ),                  &
            &      bc      = scheme( iScheme )%field( iField )%bc( iBC ),      &
            &      tree    = geometry%tree,                                    &
            &      layout  = scheme( iScheme )%layout,                         &
            &      prefix  = scheme( iScheme )%field( iField )%label,          &
            &      stencil = scheme( iScheme )%layout%stencil(                 &
            &          scheme( iScheme )%field( iField )%bc( iBC )%stencilPos ))
        enddo
      enddo

      ! 1. store treeIDs which use this stencil in the
      !    levelDesc( elemLevel )%elem
      ! 2. store the neighbors which can be loaded from the disc or accessed
      !    directly
      ! THIS MEANS THAT FOR ALL ELEMENTS (INCL. BOUNDARY ELEMENTS) NEIGHBORS
      ! ARE STORED BUT THEY DO NOT NEED TO EXIST IN THE MESH AT THIS POINT!!!
      write(dbgUnit(6),*) 'before routine tem_init_elemLevels'
      write(logUnit(6),*) 'before routine tem_init_elemLevels'
      call tem_init_elemLevels(                                                &
        &                me             = scheme( iScheme )%glob%levelDesc,    &
        &                boundary       = geometry%boundary,                   &
        &                tree           = geometry%tree,                       &
        &                computeStencil = scheme( iScheme )%layout%stencil(:))

      ! 1. build all dependencies for fluid, halos and ghosts
      ! 2. build the pointers for each element to its neighbors/stencil elements
      ! All this information is stored in tem_levelDesc_type.
      !
      ! THIS MEANS THAT FOR ALL ELEMENTS FOUND IN em_init_elemLevels THE CORRECT
      ! DEPENDENCIES AND NEIGHBORS ARE SET
      !
      ! This is the heart of the topologic data structure creation and
      ! takes most of the compute time, depending on the mesh you have
      !
      write(logUnit(1),*) 'Creating level descriptor ...'
      write(dbgUnit(6),*) 'before find_allElements'
      ! KJ: The level descriptor is created from the below routine. LD is the heart 
      ! of the code which is further processed and used in Musubi. First of all 
      ! the neigh array is created using the LD and communication buffers are filled up. 
      call tem_find_allElements(                                               &
        &                  tree            = geometry%tree,                    &
        &                  levelDesc       = scheme( iScheme )%glob%levelDesc, &
        &                  levelPointer    = pdfGlobal%levelPointer,           &
        &                  computeStencil  = scheme(iScheme)%layout%stencil(:),&
        &                  commPattern     = params%general%commPattern,       &
        &                  cleanup         = .true.,                           &
        &                  reqNesting      = params%nNesting,                  &
        &                  proc            = params%general%proc )

      write(dbgUnit(6),*) 'before horizontal dep'
      ! For each of the neighbor lists create the horizontal (neighbor)
      ! relations of the size of the total number of stencils used in this
      ! scheme
      do iLevel = minLevel, maxLevel
        allocate( scheme( iScheme )%glob%levelDesc( iLevel )%                  &
          &         neigh( scheme( iScheme )%layout%nStencils ))
      enddo
      do iStencil = 1, scheme( iScheme )%layout%nStencils
        call tem_build_horizontalDependencies(                                 &
          &       iStencil       = iStencil,                                   &
          &       levelDesc      = scheme( iScheme )%glob%levelDesc,           &
          &       tree           = geometry%tree,                              &
          &       computeStencil = scheme( iScheme )%layout%stencil(iStencil), &
          &       check          = main_debug%checkDependencies )
      enddo

      ! Communicate property after setting missing Neighbor in
      ! tem_build_horizontalDependencies routine

      do iLevel = minLevel, maxLevel
        call communicate_property(                                             &
          &   send     = scheme( iScheme )%glob%levelDesc( iLevel )%sendbuffer,&
          &   recv     = scheme( iScheme )%glob%levelDesc( iLevel )%recvbuffer,&
          &   property = scheme( iScheme )%glob%levelDesc( iLevel )%property,  &
          &   flag     = iLevel,                                               &
          &   proc     = params%general%proc,                                  &
          &   pattern  = params%general%commPattern )
        call communicate_property(                                             &
          &   send     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            sendbufferFromCoarser,  &
          &   recv     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            recvbufferFromCoarser,  &
          &   property = scheme( iScheme )%glob%levelDesc( iLevel )%property,  &
          &   flag     = iLevel,                                               &
          &   proc     = params%general%proc,                                  &
          &   pattern  = params%general%commPattern )
        call communicate_property(                                             &
          &   send     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            sendbufferFromFiner,    &
          &   recv     = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                            recvbufferFromFiner,    &
          &   property = scheme( iScheme )%glob%levelDesc( iLevel )%property,  &
          &   flag     = iLevel,                                               &
          &   proc     = params%general%proc,                                  &
          &   pattern  = params%general%commPattern )
      end do

      call tem_horizontalSpacer( after = 1, fUnit = logUnit(1) )

      ! Update the positions stored in the bc%elem lists to the positions in the
      ! levelwise total list
      ! 1. allocate and initialize the pre- and post collision buffers as well
      !    as the bcBuffer for the boundary elements
      ! 2. Update the positions stored in the bc%elems lists to the positions
      !    in the levelwise total list as well as the positions of the neighbors
      !    in bc%elems%neigh
      ! IN THIS ROUTINE THE ALREADY EXISTING NEIGHBORS ARE COPIED AND THE ONES
      ! WHICH DO NOT EXIST ARE SET TO REASONABLE ELEMENTS (CURRENT ELEMENT OR
      ! LAST VALID NEIGHBOR)
      call update_BClists( scheme    = scheme( iScheme ),                      &
        &                  pdfGlobal = pdfGlobal )

      write(dbgUnit(6),*) 'before routine: tem_build_vertical dependencies'
      ! Get vertical dependencies (between levels for interpolation)
      call tem_build_verticalDependencies(                                     &
        &        levelDesc = scheme( iScheme )%glob%levelDesc,                 &
        &        minlevel  = minLevel )

      write(dbgUnit(6),*) 'cleanup'
      ! KJ: \todo de-activate for adaptive grid refinement
      call tem_cleanup_arrays( levelDesc = scheme( iScheme )%glob%levelDesc )

      write(dbgUnit(6),*) 'after cleanup'

      call calculate_nElems( scheme    = scheme( iScheme ),   &
        &                    iScheme   = iScheme,             &
        &                    proc      = params%general%proc, &
        &                    minLevel  = minLevel, &
        &                    maxLevel  = maxLevel  )

      write(logUnit(4),*) 'Allocating PDF state and neighbor array'
      ! Allocate the PDF state array
      do iLevel = minLevel, maxLevel
        allocate( scheme( iScheme )%pdf( iLevel )%state(                       &
          & scheme( iScheme )%pdf( iLevel )%nElems_local *                     &
          & scheme( iScheme )%varSys%nScalars, 2 ))
        ! allocate the connectivity array
        !\todo KM: use single neighbor connectivity for all fields/pdfs
        !      SZ: Is this possible in case more state variables exists?
        !          For non-newtonian fluids we might add omega as an additional
        !          state variable.
        !      JQ: Omega does not require a neighbor.
        !          It should not be a problem when using nScalars.
        allocate( scheme( iScheme )%pdf( iLevel )%neigh(                       &
          & scheme( iScheme )%pdf( iLevel )%nElems_local *                     &
          & scheme( iScheme )%varSys%nScalars ))
      end do


      ! reconstruct connectivity vector pdf( iLevel )%neigh including
      ! bounce back rules
      call reconstruct_connectivity( scheme   = scheme( iScheme ),  &
        &                            minLevel = minLevel, &
        &                            maxLevel = maxLevel  )

      write(dbgUnit(6),*)  'before update deps'
      ! Only if multilevel
      if ( minLevel /= maxLevel ) then
        if ( trim(intp%label) /= 'compact' ) then
        ! Add additional source elements for fromCoarser ghost elements
        call mus_intp_complete_fromCoarserDep(                                 &
          &                        intp      = intp,                           &
          &                        scheme    = scheme( iScheme ),              &
          &                        pdfGlobal = pdfGlobal )

        ! Update the dependencies fromFiner (in 2d, 4 of 8 get removed)
        call mus_intp_update_fromFinerDep(                                     &
          &                        intp      = intp,                           &
          &                        scheme    = scheme( iScheme ),              &
          &                        pdfGlobal = pdfGlobal )
        end if
      end if

      ! Set halo elements which require all links for communication.
      call set_halo_commLinks( scheme( iScheme ), pdfGlobal,                   &
        &                      params%comm_reduced )

      write(dbgUnit(6),*)  'before init levelBuffers'
      ! Create communication buffers
      pdfGLobal%stat%nLinks_comm  = 0
      pdfGLobal%stat%nLinks_total = 0
      requireAll_loc = any( scheme( iScheme )%layout%stencil(:)%requireAll )
      do iLevel = minLevel, maxLevel
        !    for multi-level, still all the links have to be communicated
        !    so, check for regular Buffer, if the max and min level is the same
        call init_levelBuffers(                                                &
          & send       = scheme( iScheme )%glob%levelDesc( iLevel )%sendbuffer,&
          & recv       = scheme( iScheme )%glob%levelDesc( iLevel )%recvbuffer,&
          & pdf        = scheme( iScheme )%pdf(iLevel),                        &
          & requireAll = requireAll_loc,                                       &
          & pattern    = params%general%commPattern,                           &
          & proc       = params%general%proc,                                  &
          & scheme     = scheme( iScheme ),                                    &
          & stat       = pdfGLobal%stat,                                       &
          & levelDesc  = scheme( iScheme )%glob%levelDesc( iLevel ),           &
          & layout     = scheme( iScheme )%layout )

        call init_levelBuffers(                                                &
          & send       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                             sendbufferFromCoarser, &
          & recv       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                             recvbufferFromCoarser, &
          & pdf        = scheme( iScheme )%pdf(iLevel),                        &
          & requireAll = requireAll_loc,                                       &
          & pattern    = params%general%commPattern,                           &
          & proc       = params%general%proc,                                  &
          & scheme     = scheme( iScheme ),                                    &
          & stat       = pdfGLobal%stat,                                       &
          & levelDesc  = scheme( iScheme )%glob%levelDesc( iLevel ),           &
          & layout     = scheme( iScheme )%layout )

        call init_levelBuffers(                                                &
          & send       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                               sendbufferFromFiner, &
          & recv       = scheme( iScheme )%glob%levelDesc( iLevel )%           &
          &                                               recvbufferFromFiner, &
          & pdf        = scheme( iScheme )%pdf(iLevel),                        &
          & requireAll = requireAll_loc,                                       &
          & pattern    = params%general%commPattern,                           &
          & proc       = params%general%proc,                                  &
          & scheme     = scheme( iScheme ),                                    &
          & stat       = pdfGLobal%stat,                                       &
          & levelDesc  = scheme( iScheme )%glob%levelDesc( iLevel ),           &
          & layout     = scheme( iScheme )%layout )
      enddo

      ! Debug the level lists, dependencies, buffers ...
      ! \todo: DEB: update these debug calls
      if( main_debug%dumpTreeIDlists ) then
        write(logUnit(1),*) 'Dumping elements list for debug purpose'
        do iLevel = minlevel, maxLevel
          call tem_elemList_dump(                                              &
            &  me         = scheme( iScheme )%glob%levelDesc( iLevel )%elem,   &
            &  nUnit      = dbgUnit(5),                                        &
            &  string     = '  writing treeID list ',                          &
            &  stencil    = .true.,                                            &
            &  stencilPos = .true.,                                            &
            &  compact    = .true. )
        end do
      end if

      if( main_debug%dumpPropBits )                                            &
        call tem_dumpPropBits( levelDesc = scheme( iScheme )%glob%levelDesc,   &
        &                      minlevel  = minLevel)

      if( main_debug%dumpAuxLists ) then
        ! Write out detailed information about the boundary lists
        !call debug_boundaries(tree, glob, pdf, glob%levelDesc, boundary, bc )

        ! Print out PDF neighbor array
        call debug_connectivity( scheme   = scheme( iScheme ),  &
          &                      minLevel = pdfGlobal%minLevel, &
          &                      maxLevel = pdfGlobal%maxLevel  )
      end if

      call tem_write_debugMesh( globtree  = geometry%tree,                     &
        &                       levelDesc = scheme(iScheme)%glob%levelDesc,    &
        &                       myPart    = params%general%proc%rank )

      if( main_debug%dumpDependencies) then
        ! Check the dependencies from finer
        call debug_dependenciesFromFiner( scheme( iScheme )%pdf, &
          &                               scheme( iScheme )%glob, &
          &                               minLevel, maxLevel )
        call debug_dependenciesFromCoarser( scheme( iScheme )%pdf, &
          &                                 scheme( iScheme )%glob, &
          &                                 minLevel, maxLevel )
        if ( size( geometry%tree%Property ) > 0 ) then
          call debug_normals( pdfGlobal = pdfGlobal,                           &
            &                 tree      = geometry%tree,                       &
            &                 scheme    = scheme( iScheme) )
        end if
      end if

      do iLevel = geometry%tree%global%minlevel, geometry%tree%global%maxLevel
        ! set the sendHalo bit for all fluid elements which are send
        ! to remote procs
        call set_sendHaloBits( levelDesc = scheme(iScheme)%glob%               &
          &                                                levelDesc( iLevel ))
      end do

    end do ! SchemeLoop

    call mus_calc_commAmount( stat = pdfGlobal%stat,     &
      &                       proc = params%general%proc )

    call tem_stopTimer( timerHandle = params%initLvlDTimer,    &
      &                 me          = params%general%timerData )

    !! \todo: move this output to a more appopriate place.
    !! \todo: except the first one, all other stencil labels are not set.
    write(logUnit(1),"(A)") 'StencilLabels: '
    do iStencil = 1, scheme(1)%layout%nStencils
      write(logUnit(1),"(A,I0)") trim(scheme(1)%layout%stencil(iStencil)%label) &
        &                     //', QQ: ', scheme(1)%layout%stencil(iStencil)%QQ
    end do

    call tem_horizontalSpacer( before = 1, fUnit = logUnit(1) )

    write(logUnit(1),*) 'Done initializing geometry. Took [s]',                &
      & real( tem_getTimerVal(params%general%timerData, params%initLvlDTimer) )
    call tem_horizontalSpacer( after  = 1, fUnit = logUnit(1) )

    write(dbgUnit(1),*)  'after the construction'

  end subroutine mus_construct
!******************************************************************************!


!******************************************************************************!
  !> write mesh information
  subroutine dump_meshHeader( tree )
    !---------------------------------------------------------------------------
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in) :: tree
    !---------------------------------------------------------------------------

    write(logUnit(1),*)        'Mesh Information'
    call tem_horizontalSpacer(fUnit = logUnit(1))
    write(logUnit(1),"(A)")    ' Mesh name: '//trim(tree%global%label)
    write(logUnit(1),"(A,I0)") ' local number of elements: ', tree%nElems
    write(logUnit(1),"(A,I0)") ' total number of elements: ', tree%global%nElems
    write(logUnit(1),"(A,I0)") ' minimum Level in tree: ', tree%global%minLevel
    write(logUnit(1),"(A,I0)") ' maximum Level in tree: ', tree%global%maxLevel
    write(logUnit(1),*) 'bounding cube origin: ', &
      &                        real(tree%global%origin)
    write(logUnit(1),*) 'bounding cube size:         ', &
      &                 real( tree%global%BoundingCubeLength )
    write(logUnit(1),*) 'coarsest dx                 ', &
      &                 real(tem_ElemSizeLevel( tree, tree%global%minLevel))
    write(logUnit(1),*) 'finest dx                   ', &
      &                 real(tem_ElemSizeLevel( tree, tree%global%maxLevel))

    if ( size(tree%Property) > 0 ) then
      write(logUnit(1),"(A,I0)") 'nElems with boundaries: ', &
        &                        tree%Property(1)%nElems
    end if
    call tem_horizontalSpacer(fUnit = logUnit(1))
  end subroutine dump_Meshheader
!******************************************************************************!


!******************************************************************************!
  !> Initialize the communication buffers for a single level
  subroutine communicate_property( send, recv, property, flag, proc, pattern)
    !---------------------------------------------------------------------------
    !> Communication structure to send and receive property
    type(tem_communication_type), intent(inout) :: send, recv
    !> property to communicate
    integer(long_k), intent(inout)              :: property(:)
    !> communication flag
    integer, intent(in)                         :: flag
    !>
    type(tem_commPattern_type), intent(in)      :: pattern
    !> Process description to use.
    type(tem_comm_env_type), intent(in)         :: proc
    !---------------------------------------------------------------------------
    call mus_init_longBuffers( send    = send,                                 &
      &                        recv    = recv,                                 &
      &                        pattern = pattern )
    call pattern%exchange_long( send         = send,                           &
      &                         recv         = recv,                           &
      &                         state        = property,                       &
      &                         message_flag = flag,                           &
      &                         proc         = proc )

  end subroutine communicate_property
!******************************************************************************!


!******************************************************************************!
  !> Initialize the communication buffers for a single level
  subroutine init_levelBuffers( send, recv, pdf, pattern, levelDesc, layout,   &
    &                           requireAll, scheme, stat, proc )
    !---------------------------------------------------------------------------
    !> Communication structure to initialize
    type(tem_communication_type)               :: send, recv
    !> iLevel pdf info with state and neigh array
    type( pdf_data_type), intent(in)           :: pdf
    !> communication pattern
    type( tem_commpattern_type ), intent(in)   :: pattern
    !> levelDesc
    type( tem_levelDesc_type ), intent(in)     :: levelDesc
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> Statistics
    type( mus_statistics_type ), intent(inout) :: stat
    !> mpi communication enviroment with mpi communicator
    type(tem_comm_env_type), intent(in)        :: proc
    !> fluid, bnd and flow info
    type( mus_scheme_type ), intent(in)        :: scheme
    !> different place to take values from for interpolation
    logical                                    :: requireAll
    !---------------------------------------------------------------------------
    ! myhalos are to be received from remote and remoteHalos are to be
    ! send to remote
    type( halo_commBitmask_type ), allocatable :: myHalos(:), remoteHalos(:)
    !---------------------------------------------------------------------------

    ! initialize bitmask array
    call init_bitmask( myHalos     = myHalos,               &
      &                remoteHalos = remoteHalos,           &
      &                send        = send,                  &
      &                recv        = recv,                  &
      &                nScalars    = scheme%varSys%nScalars )

    ! only initialize recvBuffers and then communicate the desired
    ! values to the neighbor processes.
    ! \todo: SZ: think about the nDOFs in case more state variables than
    !            the QQ PDFs exist
    call init_recvBuffers(me         = recv,                 &
      &                   myhalos    = myhalos,              &
      &                   pdf        = pdf,                  &
      &                   nDOFs      = layout%stencil(1)%QQ, &
      &                   requireAll = requireAll,           &
      &                   levelDesc  = levelDesc,            &
      &                   stat       = stat,                 &
      &                   scheme     = scheme,               &
      &                   pattern    = pattern               )

    ! communicate bitmasks 
    call communicate_bitmask( myHalos     = myHalos,               &
      &                       remoteHalos = remoteHalos,           &
      &                       send        = send,                  &
      &                       recv        = recv,                  &
      &                       proc        = proc,                  &
      &                       nScalars    = scheme%varSys%nScalars )

    call init_sendBuffers(send        = send,                 &
      &                   remoteHalos = remoteHalos,          &
      &                   pdf         = pdf,                  &
      &                   nDOFs       = layout%stencil(1)%QQ, &
      &                   requireAll  = requireAll,           &
      &                   scheme      = scheme,               &
      &                   pattern     = pattern               )

    ! destroy send and recv bitmask since its not needed anymore
    call destroy_bitmask( myHalos, remoteHalos, send, recv )  

  end subroutine init_levelBuffers
!******************************************************************************!


!******************************************************************************!
  !> Calculate level-wise, total and local as well as global numbers of elements
  !!
  subroutine calculate_nElems( scheme, iScheme, proc, minLevel, maxLevel )
    !---------------------------------------------------------------------------
    !>
    type( mus_scheme_type ), intent(inout)  :: scheme
    !> number of current scheme
    integer, intent(in)                     :: iScheme
    !> mpi communication enviroment with mpi communicator
    type(tem_comm_env_type), intent(in)     :: proc
    !> Level range
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !---------------------------------------------------------------------------
    integer :: iLevel !< level of source elements
    integer(kind=long_k), allocatable :: ElemCount(:)
    integer :: iErr
    ! integer :: nElems_solve
    !---------------------------------------------------------------------------
    ! Count the amount of different element types
    ! counting levelwise elements ...
    scheme%glob%nElems_localFluid = 0
    scheme%glob%nElems_localSolve = 0
    call tem_horizontalSpacer(fUnit = logUnit(3))
    write(logUnit(3),"(A,I0)") 'Elements per level for scheme: ', iScheme

    !write(logUnit(1),*)'WARNING: Not computing on ghostFromFiner!!'
    do iLevel = minLevel, maxLevel
      scheme%glob%nElems_localFluid = scheme%glob%nElems_localFluid            &
        &          + scheme%glob%levelDesc(iLevel)%nElems_fluid

      scheme%glob%nElems_localSolve = scheme%glob%nElems_localSolve            &
        &          + scheme%glob%levelDesc(iLevel)%nElems_fluid                &
        &          + scheme%glob%levelDesc(iLevel)%nElems_ghostFromCoarser

      scheme%pdf( iLevel )%nElems_fluid                                        &
        &          = scheme%glob%levelDesc( iLevel )%nElems_fluid

      scheme%pdf( iLevel )%nElems_ghostFromCoarser                             &
        &          = scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser

      scheme%pdf( iLevel )%nElems_ghostFromFiner                               &
        &          = scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner

      scheme%pdf( iLevel )%nElems_ghost                                        &
        &          = scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser   &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner

      scheme%pdf( iLevel )%nElems_halo                                         &
        &          = scheme%glob%levelDesc( iLevel )%nElems_halo

      scheme%pdf( iLevel )%nElems_local                                        &
        &          = scheme%glob%levelDesc( iLevel )%nElems_fluid              &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner     &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser   &
        &          + scheme%glob%levelDesc( iLevel )%nElems_halo

      scheme%pdf( iLevel )%nElems_solve =                                      &
        &            scheme%glob%levelDesc( iLevel )%nElems_fluid              &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromFiner     &
        &          + scheme%glob%levelDesc( iLevel )%nElems_ghostFromCoarser


      ! print out number of elements of all kinds
      write(logUnit(6),"(A,I0)") '  Number of local elements on level: ',iLevel
      write(logUnit(6),"(A,I0)") '    fluid: ', &
        &                        scheme%pdf( iLevel )%nElems_fluid
      write(logUnit(6),"(A,I0)") '    ghostFromCoarser: ', &
        &                        scheme%pdf( iLevel )%nElems_ghostFromCoarser
      write(logUnit(6),"(A,I0)") '    ghostFromFiner:   ', &
        &                        scheme%pdf( iLevel )%nElems_ghostFromFiner
      write(logUnit(6),"(A,I0)") '    halo: ', &
        &                        scheme%pdf( iLevel )%nElems_halo

    end do
    call tem_horizontalSpacer(fUnit = logUnit(6))
    ! done counting levelwise  elements

    ! Calculate the total number of fluid elements in ALL LEVELS and processes
    allocate( ElemCount( maxLevel ))
    do iLevel = minLevel,MaxLevel
      ElemCount( iLevel ) = scheme%pdf( iLevel )%nElems_solve
    end do
    call mpi_allreduce( ElemCount, scheme%glob%nElems_solveLevel(:),           &
      & maxLevel, mpi_integer8, mpi_sum, proc%comm, iErr )

    do iLevel = minLevel, MaxLevel
      ElemCount( iLevel ) = scheme%pdf( iLevel )%nElems_local
    end do
    call mpi_allreduce( ElemCount, scheme%glob%nElems_allLevel(:),             &
      &  maxLevel, mpi_integer8, mpi_sum, proc%comm, iErr )

    do iLevel = minLevel, MaxLevel
      ElemCount( iLevel ) = scheme%pdf( iLevel )%nElems_fluid
    end do
    call mpi_allreduce( ElemCount, scheme%glob%nElems_totalLevel(:),           &
      &  maxLevel, mpi_integer8, mpi_sum, proc%comm, iErr )
    deallocate(ElemCount)

    scheme%glob%nElems_totalFluid = 0_long_k
    scheme%glob%nElems_totalSolve = 0_long_k
    scheme%glob%nElems_overall =                                               &
      &     sum( scheme%glob%nElems_allLevel( minLevel:maxLevel ))
    do iLevel = minLevel, maxLevel
      scheme%glob%nElems_totalFluid = scheme%glob%nElems_totalFluid            &
                             + scheme%glob%nElems_totalLevel( iLevel )

      scheme%glob%nElems_totalSolve = scheme%glob%nElems_totalSolve            &
                             + scheme%glob%nElems_solveLevel( iLevel )
      write(logUnit(1),"(A,I0)") ' Number of elements on Level: ', iLevel
      write(logUnit(1),"(A,I0)") '   Fluid:   ', &
        &                        scheme%glob%nElems_totalLevel( iLevel )
      write(logUnit(1),"(A,I0)") '   toSolve: ', &
        &                        scheme%glob%nElems_solveLevel( iLevel )
    end do
    write(logUnit(1),"(A,I0)") ' Numbers of elements for scheme: ', iScheme
    write(logUnit(1),"(A,I0)") '   Total Fluids:  ', &
      &                        scheme%glob%nElems_totalFluid
    write(logUnit(1),"(A,I0)") '   Total toSolve: ', scheme%glob%nElems_totalSolve
    write(logUnit(1),"(A,I0)") '   Total overall: ', scheme%glob%nElems_overall

    write(logUnit(1),*) ''

  end subroutine calculate_nElems
!******************************************************************************!



!******************************************************************************!
  !> Complete the required interpolation dependencies and calculate the
  !! distances for the weights.
  !!
  !! The depdendencies for ghostFromCoarser elements initially only include its
  !! parent element.
  !! Additional dependencies Y for the element x are found which were not
  !! included in the construction routine before hand. However, they are
  !! defined to be inherently part of the stencil and hence exist locally and
  !! can simply be added to the dependencies. Two different ways are
  !! implemented.
  !! - compact interpolation stencil
  !! \verbatim
  !!  Y       Y
  !!
  !!        x   o
  !!  Y       O
  !!        o   o
  !! \endverbatim
  !! - extended interpolation stencil
  !!   the complete compute stencil is included into the dependency list
  !! \verbatim
  !!  Y       Y       Y
  !!
  !!        x   o
  !!  Y       O       Y
  !!        o   o
  !!
  !!  Y       Y       Y
  !! \endverbatim
  !!
  !! Then, calculate the weights based on  barycenter distance of source and
  !! target elements.
  !!
  !! Note: The view here is according to how the interpolation will be done.
  !!       The coarse level does the interpolation for the finer level.
  !!       Thus, we access the information from the levelDesc with the
  !!       targetLevel and sourceLevel perspective.
  !!
  subroutine mus_intp_complete_fromCoarserDep( intp, scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !>
    type( mus_scheme_type )                   :: scheme
    !>
    type( interpolation_type ), intent(inout) :: intp
    !>
    type( pdf_globInfo_type), intent(in)      :: pdfGlobal

    !---------------------------------------------------------------------------
    integer(kind=long_k) :: nTreeID, tOffset, curTreeID
    integer :: sourceLevel  ! level of source elements
    integer :: sourceElem   ! treeId of current source element
    integer :: targetLevel  ! level of target elements
    integer :: targetElem   ! treeId of current source element
    integer :: iElem        ! current target element (for outer loop)
    integer :: iSourceElem  ! current source element (for inner loop)
    integer :: nSourceElems ! number of source elements for the current target
    ! maximum number of neighbor elements
    ! Compact stencil -> 8, quadraticExtended = stencil%QQ
    integer :: nMaxElems
    integer :: nNeighElems
    integer :: elemPos
    integer :: childNum, iNeigh
    integer :: neighOffset(3), x(4), parentPos( 3,8)
    integer :: iStencilElem, iLoop
    integer :: foundDirInStencil
    logical :: force_compact
    logical :: completed
    logical :: validElement
    integer :: dPos, iPos ! position, where a source element was stored
    real(kind=rk) :: targetBary(3) ! barycenter of current target element
    ! weight for each source element depending on distance to target element
    real(kind=rk),allocatable :: weight(:)
    ! resulting distance of bary centers from target to uncle
    real(kind=rk)             :: distance(3)
    integer, allocatable      :: stencil(:,:)
    ! intermediate storage
    type( grw_intArray_type ) :: tElem
    type( dyn_intArray_type ) :: tSourceFromCoarser
    real(kind=rk)             :: sumWeight
    integer(kind=long_k )     :: targetTID
    !---------------------------------------------------------------------------

! write(*,"(A)") "Inside mus_intp_complete_fromCoarserDep routine"

    if( intp%compact ) then
      nMaxElems = max( 1, intp%nSourcesCompactFromCoarser_required)
    else
      nMaxElems = scheme%layout%stencil(1)%QQ
    end if

!    call map_compactToComputeStencil( intp = intp, &
!      &            stencil = scheme%layout%stencil(1))

    allocate( stencil( 3, nMaxElems ))
    allocate( weight( nMaxElems ))

    do sourceLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel - 1
      targetLevel = sourceLevel + 1
      ! initialize the levelwise source element list for the buffer
      call init( me = scheme%glob%levelDesc( targetLevel )%sourceFromCoarser,  &
        &        unique = .true. )
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromCoarser_high)
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromCoarser_low )

      ! Treat all the fromCoarser elements
      do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromCoarser
        ! Get the target element which has to be updated
        targetElem = iElem + scheme%glob%levelDesc( targetLevel )%             &
          &                                offset( 1, eT_ghostFromCoarserElem )
        targetTID = scheme%glob%levelDesc( targetLevel )%total( targetElem )

!        write(dbgUnit(10),*) 'found the targetTID'

        ! Re-assign initial nMaxElems definition
        nNeighElems = nMaxElems
        force_compact = .false.
        completed = .false.
        ! Loop twice: first, try to find the extended stencil
        !             second: if not all elements were found, try to find the
        !                     compact stencil
        do iLoop = 1, 2
          ! If all source elements were found in the first round, then exit.
          ! No second loop required.
          if( completed ) exit
          if( iLoop ==1 ) then
            ! Store intermediate element dependencies (store complete grw type)
            tElem = scheme%glob%levelDesc( targetLevel )%                      &
              &                                    depFromCoarser( iElem )%elem
            ! And the list of all so far detected source elements (might
            ! be deleted)
            tSourceFromCoarser = scheme%glob%levelDesc( targetLevel )%         &
              &                                               sourceFromCoarser
            if( tSourceFromCoarser%nvals > 0 ) then
!              write(dbgUnit(10),*) '1st loop: shelving srcFromCoarser',       &
!                &          tSourceFromCoarser%val( 1: tSourceFromCoarser%nvals)
            end if
          end if
          ! If not all source elements were found in the first round, it means
          ! that the compact neighborhood must be reconstructed.
          if( iLoop == 2 ) then
!            write(dbgUnit(10),*)'2nd loop: '
            ! If 2nd loop: use compat stencil definition
            nNeighElems = intp%nSourcesCompactFromCoarser_required
            force_compact = .true.
            ! re-create the lists for this element
            call destroy( me = scheme%glob%levelDesc( targetLevel )%           &
              &                              depFromCoarser(iElem)%elemBuffer )
            call destroy( me = scheme%glob%levelDesc( targetLevel )%           &
              &                                    depFromCoarser(iElem)%elem )
            call destroy( me = scheme%glob%levelDesc( targetLevel )%           &
              &                                             sourceFromCoarser )
            ! recover the original element dependencies of before the first round
            ! replace current empty dyn list with the one stored above
            scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%elem  &
              &                                                         = tElem
            ! recover list of all so far detected source elements from stored
            ! above
            scheme%glob%levelDesc( targetLevel )%sourceFromCoarser             &
              &                                            = tSourceFromCoarser
          end if

          ! initialize array for storing the positions of the source elements in
          ! the levelwise buffer
          call init( me     = scheme%glob%levelDesc( targetLevel )%            &
            &                               depFromCoarser(iElem)%elemBuffer,  &
            &        length = nNeighElems )
          stencil = 0

          ! Run over all the source elements which are there
          if( scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )     &
            &            %elem%nVals > 1) then
            write(logUnit(1),*) 'Warning: encountered more than one parent '// &
              &             'elements, for which we are completing the '//     &
              &             'dependencies'
            write(logUnit(1),*) ' depFromCoarser ',                            &
              &        scheme%glob%levelDesc( targetLevel )%depFromCoarser     &
              &        ( iElem )%elem%val( 1:scheme%glob%levelDesc             &
              &        ( targetLevel )%depFromCoarser( iElem )%elem%nVals )
          end if

          !--------------- treat   parent element (first source) --------------
          ! There is only one source element, namely the parent one.
          ! Therefore I need to fill up the neighbors of the parent according
          ! to the stencil
          iSourceElem = 1
          ! Reset the source element entries
          nSourceElems = scheme%glob%levelDesc( targetLevel )                  &
            &                               %depFromCoarser( iElem )%elem%nVals

!          write(dbgUnit(10),*)'Initial entries in deps: ', nSourceElems
!          write(dbgUnit(10),*)'content: ', scheme%glob%levelDesc(             &
!            & targetLevel )%depFromCoarser( iElem )%elem%val( 1: nSourceElems )

          scheme%glob%levelDesc( targetLevel )                                 &
            &          %depFromCoarser( iElem )%elem%val( nSourceElems+1:) = -1
          weight = 0._rk
          ! start with treating the parent element, which is located at the
          ! first position in the dependency list
          ! Later on, proceed to siblings of parent in the target element's
          ! directions from its parent element
          sourceElem = scheme%glob%levelDesc( targetLevel )                    &
            &                  %depFromCoarser( iElem )%elem%val( iSourceElem )
          ! add the parent to the list of source elements on this level
          ! delete parent element from the list
          call empty( me = scheme%glob%levelDesc( targetLevel )                &
            &                                   %depFromCoarser( iElem )%elem )
          !--------------- treated parent element (first source) ---------------
!          write(dbgUnit(10),*) iLoop, 'content of dep after adding parent',   &
!            & scheme%glob%levelDesc( targetLevel )%depFromCoarser             &
!            &       ( iElem )%elem%val( 1:scheme%glob%levelDesc               &
!            &       ( targetLevel )%depFromCoarser( iElem )%elem%nVals )

          !--------------- treat next source elements --------------------------
          ! Get treeID of first source element
          nTreeID = scheme%glob%levelDesc( sourceLevel )%total( sourceElem )
          x = tem_CoordOfId( nTreeID )
          tOffset = tem_FirstIdAtLevel( x(4) )

          ! target (ghost) treeID, to find out into which direction the
          ! neighbors have to be searched
          curTreeID = scheme%glob%leveldesc( targetlevel )%total( targetElem )
          childNum = tem_ChildNumber( curTreeID )
          ! Store the child number
          scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%childNum&
            &                                                        = childNum

          ! barycenter computation:
          ! Distance from one barycenter to the next = 1
          targetBary = 0.25_rk * real(childPosition(childNum, 1:3), kind=rk)
          ! Store the relative position for later interpolation.
          ! We store the position relative to the parent one.

          if( intp%compact .or. force_compact ) then
          ! Here we store the position relative to the enclosing elements on
          ! the parent level. The children on the right, which have local coord
          ! -0.25 become 0.75, children on the left with +0.25 become 0.25
            scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )       &
              &                           %coord(:) = abs( targetBary - 0.5_rk )
          else
            scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )       &
              &                                           %coord(:) = targetBary
          end if
          ! Now loop over the maximum number of elements = computeStencil
          do iNeigh = 1, nNeighElems
            if( intp%compact .or. force_compact ) then
!              foundDirInStencil = get_DirFrom
              ! If interpolation stencil /= compute stencil, try to match the
              ! offsets
              !\todo: move this into a function, as it is needed above in
              ! mapping compact->ext
              ! dir = map_compactToStencil( iNeigh, childNum,                  &
              !   &                                  scheme%layout%stencil(1) )
              x = tem_coordOfId( int(iNeigh, kind=long_k))*2 -1
              parentPos(:, iNeigh)  =                                          &
                &           matmul(tem_rotationMatrix(:,:, ChildNum ), x(1:3) )
              neighOffset  = (parentPos(:,iNeigh)  - parentPos(:, 1))/2
!              write(dbgUnit(10),*) iLoop,' original coord ', x(1:3)
!              write(dbgUnit(10),*) iLoop,' parentOffset   ',parentPos(:,iNeigh)
!              write(dbgUnit(10),*) iLoop,' neighOffset    ',neighOffset
!              write(dbgUnit(10),*) iLoop,' content of dep ',                   &
!                & scheme%glob%levelDesc( targetLevel )%depFromCoarser          &
!                &       (iElem )%elem%val( 1:scheme%glob%levelDesc             &
!                &       ( targetLevel )%depFromCoarser( iElem )%elem%nVals )

              ! Match the found neighOffset with the stencil. Is this direction
              ! available?
              foundDirInStencil = -1
              do iStencilElem = 1, scheme%layout%stencil(1)%QQ
                if( neighOffset(1) == scheme%layout%stencil(1)%cxDir( 1,     &
                  &                                       iStencilElem ) .and. &
                  & neighOffset(2) == scheme%layout%stencil(1)%cxDir( 2,     &
                  &                                       iStencilElem ) .and. &
                  & neighOffset(3) == scheme%layout%stencil(1)%cxDir( 3,     &
                  &                                        iStencilElem )) then
                  foundDirInStencil = iStencilElem
                end if
              end do
            else
              ! If interpolation stencil = compute stencil, simply take the
              ! offsets
              neighOffset = scheme%layout%stencil(1)%cxDir( :, iNeigh )
!              write(dbgUnit(10),*) iLoop,' Get offset from stencil ', iNeigh
              foundDirInStencil = iNeigh
            end if

            if( foundDirInStencil > 0) then
              ! Get the corresponding treeID from the stencil neighbor
              ! information
              distance(1:3) = real(neighOffset(1:3),kind=rk) - targetBary
              if(foundDirInStencil==scheme%layout%Stencil(1)%restPosition)then
                ! it's the parent element. -> no neighbor info required or
                ! available
                elemPos = sourceElem
              else
                ! it's a neihgbor of the parent element ->
                ! access neighbor information.
                elemPos = scheme%glob%levelDesc( sourceLevel )%neigh(1)%       &
                  &                   nghElems( foundDirInStencil, sourceElem )
              end if
              if( elemPos > 0 ) then
!                write(dbgUnit(10),*)iLoop,' valid Elem ',                      &
!                  &     scheme%glob%levelDesc( sourceLevel )%total( elemPos )

                validElement = .true.
              else
                validElement = .false.
!                write(dbgUnit(10),*)iLoop,' INVALID Elem ',                    &
!                  &     scheme%glob%levelDesc( sourceLevel )%total( elemPos )
              end if
              if( validElement ) then
                ! Add the found element to the dependency element list
                call append( me  = scheme%glob%levelDesc( targetLevel )%       &
                  &                              depFromCoarser( iElem )%elem, &
                  &          val = elemPos )
                dPos = scheme%glob%levelDesc( targetLevel )%depFromCoarser(    &
                  &                                          iElem )%elem%nVals
                ! Store to the levelwise source element list
                call append( me  = scheme%glob%levelDesc( targetLevel )%       &
                  &                                         sourceFromCoarser, &
                  &          val = elemPos,                                    &
                  &          pos = iPos )
                ! ... and store the position in the levelwise source element
                ! list for each ghost dep
                call append( me  = scheme%glob%levelDesc( targetLevel )%       &
                  &                         depFromCoarser( iElem )%elemBuffer,&
                  &          val = iPos )
!                write(dbgUnit(10),*)'tgtID loop ',iLoop, iNeigh,              &
!                  & ' dPos ', dPos,' nMaxElems ', nMaxElems
                if( intp%compact ) then
                  stencil( :, dPos ) = (parentPos( :, iNeigh) +1)/2
                else
                  stencil(:,dPos ) = neighOffset
                end if
                weight( dPos ) = (1._rk-abs(distance(1)))                      &
                  &            * (1._rk-abs(distance(2)))                      &
                  &            * (1._rk-abs(distance(3)))
              end if
            end if
          end do
          nSourceElems = scheme%glob%levelDesc( targetLevel )%depFromCoarser(  &
            &                                                iElem )%elem%nVals

          if ( nSourceElems < intp%nSourcesFromCoarser_required ) then
            completed = .false.
          else
            completed = .true.
          end if
        end do
        if( .not. completed .and. iLoop == 1) then
          write(logUnit(1),*) 'I got here although it shouldnt have happened ' &
            &               //'inside intp_update_FromCoarser'
          call tem_abort
        end if
        if ( nSourceElems < intp%nSourcesFromCoarser_required ) then
          ! not enough source elements -> assign to lower order interpolation
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                             intpFromCoarser_low, &
            &          val = iElem )
        else
          ! enough source elements -> assign to higher order interpolation
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                            intpFromCoarser_high, &
            &          val = iElem )
        end if
        allocate( scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%&
          &                                         stencil( 3, nSourceELems ))
        scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%          &
          &         stencil( :, 1:nSourceELems ) = stencil( :, 1:nSourceElems )
        ! Now I have found all the neighbors there are.
        ! Continue with a calculation of the weights
        allocate( scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem ) &
          &             %weight( 1:nSourceElems ))

        ! Normalize weight, so that sum(weight) = 1
        sumWeight = sum( weight( 1:nSourceElems ))
        weight( 1:nSourceElems ) = weight( 1:nSourceElems ) / sumWeight

        ! ... and store, but in the order according to the rotated coordinate
        ! system
        scheme%glob%levelDesc( targetLevel )%depFromCoarser( iElem )%          &
          &                 weight( 1:nSourceElems ) = weight( 1:nSourceElems )
        if( intp%requireWeights .or. force_compact) then
          if( abs( sum( scheme%glob%levelDesc( targetLevel )%                  &
            &    depFromCoarser( iElem )%weight( 1:nSourceElems )) - 1._rk )   &
            &      >= eps ) then

            write(logUnit(1),*)'Error: sum of interpolation weights not 1!'
            call tem_abort
          end if
        end if

      end do ! source%elem loop
      write(logUnit(1),"(A)")    ' fromCoarser Interpolation source element information'
      write(logUnit(1),"(A,I0)") '   targetLevel: ', targetLevel
      write(logUnit(1),"(A,I0)") '   number of ghost elements: ', &
        &               scheme%pdf( targetLevel )%nElems_ghostFromCoarser
      write(logUnit(1),"(A,I0)") '   higher order elements:    ', &
        &     scheme%glob%levelDesc( targetLevel )%intpFromCoarser_high%nVals
      write(logUnit(1),"(A,I0)") '   lower  order elements:    ', &
        &      scheme%glob%levelDesc( targetLevel )%intpFromCoarser_low%nVals

    end do ! level loop

end subroutine mus_intp_complete_fromCoarserDep
!******************************************************************************!


!******************************************************************************!
  !> Update the dependencies for ghostfromFiner
  !!
  subroutine mus_intp_update_fromFinerDep( intp, scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !>
    type( mus_scheme_type )                :: scheme
    !>
    type( interpolation_type ), intent(in) :: intp
    !>
    type( pdf_globInfo_type), intent(in)   :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: sourceLevel         ! level of source elements
    integer :: targetLevel         ! level of target elements
    integer :: iElem               ! current target element (for outer loop)
    integer :: iSourceElem         ! current source element
    integer :: nSourceElems        ! number of source elements
    integer :: dPos                ! storage position of source element
    !---------------------------------------------------------------------------

    do targetLevel = pdfGlobal%minLevel, pdfGLobal%maxLevel-1
      call init( me = scheme%glob%levelDesc( targetLevel )%sourceFromFiner,    &
        &        unique = .true. )
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromFiner_high )
      call init( me = scheme%glob%levelDesc( targetLevel )%intpFromFiner_low )
      sourceLevel = targetLevel + 1
      do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromFiner
        if( pdfGlobal%nDims == 2 ) then
          ! in two dimensional cases, remove the additional source elements
          ! for the fromFiner ghosts.
          ! There are always eight, provided by tem_construction routine.
          ! For the interpolation we only require 4.
          ! Only consider cases where more than 4 source elements are present
          if( scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%      &
            &                                          elem%nVals > 4 ) then
            scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%        &
              &                                                  elem%nVals = 4
          end if
        elseif( pdfGlobal%nDims == 1 ) then
          ! Accordingly in one dimensions: only 2 elements required to
          ! interpolate from finer
          if( scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%      &
            &                                          elem%nVals > 2 ) then
            scheme%glob%levelDesc( targetLevel )%depFromFiner( iElem )%        &
              &                                                  elem%nVals = 2
          end if
        end if
        nSourceElems = scheme%glob%levelDesc( targetLevel )%                   &
          &                                    depFromFiner( iElem )%elem%nVals
        call init( me     = scheme%glob%levelDesc( targetLevel )%              &
          &                                    depFromFiner(iElem)%elemBuffer, &
          &        length = 8 )
        do iSourceElem = 1, scheme%glob%levelDesc( targetLevel )%              &
          &                                    depFromFiner( iElem )%elem%nVals
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                                 sourceFromFiner, &
            &          val = scheme%glob%levelDesc( targetLevel )              &
            &                  %depFromFiner( iElem )%elem%val( iSourceElem ), &
            &          pos = dPos )
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                depFromFiner( iElem )%elemBuffer, &
            &          val = dPos )
        end do
        if ( nSourceElems < intp%nSourcesFromFiner_required ) then
          ! not enough source elements -> assign to lower order interpolation
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                               intpFromFiner_low, &
            &          val = iElem )
        else
          ! enough source elements -> assign to higher order interpolation
          ! scheme
          call append( me  = scheme%glob%levelDesc( targetLevel )%             &
            &                                              intpFromFiner_high, &
            &          val = iElem )
        end if
      end do ! source%elem loop
      write(logUnit(1),"(A)")   ' fromFiner Interpolation source element information'
      write(logUnit(1),"(A,I0)")'   targetLevel: ', targetLevel
      write(logUnit(1),"(A,I0)")'   higher order elements: ', &
        &        scheme%glob%levelDesc( targetLevel )%intpFromFiner_high%nVals
      write(logUnit(1),"(A,I0)")'   lower  order elements: ', &
        &        scheme%glob%levelDesc( targetLevel )%intpFromFiner_low%nVals
    end do

  end subroutine mus_intp_update_fromFinerDep
!******************************************************************************!


!******************************************************************************!
  !> Construct the propagation stencil for each element
  !!
  !! the bounce back rules have to be applied here
  !!
  subroutine reconstruct_connectivity( scheme, minLevel, maxLevel )
    !---------------------------------------------------------------------------
    !> Current scheme
    type( mus_scheme_type ) :: scheme
    !> Minlevel and maxLevel
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !---------------------------------------------------------------------------
    integer :: iElem  ! element counter for current element in treeID list
    integer :: iLevel ! level counter
    integer :: iField ! field counter
    integer :: idx_idir ! current direction in global varsys
    integer :: iDir ! counter for current direction in iElem
    integer :: invDir ! inverse direction for iDir
    ! result: from which direction to take the pdf from (inverse for bounceback)
    integer :: GetFromDir
    ! result: from which element to take the pdf from (local for bounceback)
    integer :: GetFromPos
    integer :: neighPos      ! position of current neighElem in treeID list
    integer(kind=long_k)  :: neighProp, elemProp
    !---------------------------------------------------------------------------

    write(logUnit(1),*) 'Reconstructing connectivity ...'

    ! JQ: each field may have several state variables
    ! first calculate number of state variables
    ! nStateVars = scheme%varSys%nVars / scheme%nFields

    levelLoop: do iLevel = minLevel, maxLevel
      elemLoop: do iElem = 1, scheme%pdf( iLevel )%nElems_local
        fieldLoop: do iField = 1, scheme%nFields

          ! pdfPos = (iField - 1) * nStateVars + 1
          ! get the element property
          elemProp = scheme%glob%levelDesc( iLevel )%property( iElem )

          ! Loop over every link
          neighLoop: do iDir  = 1, scheme%layout%stencil(1)%QQ
            ! check if the direction equals to the rest direction
            ! if available the rest direction is at position QQ of the stencil
            ! (e.g. the d3q6 stencil does not have a rest direction at all)
            if ( iDir == scheme%layout%stencil(1)%restPosition ) then
              ! Rest density takes from itself, iElem and iDir
              GetFromPos = iElem
              GetFromDir = scheme%varSys%Variable%val(iField)%varPos( iDir )
            else ! iDir not rest density

              ! For push we need to find different neighbors than for pull
              ! -> ngDir
              neighPos = scheme%glob%levelDesc( iLevel )%neigh(1)%            &
                &  nghElems( ?NgDir?( iDir, scheme%layout%stencil(1) ), iElem )

              neighProp = 0_long_k
              ! JQ: get neighbor's property
              if( neighPos > 0 ) then
                neighProp = scheme%glob%levelDesc( iLevel )%property(neighPos)
              end if

              !KM: Treat missing neighbor element as solid
              ! neighPos < 0 : neighbor is a boundary
              ! neighPos = 0 : neighbor is nonExisting element
              ! neighProp solid: neighbor was solidified or has missing neigh
              ! elemProp solid: myself was solidified or has missing neigh
              if( (neighPos <= 0) .or. (btest(neighProp, prp_solid))           &
                &  .or. (btest(elemProp, prp_solid))                           &
                &  .or. (btest(neighProp, prp_missingNeigh))                   &
                &  .or. (btest(elemProp, prp_missingNeigh))                    &
                & ) then
                ! 1. Fluid element has boundary in current link direction
                ! 2. neighbor element has fluidify property (still solid now)
                ! Get inverse direction of LOCAL element
                invDir = scheme%layout%stencil(1)%cxDirInv( iDir )
                GetFromDir = scheme%varSys%Variable%val(iField)%varPos(invDir)
                GetFromPos = iElem
                ! I set the source to the target link for halo cells
                ! with no neighbor in that direction
              else
                ! link to get from neighbor
                GetFromDir = scheme%varSys%Variable%val(iField)%varPos(iDir)
                GetFromPos = neighPos
              end if ! has no boundary -> regular treatment
            end if ! not the 0 velocity

            idx_idir = scheme%varSys%Variable%val(iField)%varPos(iDir)
            scheme%pdf( iLevel )%neigh(                                        &
              &          ?NGPOS?( idx_idir, iElem, scheme%varSys%nScalars )) = &
              &          ?IDX?( GetFromDir, GetFromPos, scheme%varSys%nScalars )

          end do neighLoop
        end do fieldLoop
      end do elemLoop
    end do levelLoop
    write(logUnit(1),*) 'Done with connectivity.'

  end subroutine reconstruct_connectivity
!******************************************************************************!


!******************************************************************************!
  !> Select the halo elements which require all links
  !!
  !! Determine which pdfs to communicate
  !!
  subroutine set_halo_commLinks( scheme, pdfGlobal, comm_reduced )
    !---------------------------------------------------------------------------
    !> Global information
    type(pdf_globInfo_type), intent(in)  :: pdfGlobal
    !> scheme information including fluid, boundary and flow information
    type(mus_scheme_type), intent(inout) :: scheme
    !> reduced communication
    logical                              :: comm_reduced
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, iSourceElem
    integer :: iBnd, iField, iNeigh, iProc
    integer :: targetElem, sourceElem, targetLevel, sourceLevel
    integer :: nSourceElems
    integer :: elemPos, haloOffset
    !---------------------------------------------------------------------------
    ! Allocate the halo comm data requirement array (logicals)
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      allocate(scheme%pdf(iLevel)%haloRequired(scheme%pdf(iLevel)%nElems_halo))
      scheme%pdf(iLevel)%haloRequired(:) = .false.
    end do
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      if( .not. comm_reduced ) then
        ! Communicate all links of all halos
        scheme%pdf(iLevel)%haloRequired(:) = .true.
      else ! Choose which ones are really required
        !-----------------------------------------------------------------------
        ! GhostFromFiner
        ! check if there are halo deps
        if( iLevel < pdfGlobal%maxLevel ) then
          targetLevel = iLevel
          sourceLevel = iLevel + 1
          haloOffset =scheme%glob%levelDesc( sourceLevel )%offset(1,eT_haloElem)

          do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromFiner
            ! position of coarser ghost element in levelDesc total list
            targetElem = iElem + scheme%glob%levelDesc( targetLevel )          &
              &                      %offset( 1, eT_ghostFromFinerElem )
            nSourceElems = scheme%glob%levelDesc( targetLevel )%               &
              &                                depFromFiner( iElem )%elem%nVals
            ! number of fine elements(childs)  
            do iSourceElem = 1, nSourceElems
              ! position of child element in levelDesc total list
              sourceElem = scheme%glob%levelDesc( targetLevel )%               &
                &                 depFromFiner( iElem )%elem%val( iSourceElem )
              call set_requiredLink(                                           &
                &          haloElem  = sourceElem,                             &
                &          offset    = haloOffset,                             &
                &          linkField = scheme%pdf( sourceLevel )%haloRequired )
            end do ! iSourceElem
          end do ! iElem
        end if !iLevel < maxLevel
        ! GhostFromFiner
        !-----------------------------------------------------------------------

        !-----------------------------------------------------------------------
        ! GhostFromCoarser
        ! check if there are halo deps
        if( iLevel < pdfGlobal%maxLevel ) then
          targetLevel = iLevel + 1
          sourceLevel = iLevel
          haloOffset =scheme%glob%levelDesc( sourceLevel )%offset(1,eT_haloElem)
          do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromCoarser
            ! position of ghost finer element in levelDesc total list
            targetElem = iElem + scheme%glob%levelDesc( targetLevel )          &
              &                      %offset( 1, eT_ghostFromCoarserElem )
            ! number of coarser elements that finer ghost dependent on.
            ! usually this is just 1.
            nSourceElems = scheme%glob%levelDesc( targetLevel )                &
              &            %depFromCoarser( iElem )%elem%nVals
            do iSourceElem = 1, nSourceElems
              sourceElem = scheme%glob%levelDesc( targetLevel )                &
                &          %depFromCoarser( iElem )%elem%val( iSourceElem )
              ! Check the type of the source element.
              ! If it is a halo -> mark as require all links
              call set_requiredLink(                                           &
                &          haloElem  = sourceElem,                             &
                &          offset    = haloOffset,                             &
                &          linkField = scheme%pdf( sourceLevel )%haloRequired )
            end do ! iSourceElem
          end do ! iElem
        end if !level<maxLevel
        ! GhostFromCoarser
        !-----------------------------------------------------------------------

        haloOffset =scheme%glob%levelDesc( iLevel )%offset(1, eT_haloElem )
        !-----------------------------------------------------------------------
        ! Run over recvbufferFromCoarser
        do iProc = 1, scheme%glob%levelDesc(iLevel)%recvBufferFromCoarser%nProcs
          do iElem = 1, scheme%glob%levelDesc( iLevel )%recvBufferFromCoarser% &
              &                                             nElemsProc( iProc )
            elemPos = scheme%glob%levelDesc( iLevel )%recvBufferFromCoarser%   &
              &                                   elemPos( iProc )%val( iElem )
            call set_requiredLink(                                             &
              &                 haloElem  = elemPos,                           &
              &                 offset    = haloOffset,                        &
              &                 linkField = scheme%pdf( iLevel )%haloRequired )
          end do
        end do
        !-----------------------------------------------------------------------

        !-----------------------------------------------------------------------
        ! Run over recvbufferFromFiner
        do iProc = 1, scheme%glob%levelDesc( iLevel )%recvBufferFromFiner%nProcs
          do iElem = 1, scheme%glob%levelDesc( iLevel )%recvBufferFromFiner%   &
            &                                               nElemsProc( iProc )
            elemPos = scheme%glob%levelDesc( iLevel )%recvBufferFromFiner%     &
              &                                   elemPos( iProc )%val( iElem )
            call set_requiredLink(                                             &
              &                 haloElem  = elemPos,                           &
              &                 offset    = haloOffset,                        &
              &                 linkField = scheme%pdf( iLevel )%haloRequired )
          end do
        end do
        !-----------------------------------------------------------------------

        !-----------------------------------------------------------------------
        ! Run over all the boundary neighbors
        ! If they require neighbor infos, it is usually integral quantities
        ! ( macrosopic)  and thus require all links
        do iField = 1, scheme%nFields
          do iBnd = 1, pdfGlobal%nBCs
            do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
              do iNeigh = 1, scheme%field( iField )%bc( iBnd )%nNeighs
                ! Get the neighbor source element
                sourceElem = scheme%field( iField )%bc( iBnd )                 &
                  &                     %elems( iLevel )%neigh( iNeigh, iElem )
                call set_requiredLink(                                         &
                  &             haloElem  = sourceElem,                        &
                  &             offset    = haloOffset,                        &
                  &             linkField = scheme%pdf( iLevel )%haloRequired )
              end do ! iNeigh
            end do ! iElem
          end do ! iBnd
        end do  ! iField
        ! Boundary neighbors
        !-----------------------------------------------------------------------
      end if ! comm_reduced?

    end do ! iLevel

  end subroutine set_halo_commLinks
!******************************************************************************!


!******************************************************************************!
  !> set haloRequired link to true 
  !!
  subroutine set_requiredLink( haloElem, offset, linkField )
    !---------------------------------------------------------------------------
    !> halo element position in the state array 
    integer, intent(in)    :: haloElem
    !> halo element offset in the state array
    !! nr. of elements in state array before halo elements
    integer, intent(in)    :: offset
    !> halo required array used in init_recvBuffers to reduce communication
    !! links
    logical, intent(inout) :: linkField(:)
    !---------------------------------------------------------------------------
    if( haloElem > offset ) then
      ! haloRequired is only nElems_halo long, so remove the preceeding
      ! elem nums
      linkField( haloElem - offset ) = .true.
    end if

  end subroutine set_requiredLink
!******************************************************************************!


!******************************************************************************!
  !> Update the neighbor stencil positions for the Boundaries
  !!
  !! Before, positions in the original treeIDs were assigned to
  !! the boundary element lists. Update these now with level-wise positions in
  !! the \ref tem_construction_module::level_descriptor_type "total list".
  !! When no element is found, the value in nghElems is 0. This has to be
  !! treated in this routine. If an empty neighbor relation is encountered, the
  !! last valid neighbor is used for all other neighbors
  !!
  subroutine update_BClists( scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !> Global information
    type(pdf_globInfo_type), intent(in)  :: pdfGlobal
    !> scheme information including fluid, boundary and flow information
    type(mus_scheme_type), intent(inout) :: scheme
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, iBnd, iField
    integer :: elemPos, iPos
    logical :: wallBnd(scheme%nFields)
    !---------------------------------------------------------------------------
    ! 1. Update the element positions from the treeID to now the
    ! levelDesc%total list
    ! 1a. Create dynamic array of bc_elemBuffer with position of
    ! treeID in the levelDescriptor%total list and store the position
    ! in globBC%elems(iLevel)%elembuffer
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      call init( me     = scheme%glob%levelDesc( iLevel )%bc_elemBuffer,       &
        &        unique = .true. )
      do iBnd = 1, pdfGlobal%nBCs
        ! set to true if BC kind specified for this boundary is wall
        wallBnd = .false.
        do iField = 1 ,scheme%nFields
          if(trim(scheme%field(iField)%bc(iBnd)%BC_kind) == 'wall') &
            & wallBnd(iField) = .true.
        end do

        ! if BC kind of all fields are wall then do nothing for this boundary
        ! i.e do not add this boundary to elemBuffer or bc_elemBuffer or
        ! neigh buffers since wall boundaries are treated explicitly as
        ! bounce back via pdf%neigh array
        if (.not. all(wallBnd)) then
          allocate( scheme%globBC( iBnd )%elems( iLevel )%elemBuffer(          &
            &                          scheme%globBC( iBnd )%nElems( iLevel )))

          ! allocate neighBuffer pre-collision and postcollision for each
          ! boundary in each field and initialize them
          ! \todo: SZ: think about this allocation in case there are more state
          !            variables than the QQ PDFs
          do iField = 1, scheme%nFields
              allocate( scheme%field( iField )%bc( iBnd )%                     &
                & elems( iLevel )%neighBufferPre(                              &
                &         scheme%field( iField )%bc( iBnd )%nNeighs,           &
                &         scheme%globBC( iBnd )%nElems( iLevel )*              &
                &         scheme%layout%stencil(1)%QQ ) )
              allocate( scheme%field( iField )%bc( iBnd )%                     &
                & elems( iLevel )%neighBufferPost(                             &
                &         scheme%field( iField )%bc( iBnd )%nNeighs,           &
                &         scheme%globBC( iBnd )%nElems( iLevel )*              &
                &         scheme%layout%stencil(1)%QQ ) )
              scheme%field(iField)%bc(iBnd)%elems(iLevel)%neighBufferPre =     &
                &                                                        -1.0_rk
              scheme%field(iField)%bc(iBnd)%elems(iLevel)%neighBufferPost=     &
                &                                                        -1.0_rk
          end do

          ! create bc_elemBuffer
          do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
            elemPos = pdfGlobal%levelPointer( scheme%globBC( iBnd )%           &
              &                                  elems( iLevel )%elem( iElem ))

            scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ) = elemPos

            ! store elemPos in bc_elemBuffer
            call append( me  = scheme%glob%levelDesc( iLevel )%bc_elemBuffer,  &
              &          val = elemPos,                                        &
              &          pos = iPos )
            ! store the position in the bc_elemBuffer in the elemBuffer
            scheme%globBC( iBnd )%elems( iLevel )%elemBuffer( iElem ) = iPos
          end do ! iElem in globBC%nElems on the current level

          ! elem and elemBuffer for corner nodes
          allocate( scheme%globBC( iBnd )%cornerBC%elems( iLevel )%elemBuffer( &
            &                 scheme%globBC( iBnd )%cornerBC%nElems( iLevel )))
          do iElem = 1, scheme%globBC( iBnd )%cornerBC%nElems( iLevel )
            elemPos = pdfGlobal%levelPointer( scheme%globBC( iBnd )%           &
              &                          cornerBC%elems( iLevel )%elem( iElem ))

            scheme%globBC( iBnd )%cornerBC%elems( iLevel )%elem( iElem )       &
              &                                                        = elemPos

            iPos = PositionOfVal( me  = scheme%glob%levelDesc( iLevel )%       &
              &                                                bc_elemBuffer,  &
              &                   val = elemPos )
            ! store the position in the bc_elemBuffer in the elemBuffer
            scheme%globBC( iBnd )%cornerBC%elems( iLevel )%                    &
              &                                     elemBuffer( iElem ) = iPos
          end do

        end if ! not all fields are wall Bnd
      end do ! iBnd

      write(dbgUnit(10),*) 'bc_elemBuffer nVals: ', &
        &                  scheme%glob%levelDesc( iLevel )%bc_elemBuffer%nVals

      ! allocate bcBuffer to store state vector values of all boundary elements
      ! for all fields in bcBuffer
      allocate( scheme%pdf( iLevel )%bcBuffer(                                 &
        &              scheme%glob%levelDesc( iLevel )%bc_elemBuffer%nVals     &
        &              *scheme%varSys%nScalars ) )
      scheme%pdf(iLevel)%bcBuffer = -1.0_rk
    end do !iLevel

    ! 2. build up neigh information for higher order boundaries
    ! loop over the field to update array boundary in each field
    do iField = 1, scheme%nFields
      do iBnd = 1, pdfGlobal%nBCs
        ! Copy the neighboring information to the bc%elem%neigh
        if( scheme%field( iField )%bc( iBnd )%nNeighs > 0 ) then
!write(dbgUnit(1),*) 'iBnd ', iBnd,  ' label ',trim(scheme%globBC( iBnd )%label)
          write(logUnit(3),'(A)') ' updating neigh array of boundary: '//trim(scheme%globBC( iBnd )%label)
          ! get position of neighbor element in the levelwise list
          call setFieldBCNeigh(                                       &
            &     fieldBC   = scheme%field( iField )%bc( iBnd ),      &
            &     globBC    = scheme%globBC( iBnd ),                  &
            &     levelDesc = scheme%glob%levelDesc(:),               &
            &     minLevel  = pdfGlobal%minLevel,                     &
            &     maxLevel  = pdfGlobal%maxLevel                      )
        end if !nNeigh>0
      end do ! iBnd
    end do ! iField

  end subroutine update_BClists
!******************************************************************************!


!******************************************************************************!
  !> This routine sets field BC neigh array with position of 
  !! neighbor element in the inward normal direction of boundary 
  !! in the levelwise list.
  !! if valid 1st neighbor does not exist return current element position.
  !! if valid higher order neighbor does not exist return last valid
  !! neighbor
  subroutine setFieldBCNeigh( fieldBC, globBC, levelDesc, minLevel, maxLevel )
    !---------------------------------------------------------------------------
    !> field boundary with boundary neighbor info
    type(boundary_type), intent(inout)    :: fieldBC
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(in)  :: globBC
    !> Level descriptor
    type(tem_levelDesc_type), intent(in)  :: levelDesc(minLevel:maxLevel)
    !> Level range
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, iCount, iNeigh
    integer :: neighVal
    !---------------------------------------------------------------------------

    ! count up the position where to store the neighborhood
    iCount = 0
    lvlLoop: do iLevel = minLevel, maxLevel
      write(logUnit(6),"(A,I0)") '  on level: ', iLevel
      allocate( fieldBC%elems( iLevel )%neigh( fieldBC%nNeighs,          &
        &                                      globBC%nElems( iLevel ) ) )

      elemLoop: do iElem = 1, globBC%nElems( iLevel )
        iCount = iCount + 1
        !write(dbgUnit(1),*) 'iElem ', iElem, ' iCount ', iCount
        !write(dbgUnit(1),*) 'elemID ', &
        !  & levelDesc(iLevel)%total(globBC%elems(iLevel)%elem(iElem))
        neighLoop: do iNeigh = 1, fieldBC%nNeighs
          ! neighbor position in the levelwise list
          neighVal = levelDesc( iLevel )%neigh( fieldBC%stencilPos )%   &
              &                            nghElems( iNeigh, iCount )
          !write(dbgUnit(1),*) 'iNeigh ', iNeigh
          !write(dbgUnit(1),*) 'neighval ', neighVal

          ! No valid neighbor found. set to boundary element pos if iNeigh = 1 
          ! else set last valid neigbor
          if ( neighVal < 1 ) then
            if ( iNeigh == 1 ) then
              !write(dbgUnit(1),*) 'setting current elemPos',&
              !                     globBC%elems( iLevel )%elem( iElem )

              ! The first neighbor already fails, so save simply current
              ! bnd elem pos (itself as neighbor)
              fieldBC%elems( iLevel )%neigh( :, iElem ) = &
                &     globBC%elems( iLevel )%elem( iElem )
            else
              !write(dbgUnit(1),*) 'neighbor fails taking last valid neighbor '
              !write(dbgUnit(1),*) 'iCount ', iCount
              !write(dbgUnit(1),*) 'setting last valid ', &
              !  &    levelDesc( iLevel )%neigh(          &
              !  &    fieldBC%stencilPos )%nghElems( iNeigh-1, iCount )

              ! Neighbor fails, so take last neighbor which was valid
              fieldBC%elems( iLevel )%neigh( iNeigh:fieldBC%nNeighs, iElem ) = &
                &              fieldBC%elems( iLevel )%neigh( iNeigh-1, iElem )
            end if ! iNeigh = 1
            exit neighLoop
          else
            !write(dbgUnit(1),*) 'setting neighVal ', neighVal

            ! regular neighbor: assign
            fieldBC%elems( iLevel )%neigh( iNeigh, iElem ) = neighVal
          end if ! neighVal < 1
        end do neighLoop

        !write(dbgUnit(1),*) 'all neighID ', &
        !  & levelDesc(iLevel)%total(fieldBC%elems(iLevel)%neigh(:,iElem))
      end do  elemLoop
    end do  LvlLoop
  end subroutine setFieldBCNeigh
!******************************************************************************!


!******************************************************************************!
  !> Assemble the level-wise list of elements which adhere to the boundary
  !! conditions.
  !!
  !! The boundaries will then be treated for each level one by one, running over
  !! the list of elements.
  !! The bitmasks are set for the directions pointing into the domain.
  !! In the LBM, the incoming densities have to be updated.
  !! As boundaries are being set before the kernel, the state arrays have to be
  !! stored to the FETCH position
  !!
  subroutine build_BClists( globBC, tree, boundary, property, pdfGlobal,       &
    &                       layout, field, proc )
    !---------------------------------------------------------------------------
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)                  :: tree
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)                :: boundary
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)               :: pdfGlobal
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in)          :: layout
    !> element property type
    type( tem_property_type ), intent(inout)            :: property(:)
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(out), allocatable  :: globBC(:)
    !> field type
    type( mus_field_type ), intent(in)                  :: field(:)
    !> mpi communication enviroment with mpi communicator
    type(tem_comm_env_type), intent(in)                 :: proc
    !---------------------------------------------------------------------------
    ! number of elem for each BC on each lvl
    integer, allocatable :: nBnds(:,:), nCornerBnds(:,:)
    integer :: iBnd, iField, iProperty
    !---------------------------------------------------------------------------

    ! property 1 defines has boundary
    iProperty = 1

    write(logUnit(1),*) 'Creating boundary lists ...'

    if( boundary%nBCtypes > 0 ) then

      ! allocate bc states for each boundary with size of nFields
      ! to collect bc states of all fields of this particular boundary
      ! into one global list. Needed when boundary condition is
      ! dependent on all fields.
      ! contains element information about each boundary type
      allocate( globBC( pdfGlobal%nBCs ) )
      do iBnd = 1, pdfGlobal%nBCs
        allocate(globBC(iBnd)%bc_states(size(field)) )
        ! copy the field wise bc_states to the global_boundary_type
        do iField = 1, size(field)
          globBC(iBnd)%bc_states(iField) = field(iField)%bc(iBnd)%bc_states
        end do

        ! copy bc label from mesh boundary to solver globBC
        globBC( iBnd )%label   = boundary%BC_label( iBnd )
      end do   

      ! Initialize Qval info
      ! boundary%hasQval is allocated only if prp_hasQval is set in seeder
      ! so initialize them first and use globBC(:)%hasQval in rest of the 
      ! routine
      globBC(:)%hasQVal = .false.
      if( allocated(boundary%hasQval) ) then
        do iBnd = 1, pdfGlobal%nBCs
          globBC(iBnd)%hasQVal = boundary%hasQval(iBnd)
        enddo
      endif

      ! 1. Identify the number of boundary condition elements of each BC type
      call countnBnds( globBC      = globBC,              &
        &              boundary    = boundary,            &
        &              property    = property(iProperty), &
        &              tree        = tree,                &
        &              layout      = layout,              &
        &              pdfGlobal   = pdfGlobal,           &
        &              nBnds       = nBnds,               &
        &              nCornerBnds = nCornerBnds,         &
        &              proc        = proc                 )

      ! 2. Allocate the BC lists
      call allocateBCList( globBC      = globBC,          &
        &                  pdfGlobal   = pdfGlobal,       &  
        &                  layout      = layout,          &
        &                  nBnds       = nBnds,           &
        &                  nCornerBnds = nCornerBnds )

      ! 3. Assign the BC lists
      ! Run over all the elements with the property boundary
      ! and check each direction.
      ! Assign all common boundaries to the level-wise representation
      call assignBCList( tree        = tree,                &
        &                pdfGlobal   = pdfGlobal,           &
        &                layout      = layout,              &
        &                boundary    = boundary,            &
        &                property    = property(iProperty), &
        &                globBC      = globBC               )



      ! 4 Normalize the normal vectors for regular and corner elements
      !   and choose the corresponding prevailing direction from the stencil
      call normalizeBC( pdfGlobal      = pdfGlobal,         &
        &               globBC         = globBC,            &
        &               layout         = layout )

    end if ! nBCs > 0 ?

  end subroutine build_BClists
!******************************************************************************!


!******************************************************************************!
  !> Identify the number of boundary condition elements of each BC type
  !! and number of elements with multiple BC types.
  !! This is 1st step in Build_BClists
  subroutine countnBnds( globBC, boundary, property, tree, layout, pdfGlobal,  &
    &                    nBnds, nCornerBnds, proc )
    !---------------------------------------------------------------------------
    !> element property type
    type( tem_property_type ), intent(in)      :: property
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)       :: boundary
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)         :: tree
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)      :: pdfGlobal
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(inout)    :: globBC(:)
    !> number of local elements of each boundary type in each level
    integer, intent(out), allocatable          :: nBnds(:,:)
    !> number of local elements with multiple boundary type in each level 
    integer, intent(out), allocatable          :: nCornerBnds(:,:)
    !> mpi communication enviroment with mpi communicator
    type(tem_comm_env_type), intent(in)        :: proc
    !---------------------------------------------------------------------------
    integer :: iElem, iDir, iBnd, level, iMeshDir, iErr
    logical :: found( boundary%nBCtypes )
    integer(kind=long_k) :: bID
    integer, allocatable :: nBnds_total(:,:)
    !---------------------------------------------------------------------------
    ! number of local elements of each boundary type in each level
    allocate( nBnds( globalMaxLevels, boundary%nBCtypes ))
    ! number of total elements(all processes) of each boundary type in each level
    allocate( nBnds_total( globalMaxLevels, boundary%nBCtypes ))
    ! number of elements with multiple boundaries in each level
    allocate( nCornerBnds( globalMaxLevels, boundary%nBCtypes ))

    nBnds       =  0
    nCornerBnds = 0

    do iElem = 1, property%nElems
      found = .false.
      level = tem_LevelOf( tree%treeID( property%ElemID(iElem)))
      do iDir = 1, layout%stencil(1)%QQN
        iMeshDir = layout%stencil(1)%map( iDir )
        ! the index in boundary_ID does not correspond to the stencil!
        bID = boundary%boundary_ID( iMeshDir, iElem )
        if( bID > 0  ) then
          ! checks if number of boundary IDs dumped by seeder is not
          ! greater than number of boundary types in bnd.lua
          if( bID > pdfGlobal%nBCs ) then
            write(logUnit(1),*) ' found BID:     ', bID
            write(logUnit(1),*) ' where the highest should be ',             &
              &                 pdfGlobal%nBCs
            write(logUnit(1),*) ' at element     ', iElem
            write(logUnit(1),*) ' found higher bID than nr of bcs exist.'
          end if

          if ( .not. found ( bID ) ) then
            ! Increase the counter only, if this bID has not been
            ! identified before at iElem
            nBnds( level, bID ) = nBnds( level, bID )+1
            found( bID ) = .true.
          end if ! found ( bID ) == .false.
        end if ! bid > 0 ?
      end do ! iDir
      ! if element is intersected by more than one boundary
      if ( count(found) > 1 ) then
        do iBnd = 1, pdfGlobal%nBCs
          if (found(iBnd)) then
            nCornerBnds(level, iBnd) = nCornerBnds(level, iBnd) + 1
          endif
        enddo
      endif !count > 1 for corner nodes
    end do ! iElem

    ! Counting number of boundary elements on local partition and global
    ! Also copies field bcstates to globBC bcstates to access all
    ! field bcstates in each boundary routines
    do iBnd = 1, pdfGlobal%nBCs
      ! number of elements for each iBnd
      globBC( iBnd )%nElems( : ) = nBnds( :, iBnd )
      ! Sum up the elements of the local partition for each iBnd
      globBC( iBnd )%nElems_local = sum(                &
        &                    globBC( iBnd )%nElems( : ) )

      ! number of elements with multiple boundaries
      globBC( iBnd )%cornerBC%nElems( : ) = nCornerBnds( :, iBnd )
    end do ! iBnd

    ! calculate total number of boundary elements in all levels and processes
    call mpi_allreduce( nBnds, nBnds_total, globalMaxLevels*pdfGlobal%nBCs,  &
      &                 mpi_integer, mpi_sum, proc%comm, iErr                )

    do iBnd = 1, pdfGlobal%nBCs
      globBC( iBnd )%nElems_totalLevel( : ) = &
        &                                          nBnds_total( :, iBnd )
      globBC( iBnd )%nElems_total = sum(                          &
        &                    globBC( iBnd )%nElems_totalLevel( : ))
    end do

   end subroutine countnBnds  
!******************************************************************************!


!******************************************************************************!
  !> Allocate BC lists, 2nd step in Build_BClists
  subroutine allocateBCList( globBC, pdfGlobal, layout, nBnds, nCornerBnds )
    !---------------------------------------------------------------------------
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(inout)    :: globBC(:)
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)      :: pdfGlobal
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> number of local elements of each boundary type in each level
    integer, intent(in)                        :: nBnds(:,:)
    !> number of local elements with multiple boundary type in each level 
    integer, intent(in)                        :: nCornerBnds(:,:)
    !---------------------------------------------------------------------------
    integer :: iBnd, iLevel
    !---------------------------------------------------------------------------
    do iBnd = 1, pdfGlobal%nBCs
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        ! in gcc-ompi/4.7... this setting caused probems in the build mode
        ! assigning the whole array like done above works
        ! stores boundary element position in the levelDesc%total
        allocate( globBC( iBnd )%elems( iLevel )%elem(             &
          &                                        nBnds( iLevel, iBnd )))

        ! bitmask is set on the opposite (incoming) direction to the boundary
        allocate( globBC( iBnd )%elems( iLevel )%bitmask(          &
          &           layout%stencil(1)%QQ, nBnds( iLevel, iBnd )))

        ! allocate qVal for each level for boudnary which has qVals
        if( globBC( iBnd )%hasQVal ) then
          allocate( globBC( iBnd )%elems( iLevel )%qVal(           &
            &         layout%stencil(1)%QQ, nBnds( iLevel, iBnd )))
        end if

        ! opposite (incoming) normal direction to the boundary for each 
        ! element
        allocate( globBC( iBnd )%elems( iLevel )%normal(           &
          &                                     3, nBnds( iLevel, iBnd )))

        ! maps opposite boundary normal to layout(1)%stencil directions
        allocate( globBC( iBnd )%elems( iLevel )%normalInd(        &
          &                                        nBnds( iLevel, iBnd )))
        if( nBnds( iLevel, iBnd ) > 0) then
          write(logUnit(10),"(A,A,I2,A,I0)") ' BC '//trim(globBC(iBnd)%label), &
            & ', level: ', iLevel, &
            & ', nElems: ', globBC( iBnd )%nElems( iLevel )
          globBC( iBnd )%elems( iLevel )%elem(:) = 0
          globBC( iBnd )%elems( iLevel )%bitmask = .false.
          globBC( iBnd )%elems( iLevel )%normal  = 0
        end if

        ! elements with multiple boundaries
        allocate( globBC(iBnd)%cornerBC%elems( iLevel ) &
          &                            %elem( nCornerBnds( iLevel, iBnd )))
        allocate( globBC(iBnd)%cornerBC%elems( iLevel )%bitmask( &
          &      layout%stencil(1)%QQ, nCornerBnds( iLevel, iBnd )))
        allocate( globBC(iBnd)%cornerBC%elems( iLevel )%normal( &
          &      3, nCornerBnds( iLevel, iBnd )))
        allocate( globBC(iBnd)%cornerBC%elems( iLevel )%normalInd( &
          &      nCornerBnds( iLevel, iBnd )))
        if (nCornerBnds(iLevel, iBnd) > 0 ) then
          globBC(iBnd)%cornerBC%elems( iLevel )%elem(:) = 0
          globBC(iBnd)%cornerBC%elems( iLevel )%bitmask = .false.
          globBC(iBnd)%cornerBC%elems( iLevel )%normal  = 0
        endif

      end do !iLevel
    end do !iBnd
  end subroutine allocateBCList

!******************************************************************************!



!******************************************************************************!
  !> This routine assigns the BC lists
  !! Run over all the elements with the property boundary
  !! and check each direction.
  !! Assign all common boundaries to the level-wise representation
  !! 3rd step in build_BCLists
  subroutine assignBCList( tree, pdfGlobal, layout, boundary, property,        &
      &                    globBC )
    !---------------------------------------------------------------------------
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)         :: tree
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)      :: pdfGlobal
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)       :: boundary
    !> element property type
    type( tem_property_type ), intent(in)      :: property
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(inout)    :: globBC(:)
    !---------------------------------------------------------------------------
    integer :: iElem, iDir, iBnd, IMeshDir
    integer(kind=long_k) :: bID
    integer :: normal(3), level
    logical :: corner_bitmask(layout%stencil(1)%QQ)
    logical :: found( boundary%nBCtypes )
    integer :: iElem_qVal
    integer :: weight, length
    ! number of elem for each BC on each lvl
    integer, allocatable :: nBnds(:,:), nCornerBnds(:,:)
    !---------------------------------------------------------------------------
    ! number of local elements of each boundary type in each level
    allocate( nBnds( globalMaxLevels, boundary%nBCtypes ))
    ! number of elements with multiple boundaries in each level
    allocate( nCornerBnds( globalMaxLevels, boundary%nBCtypes ))
    !Reset the counters
    nBnds       = 0
    nCornerBnds = 0
    iElem_qVal  = 0

    do iElem = 1, property%nElems
      ! Count the elem with qVal property to access the qVal from
      ! boundary%qVal
      if( btest(tree%elemPropertyBits( Property%ElemID(iElem)), prp_hasQVal) ) &
        & iElem_qVal = iElem_qVal + 1

      found = .false.
      level = tem_LevelOf( tree%treeID( property%ElemID( iElem )))
      ! run over each direction in the stencil
      do iDir = 1, layout%stencil(1)%QQN
        ! Stencil order of directions might differ from
        ! the treelm numbering. -> look up correct treelm direction number
        iMeshDir = layout%stencil(1)%map( iDir )
        ! Get the boundary ID as stored in the mesh
        bID      = boundary%boundary_ID( iMeshDir, iElem )
        ! Exclude no-boundary or periodic boundaries
        if( bID > 0  ) then
          ! If this boundary ID was not found for current element yet,
          ! assign to list
          if ( .not. found ( bID ) ) then
            ! count up and add, only if it has not been added before for iElem
            ! this is before all the helper elements have been identified
            ! therefore only the pointers to the original treeID list are
            ! added. This is done later in the update_BC routine.
            ! Wouldn't it be better to have a dynamic array here instead of
            ! doing it manually
            nBnds( level, bID ) = nBnds( level, bID )+1

            if( .not. allocated( globBC(bID)%elems(level)%elem))then
              write(dbgUnit(1),*) 'Error: bc not allctd lvl ', level,        &
                &                     ' bid ', bID
            end if
            ! Assign the position of the treeID in tree%treeID list 
            ! to the preliminary boundary lists
            globBC( bID )%elems( level )%elem( nBnds( level, bID )) = &
              &                         property%ElemID( iElem )
            found( bID ) = .true.
          end if
          
          ! Set the bitmask for the incoming directions which have to be
          ! updated
          globBC( bID )%elems( level )%bitmask(                       &
            ! Set bitmask for incoming direction, that is opposite
            ! of the one stored in the treelmesh
            & layout%stencil(1)%cxDirInv(iDir),                       &
            ! nBnds points to current boudnary element
            & nBnds( level, bID ) ) = .true.

          if( globBC( bID )%hasQVal ) then
            ! Set the qvalue from the boundary array
            ! q-values stored in outgoing direction
            globBC( bID )%elems( level )%qVal( iDir, nBnds(level,bID))&
              &                       = boundary%qVal( iMeshDir, iElem_qVal )
          end if

          ! Get length of currently treated direction and determine
          ! If it is one of the main coordinate axis
          length =   layout%stencil(1)%cxDir( 1, iDir )**2 &
            &      + layout%stencil(1)%cxDir( 2, iDir )**2 &
            &      + layout%stencil(1)%cxDir( 3, iDir )**2

          ! For the normals, we weight the orthogonal directions
          ! of the stencil more heavily
          if( length == 1) then
            weight = 2  ! axis orthogonal weighs twice as much as all others
          else
            weight = 1  ! non-axis orthogonal
          end if

          ! Add this boundary direction to the sum for the normal vector
          ! use the minus, as we want to go into the flow domain, and the
          ! stencil is pointing outside
          globBC( bID )%elems( level )%normal(:, nBnds( level, bID ))        &
            &  =  globBC( bID )%elems( level )%normal(:, nBnds( level, bID ))&
            &  -  weight*layout%stencil(1)%cxDir( :, iDir )
        end if
      end do  ! iDir
      ! if element is intersected by more than one boundary
      if ( count(found) > 1 ) then
        corner_bitmask = .false.
        normal = 0
        do iBnd = 1, pdfGlobal%nBCs
          if (found(iBnd)) then
            nCornerBnds( level, iBnd ) = nCornerBnds( level, iBnd ) + 1
            ! Assign the position of the treeID to the corner elem list
            globBC( iBnd )%cornerBC%elems( level )              &
              & %elem( nCornerBnds( level, iBnd ) ) =           &
              &         property%ElemID( iElem )

            corner_bitmask = corner_bitmask .or. &
              & globBC( iBnd )%elems(level)%bitmask(:, nBnds(level, iBnd))
             normal = normal + &
              & globBC( iBnd )%elems(level)%normal(:, nBnds( level, iBnd ))
          endif
        end do

        do iBnd = 1, pdfGlobal%nBCs
          if (found(iBnd)) then
            globBC( iBnd )%cornerBC%elems( level )       &
              & %bitmask( :, nCornerBnds( level, iBnd ) ) = corner_bitmask
            globBC( iBnd )%cornerBC%elems( level )       &
              & %normal( :, nCornerBnds( level, iBnd ) )  = normal
          endif
        end do
      endif !count > 1
    end do ! iElem

  end subroutine assignBCList

!******************************************************************************!


!******************************************************************************!
  !> This routine normalizes the normal vectors of boundary elements including
  !! the corner elements as well as assigns the corresponding prevailing 
  !! direction from the stencil
  subroutine normalizeBC( pdfGlobal, globBC, layout )
    !---------------------------------------------------------------------------
    !> contains pdf global information
    type( pdf_globInfo_type ), intent(in)      :: pdfGlobal
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(inout)    :: globBC(:)
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in) :: layout
    !---------------------------------------------------------------------------
    integer :: iBnd, iLevel, iElem, iDir
    !---------------------------------------------------------------------------
    ! normalize all boundary normals
    do iBnd = 1, pdfGlobal%nBCs
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        do iElem = 1, globBC( iBnd )%nElems( iLevel )
          call tem_determine_discreteVector(                         &
            &   globBC( iBnd )%elems( iLevel )%normal( :, iElem ),   &
            &   layout%prevailDir )
          ! Find the index in the stencil corresponding to the normal
          ! direction
          do iDir = 1, layout%stencil(1)%QQ
            if( layout%stencil(1)%cxDir(1, iDir ) ==                 &
              & globBC( iBnd )%elems( iLevel )%normal( 1, iElem )    &
              & .and.                                                &
              & layout%stencil(1)%cxDir(2, iDir ) ==                 &
              & globBC( iBnd )%elems( iLevel )%normal( 2, iElem )    &
              & .and. layout%stencil(1)%cxDir(3, iDir ) ==           &
              & globBC( iBnd )%elems( iLevel )%normal( 3, iElem ))   &
              & then
              globBC( iBnd )%elems( iLevel )%normalInd( iElem ) = iDir
            end if
          end do
        end do  ! iElem
      end do  ! iLevel
    end do ! iBnd

    ! normalize normal of elements with multiple boundaries
    do iBnd = 1, pdfGlobal%nBCs
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        do iElem = 1, globBC(iBnd)%cornerBC%nElems( iLevel )
          call tem_determine_discreteVector(                                   &
            &   globBC( iBnd)%cornerBC%elems( iLevel )%normal( :, iElem ),     &
            &   layout%prevailDir)
          ! Find the index in the stencil corresponding to the normal
          ! direction
          do iDir = 1, layout%stencil(1)%QQ
            if( layout%stencil(1)%cxDir(1, iDir) ==                            & 
              & globBC(iBnd)%cornerBC%elems(iLevel)%normal(1, iElem)           &
              & .and.                                                          &
              & layout%stencil(1)%cxDir(2, iDir) ==                            &
              & globBC(iBnd)%cornerBC%elems(iLevel)%normal(2, iElem)           &
              & .and.                                                          &
              & layout%stencil(1)%cxDir(3, iDir) ==                            &
              & globBC(iBnd)%cornerBC%elems(iLevel)%normal(3, iElem))          &
              & then
            globBC(iBnd)%cornerBC%elems(iLevel)%normalInd(iElem) = iDir
            end if
          end do
        end do ! iElem
      end do ! iLevel
    end do ! iBnd

  end subroutine normalizeBC
!******************************************************************************!


!******************************************************************************!
  !> Initialize myHalos and remoteHalos bitmasks.
  !! myHalos bitmask are from my halo elements which are to be
  !! received from remote process and remoteHalos bitmask are 
  !! remote process halos which are requested by remote process.
  subroutine init_bitmask( myHalos, remoteHalos, send, recv, nScalars )
    !---------------------------------------------------------------------------
    !> contains bitmask of my halos with array size of recv%nProcs
    type(halo_commBitmask_type), allocatable, intent(out) :: myHalos(:)
    !> contains bitmask of remote halos with array size send%nProcs
    type(halo_commBitmask_type), allocatable, intent(out) :: remoteHalos(:)
    !> Communication structure to initialize
    type(tem_communication_type), intent(in)              :: send, recv
    !> number of scalars in variable system to communicate
    integer                                               :: nScalars
    !---------------------------------------------------------------------------
    integer :: iProc
    !---------------------------------------------------------------------------
    ! number of process my halos depends on
    allocate(myHalos(recv%nProcs))
 
    ! number of remote process contains halo elements which are my 
    ! local elements
    allocate(remoteHalos(send%nProcs))

    ! Allocate the bitmask for elements to be communicated
    ! to the target proc
    do iProc = 1, send%nProcs
      allocate( remoteHalos(iProc)%bitmask( nScalars, send%nElemsProc(iProc) ) )
      remoteHalos(iProc)%bitmask = .false.
    end do

    ! Allocate the bitmask for elements to be communicated
    ! from the provider proc
    do iProc = 1, recv%nProcs
      allocate( myHalos(iProc)%bitmask( nScalars, recv%nElemsProc(iProc) ) )
      myHalos(iProc)%bitmask = .false.
    end do

  end subroutine init_bitmask
!******************************************************************************!


!******************************************************************************!
  subroutine communicate_bitmask( myHalos, remoteHalos, send, recv, proc,      &
    &                             nScalars )
    !---------------------------------------------------------------------------
    !> contains bitmask of my halos with array size of recv%nProcs
    type(halo_commBitmask_type), intent(in)     :: myHalos(:)
    !> contains bitmask of remote halos with array size send%nProcs
    type(halo_commBitmask_type), intent(inout)  :: remoteHalos(:)
    !> Communication structure to initialize
    type(tem_communication_type), intent(in)    :: send, recv
    !> mpi communication enviroment with mpi communicator
    type(tem_comm_env_type), intent(in)         :: proc
    !> number of scalars in variable system to communicate
    integer                                     :: nScalars
    !---------------------------------------------------------------------------
    integer :: iProc
    !------ MPI variables
    ! send%nProcs + recv%nProcs )
    integer,allocatable :: rq_handle(:)
    ! mpi_status_size, recv%nProcs + send%nProcs )
    integer,allocatable :: status(:,:)
    integer :: iErr
    integer :: message_flag
    !---------------------------------------------------------------------------
    ! Choosing random message flag
    message_flag = 1
    allocate( rq_handle( send%nProcs + recv%nProcs))
    allocate( status( mpi_status_size, send%nProcs + recv%nProcs))

    ! First start receive communication of the 'send' bitmasks from the target
    ! Procs. Consumer view: receive from the target Process
    do iProc = 1, send%nProcs
      call mpi_irecv(                              &
        &      remoteHalos( iProc )%bitMask,       & ! me
        &      send%nElemsProc( iProc )*nScalars,  & ! me size
        &      MPI_LOGICAL,                        & ! data type
        &      send%proc( iProc ),                 & ! target me
        &      message_flag,                       & ! flag
        &      proc%comm,                          & ! communicator
        &      rq_handle(iProc),                   & ! handle
        &      iErr  )                               ! error status
    end do

    ! then send the 'recv' bitmask arrays to the neighbor procs
    ! Provider view: send to the source processes
    do iProc = 1, recv%nProcs
      call mpi_isend(                              &
       &      myHalos( iProc )%bitMask,            & ! me
       &      recv%nElemsProc( iProc )*nScalars,   & ! me size
       &      MPI_LOGICAL,                         & ! data type
       &      recv%proc( iProc ),                  & ! target me
       &      message_flag,                        & ! flag
       &      proc%comm,                           & ! communicator
       &      rq_handle( iProc + send%nProcs ),    & ! handle
       &      iErr                    )              ! error status
    end do ! iProc

    ! Wait for above communications to complete
    call mpi_waitall(               &
      & send%nProcs+recv%nProcs,    & ! total number of comm.'s to wait for
      & rq_handle,                  & ! request handles
      & status,                     & ! mpi status
      & iErr         )                ! error status
    !--------------------------------------------------------------------------
    ! Status: We received the bitmasks from the Provider processes,
    !         from which we can now construct the correct send buffers
    !--------------------------------------------------------------------------

  end subroutine communicate_bitmask
!******************************************************************************!


!******************************************************************************!
  !> Destroy allocated send and recv bitmask buffer
  subroutine destroy_bitmask( myHalos, remoteHalos, send, recv )
    !---------------------------------------------------------------------------
    !> contains bitmask of my halos with array size of recv%nProcs
    type(halo_commBitmask_type), allocatable, intent(inout) :: myHalos(:)
    !> contains bitmask of remote halos with array size send%nProcs
    type(halo_commBitmask_type), allocatable, intent(inout) :: remoteHalos(:)
    !> Communication structure to initialize
    type(tem_communication_type), intent(in)                :: send, recv
    !---------------------------------------------------------------------------
    integer :: iProc
    !---------------------------------------------------------------------------
    do iProc = 1, send%nProcs
      deallocate( remoteHalos( iProc )%bitmask )
    end do  
    do iProc = 1, recv%nProcs
      deallocate( myHalos( iProc )%bitmask )
    end do  

    deallocate( remoteHalos )
    deallocate( myHalos )

  end subroutine destroy_bitmask
!******************************************************************************!


!******************************************************************************!
  !> Create the communication buffers
  !!
  !! Assign the positions in state vectors, where to fetch from or save to
  !! Solver specific part to identify the actual data locations
  !! to send use within the communication buffers.
  !! \todo: this should be pattern specific, with typed exchange we
  !!        do not need to store the pos, as it is done inside MPI
  !!        with the defined datatype
  !!
  !! Access to state array: SAVE
  !! ---------------------------
  !! Why? Because we need to work on the pdfs on which the
  !! compute kernel was working. The kernel stored to SAVE
  !! All actions taking place on the state array after the kernel
  !! need to be performed with SAVE access
  !!
  subroutine init_recvBuffers( me, myHalos, pdf, nDOFs, pattern, requireAll,   &
    &                          levelDesc, scheme, stat )
    !---------------------------------------------------------------------------
    !> receive communication buffer
    type(tem_communication_type), intent(inout) :: me
    !> contains bitmask of my halos with array size of recv%nProcs
    type(halo_commBitmask_type), intent(inout)  :: myHalos(:)
    !> communication pattern
    type(tem_commPattern_type), intent(in)      :: pattern
    !> contains solver specific arrays like state and neigh
    type(pdf_data_type), intent(in)             :: pdf
    !> how many links per element? QQ!
    integer, intent(in)                         :: nDOFs
    !> requires complete element information or only required links?
    logical, intent(in)                         :: requireAll
    !> Level descriptor
    type(tem_levelDesc_type), intent(in)        :: levelDesc
    !> Scheme information on fluid, boundary and flow properties
    type( mus_scheme_type ), intent(in)         :: scheme
    !> Statistics
    type(mus_statistics_type), intent(inout)    :: stat
    !---------------------------------------------------------------------------
    ! positions, in state vectors, of links with fluid neighbors
    integer, allocatable :: pos(:)
    integer :: iElem ! count variable for elements
    integer :: iDir, iProc, elemPos
    integer :: counter ! number of total links that acturally needs to update
    integer :: iField, nScalars, varPos(nDOFs)
    ! Include the current direction for the communciation
    logical :: includeInComm
    integer :: nghElem ! neighbor element position in total list
    integer :: neighPos ! position of neighbor in the neigh array list
    !---------------------------------------------------------------------------

    nScalars = scheme%varSys%nScalars
    allocate(pos(maxval(me%nElemsProc*nScalars)))

    allocate( me%buf_real( me%nProcs ))

    do iProc = 1, me%nProcs
      counter = 0
      do iElem = 1, me%nElemsProc( iProc )
        ! halo element position in the levelDesc total list
        elemPos = me%elemPos( iProc )%val( iElem )
        ! Here the check for the bitmasks has to be performed
        do iField = 1, scheme%nFields
          varPos = scheme%varSys%Variable%val( iField )%varPos
          do iDir = 1, nDOFs
            includeInComm = .false.
            ! Now check all the neighbors of the current halo element
            ! - If we encounter fluid or ghost element, then the direction is
            !   required to exchange for      -> bitmask( iDir, iElem ) = .true.
            ! - If a halo element is ecountered, there is no need for
            !   communication                -> bitmask( iDir, iElem ) = .false.
            ! Get the element position of the neighbor element
            ! By accessing the neighbor array, get the position of the neighbor
            ! element's direction and reconstruct the neighbor element position
            ! in the total List
            neighPos =                                                         &
& ?NGPOS?( varPos(?NgDir?(iDir,scheme%layout%stencil(1))), elemPos, nScalars )
            nghElem = ?ElemIDX?( pdf%neigh( neighPos ), nScalars)
            ! Now check, what kind of element it is. Must be anything but halo.
            ! requireAll is true only for IBM stencil
            if( nghElem <= levelDesc%offset( 1, eT_haloElem ) .or.           &
              &                                               requireAll ) then
              includeInComm = .true.
            end if

            ! If it's a ghostHalo or boundary halo, then communicate all
            if( pdf%haloRequired( elemPos - levelDesc%                       &
              &                               offset( 1, eT_haloElem ))) then
              includeInComm = .true.
            end if

            ! If current neighbor is a ghosthalo or boundary halo, 
            ! then communicate all
            if( nghElem > levelDesc%offset( 1, eT_haloElem ) ) then
              if( pdf%haloRequired( nghElem  - levelDesc%                      &
                &                               offset( 1, eT_haloElem ))) then
                includeInComm = .true.
              end if
            end if

            if( includeInComm ) then
              stat%nLinks_comm = stat%nLinks_comm + 1
              counter = counter + 1
              ! ? SAVE ? ... for regular halos
              ! ? IDX ? ... for ghost-halos
              ! - > These two are the same for PULL!! so, no difference.
              ! Why SAVE? Because we need to work on the pdfs on which the
              ! compute kernel was working. The kernel stored to SAVE
              ! All actions taking place on the state array after the kernel
              ! need to be performed with SAVE access
              pos(counter) = ?SAVE?( varPos(iDir), elemPos, nScalars, pdf%neigh)
              myHalos( iProc )%bitMask( varPos(iDir), iElem ) = .true.
            endif
            stat%nLinks_total = stat%nLinks_total + 1
          end do ! iDir
        end do ! field loop
      end do ! iElem

      ! copy position array to me%pos, allocate me%val array
      call pattern%initBuf_real( me%buf_real( iProc ), pos, counter )

    end do ! iProc
    deallocate(pos)

  end subroutine init_recvBuffers
!******************************************************************************!


!******************************************************************************!
  !> Create the communication buffers
  !!
  !! Receive buffers were before created. Now receive the buffers as
  !! send buffers from the remote processes.
  !! Assign the positions, where to fetch from and where to save to
  !! Solver specific part to identify the actual data locations
  !! to send use within the communication buffers.
  !! \todo: this should be pattern specific, with typed exchange we
  !!        do not need to store the pos, as it is done inside MPI
  !!        with the defined datatype
  !!
  !! Access to state array: SAVE
  !! ---------------------------
  !! Why? Because we need to work on the pdfs on which the
  !! compute kernel was working. The kernel stored to SAVE
  !! All actions taking place on the state array after the kernel
  !! need to be performed with SAVE access
  !!
  subroutine init_sendBuffers( send, remoteHalos, pdf, nDOFs, pattern,         &
    &                          requireAll, scheme )
    !---------------------------------------------------------------------------
    !> send communication buffer
    type(tem_communication_type), intent(inout) :: send
    !> contains bitmask of remote halos with array size send%nProcs
    type(halo_commBitmask_type), intent(in)     :: remoteHalos(:)
    !> contains state and neigh array
    type( pdf_data_type), intent(in)            :: pdf
    !> communication pattern
    type(tem_commPattern_type), intent(in)      :: pattern
    !> how many links per element? QQ!
    integer, intent(in)                         :: nDOFs
    !> if true, then take not SAVE but IDX
    logical, intent(in)                         :: requireAll
    !> fluid, boundary and flow information
    type( mus_scheme_type ), intent(in)         :: scheme
    !---------------------------------------------------------------------------
    ! Element positions
    integer, allocatable :: pos(:)
    integer :: iElem ! count variable for elements
    integer :: iDir, iProc, counter, elemPos
    integer :: iField, nScalars, varPos
    !---------------------------------------------------------------------------
    nScalars = scheme%varSys%nScalars
    allocate( pos(maxval(send%nElemsProc*nScalars)))
    allocate( send%buf_real( send%nProcs ))

    ! Then assign the send buffer positions
    do iProc = 1, send%nProcs
      counter = 0
      do iElem = 1, send%nElemsProc( iProc )
        ! Now assign the exact positions in the sendBuf
        elemPos = send%elemPos( iProc )%val( iElem )
        do iField = 1, scheme%nFields
          do iDir = 1, nDOFs
            varPos = scheme%varSys%Variable%val( iField )%varPos(iDir)
            if( remoteHalos( iProc )%bitMask( varPos, iElem )) then
              counter = counter + 1
              ! ? SAVE ? ... for regular halos
              ! ? IDX ? ... for ghost-halos
              ! -> These two are the same for PULL!! so, no difference.
              ! Why SAVE? Because we need to work on the pdfs on which the
              ! compute kernel was working. The kernel stored to SAVE
              ! All actions taking place on the state array after the kernel
              ! need to be performed with SAVE access
              pos( counter ) = ?SAVE?( varPos, elemPos, nScalars, pdf%neigh )
            endif
          end do
        end do! field loop
      end do ! iElem

      call pattern%initBuf_real( send%buf_real( iProc ), pos, counter )
    end do ! iProc

    deallocate(pos)

  end subroutine init_sendBuffers
!******************************************************************************!


!******************************************************************************!
  !> subroutine to find neighbours of element with individual (for each element)
  !! stencil definitions
  !!
  subroutine mus_build_BCStencils( globBC, bc, tree, layout, prefix, stencil )
    !---------------------------------------------------------------------------
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(in)        :: globBC
    !> field boundary with boundary neighbor info
    type(boundary_type), intent(inout)          :: bc
    !> subset of tree ids for which the neighbours will be specified
    type(treelmesh_type), intent(in)            :: tree
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in)  :: layout
    !> field label
    character(len=*), intent(in)                :: prefix
    !> Stencil definition
    type(tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------
    integer              :: iElem, iPos, iLevel, iCount
    integer              :: normal(3)
    integer              :: iNeigh
    !---------------------------------------------------------------------------
    stencil%label = trim(prefix)//trim(bc%label)
    stencil%useAll = .false.
    stencil%requireNeighNeigh = .true.
    stencil%QQ     = bc%nNeighs
    stencil%QQN    = bc%nNeighs
    stencil%nElems = globBC%nElems_local
    if( allocated( stencil%elem ))then
      deallocate( stencil%elem )
    end if
    if( allocated( stencil%elemLvl ))then
      deallocate( stencil%elemLvl )
    end if
    if( allocated( stencil%cxDir ))then
      deallocate( stencil%cxDir )
    end if
    allocate( stencil%elem( stencil%nElems ))
    allocate( stencil%elemLvl( tree%global%minLevel: tree%global%maxLevel))
    allocate( stencil%cxDir( 3, stencil%QQN ))
    ! Reset the stencil, so that if nothing is defined, no additional
    ! elements are recovered
    stencil%cxDir(:,:) = 0
    iCount = 0
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      call init( me = stencil%elemLvl( iLevel ), length = 8 )
      do iElem = 1, globBC%nElems( iLevel )
        iCount = iCount + 1
        ! Get the position of the treeID in the primal mesh
        iPos = globBC%elems( iLevel )%elem( iElem )
        stencil%elem( iCount ) = iPos
        call append( me = stencil%elemLvl( iLevel ), val = iPos )
        ! Get the normal direction of this boundary
        ! only for higher order boundaries
        if( bc%nNeighs > 0) then
          normal(:) = globBC%elems( iLevel )%normal( :, iElem )
          ! And store the normal direction into the stencil
          ! QQN stands for the number of neighbors
          ! Here it means, how many neighbors in the normal direction
          do iNeigh = 1, stencil%QQN
            ! Check if there is a boundary here! ??MH: do we have to do that
            ! here?
            ! I think it should be added, that we check for a boundary in the
            ! current search direction.
            !
            ! ?: what is the purpose of multiplying iStencilElem with normal
            !    direction?
            ! !: We are going along the normal direction for more than
            ! one neighbor in one direction (required for higher order boundary)
            stencil%cxDir( :, iNeigh ) = stencil%cxDir( :, iNeigh )            &
              &                        + normal(:)*iNeigh
          enddo ! iNeigh
        endif
      enddo ! iElem
    enddo ! iLevel

    ! KM: QUICK HACK: Average the normal of each boundary element to obtain
    ! stencil for boundary
    if( bc%nNeighs > 0) then
      do iNeigh = 1, stencil%QQN
        call tem_determine_discreteVector( stencil%cxDir( : , iNeigh ),        &
          &                                                 layout%prevailDir )
        ! KM: Fix Using tem_determine_discreteVector for averaging results in
        ! vector length of prevailDir. Hence iNeigh must be used to get current
        ! cxDir
        stencil%cxDir(:, iNeigh) =  stencil%cxDir(:, iNeigh)*iNeigh
      end do
    endif

  end subroutine mus_build_BCStencils
!******************************************************************************!

!******************************************************************************!
  !>  map the compact stencil to the compute stencil
  !!
  !! The result of this routine is the compact interpolation index for each of
  !! the eight possible occurring cases for fromCoarser interpolations.
  !! There are eight child positions and hence eight different compact stencil
  !! configurations.
  !! For each configuration, the neighbor coordinates need to be known.
  !! MH: Note that this is currently only being used in
  !! intp_quadratic_compactReduced
  !!
  subroutine map_compactToComputeStencil( intp, stencil )
    !---------------------------------------------------------------------------
    !> boundaries for the elements with bnd property set
    type(interpolation_type), intent(inout)     :: intp
    !> Stencil definition
    type(tem_stencilHeader_type), intent(inout) :: stencil
    !---------------------------------------------------------------------------
    integer :: iCase, childNum, iNeigh, iRow
    integer :: neighOffset(3), x(4), parentPos( 3,8)
    integer :: foundDirInStencil
    integer :: iStencilElem
    !---------------------------------------------------------------------------
    ! size: number of compact stencil entries, 8 (=number of child cases )
    ! only to be done for noncompact interpolation schemes
    allocate( intp%compactInd( intp%nSourcesCompactFromCoarser_required, 8 ))
    intp%compactInd = -1
    ! map compact to compute stencil
    do iCase = 1, 8
      childNum = iCase
      write(logUnit(1),*)'icase', icase
      ! For each case, identify all the compact neighbors
      do iNeigh = 1, 8
        x = tem_coordOfId( int(iNeigh, kind=long_k))*2 -1
        parentPos(:, iNeigh)  = matmul(tem_rotationMatrix(:,:, ChildNum),x(1:3))
        neighOffset  = (parentPos(:,iNeigh)  - parentPos(:, 1))/2

        write(logUnit(1),*)'coordinate of neigh ', x
        write(logUnit(1),*)'parentPos           ', parentPos(:, iNeigh)
        write(logUnit(1),*) iNeigh, ' neighOff ', neighOffset

        ! Match the found neighOffset with the stencil. Is this direction
        ! available?
        foundDirInStencil = -1
        do iStencilElem = 1, stencil%QQ

          write(logUnit(1),*) iStencilElem, ' vs ',                            &
            &                 stencil%cxDir( :, iStencilElem )

          if( neighOffset(1) == stencil%cxDir( 1, iStencilElem ) .and.         &
            & neighOffset(2) == stencil%cxDir( 2, iStencilElem ) .and.         &
            & neighOffset(3) == stencil%cxDir( 3, iStencilElem )) then
            foundDirInStencil = iStencilElem
            write(logUnit(1),*)'found ', foundDirInStencil
          end if
        end do
        write(logUnit(1),*) 'iNeigh ', iNeigh, 'found dir ', foundDirInStencil
        write(logUnit(1),*) 'iNeigh ', iNeigh, 'req ',                         &
          &                  intp%nSourcesCompactFromCoarser_required

        do iRow = 1, size( intp%compactInd, 2)
          write(logUnit(1),*)intp%compactInd(:,iRow )
        end do

        if( iNeigh <= intp%nSourcesCompactFromCoarser_required )             &
          &                intp%compactInd( iNeigh, iCase ) = foundDirInStencil
      end do
      if( minval( intp%compactInd( 1:intp%nSourcesCompactFromCoarser_required, &
        &                                                iCase )) < 1 ) then
        write(logUnit(1),*)'found ', minval(                                   &
          &  intp%compactInd(1:intp%nSourcesCompactFromCoarser_required,iCase))
        write(logUnit(1),*) 'at ', minloc( intp%compactInd(1:intp%             &
          &                 nSourcesCompactFromCoarser_required,iCase))
        write(logUnit(1),*) 'Error: did not find all the required compact '//  &
          &             'neighbors in the given stencil'
        call tem_abort
      end if
    end do

  end subroutine map_compactToComputeStencil
!******************************************************************************!


!******************************************************************************!
  !> This subroutine acts as a destructor for the construct routine
  subroutine mus_construct_cleanup( scheme, pdfGlobal, intp, geometry )
    !---------------------------------------------------------------------------
    !> scheme information including fluid, boundary and flow information
    type( mus_scheme_type ), allocatable, intent(inout) :: scheme(:)
    !> global information
    type( pdf_globInfo_type ), intent(inout)            :: pdfGlobal
    !> interpolation information
    type( interpolation_type ),           intent(inout) :: intp
    !> geometric information
    type( mus_geom_type ), intent(inout)                :: geometry
    !---------------------------------------------------------------------------
    integer :: iScheme, iLevel, iField, iBnd, iStencil
    !---------------------------------------------------------------------------
    if( allocated(scheme) )then
      ! loop over the schemes and ...
      do iScheme = 1, size( scheme )
        ! ... deallocate the globBC array
        if( allocated( scheme( iScheme )%globBC ))then
          deallocate( scheme( iScheme )%globBC )
        end if
        ! ... deallocate the level descriptor array
        if( allocated( scheme( iScheme )%glob%levelDesc ))then
          deallocate( scheme( iScheme )%glob%levelDesc )
        end if
        ! ... deallocate the pdf array in each scheme
        if( allocated( scheme( iScheme )%pdf ))then
          deallocate( scheme( iScheme )%pdf )
        end if
        ! ... loop over all stencils and ...
        do iStencil = 1, scheme(iScheme)%layout%nStencils
          if( allocated( scheme(iScheme)%layout%stencil(iStencil)%elemLvl ))then
            deallocate( scheme(iScheme)%layout%stencil(iStencil)%elemLvl )
          end if
          ! ... deallocate stencil elem list
          if( allocated( scheme(iScheme)%layout%stencil(iStencil)%elem ))then
            deallocate( scheme(iScheme)%layout%stencil(iStencil)%elem )
          end if
        end do

        ! ... deallocate the global levelPointer
        if( allocated( pdfGlobal%levelPointer ))then
          deallocate( pdfGlobal%levelPointer )
        end if

        ! ... loop over all fields and ...
        do iField = 1, scheme(iScheme)%nFields
          ! ... loop over all boundaries and ...
          do iBnd = 1, pdfGlobal%nBCs
            do iLevel = 1, globalMaxLevels
              ! ... deallocate the neighBufferPre array
              if( allocated( scheme( iScheme )%field( iField )%bc( iBnd )%     &
                &                        elems( iLevel )%neighBufferPre ))then
                deallocate( scheme( iScheme )%field( iField )%bc( iBnd )%      &
                  &                           elems( iLevel )%neighBufferPre )
              end if
              ! ... deallocate the neighBufferPost array
              if( allocated( scheme( iScheme )%field( iField )%bc( iBnd )%     &
                &                        elems( iLevel )%neighBufferPost ))then
                deallocate( scheme( iScheme )%field( iField )%bc( iBnd )%      &
                  &                           elems( iLevel )%neighBufferPost )
              end if
              ! ... deallocate the neigh array
              if( allocated( scheme( iScheme )%field( iField )%bc( iBnd )%     &
                &                             elems( iLevel )%neigh ))then
                deallocate( scheme( iScheme )%field( iField )%bc( iBnd )%      &
                  &                           elems( iLevel )%neigh )
              end if
              ! ... deallocate the moments array
              if( allocated( scheme( iScheme )%field( iField )%bc( iBnd )%     &
                &                             elems( iLevel )%moments ))then
                deallocate( scheme( iScheme )%field( iField )%bc( iBnd )%      &
                  &                           elems( iLevel )%moments )
              end if
            end do ! iLevel
          end do ! iBnd
        end do ! iField
      end do ! iScheme
    end if ! allocated scheme?

    if( allocated(intp%fromFiner) )   deallocate ( intp%fromFiner )
    if( allocated(intp%fromCoarser) ) deallocate ( intp%fromCoarser)

    if( allocated( geometry%boundary%BC_label )) &
      &            deallocate( geometry%boundary%BC_label )

    if( allocated( geometry%boundary%hasQVal )) &
      &            deallocate( geometry%boundary%hasQVal )

    if( allocated( geometry%boundary%qVal )) &
      &            deallocate( geometry%boundary%qVal )

    if( allocated( geometry%boundary%boundary_ID )) &
      &            deallocate( geometry%boundary%boundary_ID )

    geometry%boundary%header => null()
    geometry%boundary%property => null()


  end subroutine mus_construct_cleanup
!******************************************************************************!


!******************************************************************************!
  !> 
  subroutine set_sendHaloBits( levelDesc )
    !---------------------------------------------------------------------------
    !> 
    type( tem_levelDesc_type ), intent(inout) :: levelDesc
    !---------------------------------------------------------------------------
    integer :: iProc, iElem
    !---------------------------------------------------------------------------

    ! loop over the integer send buffers ...
    do iProc = 1, levelDesc%sendBuffer%nProcs
      ! ... and over the corresponding elements ...
      do iElem = 1, levelDesc%sendBuffer%elemPos(iProc)%nVals
        ! ... and set the property bits
        levelDesc%property( levelDesc%sendBuffer%elemPos(iProc)%val(iElem)) =  &
          &  ibset(levelDesc%property( levelDesc%sendBuffer%elemPos(iProc)%    &
          &                                         val(iElem)), prp_sendHalo )
      end do
    end do

  end subroutine set_sendHaloBits
!******************************************************************************!


end module mus_construction_module
!******************************************************************************!
