!******************************************************************************!
!> In this module, the control structure for computing each time step is
!! established.
!!
!! Various different control structures can be defined and set by function
!! pointers.
!! The current main routine is recursive_multilevel which recursively computes
!! the new time step for all the schemes defined on all levels, starting from
!! the coarsest.
!!
module mus_control_module

  ! include treelm modules
  use env_module,         only: pathLen
  use tem_timer_module,   only: tem_startTimer, tem_stopTimer
  use tem_debug_module,   only: main_debug, dbgUnit
  use tem_time_module,    only: tem_time_advance, tem_time_dump
  use tem_logging_module, only: logUnit

  ! include musubi modules
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_param_module,              only: mus_param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_comm_module,               only: mus_exchange
  use mus_bc_general_module,         only: set_boundary
  use mus_aux_module,                only: check_flow_status, &
    &                                      mus_setParameters, &
    &                                      mus_addExternalForceToState
  use mus_interpolate_header_module, only: interpolation_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_source_module,             only: mus_update_source
  use mus_debug_module,              only: dump_debug_Info

  implicit none

  private

  public :: mus_control_type
  public :: mus_init_control

  !> Datatype containing mapping of control routines to function pointers
  type mus_control_type
    procedure( controlRoutine ), pointer :: do_computation => null()
  end type mus_control_type

  abstract interface
    !> Interface describes the main control routine which does computation
    !! set boundary and check flow status
    subroutine controlRoutine( me, scheme, geometry, params, intp, iLevel,     &
      &                        pdfGlobal )
      import :: mus_scheme_type, mus_geom_type, mus_param_type, &
        & interpolation_type, pdf_globInfo_type, mus_control_type
      !> self control type
      class( mus_control_type ) :: me
      !> containers for the different schemes
      type( mus_scheme_type ), intent(inout) :: scheme(:)
      !> geometry infomation
      type( mus_geom_type ), intent(in) :: geometry
      !> global parameters
      type( mus_param_type ), intent(inout) :: params
      !> Interpolation control
      type( interpolation_type ), intent(in) :: intp
      integer, intent(in) :: iLevel     !< Level counter variable
      !> global flow quantity properties
      type( pdf_globInfo_type ), intent(inout) :: pdfGlobal
    end subroutine controlRoutine
  end interface

  contains

!******************************************************************************!
  !> This routines sets the function pointer to main control routine
  !!
  !! control routine is chosen based on the type of simulation.
  !! like single level, multi-level because multilevel requires
  !! recursive routine
  !!
  !! in the lua file you can now select
  !! \verbatim control_routine = '...'\endverbatim
  !! where possible values are currently
  !! - `benchmark`: strongly reduced control routine with only single level,
  !!                no sources, etc.
  !! mainly for benchmarking
  !! - `multischeme`: full multilevel, multischeme routine
  !! - if nothing is given, the full multilevel, multischeme routine is chosen
  subroutine mus_init_control( params, me )
    !---------------------------------------------------------------------------
    !> global parameters
    type( mus_param_type ), intent(in) :: params
    !> contains function pointer to point control routine
    type( mus_control_type ), intent(out) :: me
    !---------------------------------------------------------------------------

    select case( trim( params%controlRoutine ))
    case('debug')
      me%do_computation => do_recursive_debug
    case('benchmark','fast')
      me%do_computation => do_fast_singleScheme
    case('multischeme', 'multilevel')
      me%do_computation => do_recursive_multiScheme
    case default !       'standard'
      me%do_computation => do_recursive_multiScheme
    end select

  end subroutine mus_init_control
!******************************************************************************!


!******************************************************************************!
  !> Main control routine: Update the time step for all levels
  !!
  recursive subroutine do_recursive_multischeme( me, scheme, geometry, params, &
    &                                            intp, iLevel, pdfGlobal )
    !--------------------------------------------------------------------------
    !> self control type
    class( mus_control_type ) :: me
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry
    !> global parameters
    type( mus_param_type ), intent(inout)       :: params
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal
    integer, intent(in) :: iLevel     !< Level counter variable
    !--------------------------------------------------------------------------
    integer :: iScheme    ! Scheme counter variable
    integer :: iField     ! Field counter variable
    integer :: depScheme  ! depending scheme
    integer :: loc_level
    !--------------------------------------------------------------------------

    loc_level = iLevel

    ! For each level, run over all the schemes defined
    do iScheme = 1, pdfGlobal%nSchemes

      ! swap double buffer index for current level
      scheme( iScheme )%pdf( iLevel )%nNow                                     &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 )+1
      scheme( iScheme )%pdf( iLevel )%nNext                                    &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNext,2 )+1

      !-------------------------------------------------------------------------
      ! set boundary for each field in current scheme
      call tem_startTimer( timerHandle = params%setBndTimer,      &
        &                  me          = params%general%timerData )
      call tem_startTimer( me = params%timer(ilevel),              &
        &                  timerHandle = params%lw_tSetBnd(iLevel) )
      call set_boundary( field     = scheme( iScheme )%field,                  &
        &                pdf       = scheme( iScheme )%pdf,                    &
        &                glob      = scheme( iScheme )%glob,                   &
        &                tree      = geometry%tree,                            &
        &                iLevel    = iLevel,                                   &
        &                pdfGlobal = pdfGlobal,                                &
        &                params    = params,                                   &
        &                layout    = scheme( iScheme )%layout,                 &
        &                physics   = params%physics,                           &
        &                varSys    = scheme( iScheme )%varSys,                 &
        &                mixture   = scheme( iScheme )%mixture,                &
        &                derVarPos = scheme( iScheme)%derVarPos,               &
        &                globBC    = scheme( iScheme )%globBC )
      call tem_stopTimer( timerHandle = params%setBndTimer,      &
        &                 me          = params%general%timerData )
      call tem_stopTimer( me = params%timer(iLevel),              &
                          timerHandle = params%lw_tSetBnd(iLevel) )
      !-------------------------------------------------------------------------

      ! Update parameters
      call mus_setParameters( scheme( iScheme ), iLevel, params )

      !-------------------------------------------------------------------------
      ! Compute current scheme of current level
      call tem_startTimer( timerHandle = params%advRelTimer,      &
        &                  me          = params%general%timerData )
      call tem_startTimer( me = params%timer(ilevel),               &
        &                  timerHandle = params%lw_tCompute(iLevel) )

!$omp parallel
      call scheme( iScheme )%compute(                                          &
        &  iScheme   = iScheme,                                                &
        &  fieldProp = scheme( iScheme )%field(:)%fieldProp,                   &
        &  inState   = scheme( iScheme )%pdf(iLevel)%state(:,                  &
        &                         scheme( iScheme )%pdf(iLevel)%nNow),         &
        &  outState  = scheme( iScheme )%pdf(iLevel)%state(:,                  &
        &                         scheme( iScheme )%pdf(iLevel)%nNext),        &
        &  neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),                 &
        &  nStart    = 1,                                                      &
        &  nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,             &
        &  level     = iLevel,                                                 &
        &  layout    = scheme( iScheme )%layout,                               &
        &  pdfGlobal = pdfGlobal,                                              &
        &  params    = params,                                                 &
        &  scheme    = scheme,                                                 &
        &  tree      = geometry%tree,                                          &
        &  varSys    = scheme( iScheme )%varSys )
!$omp end parallel

      call tem_stopTimer( timerHandle = params%advRelTimer,                    &
        &                 me          = params%general%timerData )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tCompute(iLevel) )
      !-------------------------------------------------------------------------

      ! update outstate with external force fields
      ! \todo KM: check when to add external force
      ! because after compute kernel, time is advanced according to LBMalgorithm
      ! by in code time is advanced only in check_flow status
      call mus_addExternalForceToState( iScheme   = iScheme,                   &
        &                               scheme    = scheme,                    &
        &                               pdfGlobal = pdfGlobal,                 &
        &                               tree      = geometry%tree,             &
        &                               iLevel    = loc_level,                 &
        &                               params    = params )

    end do ! scheme loop


    call tem_startTimer( me = params%timer(ilevel),            &
      &                  timerHandle = params%lw_tIntp(iLevel) )
    ! when not on finest level, go to next level
    if( iLevel < pdfGlobal%maxLevel ) then
      ! Do computation on finer level (L+1)
      ! Fill my coarser element (L) from finer (L+1)
      call do_computeFinerIntpandExchange( me, scheme, geometry, params, &
        &                                  intp, iLevel, pdfGlobal )
    end if
    call tem_stopTimer( me = params%timer(iLevel),            &
                        timerHandle = params%lw_tIntp(iLevel) )

    ! Communicate the halo elements of each scheme on current level
    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( timerHandle = params%commTimer,        &
        &                  me          = params%general%timerData )

      ! communicate halo elements
      call mus_exchange(                                                       &
        &  send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer,      &
        &  recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer,      &
        &  state   = scheme( iScheme )%pdf(iLevel)%state(:,                    &
        &                                scheme( iScheme )%pdf(iLevel)%nNext), &
        &  pattern = params%general%commPattern,                               &
        &  level   = iLevel,                                                   &
        &  proc    = params%general%proc)

      ! communicate my finer ghost element from coarse level
      if ( iLevel > pdfGlobal%minLevel ) then
        call mus_exchange(                                                     &
          & send    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             sendbufferFromCoarser, &
          & recv    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             recvbufferFromCoarser, &
          & state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )% &
          &                                                pdf(iLevel)%nNext), &
          & pattern = params%general%commPattern,                              &
          & level   = iLevel,                                                  &
          & proc    = params%general%proc)
      end if

      call tem_stopTimer( timerHandle = params%commTimer,        &
        &                 me          = params%general%timerData )
    end do ! scheme loop

    !---------------------------------------------------------------------------
    ! Interpolate the ghost elements on the finer level(L+1) with data provided
    ! from current level(L).
    ! This only needs to be done once in two finer time steps
    call tem_startTimer( me = params%timer(ilevel),            &
      &                  timerHandle = params%lw_tIntp(iLevel) )

    if ( iLevel < pdfGlobal%maxLevel ) then
      call do_intpCoarserandExchange( scheme, params, intp, iLevel, pdfGlobal )
    end if ! if not on finest level

    call tem_stopTimer( me = params%timer(iLevel),            &
                        timerHandle = params%lw_tIntp(iLevel) )
    !---------------------------------------------------------------------------

    ! update the time counters. MH: checked. Please dont move
    ! Increasing with the smallest time step (maxLevel)
    ! KM: time is advanced here since new time is required to update sources.
    if( iLevel == pdfGlobal%maxLevel ) then
      call tem_time_advance( &
        &    me = params%general%simControl%now, &
        &    sim_dt = params%physics%dtLvl( geometry%tree%global%maxLevel ))
    endif

    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( timerHandle = params%sourceTimer,                   &
        &                  me          = params%general%timerData )
      if( allocated(scheme( iScheme )%depend) )then
        depScheme = scheme( iScheme )%depend( 1 )%useScheme
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                  me            = scheme(iScheme)%source,           &
          &                  pdf           = scheme(iScheme)%pdf,              &
          &                  depPdf        = scheme(depScheme)%pdf,            &
          &                  time          = params%general%simControl%now,    &
          &                  global_varSys = scheme(iScheme)%varSys,           &
          &                  global_tree   = geometry%tree,                    &
          &                  stencil       = scheme(iScheme)%layout%stencil(1))
      else
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                  me            = scheme(iScheme)%source,           &
          &                  pdf           = scheme(iScheme)%pdf,              &
          &                  time          = params%general%simControl%now,    &
          &                  global_varSys = scheme(iScheme)%varSys,           &
          &                  global_tree   = geometry%tree,                    &
          &                  stencil       = scheme(iScheme)%layout%stencil(1))
      end if
      call tem_stopTimer(timerHandle = params%sourceTimer,                     &
        &                me          = params%general%timerData)
    end do ! scheme loop

!! !$omp parallel

!! !$omp end parallel

    ! check flow status including perform checks, tracking, restart and geomInc
    ! only at minLevel to complete one multilevel cycle
    ! this check can also be done using mus_time_modulo(now, reqInterval)
    if( iLevel == pdfGlobal%minLevel ) then
      ! check if some action has to be taken based on timeControl:
      ! tracking, global reduction operations, restart
      call check_flow_status( scheme    = scheme,                              &
        &                     params    = params,                              &
        &                     geometry  = geometry,                            &
        &                     pdfGlobal = pdfGlobal )
    end if

  end subroutine do_recursive_multischeme
!******************************************************************************!


!******************************************************************************!
  !> This routine does:
  !! 1. computation twice at finer level (iLevel+1),
  !! 2. interpolate my coarse ghost element (iLevel) from finer level (iLevel+1)
  !! 3. exchange the data of my coarse ghost elements between process
  recursive subroutine do_computeFinerIntpandExchange( me, scheme,             &
    &                               geometry, params, intp, iLevel, pdfGlobal )
    !---------------------------------------------------------------------------
    class( mus_control_type ) :: me !< contains control routine function pointer
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry
    !> global parameters
    type( mus_param_type ), intent(inout)       :: params
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal
    integer, intent(in) :: iLevel     !< Level counter variable
    !---------------------------------------------------------------------------
    integer :: iScheme
    !> diffusive or acoustic nesting loop
    integer :: iNestingLoop
    character(len=33) :: buffer
    !---------------------------------------------------------------------------

    ! Perform the number of nested time steps on the finer level L+1
    ! according to scaling type :
    !   diffusive: nNesting = 3
    !   acoustic:  nNesting = 1
    do iNestingLoop = 1, params%nNesting
      ! Compute and communicate
      call me%do_computation( scheme, geometry, params, intp, iLevel+1, &
        &                     pdfGlobal )
    end do
    ! Compute and communicate
    call me%do_computation( scheme, geometry, params, intp, iLevel+1, &
      &                     pdfGlobal )

!$omp single
    ! Interpolate my coarse ghost elements from the finer ones
    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( timerHandle = params%intpTimer_low,    &
        &                  me          = params%general%timerData )

      !-------------------------------------------------------------------------
      ! first do low order interpolation
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tIntp(iLevel) )
      call intp%fillMineFromFiner_low(                                         &
        &                  pdf       = scheme( iScheme )%pdf,                  &
        &                  glob      = scheme( iScheme )%glob,                 &
        &                  scheme    = scheme( iScheme ),                      &
        &                  fieldProp = scheme( iScheme )%field(:)%fieldProp,   &
        &                  level     = iLevel,                                 &
        &                  intp      = intp,                                   &
        &                  ind       = scheme( iScheme )%glob%                 &
        &                                 levelDesc(iLevel)%intpFromFiner_low, &
        &                  layout    = scheme( iScheme )%layout,               &
        &                  varSys    = scheme( iScheme )%varSys,               &
        &                  physics   = params%physics,                         &
        &                  derVarPos = scheme( iScheme )%derVarPos  )
      call tem_stopTimer( timerHandle = params%intpTimer_low,    &
        &                 me          = params%general%timerData )
      !-------------------------------------------------------------------------

      !-------------------------------------------------------------------------
      ! then do high order interpolation
      call tem_startTimer( timerHandle = params%intpTimer_hgh,    &
        &                  me          = params%general%timerData )
      call intp%fillMineFromFiner(                                             &
        &                  pdf       = scheme( iScheme )%pdf,                  &
        &                  glob      = scheme( iScheme )%glob,                 &
        &                  scheme    = scheme( iScheme ),                      &
        &                  fieldProp = scheme( iScheme )%field(:)%fieldProp,   &
        &                  level     = iLevel,                                 &
        &                  intp      = intp,                                   &
        &                  ind       = scheme( iScheme )%glob%                 &
        &                                levelDesc(iLevel)%intpFromFiner_high, &
        &                  layout    = scheme( iScheme )%layout,               &
        &                  varSys    = scheme( iScheme )%varSys,               &
        &                  physics   = params%physics,                         &
        &                  derVarPos    = scheme( iScheme )%derVarPos)
      call tem_stopTimer( timerHandle = params%intpTimer_hgh,    &
        &                 me          = params%general%timerData )
      !-------------------------------------------------------------------------

      call tem_stopTimer( me = params%timer(iLevel),            &
                          timerHandle = params%lw_tIntp(iLevel) )

      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer=' after fill Mine FromFiner       '
        call dump_debug_info( buffer, scheme, params, iLevel, iScheme )
      end if

      !-------------------------------------------------------------------------
      ! Exchange the coarse ghost elements
      call tem_startTimer( timerHandle = params%commTimer,        &
        &                  me          = params%general%timerData )

      call mus_exchange(                                                  &
        &        send    = scheme( iScheme )%glob%levelDesc(iLevel)%      &
        &                                            sendbufferFromFiner, &
        &        recv    = scheme( iScheme )%glob%levelDesc(iLevel)%      &
        &                                            recvbufferFromFiner, &
        &        state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme(  &
        &                                   iScheme )%pdf(iLevel)%nNext), &
        &        pattern = params%general%commPattern,                    &
        &        level   = iLevel,                                        &
        &        proc    = params%general%proc)

      call tem_stopTimer( timerHandle = params%commTimer,        &
        &                 me          = params%general%timerData )
      !-------------------------------------------------------------------------

      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer = ' after exch Mine FromFiner'
        call dump_debug_info( buffer, scheme, params, iLevel, iScheme )
      end if

    end do
!$omp end single

  end subroutine do_computeFinerIntpandExchange
!******************************************************************************!


!******************************************************************************!
  !> This routine utilizes fluid elements on my level (L) to fill finer
  !! ghost elements on next level (L+1).
  !! Then it exchanges the datas of finer ghost elements (L+1) between process.
  subroutine do_intpCoarserandExchange( scheme, params, intp, iLevel,          &
    &                                   pdfGlobal )
    !---------------------------------------------------------------------------
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)
    !> global parameters
    type( mus_param_type ), intent(inout)       :: params
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(in)  :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iScheme
    character(len=33) :: buffer
    !---------------------------------------------------------------------------

    do iScheme = 1, pdfGlobal%nSchemes

      ! Fill finer ghost elements (L+1) by low order interpolation
      call tem_startTimer( timerHandle = params%intpTimer_low,    &
        &                  me          = params%general%timerData )
      call intp%fillFinerFromMe_low(                                           &
        &               pdf       = scheme( iScheme )%pdf,                     &
        &               glob      = scheme( iScheme )%glob,                    &
        &               scheme    = scheme( iScheme ),                         &
        &               fieldProp = scheme( iScheme )%field(:)%fieldProp,      &
        &               level     = iLevel,                                    &
        &               intp      = intp,                                      &
        &               ind       = scheme( iScheme )%glob%                    &
        &                            levelDesc( iLevel+1)%intpFromCoarser_low, &
        &               layout    = scheme( iScheme )%layout,                  &
        &               varSys    = scheme( iScheme )%varSys,                  &
        &               physics   = params%physics,                            &
        &               derVarPos = scheme( iScheme )%derVarPos )
      call tem_stopTimer( timerHandle = params%intpTimer_low,    &
        &                 me          = params%general%timerData )

      ! Fill finer ghost elements (L+1) by high order interpolation
      call tem_startTimer( timerHandle = params%intpTimer_hgh,    &
        &                  me          = params%general%timerData )
      call intp%fillFinerFromMe(                                               &
        &               pdf       = scheme( iScheme )%pdf,                     &
        &               glob      = scheme( iScheme )%glob,                    &
        &               scheme    = scheme( iScheme ),                         &
        &               fieldProp = scheme( iScheme )%field(:)%fieldProp,      &
        &               level     = iLevel,                                    &
        &               intp      = intp,                                      &
        &               ind       = scheme( iScheme )%glob%                    &
        &                           levelDesc( iLevel+1)%intpFromCoarser_high, &
        &               layout    = scheme( iScheme )%layout,                  &
        &               varSys    = scheme( iScheme )%varSys,                  &
        &               physics   = params%physics,                            &
        &               derVarPos = scheme( iScheme )%derVarPos )
      call tem_stopTimer( timerHandle = params%intpTimer_hgh,    &
        &                 me          = params%general%timerData )

      ! Debug output
      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer = ' after fillFinerFromMe'
        call dump_debug_info( buffer, scheme, params, iLevel+1, iScheme )
      end if

      call tem_startTimer( timerHandle = params%commTimer,        &
        &                  me          = params%general%timerData )
      ! Exchange the fine ghost elements
      call mus_exchange(                                                       &
        &            send    = scheme( iScheme )%glob                          &
        &                       %levelDesc(iLevel+1)%sendbufferFromCoarser,    &
        &            recv    = scheme( iScheme )%glob                          &
        &                       %levelDesc(iLevel+1)%recvbufferFromCoarser,    &
        &            state   = scheme( iScheme )%pdf(iLevel+1)%                &
        &                      state(:,scheme( iScheme )%pdf(iLevel+1)%nNext), &
        &            pattern = params%general%commPattern,                     &
        &            level   = iLevel,                                         &
        &            proc    = params%general%proc)
      call tem_stopTimer( timerHandle = params%commTimer,        &
        &                 me          = params%general%timerData )

      ! Debug output
      if ( main_debug%checkEachAlgorithmicStep ) then
        buffer = '  after exchFinerFromMe'
        call dump_debug_info( buffer, scheme, params, iLevel+1, iScheme )
      end if

    end do
  end subroutine do_intpCoarserandExchange
!******************************************************************************!


!******************************************************************************!
  !> Control routine for an optimized workflow with reduced functionality.
  !!
  !! No sources, no multilevel, no multischeme.
  !! Use for benchmarking
  !!
  subroutine do_fast_singleScheme( me, scheme, geometry, params, intp, iLevel, &
    &                              pdfGlobal )
    !---------------------------------------------------------------------------
    !> self control type
    !! dummy variable in this routine, required by interface
    class( mus_control_type ) :: me
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry
    !> global parameters
    type( mus_param_type ), intent(inout)       :: params
    !> interpolation type
    !! dummy variable in this routine, required by interface
    type( interpolation_type ), intent(in) :: intp
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer, parameter  :: iScheme = 1   ! Scheme counter variable
    !---------------------------------------------------------------------------
!$omp single
    ! swap double buffer index for current level
    scheme( iScheme )%pdf( iLevel )%nNow                                       &
      &                   = mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 ) + 1
    scheme( iScheme )%pdf( iLevel )%nNext                                      &
      &                   = mod( scheme( iScheme )%pdf( iLevel )%nNext,2 ) + 1
    !set boundary for each field in current scheme
    call tem_startTimer( timerHandle =  params%setBndTimer,     &
        &                me          = params%general%timerData )
    call tem_startTimer( me = params%timer(ilevel),                          &
      &                  timerHandle = params%lw_tSetBnd(iLevel) )
    call set_boundary( field     = scheme( iScheme )%field,                    &
      &                pdf       = scheme( iScheme )%pdf,                      &
      &                glob      = scheme( iScheme )%glob,                     &
      &                tree      = geometry%tree,                              &
      &                iLevel    = iLevel,                                     &
      &                pdfGlobal = pdfGlobal,                                  &
      &                params    = params,                                     &
      &                layout    = scheme( iScheme )%layout,                   &
      &                physics   = params%physics,                             &
      &                varSys    = scheme( iScheme )%varSys,                   &
      &                mixture   = scheme( iScheme )%mixture,                  &
      &                derVarPos = scheme( iScheme)%derVarPos,                 &
      &                globBC    = scheme( iScheme )%globBC )
    call tem_stopTimer( timerHandle = params%setBndTimer,      &
        &               me          = params%general%timerData )
    call tem_stopTimer( me = params%timer(iLevel),                             &
                        timerHandle = params%lw_tSetBnd(iLevel) )

    ! Update parameters, relaxation time .etc
    call mus_setParameters( scheme( iScheme ), iLevel, params )

    ! Compute current scheme of current level
    call tem_startTimer( timerHandle = params%advRelTimer,      &
        &                me          = params%general%timerData )
    call tem_startTimer( me = params%timer(ilevel),                            &
      &                  timerHandle = params%lw_tCompute(iLevel) )
!$omp end single

    call scheme( iScheme )%compute(                                            &
      &           iScheme   = iScheme,                                         &
      &           fieldProp = scheme( iScheme )%field(:)%fieldProp,            &
      &           inState   = scheme( iScheme )%pdf(iLevel)%state(:,           &
      &                                  scheme( iScheme )%pdf(iLevel)%nNow),  &
      &           outState  = scheme( iScheme )%pdf(iLevel)%state(:,           &
      &                                  scheme( iScheme )%pdf(iLevel)%nNext), &
      &           neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),          &
      &           nStart    = 1,                                               &
      &           nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,      &
      &           level     = iLevel,                                          &
      &           layout    = scheme( iScheme )%layout,                        &
      &           pdfGlobal = pdfGlobal,                                       &
      &           params    = params,                                          &
      &           scheme    = scheme,                                          &
      &           tree      = geometry%tree,                                   &
      &           varSys    = scheme( iScheme )%varSys )
    call tem_stopTimer( timerHandle = params%advRelTimer,      &
        &               me          = params%general%timerData )
    call tem_stopTimer( me = params%timer(iLevel),                             &
                        timerHandle = params%lw_tCompute(iLevel) )

    ! update outstate with external force fields
    ! \todo KM: check when to add external force
    ! because after compute kernel, time is advanced according to LBMalgorithm
    ! by in code time is advanced only in check_flow status
    call mus_addExternalForceToState( iScheme   = iScheme,                     &
      &                               scheme    = scheme,                      &
      &                               pdfGlobal = pdfGlobal,                   &
      &                               tree      = geometry%tree,               &
      &                               iLevel    = iLevel,                      &
      &                               params    = params )

!$omp single
    ! Communicate the halo elements of each scheme on current level
    call tem_startTimer( timerHandle = params%commTimer,        &
        &                me          = params%general%timerData )
    call mus_exchange(                                                         &
      &         send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer, &
      &         recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer, &
      &         state   = scheme( iScheme )%pdf(iLevel)%state(:,               &
      &                                  scheme( iScheme )%pdf(iLevel)%nNext), &
      &         pattern = params%general%commPattern,                          &
      &         level   = iLevel,                                              &
      &         proc    = params%general%proc)
    call tem_stopTimer( timerHandle = params%commTimer,        &
        &               me          = params%general%timerData )

    ! MH: Currently using the dt from the first scheme \todo add correct dt. Which one?
    ! update the time counters. MH: checked. Please dont move
    ! Increasing with the smallest time step (maxLevel)
    ! KM: time is advanced here since new time is required to update sources.
    call tem_time_advance( me = params%general%simControl%now,                 &
      &                    sim_dt = params%physics%dtLvl(                      &
      &                                     geometry%tree%global%maxLevel ))

    ! check if some action has to be taken based on timeControl:
    ! tracking, global reduction operations, restart
    call check_flow_status( scheme    = scheme,                                &
      &                     params    = params,                                &
      &                     geometry  = geometry,                              &
      &                     pdfGlobal = pdfGlobal )
!$omp end single

  end subroutine do_fast_singleScheme
!******************************************************************************!


!******************************************************************************!
  !> Main control routine: Update the time step for all levels
  !!
  recursive subroutine do_recursive_debug( me, scheme, geometry, params,       &
    &                                      intp, iLevel, pdfGlobal )
    !---------------------------------------------------------------------------
    !> self control type
    class( mus_control_type ) :: me
    !> containers for the different schemes
    type( mus_scheme_type ), intent(inout)  :: scheme(:)
    !> geometry infomation
    type( mus_geom_type ), intent(in)        :: geometry
    !> global parameters
    type( mus_param_type ), intent(inout)       :: params
    !> Interpolation control
    type( interpolation_type ), intent(in) :: intp
    !> global flow quantity properties
    type( pdf_globInfo_type ), intent(inout)  :: pdfGlobal
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iScheme    ! Scheme counter variable
    integer :: depScheme    ! depending scheme
    character(len=33) :: buffer
    !---------------------------------------------------------------------------
    if( iLevel == pdfGlobal%minLevel ) then
      call tem_time_dump(params%general%simControl%now, dbgUnit(1))
    endif
    ! For each level, run over all the schemes defined
    do iScheme = 1, pdfGlobal%nSchemes
!$omp single
      write(buffer,'(a,i2,a,i3)')' before advrel L = ', iLevel, 't = ',        &
        &                                   params%general%simControl%now%iter
      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )

      ! swap double buffer index for current level
      scheme( iScheme )%pdf( iLevel )%nNow                                     &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 ) + 1
      scheme( iScheme )%pdf( iLevel )%nNext                                    &
        &                   = mod( scheme( iScheme )%pdf( iLevel )%nNext,2 ) + 1

      !set boundary for each field in current scheme
      call tem_startTimer( timerHandle = params%setBndTimer,      &
        &                  me          = params%general%timerData )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tSetBnd(iLevel) )
      call set_boundary( field     = scheme( iScheme )%field,                  &
        &                pdf       = scheme( iScheme )%pdf,                    &
        &                glob      = scheme( iScheme )%glob,                   &
        &                tree      = geometry%tree,                            &
        &                iLevel    = iLevel,                                   &
        &                pdfGlobal = pdfGlobal,                                &
        &                params    = params,                                   &
        &                physics   = params%physics,                           &
        &                layout    = scheme( iScheme )%layout,                 &
        &                varSys    = scheme( iScheme )%varSys,                 &
        &                mixture   = scheme( iScheme )%mixture,                &
        &                derVarPos = scheme( iScheme)%derVarPos,               &
        &                globBC    = scheme( iScheme )%globBC )
      call tem_stopTimer( timerHandle = params%setBndTimer,      &
        &                 me          = params%general%timerData )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tSetBnd(iLevel) )

      ! Update parameters
      call mus_setParameters( scheme( iScheme ), iLevel, params )

      ! Compute current scheme of current level
      call tem_startTimer( timerHandle = params%advRelTimer,      &
        &                  me          = params%general%timerData )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tCompute(iLevel) )
!$omp end single

      call scheme( iScheme )%compute(                                          &
        &         iScheme   = iScheme,                                         &
        &         fieldProp = scheme( iScheme )%field(:)%fieldProp,            &
        &         inState   = scheme( iScheme )%pdf(iLevel)%state(:,           &
        &                                scheme( iScheme )%pdf(iLevel)%nNow),  &
        &         outState  = scheme( iScheme )%pdf(iLevel)%state(:,           &
        &                                scheme( iScheme )%pdf(iLevel)%nNext), &
        &         neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),          &
        &         nStart    = 1,                                               &
        &         nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,      &
        &         level     = iLevel,                                          &
        &         layout    = scheme( iScheme )%layout,                        &
        &         pdfGlobal = pdfGlobal,                                       &
        &         params    = params,                                          &
        &         scheme    = scheme,                                          &
        &         tree      = geometry%tree,                                   &
        &         varSys    = scheme( iScheme )%varSys )
      call tem_stopTimer( timerHandle = params%advRelTimer,      &
        &                 me          = params%general%timerData )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tCompute(iLevel) )
      write(buffer,'(a,i2,a,i3)')' after advrel L = ', iLevel, 't = ',         &
        &                                   params%general%simControl%now%iter

      ! update outstate with external force fields
      call mus_addExternalForceToState( iScheme   = iScheme,                   &
        &                               scheme    = scheme,                    &
        &                               pdfGlobal = pdfGlobal,                 &
        &                               tree      = geometry%tree,             &
        &                               iLevel    = iLevel,                    &
        &                               params    = params )


      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )

    end do ! scheme loop


    ! Check, if we are on the finest level
    if( iLevel < pdfGlobal%maxLevel ) then
      ! perform interpolation and recursive calls
      call do_computeFinerIntpandExchange( me, scheme, geometry, params,       &
        &                                  intp, iLevel, pdfGlobal )
    end if

!$omp single
    ! Communicate the halo elements of each scheme on current level
    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer( timerHandle = params%commTimer,        &
        &                  me          = params%general%timerData )
      call mus_exchange(                                                       &
        &  send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer,      &
        &  recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer,      &
        &  state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )%  &
        &                                                  pdf(iLevel)%nNext), &
        &  pattern = params%general%commPattern,                               &
        &  level   = iLevel,                                                   &
        &  proc    = params%general%proc)
      write(buffer,'(a,i2)')' after exchange buffer L      ', iLevel
      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )
      call tem_stopTimer( timerHandle = params%commTimer,        &
        &                 me          = params%general%timerData )
      if( iLevel > pdfGlobal%minLevel ) then
        call tem_startTimer( timerHandle = params%commTimer,        &
        &                    me          = params%general%timerData )
        call mus_exchange(                                                     &
          & send    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             sendbufferFromCoarser, &
          & recv    = scheme( iScheme )%glob%levelDesc(iLevel)%                &
          &                                             recvbufferFromCoarser, &
          & state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )% &
          &                                                pdf(iLevel)%nNext), &
          & pattern = params%general%commPattern,                               &
          & level   = iLevel,                                                  &
          & proc    = params%general%proc)
        if ( main_debug%checkEachAlgorithmicStep ) then
          write(buffer,'(a,i2)')' after exch gFC buffer L      ', iLevel
          call dump_debug_info( buffer, scheme, params, iLevel, iScheme )
        end if
        call tem_stopTimer( timerHandle = params%commTimer,        &
        &                   me          = params%general%timerData )
      end if
    end do ! scheme loop

    if( iLevel < pdfGlobal%maxLevel ) then
      ! Now when all coarser ghosts are present, we can interpolate the other
      ! way round.
      ! Interpolate the ghost elements on the finer level with data provided
      ! from current level (This only needs to be done once in two finer time
      ! steps)
      call do_intpCoarserandExchange( scheme, params, intp, iLevel, pdfGlobal )
     ! buffer=' after fillFinerFromMe           '
     ! iScheme = 1
     ! call dump_debug_info( buffer, scheme, params, iLevel, iScheme )
    end if ! if not on finest level

    ! MH: Currently using the dt from the first scheme \todo add correct dt. Which one?
    ! update the time counters. MH: checked. Please dont move
    ! Increasing with the smallest time step (maxLevel)
    ! KM: time is advanced here since new time is required to update sources.
    if( iLevel == pdfGlobal%maxLevel ) then
      call tem_time_advance( me = params%general%simControl%now,               &
        &                    sim_dt = params%physics%dtLvl(                    &
        &                                     geometry%tree%global%maxLevel ))
    endif

    do iScheme = 1, pdfGlobal%nSchemes
      call tem_startTimer(timerHandle = params%sourceTimer,      &
        &                 me          = params%general%timerData )
      if( allocated(scheme( iScheme )%depend))then
        depScheme = scheme( iScheme )%depend( 1 )%useScheme
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                   me            = scheme(iScheme)%source,          &
          &                   pdf           = scheme(iScheme)%pdf,             &
          &                   depPdf        = scheme(depScheme)%pdf,           &
          &                   time          = params%general%simControl%now,   &
          &                   global_varSys = scheme(iScheme)%varSys,          &
          &                   global_tree   = geometry%tree,                   &
          &                   stencil       = scheme(iScheme)%layout%stencil(1))
      else
        ! update the source terms for each field and the global one for the
        ! scheme
        call mus_update_source(                                                &
          &                   me            = scheme(iScheme)%source,          &
          &                   pdf           = scheme(iScheme)%pdf,             &
          &                   time          = params%general%simControl%now,   &
          &                   global_varSys = scheme(iScheme)%varSys,          &
          &                   global_tree   = geometry%tree,                   &
          &                   stencil       = scheme(iScheme)%layout%stencil(1))
      end if
      call tem_stopTimer( timerHandle = params%sourceTimer,      &
        &                 me          = params%general%timerData )
    end do

    ! check flow status including perform checks, tracking, restart and geomInc
    ! only at minLevel to complete one multilevel cycle
    if( iLevel == pdfGlobal%minLevel ) then
      ! check if some action has to be taken based on timeControl:
      ! tracking, global reduction operations, restart
      call check_flow_status( scheme    = scheme,                              &
        &                     params    = params,                              &
        &                     geometry  = geometry,                            &
        &                     pdfGlobal = pdfGlobal )
    end if
!$omp end single

  end subroutine do_recursive_debug
!******************************************************************************!


end module mus_control_module
!******************************************************************************!


!******************************************************************************!
!>\page computationProcedure Description of the multi-level algorithm
!!
!!\section multilevel_introduction Introductory comments
!! Here we describe the recursive processing of the level-wise sorted list of
!! elements through the kernels.
!! From the mesh on disk, which was stored as treeIDs, a level-wise collection
!! of element lists are created before the computation can be started. These
!! lists include ghost and halo cells to account for elements which are missing
!! locally but are calculated on a remote partition as well as missing neighbor
!! elements due to level interfaces.
!! A detailed description of the mesh generation process and the employed data
!! structures can be found where.
!!
!!
!! All levels are updated starting from the coarsest,
!! updating all schemes. For each scheme the following procedure is performed.
!! First, the double buffer index is swapped and the boundaries are set.
!! Then the flow field is computed for the current time step (for this level and this scheme).
!! Next, communication with adjacent partitions for current level is performed.
!! If multiple levels are present, a recursive part is entered.
!! On a level n, 2*(n-1) times the time steps of the coarsest level have to be computed.
!! Therefore, the compute routine is called twice for the next finer level.
!! Afterwards, the ghost elements have to be updated again to provide valid values for
!! the fluid elements. First the coarse ghost elements are updated from fluid elements
!! on the finer level, then they have to be communicated.
!! Same applies for the finer ghost elements.
!! After all schemes and levels are done, a routine handling output, check and restart
!! is called and the time step is updated.
!!\verbatim
!! recursive subroutine compute( iLevel )
!!   compute_kernel
!!   if( iLevel < maxLevel)
!!     recursive call compute( iLevel+1 )
!!     comm FromCoarser
!!     recursive call compute( iLevel+1 )
!!     fillFromFiner
!!   endif
!!   comm FromFiner
!!   comm
!!   if( iLevel < maxLevel)
!!     fillFromCoarser( iLevel+1)
!!     comm FromCoarser( iLevel+1)
!!   endif
!! end subroutine compute
!!\endverbatim
!!
!! \image html multilevel_recursivealgorithm.png
