?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Collection of debugging functions
!!
!! 
module mus_debug_module

  ! include treelm subroutines
  use env_module,              only: long_k
  use tem_timer_module,        only: tem_startTimer, tem_stopTimer
  use tem_param_module,        only: qDirName
  use tem_construction_module, only: tem_levelDesc_type
  use treelmesh_module,        only: treelmesh_type
  use tem_bc_prop_module,      only: tem_bc_prop_type
  use tem_element_module,      only: eT_ghostFromCoarserELem, &
    &                                eT_ghostFromFinerElem
  use tem_property_module,     only: prp_hasBnd
  use tem_topology_module,     only: tem_CoordOfId, tem_IdOfCoord, &
    &                                tem_firstIdAtLevel
  use tem_geometry_module,     only: tem_baryOfId
  use tem_debug_module,        only: dbgUnit, main_debug
  use tem_tools_module,        only: tem_horizontalSpacer
  use tem_logging_module,      only: logUnit
  use tem_time_module,         only: tem_time_advance

  ! include musubi subroutines
  use mus_pdf_module,                only: pdf_data_type, pdf_global_type, &
    &                                      pdf_globInfo_type
  use mus_param_module,              only: mus_param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_comm_module,               only: mus_exchange
  use mus_bc_general_module,         only: set_boundary
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_debug_tools_module,        only: dumpPdfAll
  use mus_interpolate_header_module, only: interpolation_type

  implicit none
  private

  public :: dump_elemDependency
  public :: debug_dependenciesFromFiner
  public :: debug_dependenciesFromCoarser
  public :: debug_connectivity
  public :: debug_normals
  public :: dump_debug_info

  contains

!******************************************************************************!
  !> Main control routine: Update the time step for all levels
  !!
  !! All levels are updated starting from the coarsest, 
  !! updating all schemes. For each scheme the following procedure is performed.
  !! First, the double buffer index is swapped and the boundaries are set.
  !! Then the flow field is computed for the current time step (for this level 
  !! and this scheme). Next, communication with adjacent partitions for current 
  !! level is performed If multiple levels are present, a recursive part is 
  !! entered. On a level n, 2*(n-1) times the time steps of the coarsest level 
  !! have to be computed. Therefore, the compute routine is called twice for the
  !! next finer level. Afterwards, the ghost elements have to be updated again 
  !! to provide valid values for the fluid elements. First the coarse ghost
  !! elements are updated from fluid elements on the finer level, then they
  !! have to be communicated. Same applies for the finer ghost elements. After
  !! all schemes and levels are done, a routine handling output, check and 
  !! restart is called and the time step is updated.
  !!
  recursive subroutine debug_recursive_multilevel( scheme, geometry, params,   &
    &                                              intp, iLevel, pdfGlobal )
    !---------------------------------------------------------------------------
    !> containers for the different schemes
    type( mus_scheme_type )    :: scheme(:)
    !> geometry infomation
    type( mus_geom_type )      :: geometry
    !> global parameters
    type( mus_param_type )     :: params
    !> Interpolation control
    type( interpolation_type ) :: intp
    !> global flow quantity properties
    type( pdf_globInfo_type )  :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: iLevel     ! Level counter variable
    integer :: iLoop      ! nesting loops for diffusive or acoustic scaling
    integer :: iScheme    ! Scheme counter variable
    character(len=34) :: buffer
    !---------------------------------------------------------------------------

    write(logUnit(1),*) ' TIMESTEP ',params%general%simControl%now%iter
    if( iLevel == pdfGlobal%minLevel ) then
      write(dbgUnit(1),*)' TIMESTEP ', params%general%simControl%now%iter
    endif
    ! For each level, run over all the schemes defined
    do iScheme = 1, pdfGlobal%nSchemes
      ! swap double buffer index for current level
!$omp single
      scheme( iScheme )%pdf( iLevel )%nNow  =                                  &
        &                     mod( scheme( iScheme )%pdf( iLevel )%nNow ,2 ) + 1
      scheme( iScheme )%pdf( iLevel )%nNext =                                  &
        &                     mod( scheme( iScheme )%pdf( iLevel )%nNext,2 ) + 1

!      buffer=' before advRel                   '
!      call dump_debug_info( buffer, scheme, pdfGlobal, params, liLevel, iScheme )
      call tem_startTimer( timerHandle = params%setBndTimer, &
        &                  me          = params%general%timerData )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tSetBnd(iLevel) )
      call set_boundary( field     = scheme( iScheme )%field,                  & 
        &                pdf       = scheme( iScheme )%pdf,                    &
        &                glob      = scheme( iScheme )%glob,                   &
        &                tree      = geometry%tree,                            &
        &                iLevel    = iLevel,                                   & 
        &                pdfglobal = pdfGlobal,                                &
        &                params    = params,                                   &
        &                physics   = params%physics,                           &
        &                layout    = scheme( iScheme )%layout,                 &
        &                varSys    = scheme( iScheme )%varSys,                 &
        &                mixture   = scheme( iScheme )%mixture,                &
        &                derVarPos = scheme( iScheme )%derVarPos,              &
        &                globBC    = scheme( iScheme )%globBC )
      call tem_stopTimer( timerHandle = params%setBndTimer,      &
        &                 me          = params%general%timerData )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tSetBnd(iLevel) )


      ! Compute current scheme of current level
      call tem_startTimer( timerHandle = params%advRelTimer,      &
        &                  me          = params%general%timerData )
      call tem_startTimer( me = params%timer(ilevel),                          &
        &                  timerHandle = params%lw_tCompute(iLevel) )
!$omp end single
      ! \todo: SZ: update this subroutine for multiple fields
      call scheme( iScheme )%compute(                                          &
        &  iScheme   = iScheme,                                                &
        &  fieldProp = scheme( iScheme )%field(:)%fieldProp,                   &
        &  inState   = scheme( iScheme )%pdf(iLevel)%state(:,scheme(           &
        &                                     iScheme )%pdf(iLevel)%nNow),     &
        &  outState  = scheme( iScheme )%pdf(iLevel)%state(:,scheme(           &
        &                                    iScheme )%pdf(iLevel)%nNext),     &
        &  neigh     = scheme( iScheme )%pdf(iLevel)%neigh(:),                 &
        &  nStart    = 1,                                                      &
        &  nEnd      = scheme( iScheme )%pdf(iLevel)%nElems_solve,             &
        &  level     = iLevel,                                                 &
        &  layout    = scheme( iScheme )%layout,                               &
        &  pdfGlobal = pdfGlobal,                                              &
        &  params    = params,                                                 &
        &  scheme    = scheme,                                                 &
        &  tree      = geometry%tree,                                          &
        &  varSys    = scheme( iScheme )%varSys ) 
      call tem_stopTimer( timerHandle = params%advRelTimer,      &
        &                 me          = params%general%timerData )
      call tem_stopTimer( me = params%timer(iLevel),                           &
                          timerHandle = params%lw_tCompute(iLevel) )
      buffer=' after  advRel                   '
      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )

      ! Check, if we are on the finest level
      if( iLevel < pdfGlobal%maxLevel ) then
        ! If not, perform interpolation and recursive calls
        do iLoop = 1, params%nNesting 
          ! Perform two time steps on the next finer grid levels
          call debug_recursive_multilevel( scheme, geometry, params, intp,     &
            &                              iLevel+1, pdfGlobal )
          ! Exchange the fine ghost elements
          call mus_exchange(                                                   &
            &    send    = scheme( iScheme )%glob%levelDesc(iLevel+1)%         &
            &                                           sendbufferFromCoarser, &
            &    recv    = scheme( iScheme )%glob%levelDesc(iLevel+1)%         &
            &                                           recvbufferFromCoarser, &
            &    state   = scheme( iScheme )%pdf(iLevel+1)%state(:,scheme(     &
            &                                  iScheme )%pdf(iLevel+1)%nNext), &
            &    pattern = params%general%commPattern,                         &
            &    level   = iLevel,                                             &
            &    proc    = params%general%proc)
        end do
        call debug_recursive_multilevel( scheme, geometry, params, intp,       &
          &                              iLevel+1, pdfGlobal )

        ! Interpolate the coarse ghost elements from the finer ones
        call intp%fillMineFromFiner_low(                                       &
          &      pdf       = scheme( iScheme )%pdf,                            &
          &      glob      = scheme( iScheme )%glob,                           &
          &      scheme    = scheme( iScheme ),                                &
          &      fieldProp = scheme( iScheme )%field(:)%fieldProp,             &
          &      level     = iLevel,                                           &
          &      intp      = intp,                                             &
          &      ind       = scheme( iScheme )%glob%levelDesc( iLevel)%        &
          &                                                 intpFromFiner_low, &
          &      layout    = scheme( iScheme )%layout,                         &
          &      varSys    = scheme( iScheme )%varSys,                         & 
          &      physics   = params%physics,                                   &
          &      derVarPos = scheme( iScheme )%derVarPos)

        ! Interpolate the coarse ghost elements from the finer ones
        call intp%fillMineFromFiner(                                           &
          &      pdf       = scheme( iScheme )%pdf,                            &
          &      glob      = scheme( iScheme )%glob,                           &
          &      scheme    = scheme( iScheme ),                                &
          &      fieldProp = scheme( iScheme )%field(:)%fieldProp,             &
          &      level     = iLevel,                                           &
          &      intp      = intp,                                             &
          &      ind       = scheme( iScheme )%glob%levelDesc( iLevel)%        &
          &                                                intpFromFiner_high, &
          &      layout    = scheme( iScheme )%layout,                         &
          &      varSys    = scheme( iScheme )%varSys,                         & 
          &      physics   = params%physics,                                   &
          &      derVarPos = scheme( iScheme )%derVarPos)

        call tem_startTimer( timerHandle = params%commTimer,        &
          &                  me          = params%general%timerData )

      buffer=' after intp MineFromFiner        '
      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )

        ! Exchange the coarse ghost elements
        call mus_exchange(                                                     &
          &   send    = scheme( iScheme )%glob%levelDesc(iLevel)%              &
          &                                               sendbufferFromFiner, &
          &   recv    = scheme( iScheme )%glob%levelDesc(iLevel)%              &
          &                                               recvbufferFromFiner, &
          &   state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme(          &
          &                                      iScheme )%pdf(iLevel)%nNext), &
          &   pattern = params%general%commPattern,                            &
          &   level   = iLevel,                                                &
          &   proc    = params%general%proc)
        call tem_stopTimer( timerHandle = params%commTimer,        &
          &                 me          = params%general%timerData )
!      buffer=' after exch FromFiner            '
!      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )
      endif

!$omp single
      ! Communicate the halo elements of current scheme on current level 
      call tem_startTimer( timerHandle = params%commTimer,        &
        &                  me          = params%general%timerData )
      call mus_exchange(                                                       &
        &   send    = scheme( iScheme )%glob%levelDesc(iLevel)%sendbuffer,     &
        &   recv    = scheme( iScheme )%glob%levelDesc(iLevel)%recvbuffer,     &
        &   state   = scheme( iScheme )%pdf(iLevel)%state(:,scheme( iScheme )% &
        &                                                  pdf(iLevel)%nNext), &
        &   pattern = params%general%commPattern,                              &
        &   level   = iLevel,                                                  &
        &   proc    = params%general%proc)
      call tem_stopTimer( timerHandle = params%commTimer,        &
        &                 me          = params%general%timerData )
!      buffer=' after exchange buffer           '
!      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )
!$omp end single
 
      if( iLevel < pdfGlobal%maxLevel ) then
        ! Now when all coarser ghosts are present, we can interpolate the other
        ! way round. Interpolate the ghost elements on the finer level with data
        ! provided from current level (This only needs to be done once in two 
        ! finer time steps)
        call intp%fillFinerFromMe_low(                                         &
          &    pdf       = scheme( iScheme )%pdf,                              &
          &    glob      = scheme( iScheme )%glob,                             & 
          &    scheme    = scheme( iScheme ),                                  &
          &    fieldProp = scheme( iScheme )%field(:)%fieldProp,               &
          &    level     = iLevel,                                             &
          &    intp      = intp,                                               &
          &    ind       = scheme( iScheme )%glob%levelDesc( iLevel)%          &
          &                                               intpFromCoarser_low, &
          &    layout    = scheme( iScheme )%layout,                           &
          &    varSys    = scheme( iScheme )%varSys,                           & 
          &    physics   = params%physics,                                     &
          &    derVarPos = scheme( iScheme )%derVarPos)

        call intp%fillFinerFromMe(                                             &
          &    pdf       = scheme( iScheme )%pdf,                              &
          &    glob      = scheme( iScheme )%glob,                             & 
          &    scheme    = scheme( iScheme ),                                  &
          &    fieldProp = scheme( iScheme )%field(:)%fieldProp,               &
          &    level     = iLevel,                                             &
          &    intp      = intp,                                               &
          &    ind       = scheme( iScheme )%glob%levelDesc( iLevel)%          &
          &                                              intpFromCoarser_high, &
          &    layout    = scheme( iScheme )%layout,                           &
          &    varSys    = scheme( iScheme )%varSys,                           & 
          &    physics   = params%physics,                                     &
          &    derVarPos = scheme( iScheme )%derVarPos)

        call tem_startTimer( timerHandle = params%commTimer,        &
          &                  me          = params%general%timerData )
      buffer=' after fillFinerFromMe           '
      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )

        ! Exchange the fine ghost elements
        call mus_exchange(                                                     &
          &        send    = scheme( iScheme )%glob%levelDesc(iLevel+1)%       &
          &                                             sendbufferFromCoarser, &
          &        recv    = scheme( iScheme )%glob%levelDesc(iLevel+1)%       &
          &                                             recvbufferFromCoarser, &
          &        state   = scheme( iScheme )%pdf(iLevel+1)%state(:,scheme(   &
          &                                    iScheme )%pdf(iLevel+1)%nNext), &
          &        pattern = params%general%commPattern,                       &
          &        level   = iLevel,                                           &
          &        proc    = params%general%proc)
        call tem_stopTimer( timerHandle = params%commTimer,        &
        &                   me          = params%general%timerData )
!      buffer=' after exch FromCoarser          '
!      call dump_debug_info( buffer, scheme, params, iLevel, iScheme )

      else  ! we are on finest level
        ! MH: Currently using the dt from the first scheme \todo add correct dt. Which one?
        ! update the time counters. MH: checked. Please dont move
        ! Increasing with the smallest time step (maxLevel)
        ! KM: time is advanced here since new time is required to update sources.
        if( iLevel == pdfGlobal%maxLevel ) then
          call tem_time_advance( me = params%general%simControl%now,           &
            &                    sim_dt = params%physics%dtLvl(             &
            &                                  geometry%tree%global%maxLevel )) 
        endif

!$omp single
        ! check if some action has to be taken:
        ! output, restart, tracking, global reduction operations
!        call check_flow_status( scheme( iScheme ), params, geometry, pdfGlobal, iScheme )
!
!        if( iScheme == pdfGlobal%nSchemes )then
!          ! The last scheme updates then the global time index
!          pdfGlobal%iTime = pdfGlobal%iTime + 1 
!        end if
!$omp end single
      endif ! if not on finest level
    end do ! scheme loop

  end subroutine debug_recursive_multilevel
!******************************************************************************!


!******************************************************************************!
  !> Dump pdf values into debug files
  !!
  subroutine dump_debug_info( buffer, scheme, params, &
    &                         iLevel, iScheme, iTime )
    !---------------------------------------------------------------------------
    character(len=*), intent(in)  :: buffer
    !> scheme type contains pdf
    type( mus_scheme_type )       :: scheme(:)
    !> global parameters
    type( mus_param_type )     :: params
    !> Level counter variable
    integer, intent(in)           :: iLevel
    !> Scheme counter variable
    integer, intent(in)           :: iScheme
    !> integer time counter variable
    integer, intent(in), optional :: iTime
    !---------------------------------------------------------------------------

    if ( main_debug%dumpLevelwiseState ) then
      ! dump current time and level
      write(dbgUnit(1),*) trim( buffer ), params%general%simControl%now%sim
      write(dbgUnit(1),*)'            level      ', iLevel

      ! dump pdf values
      call dumpPdfAll( pdf       = scheme( iScheme )%pdf,                      &
        &              glob      = scheme( iScheme )%glob,                     &
        &              layout    = scheme( iScheme )%layout,                   &
        &              iLevel    = iLevel,                                     &
        &              text      = buffer,                                     &
        &              iTime     = iTime,                                      &
        &              dumpHalos = main_debug%dumpHaloState )
    end if

  end subroutine dump_debug_info
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine dumpBufferFromFiner( scheme, text, iLevel )
    !---------------------------------------------------------------------------
    type( mus_scheme_type ) :: scheme
    character(len=*) :: text
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iElem, elemPos, iDir, pos, iField
    !---------------------------------------------------------------------------

    write(dbgUnit(1),*) 'bufferFromFiner:   '//trim( text )
    do iField = 1, scheme%nFields
      write(dbgUnit(1),*)'field: ', iField
      do iElem = 1, scheme%pdf( iLevel )%nElems_GhostFromFiner
        elemPos = iElem + scheme%glob%levelDesc( iLevel )%                     &
          &                                  offset(1, eT_ghostFromFinerElem)
        write(dbgUnit(1),*) 'TID ', scheme%glob%                               &
          &                     levelDesc( iLevel )%total( elemPos )
        do iDir = 1, scheme%layout%stencil(1)%QQ
          pos = scheme%varSys%Variable%val( iField )%varPos(iDir)
          write(dbgUnit(1),*) qDirName( iDir ), scheme%pdf( iLevel )%state(    &
    ?SAVE?( pos, elemPos, scheme%varSys%nScalars, scheme%pdf( iLevel )%neigh ) &
                                           ,scheme%pdf( iLevel )%nNext )
        enddo
      enddo
    enddo
  end subroutine dumpBufferFromFiner
!******************************************************************************!


!******************************************************************************!
  !>
  !!
  subroutine dumpBufferFromCoarser( scheme, text, iLevel )
    !---------------------------------------------------------------------------
    type( mus_scheme_type ) :: scheme
    character(len=*) :: text
    !> Level counter variable
    integer, intent(in) :: iLevel
    !---------------------------------------------------------------------------
    integer :: iElem, elemPos, iDir, pos, iField
    !---------------------------------------------------------------------------

    write(dbgUnit(1),*)'bufferFromCoarser: '//trim( text )
    do iField = 1, scheme%nFields
      write(dbgUnit(1),*)'field: ', iField
      do iElem = 1, scheme%pdf( iLevel+1 )%nElems_GhostFromCoarser
        elemPos = iElem + scheme%glob%levelDesc( iLevel+1 )%                   &
          &                                offset(1, eT_ghostFromCoarserElem)
        write(dbgUnit(1),*)'TID ', scheme%glob%                                &
          &                     levelDesc( iLevel+1 )%total( elemPos )
        do iDir = 1, scheme%layout%stencil(1)%QQ 
          pos = scheme%varSys%Variable%val( iField )%varPos(iDir)
          write(dbgUnit(1),*)qDirName( iDir ), scheme%pdf( iLevel+1 )%state(   &
  ?SAVE?( pos, elemPos, scheme%varSys%nScalars, scheme%pdf( iLevel+1 )%neigh ) &
                                           ,scheme%pdf( iLevel+1 )%nNext )
        enddo
      enddo
    enddo
  end subroutine dumpBufferFromCoarser
!******************************************************************************!


!******************************************************************************!
  !> Debug the dependencies for the ghostFromFiner elements
  !!
  subroutine debug_dependenciesFromFiner( pdf, glob, minLevel, maxLevel )
    !---------------------------------------------------------------------------
    !> global pdf parameters
    type( pdf_data_type ), intent(in) :: pdf(:)
    type( pdf_global_type ), intent(in) :: glob
    !> Level range
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !---------------------------------------------------------------------------
    integer :: sourceLevel  ! level of source elements
    integer :: sourceElem   ! treeId of current source element
    integer :: targetLevel  ! level of target elements
    integer :: targetElem   ! treeId of current source element
    integer :: iElem        ! current target element (for outer loop)
    integer :: iLevel
    integer :: iSourceElem  ! current source element (for inner loop)
    integer :: nSourceElems ! number of source elements for the current target
    character(len=256) :: buffer
    !---------------------------------------------------------------------------

    write(dbgUnit(1),'(A)') 'Inside routine: debug_dependenciesFromFiner'

    do iLevel = minLevel, maxLevel-1
      write(dbgUnit(1),"(A,I0)") 'Ghost on level: ', iLevel
      ! Treat every coarse target element:
      do iElem = 1, pdf( iLevel )%nElems_ghostFromFiner

        ! find out which is the next finer level
        sourceLevel = glob%levelDesc( iLevel )%depFromFiner( iElem )%           &
          &                                                     dependencyLevel
        ! level of ghostFromFiner element
        targetLevel = iLevel
        if( sourceLevel - targetLevel /= 1 ) then
          write(dbgUnit(1),*)'error in level difference on level ', iLevel,     &
            &                ' elem ', iElem, ' sourceLevel ', sourceLevel
          stop
        endif

        ! find out the treeId of the current target element that we are
        ! treating in this iteration
        targetElem = iElem + glob%levelDesc( targetLevel )%                    &
          &                                  offset( 1, eT_ghostFromFinerElem )
        ! find out, how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
          &                                                          elem%nVals
        write( buffer,'(A,I8,A)')  'ghostID: ', &
          &            glob%levelDesc( targetLevel )%total( targetElem ), &
          &            ', sourceID:'
        !... and all the fine source elements...
        do iSourceElem = 1, nSourceElems
          ! Find out the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromFiner( iElem )%    &
            &                                           elem%val( iSourceElem )
          if( sourceElem > 0 ) then
            write(buffer, '(2a,i8)')  trim(buffer),&
              &   ' ', glob%levelDesc( sourceLevel )%total( sourceElem )
          else
            write(buffer, '(2a,i8)')  trim(buffer), ' x', sourceElem
          endif
        enddo ! iSourceElem
        write(dbgUnit(1),*) trim( buffer )
      enddo ! iElem
    enddo ! iLevel

    write(dbgUnit(1),*) 'Out of routine: debug_dependenciesFromFiner'
    write(dbgUnit(1),*) ''

  end subroutine debug_dependenciesFromFiner
!******************************************************************************!


!******************************************************************************!
  !> Interpolation routine that is based on a simple weighted average of source
  !! nodes. This is the interpolation coarse-> fine. Weights are needed here,
  !! as the distances source <-> target are different for the source nodes.
  !!
  subroutine debug_dependenciesFromCoarser( pdf, glob, minLevel, maxLevel )
    !---------------------------------------------------------------------------
    !> global fluid parameters
    type( pdf_data_type ),   intent(in) :: pdf(:)
    type( pdf_global_type ), intent(in) :: glob
    !> Level range
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !---------------------------------------------------------------------------
    integer :: sourceLevel   ! level of source elements
    integer :: sourceElem    ! treeId of current source element
    integer :: targetLevel   ! level of target elements
    integer :: targetElem    ! treeId of current source element
    integer :: iElem         ! current target element (for outer loop)
    integer :: iSourceElem   ! current source element (for inner loop)
    integer :: nSourceElems  ! number of source elements for the current target
    integer :: iLevel
    character(len=256) :: buffer
    !---------------------------------------------------------------------------

    write(dbgUnit(1),"(A)") 'Inside routine: debug_dependenciesFromCoarser'

    do iLevel = minLevel+1, maxLevel
      sourceLevel = iLevel - 1
      targetLevel = iLevel     !\todo: why not use depFromFiner?

      write(dbgUnit(1),"(A,I0)") 'Ghost on level: ', targetLevel

      ! Treat every finer target element:
      do iElem = 1, pdf( targetLevel )%nElems_ghostFromCoarser
        ! find out the treeId of the current target element
        targetElem = iElem + glob%levelDesc( targetLevel )%                    &
          &                                offset( 1, eT_ghostFromCoarserElem )

        ! number of source elements, 27 (3D), 9 (2D)
        ! but can differ at corners, obstacles, boundaries...
        nSourceElems = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
          &                                                          elem%nVals

        write( buffer,'(A,I8,A)')  'GhostID: ', &
          &       glob%levelDesc(targetLevel)%total( targetElem ), &
          &       ', sourceID:'
        !... and all the fine source elements...
        do iSourceElem = 1, nSourceElems
          ! Find out the source element's treeId
          sourceElem = glob%levelDesc( targetLevel )%depFromCoarser( iElem )%  &
            &                                           elem%val( iSourceElem )

          if( sourceElem > 0 ) then
            write( buffer, '(2a,i8)' )  trim(buffer), &
              &  ' ', glob%levelDesc( sourceLevel )%total( sourceElem )
          else
            write( buffer, '(2a,i8)' )  trim(buffer), ' x', sourceElem
          endif
        enddo ! iSourceElem
        write(dbgUnit(1),*) trim( buffer )

      end do ! iElem
    enddo ! iLevel

    write(dbgUnit(1),"(A)") 'Out of routine: debug_dependenciesFromCoarser'
    write(dbgUnit(1),"(A)") ''

  end subroutine debug_dependenciesFromCoarser
!******************************************************************************!


!******************************************************************************!
  !> 
  !!
  subroutine dump_elemDependency( levelDesc, ghostElem, &
    &                             targetLevel, minlevel )
    !---------------------------------------------------------------------------
    integer, intent(in) :: minlevel
    type( tem_levelDesc_type ), intent(in) :: levelDesc(minlevel:)
    integer, intent(in) :: ghostElem
    integer, intent(in) :: targetLevel 
    !---------------------------------------------------------------------------
    integer :: sourceLevel   ! level of source elements
    integer :: sourceElem    ! treeId of current source element
    integer :: targetElem    ! treeId of current source element
    integer :: iSourceElem   ! current source element (for inner loop)
    integer :: nSourceElems  ! number of source elements for the current target
    character(len=256) :: buffer
    !---------------------------------------------------------------------------

    ! find out the treeId of the current target element that we are treating in
    ! this iteration
    targetElem = ghostElem + levelDesc( targetLevel )%                         &
      &                                    offset( 1, eT_ghostFromCoarserElem )
    sourceLevel = targetLevel - 1
    ! find out, how many coarse source elements we have for interpolation.
    ! Usually 8, but can differ at corners, obstacles, boundaries...
    nSourceElems = levelDesc( targetLevel )%depFromCoarser( ghostElem )%       &
      &                                                              elem%nVals

    write( buffer,'(a, i8)')  'FromCoarser ', levelDesc( targetLevel )%        &
      &                                                     total( targetElem )
    !... and all the fine source elements...
    do iSourceElem = 1, nSourceElems
      ! Find out the source element's treeId
      sourceElem = levelDesc( targetLevel )%depFromCoarser( ghostElem )%       &
        &                                               elem%val( iSourceElem )

      if( sourceElem > 0 ) then
        write( buffer, '(2a,i8)' )  trim(buffer), ' ', levelDesc(              &
          &                                   sourceLevel )%total( sourceElem )
      else
        write( buffer, '(2a,i8)' )  trim(buffer), 'x', sourceElem
      endif
    enddo
    write(dbgUnit(1),*) trim( buffer )

  end subroutine dump_elemDependency
!******************************************************************************!


!******************************************************************************!
  !> Debug the finding of neighbors
  !!
  subroutine debug_find_neighbor( nSides, boundary, treeidsubset,              &
    &                             propertylistsubset, nElemsSubset )
    !---------------------------------------------------------------------------
    integer, intent(in)                  :: nSides
    type( tem_bc_prop_type ), intent(in) :: boundary
    integer(kind=long_k), intent(in)     :: treeidsubset(:)
    integer(kind=long_k), intent(in)     :: propertylistsubset(:)
    integer, intent(in)                  :: nElemsSubset
    !---------------------------------------------------------------------------
    integer              :: iElem
    integer(kind=long_k) :: myProperty
    integer              :: x(4)
    integer :: nElemsBnd, iDirection
    logical :: hasBoundary, hasBndInDir
    integer(kind=long_k) :: tOffset, origID    ! List tracking pointer
    character(len=1024) :: buffer
    !---------------------------------------------------------------------------
    nElemsBnd = 0

    write(logUnit(1),*) 'Debugging neighbor search  ...' 
    write(dbgUnit(1),*) 
    write(dbgUnit(1),*) 'List of treeIDs' 
    write(dbgUnit(1),*) int( treeidsubset )
    write(dbgUnit(1),*) 
    write(dbgUnit(1),*) 'Debugging neighbor search ... ' 
    write(dbgUnit(1),*) 

    elemLoop: do iElem = 1, nElemsSubset

      x = tem_CoordOfId( treeidsubset( iElem ))
      write(dbgUnit(1),*)'treeID ',int( treeIDsubset( iElem ))
      tOffset = tem_FirstIDAtLevel( x(4) )
      origID = tem_IdOfCoord( x, tOffset ) !< MH: was passiert hier?
      myProperty = propertylistsubset( iElem )
      hasBoundary = .false.

      !
      ! First check, if I have boundaries
      ! If yes, there might be no neighbors
      if( btest( myProperty, prp_hasBnd )) then
        hasBoundary = .true.
        nElemsBnd = nElemsBnd +1
      endif
      !
      ! start searching for neighbors stored in nTreeID
      ! and create linked lists in neighborList
      !
      !
      buffer = ''
      directionLoop: do iDirection = 1, nSides
        hasBndInDir = .false.
        if( hasBoundary ) then
          if( boundary%boundary_ID(( iDirection ),nElemsBnd ) > 0) then
            hasBndInDir = .true. !< PUSH
          endif
        endif
        !
        ! If there is NO boundary in the current direction
        if( .not. hasBndInDir ) then
!          write(buffer,'(3a,i4)') trim(buffer), qDirName(iDirection), &
!          '  ',int(neighbor(ielem)%elemId(iDirection))
        else
          write(buffer,'(3a,i4)') trim( buffer ), qDirName( iDirection ),' bnd '
        endif
      end do directionLoop
      write(dbgUnit(1),*) trim( buffer )
    end do elemLoop
    write(dbgUnit(1),*)
    write(dbgUnit(1),*) 'Finished debugging neighbor search.' 
    call tem_horizontalSpacer( fUnit = dbgUnit(1))
    write(dbgUnit(1),*) 

  end subroutine debug_find_neighbor
!******************************************************************************!


!******************************************************************************!
  !> \brief Dump the neighbors found for the boundaries with neighbor elements
  !!
  subroutine debug_normals( tree, scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    type( treelmesh_type ), intent(in)    :: tree  
    type(pdf_globInfo_type), intent(in)   :: pdfGlobal
    type( mus_scheme_type ) :: scheme
    !--------------------------------------------------------------------------- 
    integer :: iElem, iLevel, iBnd, iNeigh
    integer(kind=long_k) :: nTreeID
    !--------------------------------------------------------------------------- 

    write(dbgUnit(1),*) 'Debugging normals...'
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      do iBnd = 1, pdfGlobal%nBCs
        write(dbgUnit(1),*)'Boundary ID ', iBnd,' at Level ', iLevel
        write(dbgUnit(1),*)'following tIDs with neighbor relations' 
        do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
          nTreeID = scheme%glob%levelDesc( iLevel )%total(                     &
            &              scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ))
          write(dbgUnit(1),*) 'tID ', nTreeID, ' pos ',                        &
            &                  tem_BaryOfId( tree, nTreeID )
          write(dbgUnit(1),*) 'normal ', scheme%globBC( iBnd )%elems( iLevel )%&
            &                                              normal( :, iElem ), &
            &                ' nrmInd ', scheme%globBC( iBnd )%elems( iLevel )%&
            &                                           normalInd( iElem ),    &
            &                ' nNeighs ', scheme%field(1)%bc( iBnd )%nNeighs 

          do iNeigh = 1, scheme%field(1)%bc( iBnd )%nNeighs
            write(dbgUnit(1),*) '   ', iNeigh, ' neigh ',                      &
              &        scheme%glob%levelDesc( iLevel )%total( scheme%field(1)% &
              &        bc( iBnd )%elems( iLevel )%neigh( iNeigh, iElem )),      &
              &        'bardofID', tem_baryOfId( tree, scheme%glob%            &
              &                        levelDesc( iLevel )%total( scheme%      &
              &                        field(1)%bc( iBnd )%elems( iLevel )%    &
              &                                     neigh( iNeigh, iElem )))
          enddo
        enddo
      enddo
    enddo

  end subroutine debug_normals
!******************************************************************************!


!******************************************************************************!
  !> Detailed Debug output to the PDF neighbor (connectivity) list
  !!
  !! currently the dubug connectivity is done only for 1st field
  subroutine debug_connectivity( scheme, minLevel, maxLevel )
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(in) :: scheme
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !---------------------------------------------------------------------------
    integer(kind=long_k) :: iElem
    integer :: iLevel, iDir, iField
    integer :: iQQ, varPos(scheme%layout%stencil(1)%QQ), nScalars
    !---------------------------------------------------------------------------

    iQQ = scheme%layout%stencil(1)%QQ
    ! take varpos of 1st field
    nScalars = scheme%varSys%nScalars

    write(dbgUnit(1),*)
    write(dbgUnit(1),*) ' Debugging connectivity for the First Field'
    write(dbgUnit(1),*)
    write(dbgUnit(1),*) 'NEIGHBOR level from ', minLevel, ' to ', maxLevel

    do iLevel = minLevel, maxLevel
      write(dbgUnit(1),*)
      write(dbgUnit(1),"(A,I0)") 'LEVEL: ', iLevel
      write(dbgUnit(1),"(A,I0)") 'nElems fluid ', scheme%pdf( iLevel )%nElems_fluid
      write(dbgUnit(1),"(A,I0)") 'nElems solve ', scheme%pdf( iLevel )%nElems_solve
      write(dbgUnit(1),"(A,I0)") 'nElems halo  ', scheme%pdf( iLevel )%nElems_halo
      write(dbgUnit(1),"(A,I0)") 'nElems ghost ', scheme%pdf( iLevel )%nElems_ghost
      write(dbgUnit(1),"(A,I0)") 'nElems local ', scheme%pdf( iLevel )%nElems_local
      write(dbgUnit(1),*) 'size of total ', &
        &                 size(scheme%glob%levelDesc( iLevel )%total)
      write(dbgUnit(1),*) 'size of state ', &
        &                 size(scheme%pdf( iLevel )%state(:,1))
      write(dbgUnit(1),*) 'size of neigh ', size(scheme%pdf( iLevel )%neigh )

      write(dbgUnit(1),*) 'Loop over all local elements:'
      write(dbgUnit(1),"(A10,A20,4A8)") 'iElem', 'treeID', 'X', 'Y', 'Z', 'level'
      do iElem = 1, scheme%pdf( iLevel )%nElems_local

        ! write element info
        write(dbgUnit(1),"(I10,I20,4I8)") iElem, &
          &  int(scheme%glob%levelDesc( iLevel )%total( iElem )), &
          &  tem_CoordOfId( scheme%glob%levelDesc( iLevel )%total( iElem ))

        ! do iField = 1, scheme%nFields

        iField = 1
        varPos = scheme%varSys%Variable%val( iField )%varPos
        ! write(dbgUnit(1),*)'Field ', iField

        write(dbgUnit(1),"(3A10,4A8)") 'varPos', 'qDirName', 'treeID', 'X', 'Y', 'Z', 'level'
        do iDir = 1, iQQ
          if( scheme%pdf( iLevel )%neigh  &
            &( ?NGPOS?( varPos(iDir), iElem, nScalars ) )  <= 0 ) then
            write(dbgUnit(1),*) 'ERROR in qDirName: ',qDirName( iDir )
          else
            write(dbgUnit(1),"(I10,A10,I10,4I8)") varPos( iDir ), &
              &                 qDirName( iDir ), &
              &  int(scheme%glob%levelDesc( iLevel )%total( &
  &?ElemIDX?(scheme%pdf( iLevel )%neigh( ?NGPOS?( varPos(iDir), iElem, nScalars )), nScalars ))),&
              &  tem_CoordOfId(   scheme%glob%levelDesc( iLevel )%total( &
  &?ElemIDX?(scheme%pdf( iLevel )%neigh( ?NGPOS?( varPos(iDir), iElem, nScalars )), nScalars )))
          endif
        end do ! iDir
        write(dbgUnit(1),*) ''

        ! end do  ! iField

      end do ! iElem

    end do ! iLevel

    write(dbgUnit(1),*) ' End Debugging connectivity '

    call tem_horizontalSpacer( fUnit = dbgUnit(1))

  end subroutine debug_connectivity
!******************************************************************************!


end module mus_debug_module
!******************************************************************************!
