! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This routine serves as a wrapper to call other routines responsible for
!! dynamic load balancing in Musubi. The dynamic load balancing algorithm
!! follows the SPARTA algorithm in an iterative way to perform balancing for
!! Multilevel simulations as well as complex boundary conditions.
!!
!! \author Kartik Jain, Simon Zimny

module mus_dynLoadBal_module

  ! include treelm modules
  use mpi
  use env_module,              only: PathLen, rk, long_k
  use tem_timer_module,        only: tem_stopTimer, tem_resetTimer
  use tem_restart_module,      only: tem_restart_readHeader, tem_init_restart, &
    &                                tem_restart_finalize
  use tem_timeControl_module,  only: tem_timeControl_start_at_sim
  use tem_geometry_module,     only: tem_setEffBoundingBox
  use tem_aux_module,          only: tem_abort
  use tem_logging_module,      only: logUnit, tem_logging_type
  use tem_stlb_io_module,      only: tem_dump_stlb
  use tem_surfaceData_module,  only: tem_freeSurfData, tem_readAndUnify_surfData
  use tem_tracking_module,     only: tem_init_tracker
  use tem_balance_module,      only: Bal_logUnit
  use tem_tools_module,        only: tem_horizontalSpacer
  use tem_global_module,       only: tem_global_type

  ! include musubi modules
  use mus_bc_general_module,         only: init_boundary
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_param_module,              only: mus_param_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_geom_module,               only: mus_geom_type, mus_geom_cleanWeight
  use mus_config_module,             only: mus_load_bc_data
  use mus_tools_module,              only: calc_MLUPS
  use mus_interpolate_header_module, only: interpolation_type
  use mus_interpolate_module,        only: mus_init_interpolate
  use mus_construction_module,       only: mus_construct,                      &
    &                                      mus_construct_cleanup
  use mus_aux_module,                only: mus_balance,                        &
    &                                      fillHelperElementsCoarseToFine,     &
    &                                      fillHelperElementsFineToCoarse
  use mus_time_module,               only: mus_timeControl_homogenize
  use mus_restart_module,            only: mus_readRestart, mus_writeRestart

  implicit none

  private

  public :: mus_perform_dynLoadBal
  public :: dump_slotMLUPS

contains

!******************************************************************************!
  !> Wrap up the routines required for dynamic load balancing
  subroutine mus_perform_dynLoadBal( scheme, pdfGlobal, params, geometry, intp )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), allocatable, intent(inout) :: scheme(:)
    !> global information
    type( pdf_globInfo_type ),  intent(inout) :: pdfGlobal
    !> Global parameters
    type( mus_param_type ),     intent(inout) :: params
    !> Treelmesh data
    type( mus_geom_type ),      intent(inout) :: geometry
    !> interpolation method
    type( interpolation_type ), intent(inout) :: intp
    !---------------------------------------------------------------------------
    integer :: iScheme, iField, fileID
    character(len=PathLen) :: stlFilename
    !---------------------------------------------------------------------------
    !                 Initialize the balance restart for writing               !
    !---------------------------------------------------------------------------
    params%general%balance%restart%controller%writeRestart = .true.
    params%general%balance%restart%controller%readRestart = .false.

    call tem_init_restart( me      = params%general%balance%restart,          &
      &                    solver  = params%general%solver,                   &
      &                    nSys    = pdfGlobal%nSchemes,                      &
      &                    varSys  = scheme(:)%varSys,                        &
      &                    tree    = geometry%tree )

    !---------------------------------------------------------------------------
    !         Balance the tree and free all arrays for reconstruction          !
    !---------------------------------------------------------------------------
    !
    ! Invoke the routines for dynamic load balancing, unload the tree
    call mus_balance( scheme    = scheme,                                      &
      &               pdfGlobal = pdfGlobal,                                   &
      &               params    = params,                                      &
      &               geometry  = geometry )

    ! Destroy the arrays which were allocated for performing new construction
    call mus_construct_cleanup( scheme    = scheme,                            &
      &                         pdfGlobal = pdfGlobal,                         &
      &                         intp      = intp,                              &
      &                         geometry  = geometry )

    !---------------------------------------------------------------------------
    !         Load the restart header, mesh and BCs for reconstruction         !
    !---------------------------------------------------------------------------
    !
    call mus_restart_dynLoadBal( params    = params,                           &
      &                          geometry  = geometry,                         &
      &                          scheme    = scheme,                           &
      &                          pdfGlobal = pdfGlobal )

    write( Bal_logUnit(3),*)'Number of local elements after reloading the '//  &
      &                     ' mesh: ', geometry%tree%nElems

    !---------------------------------------------------------------------------
    !           Reconstruct the levelDesc, reinitialize the simulation         !
    !---------------------------------------------------------------------------
    !
    call mus_construct( scheme    = scheme,                                    &
      &                 geometry  = geometry,                                  &
      &                 params    = params,                                    &
      &                 pdfGlobal = pdfGlobal,                                 &
      &                 intp      = intp )

    call mus_init_dynLoadBal(  scheme    = scheme,                             &
      &                        geometry  = geometry,                           &
      &                        params    = params,                             &
      &                        pdfGlobal = pdfGlobal,                          &
      &                        intp      = intp )

  end subroutine mus_perform_dynLoadBal
!******************************************************************************!


!******************************************************************************!
  !> This subroutine reads the restart header file, the mesh and the boundary
  !! for dyn load balancing.
  !!
  subroutine mus_restart_dynLoadBal( params, geometry, scheme, pdfGlobal )
    !---------------------------------------------------------------------------
    !> Global parameters
    type( mus_param_type ),     intent(inout) :: params
    !> Treelmesh data
    type( mus_geom_type ),      intent(inout) :: geometry
    !> scheme type
    type( mus_scheme_type ), allocatable, intent(inout) :: scheme(:)
    !> global information
    type( pdf_globInfo_type ),  intent(inout) :: pdfGlobal
    !---------------------------------------------------------------------------
    ! tmp global type for keeping old tree information
    type( tem_global_type ) :: locGlobal
    character(len=PathLen)    :: filename
    !---------------------------------------------------------------------------

    !---------------------------------------------------------------------------
    !     Load the restart header and the mesh for dyn load balancing          !
    !---------------------------------------------------------------------------
    ! set the read filename
    params%general%balance%restart%controller%readFileName =                   &
      &       trim( params%general%balance%restart%controller%writePrefix)//   &
      &       trim( params%general%solver%simName )//                          &
      &       '_lastHeader.lua'
    write(logUnit(3),*)'Reading restart file for dynamic load balancing'
    ! read the header to get various parameters and read the tree
    call tem_restart_readHeader(                                               &
      &                        me          = params%general%balance%restart,   &
      &                        tree        = geometry%tree,                    &
      &                        timing      = params%general%simControl%now,    &
      &                        offset      = geometry%offset,                  &
      &                        proc        = params%general%proc,              &
      &                        readMesh    = .true.)

    !---------------------------------------------------------------------------
    !               Initialize the balance restart for reading                 !
    !---------------------------------------------------------------------------
    call tem_restart_finalize( me = params%general%balance%restart )
    params%general%balance%restart%controller%writeRestart = .false.
    params%general%balance%restart%controller%readRestart = .true.

    call tem_init_restart( me      = params%general%balance%restart,           &
      &                    solver  = params%general%solver,                    &
      &                    nSys    = pdfGlobal%nSchemes,                       &
      &                    varSys  = scheme(:)%varSys,                         &
      &                    tree    = geometry%tree )


    ! If there is a restart, the timings in the params type have to be updated
    ! to those read from the restart
    call tem_timeControl_start_at_sim(                                         &
      & me = params%general%simControl%timeControl,                            &
      & now = params%general%simControl%now )

    ! calculate the effective origin and effective length
    ! KM :\todo get effective bounding box from seeder header
    call tem_setEffBoundingBox( geometry%tree )

    ! copy the current tree global info
    locGlobal = geometry%tree%global
    write(filename,'(a,i1,a)')                                                 &
      &           trim(params%general%balance%restart%controller%writePrefix), &
      &           params%general%balance%fileID,'_'
    ! set the new filename
    geometry%tree%global%dirname    = trim(filename)

    ! Boundary conditions are loaded even in the case of dynamic load balancing
    ! when the restart is done at requisite load balancing intervals
    !
    ! bnd.lsb and qVal.lsb is read in this routine
    !
    call mus_load_bc_data( geometry, params, .false. )

    ! copy back original global type of tree
    geometry%tree%global = locGlobal

  end subroutine mus_restart_dynLoadBal
!******************************************************************************!


!******************************************************************************!
  !> This subroutine initializes musubi after a dynamic load balancing is
  !! performed.
  !!
  subroutine mus_init_dynLoadBal( scheme, pdfGlobal, params, geometry, intp )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), allocatable, intent(inout) :: scheme(:)
    !> global information
    type( pdf_globInfo_type ),  intent(inout) :: pdfGlobal
    !> Global parameters
    type( mus_param_type ),     intent(inout) :: params
    !> Treelmesh data
    type( mus_geom_type ),      intent(inout) :: geometry
    !> interpolation method
    type( interpolation_type ), intent(inout) :: intp
    !---------------------------------------------------------------------------
    integer :: iScheme, iLevel, iTimeLayer, iTrack
    integer :: nTracks
    integer :: tLayerMin, tLayerMax
    !---------------------------------------------------------------------------

    !---------------------------------------------------------------------------
    !              Load the restart data for dyn load balancing                !
    !---------------------------------------------------------------------------
    !
    write(logUnit(1),*) 'Reading restart for dyn load balancing ...'
!    write(logUnit(1),*) 'Filename: '//trim(params%general%balance%restart%controller%readFileName)
!    write(logUnit(1),*) 'Filename: '//trim(params%general%balance%restart%header%binName(1))
    !\todo KM: why do we need this call? balance%restart%controller%timeControl is never used
!KM!    call mus_timeControl_homogenize(                                           &
!KM!      &    me  = params%general%balance%restart%controller%timeControl,        &
!KM!      &    dt  = params%physics%dtLvl( geometry%tree%global%maxLevel )  )
    call mus_readRestart( pdfGlobal = pdfGlobal,                               &
      &                   restart   = params%general%balance%restart,          &
      &                   scheme    = scheme,                                  &
      &                   tree      = geometry%tree )
    write(logUnit(1),*) 'Done reading restart for dyn load balancing ...'

    !---------------------------------------------------------------------------
    !              Debug write the restart after reading it                    !
    !---------------------------------------------------------------------------
    !
!    call tem_restart_finalize( me      = params%general%balance%restart )

!    params%general%balance%restart%controller%writeRestart = .true.
!    params%general%balance%restart%controller%readRestart = .false.
!
!    call tem_init_restart( me      = params%general%balance%restart,           &
!      &                    solver  = params%general%solver,                    &
!      &                    nSys    = pdfGlobal%nSchemes,                       &
!      &                    varSys  = scheme(:)%varSys,                         &
!      &                    tree    = geometry%tree )
!
!
!    call mus_writeRestart( pdfGlobal = pdfGlobal,                              &
!      &                    restart   = params%general%balance%restart,         &
!      &                    scheme    = scheme,                                 &
!      &                    tree      = geometry%tree,                          &
!      &                    timing    = params%general%simControl%now,          &
!      &                    timerData = params%general%timerData,               &
!      &                    timerHandle = params%wRestartTimer                  )
!
!    call tem_restart_finalize( me      = params%general%balance%restart )
!
!!    call tem_abort()

    !---------------------------------------------------------------------------
    !                      Reinitialize the global restart                     !
    !---------------------------------------------------------------------------
    ! 1. finalize restart by freeing all mpi types
    call tem_restart_finalize( me = params%general%restart )
    ! 2. initialize restart again
    call tem_init_restart( me      = params%general%restart,                   &
      &                    solver  = params%general%solver,                    &
      &                    nSys    = pdfGlobal%nSchemes,                       &
      &                    varSys  = scheme(:)%varSys,                         &
      &                    tree    = geometry%tree )



    !---------------------------------------------------------------------------
    !              Initialize tracking, interpolation and boundaries           !
    !---------------------------------------------------------------------------

    ! \todo: SZ: needs to be added?
    !  - reinitialize the sources here?

    ! loop over the schemes
    do iScheme = 1, pdfGlobal%nSchemes

      !-------------------------------------------------------------------------
      !                  Reinitialize the tracking objects                     !
      !-------------------------------------------------------------------------
      write(logUnit(1),*) 'Resetting the tracker ...'
      nTracks = size( params%general%balance%trackHeaders(iScheme)%header )

      ! 1. deallocate the tracker
      if( allocated( scheme( iScheme )%track ))then
        deallocate( scheme( iScheme )%track )
      end if
      ! 2. reallocate the tracker with the number of tracking headers stored
      !    during load from config ( = #Tracking objects in the config file ) 
      allocate( scheme( iScheme )%track( nTracks ))

      ! 3. loop over all tracking objects ...
      do iTrack = 1, nTracks
        ! ... and copy the header information
        scheme( iScheme )%track( iTrack )%header =                             &
          &       params%general%balance%trackHeaders(iScheme)%header( iTrack )
        ! ... and homogenize the timeControl
        call mus_timeControl_homogenize(                                       &
          &        me = scheme( iScheme )%track( iTrack )%header%timeControl,  &
          &        dt = params%physics%dtLvl( geometry%tree%global%maxLevel ), &
          &        reqInt = params%reqInterval )
      end do

      write(logUnit(1),*) 'Initializing tracker ...'
      call tem_init_tracker( me           = scheme(iScheme)%track,             &
        &                    tCtrl        = scheme(iScheme)%trackingControl,   & 
        &                    levelPointer = pdfGlobal%levelPointer,            &
        &                    tree         = geometry%tree,                     &
        &                    proc         = params%general%proc,               &
        &                    solver       = params%general%solver,             &
        &                    varSys       = scheme( iScheme )%varSys)

      !-------------------------------------------------------------------------
      !                    Reinitialize the interpolation                      !
      !-------------------------------------------------------------------------

      write(logUnit(1),*) 'Initializing interpolation...'
      if( pdfGlobal%minLevel /= pdfGlobal%maxlevel ) then
        ! initialize the interpolation
        call mus_init_interpolate(                                             &
          &                  intp      = intp,                                 &
          &                  glob      = scheme(iScheme)%glob,                 &
          &                  scheme    = scheme( iScheme ),                    &
          &                  fieldProp = scheme( iScheme )%field(:)%fieldProp, &
          &                  tree      = geometry%tree,                        &
          &                  layout    = scheme( iScheme )%layout,             &
          &                  pdfGlobal = pdfGlobal, params = params )
      end if

      ! Fill all elements (ghost, halo) with valid values from fluid elements
      ! for both time layers now and next
      tLayerMin = 1
      if( params%init_allTimeLayers ) then
        ! Initialize all time layers
        tLayerMax = 2
      else
        ! Initialize only nNext ( iTimeLayer = 1 )
        tLayerMax = 1
      end if

!      write(logUnit(1),*) 'Filling helper elements ...'
!      write(logUnit(1),*) 'Init all timeLayers: ', params%init_allTimeLayers
!      write(logUnit(1),*) 'tLayerMin:           ', tLayerMin
!      write(logUnit(1),*) 'tLayerMax:           ', tLayerMax
      do iTimeLayer = tLayerMax, tLayerMin, -1
        call fillHelperElementsFineToCoarse( scheme     = scheme(iScheme),     &
          &                                  geometry   = geometry,            &
          &                                  params     = params,              &
          &                                  iLevel     = pdfGlobal%minLevel,  &
          &                                  intp       = intp,                &
          &                                  pdfGlobal  = pdfGlobal,           &
          &                                  iTimeLayer = iTimeLayer  )
!        write(logUnit(1),*) 'Done filling helper elements FineToCoarse.'
        call fillHelperElementsCoarseToFine( scheme     = scheme(iScheme),     &
          &                                  geometry   = geometry,            &
          &                                  params     = params,              &
          &                                  iLevel     = pdfGlobal%minLevel,  &
          &                                  intp       = intp,                &
          &                                  pdfGlobal  = pdfGlobal,           &
          &                                  iTimeLayer = iTimeLayer  )
!        write(logUnit(1),*) 'Done filling helper elements CoarseToFine.'
      end do

!      write(logUnit(1),*) 'Done filling helper elements.'

      ! And set back the time layer to standard to interpolate the correct
      ! time layer in the interpolation module
      intp%timeLayer = 1

      !-------------------------------------------------------------------------
      !                    Reinitialize the boundaries                         !
      !-------------------------------------------------------------------------

      ! initialize each field boundary by looping over the field inside
      ! init_boundary_field
!      write(logUnit(1),*) 'Initializing boundary conditions...'
      call init_boundary( field        = scheme(iScheme)%field,                &
        &                 pdf          = scheme(iScheme)%pdf,                  &
        &                 tree         = geometry%tree,                        &
        &                 glob         = scheme(iScheme)%glob,                 &
        &                 pdfGlobal    = pdfGlobal,                            &
        &                 layout       = scheme(iScheme)%layout,               &
        &                 schemeHeader = scheme(iScheme)%header,               & 
        &                 varSys       = scheme(iScheme)%varSys,               &
        &                 derVarPos    = scheme( iScheme)%derVarPos,           &
        &                 globBC       = scheme( iScheme )%globBC,             &
        &                 bc_prop      = geometry%boundary )
!      write(logUnit(1),*) 'Done initializing boundary conditions.'

    end do !scheme loop

    if( allocated(params%level_weight) ) deallocate( params%level_weight )
    call mus_geom_cleanWeight( geometry )

    ! Reset all the DLB timer values to zero except for the last restart
    ! slot
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      params%timer(iLevel)%value(:)           = 0.
    end do

  end subroutine mus_init_dynLoadBal
!******************************************************************************!


!******************************************************************************!
  !> Calculate and dump the MLUPS or MFLUPS for the simulation.
  !! This is used when dumping the mlups for the individual dynLB slots.
  !!
  subroutine dump_slotMLUPS( scheme, pdfGlobal, params, last )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), intent(in) :: scheme(:)
    !> global information
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !> global parameter
    type(mus_param_type), intent(inout) :: params
    !> is this routine called for the last time?
    logical, intent(in) :: last
    !---------------------------------------------------------------------------
    real(kind=rk) :: locMLUPS
    integer :: lastIter
    integer :: ierr
    !---------------------------------------------------------------------------

    ! Update the dynLB slot
    params%general%balance%iSlot = params%general%balance%iSlot + 1

    ! mpi barrier to make sure that timings are correct
    call mpi_barrier( params%general%proc%comm, ierr )

    ! stop the timer for this slot
    call tem_stopTimer( timerHandle = params%general%balance%slotTimer,        &
      &                 me          = params%general%balance%timings )

    if( params%general%proc%rank == params%general%proc%root ) then

      call tem_horizontalSpacer( params%general%balance%logger%funit(0) )

      lastIter = params%general%simControl%now%Iter &
        &      - params%general%balance%lastIter
      ! calculate the MLUPS for the last dynLB slot and dump it to file
      locMLUPS = calc_MLUPS(                                                     &
        &             pdfGlobal    = pdfGlobal,                                  &
        &             params       = params,                                     &
        &             nElems       = scheme(1)%glob%nElems_solveLevel,           &
        &             iter         = lastIter,                                   &
        &             mainloopTime = params%general%balance%timings%value(       &
        &                                params%general%balance%slotTimer ))
      write(params%general%balance%logger%funit(1),'(a15,i3,a9,i5,a11,en14.2)')  &
        &           'MLUPS for slot ', params%general%balance%iSlot,             &
        &                      ' ran for ', lastIter, ' iters is: ', locMLUPS

      call tem_horizontalSpacer( params%general%balance%logger%funit(0) )

    end if

    if( .not. last )then
      ! update the lastIter with the current iteration number
      params%general%balance%lastIter = params%general%simControl%now%iter
      ! reset the timer for the next slot
      call tem_resetTimer( timerHandle = params%general%balance%slotTimer, &
        &                  me          = params%general%balance%timings )
    end if

  end subroutine dump_slotMLUPS
!******************************************************************************!


end module mus_dynLoadBal_module
!******************************************************************************!
