!******************************************************************************!
!> This module contains data types and routines to define external force fields
module mus_externalForce_module
  use env_module,               only: rk, globalMaxLevels
  use tem_tools_module,         only: tem_horizontalSpacer
  use tem_aux_module,           only: tem_abort
  use tem_spacetime_fun_module, only: tem_spacetime_fun_type, tem_load_spacetime
  use tem_logging_module,       only: logUnit

  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode
  use aotus_module,     only: flu_State, aot_get_val,                          &
    &                         aoterr_Fatal, aoterr_NonExistent,                &
    &                         aoterr_WrongType, flu_State

  use mus_physics_module, only: mus_physics_type
  use mus_pdf_module,     only: pdf_globInfo_type
implicit none

private

public :: mus_externalForce_type
public :: mus_load_externalForce

!> contains all externalForce fields
!! \todo replace scalar spacetime function to vector space time
type mus_externalForce_type
  !> external electrical externalForce
  type(tem_spacetime_fun_type) :: electricField(3)
  !> active electric Field
  logical :: elecFld_active
  !> gravitational externalForce
  type(tem_spacetime_fun_type) :: gravityField(3)
  !> active gravity Field
  logical :: gravFld_active
  !> set active = .true. if externalForce table is defined
  logical :: active
end type mus_externalForce_type

contains

!******************************************************************************!
  !> This routine load externalForce table from scheme table
  !! Currently supported forces: electric field [N/Coulomb] and 
  !! gravity field [m/s^2].
  subroutine mus_load_externalForce( me, conf, path, parent, pdfGlobal, physics )
    !--------------------------------------------------------------------------!
    !> contains force information
    type( mus_externalForce_type ), intent(out) :: me
    type( flu_State ) :: conf !< flu state
    !> lua path
    type( aot_path_type ),intent(in) :: path
    integer, intent(in), optional :: parent !< parent lua handle
    type( pdf_globInfo_type), intent(in) :: pdfGlobal !< global pdf info
    !> physics type to convert physics to lattice unit or vice versa
    type( mus_physics_type ), intent(in) :: physics
    !--------------------------------------------------------------------------!
    integer :: extForce_handle
    integer :: errFatal(3)
    ! aotus paths for forces
    type( aot_path_type ) :: extForce_path
    logical :: field_active(3)
    integer :: iComp
    !--------------------------------------------------------------------------!
    errFatal = aotErr_Fatal

    if(present(parent)) then
      call aot_table_open( L = conf, parent = parent,                          &
        &                  thandle = extForce_handle,                          &
        &                  key = 'external_force')
    else
      call aot_table_open( L = conf, thandle = extForce_handle,                &
        &                  key = 'external_force')
    endif

    !> if mixture handle is not defined  
    if ( extForce_handle == 0 ) then
      me%active = .false.
      write(logUnit(10),*) ' No external force table defined'
    else
      call tem_horizontalSpacer(fUnit=logUnit(1))
      write(logUnit(1),*)' Loading external force information'
      me%active = .true.
      !set path for externalForce field table
      extForce_path = path
      call aot_path_addNode(me = extForce_path, NodeType = 'table', &
        &                   key = 'external_force') 
      !load external electric field
      call tem_load_spacetime( me     = me%electricField(1),                   &
        &                      conf   = conf,                                  &
        &                      path   = extForce_path,                         &
        &                      parent = extForce_handle,                       &
        &                      key    = 'electric_fieldX')

      call tem_load_spacetime( me     = me%electricField(2),                   &
        &                      conf   = conf,                                  &
        &                      path   = extForce_path,                         &
        &                      parent = extForce_handle,                       &
        &                      key    = 'electric_fieldY')

      call tem_load_spacetime( me     = me%electricField(3),                   &
        &                      conf   = conf,                                  &
        &                      path   = extForce_path,                         &
        &                      parent = extForce_handle,                       &
        &                      key    = 'electric_fieldZ')
      
      ! check whether this field is defined
      me%elecFld_active = .false.
      field_active = .true.
      do iComp = 1,3
        if(trim(me%electricField(iComp)%fun_kind) == 'none')                 &
          &  field_active = .false.
      end do

      if (all(field_active)) me%elecFld_active = .true.

      !load gravitational field
      call tem_load_spacetime( me     = me%gravityField(1),                    &
        &                      conf   = conf,                                  &
        &                      path   = extForce_path,                         &
        &                      parent = extForce_handle,                       &
        &                      key    = 'gravity_fieldX')

      call tem_load_spacetime( me     = me%gravityField(2),                    &
        &                      conf   = conf,                                  &
        &                      path   = extForce_path,                         &
        &                      parent = extForce_handle,                       &
        &                      key    = 'gravity_fieldY')

      call tem_load_spacetime( me     = me%gravityField(3),                    &
        &                      conf   = conf,                                  &
        &                      path   = extForce_path,                         &
        &                      parent = extForce_handle,                       &
        &                      key    = 'gravity_fieldZ')

      ! check whether this field is defined
      me%gravFld_active = .false.
      field_active = .true.
      do iComp = 1,3
        if(trim(me%gravityField(iComp)%fun_kind) == 'none')                  &
          &  field_active = .false.
      end do

      if (all(field_active)) me%gravFld_active = .true.

    endif   
    call aot_table_close(L=conf, thandle=extForce_handle)
    call tem_horizontalSpacer(fUnit=logUnit(1))

  end subroutine mus_load_externalForce
!******************************************************************************!

end module mus_externalForce_module

