!******************************************************************************!
!> This module contains information about all fields like fluid, 
!! species, temperature etc. This field type will be used for
!! multispecies and passive scalar transport.
!!
!! \author Kannan Masilamani, Simon Zimny
!!
!! [mus_field_prop]: @ref mus_field_prop_module "mus_field_prop_module"
!!
module mus_field_module

  ! include treelm modules
  use env_module,               only: labelLen
  use treelmesh_module,         only: treelmesh_type
  use tem_bc_prop_module,       only: tem_bc_prop_type
  use tem_tools_module,         only: tem_horizontalSpacer
  use tem_aux_module,           only: tem_abort, tem_CheckLabel
  use tem_ini_condition_module, only: tem_ini_condition_type, tem_load_ic
  use tem_source_module,        only: tem_source_type, tem_load_source
  use tem_restart_module,       only: tem_restart_type
  use tem_logging_module,       only: logUnit

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val, aoterr_NonExistent
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode
  use aot_out_module,   only: aot_out_type, aot_out_open_table,                &
    &                         aot_out_close_table, aot_out_val

  ! include musubi modules
  use mus_bc_header_module,     only: boundary_type, mus_load_bc,              &
    &                                 glob_boundary_type
  use mus_scheme_header_module, only: mus_scheme_header_type
  use mus_field_prop_module,    only: mus_field_prop_type, mus_field_prop_out, &
    &                                 mus_load_field_prop
  use mus_species_module,       only: compute_molWeightRatio,                  &
    &                                 compute_bulkViscOmega
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_physics_module,       only: mus_physics_type
  use mus_mixture_module,       only: mus_mixture_type, mus_load_mixture


  implicit none

  private

  public :: mus_field_type
  public :: mus_load_fields, mus_field_out

  !> This type contains all information on fields with ic and bc
  !! Example fields: fluid, species etc.
  !! Each field contains one initial condition and array of 
  !! boundary conditions
  !!
  type mus_field_type
    !> field label. Should be unique for each field
    character(len=labelLen) :: label
    !> lbm parameters (fluid and species) for field
    type( mus_field_prop_type ) :: fieldProp
    !> array of fieldboundary types for each field
    !! size: #BCs in the boundary_condition table
    type( boundary_type ), allocatable :: bc(:)
    !> initialization case, one initial condition for each field
    type( tem_ini_condition_type ) :: ic
    !> \todo: SZ: add comment
    logical :: setThirdOrder = .false.
    !> An instance of restart type
    type( tem_restart_type ) :: restart
  end type mus_field_type

  !> Interface for dumping a single field or a set of fields in a file in lua 
  !! format.
  !!
  interface mus_field_out
    module procedure mus_field_out_vec
    module procedure mus_field_out_scal
  end interface mus_field_out

contains

!******************************************************************************!
  !> Subroutine to load the field table from the lua configuration file.
  !!
  !! If field table is not defined than load bc, ic, fluid, species from scheme
  !! table. If scheme table is not defined than load field variables from
  !! config parent.
  !!
  subroutine mus_load_fields( me, nFields, mixture, boundary, conf, path,      &
    &                         parent, pdfGlobal, schemeHeader, source,         &
    &                         physics, tree, rank, scaling )
    !--------------------------------------------------------------------------
    !> array of field type
    type( mus_field_type ), allocatable, intent(out) :: me(:)
    integer, intent(out) :: nFields !< number of fields defined in lua file
    !> contains mixture information
    type( mus_mixture_type ), intent(out) :: mixture
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in):: boundary
    !> flu state
    type( flu_State ), intent(inout) :: conf
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !> global pdf info
    type( pdf_globInfo_type) :: pdfGlobal
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !> parent handle if scheme table is defined
    integer, intent(in), optional :: parent
    !> array of source information
    type( tem_source_type ), allocatable, intent(out) :: source(:)
    !> physics type to convert physics to lattice unit or vice versa
    type( mus_physics_type ), intent(in) :: physics
    !> global treelm mesh
    type( treelmesh_type), intent(inout) :: tree
    !> the current rank
    integer, intent(in) :: rank
    !> scaling type
    character(len=labelLen), intent(in) :: scaling
    !---------------------------------------------------------------------------
    ! counter variables
    integer :: iField, iError 
    ! aotus handles
    integer :: field_handle, field_sub_handle
    ! aotus paths for field and field sub table
    type( aot_path_type ) :: fields_path, field_sub_path 
    !---------------------------------------------------------------------------

    call tem_horizontalSpacer(fUnit = logUnit(1))
    write(logUnit(1),*) 'Loading the fields ...' 
    fields_path = path
    if(present(parent)) then
      call aot_table_open( L       =conf,                                      &
        &                  parent  = parent,                                   &
        &                  thandle = field_handle,                             &
        &                  key     = 'field' ) 
    else
      call aot_table_open( L       = conf,                                     &
        &                  thandle = field_handle,                             &
        &                  key     = 'field' )
    endif

    !get field path
    call aot_path_addNode( me       = fields_path,                             &
      &                    NodeType = 'table',                                 &
      &                    key      = 'field' )

    ! allocate the fields
    nFields = 1
    allocate( me( nFields ) )
    me(1)%label = ''

    ! allocate the sources (size: nFields + 1, for the entire scheme)
    allocate( source( nFields + 1 ))

    if ( field_handle == 0 ) then
      if (present(parent)) then
        ! field table is not defined load field variable from scheme 
        write(logUnit(1),*) 'No field table defined.'//                            &
          &             'Loading field variable from scheme' 

        call mus_load_field_single( me = me(1),                                &
          &                         nFields = nFields,                         &
          &                         boundary = boundary,                       &
          &                         conf = conf,                               &
          &                         path = path,                               &
          &                         pdfGlobal = pdfGlobal,                     &
          &                         parent = parent,                           &
          &                         source = source( 1 ),                      &
          &                         physics = physics,                         &
          &                         schemeHeader = schemeHeader,               &
          &                         tree = tree,                               &
          &                         scaling = scaling,                         &
          &                         rank = rank )

      ! field table is not defined load field variable from config
      else
        write(logUnit(1),*) 'No field table and scheme table are defined.'//       &
          &             'Loading field variable from config' 

        call mus_load_field_single( me = me(1),                                &
          &                         nFields = nFields,                         &
          &                         boundary = boundary,                       &
          &                         conf = conf,                               &
          &                         path = path,                               &
          &                         pdfGlobal = pdfGlobal,                     &
          &                         source = source( 1 ),                      &
          &                         physics = physics,                         &
          &                         schemeHeader = schemeHeader,               &
          &                         tree = tree,                               &
          &                         scaling = scaling,                         &
          &                         rank = rank )
      endif ! present parent?
    else ! check whether field is a single table or multiple table

      call aot_table_open( L       = conf,                                     &
        &                  parent  = field_handle,                             & 
        &                  thandle = field_sub_handle,                         &
        &                  pos     = 1 )
      if ( field_sub_handle == 0 ) then
      ! field is a single table
        call aot_table_close( L=conf, thandle=field_sub_handle )
        write(logUnit(1),*) 'Field is a single table' 

        call aot_get_val( L       = conf,                                      &
          &               thandle = field_handle,                              &
          &               key     = 'label',                                   &
          &               val     = me(1)%label,                               &
          &               default = '',                                        &
          &               ErrCode = iError )
        if( trim(me(1)%label) /= '')                                           &
          & me(1)%label = trim( me(1)%label )//'_'
        call mus_load_field_single( me = me(1),                                &
          &                         nFields = nFields,                         &
          &                         boundary = boundary,                       &
          &                         conf = conf,                               &
          &                         path = fields_path,                        &
          &                         parent = field_handle,                     &
          &                         pdfGlobal = pdfGlobal,                     &
          &                         source = source( 1 ),                      &
          &                         physics = physics,                         &
          &                         schemeHeader = schemeHeader,               &
          &                         tree = tree,                               &
          &                         scaling = scaling,                         &
          &                         rank = rank )

      else ! field is multiple table
        call aot_table_close( L=conf, thandle=field_sub_handle )
        nFields = aot_table_length( L=conf, thandle=field_handle )
        write(logUnit(1),*) 'Number of fields: ', nFields 
        ! reallocate the fields
        deallocate( me )
        allocate( me( nFields ))
        ! reallocate the sources
        deallocate( source )
        allocate( source( nFields + 1 ))
        do iField = 1, nFields
          call aot_table_open( L       = conf,                                 &
            &                  parent  = field_handle,                         & 
            &                  thandle = field_sub_handle,                     &
            &                  pos     = iField )

          ! update the lua path
          field_sub_path = fields_path
          call aot_path_addNode( me       = field_sub_path,                    &
            &                    NodeType = 'table',                           &
            &                    pos      = iField ) 

          write(logUnit(1),*) 'Properties for Field ', iField 
          call aot_get_val( L       = conf,                                    &
            &               thandle = field_sub_handle,                        &
            &               key     = 'label',                                 &
            &               val     = me(iField)%label,                        &
            &               default = '',                                      &
            &               ErrCode = iError )

          ! add underscore here to differentiate state variables and derirved 
          ! variables of each field
          if( trim(me(iField)%label) /= '' )                                   &
            &          me(iField)%label = trim( me(iField)%label )//'_'

          if (btest(iError, aoterr_NonExistent)) then
            write(logUnit(1),*) 'field label is not specified' 
            write(logUnit(1),*) 'field label is neccessary when nFields>1' 
            call tem_abort()
          end if  

          call tem_horizontalSpacer(fUnit = logUnit(1))
          call mus_load_field_single( me = me(iField),                         &
            &                         nFields = nFields,                       &
            &                         boundary = boundary,                     &
            &                         conf = conf,                             &
            &                         path = field_sub_path,                   &
            &                         pdfGlobal = pdfGlobal,                   &
            &                         parent = field_sub_handle,               &
            &                         source = source( iField ),               &
            &                         physics = physics,                       &
            &                         schemeHeader = schemeHeader,             &
            &                         tree = tree,                             &
            &                         scaling = scaling,                       &
            &                         rank = rank )

          ! Check, if the names of the fields are unique. If not, add an
          ! increasing number
          call tem_checkLabel( label = me(:)%label, nLabels = iField )

          call aot_table_close( L=conf, thandle=field_sub_handle )
        end do ! number of fields
      end if ! if field table is present
    end if

    call aot_table_close( L=conf, thandle=field_handle )

    !load scheme specific field
    select case(trim(schemeHeader%kind))
    case('multi-species_gas','multi-species_liquid')
      !nFields must be > 1 for multispecies
      if( nFields == 1 ) then
        write(logUnit(1),*) ' ERROR: Number of fields(species) defined for '
        write(logUnit(1),*) '        multi-species scheme is 1. '
        write(logUnit(1),*) '        Specify nFields > 1. Aborting'
        call tem_abort()
      endif

      !compute molecular weight ratios incase of multispecies simulation
      call compute_molWeightRatio(                                             &
        &                  molWeights = me(:)%fieldProp%species%molWeight,      &
        &                  molWeigRatios = me(:)%fieldProp%species%molWeigRatio )

      !load mixture table and set bulk viscosity relaxation time
      ! for each species which is depends on molecular weight ratios
      call mus_load_mixture( me = mixture, conf = conf, path = path,           &
        &                    parent = parent, schemeHeader = schemeHeader,     &
        &                    pdfGlobal = pdfGlobal, physics = physics,         &
        &                    field_ic = me(:)%ic, nFields = nFields,           &
        &                    molWeights = me(:)%fieldProp%species%molWeight )

      !compute bulk viscosity omega for each field
      write(logUnit(1),*)'  Bulk omega for each species:'
      do iField = 1,nFields
        write(logUnit(1),*) '   species ', iField
        call compute_bulkViscOmega( species = me(iField)%fieldProp%species,    &
          &                         bulkvisc = mixture%bulk_viscosityLB,       &
          &                         bulkviscLvl = mixture%relaxLvl(:)%bulkVisc,&
          &                         pdfGlobal = pdfGlobal )
      end do
    end select

  end subroutine mus_load_fields
!******************************************************************************!


!******************************************************************************!
  !> load a single field table
  !!
  subroutine mus_load_field_single( me, nFields, boundary, conf, path, parent, &
    &                               pdfGlobal, source, physics, schemeHeader,  &
    &                               tree, rank, scaling )
    !--------------------------------------------------------------------------
    !> field type
    type( mus_field_type ), intent(inout) :: me
    integer, intent(in) :: nFields !< number of fields defined in lua file
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in):: boundary
    !> flu state
    type( flu_State ), intent(inout) :: conf
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !> parent handle if scheme table is defined
    integer, intent(in), optional :: parent
    !> global pdf info
    type( pdf_globInfo_type) :: pdfGlobal
    !> source information for the current field
    type( tem_source_type ), intent(out) :: source
    !> physics type to convert physics to lattice unit or vice versa
    type( mus_physics_type ), intent(in) :: physics
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !> global treelm mesh
    type( treelmesh_type), intent(inout) :: tree
    !> the current rank
    integer, intent(in) :: rank
    !> scaling type
    character(len=labelLen), intent(in) :: scaling
    !--------------------------------------------------------------------------
    character(len=labelLen), allocatable :: ic_states(:)
    !--------------------------------------------------------------------------

    write(logUnit(1),*)'Field label '//trim(me%label)
    ! load field properties
    call mus_load_field_prop( me           = me%fieldProp,                     &
      &                       nFields      = nFields,                          &
      &                       conf         = conf,                             &
      &                       pdfGlobal    = pdfGlobal,                        &
      &                       parent       = parent,                           &
      &                       path         = path,                             &
      &                       physics      = physics,                          &
      &                       scaling      = scaling,                          &
      &                       schemeHeader = schemeHeader )

    ! load the global source variables
    call tem_load_source( me     = source,                                     &
      &                   conf   = conf,                                       &
      &                   parent = parent,                                     &
      &                   key    = 'source',                                   &
      &                   path   = path )

    ! load boundary
    call mus_load_bc( me           = me%bc,                                    &
      &               boundary     = boundary,                                 &
      &               conf         = conf,                                     &
      &               path         = path,                                     &
      &               parent       = parent )

    ! load scheme specific field
    select case(trim(schemeHeader%kind))
    case('lbm', 'lbm_incomp','lbm_ps')
      allocate(ic_states(10))
      ic_states = (/ 'pressure ', 'velocityX', 'velocityY', 'velocityZ',&
        &            'Sxx      ', 'Syy      ', 'Szz      ', 'Sxy      ',&
        &            'Syz      ', 'Sxz      ' /)
    case('lbm_nNwtn', 'lbm_incomp_nNwtn')
      allocate(ic_states(11))
      ic_states = (/ 'pressure ', 'velocityX', 'velocityY', 'velocityZ',&
        &            'Sxx      ', 'Syy      ', 'Szz      ', 'Sxy      ',&
        &            'Syz      ', 'Sxz      ', 'omega    '/)
    case('multi-species_gas')
      allocate( ic_states(4) )
      ic_states = (/ 'pressure ', 'velocityX', 'velocityY',&
        &            'velocityZ' /)
    case('multi-species_liquid')
      allocate( ic_states(4) )
      ic_states = (/ 'molefraction', 'velocityX   ', 'velocityY   ', &
        &            'velocityZ   ' /)
    case default
          write(logUnit(1),*)'The selected scheme kind model is not &
             & supported to initialize ic state names: '&
             &//trim( schemeHeader%kind )
          call tem_abort()
    end select

    ! Read initial condition
    call tem_load_ic( me=me%ic, conf=conf,                      &
      &               parent=parent, key = 'initial_condition', &
      &               path = path,                              &
      &               StateName = ic_states )

    call mus_load_ic_prop( me     = me,                  &
      &                    key    = 'initial_condition', &
      &                    conf   = conf,                &
      &                    parent = parent,              &
      &                    path   = path )

    call tem_horizontalSpacer(fUnit = logUnit(1))

  end subroutine mus_load_field_single
!******************************************************************************!


!******************************************************************************!
  !> load fluid properties like fluid and species table from
  !! lua file based on the scheme kind
  !! KM: \todo what is the purpose of this routine? what is thirdOrder in 
  !! initial condition
  !!
  subroutine mus_load_ic_prop( me, conf, parent, path, key )
    !---------------------------------------------------------------------------
    !> field property type
    type( mus_field_type ), intent(inout) :: me
    !> flu state
    type( flu_State ), intent(inout) :: conf 
    !> parent lua handle
    integer, intent(in), optional :: parent 
    !> aotus lua path
    type( aot_path_type ),intent(in) :: path 
    character(len=*), optional :: key
    !---------------------------------------------------------------------------
    character(len=labelLen) :: locKey
    integer :: ic_table, iError
    !---------------------------------------------------------------------------
    write(logUnit(1),*)'Loading IC properties ..'
    if( present( key )) then
      locKey = key
    else
      locKey = 'initial_condition'
    end if
    if ( present(parent) ) then
      call aot_table_open(L=conf, parent=parent, thandle=ic_table, key=locKey)
    else
      call aot_table_open(L=conf, thandle=ic_table, key=locKey)
    end if
    if (ic_table == 0) then
      ! Interpret the initial_condition entry as name for the restart file,
      ! if it is not a table.
      write(logUnit(1),*) 'Initial condition NOT a table!'
    else
      ! First check if a restart file is present:
      call aot_get_val( L       = conf,                                        &
        &               thandle = ic_table,                                    &
        &               key     = 'setThirdOrder',                             &
        &               val     = me%setThirdOrder,                            &
        &               default = .false.,                                     &
        &               ErrCode = iError )
    end if
    call aot_table_close(L=conf, thandle=ic_table)
    if( me%setThirdOrder ) then
      write(logUnit(1),*)'  Setting third order in initial conditions'
    end if

  end subroutine mus_load_ic_prop
!******************************************************************************!


!******************************************************************************!
  !> write array of fields into a lua file
  !!
  subroutine mus_field_out_vec( me, conf, schemeHeader )
    !---------------------------------------------------------------------------
    !> array of field type
    type( mus_field_type ), intent(in) :: me(:)
    !> aotus out type
    type( aot_out_type ), intent(inout) :: conf
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !---------------------------------------------------------------------------
    integer :: iField ! counter variable
    !---------------------------------------------------------------------------

    call aot_out_val( put_conf = conf,                                         &
      &               val      = size(me),                                     &
      &               vname    = 'nFields' )

    ! write out the field prefixes for more than one field
    if(size(me) > 1) then
      call aot_out_val( put_conf     = conf,                                   &
        &               val          = me(:)%label,                            &
        &               vname        = 'fieldPrefixes',                        &
        &               max_per_line = 3 )
    end if
    do iField = 1, size(me)
      call mus_field_out_scal( me           = me( iField ),                    &
        &                      conf         = conf,                            &
        &                      schemeHeader = schemeHeader )
    end do

  end subroutine mus_field_out_vec
!******************************************************************************!


!******************************************************************************!
  !> write single field into a lua file
  !!
  subroutine mus_field_out_scal( me, conf, schemeHeader )
    !---------------------------------------------------------------------------
    !> single field type
    type( mus_field_type ), intent(in) :: me
    !> aotus out type
    type( aot_out_type ), intent(inout) :: conf
    !> identifier of the scheme
    type( mus_scheme_header_type ), intent(in) :: schemeHeader
    !---------------------------------------------------------------------------

    call aot_out_open_table(put_conf = conf, tname = trim(me%label)//'field')
    call mus_field_prop_out( me           = me%fieldProp,                      &
      &                      conf         = conf,                              &
      &                      schemeHeader = schemeHeader )
    call aot_out_close_table(put_conf = conf)

  end subroutine mus_field_out_scal
!******************************************************************************!


end module mus_field_module
!******************************************************************************!
