?? include 'header/lbm_macros.inc'
?? include 'treelm/source/logMacros.inc'
!******************************************************************************!
!> Flow-related routines such as initialization of the flow field and parameters
!!
module mus_flow_module

  ! include treelm modules
  use env_module,              only: rk, io_buffer_size, long_k, eps, single_k,&
    &                                eps_single_k
  use tem_param_module,        only: cs2inv, cs2, t2cs2inv, t2cs4inv
  use tem_aux_module,          only: tem_abort
  use tem_geometry_module,     only: tem_baryOfId
  use tem_topology_module,     only: tem_idOfCoord
  use tem_spatial_module,      only: tem_spatial_for
  use tem_restart_module,      only: tem_restart_readHeader,                   &
    &                                tem_init_restart_alloc,                   &
    &                                tem_init_restart_create_types,            &
    &                                tem_restart_getTotalChunks,               &
    &                                tem_restart_openRead_single,              &
    &                                tem_restart_readData_single,              &
    &                                tem_restart_closeRead,                    &
    &                                tem_restart_finalize_single
  use tem_depend_module,       only: tem_init_depend
  use tem_subTree_type_module, only: tem_subTree_type, tem_local_subTree_from, &
    &                                tem_destroy_subTree
  use tem_logging_module,      only: logUnit
  use tem_debug_module,        only: dbgUnit
  use tem_tools_module,        only: tem_horizontalSpacer
  use tem_math_module,         only: invert_matrix
  use tem_ini_condition_module, only: tem_ini_condition_type
  use tem_var_system_module, only: tem_var_system_type

  ! include musubi modules
  use mus_param_module,              only: mus_param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_derivedQuantities_module2, only: getNEq_diffusive,                   &
    &                                      getNEq_acoustic, convPrePost,       &
    &                                      setThirdOrder_diffusive
  use mus_buffer_module,             only: mus_rest_unserialize
  use mus_field_module,              only: mus_field_type
  use mus_interpolate_header_module, only: interpolation_type
  use mus_mixture_module,            only: mus_mixture_type
  use mus_initLBM_module,            only: mus_init_advRel_LBM
  use mus_initLBMIncomp_module,      only: mus_init_advRel_LBMIncomp
  use mus_initMultispecies_module,   only: mus_init_advRel_multispecies_gas,   &
    &                                      mus_init_advRel_multispecies_liquid
  use mus_initLBMnNwtn_module,       only: mus_init_advRel_LBM_nNwtn, &
    &                                      mus_init_advRel_LBM_incomp_nNwtn
  use mus_eNRTL_module,              only: mus_calc_thermFactor,               &
    &                                      mus_calc_MS_DiffMatrix

  implicit none

  private

  public :: mus_init_flow
  public :: init_advRel

contains

!******************************************************************************!
  !> Initialize  the relaxation parameters and choose the relaxation model
  !!
  subroutine init_advRel( scheme )
    !---------------------------------------------------------------------------
    !> scheme type
    type(mus_scheme_type) :: scheme
    !---------------------------------------------------------------------------
    integer :: iField
    !---------------------------------------------------------------------------
    if( index( scheme%header%relaxation, 'fixedVel') > 0 ) then
      write(logUnit(1),*)' FixedVelocity: setting spatial velocity '//         &
        &                'distribution to be stored'
      do iField = 1, scheme%nFields
        scheme%field(iField)%fieldProp%fluid%velSpatial(:)%toStore = .true.
      end do
    end if

    write(logUnit(1),*) 'Initialize compute kernel for scheme kind: '&
      &                   //trim( scheme%header%kind )

    select case( trim( scheme%header%kind ) )
    case( 'lbm' )
      call mus_init_advRel_lbm( scheme )
    case( 'lbm_incomp' )
      call mus_init_advRel_lbmIncomp( scheme )
    case( 'multi-species_gas' )
      call mus_init_advRel_multispecies_gas( scheme )
    case ( 'multi-species_liquid' )
      call mus_init_advRel_multispecies_liquid( scheme )
    case( 'lbm_nNwtn' )
      ! lattice boltzmann nonNewtonian fluid model
      call mus_init_advRel_lbm_nNwtn( scheme )
    case( 'lbm_incomp_nNwtn' )
      ! lattice boltzmann nonNewtonian fluid model
      call mus_init_advRel_lbm_incomp_nNwtn( scheme )
    case default
      write(logUnit(1),*) 'The selected scheme kind model is not supported: '//&
        &                 trim( scheme%header%kind )
      call tem_abort()
    end select

  end subroutine init_advRel
!******************************************************************************!

!******************************************************************************!
  !> Initialize flow field by calling corresponding routine according to scheme
  !! kind.
  !!
  subroutine mus_init_flow( scheme, geometry, params, pdfGlobal, intp )
    !---------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type), intent(inout) :: scheme(:)
    !> Global parameters
    type(mus_param_type), intent(inout) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> global info type
    type(pdf_globInfo_type) :: pdfGlobal
    !> information on interpolation
    type(interpolation_type), intent(in) :: intp
    !---------------------------------------------------------------------------
    integer :: iScheme, nElems, iField, iLevel
    integer :: tLayerMin, tLayerMax
    !---------------------------------------------------------------------------
    call tem_horizontalSpacer(fUnit = logUnit(1))

    write(logUnit(1),*) "Initialize fluid initial conditions..."
    !Normally one would expect a call to open config here. However, with the
    !implementation of dynamic load balancing we have moved that call to
    !mus_load_config. It was always there but the conf was closed in the end and
    !reopened here. This does not make much sense so this strategy has been
    !followed. 

    do iScheme = 1, pdfGlobal%nSchemes

      write(logUnit(1),*) "Initialize flow on scheme:", scheme(iScheme)%schemeID

      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxlevel

        write(logUnit(1),*) ' Initializing level ', iLevel

        ! For debugging purposes, set complete flow field to invalid
        scheme( iScheme )%pdf(iLevel)%state(:,:) =  -1000000.0_rk

        ! Set number of elements to fill (only fluid or also helper elements?)
        if( .not. params%init_allElems ) then
          nElems =  scheme(iScheme)%pdf( iLevel )%nElems_fluid
        else
          nElems =  scheme(iScheme)%pdf( iLevel )%nElems_solve
        end if 
?? IF (loglvl >= 3) THEN
        if( nElems /= scheme(iScheme)%pdf( iLevel )%nElems_fluid ) then
          write(logUnit(3),*)
          write(logUnit(3),*)'WARNING!!!!'
          write(logUnit(3),*) 'For debugging purposes, in addition to fluid '//&
            &             'elements'
          write(logUnit(3),*) 'more than fluid elements are filled with Initial '//&
            &             'values'
          write(logUnit(3),*)
        end if
?? ENDIF

        ! how many time laywers to initialize
        if( params%init_allTimeLayers ) then
          tLayerMin = 1
          tLayerMax = size( scheme(iScheme)%pdf( iLevel )%state, 2 )
        else
          tLayerMin = scheme(iScheme)%pdf( iLevel )%nNext
          tLayerMax = scheme(iScheme)%pdf( iLevel )%nNext
?? IF (loglvl >= 3) THEN
          write(logUnit(3),*)
          write(logUnit(3),*)'WARNING!!!!'
          write(logUnit(3),*)'For debugging purposes, only one time layer' // &
          &  'being initialized.'
          write(logUnit(3),*)
?? ENDIF
        end if

        select case( trim(scheme( iScheme )%header%kind) )
        case( 'lbm', 'lbm_incomp', 'lbm_nNwtn', 'lbm_incomp_nNwtn' )
          do iField = 1, scheme(iScheme)%nFields
            ! If we are restarting the simulation, the flow parameters are
            ! initialized from restart file and we don't have to calculate
            ! find chunksize and number of chunks required for initialzation
            if ( .not. scheme(iScheme)%field(iField)%ic%isRestart ) then
              ! Invoke the routine which initializes flow with calculated
              ! quantities

              call mus_init_lbm_withFlowParams( me = scheme(iScheme),          &
                &                geometry  = geometry,                         &
                &                params    = params,                           &
                &                nElems    = nElems,                           &
                &                iField    = iField,                           &
                &                iLevel    = iLevel,                           &
                &                field     = scheme(iScheme)%field(iField),    &
                &                tLayerMin = tLayerMin,                        &
                &                tLayerMax = tLayerMax )

              ! initilize omega for nonNewtonian fluid model
              select case ( trim(scheme(iScheme)%header%kind ))
              case ( 'lbm_nNwtn', 'lbm_incomp_nNwtn' )

                ! initialize omega when nonNewtonian feature is on
                call mus_init_omega_byOmega(                      &
                  &    state = scheme(iScheme)%pdf(iLevel)%state, &
                  &    params = params, &
                  &    geometry  = geometry, &
                  &    nElems = nElems, &
                  &    varSys = scheme(iScheme)%varSys,    &
                  &    total = scheme(iScheme)%glob%levelDesc(iLevel)%total, &
                  &    ic = scheme(iScheme)%field(iField)%ic, &
                  &    tLayerMin = tLayerMin, &
                  &    tLayerMax = tLayerMax )

              end select

            else
              ! Invoke the routine which initializes flow from restart file
              call mus_init_withRestart( scheme   = scheme,                    &
                &                        geometry = geometry,                  &
                &                        params   = params,                    &
                &                        pdfGlobal= pdfGlobal,                 &
                &                        field = scheme(iScheme)%field(iField) )
            end if ! not restart
          end do ! ifield
        case( 'multi-species_gas', 'multi-species_liquid' )
          ! read restart from ic for multispecies
          ! only If all field are initialized with restart
          ! then we are restarting the simulation, the flow parameters are
          ! initialized from restart file and we don't have to calculate
          if ( all(scheme(iScheme)%field(:)%ic%isRestart) ) then
            do iField = 1, scheme(iScheme)%nFields
              ! Invoke the routine which initializes flow from restart file
              call mus_init_withRestart( scheme   = scheme,                    &
                &                        geometry = geometry,                  &
                &                        params   = params,                    &
                &                        pdfGlobal= pdfGlobal,                 &
                &                        field = scheme(iScheme)%field(iField) )
            end do
          else
            ! Invoke the routine which initializes flow with calculated
            ! quantities
            call mus_init_multispecies_withFlowParams( me = scheme(iScheme),   &
              &                            geometry  = geometry,               &
              &                            params    = params,                 &
              &                            nElems    = nElems,                 &
              &                            mixture   = scheme(iScheme)%mixture,&
              &                            iLevel    = iLevel,                 &
              &                            field     = scheme(iScheme)%field,  &
              &                            tLayerMin = tLayerMin,              &
              &                            tLayerMax = tLayerMax )
          endif
        case default
          write(logUnit(1),*)'The selected scheme kind model is not '//        &
             & 'supported to initialize flow with equilibrium function: '//    &
             & trim( scheme( iScheme )%header%kind )
          call tem_abort()
        end select
      end do !iLevel

      select case( trim( scheme(iScheme)%header%kind ) )
        ! Check the shear stress only for fluid simulation
        case( 'lbm', 'lbm_incomp', 'lbm_nNwtn', 'lbm_incomp_nNwtn' )
          if( intp%requireMoments ) then
            call mus_checkShearStress( scheme    = scheme( iScheme ),          &
              &                        geometry  = geometry,                   &
              &                        params    = params,                     &
              &                        pdfGlobal = pdfGlobal )
          else
            write(logUnit(3),*)'WARNING: Check for shear stress is not available'
            write(logUnit(3),*)'         for interpolation schemes not using '
            write(logUnit(3),*)'         moments'
          end if
        case('lbm_passive_scalar', 'multi-species_gas',               &
          &  'multi-species_liquid')
          write(logUnit(3),*)'WARNING: Check for shear stress is not available'
          write(logUnit(3),*)'         for scheme kind'//                      &
            &                trim(scheme(iScheme)%header%kind)
      end select

      ! Choose the advection relaxation scheme
      call init_advRel( scheme = scheme(iScheme) )

    end do ! schemes

  end subroutine mus_init_flow
!******************************************************************************!


!******************************************************************************!
  !> Initialize the flow from pressure, velocity and strain rate.\n
  !! First equilibirium pdf (fEq) is calculated from pressure and velocity.
  !! Then non-equilibirium (fnEq) is calculated from strain rate.
  !! At last set the pdf of each element by sum up these two parts (fEq+fnEq).
  !!
  subroutine mus_init_lbm_withFlowParams( me, geometry, params, Field, iField, &
    &                                     nElems, iLevel, tLayerMin, tLayerMax )
    !--------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type), intent(inout) :: me
    !> Global parameters
    type(mus_param_type), intent(in) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> Field type
    type(mus_field_type), intent(inout) :: field
    !> Field index
    integer, intent(in)                 :: iField
    !> Number of Elements
    integer, intent(in)                 :: nElems
    !> Level index
    integer, intent(in)                 :: iLevel
    !> min time layer of state to initialize
    integer, intent(in) :: tLayerMin
    !> max time layer of state to initialize
    integer, intent(in) :: tLayerMax
    !---------------------------------------------------------------------------
    integer :: iDir, idx_iDir, iElem
    real(kind=rk), allocatable :: fEq(:), fnEq(:), press(:)
    real(kind=rk), allocatable :: xc(:,:), ux(:), uy(:), uz(:), rho(:)
    ! shear rate tensor
    real(kind=rk), allocatable :: Sxx(:,:) ! Sxx, Syy, Szz, Sxy, Syz, Sxz
    real(kind=rk), allocatable :: input_loc(:)
    integer :: iChunk, nChunks, chunkSize, nChunkElems, elemPos
    integer :: elemoff
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    ! find chunksize and number of chunks required for initialzation
    chunkSize = io_buffer_size/me%layout%stencil(1)%QQ
    nChunks = ceiling( real(nElems, kind=rk)/real(chunkSize, kind=rk))

    do iChunk = 1, nChunks
      ! Number of elements read so far in previous chunks.
      elemOff = ( (iChunk-1)*chunksize )
      nChunkElems = min(chunkSize, nElems - elemOff)

      !allocate memory for density, vel, eq, coord
      allocate(xc(nChunkElems, 3))
      allocate(rho(nChunkElems))
      allocate(press(nChunkElems))
      allocate(ux(nChunkElems))
      allocate(uy(nChunkElems))
      allocate(uz(nChunkElems))
      allocate(Sxx(6,nChunkElems))  ! Sxx, Syy, Szz, Sxy, Syz, Sxz
      allocate(input_loc(nChunkElems*4))
      ! Equilibrium part of the pdf
      allocate(fEq(me%layout%stencil(1)%QQ*nChunkElems))
      ! non-Equilibrium part of the pdf
      allocate(fnEq(me%layout%stencil(1)%QQ*nChunkElems))
      fnEq = 0._rk

      do iElem = 1, nChunkElems
        ! Calculate the coordinates
        elemPos = elemOff + iElem
        xc(iElem,1:3) = tem_BaryOfId( geometry%tree,     &
          &                           me%glob%levelDesc(iLevel)%total(elemPos) )
      end do

      press = tem_spatial_for( conf  = params%general%solver%conf(1),          &
        &                      me    = field%ic%ini_state(1),                  &
        &                      coord = xc,                                     &
        &                      n     = nChunkElems )
      ux = tem_spatial_for( conf  = params%general%solver%conf(1),             &
        &                   me    = field%ic%ini_state(2),                     &
        &                   coord = xc,                                        &
        &                   n     = nChunkElems )
      uy = tem_spatial_for( conf  = params%general%solver%conf(1),             &
        &                   me    = field%ic%ini_state(3),                     &
        &                   coord = xc,                                        &
        &                   n     = nChunkElems )
      uz = tem_spatial_for( conf  = params%general%solver%conf(1),             &
        &                   me    = field%ic%ini_state(4),                     &
        &                   coord = xc,                                        &
        &                   n     = nChunkElems )
      ! Read in the shear rate tensor
      ! This corresponds to S = grad(u) + grad(u)^T
      Sxx = 0._rk
      Sxx(1,:) = tem_spatial_for( conf  = params%general%solver%conf(1),       &
        &                         me    = field%ic%ini_state(5),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(2,:) = tem_spatial_for( conf  = params%general%solver%conf(1),       &
        &                         me    = field%ic%ini_state(6),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(3,:) = tem_spatial_for( conf  = params%general%solver%conf(1),       &
        &                         me    = field%ic%ini_state(7),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(4,:) = tem_spatial_for( conf  = params%general%solver%conf(1),       &
        &                         me    = field%ic%ini_state(8),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(5,:) = tem_spatial_for( conf  = params%general%solver%conf(1),       &
        &                         me    = field%ic%ini_state(9),               &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )
      Sxx(6,:) = tem_spatial_for( conf  = params%general%solver%conf(1),       &
        &                         me    = field%ic%ini_state(10),              &
        &                         coord = xc,                                  &
        &                         n     = nChunkElems )

      ! convert these quantities from physics to LB
      ! 1. velocity
      ux = ux/params%physics%fac( iLevel )%vel
      uy = uy/params%physics%fac( iLevel )%vel
      uz = uz/params%physics%fac( iLevel )%vel
      ! 2. density / pressure
      ! for lbm:        rho0 = 0
      ! for lbm_incomp: rho0 = 1
      rho = press * cs2inv / params%physics%fac( iLevel )%press &
        &   + field%fieldProp%fluid%rho0

      ! 3. strain rate
      Sxx = Sxx / params%physics%fac( iLevel )%strainRate

      ! linearized the input for equilibrium function pointer
      ! with following order
      ! | dens_1 | dens_2 | .. | dens_nElems | ux_1 | uy_1 | uz_1 | ..
      ! ! .. | ux_nelems | uy_nelems | uz_nelems |
      input_loc = 0._rk

      ! transfer density
      input_loc(1:nChunkElems) = rho

      ! transfer velocity
      do iElem = 1, nChunkElems
        ! MH: Fix for the intel compiler:
        ! entries have to be assigned one by one instead of
        ! assigning them as an array of (/ux, uy, uz/)
        input_loc( nChunkElems + (iElem-1)*3 + 1) = ux( iElem )
        input_loc( nChunkElems + (iElem-1)*3 + 2) = uy( iElem )
        input_loc( nChunkElems + (iElem-1)*3 + 3) = uz( iElem )
      end do

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = nChunkElems

      ! Calculate the equilibrium pdf (fEq)
      call me%varSys%derive( me%derVarPos(iField)%equilFromMacro )%do(         &
        &              input   = input_loc,                                    &
        &              bound   = bound,                                        &
        &              tree    = geometry%tree,                                &
        &              subTree = subTree,                                      &
        &              requVar = me%varSys%Variable%val(                       &
        &                                me%derVarPos(iField)%equilFromMacro), &
        &              globSys = me%varSys,                                    &
        &              stencil = me%layout%stencil(1),                         &
        &              conf    = me%varSys%solSpec%conf,                       &
        &              res     = fEq )


      select case( trim( me%header%kind ) )
      ! KM: compute fNeq only for lbm and lbm_incomp, for multispecies
      ! and passive scalar it is initialized with zero
      case( 'lbm', 'lbm_incomp', 'lbm_nNwtn', 'lbm_incomp_nNwtn' )
        write(logUnit(3),*) ' Initialize pdf with fNeq by scaling:'&
          &                 //trim(params%scaling)
        select case( trim( params%scaling ))
        case('acoustic')
          ! Acoustic fNeq with defining the stress tensor instead of the
          ! shear rate tensor in the lua file
          do iElem = 1, nChunkElems
            fNeq(   (iElem-1)*me%layout%stencil(1)%QQ+1             &
              &   : iElem*me%layout%stencil(1)%QQ ) =               &
              &   getnEq_acoustic(                                  &
              &     layout = me%layout,                             &
              &     rho0   = field%fieldProp%fluid%rho0,            &
              &     omega  = field%fieldProp%fluid%omLvl(iLevel),   &
              &     Sxx    = [ Sxx(1,iElem), Sxx(4,iElem),          &
              &                Sxx(6,iElem), Sxx(4,iElem),          &
              &                Sxx(2,iElem), Sxx(5,iElem),          &
              &                Sxx(6,iElem ), Sxx(5,iElem),         &
              &                Sxx(3,iElem) ] )
          end do ! iElem
        case('diffusive')
          ! Diffusive non-equilibrium part with the shear rate tensor
          ! given as an input
          do iElem = 1, nChunkElems
            fNeq(   (iElem-1)*me%layout%stencil(1)%QQ + 1       &
            &     : (iElem  )*me%layout%stencil(1)%QQ     ) =   &
            &  getnEq_diffusive(                                &
            &     layout = me%layout,                           &
            &     omega  = field%fieldProp%fluid%omLvl(iLevel), &
            &     Sxx    = [ Sxx(1,iElem), Sxx(4,iElem),        &
            &                Sxx(6,iElem), Sxx(4,iElem),        &
            &                Sxx(2,iElem), Sxx(5,iElem),        &
            &                Sxx(6,iElem), Sxx(5,iElem),        &
            &                Sxx(3,iElem) ] )
            if( field%setThirdOrder ) then
              fNeq(   (iElem-1)*me%layout%stencil(1)%QQ+1               &
                &   : (iElem  )*me%layout%stencil(1)%QQ   )             &
                & = fNeq(  (iElem-1)*me%layout%stencil(1)%QQ+1          &
                &         :(iElem  )*me%layout%stencil(1)%QQ   )        &
                &   + setThirdOrder_diffusive(                          &
                &         layout = me%layout,                           &
                &         omega  = field%fieldProp%fluid%omLvl(iLevel), &
                &         x      = xc(iElem,:),                         &
          ! uLB Velocity in lattice units = u0Phys *dt/dx , u0Phys = 1 for TGV
                &         dx     = params%physics%dxLvl( iLevel ),      &
                &         dt     = params%physics%dtLvl( iLevel ) )
            end if
          end do ! iElem
        end select ! scaling
      case default
        write(logUnit(6),*) ' Initialize pdf without fNeq calculation.'
      end select ! me%header%kind

      ! assign pdf by fEq + fNeq
      do iElem = 1, nChunkElems
        elemPos = elemOff + iElem
        do iDir = 1, me%layout%stencil(1)%QQ
          idx_iDir = me%varSys%Variable%val(iField)%varPos( iDir )

          me%pdf(iLevel)%state( ?IDX?( idx_iDir, elemPos, me%varSys%nScalars), &
          ! KM: initialize both state vectors for boundaries
          !\todo need to check for multilevel
            &   tLayerMin:tLayerMax )                               &
            & =   fEq( (iElem-1)*me%layout%stencil(1)%QQ + iDir)    &
            &   + fNeq((iElem-1)*me%layout%stencil(1)%QQ + iDir)
        end do ! iDir
      end do ! iElem

      ! deallocate memory for next chunk
      deallocate(xc)
      deallocate(rho)
      deallocate(press)
      deallocate(ux)
      deallocate(uy)
      deallocate(uz)
      deallocate(Sxx)
      deallocate(input_loc)
      deallocate(fEq)
      deallocate(fnEq)

    end do ! chunk

  end subroutine mus_init_lbm_withFlowParams
!******************************************************************************!

!******************************************************************************!
  !> Initialize the flow from calculated quantitites like density, velocity etc.
  !! for multispecies lbm
  subroutine mus_init_multispecies_withFlowParams( me, geometry, params,       &
    &                                Field, mixture, nElems, iLevel,           &
    &                                tLayerMin, tLayerMax )
    !--------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: me !< Scheme type
    type(mus_param_type), intent(in) :: params  !< Global parameters
    type(mus_geom_type), intent(in) :: geometry !< Geometry information
    type(mus_field_type), intent(inout) :: field(:) !< Field type
    type(mus_mixture_type), intent(inout) :: mixture !< mixture type
    integer, intent(in)                 :: nElems !< Number of Elements
    integer, intent(in)                 :: iLevel !< Level index
    !> min time layer of state to initialize 
    integer, intent(in) :: tLayerMin
    !> max time layer of state to initialize 
    integer, intent(in) :: tLayerMax
    !--------------------------------------------------------------------------
    select case( trim(me%header%kind) )
    case( 'multi-species_liquid' )
      call mus_init_MSLiquid_withFlowParams( me        = me,                   &
        &                                    geometry  = geometry,             &
        &                                    params    = params,               &
        &                                    nElems    = nElems,               &
        &                                    mixture   = mixture,              &
        &                                    iLevel    = iLevel,               &
        &                                    field     = field,                &
        &                                    tLayerMin = tLayerMin,            &
        &                                    tLayerMax = tLayerMax )
    case( 'multi-species_gas' )
      call mus_init_MSGas_withFlowParams( me    = me,                          &
        &                                 geometry  = geometry,                &
        &                                 params    = params,                  &
        &                                 nElems    = nElems,                  &
        !&                                 mixture   = mixture,                 &
        &                                 iLevel    = iLevel,                  &
        &                                 field     = field,                   &
        &                                 tLayerMin = tLayerMin,               &
        &                                 tLayerMax = tLayerMax )
    end select
  end subroutine mus_init_multispecies_withFlowParams
!******************************************************************************!

!******************************************************************************!
  !> Initialize the flow from calculated quantitites like density, velocity etc.
  !! for multispecies lbm
  subroutine mus_init_MSLiquid_withFlowParams( me, geometry, params,           &
    &                           Field, mixture, nElems, iLevel,                &
    &                           tLayerMin, tLayerMax )
    !--------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: me !< Scheme type
    type(mus_param_type), intent(in) :: params  !< Global parameters
    type(mus_geom_type), intent(in) :: geometry !< Geometry information
    type(mus_field_type), intent(inout) :: field(:) !< Field type
    type(mus_mixture_type), intent(inout) :: mixture !< mixture type
    integer, intent(in)                 :: nElems !< Number of Elements
    integer, intent(in)                 :: iLevel !< Level index
    !> min time layer of state to initialize 
    integer, intent(in) :: tLayerMin
    !> max time layer of state to initialize 
    integer, intent(in) :: tLayerMax
    !--------------------------------------------------------------------------
    integer :: iDir, idx_iDir, iElem, iField, nFields
    real(kind=rk), allocatable :: fEqStar(:), fEq(:), press(:)
    real(kind=rk), allocatable :: xc(:,:), ux(:,:), uy(:,:), uz(:,:)
    real(kind=rk), allocatable :: rho(:,:), moleFrac(:,:)
!    real(kind=rk), allocatable :: input_loc(:)
    real(kind=rk) :: phi(me%nFields), tot_massDens
    !real(kind=rk) :: mixMolWeight
    integer :: iChunk, nChunks, chunkSize, nChunkElems, elemPos, elemOff
!    integer :: offVel, posVel
    ! local DUMMY subTree needed to use the function pointers (not filled)
!    type(tem_subTree_type) :: subTree
!    integer :: bound(2)
    real(kind=rk), dimension(3) :: eqVel, velAvg, velQuadStar, velQuad
    real(kind=rk) :: vel(3, me%nFields)
    integer :: iField_2, iField_3
    real(kind=rk) :: resi_coeff(me%nFields, me%nFields)
    real(kind=rk) :: diff_coeff(me%nFields, me%nFields)
    real(kind=rk) :: ucx, ucxStar, usq, usqStar, ucxQuad, ucxQuadStar
    real(kind=rk), dimension(me%nFields, me%nFields) :: &
      & thermodynamic_fac, inv_thermodyn_fac
    integer :: restPosition  
    real(kind=rk) :: diff_coeff_3Spc(3,3), tdf_3Spc(3,3)
    real(kind=rk) :: moleFrac_3Spc(3)
    !--------------------------------------------------------------------------

    nFields = me%nFields
    ! molecular weight ratios
    phi = field(:)%fieldProp%species%molWeigRatio
    ! pdf stencil at rest (center)
    restPosition = me%layout%stencil(1)%restPosition

    !resistivities
    do iField = 1, nFields
      resi_coeff(iField,:) = field(iField)%fieldProp%species%resi_coeff
    end do  

    ! find chunksize and number of chunks required for initialzation
    chunkSize = io_buffer_size/me%layout%stencil(1)%QQ
    nChunks = ceiling( real(nElems, kind=rk)/real(chunkSize, kind=rk))

    do iChunk = 1, nChunks
      ! Number of elements read so far in previous chunks.
      elemOff = ( (iChunk-1)*chunksize )
      nChunkElems = min(chunkSize, nElems - elemOff)

      !allocate memory for density, vel, eq, coord
      allocate(xc(nChunkElems, 3))
      allocate(moleFrac(nFields, nChunkElems))
      allocate(rho(nFields, nChunkElems))
      allocate(ux(nFields, nChunkElems))
      allocate(uy(nFields, nChunkElems))
      allocate(uz(nFields, nChunkElems))
      allocate(Press(nChunkElems))
      ! Equilibrium part of the pdf
      allocate(fEqStar(me%layout%stencil(1)%QQ*nChunkElems))
      allocate(fEq(me%layout%stencil(1)%QQ*nChunkElems))
      !input array to calculate equilibrium from macro
      !This array contains all species density and velocities
      !allocate( input_loc( nFields*nChunkElems*4 ) )
      
      do iElem = 1, nChunkElems
        ! Calculate the coordinates
        elemPos = elemOff + iElem
        xc(iElem,1:3) = tem_BaryOfId( geometry%tree, &
          & me%glob%levelDesc( iLevel )%total( elemPos ) )
      end do

      do iField = 1, nFields 
        ! read velocity
        ux(iField, :) = tem_spatial_for( conf = params%general%solver%conf(1), &
          &                      me = field(iField)%ic%ini_state(2),           &
          &                      coord = xc, n = nChunkElems )
!write(dbgUnit(1),*) iField, 'ux_p ', ux(iField, :)
        ux(iField, :) = ux(iField, :)/params%physics%fac( iLevel )%vel
!write(dbgUnit(1),*) iField, 'ux_l ', ux(iField, :)

        uy(iField, :) = tem_spatial_for( conf = params%general%solver%conf(1), &
          &                      me = field(iField)%ic%ini_state(3),           &
          &                      coord = xc, n = nChunkElems )
!write(dbgUnit(1),*) iField, 'uy_p ', uy(iField, :)
        uy(iField, :) = uy(iField, :)/params%physics%fac( iLevel )%vel
!write(dbgUnit(1),*) iField, 'uy_l ', uy(iField, :)

        uz(iField, :) = tem_spatial_for( conf = params%general%solver%conf(1), &
          &                      me = field(iField)%ic%ini_state(4),           &
          &                      coord = xc, n = nChunkElems )
        uz(iField, :) = uz(iField, :)/params%physics%fac( iLevel )%vel

        ! mole fraction of each species
        moleFrac(iField, :) = tem_spatial_for(                                 &
          &                      conf = params%general%solver%conf(1),         &
          &                      me = field(iField)%ic%ini_state(1),           &
          &                      coord = xc, n = nChunkElems )
      end do

      ! read mixture hydrodynamic pressure
      Press = tem_spatial_for( conf = params%general%solver%conf(1),           &
        &                          me = mixture%ic%ini_state(1),               & 
        &                          coord = xc, n = nChunkElems )

      !convert to lattice unit
      press = press/params%physics%fac(iLevel)%press

      !compute species density from moleFraction and mixture number density
      !\rho_i = n0 * \chi_i * m_i + mass_fraction_i*rho0*KinePress/(cs^2*phi_i)
      !\rho_i = n0 * \chi_i * m_i + mass_fraction_i*rho0*Press/rho/(cs^2*phi_i)
      !\rho_i = n0 * \chi_i * m_i + mass_fraction_i*Press/(cs^2*phi_i)
      do iElem = 1, nChunkElems
        ! check if physical constraint sum(MolFrac) = 1
        ! check upto single precision
        if( abs( 1.0_single_k - real( sum(moleFrac(:,iElem)), kind=single_k ) )&
          & > eps_single_k ) then
          write(logUnit(1),*)'Error: Initial sum(molefraction) of all fields is'
          write(logUnit(1),*)'       not equal to 1 for iElem', iElem
          write(logUnit(1),*)' totMolFrac:', sum(moleFrac(:,iElem)) 
          write(logUnit(1),*)' MoleFracs: ', moleFrac(:,iElem) 
          call tem_abort()
        end if 
        !mixMolWeight = sum( chi_i * m_i )
        !mixMolWeight = sum( moleFrac(:, iElem)                                &
        !  &        * field(:)%fieldProp%species%molWeight )
        do iField = 1, nFields 
!         rho(iField, iElem) = mixture%rho0LB * moleFrac(iField, iElem)        &
!           & * field(iField)%fieldProp%species%molWeight / mixMolWeight
          rho(iField, iElem) = mixture%moleDens0LB * moleFrac(iField, iElem)   &
            & * field(iField)%fieldProp%species%molWeight
        end do
        ! add kinematic pressure term to density
        tot_massDens = sum(rho(:,iElem))
        do iField = 1, nFields
          rho(iField, iElem) = rho(iField, iElem) &
            & + ( cs2inv*(rho(iField,iElem)/tot_massDens)*Press(iElem)  &
            & / phi(iField) )
        end do  

!write(*,*) iElem, ' rho ', rho(:, iElem)          
!write(*,*) 'totMolDens ', mixture%rho0 &
! & / sum( moleFrac(:, iElem) * field(:)%fieldProp%species%molWeight )
      end do !iElem
!write(*,*) 'rho0', mixture%rho0
!write(*,*) 'molFrac', moleFrac(:, 1)
!write(*,*) 'molWeight', field(:)%fieldProp%species%molWeight
!write(*,*) 'mixtureMolWeight',  sum( moleFrac(:, 1)*field(:)%fieldProp%species%molWeight)
!write(*,*) 'rho tol ', rho(:,1), sum(rho(:, 1))
!stop

!      input_loc = 0.0_rk
!      ! input array contains density and velocity in the following order
!      ! |dens_1_sp1 | dens_1_sp2 | dens_2_sp1 | dens_2_sp2| ... 
!      ! | dens_nElems_sp1 | dens_nElems_sp2 
!      ! | u_x_1_sp1 | u_y_1_sp1 | u_z_1_sp1 | 
!      ! !u_x_1_sp2 | u_y_1_sp2 | u_z_1_sp2 |....
!      ! .. | u_x_nElems_sp1 | u_y_nElems_sp1 | u_z_nElems_sp1 | 
!      ! |u_x_nElems_sp2 | u_y_nElems_sp2 | u_z_nElems_sp2 |
!      offVel = nChunkElems*nFields
!      do iElem = 1, nChunkElems
!        posVel = offVel + (iElem-1)*nFields*3 
!        do iField = 1, nFields
!          input_loc( (iElem-1)*nFields + iField ) = rho(iField, iElem)  
!          input_loc( posVel + (iField-1)*3 + 1 ) = ux(iField, iElem)
!          input_loc( posVel + (iField-1)*3 + 2 ) = uy(iField, iElem)
!          input_loc( posVel + (iField-1)*3 + 3 ) = uz(iField, iElem)
!        end do  
!      end do
!
!      ! initialize the lower and upper bound
!      bound(1) = 1
!      bound(2) = nChunkElems
      do iField = 1, nFields
        ! Calculate the equilibrium distribution
        fEqStar = 0.0_rk
        fEq = 0.0_rk
        ! feq = feq_L(rho_k) + feq_BL(rho_k, v^star_k) + feq_Q(rho_k, v,v)
        ! v^star_k -> equilibrium velocity and
        ! v -> mass averaged mixture velocity are computed inside the 
        ! derived quantities subroutine
        !call me%varSys%derive( me%derVarPos(iField)%equilFromMacro )%do(         &
        !  &                    input = input_loc,                                &
        !  &                    bound = bound,                                    &
        !  &                    tree = geometry%tree,                             &
        !  &                    subTree = subTree,                                &
        !  &                    requVar = me%varSys%Variable%val(                 &
        !  &                              me%derVarPos(iField)%equilFromMacro ),  &
        !  &                    globSys = me%varSys,                              &
        !  &                    stencil = me%layout%stencil(1),                   &
        !  &                    res = fEqStar,                                    &
        !  &                    conf = me%varSys%solSpec_conf )

        do iElem = 1, nChunkElems
          do iField_2 = 1, nFields
           vel(1, iField_2) = ux(iField_2, iElem)
           vel(2, iField_2) = uy(iField_2, iElem)
           vel(3, iField_2) = uz(iField_2, iElem)
          end do
!write(dbgUnit(1),*) iElem, 'ux ', ux(:, iElem)
!write(dbgUnit(1),*) iElem, 'velocity ', vel
!write(dbgUnit(1),*) 'initial rho ', rho(iField, iElem), 'vel ', vel(:, iField)
          ! if scheme relaxation is bgk_withthermodynfac then
          ! calculate thermodynamic factor from c++ code and
          ! compute equilibrium velocity from thermodynamic factor
          if(trim(me%header%relaxation) == 'bgk_withthermodynfac') then
            if( nFields == 2) then
              moleFrac_3Spc(1) = moleFrac(1, iElem)
              moleFrac_3Spc(2) = moleFrac(2, iElem)
              moleFrac_3Spc(3) = moleFrac(2, iElem)

              call mus_calc_MS_DiffMatrix(nFields+1, me%mixture%temp0,           &
                &                        me%mixture%atm_press,                   &
                &                        moleFrac_3Spc*me%mixture%moleDens0,     &
                &                        diff_coeff_3Spc )

              call mus_calc_thermFactor( nFields+1, me%mixture%temp0,            &
                &                        me%mixture%atm_press,                   &
                &                        moleFrac_3Spc,                          &
                &                        tdf_3Spc )
!write(*,*) 'moleFrac_3Spc ', moleFrac_3Spc                
!write(*,*) 'diff_coeff_3Spc ', diff_coeff_3Spc                
!write(*,*) 'thermodynamic_fac_3Spc', tdf_fac_3Spc
              diff_coeff = 1.0_rk
              diff_coeff(1,2) = (diff_coeff_3Spc(1,2)*diff_coeff_3Spc(1,3))      &
                &             / (diff_coeff_3Spc(1,2)+diff_coeff_3Spc(1,3))
              diff_coeff(2,1) = diff_coeff(1,2)
!write(*,*) 'diff_coeff ', diff_coeff 
              thermodynamic_fac(1,1) = tdf_3Spc(1,1)
              thermodynamic_fac(1,2) = tdf_3Spc(1,2)
              thermodynamic_fac(2,1) = tdf_3Spc(2,1)
              thermodynamic_fac(2,2) = tdf_3Spc(2,2)
!write(*,*) 'thermodynamic_fac ', thermodynamic_fac
!stop
            else    
              call mus_calc_MS_DiffMatrix(nFields, me%mixture%temp0,             &
                &                        me%mixture%atm_press,                   &
                &                        moleFrac(:, iElem)*me%mixture%moleDens0,&
                &                        diff_coeff )
              call mus_calc_thermFactor( nFields, me%mixture%temp0,              &
                &                        me%mixture%atm_press,                   &
                &                        moleFrac(:, iElem),                     &
                &                        thermodynamic_fac )
!write(*,*) 'TODO: check mixture%moleDens0 and physics%moleDens0'
!stop
            endif    

            ! calculate resi_coeff from diff_coeff from C-code
            resi_coeff = params%physics%fac(ilevel)%diffusivity/diff_coeff
            do iField_2 = 1, nFields
              resi_coeff(iField_2, iField_2) = mixture%paramB
            end do
!write(*,*) 'diff_coeff ', diff_coeff
!write(dbgUnit(1),*) 'resi_coeff ', resi_coeff
!write(*,*) 'thermodynamic_fac ', thermodynamic_fac
!stop

            inv_thermodyn_fac = invert_matrix( thermodynamic_fac )

            eqVel = rho(iField, iElem)*vel(:, iField)
            do iField_2 = 1, nFields
              do iField_3 = 1, nFields
                eqVel(:) = eqVel(:)                                            &
                  &      + inv_thermodyn_fac(iField, iField_2)                 & 
                  &      * rho(iField_2, iElem)                                &  
                  &      * resi_coeff(iField_2, iField_3)                      &
                  &      * phi(iField_2) * moleFrac(iField_3, iElem)           &
                  &      * ( vel(:, iField_3) - vel( :, iField_2 ) )           &
                  &      / me%mixture%paramB
              end do    
            end do
          else
            eqVel = vel(:, iField)
            do iField_2 = 1, nFields
              eqVel(:) = eqVel(:) + resi_coeff(iField, iField_2)               & 
                &      * phi(iField) * moleFrac(iField_2, iElem)               &
                &      * ( vel(:, iField_2) - vel( :, iField ) )               &
                &      / me%mixture%paramB
            end do
            eqVel = rho(iField, iElem)*eqVel
          endif
          
          tot_massDens = sum(rho(:,iElem))
          ! mass averaged mixture velocity
          velAvg(1) = dot_product( rho(:, iElem), vel(1, :) )/tot_massDens
          velAvg(2) = dot_product( rho(:, iElem), vel(2, :) )/tot_massDens
          velAvg(3) = dot_product( rho(:, iElem), vel(3, :) )/tot_massDens

          ! velocity in quadratic term of equilibrium
          ! eqVel is div by rho since rho is multiplied in the
          ! calculate of eqVel with thermodynamic factor
          ! due to this reason Bilinear part in fEq is not multiplied
          ! by rho in fEq calculation.
          velQuadStar(:) = me%mixture%theta_eq*velAvg(:)                       &
            &            + (1.0_rk-me%mixture%theta_eq)                        &
            &            * ( eqVel(:) / rho(iField, iElem) )
          velQuad(:) = me%mixture%theta_eq*velAvg(:)                           &
            &            + (1.0_rk-me%mixture%theta_eq)*vel(:, iField)

          ! compute Eq with eqVel and initial velocity
          usqStar = dot_product(velQuadStar, velQuadStar)*t2cs2inv
          usq = dot_product(velQuad, velQuad)*t2cs2inv
    
          do iDir =1,me%layout%stencil(1)%QQ 
            ucx = dot_product(me%layout%stencil(1)%cxDir(:, iDir),             &
              &               vel(:,iField))
            ucxQuad = dot_product(me%layout%stencil(1)%cxDir(:, iDir),         &
              &               velQuad)

            ucxStar = dot_product(me%layout%stencil(1)%cxDir(:, iDir),         &
              &                   eqVel)
            ucxQuadStar = dot_product(me%layout%stencil(1)%cxDir(:, iDir),     &
              &                   velQuadStar)

            ! eqVel is actually is rho_i*eqVel so ucxStar is not multiplied
            ! with rho in below equation
            fEqStar(?IDX?(iDir, iElem, me%layout%stencil(1)%QQ))               & 
              & = me%layout%weight(iDir) * ( rho(iField, iElem) * ( phi(iField)&
              & + ucxQuadStar * ucxQuadStar * t2cs4inv - usqStar )             &
              & + ucxStar * cs2inv )

            fEq(?IDX?(iDir, iElem, me%layout%stencil(1)%QQ))                   &
              & = me%layout%weight(iDir) * rho(iField, iElem)                  &
              & * ( phi(iField) + ucx * cs2inv + ucxQuad * ucxQuad * t2cs4inv  &
              & - usq )
          end do! iDir

          ! equilibrium at rest
          select case( trim(me%header%layout) )
          case('d2q9')
            fEqStar(?IDX?(restPosition, iElem, me%layout%stencil(1)%QQ))       &
              & = me%layout%weight(restPosition) * rho(iField, iElem)          &
              & * ( (9._rk - 5._rk*phi(iField)) / 4.0_rk - usqStar )
            fEq(?IDX?(restPosition, iElem, me%layout%stencil(1)%QQ))           &
              & = me%layout%weight(restPosition) * rho(iField, iElem)          &
              & * ( (9._rk - 5._rk*phi(iField)) /4.0_rk - usq )
          case('d3q19')
            fEqStar(?IDX?(restPosition, iElem, me%layout%stencil(1)%QQ))       &
              & = me%layout%weight(restPosition) * rho(iField, iElem)          &
              & * ( (3._rk - 2._rk*phi(iField)) - usqStar )
            fEq(?IDX?(restPosition, iElem, me%layout%stencil(1)%QQ))           &
              & = me%layout%weight(restPosition) * rho(iField, iElem)          &
              & * ( (3._rk - 2._rk*phi(iField)) - usq )
          case default
            write(logUnit(1),*)'ERROR: initializing Multispecies LBM '
            write(logUnit(1),*)'Unknown stencil layout'
            call tem_abort()
          end select

        end do !iElem

        ! set initial pdf  
!write(dbgUnit(1),*) 'iField ', iField          
        do iElem = 1, nChunkElems
          elemPos = elemOff + iElem
!write(dbgUnit(1),*) 'iElem ', iElem, ' xc ', xc(iElem, :) 
!write(dbgUnit(1),*) ' rho ', rho(iField, iElem)
          do iDir =1,me%layout%stencil(1)%QQ 
            idx_iDir = me%varSys%Variable%val(iField)%varPos( iDir )
!write(dbgUnit(1),*) ' idir ', iDir, 'idx_idir', iDir
            ! In general we want to set for each element a valid entry.
            ! This was  IDX, but for PULL, it doesnt change anything
            ! for PUSH however, we need to set it to save!
            !KM: initialize both state vectors for boundaries
            !\todo need to check for multilevel
            me%pdf( iLevel )%state( &
& ?IDX?( idx_iDir, elemPos, me%varSys%nScalars, me%pdf(iLevel)%neigh),         &
              &   tLayerMin:tLayerMax )        &
              & = fEq(?IDX?(iDir, iElem, me%layout%stencil(1)%QQ))             &
              & + me%mixture%omega_diff*0.5_rk                                 &
              & * ( fEq(?IDX?(iDir, iElem, me%layout%stencil(1)%QQ))           &
              & - fEqStar(?IDX?(iDir, iElem, me%layout%stencil(1)%QQ)) )
!write(dbgUnit(1),*) 'fEq ', fEq(?IDX?(iDir, iElem, me%layout%stencil(1)%QQ)) 
!write(dbgUnit(1),*) 'fEqStar ', fEqStar(?IDX?(iDir, iElem, me%layout%stencil(1)%QQ)) 
!write(dbgUnit(1),*) 'fInit ', me%pdf( iLevel )%state( &
!    & ?IDX?( idx_iDir, elemPos, me%varSys%nScalars, me%pdf(iLevel)%neigh), &
!                &   1 )

          end do !idir
        end do ! iElem
      end do ! iField

      ! deallocate memory for next chunk  
      deallocate(xc)
      deallocate(rho)
      deallocate(moleFrac)
      deallocate(ux)
      deallocate(uy)
      deallocate(uz)
      deallocate(press)
      deallocate(fEqStar)
      deallocate(fEq)
!      deallocate(input_loc)

    end do ! chunk

  end subroutine mus_init_MSLiquid_withFlowParams
!******************************************************************************!

!******************************************************************************!
  !> Initialize the flow from calculated quantitites like density, velocity etc.
  !! for multispecies lbm
  subroutine mus_init_MSGas_withFlowParams( me, geometry, params,              &
    &                                       Field, nElems, iLevel,          &
    &                                       tLayerMin, tLayerMax )
    !--------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: me !< Scheme type
    type(mus_param_type), intent(in) :: params  !< Global parameters
    type(mus_geom_type), intent(in) :: geometry !< Geometry information
    type(mus_field_type), intent(inout) :: field(:) !< Field type
    ! type(mus_mixture_type), intent(inout) :: mixture !< mixture type
    integer, intent(in)                 :: nElems !< Number of Elements
    integer, intent(in)                 :: iLevel !< Level index
    !> min time layer of state to initialize 
    integer, intent(in) :: tLayerMin
    !> max time layer of state to initialize 
    integer, intent(in) :: tLayerMax
    !--------------------------------------------------------------------------
    integer :: iDir, idx_iDir, iElem, iField, nFields
    real(kind=rk), allocatable :: fEq(:)
    real(kind=rk), allocatable :: xc(:,:), ux(:,:), uy(:,:), uz(:,:)
    real(kind=rk), allocatable :: rho(:,:)
    real(kind=rk), allocatable :: input_loc(:)
    real(kind=rk) :: phi(me%nFields)
    integer :: iChunk, nChunks, chunkSize, nChunkElems, elemPos, elemOff
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !--------------------------------------------------------------------------

    nFields = me%nFields
    ! molecular weight ratios
    phi = field(:)%fieldProp%species%molWeigRatio
!write(dbgUnit,*) 'theta_eq ', theta_eq    

    ! find chunksize and number of chunks required for initialzation
    chunkSize = io_buffer_size/me%layout%stencil(1)%QQ
    nChunks = ceiling( real(nElems, kind=rk)/real(chunkSize, kind=rk))

    do iChunk = 1, nChunks
      ! Number of elements read so far in previous chunks.
      elemOff = ( (iChunk-1)*chunksize )
      nChunkElems = min(chunkSize, nElems - elemOff)

      !allocate memory for density, vel, eq, coord
      allocate(xc(nChunkElems, 3))
      allocate(rho(nFields, nChunkElems))
      allocate(ux(nFields, nChunkElems))
      allocate(uy(nFields, nChunkElems))
      allocate(uz(nFields, nChunkElems))
      allocate(input_loc( nChunkElems*4))
      ! Equilibrium part of the pdf
      allocate(fEq(me%layout%stencil(1)%QQ*nChunkElems))
      
      do iElem = 1, nChunkElems
        ! Calculate the coordinates
        elemPos = elemOff + iElem
        xc(iElem,1:3) = tem_BaryOfId( geometry%tree, &
          & me%glob%levelDesc( iLevel )%total( elemPos ) )
      end do
      

      do iField = 1, nFields 
        !partial pressure
        rho(iField, :) = tem_spatial_for( conf = params%general%solver%conf(1),&
          &                      me = field(iField)%ic%ini_state(1),           &
          &                      coord = xc, n = nChunkElems )
        ! read velocity
        ux(iField, :) = tem_spatial_for( conf = params%general%solver%conf(1), &
          &                      me = field(iField)%ic%ini_state(2),           &
          &                      coord = xc, n = nChunkElems )
        ux(iField, :) = ux(iField, :)/params%physics%fac( iLevel )%vel

        uy(iField, :) = tem_spatial_for( conf = params%general%solver%conf(1), &
          &                      me = field(iField)%ic%ini_state(3),           &
          &                      coord = xc, n = nChunkElems )
        uy(iField, :) = uy(iField, :)/params%physics%fac( iLevel )%vel

        uz(iField, :) = tem_spatial_for( conf = params%general%solver%conf(1), &
          &                      me = field(iField)%ic%ini_state(4),           &
          &                      coord = xc, n = nChunkElems )
        uz(iField, :) = uz(iField, :)/params%physics%fac( iLevel )%vel
      end do

      !partial pressure
      do iField = 1, nFields 
        !p_sigma = r_sigma * cs2 * phi_sigma
        !rho = press * cs2inv / phi
        rho(iField, :) = ( rho(iField, :) * cs2inv / phi(iField) ) &
          &            / params%physics%fac( iLevel )%press
      end do    

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = nChunkElems
      do iField = 1, nFields
!write(dbgUnit,*) 'iField ', iField      
        !linearized the input for equilibrium function pointer
        !with following order
        !|dens_1 | dens_2 | .. | dens_nElems | ux_1 | uy_1 | uz_1 | ..
        !!.. | ux_nelems| uy_nelems| uz_nelems|
        input_loc = 0._rk
        input_loc(1:nChunkElems) = rho(iField, :)
        do iElem = 1, nChunkElems
          ! MH: Fix for the intel compiler:
          ! entries have to be assigned one by one instead of 
          ! assigning them as an array of (/ux, uy, uz/)
          input_loc( nChunkElems + (iElem-1)*3 + 1) = ux( iField, iElem )
          input_loc( nChunkElems + (iElem-1)*3 + 2) = uy( iField, iElem )
          input_loc( nChunkElems + (iElem-1)*3 + 3) = uz( iField, iElem )
        end do

        ! Calculate the equilibrium distribution
        fEq = 0.0_rk
        call me%varSys%derive( me%derVarPos(iField)%equilFromMacro )%do(         &
          &                    input = input_loc,                                &
          &                    bound = bound,                                    &
          &                    tree = geometry%tree,                             &
          &                    subTree = subTree,                                &
          &                    requVar = me%varSys%Variable%val(                 &
          &                              me%derVarPos(iField)%equilFromMacro ),  &
          &                    globSys = me%varSys,                              &
          &                    stencil = me%layout%stencil(1),                   &
          &                    res = fEq,                                        &
          &                    conf = me%varSys%solSpec%conf )

        do iElem = 1, nChunkElems
          elemPos = elemOff + iElem
          do iDir =1,me%layout%stencil(1)%QQ 
            idx_iDir = me%varSys%Variable%val(iField)%varPos( iDir )
            me%pdf( iLevel )%state( &
            ! In general we want to set for each element a valid entry.
            ! This was  IDX, but for PULL, it doesnt change anything
            ! for PUSH however, we need to set it to save!
            !KM: initialize both state vectors for boundaries
    & ?IDX?( idx_iDir, elemPos, me%varSys%nScalars, me%pdf(iLevel)%neigh), &
            !\todo need to check for multilevel
              &   tLayerMin:tLayerMax )        &
              & = fEq((iElem-1)*me%layout%stencil(1)%QQ + iDir) 
          end do !idir
        end do !ielem
      end do !iField  

      ! deallocate memory for next chunk  
      deallocate(xc)
      deallocate(rho)
      deallocate(ux)
      deallocate(uy)
      deallocate(uz)
      deallocate(input_loc)
      deallocate(fEq)

    end do ! chunk

  end subroutine mus_init_MSGas_withFlowParams
!******************************************************************************!

!******************************************************************************!
  !> Initialize flow field 
  !!
  subroutine mus_init_withRestart( scheme, geometry, params, pdfGlobal, Field )
    !---------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type), intent(inout) :: scheme(:)
    !> Global parameters
    type(mus_param_type), intent(inout) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> global info type
    type(pdf_globInfo_type) :: pdfGlobal
    !> Field type
    type(mus_field_type), intent(inout) :: Field
    !---------------------------------------------------------------------------
    integer :: iChunk
    integer :: offset, iSys, nDofs, nChunkElems_res, elemoff
    ! array for chunkwise memory bounded output
    real(kind=rk), allocatable :: chunk(:)
    !---------------------------------------------------------------------------
    allocate( chunk( io_buffer_size ))
    ! The file is read in ini condition type. Assign the file name to
    ! restart controller
    field%restart%controller%readFileName = field%ic%RestartFile

    ! Load the restart header for this field
    call tem_restart_readHeader( me       = field%restart,                     &
      &                          tree     = geometry%tree,                     &
      &                          timing   = params%general%simControl%now,     &
      &                          proc     = params%general%proc,               &
      &                          readMesh = .false. )
    
    ! Initialize the depend, to get the position of restart varsys
    call tem_init_depend( medepend    = field%ic%depend,                       &
      &                   varSys      = scheme(:)%varSys,                      &
      &                   resSyslabel = 'restart',                             &
      &                   label_list  = scheme(:)%header%label,                &
      &                   iScheme     = 0 )

    ! Initialize the restart for this variable system. The scheme on 
    ! which the requested system is dependent is passed as an input
    ! to find the position of varsys in the global system
    !\todo: depno: KJ: multiple depend tables are not making much sense here
    ! it has to be discussed and resolved, right now passing first 
    ! instance of depend array
    iSys = field%ic%depend(1)%useScheme

    ! Invoke the routine to allocate restart datatypes
    call tem_init_restart_alloc( me      = field%restart,                      &
      &                          solver  = params%general%solver,              &
      &                          tree    = geometry%tree )

    ! Create MPI datatypes for setting views on files
    call tem_init_restart_create_types( me     = field%restart,                &
      &                                 varSys = field%restart%varSys,         &
      &                                 tree   = geometry%tree,                &
      &                                 iSys   = iSys )

    call tem_restart_getTotalChunks( restart = field%restart,                  &
      &                              tree    = geometry%tree )

    ! Open the restart file and set the processes' view on the file
    !\todo: depno
    call tem_restart_openRead_single( me = field%restart, iSys = iSys )

    ! Read the data from restart file into the chunk array
    ! first calculate the offset which is dependent on the positon
    ! of the variable system

!    elemOff = 0
    write(logUnit(1),*) 'Reading restart...' 

    do iChunk = 1, field%restart%nChunks
      
      elemOff = ((iChunk-1)*field%restart%chunkSize )
      field%restart%nChunkElems = min( field%restart%chunkSize,                &
        &      geometry%tree%nElems-elemOff)

      nDofs = field%restart%nDofs
      nChunkElems_res = field%restart%nChunkElems
      if( iSys == 1 ) then
        offset = 1
      else
        offset = 1 + sum( field%restart%varSys                                 &
          &                           (1:iSys)%nScalars*nChunkElems_res*nDofs )
      end if

      call tem_restart_readData_single( restart = field%restart,               &
        &                               chunk   = chunk,                       &
        &                               iSys    = iSys,                        &
        &                               offset  = offset )

      ! \todo: KJ: Invoke the unserialization routine to put the data in chunk
      ! in state arrays. This is quite a work around at this moment
      ! and the index of chunk needs to be thought of in detail
      
      call mus_rest_unserialize( scheme    = scheme,                           &
        &                        pdfGlobal = pdfGlobal,                        &
        &                        restart   = field%restart,                    &
        &                        tree      = geometry%tree, chunk=chunk,       &
        &                        iChunk    = iChunk,                           &
        &                        iSys_idx  = iSys )

!      elemOff = elemOff + field%restart%nChunkElems

    end do !nChunks within restart

    !close the restart file
    call tem_restart_closeRead( me = field%restart, iSys_idx = iSys)

    ! Free the datatypes by finalizing the restart
    call tem_restart_finalize_single(me = field%restart, iSys = iSys)

    deallocate( chunk )

  end subroutine mus_init_withRestart
!******************************************************************************!

  !****************************************************************************!
  !> This routine initialize omega values in state vector for nonNewtonian fluid
  !!
  subroutine mus_init_omega_byOmega( state, params, geometry, nElems, &
    &                        varSys, total, ic, tLayerMin, tLayerMax )
    !---------------------------------------------------------------------------
    real(kind=rk), dimension(:,:) :: state !< state vector
    !> Global parameters
    type( mus_param_type ), intent(in) :: params
    !> Geometry information
    type( mus_geom_type ), intent(in) :: geometry
    !> initial condition
    type( tem_ini_condition_type ) :: ic
    !> Number of Elements
    integer, intent(in) :: nElems
    !> variable system
    type( tem_var_system_type ) :: varSys
    !> total list of elements
    integer(kind=long_k), intent(in) :: total(:)
    !> min time layer of state to initialize
    integer, intent(in) :: tLayerMin
    !> max time layer of state to initialize
    integer, intent(in) :: tLayerMax
    !---------------------------------------------------------------------------
    real(kind=rk), allocatable :: omegas(:) ! omega values from ic
    real(kind=rk), allocatable :: coord(:,:) ! coordinates
    integer :: iElem, iChunk, nChunks, chunkSize, nChunkElems, elemPos, elemoff
    integer :: nScalars, omegaPos
    !---------------------------------------------------------------------------

    write(logUnit(1),*) " Initialize omega by omegas from initial condition."

    nScalars = varSys%nScalars
    omegaPos = varSys%variable%val(2)%varPos(1)

    ! find chunksize and number of chunks required for initialzation
    chunkSize = io_buffer_size / nScalars
    nChunks = ceiling( real(nElems, kind=rk) / real(chunkSize, kind=rk) )

    do iChunk = 1, nChunks
      ! Number of elements read so far in previous chunks.
      elemOff = ( (iChunk-1)*chunksize )
      nChunkElems = min(chunkSize, nElems - elemOff)
      allocate( omegas(nChunkElems) )
      allocate( coord( nChunkElems, 3 ) )

      do iElem = 1, nChunkElems
        ! Calculate the coordinates
        elemPos = elemOff + iElem
        coord(iElem,1:3) = tem_BaryOfId( geometry%tree, total( elemPos ) )
      end do

      ! obtain omega values from initial condition
      omegas(:) = tem_spatial_for(                         &
        &           conf  = params%general%solver%conf(1), &
        &           me    = ic%ini_state(11),              &
        &           coord = coord,                         &
        &           n     = nChunkElems )

      do iElem = 1, nChunkElems
        elemPos = elemOff + iElem

        ! assign omega
state( ?IDX?( omegaPos, elemPos, nScalars), tLayerMin:tLayerMax ) = omegas(iElem)

      end do ! iElem

      deallocate(omegas)
      deallocate(coord)

    end do ! iChunk

  end subroutine mus_init_omega_byOmega
  !****************************************************************************!


!******************************************************************************!
  !> test the shear stress setting with the non-eq part and re-computing it with
  !! the derive routine
  !! In the first part, this routine generates pressure, velocity, and shear
  !! stress from random number, then fEq is calculated from pressure and
  !! velocity, fNeq is calculated from shear stress. Next fTmp (=fEq+fNeq) is
  !! utilized to calculate fNeq reversily, to double check with the fNeq
  !! calculation previously.
  !!
  !! \todo Make this routine a unit testcase.
  !!
  subroutine mus_checkShearStress( scheme, geometry, params, pdfGlobal )
    !---------------------------------------------------------------------------
    !> Scheme type
    type(mus_scheme_type), intent(inout) :: scheme
    !> Global parameters
    type(mus_param_type), intent(in) :: params
    !> Geometry information
    type(mus_geom_type), intent(in) :: geometry
    !> global info type
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !---------------------------------------------------------------------------
    integer :: iLevel, iVal, jVal
    real(kind=rk), allocatable :: fEq(:), fnEq(:), fTmp(:), fnEqDerived(:)
    real(kind=rk), allocatable :: tMom(:)
    ! Physical quantities
    real(kind=rk) :: pPhRef,    uxPhRef(3), mSxxPhRef(3,3), SxxPhRef(6)
    real(kind=rk) :: pPhres(1), uxPhres(3), SxxPhres(6),    strPhRes(6)
    ! Lattice  quantities
    real(kind=rk) :: rhoLBref, uxLBref(3), mSxxLBref(3,3), SxxLBref(6)
    real(kind=rk) :: rhoLBres, uxLBres(3), SxxLBres(6), strLBres(6)
    ! errors
    real(kind=rk) :: errpPh, errVel(3), errSxx(6)
    ! temporary variables
    real(kind=rk) :: input_loc(6), omega, rho0, viscLB, rho0div
    integer :: nStresses, iShearMin, iShearMax
    integer(kind=long_k) :: targetID
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    integer :: iField = 1
    !---------------------------------------------------------------------------

    ! only do this on root rank
    if( params%general%proc%rank == 0 ) then
      write(logUnit(6),*) 'Checking shear stress computation for scaling: '//&
        &                                                 trim(params%scaling)

      ! loop over all levels
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel

        targetID = tem_idOfCoord( [1,1,1,iLevel] )

        ! build the subTree from the global information and the targetTreeID
        subTree%global = geometry%tree%global
        call tem_local_subTree_from( me = subTree, treeID = (/targetID/))

        ! get omega and viscosity
        omega = scheme%field( iField )%fieldProp%fluid%omLvl( iLevel )
        rho0  = scheme%field( iField )%fieldProp%fluid%rho0
        viscLB = cs2 * ( 1._rk/omega - 0.5_rk )

        write(logUnit(6), "(A,I0)")   '   on level: ', iLevel
        write(logUnit(6), "(A,F6.4)") '      omega: ', omega
        write(logUnit(6), "(A,F6.4)") '  viscosity: ', viscLB

        ! allocate arrays
        if( allocated( fTmp )) deallocate( fTmp )
        if( allocated( fNeq )) deallocate( fNeq )
        if( allocated( fEq ))  deallocate( fEq )
        if( allocated( fnEqDerived ))  deallocate( fnEqDerived )
        if( allocated( tMom ))  deallocate( tMom )
        allocate( fTmp( scheme%layout%stencil(1)%QQ ))
        allocate(  fEq( scheme%layout%stencil(1)%QQ ))
        allocate( fNeq( scheme%layout%stencil(1)%QQ ))
        allocate( fnEqDerived( scheme%layout%stencil(1)%QQ ))
        allocate( tMom( scheme%layout%stencil(1)%QQ ))

        ! generate a random pressure: pPhRef
        call random_number(pPhRef)

        uxPhRef  = 0._rk
        ! generate random velocities: uxPhRef
        do jVal = 1, pdfGlobal%nDims
          call random_number( uxPhRef( jVal ))
        end do

        mSxxPhRef = 0._rk
        ! 3D: nStresses = 6, iShearMin = 5, iShearMax = 10
        ! 2D: nStresses = 4, iShearMin = 4, iShearMax = 7
        ! 1D: nStresses = 1, iShearMin = 3, iShearMax = 3
        nStresses = max( 3*(pdfGlobal%nDims-1), 1 )
        iShearMin = pdfGlobal%nDims+2
        iShearMax = iShearMin + nStresses - 1

        ! generate random velocities: uxPhRef
        do jVal = 1, pdfGlobal%nDims
          call random_number( uxPhRef( jVal ))
        end do
        ! generate random strain rate: mSxxPhRef
        do jVal = 1, pdfGlobal%nDims
          do iVal = 1, pdfGlobal%nDims
            call random_number( mSxxPhRef( iVal, jVal ))
          end do
        end do
        mSxxPhRef(1,2) = mSxxPhRef(2,1)
        mSxxPhRef(2,3) = mSxxPhRef(3,2)
        mSxxPhRef(1,3) = mSxxPhRef(3,1)

        ! Physical to LB units
        mSxxLBref = mSxxPhRef / params%physics%fac( iLevel )%strainRate
        uxLBref   = uxPhRef   / params%physics%fac( iLevel )%vel
        rhoLBref  = pPhRef    / params%physics%fac( iLevel )%press*cs2inv
        if( trim( scheme%header%kind ) == 'lbm_incomp' ) then
          rhoLBref  = rhoLBref + rho0
        end if

        ! linearized strain rate matrix (m)
        SxxPhRef(:) = 0.0_rk
        SxxLBRef(:) = 0.0_rk
        if ( scheme%layout%stencil(1)%nDims == 3 ) then
          SxxPhRef(1) = mSxxPhRef(1,1)
          SxxPhRef(2) = mSxxPhRef(2,2)
          SxxPhRef(3) = mSxxPhRef(3,3)
          SxxPhRef(4) = mSxxPhRef(1,2)
          SxxPhRef(5) = mSxxPhRef(2,3)
          SxxPhRef(6) = mSxxPhRef(1,3)
          SxxLBRef(1) = mSxxLBRef(1,1)
          SxxLBRef(2) = mSxxLBRef(2,2)
          SxxLBRef(3) = mSxxLBRef(3,3)
          SxxLBRef(4) = mSxxLBRef(1,2)
          SxxLBRef(5) = mSxxLBRef(2,3)
          SxxLBRef(6) = mSxxLBRef(1,3)
        else
          SxxPhRef(1) = mSxxPhRef(1,1)
          SxxPhRef(2) = mSxxPhRef(2,2)
          SxxPhRef(3) = mSxxPhRef(1,2)
          SxxLBRef(1) = mSxxLBRef(1,1)
          SxxLBRef(2) = mSxxLBRef(2,2)
          SxxLBRef(3) = mSxxLBRef(1,2)
        end if

        input_loc = 0._rk
        input_loc(1) = rhoLBref
        input_loc(2:4) = uxLBref(:)

        ! initialize the lower and upper bound
        bound(1) = 1
        bound(2) = 1

        ! Calculate the equilibrium distribution
        write(logUnit(7), "(A)") 'Calculate fEq from rho and velocity'
        call scheme%varSys%derive(scheme%derVarPos(iField)%equilFromMacro)%do( &
          &          input   = input_loc,                                      &
          &          bound   = bound,                                          &
          &          tree    = geometry%tree,                                  &
          &          subTree = subTree,                                        &
          &          requVar = scheme%varSys%Variable%val(                     &
          &                         scheme%derVarPos(iField)%equilFromMacro),  &
          &          globSys = scheme%varSys,                                  &
          &          stencil = scheme%layout%stencil(1),                       &
          &          conf    = scheme%varSys%solSpec%conf,                     &
          &          res     = fEq )

        ! Calculate non-equilibrium
        write(logUnit(7), "(A)") 'Calculate fNeq from Sxx by '//trim(params%scaling)
        select case( trim( params%scaling ))
        case('acoustic')
          ! Acoustic fNeq with defining the stress tensor instead of the
          ! shear rate tensor in the lua file
          fNeq(1:scheme%layout%stencil(1)%QQ) =                                &
            &  getnEq_acoustic(                                                &
            &              layout = scheme%layout,                             &
            &              rho0   = scheme%field(iField)%fieldProp%fluid%rho0, &
            &              omega  = omega,                                     &
            &              Sxx    = mSxxLBref  )
        case('diffusive')
          ! Diffusive non-equilibrium part with the shear rate tensor
          ! given as an input
          fNeq( 1:scheme%layout%stencil(1)%QQ) =                               &
            &  getnEq_diffusive(                                               &
            &     layout = scheme%layout,                                      &
            &     omega  = scheme%field(iField)%fieldProp%fluid%omLvl(iLevel), &
            &     Sxx    = mSxxLBref )
        end select

        ! combine fEq and fNeq to get f
        write(logUnit(7), "(A)") 'Calculate PDF = fEq + fNeq'
        fTmp = fEq + fNeq

      ! double check fNeq by utilizing fTmp to re-calculate fNeq
      write(logUnit(7), "(A)") 'Calculate fNeqDerived from PDF'
      call scheme%varSys%derive(scheme%derVarPos(iField)%nonEquilfromState)%do(&
          &      input   = fTmp,                                               &
          &      bound   = bound,                                              &
          &      tree    = geometry%tree,                                      &
          &      subTree = subTree,                                            &
          &      requVar = scheme%varSys%Variable%val(                         &
          &                      scheme%derVarPos(iField)%nonEquilFromState ), &
          &      globSys = scheme%varSys,                                      &
          &      stencil = scheme%layout%stencil(1),                           &
          &      conf    = scheme%varSys%solSpec%conf,                         &
          &      res     = fNeqDerived )

        ! calcualte error between fNeq and fNeqDerived
        write(logUnit(7), "(A)") 'Calculate error bewtween fNeqDerived and fNeq'
        write(logUnit(7),*) 'fNeq Error abs = ', maxval(abs((fNeq-fNeqDerived)))
        if ( maxval(abs((fNeq-fNeqDerived))) > 1000._rk*eps ) then
          write(logUnit(1),*) 'Deviation between the nEq-computation in '//    &
            &                 'the derive module and a manual computation '//  &
            &                 'by fNeq = f-fEq. Please check!'
          write(logUnit(1),*) 'fNeq:    ', fNeq
          write(logUnit(1),*) 'fNeqDer: ', fNeqDerived
          write(logUnit(1),*) 'maxError:', maxval(abs((fNeq-fNeqDerived)))
        else
          write(logUnit(7),"(A)") 'Within tolerance.'
        end if

        !------------------------------------------------------------------------
        !! JQ: tMom is actually not used in this routine
        ! calculate moments from pdf
        ! m(1) = rho
        ! m(2) = velX
        ! m(3) = velY
        ! m(4) = velZ
        ! m(5) = pressure + rho*velX*velX - tauXX
        ! m(6) = pressure + rho*velY*velY - tauYY
        ! m(7) = pressure + rho*velZ*velZ - tauZZ
        ! m(8) =            rho*velX*velY - tauXY
        write(logUnit(6), "(a)") 'calculate moments from PDF'
        tMom(:) = matmul( scheme%layout%moment%toMoments%A, fTmp)

        if( trim( scheme%header%kind ) == 'lbm' ) then
          rho0div = 1._rk / tMom(1)
        else
          ! \todo: is this correct? should be rho0div = 1._rk / rho0?
          ! though rho0 = 1 for lbm incompressible
          rho0div = rho0
        end if

        ! The following moments calculation actually NOT used
        ! convert moments(5:10) to shear stress (tau)
        if( scheme%layout%stencil(1)%nDims == 2 ) then
          tMom( 4 ) = tMom( 4 ) - cs2*tMom( 1) - tMom( 2)*tMom( 2)*rho0div
          tMom( 5 ) = tMom( 5 ) - cs2*tMom( 1) - tMom( 3)*tMom( 3)*rho0div
          tMom( 6 ) = tMom( 6 ) - tMom( 2)*tMom( 3)*rho0div
          tMom(4:6) = -( 1._rk - omega*0.5_rk )*tMom(4:6)*convPrePost( omega )
          ! Convert shear stress to physical strain rate
          tMom(4:6) =   tMom(4:6) / viscLB * rho0div &
            &         * params%physics%fac( iLevel )%strainRate
        else
          tMom(  5 ) = tMom(  5 ) - cs2*tMom( 1 ) - tMom( 2 )*tMom( 2 )*rho0div
          tMom(  6 ) = tMom(  6 ) - cs2*tMom( 1 ) - tMom( 3 )*tMom( 3 )*rho0div
          tMom(  7 ) = tMom(  7 ) - cs2*tMom( 1 ) - tMom( 4 )*tMom( 4 )*rho0div
          tMom(  8 ) = tMom(  8 ) - tMom( 2 )*tMom( 3 )*rho0div
          tMom(  9 ) = tMom(  9 ) - tMom( 3 )*tMom( 4 )*rho0div
          tMom( 10 ) = tMom( 10 ) - tMom( 4 )*tMom( 2 )*rho0div
          tMom(5:10) = -( 1._rk - omega*0.5_rk )*tMom(5:10)*convPrePost( omega )
          ! Convert shear stress to physical strain rate
          tMom(5:10) =   tMom(5:10) / viscLB * rho0div * 0.5_rk &
            &          * params%physics%fac( iLevel )%strainRate
        end if
        !------------------------------------------------------------------------

        ! Get pressure(pPhRes), velocity(uxPhRes) and shearstress (strPhRes)
        ! from the given distribution fTmp the initial quantities again
        call scheme%varSys%derive( scheme%derVarPos(iField)%pressure )%do(     &
          &                    input   = fTmp,                                 &
          &                    bound   = bound,                                &
          &                    tree    = geometry%tree,                        &
          &                    subTree = subTree,                              &
          &                    requVar = scheme%varSys%Variable%val(           &
          &                               scheme%derVarPos(iField)%pressure ), &
          &                    globSys = scheme%varSys,                        &
          &                    stencil = scheme%layout%stencil(1),             &
          &                    conf    = scheme%varSys%solSpec%conf,           &
          &                    res     = pPhres )
        if( trim( scheme%header%kind ) == 'lbm' ) then
          pPhRes = pPhRes - cs2*rho0
        end if
        rhoLBres = pPhRes(1) * cs2inv
        pPhRes   = pPhRes * params%physics%fac( iLevel )%press

        call scheme%varSys%derive( scheme%derVarPos(iField)%velocity )%do(     &
          &                     input   = fTmp,                                &
          &                     bound   = bound,                               &
          &                     tree    = geometry%tree,                       &
          &                     subTree = subTree,                             &
          &                     requVar = scheme%varSys%Variable%val(          &
          &                               scheme%derVarPos(iField)%velocity ), &
          &                     globSys = scheme%varSys,                       &
          &                     stencil = scheme%layout%stencil(1),            &
          &                     conf    = scheme%varSys%solSpec%conf,          &
          &                     res     = uxLBres(:) )
        uxPhRes = uxLBres * params%physics%fac( iLevel )%vel

        ! get shear stress LB: strLBres
        call scheme%varSys%derive( scheme%derVarPos(iField)%shearstress )%do(  &
          &                       input   = fTmp,                              &
          &                       bound   = bound,                             &
          &                       tree    = geometry%tree,                     &
          &                       subTree = subTree,                           &
          &                       requVar = scheme%varSys%Variable%val(        &
          &                            scheme%derVarPos(iField)%shearstress ), &
          &                       globSys = scheme%varSys,                     &
          &                       stencil = scheme%layout%stencil(1),          &
          &                       conf    = scheme%varSys%solSpec%conf,        &
          &                       res     = StrLBres(:) )
        ! get strain rate LB: sxxLBres
        sxxLBres    = strLBres / viscLB * rho0div * 0.5_rk
        ! get train rate phy: sxxPhres
        sxxPhRes    = SxxLBres * params%physics%fac( iLevel )%strainRate
        ! get shear stress phy: strPhyres
        strPhres    = strLBres * params%physics%fac( iLevel )%press

        !-----------------------------------------------------------------------
        !   Error evaluation
        !-----------------------------------------------------------------------
        errpPh = 0._rk
        errVel = 0._rk
        errSxx = 0._rk

        write(logUnit(7),"(A)") 'Calculate error between ref pressure, velocity,&
          & strain rate and derived results of those.'
        ! calculate pressure error
        if( abs(pPhRef) > eps ) then
          errpPh = abs((pPhRef-pPhRes(1))/pPhRef)
        end if
        ! calculate velocity error
        do jVal = 1,pdfGlobal%nDims
          if( abs(uxPhRef(jVal)) > eps ) then
            errVel(jVal) = abs((uxPhRef(jVal)-uxPhRes(jVal))/uxPhRef(jVal))
          end if
        end do
        ! calculate shear stress error
        do jVal = 1, size( sxxPhRef )
          if( abs(sxxPhRef(jVal)) > eps ) then
            errSxx(jVal) = abs((SxxPhRef(jVal)-SxxPhRes(jVal))/SxxPhRef(jVal))
          end if
        end do

        ! check if error exceeds tolerance
        if( maxval( abs( [errpPh, errVel, errSxx] )) > 100000._rk*eps ) then

          write(logUnit(1),*)'max error:   ',   &
            &                 maxval(abs( [errpPh, errVel, errSxx] ))
          write(logUnit(1),*)'error limit: ', eps*100000._rk
          write(logUnit(1),*)'error in shear stress computation.'
          write(logUnit(1),*)'Larger deviation than allowed!'

          write(logUnit(10),*)'    input data'
          write(logUnit(10),*)'    rho        ux                 Sxx      '
          write(logUnit(10),*)'-----------Physical units-------------------------'
          write(logUnit(10),*)'ref ', pPhRef, uxPhRef, SxxPhRef
          write(logUnit(10),*)'res ', pPhRes, uxPhRes, SxxPhRes
          write(logUnit(10),*)'err ', errpPh, errVel, errSxx
          write(logUnit(10),*)'-----------Lattice units--------------------------'
          write(logUnit(10),*)'ref ', rhoLBRef, uxLBRef, SxxLBRef
          write(logUnit(10),*)'val ', rhoLBRes, uxLBRes, SxxLBRes

        else
          write(logUnit(7),"(A)") 'Within tolerance.'
        end if ! error within tolerance?

        call tem_destroy_subTree(subTree)

      end do
    end if

  end subroutine mus_checkShearStress
!******************************************************************************!


end module  mus_flow_module
!******************************************************************************!
