! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the definition and methods for
!! boundarys.
module mus_param_module

  ! include musubi modules
  use mus_physics_module,     only: mus_physics_type

  ! include treelm modules
  use env_module,             only: labelLen, rk
  use tem_timer_module,       only: tem_addTimer, tem_timer_type
  use tem_logging_module,     only: logUnit
  use tem_general_module,     only: tem_general_type
  use tem_tools_module,       only: tem_horizontalSpacer

  ! include aotus modules
  use aotus_module, only: flu_state, aot_get_val

  implicit none

  private

  public :: mus_init_timers
  public :: mus_init_leveltimers
  public :: mus_param_type
  public :: mus_load_param


  !> Global parameter type definition, filled with
  type mus_param_type
    !> Treelm param parameter type
    type( tem_general_type ) :: general
    !> contains basic SI units to convert from lattice to physical and
    !! vice versa
    type(mus_physics_type) :: physics
    ! Handles for timing objects
    !> timer handle for advection relaxation
    integer :: advrelTimer
    !> timer handle for communication
    integer :: commTimer
    !> timer handle for interpolation
    integer :: intpTimer_low, intpTimer_hgh
    !> timer handle for the dyn_loadBal routine
    integer :: dyn_loadBalTimer
    !> timer handle for the complete mainloop
    integer :: mainloopTimer
    !> timer handle for initialising the levelDescriptor
    integer :: initLvlDTimer
    !> timer handle for loading / creating the mesh and config
    integer :: loadMeshTimer
    !> timer handle for writing restart
    integer :: wRestartTimer
    !> timer handle for source terms
    integer :: sourceTimer
    !> timer handle for checking nans
    integer :: isNanTimer
    !> timer handle for setboundary terms
    integer :: setBndTimer
    !> timer handle for levelwise interpolation
    integer, allocatable :: lw_tIntp(:)
    !> timer handle for levelwise computation
    integer, allocatable :: lw_tCompute(:)
    !> timer handle for levelwise set boundary
    integer, allocatable :: lw_tSetBnd(:)
    !> timer handle for IBM (levelwise is not used presently)
    integer, allocatable :: lw_tIBM(:)
    !> Array of timers for level wise timing
    type(tem_timer_type), allocatable :: timer(:)
    !> weights obtained for load balancing for each level
    real(kind=rk), allocatable :: level_weight(:)
    !> weights obtained for load balancing for each level boundary
    real(kind=rk), allocatable :: level_weight_bnd(:)
    !> weights obtained for load balancing for each level IBM 
    real(kind=rk), allocatable :: level_weight_IBM(:)
    !> Variable to capture the min time for communication
    real(kind=rk)              :: mincomm
    !> Variable to capture the min time for communication
    real(kind=rk)              :: maxcomm
    !> type of the control routine
    character(len=labelLen) :: controlRoutine
    !> initialize all elements with valid entries?
    !! This should only be activated for debugging, 
    !! as it needs to be ensured that all helper elements
    !! are filled by communication and interpolation instead
    !! of filling initial values (consider restart!)
    logical :: init_allElems = .false.
    !> Initialize both time layers nNext and nNow during the initialization
    !! phase? This is needed for boundary conditions requiring old time step
    !! information. For debugging multilevel, it should be turned off.
    logical :: init_allTimeLayers = .true.
    !> type of the scaling:
    !! * acoustic
    !! * diffusive
    character(len=labelLen) :: scaling
    !> Temporal scaling factor for the scaling. Acoustic = 2, Diffusive = 4
    integer :: scaleFactor
    !> Nesting: 1: acoustic, 3: diffusive
    integer :: nNesting = 1
    !> Required interval, in which the update MUST occur.
    !! This is required for the musubi multilevel, where the time step should
    !! only be determined active, when the end of the largest cycle is reached.
    integer :: reqInterval
    logical :: comm_reduced = .true. !< Communicate all links?
    !> need to set solver version in  general%solver%version
    character(len=labelLen) :: version = 'v1.1'
    !> active when restart is triggered by restart timeControl
    !! dump restart when simulation reached end only when
    !! restart is not triggered by its timeControl before
    logical :: restart_triggered = .false.
  end type mus_param_type

contains

!******************************************************************************!
  !> load global parameter from conf
  subroutine mus_load_param( params, conf )
    !---------------------------------------------------------------------------
    !> global parameter info
    type(mus_param_type), intent(inout)  :: params
    !> lua state
    type(flu_state) :: conf
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------
    write(logUnit(1),*) 'Loading general solver params: '
    ! Load the control routine
    call aot_get_val( L       = conf,                                          &
      &               key     = 'control_routine',                             &
      &               val     = params%controlRoutine,                         &
      &               ErrCode = iError,                                        &
      &               default = 'standard' )
    write(logUnit(1),*)'Choosing Control Routine: ', trim(params%controlRoutine)

    ! Decide, which scaling
    ! KM: \todo move scaling inside physics table because for lbm
    ! default scaling is acoustic 
    call aot_get_val( L       = conf,                                          &
      &               key     = 'scaling',                                     &
      &               val     = params%scaling,                                &
      &               ErrCode = iError,                                        &
      &               default = 'acoustic')
    write(logUnit(1),*) 'Setting SCALING to:',trim( params%scaling )

    ! Set acoustic or diffusive scaling parameters
    select case( trim( params%scaling ))
    case('acoustic')
      params%nNesting = 1
      params%scaleFactor = 2
    case('diffusive')
      params%nNesting = 3
      params%scaleFactor = 4
    end select  

    ! Communicate reduced set of links only? 
    call aot_get_val( L       = conf,                                          &
      &               key     = 'comm_reduced',                                &
      &               val     = params%comm_reduced,                           &
      &               ErrCode = iError,                                        &
      &               default = .true.)


    ! initialize all elements? this includes not only fluid but also ghost and
    ! halo
    call aot_get_val( L       = conf,                                          &
      &               key     = 'init_allElems',                               &
      &               val     = params%init_allElems,                          &
      &               ErrCode = iError,                                        &
      &               default = .false.)
    if( params%init_allElems ) then
      write(logUnit(1),*) 'WARNING: Initializing all (including helper '//     &
        &                           'elements) with valid data' 
      write(logUnit(1),*) '         Only activate this for debugging purposes!!  '
    end if

    ! initialize all time layers?
    call aot_get_val( L       = conf,                                          &
      &               key     = 'init_allTimeLayers',                          &
      &               val     = params%init_allTimeLayers,                     &
      &               ErrCode = iError,                                        &
      &               default = .true.)
    if( params%init_allTimeLayers ) then
      write(logUnit(1),*) 'NOTE:    Initializing all Time layers with '//      &
        &                          'valid data.'
    else
      write(logUnit(1),*) 'WARNING: Only initializing nNext time layer.'
      write(logUnit(1),*) '         Boundaries using old time layer info will '&
        &             //'not work and might cause a crash in the beginning' 
    end if

    call tem_horizontalSpacer(fUnit = logUnit(1))

  end subroutine mus_load_param
!******************************************************************************!

!******************************************************************************!
  !> Timers initialization routine for whatever 
  !!
  subroutine mus_init_timers( params )
    !---------------------------------------------------------------------------
    !> global parameter info
    type(mus_param_type), intent(inout)  :: params
    !---------------------------------------------------------------------------

    ! add timer handles to measure wall clock time of different routines
    call tem_addTimer( timerHandle = params%mainloopTimer,    &
      &                timerName   = 'MainLoop',              &
      &                me          = params%general%timerData )
    call tem_addTimer( timerHandle = params%advrelTimer,      &
      &                timerName   = 'Comp',                  &
      &                me          = params%general%timerData )
    call tem_addTimer( timerHandle = params%setBndTimer,      &
      &                timerName   = 'SetBnd',                &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%commTimer,        &
      &                timerName   = 'Comm',                  &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%loadMeshTimer,    &
      &                timerName   = 'LoadMesh',              &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%intpTimer_hgh,    &
      &                timerName   = 'Intp_hgh',              &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%intpTimer_low,    &
      &                timerName   = 'Intp_low',              &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%initLvlDTimer,    &
      &                timerName   = 'InitLvlD',              &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%wRestartTimer,    &
      &                timerName   = 'wRestart',              &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%sourceTimer,      &
      &                timerName   = 'source',                &
      &                me          = params%general%timerData )

    call tem_addTimer( timerHandle = params%isNanTimer,       &
      &                timerName   = 'isNan',                 &
      &                me          = params%general%timerData )

  end subroutine mus_init_timers
!******************************************************************************!
 
!******************************************************************************!
  !> Timers initialization routine for whatever 
  !!
  subroutine mus_init_leveltimers( params, minLevel, maxLevel )
    !---------------------------------------------------------------------------
    !> global parameter info
    type(mus_param_type), intent(inout)  :: params
    !> minimum level of the solver mesh
    integer, intent(in) :: minLevel
    !> maximum level of the solver mesh
    integer, intent(in) :: maxLevel
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    ! allocate levelwise timer and handles, this will be used in 
    ! dynamic load  balancing
    allocate (params%timer      (minLevel:maxLevel) )
    allocate (params%lw_tIntp   (minLevel:maxLevel) )
    allocate (params%lw_tCompute(minLevel:maxLevel) )
    allocate (params%lw_tSetBnd (minLevel:maxLevel) )
    allocate (params%lw_tIBM    (minLevel:maxLevel) )

    ! Add the main timer of dynamic load balancing
    call tem_addTimer( timerHandle = params%dyn_loadBalTimer,                &
      &                timerName   = 'postDLB',                              &
      &                me          = params%general%timerData )

    do iLevel = minLevel, maxLevel
    ! add a levelwise timer for load balancing
      call tem_addTimer( me          = params%timer(iLevel),                 &
        &                timerHandle = params%lw_tIntp(iLevel),              &
        &                timerName   = 'tIntp' )

      call tem_addTimer( me          = params%timer(iLevel),                 &
        &                timerHandle = params%lw_tCompute(iLevel),           &   
        &                timerName   = 'tCompute' )

      call tem_addTimer( me          = params%timer(iLevel),                 &
        &                timerHandle = params%lw_tSetBnd(iLevel),            &  
        &                timerName   = 'tSetBnd' )
    end do    


  end subroutine mus_init_leveltimers
!******************************************************************************!


end module mus_param_module
!******************************************************************************!

