! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!?? include 'header/lbm_macros.inc'
!> Definitions for the main state and neighbor arrays
!!
module mus_pdf_module

  ! include treelm modules
  use env_module,              only: rk, globalMaxLevels, long_k
  use tem_construction_module, only: tem_levelDesc_type
  use tem_comm_module,         only: tem_communication_type

  ! include musubi modules
  use mus_statistics_module,   only: mus_statistics_type
  use mus_physics_module,      only: set_values_by_levels

  implicit none
  private

  public :: pdf_data_type
  public :: pdf_global_type
  public :: pdf_globInfo_type
  public :: mus_globInfo_set_levels
  public :: mus_globInfo_set_dxdt

  !> This data type contains the Musubi-specific arrays which 
  !! are fed into the kernel. The solver updates the state vector
  !! and finds the position of the neighbor vectors by looking it 
  !! up in the neigh array
  type pdf_data_type
    ! Per level element numbers
    !> number of local fluid elements (inner+exposed)
    !! inner: elements which do not require communication
    !!        (neighbors are not halos)
    !! exposed: elements which require communication 
    !!        (neighbors are halos)
    integer :: nElems_fluid
    !> fluid elements + ghost elements (elements for solver)
    integer :: nElems_solve
    !> number of halo elements (from remote processes)
    integer :: nElems_halo
    !> number of ghost elements (from other levels)
    integer :: nElems_ghost
    !> number of ghost elements (from other levels)
    integer :: nElems_ghostFromCoarser
    !> number of ghost elements (from other levels)
    integer :: nElems_ghostFromFiner
    !> number of local elements (fluid+halos+ghost)
    integer :: nElems_local
    !> number of inner fluid elements 
    !! KM: Not used. Can be removed
    integer :: nElems_inner
    !> number of exposed fluid elements
    !! KM: Not used. Can be removed
    integer :: nElems_exposed

    !> which buffer to use for current time step
    integer :: nNow  = 1
    !> which buffer to use for next time step 
    integer :: nNext = 2

    !> Data vector containing the pdf state 
    !! Double buffering requires two arrays
    !! Size: scheme%varSys%nScalars * nElems_local, 2
    real(kind=rk),allocatable,dimension(:,:) :: state

    !> Additional values for each element
    !! Size: glob%nAuxVars, nElems_local
    real(kind=rk),allocatable,dimension(:,:) :: aux

    !> Connectivity array 
    !! Points to where to send respective pdfs 
    !! Access in a neigh way
    integer,allocatable,dimension(:) :: neigh

    !> containing state vector values of elements which have a boundary
    real(kind=rk),allocatable, dimension(:) :: bcBuffer

    !> Which halo requires all informations?
    logical, allocatable :: haloRequired(:)

  end type pdf_data_type

  !> Definition of a block containing the 1D vector and more information 
  type :: pdf_global_type
    !> number of total fluid elements (inner+exposed elements)
    integer(kind=long_k) :: nElems_totalFluid
    !> number of to-solve elements (inner+exposed+ghost elements)
    integer(kind=long_k) :: nElems_totalSolve
    !> number of total elements amoung all process in all levels 
    !! (inner+exposed+ghost+halos elements)
    !! i.e sum(nElems_allLevel)
    integer(kind=long_k) :: nElems_overall
    !> number of total fluid elements per level
    integer(kind=long_k) :: nElems_totalLevel( globalMaxLevels ) 
    !> number of total elements per level (fluids+ghosts+halos)
    integer(kind=long_k) :: nElems_allLevel( globalMaxLevels )
    !> number of elements which have to be solved (fluids+ghosts)
    integer(kind=long_k) :: nElems_solveLevel( globalMaxLevels ) 

    !> Local number of fluid elements
    integer :: nElems_localFluid
    !> Local number of elements to solve (fluids+ghosts)
    integer :: nElems_localSolve 

    real(kind=rk) :: totalDens !< overall integrated density

    !> index of the first fluid cell in the 
    !! local partition using global coordinates
    integer :: globalPosLocal

    !> Detailed information about the level-wise treeID lists
    type(tem_levelDesc_type),allocatable  :: levelDesc(:)

  end type pdf_global_type

  !> This datatype contains all scheme independent informations
  !! \todo KM: move general param to mus_param_type
  type :: pdf_globInfo_type
    !> number of dimensions for current run
    !! standard: 3, but if a 2d stencil was chosen, set to 2
    integer :: nDims = 3
    !> total number of used schemes
    integer :: nSchemes = 0
    !> lattice unit time steps
    real(kind=rk) :: dtLB(globalMaxLevels) = -1._rk
    !> lattice unit spatial steps
    real(kind=rk) :: dxLB(globalMaxLevels) = -1._rk
    !> number of levels
    integer :: nLevels
    !> global minimal level
    integer :: minLevel
    !> global maximal level
    integer :: maxLevel
    !> is this level being updated at the current time step?
    !! Not all levels get updated with the nested time stepping
    !! todo: is this used somewhere?
    logical :: levelActive( globalMaxLevels ) = .false.
    !> number of boundary conditions
    integer :: nBCs
    !> number of local boundary conditions
    integer :: nBCs_local
    !> pointer from the original treeIDlist to the levelwise fluid list
    !! it has a size of tree%nElems
    integer, allocatable :: levelPointer(:)
    type( mus_statistics_type ) :: stat
  end type pdf_globInfo_type

contains

  subroutine mus_globInfo_set_levels( me, minLevel, maxLevel )
    type( pdf_globInfo_type ) :: me
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel

    me%minLevel = minLevel
    me%maxLevel = maxLevel
    me%nLevels  = maxLevel - minLevel + 1
  end subroutine mus_globInfo_set_levels

  subroutine mus_globInfo_set_dxdt( me, scaleFacLvl )
    type( pdf_globInfo_type ) :: me
    integer, intent(in)  :: scaleFacLvl !< dt scale factor between levels

    me%dxLB(me%minLevel:me%maxLevel) = &
      &                     set_values_by_levels( valMinLevel = 1.0_rk,      &
      &                                           minLevel    = me%minLevel, &
      &                                           maxLevel    = me%maxLevel, &
      &                                           scaleFac    = 2            )

    me%dtLB(me%minLevel:me%maxLevel) = &
      &                     set_values_by_levels( valMinLevel = 1.0_rk,      &
      &                                           minLevel    = me%minLevel, &
      &                                           maxLevel    = me%maxLevel, &
      &                                           scaleFac    = scaleFacLvl  )

  end subroutine mus_globInfo_set_dxdt

end module mus_pdf_module
!******************************************************************************!


!******************************************************************************!
!>\page datastructures Data structures
!! The Octree data structure is mapped to a one-dimensional array in order to
!! have an efficient data structure on which the solver can act on in a 
!! performant way. An efficient representation of the elements and their 
!! neighbor relations is chosen.
!! The fluid elements are mapped to a one-dimensional array and the neighboring 
!! relations are introduced by an additional connectivity array. The access of a
!! neighbor element is performed by looking up the correct position of an 
!! element's link neighbor in the connectivity array, thus constituting an
!! indirect access.
!!
!! \image html  statevector_neighborlist.png
!!
!! The different dependencies of each link in an element require a thorough
!! treatment, when data is exchanged at domain boundaries. Only the links, which
!! point outside the domain have to be sent to neighbor partitions, and the 
!! links pointing inwards have to be filled with valid values from these.
