! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains data types and routines used by musubi main program
!! i.e unifying routines for apesmate
!!
!! \author Kannan Masilamani
module mus_program_module

  ! include treelm modules
  use mpi
  use env_module,             only: rk, long_k
  use tem_timer_module,       only: tem_startTimer, tem_stopTimer,             &
    &                               tem_resetTimer
  use tem_status_module,      only: tem_status_run_end, tem_status_dump,       &
    &                               tem_status_run_terminate
  use tem_simControl_module,  only: tem_simControl_clearStat
  use tem_tracking_module,    only: tem_tracking_finalize
  use tem_restart_module,     only: tem_restart_finalize
  use tem_balance_module,     only: tem_balance_evaluateImbalance
  use tem_logging_module,     only: logUnit, tem_logging_init, tem_logging_type
  use tem_debug_module,       only: tem_reportStatus, tem_debug_type
  use tem_timeControl_module, only: tem_timeControl_globalTriggered,           &
    &                               tem_timeControl_check
  use tem_time_module,        only: tem_time_dump, tem_time_gt_trigger

  ! include musubi modules
  ! RESPECT THE ORDER !!!!!!!!
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_param_module,              only: mus_param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_tools_module,              only: check_totalDensity, dump_timing
  use mus_interpolate_header_module, only: interpolation_type
  use mus_flow_module,               only: mus_init_flow
  use mus_dynLoadBal_module,         only: mus_perform_dynLoadBal,             &
    &                                      dump_slotMLUPS
  use mus_time_module,               only: mus_timeControl_homogenize
  ! ESPECIALLY OF THE FOLLOWING MODULES
  use mus_construction_module,       only: mus_construct
  use mus_control_module,            only: mus_control_type, mus_init_control
  use mus_aux_module,                only: mus_init_aux
  use mus_restart_module,            only: mus_writeRestart

  ! include aotus nmodules
  use aotus_module, only: close_config

  implicit none

  private

  public :: mus_initialize
  public :: mus_solve
  public :: mus_finalize


contains

!******************************************************************************!
  !> This routine load musubi configuration file and initialize construction
  !! flow, auxilary and main control routines
  subroutine mus_initialize( scheme, geometry, params, pdfGlobal, intp,        &
    &                        control )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), intent(inout), allocatable :: scheme(:)
    !> Treelmesh data
    type( mus_geom_type ),      intent(inout) :: geometry
    !> Global parameters
    type( mus_param_type ),     intent(inout) :: params
    !> global information
    type( pdf_globInfo_type ),  intent(inout) :: pdfGlobal
    !> interpolation method
    type( interpolation_type ), intent(inout) :: intp
    !> control routine
    type( mus_control_type ),   intent(out)   :: control
    !-------------------------------------------------------------------------!

    ! construct levelDescriptor, connectivity array and boundary elements
    call mus_construct( scheme    = scheme,     &
      &                 geometry  = geometry,   &
      &                 params    = params,     &
      &                 pdfGlobal = pdfGlobal,  &
      &                 intp      = intp        )

    ! initialize state vector depends on scheme type
    call mus_init_flow( scheme    = scheme,    &
      &                 geometry  = geometry,  &
      &                 params    = params,    &
      &                 pdfGlobal = pdfGlobal, &
      &                 intp      = intp       )

    ! Init auxiliary features such as interpolation, boundaries, restart
    ! and the tracker
    call mus_init_aux( scheme    = scheme,    &
      &                geometry  = geometry,  &
      &                params    = params,    &
      &                pdfGlobal = pdfGlobal, &
      &                intp      = intp       )

    ! choose main control routine function pointer
    call mus_init_control( params = params, me = control )

  end subroutine mus_initialize
!******************************************************************************!


!******************************************************************************!
  !> This routine does the main musubi computation loop
  subroutine mus_solve( scheme, geometry, params, pdfGlobal, intp, control )
    !--------------------------------------------------------------------------!
    !> scheme type
    type( mus_scheme_type ), allocatable, intent(inout) :: scheme(:)
    !> Treelmesh data
    type( mus_geom_type ),      intent(inout) :: geometry
    !> Global parameters
    type( mus_param_type ),     intent(inout) :: params
    !> global information
    type( pdf_globInfo_type ),  intent(inout) :: pdfGlobal
    !> interpolation method
    type( interpolation_type ), intent(inout) :: intp
    !> control routine
    type( mus_control_type ),   intent(in)    :: control
    !--------------------------------------------------------------------------!
    integer :: iErr
    logical :: balance_triggered
    type(tem_debug_type) :: debug
    !--------------------------------------------------------------------------!
?? if ( SCOREP_USER_ENABLE ) then
   integer, parameter :: scorep_user_invalid_region = -1
   integer, parameter :: scorep_user_region_type_common = 0
   integer, parameter :: scorep_user_region_type_dynamic = 4
   !SCOREP_USER_REGION_DEFINE(computation)
   integer(kind=long_k) :: phase = scorep_user_invalid_region
   !SCOREP_USER_REGION_DEFINE(computation)
   integer(kind=long_k) :: computation = scorep_user_invalid_region 
   !SCOREP_USER_REGION_DEFINE(loadbalancing)
   integer(kind=long_k) :: loadbalancing = scorep_user_invalid_region
?? endif
  !-----------------------------------------------------------------------------

    ! Main time loop
    !
    call mpi_barrier( params%general%proc%comm, ierr )
    call tem_startTimer( timerHandle = params%mainloopTimer,    &
      &                  me          = params%general%timerData )
    !Start a timer if dynamic load balancing is active
    if ( params%general%balance%dynamic ) then
      call tem_startTimer( timerHandle = params%dyn_loadBalTimer,              &
        &                  me          = params%general%timerData )
      call tem_logging_init( me    = debug%logger,                             &
        &                    level = 10,                                       &
        &                    rank  = 0,                                        &
        &                    filename = 'DLB_MEM.res' )
      call tem_startTimer( timerHandle = params%general%balance%slotTimer,     &
        &                  me          = params%general%balance%timings )
    end if

?? if( SCOREP_USER_ENABLE ) then
   ! SCOREP_USER_REGION_BEGIN(phase, "LB phase",SCOREP_USER_REGION_TYPE_DYNAMIC)
   call SCOREP_F_Begin(phase,"LB phase",scorep_user_region_type_dynamic,"",?line?)
?? endif

    mainloop: do

      ! clear status bit
      call tem_simControl_clearStat( me = params%general%simControl )

?? if( SCOREP_USER_ENABLE )then
    !SCOREP_USER_REGION_BEGIN(computation, "do_computation",SCOREP_USER_REGION_TYPE_COMMON)
    call SCOREP_F_Begin(computation,"do_computation",scorep_user_region_type_common,"",?line?)
?? endif

      ! Iterate through the elements in a level-wise fashion
      ! and advance the time steps
      call control%do_computation( scheme    = scheme,                         &
        &                          geometry  = geometry,                       &
        &                          params    = params,                         &
        &                          intp      = intp,                           &
        &                          iLevel    = pdfGlobal%minLevel,             &
        &                          pdfGlobal = pdfGlobal )

?? if( SCOREP_USER_ENABLE )then
    !SCOREP_USER_REGION_END(computation)
    call SCOREP_F_RegionEnd(computation)
?? endif

      ! Perform the dynamic load balancing if interval has been reached
      if( params%general%balance%dynamic ) then
        call tem_timeControl_check( me   = params%general%balance%timeControl, &
          &                         now  = params%general%simControl%now,      &
          &                         comm = params%general%proc%comm,           &
          &                         triggered = balance_triggered )

        if( balance_triggered ) then
          ! dump the mlups of the current slot to disc
          call dump_slotMLUPS( scheme    = scheme,                             &
            &                  pdfGlobal = pdfGlobal,                          &
            &                  params    = params,                             &
            &                  last      = .false. )

?? if( SCOREP_USER_ENABLE ) then
    !SCOREP_USER_REGION_END(phase)
    call SCOREP_F_RegionEnd(phase)
    !SCOREP_USER_REGION_BEGIN(phase, "LB phase",SCOREP_USER_REGION_TYPE_DYNAMIC)
    call SCOREP_F_Begin(phase,"LB phase",scorep_user_region_type_dynamic,"",?line?)
    !SCOREP_USER_REGION_BEGIN(loadbalancing, "loadbalancing",SCOREP_USER_REGION_TYPE_DYNAMIC)
    call SCOREP_F_Begin(loadbalancing,"loadbalancing",scorep_user_region_type_common,"",?line?)
?? endif

          write(logUnit(1),*)'Performing dynamic load balancing ...'
          call mus_perform_dynLoadBal( scheme    = scheme,                     &
            &                          pdfGlobal = pdfGlobal,                  &
            &                          params    = params,                     &
            &                          geometry  = geometry,                   &
            &                          intp      = intp )

          call tem_reportStatus( debug = debug, level = 1, &
                                 text = 'VmRSS', string = 'VmRSS')
          call tem_reportStatus( debug = debug, level = 1, &
                                 text = 'VmHWM', string = 'VmHWM')

          ! Reset the loadBal timer such that it contains only post DLB time
          call tem_resetTimer( timerHandle  = params%dyn_loadBalTimer,         &
            &                   me          = params%general%timerData )

          ! restart the slotTimer for dynLB
          call tem_startTimer( timerHandle = params%general%balance%slotTimer, &
            &                  me          = params%general%balance%timings )

          write(logUnit(1),*)'Done with dynamic load balancing.'

?? if( SCOREP_USER_ENABLE ) then
    !SCOREP_USER_REGION_END(loadbalancing)
    call SCOREP_F_RegionEnd(loadbalancing)
?? endif

        end if ! balance triggered
      end if !dynamic

      if( tem_status_run_end(params%general%simControl%status) .or.            &
        & tem_status_run_terminate(params%general%simControl%status) ) then
        exit mainLoop
      end if

    enddo mainloop
    !
    ! Finish main loop
    !-----------------------------------------------------------------------------
?? if( SCOREP_USER_ENABLE ) then
    !SCOREP_USER_REGION_END(phase)
    call SCOREP_F_RegionEnd(phase)
?? endif

    call tem_stopTimer( timerHandle = params%mainloopTimer,    &
      &                 me          = params%general%timerData )
    ! Stop the load balancing timer
    if( params%general%balance%dynamic ) then 
      call tem_stopTimer( timerHandle  = params%dyn_loadBalTimer,              &
        &                  me          = params%general%timerData )
      ! dump the mlups of the last slot to disc
      call dump_slotMLUPS( scheme    = scheme,                                 &
        &                  pdfGlobal = pdfGlobal,                              &
        &                  params    = params,                                 &
        &                  last      = .true. )
    end if

  end subroutine mus_solve
!******************************************************************************!


!******************************************************************************!
  !> Do final check on check on total density,
  !! Close auxiliary stuff such as restart and the tracker,
  !! finalize treelm, dump timing and finialize mpi with fin_env
  !!
  subroutine mus_finalize( scheme, params, geometry, pdfGlobal )
    !--------------------------------------------------------------------------!
    !> scheme type
    type( mus_scheme_type ), intent(inout) :: scheme(:)
    !> Global parameters
    type( mus_param_type ),intent(inout) :: params
    !> geometry infomation
    type(mus_geom_type),intent(in) :: geometry
    !> global information
    type( pdf_globInfo_type ), intent(in) :: pdfGlobal
    !--------------------------------------------------------------------------!
    integer           :: iScheme
    real(kind=rk)     :: mainloopTime, commTime, cpuCost
    !--------------------------------------------------------------------------!

    ! dump status information on how simulation was terminated
    call tem_status_dump( me = params%general%simControl%status,               &
      &                   outUnit = logUnit(1) )

    write(logUnit(1),*)'Finished main loop with sim:',                         &
      &                             real(params%general%simControl%now%sim)
    write(logUnit(1),*)'                       iter:',                         &
      &             params%general%simControl%now%iter

    ! check final total density
    call check_totalDensity( scheme, pdfGlobal, params%general )

    ! write restart file when it is not triggered by restart timecontrol
    ! before exiting mainloop.
    ! restart will be dumped even if timeControl table is not defined
    ! for restart and only writeRestart is active
    if (params%general%restart%controller%writeRestart .and. &
      & .not. params%restart_triggered) then
      write(logUnit(1),*) 'Writing restart at final time:'
      call tem_time_dump( params%general%simControl%now, logUnit(1) )
      call mus_writeRestart( pdfGlobal = pdfGlobal,                          &
        &                    restart   = params%general%restart,             &
        &                    scheme    = scheme,                             &
        &                    tree      = geometry%tree,                      &
        &                    timing    = params%general%simControl%now,      &
        &                    timerData = params%general%timerData,           &
        &                    timerHandle = params%wRestartTimer              )
      write(logUnit(1),*) 'Done writing'
    end if

    ! close musubi main configuration file
    call close_config( L = params%general%solver%conf(1) )

    if( params%general%restart%controller%writeRestart .or.                    &
      & params%general%restart%controller%readRestart ) then
      call tem_restart_finalize( params%general%restart )
    end if

    ! Evaluate the load imbalance
    mainloopTime = 0.0_rk
    commTime     = 0.0_rk
    cpuCost      = 0.0_rk
    if( params%general%balance%static .or. params%general%balance%dynamic) then
      mainloopTime = params%general%timerData%value(params%mainloopTimer)
      commTime     = params%general%timerData%value(params%commTimer)
      ! The actual computational cost is the total mainloop time
      ! minus the time it spent inside the communication routines
      cpuCost = mainLoopTime - commTime
      call tem_balance_evaluateImbalance(                                      &
        &                          me      = params%general%balance,           &
        &                          CPUcost = cpuCost,                          &
        &                          nParts  = params%general%proc%comm_size,    &
        &                          comm    = params%general%proc%comm )
    endif

    do iScheme = 1, pdfGlobal%nSchemes
      call tem_tracking_finalize( scheme( iScheme )%track )
      ! close the solver specific string
      call close_config( L = scheme(iScheme)%varSys%solSpec%conf )
    end do

    ! dump timing
    call dump_timing( scheme, params, pdfGlobal)
  end subroutine mus_finalize
!******************************************************************************!


end module mus_program_module
!******************************************************************************!
