! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!>  Module containing subroutines for building MUSUBI specific source 
!!  variable systems, initialize and update the source terms.
!! 
!! \author Simon Zimny 
!!
!! The source type and loading routine are described in the [tem_source].
!!
!! [tem_source]: @ref tem_source_module "tem_source_module"
!!
?? include 'header/lbm_macros.inc'

module mus_source_module

  ! include treelm modules
  use env_module,               only: rk, long_k, labelLen
  use tem_aux_module,           only: tem_abort
  use tem_var_system_module,    only: tem_var_system_type, tem_var_append,     &
    &                                 tem_var_sys_debug, tem_var_sys_mark,     &
    &                                 tem_var_sys_init
  use tem_variable_module,      only: tem_variable_type
  use tem_stencil_module,       only: tem_stencilHeader_type
  use tem_source_module,        only: tem_source_type, tem_sourceMap_type
  use tem_time_module,          only: tem_time_type
  use tem_spacetime_fun_module, only: tem_spacetime_for
  use tem_topology_module,      only: tem_LevelOf
  use treelmesh_module,         only: treelmesh_type
  use tem_derived_module,       only: evaluateSources_single
  use tem_var_module,           only: tem_var_type
  use tem_subTree_type_module,  only: tem_subTree_type, tem_treeIDfrom_subTree
  use tem_logging_module,       only: logUnit

  ! include musubi modules
  use mus_pdf_module,                only: pdf_data_type
  use mus_derivedQuantities_module2, only: getVelocityPre

  implicit none
  private

  public :: mus_build_sourceVarSys
  public :: mus_init_source
  public :: mus_update_source

  type( tem_var_type ) :: equalInjection_sourceVar
  type( tem_var_type ) :: injection_sourceVar

  contains


!******************************************************************************!
  !> Build a variable system of all possible source terms for the given 
  !! schemeKind
  !!
  subroutine mus_build_sourceVarSys( me, sourceVarList, stencil, schemeKind )
    !--------------------------------------------------------------------------- 
    !> array of all possible derived variables for this scheme kind 
    !! (without prefixes)
    type(tem_source_type), intent(inout) :: me(:)
    !> list of all possible source variables for a given scheme kind
    type(tem_var_type), allocatable :: sourceVarList(:) 
    type(tem_stencilHeader_type), intent(in) :: stencil !< stencil informations
    !> scheme kind to decide how to fill the sourceVarList
    character(len=*) :: schemeKind
    !--------------------------------------------------------------------------- 
    integer :: pos
    integer :: iField, iVar ! counter variables
    !--------------------------------------------------------------------------- 

    ! initialise the derived quantities
    ! the variables are stored in their 'pure' form without prefixes

    ! injection (updated with local velocity)
    injection_sourceVar%variable%label = 'injection'
    injection_sourceVar%variable%nComponents = stencil%QQ
    injection_sourceVar%do => evaluateSources_single

    ! equaly distributed injection
    equalInjection_sourceVar%variable%label = 'equalInjection'
    equalInjection_sourceVar%variable%nComponents = stencil%QQ
    equalInjection_sourceVar%do => evaluateSources_single

    write(logUnit(1),*)'Building the source variable list ...'


    ! build the right variable list depending on the scheme kind
    select case( trim( schemeKind )) 
      case('lbm', 'lbm_incomp')
      
        allocate(sourceVarList(2))
        sourceVarList(1) = injection_sourceVar
        sourceVarList(2) = equalInjection_sourceVar
    
      case default
        write(logUnit(1),*) 'The selected scheme kind is unknown '//           &
          &             trim(schemeKind) 
        call tem_abort()
    end select 

    do iField = 1, size( me )
      ! initialise the variable system of all possible variables
      call tem_var_sys_init( me(iField)%poss_varSys )

      ! build a variable system out of the sourceVarList
      do iVar = 1, size(sourceVarList)
        ! append the source variables for this type
        call tem_var_append(                                                   &
          &            sys         = me(iField)%poss_varSys,                   &
          &            varName     = sourceVarList(iVar)%variable%label,       &
          &            nComponents = sourceVarList(iVar)%variable%nComponents, &
          &            pos         = pos )
      end do
     
      call tem_var_sys_mark( sys        = me(iField)%poss_varSys,              &
        &                    SystemName = 'possSources' )

      ! debug output
      call tem_var_sys_debug( sys   = me(iField)%poss_varSys,                  &
        &                     level = 3 )

    end do  ! iField

  end subroutine mus_build_sourceVarSys
!******************************************************************************!


!******************************************************************************!
  !> Initialize all source terms for the fields and the one for the scheme
  !!
  subroutine mus_init_source( me, global_varSys, sourceMap, prefix )
    !---------------------------------------------------------------------------
    !> array of all possible derived variables for this scheme kind 
    !! (without prefixes)
    type(tem_source_type), intent(in) :: me(:)
    !> global variable system
    type(tem_var_system_type), intent(inout) :: global_varSys
    !> source map to store the source type and the position of this source
    !! variable in the global system
    type(tem_sourceMap_type), intent(inout) :: sourceMap
    !> optional argument containing a prefix for the source variables
    character(len=*), intent(in), optional :: prefix(:)
    !--------------------------------------------------------------------------- 
    integer :: pos
    integer :: iVar, iSource ! counter variables
    character(len=labelLen), allocatable :: local_prefix(:)
    !--------------------------------------------------------------------------- 

    ! allocate the local_prefix with the number of source types
    ! (= number of prefixes)
    allocate(local_prefix( size( me )))

    if(present(prefix))then
      local_prefix = prefix
    else
      local_prefix = ''
    end if

    ! append all intersected variables to the global varSys, this assures that
    ! the source terms can be tracked when appending them to the global varSys
    do iSource = 1, size(me)
      do iVar = 1, me(iSource)%varSys%nVars
        call tem_var_append(                                                   &
          &  sys         = global_varSys,                                      &
          &  varName     = trim(local_prefix(iSource))//                       &
          &                       me(iSource)%varSys%variable%val(iVar)%label, &
          &  nComponents = me(iSource)%varSys%variable%val(iVar)%nComponents,  &
          &  pos         = pos )
        sourceMap%map( pos )    =  iVar
        sourceMap%source( pos ) = me(iSource)
      end do  ! iVar
    end do  ! iSource

  end subroutine mus_init_source
!******************************************************************************!


!******************************************************************************!
  !> Update all source terms for the fields and the one for the scheme for the
  !! NEXT timestep t+1.
  !!
  subroutine mus_update_source( me, pdf, depPdf, time, global_varSys,          &
    &                           global_tree, stencil )
    !---------------------------------------------------------------------------
    !> source variable information (incl. st-function)
    type(tem_source_type), intent(in) :: me(:)
    !> pdf datatype containing the state vector to update
    type(pdf_data_type), intent(inout) :: pdf(:)
    !> state vector for a dependent scheme
    type(pdf_data_type), intent(inout), optional :: depPdf(:)
    !> current timing information
    type(tem_time_type), intent(in) :: time
    !> global variable system
    type(tem_var_system_type), intent(in) :: global_varSys
    !> global tree information
    type(treelmesh_type), intent(in) :: global_tree
    !> stencil information
    type(tem_stencilHeader_type), intent(in) :: stencil
    !--------------------------------------------------------------------------- 
    ! counter variables
    integer :: iSource, iSourceVar, iFun
    !--------------------------------------------------------------------------- 

    ! loop over all sources
    do iSource = 1, size(me)
      ! loop over all sourceVars
      do iSourceVar = 1, size(me(iSource)%sourceVar)
        ! loop over all spacetime functions
        do iFun = 1, size( me(iSource)%sourceVar(iSourceVar)%st_fun )
            call mus_updateSourceElems(                                        &
              &     me            = me(iSource)%sourceVar(iSourceVar),         &
              &     glob_tree     = global_tree,                               &
              &     subTree       = me( iSource )%sourceVar(iSourceVar)%       &
              &                                          st_fun(iFun)%subTree, &
              &     pdf           = pdf,                                       &
              &     depPdf        = depPdf,                                    &
              &     time          = time,                                      &
              &     global_varSys = global_varSys,                             &
              &     stencil       = stencil,                                   &
              &     iFun          = iFun,                                      &
              &     field_source  = (iSource < size(me)),                      &
              &     varPos        = global_varSys%Variable%val(iSource)%varPos )
        end do  ! iFun
      end do  ! iSourceVar
    end do  ! iSource

  end subroutine mus_update_source
!******************************************************************************!


!******************************************************************************!
  !> Initialize all source terms for the fields and the one for the scheme
  !!
  subroutine mus_updateSourceElems( me, glob_tree, subTree, pdf, depPdf, time, &
    &                               global_varSys, stencil, iFun, field_source,&
    &                               varPos )
    !---------------------------------------------------------------------------
    !> variable information (incl. st-function)
    type(tem_variable_type), intent(in) :: me
    !> global tree information
    type(treelmesh_type)  :: glob_tree
    !> subTree information
    type(tem_subTree_type)  :: subTree
    !> pdf datatype containing the state vector to update
    type(pdf_data_type), intent(inout) :: pdf(:)
    !> state vector for a dependent scheme
    type(pdf_data_type), intent(inout), optional :: depPdf(:)
    !> current timing information
    type(tem_time_type), intent(in) :: time
    !> global variable system
    type(tem_var_system_type), intent(in) :: global_varSys
    !> stencil information
    type(tem_stencilHeader_type), intent(in) :: stencil
    !> current lua function
    integer, intent(in) :: iFun
    !> work on single field??
    logical, intent(in) :: field_source
    !> position of variable in global system, if not provided work on all 
    !! fields
    integer, intent(in) :: varPos( stencil%QQ )
    !--------------------------------------------------------------------------- 
    ! counter variables
    integer :: iElem, iDir, iVar, iPos, iLevel
    ! array for the lua function results for a set of elements
    real(kind=rk), allocatable :: luaRes(:)
    ! local velocity
    real(kind=rk) :: local_vel(3)
    ! linkwise lua results for a single element
    real(kind=rk) :: linkwise_luaRes( stencil%QQ )
    ! position of the directions of a variable in the global system
    integer :: tmp_varPos( stencil%QQ )
    ! number of scalar quantities in global variable system
    integer :: nScalars
    ! list of treeIDs to work on
    integer(kind=long_k), allocatable :: treeID(:)
    ! lower and upper bound for subTree array
    integer :: bound(2)
    !--------------------------------------------------------------------------- 
    ! set the bound array to cover the complete subTree
    bound(1) = 1
    bound(2) = subTree%nElems

    ! get the number of scalars in the global variable system
    nScalars = global_varSys%nScalars

    call tem_treeIDfrom_subTree( subTree, glob_tree, treeID, bound )

    ! allocate a temporary array for the lua result for every element 
    ! in this sourceVar
    allocate(luaRes( glob_tree%nElems ))
    luaRes = 0._rk
    ! evaluate the lua function
    luaRes = tem_spacetime_for( me      = me%st_fun(iFun),                     &
      &                         treeIDs = treeID,                              &
      &                         time    = time,                                &
      &                         tree    = glob_tree,                           &
      &                         n       = subTree%nElems )
  
    ! now update the state vector for all elements
    ! the source types are stored like follows:
    ! -------------------------------------------------------
    ! | source_field1 | ... | source_fieldN | source_scheme |
    ! -------------------------------------------------------
    do iElem = 1, subTree%nElems
      linkwise_luaRes = 0._rk
      iPos = subTree%map2global( iElem )
      ! get the level of the treeID
      iLevel = tem_LevelOf( treeID( iElem ))
      
      ! the source terms are only applied on the individual fields
      if( field_source ) then
        ! save the varPos to the local varPos
        tmp_varPos = varpos
  
        ! \todo: SZ: for now work with a select case, function pointers
        !            might be useful later on
  
        ! modify the lua res according to the type of injection
        select case( trim(me%label ))
        
        case('equalInjection')
          ! compute the linkwise equal distribution
          do iDir = 1, stencil%QQ
            linkwise_luaRes(iDir) = luaRes(iElem) / real(stencil%QQ, kind=rk)
          end do
        case('injection')
          local_vel = 0._rk                
          ! get the local velocity for this element from the dependend
          ! scheme (fluid scheme)
          local_vel =  getVelocityPre(                                         &
            &     state    = depPdf(iLevel)%state( :, pdf(iLevel)%nNext),      &
            &     elem     = iPos,                                             &
            &     stencil  = stencil,                                          &
            &     varPos   = tmp_varPos,                                       &
            &     nScalars = nScalars,                                         &
            &     neigh    = pdf(iLevel)%neigh)
         
          do iDir = 1, stencil%QQ
            linkwise_luaRes(iDir) =                                            &
              & luaRes(iElem) /  6._rk * (1._rk + 3._rk * (                    &
              & real(stencil%cxDir(1,iDir), kind =rk) * local_vel(1) +         &
              & real(stencil%cxDir(2,iDir), kind =rk) * local_vel(2) +         &
              & real(stencil%cxDir(3,iDir), kind =rk) * local_vel(3)))
          end do
        end select
  
        ! loop over all directions
        do iDir = 1, stencil%QQ
        ! add the linkwise lua res for the current element 
 pdf(iLevel)%state( ?FETCH?( tmp_varPos(iDir), iPos, nScalars, pdf(iLevel)%neigh ), pdf(iLevel)%nNext) =  &
  pdf(iLevel)%state( ?FETCH?( tmp_varPos(iDir), iPos, nScalars, pdf(iLevel)%neigh ), pdf(iLevel)%nNext) + &
          &  linkwise_luaRes(iDir) 
        end do
        ! the source terms are applied on the scheme (= on all state
        ! pdfs), only the last source
        else
          ! loop over all state variables
          do iVar = 1, global_varSys%nVars
            ! get the position of the variable in the global system
            tmp_varPos = global_varSys%Variable%val( iVar )%varPos
  
            ! \todo: SZ: for now work with a select case, function
            !            pointers might be useful later on
           
            ! modify the lua res according to the type of injection
            select case( trim(me%label ))
            
            case('equalInjection')
              ! compute the linkwise equal distribution
              do iDir = 1, stencil%QQ
                linkwise_luaRes(iDir) = luaRes(iElem) / real(stencil%QQ,kind=rk)
              end do
           
            case('injection')
              ! \todo: SZ: one might also work with pdf if depPdf is not
              !            provided!! Maybe use pointer
              if(present(depPdf))then
                ! get the local velocitry for this element
                local_vel =  getVelocityPre(                                   &
                  &    state    = depPdf(iLevel)%state( :, pdf(iLevel)%nNext), &
                  &    elem     = iPos,                                        &
                  &    stencil  = stencil,                                     &
                  &    varPos   = tmp_varPos,                                  &
                  &    nScalars = nScalars,                                    &
                  &    neigh    = pdf(iLevel)%neigh)
             
                do iDir = 1, stencil%QQ
                  linkwise_luaRes(iDir) =                                      &
                  &luaRes(iElem) / 6._rk * (1._rk + 3._rk * (                  &
                  &real(stencil%cxDir(1,iDir), kind=rk) * local_vel(1) +       &
                  &real(stencil%cxDir(2,iDir), kind=rk) * local_vel(2) +       &
                  &real(stencil%cxDir(3,iDir), kind=rk) * local_vel(3) ))
                end do
              else
                write(logUnit(1),*) 'No dependend scheme availablefor'//       &
                  &             'velocity calculation available!!' 
              end if
            end select
            ! loop over all directions
            do iDir = 1, stencil%QQ
  
            ! add the linkwise lua res to the current distribution
  pdf(iLevel)%state( ?FETCH?( tmp_varPos(iDir), iPos, nScalars, pdf(iLevel)%neigh ), pdf(iLevel)%nNext) = &
  pdf(iLevel)%state( ?FETCH?( tmp_varPos(iDir), iPos, nScalars, pdf(iLevel)%neigh ), pdf(iLevel)%nNext) + &
             &  linkwise_luaRes(iDir)
  
             end do  ! iDir
           end do  ! iVar
         end if
       end do  ! iElem
       ! deallocate the lua result array
       if( allocated(luaRes) ) deallocate( luaRes )

  end subroutine mus_updateSourceElems
!******************************************************************************!


end module mus_source_module
!******************************************************************************!
