! See copyright notice in the COPYRIGHT file.
!!******************************************************************************!
!> Some generic handy check routines to check the properties of the flow field 
!! and the current run
!! 
!! Performance estimation, check of the total density, check for NaNs
!!\author Manuel Hasert
?? include 'header/lbm_macros.inc'
module mus_tools_module

  ! include treelm modules
  use mpi
!$ use omp_lib
!  use, intrinsic :: ieee_arithmetic
!  use, intrinsic :: ieee_exceptions
  use env_module,             only: rk, long_k, newunit, PathLen
  use tem_time_module,        only: tem_time_dump
  use tem_timeControl_module, only: tem_timeControl_globalTriggered
  use tem_timer_module,       only: tem_startTimer, tem_stopTimer
  use tem_status_module,      only: tem_stat_nan_detected, tem_stat_nonPhysical
  use tem_isNaN_module,       only: tem_isNaN
  use tem_debug_module,       only: main_debug
  use tem_logging_module,     only: logUnit
  use tem_comm_env_module,    only: tem_comm_env_type
  use tem_general_module,     only: tem_general_type

  ! include musubi modules
  use mus_pdf_module,         only: pdf_globInfo_type
  use mus_param_module,       only: mus_param_type
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_time_module,        only: mus_time_modulo

  implicit none

  private

  public :: perform_checks
  public :: check_totalDensity
  public :: dump_timing
  public :: check_streaming_layout
  public :: mus_min_max_commTime
  public :: calc_MLUPS

  interface check_totalDensity
    module procedure check_scheme_density
    module procedure check_all_density
  end interface

contains

!******************************************************************************!
  !> Perform run-time checks if interval is active
  !!
  subroutine perform_checks( scheme, pdfGlobal, params)
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: scheme
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    type(mus_param_type), intent(inout) :: params
    !---------------------------------------------------------------------------
    ! Check for NaNs. This check might be expensive so perform only for debug
    if ( main_debug%checkNans ) then
      call tem_startTimer( timerHandle = params%isNanTimer, &
        &                  me          = params%general%timerData )
      call check_NaNs( scheme, pdfGlobal,                                      &
        &             params%general%simControl%status%bits(                   &
        &                                      tem_stat_nan_Detected),         &
        &             params%general%proc )
      call tem_stopTimer( timerHandle = params%isNanTimer,       &
        &                 me          = params%general%timerData )
      if( params%general%simControl%status%bits( tem_stat_nan_detected ) ) then
        write(logUnit(1),*) '   Found NaN at time = ',                         &
          &               params%general%simControl%now%sim,                   &
          &               '       iTime = ', params%general%simControl%now%iter
      end if
    end if

    !  Check and output in the main intervals
    if ( tem_timeControl_globalTriggered(                                      &
      &                            me = params%general%simControl%timeControl, &
      &                            now = params%general%simControl%now,        &
      &                            comm = params%general%proc%comm )           &
      &)then

      call check_totalDensity( scheme, pdfGlobal, params%general )
    end if

  end subroutine perform_checks
!******************************************************************************!



!******************************************************************************!
  !> Check the state array for nans
  !!
  subroutine check_NaNs( scheme, pdfGlobal, nan_Detected, proc )
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(in) :: scheme
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    logical, intent(inout)  :: nan_Detected
    !> mpi communication enviroment with mpi communicator
    type(tem_comm_env_type), intent(in) :: proc
    !---------------------------------------------------------------------------
    real(kind=rk)          :: value
    integer iDir, iLevel, iElem, iField, pos
    integer                :: found_nan
    !---------------------------------------------------------------------------

    found_nan = 0
    ! Check for nans only if none was found yet to avoid multiple outputs
    ! First appearing nan element is crucial!
    if( .not. nan_Detected ) then
      !> run over each entry in the state array and check if a nan was found
      do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
        found_nan = 0
        do iElem = 1, scheme%pdf( iLevel )%nElems_local
          found_nan = 0
          do iField = 1, scheme%nFields
            do iDir = 1, scheme%layout%stencil(1)%QQ
              pos = scheme%varSys%Variable%val( iField )%varPos( iDir )
              ! Calculate the density for each element each field
              ! Use post-collsion value to compute density
              value = scheme%pdf( iLevel )%state(                              &
   & ?SAVE?( pos, iElem, scheme%varSys%nScalars, scheme%pdf(iLevel)%neigh ),   &
                & scheme%pdf( iLevel )%nNext )
!              if( ieee_is_nan ( value )) then
              if( tem_isnan ( value )) then
                found_nan = iElem
              end if
            end do ! iDir
          end do ! iField
        end do ! iElem
        if( found_nan > 0 ) then
          write(logUnit(1),*) ' rank ', proc%rank,' NaN found at treeID ',     &
            &                 scheme%glob%levelDesc(iLevel)%total( found_nan )
          nan_Detected = .true.
        end if
      end do ! iLevel
    end if

  end subroutine check_NaNs 
!******************************************************************************!


!******************************************************************************!
  !> Check the total density for a selected scheme and write to unit
  !!
  !! The output might be delayed by using arrays which are then dumped 
  !! to keep disc access more restricted
  !!
  subroutine check_scheme_density( scheme, pdfGlobal, general )
    !---------------------------------------------------------------------------
    !> scheme type
    type(mus_scheme_type), intent(inout)   :: scheme
    !> global scheme independent information
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    type(tem_general_type), intent(inout) :: general
    !---------------------------------------------------------------------------
    real(kind=rk), allocatable :: dens(:) , total_dens(:)
    real(kind=rk) :: dens_field
    integer iDir, iLevel, ierror, iElem, iField, pos
    character(len=PathLen) :: outstring, buffer
    !--------------------------------------------------------------------------
    allocate( dens( scheme%nFields ) )
    allocate( total_dens( scheme%nFields ) )
    total_dens = 0._rk
    dens = 0._rk
    ! Sum up all values in the complete vector to get the total density
    lvlLoop: do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      do iElem = 1, scheme%pdf( iLevel )%nElems_fluid
        ! compute density for each field
        do iField = 1, scheme%nFields
          ! field density
          dens_field = 0.0_rk
          do iDir = 1, scheme%layout%stencil(1)%QQ
            pos = scheme%varSys%Variable%val( iField )%varPos( iDir )
            ! Calculate the density for each element each field
            ! Use post-collsion value to compute density
            dens_field = dens_field +                                          &
              & scheme%pdf( iLevel )%state(                                    &
 & ?SAVE?( pos, iElem, scheme%varSys%nScalars, scheme%pdf(iLevel)%neigh ),     &
              & scheme%pdf( iLevel )%nNext )
          end do ! iDir
          ! set nonPhysical if local field density is negative
          ! density can never be negative
          if (dens_field <= 0.0_rk) then
            write(logUnit(1),'(a,i3)') 'ERROR: Negative density for field', iField
            write(logUnit(1),*) 'Terminating check total density calculation'
            write(logUnit(1),*) 'non-physical density=',dens_field
            general%simControl%status%bits(tem_stat_nonPhysical) = .true. 
            exit lvlLoop
          end if  
          ! total density
          dens( iField ) = dens( iField ) + dens_field 
        end do ! iField 
      end do ! iElem 
    end do lvlLoop ! iLevel
 
    ! All ranks check, if there are NaN entries
    if( any( tem_isnan( dens ))  ) then
      write(logUnit(1),*) ' Error: Total density is NaN on rank ',             &
        &                 general%proc%rank
      general%simControl%status%bits(tem_stat_nan_detected) = .true.
    endif

    ! compute total_dens
    call mpi_reduce( dens, total_dens,scheme%nFields, mpi_double_precision,    &
                     mpi_sum, 0, general%proc%comm, ierror )
    outstring = ''
    if ( general%proc%rank == general%proc%root ) then
      ! dump time
      call tem_time_dump(me = general%simControl%now, outUnit = logUnit(1))
      ! dump density
      write(outString, '(a,10a,22a)') '     field ','| total density',         &
        & new_line('A')
      buffer = ''
      do iField = 1, scheme%nFields
        write(buffer, '(a,i10,a,en20.10,a,a)' )                                &
          &        trim(buffer),iField,' | ', total_dens(iField), new_line('A')
      end do
      write(outString, '(a)') trim(outString)//trim(buffer)
      scheme%glob%totalDens = sum(total_dens)
    endif
    write(logUnit(1),*) trim(outstring)

  end subroutine check_scheme_density
!******************************************************************************!


!******************************************************************************!
  !> Check the total density for all schemes
  !!
  subroutine check_all_density( scheme, pdfGlobal, general )
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: scheme(:)
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    type(tem_general_type), intent(inout) :: general
    !---------------------------------------------------------------------------
    integer :: iScheme
    !---------------------------------------------------------------------------

    do iScheme = 1, pdfGlobal%nSchemes
      call check_scheme_density( scheme    = scheme( iScheme ),                &
        &                        pdfGlobal = pdfGlobal,                        &
        &                        general = general )
    enddo

  end subroutine check_all_density
!******************************************************************************!


!******************************************************************************!
  !> Get the min and max communication times
  !!
  subroutine mus_min_max_commTime(params)
    !---------------------------------------------------------------------------
    type(mus_param_type), intent(inout) :: params
    !---------------------------------------------------------------------------
    integer                             :: iError
    !---------------------------------------------------------------------------

    ! Get the min comm time
    call mpi_allReduce( params%general%timerData%value(params%commTimer),      &
      &                 params%mincomm, 1,                                     &
      &                 mpi_double_precision, mpi_min,                         &
      &                 params%general%proc%comm, iError )

    ! Get the max comm time
    call mpi_allReduce( params%general%timerData%value(params%commTimer),      &
      &                 params%maxcomm, 1,                                     &
      &                 mpi_double_precision, mpi_max,                         &
      &                 params%general%proc%comm, iError )

  end subroutine mus_min_max_commTime
!******************************************************************************!


!******************************************************************************!
  !> Performance results (MLUPs) are written to a file for statistical review
  !! The file-format is simple can be evaluated with gnuplot
  !!
  subroutine dump_timing( scheme, params, pdfGlobal )
    !---------------------------------------------------------------------------
    type(mus_scheme_type), intent(inout) :: scheme(:)
    !> global parameter
    type(mus_param_type), intent(inout) :: params
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !---------------------------------------------------------------------------
    real(kind=rk)        :: mainloopTime, loopTime_dyn
    real(kind=rk)        :: mlups, mflups, mlups_dyn
    logical              :: file_exists
    character(15)        :: filename
    integer              :: fileunit, iTimer, iScheme
    integer              :: iter_dyn
    ! iterarions of current run
    integer              :: sim_iter
    real(kind=rk),allocatable        :: timerVal(:)
    character(len=40),allocatable    :: timerLabel(:)
    character(len=PathLen)    :: header
    character(len=PathLen)    :: output
    !---------------------------------------------------------------------------


    ! stop musubi timer before dump timing
    call tem_stopTimer(timerHandle = params%general%solver%timerHandle,        &
      &                me          = params%general%timerData)
    ! Dynamic load balancing: subtract the time used in balancing from the
    ! main time for performance metrics calculation
    if( params%general%balance%dynamic ) then
      iter_dyn     = params%general%simControl%now%iter -                      &
        &            params%general%balance%lastIter
      loopTime_dyn = params%general%timerData%value( params%dyn_LoadBalTimer )
    end if

    if( params%general%proc%rank == params%general%proc%root ) then
      allocate( timerVal( params%general%timerData%nTimers ) )
      allocate( timerLabel( params%general%timerData%nTimers ) )

      do iScheme = 1, pdfGlobal%nSchemes

        fileunit = newunit()

        mainloopTime = 0._rk
        header = ''
        write(header,'(a,a1)') trim(header), '#'
        write(header,'(a,a16)') trim(header), 'Revision'
        write(header,'(a,a7)') trim(header), 'Scheme'
        write(header,'(a,a17)') trim(header), 'DomSize'
        write(header,'(a,a10)') trim(header), 'nProcs'
!$      write(header,'(a,a10)') trim(header), 'nThreads'
        write(header,'(a,a14)') trim(header), 'MFLUPs'
        write(header,'(a,a14)') trim(header), 'MLUPs'
        write(header,'(a,a14)') trim(header), 'MLUPs_dyn'

        write(header,'(a,a14)') trim(header), 'timeTotal'
        write(header,'(a,a11)') trim(header), 'maxIter'
        write(header,'(a,a19)') trim(header), 'totalDens'
        write(header,'(a,a21)') trim(header), 'Casename'

        ! Get the timer values and corresponding labels
        do iTimer = 1, params%general%timerData%nTimers
          timerVal( iTimer )   = params%general%timerData%value( iTimer )
          timerLabel( iTimer ) = 'time'//                                      &
            &                    trim(params%general%timerData%label( iTimer ))
          write(header,'(a,a16)') trim(header),trim(timerLabel( iTimer ))
        enddo
        mainloopTime = params%general%timerData%value(params%mainloopTimer) 
        
        !calculate the actual number of iterations of current run 
        !it might difference from now%iter because of restart
        sim_iter = params%general%simControl%now%iter                          &
          &                    - params%general%simControl%timeControl%min%iter

        mlups  = calc_MLUPS(                                                   &
          &           nElems       = scheme( iScheme )%glob%nElems_solveLevel, &
          &           pdfGlobal    = pdfGlobal,                                &
          &           params       = params,                                   &
          &           iter         = sim_iter,                                 &
          &           mainloopTime = mainloopTime )
        mflups = calc_MLUPS(                                                   &
          &           nElems       = scheme( iScheme )%glob%nElems_totalLevel, &
          &           pdfGlobal    = pdfGlobal,                                &
          &           params       = params,                                   &
          &           iter         = sim_iter,                                 &
          &           mainloopTime = mainloopTime )
        if( params%general%balance%dynamic ) then
          mlups_dyn = calc_MLUPS(                                              &
            &         nElems       = scheme( iScheme )%glob%nElems_solveLevel, &
            &         pdfGlobal    = pdfGlobal,                                &
            &         params       = params,                                   &
            &         iter         = iter_dyn,                                 &
            &         mainloopTime = loopTime_dyn )
        else
          mlups_dyn = 0._rk
        end if

        output = ''
        write(output, '(a,a16)')    trim(output),                              &
          &                         trim(params%general%solver%revision)
        write(output, '(a,i7)')     trim(output), iScheme
        write(output, '(a,i18)')    trim(output),                              &
          &                              scheme(iScheme)%glob%nElems_totalFluid
        write(output, '(a,i10)')    trim(output), params%general%proc%comm_size
!$      write(output, '(a,i10)')    trim(output), omp_get_max_threads()
        write(output, '(a,en14.2)') trim(output), mflups
        write(output, '(a,en14.2)') trim(output), mlups
        write(output, '(a,en14.2)') trim(output), mlups_dyn
        write(output, '(a,en14.4)') trim(output), mainloopTime
        write(output, '(a,en11.0)') trim(output),                              &
          &                   real(params%general%simControl%now%iter, kind=rk)
        write(output, '(a,en19.9)') trim(output), scheme(iScheme)%glob%totalDens
        write(output, '(a,a1,a20)') trim(output), ' ',                         &
          &                         trim(params%general%solver%simName)

        do iTimer = 1, params%general%timerData%nTimers
          write(output,'(a,f16.4)') trim(output),timerVal( iTimer )
        enddo

        filename='timing.res'

        inquire(file=filename,exist=file_exists)
        open(unit=fileunit,file=trim(filename),position='append')

        if( .not. file_exists ) then
           write(fileunit,'(a)') trim(header)
        end if
        write(fileunit,'(a)') trim(output)
        close(fileunit)
      end do
    endif

  end subroutine dump_timing
!******************************************************************************!


!******************************************************************************!
  !> Calculate the MLUPS or MFLUPS for the simulation
  !!
  function calc_MLUPS( pdfGlobal, params, nElems, iter, mainloopTime )         &
    &                  result(mlups)
    !---------------------------------------------------------------------------
    !> global information (here levels are used)
    type(pdf_globInfo_type) :: pdfGlobal
    !> global parameter
    type(mus_param_type)    :: params
    !> array of nElems levelwise
    integer(kind=long_k)    :: nElems(:)
    !> number of total iterations performed
    integer                 :: iter
    !> time consumed for running iter iterations
    real(kind=rk)           :: mainloopTime
    !> resulting mlups
    real(kind=rk)           :: mlups
    !---------------------------------------------------------------------------
    integer :: iLevel
    integer(kind=long_k) :: cellUpdates
    !---------------------------------------------------------------------------

    cellUpdates = 0_long_k

    ! Calculate the mlups with the level wise time steps
    ! nElems divided by scaleFactor on the level is same as
    ! multiplying scaleFactor to number of time steps.
    ! here, the scalefactor for that level is defined respect to
    ! finest level since iter - defines number of iterations performed
    ! in the finest level
    do iLevel = pdfGlobal%minLevel, pdfGlobal%maxLevel
      cellUpdates = cellUpdates + nElems(iLevel)                               &
        &         / int( params%scaleFactor**(                                 &
        &                       pdfGlobal%maxLevel - iLevel ), kind=long_k )
    enddo

    mlups = real(cellUpdates*iter,kind=rk) / (mainloopTime*1000000._rk)

  end function calc_MLUPS
!******************************************************************************!


!******************************************************************************!
  !> Check for the streaming layout.
  !!
  subroutine check_streaming_layout( pdfGlobal )
    !---------------------------------------------------------------------------
    type(pdf_globInfo_type), intent(in) :: pdfGlobal
    !---------------------------------------------------------------------------

    ! Check, if advection layout is PUSH. if so, abort, because the current code
    ! can only handle PUSH with single level
    if( "?StreamName?" == 'push') then
      write(logUnit(1),*)' Pushing pdfs to neighbor elements after collision '
      if( pdfGlobal%minLevel /= pdfGlobal%maxLevel ) then
        write(logUnit(1),*) ' ----------------------------------------------'//&
          &                 '------------ '
        write(logUnit(1),*) '           WARNING !!'
        write(logUnit(1),*) '    Push is not possible with multi-level '//     &
          &                 'execution.'
        write(logUnit(1),*) ' ----------------------------------------------'//&
          &                 '------------ '
        write(logUnit(1),*)

      endif
    else
      write(logUnit(1),*)' Pulling pdfs from neighbor elements before collision'
    endif

  end subroutine check_streaming_layout
!******************************************************************************!

end module mus_tools_module
!******************************************************************************!
