!******************************************************************************!
!> M U S U B I
!! The lattice Boltzmann solver within the APES suite
!! (c) 2012 German Research School for Simulation Sciences GmbH
!!
!! For a documentation, run ./waf gendoxy and find the documentation at
!! ./Documentation/html/index.html
program musubi
  ! include treelm modules
  use mpi
  use tem_general_module,         only: tem_start, tem_finalize

  ! include musubi modules
  ! RESPECT THE ORDER !!!!!!!!
  use mus_aux_module,                only: mus_banner
  use mus_config_module,             only: mus_load_config
  use mus_pdf_module,                only: pdf_globInfo_type
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_param_module,              only: mus_param_type,                     &
    &                                      mus_init_timers, mus_init_leveltimers
  use mus_geom_module,               only: mus_geom_type
  use mus_interpolate_header_module, only: interpolation_type
  ! ESPECIALLY OF THE FOLLOWING MODULES
  use mus_control_module,            only: mus_control_type
  use mus_program_module,            only: mus_initialize, mus_solve,          &
    &                                      mus_finalize

  implicit none
  !-----------------------------------------------------------------------------
  type(mus_scheme_type), allocatable :: scheme(:) !< scheme types
  type(mus_geom_type)                :: geometry
  type(mus_param_type)               :: params
  type(interpolation_type)           :: intp
  type(pdf_globInfo_type)            :: pdfGlobal
  type(mus_control_type)             :: control
  integer :: ierr
  !-----------------------------------------------------------------------------

  ! Initialize environment
  call tem_start(codeName   = 'Musubi',                 &
    &            version    = params%version,           &
    &            general    = params%general,           &
    &            simControl = params%general%simControl )

  call mus_banner(solver = params%general%solver)

  call mus_init_timers( params )
  ! load configuration file
  call mus_load_config( scheme    = scheme,                                 &
    &                   geometry  = geometry,                               &
    &                   params    = params,                                 &
    &                   pdfGlobal = pdfGlobal,                              &
    &                   intp      = intp )

  ! initialize global params
  call mus_init_leveltimers( params, pdfGlobal%minLevel, pdfGlobal%maxLevel )

  ! initialize musubi
  call mus_initialize( scheme    = scheme,                                     &
    &                  geometry  = geometry,                                   &
    &                  params    = params,                                     &
    &                  pdfGlobal = pdfGlobal,                                  &
    &                  intp      = intp,                                       &
    &                  control   = control )

  call mpi_barrier( MPI_COMM_WORLD, ierr )

  ! do main loop
  call mus_solve( scheme    = scheme,                                          &
    &             geometry  = geometry,                                        &
    &             params    = params,                                          &
    &             pdfGlobal = pdfGlobal,                                       &
    &             intp      = intp,                                            &
    &             control   = control )

  ! finialize musubi
  call mus_finalize( scheme    = scheme,   &
    &                params    = params,   &
    &                geometry  = geometry, &
    &                pdfGlobal = pdfGlobal )

  ! finalize treelm function like print run time info and mpi
  call tem_finalize(params%general)

end program musubi
