?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Definition of the datatypes for the scheme implementation.
!!
!! For a detailed description see
!! \ref scheme_implementation "Scheme Implementation".
!!
!! \author Simon Zimny, Kannan Masilamani
!!
module mus_scheme_module

  ! include treelm modules
  use env_module,            only: rk, newUnit, LabelLen, PathLen
  use tem_aux_module,        only: tem_abort, tem_CheckLabel
  use treelmesh_module,      only: treelmesh_type
  use tem_bc_prop_module,    only: tem_bc_prop_type
  use tem_restart_module,    only: tem_restart_type
  use tem_tools_module,      only: tem_horizontalSpacer
  use tem_tracking_module,   only: tem_load_tracking
  use tem_variable_module,   only: tem_variable_type, tem_variable_from
  use tem_var_system_module, only: tem_var_system_type, tem_var_append,        &
    &                              tem_var_sys_mark, tem_var_intersect,        &
    &                              tem_var_sys_init
  use tem_solveData_module,  only: tem_solveData_load
  use tem_source_module,     only: tem_sourceMap_type, tem_load_source,        &
    &                              tem_reduce_sourceVars
  use tem_derived_module,    only: access_stateVar
  use tem_meshInfo_module,   only: nMeshVars
  use tem_depend_module,     only: tem_load_depend_single, tem_init_depend
  use tem_logging_module,    only: logUnit

  ! include aotus modules
  use aot_out_module,   only: aot_out_type, aot_out_open, aot_out_close
  use aotus_module,     only: flu_State, open_config_chunk
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode

  ! include musubi modules
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_param_module,         only: mus_param_type
  use mus_scheme_layout_module, only: mus_load_newLayout, mus_weights_out,     &
    &                                 mus_define_d3q19, mus_define_d3q7,       &
    &                                 mus_define_d2q9, mus_define_d3q6,        &
    &                                 mus_define_d1q3, mus_define_d3q27,       &
    &                                 mus_init_stencil, mus_finish_stencil
  use mus_scheme_header_module, only: mus_scheme_header_type,                  &
    &                                 mus_load_scheme_header,                  &
    &                                 mus_scheme_header_out
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_moments_module,       only: mus_init_moments
  use mus_field_module,         only: mus_field_type, mus_load_fields,         &
    &                                 mus_field_out
  use mus_variable_module,      only: mus_init_varSys, mus_build_derVarList,   &
    &                                 mus_init_variables, mus_store_derVarPos, &
    &                                 mus_nGenVars
  use mus_source_module,        only: mus_build_sourceVarSys, mus_init_source
  use mus_physics_module,       only: mus_physics_type, mus_physics_out
  use mus_mixture_module,       only: mus_mixture_type, mus_mixture_out
  use mus_externalForce_module, only: mus_load_externalForce
  use mus_bc_header_module,     only: mus_setBC_stencilPos

  ! include aotus modules
  use aotus_module, only: flu_state, aot_get_val

  implicit none

  private

  public :: mus_readSolSpecChar
  public :: mus_writeSolSpecChar
  public :: mus_load_schemes
  public :: mus_init_sources
  public :: mus_init_stateVars
  public :: mus_kill_stateVars

  interface mus_writeSolSpecChar
    module procedure mus_writeSolSpecChar_array
    module procedure mus_writeSolSpecChar_single
  end interface

contains

!******************************************************************************!
  !> load different schemes defined in lua file
  !!
  !! This routines checks whether schemes table
  !! \verbatim
  !! scheme = { boundary_conditions = {...},
  !!            ...  }
  !! \endverbatim
  !! is defined. If yes then it will
  !! load schemes table and set mus_scheme_type.
  !! If no special scheme table is defined, default tables are loaded from the
  !! root level of the lua file
  !! fluid, boundary conditions and initial condiitions
  !! \verbatim
  !! fluid = { ... }
  !! boundary_conditions = {...}
  !! initial_conditions = {...}
  !! \endverbatim
  !!
  subroutine mus_load_schemes( me, boundary, conf, params, pdfGlobal, tree,    &
    &                          path )
    !---------------------------------------------------------------------------
    !> array of scheme types to be filled
    type( mus_scheme_type ), intent(out), allocatable :: me(:)
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in):: boundary
    !> global parameter type
    type( mus_param_type ), intent(inout) :: params
    !> flu state
    type( flu_State ) :: conf
    !> global pdf info
    type( pdf_globInfo_type) :: pdfGlobal
    !> global treelm mesh
    type( treelmesh_type), intent(inout) :: tree
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !---------------------------------------------------------------------------
    integer :: nSchemes, iScheme
    integer :: scheme_handle
    integer :: scheme_sub_handle
    ! lua paths
    type( aot_path_type ) :: schemes_path ! path to the scheme table
    type( aot_path_type ) :: scheme_sub_path ! path to entry in scheme table
    !---------------------------------------------------------------------------

    call tem_horizontalSpacer(fUnit = logUnit(1))
    write(logUnit(1),*) 'Loading the schemes ...'

    ! initialize all musubi derived variables
    call mus_init_variables()

    ! try to open the scheme table
    call aot_table_open( L=conf, thandle=scheme_handle, key='scheme' )
    ! set schemes path
    schemes_path = path
    call aot_path_addNode( me       = schemes_path,                            &
      &                    NodeType = 'table',                                 &
      &                    key      = 'scheme' )

    nSchemes = 1
    pdfGlobal%nSchemes = nSchemes
    allocate( me( nSchemes ))
    me(1)%schemeID = 1

    if (scheme_handle == 0) then
      ! scheme table is not defined
      ! -> load single scheme with passing scheme handle
      write(logUnit(1),*) 'No scheme table defined. &
       &  Setting number of schemes: ', nSchemes
      call mus_load_scheme_single( me        = me(1),                          &
        &                          boundary  = boundary,                       &
        &                          conf      = conf,                           &
        &                          params    = params,                         &
        &                          pdfGlobal = pdfGlobal,                      &
        &                          tree      = tree,                           &
        &                          path      = path )
    else
      ! check whether scheme is a single table or multiple table
      call aot_table_open( L       = conf,                                     &
        &                  parent  = scheme_handle,                            &
        &                  thandle = scheme_sub_handle,                        &
        &                  pos     = 1 )
      if ( scheme_sub_handle == 0 ) then
        ! scheme is a single table, load its contents
        call aot_table_close( L=conf, thandle=scheme_sub_handle )
        write(logUnit(1),*) 'Scheme is a single table.'
        call mus_load_scheme_single( me        = me(1),                        &
          &                          boundary  = boundary,                     &
          &                          conf      = conf,                         &
          &                          params    =  params,                      &
          &                          pdfGlobal = pdfGlobal,                    &
          &                          tree      = tree,                         &
          &                          path      = schemes_path,                 &
          &                          parent    = scheme_handle )
      else
        ! scheme is a multiple table
        call aot_table_close( L=conf, thandle=scheme_sub_handle )
        nSchemes = aot_table_length( L=conf, thandle=scheme_handle )
        write(logUnit(1),*) 'Number of schemes used: ', nSchemes
        deallocate( me )
        allocate( me( nSchemes ))
        pdfGlobal%nSchemes = nSchemes

        do iScheme = 1, nSchemes
          ! storing the number of the current scheme
          me(iScheme)%schemeID = iScheme

          call aot_table_open( L       = conf,                                 &
            &                  parent  = scheme_handle,                        &
            &                  thandle = scheme_sub_handle,                    &
            &                  pos     = iScheme )
          scheme_sub_path = schemes_path
          call aot_path_addNode( me       = scheme_sub_path,                   &
            &                    NodeType = 'table',                           &
            &                    pos      = iScheme )
          !write(logUnit(1),*) 'Properties for scheme', iScheme
          call tem_horizontalSpacer(fUnit = logUnit(1))
          call mus_load_scheme_single( me        = me( iScheme ),              &
            &                          boundary  = boundary,                   &
            &                          conf      = conf,                       &
            &                          params    = params,                     &
            &                          pdfGlobal = pdfGlobal,                  &
            &                          tree      = tree,                       &
            &                          path      = scheme_sub_path,            &
            &                          parent    = scheme_sub_handle )
          ! Check, if the names of the schemes are unique.
          ! If not, add an increasing number
          call tem_checkLabel( label = me(:)%header%label, nLabels = iScheme )
          call aot_table_close( L=conf, thandle=scheme_sub_handle )
        end do ! number of schemes

        ! if multiple schemes are defined.
        ! it is done after loading all scheme table since the global system
        ! in scheme might differ based on their scheme kind and in
        ! scheme depend we need to find the interection of required
        ! variable system from global var system of the depend scheme
        ! load scheme dependency for each scheme from lua file
        call mus_load_scheme_depend( me     = me,                              &
          &                          conf   = conf,                            &
          &                          parent = scheme_handle)

      end if !if scheme table is single or multiple
    end if ! if scheme is a table

    call aot_table_close( L=conf, thandle=scheme_handle )

    ! initialize the scheme depend for all schemes
    do iScheme = 1, nSchemes
      if ( allocated( me(iScheme)%depend) ) then
        call tem_init_depend( medepend    = me(iScheme)%depend,                &
          &                   varSys      = me(:)%varSys,                      &
          &                   resSysLabel = me(iScheme)%header%label,          &
          &                   label_list  = me(:)%header%label,                &
          &                   iScheme     = iScheme )
      end if
    end do

    ! write solver-specific character for passing solver-stuff to the
    ! treelm restart and tracking files
    call mus_writeSolSpecChar( scheme = me, tree = tree, params = params )

  end subroutine mus_load_schemes
!******************************************************************************!


!******************************************************************************!
  !> load single scheme defined in lua file with or without scheme handle
  !!
  subroutine mus_load_scheme_single( me, boundary, conf, params, pdfGlobal,    &
    &                                tree, path, parent )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), intent(inout) :: me
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in):: boundary
    !> global parameter type
    type( mus_param_type ), intent(inout) :: params
    !> flu state
    type( flu_State ) :: conf
    !> global pdf info
    type( pdf_globInfo_type) :: pdfGlobal
    !> global treelm mesh
    type( treelmesh_type), intent(inout) :: tree
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !> parent handle if scheme table is defined
    integer, intent(in), optional :: parent
    !---------------------------------------------------------------------------
    type( tem_variable_type ), allocatable :: tVar(:)
    integer :: loc_nTotal
    integer :: iField, iError
    !---------------------------------------------------------------------------

    ! load scheme identifier table
    call mus_load_scheme_header( me      = me%header,                          &
      &                          conf    = conf,                               &
      &                          parent  = parent,                             &
      &                          scaling = params%scaling )

    ! load fields from parent, including possible sources for each field
    call mus_load_fields( me           = me%field,                             &
      &                   nFields      = me%nFields,                           &
      &                   mixture      = me%mixture,                           &
      &                   boundary     = boundary,                             &
      &                   conf         = conf,                                 &
      &                   pdfGlobal    = pdfGlobal,                            &
      &                   path         = path,                                 &
      &                   parent       = parent,                               &
      &                   schemeHeader = me%header,                            &
      &                   source       = me%source,                            &
      &                   physics      = params%physics,                       &
      &                   tree         = tree,                                 &
      &                   scaling      = params%scaling,                       &
      &                   rank         = params%general%proc%rank )

    ! load force fields
    call mus_load_externalForce( me        = me%ext_force,                     &
      &                          conf      = conf,                             &
      &                          path      = path,                             &
      &                          parent    = parent,                           &
      &                          pdfGlobal = pdfGlobal,                        &
      &                          physics   = params%physics )

    ! load the global source variables
    call tem_load_source( me     = me%source(me%nFields + 1),                  &
      &                   conf   = conf,                                       &
      &                   parent = parent,                                     &
      &                   key    = 'source',                                   &
      &                   path   = path )

    ! load tracking
    call tem_load_tracking( tCtrl  = me%trackingControl, &
      &                     me     = me%track,           &
      &                     conf   = conf,               &
      &                     parent = parent,             &
      &                     path   = path                )

    ! set the number of stencils in this scheme:
    ! 1 for the fluid stencil
    ! + number of fields * number of BCs
    me%layout%nStencils = 1                                                    &
      &                 + pdfGlobal%nBCs*me%nFields

    write(logUnit(1),*) 'Number of stencils: ', me%layout%nStencils

    ! define the scheme stencil layout(set parameters, offsets, ...)
    call mus_init_stencil( stencil   = me%layout%stencil,                      &
      &                    nStencils = me%layout%nStencils )

    ! for all fields
    do iField = 1, me%nFields
      ! ... set the stencil positions for the individual boundary conditions
      call mus_setBC_stencilPos( me     = me%field( iField )%bc,               &
        &                        iField = iField,                              &
        &                        nBCs   = pdfGlobal%nBCs )
    end do

    select case ( trim( me%header%layout ) )
      case( 'd2q9' )
        pdfGlobal%nDims = 2
        call mus_define_d2q9( layout = me%layout, nElems = tree%nElems )
      case( 'd3q19' )
        pdfGlobal%nDims = 3
        call mus_define_d3q19( layout = me%layout, nElems = tree%nElems )
      case( 'd3q27' )
        pdfGlobal%nDims = 3
        call mus_define_d3q27( layout = me%layout, nElems = tree%nElems )
      case( 'd3q7' )
        pdfGlobal%nDims = 3
        call mus_define_d3q7( layout = me%layout, nElems = tree%nElems )
      case( 'd3q6', 'flekkoy' )
        pdfGlobal%nDims = 3
        call mus_define_d3q6( layout = me%layout, nElems = tree%nElems )
      case( 'd1q3' )
        pdfGlobal%nDims = 1
        call mus_define_d1q3( layout = me%layout, nElems = tree%nElems )
      case ( 'new_stencil' )
        call tem_horizontalSpacer(fUnit = logUnit(1))
        write(logUnit(1),*) 'Reading the new layout...'
        call mus_load_newLayout( me            = me%layout,                    &
          &                      parent_handle = parent,                       &
          &                      conf          = conf )
      case default
        call tem_horizontalSpacer(fUnit = logUnit(1))
        write(logUnit(1),*) 'The chosen scheme layout is not available. '
        write(logUnit(1),*) 'STOPPING!'
        write(logUnit(1),*) 'Choose one of:  '
        write(logUnit(1),*) '  - d2q9 '
        write(logUnit(1),*) '  - d3q19 '
        write(logUnit(1),*) '  - d3q7  '
        write(logUnit(1),*) '  - d3q6  '
        write(logUnit(1),*) '  - new_stencil  '
        call tem_abort()
    end select
    ! finish stencil definition
    call mus_finish_stencil( layout = me%layout )

    write(logUnit(1),*) 'Chosen stencil layout '//trim( me%header%layout )
    write(logUnit(1),*) '  Corresponding nDims ', pdfGlobal%nDims

    ! To initialiaze moments 
    if (present(parent)) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = parent,                                      &
        &               key     = 'require_moments',                           &
        &               val     = me%requireMoments,                           &
        &               ErrCode = iError,                                      &
        &               default = .true.)
    else
      call aot_get_val( L       = conf,                                        &
        &               key     = 'require_moments',                           &
        &               val     = me%requireMoments,                           &
        &               ErrCode = iError,                                      &
        &               default = .true.)
    end if    

    ! initialize the moments if requireMoments is true
    ! KM: Initialize moments independent whether its required for
    ! interpolation or mrt
    call mus_init_moments( me             = me%layout%moment, &
      &                    layout         = me%layout,        &
      &                    requireMoments = me%requireMoments )

    call tem_horizontalSpacer(fUnit = logUnit(1))

    ! add analytical function loaded from lua file
    ! and in the end add error variable without dependency
    call tem_variable_from( me     = tVar,                                     &
      &                     conf   = conf,                                     &
      &                     parent = parent,                                   &
      &                     nTotal = loc_nTotal,                               &
      &                     key    = 'add_variable',                           &
      &                     path   = path )

    ! build a list of all derived variables (without prefixes)
    call mus_build_derVarList(                                                 &
      &    derVarList = me%derVarList,                                         &
      &    schemeHeader = me%header,                                           &
      &    stencil    = me%layout%stencil(1) )

    ! Initialize the variable system
    call tem_var_sys_init( me%varSys )

    ! initialize musubi state variables to the scheme variable system
    call mus_init_stateVars( me, size(tVar) )

    ! Now initialize all the available derived quantities
    call mus_init_varSys( derVarList = me%derVarList,                          &
      &                   var_sys    = me%varSys,                              &
      &                   label      = me%field(:)%label,                      &
      &                   nFields    = me%nFields,                             &
      &                   tVar       = tVar )

    ! build derVarPos
    call mus_store_derVarPos( derVarPos = me%derVarPos,                        &
      &                       var_sys   = me%varSys,                           &
      &                       label     = me%field(:)%label )

  end subroutine mus_load_scheme_single
!******************************************************************************!

!******************************************************************************!
  !> load multiple dependent variables of the scheme
  !!
  subroutine mus_load_scheme_depend( me, conf, parent )
    !---------------------------------------------------------------------------
    !> scheme type
    type( mus_scheme_type ), intent(inout) :: me(:)
    !> flu state
    type( flu_State ) :: conf
    !>
    integer, intent(in) :: parent
    !---------------------------------------------------------------------------
    integer :: scheme_sub_handle, depend_handle, depend_sub_handle
    integer :: iScheme, iDepend, nDepend
    !---------------------------------------------------------------------------
    write(logUnit(1),*) 'loading scheme depend table'
    do iScheme = 1, size(me)
      call aot_table_open( L       = conf,                                     &
        &                  parent  = parent,                                   &
        &                  thandle = scheme_sub_handle,                        &
        &                  pos     = iScheme )

      call aot_table_open( L       = conf,                                     &
        &                  parent  = scheme_sub_handle,                        &
        &                  thandle = depend_handle,                            &
        &                  key     = 'depend' )
      if( depend_handle > 0 ) then
        !check single table or multiple table
        call aot_table_open( L       = conf,                                   &
          &                  parent  = depend_handle,                          &
          &                  thandle = depend_sub_handle,                      &
          &                  pos     = 1 )
        if ( depend_sub_handle == 0 ) then
          allocate( me(iScheme)%depend(1) )
          call tem_load_depend_single( medepend = me(iScheme)%depend(1),       &
            &                          label    = me(iScheme)%header%label,    &
            &                          conf     = conf,                        &
            &                          thandle  = depend_handle )
        else
          call aot_table_close( L=conf, thandle=depend_sub_handle )
          nDepend = aot_table_length( L=conf, thandle=depend_handle )
          allocate( me(iScheme)%depend(nDepend) )
          do iDepend = 1, nDepend
            call aot_table_open( L       = conf,                               &
              &                  parent  = depend_handle,                      &
              &                  thandle = depend_sub_handle,                  &
              &                  pos     = iDepend )
            ! Load  variable system
            !\todo: SZ: why mus_load_depend_single and not mus_load_depend??
            call tem_load_depend_Single(                                       &
              &                       medepend = me(iScheme)%depend(iDepend),  &
              &                       label    = me(iScheme)%header%label,     &
              &                       conf     = conf,                         &
              &                       thandle  = depend_sub_handle )
            call aot_table_close( L = conf, thandle = depend_sub_handle )
          end do
        end if
      else
        write(logUnit(1),*) 'No depend table defined for scheme: '//           &
          & trim(me(iScheme)%header%label)
      end if
      call aot_table_close( L=conf, thandle=depend_handle )

      call aot_table_close( L=conf, thandle=scheme_sub_handle )
    end do ! number of schemes

  end subroutine mus_load_scheme_depend
!******************************************************************************!


!******************************************************************************!
  !> Write solver-specific information (for e.g. the restart header file, which
  !! is being handled in the Treelm) into a scratch file.
  !! The scratch file contents will then be read by the Treelm
  !! \todo KM: write stencil layout, fieldprop to solverspecific unit
  !!
  subroutine mus_writeSolSpecChar_array( scheme, tree, params )
    !----------------------------------------------------------------------------
    !> type holding state array, global information etc.
    type( mus_scheme_type ), intent(inout) :: scheme(:)
    !> global treelm mesh
    type( treelmesh_type), intent(in) :: tree
    !> global parameter type
    type( mus_param_type ), intent(inout) :: params
    !----------------------------------------------------------------------------
    integer :: iScheme
    !----------------------------------------------------------------------------
    ! Assemble solver specific lua table
    ! this might include a table of the field properties
    do iScheme = 1, size(scheme)
      call mus_writeSolSpecChar_single( scheme = scheme( iScheme ),            &
        &                               tree = tree, params = params )
    enddo

  end subroutine mus_writeSolSpecChar_array
!******************************************************************************!

!******************************************************************************!
  !> Write solver-specific information (for e.g. the restart header file, which
  !! is being handled in the Treelm) into a scratch file.
  !! The scratch file contents will then be read by the Treelm
  !! \todo KM: write stencil layout, fieldprop to solverspecific unit
  !!
  subroutine mus_writeSolSpecChar_single( scheme, tree, params )
    !---------------------------------------------------------------------------
    !> type holding state array, global information etc.
    type( mus_scheme_type ), intent(inout) :: scheme
    !> global treelm mesh
    type( treelmesh_type), intent(in) :: tree
    !> global parameter type
    type( mus_param_type ), intent(inout) :: params
    !---------------------------------------------------------------------------
    integer :: nUnit
    type(aot_out_type) :: conf
    !---------------------------------------------------------------------------

    ! Assemble solver specific lua table
    ! this might include a table of the field properties
    if (params%general%proc%rank == 0) then

      nUnit = newUnit()
      open(unit = nUnit, status='scratch')
      call aot_out_open( put_conf =  conf, outUnit = nUnit )

      !\todo write species properties
      call mus_scheme_header_out( me = scheme%header, conf = conf )

      call mus_field_out( me           = scheme%field,                         &
        &                 conf         = conf,                                 &
        &                 schemeHeader = scheme%header )

      call mus_mixture_out( me = scheme%mixture, conf = conf, &
       & schemeHeader = scheme%header )

      call mus_weights_out( me = scheme%layout%weight, conf = conf )

      call mus_physics_out( me   = params%physics,           &
        &                   conf = conf,                     &
        &                   minLevel = tree%global%minLevel, &
        &                   maxLevel = tree%global%maxLevel  )

      call aot_out_close( put_conf =  conf )

    end if

    ! transform the solver specific character to a single string on all
    ! processes
    call tem_solveData_load(me        = scheme%varSys%solSpec,                 &
      &                     file_unit = nUnit,                                 &
      &                     rank      = params%general%proc%rank,              &
      &                     comm      = params%general%proc%comm)
    if (params%general%proc%rank == 0) close(nUnit)

    write(logUnit(10),*) 'solver specific string:'
    write(logUnit(10),*) trim(scheme%varSys%solspec%String)

  end subroutine mus_writeSolSpecChar_single
!******************************************************************************!

!******************************************************************************!
  !\todo: what about this function??,
  !> Take the solver specific information (from e.g. the restart header),
  !! and read it into the data structures using the lua read functions
  !!
  !! The information is stored in a lua style string.
  !!
  subroutine mus_readSolSpecChar( restart, scheme )
   !----------------------------------------------------------------------------
   !> the solver specific character
   type( tem_restart_type ) :: restart
   !> type holding state array, global information etc.
   type( mus_scheme_type ) :: scheme(:)
   !----------------------------------------------------------------------------
   integer :: iScheme, nSchemes ! scheme counter
   !----------------------------------------------------------------------------
   if( restart%comm%rank == 0 ) then
     nSchemes = size( scheme )
     ! Read fluid data etc. from lua header file
     do iScheme = 1, nSchemes
       write(logUnit(1),*)' Must read fluid props from restart header here'
     enddo
   endif

  end subroutine mus_ReadSolSpecChar
!******************************************************************************!

!******************************************************************************!
  !> Initialize the global variable system for this scheme.
  !!
  !! allocate function pointers, and append pdf to scheme%varSys
  subroutine mus_init_stateVars( scheme, extraVar )
    !---------------------------------------------------------------------------
    !> type holding state array, global information etc.
    type( mus_scheme_type ) :: scheme
    !> additional variables defined for this scheme
    integer, intent(in) :: extraVar
    !---------------------------------------------------------------------------
    integer :: pos, iField
    integer :: nDeriveFuns, nDerVarPos
    character(len=labelLen) ::  buffer
    !---------------------------------------------------------------------------

    write(logUnit(1),*) 'Initializing global variable system for scheme: ',&
      &                              scheme%schemeID

    ! allocate state variables and derived functions
    nDeriveFuns = 1 + size( scheme%derVarList )
    nDerVarPos = 1

    ! JQ: allocate derive for state variable omega ( nonNewtonian feature )
    select case ( trim(scheme%header%kind) )
      case ( 'lbm_nNwtn', 'lbm_incomp_nNwtn' )
        do iField = 1, scheme%nFields
          nDeriveFuns = nDeriveFuns + 1
        end do
    end select

    if ( scheme%nFields > 1 ) then
      ! each field has a (derVarList + state_var), +1 derVarList for mixture
      nDeriveFuns = nDeriveFuns * scheme%nFields + size( scheme%derVarList )
      nDerVarPos = nDerVarPos * ( scheme%nFields + 1 )
    end if

    ! +extraVar +number of generic variables + number of mesh variables
    nDeriveFuns = nDeriveFuns + extraVar + mus_nGenVars + nMeshVars

    allocate( scheme%varSys%derive( nDeriveFuns ))
    allocate( scheme%derVarPos( nDerVarPos ) )

    ! append pdf for each field
    do iField = 1, scheme%nFields
      write(buffer,'(a)') trim( scheme%field( iField )%label )//'pdf'
      call tem_var_append( sys         = scheme%varSys,               &
        &                  varName     = trim(buffer),                &
        &                  nComponents = scheme%layout%stencil(1)%QQ, &
        &                  pos         = pos                          )
      scheme%varSys%derive( pos )%do => access_stateVar
    end do

    ! when nonNewtonian feature is activated,
    ! append omage as state variable.
    select case ( trim(scheme%header%kind) )
      case ( 'lbm_nNwtn', 'lbm_incomp_nNwtn' )
        do iField = 1, scheme%nFields
          write(buffer,'(a)') trim( scheme%field( iField )%label )//'omega'
          call tem_var_append( sys         = scheme%varSys, &
            &                  varName     = trim(buffer),  &
            &                  nComponents = 1,             &
            &                  pos         = pos )
          scheme%varSys%derive( pos )%do => access_stateVar
        end do
    end select

    ! Mark system as complete, so we can allocate the state vector
    ! and set the name of the variable system to the scheme label
    call tem_var_sys_mark( sys        = scheme%varSys,                         &
      &                    SystemName = trim( scheme%header%label ))

    ! debug output
    write(logUnit(10),*) '  nVars in varSys: ', scheme%varSys%nVars
    write(logUnit(10),*) '  nScalars in varSys: ', scheme%varSys%nScalars
    write(logUnit(10),*) '  num of derive functions: ', nDeriveFuns
    write(logUnit(10),*) '  num of state variables: ', nDerVarPos


  end subroutine mus_init_stateVars
!******************************************************************************!

!******************************************************************************!
  !> Destroy the arrays initialized in mus_init_stateVars
  !!
  subroutine mus_kill_stateVars( scheme )
    !---------------------------------------------------------------------------
    !> type holding state array, global information etc.
    type( mus_scheme_type ) :: scheme
    !---------------------------------------------------------------------------

    deallocate( scheme%varSys%derive)
    deallocate( scheme%derVarPos )
  end subroutine mus_kill_stateVars
!******************************************************************************!

!******************************************************************************!
  !> Initialize all source terms for the fields and the one for the scheme
  !!
  subroutine mus_init_sources( me, tree )
    !---------------------------------------------------------------------------
    !> scheme information
    type(mus_scheme_type) :: me
    !> The tree
    type(treelmesh_type)  :: tree
    !---------------------------------------------------------------------------
    integer :: iField
    integer :: mapSize
    character(len=labellen), allocatable :: tmp_prefix(:)
    !---------------------------------------------------------------------------

    write(logUnit(1),*)'Initializing the source terms ...'
    ! allocate and initialize the temporary array of prefixes
    ! size = nFields + 1 (for the scheme)
    allocate(tmp_prefix(me%nFields+1))

    tmp_prefix( 1:me%nFields )  = me%field( : )%label
    tmp_prefix(me%nFields + 1)  = ''

    ! initialize the basic source variable systems for each field
    call mus_build_sourceVarSys( me            = me%source,                    &
      &                          sourceVarList = me%sourceVarList,             &
      &                          stencil       = me%layout%stencil(1),         &
      &                          schemeKind    = trim(me%header%kind))

    do iField = 1, me%nFields + 1
      ! intersect the requested and the possible source variable systems
      call tem_var_intersect( systemA     = me%source(iField)%requ_varSys,     &
        &                     systemB     = me%source(iField)%poss_varSys,     &
        &                     resSys      = me%source(iField)%varSys,          &
        &                     resSysLabel = 'possInterRequ' )

    end do

    ! allocate the sourceMap with the number of variables in the global varSys
    ! plus the sum of those ones appended by the individual variable systems in
    ! the sources

    mapSize = me%varSys%variable%nVals
    do iField = 1, me%nFields + 1
      mapSize = mapSize + me%source( iField )%varSys%variable%nVals
    end do

    allocate( me%sourceMap%map( mapSize ))
    me%sourceMap%map = 0
    allocate( me%sourceMap%source( mapSize ))

    !\todo: SZ: remove the verbose statement
    call tem_reduce_sourceVars( me          = me%source,                       &
      &                         global_tree = tree )

    ! initialize the sources for all fields including the prefixes
    call mus_init_source( me            = me%source,                           &
      &                   global_varSys = me%varSys,                           &
      &                   sourceMap     = me%sourceMap,                        &
      &                   prefix        = tmp_prefix )

    ! copy the information back
!    me%source = tmp_sources

    ! update the varmap with the right function pointers
!    call mus_update_varmap( me = me%var,                     &
!      &                     varSys = me%varSys,              &
!      &                     sourceMap = me%sourceMap,        &
!      &                     sourceVarList = me%sourceVarList)

  end subroutine mus_init_sources
!******************************************************************************!

end module mus_scheme_module
!******************************************************************************!


!******************************************************************************!
!> \page scheme_implementation Scheme Implementation
!! The concept of schemes provides the user with a bigger flexibility. It is now
!! possible to run multiple simulations with different layouts on the same mesh.
!! This is the basis for LBM passive scalar transport simulations in MUSUBI.
!! In general the scheme implementation is able to run on old and new lua files.
!!
!! \subsection scheme_definition Scheme Definition
!!
!! The scheme is part of the individual solvers and contains all information about
!! initial conditions, boundary conditions, the \ref mus_scheme_layout_module
!! "Scheme Layout", fluid and flow properties.
!!
!! \subsection scheme_usage Usage
!!
!! To start a flow simulation one can either use the 'old' lua files or add the
!! table species with the following quantities:
!!
!! - the scheme identifier which include label, kind, relaxation, layout
!! - the initial conditions
!! - the boundary conditions
!! - the fluid quantities
!!
!! \verbatim
!! -- Sample Flow Scheme Definition
!! scheme = {
!!   {
!!     identify = {
!!       label = 'fluid',
!!       kind = 'lbm',
!!       relaxation = 'bgk',
!!       layout = 'd3q19',
!!     }, -- identify table
!!
!!    -- Initial condition
!!    initial_condition = {
!!      density = 1.0,
!!      velocityX = 0.0,
!!      velocityY = 0.0,
!!      velocityZ = 0.0,
!!      Sxx = 0.0, Syy = 0.0, Szz = 0.0, Sxy = 0.0, Syz = 0.0, Sxz = 0.0,
!!    },
!!
!!    -- Boundary conditions
!!    boundary_condition = {
!!    { label = 'wall',
!!      kind = 'inlet_ubb',
!!      velocityX = 0.03, velocityY = 0.0, velocityZ = 0.0 },
!!    { label = 'wall',
!!      kind = 'wall',
!!      velocityX = 0.0, velocityY = 0.0, velocityZ = 0.0 }
!!    }, -- boundary table
!!
!!    fluid = {
!!       omega = 1.8,
!!       rho0 = 1.0
!!    }, -- fluid table
!!   }
!! }
!! \endverbatim
!!
!! Full example:
!!
!! This example runs a multi scheme simulation (flow & passive scalar) on one
!! grid:
!!
!! \verbatim
!!
!! \endverbatim

