! This file contains the source code for growing and dynamic arrays.
! This is used for arrays of primitives (int, long_int, real, ...) as well as
! for arrays of derived datatypes (tem_variable_type,...).
!
! To use these macros include the following to your source file.
!
! Smart growing array (GA) for ?tstring?
! Growing Arrays:
!
! declaration
!
?? text :: GA_decltxt(tname, tstring)
  !> Growing array type for ?tstring?
  type grw_?tname?Array_type
    integer :: nVals = 0
    integer :: ContainerSize = 0
    ?tstring?, allocatable :: Val(:)
  end type

  !> Initialize the dynamic array
  interface init
    module procedure init_GA_?tname?
  end interface

  !> Truncate the array, meaning
  !! cut off the trailing empty entries
  interface truncate
    module procedure truncate_GA_?tname?
  end interface

  !> Empty the entries  without changing arrays
  interface empty
    module procedure empty_GA_?tname?
  end interface

  !> Destroy the dynamic array
  interface destroy
    module procedure destroy_GA_?tname?
  end interface

  !> Append a value to the dynamic array
  !! and return its position.
  interface append
    module procedure insert_GA_?tname?
    module procedure append_GA_?tname?
    module procedure append_GA_?tname?_vec
  end interface

  !> Increase the size of the container
  !! for the array.
  interface expand
    module procedure expand_GA_?tname?
  end interface
?? end text GA_decltxt
!
! implementation
!
?? text :: GA_impltxt(tname, tstring)
  subroutine init_GA_?tname?(me, length)
    type(grw_?tname?Array_type), intent(out) :: me !< Dynamic Array to init
    integer, intent(in), optional :: length !< Initial length of the container

    if (present(length)) then
      me%ContainerSize = length
    else
      me%ContainerSize = zeroLength
    end if
    ! Deallocate ...
    if( allocated( me%Val ))     &
      deallocate(me%Val)
    ! ... and reallocate
    allocate(me%Val(me%ContainerSize))

  end subroutine init_GA_?tname?

  subroutine destroy_GA_?tname?(me)
    type(grw_?tname?Array_type), intent(inout) :: me !< Dynamic Array to init

    me%ContainerSize = 0 
    me%nVals = 0 
    if( allocated( me%Val ))     &
      deallocate(me%Val)

  end subroutine destroy_GA_?tname?


  subroutine truncate_GA_?tname?(me)
    !------------------------------------------------------------------------
    type(grw_?tname?Array_type) :: me !< Array to sortTruncate
    !------------------------------------------------------------------------
    ?tstring?, allocatable :: tArray(:)
    !------------------------------------------------------------------------

    ! Nothing to do if container size is not larger than the number of values
    ! in the array.
    if (me%containerSize > me%nVals) then
      allocate(tArray(me%nVals))
      tArray = me%val(:me%nVals)

      deallocate(me%val)
      me%containerSize = me%nVals
      allocate(me%val(me%containerSize))

      me%val = tArray
      deallocate(tArray)
    end if

  end subroutine truncate_GA_?tname?

  subroutine empty_GA_?tname?(me)
    !------------------------------------------------------------------------
    type(grw_?tname?Array_type) :: me !< Array to sortTruncate
    !------------------------------------------------------------------------

    me%nVals = 0

  end subroutine empty_GA_?tname?


  subroutine insert_GA_?tname?(me, val, pos)
    type(grw_?tname?Array_type) :: me !< Array to insert the value into
    ?tstring?, intent(in) :: val !< Value to insert
    integer, intent(in) :: pos !< optional predefined position


    ! Value to append is larger than all existing ones,
    ! just put it to the end of the list, this captures
    ! also the case of empty lists.
    ! In this case FoundPos = me%nVals + 1 holds.
    if (Pos > me%ContainerSize) then
      ! Expand the array, if its boundary is reached
      call expand(me, pos)
    end if

    me%nVals = max( Pos, me%nVals )
    me%val(Pos) = val

  end subroutine insert_GA_?tname?

  subroutine append_GA_?tname?(me, val)
    type(grw_?tname?Array_type) :: me !< Array to append the value to
    ?tstring?, intent(in) :: val !< Value to append

    ! Value to append is larger than all existing ones,
    ! just put it to the end of the list, this captures
    ! also the case of empty lists.
    ! In this case FoundPos = me%nVals + 1 holds.
    if (me%nVals+1 > me%ContainerSize) then
      ! Expand the array, if its boundary is reached
      call expand(me)
    end if

    me%nVals = me%nVals+1
    me%val(me%nVals) = val

  end subroutine append_GA_?tname?

  subroutine append_GA_?tname?_vec(me, val, pos)
    type(grw_?tname?Array_type) :: me !< Array to append the value to
    ?tstring?, intent(in) :: val(:) !< Values to append
    integer, intent(in), optional :: pos !< optional predefined position

    integer :: FoundPos
    integer :: ub
    integer :: n

    n = size(val)
    if( present( pos ) ) then
      ! Do it smart array style, just use FoundPos
      FoundPos = pos
    else
      FoundPos = me%nVals + 1
    end if
    ub = FoundPos+n-1

    ! Value to append is larger than all existing ones,
    ! just put it to the end of the list, this captures
    ! also the case of empty lists.
    ! In this case FoundPos = me%nVals + 1 holds.
    if (ub > me%ContainerSize) then
      ! Expand the array, if its boundary is reached
      call expand(me, ub)
    end if

    me%nVals = max( ub, me%nVals )
    me%val(FoundPos:ub) = val

  end subroutine append_GA_?tname?_vec


  subroutine expand_GA_?tname?(me, pos)
    type(grw_?tname?Array_type) :: me !< Array to resize
    integer, intent(in), optional :: pos !< optional predefined position 

    ?tstring?, allocatable :: swpval(:)

    if( allocated(me%Val)) then
      allocate(swpval(me%nVals))
      swpval(:me%nVals) = me%Val(:me%nVals)
      deallocate(me%Val)
    end if

    if( present( pos ) ) then
      me%ContainerSize = max(minLength, me%ContainerSize*2, pos)
    else
      me%ContainerSize = max(minLength, me%ContainerSize*2)
    endif
    allocate(me%Val(me%ContainerSize))

    if( allocated(swpval)) then
      me%Val(:me%nVals) = swpval(:me%nVals)
      deallocate(swpval)
    end if

  end subroutine expand_GA_?tname?
?? end text GA_impltxt
!
!------------------------------------------------------------------------------
!
! dynamic Arrays:
!
! declaration
!
?? text :: DA_decltxt(tname, tstring)
! \brief Smart dynamic array (DA) for ?tstring?
!
! This datatype implements a dynamic array, which is capable of
! growing and adding of unique elements. It is available for
! various types, here we deal with $tstring$.
! Sorted array contains the pointers to Val array, instead of
! the actual values in Val array. For example:
! Val:     8, 6, 7, 5
! sorted:  4, 2, 3, 1
  !> dynamic array (DA) type for ?tstring?
  type dyn_?tname?Array_type
    integer :: nVals = 0
    integer :: ContainerSize = 0
    logical :: Unique
    ?tstring?, allocatable :: Val(:)
    integer, allocatable :: sorted(:) !< pointers, NOT values
  end type

  !> Initialize the dynamic array
  interface init
    module procedure init_DA_?tname?
  end interface

  !> Destroy the dynamic array
  interface destroy
    module procedure destroy_DA_?tname?
  end interface

  !> Append a value to the dynamic array
  !! and return its position.
  interface append
    module procedure append_DA_?tname?
  end interface

  !> Truncate the array, meaning
  !! cut off the trailing empty entries
  interface truncate
    module procedure truncate_DA_?tname?
  end interface

  !> Truncate the array, meaning
  !>TODO: comments might be wrong
  !! cut off the trailing empty entries
  interface empty
    module procedure empty_DA_?tname?
  end interface


  !> Fix the dynamic array, meaning:
  !! store the array in the sorted order and cut
  !! off the trailing empty entries
  interface sortTruncate
    module procedure sortTruncate_DA_?tname?
  end interface

  !> Increase the size of the container
  !! for the array.
  interface expand
    module procedure expand_DA_?tname?
  end interface

  !> Return the position of a given value
  !! in the array val, which is what you usually want to know.
  !! It is the index of a given value
  interface PositionOfVal
    module procedure PosOfVal_?tname?
  end interface

  !> Return the position of a given value
  !! in the list 'sorted'. This is mainly for internal usage.
  !! The sorted list is only a pointer list to the actual values
  !! Thus, in order to get the index of a given value, you
  !! need to look up the entry in the sorted list.
  !! This is done by the PositionOfVal routine
  interface SortedPosOfVal
    module procedure SortPosOfVal_?tname?
  end interface
?? end text DA_decltxt
!
! implementation
!
?? text :: DA_impltxt(tname, tstring, argstring)
!******************************************************************************!
  !> Initialization of a dynamic array
  !!
  !! Before a dynamic array can be used, it has to be initialized
  !! with this routine. The initial length provided here, can
  !! avoid reallocations and memory copying, if approximated
  !! correctly enough. If none is specified, the provided container
  !! initially will be of size 0.
  subroutine init_DA_?tname?(me, unique, length)
    !-----------------------------------------------------------------
    type(dyn_?tname?Array_type), intent(out) :: me !< Dynamic Array to init
    logical, intent(in) :: unique !< Flag if the values have to be unique
    integer, intent(in), optional :: length !< Initial length of the container
    !-----------------------------------------------------------------

    if (present(length)) then
      me%ContainerSize = length
    else
      me%ContainerSize = zeroLength
    end if

    me%Unique = unique

    ! Deallocate ...
    if( allocated( me%Val ))     &
      deallocate(me%Val)
    if( allocated( me%sorted ))     &
      deallocate(me%sorted)
    ! ... and reallocate
    allocate(me%Val(me%ContainerSize))
    allocate(me%sorted(me%ContainerSize))

  end subroutine init_DA_?tname?

  !> Destruction of a dynamic array
  !!
  !! This subroutine takes care of a proper destruction of a
  !! dynamic array, it frees the allocated memory and resets
  !! the internal counts to 0.
  subroutine destroy_DA_?tname?(me)
    type(dyn_?tname?Array_type), intent(inout) :: me !< Dynamic Array to init

    me%ContainerSize = 0
    me%nVals         = 0

    if( allocated( me%Val ))     &
      deallocate(me%Val)
    if( allocated( me%sorted ))     &
      deallocate(me%sorted)
  end subroutine destroy_DA_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Appending a value to the dynamic array
  !!
  !! With this subroutine, a given value can be added to the
  !! dynamic array. The actual position of this value in the
  !! dynamic array will be returned, so it can be found again
  !! easily later. With the wasAdded flag, it is indicated,\n
  !! wasAdded = true,  if this entry had to be added,\n
  !! wasAdded = false, if this was already found in the array.
  subroutine append_DA_?tname?(me, val, pos, wasAdded )
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type) :: me   !< Array to append the value to
    ?argstring?, intent(in)     :: val  !< Value to append
    integer, intent(out)  :: pos !< Position in the array, the value is found
    !> Flag to indicate, if val was newly added
    logical, intent(out), optional :: wasAdded
    !------------------------------------------------------------------------
    integer :: FoundPos
    integer :: i
    !------------------------------------------------------------------------

    ! Do a binary search on existing entries (returns closest entry next to
    ! it if not found).
    FoundPos = SortedPosOfVal(me, val, .true.)
    if( present( wasAdded ) ) wasAdded = .false.

    ! If it found the value, the position is smaller than nVals
    if (FoundPos <= me%nVals) then

      ! The returned position might actually be the right entry already or
      ! not, check for this here.
      if ((me%val(me%sorted(FoundPos)) == val) &
        & .and. me%Unique) then

        ! Found the value in a list of unique values,
        ! nothing to do, just return its position.
        pos = me%sorted(FoundPos)

      else

        ! Need to append a new value!

        if( present( wasAdded ) ) wasAdded = .true.
        if (me%nVals == me%ContainerSize) then

          ! Container is full, need to expand it
          call expand(me)
        end if
        me%nVals = me%nVals + 1

        ! Put the new value into the last position in the
        ! array.
        me%val(me%nVals) = val
        do while( (FoundPos < me%nVals) &
          & )
          if(me%val(me%sorted(FoundPos)) /= val) then
           exit
          end if
          ! In case of multiple entries with the same value
          ! move on to the first differing entry.
          FoundPos = FoundPos + 1
        end do
        ! Shift the sorted list of indices, to create a
        ! whole for the value to be inserted, at position
        ! FoundPos.
        do i=me%nVals-1,FoundPos,-1
          me%sorted(i+1) = me%sorted(i)
        end do
        ! Put the index of the new value into the
        ! sorted list at the now freed position.
        me%sorted(FoundPos) = me%nVals
        pos = me%nVals

      end if

    else

      ! Value to append is larger than all existing ones,
      ! just put it to the end of the list, this captures
      ! also the case of empty lists.
      ! In this case FoundPos = me%nVals + 1 holds.
      if( present( wasAdded ) ) wasAdded = .true.
      if (FoundPos > me%ContainerSize) then
        ! Expand the array, if its boundary is reached
        call expand(me)
      end if
      me%nVals = FoundPos
      me%val(FoundPos) = val
      me%sorted(FoundPos) = FoundPos
      pos = FoundPos

    end if

  end subroutine append_DA_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Truncate the array after the last valid entry and hence cut off the empty
  !! trailing empty entries
  !!
  subroutine truncate_DA_?tname?(me)
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type) :: me !< Array to sortTruncate
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type) :: tArray ! temporary Array
    integer :: iVal
    integer :: dPos
    !------------------------------------------------------------------------
    ! Allocate the temporary array
    call init( me = tArray, length = me%nVals, unique = me%unique )
    ! Copy the entries in a sorted fashion into the temporary array
    do iVal = 1, me%nVals
      call append( me = tArray, val = me%val( iVal ), &
           &       pos = dPos)
    enddo
    call destroy( me = me )

    me = tArray
    call destroy( me = tArray )

  end subroutine truncate_DA_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Empty all contents of the array without changing the size or status of any
  !! array
  !!
  subroutine empty_DA_?tname?(me)
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type) :: me !< Array to sortTruncate
    !------------------------------------------------------------------------
    ! Reset the number of entries 
    me%nVals = 0
  end subroutine empty_DA_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Fixing the dynamic array
  !!
  !! Truncate the array after the last valid entry and hence cut off the empty
  !! trailing empty entries
  !! Store the array in the sorted order according to the sorted( ) array
  !!
  subroutine sortTruncate_DA_?tname?(me)
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type) :: me !< Array to sortTruncate
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type) :: tArray !< temporary Array
    integer :: iVal
    integer :: dPos
    !------------------------------------------------------------------------
    ! Allocate the temporary array
    call init( me = tArray, length = me%nVals, unique = me%unique )
    ! Copy the entries in a sorted fashion into the temporary array
    do iVal = 1, me%nVals
      call append( me = tArray, val = me%val( me%sorted( iVal )), &
           &       pos = dPos)
    enddo
    call destroy( me = me )

    me = tArray
    call destroy( me = tArray )

  end subroutine sortTruncate_DA_?tname?
!******************************************************************************!




!******************************************************************************!
  !> Expanding the dynamic array
  !!
  !! This is a helping subroutine, which doubles the container
  !! of the given dynamic array. As the container might be
  !! initially 0-sized, a module variable minLength has been introduced, which
  !! is used here, to at least create a container of this size.
  subroutine expand_DA_?tname?(me)
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type) :: me !< Array to resize
    !------------------------------------------------------------------------
    ?tstring?, allocatable :: swpval(:)
    integer, allocatable :: swpsort(:)
    !------------------------------------------------------------------------
    allocate(swpval(me%nVals))
    swpval = me%Val(1:me%nVals)
    if( allocated( me%val)) &
    deallocate(me%Val)

    me%ContainerSize = max(minLength, me%ContainerSize*2)
    allocate(me%Val(me%ContainerSize))
    me%Val(1:me%nVals) = swpval
    deallocate(swpval)

    allocate(swpsort(me%nVals))
    swpsort = me%sorted(:me%nVals)
    if( allocated( me%sorted)) &
      deallocate(me%sorted)
    allocate(me%sorted(me%ContainerSize))
    me%sorted(1:me%nVals) = swpsort
    deallocate(swpsort)
  end subroutine expand_DA_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Return the sorted position of a value in the given dynamic array
  !!
  !! If the value was not found,
  !!  - return 0 if nextIfNotFound = .false.
  !!  - return position at the end if nextIfNotFound = .true.
  function SortPosOfVal_?tname?(me, val, nextIfNotFound, lower, upper) result(pos)
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type), intent(in) :: me !< Dynamic Array
    ?argstring?, intent(in) :: val !< Value to look for
    !> Flag to indicate, if the next entry in the list should be returned,
    !! if the searched one is not found.
    logical, intent(in), optional :: nextIfNotFound
    integer, intent(in), optional :: lower !< lower search limit
    integer, intent(in), optional :: upper !< upper search limit
    integer :: pos !< Position of val in the sorted list, 0 if not found
    !------------------------------------------------------------------------
    logical :: retNext
    integer :: lb, ub
    integer :: mid
    ?tstring? :: lb_val, ub_val
    ?tstring? :: mid_val
    !------------------------------------------------------------------------

    retNext = .false.
    if (present(nextIfNotFound)) retNext = nextIfNotFound

    lb = 1
    ub = me%nVals

    if( present( lower ) ) lb = lower
    if( present( upper ) ) ub = upper

    pos = 0
    if (retNext) pos = lb

    !> Binary search on sorted list
    do while(ub >= lb)
      lb_val = me%Val(me%sorted(lb))

      if (val < lb_val) then
        if (retNext) pos = lb
        exit
      end if

      ub_val = me%Val(me%sorted(ub))

      if (val > ub_val) then
        if (retNext) pos = ub+1
        exit
      end if

      mid = (lb+ub) / 2
      mid_val = me%Val(me%sorted(mid))
      if (val == mid_val) then
        pos = mid
        exit
      end if
      if (val > mid_val) then
        lb = mid + 1
      else
        ub = mid - 1
      end if
    end do
  end function SortPosOfVal_?tname?
!******************************************************************************!


!******************************************************************************!
  !> The actual position of a given value in the dynamic array
  !!
  !! Most likely this, is what you need in codes, using this
  !! data structure, it first does the binary search on the sorted
  !! values with SortPosOfVal_?tname? and then returns the looked
  !! up position in the original unsorted array, which corresponds
  !! to the position returned by the append routine.
  function PosOfVal_?tname?(me, val, nextIfNotFound, lower, upper) result(pos)
    !------------------------------------------------------------------------
    type(dyn_?tname?Array_type), intent(in) :: me !< Dynamic Array
    ?argstring?, intent(in) :: val !< Value to search for
    !> Flag to indicate, if the position of the next entry in the sorted
    !! list should be returned instead, if val is not found.
    logical, intent(in), optional :: nextIfNotFound
    integer, intent(in), optional :: lower !< lower search limit
    integer, intent(in), optional :: upper !< upper search limit
    integer :: pos !< Position in the array of the searche value, 0 if not found
    !------------------------------------------------------------------------
    integer :: sortpos
    integer :: lb, ub
    !------------------------------------------------------------------------

    lb = 1
    ub = me%nVals

    if( present( lower ) ) lb = lower
    if( present( upper ) ) ub = upper

    sortpos = SortedPosOfVal(me, val, nextIfNotFound, lb, ub)

    ! If result (sorted pos)
    if ((sortpos <= me%nVals) .and. (sortpos > 0)) then
      pos = me%sorted(sortpos)
    else
      pos = sortpos
    end if
  end function PosOfVal_?tname?
!******************************************************************************!
?? end text DA_impltxt
! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! Copyright 2012 German Research School for Simulation Sciences GmbH, Aachen.
! All rights reserved.
! 
! Redistribution and use in source and binary forms, with or without
! modification, are permitted provided that the following conditions are met:
! 
! 1. Redistributions of source code must retain the above copyright notice, this
! list of conditions and the following disclaimer.
! 
! 2. Redistributions in binary form must reproduce the above copyright notice,
! this list of conditions and the following disclaimer in the documentation
! and/or other materials provided with the distribution.
! 
! THIS SOFTWARE IS PROVIDED BY GERMAN RESEARCH SCHOOL FOR SIMULATION SCIENCES
! GMBH, AACHEN “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
! LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
! PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GERMAN RESEARCH SCHOOL
! FOR SIMULATION SCIENCES GMBH, AACHEN OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
! INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
! BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
! DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
! OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
! NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
! EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
! 
! The views and conclusions contained in the software and documentation are
! those of the authors and should not be interpreted as representing official
! policies, either expressed or implied, of German Research School for
! Simulation Sciences GmbH, Aachen.
! !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
