!******************************************************************************!
!> This module adapts(refines or coarsens) the mesh dynamically during the 
!! simulation. The newly created elements are appended to the level descriptor.
!! The already existing helper elements can be converted to fluid in case of
!! dynamic refinement. While coarsening, 8 fluid elements will turn into one 
!! coarse element, and thus the "freed" elements will be appended to ghost/halo
!! list. 

!! There will be a destruction/cleaning up routine which will destroy all the 
!! extra elements which were introduced during this process.

module tem_adaptation_module

  ! include treelm modules
  use mpi
  use env_module,               only: rk, long_k
  use treelmesh_module,         only: treelmesh_type
  use tem_topology_module,      only: tem_directChildren, tem_levelOf,         &
    &                                 tem_PathOf, tem_path_type, tem_coordOfId
  use tem_stencil_module,       only: tem_stencilHeader_type
  use tem_construction_module,  only: tem_levelDesc_type                         
  use tem_element_module,       only: tem_element_type, tem_element_dump,      &
    &                                 init, append, PositionOfVal, truncate,   &
    &                                 eT_fluidElem, eT_nonExistingElem,        &
    &                                 eT_haloElem, destroy,                    &
    &                                 eT_ghostFromCoarserElem, eT_minNumber,   &
    &                                 eT_ghostFromFinerElem, eT_maxNumber,     &
    &                                 eT_minNumRelevant, eT_undefinedElem,     &
    &                                 eT_distributedGhostFromFinerElem
  use tem_geometry_module,      only: tem_posOfId
  use tem_comm_env_module,      only: tem_comm_env_type
  use tem_construction_module,  only: identify_stencilNeigh 

  ! include musubi modules
!HK: treelm can not depend on Solver routines!
!HK!  use mus_scheme_type_module,   only: mus_scheme_type
!HK!  use mus_geom_module,          only: mus_geom_type

  implicit none
  
  private

  public :: tem_adapt_refine
!  public :: tem_adapt_coarsen

  contains

!*******************************************************************************
  !> This routine converts a coarse element into 8 fine elements dynamically
  subroutine tem_adapt_refine(treeID, level, levelDesc, proc, computeStencil, &
    &                         tree )
    !---------------------------------------------------------------------------
    integer(kind=long_k)        , intent(in)        :: treeID
    integer                     , intent(in)        :: level
    type(tem_levelDesc_type)    , intent(inout)     :: levelDesc(:)
    type(tem_comm_env_type)     , intent(in)        :: proc
    type(tem_stencilHeader_type), intent(in)        :: computeStencil
    !> Treelmesh data
    type(treelmesh_type), intent(inout) :: tree
    !! HK: Nope obviously this is some musubi information...
!HK!    type( mus_geom_type )       , intent(inout)     :: geometry
    !---------------------------------------------------------------------------
    integer(kind=long_k) :: children(8)
    integer              :: iChild, iStencilElem
    integer              :: tPos, neighIDpos
    integer              :: posOfId
    integer              :: neightID
    logical              :: wasAdded
    type(tem_path_type), allocatable :: pathFirst(:), pathLast(:)
    !---------------------------------------------------------------------------
    wasAdded             = .false.
    ! first and last path in every process
    ! which helps in determining the process location of a given treeID
    allocate(pathFirst( tree%global%nParts ))
    allocate(pathLast( tree%global%nParts ))
    ! Fill the pathList for each element in the treeID list
    pathFirst     =  tem_PathOf( tree%Part_First )
    pathLast      =  tem_PathOf( tree%Part_Last )
    !---------------------------------------------------------------------------
    ! Convert the sacrificed element to a ghost and set its property
    posOfId  = tem_posOfId(treeID, levelDesc(level)%elem%tID%val)
write(*,*) 'POS OF ID IS', posOfID    
    if (posOfId > 0) then
      levelDesc(level)%elem%eType%val(posOfId) = eT_ghostFromCoarserElem
    end if

    ! Get the kids of sacrificed treeID 
    children = tem_directChildren( TreeID )

    ! Append the newly created children to the end of LD
!write(*,*) 'SIZE OF LD BEFORE', levelDesc(1)%elem%tID%nVals
!write(*,*) 'SIZE OF LD BEFORE', levelDesc(2)%elem%tID%nVals
    do iChild = 1, 2
      call append( me         = levelDesc(level+1)%elem,                   &
        &          tID        = children(iChild),                          &
        &          eType      = eT_fluidElem,                              &
        &          pos        = tPos,                                      &
        &          wasAdded   = wasAdded )
write(*,*) 'child, tpos', children(iChild), tpos
write(*,*) 'SIZE OF ELEM STENCIL', size(levelDesc(level)%elem%stencil%val)
write(*,*) 'SIZE OF ELEM STENCIL', size(levelDesc(level+1)%elem%stencil%val)
      call identify_stencilNeigh( iElem          = tPos,               &
        &                         iLevel         = level+1,          &
        &                         tree           = tree,    &
        &                         iStencil       = 1,                &
        &                         pathFirst      = pathFirst,        &
        &                         pathLast       = pathLast,         &
        &                         levelDesc      = levelDesc,        &
        &                         proc           = proc,             &
        &                         computeStencil = computeStencil,   &
        &                         nesting        = 0 )
      end do
!write(*,*) 'SIZE OF LD AFTER', levelDesc(1)%elem%tID%nVals
!write(*,*) 'SIZE OF LD AFTER', levelDesc(2)%elem%tID%nVals
!write(*,*) 'TPOS'            , tPos
    ! Build the neighboring stencil of added children
write(*,*) 'CALLING IDENTIFY STENCIL NEIGH'    



  end subroutine tem_adapt_refine
!*******************************************************************************
end module tem_adaptation_module
