! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the datatypes for load balancing algorithms.
!!
!! \author Daniel Harlacher, Manuel Hasert, Kartik Jain, Simon Zimny
!!
module tem_balance_module

  ! include treelm modules
  use mpi
  use env_module,             only: rk, long_k, rk_mpi, labelLen,              &
    &                               tem_connect_toNull
  use flu_binding,            only: flu_State
  use tem_time_module,        only: tem_time_never
  use tem_timeControl_module, only: tem_timeControl_type, tem_TimeControl_load
  use tem_tools_module,       only: tem_horizontalSpacer
  use tem_logging_module,     only: tem_logging_type, logUnit, tem_last_lu,    &
    &                               tem_logging_load, tem_logging_init
  use tem_restart_module,     only: tem_restart_type
  use tem_tracking_module,    only: tem_trackingHeader_type, tem_tracking_type
  use tem_aux_module,         only: tem_abort
  use tem_timer_module,       only: tem_timer_type, tem_addTimer

  ! include aotus modules
  use aotus_module,     only: aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val

  implicit none
  private

  public :: tem_balance_type
  public :: tem_balance_evaluateImbalance
  public :: tem_balance_loadConfig
  public :: tem_balance_initTracks

  !> log units for the load balancing
  integer, save, public :: Bal_logUnit(0:tem_last_lu)

  type tem_trackingHeaders_type
    ! trackingHeader for each tracking object
    ! size: nTracks
    type( tem_trackingHeader_type ), allocatable :: header(:)
  end type tem_trackingHeaders_type

  !> Load balancing information and control 
  type tem_balance_type
    !> load balancing type
    character(len=labelLen) :: balanceKind
    !> is static load balancing activated?
    logical :: static
    !> is dynamic load balancing activated?
    logical :: dynamic
    !> Imbalance: max( timeMainLoop( iProc )) - average( timeMainLoop )
    real( kind=rk ) :: imbalanceAbs
    !> Imbalance: max( timeMainLoop( iProc )) - average( timeMainLoop )/average
    real( kind=rk ) :: imbalanceRel
    !> base weight for each element which is set to be
    !! weight( element ) = baseWeight^( level(element) -refLevel )
    real( kind=rk ) :: baseWeight
    !> control about when to do dynamic balancing
    type(tem_timeControl_type) :: timeControl
    !> logging type for debug output to disc (i.e. weights)
    type( tem_logging_type ) :: logBal
    !> restart type for dumping the intermediate restarts
    type( tem_restart_type ) :: restart
    !> integer defining the suffix for the dyn load balancing restart files
    integer :: fileID
    !> trackingHeaders for all schemes and tracking objects
    !! size: nSchemes
    type( tem_trackingHeaders_type ), allocatable :: trackHeaders(:)
    !> logging type for dumping the timings for the individual slots
    type( tem_logging_type ) :: logger
    !> last iteration the dynamic load balancing was called
    integer :: lastIter = 0
    !> timer type for the dynamic load balancing
    type( tem_timer_type ) :: timings
    !> timer handle for the last slot of dyn LB
    integer :: slotTimer
    !> current slot
    integer :: iSlot
  end type tem_balance_type

  contains

!******************************************************************************!
  !> Read all the configuration options for load balancing from the
  !! configuration.
  !!
  !! To activate the load balancing, use 
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! balance = {
  !!   static  = false,         -- perform static load balancing
  !!   dynamic = true,          -- perform dynamic load balancing
  !!   kind = 'IBM',            -- kind = 'levelwise','boundary','IBM'
  !!   time_control = {         -- time control for slots when to perform load
  !!        min      = 0.01,    --   balancing
  !!        interval = 0.01,
  !!        max      = 1.0 
  !!   },
  !!   balanceDir = 'balance/', -- directory where to store tmp data
  !!   logging    = {           -- logging table for load balancing debug
  !!                 ...        --   output (see tem_logging_module)
  !!   }
  !! }
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! 
  subroutine tem_balance_loadConfig(me, conf, rank, parent, key, defaultActive )
    !---------------------------------------------------------------------------
    !> The load balancing type
    type(tem_balance_type), intent(inout) :: me
    !> lua configuration handle
    type(flu_state)                       :: conf
    !> the current rank
    integer, intent(in) :: rank
    !> parent of the balance table
    integer, intent(in), optional         :: parent
    !> key in the lua, if not balance
    character(len=*),intent(in), optional :: key
    !> The default option for activating load balancing 
    !! if nothing is given in the lua file
    logical, intent(in), optional         :: defaultActive
    !---------------------------------------------------------------------------
    integer :: iError, thandle, log_handle
    logical :: local_defaultActive
    character(len=labelLen) :: localKey
    character(len=labelLen) :: local_defaultString
    !---------------------------------------------------------------------------
    if( present(key) ) then
      localKey = key
    else
      localKey = 'balance'
    endif
    if( present(defaultActive) ) then
      local_defaultActive = defaultActive
    else
      local_defaultActive = .false.
    endif
    local_defaultString = 'simple'
    me%dynamic = .false.

    call aot_table_open(L       = conf,                                        &
      &                 parent  = parent,                                      &
      &                 thandle = thandle,                                     &
      &                 key     = localKey )
    ! Check here if the table was successfully opened. 
    ! If not, try to read the keyword directly
    if( thandle .ne. 0 ) then
      ! table is open !
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'static',                                    &
        &               val     = me%static,                                   &
        &               ErrCode = iError,                                      &
        &               default = local_defaultActive )

      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'dynamic',                                   &
        &               val     = me%dynamic,                                  &
        &               ErrCode = iError,                                      &
        &               default = .false. )
    
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'kind',                                      &
        &               val     = me%balanceKind,                              &
        &               ErrCode = iError,                                      &
        &               default = local_defaultString )

      me%restart%controller%writeRestart = .false.
      me%restart%controller%readRestart  = .false.

      ! read the directory for dumping restart files of load balancing
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               key     = 'balanceDir',                                &
        &               val     = me%restart%controller%writePrefix,           &
        &               ErrCode = iError )

      me%restart%controller%writeRestart = (iError == 0)

      ! if the balance directory is provided ...
      if( me%restart%controller%writeRestart )then
        write(logUnit(3),*)'Balance directory: '//                             &
          &                trim(me%restart%controller%writePrefix)
      else
        ! ... no dyn balancing can be done, stop the code
        write(logUnit(0),*)'No balanceDir provided!'
        write(logUnit(0),*)'STOPPING'
        call tem_abort()
      end if

   
      ! Load the time controls for dynamic load balancing
      call tem_timeControl_load( me     = me%timeControl,                      &
        &                        conf   = conf,                                &
        &                        parent = thandle )

      ! open the logging table if available
      call aot_table_open( L       = conf,                                     &
        &                  parent  = thandle,                                  &
        &                  thandle = log_handle,                               &
        &                  key     = 'logging' )

      if( log_handle > 0 )then

        ! load the logging type
        call tem_logging_load( conf    = conf,                                 &
          &                    thandle = log_handle,                           &
          &                    rank    = rank,                                 &
          &                    me      = me%logBal )
        ! now set the module variable Bal_logUnits for the Bal logging type
        Bal_logUnit = me%logBal%fUnit
      else
        ! connect the module variable to the null device
        call tem_connect_toNull(Bal_logUnit(0))
        Bal_logUnit(1:tem_last_lu) = Bal_logUnit(0)
      end if
      call aot_table_close( L = conf, thandle = log_handle )

    else
      ! Now test if balance is given as a logical 
      call aot_get_val( L       = conf,                                        &
        &               key     = localKey,                                    &
        &               val     = me%static,                                   &
        &               ErrCode = iError,                                      &
        &               default = local_defaultActive )
      ! Was it given as a logical?
      if( iError .ne. 0 ) then
        ! There was an error, so not given as logical
        call aot_get_val( L       = conf,                                      &
          &               key     = localKey,                                  &
          &               val     = me%balanceKind,                            &
          &               ErrCode = iError,                                    &
          &               default = local_defaultString )
      end if
      ! connect the module variable to the null device
      call tem_connect_toNull(Bal_logUnit(0))
      Bal_logUnit(1:tem_last_lu) = Bal_logUnit(0)
    end if
    call aot_table_close(L=conf, thandle=thandle)
    ! If the balance type is none, deactivate it
    if( trim( me%balanceKind ) == 'none' )  then
      me%static  = .false.
      me%dynamic = .false.
    end if

    if ( .not. me%dynamic ) then
      me%timeControl%Interval = tem_time_never()
      me%timeControl%Min      = tem_time_never()
      me%timeControl%Max      = tem_time_never()
    end if

    if ( me%static .or. me%dynamic ) then
      write(logUnit(1),*) 'Load Balancing is active with kind '//              & 
        &                    trim( me%balanceKind )
    endif

    ! add the timer for the last dynLB slot
    call tem_addTimer( me          = me%timings,                               &
      &                timerHandle = me%slotTimer,                             &
      &                timerName   = 'lastSlot' )

    if ( me%dynamic )then
      ! initialize the logger
      ! \todo: maybe get the name from the balanceDir??
      call tem_logging_init( me    = me%logger,                                &
        &                    level = 10,                                       &
        &                    rank  = 0,                                        &
        &                    filename = trim(me%restart%controller%writePrefix)&
        &                                                  //'MLUPS_dynLB.res' )
    end if

  end subroutine tem_balance_loadConfig
!******************************************************************************!


!******************************************************************************!
  !> This routine copies the global tracking information to the tracking header
  !! type in the tem_balance_type.
  !!
  !! ATTENTION: Before this routine works as expected the array 'trackHeaders'
  !!            has to be allocated (for MUSUBI: with nSchemes) and called
  !!            BEFORE (!!!!) tem_init_tracker, since this call modifies the
  !!            number tracking objects (and by this their headers).
  !!
  subroutine tem_balance_initTracks( me, track )
    !---------------------------------------------------------------------------
    !> The tracking headers to be copied to
    type(tem_trackingHeader_type), allocatable, intent(inout) :: me(:)
    !> tracking entities incl headers to be copied
    type(tem_tracking_type), allocatable, intent(in)          :: track(:)
    !---------------------------------------------------------------------------
    integer :: iLog
    !---------------------------------------------------------------------------

    if( allocated( track ))then
      allocate( me( size( track )))
      do iLog = 1, size( track )
        me( iLog ) = track( iLog )%header
      end do
    end if

  end subroutine tem_balance_initTracks
!******************************************************************************!


!******************************************************************************!
  !> Evaluate the imbalance of all the processes by each rank.
  !!
  subroutine tem_balance_evaluateImbalance( me, CPUcost, nParts, comm)
    !---------------------------------------------------------------------------
    !> Load balance type to store the result to
    type(tem_balance_type), intent(inout) :: me
    !> Number of partitions, the mesh is partitioned into (= Number of MPI
    !! processes in comm).
    integer, intent(in) :: nParts
    !> MPI Communicator
    integer, intent(in) :: comm
    !> each process's cpu cost. Basis to evaluate the imbalance
    real(kind=rk), intent(in) :: CPUcost
    !---------------------------------------------------------------------------
    ! computation cost over all processes
    real(kind=rk) :: totalCPUcost
    ! maximum computation cost in one rank
    real(kind=rk) :: maxCPUcost
    ! average computation cost over all processes
    real(kind=rk) :: average
    integer :: iErr ! error handle
    !---------------------------------------------------------------------------

    ! Get the sum of total CPU cost
    call mpi_reduce( CPUcost, totalCPUcost, 1, rk_mpi, mpi_sum, 0, comm, iErr)
    average = totalCPUcost/real( nParts, kind=rk )

    ! Get the maximum process CPU cost
    call mpi_reduce( CPUcost, maxCPUcost, 1, rk_mpi, mpi_max, 0, comm, iErr)
    me%imbalanceAbs = maxCPUcost - average
    if( average .gt. 0._rk ) then
      me%imbalanceRel = (maxCPUcost - average)/average
    else
      me%imbalanceRel = 0._rk
    end if

    if( nParts .gt. 1 ) then
      call tem_horizontalSpacer(fUnit=logUnit(1))
      write(logUnit(1),*)' Load-balancing evaluation '
      write(logUnit(1),*)'   max. absolute imbalance  [s]:',                   &
        &                real(me%imbalanceAbs )
      write(logUnit(1),*)'   max. relative imbalance  [%]:',                   &
        &                real(me%imbalanceRel*100._rk )
      call tem_horizontalSpacer(fUnit=logUnit(1))
    end if

  end subroutine tem_balance_evaluateImbalance
!******************************************************************************!


end module tem_balance_module
!******************************************************************************!
