! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the user with a simple geometrical object like point,
!! line, plane and box.
!!
!! \author Simon Zimny, Kartik Jain
!!
!! Detail description of the canonical shapes can be found
!! [here] (@ref canoND_shapes).
!!
module tem_canonicalND_module

  ! include treelm modules
  use env_module,            only: rk, long_k, globalMaxLevels, labelLen
  use tem_aux_module,        only: tem_abort
  use treelmesh_module,      only: treelmesh_type
  use tem_geometry_module,   only: tem_CoordOfReal, tem_PosOfId
  use tem_dyn_array_module,  only: dyn_intArray_type, append
  use tem_topology_module,   only: tem_levelOf, tem_IDofCoord
  use tem_logging_module,    only: logUnit

  ! include aotus modules
  use aotus_module,     only: aot_get_val, aoterr_Fatal, aoterr_WrongType,     &
    &                         flu_State
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open_table, aot_out_close_table

  implicit none
  private

  public :: tem_canonicalND_type
  public :: tem_load_canonicalND
  public :: tem_getNextCoordOfcanonicalND
  public :: tem_canonicalND_out
  public :: tem_cano_initSubTree

  !> Definition of the canonicalND
  type tem_canonicalND_type
    real(kind=rk) :: origin(3)   !< origin of the canonicalND
    !< vector along the edge A (also defines size)
    !!1st dimension defines x,yz coord
    !! 2nd dimension vec number
    real(kind=rk) :: vec(3,3)    
    !> how many discrete points the canonicalND is divided into
    integer ::       segments(3)    
    !> spatial distribution of the points
    character(len=labellen) :: distribution 
    !> kind of canonicalND (line, plane, point, box)
    character( len=labellen ) :: kind
    logical :: active(3)     !< identify which vectors are active (not equal 0)
    !> dimension of canonical object 
    !! nDim=0 - point
    !! nDim=1 - line
    !! nDim=2 - plane
    !! nDim=3 - box
    integer :: nDim
    !> To choose what to do with intersection of box
    !! if only_surface = true than the only the surface of the object
    !! is intersected
    !! if only_surface = false then the whole object is intersected
    !! default is set to false
    logical :: only_surface
  end type tem_canonicalND_type

  !> interface to write out canonical shape(s) in lua format to a file 
  interface tem_canonicalND_out
    module procedure tem_canonicalND_out_scal
    module procedure tem_canonicalND_out_vec
  end interface tem_canonicalND_out

  !> interface to load canonical objects
  interface tem_load_canonicalND
    module procedure tem_load_onecanonicalND
    module procedure tem_load_canonicalND_vec
  end interface tem_load_canonicalND

  contains

!******************************************************************************!
  !> Loading canonicalNDs from the config file valid definitions:
  !!
  subroutine tem_load_canonicalND_vec( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> the array of canonical objects which to read in (and first allocate)
    type(tem_canonicalND_type), allocatable, intent(out) :: me(:)
    !> lua config handle
    type(flu_state) :: conf 
    !> table handle from which to read
    integer, intent(in) :: thandle 
    !---------------------------------------------------------------------------
    ! lua handles
    integer :: canonicalNDs_thandle, single_thandle
    integer :: canonicalND_entries
    integer :: iCan
    integer :: ncanonicalNDs
    !---------------------------------------------------------------------------

    write(logUnit(3),*) 'Trying to read canoND shape...'


    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = canonicalNDs_thandle,                       &
      &                  key     = 'object' )
    canonicalND_entries = aot_table_length( L       = conf,                    &
      &                                     thandle = canonicalNDs_thandle )

    ! Now check, if the first entry is a table itself:
    call aot_table_open( L       = conf,                                       &
      &                  parent  = canonicalNDs_thandle,                       &
      &                  thandle = single_thandle,                             &
      &                  pos     = 1 )

    if (single_thandle == 0) then
      ! The entries are not tables themselves, try to interpret it as
      ! single canonical object.
      call aot_table_close( L = conf, thandle = single_thandle )
      ncanonicalNDs = 1
      allocate(me( ncanonicalNDs ))
      call tem_load_oneCanonicalND( me      = me(1),                           &
        &                           conf    = conf,                            &
        &                           thandle = canonicalNDs_thandle )
    else
      ! First entry is a table, assume all others are also tables,
      ! and properly define canonicalNDs coordinates, go on reading them.
      call aot_table_close( L = conf, thandle = single_thandle )
      ncanonicalNDs = canonicalND_entries
      allocate( me( ncanonicalNDs ))
      do iCan=1, ncanonicalNDs
        write(logUnit(3),*) 'Read Object: ', iCan, ' with ' 
        call aot_table_open( L       = conf,                                   &
          &                  parent  = canonicalNDs_thandle,                   &
          &                  thandle = single_thandle,                         &
          &                  pos     = iCan )
        call tem_load_oneCanonicalND( me      = me( iCan ),                    &
          &                           conf    = conf,                          &
          &                           thandle = single_thandle )
        call aot_table_close( L = conf, thandle = single_thandle )
      end do
    end if
    
    call aot_table_close( L = conf, thandle = canonicalNDs_thandle )


  end subroutine tem_load_canonicalND_vec
!******************************************************************************!


!******************************************************************************!
  !> Read one canonical object definition into a tem_canonicalND_type from a lua
  !! table.
  !!
  subroutine tem_load_oneCanonicalND( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> contains canonicalND data
    type(tem_canonicalND_type), intent(out) :: me 
    !> lua state
    type(flu_state) :: conf 
    !> lua table identification
    integer, intent(in) :: thandle 
    !> output messages?
    !---------------------------------------------------------------------------
    ! error identifiers
    integer :: iError
    integer :: vError(3), errfatal(3)
    ! counter and total number of vectors
    integer :: iVec, nVecs   
    ! lua handles
    integer :: vec_thandle, halfvec_thandle, single_thandle, halfsingle_thandle
    real(kind=rk), allocatable :: local_vecs(:,:)
    real(kind=rk) :: local_length
    real(kind=rk) :: vecAbsSum
    character(len=labelLen) :: buffer
    !---------------------------------------------------------------------------

    errfatal = aotErr_Fatal

    ! initialize the array of vectors and the array of segments
    me%vec = 0._rk

    ! open the vec and halvec table and get its size
    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = vec_thandle,                                &
      &                  key     = 'vec' )
    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = halfvec_thandle,                            &
      &                  key     = 'halfvec' )
    ! check if the vec table is given
    if( vec_thandle .ne. 0 ) then
      nVecs = aot_table_length( L = conf, thandle = vec_thandle )

      ! Now check, if the first entry is a table itself:
      call aot_table_open( L       = conf,                                     &
        &                  parent  = vec_thandle,                              &
        &                  thandle = single_thandle,                           &
        &                  pos     = 1 )

      if (single_thandle == 0) then
        ! The entries are not tables themselves, try to interpret it as
        ! single vector.
        call aot_table_close( L = conf, thandle = single_thandle )
        nVecs = 1
        allocate(local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               val     = local_vecs(:,1),                           &
          &               key     = 'vec',                                     &
          &               ErrCode = vError )
        if( any( btest( vError, errFatal ))) then
          write(logUnit(1),*)' FATAL ERROR: in loading vec'
          write(logUnit(1),*)' You should provide 3 numbers for this vector.'
          call tem_abort()
        end if
      else
        ! First entry is a table, assume all others are also tables,
        ! and properly define canonicalNDs coordinates, go on reading them.
        call aot_table_close( L = conf, thandle = single_thandle )
        allocate( local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        do iVec=1, nVecs
          call aot_get_val( L       = conf,                                    &
            &               thandle = vec_thandle,                             &
            &               val     = local_vecs(:,iVec),                      &
            &               pos     = iVec,                                    &
            &               ErrCode = vError )
        end do
      end if
      if( any( btest( vError, errFatal ))) then
        write(logUnit(1),*) ' FATAL ERROR: in loading vec' 
        call tem_abort()
      endif
    ! if vec table is not set try to read halfvec
    else if( halfvec_thandle .ne. 0 ) then
      call aot_table_open( L       = conf,                                     &
        &                  parent  = halfvec_thandle,                          &
        &                  thandle = halfsingle_thandle,                       &
        &                  pos     = 1 )
      nVecs = aot_table_length( L = conf, thandle = halfvec_thandle )
      if( halfsingle_thandle == 0 ) then
        ! The entries are not tables themselves, try to interpret it as
        ! single vector.
        call aot_table_close( L = conf, thandle = halfsingle_thandle )
        nVecs = 1
        allocate( local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               val     = local_vecs(:,1),                           &
          &               key     = 'halfvec',                                 &
          &               ErrCode = vError )
        local_vecs = local_vecs * 2._rk
      else
        ! First entry is a table, assume all others are also tables,
        ! and properly define canonicalNDs coordinates, go on reading them.
        call aot_table_close( L = conf, thandle = halfsingle_thandle )
        allocate( local_vecs( 3, nVecs ))
        local_vecs = 0._rk
        do iVec=1, nVecs
          call aot_get_val( L       = conf,                                    &
            &               thandle = halfvec_thandle,                         &
            &               val     = local_vecs(:,iVec),                      &
            &               pos     = iVec,                                    &
            &               ErrCode = vError )
        end do
        local_vecs = local_vecs * 2._rk
      end if
      if( any( btest( vError, errFatal ))) then
        write(logUnit(1),*) ' FATAL ERROR: in loading halfvec' 
        call tem_abort()
      endif
    ! if vec and halfvec table is not set, try to read length (for a cube)
    else
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = local_length,                                &
        &               key     = 'length',                                    &
        &               default = 0.0_rk,                                      &
        &               ErrCode = iError )
      ! set the axis parallel entries of local_vecs to local_length
      nVecs = 3
      allocate( local_vecs( 3, nVecs ))
      local_vecs = 0._rk
      local_vecs(1,1) = local_length
      local_vecs(2,2) = local_length
      local_vecs(3,3) = local_length
    end if

    call aot_table_close( L = conf, thandle = vec_thandle )
    call aot_table_close( L = conf, thandle = halfvec_thandle )


    ! copy back the information from the local vector array
    me%vec(:,:nVecs) = local_vecs
    me%vec(:,nVecs+1:) = 0.0_rk

    ! read origin of the canonicalND
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%origin,                                     &
      &               key     = 'origin',                                      &
      &               ErrCode = vError )
    if( any( btest( vError, errFatal ))) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%origin,                                   &
        &               key     = 'center',                                    &
        &               ErrCode = vError )
      if (any(btest( vError, errFatal )) ) then
        write(logUnit(1),*) 'ERROR reading the canonical objects, '//          &
          &             'origin/center is not well defined. STOPPING' 
        call tem_abort()
      endif
      me%origin = me%origin - me%vec(:,1)/2.0_rk - me%vec(:,2)/2.0_rk -        &
        &         me%vec(:,3)/2.0_rk
    end if

    me%segments = 1
    ! try loading segments as integer if fails try
    ! to loading segments as vector integer
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%segments(1),                                &
      &               key     = 'segments',                                    &
      &               ErrCode = iError )
    if( btest( iError, aoterr_Fatal )) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%segments,                                 &
        &               key     = 'segments',                                  &
        &               default = [ 1, 1, 1 ],                                 &
        &               ErrCode = vError )
      if( any( btest( vError, errFatal ))) then
        write(logUnit(1),*) ' FATAL ERROR: in loading segments' 
        call tem_abort()
      endif
    endif

    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%distribution,                               &
      &               ErrCode = iError,                                        &
      &               key     = 'distribution',                                &
      &               default = 'equal' )

    me%active = .false.
    me%nDim = 0

    ! check the kind and the active vectors
    do iVec = 1, 3
      vecAbsSum = me%vec(1,iVec) + me%vec(2,iVec) + me%vec(3,iVec)
      if (vecAbsSum .ne. 0._rk) then
        me%active(iVec) = .true.
        me%nDim = me%nDim + 1
      end if
    end do

    if(me%nDim .eq. 0) me%kind = 'point'
    if(me%nDim .eq. 1) me%kind = 'line'
    if(me%nDim .eq. 2) me%kind = 'plane'
    if(me%nDim .eq. 3) me%kind = 'box'

    !if canonical object is box then check for only_surface
    if(me%kind=='box') then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%only_surface,                             &
        &               ErrCode = iError,                                      &
        &               key     = 'only_surface',                              &
        &               pos     = 4,                                           &
        &               default = .false. )
   
      if (btest(iError, aoterr_WrongType)) then
        write(logUnit(1),*) 'Error occured, while retrieving canonical box '// &
          &             'only_surface' 
        write(logUnit(1),*) 'Variable has wrong type!' 
        write(logUnit(1),*) 'Should be a LOGICAL!' 
        call tem_abort()
      endif  
    endif

    ! \todo: DEB: Update this when converting arrays to strings is available
    write(logUnit(6),*) ' kind: '//trim( me%kind )
    write(logUnit(6),*)' origin:', me%origin 
    do iVec=1,me%nDim
      write(buffer,'(a4,i1)') 'vec ', iVec
      write(logUnit(6),*) trim(buffer), me%vec(:,iVec) 
    enddo
    do iVec=1,me%nDim
      write(buffer,'(a10,i1)') 'segments ', iVec
      write(logUnit(6),*) trim(buffer), me%segments(iVec) 
    enddo
    write(logUnit(6),*) ' distribution '//trim( me%distribution )

  end subroutine tem_load_oneCanonicalND
!******************************************************************************!


!******************************************************************************!
  !> Return the next coordinate of the canonical shape.
  !!
  function tem_getNextCoordOfcanonicalND( me, iElem ) result( coord )
    !---------------------------------------------------------------------------
    !> current element within total amount of elements to process
    integer, intent(in) :: iElem
    !> current canonical type
    type( tem_canonicalND_type ) :: me 
    !> calulated real-world coordinate, which is returned
    real(kind=rk)       :: coord(3) 
    !---------------------------------------------------------------------------

      ! Start from the origin, step through the segments on the two
      ! defining vectors successively

      coord = me%origin                                                        &
            ! offset in direction of vector A                                  
        &   + real(mod(( iElem-1 ), me%segments( 1 )), kind=rk)                &
        &   / real( max(me%segments( 1 )-1,1), kind=rk)                        &
        &   * me%vec(:,1)                                                      &
            ! offset in direction of vector B
            + real(mod( (iElem-1)/me%segments( 1 ), me%segments( 2 )),         &
        &          kind=rk)                                                    &
        &   / real( max(me%segments( 2 )-1,1), kind=rk)                        &
        &   * me%vec(:,2)                                                      &
            ! offset in direction of vector C
            + real(mod( (iElem-1)/(me%segments( 1 )*me%segments( 2 )),         &
        &               me%segments( 3 )), kind=rk)                            &
        &   / real( max(me%segments( 3 )-1,1), kind=rk)                        &
        &   * me%vec(:,3)

  end function tem_getNextCoordOfcanonicalND
!******************************************************************************!


!******************************************************************************!
  !> Write out an array of canonical shapes in lua format
  !!
  subroutine tem_canonicalND_out_vec( me, conf )
    !---------------------------------------------------------------------------
    !> canonicalND types to write out
    type( tem_canonicalND_type ), intent(in) :: me(:)
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------
    ! counter
    integer :: i
    !---------------------------------------------------------------------------

    ! create a table with name canonicalND
    call aot_out_open_table( put_conf = conf, tname = 'canonicalND' )

    do i = 1, size(me)
      call tem_canonicalND_out_scal( me(i), conf )
    end do

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_canonicalND_out_vec
!******************************************************************************!


!******************************************************************************!
  !> Write out a canonicalND shape in lua format
  !!
  subroutine tem_canonicalND_out_scal( me, conf )
    !---------------------------------------------------------------------------
    !> canonicalND types to write out
    type( tem_canonicalND_type ), intent(in) :: me
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------
    integer :: iVec
    !---------------------------------------------------------------------------

    ! create a table with name canonicalND if not exist
    if( conf%level .eq. 0 ) then
      call aot_out_open_table( put_conf = conf, tname = 'canonicalND' )
    else
      call aot_out_open_table( put_conf = conf )
    end if
    !write origin
    call aot_out_val( put_conf = conf, vname = 'origin', val = me%origin )

    !write vec
    if(me%nDim >0) then
      call aot_out_open_table( put_conf = conf, tname = 'vec' )
      do iVec=1,me%nDim
        call aot_out_val( put_conf = conf, val = me%vec(:,iVec) )
      enddo
      call aot_out_close_table( put_conf = conf )
     
      !write segments
      call aot_out_open_table( put_conf = conf, tname = 'segments' )

      do iVec=1,me%nDim
        call aot_out_val( put_conf = conf, val = me%segments(iVec) )
      enddo
      call aot_out_close_table( put_conf = conf )
    endif

    !write distribution
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'distribution',                               &
      &               val      = me%distribution )

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_canonicalND_out_scal
!******************************************************************************!


!******************************************************************************!
  subroutine tem_cano_initSubTree( me, inTree, countElems, map2global )
    !---------------------------------------------------------------------------
    !> canonicalND objects on which to work
    type(tem_canonicalND_type ),intent(inout) :: me(:)
    !> Global tree
    type(treelmesh_type), intent(in) :: inTree
    !> How many elements there will be for each level in the track
    integer, intent( out ) :: countElems( globalMaxLevels )
    !> growing array for the map2global
    type(dyn_intArray_type), intent(inout) :: map2global
    !---------------------------------------------------------------------------
    integer :: iCano, tLevel, nElems, elemPos, iElem, dPos
    real(kind=rk) :: coord(3)
    integer(kind=long_k) :: treeID
    logical :: wasAdded
    !---------------------------------------------------------------------------

    write(logUnit(1),*) 'Initializing canonical shapes'

    !Initialize countElems
    countElems = 0

    ! loop over all canoND objects
    do iCano = 1, size(me)
      ! \todo: SZ: set an upper limit in the load routine of the shapes,
      !            limit(*) = maximum number of elements in the directions
      !                       depending on minimum dx in the whole mesh
      !                       and edge size(*), *=x,y,z
      ! total number of elements
      nElems = me( iCano )%segments(1)                                         &
        &    * me( iCano )%segments(2)                                         &
        &    * me( iCano )%segments(3)

      ! Count how many elements are really local
      do iElem = 1, nElems
        coord(1:3) = tem_getNextCoordOfcanonicalND( me    = me( iCano ),       &
          &                                         iElem = iElem )
        ! Get the treeID on the highest level
        treeID = tem_IdOfCoord( tem_CoordOfReal( inTree, coord( 1:3 )))
        ! get position of the treeID in inTree
        elemPos = tem_PosOfId( treeID, inTree%treeID )
        if( elemPos .gt. 0 ) then
          ! append the position in inTree to the map (note that already existing
          ! ones are omitted)
          call append( me       = map2global,                                  &
            &          pos      = dpos,                                        &
            &          val      = elemPos,                                     &
            &          wasAdded = wasAdded )
          ! Count up if it was added
          if( wasAdded ) then
            tLevel   = tem_levelOf( treeID )
            countElems( tLevel ) = countElems( tLevel ) + 1
          endif ! wasAdded
        endif ! elemPos > 0
      end do ! iElem
    end do ! iCano
  end subroutine tem_cano_initSubTree
!******************************************************************************!


end module tem_canonicalND_module
!******************************************************************************!
