! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the data structure for the communication
!! during the simulation.
!!
!! Several exchange methods are implemented. CoCo is heavily used here to allow
!! for a concise definition of exchanges for various data types.
!! The basic idea is to initialize the buffers, use them in exchanges and
!! finalize them when not needed anymore.
!! In the definition of the buffers, an array of positions in the original
!! linearized data array is used to describe the origin or target positions for
!! communicated data.
!!
!!\note If you introduce a new type to exchange, you will need to introduce
!!      appropriate CoCo copy statements everywhere.
!!
module tem_comm_module

  ! include treelm modules
  use mpi
  use env_module,            only: rk, rk_mpi, long_k, long_k_mpi
  use tem_aux_module,        only: tem_abort
  use tem_logging_module,    only: logUnit
  use tem_comm_env_module,   only: tem_comm_env_type
  use tem_grow_array_module, only: grw_intArray_type,     &
    &                              init, append, destroy

  ! include aotus modules
  use flu_binding,  only: flu_State
  use aotus_module, only: aot_get_val

  implicit none

  private

  public :: tem_communication_type
  public :: tem_commPattern_type
  public :: tem_comm_dumpType
  public :: tem_load_commPattern

! Use CoCo here to make this buffer type generic for different types
?? text :: buffer_txt(tname, tstring)
  public :: tem_?tname?buffer_type

  !> Process-wise buffer for data of type ?tstring?
  !!
  !! This datatype is used to describe the exchange with a specific process, in
  !! case of explicit buffers it provides the memory for them.
  type tem_?tname?buffer_type

    !> Explicit buffer for data to be transferred
    ?tstring?, allocatable :: val(:)

    !> position in the input vector from where to read the entries in
    !! val_?tname?
    !!
    !!\note JZ: in ATELES we use this to specify the positions of the cell
    !!          states that have to be sent.
    integer, allocatable :: pos(:)

    !> Number of values to exchange
    !!
    !!\note JZ: in ATELES this variable stores the number of coefficients we
    !!          transfer, i.e. number of cells to transfer times number of
    !!          degree of freedoms per cell times the number of scalar
    !!          variables.
    integer :: nVals

    !> Number of elements involved in the exchange.
    !!
    !!\todo HK: Not sure if this is actually used anywhere?
    integer :: nElems

    !> Handle for the MPI-Datatype to describe the memory access,
    !! without explicit copying in the application.
    integer :: memIndexed
  end type tem_?tname?buffer_type
?? end text buffer_txt

! Let coco actually implement the buffer type given above for specific data
! types
?? copy :: buffer_txt(long, integer(kind=long_k))
?? copy :: buffer_txt(int, integer)
?? copy :: buffer_txt(real, real(kind=rk))

  !----------------------------------------------------------------------------!
  !> Description of communication data
  type tem_communication_type
    integer :: nProcs=0     !< amount of partitions to send to
    integer,allocatable :: proc(:) !< adjacent partitions
    !> How many data elements need to be exchanged with proc (per process).
    integer,allocatable :: nElemsProc(:)

    !> Data element positions in the actual arrays, used to built the pos
    !! information in the actual buffers (per process).
    type(grw_intArray_type), allocatable :: elemPos(:)

    !> declare communication buffers for each variable type
?? text :: buf_component_txt(tname)
    type( tem_?tname?buffer_type ), allocatable :: buf_?tname?(:)
?? end text buf_component_txt

?? copy :: buf_component_txt(long)
?? copy :: buf_component_txt(int)
?? copy :: buf_component_txt(real)
  end type tem_communication_type
  !----------------------------------------------------------------------------!


  !----------------------------------------------------------------------------!
  !> General description of the communication pattern to use.
  !!
  !! Depending on the chosen style, different exchange implementations are
  !! used. This data type provides the appropriate function pointers for
  !! initialization, finalization and exchange of the buffers.
  type tem_commPattern_type
    character(len=40) :: style

?? text :: exch_routine_txt(tname)
  procedure(tem_exchange_?tname?), nopass, pointer :: exchange_?tname?
  procedure(tem_commbuf_?tname?_init), nopass, pointer :: initBuf_?tname?
  procedure(tem_commbuf_?tname?_fin), nopass, pointer :: finBuf_?tname?
?? end text exch_routine_txt

?? copy :: exch_routine_txt(long)
?? copy :: exch_routine_txt(int)
?? copy :: exch_routine_txt(real)
  end type tem_commPattern_type
  !----------------------------------------------------------------------------!

! Definition of the abstract interfaces for the function pointers in the
! TEM_commPattern_type
?? text :: exchange_interface_txt(tname, tstring)
  abstract interface
    subroutine tem_exchange_?tname?(send, recv, state, message_flag,           &
      &                             send_state, proc)
      import :: rk, long_k, tem_communication_type, tem_comm_env_type
      type(tem_communication_type), intent(inout) :: send, recv
      ?tstring?, intent(inout) :: state(*)
      integer, intent(in) :: message_flag
      ?tstring?, intent(in), optional :: send_state(*)
      !> Process description to use.
      type(tem_comm_env_type), intent(in) :: proc 
    end subroutine tem_exchange_?tname?

    subroutine tem_commbuf_?tname?_init(me, pos, nVals)
      import :: tem_?tname?buffer_type
      type(tem_?tname?buffer_type), intent(inout) :: me
      integer, intent(in) :: nVals
      integer, intent(in) :: pos(nVals)
    end subroutine tem_commbuf_?tname?_init

    subroutine tem_commbuf_?tname?_fin(me)
      import :: tem_?tname?buffer_type
      type(tem_?tname?buffer_type), intent(inout) :: me
    end subroutine tem_commbuf_?tname?_fin
  end interface
?? end text exchange_interface_txt

?? copy :: exchange_interface_txt(long, integer(kind=long_k))
?? copy :: exchange_interface_txt(int, integer)
?? copy :: exchange_interface_txt(real, real(kind=rk))


contains

!******************************************************************************!
  !> This subroutine loads the communication pattern from a Lua script
  !! and sets the exchange routine to be used accordingly.
  !!
  !! The variable read from the script is "commpattern".
  !! Several patterns are available:
  !! * isend_irecv - Use explicit buffers, copy first the outgoing ones, then
  !!      post irecvs and isends, wait on all, and copy the incoming buffers to
  !!      their final locations (default).
  !! * isend_irecv_overlap - Similar to isend_irecv, but directly post sends,
  !!      after filling the outgoing buffer to each process, wait on any
  !!      incoming messages and only wait on sends, after everything is copied
  !!      into the final location.
  !! * typed_isend_irecv - Instead of copying the memory around, define a
  !!      indexed MPI datatype and use that in the exchange.
  !! * gathered_type - Similar to typed_isend_irecv, but with minimal number of
  !!      blocks in the indexed type, by tracking only contiguous blocks in the
  !!      memory layout.
  !!
  !! Instead of reading the style from a configuration script, it can also be
  !! directly set by the caller. If nothing is specified by the caller, the
  !! style will default to isend_irecv.
  !!
  !! Usage:
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! commpattern = 'isend_irecv'
  !! ~~~~~~~~~~~~~~~~~~~~~
  !!
  subroutine tem_load_commPattern( me, conf, style )
    !---------------------------------------------------------------------------
    !> commpattern to set
    type(tem_commPattern_type), intent(out) :: me
    !> handle to the Lua script
    type(flu_State), optional :: conf
    !> optional communication style
    character(len=*), intent(in), optional :: style
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------

    write(logUnit(1),*)"Loading communication pattern:"

    if (present(conf)) then
      ! If a configuration is given, this trumps any other setting.
      ! Defaults to isend_irecv.
      call aot_get_val( L       = conf,                                        &
        &               key     = 'commpattern',                               &
        &               val     = me%style,                                    &
        &               ErrCode = iError,                                      &
        &               default = 'isend_irecv' )
    else if (present(style)) then
      ! If a style is given directly by the caller, use that one.
      me%style = style
    else
      ! Default to isend_irecv if nothing provided by the caller.
      me%style = 'isend_irecv'
    end if

?? text :: exch_pointer_txt(routinename, initname, finname, tname)
      me%exchange_?tname? => ?routinename?_?tname?
      me%initBuf_?tname? => tem_commbuf_?tname?_?initname?
      me%finBuf_?tname? => tem_commbuf_?tname?_?finname?
?? end text exch_pointer_txt

    select case(trim(me%style))
    case ('isend_irecv')
?? copy :: exch_pointer_txt(comm_isend_irecv, fillPos, finPos, long)
?? copy :: exch_pointer_txt(comm_isend_irecv, fillPos, finPos, int)
?? copy :: exch_pointer_txt(comm_isend_irecv, fillPos, finPos, real)

    case ('isend_irecv_overlap')
?? copy :: exch_pointer_txt(comm_isend_irecv_overlap, fillPos, finPos, long)
?? copy :: exch_pointer_txt(comm_isend_irecv_overlap, fillPos, finPos, int)
?? copy :: exch_pointer_txt(comm_isend_irecv_overlap, fillPos, finPos, real)

    case ('typed_isend_irecv')
?? copy :: exch_pointer_txt(comm_typed_isend_irecv, fillIndexed, finTyped, long)
?? copy :: exch_pointer_txt(comm_typed_isend_irecv, fillIndexed, finTyped, int)
?? copy :: exch_pointer_txt(comm_typed_isend_irecv, fillIndexed, finTyped, real)

    case ('gathered_type')
?? copy :: exch_pointer_txt(comm_typed_isend_irecv, gatherIndexed, finTyped, long)
?? copy :: exch_pointer_txt(comm_typed_isend_irecv, gatherIndexed, finTyped, int)
?? copy :: exch_pointer_txt(comm_typed_isend_irecv, gatherIndexed, finTyped, real)

    case default
      write(logUnit(1),*)"ERROR, unknown commpattern: "//trim(me%style)
      write(logUnit(1),*)"available are: "
      write(logUnit(1),*)"* isend_irecv"
      write(logUnit(1),*)"* isend_irecv_overlap"
      write(logUnit(1),*)"* typed_isend_irecv"
      write(logUnit(1),*)"* gathered_type"
      call tem_abort()
    end select

    write(logUnit(1),*)trim(me%style)

  end subroutine tem_load_commPattern
!******************************************************************************!

! The following template defines the routines for various communication
! patterns, if you want to add a new pattern, you have to define an init,
! finalize and exchange routine in this template.
! To add a new data type, just add the appropriate CoCo copy line after this
! template.
?? text :: exchange_impl_txt(tname, tstring, vartype_mpi )
!******************************************************************************!
  !> Fill the positions that describe how the data in the state
  !! vector relates to the entries in the buffer.
  !!
  subroutine tem_commbuf_?tname?_fillPos( me, pos, nVals )
    !---------------------------------------------------------------------------
    type(tem_?tname?buffer_type), intent(inout) :: me
    integer, intent(in) :: nVals
    integer, intent(in) :: pos(nVals)
    !---------------------------------------------------------------------------

    me%nVals = nVals
    allocate(me%pos(nVals))
    me%pos = pos
    allocate(me%val(nVals))

  end subroutine tem_commbuf_?tname?_fillPos
!******************************************************************************!


!******************************************************************************!
  !> Free the communication buffer allocated by the fillPos routine again.
  !!
  subroutine tem_commbuf_?tname?_finPos(me)
    !---------------------------------------------------------------------------
    type(tem_?tname?buffer_type), intent(inout) :: me
    !---------------------------------------------------------------------------

    me%nVals = 0
    if( allocated( me%pos )) &
      deallocate(me%pos)
    if( allocated( me%val )) &
      deallocate(me%val)

  end subroutine tem_commbuf_?tname?_finPos
!******************************************************************************!


!******************************************************************************!
  !> Fill the indexed MPI datatype, which describes how the data in the state
  !! vector relates to the entries in the buffer.
  subroutine tem_commbuf_?tname?_fillIndexed(me, pos, nVals)
    !---------------------------------------------------------------------------
    type(tem_?tname?buffer_type), intent(inout) :: me
    integer, intent(in) :: nVals
    integer, intent(in) :: pos(nVals)
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------

    me%nVals = nVals
    ! CALL MPI_TYPE_CREATE_INDEXED_BLOCK(COUNT, BLOCKLENGTH, &
    !   &                   ARRAY_OF_DISPLACEMENTS, &
    !   &                   OLDTYPE, NEWTYPE, IERROR)
    call MPI_Type_create_indexed_block( nVals, 1, pos - 1,                     &
      &                                 ?vartype_mpi?, me%memIndexed, iError )
    call MPI_Type_commit(me%memIndexed, iError)

  end subroutine tem_commbuf_?tname?_fillIndexed
!******************************************************************************!


!******************************************************************************!
  !> Free the communication buffer allocated by the fillIndexed routine again.
  !!
  subroutine tem_commbuf_?tname?_finTyped(me)
    !---------------------------------------------------------------------------
    type(tem_?tname?buffer_type), intent(inout) :: me
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------

    me%nVals = 0
    call MPI_Type_free(me%memIndexed, iError)

  end subroutine tem_commbuf_?tname?_finTyped
!******************************************************************************!


!******************************************************************************!
  !> Gather the indexed MPI datatype, which describes how the data in the state
  !! vector relates to the entries in the buffer.
  !! In contrast to the simple indexed type above, we try to minimize the number
  !! of blocks here, and gather contiguous blocks of memory together.
  !!
  subroutine tem_commbuf_?tname?_gatherIndexed( me, pos, nVals )
    !---------------------------------------------------------------------------
    type(tem_?tname?buffer_type), intent(inout) :: me
    integer, intent(in) :: nVals
    integer, intent(in) :: pos(nVals)
    !---------------------------------------------------------------------------
    type(grw_IntArray_type) :: blocklength
    type(grw_IntArray_type) :: displ
    integer :: iVal, counter
    integer :: iError
    !---------------------------------------------------------------------------

    me%nVals = nVals

    ! Initialize growing arrays, a kB should be fine to start with...
    call init(blocklength, 256)
    call init(displ, 256)

    if (nVals > 0) then

      ! Start with the displacement of the first entry in the list
      call append(displ, pos(1)-1)
      counter = 1

      do iVal=2,nVals
        if (pos(iVal) == pos(iVal-1)+1) then
          ! Contiguous memory location following the previous one, increase the
          ! the blocklength.
          counter = counter + 1
        else
          ! New block encountered, record the block found so far
          call append(blocklength, counter)

          ! Start new block
          call append(displ, pos(iVal)-1)
          counter = 1
        end if
      end do

      ! Finish the last block, by recording its found length:
      call append(blocklength, counter)

    end if

    ! CALL MPI_TYPE_INDEXED(COUNT, ARRAY_OF_BLOCKLENGTHS, &
    !   &                   ARRAY_OF_DISPLACEMENTS, OLDTYPE, NEWTYPE, IERROR)
    call MPI_Type_indexed(displ%nVals, blocklength%Val, displ%Val,             &
      &                   ?vartype_mpi?, me%memIndexed, iError)
    call MPI_Type_commit(me%memIndexed, iError)

    call destroy(displ)
    call destroy(blocklength)

  end subroutine tem_commbuf_?tname?_gatherIndexed
!******************************************************************************!


!******************************************************************************!
  !> Exchange the communication buffers
  !! with a non-blocking MPI communication using preposted iRecv and iSend
  !! with a waitall
  !!
  subroutine comm_isend_irecv_?tname?( send, recv, state, message_flag,        &
    &                                  send_state, proc )
    !---------------------------------------------------------------------------
    type( tem_communication_type ), intent(inout) :: send
    type( tem_communication_type ), intent(inout) :: recv
    ?tstring?, intent(inout) :: state(*)  !< State vector to update
    integer, intent(in) :: message_flag
    ?tstring?, intent(in), optional :: send_state(*)  !< Data to send
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc 
    !---------------------------------------------------------------------------
    ! request handle for messages
    ! \todo request handle array could exist during complete code runtime
    integer :: rq_handle( recv%nProcs + send%nProcs )
    integer :: status( mpi_status_size, recv%nProcs + send%nProcs )
    integer :: iErr             !< error flag
    integer :: iProc, iVal
    integer :: comm
    !---------------------------------------------------------------------------
    comm = proc%comm

    if (present(send_state)) then
      do iProc = 1, send%nProcs
        ! Fill communication message
        do iVal = 1, send%buf_?tname?( iProc )%nVals
          send%buf_?tname?( iProc )%val( iVal )                                &
            &  = send_state( send%buf_?tname?( iProc )%pos( iVal ) )
        end do
      end do
    else
      do iProc = 1, send%nProcs
        ! Fill communication message
        do iVal = 1, send%buf_?tname?( iProc )%nVals
          send%buf_?tname?( iProc )%val( iVal )                                &
            &  = state( send%buf_?tname?( iProc )%pos( iVal ) )
        end do
      end do
    end if

    do iProc = 1, recv%nProcs
      ! Start receive communications
      call mpi_irecv(                           &
       &      recv%buf_?tname?( iProc )%val,    & ! me
       &      recv%buf_?tname?( iProc )%nVals,  & ! me size
       &      ?vartype_mpi?,                    & ! data type
       &      recv%proc(iProc),                 & ! target me
       &      message_flag,                     & ! flag
       &      comm,                             & ! communicator
       &      rq_handle(iProc),                 & ! handle
       &      iErr )                              ! error status
    enddo

    !  Start the sending communications
    do iProc = 1, send%nProcs
      call mpi_isend(                           &
       &      send%buf_?tname?( iProc )%val,    & ! buffer
       &      send%buf_?tname?( iProc )%nVals,  & ! count
       &      ?vartype_mpi?,                    & ! data type
       &      send%proc(iProc),                 & ! target
       &      message_flag,                     & ! tag
       &      comm,                             & ! communicator
       &      rq_handle( iProc + recv%nProcs ), & ! handle
       &      iErr )                              ! error status
    enddo ! iProc

    ! Wait for above communications to complete
    call mpi_waitall(send%nProcs+recv%nProcs,   & ! count
      &              rq_handle,                 & ! request handles
      &              status,                    & ! mpi status
      &              iErr )                       ! error status

    ! Now values from recv me can be copied to the actual state array
    do iProc = 1, recv%nProcs
      do iVal = 1, recv%buf_?tname?( iProc )%nVals
        ! write the values from the recv me to the state array
        ! to positions given in recvme%pos
        state( recv%buf_?tname?( iProc )%pos( iVal ) )                         &
          &  = recv%buf_?tname?( iProc )%val( iVal )
      end do
    end do

  end subroutine comm_isend_irecv_?tname?
!******************************************************************************!


!******************************************************************************!
  subroutine comm_isend_irecv_overlap_?tname?( send, recv, state,              &
    &                                          message_flag, send_state, proc )
    !---------------------------------------------------------------------------
    type( tem_communication_type), intent(inout) :: send, recv
    ?tstring?, intent(inout) :: state(*)
    integer, intent(in) :: message_flag
    ?tstring?, intent(in), optional :: send_state(*)  !< Data to send
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc 
    !---------------------------------------------------------------------------
    integer :: iProc, iVal  ! counter for neigbor mees
    integer :: finProc
    integer :: rq_handle( recv%nProcs + send%nProcs )
    integer :: status( mpi_status_size, recv%nProcs + send%nProcs )
    integer :: iErr             ! error flag
    integer :: comm
    !---------------------------------------------------------------------------
    comm = proc%comm

    ! Pre-post iRecv
    do iProc = 1, recv%nProcs
      call mpi_irecv(                          &
       &      recv%buf_?tname?( iProc )%val,   & ! buffer
       &      recv%buf_?tname?( iProc )%nVals, & ! count
       &      ?vartype_mpi?,                   & ! data type
       &      recv%proc(iProc),                & ! target me
       &      message_flag,                    & ! tag
       &      comm,                            & ! communicator
       &      rq_handle(iProc),                & ! handle
       &      iErr                             ) ! error status
    end do

    !> Fill send buffers and start Sending
    do iProc = 1, send%nProcs
      if (present(send_state)) then
        do iVal = 1, send%buf_?tname?( iProc )%nVals
          send%buf_?tname?( iProc )%val( iVal )                                &
            &  = send_state( send%buf_?tname?( iProc )%pos( iVal ) )
        end do
      else
        do iVal = 1, send%buf_?tname?( iProc )%nVals
          send%buf_?tname?( iProc )%val( iVal )                                &
            &  = state( send%buf_?tname?( iProc )%pos( iVal ) )
        end do
      end if
      call mpi_isend(                               &
       &      send%buf_?tname?( iProc )%val,        & ! buffer
       &      send%buf_?tname?( iProc )%nVals,      & ! count
       &      ?vartype_mpi?,                        & ! data type
       &      send%proc(iProc),                     & ! target
       &      message_flag,                         & ! tag
       &      comm,                                 & ! comm
       &      rq_handle( iProc + recv%nProcs ),     & ! handle
       &      iErr )                                  ! error status
    end do

    do iProc = 1, recv%nProcs
      ! Wait for above communications to complete
      call mpi_waitany(   &
        &    recv%nProcs, & ! count
        &    rq_handle,   & ! request handles
        &    finProc,     & ! process that finished
        &    status(:,1), & ! mpi status
        &    iErr )         ! error status
      do iVal = 1, recv%buf_?tname?( finProc )%nVals
        ! write the values from the recv me to the state array
        ! to positions given in recvme%pos
        state( recv%buf_?tname?( finProc )%pos( iVal ) )                       &
          &  = recv%buf_?tname?( finProc )%val( iVal )
      end do
    end do

    if (send%nProcs .gt. 0) then
      ! Wait for above communications to complete
      call mpi_waitall(                  &
        &    send%nProcs,                & ! total number of comm.'s to wait for
        &    rq_handle( recv%nProcs+1:), & ! request handles
        &    status( :,recv%nProcs+1:),  & ! mpi status
        &    iErr )                        ! error status
    end if

  end subroutine comm_isend_irecv_overlap_?tname?
!******************************************************************************!


!******************************************************************************!
  !> Exchange the communication mes
  !! with a non-blocking MPI communication using preposted iRecv and iSend
  !! with a waitall
  !!
  subroutine comm_typed_isend_irecv_?tname?( send, recv, state,                &
    &                                        message_flag, send_state, proc )
    !---------------------------------------------------------------------------
    type( tem_communication_type ), intent(inout) :: send, recv
    ?tstring?, intent(inout) :: state(*)  !< Current state vector
    integer, intent(in) :: message_flag
    ?tstring?, intent(in), optional :: send_state(*)  !< Data to send
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc 
    !---------------------------------------------------------------------------
    ! \todo request handle array could exist during complete code runtime
    integer :: rq_handle( recv%nProcs + send%nProcs )
    integer :: status( mpi_status_size, recv%nProcs + send%nProcs )
    integer :: iErr             ! error flag
    integer :: iProc
    integer :: comm
    !---------------------------------------------------------------------------
    comm = proc%comm

    !> Values for send me must have been copied from the actual state array
    do iProc = 1, recv%nProcs
      !> Start receive communications
      call mpi_irecv(                                     &
       &              state,                              & ! buffer
       &              1,                                  & ! count
       &              recv%buf_?tname?(iProc)%memIndexed, & ! type
       &              recv%proc(iProc),                   & ! source
       &              message_flag,                       & ! tag
       &              comm,                               & ! comm
       &              rq_handle(iProc),                   & ! request handle
       &              iErr )

    end do

    !>  Start the sending communications
    if (present(send_state)) then
      do iProc = 1, send%nProcs
        call mpi_isend(                                     &
          &             send_state,                         & ! buffer
          &             1,                                  & ! count
          &             send%buf_?tname?(iProc)%memIndexed, & ! type
          &             send%proc(iProc),                   & ! target
          &             message_flag,                       & ! tag
          &             comm,                               & ! comm
          &             rq_handle( iProc + recv%nProcs ),   & ! handle
          &             iErr )
      end do !< iProc
    else
      do iProc = 1, send%nProcs
        call mpi_isend(                                     &
          &             state,                              & ! buffer
          &             1,                                  & ! count
          &             send%buf_?tname?(iProc)%memIndexed, & ! type
          &             send%proc(iProc),                   & ! target
          &             message_flag,                       & ! tag
          &             comm,                               & ! comm
          &             rq_handle( iProc + recv%nProcs ),   & ! handle
          &             iErr )
      end do !< iProc
    end if

    !> Wait for above communications to complete
    call mpi_waitall(                          &
      &               send%nProcs+recv%nProcs, & ! count
      &               rq_handle,               & ! request handles
      &               status,                  & ! statuses
      &               iErr )

  end subroutine comm_typed_isend_irecv_?tname?
!******************************************************************************!
?? end text exchange_impl_txt

?? copy :: exchange_impl_txt(long, integer(kind=long_k), long_k_mpi)
?? copy :: exchange_impl_txt(int, integer, mpi_integer)
?? copy :: exchange_impl_txt(real, real(kind=rk), rk_mpi )


!******************************************************************************!
  !> Write communication type data to nUnit (debugging routine)
  !!
  subroutine tem_comm_dumpType(me, nUnit)
    !---------------------------------------------------------------------------
    type( tem_communication_type ), intent(in) :: me
    integer, intent(in) :: nUnit
    !---------------------------------------------------------------------------

    write(nUnit,*) '----Communication type-----------------------'
    write(nUnit,*) '   nProcs      ', me%nProcs
    write(nUnit,*) '   proc        ', me%proc
    write(nUnit,*) '   nElemsProc  ', me%nElemsProc
    write(nUnit,*) '---------------------------------------------'

  end subroutine tem_comm_dumpType
!******************************************************************************!


end module tem_comm_module
!******************************************************************************!
