! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains routines to load the depend table and various components
!! inside it. Depend table is usually defined in [scheme_module] and 
!! [geometry_increase] module. Thus, the routines within this module are called 
!! in [scheme_module] and [geometry_increase] module as required
!!
!! \author Simon Zimny, Kartik Jain
!!
!! [scheme_module]: @ref mus_scheme_module
!! [geometry_increase]: @ref mus_geomIncr_module
!!
module tem_depend_module

  ! include treelm modules
  use env_module,            only: labelLen
  use tem_var_system_module, only: tem_var_system_type, tem_var_sys_debug,     &
    &                              tem_var_intersect, tem_var_sys_load
  use tem_aux_module,        only: tem_abort
  use tem_condition_module,  only: tem_condition_type, tem_load_cond_single
  use tem_logging_module,    only: logUnit

  ! include aotus modules 
  use aotus_module,     only: flu_State, aot_get_val, aoterr_NonExistent
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length, aot_get_val

  implicit none
  private

  public :: tem_depend_type
  public :: tem_load_depend 
  public :: tem_load_depend_single
  public :: tem_init_depend
  
  !> Datatype containing information about dependency
  !! of geomIncrease header on other scheme(s)
  type tem_depend_type
    !> Contains variables which are requested in lua file
    type(tem_var_system_type) :: requestedSys
    type(tem_var_system_type) :: resSys
    !> Pointer to the parent scheme
    integer :: useScheme
    !> Name of the parent scheme
    character(len=labelLen) :: parent_scheme
    !> An instance of the condition type 
    type(tem_condition_type), allocatable :: cond(:)
  end type tem_depend_type

  contains

!******************************************************************************!
  !> Load variables, parent scheme and conditions defined in lua file.
  !! This routine serves as a wrapper and calls the single routine which loads
  !! the various arguments
  !!
  subroutine tem_load_depend(medepend, conf, parent)
    !---------------------------------------------------------------------------
    !> list of depend types to be filled
    type(tem_depend_type), allocatable, intent(inout) :: medepend(:)
    !> lua state to read from
    type(flu_state), intent(in) :: conf
    !> parent table identifier
    integer, intent(in)        :: parent
    !---------------------------------------------------------------------------
    integer :: iDep, nDeps
    integer :: dep_handle ! local depend table handle
    integer :: dep_sub_handle ! local depend table handle
    !---------------------------------------------------------------------------
    
    ! Open the depend table
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent,                                     &
      &                  thandle = dep_handle,                                 &
      &                  key     = 'depend' )

    if (dep_handle .ne. 0) then
      ! Check if there are multiple members inside depend
      call aot_table_open( L       = conf,                                     &
        &                  parent  = dep_handle,                               &
        &                  thandle = dep_sub_handle,                           &
        &                  pos     = 1 )
      ! If there is only one member, call the load routine once
      if (dep_sub_handle .eq. 0) then
        allocate(medepend(1))
        call aot_table_close( L = conf, thandle = dep_sub_handle )
        call tem_load_depend_single( medepend = medepend(1),                   &
          &                          conf     = conf,                          &
          &                          thandle  = dep_handle,                    &
          &                          label    = 'geomincr' )
      else
        call aot_table_close( L = conf, thandle = dep_sub_handle )
        nDeps = aot_table_length( L = conf, thandle = dep_handle )
        allocate( medepend( nDeps ))

        ! Read the subtables individually
        do iDep = 1, nDeps
          call aot_table_open( L       = conf,                                 &
            &                  parent  = dep_handle,                           &
            &                  thandle = dep_sub_handle,                       &
            &                  pos     = iDep )

          call tem_load_depend_single( medepend = medepend(iDep),              &
            &                          conf     = conf,                        &
            &                          thandle  = dep_sub_handle,              &
            &                          label    = 'geomincr')

          call aot_table_close( L = conf, thandle = dep_sub_handle )
        end do
      end if
    end if

    call aot_table_close( L = conf, thandle = dep_handle )

  end subroutine tem_load_depend
!******************************************************************************!


!******************************************************************************!
  !> Load single dependent variable of the scheme, in case of geomIncr: load the
  !! dependent variable as well as the conditions to be imposed for the geometry
  !! increase to take place.
  !!
  subroutine tem_load_depend_single( medepend, conf, thandle, label )
    !---------------------------------------------------------------------------
    !> depend type to be filled
    type(tem_depend_type), intent(inout) :: medepend
    !> lua state to read from
    type(flu_state), intent(in)          :: conf
    !> handle of parent table
    integer, intent(in)                  :: thandle
    !> label of the table
    character(len=*), intent(in) :: label
    !---------------------------------------------------------------------------
    integer :: iError, iCond, nCond
    integer :: cond_handle, sub_cond_handle
    !---------------------------------------------------------------------------

    ! Load the variable system
    call tem_var_sys_load( conf    = conf,                                     &
      &                    sys     = medepend%requestedSys,                    &
      &                    thandle = thandle,                                  &
      &                    sysName = label//'_sys' )

    call tem_var_sys_debug( sys   = medepend%requestedSys,                     &
      &                     level = 1 )
    
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'usescheme',                                   &
      &               val     = medepend%parent_Scheme,                        &
      &               ErrCode = iError )

    if( btest( iError, aoterr_NonExistent ) ) then
      write(logUnit(1),*) 'usescheme is not defined in depend table in '//     &
        &             'scheme '//label 
      call tem_abort()
    end if

    ! The below code snippet loads the conditions defined within depend table
    ! in the case of geometry increase, whereas in case this routine is called
    ! to load the depend table within schemes then conditions are not available
    ! and this load will not be effective in that case 
    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = cond_handle,                                &
      &                  key     = 'condition')
    if (cond_handle .ne. 0) then
      !> check single or multiple table
      call aot_table_open( L       = conf,                                     &
        &                  parent  = cond_handle,                              &
        &                  thandle = sub_cond_handle,                          &
        &                  pos     = 1 )
      if (sub_cond_handle .eq. 0) then
        allocate (medepend%cond(1))
        call tem_load_cond_single(medepend%cond(1), conf, cond_handle)
      else
        call aot_table_close( L=conf, thandle=sub_cond_handle)
        nCond = aot_table_length( L=conf, thandle=cond_handle)
        allocate (medepend%cond(nCond))
        do iCond = 1, nCond
          call aot_table_open( L       = conf,                                 &
            &                  parent  = cond_handle,                          &
            &                  thandle = sub_cond_handle,                      &
            &                  pos     = iCond )
          call tem_load_cond_single(medepend%cond(iCond), conf, sub_cond_handle)
        end do
      end if ! sub condition check
    end if ! Condition check

  end subroutine tem_load_depend_single
!******************************************************************************!


!******************************************************************************!
  !> This subroutine initializes the loaded depend table
  !!
  subroutine tem_init_depend( medepend, varSys, resSyslabel, label_list,       &
    &                         iScheme )
    !---------------------------------------------------------------------------
    !> list of depend types to be filled
    type(tem_depend_type), intent(inout)  :: medepend(:)
    !> list of all global variable systems
    type(tem_var_system_type), intent(in) :: varSys(:)
    !> label for the resulting variable system (intersection of varSys and 
    !! the requested varSys)
    character(len=*), intent(in)          :: resSyslabel
    !> list of labels of all the schemes
    character(len=labelLen), intent(in)   :: label_list(:)
    !> iScheme will be 0 in case of geomIncr and restart
    integer, intent(in)                   :: iScheme
    !---------------------------------------------------------------------------
    integer :: iDepend, nDepends, nSchemes, dScheme, ii
    logical :: found
    !---------------------------------------------------------------------------
    nDepends = size(medepend)
    ! The number of schemes is the length of label lists
    nSchemes = size(label_list)

    do iDepend = 1, nDepends
      ! Get the dependent scheme ID
      !\todo: KJ: may be make a function for this ?
      dScheme = 1
      found = .false.
      ! Calculate the dependent scheme ID
      do ii = 1, nSchemes
        ! in case of scheme-dependency, it is not allowed that the scheme
        ! depends on itself (given by iScheme)
        if( ii .ne. iScheme .and. .not. found ) then
          if ( trim( medepend( iDepend )%parent_scheme )                       &
            &  .eq. trim(label_list(ii) ) ) then
            dScheme = ii
            found = .true.
          end if
        end if
      end do
      if( .not. found ) then
        write(logUnit(1),*)' WARNING: The usescheme was not found!  '//        &
          &            trim(medepend(iDepend)%parent_scheme)
        ! Only abort if there are more than one scheme, in which case it is not
        ! clear which value should be taken.
        if( nSchemes .gt. 1 ) then
          call tem_abort()
        else ! didn't find scheme, so assign it to the first one
          write(logUnit(1),*)' Only one scheme present, so setting it to first'
          dScheme = 1
        end if
      end if
      ! Initialize the useScheme inside depend table to parent scheme
      medepend%useScheme = dScheme
      
      ! find intersection of global varSys and required varSys for all depends
      call tem_var_intersect(                                                  &
        &                   systemA     = medepend(iDepend)%requestedSys,      &
        &                   systemB     = varSys(medepend(iDepend)%usescheme), &
        &                   resSys      = medepend(iDepend)%resSys,            &
        &                   resSysLabel = resSyslabel//'_dependresSys' )

      call tem_var_sys_debug( sys   = medepend(iDepend)%resSys,                &
        &                     level = 1 )
    end do ! iDepend
  

  end subroutine tem_init_depend
!******************************************************************************!


end module tem_depend_module
!******************************************************************************!
