! See copyright notice in the COPYRIGHT file.
?? include 'treelm/source/deriveMacros.inc'
!******************************************************************************!
!> This module provides the common functions to evaluate derived quantities
!! and accessing input variables for musubi and ateles.
!!
!! \author Simon Zimny, Jiaxing Qi
!!
module tem_derived_module

  ! include treelm modules
  use env_module,               only: rk
  use tem_var_system_module,    only: tem_var_system_type, &
    &                                 tem_var_append
  use tem_variable_module,      only: tem_variable_type, tem_variable_from,    &
    &                                 tem_varError_nonAppend
  use tem_stencil_module,       only: tem_stencilHeader_type
  use tem_time_module,          only: tem_time_type
  use tem_spacetime_fun_module, only: tem_spacetime_for
  use treelmesh_module,         only: treelmesh_type
  use tem_subTree_type_module,  only: tem_subTree_type
  use tem_logging_module,       only: logUnit

  ! include aotus modules
  use aotus_module,             only: flu_state
  use aot_path_module,          only: aot_path_type


  implicit none

  private

  public :: access_StateVar
  public :: evaluateSources_single
  public :: evalSpacetime
  public :: evalDiff
  public :: evalRelDiff
  public :: evalMag
  public :: evalGrad
  public :: evalMagGrad
  public :: tem_append_inp_vars

contains

!******************************************************************************!
  !> Return the solver input of a variable for a given set of elements
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt(access_stateVar)
    !---------------------------------------------------------------------------
    integer :: iElem, iDir, pos, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    if( present(res) )then
      local_res => res
      ! copy the information to the output array
      do iElem = 1, nElems
        do iDir = 1, requVar%nComponents
          pos = globSys%variable%val( requVar%mapping )%varPos( iDir )
          local_res( ( iElem - 1 ) * requVar%nComponents + iDir ) =            &
            &             input(( iElem - 1 ) * globSys%nScalars + pos )
        end do
      end do
    end if
    ! else case is useless since all information would be just copied back
    ! at the same position in input

   end subroutine access_stateVar
!******************************************************************************!


!******************************************************************************!
  !> Empty function to evaluate the sources
  !!
  !! \todo Please elaborate a little on why this empty procedure is needed.
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt(evaluateSources_single)
    !-------------------------------------------------------------------------
!    integer :: iSourceVar, pos, nRes !\ todo: where to get nRes from??
!    integer :: sourcePos
!    integer :: treeIDpos
!    integer :: iFun
!    real(kind=rk), allocatable :: luaRes(:)
!    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
!    !-------------------------------------------------------------------------
!    if( present(res) )then
!      local_res => res
!    else
!      local_res => input
!    end if
!
!    allocate(luaRes(nElems*nRes))
!
!    ! initialize result array with 0
!    luaRes = 0.0
!    ! get the position of the requested variable in the global varSys
!    pos = varSys%variable%val( iVar )%mapping
!
!    ! get the positions in the list of source variables to catch all
!    ! sources of the same type (e.g. all injections)
!    ! loop over the source variables and check wether the treeID is
!    ! part of the different sources
!    ! * if yes -> evaluate the function and add the result to res
!    ! * if no -> add 0
!
!    do iSourceVar = 1, sourceMap%source( pos )%map( sourceMap%map( pos ))%nVals
!      do iFun = 1, size( sourceMap%source( pos )%sourceVar( sourceMap%source( pos )% &
!          &              map( sourceMap%map( pos ))%val( iSourceVar ))%st_fun )
!        ! check if the treeID is part of the tree of the current source variable
!        treeIDpos = tem_PosOfId( treeID, &
!          &     sourceMap%source( pos )%sourceVar(              &
!          &       sourceMap%source( pos )%map( sourceMap%map( pos ))%val( iSourceVar )) &
!          &     %st_fun(iFun)%tree%treeID )
!
!        if(treeIDpos .gt. 0)then
!          luaRes = tem_spacetime_for(                                                      &
!            &    me =  sourceMap%source( pos )%sourceVar( sourceMap%source(                &
!            &            pos )%map( sourceMap%map( pos ))%val( iSourceVar ))%st_fun(iFun), &
!            &    treeIDs = local_treeID,                                                   &
!            &    time = time,                                                              &
!            &    tree = sourceMap%source( pos )%sourceVar( sourceMap%source(               &
!            &       pos )%map( sourceMap%map( pos ))%val( iSourceVar ))%st_fun(iFun)%tree, &
!            &    n = 1 )
!          do iElem = 1, nElems
!            local_res(1:nElems*nRes) = res(1:nElems*nRes) + luaRes
!          end do
!        end if
!      end do
!    end do

   end subroutine evaluateSources_single
!******************************************************************************!


!******************************************************************************!
  !> Call the spacetime function, which are stored in variable%st_fun
  !! This spacetime function can be used as a analytical solution for refenrece
  !!
  !! \verbatim
  !! -- in lua file, one can define it as following:
  !! tracking = {
  !!   label = 'point',
  !!   variable = {
  !!     { name = 'spacetime', ncomponents=1, spacetime = luaFun },
  !!   },
  !!   ...
  !!  }
  !! \endverbatim
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt(evalSpacetime)
    !---------------------------------------------------------------------------
    integer :: nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------
    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! evaluate the spacetime function for all given elements
    local_res(1:nElems) = tem_spacetime_for(                                   &
      &    globSys%variable%val( requVar%mapping )%st_fun(1),                  &
      &    bary, time, nElems, conf )

  end subroutine evalSpacetime
!******************************************************************************!


!******************************************************************************!
  !> Evaluate the function pointers of the dependent variables,
  !! and then calculate the difference between these two. ( only scalar )
  !! In lua file, one can first define a variable (e.g. density) to track,
  !! then define a spacetime variable as analytical solution,
  !! at last, define a variable called difference, which depend on density
  !! and spacetime. In this way, one can get an error between simulation
  !! results and analytical solution.
  !!
  !! \verbatim
  !! -- in lua file, one can define as following:
  !! tracking = {
  !!   label = 'point',
  !!   variable = {
  !!     { name = 'density', ncomponents=1 },
  !!     { name = 'spacetime', ncomponents=1, spacetime = luaFun },
  !!     { name = 'difference', ncomponents=1, dep = {'density','spacetime'}},
  !!   },
  !!   folder = 'tracking/',
  !!   shape = {kind = 'canoND', object = {origin = {3.0,3.1,3.0} } },
  !!   format = 'ascii',
  !!   time = {min = 0, max = tmax, interval = 1},
  !!  }
  !! \endverbatim
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt(evalDiff)
    !---------------------------------------------------------------------------
    integer :: iDep, posDepVar, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk), allocatable :: tmpRes(:,:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(tmpRes( nElems, 2 ))

    do iDep = 1, 2
      ! get position of dependent var from reqSys
      ! which gives the position of dependent vat in the global sys
      !reqsys = globSys since posDepVar map to globsys
      ! save results in tmpRes
      posDepVar = requVar%varDep(iDep)

      call globSys%derive( posDepVar )%do(                                     &
        &                           input   = input,                           &
        &                           bound   = bound,                           &
        &                           tree    = tree,                            &
        &                           subTree = subTree,                         &
        &                           requVar = globSys%Variable%val(posDepVar), &
        &                           globSys = globSys,                         &
        &                           stencil = stencil,                     &
        &                           conf    = conf,                            &
        &                           res     = tmpRes( :, iDep ),               &
        &                           bary    = bary,                            &
        &                           time    = time )

    end do

    local_res(:)  = tmpRes(:,1) - tmpRes(:,2)

    deallocate(tmpRes)

  end subroutine evalDiff
!******************************************************************************!

!******************************************************************************!
  !> Evaluate the function pointers of the two dependent variables,
  !! and then calculate the relative difference between them. ( only scalar )
  !! In lua file, one can first define a variable (e.g. density) to track,
  !! then define a spacetime variable as analytical solution,
  !! at last, define a variable called relDiff, which depend on density
  !! and spacetime. In this way, one can get an relative error between numerical
  !! result and analytical solution.
  !!
  !! \verbatim
  !! -- in lua file, one can define as following:
  !! tracking = {
  !!   label = 'point',
  !!   variable = {
  !!     { name = 'relDiff', ncomponents=1, dep = {'density','spacetime'}},
  !!   },
  !!   folder = 'tracking/',
  !!   shape = {kind = 'canoND', object = {origin = {3.0,3.1,3.0} } },
  !!   format = 'ascii',
  !!   time = {min = 0, max = tmax, interval = 1},
  !! }
  !! \endverbatim
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt( evalRelDiff )
    !---------------------------------------------------------------------------
    integer :: iDep, posDepVar, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk), allocatable :: tmpRes(:,:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    allocate(tmpRes( nElems, 2 ))

    do iDep = 1, 2
      ! get position of dependent var from reqSys
      ! which gives the position of dependent vat in the global sys
      ! reqsys = globSys since posDepVar map to globsys
      ! save results in tmpRes
      posDepVar = requVar%varDep( iDep )

      call globSys%derive( posDepVar )%do(                                     &
        &                           input   = input,                           &
        &                           bound   = bound,                           &
        &                           tree    = tree,                            &
        &                           subTree = subTree,                         &
        &                           requVar = globSys%Variable%val(posDepVar), &
        &                           globSys = globSys,                         &
        &                           stencil = stencil,                         &
        &                           conf    = conf,                            &
        &                           res     = tmpRes( :, iDep ),               &
        &                           bary    = bary,                            &
        &                           time    = time )

    end do ! iDep

    local_res(:)  = (tmpRes(:,1) - tmpRes(:,2)) / tmpRes(:,2)

    deallocate(tmpRes)

  end subroutine evalRelDiff
!******************************************************************************!

!******************************************************************************!
  !> Evaluate the function pointers of the two dependent variables,
  !! and then calculate the relative difference between them. ( only scalar )
  !! In lua file, one can first define a variable (e.g. density) to track,
  !! then define a spacetime variable as analytical solution,
  !! at last, define a variable called relDiff, which depend on density
  !! and spacetime. In this way, one can get an relative error between numerical
  !! result and analytical solution.
  !!
  !! \verbatim
  !! -- in lua file, one can define as following:
  !! tracking = {
  !!   label = 'point',
  !!   variable = {
  !!     { name = 'magnitude', ncomponents=1, dep = {'velocity'} },
  !!   },
  !!   folder = 'tracking/',
  !!   shape = {kind = 'canoND', object = {origin = {3.0,3.1,3.0} } },
  !!   format = 'ascii',
  !!   time = {min = 0, max = tmax, interval = 1},
  !! }
  !! \endverbatim
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt( evalMag )
    !---------------------------------------------------------------------------
    integer :: iElem, nElems, depVar_pos, depVar_nComps, offset
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk), allocatable :: tmpRes(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the position of dependent variable, assuming only one
    depVar_pos = requVar%varDep(1)
    ! get the nComp of dependent variable
    depVar_nComps = globSys%variable%val( depVar_pos )%nComponents
    ! allocate array to save results
    allocate(tmpRes( depVar_nComps * nElems ))

    call globSys%derive( depVar_pos )%do(                                    &
      &                           input   = input,                           &
      &                           bound   = bound,                           &
      &                           tree    = tree,                            &
      &                           subTree = subTree,                         &
      &                           requVar = globSys%Variable%val(depVar_pos),&
      &                           globSys = globSys,                         &
      &                           stencil = stencil,                         &
      &                           conf    = conf,                            &
      &                           res     = tmpRes,                          &
      &                           bary    = bary,                            &
      &                           time    = time )

    tmpRes = tmpRes * tmpRes
    do iElem = 1, nElems
      offset = (iElem-1) * depVar_nComps
      local_res(iElem) = sqrt(sum(tmpRes(offset+1:offset+depVar_nComps)))
    end do

    deallocate(tmpRes)

  end subroutine evalMag
!******************************************************************************!

!******************************************************************************!
  !> Calculate the spatial gradient for a given quantity which has to be a
  !! scalar.\n
  !! It utilizes the subtree to find its neighbor, then perform central
  !! difference (2nd order) on all 3 directions when neighbors are available.
  !! If neighbor on one side is missing, it calculates the difference only on
  !! the other side (1st order).\n
  !! \todo: NOT work yet!
  !!
  !! \verbatim
  !! -- in lua file, one can define as following:
  !! tracking = {
  !!   label = 'point',
  !!   variable = {
  !!     { name = 'gradient', ncomponents=3, dep = {'velocity'} },
  !!   },
  !!   folder = 'tracking/',
  !!   shape = {kind = 'canoND', object = {origin = {3.0,3.1,3.0} } },
  !!   format = 'ascii',
  !!   time = {min = 0, max = tmax, interval = 1},
  !! }
  !! \endverbatim
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt( evalGrad )
    !---------------------------------------------------------------------------
    integer :: iElem, nElems, depVar_pos, depVar_nComps
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk), allocatable :: tmpRes(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the position of dependent variable, assuming only one
    depVar_pos = requVar%varDep(1)
    ! get the nComp of dependent variable
    depVar_nComps = globSys%variable%val( depVar_pos )%nComponents
    ! allocate array to save results
    allocate(tmpRes( depVar_nComps * nElems ))

    call globSys%derive( depVar_pos )%do(                                    &
      &                           input   = input,                           &
      &                           bound   = bound,                           &
      &                           tree    = tree,                            &
      &                           subTree = subTree,                         &
      &                           requVar = globSys%Variable%val(depVar_pos),&
      &                           globSys = globSys,                         &
      &                           stencil = stencil,                         &
      &                           conf    = conf,                            &
      &                           res     = tmpRes,                          &
      &                           bary    = bary,                            &
      &                           time    = time )

    tmpRes = tmpRes * tmpRes
    do iElem = 1, nElems
      local_res(iElem) = sqrt(sum(tmpRes((iElem-1)+1:(iElem-1)+depVar_nComps)))
    end do

    deallocate(tmpRes)

  end subroutine evalGrad
!******************************************************************************!

!******************************************************************************!
  !> Calculate the magnitude of gradient for a given quantity.
  !! \todo: NOT work yet.
  !!
  !! \verbatim
  !! -- in lua file, one can define as following:
  !! tracking = {
  !!   label = 'point',
  !!   variable = {
  !!     { name = 'magGrad', ncomponents=1, dep = {'velocity'} },
  !!   },
  !!   folder = 'tracking/',
  !!   shape = {kind = 'canoND', object = {origin = {3.0,3.1,3.0} } },
  !!   format = 'ascii',
  !!   time = {min = 0, max = tmax, interval = 1},
  !! }
  !! \endverbatim
  !!
  !! The interface has to comply to the abstract interface
  !! tem_var_system_mdoule#derive "derive".
  !!
?? copy :: derive_headtxt( evalMagGrad )
    !---------------------------------------------------------------------------
    integer :: iElem, nElems, depVar_pos, depVar_nComps
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk), allocatable :: tmpRes(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! get the position of dependent variable, assuming only one
    depVar_pos = requVar%varDep(1)
    ! get the nComp of dependent variable
    depVar_nComps = globSys%variable%val( depVar_pos )%nComponents
    ! allocate array to save results
    allocate(tmpRes( depVar_nComps * nElems ))

      call globSys%derive( depVar_pos )%do(                                    &
        &                           input   = input,                           &
        &                           bound   = bound,                           &
        &                           tree    = tree,                            &
        &                           subTree = subTree,                         &
        &                           requVar = globSys%Variable%val(depVar_pos),&
        &                           globSys = globSys,                         &
        &                           stencil = stencil,                         &
        &                           conf    = conf,                            &
        &                           res     = tmpRes,                          &
        &                           bary    = bary,                            &
        &                           time    = time )

    tmpRes = tmpRes * tmpRes
    do iElem = 1, nElems
      local_res(iElem) = sqrt(sum(tmpRes((iElem-1)+1:(iElem-1)+depVar_nComps)))
    end do

    deallocate(tmpRes)

  end subroutine evalMagGrad
!******************************************************************************!

!******************************************************************************!
  !> subroutine to add the variables from the input lua script to the varsys
  !!
  !! \verbatim
  !! -- in lua file, one can add additional variables by:
  !! add_variable = {
  !!   { name = 'vel_an', ncomponents=1, vel_an=velAn },
  !! }
  !!
  !! function velAn(x, y, z, t)
  !!    return x+y+z
  !! end funciton
  !! \endverbatim
  !!
  !! \todo: appending space time variables to varsys should be seperated into a
  !! subroutine.
  subroutine tem_append_inp_vars(conf, path, varSys )
    !> flu state
    type( flu_State ) :: conf
    !> lua path
    type( aot_path_type ),intent(in) :: path
    !> The resulting variable system used in the Euler equations
    type(tem_var_system_type), intent(inout)   :: varSys
    !--------------------------------------------------------------------------
    type( tem_variable_type ), allocatable :: tVar(:)
    integer :: loc_nTotal, pos, iVar
    !--------------------------------------------------------------------------

    ! create variables(tVars) by reading the add_variable table from lua file
    call tem_variable_from( me     = tVar,                                     &
      &                     conf   = conf,                                     &
      &                     nTotal = loc_nTotal,                               &
      &                     key    = 'add_variable',                           &
      &                     path   = path )

    ! Append these variables to variable system, only if they have a space time
    ! function. Otherwise they are useless.
    do iVar = 1, size(tVar)
      if ( trim(tVar(iVar)%st_fun(1)%fun_kind) .ne. 'none') then
        call tem_var_append( sys = varSys, var = tVar( iVar ), pos = pos)
        ! check if the variable has been added (only appended if requirements are
        ! fullfilled)
        if ( pos == 0 ) then
          call tem_varError_nonAppend( variable = tVar(ivar),                  & 
            &                          fUnit = logUnit(1) )
        ! check if the variable to append is already a state variable
        else if( pos .gt. varSys%nVars )then
          varSys%derive(pos)%do => evalSpacetime
          varSys%variable%val( pos )%mapping = pos
        end if
      endif
    enddo

  end subroutine tem_append_inp_vars
  !*****************************************************************************

end module tem_derived_module
!******************************************************************************!
