  ! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module  provides descriptions for elements
!!
?? include 'arrayMacros.inc'
module tem_element_module

  ! include treelm modules
  use env_module,            only: long_k, minLength, zeroLength
  use tem_dyn_array_module,  only: dyn_longArray_type, init, append, expand,   &
    &                              destroy, truncate, SortedPosOfVal,          &
    &                              PositionOfVal
  use tem_grow_array_module, only: grw_longArray_type, truncate, destroy,      &
    &                              grw_logicalArray_type, init, append, expand,&
    &                              grw_intArray_type
  use tem_stencil_module,    only: tem_stencilElement_type, tem_stencil_dump,  &
    &                              grw_stencilElementArray_type, init, append, &
    &                              truncate, destroy

  implicit none

  private

  public :: tem_element_type
  public :: tem_element_dump
  public :: grw_dyn_longArray_type
  !> dynamic array of variable types
  public :: init, destroy, append, expand, getSize, getReqSize,                &
    &       PositionOfVal, sortedPosOfVal, truncate

  public :: eT_fluidElem, eT_haloElem, eT_nonExistingElem
  public :: eT_ghostFromCoarserElem, eT_ghostFromFinerElem
  public :: eT_distributedGhostFromFinerElem
  public :: eT_undefinedElem
  public :: eT_minNumRelevant, eT_maxNumRelevant
  public :: eT_minNumber, eT_maxNumber

  integer, parameter :: eT_undefinedElem                  = -1
  integer, parameter :: eT_nonExistingElem                =  0
  integer, parameter :: eT_fluidElem                      =  1
  integer, parameter :: eT_ghostFromCoarserElem           =  2
  integer, parameter :: eT_ghostFromFinerElem             =  3
  integer, parameter :: eT_haloElem                       =  4
  integer, parameter :: eT_distributedGhostFromFinerElem  =  5
  !> Element properties which are created while adaptive refinement
  integer, parameter :: eT_sacrificedElem                 =  6
  integer, parameter :: eT_newFluid                       =  7
  !> eType integer values must be in the range of 
  !! eT_minNumber <= eT_val <= eT_maxNumber
  integer, parameter :: eT_minNumber                      = 0
  integer, parameter :: eT_maxNumber                      = 5
  !> Relevant number of element types.
  !! As the distributed ghost from finer are in the assemble list step collapsed
  !! with the normal ghosFromFiner, only eType = 1 to 4 needs to be treated
  !! afterwards
  integer, parameter :: eT_minNumRelevant      = 1
  integer, parameter :: eT_maxNumRelevant      = 4

?? copy :: GA_decltxt(grw_stencilElement, type(grw_stencilElementArray_type))
?? copy :: GA_decltxt(dyn_long, type(dyn_longArray_type))

  !KJ: The tem_element_type contains the attributes of an element
  ! inside the level descriptor (usually). 

  type tem_element_type
    !> Tree ID 
    type(dyn_longArray_type)                :: tID
    !> Property
    type(grw_longArray_type)                :: property
    !> element type: one of fluid, ghostFromCoarser,
    !! ghostFromFiner, halo
    type(grw_intArray_type)                 :: eType
    !> Pointer to the original treeID list
    type(grw_intArray_type)                 :: pntTID
    !> Stencils defined for this element
    type(grw_grw_stencilElementArray_type ) :: stencil
    !> neighbor treeIDs coming from the stencil definitions
    type(grw_dyn_longArray_type )           :: neighID
    !> source partition (starts at 1)
    type(grw_intArray_type)                 :: sourceProc
    !> position of the boundary_ID( nDir, nElems ) in bc_prop_type
    type(grw_intArray_type)                 :: boundaryPos
    !> nesting (only relevant for halos, to include their neighborhood)
    type(grw_intArray_type)                 :: haloNesting
    !> does this element need an update
    type(grw_logicalArray_type)             :: needsUpdate
  end type tem_element_type

  interface init
    module procedure init_element
  end interface

  interface append
    module procedure append_element
  end interface

  interface truncate
    module procedure truncate_element
  end interface

  interface destroy 
    module procedure destroy_element
  end interface

  interface getSize
    module procedure getSize_element
  end interface

  interface getReqSize
    module procedure getReqSize_element
  end interface

  contains

!******************************************************************************!
  !> Include the subroutines for the dynamic array.
  !!
?? copy :: GA_impltxt(grw_stencilElement, type(grw_stencilElementArray_type))
?? copy :: GA_impltxt(dyn_long, type(dyn_longArray_type))

!******************************************************************************!
  !> initialize an element and optionally set contents
  !!
  subroutine init_element( me, length )
    !---------------------------------------------------------------------------
    type( tem_element_type ), intent(out) :: me
    integer, intent(in), optional :: length
    !---------------------------------------------------------------------------
    call init( me = me%tID, length = length, unique = .true. )
    call init( me = me%property, length = length )
    call init( me = me%eType, length = length )
    call init( me = me%pntTID, length = length )
    call init( me = me%stencil, length = length )
    call init( me = me%neighID, length = length )
    call init( me = me%sourceProc, length = length )
    call init( me = me%boundaryPos, length = length )
    call init( me = me%haloNesting, length = length )
    call init( me = me%needsUpdate, length = length )

  end subroutine init_element
!******************************************************************************!


!******************************************************************************!
  !> truncate all the lists in the element object
  !!
  subroutine truncate_element( me )
    !---------------------------------------------------------------------------
    !> element object
    type( tem_element_type ), intent(inout) :: me
    !---------------------------------------------------------------------------
    integer :: iVal
    !---------------------------------------------------------------------------
    call truncate( me = me%tID )
    call truncate( me = me%property )
    call truncate( me = me%eType )
    call truncate( me = me%pntTID )
    call truncate( me = me%stencil )
    do iVal = 1, me%stencil%nVals
      call truncate( me = me%stencil%val( iVal ) )
    end do
    call truncate( me = me%neighID )
    do iVal = 1, me%neighID%nVals
      call truncate( me = me%neighID%val( iVal ) )
      deallocate( me%neighID%val(iVal)%sorted)
    end do
    call truncate( me = me%sourceProc )
    call truncate( me = me%boundaryPos )
    call truncate( me = me%haloNesting )
    call truncate( me = me%needsUpdate )

  end subroutine truncate_element
!******************************************************************************!


!******************************************************************************!
  !> return the number of actual used memory (bytes) per element
  !!
  subroutine getReqSize_element( me, elemSize )
    !---------------------------------------------------------------------------
    !> element object
    type( tem_element_type ), intent(inout) :: me
    !> number of total entries to be returned
    integer(kind=long_k), intent(out) :: elemSize
    !---------------------------------------------------------------------------
    integer :: iVal, nSize
    !---------------------------------------------------------------------------
    elemSize = 0
    elemSize = elemSize + int( me%tID%nVals*12, kind=long_k)
    elemSize = elemSize + int( me%property%nVals*8, kind=long_k)
    elemSize = elemSize + int( me%eType%nVals*4, kind=long_k)
    elemSize = elemSize + int( me%pntTID%nVals*4, kind=long_k)
    
    do iVal = 1, me%stencil%nVals
      if( me%stencil%val( iVal )%nVals .gt. 0 ) then
        nSize = size( me%stencil%val( iVal )%val(1)%tIDpos )*8
      else
        nSize = 0
      end if
      elemSize = elemSize + int( me%stencil%val(iVal)%nVals*nSize, kind=long_k)
    end do

    do iVal = 1, me%neighID%nVals
      elemSize = elemSize + int( me%neighID%val( iVal )%nVals*12, kind=long_k)
    end do

    elemSize = elemSize + int( me%sourceProc%nVals*4, kind=long_k)
    elemSize = elemSize + int( me%boundaryPos%nVals*4, kind=long_k)
    elemSize = elemSize + int( me%haloNesting%nVals*4, kind=long_k)
    elemSize = elemSize + int( me%needsUpdate%nVals*4, kind=long_k)

  end subroutine getReqSize_element
!******************************************************************************!


!******************************************************************************!
  !> return the number of total allocated memory (bytes) per element
  !!
  subroutine getSize_element( me, elemSize )
    !---------------------------------------------------------------------------
    !> element object
    type( tem_element_type ), intent(inout) :: me
    !> number of total entries to be returned
    integer(kind=long_k), intent(out) :: elemSize
    !---------------------------------------------------------------------------
    integer :: iVal, nSize
    !---------------------------------------------------------------------------
    elemSize = 0
    elemSize = elemSize + int( me%tID%containerSize*12, kind=long_k)
    elemSize = elemSize + int( me%property%containerSize*8, kind=long_k)
    elemSize = elemSize + int( me%eType%containerSize*4, kind=long_k)
    elemSize = elemSize + int( me%pntTID%containerSize*4, kind=long_k)
    
    do iVal = 1, me%stencil%nVals
      if( me%stencil%val( iVal )%nVals .gt. 0 ) then
        nSize = size( me%stencil%val( iVal )%val(1)%tIDpos )*8
      else
        nSize = 0
      end if
      elemSize = elemSize                                                      &
        &      + int( me%stencil%val( iVal )%containerSize*nSize, kind=long_k)
    end do

    do iVal = 1, me%neighID%nVals
      elemSize = elemSize                                                      &
        &      + int( me%neighID%val( iVal )%containerSize*12, kind=long_k)
    end do

    elemSize = elemSize + int( me%sourceProc%containerSize*4, kind=long_k)
    elemSize = elemSize + int( me%boundaryPos%containerSize*4, kind=long_k)
    elemSize = elemSize + int( me%haloNesting%containerSize*4, kind=long_k)
    elemSize = elemSize + int( me%needsUpdate%containerSize*4, kind=long_k)

  end subroutine getSize_element
!******************************************************************************!


!******************************************************************************!
  !> destroy all the lists in the element object
  !!
  subroutine destroy_element( me )
    !---------------------------------------------------------------------------
    !> element object
    type( tem_element_type ), intent(inout) :: me
    !---------------------------------------------------------------------------
    integer :: iVal
    !---------------------------------------------------------------------------
    call destroy( me = me%tID )
    call destroy( me = me%property )
    call destroy( me = me%eType )
    call destroy( me = me%pntTID )
    do iVal = 1, me%stencil%nVals
      call destroy( me = me%stencil%val( iVal ) )
    end do
    call destroy( me = me%stencil )
    do iVal = 1, me%neighID%nVals
      call destroy( me = me%neighID%val( iVal ) )
    end do
    call destroy( me = me%neighID )
    call destroy( me = me%sourceProc )
    call destroy( me = me%boundaryPos )
    call destroy( me = me%haloNesting )
    call destroy( me = me%needsUpdate )

  end subroutine destroy_element
!******************************************************************************!


!******************************************************************************!
  !> initialize an element and optionally set contents
  !!
  subroutine append_element( me, tID, property, eType, pntTID, sourceProc,     &
    &                       nNeighIDs, boundaryPos, haloNesting, needsUpdate,  &
    &                       stencil, pos, wasAdded )
    !---------------------------------------------------------------------------
    !> element object
    type( tem_element_type ), intent(inout)     :: me
    !> tree ID to be added
    integer(kind=long_k), intent(in)            :: tID
    !> position of treeID 
    integer, intent(out)                        :: pos
    !> property associated with the treeID
    integer(kind=long_k), intent(in), optional  :: property
    !> element property bit
    integer, intent(in), optional               :: eType
    !> pointer of the treeID 
    integer, intent(in), optional               :: pntTID
    !> boundary position
    integer, intent(in), optional               :: boundaryPos
    !> nesting level for haloElems
    integer, intent(in), optional               :: haloNesting
    !> the procID which is adding the element
    integer, intent(in), optional               :: sourceProc
    !> 
    logical, intent(in), optional               :: needsUpdate
    !>
    integer, intent(in), optional               :: nNeighIDs
    !>
    logical, intent(out),optional               :: wasAdded
    !>
    type(tem_stencilElement_type), intent(in), optional :: stencil(:)
    !---------------------------------------------------------------------------
    integer :: iStencil    
    integer :: neighIDsize
    logical :: wasAdded_loc
    type(grw_stencilElementArray_type) :: stencilElement
    type(dyn_longArray_type) :: tneighID
    !---------------------------------------------------------------------------
    call append( me = me%tID, val = tID, pos = pos, wasAdded = wasAdded_loc )
    if( wasAdded_loc ) then
      ! was not in list before, but added. Add all further handed in contents.
      if( present( property )) then
        call append( me = me%property, val = property )
      else
        call append( me = me%property, val = 0_long_k )
      end if

      if( present( eType )) then
        call append( me = me%eType, val = eType )
      else
        call append( me = me%eType, val = eT_undefinedElem )
      end if

      if( present( pntTID )) then
        call append( me = me%pntTID, val = pntTID )
      else
        call append( me = me%pntTID, val = -1 )
      end if

      if( present( sourceProc )) then
        call append( me = me%sourceProc, val = sourceProc )
      else
        call append( me = me%sourceProc, val = -1 )
      end if

      if( present( boundaryPos )) then
        call append( me = me%boundaryPos, val = boundaryPos )
      else
        call append( me = me%boundaryPos, val = -1 )
      end if

      if( present( haloNesting )) then
        call append( me = me%haloNesting, val = haloNesting )
      else
        call append( me = me%haloNesting, val = 1 )
      end if

      if( present( needsUpdate )) then
        call append( me = me%needsUpdate, val = needsUpdate )
      else
        call append( me = me%needsUpdate, val = .false. )
      end if

      call init( me = stencilElement, length = 1 )
      call append( me = me%stencil, val = stencilElement )
      if( present( stencil )) then
        do iStencil = 1, size( stencil )
          if( stencil(iStencil)%QQN .gt. 0 )                                   &
            &           call append( me  = me%stencil%val( pos ),              &
            &                        val = stencil( iStencil ))
        end do
      end if

      if( present( nNeighIDs )) then
        neighIDsize = max( 0, nNeighIDs )
      else
        neighIDsize = 1
      end if
      ! add empty dynamic array of length neighIDsize for the neighbors
      call init( me = tNeighID, unique = .true., length = neighIDsize )
      call append( me = me%neighID, val = tNeighID )
    else ! was already in list. update contents.
    end if

    if( present( wasAdded )) then
      wasAdded = wasAdded_loc
    end if

  end subroutine append_element
!******************************************************************************!


!******************************************************************************!
  !> Write element information to disk
  subroutine tem_element_dump( me, elemPos, nUnit, compact, header, stencil,   &
    &                          stencilPos )
    !---------------------------------------------------------------------------
    !> element object
    type( tem_element_type ), intent(in) :: me
    !>
    integer, intent(in) :: elemPos
    !>
    integer, intent(in) :: nUnit
    !>
    logical, intent(in), optional :: compact
    !>
    logical, intent(in), optional :: header
    !>
    logical, intent(in), optional :: stencil
    !>
    logical, intent(in), optional :: stencilPos
    !---------------------------------------------------------------------------
    integer :: iStencil
    logical :: locCompact, locStencil, locStencilpos
    !---------------------------------------------------------------------------
    if( present( stencil )) then
      locStencil= stencil
    else
      locStencil = .false.
    endif
    if( present( stencilPos )) then
      locStencilPos = stencilPos
    else
      locStencilPos = .false.
    endif
    if( present( compact )) then
      locCompact = compact
    else
      locCompact = .true.
    endif
    if( locCompact ) then
      if( present( header )) then 
        if( header ) then
      write(nUnit,'(a15,6(a4))') 'tID', 'prp', 'eT', 'prc', 'nst', 'pTD', 'stc'
        endif
      endif
      write(nUnit,'(i15,6i4)') me%tID%val( elemPos ),                          &
        &                      me%property%val( elemPos ),                     &
        &                      me%eType%val( elemPos ),                        &
        &                      me%sourceProc%val( elemPos ),                   &
        &                      me%haloNesting%val( elemPos ),                  &
        &                      me%pntTID%val( elemPos ),                       &
        &                      me%stencil%val( elemPos )%nVals
    else
      write(nUnit,*) '---------------------------------------------'
      write(nUnit,*) 'Element'
      write(nUnit,*) '   treeID      ', me%tID%val( elemPos )
      write(nUnit,*) '   property    ', me%property%val( elemPos )
      write(nUnit,*) '   eType       ', me%eType%val( elemPos )
      write(nUnit,*) '   srcProc     ', me%sourceProc%val( elemPos )
      write(nUnit,*) '   haloNesting ', me%haloNesting%val( elemPos )
      write(nUnit,*) '---------------------------------------------'
    end if
    do iStencil = 1, me%stencil%val( elemPos )%nVals
      call tem_stencil_dump(                                                   &
        &                 me      = me%stencil%val( elemPos )%val( iStencil ), &
        &                 nUnit   = nUnit,                                     &
        &                 neighID = me%neighID%val( elemPos )%val,             &
        &                 tIDonly = .false. )
    enddo

  end subroutine tem_element_dump
!******************************************************************************!


end module tem_element_module
!*******************************************************************************
