! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Initialize general infrastructure settings of treelm.
!!
!! \author Harald Klimach
!!
!! This module provides a central point to initialize the
!! various general settings in treelm.
module tem_general_module

  ! include aotus modules
  use aotus_module, only: flu_State, aot_get_val

  ! include treelm modules
  use env_module,             only: rk, tem_load_env_params, io_buffer_size,   &
    &                               printRuntimeInfo, init_env, fin_env,       &
    &                               print_self_status
  use tem_logging_module,     only: tem_logging_load_primary
  use tem_comm_module,        only: tem_commpattern_type, tem_load_commpattern
  use tem_debug_module,       only: tem_debug_load_main
  use tem_logging_module,     only: logUnit, tem_logging_init
  use tem_solveHead_module,   only: tem_solveHead_type, tem_init_solveHead
  use tem_restart_module,     only: tem_restart_type
  use tem_timer_module,       only: tem_addTimer, tem_startTimer,              &
    &                               tem_stopTimer, tem_getTimerVal,            &
    &                               tem_getTimerName, tem_timer_type
  use tem_balance_module,     only: tem_balance_type, tem_balance_loadConfig
  use tem_status_module,      only: tem_status_type
  use tem_comm_env_module,    only: tem_comm_env_init, tem_comm_env_fin,       &
    &                               tem_comm_env_type
  use tem_tools_module,       only: tem_horizontalSpacer 
  use tem_simControl_module,  only: tem_simControl_type, tem_simControl_load,  &
    &                               tem_simControl_dump, tem_simControl_start


  use tem_precice_module, only: precice_available, tem_precice_load

  implicit none

  private

  public :: tem_start
  public :: tem_finalize
  public :: tem_load_general
  public :: tem_general_type

  !> Global parameter type contains all general information
  !! needed for all solvers
  type tem_general_type

    !> General description of the deployed solver.
    type(tem_solveHead_type) :: solver

    !> contains current simulation time, timeControl, abortCriteria and 
    !! simulation status
    type(tem_simControl_type) :: simControl

    !> MPI communication enviroment including MPI communicator.
    !!
    !!\todo HK: not sure, if this should be here!
    type(tem_comm_env_type) :: proc

    !> MPI communication pattern type.
    type(tem_commPattern_type)    :: commPattern

    !> Global restart type for all schemes
    !!
    !!\todo HK: not sure, if this should be here!
    type(tem_restart_type) :: restart

    !> Load balancing information.
    !!
    !!\todo HK: not sure, if this should be here!
    type(tem_balance_type) :: balance

    !> A timer type for internal timing analyses
    type(tem_timer_type) :: timerData

  end type tem_general_type


contains


  !****************************************************************************!
  !> Load general treelm settings from the Lua script in conf.
  !!
  subroutine tem_load_general( me, conf )
    !---------------------------------------------------------------------------
    !> global general parameter
    type( tem_general_type ), intent(inout) :: me
    !> Handle to the Lua script containing the configuration.
    type(flu_state) :: conf
    !---------------------------------------------------------------------------
    integer :: iError
    !---------------------------------------------------------------------------

    call tem_horizontalSpacer(fUnit = logUnit(1))

    write(logUnit(1),*) 'Loading general parameters:'

    ! load and initialize logUnit
    call tem_logging_load_primary(conf = conf, rank = me%proc%rank)

    ! load and initialize debug unit
    call tem_debug_load_main(conf = conf, rank = me%proc%rank)

    ! load global enviromental parameters from config file
    call tem_load_env_params(conf)

    write(logUnit(1),*)'Size of the IO Buffer (in size of 8 Bytes): ', io_buffer_size

    ! load simulation name
    call aot_get_val( L       = conf,                                          &
      &               key     = 'simulation_name',                             &
      &               val     = me%solver%simName,                             &
      &               ErrCode = iError,                                        &
      &               default = 'simulation')

    write(logUnit(1),*) 'Simulation Name: '// trim( me%solver%simName )

    ! load simulation time control
    call tem_simControl_load( me = me%simControl, conf = conf ) 

    call tem_simControl_dump(me=me%simControl, outUnit=logUnit(1))

    ! load communication pattern and initialize commuication 
    ! infrastructure
    call tem_load_commpattern(conf = conf, me = me%commpattern)

    ! Get load balancing config
    !>\todo HK: the balance type does have a timeControl on its own,
    !!          why did it need the passing in of a global time?
    !! \todo: SZ: tem_balance_loadConfig is called in mus_load_config as well
    !!            useless here?
    call tem_balance_loadConfig( conf          = conf,                         &
      &                          me            = me%balance,                   &
      &                          rank          = me%proc%rank,                 &
      &                          defaultActive = .true.      )

    ! load precice if available
    if (precice_available) then
      write(logUnit(1),*)'Loading precice data' 
      call tem_precice_load(conf = conf)
    end if
    call tem_horizontalSpacer(fUnit = logUnit(1))

  end subroutine tem_load_general
  !****************************************************************************!
  
  
  !****************************************************************************!
  !> Initialize the environment. Should be the very first call in the program.
  !!
  subroutine tem_start(codeName, version, general, comm, simControl)
    !---------------------------------------------------------------------------
    !> name of code
    character(len=*), intent(in) :: codeName
    !> version of the code
    character(len=*), intent(in) :: version
    !> encapsulates global parameters which are common for all solvers
    type(tem_general_type), intent(out) :: general
    !> mpi communicator if it is predefined as in apesmate
    integer, intent(in), optional :: comm
    !> simulation control to initialize
    type(tem_simControl_type), intent(out), optional :: simControl
    !---------------------------------------------------------------------------
    integer :: nProcs, rank, nThreads
 
    ! if comm is present initialize environment already called
    ! so should not be called again
    if(.not. present(comm)) call init_env()

    ! initialize mpi environment
    call tem_comm_env_init(general%proc, comm)
    nProcs   = general%proc%comm_size
    rank     = general%proc%rank
    nThreads = general%proc%nThreads

    ! initialize solverHead
    call tem_init_solveHead( me = general%solver, solName = codeName,          &
      &                      version = version )

    ! initialize logging primary with level 1 
    call tem_logging_init(level = 1, rank = rank)

    if (present(simControl)) call tem_simControl_start(simControl)

    write(logUnit(1),*) "Starting up "//trim(codeName)//" with nprocs: ", nProcs

    !$ write(logUnit(1),*)"               and nThreads pp: ", nThreads

    call tem_addTimer(timerHandle = general%solver%timerHandle, &
      &               timerName   = trim(codeName),             &
      &               me          = general%timerData    )
    call tem_startTimer(timerHandle = general%solver%timerHandle, &
      &                 me          = general%timerData    )

  end subroutine tem_start
  !****************************************************************************!


  !****************************************************************************!
  !> Finalize the environment, should be the very last call in the program.
  subroutine tem_finalize(general)
    !--------------------------------------------------------------------------
    !> encapsulates global parameters which are common for all solvers
    type(tem_general_type), intent(inout) :: general
    !---------------------------------------------------------------------------
    character(len=64) :: timerName
    real(kind=rk) :: timerValue
    !---------------------------------------------------------------------------

    if (general%proc%rank == 0) then
      if ( printRuntimeInfo ) call print_self_status()
    
      timerName  =  tem_getTimerName(timerHandle = general%solver%timerHandle, &
        &                            me          = general%timerData           )

      timerValue =  tem_getTimerVal(timerHandle = general%solver%timerHandle,  &
        &                           me          = general%timerData            )
      
      write(logUnit(1),*)
      write(logUnit(1),*)" Done with " // trim(timerName) // " in [s]",        &
        &            real(timerValue)
    end if    

    ! finialize mpi
    call fin_env()

  end subroutine tem_finalize
  !****************************************************************************!


end module tem_general_module
!******************************************************************************!
