! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the functionality to fill the initial values in the
!! computational domain.
module tem_ini_condition_module

  ! include treelm modules
  use env_module,            only: rk, labelLen
  use tem_tools_module,      only: tem_horizontalSpacer
  use tem_logging_module,    only: logUnit
  use tem_spatial_module,    only: load_spatial, tem_spatial_type
  use tem_depend_module,     only: tem_depend_type, tem_load_depend

  ! include aotus modules
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode
  use aotus_module,     only: flu_State

  implicit none

  private

  public :: tem_ini_condition_type
  public :: tem_load_ic

  !> Definition of the initial condition.
  !! The ini_state must exist for each variable
  type tem_ini_condition_type
    !>
    logical :: isRestart
    !>
    character(len=256) :: RestartFile
    !>
    character(len=32), allocatable :: StateName(:)
    !> initial state of each variable. size is nVars
    type(tem_spatial_type), allocatable :: ini_state(:)
    !> An instance of depend type in case of restart
    type(tem_depend_type), allocatable :: depend(:)
  end type tem_ini_condition_type

contains

!******************************************************************************!
  !> Load initial condition
  !!
  !! Check if restart file is defined. Then use restart file to initialize
  !! variables. If not then load state variables either as constant or function
  !! \author Kannan Masilamani
  !!
  subroutine tem_load_ic(me, conf, parent, path, StateName, key )
    !---------------------------------------------------------------------------
    !> Initial condition type
    type(tem_ini_condition_type), intent(out) :: me
    !> lua state type
    type(flu_State) :: conf
    !> aotus path to the function in the lua
    type( aot_path_type ) :: path
    !> array of defined initial state variables
    character(len=*), intent(in) :: StateName(:)
    !>
    integer, intent(in), optional :: parent
    !> key for the initial cond.
    character(len=*), intent(in), optional :: key
    !---------------------------------------------------------------------------
    integer :: iState
    integer :: ic_table
    integer :: nStates
    integer :: iError
    type( aot_path_type ) :: ic_path ! lua path to the entry in the scheme table
    character(len=labelLen) :: lockey ! key for the initial cond.
    !---------------------------------------------------------------------------

    nStates = size(StateName)

    allocate(me%StateName(nStates))
    allocate(me%ini_state(nStates))
    me%StateName = StateName
    if ( present(key   ) ) then
      lockey = key
    else
      lockey = 'initial_condition'
    end if
    call aot_table_open(L=conf, parent=parent, thandle=ic_table, key=lockey)
    ic_path = path
    call tem_horizontalSpacer(fUnit=logUnit(1))
    write(logUnit(1),*) ' Loading initial conditions' 
    call aot_path_addNode( me       = ic_path,                                 &
      &                    NodeType = 'table',                                 &
      &                    key      = lockey )
    if (ic_table == 0) then
      ! Interpret the initial_condition entry as name for the restart file,
      ! if it is not a table.
      write(*,*) 'Initial condition NOT a table!'
    else
      ! First check if a restart file is present:
      call aot_get_val( L       = conf,                                        &
        &               thandle = ic_table,                                    &
        &               key     = 'file',                                      &
        &               val     = me%RestartFile,                              &
        &               default = '',                                          &
        &               ErrCode = iError)

      ! Try loading the depend table within restart 
      call tem_load_depend( medepend = me%depend,                              &
        &                   conf     = conf,                                   &
        &                   parent   = ic_table )

      ! \todo: SZ: this is not used at all in the current implementation since
      !            the restart data overrides the ini condition data in
      !            mus_init_aux
      !            BUT this needs to be extended for the new restart
      me%isRestart = (trim(me%RestartFile) /= '')
    end if

    if (.not. me%isRestart) then
      ! We are not doing a restart, need to find initial conditions for all
      ! requested states.
      do iState=1,nStates
         call load_spatial( me           = me%ini_state(iState),               &
           &                conf         = conf,                               &
           &                parent       = ic_table,                           &
           &                key          = trim(me%stateName(iState)),         &
           &                path         = ic_path,                            &
           &                defaultValue = 0._rk )
      end do

    end if
    call aot_table_close(L=conf, thandle=ic_table)

  end subroutine tem_load_ic
!******************************************************************************!


end module tem_ini_condition_module
!******************************************************************************!
