! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!>
!!
module tem_line_module

  ! include treelm modules
  use env_module,         only: rk
  use tem_aux_module,     only: tem_abort
  use tem_logging_module, only: logUnit, tem_toStr

  ! include aotus modules
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_out_module,   only: aot_out_type, aot_out_val, aot_out_open_table,   &
    &                         aot_out_close_table
  use aotus_module,     only: aot_get_val, aoterr_Fatal, flu_State

  implicit none
  private

  public :: tem_line_type
  public :: tem_load_line
  public :: tem_line_out
  public :: tem_getNextCoordOfLine
  public :: lenOfProjection

  !> Definition of a line object
  type tem_line_type
    !> origin of tracking line
    real(kind=rk) :: origin(3)
    !> direction vector of tracking line
    real(kind=rk) :: direction(3)
    !> how many segments the line is divided into
    integer ::       segments
    !> spatial distribution of the lines on the line
    character(len=40) :: distribution
  end type tem_line_type

  interface tem_line_out
    module procedure tem_line_out_scal
    module procedure tem_line_out_vec
  end interface tem_line_out

contains

!******************************************************************************!
  !> Loading lines from the config file
  !! valid definitions:
  !! * line = { origin = {1.0, 0.0, -1.0}, direction = { ...} } -- single line
  !! -- single line defined by center
  !! * line = { center = {1.0, 0.0, -1.0}, direction = { ...} }
  !! * line = { { origin = {1.0, 0.0, -1.0}, direction = { ...}} } -- single
  !!                                                         line within a table
  !! * line = { { origin = {0.0, 0.0, -1.0}, direction = { ...}} , --  multiple
  !!                                                                       lines
  !! *          { origin = {1.0, 0.0, -1.0}, direction = { ...}} }
  !!
  subroutine tem_load_line( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> Arrays of Lines to read in, which first has to be allocated
    type(tem_line_type), allocatable, intent(out) :: me(:)
    !>
    type(flu_state) :: conf
    !>
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    integer :: lines_thandle
    integer :: single_thandle
    integer :: line_entries
    integer :: iLine
    integer :: nLines
    !---------------------------------------------------------------------------

    call aot_table_open( L       = conf,                                       &
      &                  parent  = thandle,                                    &
      &                  thandle = lines_thandle,                              &
      &                  key     = 'line' )

    line_entries = aot_table_length( L = conf, thandle = lines_thandle )

    ! Now check, if the first entry is a table itself:
    call aot_table_open( L       = conf,                                       &
      &                  parent  = lines_thandle,                              &
      &                  thandle = single_thandle,                             &
      &                  pos     = 1 )

    if (single_thandle == 0) then
      ! The entries are not tables themselves, try to interpret it as
      ! single line.
      call aot_table_close(L=conf, thandle=single_thandle)
      nLines = 1
      allocate(me( nLines ))
      call get_one_line(me(1), conf, lines_thandle)

    else
      ! First entry is a table, assume all others are also tables,
      ! and properly define line coordinates, go on reading them.
      call aot_table_close(L=conf, thandle=single_thandle)
      nLines = line_entries
      allocate(me(nLines))
      do iLine = 1, nLines
        call aot_table_open( L       = conf,                                   &
          &                  parent  = lines_thandle,                          &
          &                  thandle = single_thandle,                         &
          &                  pos     = iLine )
        call get_one_line(me(iLine), conf, single_thandle)
        call aot_table_close(L=conf, thandle=single_thandle)
      end do
    end if
    call aot_table_close(L=conf, thandle=lines_thandle)

    do iLine = 1, nLines
      write(logUnit(1),*) ' Read line: ',iLine,' with'
      write(logUnit(1),*) '  origin       '//trim( tem_toStr(                  &
        &                   val = me( iLine )%origin,                          &
        &                   sep = ',' ))
      write(logUnit(1),*) '  direction    '//trim( tem_toStr(                  &
        &                   val = me( iLine )%direction,                       &
        &                   sep = ',' ))
      write(logUnit(1),*) '  segments     ',me( iLine )%segments
      write(logUnit(1),*) '  distribution '//trim( me( iLine )%distribution)
    end do

  end subroutine tem_load_line
!******************************************************************************!


!******************************************************************************!
  !> Read in one complete line definition from a specified handle of a lua file
  !!
  !! Parameters for a line are the following.
  !! It is the defintion for a line starting at {0.,0.,0.} ending at {1.,0.,0.}
  !! which is discretized by 10 segments, meaning eleven points, on which then
  !! corresponding elements are identified.
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! line = { origin = {0.,0.,0.}, -- the origin of the line
  !!          direction = {1.,0.,0.}, -- the length and direction vector of the line
  !!          segments = 10,  -- with how many points does the contiguous line
  !!                          -- have to be segmented
  !!          distribution = 'equal' -- how are the discrete points distributed
  !!        }                        -- on the contiguous line
  !! ~~~~~~~~~~~~~~~~~~~~~
  !!
  subroutine get_one_line( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> the line type to read into
    type(tem_line_type), intent(out) :: me
    !> the lua config
    type(flu_state) :: conf
    !> the table handle from where to read
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    integer :: iError
    integer :: vError(3), errfatal(3)
    !---------------------------------------------------------------------------
    errfatal = aotErr_Fatal

    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%segments,                                   &
      &               ErrCode = iError,                                        &
      &               key     = 'segments',                                    &
      &               default = 10 )
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%distribution,                               &
      &               ErrCode = iError,                                        &
      &               key     = 'distribution',                                &
      &               default = 'equal' )

    ! Load the direction of the line
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%direction,                                  &
      &               key     = 'direction',                                   &
      &               ErrCode = vError )
    ! if direction is not defined look for halfdirection and compute
    ! direction from half direction
    if (any(btest( vError, errFatal )) ) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%direction,                                &
        &               key     = 'halfdirection',                             &
        &               ErrCode = vError )
      if (any(btest( vError, errFatal )) ) then
        write(logUnit(1),*)'ERROR reading the line, direction/halfdirection '//&
          &                'is not well'
        write(logUnit(1),*)'defined. STOPPING' 
        call tem_abort()
      endif
      me%direction = me%direction * 2._rk
    end if

    ! Load the origin of the line
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me%origin,                                     &
      &               key     = 'origin',                                      &
      &               ErrCode = vError )

    if (any(btest( vError, errFatal )) ) then
      call aot_get_val( L       = conf,                                        &
        &               thandle = thandle,                                     &
        &               val     = me%origin,                                   &
        &               key     = 'center',                                    &
        &               ErrCode = vError )
      if (any(btest( vError, errFatal )) ) then
        write(logUnit(1),*) 'ERROR reading the line, origin/center is not well'
        write(logUnit(1),*) 'defined. STOPPING' 
        call tem_abort()
      end if
      me%origin = me%origin - me%direction/2.0_rk
    end if

  end subroutine get_one_line
!******************************************************************************!


!******************************************************************************!
  !> Return the next coordinate of the line
  !!
  !! More complex shapes are identified by a point-wise processing of the shape
  !! The objects are described by a set of points. These sets are defined by an
  !! amount of subdivisions
  !!
  function tem_getNextCoordOfLine( me, iElem ) result(coord)
    !---------------------------------------------------------------------------
    !> current element within total amount of elements to process
    integer, intent(in) :: iElem
    !> the line
    type(tem_line_type ) :: me
    !> calulated real-world coordinate, which is returned
    real(kind=rk)       :: coord(3)
    !---------------------------------------------------------------------------

    ! Start from the origin, step through the segments on the line
    coord = me%origin                                                          &
      &   + real(iElem-1,kind=rk)/real(me%segments-1,kind=rk)                  &
      &   * me%direction

  end function tem_getNextCoordOfLine
!******************************************************************************!


!******************************************************************************!
  !> This function returns the length of projection of vector a on vector b
  !!
  !! \f$ proj_b a = \frac{a \cdot b}{|b|}\f$
  !!
  function lenOfProjection(vecA, vecB) result (res)
    !---------------------------------------------------------------------------
    !> vector to project
    real( kind=rk ) :: vecA(3)
    !> vector on which vecA to project
    real( kind=rk ) :: vecB(3)
    !> return value of a function
    real( kind=rk ) :: res
    !---------------------------------------------------------------------------

    res = dot_product( vecA, vecB )/sqrt(dot_product( vecB, vecB ))

  end function lenOfProjection
!******************************************************************************!


!******************************************************************************!
  !> Write a array of lines to lua file
  !!
  subroutine tem_line_out_vec( me, conf )
    !--------------------------------------------------------------------------
    !> line types to write out
    type( tem_line_type ), intent(in) :: me(:)
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !--------------------------------------------------------------------------
    !local variable
    integer :: i
    !--------------------------------------------------------------------------

    ! create a table with name line
    call aot_out_open_table( put_conf = conf, tname = 'line' )

    do i = 1, size(me)
      call tem_line_out_scal( me(i), conf )
    end do

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_line_out_vec
!******************************************************************************!


!******************************************************************************!
  !> Write a line to lua file
  !!
  subroutine tem_line_out_scal( me, conf )
    !---------------------------------------------------------------------------
    !> line type to write out
    type( tem_line_type ), intent(in) :: me
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !---------------------------------------------------------------------------

    ! create a table with name line if not exist
    if( conf%level .eq. 0 ) then
      call aot_out_open_table( put_conf = conf, tname = 'line' )
    else
      call aot_out_open_table( put_conf = conf )
    end if
    ! write origin
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'origin',                                     &
      &               val      = me%origin )

    ! write direction
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'direction',                                  &
      &               val      = me%direction )

    ! write segments
    call aot_out_val( put_conf = conf,                                         &
      &                 vname    = 'segments',                                 &
      &                 val      = me%segments )
    ! write distribution
    call aot_out_val( put_conf = conf,                                         &
      &               vname    = 'distribution',                               &
      &               val      = me%distribution )

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_line_out_scal
!******************************************************************************!


end module tem_line_module
!******************************************************************************!
