! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module provides the functionality to derive mesh dependend information
!! for tracking output.
!!
!! \author Simon Zimny
!!
?? include 'treelm/source/deriveMacros.inc'
module tem_meshInfo_module

  ! include treelm modules
  use env_module,              only: rk, long_k
  use tem_var_system_module,   only: tem_var_system_type
  use tem_variable_module,     only: tem_variable_type
  use tem_var_module,          only: tem_var_type
  use tem_stencil_module,      only: tem_stencilHeader_type
  use tem_time_module,         only: tem_time_type
  use treelmesh_module,        only: treelmesh_type
  use tem_geometry_module,     only: tem_ElemSize, tem_posOfID
  use tem_property_module,     only: prp_fluidify, prp_hasIBM, prp_solid,      &
    &                                prp_sendHalo
  use tem_subTree_type_module, only: tem_subTree_type, tem_treeIDfrom_subTree

  ! include aotus modules
  use aotus_module, only: flu_state

  implicit none

  private

  public :: nMeshVars
  public :: tem_initMeshInfoList

  !> treeID
  type(tem_var_type), SAVE :: treeID_var
  !> MPI process
  type(tem_var_type), SAVE :: mpiProc_var
  !> element volume
  type(tem_var_type), SAVE :: elemVol_var
  !> fraction element to total volume of the shape
  type(tem_var_type), SAVE :: volFrac_var
  !> derive wether prp_fluidify is set or not
  type(tem_var_type), SAVE :: fluidify_var
  !> derive wether prp_hasIBM is set or not
  type(tem_var_type), SAVE :: hasIBM_var
  !> derive wether prp_hasIBM is set or not
  type(tem_var_type), SAVE :: hasSendHalo_var
  !> derive wether prp_fluidify is set or not
  type(tem_var_type), SAVE :: solid_var

  !> number of mesh variables
  integer, parameter :: nMeshVars = 8

  contains


!******************************************************************************!
  !> This subroutine initializes the list of meshInfo variables (e.g. element 
  !! volume, element volume fraction, treeID, ... )
  !!
  subroutine tem_initMeshInfoList( meshInfoList )
    !---------------------------------------------------------------------------
    !> 
    type(tem_var_type), allocatable, intent(inout) :: meshInfoList(:)
    !---------------------------------------------------------------------------
!    nMeshVars = 0

    ! treeID
    treeID_var%variable%label = 'treeID'
    treeID_var%variable%nComponents = 1
    treeID_var%do => deriveTreeID

    ! MPI Process
    mpiProc_var%variable%label = 'process'
    mpiProc_var%variable%nComponents = 1
    mpiProc_var%do => deriveMPIproc

    ! element volume
    elemVol_var%variable%label = 'elemVol'
    elemVol_var%variable%nComponents = 1
    elemVol_var%do => deriveElemVol

    ! volume fraction
    volFrac_var%variable%label = 'volFrac'
    volFrac_var%variable%nComponents = 1
    volFrac_var%do => deriveVolFrac

    ! prp_fluidify
    fluidify_var%variable%label = 'fluidify'
    fluidify_var%variable%nComponents = 1
    fluidify_var%do => deriveFluidify

    ! prp_solid
    solid_var%variable%label = 'solid'
    solid_var%variable%nComponents = 1
    solid_var%do => deriveSolid

    ! prp_hasIBM
    hasIBM_var%variable%label = 'hasIBM'
    hasIBM_var%variable%nComponents = 1
    hasIBM_var%do => deriveIBM

    ! prp_sendHalo
    hasSendHalo_var%variable%label = 'hasSendHalos'
    hasSendHalo_var%variable%nComponents = 1
    hasSendHalo_var%do => deriveSendHalo

    allocate( meshInfoList( nMeshVars ))

    meshInfoList = (/ treeID_var, mpiProc_var, elemVol_var, volFrac_var,       &
      &               fluidify_var, solid_var, hasIBM_var, hasSendHalo_var /)

  end subroutine tem_initMeshInfoList
!******************************************************************************!


!******************************************************************************!
  !> Get the volume of an element
  !!
?? copy :: derive_headtxt(deriveTreeID) 
    !---------------------------------------------------------------------------
    ! local variable
    integer :: iElem, nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    integer(kind=long_k), allocatable :: treeID(:)
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1
    ! get the treeIDs
    call tem_treeIDfrom_subTree( subTree, tree, treeID, bound )

    do iElem = 1, nElems
      local_res( iElem ) = real( treeID( iElem ) , kind=rk)
    end do  ! iElem

  end subroutine deriveTreeID
!******************************************************************************!


!******************************************************************************!
  !> Get the MPI rank to which the corresponding treeID belongs.
  !! This routine is used to visualize the rank affinity of the mesh elements
  !!
?? copy :: derive_headtxt(deriveMPIproc) 
    !---------------------------------------------------------------------------
    ! local variable
    integer :: iElem
    integer :: nElems
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      local_res( iElem ) = real( tree%global%myPart, kind=rk)
    end do  ! iElem

  end subroutine deriveMPIproc
!******************************************************************************!


!******************************************************************************!
  !> Get the volume of an element
  !!
?? copy :: derive_headtxt(deriveElemVol)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nElems
    integer :: pos
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dx
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    do iElem = 1, nElems
      !\todo: SZ: this might be pretty expansive
      ! first get the position of the treeID in the tree
      if( subTree%useGlobalMesh )then
        pos = iElem
      else if( allocated( subTree%treeID ))then
        pos = tem_PosOfId( subTree%treeID(iElem), tree%treeID )
      else
        pos = subTree%map2global( iElem )
      end if

      ! if this exists ... 
      if( pos .gt. 0 )then
        ! ... check for the property
        ! \todo: SZ: USE THE SUBTREE OR GLOBAL TREE HERE?? 
        !            Maybe need to distinguish between different propertyBits
        !            arrays
        if( .not. btest( tree%ElemPropertyBits( pos ), prp_fluidify ))then
          ! calculate without qVals
          dx = tem_ElemSize( tree, tree%treeID( pos )) 
          local_res( iElem ) = dx*dx*dx
        else
          ! calculate using qVals
          local_res( iElem ) = 0._rk
        end if
      else
        local_res( iElem ) = 0._rk
      end if
    end do  ! iElem

  end subroutine deriveElemVol
!******************************************************************************!


!******************************************************************************!
  !> compute the volume of the current element as a fraction of a reference
  !! bounding hexahedral volume (stored in tree%global%effLength).
  !! The sum of all element fractions then gives the fractional fluid volume
  !! occupation inside the reference volume, i.e. the porosity
  !!
?? copy :: derive_headtxt(deriveVolFrac)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nElems
    integer :: pos
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    real(kind=rk) :: dx
    real(kind=rk) :: boundVol
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! calculate the volume of the shape defined by the tree (only rectangulars)
    boundVol = subTree%global%effLength(1) * subTree%global%effLength(2)       &
      &      * subTree%global%effLength(3)

    do iElem = 1, nElems
      !\todo: SZ: this might be pretty expensive
      ! first get the position of the treeID in the tree
      if( subTree%useGlobalMesh )then
        pos = iElem
      else if( allocated( subTree%treeID ))then
        pos = tem_PosOfId( subTree%treeID(iElem), tree%treeID )
      else
        pos = subTree%map2global( iElem )
      end if
      ! if this exists ... 
      if( pos .gt. 0 )then
        ! ... check for the property
        if( .not. btest( tree%ElemPropertyBits( pos ), prp_fluidify ))then
          dx = tem_ElemSize( tree, tree%treeID( pos )) 
          local_res( iElem ) = dx*dx*dx/boundVol
        else
          local_res( iElem ) = 0._rk
        end if
      else
        local_res( iElem ) = 0._rk
      end if
    end do  ! iElem

  end subroutine deriveVolFrac
!******************************************************************************!


!******************************************************************************!
  !>  Evaluate if the element is fluidifiable or not
  !!
?? copy :: derive_headtxt(deriveSolid)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nElems
    integer :: pos
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! \todo: SZ: CHECK subTree vs global tree again
    do iElem = 1, nElems
      if( subTree%useGlobalMesh )then
        pos = iElem
      else if( allocated( subTree%treeID ))then
        pos = tem_PosOfId( subTree%treeID(iElem), tree%treeID )
      else
        pos = subTree%map2global( iElem )
      end if

      if( btest( tree%ElemPropertyBits( pos ), prp_solid ))then
        local_res( iElem ) = 1.0_rk
      else
        local_res( iElem ) = 0.0_rk
      end if
    end do  ! iElem

  end subroutine deriveSolid
!******************************************************************************!


!******************************************************************************!
  !>  Evaluate if the element is fluidifiable or not
  !!
?? copy :: derive_headtxt(deriveFluidify)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nElems
    integer :: pos
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! \todo: SZ: CHECK subTree vs global tree again
    do iElem = 1, nElems
      if( subTree%useGlobalMesh )then
        pos = iElem
      else if( allocated( subTree%treeID ))then
        pos = tem_PosOfId( subTree%treeID(iElem), tree%treeID )
      else
        pos = subTree%map2global( iElem )
      end if

      if( btest( tree%ElemPropertyBits( pos ), prp_fluidify ))then
        local_res( iElem ) = 1.0_rk
      else
        local_res( iElem ) = 0.0_rk
      end if
    end do  ! iElem

  end subroutine deriveFluidify
!******************************************************************************!


!******************************************************************************!
  !>  Evaluate if the element has an IBM or not
  !!
?? copy :: derive_headtxt(deriveIBM)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nElems
    integer :: pos
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! \todo: SZ: CHECK subTree vs global tree again
    do iElem = 1, nElems
      if( subTree%useGlobalMesh )then
        pos = iElem
      else if( allocated( subTree%treeID ))then
        pos = tem_PosOfId( subTree%treeID(iElem), tree%treeID )
      else
        pos = subTree%map2global( iElem )
      end if
      if( btest( tree%ElemPropertyBits( pos ), prp_hasIBM ))then
        local_res( iElem ) = 1.0_rk
      else
        local_res( iElem ) = 0.0_rk
      end if
    end do  ! iElem

  end subroutine deriveIBM
!******************************************************************************!


!******************************************************************************!
  !>  Evaluate if the element has an sendHalo bit set or not
  !!
?? copy :: derive_headtxt(deriveSendHalo)
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nElems
    integer :: pos
    real(kind=rk), pointer, dimension(:) :: local_res => NULL()
    !---------------------------------------------------------------------------

    if( present(res) )then
      local_res => res
    else
      local_res => input
    end if

    ! total number of elements (+1 due to the definition of bound)
    nElems = bound(2)-bound(1)+1

    ! \todo: SZ: CHECK subTree vs global tree again
    do iElem = 1, nElems
      if( subTree%useGlobalMesh )then
        pos = iElem
      else if( allocated( subTree%treeID ))then
        pos = tem_PosOfId( subTree%treeID(iElem), tree%treeID )
      else
        pos = subTree%map2global( iElem )
      end if
      if( btest( tree%ElemPropertyBits( pos ), prp_sendHalo ))then
        local_res( iElem ) = 1.0_rk
      else
        local_res( iElem ) = 0.0_rk
      end if
    end do  ! iElem

  end subroutine deriveSendHalo
!******************************************************************************!


end module tem_meshInfo_module
!******************************************************************************!
