!> This module applies operator functionality to a set of tracking quantities
!! \author Nikhil Anand
!!
module tem_operator_module

  ! include treelm modules
  use mpi
  use env_module,            only: rk, rk_mpi, labelLen, long_k
  use tem_comm_env_module,   only: tem_comm_env_type
  use treelmesh_module,      only: treelmesh_type
  use tem_tools_module,      only: upper_to_lower
  use tem_geometry_module,   only: tem_GetLocalBoundingCube, tem_elemSize
  use tem_var_system_module, only: tem_var_system_type, tem_var_sys_init,     &
                                   tem_var_append, tem_var_sys_mark     
  use tem_logging_module,    only: logUnit

  ! include aotus modules
  use aotus_module,     only: aot_get_val, flu_State
  use aot_table_module, only: aot_table_open, aot_table_close,                 &
    &                         aot_table_length
  use tem_variable_module, only: tem_variable_type
  implicit none

  private

  public :: tem_operator_type
  public :: tem_operator_init
  public :: tem_load_operator

  !> This data type is providing the input for the reduction routines
  !! It must be filled by the solver, before the reduction is called
  !! It exists on each process
  type tem_operator_type
    !> Which operation to perform on the  list of elements
    character(len=labelLen) :: operation = ''
    !> solver-provided variable systems
    type(tem_var_system_type ) :: varSys
  end type tem_operator_type

contains

!******************************************************************************!
  !> read configuration file
  !!
  subroutine tem_load_operator(conf, me, parent, key, iError )
    !---------------------------------------------------------------------------
    !> handle for lua file
    type(flu_State)        :: conf
    !> the reduction file to fill
    type(tem_operator_type),intent(inout) :: me
    !> handle for reduce table
    integer, optional,intent(in)   :: parent
    !> which key to open
    character(len=*),optional,intent(in) :: key
    !> handle for reduce table
    integer, optional,intent(out)        :: iError
    !---------------------------------------------------------------------------
    integer          :: iErr
    character(len=labelLen) :: localKey
    !---------------------------------------------------------------------------
    if( present( key )) then
      localKey = key
    else
      localKey = 'operator'
    endif

    ! load operator which is always scalar.
    ! same operator is applied to all variables within a tracking entry
    call aot_get_val( L       = conf,                                          &
      &               key     = localKey,                                      &
      &               thandle = parent,                                        &
      &               val     = me%operation,                                  &
      &               ErrCode = iErr,                                          &
      &               default = 'none' )

    me%operation = upper_to_lower( me%operation ) 
    ! Check the chosen operation
    select case( trim(me%operation) )
      case('grad','gradient')
        me%operation = 'grad'
      case('none')
      case default
        write(logUnit(1),*)' Error: The chosen operator '//trim(me%operation) &
          &            //' is not defined.'
        iErr = -1
    end select

    if( present( iError )) iError = iErr

  end subroutine tem_load_operator
!******************************************************************************!

!******************************************************************************!
  !> Initialize operator objects according to the variable systems
  !!
  subroutine tem_operator_init( trackVarSys, operVarSys )
    !---------------------------------------------------------------------------
    !> tracking_type to initialize operators
    type(tem_var_system_type ), intent(in) :: trackVarSys
    type(tem_var_system_type ), intent(inout) :: operVarSys
    !---------------------------------------------------------------------------
    integer :: iVar, nComp
    !---------------------------------------------------------------------------
    
    write(logUnit(1),*)' Initializing operator ...'

    !Initialize the operator Varsys!
    call tem_var_sys_init( sys = operVarSys )
    ! Build up the operator varsys !
    do iVar = 1, trackVarsys%nVars
      nComp      = trackVarsys%variable%val(iVar)%nComponents
      call build_opVarsys(me  = trackVarSys, &
                 &   me_temp  = operVarSys, &
                 &    iVar    = iVar, &
                 &nComponents = nComp)
    enddo

  end subroutine tem_operator_init
!******************************************************************************!



!******************************************************************************!
  subroutine build_opVarsys(me, me_temp, iVar, nComponents)
    type( tem_var_system_type ), intent(in) :: me
    type( tem_var_system_type ), intent(inout):: me_temp
    integer, intent(in) :: nComponents, iVar
    character(len=25)  :: Varname
    integer :: pos
    varname = 'grad_'//me%variable%val(iVar)%label
    call tem_var_append( sys = me_temp , &
                       varname = trim(Varname), &
                       nComponents = 3*nComponents, &
                       pos = pos )
    call tem_var_sys_mark(sys = me_temp, &
                        systemName = me_temp%SystemName)

  end subroutine build_opVarsys


end module tem_operator_module
!******************************************************************************!
