! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> point type definition and routines used to generate points
!!
module tem_point_module

  ! include treelm modules
  use env_module,         only: rk
  use tem_logging_module, only: tem_toStr, logUnit

  ! include aotus modules
  use aotus_module,     only: aot_get_val, flu_State, aoterr_Fatal
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open_table, aot_out_close_table

  implicit none
  private

  public :: tem_point_type
  public :: tem_load_point
  public :: tem_getNextCoordOfPoint
  public :: tem_point_out

  !> Definition of a point object
  type tem_point_type
    !> real world coordinates of tracking point
    real(kind=rk) :: coord(3)
  end type tem_point_type

  interface tem_point_out
    module procedure tem_point_out_scal
    module procedure tem_point_out_vec
  end interface tem_point_out

contains

!******************************************************************************!
  !> Loading points from the config file
  !! valid definitions:
  !! * None at all, picks the default (origin of bounding cube)
  !! * point = { 1.0, 0.0, -1.0 } -- single point
  !! * point = { {1.0, 0.0, -1.0} } -- single point
  !! * point = { {1.0, 0.0, -1.0},
  !!             {2.0, 0.0, -1.0} } -- multiple points
  !!
  subroutine tem_load_point( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> the array of points which to read in (and first allocate)
    type(tem_point_type), allocatable, intent(out) :: me(:)
    !> the lua config handle
    type(flu_state) :: conf
    !> table handle from which to read
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    integer :: points_thandle
    integer :: iError(3), errFatal(3)
    integer :: iPoint
    integer :: nPoints
    !---------------------------------------------------------------------------
    errFatal = aoterr_fatal

    !assume it is single point and try to load the point
    nPoints = 1
    allocate(me(1))
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               val     = me(1)%coord,                                   &
      &               key     = 'point',                                       &
      &               ErrCode = iError )
    if (any(btest( iError, errFatal )) ) then
      ! try to load points as multiple table
      deallocate(me)
      call aot_table_open( L       = conf,                                     &
        &                  parent  = thandle,                                  &
        &                  thandle = points_thandle,                           &
        &                  key     = 'point' )
      nPoints = aot_table_length( L = conf, thandle = points_thandle )
      allocate(me(nPoints))
      do iPoint=1,nPoints
        call aot_get_val( L       = conf,                                      &
          &               thandle = points_thandle,                            &
          &               val     = me(iPoint)%coord,                          &
          &               pos     = iPoint,                                    &
          &               ErrCode = iError )
        if (any(btest( iError, errFatal )) ) then
          write(*,*) 'ERROR reading the point, coordinates not well'
          write(*,*) 'defined. STOPPING'
          STOP
        end if
      enddo
    end if

    do iPoint = 1, nPoints
      write(logUnit(1),*) ' Read Point: '//trim(tem_toStr(iPoint)//' with ' )
      write(logUnit(1),*) '  origin       '//trim( tem_toStr(                &
        &                   val = me(iPoint)%coord,                            &
        &                   sep = ',' ))
    enddo

  end subroutine tem_load_point
!******************************************************************************!


!******************************************************************************!
  !> Return the next coordinate of the point shape
  !!
  !! It actually returns just the point coordinate. However, this function is
  !! used for other shapes as well, where more than one point have to be
  !! returned.
  function tem_getNextCoordOfPoint( me ) result( coord )
    !---------------------------------------------------------------------------
    !> current point
    type(tem_point_type ) :: me
    !> calulated real-world coordinate, which is returned
    real(kind=rk)       :: coord(3)
    !---------------------------------------------------------------------------

    coord = me%coord

  end function tem_getNextCoordOfPoint
!******************************************************************************!


!******************************************************************************!
  !> Write a array of points to lua file
  !!
  subroutine tem_point_out_vec( me, conf )
    !--------------------------------------------------------------------------
    !> point types to write out
    type( tem_point_type ), intent(in) :: me(:)
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !--------------------------------------------------------------------------
    integer :: i
    !--------------------------------------------------------------------------

    ! create a table with name point
    call aot_out_open_table( put_conf = conf, tname = 'point' )

    do i = 1, size(me)
      call tem_point_out_scal( me(i), conf )
    end do

    call aot_out_close_table( put_conf = conf )

  end subroutine tem_point_out_vec
!******************************************************************************!


!******************************************************************************!
  !> Write a point to lua file
  !!
  subroutine tem_point_out_scal( me, conf )
    !--------------------------------------------------------------------------
    !> point types to write out
    type( tem_point_type ), intent(in) :: me
    !> Aotus type handling the output to the file in lua format
    type(aot_out_type), intent(inout) :: conf
    !--------------------------------------------------------------------------

    ! create a table with name point if not exist
    if( conf%level .eq. 0 ) then
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'point',                                    &
        &               val      = me%coord )
    else
      call aot_out_val( put_conf = conf,                                       &
        &               val      = me%coord )
    end if

  end subroutine tem_point_out_scal
!******************************************************************************!


end module tem_point_module
!******************************************************************************!
