!> Description of solver specific data that has to be passed to treelm
!! functions.
!!
!! To use it, create the data using aot_out to write the data as Lua tables
!! to a scratch file (should only be done by the root process).
!! Then load this data into the tem_solveData_type, using tem_solveData_load.
module tem_solveData_module
  use mpi

  use env_module, only: SolSpecLen, PathLen
  use aotus_module, only: flu_State, aot_get_val, open_config_chunk
  use aot_out_module, only: aot_out_type, aot_out_open_table, &
    &                       aot_out_close_table
  use aot_table_module, only: aot_table_open, aot_table_close
  use tem_tools_module, only: tem_file_to_string

  implicit none

  private

  type tem_solveData_type
    !> Character string containing solver specific data as a Lua script.
    character(len=SolSpecLen) :: string = ''

    !> Handle to the opened Lua script with the solver specific data.
    type(flu_State) :: conf
  end type tem_solveData_type

  interface tem_solveData_load
    module procedure solveData_load_fromConf
    module procedure solveData_load_dist_fromUnit
  end interface tem_solveData_load

  public :: tem_solveData_type
  public :: tem_solveData_load
  public :: tem_solveData_as_Lua_string

contains

  !> Load solver specific data from a configuration file.
  !!
  !! \todo HK: why does this solver specific string have to be inside a table?
  subroutine solveData_load_fromConf(me, key, conf, parent)
    !> Solver specific data to fill
    type(tem_solveData_type), intent(out) :: me
    !> Name of the solver specific data in the given conf
    character(len=*), intent(in) :: key
    !> Lua script to load the solver specific data from
    type(flu_state) :: conf
    !> Handle to the parent table
    integer, intent(in), optional :: parent

    integer :: solSpec_handle
    integer :: iError

    ! open the solver definition table
    call aot_table_open( L       = conf,                                       &
      &                  parent  = parent,                                     &
      &                  thandle = solSpec_handle,                             &
      &                  key     = key )

    call aot_get_val( L       = conf,                                          &
      &               thandle = solSpec_handle,                                &
      &               val     = me%String,                                     &
      &               ErrCode = iError,                                        &
      &               pos     = 1,                                             &
      &               default = '' ) 
    
    ! close the solver definition table again
    call aot_table_close( L= conf, thandle = solSpec_handle )

    call open_config_chunk( L     = me%conf,                                   &
      &                     chunk = trim( me%String ))
  end subroutine solveData_load_fromConf
  !----------------------------------------------------------------------------!
  !----------------------------------------------------------------------------!


  !----------------------------------------------------------------------------!
  !> Load the solver specific data from a unit on the root rank of comm.
  !!
  !! Rank 0 has to have the file_unit opened and filled with a Lua script,
  !! describing the solver specific data.
  !! This is then used to fill the solver specific data of me on all ranks.
  subroutine solveData_load_dist_fromUnit(me, file_unit, rank, comm)
    !> Solver specific data to fill
    type(tem_solveData_type), intent(out) :: me
    !> File unit on rank 0, that contains the solver specific data
    integer, intent(in) :: file_unit
    !> Rank of the executing process in the communicator comm
    integer :: rank
    !> Handle for the MPI communicator
    integer :: comm

    integer :: iError
    integer :: actlen

    if (rank == 0) then
      call tem_file_to_string(file_unit, me%string, iError)
      actlen = len_trim(me%string)
    end if
    call MPI_Bcast(actlen, 1, MPI_INTEGER, 0, comm, iError)
    call MPI_Bcast(me%string, solspeclen, MPI_CHARACTER, 0, comm, iError)

    call open_config_chunk( L     = me%conf,                                   &
      &                     chunk = trim( me%String ))
  end subroutine solveData_load_dist_fromUnit
  !----------------------------------------------------------------------------!
  !----------------------------------------------------------------------------!


  !----------------------------------------------------------------------------!
  !> Write the solver specific data as a string to a Lua script.
  !!
  !! The data is put into quotations and written to the given out_unit, under
  !! the name given as key.
  !! If the solver specific data is empty, nothing will be written.
  subroutine tem_solveData_as_Lua_string(me, conf, key)
    !> Solver specific data to write
    type(tem_solveData_type), intent(in) :: me
    !> File to write to
    type(aot_out_type) :: conf
    !> Name for the solver specific data to use in the Lua script
    character(len=*), intent(in) :: key

    character(len=PathLen) :: line
    character(len=2) :: concat
    integer :: lb_pos
    integer :: last_pos
    integer :: maxlen

    maxlen = len_trim(me%string)
    if (maxlen > 0) then
      call aot_out_open_table(put_conf = conf, tname = trim(key))
      concat = ''
      last_pos = 0
      do
        lb_pos = index(me%string(last_pos+1:), new_line('A'))
        if (lb_pos > 0) then
          line = me%string(last_pos+1:last_pos+lb_pos-1)
          last_pos = min(last_pos + lb_pos, maxlen)
          write(conf%outUnit,'(a)') trim(concat)//'"'//trim(line)//'"'
          concat = '..'
        else
          if (len_trim(me%string(last_pos+1:)) > 0) then
            write(conf%outUnit,'(a)') trim(concat) &
              &                       //'"'//trim(me%string(last_pos+1:)) &
              &                       //'"'
          end if
          EXIT
        end if
      end do
      call aot_out_close_table(conf)
    end if

  end subroutine tem_solveData_as_Lua_string

end module tem_solveData_module
