! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module contains subroutines needed to create and update subTrees from 
!! corresponding global trees.
!! 
!! The subroutine tem_create_subTree_of creates a subTree from a global tree 
!! using an array of shapes. This functionality is used when certain operations
!! (e.g. tracking, source terms) shall be executed only on a subset of the fluid
!! tree (or another tree).
!!
module tem_subTree_module

  ! include treelm modules
  use mpi
  use env_module,              only: long_k, globalMaxLevels, pathLen
  use tem_aux_module,          only: tem_abort
  use tem_tools_module,        only: tem_horizontalSpacer
  use tem_debug_module,        only: tem_debug_type, main_debug
  use treelmesh_module,        only: treelmesh_type
  use tem_canonicalND_module,  only: tem_cano_initSubTree
  use tem_dyn_array_module,    only: dyn_intArray_type, init, append, destroy
  use tem_property_module,     only: gather_property
  use tem_shape_module,        only: tem_shape_type, tem_global_shape,         &
    &                                tem_canonicalND_shape, tem_property_shape,&
    &                                tem_local_shape
  use tem_geometry_module,     only: tem_setEffBoundingBox, tem_baryOfId
  use tem_topology_module,     only: tem_levelOf
  use tem_construction_module, only: tem_levelDesc_type
  use tem_subTree_type_module, only: tem_subTree_type, tem_subTree_from,       &
    &                                tem_dump_subTree
  use tem_logging_module,      only: logUnit

  implicit none

  private

  public :: tem_updatePropertyBits
  public :: tem_create_subTree_of
  public :: tem_write_debugMesh

  !> interface for copying the property bits from the global tree or the 
  !! level descriptor
  interface tem_copyPropertyBits
    module procedure tem_copyPropertyBitsFromLevelDesc
    module procedure tem_copyPropertyBitsFromTree
  end interface

  contains

!******************************************************************************!
  !> Update the property of subTree with the ones from inTree, if something 
  !! changed update the logical meshChange.
  !!
  subroutine tem_updatePropertyBits( inTree, subTree )
    !---------------------------------------------------------------------------
    !> tree to get information from
    type(treelmesh_type), intent(in) :: inTree
    !> tree to pass information to
    type(tem_subTree_type), intent(inout) :: subTree
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: elemPos
    !---------------------------------------------------------------------------

    do iElem = 1, subTree%nElems
      ! get the position in the global tree
      elemPos = subTree%map2global(iElem)
      ! test the individual property bits
      if ( inTree%ElemPropertyBits(elempos)   &
        &  /= subTree%ElemPropertyBits(iElem) ) then
        ! update the propertyBits in the subTree
        subTree%ElemPropertyBits(iElem) = inTree%ElemPropertyBits(elemPos)
        subTree%global%meshChange = .true.
      end if
    end do

  end subroutine tem_updatePropertyBits
!******************************************************************************!


!******************************************************************************!
  !> Copy the properties of the level decriptor to the ones in subTree.
  !!
  subroutine tem_copyPropertyBitsFromLevelDesc( levelDesc, subTree )
    !---------------------------------------------------------------------------
    !> level descriptor including all elements (fluid, ghost, halo)
    type(tem_levelDesc_type ),intent(in) :: levelDesc(:)
    !> tree to pass information to
    type(tem_subTree_type), intent(inout) :: subTree
    !---------------------------------------------------------------------------
    integer :: iLevel
    integer :: nElems
    integer :: elemCounter
    !---------------------------------------------------------------------------
    allocate( subTree%ElemPropertyBits( subTree%nElems ))

    elemCounter = 0
    do iLevel = 1, size(levelDesc)
      nElems = size(levelDesc(iLevel)%property(:))
      subTree%ElemPropertyBits(elemCounter+1 : elemCounter+nElems) =           &
        &                                         levelDesc(iLevel)%property(:)
      elemCounter = elemCounter + nElems
    end do
  end subroutine tem_copyPropertyBitsFromLevelDesc
!******************************************************************************!


!******************************************************************************!
  !> Copy the properties of inTree to the ones in subTree.
  !!
  subroutine tem_copyPropertyBitsFromTree( inTree, subTree )
    !---------------------------------------------------------------------------
    !> tree to get information from
    type(treelmesh_type), intent(in) :: inTree
    !> tree to pass information to
    type(tem_subTree_type), intent(inout) :: subTree
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: elemPos
    integer :: iProp
    !---------------------------------------------------------------------------
    ! allocate the bitfield
    allocate(subTree%elemPropertyBits(subTree%nElems))

    ! initialize the bitfield with 0
    subTree%elemPropertyBits = 0

    do iElem = 1, subTree%nElems
      ! get the position in the global tree
      elemPos = subTree%map2global(iElem)
      ! copy the propertyBits in the subTree
      subTree%ElemPropertyBits(iElem) = inTree%ElemPropertyBits(elemPos)
    end do

    ! update the property information
    allocate(subTree%property(subTree%global%nProperties))

    do iProp=1,subTree%global%nProperties
      call gather_Property( Property = subTree%Property(iProp),                &
        &                   Header   = subTree%global%Property(iProp),         &
        &                   BitField = subTree%ElemPropertyBits,               &
        &                   comm     = subTree%global%comm )
    end do

  end subroutine tem_copyPropertyBitsFromTree
!******************************************************************************!


!******************************************************************************!
  !> Create a subTree from a given inTree and an array of shapes.
  !!
  !! All elements to the inTree%treeID are collected into the subTree%map2global
  !! array and a sub-communicator with participating processes is created.
  !! Additionally how many and which elements exist on my local process and are
  !! requested from the shapes is identified.
  !!
  subroutine tem_create_subTree_of( inTree, subTree, inShape, levelDesc,       &
    &                               prefix )
    !---------------------------------------------------------------------------
    !> Global mesh from which the elements are identified and then stored to
    type(treelmesh_type), intent(in)                :: inTree
    !> new mesh
    type(tem_subTree_type), intent(out)             :: subTree
    !> shape objects on which to work
    type(tem_shape_type ),intent(inout)             :: inShape(:)
    !> optional level descriptor needed for local shape
    type(tem_levelDesc_type ), optional, intent(in) :: levelDesc(:)
    !> prefix for the subTree
    character(len=*), optional, intent(in)          :: prefix
    !---------------------------------------------------------------------------
    ! local growing array for the map to the global tree
    type(dyn_intArray_type) :: local_map2global
    integer, allocatable :: map2globalTemp(:)
    integer(kind=long_k), allocatable :: tmp_treeID(:)
    integer :: iShape, iElem
    ! mpi variables for parallel writing of the log-wise trees
    integer :: color, iError
    ! if the local rank has part of the tracking object
    logical :: participateInMesh
    integer :: local_countElems( globalMaxLevels )
    !--------------------------------------------------------------------------

    subTree%useGlobalMesh = .false.
    subTree%useLocalMesh = .false.

    ! Prepare output structure
    ! init here the sub-tree mesh
    subTree%global = inTree%global
    ! initialize the number of elements on this partition with 0
    subTree%nElems = 0
    subTree%global%nElems = 0
    if( present( prefix ))                                                     &
      &   subTree%global%dirname = trim(prefix)

    ! Reset counter for actual elements for this shape object
    local_countElems = 0
    call init( local_map2global, unique = .true., length = 128)
    do iShape = 1, size( inShape )
      select case( inShape(iShape)%shapeID )
      ! Use elements belonging to a canonical n-dimensional shape
      case( tem_canonicalND_shape )
        write(logUnit(1),*) 'Found canonical shape ', iShape
        call tem_cano_initSubTree( inShape( iShape )%canoND(:), inTree,        &
          &                        local_countElems, local_map2global )
      ! Use all elements for this shape
      case( tem_global_shape )
        write(logUnit(1),*) 'Found global shape ', iShape
        ! Global shape means, all elements are part of the shape.
        ! so nothing to do, just continue with the next shape
        local_countElems = 1
        subTree%useGlobalMesh = .true.
        exit
      ! Only use elements with a certain property
      case( tem_property_shape )
        call tem_shape_initPropElements( inShape(iShape), inTree,              &
          &                              local_countElems, local_map2global )

      ! elements on the local partition
      case( tem_local_shape )
        write(logUnit(1),*) 'Found local shape ', iShape
        if(present(levelDesc)) then
          subTree%useLocalMesh = .true.
          call tem_shape_initLocal( levelDesc, tmp_treeID )
        else
          write(logUnit(1),*)'ERROR:levelDesc is not passed to '//             &
            &            'tem_creat_subTree_of'
          write(logUnit(1),*)'      levelDesc is required to create '//        &
            &                'subTree for tem_local_shape'
          call tem_abort()
        end if  
        local_countElems = 1
        ! update the communicator
        subTree%global%comm = mpi_comm_self
        exit
      end select ! shapeID
    end do ! iShape

    if( .not. subTree%useGlobalMesh ) then
      subTree%global%predefined = ''
    end if

    ! Exchange information about the mesh
    ! mpi_exscan: track( iLog )%tree%offset
    ! Now create a new communicator with ranks that own parts of the sub-mesh
    ! This is done with the color
    participateInMesh = ( sum( local_countElems ) > 0 )

    if( participateInMesh ) then
      color = 1
    else
      ! Not participating, set color to undefined
      color = MPI_UNDEFINED
    end if

    ! Input rank as a key to keep sorting as in global communicator
    ! reorder process ranks into tracking communicator
    call mpi_comm_split( inTree%global%comm, color, inTree%global%myPart,      &
      &                  subTree%global%comm, iError  )
    if( iError .ne. mpi_success ) then
      write(logUnit(1),*)' There was an error splitting the communicator'
      write(logUnit(1),*)' for the tracking subset. Aborting...'
      call tem_abort()
    end if
    if( participateInMesh ) then
      call mpi_comm_size( subTree%global%comm, subTree%global%nParts, iError )
      call mpi_comm_rank( subTree%global%comm, subTree%global%myPart, iError )
      if( .not. subTree%useGlobalmesh) then

        if( subTree%useLocalMesh ) then
          call tem_subTree_from( me     = subTree,                             &
            &                    treeID = tmp_treeID )
          ! update the property list of the newly defined tree according to the 
          ! the local shape (simply copy the bits)          
          call tem_copyPropertyBits( levelDesc = levelDesc, subTree = subTree )
        else
          ! Copy the created dynamic local_tIDlist to a long integer array
          allocate( map2globalTemp( local_map2global%nVals ))
          ! Assign list of links to the temporary list, first sorting
          do iElem = 1, local_map2global%nVals
            map2globalTemp( iElem ) = local_map2global%val(                    &
              &                               local_map2global%sorted( iElem ))
          end do
          
          call tem_subTree_from( me         = subTree,                         &
            &                    map2global = map2globalTemp )
          ! the global tree
          call tem_copyPropertyBits( inTree = inTree, subTree = subTree )
         deallocate( map2globalTemp )
        end if ! useLocalMesh with level desc

        ! Calculate the effective extents of the tracking sub-tree bounding box
        call tem_setEffBoundingBox( subTree, inTree )

        if( subTree%global%myPart == 0 ) then
          write(*,*) 'Found corresponding nElems = ', subTree%global%nElems
          if( subTree%global%nElems == 1 .and.                                 &
            & allocated(subTree%map2global))then
            write(*,*) 'Barycenter of tracking element: ',                     &
              &        tem_baryOfId(inTree,                                    &
              &                     inTree%treeID(subTree%map2global(1)))
          end if
        end if
      else
        ! assign the global properties to the subTree
        subTree%nElems = inTree%nElems
        subTree%global%nProperties = inTree%global%nProperties
        allocate(subTree%property(subTree%global%nproperties))
        subTree%Property = inTree%Property
        allocate(subTree%ElemPropertyBits(subTree%nElems))
        subTree%ElemPropertyBits = inTree%ElemPropertyBits
      end if ! not global mesh?
    end if ! Participate in mesh?

    call destroy( me = local_map2global )
    call tem_horizontalSpacer( fUnit= logUnit(1) )

  end subroutine tem_create_subTree_of
!******************************************************************************!


!******************************************************************************!
  !> This routine collects the positions of all elements with a certain property
  !! in the inTree%treeID array (e.g. for shape kind='property').
  !!
  subroutine tem_shape_initPropElements( me, inTree, countElems, map2global )
    !---------------------------------------------------------------------------
    !> shape objects on which to work
    type(tem_shape_type ) :: me
    !> Global mesh from which the elements are identified and then stored to
    type(treelmesh_type), intent(in) :: inTree
    !> How many elements there will be for each level in the track
    integer, intent( out ) :: countElems( globalMaxLevels )
    !> growing array for the map2global
    type(dyn_intArray_type), intent(inout) :: map2global
    !---------------------------------------------------------------------------
    integer(kind=long_k)  :: treeID, elemProp, match
    integer :: tLevel, dPos, iElem
    logical :: wasAdded
    !---------------------------------------------------------------------------

    write(logUnit(1),*) 'Initializing shapes of elements that have a '//       &
      &             'certain property' 

    ! Count how many elements are really local
    do iElem = 1, inTree%nElems

      ! get its property bits
      elemProp = inTree%elemPropertyBits( iElem )
      ! check whether its property
      match =  iand( me%propBits, elemProp )
      if( match > 0_long_k ) then
        ! Get its TreeID
        treeID  = inTree%treeID( iElem )
        ! Append to treeID list (note that already existing ones are
        ! omitted)
        call append( me       = map2global,                                    &
          &          pos      = dPos,                                          &
          &          val      = iElem,                                         &
          &          wasAdded = wasAdded )
        ! Count up if it was added
        if( wasAdded ) then
          tLevel   = tem_levelOf( treeID )
          countElems( tLevel ) = countElems( tLevel ) + 1
        end if ! wasAdded
      end if ! match > 0
    end do ! iElem

  end subroutine tem_shape_initPropElements
!******************************************************************************!


!******************************************************************************!
  !> This subroutine collects the treeIDs of all elements in the level 
  !! descriptor on the local partition (e.g. used in debug mesh, no general
  !! option for now).
  !!
  subroutine tem_shape_initLocal( levelDesc, treeID )
    !---------------------------------------------------------------------------
    !> level descriptor including all elements (fluid, ghost, halo)
    type(tem_levelDesc_type ),intent(in) :: levelDesc(:)
    !> temporary array of treeIDs
    integer(kind=long_k), allocatable, intent(out) :: treeID(:) 
    !---------------------------------------------------------------------------
    integer :: nTotalElems
    integer :: iLevel
    integer :: iElem
    !---------------------------------------------------------------------------

    write(logUnit(1),*) 'Initializing shape with all elements on this partition' 

    ! get the total number of elements in the level descriptor
    nTotalElems = sum(levelDesc(:)%nElems)

    ! allocate the array of treeIDs
    allocate(treeID(nTotalElems))

    ! copy the treeIDs to the array
    do iLevel = 1,size(levelDesc)
      do iElem = 1, size( levelDesc( iLevel )%total )
        treeID((iLevel-1)*size(levelDesc)+iElem) =                             &
          &                                      levelDesc(iLevel)%total(iElem)
      end do
    end do

  end subroutine tem_shape_initLocal
!******************************************************************************!


!******************************************************************************!
  !> Write the complete mesh including fluid, ghosts and halo elements to disk
  !!
  !! Define in the \ref tem_debug_module::tem_load_debug "debug table" as
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !!   debug = { debugMode = true,
  !!             debugMesh = true }
  !! ~~~~~~~~~~~~~~~~~~~~~
  !!
  subroutine tem_write_debugMesh( globtree, levelDesc, debug, myPart )
    !---------------------------------------------------------------------------
    !> mesh to locate the point in
    type(treelmesh_type), intent(in) :: globtree
    !> current level descriptor
    type(tem_levelDesc_type ),intent(inout) :: levelDesc(:)
    !> debug info
    type(tem_debug_type ), optional, intent(inout) :: debug
    !> Partition to use on the calling process (= MPI Rank in comm)
    integer, intent(in) :: myPart
    !---------------------------------------------------------------------------
    type(tem_shape_type ) :: inShape(1)
    character(len=pathLen) :: dirname
    type(tem_debug_type ) :: local_debug
    ! local subTree that is dumped to disc
    type( tem_subTree_type ) :: subTree
    !---------------------------------------------------------------------------

    if( present( debug ))then
      local_debug = debug
    else
      local_debug = main_debug
    end if

    if( trim( local_debug%debugMesh ) .ne. '')then
      write(logUnit(1),*)' Writing debug mesh to disk '
      inShape(1)%shapeID = tem_local_shape
      write(dirname ,'(a,i6.6,a)')trim( local_debug%debugMesh ),myPart,'_'

      call tem_create_subTree_of( inTree    = globtree,                        &
        &                         subTree   = subTree,                         &
        &                         inShape   = inShape,                         &
        &                         levelDesc = levelDesc,                       &
        &                         prefix    = trim( dirname ))

      call tem_dump_subTree( subTree, globtree, root_only = .false. )

    endif ! output debugmesh

  end subroutine tem_write_debugMesh
!******************************************************************************!


end module tem_subTree_module
!******************************************************************************!
