! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Timer infrastructure methods for the TreElM module
!!
!! This module provides a convenience timing functionality for coarse grain
!! measurements by encapsulating mpi_wtime and providing a set of timers to
!! track.
!!
module tem_timer_module

  ! include treelm modules
  use mpi
  use env_module,         only: rk, rk_mpi
  use tem_logging_module, only: logUnit

  implicit none
  private

  public :: tem_addTimer
  public :: tem_startTimer
  public :: tem_stopTimer
  public :: tem_resetTimer
  public :: tem_writeTimer
  public :: tem_writeTimers
  public :: tem_getTimerName
  public :: tem_getTimerVal
  public :: tem_timer_type
  public :: tem_getMaxTimerVal
  public :: tem_getMinTimerVal
  public :: tem_getAveTimerVal

  integer, parameter :: maxTimers = 64
  !> Timer handle for the overall run time
  !integer :: globalTimerHandle

  type tem_timer_type
    !> number of used timers
    integer :: nTimers = 0

    !> Start timing values
    real(kind=rk) :: tStart(maxTimers)     = 0._rk

    !>  timer running?
    logical(kind=rk) :: running(maxTimers) = .false.

    !> timing value
    real(kind=rk) :: value(maxTimers)      = 0._rk

    !> Fluctuations in a timer value (parallel)
    real(kind=rk) :: fluc (maxTimers)      = 0._rk

    !> timing object labels
    character(len=40) :: label(maxTimers)
  end type tem_timer_type

  ! KM: timerData is moved to general type
  !type(tem_timer_type), save :: timerData
  ! A timer type for internal timing analyses, this will not be written to disk

  contains

!******************************************************************************!
  !> Setup a new timer object, reset the values and give it a label
  !! for later identification
  !!
  subroutine tem_addTimer( me, timerHandle, timerName )
    !---------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> A handle to reference this timer
    integer, intent(out) :: timerHandle
    !> Name for this timer
    character(len=*), intent(in) :: timerName
    !---------------------------------------------------------------------------

    if(me%nTimers .ge. maxTimers) then
      write(logUnit(1),*) "Exceeded number of timer slots. stopping." 
      stop
    endif
    me%nTimers = me%nTimers + 1
    timerHandle = me%nTimers
    ! created a new timer handle and set the name
    me%label(timerHandle) = timerName

  end subroutine tem_addTimer
!******************************************************************************!


!******************************************************************************!
  !> Start the timer for the given timer handle
  !!
  subroutine tem_startTimer( me, timerHandle )
    !--------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> Handle of the timer to start
    integer, intent(in) :: timerHandle
    !--------------------------------------------------------------------------

    me%tStart(timerHandle)  = mpi_wtime()
    me%running(timerHandle) = .true.

  end subroutine tem_startTimer
!******************************************************************************!


!******************************************************************************!
  !> Reset the timer to zero for the given timer handle
  !!
  subroutine tem_resetTimer( me, timerHandle )
    !--------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> Handle of the timer to start
    integer, intent(in) :: timerHandle
    !--------------------------------------------------------------------------

    me%tStart(timerHandle) = mpi_wtime()
    me%value(timerHandle)  = 0.

  end subroutine tem_resetTimer
!******************************************************************************!


!******************************************************************************!
  !> Stop the timer for the given timer handle and update the timings
  !!
  subroutine tem_stopTimer( me, timerHandle )
    !---------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> Handle of the timer to stop
    integer, intent(in) :: timerHandle
    !---------------------------------------------------------------------------
    real(kind=rk) :: endTime
    !---------------------------------------------------------------------------

    ! Get the current time
    endTime = mpi_wtime()
    me%running(timerHandle) = .false.
    ! And add the measured difference to the last value
    me%value(timerHandle) =  me%value(timerHandle) + endTime               &
      &                   -  me%tStart(timerHandle)

  end subroutine tem_stopTimer
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its value
  !!
  subroutine tem_writeTimer( me, timerHandle, outunit )
    !--------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(in) :: me
    !> handle of the timer to be dumped
    integer, intent(in) :: timerHandle
    !> unit to write to
    integer, optional, intent(in) :: outunit
    !--------------------------------------------------------------------------
    integer :: loc_unit
    !--------------------------------------------------------------------------

    if( present( outunit ))then
      loc_unit = outunit
    else
      ! default is the screen output (unit 6)
      loc_unit = 6
    end if

    write(loc_unit,'(3a,f15.4,a)') ' Run-time ',                               &
      &                            trim(me%label(timerHandle)),                &
      &                            ':', me%value(timerHandle), ' s'

  end subroutine tem_writeTimer
!******************************************************************************!


!******************************************************************************!
  !> Write out all timers including name and values to the given outunit
  !!
  subroutine tem_writeTimers( me, outunit )
    !---------------------------------------------------------------------------
    !> timer object 
    type(tem_timer_type), intent(in) :: me
    !> unit to dump timers to
    integer, optional, intent(in) :: outunit
    !---------------------------------------------------------------------------
    integer :: loc_unit
    integer :: iTimer
    !---------------------------------------------------------------------------

    if( present( outunit ))then
      loc_unit = outunit
    else
      ! default is the screen output (unit 6)
      loc_unit = 6
    end if

    do iTimer = 1, me%nTimers
      call tem_writeTimer( me, iTimer, loc_unit)
    end do

  end subroutine tem_writeTimers
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its value
  !!
  function tem_getTimerName( me, timerHandle ) result( timerName )
    !---------------------------------------------------------------------------
    !> timer object 
    type(tem_timer_type), intent(inout) :: me
    !> timer handle
    integer, intent(in) :: timerHandle
    !> name of the timer
    character(len=64) :: timerName
    !---------------------------------------------------------------------------
    timerName = trim( me%label(timerHandle) )

  end function tem_getTimerName
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its value
  !!
  function tem_getTimerVal( me, timerHandle ) result( retValue )
    !---------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> timer handle
    integer, intent(in) :: timerHandle
    !> timer value
    real(kind=rk) :: retValue
    !---------------------------------------------------------------------------
    real(kind=rk) :: endTime
    !---------------------------------------------------------------------------
    if( .not. me%running(timerHandle) ) then
      retValue = me%value(timerHandle)
    else
      endTime = mpi_wtime()
      ! And add the measured difference to the last value
      retValue = endTime - me%tStart(timerHandle)
    end if

  end function tem_getTimerVal
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its max value across all partitions.
  !!
  function tem_getMaxTimerVal( me, timerHandle, comm ) result( retValue )
    !---------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> timer handle
    integer, intent(in) :: timerHandle
    !> communicator handle
    integer, intent(in) :: comm
    !> timer value
    real(kind=rk) :: retValue
    !---------------------------------------------------------------------------
    real(kind=rk) :: send, recv
    real(kind=rk) :: iError
    !---------------------------------------------------------------------------

    send = tem_getTimerVal( me, timerHandle )

    call MPI_Allreduce( send, recv, 1, rk_mpi, MPI_MAX, comm, iError )

    retValue = recv

  end function tem_getMaxTimerVal
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its min value across all partitions.
  !!
  function tem_getMinTimerVal( me, timerHandle, comm ) result( retValue )
    !---------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> timer handle
    integer, intent(in) :: timerHandle
    !> communicator handle
    integer, intent(in) :: comm
    !> timer value
    real(kind=rk) :: retValue
    !---------------------------------------------------------------------------
    real(kind=rk) :: send, recv
    real(kind=rk) :: iError
    !---------------------------------------------------------------------------

    send = tem_getTimerVal( me, timerHandle )

    call MPI_Allreduce( send, recv, 1, rk_mpi, MPI_MIN, comm, iError )

    retValue = recv

  end function tem_getMinTimerVal
!******************************************************************************!


!******************************************************************************!
  !> Write out the timer name and its average value across all partitions.
  !!
  function tem_getAveTimerVal( me, timerHandle, comm, nProcs ) result(retValue)
    !---------------------------------------------------------------------------
    !> timer object
    type(tem_timer_type), intent(inout) :: me
    !> timer handle
    integer, intent(in) :: timerHandle
    !> communicator handle
    integer, intent(in) :: comm
    !> communicator handle
    integer, intent(in) :: nProcs
    !> timer value
    real(kind=rk) :: retValue
    !---------------------------------------------------------------------------
    real(kind=rk) :: send, recv
    real(kind=rk) :: iError
    !---------------------------------------------------------------------------

    send = tem_getTimerVal( me, timerHandle )

    call MPI_Allreduce( send, recv, 1, rk_mpi, MPI_SUM, comm, iError )

    retValue = recv/nProcs

  end function tem_getAveTimerVal
!******************************************************************************!


end module tem_timer_module
!******************************************************************************!
