! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module gathers different transient effect treatments.
!! like: constant, lua function and predefined Fortran function. \n
!! Currently supports:
!! \li const: data is a constant value
!! \li lua_fun: data is a lua function
!! \li linear: data is ramped linearly
!! \li smooth: data is ramped with a smooth function
!! \li from_file: data is read from file (supports periodicity)
!!
!! \note Boundaries can be defined with superposed conditions in
!! space and time. Therefore the function defined here acts as a
!! factor applied to the overall boundary value.
!!
!! \author Kannan Masilamani, Simon Zimny
!!
module tem_transient_module

  ! include treelm modules
  use env_module,            only: rk, LabelLen, PathLen, newunit, eps
  use tem_aux_module,        only: tem_abort
  use tem_param_module,      only: PI
  use tem_time_module,       only: tem_time_type
  use tem_grow_array_module, only: grw_real2darray_type,                       &
    &                              init, append, expand, destroy
  use tem_logging_module,    only: logUnit

  ! include aotus modules
  use aot_path_module,  only: aot_path_type, assignment(=), aot_path_addNode,  &
    &                         aot_path_open, aot_path_close
  use aotus_module,     only: aoterr_Fatal, aoterr_NonExistent, flu_State,     &
    &                         aoterr_WrongType, aot_top_get_val, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close
  use aot_fun_module,   only: aot_fun_type, aot_fun_open, aot_fun_close,       &
    &                         aot_fun_put, aot_fun_do

  implicit none

  private

  public :: tem_transient_type
  public :: load_transient, tem_transient_for

  !> contains information for predefined transient functions
  type tem_predef_transient_type
    !> start time
    real(kind=rk) :: from_time
    !> end time
    real(kind=rk) :: to_time
    !> minimum factor at start time
    real(kind=rk) :: min_factor
    !> maximum factor at end time
    real(kind=rk) :: max_factor
  end type tem_predef_transient_type

  !> contains information for loading inlet velocities from a datafile
  !! The data has to be stored as tuples (time,velocity) columnwise.
  !!         t1 v1
  !!         t2 v2
  !!         ...
  !!         tn vn
  !! and v1 .eq. vn has to be fullfilled.
  !! The data has to be provided in the format '(e15.8)'
  type tem_from_file_transient_type
    !> filename of the data
    character(len=PathLen) :: datafile
    !> interpolate linearly between the data
    character(len=LabelLen) :: intp
    !> growing array of tuples (time, velocity)
    type( grw_real2darray_type ) :: signal
    !> ramping active?
    logical :: ramp = .false.
    !> ramping value at the end of rampT
    real(kind=rk) :: rampVal
    !> ramping time
    real(kind=rk) :: rampT 
    !> factor to multiply data with
    real(kind=rk) :: fac
    !> is the data periodic?
    logical :: periodic
  end type tem_from_file_transient_type

  !> defines different transient types like const, lua func or predefined func
  type tem_transient_type
    !> transient kind
    character(len=LabelLen) :: kind
    !> constant value
    real(kind=rk) :: const
    !> stores path to lua function
    type( aot_path_type ) :: LuaPath
    !> contains information for predefined functions
    type( tem_predef_transient_type )   :: predef_fun
    !> contains information for reading the data from file
    type( tem_from_file_transient_type) :: from_file
  end type tem_transient_type

  contains

!******************************************************************************!
  !> This subroutine load transient table defined for a boundary.\n
  !!
  !! If transient is defined as lua function then set kind = transient_lua
  !! else if transient block is defined then load transient table for predefined
  !! Fortran function variables and set kind = transient_\a function_name
  !! else transient is a constant value and set kind = transient_const.\n
  !! \n
  !! Valid definitions:
  !! \li Constant
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! transient = 1.0
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! \li lua_function
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! transient = 'linear'
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! Example: \a linear lua function
  !! \verbatim
  !! function linear(iTime)
  !!  local to_time = 1000
  !!  if iTime < to_time then
  !!    return iTime/to_time
  !!  else
  !!    return 1.0
  !!  end
  !!end
  !! \endverbatim
  !! \li Predefined Fortran function
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! transient = {predefined="Fortranfun_name",
  !!              min_factor = 0.0, max_factor = 1.0,
  !!              from_time = 0.0, to_time = 1000.0}
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! \li Data from a file (periodic data supported)
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! transient = {predefined="datafile",
  !!              filename='data.dat',      -- path/name of the datafile
  !!              intp='linear',            -- interpolation between the time tics ('linear','none')
  !!              periodic= true}           -- is the data periodic?
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! \n
  subroutine load_transient( me, conf, parent, path, key )
    !---------------------------------------------------------------------------
    !> boundary transient type
    type(tem_transient_type), intent(out) :: me
    !> lua state
    type(flu_State) :: conf
    !> parent handle contains transient table
    integer, intent(in) :: parent
    !> lua path to transient lua functin
    type(aot_path_type), intent(in) :: path
    !> state variable key string defined in lua
    character(len=*), intent(in), optional :: key
    !---------------------------------------------------------------------------
    type(aot_fun_type) :: fun
    integer :: thandle
    integer :: iError
    character(len=labelLen) :: local_key
    !---------------------------------------------------------------------------

    if(present(key)) then
      local_key = trim(key)
    else
      local_key = 'transient'
    endif

    ! First test for a lua function
    call aot_fun_open( L      = conf,                                          &
      &                parent = parent,                                        &
      &                fun    = fun,                                           &
      &                key    = trim( local_key ))

    if (fun%handle /= 0) then
      ! This transient modifier is defined as a Lua function
      me%kind = 'lua_fun'
      me%LuaPath = path
      write(logUnit(1),*)'    Defined lua transient function'
      call aot_path_addNode( me       = me%LuaPath,                            &
        &                    NodeType = 'function',                            &
        &                    key      = trim( local_key ))
    else
      ! It is not defined as a function, try to interpret it as a table
      call aot_table_open( L       = conf,                                     &
        &                  thandle = thandle,                                  &
        &                  parent  = parent,                                   &
        &                  key     = trim( local_key ))
      if (thandle /= 0) then
        call aot_get_val( L       = conf,                                      &
          &               thandle = thandle,                                   &
          &               key     = 'predefined',                              &
          &               val     = me%kind,                                   &
          &               default = 'unknown',                                 &
          &               ErrCode = iError)

        write(logUnit(1),*)'    Defined predefined transient function '//      &
          &            '"'//trim(me%kind)//'"'
        select case(trim(me%kind))
        case('linear', 'smooth')
          ! Load the standard parameters necessary to describe the
          ! transient behavior of the BC.
          call load_transient_predef_fun( me      = me%predef_fun,             &
            &                             conf    = conf,                      &
            &                             thandle = thandle )
        case('datafile')
          ! Load the filename for the datafile
          call load_transient_from_file( me      = me%from_file,               &
            &                            conf    = conf,                       &
            &                            thandle = thandle )
        case default
          write(logUnit(1),*)'ERROR in definition of the transient '//         &
            &            ' boundary Conditions:'
          write(logUnit(1),*)'Selected an unknown transient boundary'
          write(logUnit(1),*)trim(me%kind)
          call tem_abort()
        end select
      else
        ! As the entry for the variable is neither a function nor a table, try
        ! to interpret it as a constant variable
        call aot_get_val( L       = conf,                                      &
          &               thandle = parent,                                    &
          &               key     = trim(local_key),                           &
          &               val     = me%const,                                  &
          &               ErrCode = iError )
        if (btest(iError, aoterr_WrongType)) then
          write(logUnit(1),*)'FATAL Error occured in definition of the '//     &
            &                'transient boundary conditions'
          write(logUnit(1),*)'while retrieving transient constant:'
          write(logUnit(1),*)'Variable has wrong type (should be a '//         &
            &                'real number)!'
          write(logUnit(1),*)'STOPPING'
          call tem_abort()
        end if
        if (btest(iError, aoterr_NonExistent)) then
          ! "transient" variable not specified
          me%kind = 'none'
          me%const = 1.0_rk
        else
          me%kind = 'const'
        end if
      end if
    end if

  end subroutine load_transient
!******************************************************************************!


!******************************************************************************!
  !> This subroutine load standard transient function variables from LUA file.
  !!
  !! Default values: 
  !! \li min_factor = 0.0
  !! \li max_factor = 1.0
  !! \li from_time = 0
  !! \li to_time = tmax/2
  !! 
  !! Valid definition:
  !! \li linear function
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! transient = {predefined="linear", min_factor = 0.0, max_factor = 1.0,
  !! from_time = 0, to_time = 1000}
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! Example: Transient inlet velocity which starts from 0 to 1000 with maximum
  !! value 0.08 is shown below for linear and smooth with definition and image.
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! boundary_condition =
  !!         {
  !!           { label = 'inlet',
  !!             kind = 'inlet_ubb',
  !!             velocityX = {ref_value = 0.08
  !!                ,transient = {predefined="linear", min_factor = 0.0,
  !!                      max_factor = 1.0, from_time = 0, to_time = 1000}}
  !!--                ,transient = {predefined="smooth", min_factor = 0.0,
  !!--                      max_factor = 1.0, from_time = 0, to_time = 1000}}
  !!             velocityY = 0.0
  !!             velocityZ = 0.0
  !!           }
  !!         }
  !! ~~~~~~~~~~~~~~~~~~~~~
  !! \image html transient.png
  !! 
  subroutine load_transient_predef_fun( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> transient predefined fun type
    type(tem_predef_transient_type),intent(inout) :: me
    !> lua state type
    type(flu_State) :: conf
    !> aotus parent handle
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iError
    !---------------------------------------------------------------------------

    !min_factor
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'min_factor',                                  &
      &               val     = me%min_factor,                                 &
      &               ErrCode = iError,                                        &
      &               default = 0.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving min_factor :'
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    !max_factor
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'max_factor',                                  &
      &               val     = me%max_factor,                                 &
      &               ErrCode = iError,                                        &
      &               default = 1.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving max_factor :'
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    !from_time
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'from_time',                                   &
      &               val     = me%from_time,                                  &
      &               ErrCode = iError,                                        &
      &               default = 0.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving from_time :'
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    ! to_time
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'to_time',                                     &
      &               val     = me%to_time,                                    &
      &               ErrCode = iError,                                        &
      &               default = 1.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving to_time :'
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

  end subroutine load_transient_predef_fun
!******************************************************************************!


!******************************************************************************!
  !> This subroutine loads the information needed to read data from a file. 
  !!
  !! Example: Read time periodic velocity signals from file 'data.dat' as inlet
  !!          condition without interpolation between the velocity signals.
  !! 
  !! ~~~~~~~~~~~~~~~~~~~~~{.lua}
  !! boundary_condition =
  !!         {
  !!           { label = 'inlet',
  !!             kind = 'inlet_ubb',
  !!             velocityX = { kind = 'combined',
  !!                           transient = { predefined = 'datafile', 
  !!                                         filename   = 'data.dat',
  !!                                         intp       = 'none',
  !!                                         periodic   = true
  !!                                       }
  !!                         },
  !!             velocityY = 0.0
  !!             velocityZ = 0.0
  !!           }
  !!         }
  !! ~~~~~~~~~~~~~~~~~~~~~
  !!
  !! The options for interpolating the values are
  !! \li \a 'none': No interpolation is used. The signals will be chosen by:
  !!                [t1,t2[ = v1, [t2,t3[ = v2, ...
  !! \li \a 'linear': A linear interpolation is done between neighboring 
  !!                  values according to the following equation:
  !!                  \f[
  !!                         \frac{v_{i+1}-v_{i}}{t_{i+1}-t_{i}} \cdot
  !!                              (t_{sim} \% t_{period} - t_{i}) + v_{i}
  !!                  \f]
  !! 
  subroutine load_transient_from_file( me, conf, thandle )
    !---------------------------------------------------------------------------
    !> transient predefined from_file type
    type(tem_from_file_transient_type),intent(inout) :: me
    !> lua state type
    type(flu_State) :: conf
    !> aotus parent handle
    integer, intent(in) :: thandle
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iError
    ! aotus handle for ramping table
    integer :: rampHandle
    !---------------------------------------------------------------------------

    ! read the filename of the datafile
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'filename',                                    &
      &               val     = me%datafile,                                   &
      &               ErrCode = iError )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving datafile:'
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    ! read the interpolation kind for the data
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'intp',                                        &
      &               val     = me%intp,                                       &
      &               ErrCode = iError,                                        &
      &               default = 'none' )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving interpolation:'
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    ! open the ramping table
    call aot_table_open( L       = conf,                                       &
      &                  thandle = rampHandle,                                 &
      &                  parent  = thandle,                                    &
      &                  key     = 'ramping')
    if (rampHandle /= 0) then
      me%ramp = .true.
      ! read the ramping time
      call aot_get_val( L       = conf,                                        &
        &               thandle = rampHandle,                                  &
        &               key     = 'rampVal',                                   &
        &               val     = me%rampVal,                                  &
        &               ErrCode = iError)
      if (btest(iError, aoterr_Fatal)) then
        write(logUnit(1),*)'FATAL Error occured, while retrieving ramping '//  &
          &                'value:'
        if (btest(iError, aoterr_NonExistent))                                 &
          & write(logUnit(1),*)'Variable not existent!'
        if (btest(iError, aoterr_WrongType))                                   &
          & write(logUnit(1),*)'Variable has wrong type!'
        write(logUnit(1),*)'STOPPING'
        call tem_abort()
      end if

      ! read the ramping time
      call aot_get_val( L       = conf,                                        &
        &               thandle = rampHandle,                                  &
        &               key     = 'rampT',                                     &
        &               val     = me%rampT,                                    &
        &               ErrCode = iError)
      if (btest(iError, aoterr_Fatal)) then
        write(logUnit(1),*)'FATAL Error occured, while retrieving ramping time:'
        if (btest(iError, aoterr_NonExistent))                                 &
          & write(logUnit(1),*)'Variable not existent!'
        if (btest(iError, aoterr_WrongType))                                   &
          & write(logUnit(1),*)'Variable has wrong type!'
        write(logUnit(1),*)'STOPPING'
        call tem_abort()
      end if
    end if

    ! read the factor
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'fac',                                         &
      &               val     = me%fac,                                        &
      &               ErrCode = iError,                                        &
      &               default = 1.0_rk )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving factor: '
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    ! read in wether the data should be treated periodic
    call aot_get_val( L       = conf,                                          &
      &               thandle = thandle,                                       &
      &               key     = 'periodic',                                    &
      &               val     = me%periodic,                                   &
      &               ErrCode = iError,                                        &
      &               default = .false. )
    if (btest(iError, aoterr_Fatal)) then
      write(logUnit(1),*)'FATAL Error occured, while retrieving periodic:'
      if (btest(iError, aoterr_NonExistent))                                   &
        & write(logUnit(1),*)'Variable not existent!'
      if (btest(iError, aoterr_WrongType))                                     &
        & write(logUnit(1),*)'Variable has wrong type!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    write(logUnit(3),*)"Using data from file: "//trim(me%datafile)
    write(logUnit(3),*)"  interpolation:      "//trim(me%intp)
    if (rampHandle /= 0) then
      write(logUnit(3),*)"  ramping time:       ",me%rampT
      write(logUnit(3),*)"  ramping value:      ",me%rampVal
    end if
    write(logUnit(3),*)"  periodic:           ",me%periodic

    ! load the data from file
    call load_datafile( me )

  end subroutine load_transient_from_file
!******************************************************************************!


!******************************************************************************!
  !> This subroutine reads the data from disc and stores it in the 
  !! tem_from_file_transient_type.
  !!
  subroutine load_datafile( me )
    !---------------------------------------------------------------------------
    !> transient predefined from_file type
    type(tem_from_file_transient_type),intent(inout) :: me
    !---------------------------------------------------------------------------
    ! unit of the datafile
    integer :: datafile_unit
    ! status of IO
    integer :: stat = 0
    ! tmp time value
    real(kind=rk) :: tmp_time
    ! tmp vel value
    real(kind=rk) :: tmp_vel
    !---------------------------------------------------------------------------
    datafile_unit = newunit()

    ! initialize the data array
    call init( me = me%signal, width = 2 )

    ! open the datafile
    open( unit = datafile_unit,                                                &
      &   file = trim(me%datafile),                                            &
      &   action = 'read',                                                     &
      &   status = 'old',                                                      &
      &   form = 'formatted',                                                  &
      &   access = 'sequential' )

    ! as long as there is no io error (end of file is not reached)
    do
      ! ... read the data
      read( unit = datafile_unit, fmt='(2e15.8)', iostat=stat ) tmp_time, tmp_vel
      ! ... check wether the file has reached its end to prevent storing
      !     the last item twice
      if ( stat .ne. 0) then
        exit
      end if
      ! ... and append them to the growing array of tuples
      call append( me = me%signal, val = (/ tmp_time, tmp_vel/) )
    end do

    ! multiply the data with the provided factor
    me%signal%val(2,:)=me%fac*me%signal%val(2,:)

    if( stat .gt. 0 ) then
      write(logUnit(1),*)'Error when reading the inlet data from file!'
      write(logUnit(1),*)'STOPPING'
      call tem_abort()
    end if

    ! close the file
    close( unit = datafile_unit )

  end subroutine load_datafile
!******************************************************************************!


!******************************************************************************!
  !> This function returns value of linear function which is defined by
  !! from_time, to_time, min_factor and max_factor
  !!
  function transient_linear_for( me, time ) result(res)
    !---------------------------------------------------------------------------
    !> transient linear type
    type(tem_predef_transient_type) :: me
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value of a function
    real(kind=rk) :: res
    !---------------------------------------------------------------------------

    if (time%sim <= me%from_time) then
      res = me%min_factor
    else if (time%sim >= me%to_time) then
      res = me%max_factor
    else
      res = me%min_factor + (me%max_factor - me%min_factor)                    &
        & * (  (time%sim - me%from_time) / (me%to_time - me%from_time) )
    end if
  end function transient_linear_for
!******************************************************************************!


!******************************************************************************!
  !> This function returns value of smooth sin function which is defined by
  !! from_time, to_time, min_factor and max_factor
  !!
  function transient_smooth_for( me, time ) result(res)
    !---------------------------------------------------------------------------
    !> transient smooth type
    type(tem_predef_transient_type) :: me
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> return value of a function
    real(kind=rk) :: res
    !---------------------------------------------------------------------------

    if (time%sim <= me%from_time) then
      res = me%min_factor
    else if (time%sim >= me%to_time) then
      res = me%max_factor
    else
      res = me%min_factor                                                      &
        & + (me%max_factor - me%min_factor)                                    &
        & * sin( PI * 0.5_rk * (time%sim - me%from_time)                       &
        &                    / (me%to_time - me%from_time))
    end if
  end function transient_smooth_for
!******************************************************************************!


!******************************************************************************!
  !> This function evaluate lua function and return its result
  !!
  function transient_lua_for(me, time, conf) result(res)
    !---------------------------------------------------------------------------
    !> lua path to transient lua function
    type(aot_path_type) :: me
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> optional lua state
    type(flu_State), optional :: conf
    !> return value
    real(kind=rk) :: res
    !---------------------------------------------------------------------------
    ! local variables
    type(aot_fun_type) :: fun
    integer :: iError
    type(flu_State) :: loc_conf
    logical :: openLua
    !---------------------------------------------------------------------------

    if (present(conf)) then
      loc_conf = conf
      openLua = .false.
    else
      openLua = .true.
    end if

    call aot_path_open( me, conf=loc_conf, fun=fun, openLua=openLua )

    call aot_fun_put( L=loc_conf, fun=fun, arg=time%sim )
    call aot_fun_do( L=loc_conf, fun=fun, nresults=1 )
    call aot_top_get_val( L=loc_conf, val=res, ErrCode=iError )

    call aot_path_close( me, conf=loc_conf, fun=fun, closeLua=openLua )

  end function transient_lua_for
!******************************************************************************!


!******************************************************************************!
  !> This function searches for the right values in the periodic data read
  !! from file and interpolates them if needed. 
  !!
  !! Two different kinds of interpolation between the data tuples are available.
  !! \li \a 'none': No interpolation between the data points is performed.
  !!                The data is evaluated as ]t1,t2] -> v1, ]t2,t3] -> v2, ...
  !! \li \a 'linear': The value is interpolated in a linear fashion from the 
  !!                  provided data.
  !!
  function transient_from_file_periodic_for( me, time, dt_phy) result(res)
    !---------------------------------------------------------------------------
    !> datatype incl. the data read from file
    type(tem_from_file_transient_type) :: me
    !> timer object incl. the current time information
    type(tem_time_type), intent(in)  :: time
    !> temporal discretization
    real(kind=rk), intent(in) :: dt_phy
    !> return value
    real(kind=rk) :: res
    !---------------------------------------------------------------------------
    ! local variables
    integer :: iData
    ! period in physical units
    real(kind=rk) :: rPeriod
    real(kind=rk) :: ratio
    ! period in lattice units
    integer :: iPeriod
    ! remainder in lattice units
    integer :: iRem
    ! remainder in physical units
    real(kind=rk) :: rRem
    ! factor to ramp for the first period
    real(kind=rk) :: fac
    !---------------------------------------------------------------------------

    ! calculate the period in physical units (last time signal 
    ! - first time signal)
    rPeriod = me%signal%val(1,me%signal%nVals) - me%signal%val(1,1)
    ratio = time%sim / rPeriod
    rRem =  (ratio - floor(ratio))*rPeriod

    ! search for the correct entry in the data read from file
    do iData = 1, me%signal%nVals-1
      if ( ( (rRem+eps) >= me%signal%val(1, iData) ) .and.                   &
        &  ( (rRem) <= me%signal%val(1, iData+1) ) )then
        exit
      end if
    end do

    if( .not. me%ramp .or. time%sim >= me%rampT)then
      select case( trim(adjustl(me%intp)) )
        case( 'none' )
          ! result is the data stored in the position directly
          res = me%signal%val(2, iData)
        case( 'linear' )
          ! result is the linear interpolation between the data from
          ! iData and iData+1
          res = ( me%signal%val(2, iData+1) - me%signal%val(2, iData) )        &
            & / ( me%signal%val(1, iData+1) - me%signal%val(1, iData) )        &
            & * (rRem - me%signal%val(1, iData) )                              &
            & + me%signal%val(2, iData)
      end select
    else
      res = me%rampVal * sin( PI * 0.5_rk * (time%sim/me%rampT))
    end if

  end function transient_from_file_periodic_for
!******************************************************************************!


!******************************************************************************!
  !> This function invokes the type of the boundary such as constant, lua or
  !! predefined Fortran function.
  !!
  !! If transient block is not defined than it returns value = 1.0.
  function tem_transient_for( transient, time, conf, dt_phy ) result( res )
    !---------------------------------------------------------------------------
    !> boundary state
    type( tem_transient_type ) :: transient
    !> timer object incl. the current time information
    type(tem_time_type), intent( in )  :: time
    !> lua state
    type(flu_State), intent(in), optional :: conf
    !> temporal discretization step in physical units
    real(kind=rk), intent(in), optional :: dt_phy
    !> return value of a function
    real(kind=rk) :: res
    !---------------------------------------------------------------------------

    select case( trim(adjustl(transient%kind)) )
    case( 'none' )
      res = transient%const
    case( 'const' )
      res = transient%const
    case( 'lua_fun' )
      res = transient_lua_for( me   = transient%LuaPath,                       &
        &                      time = time,                                    &
        &                      conf = conf )
    case( 'linear' )
      res = transient_linear_for( me   = transient%predef_fun,                 &
        &                         time = time )
    case( 'smooth' )
      res = transient_smooth_for( me   = transient%predef_fun,                 &
        &                         time = time )
    case( 'datafile' )
      if ( transient%from_file%periodic )then
        res = transient_from_file_periodic_for( me     = transient%from_file,  &
          &                                     time   = time,                 &
          &                                     dt_phy = dt_phy )
      else
        write(logUnit(1),*)'Only Periodic datasets are currently implemented!'
        write(logUnit(1),*)'STOPPING'
        call tem_abort()
      end if
    end select

   end function tem_transient_for
!******************************************************************************!


end module tem_transient_module
!******************************************************************************!
