! This utest program checks if the bgk d3Q19 optimized(Op) and explicit compute
! kernel output the same results. Input pdf is initilized by equilibrium with
! randomly generated density and velocity values.
program mus_bgk_d3q19_compare_test
  use env_module, only: rk, eps, labelLen, init_random_seed
  use treelmesh_module, only: treelmesh_type
  use tem_tools_module, only: tem_horizontalSpacer
  use tem_logging_module, only: logUnit
  use tem_general_module, only: tem_start, tem_finalize
  use tem_var_system_module, only: tem_var_system_type, tem_var_append,  &
    &                               tem_var_sys_init, tem_var_sys_mark

  use mus_param_module,  only: mus_param_type
  use mus_pdf_module,    only: pdf_globInfo_type
  use mus_scheme_type_module, only: mus_scheme_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type, &
    &                                 mus_init_stencil, mus_define_d3q19
  use mus_field_prop_module, only: mus_field_prop_type
  use mus_d3q19_module, only: bgk_advRel_d3q19
  use mus_bgk_module,   only: bgk_advRel_explicit
  use mus_derivedQuantities_module2, only: getEquilibrium

  implicit none

  integer,  parameter :: QQ  = 19   !> number of pdf directions

  logical :: error
  real(kind=rk) :: tolerance, max_error
  integer :: iDir

  type( mus_field_prop_type ), allocatable :: fieldProp(:)
  real(kind=rk) :: omega, rho0, velocity(3), density
  real(kind=rk) :: instate(QQ), outEx(QQ), outOp(QQ), diff(QQ)
  integer :: neigh(QQ)
  integer :: nStart, nEnd
  type( mus_scheme_layout_type ) :: layout
  type( tem_var_system_type ) :: varSys
  character(len=labelLen) :: sysName
  integer :: clock

  ! dummy variables
  integer :: level
  type( pdf_globInfo_type ) :: globInfo
  type( mus_param_type ) :: params
  type( mus_scheme_type ), allocatable :: scheme(:)
  type( treelmesh_type ) :: tree

  call tem_start('BGK D3Q19 kernels comparison utest', 'utest', params%general)
  error = .true.
  tolerance = eps * 2500._rk
  write(*,*) 'tolerance = ', tolerance

  CALL SYSTEM_CLOCK( COUNT = clock )
  call init_random_seed( clock )
  call random_number( omega )
  omega = omega * 2.0_rk ! omega range [0,2)
  write( logUnit(1), "(' omega = ', F5.3)" ) omega

  ! generate density within range [0.9, 1.1]
  call random_number( density )
  density = density * 0.2 + 0.9_rk
  write( logUnit(1), *) ' density = ', density

  ! generate velocity within range [-0.1, 0.1]
  call random_number( velocity(1) )
  call random_number( velocity(2) )
  call random_number( velocity(3) )
  velocity = velocity * 0.2_rk - 0.1_rk
  write( logUnit(1), *) ' velocity = ', velocity

  ! fill fieldProp with omega and rho0
  allocate( fieldProp(1) )
  rho0  = 1.0_rk
  call init_fieldProp( fieldProp(1), omega, rho0 )

  neigh = [ (iDir, iDir = 1, QQ) ]

  nStart = 1
  nEnd   = 1

  ! init layout
  call mus_init_stencil( stencil = layout%stencil, nStencils = 1 )
  call mus_define_d3q19( layout = layout, nElems = 1 )

  ! init variable system
  sysName = 'var_system'
  call init_varSys( varSys, sysName )

  level = 1
  allocate( scheme(1) )

  call tem_horizontalSpacer(fUnit = logUnit(1))
  write( logUnit(1), *) 'Input:  inState initilized by Eq from random den and vel'
  write( logUnit(1), *) 'Expect: outState from optimized kernel and explicity'
  write( logUnit(1), *) '        kernel be the same as inState'
  ! initialize pdf by calculating equilibirum
  inState =  getEquilibrium( density, velocity, layout )

  ! call optimized compute kernel
  outOP = -1.0_rk
  write( logUnit(1), *) 'Calling Optimized compute kernel routine.'
  call bgk_advRel_d3q19( iScheme   = 1,        &
    &                    fieldProp = fieldProp,&
    &                    inState   = inState,  &
    &                    outState  = outOP,    &
    &                    neigh     = neigh,    &
    &                    nStart    = nStart,   &
    &                    nEnd      = nEnd,     &
    &                    level     = level,    &
    &                    layout    = layout,   &
    &                    pdfGlobal = globInfo, &
    &                    params    = params,   &
    &                    scheme    = scheme,   &
    &                    tree      = tree,     &
    &                    varSys    = varSys    )

  ! call explicit compute kernel
  outEx = -1.0_rk
  write( logUnit(1), *) 'Calling Explicit compute kernel routine.'
  call bgk_advRel_explicit( iScheme   = 1,     &
    &                    fieldProp = fieldProp,&
    &                    inState   = inState,  &
    &                    outState  = outEx,    &
    &                    neigh     = neigh,    &
    &                    nStart    = nStart,   &
    &                    nEnd      = nEnd,     &
    &                    level     = level,    &
    &                    layout    = layout,   &
    &                    pdfGlobal = globInfo, &
    &                    params    = params,   &
    &                    scheme    = scheme,   &
    &                    tree      = tree,     &
    &                    varSys    = varSys    )

  write( logUnit(1), *) 'Calculating errors between results from two kernels.'
  diff = abs(outEx - outOp)
  max_error = maxval( diff(:) )
  if ( max_error > tolerance ) then
    write(logUnit(1), *) "Max error exceeds tolerance!"
    write(logUnit(1), *) "  inState                 outState               diff"
    do iDir = 1, QQ
      if ( diff(idir) > tolerance ) &
        & write(logUnit(1),*) iDir, inState(iDir), outEx(iDir), outOp(iDir), &
        &                     diff(iDir)
    end do
  else
    error = .false.
    write(logUnit(1), *) "Max error within tolerance!"
    write(logUnit(1), *) "  inState        outEx              outOp         diff"
    do iDir = 1, QQ
      write(logUnit(1),*) iDir, inState(iDir), outEx(iDir), outOp(iDir), &
                          diff(iDir)
    end do
  end if

  deallocate( fieldProp )
  deallocate( scheme    )
  call tem_finalize(params%general)

  if (.not. error) then
    write(*,*) 'PASSED'
  else
    write(*,*) 'FAILED'
  end if

  !*****************************************************************************

contains

  ! fill fieldProp with values of omega and rho0
  subroutine init_fieldProp( fieldProp, omega, rho0 )
    type( mus_field_prop_type ) :: fieldProp
    real(kind=rk) :: omega
    real(kind=rk) :: rho0

    fieldProp%fluid%omega    = omega
    fieldProp%fluid%omLvl(:) = omega
    fieldProp%fluid%rho0     = rho0
  end subroutine init_fieldProp

  ! append pdf to variable system
  subroutine init_varSys( varSys, sysName )
    type( tem_var_system_type ) :: varSys
    character(len=labelLen) :: sysName

    integer :: pdfPos
    logical :: wasAdded

    call tem_var_sys_init( varSys )
    call tem_var_append( sys = varSys, &
      &                  varName = 'pdf', &
      &                  nComponents = QQ, &
      &                  pos = pdfPos,  &
      &                  wasAdded = wasAdded )
    call tem_var_sys_mark( varSys, sysName )

  end subroutine init_varsys

end program mus_bgk_d3q19_compare_test
!******************************************************************************
