! This utest program compare the outputs of nonNewtonian CY optimized kernel and
! explicit kernel.
! input pdf is initialized by equilibrium function, where density and velocity
! is randomly generated.
program mus_nNwtn_CY_compare_test
  use env_module, only: rk, eps, labelLen, init_random_seed, newUnit
  use treelmesh_module, only: treelmesh_type
  use tem_tools_module, only: tem_horizontalSpacer, tem_file_to_string
  use tem_logging_module, only: logUnit
  use tem_general_module, only: tem_start, tem_finalize
  use tem_var_system_module, only: tem_var_system_type, tem_var_append, &
    &                               tem_var_sys_init, tem_var_sys_mark

  use aotus_module, only: open_config_chunk
  use aot_out_module,   only: aot_out_type, aot_out_open,    &
    &                         aot_out_close

  use mus_param_module,         only: mus_param_type
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type, &
    &                                 mus_init_stencil, mus_define_d3q19, &
    &                                 mus_set_weights_d3q19, mus_finish_stencil
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_d3q19_module,         only: bgk_advRel_d3q19_nNwtn_CY
  use mus_bgk_module,           only: bgk_nNwtn_CY_explicit
  use mus_physics_module,       only: mus_physics_type, &
    &                                 mus_set_convFac, mus_physics_out, &
    &                                 set_values_by_levels, &
    &                                 mus_physics_dump2outUnit
  use mus_nonNewtonian_module,  only: mus_nNwtn_CY
  use mus_derivedQuantities_module2, only: getEquilibrium

  implicit none

  integer, parameter :: QQ = 19   !> number of pdf directions
  integer, parameter :: nScalars = QQ + 1
  real(kind=rk), parameter :: rho0    = 1.0_rk
  real(kind=rk), parameter :: densPhy = 1050._rk
  real(kind=rk), parameter :: nuPhy   = 0.0035_rk / densPhy
  character(len=labelLen), parameter :: scaling = 'diffusive'
  integer, parameter :: scaleFactor = 4

  logical :: error = .true.
  real(kind=rk) :: tolerance, max_error
  integer :: iDir

  type( mus_scheme_type ), allocatable :: scheme(:)
  real(kind=rk) :: omega
  real(kind=rk) :: instate(nScalars), diff(nScalars)
  real(kind=rk) :: outOP(nScalars), outEx(nScalars)
  real(kind=rk) :: density, velocity(3)
  integer :: neigh(nScalars)
  integer :: nStart = 1
  integer :: nEnd   = 1
  integer :: level  = 1
  type( mus_field_prop_type ), allocatable :: fieldProp(:)
  type( mus_scheme_layout_type ) :: layout
  type( tem_var_system_type ) :: varSys
  character(len=labelLen) :: sysName
  integer :: clock
  integer :: nUnit
  type( aot_out_type ) :: conf
  integer :: iError
  type( treelmesh_type ) :: tree

  ! dummy variables
  type( pdf_globInfo_type ) :: globInfo
  type( mus_param_type ) :: param

  ! start program, set up error tolerance
  call tem_start('BGK D3Q19 nonNewtonian CY kernel compare utest', &
    &            'utest', param%general)
  tolerance = eps * 10000._rk
  write(logUnit(1), "('tolerance = ', ES11.4 )" ) tolerance

  ! generate random omega
  CALL SYSTEM_CLOCK( COUNT = clock )
  call init_random_seed( clock )
  call random_number( omega )
  omega = omega * 2.0_rk ! omega range [0,2)
  write( logUnit(1), "('omega = ', F12.7)") omega

  ! generate density within range [0.9, 1.1]
  call random_number( density )
  density = density * 0.2 + 0.9_rk
  write( logUnit(1), "('density = ', F5.3)" ) density

  ! generate velocity within range [-0.1, 0.1]
  call random_number( velocity(1) )
  call random_number( velocity(2) )
  call random_number( velocity(3) )
  velocity = velocity * 0.2_rk - 0.1_rk
  write( logUnit(1), "('velocity = ', 3F7.3) " ) velocity

  allocate( scheme(1) )
  allocate( scheme(1)%field(1) )

  ! fill param
  param%scaling = scaling
  param%scaleFactor = scaleFactor

  ! fill minLevel and maxLevel in tree
  tree%global%minLevel = level
  tree%global%maxLevel = level

  ! fill fieldProp
  write( logUnit(1), '(A)') 'fill fieldProp'
  allocate( fieldProp(1) )
  call init_fieldProp( fieldProp(1), omega, rho0 )

  ! fill nonNewtonian model parameters
  scheme(1)%field(1)%fieldProp%fluid%nNwtn%active = .true.
  scheme(1)%field(1)%fieldProp%fluid%nNwtn%model  = mus_nNwtn_CY

  ! fill neigh array
  write( logUnit(1), '(A)') 'fill neigh array'
  neigh = [ (iDir, iDir = 1, nScalars) ]

  ! init layout
  write( logUnit(1), '(A)') 'fill layout'
  call mus_init_stencil( stencil = layout%stencil, nStencils = 1 )
  call mus_define_d3q19( layout = layout, nElems = 1 )
  call mus_finish_stencil( layout )

  write( logUnit(1), '(A)') 'fill variable system'
  sysName = 'var_system'
  call init_varSys( varSys, sysName )

  ! set up dx, dt, physics table, write to solSpec string
  write( logUnit(1), '(A)') 'set physics table'
  call set_physics( param%physics, omega, level, level )

  ! write physics table to solSpec string
  write( logUnit(1), '(A)') 'write physics table to solve string'
  nUnit = newUnit()
  open(unit = nUnit, status='scratch')
  call aot_out_open( put_conf = conf, outUnit = nUnit )
  call mus_physics_out( me   = param%physics, &
    &                   conf = conf,           &
    &                   minLevel = level,      &
    &                   maxLevel = level       )
  call aot_out_close( put_conf = conf )
  call tem_file_to_string(nUnit, varSys%solSpec%string, iError)
  call open_config_chunk( L     = varSys%solSpec%conf,         &
      &                   chunk = trim( varSys%solSpec%String ))

  call tem_horizontalSpacer(fUnit = logUnit(1))
  write( logUnit(1), '(A)') 'Input:  inState initialized by equilibrium'
  write( logUnit(1), '(A)') 'Expect: outState from two kernels should be the same.'
  ! initialize pdf by calculating equilibirum
  inState(1:QQ) = getEquilibrium( density, velocity, layout )
  inState(nScalars) = omega

  outOP = -1.0_rk
  ! call Optimized compute kernel
  write( logUnit(1), '(A)') 'Calling compute kernel routine.'
  call bgk_advRel_d3q19_nNwtn_CY( iScheme   = 1,        &
    &                             fieldProp = fieldProp,&
    &                             inState   = inState,  &
    &                             outState  = outOP,    &
    &                             neigh     = neigh,    &
    &                             nStart    = nStart,   &
    &                             nEnd      = nEnd,     &
    &                             level     = level,    &
    &                             layout    = layout,   &
    &                             pdfGlobal = globInfo, &
    &                             params    = param,    &
    &                             scheme    = scheme,   &
    &                             tree      = tree,     &
    &                             varSys    = varSys    )

  ! call Explicit compute kernel
  outEx = -1.0_rk
  write( logUnit(1), '(A)') 'Calling compute kernel routine.'
  call bgk_nNwtn_CY_explicit( iScheme   = 1,        &
    &                         fieldProp = fieldProp,&
    &                         inState   = inState,  &
    &                         outState  = outEx,    &
    &                         neigh     = neigh,    &
    &                         nStart    = nStart,   &
    &                         nEnd      = nEnd,     &
    &                         level     = level,    &
    &                         layout    = layout,   &
    &                         pdfGlobal = globInfo, &
    &                         params    = param,    &
    &                         scheme    = scheme,   &
    &                         tree      = tree,     &
    &                         varSys    = varSys    )

  write( logUnit(1), '(A)') 'Calculating errors.'
  diff = outOp - outEx
  max_error = maxval( abs(diff(:)) )
  if ( max_error > tolerance ) then
    write(logUnit(1), *) "Max error exceeds tolerance: "
    write(logUnit(1), *) " Max error = ", max_error
    write(logUnit(1), *) " Tolerance = ", tolerance
    write(logUnit(1), '(A4,4A10)') "iDir", "inState", "outOp", "outEx", "diff"
    do iDir = 1, nScalars
      write(logUnit(1),*) iDir, inState(iDir), &
        &                             outOp(iDir), outEx(iDir), diff(iDir)
    end do
  else
    error = .false.
    write(logUnit(1), '(A)') "Max error within tolerance!"
    write(logUnit(1), *) " Max error = ", max_error
    write(logUnit(1), *) " Tolerance = ", tolerance
  end if

  deallocate( fieldProp )
  deallocate( scheme    )
  call tem_finalize(param%general)

  if (.not. error) then
    write(*,'(A)') 'PASSED'
  else
    write(*,'(A)') 'FAILED'
  end if

  !*****************************************************************************

contains

  ! fill fieldProp with values of omega and rho0
  subroutine init_fieldProp( fieldProp, omega, rho0 )
    type( mus_field_prop_type ) :: fieldProp
    real(kind=rk) :: omega
    real(kind=rk) :: rho0

    fieldProp%fluid%omega    = omega
    fieldProp%fluid%omLvl(:) = omega
    fieldProp%fluid%rho0     = rho0
  end subroutine init_fieldProp

  ! append pdf and omega to variable system
  subroutine init_varSys( varSys, sysName )
    type( tem_var_system_type ) :: varSys
    character(len=labelLen) :: sysName

    integer :: pdfPos
    logical :: wasAdded

    call tem_var_sys_init( varSys )

    call tem_var_append( sys = varSys,        &
      &                  varName = 'pdf',     &
      &                  nComponents = QQ,    &
      &                  pos = pdfPos,        &
      &                  wasAdded = wasAdded  )

    call tem_var_append( sys = varSys,      &
      &                  varName = 'omega', &
      &                  nComponents = 1,   &
      &                  pos = pdfPos,      &
      &                  wasAdded = wasAdded )

    call tem_var_sys_mark( varSys, sysName )

  end subroutine init_varsys

  subroutine set_physics( physics, omega, minLevel, maxLevel )
    type( mus_physics_type ) :: physics
    real(kind=rk), intent(in) :: omega
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel

    real(kind=rk) :: dx, dt, nuLB

    dx = 1.0_rk
    nuLB = ( 1.0 / omega - 0.5 ) / 3.0
    dt = nuLB * dx * dx / nuPhy

    physics%dt = dt
    physics%dx = dx
    physics%rho0 = densPhy

    ! Assign dx and dt for each level
    allocate(physics%dxLvl( minLevel:maxLevel ))
    allocate(physics%dtLvl( minLevel:maxLevel ))
    allocate(physics%fac(   minLevel:maxLevel ))

    physics%dxLvl( minLevel:maxLevel ) = &
      &       set_values_by_levels( physics%dx, minLevel, maxLevel, 2 )
    physics%dtLvl( minLevel:maxLevel ) = &
      &       set_values_by_levels( physics%dt, minLevel, maxLevel, scaleFactor)

    call mus_set_convFac( physics, minLevel, maxLevel )
    call mus_physics_dump2outUnit( physics, logUnit(1), minLevel, maxLevel )

  end subroutine set_physics

end program mus_nNwtn_CY_compare_test
!******************************************************************************
