?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> Boundary condition treatment routines for multispecies simulation
!!
module mus_bc_species_module

  ! include treelm modules
  use env_module,               only: rk, labelLen
  use tem_param_module,         only: div1_3, div1_36, div1_8, div3_4h,        &
    &                                 div1_4, div3_8, div3_2, div9_16, div3_16,&
    &                                 cs2inv, div1_6, cs4inv,                  &
    &                                 t2cs2inv, t2cs4inv, div1_18, cs2
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_var_system_module,    only: tem_var_system_type
  use tem_subTree_type_module,  only: tem_subTree_type
  use tem_math_module,          only: invert_matrix
  use tem_property_module,      only: prp_missingNeigh
  use tem_ini_condition_module, only: tem_ini_condition_type
  use tem_spatial_module,       only: tem_spatial_for
  use tem_spacetime_fun_module, only: tem_spacetime_for
  use tem_isNaN_module,         only: tem_isNaN
  use tem_debug_module,         only: dbgUnit

  ! include musubi modules
  use mus_bc_header_module,      only: boundary_type, glob_boundary_type
  use mus_scheme_layout_module,  only: mus_scheme_layout_type 
  use mus_field_prop_module,     only: mus_field_prop_type
  use mus_pdf_module,            only: pdf_global_type
  use mus_variable_module,       only: mus_derVarPos_type
  use mus_param_module,          only: mus_param_type
  use mus_physics_module,        only: mus_physics_type
  use mus_mixture_module,        only: mus_mixture_type
  use mus_solSpecHelpers_module, only: getFieldVariable, getFieldPrefixes
  use mus_eNRTL_module,          only: mus_calc_thermFactor,                   &
    &                                  mus_calc_MS_DiffMatrix


  implicit none

  private

  public :: spc_inlet_eq, spc_vel_bb

  public :: spc_outlet_zero_prsgrd
  public :: spc_outlet_eq, spc_outlet_vel, spc_outlet_expol
  public :: spc_moleFrac, spc_moleDiff_Flux
  public :: spc_moleFlux, spc_moleFlux_eq
  public :: spc_blackbox_mem_ion, spc_blackbox_mem_solvent
  public :: spc_moleFrac_wtdf
  ! moments BC
  public :: spc_moments_moleFrac
  public :: spc_moments_moleFlux
  public :: spc_moments_wall
  public :: spc_moments_vel
  public :: spc_moments_blackbox_mem_ion

contains
!******************************************************************************!
  !> Outlet boundary conditions with zero pressure gradient.
  !! \author Kannan Masilamani
  !! 
  !! These boundary conditions use the neighbor cells density and velocity in
  !! the equilibrium function. It is not necessary to specify density at 
  !! boundary in the lua configuration file 
  !! \f$ f = f^{eq}(\rho_{b-1},u_{b-1}) \f$
  !!
  ! subroutine spc_outlet_zero_prsgrd
?? copy :: bc_routineHeader( spc_outlet_zero_prsgrd )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rho, ux(3), usq, m_ratio
    integer :: iElem, iDir, iField_2, pos
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmpFields( nScalars )
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    !---------------------------------------------------------------------------

    m_ratio = fieldProp%species%molWeigRatio

    do iElem = 1, globBC%nElems( iLevel )
      !Calculate local density and velocity moments
      rho = 0._rk
      do iDir = 1,layout%stencil(1)%QQ
        ! Only changed from save to fetch and added currT
        fTmp( iDir ) = bcBuffer( &
          & ?IDX?( varPos(iDir), globBC%elems(iLevel)%elemBuffer(iElem), nScalars ) )
        rho = rho + fTmp(iDir)
      enddo
      do iField_2 = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iField_2)%varPos(iDir)
          fTmpFields(pos) = bcBuffer( &
          & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer(iElem), nScalars ) )
        end do
      end do  
      ux  = 0._rk
      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      call varSys%derive( derVarPos%velocity )%do(                             &
        &                  input   = fTmpFields,                               &
        &                  bound   = bound,                                    &
        &                  tree    = tree,                                     &
        &                  subTree = subTree,                                  &
        &                  requVar = varSys%Variable%val( derVarPos%velocity ),&
        &                  globSys = varSys,                                   &
        &                  stencil = layout%stencil(1),                        &
        &                  conf    = varSys%solSpec%conf,                      &
        &                  res     = ux )

      ! get the actual velocity out of the moments
      ux = ux
      usq = ux(1)*ux(1) + ux(2)*ux(2) + ux(3)*ux(3) 

      
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Only changed from save to fetch and added currT
          state(                                                               &
& ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)  = &
            &       layout%weight( iDir )*rho*(2.0_rk*m_ratio+9._rk*( (        &
            &       layout%stencil(1)%cxDir( 1, layout%stencil(1)%             &
            &                                          cxDirInv( iDir ))*ux(1) &
            &    +  layout%stencil(1)%cxDir( 2, layout%stencil(1)%             &
            &                                          cxDirInv( iDir ))*ux(2) &
            &    +  layout%stencil(1)%cxDir( 3, layout%stencil(1)%             &
            &                                          cxDirInv( iDir ))*ux(3) &
            &    )**2 -  div1_3*usq))                                          &
            &    - state(                                                      &
& ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ),nScalars, neigh ), currT)
        endif
      enddo
    enddo

  end subroutine spc_outlet_zero_prsgrd
!******************************************************************************!

!******************************************************************************!
  !> Inlet species velocity equilibrium boundary with specified
  !! mixture averaged mass velocity and its molefraction
  !! mixture kinematic pressure is extrapolated here.
  !! Density and velocity of all fields are used to compute equilibrium
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_inlet_eq', 
  !!    velocityX = 0.0,
  !!    velocityY = 0.0,
  !!    velocityZ = 0.0,
  !!    moleFraction = 1.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_inlet_eq )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: rho_spc
    real(kind=rk) :: mass_dens( varSys%nVars )
    real(kind=rk) :: tot_MassDens !, tot_NumDens
    real(kind=rk) :: ux(globBC%nElems(iLevel)),uy(globBC%nElems(iLevel)),&
                    &uz(globBC%nElems(iLevel))
    real(kind=rk) :: moleFrac(globBC%nElems(iLevel))                
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
    real(kind=rk) :: kinePress(globBC%nElems(iLevel))
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    integer :: offVel, posVel
!    real(kind=rk), dimension(varSys%nVars, globBC%nElems(iLevel)) :: uxB, uyB, &
!      &            uzB, rhoB, moleFracB
    !------------------------------------------------------------------------ 
    nFields = varSys%nVars

    ! compute velocity and mole fraction of all fields defined at the boundary

    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,               &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
    end do    

    ! compute molefraction to compute density of all fields and velocities
    ! of all fields and store them in input array for function pointer to
    ! compute equilibrium distribution
    offVel = globBC%nElems(iLevel)*nFields
    input_loc = 0.0_rk
    do iFieldLoc = 1, nFields
      ! Get velocity which might be defined to vary in space and time
      ux = tem_spacetime_for(                                                  &
        & me      = globBC%BC_states(iFieldLoc)%velocity(1)%spacetime,         &
        & treeIDs = glob%levelDesc( iLevel )%total(                            &
        &                                       globBC%elems(iLevel)%elem(:) ),&
        & tree    = tree,                                                      &
        & time    = sim_time,                                                  &
        & n       = globBC%nElems(iLevel),                                     &
        & conf    = params%general%solver%conf(1),                             &
        & dt_phy  = physics%dt )

      uy = tem_spacetime_for(                                                  &
        & me      = globBC%BC_states(iFieldLoc)%velocity(2)%spacetime,         &
        & treeIDs = glob%levelDesc( iLevel )%total(                            &
        &                                       globBC%elems(iLevel)%elem(:) ),&
        & tree    = tree,                                                      &
        & time    = sim_time,                                                  &
        & n       = globBC%nElems(iLevel),                                     &
        & conf    = params%general%solver%conf(1),                             &
        & dt_phy  = physics%dt )

      uz = tem_spacetime_for(                                                  &
        & me      = globBC%BC_states(iFieldLoc)%velocity(3)%spacetime,         &
        & treeIDs = glob%levelDesc( iLevel )%total(                            &
        &                                       globBC%elems(iLevel)%elem(:) ),&
        & tree    = tree,                                                      &
        & time    = sim_time,                                                  &
        & n       = globBC%nElems(iLevel),                                     &
        & conf    = params%general%solver%conf(1),                             &
        & dt_phy  = physics%dt )

      ! If physical quantities are given, transform to lattice units by division
      ! with the conversion factor
!write(*,*) 'vel ', ux(globBC%nElems(iLevel)/2), physics%fac( iLevel )%vel
      ux = ux/physics%fac( iLevel )%vel
      uy = uy/physics%fac( iLevel )%vel
      uz = uz/physics%fac( iLevel )%vel

      ! store velocity in input_loc array
      do iElem = 1, globBC%nElems(iLevel)
        posVel = offVel + (iElem-1)*nFields*3
        input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = ux(iElem)
        input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uy(iElem)
        input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uz(iElem)
      end do
    end do ! iField Loc  

    ! compute kinematic pressure 
    kinePress = 0.0_rk
    do iElem = 1, globBC%nElems(iLevel)
      ! Calculate the density of current element
      mass_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  

      !total mass density
      tot_massDens = sum(mass_dens)

      ! compute mixture kinematic pressure from extrapolate density 
      ! p = cs2*(sum(phi_k*rho_k) - min_a(m_a)*n0)/rho0
      kinePress(iElem) = cs2 * ( dot_product(phi, mass_dens)                   &
        &       - minval(molWeight)*mixture%moleDens0LB )/tot_massDens 

    enddo !iElem

    do iFieldLoc = 1, nFields
      ! mole fraction
      moleFrac = tem_spacetime_for(                                            &
        & me      = globBC%BC_states(iFieldLoc)%moleFrac%spacetime,            &
        & treeIDs = glob%levelDesc( iLevel )%total(                            &
        &                                       globBC%elems(iLevel)%elem(:) ),&
        & tree    = tree,                                                      &
        & time    = sim_time,                                                  &
        & n       = globBC%nElems(iLevel),                                     &
        & conf    = params%general%solver%conf(1),                             &
        & dt_phy  = physics%dt )

      ! compute current species mass density from specified molefraction at
      ! boundary
      ! rho = n_t * chi_i * m_i + massFrac_i * rho0 * KinePress/(cs2*phi_i)
      ! massFrac_i = rho_i/rho
      ! rho = n_t * chi_i * m_i + rho_i * KinePress/(cs2*phi_i)
      ! 1st term is zero order density
      ! 2nd term is second order density with kinematic mixture pressure
      !rho = tot_NumDens * moleFrac(iElem) * fieldProp%species%molWeight       &
      !KM: using inital mixture number density rho0/mixtureMOlWeight
      !Using local tot_NuMdens increases density over time
      do iElem = 1, globBC%nElems(iLevel)
        rho_spc = mixture%moleDens0LB * moleFrac(iElem) * molWeight(iFieldLoc)

        !rho_spc = rho_spc - cs2inv * rho_spc * kinePress(iElem) / phi(iFieldLoc)

        ! copy density defined at the inlet for this boundary  
        !mass_dens(iField) = rho
        !store in input_loc to compute equilibrium from state
        !assume all fields have same velocity at inlet
        input_loc( (iElem-1)*nFields + iFieldLoc ) = rho_spc
      enddo !iElem#
    end do !iField  

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input = input_loc,                                      &
      &                bound = bound,                                          &
      &                tree = tree,                                            &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf = varSys%solSpec%conf,                             & 
      &                res = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )

!          if (fEq( ?IDX?(iDir,iElem, layout%stencil(1)%QQ) ) < 0.0_rk) then
!            write(*,*)'BC Inlet: time ', sim_time%iter
!            write(*,*) 'found negative', iElem, iField, &
!            &  fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
!            stop
!          endif
        endif
      enddo
      endif  
    enddo
  end subroutine spc_inlet_eq
!******************************************************************************!

!******************************************************************************!
  !> Inlet species velocity bounce back boundary with specified
  !! mixture averaged mass velocity and its molefraction
  !! mixture kinematic pressure is extrapolated here.
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_vel_bb', 
  !!    velocityX = 0.0,
  !!    velocityY = 0.0,
  !!    velocityZ = 0.0,
  !!    moleFraction = 1.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_vel_bb )
    !------------------------------------------------------------------------ 
!    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
!    real(kind=rk) :: fEq
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: rho_spc
    real(kind=rk) :: mass_dens( varSys%nVars )
    real(kind=rk) :: tot_MassDens
    real(kind=rk) :: ux(globBC%nElems(iLevel)),uy(globBC%nElems(iLevel)),&
                    &uz(globBC%nElems(iLevel))
!    real(kind=rk) :: moleFrac(globBC%nElems(iLevel))                
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: kinePress
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    !------------------------------------------------------------------------ 
    nFields = varSys%nVars

    ! compute velocity and mole fraction of all fields defined at the boundary

    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,               &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
    end do    

    ! compute molefraction to compute density of all fields and velocities
    ! of all fields and store them in input array for function pointer to
    ! compute equilibrium distribution
    ! Get velocity which might be defined to vary in space and time
    ux = tem_spacetime_for(                                                    &
      & me      = me%BC_states%velocity(1)%spacetime,                          &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                       globBC%elems(iLevel)%elem(:) ),  &
      & tree    = tree,                                                        &
      & time    = sim_time,                                                    &
      & n       = globBC%nElems(iLevel),                                       &
      & conf    = params%general%solver%conf(1),                               &
      & dt_phy  = physics%dt )

    uy = tem_spacetime_for(                                                    &
      & me      = me%BC_states%velocity(2)%spacetime,                          &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                       globBC%elems(iLevel)%elem(:) ),  &
      & tree    = tree,                                                        &
      & time    = sim_time,                                                    &
      & n       = globBC%nElems(iLevel),                                       &
      & conf    = params%general%solver%conf(1),                               & 
      & dt_phy  = physics%dt )

    uz = tem_spacetime_for(                                                    &
      & me      = me%BC_states%velocity(3)%spacetime,                          &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                       globBC%elems(iLevel)%elem(:) ),  &
      & tree    = tree,                                                        &
      & time    = sim_time,                                                    &
      & n       = globBC%nElems(iLevel),                                       &
      & conf    = params%general%solver%conf(1),                               & 
      & dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel

    ! mole fraction
!!    moleFrac = tem_spacetime_for(                                              &
!!      & me      = me%BC_states%moleFrac%spacetime,                             &
!!      & treeIDs = glob%levelDesc( iLevel )%total(                              &
!!      &                                       globBC%elems(iLevel)%elem(:) ),  &
!!      & tree    = tree,                                                        &
!!      & time    = sim_time,                                                    &
!!      & n       = globBC%nElems(iLevel),                                       &
!!      & conf    = params%general%solver%conf(1),                               & 
!!      & dt_phy  = physics%dt )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      ! Calculate the density of current element
      mass_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  

!!      !!total mass density
!!      tot_massDens = sum(mass_dens)
!!
!!      !! compute mixture kinematic pressure from extrapolate density 
!!      !! p = cs2*(sum(phi_k*rho_k) - min_a(m_a)*n0)/rho0
!!      kinePress = cs2 * ( dot_product(phi, mass_dens)                          &
!!        &       - minval(molWeight)*mixture%moleDens0LB )/tot_massDens 
!!
!!      ! compute current species mass density from specified molefraction at
!!      ! boundary
!!      ! rho = n_t * chi_i * m_i + massFrac_i * rho0 * p/(cs2*phi_i)
!!      ! massFrac_i = rho_i/rho
!!      ! rho = n_t * chi_i * m_i + rho_i * p/(cs2*phi_i)
!!      ! 1st term is zero order density
!!      ! 2nd term is second order density with kinematic mixture pressure
!!      !rho = tot_NumDens * moleFrac(iElem) * fieldProp%species%molWeight       &
!!      !KM: using inital mixture number density rho0/mixtureMOlWeight
!!      !Using local tot_NuMdens increases density over time
!!
!!      rho_spc = mixture%moleDens0LB * moleFrac(iElem)                            &
!!        &     * fieldProp%species%molWeight
!!      !rho_spc = mass_dens(iField)
!!
!!      ! KM: No need to add kinematic pressure term since it is part of 
!!      ! linear term of fEq
!!!write(*,*) iElem, 'before rho_spc ', rho_spc      
!!      rho_spc = rho_spc - cs2inv * rho_spc * kinePress                         &
!!        &     / fieldProp%species%molWeigRatio
!!!write(*,*) iElem, 'after rho_spc ', rho_spc      

      do iDir = 1,layout%stencil(1)%QQ
        fTmp(iDir) = bcBuffer( & 
          &?IDX?(varpos(iDir),globBC%elems(iLevel)%elemBuffer(iElem),nScalars))
      end do

      do iDir = 1,layout%stencil(1)%QQ
        !if ( iDir == layout%stencil(1)%restPosition ) then      
        !  fEq = layout%weight( iDir )*rho_spc*((3.0_rk-2.0_rk*fieldProp%species%molWeigRatio))
        !else
        !  fEq = layout%weight( iDir )*rho_spc*( fieldProp%species%molWeigRatio &
        !      & + cs2inv                          &
        !      &   * ( layout%stencil(1)%cxDir( 1, iDir )*ux(iElem)            &
        !      &    +   layout%stencil(1)%cxDir( 2, iDir )*uy(iElem)            &
        !      &    +   layout%stencil(1)%cxDir( 3, iDir )*uz(iElem) ) )
        !   
        !end if
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
&?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
!              & = fEq
              & = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
!!              & - layout%weight( iDir )*6._rk*rho_spc                          &
!!            &       * ( layout%stencil(1)%cxDir( 1, layout%stencil(1)%         &
!!            &                                       cxDirInv( iDir ))*ux(iElem) &
!!            &       +   layout%stencil(1)%cxDir( 2, layout%stencil(1)%         &
!!            &                                       cxDirInv( iDir ))*uy(iElem) &
!!            &       +   layout%stencil(1)%cxDir( 3, layout%stencil(1)%         &
!!            &                                       cxDirInv( iDir ))*uz(iElem))

!!              & + layout%weight( iDir )*6._rk*rho_spc                          &
!!              &    * ( layout%stencil(1)%cxDir( 1, iDir )*ux(iElem)            &
!!              &    +   layout%stencil(1)%cxDir( 2, iDir )*uy(iElem)            &
!!              &    +   layout%stencil(1)%cxDir( 3, iDir )*uz(iElem))

              & + layout%weight( iDir )*2._rk*cs2inv*mass_dens(iField) &
              !&    * (1.0_rk + mixture%omega_diff*0.5_rk)              &
              &    * ( layout%stencil(1)%cxDir( 1, iDir )*ux(iElem)    &
              &    +   layout%stencil(1)%cxDir( 2, iDir )*uy(iElem)    &
              &    +   layout%stencil(1)%cxDir( 3, iDir )*uz(iElem))
        endif
      enddo
      endif  
    enddo
  end subroutine spc_vel_bb
!******************************************************************************!

!******************************************************************************!
!> species Outlet Pressure extrapolation boundary
!!
!! This is taken from the paper: 
!! M. Junk and Z. Yang, 
!! Asymptotic Analysis of Lattice Boltzmann Outflow Treatments
!! Commun. Comput. Phys., pp. 1–11, 2011.
!!
!! * Pressure as defined in the configuration file
!! Usage 
!! -----
!!\verbatim
!!boundary_condition = {  
!! { label = 'outlet', 
!!    kind = 'spc_outlet_expol', 
!!    pressure = 1.0
!!     }
!!\endverbatim
?? copy :: bc_routineHeader( spc_outlet_expol )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: fTmpAll( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
    real(kind=rk) :: fTmp_1
    real(kind=rk) :: fTmp_2
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    integer :: iDir              !< Direction counter
    integer :: iElem             !< Element counter
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    integer :: pos, ifieldloc, nFields, neighPos
    !------------------------------------------------------------------------ 
    nFields = varSys%nVars
    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems(iLevel)
      neighPos = me%elems(iLevel)%neigh(1,iElem)
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          !neighbor element
          fTmpAll( ?IDX?( pos, iElem, nScalars ) ) = &
            & state( ?SAVE?( pos, neighPos, nScalars, neigh ), currT)
          !local element
          !fTmpAll( ?IDX?( pos, iElem, nScalars ) ) = bcBuffer(                    &
          !  & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField
    enddo !iElem

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Extrapolate equilibrium from local
    call varSys%derive( derVarPos%equilFromState )%do(                         &
      &                       input = fTmpAll,                                 &
      &                       bound = bound,                                   &
      &                       tree = tree,                                     &
      &                       subTree = subTree,                               &
      &                       requVar = varSys%Variable%val(                   &
      &                                 derVarPos%equilFromState ),            &
      &                       globSys = varSys, stencil = layout%stencil(1),   &
      &                       res = fEq,                                       &
      &                       conf = varSys%solSpec%conf )

    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      ! Treat all directions, but actually we only want to treat 
      ! non-normal directions
      ! Equation (3.2b)
      do iDir = 1,layout%stencil(1)%QQ
        ! the bitmask points into the incoming direction, which must
        ! be updated. As the kernel reads this value with FETCH,
        ! we store it to the FETCH position. 
        ! This results for 
        ! - PUSH in storing the density to where the bitmask 
        !   points, i.e. incoming
        ! - PULL storing the density pointing outwards, but as 
        !   this density is being bounced back before collision in 
        !   the kernel, this is the correct position
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
        ! KM: pre-collision of current time step must be used for extrapolation
          fTmp_1 = me%elems( iLevel )%neighBufferPre( 1, &
            & ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
          fTmp_2 = me%elems( iLevel )%neighBufferPre( 2, &
            & ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )

          ! update the incoming velocity direction  
state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
          !& 1.5_rk*fTmp_1 - 0.5_rk*fTmp_2
           &  feq( ?IDX?(iDir, iElem, layout%stencil(1)%QQ) )
        endif
      enddo

      ! then overwrite the normal direction with special treatment
      ! Equation (3.2a)
!      iDir = globBC%elems( iLevel )%normalInd( iElem )
!state( ?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), currT)  = &
!        &  feq( ?IDX?(iDir, iElem, layout%stencil(1)%QQ) ) !&
        !& +(fTmpAll(?IDX?(varPos(layout%stencil(1)%cxDirInv(iDir)), iElem, nScalars))&
        !& -feq( ?IDX?(layout%stencil(1)%cxDirInv(iDir), iElem, layout%stencil(1)%QQ) ) )
        !& fTmpAll(?IDX?(varPos(iDir), iElem, nScalars))
      endif 
    enddo !iElem

  end subroutine spc_outlet_expol
!******************************************************************************!


!******************************************************************************!
  !> Outlet species velocity equilibrium boundary with specified
  !! mixture averaged mass velocity.
  !! molefraction is extrapolated here.
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_outlet_vel', 
  !!    velocityX = 0.0,
  !!    velocityY = 0.0,
  !!    velocityZ = 0.0,
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_outlet_vel )
    !---------------------------------------------------------------------------
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
!    real(kind=rk) :: rho
    real(kind=rk) :: mass_dens( varSys%nVars ) !, num_dens( varSys%nVars )
!    real(kind=rk) :: tot_NumDens, tot_MassDens
    real(kind=rk) :: ux(globBC%nElems(iLevel)),                                &
      &              uy(globBC%nElems(iLevel)),                                &
      &              uz(globBC%nElems(iLevel))
    real(kind=rk) :: IC_moleFrac(varSys%nVars, globBC%nElems(iLevel))                
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
!    real(kind=rk) :: kinePress, tot_NumDens0
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    integer :: offVel, posVel
    !--------------------------------------------------------------------------- 
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      IC_moleFrac(iFieldLoc, :) = tem_spatial_for(                             &
        & conf = params%general%solver%conf(1),                                &
        & me = IC_fields(iFieldLoc)%ini_state(1),                              &
        & treeIDs = glob%levelDesc( iLevel )%total(                            &
        &                           globBC%elems(iLevel)%elem(:) ),            &
        & tree = tree,                                                         &
        & n = globBC%nElems(iLevel) )
    end do    

!write(*,*) 'molWeight Inlet BC', molWeight
!stop
    ! Get velocity which might be defined to vary in space and time
    ux = tem_spacetime_for(                                                    &
      & me = me%BC_states%velocity(1)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    uy = tem_spacetime_for(                                                    &
      & me =  me%BC_states%velocity(2)%spacetime,                              &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    uz = tem_spacetime_for(                                                    &
      & me = me%BC_states%velocity(3)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
!write(*,*) 'vel ', ux(globBC%nElems(iLevel)/2), physics%fac( iLevel )%vel
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel
!write(*,*) 'vel ', ux(globBC%nElems(iLevel)/2)

!stop

    input_loc = 0.0_rk
    offVel = globBC%nElems(iLevel)*nFields
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      mass_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
!        num_dens(iFieldLoc) = mass_dens(iFieldLoc)/molWeight(iFieldLoc)
      enddo 

!      !total number density
!      tot_NumDens = sum(num_dens)
!      !total mass density
!      tot_massDens = sum(mass_dens)
!
!      tot_NumDens0 = mixture%rho0LB &
!        &          / dot_product( IC_moleFrac(:, iElem), molWeight )
!
!      ! compute mixture kinematic pressure from extrapolate density 
!      ! p = cs2*(sum(phi_k*rho_k) - min_a(m_a)*n0)/rho0
!      ! \todo check kinePress for increase in density as in spc_molefrac bc
!      kinePress = cs2 * ( dot_product(phi, mass_dens)                          &
!        &       - minval(molWeight)*tot_NumDens0 )/mixture%rho0LB

      ! compute current species mass density from specified molefraction at
      ! boundary
      ! rho = n_t * chi_i * m_i + massFrac_i * rho0 * p/(cs2*phi_i)
      ! 1st term is zero order density
      ! 2nd term is second order density with kinematic mixture pressure
      !rho = tot_NumDens * moleFrac(iElem) * fieldProp%species%molWeight        &
      !KM: using inital mixture number density rho0/mixtureMOlWeight
      !Using local tot_NuMdens increases density over time
!      rho = mass_dens(iField)
        !& + cs2inv * ( mass_dens(iField)/tot_massDens ) * mixture%rho0LB         &
        !& * kinePress / fieldProp%species%molWeigRatio
      
      !store in input_loc to compute equilibrium from state
      !assume all fields have same velocity at outlet
      posVel = offVel + (iElem-1)*nFields*3
      do iFieldLoc = 1, nFields
        input_loc( (iElem-1)*nFields + iFieldLoc ) = mass_dens(iFieldLoc)
        input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = ux(iElem)
        input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uy(iElem)
        input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uz(iElem)
      end do  

!write(*,*) 'totNumDens ', tot_NumDens
!write(*,*) 'totMassDens ', tot_massDens
!write(*,*) 'kinePress ', kinePress
!write(*,*) 'rho_k ', rho 
!stop
    enddo !iElem

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input = input_loc,                                      &
      &                bound = bound,                                          &
      &                tree = tree,                                            &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf = varSys%solSpec%conf,                             & 
      &                res = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(                         &
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) )   &
        & then

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
      endif  
    enddo

  end subroutine spc_outlet_vel
!******************************************************************************!


!******************************************************************************!
  !> Outlet mixture pressure species equilibrium boundary
  !! kinematic pressure is computed from pressure 
  !! species density and velocity are extrapolated
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_outlet_eq', 
  !!    pressure = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_outlet_eq )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    !store pdf of all fields for derive function pointer input
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
    real(kind=rk) :: fTmpNeigh_1( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
    real(kind=rk) :: fTmpNeigh_2( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
!    real(kind=rk) :: uxB_1(3*globBC%nElems(iLevel))   ! Velocity on boundary element
!    real(kind=rk) :: uxB_2(3*globBC%nElems(iLevel))   ! Velocity on boundary element
!    real(kind=rk) :: uxB(3*globBC%nElems(iLevel))   ! Velocity on boundary element
    real(kind=rk) :: mass_dens( varSys%nVars )
    real(kind=rk) :: phi(varSys%nVars)
    integer :: iDir              ! Direction counter
    integer :: iElem             ! Element counter
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    integer :: nFields, iFieldLoc, pos
    real(kind=rk) :: press(globBC%nElems(iLevel))
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
!    integer :: offVel, posVel
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
    !--------------------------------------------------------------------------- 
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
    end do    

    ! \todo KM: use pressure term to set species density at boundary
    ! get density value from the definition in the lua file
    press = tem_spacetime_for(                                                 &
      & me = me%BC_states%pressure%spacetime,                                  &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    ! kinematic pressure = pressure/rho0
    press = (press/mixture%rho0)/(physics%fac( iLevel )%press/physics%rho0)
 
    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )

    input_loc = 0.0_rk
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      mass_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp( ?IDX?( pos, iElem, nScalars ) ) = bcBuffer(                    &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
!          fTmpNeigh_1( ?IDX?( pos, iElem, nScalars ) ) = state( &
!          & ?IDX?(pos, globBC%elems(iLevel)%elem(iElem), nScalars, neigh), currT)
          fTmpNeigh_1( ?IDX?( pos, iElem, nScalars ) ) = state( &
          & ?IDX?(pos, me%elems(iLevel)%neigh(1,iElem), nScalars, neigh), currT)
          fTmpNeigh_2( ?IDX?( pos, iElem, nScalars ) ) = state( &
          & ?IDX?(pos, me%elems(iLevel)%neigh(2,iElem), nScalars, neigh), currT)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) +                        &
            & fTmp( ?IDX?( pos, iElem, nScalars ) )
        end do

        input_loc( (iElem-1)*nFields + iFieldLoc ) = mass_dens(iFieldLoc)
      enddo 
!      write(*,*) me%elems(iLevel)%neigh(1,iElem),me%elems(iLevel)%neigh(2,iElem),&
!        & globBC%elems(iLevel)%elem( iElem )
    end do !iElem
!    stop
    ! Extrapolate equilibrium from local
    call varSys%derive( derVarPos%equilFromState )%do(                      &
      &                       input = fTmp,                                    &
      &                       bound = bound,                                   &
      &                       tree = tree,                                     &
      &                       subTree = subTree,                               &
      &                       requVar = varSys%Variable%val(                   &
      &                                 derVarPos%equilFromState ),         &
      &                       globSys = varSys, stencil = layout%stencil(1),   &
      &                       res = fEq,                                       &
      &                       conf = varSys%solSpec%conf )

!    offVel = globBC%nElems(iLevel)*nFields
!    uxB = 0.0_rk
!    uxB_1 = 0.0_rk
!    uxB_2 = 0.0_rk
!!    do iFieldLoc = 1, nFields
!!      call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
!!        &                       input = fTmpNeigh_1,                           &
!!        !&                       input = fTmp,                           &
!!        &                       bound = bound,                                 &
!!        &                       tree = tree,                                   &
!!        &                       subTree = subTree,                             &
!!        &                       requVar = varSys%Variable%val(                 &
!!        &                                 derVarPosAll(iFieldLoc)%velocity ),  &
!!        &                       globSys = varSys, stencil = layout%stencil(1), &
!!        &                       res = uxB_1,                                   &
!!        &                       conf = varSys%solSpec%conf )
!!
!!      call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
!!        &                       input = fTmpNeigh_2,                           &
!!        &                       bound = bound,                                 &
!!        &                       tree = tree,                                   &
!!        &                       subTree = subTree,                             &
!!        &                       requVar = varSys%Variable%val(                 &
!!        &                                 derVarPosAll(iFieldLoc)%velocity ),  &
!!        &                       globSys = varSys, stencil = layout%stencil(1), &
!!        &                       res = uxB_2,                                   &
!!        &                       conf = varSys%solSpec%conf )
!!
!!      uxB = 1.5_rk*uxB_1 - 0.5_rk*uxB_2
!!      do iElem = 1, globBC%nElems(iLevel)
!!        posVel = offVel + (iElem-1)*nFields*3
!!        input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = uxB((iElem-1)*3+1)
!!        input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uxB((iElem-1)*3+2)
!!        input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uxB((iElem-1)*3+3)
!!      end do    
!!    end do  
!     call varSys%derive( derVarPosAll(nFields+1)%velocity )%do(               &
!        !&                       input = fTmpNeigh_1,                           &
!        &                       input = fTmp,                           &
!        &                       bound = bound,                                 &
!        &                       tree = tree,                                   &
!        &                       subTree = subTree,                             &
!        &                       requVar = varSys%Variable%val(                 &
!        &                                 derVarPosAll(nFields+1)%velocity ),  &
!        &                       globSys = varSys, stencil = layout%stencil(1), &
!        &                       res = uxB_1,                                   &
!        &                       conf = varSys%solSpec%conf )
!
!    call varSys%derive( derVarPosAll(nFields+1)%velocity )%do(               &
!        &                       input = fTmpNeigh_2,                           &
!        &                       bound = bound,                                 &
!        &                       tree = tree,                                   &
!        &                       subTree = subTree,                             &
!        &                       requVar = varSys%Variable%val(                 &
!        &                                 derVarPosAll(nFields+1)%velocity ),  &
!        &                       globSys = varSys, stencil = layout%stencil(1), &
!        &                       res = uxB_2,                                   &
!        &                       conf = varSys%solSpec%conf )
!
!      uxB = 1.0_rk*uxB_1 !- 0.5_rk*uxB_2
!    do iFieldLoc = 1, nFields
!      do iElem = 1, globBC%nElems(iLevel)
!        posVel = offVel + (iElem-1)*nFields*3
!        input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = uxB((iElem-1)*3+1)
!        input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uxB((iElem-1)*3+2)
!        input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uxB((iElem-1)*3+3)
!      end do    
!    end do    
!
!
!    fEq = 0.0_rk
!    ! Calculate the equilibrium distribution
!    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
!      &                input = input_loc,                                      &
!      &                bound = bound,                                          &
!      &                tree = tree,                                            &
!      &                subTree = subTree,                                      &
!      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
!      &                globSys = varSys,                                       &
!      &                stencil = layout%stencil(1),                            &
!      &                conf = varSys%solSpec%conf,                             & 
!      &                res = fEq )

    ! Run over all elements in list for this boundary
    do iElem = 1, globBC%nElems( iLevel )
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      do iDir = 1,layout%stencil(1)%QQ
        ! the bitmask points into the incoming direction, which must
        ! be updated. As the kernel reads this value with FETCH,
        ! we store it to the FETCH position. 
        ! This results for 
        ! - PUSH in storing the density to where the bitmask 
        !   points, i.e. incoming
        ! - PULL storing the density pointing outwards, but as 
        !   this density is being bounced back before collision in 
        !   the kernel, this is the correct position
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Now assign the values
          state( &
&?FETCH?( varPos(iDir), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ),&
&         currT)  = &
            & fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
!          if (fEq( ?IDX?(iDir,iElem, layout%stencil(1)%QQ) ) < 0.0_rk) then
!            write(*,*)'BC Outlet: time ', sim_time%iter
!            write(*,*) 'found negative', iElem, iField, &
!            &  fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
!            stop
!          endif
        endif
      enddo
      endif
    enddo


  end subroutine spc_outlet_eq
!******************************************************************************!


!******************************************************************************!
  !> Mole fraction boundary condition
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moleFrac', 
  !!    moleFraction = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moleFrac )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: moleFrac(globBC%nElems(iLevel))                
    real(kind=rk) :: rho, press
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: fEq, fEqStar, ucx, ucxStar, ucxQuad, ucxQuadStar
    real(kind=rk) :: usq, usqStar
    real(kind=rk) :: velAvg(3), velQuad(3), velQuadStar(3), eqVel(3)
    real(kind=rk) :: vel(3,varSys%nVars), spc_vel(3)
    real(kind=rk) :: mom(3,varSys%nVars), spc_mom(3), mom_inG(3), vel_inG(3)
    real(kind=rk) :: mass_dens(varSys%nVars), num_dens(varSys%nVars)
    real(kind=rk) :: totMassDens, totNumDens
    real(kind=rk) :: moleFrac_loc(varSys%nVars)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    real(kind=rk) :: resi_coeff( varSys%nVars, varSys%nVars )
    real(kind=rk) :: weight0_inv
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: fEq_mom( layout%stencil(1)%QQ )
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !--------------------------------------------------------------------------- 
!    write(*,*) 'boundary label ', trim(me%label)
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    end do    
    weight0_inv = 1.0_rk/layout%weight(layout%stencil(1)%restPosition)

    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    ! mole fraction
    moleFrac = tem_spacetime_for(                                              &
      & me = me%BC_states%moleFrac%spacetime,                                  &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                     &
      & dt_phy  = physics%dt )

    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      ! local state vector to compute density and velocity of other species
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer( ?IDX?( pos, globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo
      enddo

      !KM: using inital mixture number density rho0/mixtureMOlWeight
      rho = mixture%moleDens0LB *  moleFrac(iElem) &
        & * fieldProp%species%molWeight
      press = cs2 * rho * fieldProp%species%molWeigRatio

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      !local density and momentum
      mass_dens = 0.0_rk
      vel = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc)               &
            & + fTmp_all(varSys%Variable%val(iFieldLoc)%varPos(iDir)) 
        end do
        num_dens(iFieldLoc) = mass_dens(iFieldLoc) &
          &                 / molWeight(iFieldLoc)
        !mass flux
        call varSys%derive( derVarPosAll(iFieldLoc)%momentum )%do(              &
          &                       input = fTmp_all,                             &
          &                       bound = bound,                                &
          &                       tree = tree,                                  &
          &                       subTree = subTree,                            &
          &                       requVar = varSys%Variable%val(                &
          &                                 derVarPosAll(iFieldLoc)%momentum ), &
          &                       globSys = varSys, stencil = layout%stencil(1),&
          &                       res = spc_mom,                                &
          &                       conf = varSys%solSpec%conf )
        mom(:, iFieldLoc) = spc_mom
      end do  

      mass_dens(iField) = rho
      num_dens(iField) = rho / fieldProp%species%molWeight

      do iFieldLoc = 1, nFields
        vel(:, iFieldLoc) =  mom(:, iFieldLoc) / mass_dens(iFieldLoc)
      end do  
 
      ! total mass density
      totMassDens = sum(mass_dens)

      ! mole fraction
      moleFrac_loc = num_dens/sum(num_dens)

      ! compute velocity of transformed variable g. Eq. which is normally
      ! used to compute actual velocity by solving LSE
      mom_inG = mom(:, iField)
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG + ( mixture%omega_diff * 0.5_rk                    &
            &               * resi_coeff(iField, iFieldLoc) * phi(iField)    &
            &               * moleFrac_loc(iFieldLoc) / mixture%paramB )     &
            &               * mom(:, iField)
      end do
      !set non-diagonal part
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG - ( mixture%omega_diff * 0.5_rk                    &
          &                 * resi_coeff(iField, iFieldLoc) * phi(iFieldLoc) &
          &                 * moleFrac_loc(iField) / mixture%paramB )        &
          &                 * mom(:,iFieldLoc)
      end do

      vel_inG = mom_inG/rho

!      vel(:, iField) = vel_inG
!
      eqVel = vel(:, iField)
      do iFieldLoc = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(iField, iFieldLoc)               & 
          &      * phi(iField) * moleFrac_loc(iFieldLoc)                  &
          &      * ( vel(:, iFieldLoc) - vel( :, iField ) )               &
          &      / mixture%paramB
      end do

      ! mass averaged mixture velocity
      velAvg(1) = sum(mom(1,:)) / totMassDens
      velAvg(2) = sum(mom(2,:)) / totMassDens
      velAvg(3) = sum(mom(3,:)) / totMassDens
        
      velQuadStar(:) = mixture%theta_eq*velAvg(:)                       &
        &            + (1.0_rk-mixture%theta_eq) * eqVel(:)
      velQuad(:) = mixture%theta_eq*velAvg(:)                           &
        &            + (1.0_rk-mixture%theta_eq)*vel(:, iField)

      usqStar = dot_product(velQuadStar, velQuadStar)*t2cs2inv
      usq = dot_product(velQuad, velQuad)*t2cs2inv

      do iDir =1,layout%stencil(1)%QQ 
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ucx = dot_product(layout%stencil(1)%cxDir(:, iDir),             &
            &               vel(:,iField))
          ucxQuad = dot_product(layout%stencil(1)%cxDir(:, iDir),         &
            &               velQuad)

          ucxStar = dot_product(layout%stencil(1)%cxDir(:, iDir),         &
            &                   eqVel)
          ucxQuadStar = dot_product(layout%stencil(1)%cxDir(:, iDir),     &
            &                   velQuadStar)

          ! eqVel is actually is rho_i*eqVel so ucxStar is not multiplied
          ! with rho in below equation
          fEqStar = layout%weight(iDir) * rho                 &
            & * ( phi(iField) + ucxStar * cs2inv                 &
            & + ucxQuadStar * ucxQuadStar * t2cs4inv - usqStar )

          fEq = layout%weight(iDir) * rho         &
            & * ( phi(iField) + ucx * cs2inv         &
            & + ucxQuad * ucxQuad * t2cs4inv - usq )

          if ( iDir == layout%stencil(1)%restPosition ) then      
            ! equilibrium at rest
            fEq = layout%weight( iDir ) * rho * ( &
                    & ( weight0_inv + (1.0_rk-weight0_inv) * phi(iField) ) - usq )
            fEqStar = layout%weight( iDir ) * rho * ( &
                    & ( weight0_inv + (1.0_rk-weight0_inv) * phi(iField) ) - usqStar )
          end if  

          ! set equilibrium
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq + mixture%omega_diff*0.5_rk*( fEq - fEqStar )
        end if   
      end do! iDir

      end if
    end do

  end subroutine spc_moleFrac
!******************************************************************************!

!******************************************************************************!
  !> Mole fraction boundary condition
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moleFrac_eq', 
  !!    moleFraction = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moleFrac_eq )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    !store pdf of all fields for derive function pointer input
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
    real(kind=rk) :: rho
    real(kind=rk) :: mass_dens( varSys%nVars )
    real(kind=rk) :: num_dens( varSys%nVars )
    real(kind=rk) :: tot_massDens, tot_numDens
    real(kind=rk) :: moleFrac(globBC%nElems(iLevel))                
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
    real(kind=rk) :: uxB(3*globBC%nElems(iLevel))   !< Velocity on boundary element
    real(kind=rk) :: kinePress !, tot_NumDens0
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    integer :: offVel, posVel
    !--------------------------------------------------------------------------- 
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
    end do    

!write(*,*) 'iField ', iField
!write(*,*) 'molWeight moleFrac BC', molWeight
!write(*,*) 'phi ', phi
!stop

    ! mole fraction
    moleFrac = tem_spacetime_for(                                              &
      & me = me%BC_states%moleFrac%spacetime,                                  &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                     &
      & dt_phy  = physics%dt )

    input_loc = 0.0_rk
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      mass_dens = 0.0_rk
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp( ?IDX?( pos, iElem, nScalars ) ) = bcBuffer(                    &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) +                        &
            & fTmp( ?IDX?( pos, iElem, nScalars ) )
        end do
        num_dens(iFieldLoc) = mass_dens(iFieldLoc)/molWeight(iFieldLoc)
      enddo 

      !total number density
      tot_NumDens = sum(num_dens)
      !total mass density
      tot_massDens = sum(mass_dens)

!      tot_NumDens0 = mixture%rho0LB &
!        &          / dot_product( IC_moleFrac(:, iElem), molWeight )
      ! compute mixture kinematic pressure from extrapolate density 
      ! p = cs2*(sum(phi_k*rho_k) - min_a(m_a)*n0)/rho0
      ! ASSUME n0 to be constant overtime. Neglecting small difference
      ! over time and using current tot_NumDensity
      ! \todo Kinematic pressure term with molefraction increase total density
      ! and does not reach steady state
      ! kinePress with initial num dens produce NaN value
      kinePress = cs2 * ( dot_product(phi, mass_dens)                          &
        &       - minval(molWeight)*mixture%moleDens0LB )/mixture%rho0LB 
      ! compute current species mass density from specified molefraction at
      ! boundary
      ! rho = n_t0 * chi_i * m_i + massFrac_i * rho0 * p/(cs2*phi_i)
      ! 1st term is zero order density
      ! 2nd term is second order density with kinematic mixture pressure
      ! \todo KM: what about kinematic pressure here?
      !KM: using inital mixture number density rho0/mixtureMOlWeight
      rho = mixture%moleDens0LB *  moleFrac(iElem) * fieldProp%species%molWeight !&
      !rho = tot_NumDens0 * moleFrac(iElem) * fieldProp%species%molWeight!       &
      !rho = tot_NumDens * moleFrac(iElem) * fieldProp%species%molWeight        &
        ! KM: using kinematic pressure here produce NAN value
      !  & + cs2inv * ( mass_dens(iField)/tot_massDens ) * mixture%rho0LB         &
      !  & * kinePress / fieldProp%species%molWeigRatio
      !set density for current field   
      mass_dens(iField) = rho
      !store in input_loc to compute equilibrium from state
      !input_loc( iElem ) = rho
      do iFieldLoc = 1, nFields
        input_loc( (iElem-1)*nFields + iFieldLoc ) = mass_dens(iFieldLoc)
      end do
!write(*,*) 'iElem ', iElem
!write(*,*) 'totNumDens ', tot_NumDens
!write(*,*) 'moleFrac ', moleFrac(iElem)
!write(*,*) 'totMassDens ', tot_massDens
!write(*,*) 'kinePress ', kinePress,  &
!  &dot_product(phi, mass_dens), minval(molWeight)*tot_NumDens0
!!write(*,*) 'rho0', mixture%rho0LB
!write(*,*) 'mass_dens all ', mass_dens
!write(*,*) 'mass_dens ', mass_dens(iField)
!if(iElem==1) write(*,*) 'rho_k ', rho, ' totNumDens ', tot_NumDens
!write(*,*) 'num_dens ', num_dens(iField)
!stop
    enddo !iElem

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Get local equilibrium velocity
!write(*,*) 'In derive EqVelState:'
    offVel = globBC%nElems(iLevel)*nFields
    do iFieldLoc = 1, nFields
      call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
        &                       input = fTmp,                                  &
        &                       bound = bound,                                 &
        &                       tree = tree,                                   &
        &                       subTree = subTree,                             &
        &                       requVar = varSys%Variable%val(                 &
        &                                 derVarPosAll(iFieldLoc)%velocity ),  &
        &                       globSys = varSys, stencil = layout%stencil(1), &
        &                       res = uxB,                                     &
        &                       conf = varSys%solSpec%conf )
      do iElem = 1, globBC%nElems(iLevel)
        posVel = offVel + (iElem-1)*nFields*3
        input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = uxB((iElem-1)*3+1)
        input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uxB((iElem-1)*3+2)
        input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uxB((iElem-1)*3+3)
      end do    
    end do  

    fEq = 0.0_rk
    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input = input_loc,                                      &
      &                bound = bound,                                          &
      &                tree = tree,                                            &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf = varSys%solSpec%conf,                             & 
      &                res = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
      endif
    enddo

  end subroutine spc_moleFrac_eq
!******************************************************************************!


!******************************************************************************!
  !> Mole fraction boundary condition with thermodynamic factor
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moleFrac_wtdf', 
  !!    moleFraction = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moleFrac_wtdf )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: fEq
    !store pdf of all fields for derive function pointer input
    real(kind=rk) :: fTmp( varSys%nScalars )
    real(kind=rk) :: mass_dens( varSys%nVars ), num_dens( varSys%nVars )
    real(kind=rk) :: tot_numDens, tot_massDens
    real(kind=rk) :: moleFrac(globBC%nElems(iLevel))                
    integer :: iElem, iDir, iFieldLoc, iField_2, iField_3, nFields, pos
    real(kind=rk), dimension(varSys%nVars) :: molWeight, phi
    real(kind=rk), dimension(varSys%nVars) :: moleFrac_loc
    real(kind=rk), dimension(3, varSys%nVars) :: first_moments, velocity
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk), &
      & dimension(varSys%nVars, varSys%nVars) :: matA, invA, resi_coeff,       &
      & thermodynamic_fac, inv_thermodyn_fac, diff_coeff
    real(kind=rk) :: temp, press, phy_moleDens_fac
    real(kind=rk) :: omega, paramB, theta_eq
    real(kind=rk) :: eqVel(3), velAvg(3), velQuad(3), usqr, ucx, ucxQuadTerm
    !------------------------------------------------------------------------ 
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    enddo    

    !KM \todo check moleDens for multilevel
    phy_moleDens_fac = physics%moleDens0
    !fixed parameter
    paramB = mixture%paramB
    !equilibrium theta
    theta_eq = mixture%theta_eq
    
    omega = mixture%relaxLvl(iLevel)%omega_diff
    ! temperature
    temp = mixture%temp0
    ! atmospheric pressure
    press = mixture%atm_press
!write(*,*) 'iField ', iField
!write(*,*) 'resi_coeff ', resi_coeff
!write(*,*) 'molWeight moleFrac BC', molWeight
!write(*,*) 'phi ', phi
!stop

    ! mole fraction
    moleFrac = tem_spacetime_for(                                              &
      & me = me%BC_states%moleFrac%spacetime,                                  &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
!write(*,*) 'IElem ', iElem    
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      mass_dens = 0.0_rk
      first_moments = 0.0_rk
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp( pos ) = bcBuffer(                                              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + fTmp( pos )

          !field momentum (rho*u)
          first_moments( 1, iFieldLoc ) = first_moments( 1, iFieldLoc )              &
            & + fTmp( pos ) * layout%stencil(1)%cxDir( 1, iDir )

          first_moments( 2, iFieldLoc ) = first_moments( 2, iFieldLoc )              &
            & + fTmp( pos ) * layout%stencil(1)%cxDir( 2, iDir )

          first_moments( 3, iFieldLoc ) = first_moments( 3, iFieldLoc )              &
            & + fTmp( pos ) * layout%stencil(1)%cxDir( 3, iDir )

        end do
        num_dens(iFieldLoc) = mass_dens(iFieldLoc)/molWeight(iFieldLoc)
      enddo 

      ! update num_dens, massDens and moleFrac of current species with specified 
      ! molefraction
      num_dens(iField) = moleFrac(iElem) * mixture%moleDens0LB
      mass_dens(iField) = num_dens(iField) * molWeight(iField)

      !total number density
      tot_NumDens = sum(num_dens)

      !total mass density
      tot_massDens = sum(mass_dens)

      !mole fraction
      moleFrac_loc(:) =  num_dens(:)/tot_NumDens
!write(*,*) 'num_dens ', num_dens
!write(*,*) 'num_dens*phy_moleDens_fac ', num_dens*phy_moleDens_fac

      ! MS-Diff coeff matrix from C++ code
      call mus_calc_MS_DiffMatrix( nFields, temp, press,                       &
        &                          num_dens*phy_moleDens_fac, diff_coeff )
!write(*,*) 'diff_coeff ', diff_coeff
      ! Convert to lattice unit
      resi_coeff = physics%fac(iLevel)%diffusivity/diff_coeff
!write(*,*) 'resi_coeff ', resi_coeff

      call mus_calc_thermFactor( nFields, temp, press, moleFrac_loc,            &
        &                        thermodynamic_fac )
!write(*,*) 'thermodyn_fac ', thermodynamic_fac

      inv_thermodyn_fac = invert_matrix( thermodynamic_fac )
!write(*,*) 'inv_thermodyn_fac ', inv_thermodyn_fac

      matA = 0.0_rk
      !build up matrix to solver LSE for actual velocity
      do iFieldLoc = 1, nFields
        !set diagonal part 
        matA(iFieldLoc, iFieldLoc) = 1.0_rk
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            matA(iFieldLoc, iField_2) = matA(iFieldLoc, iField_2) + omega * 0.5_rk&
              &                    * inv_thermodyn_fac(iFieldLoc, iField_2)       &
              &                    * resi_coeff(iField_2, iField_3)               &
              &                    * phi(iField_2) * moleFrac_loc(iField_3)       &
              &                    / paramB
          end do
        end do
        !set non-diagonal part
        do iField_2 = 1, nFields
          do iField_3 = 1, nFields
            matA(iFieldLoc, iField_3) = matA(iFieldLoc, iField_3) - omega * 0.5_rk&
              &                    * inv_thermodyn_fac(iFieldLoc, iField_2)       &
              &                    * resi_coeff(iField_2, iField_3)               &
              &                    * phi(iField_3) * moleFrac_loc(iField_2)       &
              &                    / paramB
          end do
        end do
      end do
!write(*,*) 'matA ', matA

      ! invert matrix
      invA = invert_matrix( matA )
!write(*,*) 'invA ', invA

      !actual velocity of all species
      velocity(1, :) = matmul( invA, first_moments(1,:) ) / mass_dens(:)
      velocity(2, :) = matmul( invA, first_moments(2,:) ) / mass_dens(:)
      velocity(3, :) = matmul( invA, first_moments(3,:) ) / mass_dens(:)
!write(*,*) 'velocity ', velocity

      ! equilibrium velocity with thermodynamic factor
      eqVel( : ) = mass_dens(iField)*velocity( :, iField )
      do iField_2 = 1, nFields
        do iField_3 = 1, nFields
          eqVel( : ) = eqVel( : )                              &
            &        + inv_thermodyn_fac(iField, iField_2)             &
            &        * mass_dens(iField_2)                             &
            &        * resi_coeff( iField_2, iField_3 ) * phi(iField_2)&
            &        * moleFrac_loc(iField_3)                          &
            &        * (velocity(:, iField_3) - velocity(:,iField_2))  &
            &        / paramB 
        end do    
      end do  
!write(*,*) 'eqVel ', eqVel
      !compute mass averaged mixture velocity
      velAvg(1) = dot_product( mass_dens, velocity(1,:) )/tot_massDens
      velAvg(2) = dot_product( mass_dens, velocity(2,:) )/tot_massDens
      velAvg(3) = dot_product( mass_dens, velocity(3,:) )/tot_massDens

      velQuad = theta_eq*velAvg                                              &
        &     + (1.0_rk - theta_eq) * eqVel(:) / mass_dens(iField)

      usqr = dot_product( velQuad, velQuad ) * t2cs2inv

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted

          ucx = dot_product(                                                   &
            & dble(layout%stencil(1)%cxDir(:, iDir)), eqVel(:) )

          ucxQuadTerm = dot_product(                                           &
            & dble(layout%stencil(1)%cxDir(:, iDir)), velQuad )

          ! Equilibrium
          feq = layout%weight(iDir) * ( mass_dens(iField) * ( phi(iField)        &
            & + ucxQuadTerm * ucxQuadTerm * t2cs4inv - usqr ) + ucx * cs2inv )

          if ( iDir == layout%stencil(1)%restPosition ) then      
          ! equilibrium at rest
            select case( trim(layout%stencil(1)%label) )
            case('d2q9')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 9._rk - 5._rk * phi(iField) )/4._rk - usqr )
            case('d3q19')
              feq = layout%weight( iDir ) * mass_dens(iField) * ( &
                    & ( 3._rk - 2._rk * phi(iField) ) - usqr )
            end select
          end if  

          ! set equilibrium
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq
!write(*,*) 'fEq ', fEq            
        endif
      enddo
      endif
    enddo

  end subroutine spc_moleFrac_wtdf
!******************************************************************************!


!******************************************************************************!
  !> molar flux equilibrium boundary condition
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moleflux', 
  !!    moleflux = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moleFlux_eq )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    !store pdf of all fields for derive function pointer input
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
!    real(kind=rk) :: rho
    real(kind=rk) :: mass_dens( varSys%nVars ), num_dens( varSys%nVars )
!    real(kind=rk) :: tot_NumDens, tot_MassDens
    real(kind=rk) :: moleFluxX(globBC%nElems(iLevel)),                         &
      &              moleFluxY(globBC%nElems(iLevel)),                         &
      &              moleFluxZ(globBC%nElems(iLevel))
    real(kind=rk) :: IC_moleFrac(varSys%nVars, globBC%nElems(iLevel))                
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
    real(kind=rk) :: uxB(3*globBC%nElems(iLevel))   !< Velocity on boundary element
    real(kind=rk) :: uxBOthers(3*globBC%nElems(iLevel))   !< Velocity on boundary element
!    real(kind=rk) :: kinePress, tot_NumDens0
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    integer :: offVel, posVel
    !------------------------------------------------------------------------ 
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      IC_moleFrac(iFieldLoc, :) = tem_spatial_for(                             &
        & conf = params%general%solver%conf(1),                                &
        & me = IC_fields(iFieldLoc)%ini_state(1),                              &
        & treeIDs = glob%levelDesc( iLevel )%total(                            &
        &                           globBC%elems(iLevel)%elem(:) ),            &
        & tree = tree,                                                         &
        & n = globBC%nElems(iLevel) )
    end do    

!write(*,*) 'molWeight moleFlux BC', molWeight
!stop
    ! Get velocity which might be defined to vary in space and time
    moleFluxX = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(1)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleFluxY = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(2)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleFluxZ = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(3)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    moleFluxX = moleFluxX/physics%fac( iLevel )%moleFlux
    moleFluxY = moleFluxY/physics%fac( iLevel )%moleFlux
    moleFluxZ = moleFluxZ/physics%fac( iLevel )%moleFlux

!write(*,*) 'physical moleFlux ', physics%fac( iLevel )%moleFlux
!write(*,*) 'moleFluxX ', moleFluxX
!write(*,*) 'moleFluxY ', moleFluxY
!write(*,*) 'moleFluxZ ', moleFluxZ
!stop
    input_loc = 0.0_rk
    uxB = 0.0_rk
    uxBOthers = 0.0_rk
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      mass_dens = 0.0_rk
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp( ?IDX?( pos, iElem, nScalars ) ) = bcBuffer(                    &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) +                        &
            & fTmp( ?IDX?( pos, iElem, nScalars ) )
        end do
        num_dens(iFieldLoc) = mass_dens(iFieldLoc)/molWeight(iFieldLoc)
      enddo 

!      !total number density
!      tot_NumDens = sum(num_dens)
!      tot_NumDens0 = mixture%rho0LB &
!        &          / dot_product( IC_moleFrac(:, iElem), molWeight )
!      ! compute mixture kinematic pressure from extrapolate density 
!      !total mass density
!      tot_massDens = sum(mass_dens)
!
!      ! compute mixture kinematic pressure from extrapolate density 
!      ! p = cs2*(sum(phi_k*rho_k) - min_a(m_a)*n0)/rho0
!      ! ASSUME n0 to be constant overtime. Neglecting small difference
!      ! over time and using current tot_NumDensity
!      ! \todo check with local number density
!      kinePress = cs2 * ( dot_product(phi, mass_dens)                          &
!        &       - minval(molWeight)*mixture%moleDens0LB )/mixture%rho0LB 
!      ! compute current species mass density from specified molefraction at
!      ! boundary
!      ! rho = n_t * chi_i * m_i + massFrac_i * rho0 * p/(cs2*phi_i)
!      ! 1st term is zero order density
!      ! 2nd term is second order density with kinematic mixture pressure
!      ! \todo KM: what about kinematic pressure here?
!      !rho = tot_NumDens0 * moleFrac_bnd(iElem) * fieldProp%species%molWeight!       &
!      rho = mass_dens(iField)!                                                  &
        !& + cs2inv * ( mass_dens(iField)/tot_massDens ) * mixture%rho0LB         &
        !& * kinePress / fieldProp%species%molWeigRatio
      
      !store in input_loc to compute equilibrium from state
 !     input_loc( iElem ) = rho
      do iFieldLoc = 1, nFields
        input_loc( (iElem-1)*nFields + iFieldLoc ) = mass_dens(iFieldLoc)
      end do

      !store velocity in input_loc
      !caulate velocity from moleFlux and num_dens
      ! moleFlux = num_dens * velocity
      uxB((iElem-1)*3+1) = moleFluxX(iElem) / num_dens(iField)
      uxB((iElem-1)*3+2) = moleFluxY(iElem) / num_dens(iField)
      uxB((iElem-1)*3+3) = moleFluxZ(iElem) / num_dens(iField)
    enddo

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    offVel = globBC%nElems(iLevel)*nFields
    do iFieldLoc = 1, nFields
      if(iField == iFieldLoc) then
        do iElem = 1, globBC%nElems(iLevel)
          posVel = offVel + (iElem-1)*nFields*3
          input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = uxB((iElem-1)*3+1)
          input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uxB((iElem-1)*3+2)
          input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uxB((iElem-1)*3+3)
        end do  
      else
        call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(              &
          &                       input = fTmp,                                 &
          &                       bound = bound,                                &
          &                       tree = tree,                                  &
          &                       subTree = subTree,                            &
          &                       requVar = varSys%Variable%val(                &
          &                                 derVarPosAll(iFieldLoc)%velocity ), &
          &                       globSys = varSys, stencil = layout%stencil(1),&
          &                       res = uxBOthers,                              &
          &                       conf = varSys%solSpec%conf )
        do iElem = 1, globBC%nElems(iLevel)
          posVel = offVel + (iElem-1)*nFields*3
          input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = uxBOthers((iElem-1)*3+1)
          input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uxBOthers((iElem-1)*3+2)
          input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uxBOthers((iElem-1)*3+3)
        end do  
      endif
    enddo  

    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input = input_loc,                                      &
      &                bound = bound,                                          &
      &                tree = tree,                                            &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf = varSys%solSpec%conf,                             & 
      &                res = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
      endif
    enddo


  end subroutine spc_moleFlux_eq
!******************************************************************************!


!******************************************************************************!
  !> molar flux boundary condition like velocity bounce back bc type
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moleflux', 
  !!    moleflux = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moleFlux )
    !------------------------------------------------------------------------ 
    !store pdf of all fields for derive function pointer input
    !real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
    !  &              * varSys%nVars )
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: mass_dens, num_dens
    real(kind=rk) :: massFlux(3)
    real(kind=rk) :: moleFluxX(globBC%nElems(iLevel)),                         &
      &              moleFluxY(globBC%nElems(iLevel)),                         &
      &              moleFluxZ(globBC%nElems(iLevel))
    integer :: iElem, iDir
    real(kind=rk) :: uxB(3)   ! Velocity on boundary element
    !------------------------------------------------------------------------ 
    ! Get velocity which might be defined to vary in space and time
    moleFluxX = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(1)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleFluxY = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(2)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleFluxZ = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(3)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    moleFluxX = moleFluxX/physics%fac( iLevel )%moleFlux
    moleFluxY = moleFluxY/physics%fac( iLevel )%moleFlux
    moleFluxZ = moleFluxZ/physics%fac( iLevel )%moleFlux

!write(*,*) 'physical moleFlux ', physics%fac( iLevel )%moleFlux
!write(*,*) 'moleFluxX ', moleFluxX
!write(*,*) 'moleFluxY ', moleFluxY
!write(*,*) 'moleFluxZ ', moleFluxZ
!stop
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      uxB = 0.0_rk
      mass_dens = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        fTmp( iDir ) = bcBuffer(                    &
       & ?IDX?( varpos(iDir), globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        mass_dens = mass_dens + fTmp( iDir )
      enddo 
      num_dens = mass_dens/fieldProp%species%molWeight
      !store velocity in input_loc
      !caulate velocity from moleFlux and num_dens
      ! moleFlux = num_dens * velocity
      uxB(1) = moleFluxX(iElem) / num_dens
      uxB(2) = moleFluxY(iElem) / num_dens
      uxB(3) = moleFluxZ(iElem) / num_dens

      !write(dbgUnit(1),*) iElem, 'num_dens ', num_dens, 'mass_dens ', mass_dens
      !write(dbgUnit(1),*) 'moleFlux ', moleFluxX(iElem), moleFluxY(iElem), moleFluxZ(iElem)
      !write(dbgUnit(1),*) 'uxB ', uxB
      !write(dbgUnit(1),*) 'uxB*rho ', mass_dens*uxB

      !do iDir = 1,layout%stencil(1)%QQ
      !  fTmp(iDir) = bcBuffer( & 
      !    &?IDX?(varpos(iDir),globBC%elems(iLevel)%elemBuffer(iElem),nScalars))
      !end do

      massFlux(1) = moleFluxX(iElem) * fieldProp%species%molWeight
      massFlux(2) = moleFluxY(iElem) * fieldProp%species%molWeight
      massFlux(3) = moleFluxZ(iElem) * fieldProp%species%molWeight
      !write(dbgUnit(1),*) 'massFlux ', massFlux

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
&?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
              & = fTmp(layout%stencil(1)%cxDirInv( iDir ))  &
              & + layout%weight( iDir )*2._rk*cs2inv &
              !&    * (1.0_rk + mixture%omega_diff*0.5_rk)              &
              &    * ( layout%stencil(1)%cxDir( 1, iDir )*massFlux(1)   &
              &    +   layout%stencil(1)%cxDir( 2, iDir )*massFlux(2)    &
              &    +   layout%stencil(1)%cxDir( 3, iDir )*massFlux(3))
 
              !& = fTmp( layout%stencil(1)%cxDirInv(iDir)) &
              !& + layout%weight( iDir )*6._rk*mass_dens                        &
              !& * ( layout%stencil(1)%cxDir( 1, iDir )*uxB(1)                  &
              !& +   layout%stencil(1)%cxDir( 2, iDir )*uxB(2)                  &
              !& +   layout%stencil(1)%cxDir( 3, iDir )*uxB(3))
        endif
      enddo
      endif
    enddo


  end subroutine spc_moleFlux
!******************************************************************************!

!******************************************************************************!
  !> molar diffusion flux boundary condition
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moleDiff_flux', 
  !!    moleDiff_flux = 0.0
  !!     }
  !!\endverbatim
  !! \todo KM Adapt moleDiff flux boundary input_loc for equilibrium function pointer
?? copy :: bc_routineHeader( spc_moleDiff_Flux )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    real(kind=rk) :: rho
    !first moments of nSpecies
    real(kind=rk) :: first_moments( 3, varSys%nVars )
    real(kind=rk) :: momentum( 3, varSys%nVars )
    real(kind=rk) :: vel( 3, varSys%nVars )
    real(kind=rk) :: mixMolvel( 3 )
!    real(kind=rk) :: uXB( 3 ) ! species velocity at boundary element
    real(kind=rk) :: eqVel( 3 ) ! species velocity at boundary element
    real(kind=rk) :: mass_dens( varSys%nVars ), num_dens( varSys%nVars )
    real(kind=rk) :: tot_NumDens, tot_MassDens
    real(kind=rk) :: moleDiff_Flux(3,globBC%nElems(iLevel))
    real(kind=rk) :: IC_moleFrac(varSys%nVars, globBC%nElems(iLevel))                
    real(kind=rk) :: moleFrac_bnd(globBC%nElems(iLevel))                
    real(kind=rk) :: moleFrac(varSys%nVars)   
    integer :: iElem, iDir, iFieldLoc, nFields, iComp, pos
    integer :: iFieldDia, iFieldNonDia
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
    real(kind=rk) :: kinePress, tot_NumDens0
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk) :: matrixA( varSys%nVars, varSys%nVars )
    real(kind=rk) :: invA( varSys%nVars, varSys%nVars )
    real(kind=rk) :: resi_coeff( varSys%nVars, varSys%nVars )
    !------------------------------------------------------------------------ 
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,               &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
      IC_moleFrac(iFieldLoc, :) = tem_spatial_for(                             &
        & conf = params%general%solver%conf(1),                                &
        & me = IC_fields(iFieldLoc)%ini_state(1),                              &
        & treeIDs = glob%levelDesc( iLevel )%total(                            &
        &                           globBC%elems(iLevel)%elem(:) ),            &
        & tree = tree,                                                         &
        & n = globBC%nElems(iLevel) )
    end do    

!write(*,*) 'molWeight moleDiffFlux BC', molWeight
!stop
    ! Get velocity which might be defined to vary in space and time
    moleDiff_Flux(1,:) = tem_spacetime_for(                                    &
      & me = me%BC_states%moleDiff_Flux(1)%spacetime,                          &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleDiff_Flux(2,:) = tem_spacetime_for(                                    &
      & me = me%BC_states%moleDiff_Flux(2)%spacetime,                          &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleDiff_Flux(3,:) = tem_spacetime_for(                                    &
      & me = me%BC_states%moleDiff_Flux(3)%spacetime,                          &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    moleDiff_Flux = moleDiff_Flux/physics%fac( iLevel )%moleFlux

!write(*,*) 'moleDiffFluxX ', moleDiff_FluxX
!write(*,*) 'moleDiffFluxY ', moleDiff_FluxY
!write(*,*) 'moleDiffFluxZ ', moleDiff_FluxZ
!stop

    ! mole fraction
    moleFrac_bnd = tem_spacetime_for( me%BC_states%moleFrac%spacetime,         &
      & treeIDs = glob%levelDesc(iLevel)%total(globBC%elems(iLevel)%elem(:)),  &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    input_loc = 0.0_rk
    ! Calculate the mole averaged mixture velocity w = sum(n_i v_i)/n
    do iElem = 1, globBC%nElems(iLevel)
      ! initialize the local velocity for every element
      first_moments = 0.0_rk
      momentum = 0.0_rk
      mass_dens = 0.0_rk
      num_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        ! all fields have some nComponents
        do iDir = 1,layout%stencil(1)%QQ
          ! get the position of the current direction of the depending variable
          ! in the global system (= position in the input)
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          ! density
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) + bcBuffer(              &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )

          first_moments( :, iFieldLoc ) = first_moments( :, iFieldLoc )          &
& + bcBuffer( ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )&
              & * real(layout%stencil(1)%cxDir( :, iDir ), kind=rk)
        end do !iDir
        ! number density
        num_dens(iFieldLoc) = mass_dens(iFieldLoc)/molWeight(iFieldLoc)
      end do !iField

      !total number density
      tot_NumDens = sum(num_dens)
      !total mass density
      tot_massDens = sum(mass_dens)
      ! mole fraction
      moleFrac(:) = num_dens(:) / tot_NumDens

      tot_NumDens0 = mixture%rho0LB &
        &          / dot_product( IC_moleFrac(:, iElem), molWeight )

      matrixA = 0.0_rk
      ! build up the equation system for momentum
      do iFieldLoc = 1, nFields
        ! set diagonal part
        matrixA( iFieldLoc, iFieldLoc ) = 1.0_rk
        do iFieldDia = 1, nFields
          matrixA( iFieldLoc, iFieldLoc ) = matrixA( iFieldLoc, iFieldLoc )    &
            & + mixture%omega_diff * 0.5_rk                          &
            & * resi_coeff( iFieldLoc, iFieldDia )                             &
            & * phi( iFieldLoc ) * moleFrac( iFieldDia ) / mixture%paramB
        end do
        ! set nonDiagonal
        do iFieldNonDia = 1,nFields
          matrixA( iFieldLoc, iFieldNonDia ) = matrixA( iFieldLoc, iFieldNonDia )&
            & - mixture%omega_diff * 0.5_rk                          &
            & * resi_coeff( iFieldLoc, iFieldNonDia )                          &
            & * phi( iFieldNonDia ) * moleFrac( iFieldLoc ) / mixture%paramB
        end do
      end do
        
      invA = invert_matrix( matrixA )

      ! momentum of all species
      do iComp = 1, 3
        momentum( iComp, : ) = matmul( invA, first_moments( iComp, : ) )
      end do
       
      !velocity of all species
      do iFieldLoc = 1,nFields
        vel( :, iFieldLoc) = momentum( :, iFieldLoc) / mass_dens(iFieldLoc)
      end do
      !mole averaged mixture velocity, w = sum(n_i*v_i)/n_T
      do iComp = 1, 3
        mixMolVel(iComp) = dot_product( num_dens, vel(iComp,:) )/tot_numDens
      end do

      ! species velocity at boundary element
      ! moleDiff_flux, J_i = n_i(v_i - w )
      ! v_i = J_i/n_i + w
!      do iComp = 1, 3
!        uxB(iComp) = moleDiff_Flux(iComp,iElem)/num_dens(iField) + mixMolVel(iComp)
!      end do
       
      ! compute equil velocity 
      do iFieldLoc = 1,nFields
        vel(:, iFieldLoc) = moleDiff_Flux(:,iElem)/num_dens(iFieldLoc) + mixMolVel
      end do

      eqVel(:) = vel(:, iField)
      do iFieldLoc = 1, nFields
        eqVel(:) = eqVel(:) + resi_coeff(iField,iFieldLoc)*phi(iField)*molefrac(iFieldLoc) &
          &      * ( vel(:, iFieldLoc) - vel(:, iField) ) / mixture%paramB
      end do

      !store velocity in input_loc
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 1 ) = eqVel(1)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 2 ) = eqVel(2)
      input_loc( globBC%nElems(iLevel) + (iElem-1)*3 + 3 ) = eqVel(3)

      ! compute mixture kinematic pressure from extrapolate density 
      ! p = cs2*(sum(phi_k*rho_k) - min_a(m_a)*n0)/rho0
      ! ASSUME n0 to be constant overtime. Neglecting small difference
      ! over time and using current tot_NumDensity
      kinePress = cs2 * ( dot_product(phi, mass_dens)                          &
        &       - minval(molWeight)*tot_NumDens0 )/mixture%rho0LB 
      ! compute current species mass density from specified molefraction at
      ! boundary
      ! rho = n_t * chi_i * m_i + massFrac_i * rho0 * p/(cs2*phi_i)
      ! 1st term is zero order density
      ! 2nd term is second order density with kinematic mixture pressure
      ! \todo KM: what about kinematic pressure here?
      !rho = tot_NumDens0 * moleFrac_bnd(iElem) * fieldProp%species%molWeight!       &
      rho = mass_dens(iField)!                                                  &
        !& + cs2inv * ( mass_dens(iField)/tot_massDens ) * mixture%rho0         &
        !& * kinePress / fieldProp%species%molWeigRatio
      
      !store in input_loc to compute equilibrium from state
      input_loc( iElem ) = rho

!write(*,*) 'totNumDens ', tot_NumDens
!write(*,*) 'totMassDens ', tot_massDens
!write(*,*) 'kinePress ', kinePress
!write(*,*) 'mass_dens ', mass_dens(iField)
!write(*,*) 'numDens ', num_dens(iField)
!write(*,*) 'rho_k ', rho 
!write(*,*) 'moleFlux ', moleDiff_Flux(:,iElem)
!write(*,*) 'MixMolVel ', mixMolVel
!write(*,*) 'velocity ', eqVel
!stop
    enddo !iElem

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Calculate the equilibrium distribution
    call varSys%derive( derVarPos%equilFromMacro )%do(                         &
      &                input = input_loc,                                      &
      &                bound = bound,                                          &
      &                tree = tree,                                            &
      &                subTree = subTree,                                      &
      &                requVar = varSys%Variable%val(derVarPos%equilFromMacro),&
      &                globSys = varSys,                                       &
      &                stencil = layout%stencil(1),                            &
      &                conf = varSys%solSpec%conf,                             & 
      &                res = fEq )

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        ! The bitmask points into the incoming direction into the flow domain, 
        !  which actually we want to update
        ! * For PUSH, we write to the incoming position, 
        !   as the kernel reads it from there without propagation.
        ! * For PULL, we need to write to the inverse direction, as the kernel 
        !   performs a bounce back before reading it. 
        !   However, this bounced back direction actually comes from the 
        !   non-existent boundary element and would point into the incoming 
        !   direction, so the value has to be treated and set as if it points 
        !   really into the incoming direction.
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions 
          ! are inverted
          state(                                                               & 
?FETCH?( varPos(iDir), globBC%elems( iLevel )%elem( iElem ), nScalars, neigh ), currT) &
            & = fEq( ?IDX?( iDir, iElem, layout%stencil(1)%QQ ) )
        endif
      enddo
      endif
    enddo

  end subroutine spc_moleDiff_Flux
!******************************************************************************!


!******************************************************************************!
  !> This routine computes mole diffusion flux of the ionic species at the 
  !! membrance using black box model and then mass density at the membrane 
  !! boundary from mole diffusion flux. Then equilibrium is set at the boundary 
  !! which is computed from mass density and velocity
  !! 
  !! Black box model: \n
  !!  - for ionic species mole diffusion flux:
  !! $J^m_k (x^m,t) = \frac{t^m_k(x^m,t)}{z_k F} i^m(x^m,t)$
  !! $J^m_k$ - mole diffusion flux of ion in the membrane 
  !! $t^m_k$ - transference number of ion of the membrane 
  !!           (selective membrane property)
  !! $z_k$ - charge number of ion
  !! $F$ - Faraday constant
  !! $i^m(x^m,t) = n^l_x_1 \cdot i^l(x,t) = - n^r_x_1 \cdot i^r(x,t)$ 
  !!    - orthogonal projection of the current density at the left $i^l$ and
  !!      right $i^r$ side of the membrance.
  !! The normal vectors $n^l_1$ and $n^r_1$ point into the membrane surface
  !! $i(x,t) = F \sum_{j=1}^{n} z_j J^e_j(x,t)$ 
  !!    - local current density [A/m^3]
  !! $J^e_k = c^e_k(v^e_k - v) $ - mole diffusion flux of ion in the electrolyte
  !! In black box model: mixture averaged velocity at the membrane is assumed to
  !! be zero
?? copy :: bc_routineHeader( spc_blackbox_mem_ion )
    !------------------------------------------------------------------------ 
!    real(kind=rk) :: fEq( layout%stencil(1)%QQ*globBC%nElems(iLevel) )
    !store pdf of all fields for derive function pointer input
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
    real(kind=rk) :: mass_dens( varSys%nVars ), num_dens( varSys%nVars )
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
    real(kind=rk) :: uxB(3*globBC%nElems(iLevel))   !< Velocity on boundary element
    real(kind=rk) :: molWeight(varSys%nVars), chargeNr(varSys%nVars)
    integer :: bound(2)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    integer :: offVel, posVel
    real(kind=rk), dimension(3, varSys%nVars):: molar_flux, velBnd
    real(kind=rk) :: loc_curr_dens(3)
    integer :: BndNormalDir
    real(kind=rk) :: mem_curr_dens, mem_molar_flux, velAvg(3)
    real(kind=rk) :: molefrac(3), tot_numDens, mem_mass_flux, elec_mass_flux(3)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    !------------------------------------------------------------------------ 
!write(dbgUnit(1),*) 'Boundary label ', trim(me%label)
!write(dbgUnit(1),*) 'iField ', iField

    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      chargeNr(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,      &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'charge_nr',                                    &
        &           fieldProp = 'species' )                           
    end do    

    do iElem = 1, globBC%nElems(iLevel)
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp( ?IDX?( pos, iElem, nScalars ) ) = bcBuffer(                    &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  
    end do !iElem

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Get local equilibrium velocity
    offVel = globBC%nElems(iLevel)*nFields
    do iFieldLoc = 1, nFields
      call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
        &                       input = fTmp,                                  &
        &                       bound = bound,                                 &
        &                       tree = tree,                                   &
        &                       subTree = subTree,                             &
        &                       requVar = varSys%Variable%val(                 &
        &                                 derVarPosAll(iFieldLoc)%velocity ),  &
        &                       globSys = varSys, stencil = layout%stencil(1), &
        &                       res = uxB,                                     &
        &                       conf = varSys%solSpec%conf )
      do iElem = 1, globBC%nElems(iLevel)
        posVel = offVel + (iElem-1)*nFields*3
        input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = uxB((iElem-1)*3+1)
        input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uxB((iElem-1)*3+2)
        input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uxB((iElem-1)*3+3)
      end do    
    end do  

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      posVel = offVel + (iElem-1)*nFields*3
      ! velocity at boundary
      do iFieldLoc = 1, nFields
        velBnd(1, iFieldLoc) = input_loc( posVel + (iFieldLoc-1)*3 + 1 )
        velBnd(2, iFieldLoc) = input_loc( posVel + (iFieldLoc-1)*3 + 2 )
        velBnd(3, iFieldLoc) = input_loc( posVel + (iFieldLoc-1)*3 + 3 )
      end do
      ! Calculate the density of current element
      mass_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) +                        &
            & fTmp( ?IDX?( pos, iElem, nScalars ) )
        end do
        num_dens(iFieldLoc) = mass_dens(iFieldLoc) / molWeight(iFieldLoc)
      end do !iField  

      tot_numDens = sum(num_dens)
      molefrac = num_dens/tot_numDens
      ! assuming mass averaged velocity at membrance is zero
      velAvg = 0.0_rk
      do iFieldLoc = 1, nFields
        velAvg(1) = velAvg(1) + mass_dens(iFieldLoc) * velBnd(1, iFieldLoc)
        velAvg(2) = velAvg(2) + mass_dens(iFieldLoc) * velBnd(2, iFieldLoc)
        velAvg(3) = velAvg(3) + mass_dens(iFieldLoc) * velBnd(3, iFieldLoc)
      end do
      velAvg = velAvg/sum(mass_dens)

      ! mole average velocity
      !do iFieldLoc = 1, nFields
      !  velAvg(1) = velAvg(1) + molefrac(iFieldLoc) * velBndX(iFieldLoc)
      !  velAvg(2) = velAvg(2) + molefrac(iFieldLoc) * velBndY(iFieldLoc)
      !  velAvg(3) = velAvg(3) + molefrac(iFieldLoc) * velBndZ(iFieldLoc)
      !end do
      ! molar diffusive flux = molar flux: Assuming mixture velocity is zero
      ! at membrane surface
      ! \todo mcheck for flux equality
      do iFieldLoc = 1, nFields
        molar_flux(:, iFieldLoc) = num_dens(iFieldLoc)                           &
          & * (velBnd(:, iFieldLoc))! - velAvg(:) )
      end do

      ! local current density
      loc_curr_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        loc_curr_dens(:) = loc_curr_dens(:) + chargeNr(iFieldLoc)                  &
          &              * molar_flux(:, iFieldLoc)
      end do

      loc_curr_dens = loc_curr_dens * mixture%faradayLB

      ! normal direction point to boundary
      bndNormalDir = layout%stencil(1)%cxDirInv( globBC%elems( iLevel )%       &
        &                                                   normalInd( iElem ))
      ! project local current density on the membrane
      mem_curr_dens = dot_product(abs(globBC%elems(iLevel)%normal(:, iElem)), &
        &                             loc_curr_dens)
      !mem_curr_dens = (layout%stencil(1)%cxDir(1, bndNormalDir )*loc_curr_dens(1) &
      !  &           + layout%stencil(1)%cxDir(2, bndNormalDir )*loc_curr_dens(2) &
      !  &           + layout%stencil(1)%cxDir(3, bndNormalDir )*loc_curr_dens(3))

      ! membrance molar flux of ionic species
      mem_molar_flux = me%blackbox_mem%transNr * mem_curr_dens                 &
        &            / ( chargeNr(iField) * mixture%faradayLB )

      ! membrance mass flux
      mem_mass_flux = mem_molar_flux*molWeight(iField)
      ! set normal direction of membrance flux
      elec_mass_flux = 0.0_rk 
      if( abs(globBC%elems(iLevel)%normal(1, iElem)) == 1 )                &
        & elec_mass_flux(1) = mem_mass_flux
      if( abs(globBC%elems(iLevel)%normal(2, iElem)) == 1 )                &
        & elec_mass_flux(2) = mem_mass_flux
      if( abs(globBC%elems(iLevel)%normal(3, iElem)) == 1 )                &
        & elec_mass_flux(3) = mem_mass_flux
           
!write(dbgUnit(1),*) 'transNr ', me%blackbox_mem%transNr
!write(dbgUnit(1),*) 'moleFlux ', molar_flux(:, iField)
!write(dbgUnit(1),*) 'moleFluxAll ', molar_flux(:, :)
!write(dbgUnit(1),*) 'loc_curr_dens ', loc_curr_dens
!write(dbgUnit(1),*) 'mem_curr_dens ', mem_curr_dens
!write(dbgUnit(1),*) 'mem_mass_flux ', mem_mass_flux
!write(dbgUnit(1),*) 'elec_mass_flux ', elec_mass_flux
      !if( abs(layout%stencil(1)%cxDir( 1, bndNormalDir )) == 1)                &
      !  &         elec_mass_flux(1)  = me%blackbox_mem%flux_dir(1)*mem_mass_flux 
      !if( abs(layout%stencil(1)%cxDir( 2, bndNormalDir )) == 1)                &
      !  &         elec_mass_flux(2)  = me%blackbox_mem%flux_dir(2)*mem_mass_flux 
      !if( abs(layout%stencil(1)%cxDir( 3, bndNormalDir )) == 1)                &
      !  &         elec_mass_flux(3)  = me%blackbox_mem%flux_dir(3)*mem_mass_flux 

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions
          ! are inverted
          state(                                                               &
& ?FETCH?(varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh), currT) = &
          ! We need to get post-collision pdf in direction
          ! alpha- outgoing direction, which is the inverse direction of bitmask
          ! For PULL this means, get the outgoing one, as this is the one which
          ! will be bounced back
          ! For PUSH this means, get the already bounced back pdf back, so take
          ! the incoming
            & fTmp(                                                            &
& ?IDX?(varPos(layout%stencil(1)%cxDirInv(iDir)), iElem, nScalars))            &
            &       + layout%weight( iDir )*2._rk*cs2inv                       &
            &       * ( layout%stencil(1)%cxDir( 1, iDir )*elec_mass_flux(1)   &
            &       +   layout%stencil(1)%cxDir( 2, iDir )*elec_mass_flux(2)   &
            &       +   layout%stencil(1)%cxDir( 3, iDir )*elec_mass_flux(3) )
        endif
      enddo
      endif
    enddo !iElem

  end subroutine spc_blackbox_mem_ion
!******************************************************************************!


!******************************************************************************!
  !> This routine computes mole diffusion flux of the solvent at the membrance 
  !! using black box model and then mass density at the membrane boundary 
  !! from mole diffusion flux. Then equilibrium is set at the boundary which is 
  !! computed from mass density and velocity
  !! \todo KM: Not completeed. Requires ionic strength at left and right side of 
  !! membrane which is not available at the moment
  !! 
  !! Black box model: \n
  !!  - for solvent mole diffusion flux:
  !! $J^m_s (x^m,t) = \frac{t^m_s(x^m,t)}{F} i^m(x^m,t) 
  !!                + LW(IS^l(x^m,t)-IS^r(x^m,t))$
  !!
  !! $J^m_s$ - mole diffusion flux of solvent through the membrane 
  !! $t^m_k$ - transference number of solvent on the membrane 
  !!           (selective membrane property)
  !! $F$ - Faraday constant
  !! $LW$ - osmotic solvent transport coefficient 
  !! $IS^l$ and $IS^r$ are ionic strength of electrolyte solution at the left
  !! and right side of membrance
  !! $IS^l(x^m,t)=\frac{1}{2}\sum^n_{k=1} z^2_k c^m_k(x^m,t)$ 
  !! $c^m_k = \rho^m_k/m_k$ - molar concentration
  !! $i^m(x^m,t) = n^l_x_1 \cdot i^l(x,t) = - n^r_x_1 \cdot i^r(x,t)$ 
  !!    - orthogonal projection of the current density at the left $i^l$ and
  !!      right $i^r$ side of the membrance.
  !! The normal vectors $n^l_1$ and $n^r_1$ point into the membrane surface
  !! $i(x,t) = F \sum_{j=1}^{n} z_j J^e_j(x,t)$ 
  !!    - local current density [A/m^3]
  !! $J^e_k = c^e_k(v^e_k - v) $ - mole diffusion flux of ion in the electrolyte
  !! In black box model: mixture averaged velocity at the membrane is assumed to
  !! be zero
?? copy :: bc_routineHeader( spc_blackbox_mem_solvent )
    !------------------------------------------------------------------------ 
    !store pdf of all fields for derive function pointer input
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ*globBC%nElems(iLevel) &
      &              * varSys%nVars )
    real(kind=rk) :: mass_dens( varSys%nVars ), num_dens( varSys%nVars )
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: input_loc(4*globBC%nElems(iLevel)*varSys%nVars)
    real(kind=rk) :: uxB(3*globBC%nElems(iLevel))   !< Velocity on boundary element
    real(kind=rk) :: molWeight(varSys%nVars), chargeNr(varSys%nVars)
    integer :: bound(2)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    integer :: offVel, posVel
    real(kind=rk), dimension(3, varSys%nVars):: molar_flux, velBnd
    real(kind=rk) :: loc_curr_dens(3)
    integer :: BndNormalDir
    real(kind=rk) :: mem_curr_dens, mem_molar_flux, velAvg(3)
    real(kind=rk) :: molefrac(3), tot_numDens, mem_mass_flux, elec_mass_flux(3)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    !------------------------------------------------------------------------ 

    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      chargeNr(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,      &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'charge_nr',                                    &
        &           fieldProp = 'species' )                           

    end do    

    do iElem = 1, globBC%nElems(iLevel)
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp( ?IDX?( pos, iElem, nScalars ) ) = bcBuffer(                    &
            & ?IDX?( pos, globBC%elems(iLevel)%elemBuffer( iElem ), nScalars ) )
        end do
      end do !iField  
    end do !iElem

    ! initialize the lower and upper bound
    bound(1) = 1
    bound(2) = globBC%nElems( iLevel )
    ! Get local equilibrium velocity
    offVel = globBC%nElems(iLevel)*nFields
    do iFieldLoc = 1, nFields
      call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
        &                       input = fTmp,                                  &
        &                       bound = bound,                                 &
        &                       tree = tree,                                   &
        &                       subTree = subTree,                             &
        &                       requVar = varSys%Variable%val(                 &
        &                                 derVarPosAll(iFieldLoc)%velocity ),  &
        &                       globSys = varSys, stencil = layout%stencil(1), &
        &                       res = uxB,                                     &
        &                       conf = varSys%solSpec%conf )
      do iElem = 1, globBC%nElems(iLevel)
        posVel = offVel + (iElem-1)*nFields*3
        input_loc( posVel + (iFieldLoc-1)*3 + 1 ) = uxB((iElem-1)*3+1)
        input_loc( posVel + (iFieldLoc-1)*3 + 2 ) = uxB((iElem-1)*3+2)
        input_loc( posVel + (iFieldLoc-1)*3 + 3 ) = uxB((iElem-1)*3+3)
      end do    
    end do  

    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      posVel = offVel + (iElem-1)*nFields*3
      ! velocity at boundary
      do iFieldLoc = 1, nFields
        velBnd(1, iFieldLoc) = input_loc( posVel + (iFieldLoc-1)*3 + 1 )
        velBnd(2, iFieldLoc) = input_loc( posVel + (iFieldLoc-1)*3 + 2 )
        velBnd(3, iFieldLoc) = input_loc( posVel + (iFieldLoc-1)*3 + 3 )
      end do
      ! Calculate the density of current element
      mass_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc) +                        &
            & fTmp( ?IDX?( pos, iElem, nScalars ) )
        end do
        num_dens(iFieldLoc) = mass_dens(iFieldLoc) / molWeight(iFieldLoc)
      end do !iField  

      tot_numDens = sum(num_dens)
      molefrac = num_dens/tot_numDens
      ! assuming mass averaged velocity at membrance is zero
      velAvg = 0.0_rk
      do iFieldLoc = 1, nFields
        velAvg(1) = velAvg(1) + mass_dens(iFieldLoc) * velBnd(1, iFieldLoc)
        velAvg(2) = velAvg(2) + mass_dens(iFieldLoc) * velBnd(2, iFieldLoc)
        velAvg(3) = velAvg(3) + mass_dens(iFieldLoc) * velBnd(3, iFieldLoc)
      end do
      ! molar diffusive flux = molar flux 
      do iFieldLoc = 1, nFields
        molar_flux(:, iFieldLoc) = num_dens(iFieldLoc)                           &
          & * (velBnd(:, iFieldLoc))! - velAvg(:) )
      end do

      ! local current density
      loc_curr_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        loc_curr_dens(:) = loc_curr_dens(:) + chargeNr(iFieldLoc)                  &
          &              * molar_flux(:, iFieldLoc)
      end do

      loc_curr_dens = loc_curr_dens * mixture%faradayLB

      ! normal direction point to boundary
      bndNormalDir = layout%stencil(1)%cxDirInv( globBC%elems( iLevel )%       &
        &                                                   normalInd( iElem ))
      ! project local current density on the membrane
      mem_curr_dens = dot_product(abs(globBC%elems(iLevel)%normal(:, iElem)), &
        &                             loc_curr_dens)
      !mem_curr_dens = layout%stencil(1)%cxDir(1, bndNormalDir )*loc_curr_densX &
      !  &           + layout%stencil(1)%cxDir(2, bndNormalDir )*loc_curr_densY &
      !  &           + layout%stencil(1)%cxDir(3, bndNormalDir )*loc_curr_densZ

      ! membrance molar flux of ionic species
      mem_molar_flux = me%blackbox_mem%transNr * mem_curr_dens                 &
        &            / ( mixture%faradayLB )

      ! membrance mass flux
      mem_mass_flux = mem_molar_flux*molWeight(iField)
      elec_mass_flux = 0.0_rk
      ! set normal direction of membrance flux
      if( abs(globBC%elems(iLevel)%normal(1, iElem)) == 1 )                &
        & elec_mass_flux(1) = mem_mass_flux
      if( abs(globBC%elems(iLevel)%normal(2, iElem)) == 1 )                &
        & elec_mass_flux(2) = mem_mass_flux
      if( abs(globBC%elems(iLevel)%normal(3, iElem)) == 1 )                &
        & elec_mass_flux(3) = mem_mass_flux
      !if( abs(layout%stencil(1)%cxDir( 1, bndNormalDir )) == 1)                &
      !  &         elec_mass_flux(1)  = me%blackbox_mem%flux_dir(1)*mem_mass_flux 
      !if( abs(layout%stencil(1)%cxDir( 2, bndNormalDir )) == 1)                &
      !  &         elec_mass_flux(2)  = me%blackbox_mem%flux_dir(2)*mem_mass_flux 
      !if( abs(layout%stencil(1)%cxDir( 3, bndNormalDir )) == 1)                &
      !  &         elec_mass_flux(3)  = me%blackbox_mem%flux_dir(3)*mem_mass_flux 

      do iDir = 1,layout%stencil(1)%QQ
        ! Write the values
        if( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
          ! Depending on PUSH or pull, use + or - for cxDir, because directions
          ! are inverted
          state(                                                               &
& ?FETCH?(varPos(iDir), globBC%elems(iLevel)%elem(iElem), nScalars, neigh), currT) = &
          ! We need to get post-collision pdf in direction
          ! alpha- outgoing direction, which is the inverse direction of bitmask
          ! For PULL this means, get the outgoing one, as this is the one which
          ! will be bounced back
          ! For PUSH this means, get the already bounced back pdf back, so take
          ! the incoming
            & fTmp(                                                            &
& ?IDX?(varPos(layout%stencil(1)%cxDirInv(iDir)), iElem, nScalars))        &
            &       + layout%weight( iDir )*2._rk*cs2inv                       &
            &       * ( layout%stencil(1)%cxDir( 1, iDir )*elec_mass_flux(1)   &
            &       +   layout%stencil(1)%cxDir( 2, iDir )*elec_mass_flux(2)   &
            &       +   layout%stencil(1)%cxDir( 3, iDir )*elec_mass_flux(3) )
        endif
      enddo
      endif
    enddo !iElem

  end subroutine spc_blackbox_mem_solvent
!******************************************************************************!


!******************************************************************************!
  !> Moment based wall boundary condition from Sam Bennent PhD thesis
  !! "A Lattice Boltzmann Model for Diffusion of Binary Gas Mixtures"
  !! \author Kannan Masilamani
  !!
?? copy :: bc_routineHeader( spc_moments_wall )
    !---------------------------------------------------------------------------
    real(kind=rk) :: rhoB, press
    integer :: iElem, iDir, tmpDir
    integer :: nLinks, iLink, elemPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp( : )
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !---------------------------------------------------------------------------
    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

!    write(*,*) 'Boundary label ', trim(me%label)
    do iElem = 1, globBC%nElems( iLevel )

      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )
!      write(*,*) 'elemPos ', elemPos
!      write(*,*) 'treeID ', glob%levelDesc(iLevel)%total(elemPos)

      iLink = 0
      fTmp = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do  
!        write(*,*) 'fTmp0 ',  fTmp

      rhoB = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          tmpDir = layout%stencil(1)%cxDirInv(iDir)
          if( globBC%elems(iLevel)%bitmask( tmpDir, iElem )) then
            tmpDir = layout%stencil(1)%cxDirInv( &
              &            globBC%elems(iLevel)%normalInd(iElem))
          endif
        else
          tmpDir = iDir
        endif
        rhoB  = rhoB + fTmp(tmpDir)
      end do

      press = cs2*fieldProp%species%molWeigRatio*rhoB
!write(*,*) 'press ', press

      ! moments for multispecies LBM model of transformed variable
      first_moments = (/ 0.0_rk, 0.0_rk, 0.0_rk /) !momX, momY, momZ 
      second_moments = (/ press, & !momXX
        &                 press, & !momYY
        &                 press, & !momZZ
        &                 0.0_rk, & !momXY
        &                 0.0_rk, & !momYZ
        &                 0.0_rk /) !momXZ
      third_moments = 0.0_rk
      fourth_moments = cs2*press !momXXYY, momYYZZ, momZZXX 
    
      ! moments for liquid mixture
      moments = 0.0_rk
      moments(1) = rhoB
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

!     write(*,*) 'Moments ', moments

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do
     !write(*,*) 'rhs ', rhs

     unKnown_fTmp = matmul( &
       & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )

     !write(*,*) 'unknown fTmp ', unKnown_fTmp

     do iLink = 1, nLinks
       state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
        & currT)  = unKnown_fTmp(iLink)
     end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

    end do

  end subroutine spc_moments_wall
!******************************************************************************!


!******************************************************************************!
  !> Mole fraction boundary condition
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moments_moleFrac', 
  !!    moleFraction = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moments_moleFrac )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: rho, press
    real(kind=rk) :: moleFrac(globBC%nElems(iLevel))                
    integer :: iElem, iDir, iFieldLoc, nFields, pos
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk) :: resi_coeff( varSys%nVars, varSys%nVars )
    integer :: nLinks, iLink, elemPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp(:)
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: vel_inG(3), velAvg(3), velQuad(3)
    real(kind=rk) :: mom(3,varSys%nVars), spc_mom(3), mom_inG(3)
    real(kind=rk) :: mass_dens(varSys%nVars), num_dens(varSys%nVars)
    real(kind=rk) :: totMassDens, totNumDens
    real(kind=rk) :: moleFrac_loc(varSys%nVars)
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !--------------------------------------------------------------------------- 
    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,                &
      &                             nFields = nFields )
    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    end do    

!write(*,*) 'iField ', iField
!write(*,*) 'molWeight moleFrac BC', molWeight
!write(*,*) 'phi ', phi
!stop

    ! mole fraction
    moleFrac = tem_spacetime_for(                                              &
      & me = me%BC_states%moleFrac%spacetime,                                  &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                     &
      & dt_phy  = physics%dt )

    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      fTmp = 0.0_rk
      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(unKnown_fTmp(nLinks))
      allocate(rhs(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink + 1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), currT) 
        end if
      end do

      ! local state vector to compute density and velocity of other species
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer( ?IDX?( pos, globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo
      enddo

      !KM: using inital mixture number density rho0/mixtureMOlWeight
      rho = mixture%moleDens0LB *  moleFrac(iElem) &
        & * fieldProp%species%molWeight
      press = cs2 * rho * fieldProp%species%molWeigRatio

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      !local density and momentum
      mass_dens = 0.0_rk
      mom = 0.0_rk
      do iFieldLoc = 1, nFields
        do iDir = 1,layout%stencil(1)%QQ
          mass_dens(iFieldLoc) = mass_dens(iFieldLoc)               &
            & + fTmp_all(varSys%Variable%val(iFieldLoc)%varPos(iDir)) 
        end do
        num_dens(iFieldLoc) = mass_dens(iFieldLoc) &
          &                 / molWeight(iFieldLoc)
        !mass flux
        call varSys%derive( derVarPosAll(iFieldLoc)%momentum )%do(              &
          &                       input = fTmp_all,                             &
          &                       bound = bound,                                &
          &                       tree = tree,                                  &
          &                       subTree = subTree,                            &
          &                       requVar = varSys%Variable%val(                &
          &                                 derVarPosAll(iFieldLoc)%momentum ), &
          &                       globSys = varSys, stencil = layout%stencil(1),&
          &                       res = spc_mom,                                &
          &                       conf = varSys%solSpec%conf )
        mom(:, iFieldLoc) = spc_mom
      end do  

      mass_dens(iField) = rho
      num_dens(iField) = rho / fieldProp%species%molWeight
 
      ! total mass density
      totMassDens = sum(mass_dens)

      ! mole fraction
      moleFrac_loc = num_dens/sum(num_dens)

!do iFieldLoc = 1, nFields
!  write(dbgUnit(1),*) iFieldLoc, 'vel ', mom(:, iFieldLoc)/mass_dens(iFieldLoc), &
!    & 'dens ', mass_dens(iFieldLoc), 'moleFrac ', moleFrac(iFieldLoc)
!enddo  

      ! compute velocity of transformed variable g. Eq. which is normally
      ! used to compute actual velocity by solving LSE
      mom_inG = mom(:, iField)
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG + ( mixture%omega_diff * 0.5_rk                    &
            &               * resi_coeff(iField, iFieldLoc) * phi(iField)    &
            &               * moleFrac_loc(iFieldLoc) / mixture%paramB )     &
            &               * mom(:, iField)
      end do
      !set non-diagonal part
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG - ( mixture%omega_diff * 0.5_rk                    &
          &                 * resi_coeff(iField, iFieldLoc) * phi(iFieldLoc) &
          &                 * moleFrac_loc(iField) / mixture%paramB )        &
          &                 * mom(:,iFieldLoc)
      end do

      vel_inG = mom_inG/rho
!write(dbgUnit(1),*) 'transformed momentum ', mom_inG
!write(dbgUnit(1),*) 'transformed velocity ', vel_inG

      ! mass averaged mixture velocity
      velAvg(1) = sum(mom(1,:)) / totMassDens
      velAvg(2) = sum(mom(2,:)) / totMassDens
      velAvg(3) = sum(mom(3,:)) / totMassDens

!      write(*,*) 'velAvg ', velAvg
      velQuad(:) = mixture%theta_eq*velAvg(:) &
        &        + (1.0_rk - mixture%theta_eq)*vel_inG

      ! moments for multispecies LBM model of transformed variable
      first_moments = mom_inG !momX, momY, momZ 
      second_moments = (/ press + rho*velQuad(1)*velQuad(1), & !momXX
        &                 press + rho*velQuad(2)*velQuad(2), & !momYY
        &                 press + rho*velQuad(3)*velQuad(3), & !momZZ
        &                 rho*velQuad(1)*velQuad(2),         & !momXY
        &                 rho*velQuad(2)*velQuad(3),         & !momYZ
        &                 rho*velQuad(3)*velQuad(1) /)         !momXZ
      third_moments = (/ cs2*mom_inG(2), & !momXXY
        &                cs2*mom_inG(3), & !momXXZ
        &                cs2*mom_inG(1), & !momYYX
        &                cs2*mom_inG(3), & !momYYZ
        &                cs2*mom_inG(1), & !momZZX
        &                cs2*mom_inG(2) /) !momZZY
      fourth_moments = (/ &
        & cs2*(press+rho*(        velQuad(1)**2 +        velQuad(2)**2 &
        &                - 0.5_rk*velQuad(3)**2 )), &!momXXYY
        & cs2*(press+rho*(-0.5_rk*velQuad(1)**2 +        velQuad(2)**2 &
        &                + velQuad(3)**2 )), &!momYYZZ
        & cs2*(press+rho*(        velQuad(1)**2 - 0.5_rk*velQuad(2)**2 &
        &                + velQuad(3)**2 ))  &!momZZXX
        & /) 
     
      ! moments for liquid mixture
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      ! compute rhs of lse to compute unknown pdfs
      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do

      ! unknown pdfs
      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
        & currT)  = unKnown_fTmp(iLink)
      end do

     deallocate(missing_links)
     deallocate(unKnown_fTmp)
     deallocate(rhs)
   end do

  end subroutine spc_moments_moleFrac
!******************************************************************************!


!******************************************************************************!
  !> molar flux boundary condition like moments velocity bc type
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moments_moleflux', 
  !!    moleflux = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moments_moleFlux )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: moleFluxX(globBC%nElems(iLevel)),                         &
      &              moleFluxY(globBC%nElems(iLevel)),                         &
      &              moleFluxZ(globBC%nElems(iLevel))
    integer :: iElem, iDir, iFieldLoc, nFields, pos, tmpDir
    real(kind=rk) :: rho, press, dens_correction
    integer :: nLinks, iLink, elemPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp(:)
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: massFlux(3), vel_inG(3)
    real(kind=rk) :: mom(3,varSys%nVars), spc_mom(3), mom_inG(3)
    real(kind=rk) :: mass_dens(varSys%nVars), num_dens(varSys%nVars)
    real(kind=rk) :: spc_num_dens(1), totMassDens, velAvg(3), velQuad(3)
    real(kind=rk) :: moleFrac(varSys%nVars)
    integer :: nField
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk) :: resi_coeff( varSys%nVars, varSys%nVars )
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    real(kind=rk) :: rhoTmp(1), uxTmp(3)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !------------------------------------------------------------------------ 
!write(dbgUnit(1),*) 'Boundary label ', trim(me%label)
!write(dbgUnit(1),*) 'iField ', iField

    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    end do    

    ! Get velocity which might be defined to vary in space and time
    moleFluxX = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(1)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleFluxY = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(2)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    moleFluxZ = tem_spacetime_for(                                             &
      & me = me%BC_states%moleFlux(3)%spacetime,                               &
      & treeIDs = glob%levelDesc( iLevel )%total(                              &
      &                                        globBC%elems(iLevel)%elem(:) ), &
      & tree = tree,                                                           &
      & time = sim_time,                                                       &
      & n = globBC%nElems(iLevel),                                             &
      & conf = params%general%solver%conf(1),                                  &
      & dt_phy  = physics%dt )

    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    moleFluxX = moleFluxX/physics%fac( iLevel )%moleFlux
    moleFluxY = moleFluxY/physics%fac( iLevel )%moleFlux
    moleFluxZ = moleFluxZ/physics%fac( iLevel )%moleFlux

!write(*,*) 'physical moleFlux ', physics%fac( iLevel )%moleFlux
!write(*,*) 'moleFluxX ', moleFluxX
!write(*,*) 'moleFluxY ', moleFluxY
!write(*,*) 'moleFluxZ ', moleFluxZ
!stop
   
    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)
    
!    write(*,*) 'iField ', iField
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
!write(dbgUnit(1),*) 'iElem ', iElem  

      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      fTmp = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do 
      !write(*,*) 'fTmp ', fTmp

      rho = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          tmpDir = layout%stencil(1)%cxDirInv(iDir)
          if( globBC%elems(iLevel)%bitmask( tmpDir, iElem )) then
            tmpDir = layout%stencil(1)%cxDirInv(globBC%elems(iLevel)%normalInd(iElem))
          endif
        else
          tmpDir = iDir
        endif
        rho  = rho + fTmp(tmpDir)
      end do
!write(dbgUnit(1),*) 'density ', rho

      massFlux(1) = moleFluxX(iElem) * fieldProp%species%molWeight
      massFlux(2) = moleFluxY(iElem) * fieldProp%species%molWeight
      massFlux(3) = moleFluxZ(iElem) * fieldProp%species%molWeight
!write(dbgUnit(1),*) 'massFlux ', massFlux

      ! add small correction term to density which comes from momentum in the
      ! flow direction while substituing unknowns pdfs terms obtained
      ! from moments BC derivation to compute density.
      ! With this correction term density is recovered correctly at this node
      dens_correction = dot_product(massFlux, globBC%elems(iLevel)%normal(:, iElem))
      ! correction term for Multispecies when flux is defined 
      ! rho_new = rho_old + dens_correction
      rho = rho + dens_correction
      !write(*,*) 'density after correction', rho

      !pressure
      press = cs2*rho*fieldProp%species%molWeigRatio
!write(dbgUnit(1),*) 'rho ', rho, 'press', press

      ! local state vector to compute density and velocity of other species
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer( ?IDX?( pos, globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo
      enddo

      !local density and momentum
      mass_dens = 0.0_rk
      mom = 0.0_rk
      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      do iFieldLoc = 1, nFields
        if(iField == iFieldLoc) then
          ! massflux = moleflux * molecular_weight
          mom(:, iField) = massFlux
          mass_dens(iField) = rho
          num_dens(iField) = rho / fieldProp%species%molWeight
        else
          do iDir = 1,layout%stencil(1)%QQ
           ! mass density
            mass_dens(iFieldLoc) = mass_dens(iFieldLoc)               &
              & + fTmp_all(varSys%Variable%val(iFieldLoc)%varPos(iDir)) 
          end do
          num_dens(iFieldLoc) = mass_dens(iFieldLoc) &
            &                 / molWeight(iFieldLoc)
          !mass flux
          call varSys%derive( derVarPosAll(iFieldLoc)%momentum )%do(              &
            &                       input = fTmp_all,                             &
            &                       bound = bound,                                &
            &                       tree = tree,                                  &
            &                       subTree = subTree,                            &
            &                       requVar = varSys%Variable%val(                &
            &                                 derVarPosAll(iFieldLoc)%momentum ), &
            &                       globSys = varSys, stencil = layout%stencil(1),&
            &                       res = spc_mom,                                &
            &                       conf = varSys%solSpec%conf )

          mom(:, iFieldLoc) = spc_mom
        endif
      enddo  

      ! mass density
      mass_dens = num_dens * molWeight
      ! total mass density
      totMassDens = sum(mass_dens)

      ! mole fraction
      moleFrac = num_dens/sum(num_dens)

!do iFieldLoc = 1, nFields
!  write(dbgUnit(1),*) iFieldLoc, 'vel ', mom(:, iFieldLoc)/mass_dens(iFieldLoc), &
!    & 'dens ', mass_dens(iFieldLoc), 'moleFrac ', moleFrac(iFieldLoc)
!enddo  
      
      ! compute velocity of transformed variable g. Eq. which is normally
      ! used to compute actual velocity by solving LSE
      mom_inG = mom(:, iField)
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG + ( mixture%omega_diff * 0.5_rk                    &
            &               * resi_coeff(iField, iFieldLoc) * phi(iField)    &
            &               * moleFrac(iFieldLoc) / mixture%paramB )         &
            &               * mom(:, iField)
      end do
      !set non-diagonal part
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG - ( mixture%omega_diff * 0.5_rk                    &
          &                 * resi_coeff(iField, iFieldLoc) * phi(iFieldLoc) &
          &                 * moleFrac(iField) / mixture%paramB )            &
          &                 * mom(:,iFieldLoc)
      end do

      vel_inG = mom_inG/rho
!write(dbgUnit(1),*) 'defined momentum ', massFlux      
!write(dbgUnit(1),*) 'transformed momentum ', mom_inG
!write(dbgUnit(1),*) 'transformed velocity ', vel_inG

      ! mass averaged mixture velocity
      velAvg(1) = sum(mom(1,:)) / totMassDens
      velAvg(2) = sum(mom(2,:)) / totMassDens
      velAvg(3) = sum(mom(3,:)) / totMassDens
!      write(*,*) 'velAvg ', velAvg
      velQuad(:) = mixture%theta_eq*velAvg(:) &
        &        + (1.0_rk - mixture%theta_eq)*vel_inG

      ! moments for multispecies LBM model of transformed variable
      first_moments = mom_inG !momX, momY, momZ 
      second_moments = (/ press + rho*velQuad(1)*velQuad(1), & !momXX
        &                 press + rho*velQuad(2)*velQuad(2), & !momYY
        &                 press + rho*velQuad(3)*velQuad(3), & !momZZ
        &                 rho*velQuad(1)*velQuad(2),         & !momXY
        &                 rho*velQuad(2)*velQuad(3),         & !momYZ
        &                 rho*velQuad(3)*velQuad(1) /)         !momXZ
      third_moments = (/ cs2*mom_inG(2), & !momXXY
        &                cs2*mom_inG(3), & !momXXZ
        &                cs2*mom_inG(1), & !momYYX
        &                cs2*mom_inG(3), & !momYYZ
        &                cs2*mom_inG(1), & !momZZX
        &                cs2*mom_inG(2) /) !momZZY
      fourth_moments = (/ &
        & cs2*(press+rho*(        velQuad(1)**2 +        velQuad(2)**2 &
        &                - 0.5_rk*velQuad(3)**2 )), &!momXXYY
        & cs2*(press+rho*(-0.5_rk*velQuad(1)**2 +        velQuad(2)**2 &
        &                + velQuad(3)**2 )), &!momYYZZ
        & cs2*(press+rho*(        velQuad(1)**2 - 0.5_rk*velQuad(2)**2 &
        &                + velQuad(3)**2 ))  &!momZZXX
        & /) 
     
      ! moments for liquid mixture
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      !write(*,*) 'Moments ', moments
      !write(*,*) 'knownMomPos ', me%elems(iLevel)%moments(iElem)%knownMom_pos

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do
      !write(*,*) 'rhs ', rhs

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )
      !write(*,*) 'unknown fTmp ', unKnown_fTmp

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

      endif
    end do

!!    if (iField == nFields) then
!!    do iElem = 1, globBC%nElems( iLevel )
!!      !ux = 0.0_rk
!!      do iFieldLoc = 1, nFields
!!        do iDir = 1,layout%stencil(1)%QQ
!!          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
!!          fTmp_all(pos) = state(                                                  &
!!&?FETCH?( pos, globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
!!        & currT)
!!        end do
!!      end do  
!!      bound(1)=iElem
!!      bound(2)=iElem
!!!write(dbgUnit(1),*) 'iElem ', iElem
!!!write(dbgUnit(1),*) 'fTmp_all ', fTmp_all
!!      ! density
!!      do iFieldLoc = 1, nFields
!!        call varSys%derive( derVarPosAll(iFieldLoc)%density )%do(        &
!!          &                 input = fTmp_all,                               &
!!          &                 bound = bound,                                  &
!!          &                 tree = tree,                                    &
!!          &                 subTree = subTree,                              &
!!          &                 requVar = varSys%Variable%val(                  &
!!          &                           derVarPosAll(iFieldLoc)%density ), &
!!          &                 globSys = varSys, stencil = layout%stencil(1),  &
!!          &                 res = rhoTmp,                             &
!!          &                 conf = varSys%solSpec%conf )
!!
!!        call varSys%derive( derVarPosAll(iFieldLoc)%velocity )%do(               &
!!          &                 input   = fTmp_all,                                  &
!!          &                 bound   = bound,                                     &
!!          &                 tree    = tree,                                      &
!!          &                 subTree = subTree,                                   &
!!          &                 requVar = varSys%Variable%val(&
!!          &                           derVarPosAll(iFieldLoc)%velocity),   &
!!          &                 globSys = varSys,                                    &
!!          &                 stencil = layout%stencil(1),                         &
!!          &                 conf    = varSys%solSpec%conf,                       &
!!          &                 res     = uxTmp )
!!!write(dbgUnit(1),*) 'iField ', iFieldLoc, 'density ', rhoTmp
!!!write(dbgUnit(1),*) 'iField ', iFieldLoc, 'velocity ', uxTmp
!!      end do  
!!    end do
!!    endif
!
!  stop

  end subroutine spc_moments_moleFlux
!******************************************************************************!


!******************************************************************************!
  !> velocity boundary condition like moments velocity bc type
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moments_vel', 
  !!    moleflux = 0.0
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moments_vel )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: ux(globBC%nElems(iLevel)), uy(globBC%nElems(iLevel)),     &
      &              uz(globBC%nElems(iLevel))
    integer :: iElem, iDir, iFieldLoc, nFields, pos, tmpDir
    real(kind=rk) :: rho, press, dens_correction
    integer :: nLinks, iLink, elemPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp(:)
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: vel(3), vel_inG(3)
    real(kind=rk) :: mom(3,varSys%nVars), spc_mom(3), mom_inG(3)
    real(kind=rk) :: mass_dens(varSys%nVars), num_dens(varSys%nVars)
    real(kind=rk) :: velAvg(3), velQuad(3)
    real(kind=rk) :: totMassDens, totNumDens
    real(kind=rk) :: moleFrac(varSys%nVars)
    integer :: nField
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk) :: resi_coeff( varSys%nVars, varSys%nVars )
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars)
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !------------------------------------------------------------------------ 
    !write(*,*) 'Boundary label ', trim(me%label)

    !write(*,*) 'iField ', iField

    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
    end do    

    ! Get velocity which might be defined to vary in space and time
     ux = tem_spacetime_for( me      = me%bc_states%velocity(1)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uy = tem_spacetime_for( me      = me%bc_states%velocity(2)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )

    uz = tem_spacetime_for( me      = me%bc_states%velocity(3)%spacetime,      &
      &                     treeIDs = glob%levelDesc( iLevel )%total( globBC%  &
      &                                               elems(iLevel)%elem(:) ), &
      &                     tree    = tree,                                    &
      &                     time    = sim_time,                                &
      &                     n       = globBC%nElems(iLevel),                   &
      &                     conf    = params%general%solver%conf(1),           &
      &                     dt_phy  = physics%dt )
    ! If physical quantities are given, transform to lattice units by division
    ! with the conversion factor
    !write(*,*) 'phy Vel', ux, uy, uz
    ux = ux/physics%fac( iLevel )%vel
    uy = uy/physics%fac( iLevel )%vel
    uz = uz/physics%fac( iLevel )%vel
   
    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)

    !write(*,*) 'iField ', iField
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
      !write(*,*) 'iElem ', iElem  
      !write(*,*) 'vel ', ux(iElem), uy(iElem), uz(iElem)
      vel = (/ ux(iElem), uy(iElem), uz(iElem) /)

      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      fTmp = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do 

      rho = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          tmpDir = layout%stencil(1)%cxDirInv(iDir)
          if( globBC%elems(iLevel)%bitmask( tmpDir, iElem )) then
            tmpDir = layout%stencil(1)%cxDirInv(globBC%elems(iLevel)%normalInd(iElem))
          endif
        else
          tmpDir = iDir
        endif
        rho  = rho + fTmp(tmpDir)
      end do
!write(dbgUnit(1),*) 'density ', rho

      ! add small correction term to density which comes from momentum in the
      ! flow direction while substituing unknowns pdfs terms obtained
      ! from moments BC derivation to compute density.
      ! With this correction term density is recovered correctly at this node
      dens_correction = dot_product(vel, globBC%elems(iLevel)%normal(:, iElem))
      ! correction term for Multispecies when vel is defined
      ! rho_new = rho_old/(1.0 - dens_correction)
      rho = rho/(1.0_rk-dens_correction)

      !write(*,*) 'density after correction', rho
      !pressure
      press = cs2*rho*fieldProp%species%molWeigRatio

      ! local state vector to compute density and velocity of other species
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer( ?IDX?( pos, globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo
      enddo

      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      !local density and momentum
      mass_dens = 0.0_rk
      mom = 0.0_rk
      do iFieldLoc = 1, nFields
        if(iField == iFieldLoc) then
          ! massflux = moleflux * molecular_weight
          mom(:, iField) = rho * vel
          mass_dens(iField) = rho
          num_dens(iField) = rho / fieldProp%species%molWeight
        else
          do iDir = 1,layout%stencil(1)%QQ
            mass_dens(iFieldLoc) = mass_dens(iFieldLoc)               &
              & + fTmp_all(varSys%Variable%val(iFieldLoc)%varPos(iDir)) 
          end do
          num_dens(iFieldLoc) = mass_dens(iFieldLoc) &
            &                 / molWeight(iFieldLoc)

          ! mass flux
          call varSys%derive( derVarPosAll(iFieldLoc)%momentum )%do(        &
            &                 input = fTmp_all,                             &
            &                 bound = bound,                                &
            &                 tree = tree,                                  &
            &                 subTree = subTree,                            &
            &                 requVar = varSys%Variable%val(                &
            &                           derVarPosAll(iFieldLoc)%momentum ), &
            &                 globSys = varSys, stencil = layout%stencil(1),&
            &                 res = spc_mom,                                &
            &                 conf = varSys%solSpec%conf )
          mom(:, iFieldLoc) = spc_mom
        end if
      end do  

      ! total mass density
      totMassDens = sum(mass_dens)

      ! mole fraction
      moleFrac = num_dens/sum(num_dens)

!do iFieldLoc = 1, nFields
!  write(dbgUnit(1),*) iFieldLoc, 'vel ', mom(:, iFieldLoc)/mass_dens(iFieldLoc), &
!    & 'dens ', mass_dens(iFieldLoc), 'moleFrac ', moleFrac(iFieldLoc)
!enddo  

      ! compute velocity of transformed variable g. Eq. which is normally
      ! used to compute actual velocity by solving LSE
      mom_inG = mom(:, iField)
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG + ( mixture%omega_diff * 0.5_rk                    &
            &               * resi_coeff(iField, iFieldLoc) * phi(iField)    &
            &               * moleFrac(iFieldLoc) / mixture%paramB )         &
            &             * mom(:, iField)
      end do
      !set non-diagonal part
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG - ( mixture%omega_diff * 0.5_rk                    &
          &                 * resi_coeff(iField, iFieldLoc) * phi(iFieldLoc) &
          &                 * moleFrac(iField) / mixture%paramB )            &
          &               * mom(:,iFieldLoc)
      end do

      vel_inG = mom_inG/rho
!write(dbgUnit(1),*) 'transformed momentum ', mom_inG
!write(dbgUnit(1),*) 'transformed velocity ', vel_inG

      ! mass averaged mixture velocity
      velAvg(1) = sum(mom(1,:)) / totMassDens
      velAvg(2) = sum(mom(2,:)) / totMassDens
      velAvg(3) = sum(mom(3,:)) / totMassDens

      velQuad(:) = mixture%theta_eq*velAvg(:) &
        &        + (1.0_rk - mixture%theta_eq)*vel_inG
!      write(*,*) 'velAvg ', velAvg

      ! moments for multispecies LBM model of transformed variable
      first_moments = mom_inG !momX, momY, momZ 
      second_moments = (/ press + rho*velQuad(1)*velQuad(1), & !momXX
        &                 press + rho*velQuad(2)*velQuad(2), & !momYY
        &                 press + rho*velQuad(3)*velQuad(3), & !momZZ
        &                 rho*velQuad(1)*velQuad(2),         & !momXY
        &                 rho*velQuad(2)*velQuad(3),         & !momYZ
        &                 rho*velQuad(3)*velQuad(1) /)         !momXZ
      third_moments = (/ cs2*mom_inG(2), & !momXXY
        &                cs2*mom_inG(3), & !momXXZ
        &                cs2*mom_inG(1), & !momYYX
        &                cs2*mom_inG(3), & !momYYZ
        &                cs2*mom_inG(1), & !momZZX
        &                cs2*mom_inG(2) /) !momZZY
      fourth_moments = (/ &
        & cs2*(press+rho*(        velQuad(1)**2 +        velQuad(2)**2 &
        &                - 0.5_rk*velQuad(3)**2 )), &!momXXYY
        & cs2*(press+rho*(-0.5_rk*velQuad(1)**2 +        velQuad(2)**2 &
        &                + velQuad(3)**2 )), &!momYYZZ
        & cs2*(press+rho*(        velQuad(1)**2 - 0.5_rk*velQuad(2)**2 &
        &                + velQuad(3)**2 ))  &!momZZXX
        & /) 
     
      ! moments for liquid mixture
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      !write(*,*) 'Moments ', moments
      !write(*,*) 'knownMomPos ', me%elems(iLevel)%moments(iElem)%knownMom_pos

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do
      !write(*,*) 'rhs ', rhs

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )
      !write(*,*) 'unknown fTmp ', unKnown_fTmp

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

      endif
    end do
    !stop

  end subroutine spc_moments_vel
!******************************************************************************!

!******************************************************************************!
  !> Black box membrane molar flux boundary condition for ionic species 
  !! like moments velocity bc type
  !! Usage 
  !! -----
  !!\verbatim
  !!boundary_condition = {  
  !! { label = 'outlet', 
  !!    kind = 'spc_moments_blackbox_mem_ion', 
  !!    transference_number = 0.9
  !!     }
  !!\endverbatim
?? copy :: bc_routineHeader( spc_moments_blackbox_mem_ion )
    !------------------------------------------------------------------------ 
    real(kind=rk) :: moleFluxX(globBC%nElems(iLevel)),                         &
      &              moleFluxY(globBC%nElems(iLevel)),                         &
      &              moleFluxZ(globBC%nElems(iLevel))
    integer :: iElem, iDir, iFieldLoc, nFields, pos, tmpDir
    real(kind=rk) :: rho, press, dens_correction
    integer :: nLinks, iLink, elemPos
    integer, allocatable :: missing_links(:)
    real(kind=rk), allocatable :: rhs(:)
    real(kind=rk), allocatable :: unKnown_fTmp(:)
    real(kind=rk) :: fTmp( layout%stencil(1)%QQ )
    real(kind=rk) :: fTmp_all( layout%stencil(1)%QQ*varSys%nVars )
    real(kind=rk) :: moments( layout%stencil(1)%QQ )
    real(kind=rk) :: massFlux(3), vel_inG(3)
    real(kind=rk) :: mom(3,varSys%nVars), spc_mom(3), mom_inG(3)
    real(kind=rk) :: mass_dens(varSys%nVars), num_dens(varSys%nVars)
    real(kind=rk) :: spc_num_dens(1), totMassDens, velAvg(3), velQuad(3)
    real(kind=rk) :: moleFrac(varSys%nVars)
    integer :: nField
    character(len=labelLen) :: fieldPrefix( varSys%nVars )
    real(kind=rk) :: resi_coeff( varSys%nVars, varSys%nVars )
    real(kind=rk) :: molWeight(varSys%nVars), phi(varSys%nVars), &
      &              chargeNr(varSys%nVars)
    real(kind=rk) :: loc_curr_dens(3)
    real(kind=rk) :: mem_curr_dens, mem_molar_flux
    ! local DUMMY subTree needed to use the function pointers (not filled)
    type(tem_subTree_type) :: subTree
    integer :: bound(2)
    real(kind=rk) :: first_moments(3), second_moments(6), third_moments(6), &
      &              fourth_moments(3)
    integer :: iMom, nMom(4)
    !------------------------------------------------------------------------ 
!write(dbgUnit(1),*) 'Boundary label ', trim(me%label)
!write(dbgUnit(1),*) 'iField ', iField

    nFields = varSys%nVars
    fieldPrefix = getFieldPrefixes( conf = varSys%solSpec%conf,               &
      &                             nFields = nFields )

    do iFieldLoc = 1, nFields
      molWeight(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,     &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeight',                                    &
        &           fieldProp = 'species' )                           
      phi(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,           &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'molWeigRatio',                                 &
        &           fieldProp = 'species' )                                
      resi_coeff(iFieldLoc, :) = getFieldVariable( conf = varSys%solSpec%conf, &
        &           varLabel = trim(fieldPrefix(iField))//'velocity',          &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'resi_coeff',                                   &
        &           fieldProp = 'species',                                     &
        &           nVals = nFields )                                    
      chargeNr(iFieldLoc) = getFieldVariable( conf = varSys%solSpec%conf,      &
        &           varLabel = trim(fieldPrefix(iFieldLoc))//'velocity',       &
        &           varName = 'velocity',                                      &
        &           fieldVar = 'charge_nr',                                    &
        &           fieldProp = 'species' )                           
    end do    

   
    nMom(1) = 1 + size(layout%moment%first_moments)
    nMom(2) = nMom(1) + size(layout%moment%second_moments)
    nMom(3) = nMom(2) + size(layout%moment%third_moments)
    nMom(4) = nMom(3) + size(layout%moment%fourth_moments)
    
!    write(*,*) 'iField ', iField
    ! Calculate the density of current element
    do iElem = 1, globBC%nElems(iLevel)
      if( .not. btest(glob%levelDesc(iLevel)%property(&
        &             globBC%elems(iLevel)%elem(iElem)), prp_missingNeigh) ) then
!write(dbgUnit(1),*) 'iElem ', iElem  

      nLinks = me%elems(iLevel)%moments(iElem)%nUnKnownPdfs
      allocate(missing_links(nLinks))
      allocate(rhs(nLinks))
      allocate(unKnown_fTmp(nLinks))
      elemPos = globBC%elems(iLevel)%elem( iElem )

      iLink = 0
      fTmp = 0.0_rk
      do iDir = 1,layout%stencil(1)%QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          iLink = iLink +1
          missing_links(iLink) = iDir
        else
          fTmp( iDir ) = &
            &state(?FETCH?( varPos(iDir), elemPos, nScalars, neigh ), &
        & currT) 
        endif
      end do 
      !write(*,*) 'fTmp ', fTmp

      ! local state vector to compute density and velocity of other species
      do iFieldLoc = 1, VarSys%nVars
        do iDir = 1,layout%stencil(1)%QQ
          pos = varSys%Variable%val(iFieldLoc)%varPos(iDir)
          fTmp_all( pos ) = &
            & bcBuffer( ?IDX?( pos, globBC%elems(iLevel)%elemBuffer(iElem), nScalars ))
        enddo
      enddo

      !local density and momentum
      mass_dens = 0.0_rk
      mom = 0.0_rk
      ! initialize the lower and upper bound
      bound(1) = 1
      bound(2) = 1
      do iFieldLoc = 1, nFields
!        if(iField == iFieldLoc) then
!          ! massflux = moleflux * molecular_weight
!          mom(:, iField) = massFlux
!          mass_dens(iField) = rho
!          num_dens(iField) = rho / fieldProp%species%molWeight
!        else
          do iDir = 1,layout%stencil(1)%QQ
           ! mass density
            mass_dens(iFieldLoc) = mass_dens(iFieldLoc)               &
              & + fTmp_all(varSys%Variable%val(iFieldLoc)%varPos(iDir)) 
          end do
          num_dens(iFieldLoc) = mass_dens(iFieldLoc) &
            &                 / molWeight(iFieldLoc)
          !mass flux
          call varSys%derive( derVarPosAll(iFieldLoc)%momentum )%do(              &
            &                       input = fTmp_all,                             &
            &                       bound = bound,                                &
            &                       tree = tree,                                  &
            &                       subTree = subTree,                            &
            &                       requVar = varSys%Variable%val(                &
            &                                 derVarPosAll(iFieldLoc)%momentum ), &
            &                       globSys = varSys, stencil = layout%stencil(1),&
            &                       res = spc_mom,                                &
            &                       conf = varSys%solSpec%conf )

          mom(:, iFieldLoc) = spc_mom
!        endif
      end do  

      ! local current density
      loc_curr_dens = 0.0_rk
      do iFieldLoc = 1, nFields
        loc_curr_dens(:) = loc_curr_dens(:) + chargeNr(iFieldLoc)                  &
          &              * mom(:, iFieldLoc) / molWeight(iFieldLoc)
      end do

      loc_curr_dens = loc_curr_dens * mixture%faradayLB

      ! project local current density on the membrane
      mem_curr_dens = dot_product(abs(globBC%elems(iLevel)%normal(:, iElem)), &
        &                             loc_curr_dens)

      mem_molar_flux = me%blackbox_mem%transNr * mem_curr_dens                 &
        &            / ( chargeNr(iField) * mixture%faradayLB )

      mom(:, iField) = mem_molar_flux * molWeight(iField)

      ! mass density
      mass_dens = num_dens * molWeight
      rho = mass_dens(iField)
      !pressure
      press = cs2*rho*fieldProp%species%molWeigRatio
!write(dbgUnit(1),*) 'rho ', rho, 'press', press

      ! total mass density
      totMassDens = sum(mass_dens)

      ! mole fraction
      moleFrac = num_dens/sum(num_dens)

!do iFieldLoc = 1, nFields
!  write(dbgUnit(1),*) iFieldLoc, 'vel ', mom(:, iFieldLoc)/mass_dens(iFieldLoc), &
!    & 'dens ', mass_dens(iFieldLoc), 'moleFrac ', moleFrac(iFieldLoc)
!enddo  
      
      ! compute velocity of transformed variable g. Eq. which is normally
      ! used to compute actual velocity by solving LSE
      mom_inG = mom(:, iField)
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG + ( mixture%omega_diff * 0.5_rk                    &
            &               * resi_coeff(iField, iFieldLoc) * phi(iField)    &
            &               * moleFrac(iFieldLoc) / mixture%paramB )         &
            &               * mom(:, iField)
      end do
      !set non-diagonal part
      do iFieldLoc = 1, nFields
        mom_inG = mom_inG - ( mixture%omega_diff * 0.5_rk                    &
          &                 * resi_coeff(iField, iFieldLoc) * phi(iFieldLoc) &
          &                 * moleFrac(iField) / mixture%paramB )            &
          &                 * mom(:,iFieldLoc)
      end do

      vel_inG = mom_inG/rho
!write(dbgUnit(1),*) 'defined momentum ', massFlux      
!write(dbgUnit(1),*) 'transformed momentum ', mom_inG
!write(dbgUnit(1),*) 'transformed velocity ', vel_inG

      ! mass averaged mixture velocity
      velAvg(1) = sum(mom(1,:)) / totMassDens
      velAvg(2) = sum(mom(2,:)) / totMassDens
      velAvg(3) = sum(mom(3,:)) / totMassDens
!      write(*,*) 'velAvg ', velAvg
      velQuad(:) = mixture%theta_eq*velAvg(:) &
        &        + (1.0_rk - mixture%theta_eq)*vel_inG

      ! moments for multispecies LBM model of transformed variable
      first_moments = mom_inG !momX, momY, momZ 
      second_moments = (/ press + rho*velQuad(1)*velQuad(1), & !momXX
        &                 press + rho*velQuad(2)*velQuad(2), & !momYY
        &                 press + rho*velQuad(3)*velQuad(3), & !momZZ
        &                 rho*velQuad(1)*velQuad(2),         & !momXY
        &                 rho*velQuad(2)*velQuad(3),         & !momYZ
        &                 rho*velQuad(3)*velQuad(1) /)         !momXZ
      third_moments = (/ cs2*mom_inG(2), & !momXXY
        &                cs2*mom_inG(3), & !momXXZ
        &                cs2*mom_inG(1), & !momYYX
        &                cs2*mom_inG(3), & !momYYZ
        &                cs2*mom_inG(1), & !momZZX
        &                cs2*mom_inG(2) /) !momZZY
      fourth_moments = (/ &
        & cs2*(press+rho*(        velQuad(1)**2 +        velQuad(2)**2 &
        &                - 0.5_rk*velQuad(3)**2 )), &!momXXYY
        & cs2*(press+rho*(-0.5_rk*velQuad(1)**2 +        velQuad(2)**2 &
        &                + velQuad(3)**2 )), &!momYYZZ
        & cs2*(press+rho*(        velQuad(1)**2 - 0.5_rk*velQuad(2)**2 &
        &                + velQuad(3)**2 ))  &!momZZXX
        & /) 
     
      ! moments for liquid mixture
      moments = 0.0_rk
      moments(1) = rho
      moments(2:nMom(1)) = &
        & first_moments(layout%moment%first_moments(:))
      moments(nMom(1)+1:nMom(2)) = &
        & second_moments(layout%moment%second_moments(:))
      moments(nMom(2)+1:nMom(3)) = &
        & third_moments(layout%moment%third_moments(:))
      moments(nMom(3)+1:nMom(4)) = &
        & fourth_moments(layout%moment%fourth_moments(:))

      !write(*,*) 'Moments ', moments
      !write(*,*) 'knownMomPos ', me%elems(iLevel)%moments(iElem)%knownMom_pos

      do iLink = 1, nLinks
        rhs(iLink) = &
          & moments( me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink) ) &
          ! convert pdf to moments
          & - dot_product(layout%moment%toMoments%A(                        &
          &    me%elems(iLevel)%moments(iElem)%knownMom_pos(iLink), :), fTmp)
      end do
      !write(*,*) 'rhs ', rhs

      unKnown_fTmp = matmul( &
        & me%elems(iLevel)%moments(iElem)%unKnownPdfs_MatInv, rhs )
      !write(*,*) 'unknown fTmp ', unKnown_fTmp

      do iLink = 1, nLinks
        state(                                                                 &
&?FETCH?( varPos(missing_links(iLink)), globBC%elems(iLevel)%elem( iElem ), nScalars, neigh ), &
         & currT)  = unKnown_fTmp(iLink)
      end do

      deallocate(missing_links)
      deallocate(rhs)
      deallocate(unKnown_fTmp)

      endif
    end do

  end subroutine spc_moments_blackbox_mem_ion
!******************************************************************************!

end module mus_bc_species_module
!******************************************************************************!
