! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> Routines and parameter definitions for the standard D3Q19 model
!! \author Manuel Hasert, Kannan Masilamani
?? include 'header/lbm_macros.inc'
module mus_d3q19_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_var_system_module, only: tem_var_system_type
  use tem_param_module,      only: div1_3, div1_36, div1_8, div3_4h, cs2inv,   &
    &                              cs4inv, t2cs2inv,t2cs4inv
  use treelmesh_module,      only: treelmesh_type

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_param_module,         only: mus_param_type
  use mus_solSpecHelpers_module, only: getConversionFac

  implicit none

  private

  public :: bgk_advRel_d3q19
  public :: bgk_advRel_d3q19_fixedVel
  public :: trt_advRel_d3q19
  public :: trt_advRel_d3q19_fixedVel
  public :: trt_advRel_d3q19_les
  public :: bgk_advRel_d3q19_incomp
  public :: trt_advRel_d3q19_incomp
  public :: bgk_advRel_d3q19_omegaSpatial
  public :: bgk_advRel_d3q19_incomp_omegaSpatial
  public :: trt_advRel_d3q19_les_omegaSpatial
  public :: bgk_advRel_d3q19_les
  public :: bgk_advRel_d3q19_incomp_les
  public :: bgk_advRel_d3q19_incomp_fixedVel
  public :: bgk_advRel_d3q19_nNwtn_PL
  public :: bgk_advRel_d3q19_nNwtn_CY
  public :: bgk_advRel_d3q19_nNwtn_CS
  public :: bgk_advRel_d3q19_incomp_nNwtn_CY
  public :: bgk_advRel_d3q19_incomp_nNwtn_PL
  public :: bgk_advRel_d3q19_incomp_nNwtn_CS

  !=============================================================================
  ! D3Q19 flow model
  !=============================================================================
  !> Definition of the discrete velocity set

  integer, parameter :: QQ  = 19   !> number of pdf directions

  integer,parameter :: q__W     = 1     !< west             x-
  integer,parameter :: q__S     = 2     !< south            y-
  integer,parameter :: q__B     = 3     !< bottom           z-
  integer,parameter :: q__E     = 4     !< east             x+
  integer,parameter :: q__N     = 5     !< north            y+
  integer,parameter :: q__T     = 6     !< top              z+
  integer,parameter :: q_BS     = 7     !<                  z-,y-
  integer,parameter :: q_TS     = 8     !<                  z+,y-
  integer,parameter :: q_BN     = 9     !<                  z-,y+
  integer,parameter :: q_TN     = 10    !<                  z+,y+
  integer,parameter :: q_BW     = 11    !<                  x-,z-
  integer,parameter :: q_BE     = 12    !<                  x+,z-
  integer,parameter :: q_TW     = 13    !<                  x-,z+
  integer,parameter :: q_TE     = 14    !<                  x+,z+
  integer,parameter :: q_SW     = 15    !<                  y-,x-
  integer,parameter :: q_NW     = 16    !<                  y+,x-
  integer,parameter :: q_SE     = 17    !<                  y-,x+
  integer,parameter :: q_NE     = 18    !<                  y+,x+
  integer,parameter :: q__0     = 19    !< rest density is last

contains

!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK.
  !! 
  !! Streaming is defined in the header/lbm_macros.inc in STREAM_PUSH = .false.
  !! This routine ONLY contains pure computation
  !! and should be implemented in a very efficient way 
  !! \f$ f_\alpha(x_i+e_{\alpha,i},t+1) = 
  !! f_\alpha(x_i,t) - \omega(f_\alpha(x_i,t)-f^{eq}_{\alpha}(x_i,t)) \f$  
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19 )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho ! inverse local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2 
    real(kind=rk) :: omega_2, cmpl_o, omega
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &              sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1,        &
      &              sum11_2, sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! read the relaxation parameter omega for the current level
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    ! pre-calculate partial collision constants
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible

      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho

      ! f = (1-w) * f + w * fEq
      outState(?SAVE?(Q__0,iElem,QQ,neigh)) =                                  &
        &                     pdfTmp(Q__0)*cmpl_o+omega*rho*(div1_3-0.5_rk*usq)

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,QQ,neigh)) =                                  &
        &                                   pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,QQ,neigh )) =                                 &
        &                                   pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,QQ,neigh )) =                                 &
        &                                   pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,QQ,neigh )) =                                 &
        &                                   pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,QQ,neigh )) =                                 &
        &                                   pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,QQ,neigh )) =                                 &
        &                                   pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,QQ,neigh )) =                                 &
        &                                   pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK.
  !! 
  !! Streaming is defined in the header/lbm_macros.inc in STREAM_PUSH = .false.
  !! This routine ONLY contains pure computation
  !! and should be implemented in a very efficient way 
  !! \f$ f_\alpha(x_i+e_{\alpha,i},t+1) = 
  !! f_\alpha(x_i,t) - \omega(f_\alpha(x_i,t)-f^{eq}_{\alpha}(x_i,t)) \f$  
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_fixedVel )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    integer, parameter :: iField = 1
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho ! inverse local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2 
    real(kind=rk) :: omega_2, cmpl_o, omega
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &              sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1,        &
      &              sum11_2, sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! read the relaxation parameter omega for the current level
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    ! pre-calculate partial collision constants
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(1)%       &
        &                                          stored( level )%val( iElem )
      u_y = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(2)%       &
        &                                          stored( level )%val( iElem )
      u_z = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(3)%       &
        &                                          stored( level )%val( iElem )

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho

      outState(?SAVE?(Q__0,iElem,QQ,neigh)) =                                  &
        &                     pdfTmp(Q__0)*cmpl_o+omega*rho*(div1_3-0.5_rk*usq)

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,QQ,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,QQ,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,QQ,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,QQ,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,QQ,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,QQ,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,QQ,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,QQ,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,QQ,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,QQ,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,QQ,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,QQ,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,QQ,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,QQ,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,QQ,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,QQ,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,QQ,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,QQ,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_fixedVel
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK for 
  !! spatially varying relaxation parameter omega
  !! Definition in the configuration file according to 
  !! \ref mus_aux_module::mus_setspatialomega "spatial omega"
?? copy :: compute_routineHeader( bgk_advRel_d3q19_omegaSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho ! inverse local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2 
    real(kind=rk) :: omega_2, cmpl_o, omega, omSrc
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &              sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1,        &
      &              sum11_2, sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! some global variables
    omSrc   = fieldProp(1)%fluid%rampOmegas( level )

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      omega   = fieldProp(1)%fluid%viscSpatial%stored(level)%val( iElem )*omSrc
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho

      outState(?SAVE?(Q__0,iElem,QQ,neigh)) =                                  &
        &                     pdfTmp(Q__0)*cmpl_o+omega*rho*(div1_3-0.5_rk*usq)

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,QQ,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,QQ,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,QQ,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,QQ,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,QQ,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,QQ,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,QQ,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,QQ,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,QQ,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,QQ,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,QQ,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,QQ,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,QQ,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,QQ,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,QQ,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,QQ,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,QQ,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,QQ,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_omegaSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK and 
  !! Smagorinsky Subgrid-scale model for large eddy simulations (LES).
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_LES )
    !---------------------------------------------------------------------------
    integer :: iElem, iDir                      ! voxel element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    ! derived constants
    real(kind=rk) :: usq, ucx
    real(kind=rk) :: fEq(19) !< equilibrium distribution
    real(kind=rk) :: nEq(19) !< non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3), omega
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const
    !---------------------------------------------------------------------------

    ! some global variables
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s
    tau0 = 1._rk/omega

    !$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! Calculate the equilibrium distribution function
      !fEq(:) = getEquilibrium( density = rho, velocity = (/u_x, u_y, u_z /), layout = layout)
      ! square of velocity
      usq = u_x*u_x + u_y*u_y + u_z*u_z
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx =   real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk) * u_x       &
          &   + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk) * u_y       &
          &   + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk) * u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * rho * ( 1.d0 + ucx*cs2inv        &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv )

      enddo

      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )

      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)
      om_turb = 1._rk/(tau0+tau_turb)

      ! relaxation
      outState(?SAVE?( Q__0,iElem,QQ, neigh)) =                                &
        &      pdfTmp( Q__0 ) - om_turb*(pdfTmp( Q__0 ) - fEq( Q__0 ))
      outState(?SAVE?(Q_NE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_NE ) - om_turb*(pdfTmp( Q_NE ) - fEq( Q_NE ))
      outState(?SAVE?(Q_SW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_SW ) - om_turb*(pdfTmp( Q_SW ) - fEq( Q_SW ))
      outState(?SAVE?(Q_NW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_NW ) - om_turb*(pdfTmp( Q_NW ) - fEq( Q_NW ))
      outState(?SAVE?(Q_SE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_SE ) - om_turb*(pdfTmp( Q_SE ) - fEq( Q_SE ))
      outState(?SAVE?(Q_TE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TE ) - om_turb*(pdfTmp( Q_TE ) - fEq( Q_TE ))
      outState(?SAVE?(Q_BW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BW ) - om_turb*(pdfTmp( Q_BW ) - fEq( Q_BW ))
      outState(?SAVE?(Q_TW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TW ) - om_turb*(pdfTmp( Q_TW ) - fEq( Q_TW ))
      outState(?SAVE?(Q_BE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BE ) - om_turb*(pdfTmp( Q_BE ) - fEq( Q_BE ))
      outState(?SAVE?(Q_TN,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TN ) - om_turb*(pdfTmp( Q_TN ) - fEq( Q_TN ))
      outState(?SAVE?(Q_BS,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BS ) - om_turb*(pdfTmp( Q_BS ) - fEq( Q_BS ))
      outState(?SAVE?(Q_TS,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TS ) - om_turb*(pdfTmp( Q_TS ) - fEq( Q_TS ))
      outState(?SAVE?(Q_BN,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BN ) - om_turb*(pdfTmp( Q_BN ) - fEq( Q_BN ))
      outState(?SAVE?(Q__N,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__N ) - om_turb*(pdfTmp( Q__N ) - fEq( Q__N ))
      outState(?SAVE?(Q__S,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__S ) - om_turb*(pdfTmp( Q__S ) - fEq( Q__S ))
      outState(?SAVE?(Q__E,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__E ) - om_turb*(pdfTmp( Q__E ) - fEq( Q__E ))
      outState(?SAVE?(Q__W,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__W ) - om_turb*(pdfTmp( Q__W ) - fEq( Q__W ))
      outState(?SAVE?(Q__B,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__B ) - om_turb*(pdfTmp( Q__B ) - fEq( Q__B ))
      outState(?SAVE?(Q__T,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__T ) - om_turb*(pdfTmp( Q__T ) - fEq( Q__T ))

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_les
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK for 
  !! incompressible lbm model
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_incomp_fixedVel )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    integer, parameter :: iField = 1
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ)    ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho0 ! inverse local density
    real(kind=rk) rho0    ! reference density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! derived constants
    real(kind=rk) usqn, usqn_o1, usqn_o2 
    real(kind=rk) omega_2, cmpl_o, omega
    real(kind=rk) coeff_1, coeff_2
    real(kind=rk) ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,   &
      &           fac_13
    real(kind=rk) sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,      &
      &           sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1, sum11_2,  &
      &           sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! some global variables
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    rho0 = fieldProp(1)%fluid%rho0
    ! inverse local density
    inv_rho0 = 1._rk / rho0
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(1)%       &
        &                                          stored( level )%val( iElem )
      u_y = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(2)%       &
        &                                          stored( level )%val( iElem )
      u_z = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(3)%       &
        &                                          stored( level )%val( iElem )

      ! transfer moments to velocities
      u_x = u_x * inv_rho0
      u_y = u_y * inv_rho0
      u_z = u_z * inv_rho0

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk*rho - 1.5_rk * usq * rho0)

      outState(?SAVE?(Q__0,iElem,QQ,neigh)) =                                  &
        &                pdfTmp(Q__0)*cmpl_o+omega*(rho*div1_3-0.5_rk*usq*rho0)

      coeff_1 = div1_8 * omega * rho0

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,QQ,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,QQ,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,QQ,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,QQ,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,QQ,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,QQ,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,QQ,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,QQ,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,QQ,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,QQ,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,QQ,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,QQ,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      coeff_2 = div1_8 * omega_2 * rho0
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,QQ,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,QQ,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,QQ,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,QQ,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,QQ,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,QQ,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_incomp_fixedVel
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK for 
  !! incompressible lbm model
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_incomp )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho0 ! inverse local density
    real(kind=rk) rho0    ! reference density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! derived constants
    real(kind=rk) usqn, usqn_o1, usqn_o2 
    real(kind=rk) omega_2, cmpl_o, omega
    real(kind=rk) coeff_1, coeff_2
    real(kind=rk) ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,   &
      &           fac_13
    real(kind=rk) sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,      &
      &           sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1, sum11_2,  &
      &           sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! some global variables
    omega = fieldProp(1)%fluid%rampOmegas( level )
    rho0 = fieldProp(1)%fluid%rho0
    ! inverse local density
    inv_rho0 = 1._rk / rho0
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! transfer moments to velocities
      u_x = u_x * inv_rho0
      u_y = u_y * inv_rho0
      u_z = u_z * inv_rho0

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (rho - 1.5_rk * usq * rho0)

      outState(?SAVE?(Q__0,iElem,QQ,neigh)) =                                  &
        &                pdfTmp(Q__0)*cmpl_o+omega*(rho*div1_3-0.5_rk*usq*rho0)

      coeff_1 = div1_8 * omega * rho0

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,QQ,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,QQ,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,QQ,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,QQ,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,QQ,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,QQ,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,QQ,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,QQ,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,QQ,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,QQ,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,QQ,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,QQ,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      coeff_2 = div1_8 * omega_2 * rho0
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,QQ,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,QQ,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,QQ,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,QQ,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,QQ,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,QQ,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_incomp
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with incompressible BGK 
  !! for spatially varying relaxation parameter omega.
  !! Definition in the configuration file according to 
  !! \ref mus_aux_module::mus_setspatialomega "spatial omega"
?? copy :: compute_routineHeader( bgk_advRel_d3q19_incomp_omegaSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem                    ! voxel element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho0 ! inverse local density
    real(kind=rk) rho0    ! reference density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! derived constants
    real(kind=rk) usqn, usqn_o1, usqn_o2 
    real(kind=rk) omega_2, cmpl_o, omega, omSrc
    real(kind=rk) coeff_1, coeff_2
    real(kind=rk) ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,   &
      &           fac_13
    real(kind=rk) sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,      &
      &           sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1, sum11_2,  &
      &           sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------

    ! some global variables
    omSrc   = fieldProp(1)%fluid%rampOmegas( level )
    rho0 = fieldProp(1)%fluid%rho0
    ! inverse local density
    inv_rho0 = 1._rk / rho0

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      omega   = fieldProp(1)%fluid%viscSpatial%stored( level )%                &
        &                                                    val( iElem )*omSrc
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! transfer moments to velocities
      u_x = u_x * inv_rho0
      u_y = u_y * inv_rho0
      u_z = u_z * inv_rho0

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk*rho - 1.5_rk * usq * rho0)

      outState(?SAVE?(Q__0,iElem,QQ,neigh)) =                                  &
        &                pdfTmp(Q__0)*cmpl_o+omega*(rho*div1_3-0.5_rk*usq*rho0)

      coeff_1 = div1_8 * omega * rho0

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,QQ,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,QQ,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,QQ,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,QQ,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,QQ,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,QQ,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,QQ,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,QQ,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,QQ,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,QQ,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,QQ,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,QQ,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      coeff_2 = div1_8 * omega_2 * rho0
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,QQ,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,QQ,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,QQ,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,QQ,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,QQ,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,QQ,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_incomp_omegaSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK incompressible 
  !! model and Smagorinsky Subgrid-scale model for large eddy simulations (LES).
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_incomp_les )
    !---------------------------------------------------------------------------
    integer :: iElem,iDir                      ! voxel element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: inv_rho0 ! inverse local density
    real(kind=rk) :: rho0    ! reference density
    ! derived constants
    real(kind=rk) :: usq, ucx
    real(kind=rk) :: fEq(19) ! equilibrium distribution
    real(kind=rk) :: nEq(19) ! non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3), omega
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const
    !---------------------------------------------------------------------------

    ! some global variables
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s
    tau0 = 1._rk/omega
    rho0 = fieldProp(1)%fluid%rho0
    inv_rho0 = 1._rk / rho0


    !$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! loal x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! transfer moments to velocities
      u_x = u_x * inv_rho0
      u_y = u_y * inv_rho0
      u_z = u_z * inv_rho0

      ! Calculate the equilibrium distribution function using
      ! incompressible formulation
      ! square of velocity
      usq = u_x*u_x + u_y*u_y + u_z*u_z
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx =   real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk) * u_x       &
          &   + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk) * u_y       &
          &   + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk) * u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * ( rho + rho0 * ( ucx*cs2inv      &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv ) )

      enddo

      !fEq(:) = getEquilibrium( density = rho, velocity = (/u_x, u_y, u_z /), layout = layout)
!      input_loc = (/ rho, u_x, u_y, u_z /)
!      call scheme( iScheme )%varSys%derive(                                    &
!        &     scheme( iScheme )%derVarPos( iField )%equilFromMacro )%do(       &
!        &            input = input_loc, nElems = nChunkElems,                  &
!        &            tree = tree,                                              &
!        &            requVar = scheme(iScheme)%varSys%Variable%val(            &
!        &                scheme(iScheme)%derVarPos(iField)%equilFromMacro ),   &
!        &            globSys = scheme(iScheme)%varSys,                         &
!        &            stencil = scheme(iScheme)%layout%stencil(1),              &
!        &            res = fEq,                                                &
!        &              treeID = tree%treeID,                                   &
!        &            conf = scheme(iScheme)%varSys%solSpec%conf )
 
      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )
      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)

      om_turb = 1._rk/(tau0+tau_turb)

      ! relaxation
      outState(?SAVE?( Q__0,iElem,QQ, neigh)) =                                &
        &      pdfTmp( Q__0 ) - om_turb*(pdfTmp( Q__0 ) - fEq( Q__0 ))
      outState(?SAVE?(Q_NE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_NE ) - om_turb*(pdfTmp( Q_NE ) - fEq( Q_NE ))
      outState(?SAVE?(Q_SW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_SW ) - om_turb*(pdfTmp( Q_SW ) - fEq( Q_SW ))
      outState(?SAVE?(Q_NW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_NW ) - om_turb*(pdfTmp( Q_NW ) - fEq( Q_NW ))
      outState(?SAVE?(Q_SE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_SE ) - om_turb*(pdfTmp( Q_SE ) - fEq( Q_SE ))
      outState(?SAVE?(Q_TE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TE ) - om_turb*(pdfTmp( Q_TE ) - fEq( Q_TE ))
      outState(?SAVE?(Q_BW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BW ) - om_turb*(pdfTmp( Q_BW ) - fEq( Q_BW ))
      outState(?SAVE?(Q_TW,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TW ) - om_turb*(pdfTmp( Q_TW ) - fEq( Q_TW ))
      outState(?SAVE?(Q_BE,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BE ) - om_turb*(pdfTmp( Q_BE ) - fEq( Q_BE ))
      outState(?SAVE?(Q_TN,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TN ) - om_turb*(pdfTmp( Q_TN ) - fEq( Q_TN ))
      outState(?SAVE?(Q_BS,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BS ) - om_turb*(pdfTmp( Q_BS ) - fEq( Q_BS ))
      outState(?SAVE?(Q_TS,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_TS ) - om_turb*(pdfTmp( Q_TS ) - fEq( Q_TS ))
      outState(?SAVE?(Q_BN,iElem,QQ,neigh))  =                                 &
        &      pdfTmp( Q_BN ) - om_turb*(pdfTmp( Q_BN ) - fEq( Q_BN ))
      outState(?SAVE?(Q__N,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__N ) - om_turb*(pdfTmp( Q__N ) - fEq( Q__N ))
      outState(?SAVE?(Q__S,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__S ) - om_turb*(pdfTmp( Q__S ) - fEq( Q__S ))
      outState(?SAVE?(Q__E,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__E ) - om_turb*(pdfTmp( Q__E ) - fEq( Q__E ))
      outState(?SAVE?(Q__W,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__W ) - om_turb*(pdfTmp( Q__W ) - fEq( Q__W ))
      outState(?SAVE?(Q__B,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__B ) - om_turb*(pdfTmp( Q__B ) - fEq( Q__B ))
      outState(?SAVE?(Q__T,iElem,QQ,neigh )) =                                 &
        &      pdfTmp( Q__T ) - om_turb*(pdfTmp( Q__T ) - fEq( Q__T ))

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_incomp_les
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with TRT collision 
  !! operator
  !!
?? copy :: compute_routineHeader( trt_advRel_d3q19 )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! derived constants
    real(kind=rk) usqn
    real(kind=rk) omega_2, cmpl_o, omega, om, omega_h, asym_omega_h
    real(kind=rk) ui
    real(kind=rk) :: asym, sym, feq_common
    ! in case of the default model, the actual values depend on density
    real(kind=rk) ::  t0,t1x2,t2x2,fac1,fac2
    ! common prefactors for calculating the equilibrium parts (D3Q19 model)
    real(kind=rk), parameter :: t0_0 = 1.d0/3.d0
    ! twice the usual value due
    real(kind=rk), parameter :: t1x2_0 = 1.d0/18.d0 * 2.d0
    ! to the way of calculations
    real(kind=rk), parameter :: t2x2_0 = 1.d0/36.d0 * 2.d0
    !---------------------------------------------------------------------------

    ! read the relaxation parameter omega for the current level
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    ! pre-calculate partial collision constants
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

    ! some global variables
    omega_h      = 0.5d0 * omega
!#ifdef TRT_ASYM_1
!    asym_omega_h = 0.5d0 * asym_omega
!#else
!#ifdef TRT_BGK
!    ! check that BGK is recovered; o.k. works fine
!    asym_omega_h = 0.5d0 * omega
!#else
!    ! magic Irina for fixed walls
    asym_omega_h = 0.5d0 * 8.d0 * (2.d0-omega)/(8.d0-omega)
!#endif   /* not TRT_BGK */
!#endif   /* not TRT_ASYM_1 */
    om = omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      feq_common = 1.d0  - 1.5d0 * usq
      t0 = t0_0 * rho

      ! let's start the relaxation process
      outstate(?SAVE?(Q__0, iElem, QQ, neigh)) =                               &
        &                            pdfTmp(Q__0 )*(1.d0-om) + om*t0*feq_common

      t2x2 = t2x2_0 * rho
      fac2 = t2x2 * t2cs4inv !inv2csq2

      ui   = u_x + u_y
      sym  = omega_h*(pdfTmp(Q_NE) + pdfTmp(Q_SW) - fac2*ui*ui -t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_NE) - pdfTmp(Q_SW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_NE, iElem, QQ, neigh)) = pdfTmp(Q_NE) - sym - asym
      outstate(?SAVE?(Q_SW, iElem, QQ, neigh)) = pdfTmp(Q_SW) - sym + asym

      ui   = u_x - u_y
      sym  = omega_h*(pdfTmp(Q_SE) + pdfTmp(Q_NW) - fac2*ui*ui -t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_SE) - pdfTmp(Q_NW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_SE, iElem, QQ, neigh)) = pdfTmp(Q_SE) - sym - asym
      outstate(?SAVE?(Q_NW, iElem, QQ, neigh)) = pdfTmp(Q_NW) - sym + asym

      ui   = u_x + u_z
      sym  = omega_h*(pdfTmp(Q_TE) + pdfTmp(Q_BW) - fac2*ui*ui -t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TE) - pdfTmp(Q_BW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TE, iElem, QQ, neigh)) = pdfTmp(Q_TE) - sym - asym
      outstate(?SAVE?(Q_BW, iElem, QQ, neigh)) = pdfTmp(Q_BW) - sym + asym

      ui   = u_x - u_z
      sym  = omega_h*(pdfTmp(Q_BE) + pdfTmp(Q_TW) - fac2*ui*ui -t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BE) - pdfTmp(Q_TW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BE, iElem, QQ, neigh)) = pdfTmp(Q_BE) - sym - asym
      outstate(?SAVE?(Q_TW, iElem, QQ, neigh)) = pdfTmp(Q_TW) - sym + asym

      ui   = u_y + u_z
      sym  = omega_h*(pdfTmp(Q_TN) + pdfTmp(Q_BS) - fac2*ui*ui -t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TN) - pdfTmp(Q_BS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TN, iElem, QQ, neigh)) = pdfTmp(Q_TN) - sym - asym
      outstate(?SAVE?(Q_BS, iElem, QQ, neigh)) = pdfTmp(Q_BS) - sym + asym

      ui   = u_y - u_z
      sym  = omega_h*(pdfTmp(Q_BN) + pdfTmp(Q_TS) - fac2*ui*ui -t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BN) - pdfTmp(Q_TS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BN, iElem, QQ, neigh)) = pdfTmp(Q_BN) - sym - asym
      outstate(?SAVE?(Q_TS, iElem, QQ, neigh)) = pdfTmp(Q_TS) - sym + asym

      t1x2 = t1x2_0 * rho
      fac1 = t1x2 * t2cs4inv !inv2csq2

      ui   = u_y
      sym  = omega_h*(pdfTmp(Q__N ) + pdfTmp(Q__S) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__N ) - pdfTmp(Q__S) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__N, iElem, QQ, neigh)) = pdfTmp(Q__N ) - sym - asym
      outstate(?SAVE?(Q__S, iElem, QQ, neigh)) = pdfTmp(Q__S ) - sym + asym

      ui   = u_x
      sym  = omega_h*(pdfTmp(Q__E ) + pdfTmp(Q__W) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__E ) - pdfTmp(Q__W ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__E, iElem, QQ, neigh)) = pdfTmp(Q__E ) - sym - asym
      outstate(?SAVE?(Q__W, iElem, QQ, neigh)) = pdfTmp(Q__W ) - sym + asym

      ui   = u_z
      sym  = omega_h*(pdfTmp(Q__T ) + pdfTmp(Q__B) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__T ) - pdfTmp(Q__B ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__T, iElem, QQ, neigh)) = pdfTmp(Q__T ) - sym - asym
      outstate(?SAVE?(Q__B, iElem, QQ, neigh)) = pdfTmp(Q__B ) - sym + asym
    enddo nodeloop

  end subroutine trt_advRel_d3q19
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with TRT collision
  !! operator
  !!
?? copy :: compute_routineHeader( trt_advRel_d3q19_fixedVel )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! derived constants
    real(kind=rk) usqn
    real(kind=rk) omega_2, cmpl_o, omega, om, omega_h, asym_omega_h
    real(kind=rk) ui
    real(kind=rk) :: asym, sym, feq_common
    ! in case of the default model, the actual values depend on density
    real(kind=rk) ::  t0,t1x2,t2x2,fac1,fac2

    ! common prefactors for calculating the equilibrium parts (D3Q19 model)
    real(kind=rk), parameter :: t0_0 = 1.d0/3.d0
    ! twice the usual value due
    real(kind=rk), parameter :: t1x2_0 = 1.d0/18.d0 * 2.d0
    ! to the way of calculations
    real(kind=rk), parameter :: t2x2_0 = 1.d0/36.d0 * 2.d0
    integer, parameter :: iField = 1
    !------------------------------------------------------------------------ 

    ! read the relaxation parameter omega for the current level
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    ! pre-calculate partial collision constants
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

    ! some global variables
    omega_h      = 0.5d0 * omega
!#ifdef TRT_ASYM_1
!    asym_omega_h = 0.5d0 * asym_omega
!#else
!#ifdef TRT_BGK
!    ! check that BGK is recovered; o.k. works fine
!    asym_omega_h = 0.5d0 * omega
!#else
!    ! magic Irina for fixed walls
    asym_omega_h = 0.5d0 * 8.d0 * (2.d0-omega)/(8.d0-omega)
!#endif   /* not TRT_BGK */
!#endif   /* not TRT_ASYM_1 */
    om = omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! inverse local density
      inv_rho = 1._rk / rho

      u_x = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(1)%       &
        &                                          stored( level )%val( iElem )
      u_y = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(2)%       &
        &                                          stored( level )%val( iElem )
      u_z = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(3)%       &
        &                                          stored( level )%val( iElem )

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      feq_common = 1.d0  - 1.5d0 * usq
      t0 = t0_0 * rho

      ! let's start the relaxation process
      outstate(?SAVE?(Q__0, iElem, QQ, neigh)) =                               &
        &                            pdfTmp(Q__0 )*(1.d0-om) + om*t0*feq_common

      t2x2 = t2x2_0 * rho
      fac2 = t2x2 * t2cs4inv !inv2csq2

      ui   = u_x + u_y
      sym  = omega_h*(pdfTmp(Q_NE) + pdfTmp(Q_SW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_NE) - pdfTmp(Q_SW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_NE, iElem, QQ, neigh)) = pdfTmp(Q_NE) - sym - asym
      outstate(?SAVE?(Q_SW, iElem, QQ, neigh)) = pdfTmp(Q_SW) - sym + asym

      ui   = u_x - u_y
      sym  = omega_h*(pdfTmp(Q_SE) + pdfTmp(Q_NW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_SE) - pdfTmp(Q_NW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_SE, iElem, QQ, neigh)) = pdfTmp(Q_SE) - sym - asym
      outstate(?SAVE?(Q_NW, iElem, QQ, neigh)) = pdfTmp(Q_NW) - sym + asym

      ui   = u_x + u_z
      sym  = omega_h*(pdfTmp(Q_TE) + pdfTmp(Q_BW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TE) - pdfTmp(Q_BW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TE, iElem, QQ, neigh)) = pdfTmp(Q_TE) - sym - asym
      outstate(?SAVE?(Q_BW, iElem, QQ, neigh)) = pdfTmp(Q_BW) - sym + asym

      ui   = u_x - u_z
      sym  = omega_h*(pdfTmp(Q_BE) + pdfTmp(Q_TW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BE) - pdfTmp(Q_TW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BE, iElem, QQ, neigh)) = pdfTmp(Q_BE) - sym - asym
      outstate(?SAVE?(Q_TW, iElem, QQ, neigh)) = pdfTmp(Q_TW) - sym + asym

      ui   = u_y + u_z
      sym  = omega_h*(pdfTmp(Q_TN) + pdfTmp(Q_BS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TN) - pdfTmp(Q_BS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TN, iElem, QQ, neigh)) = pdfTmp(Q_TN) - sym - asym
      outstate(?SAVE?(Q_BS, iElem, QQ, neigh)) = pdfTmp(Q_BS) - sym + asym

      ui   = u_y - u_z
      sym  = omega_h*(pdfTmp(Q_BN) + pdfTmp(Q_TS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BN) - pdfTmp(Q_TS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BN, iElem, QQ, neigh)) = pdfTmp(Q_BN) - sym - asym
      outstate(?SAVE?(Q_TS, iElem, QQ, neigh)) = pdfTmp(Q_TS) - sym + asym

      t1x2 = t1x2_0 * rho
      fac1 = t1x2 * t2cs4inv !inv2csq2

      ui   = u_y
      sym  = omega_h*(pdfTmp(Q__N ) + pdfTmp(Q__S) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__N ) - pdfTmp(Q__S ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__N, iElem, QQ, neigh)) = pdfTmp(Q__N ) - sym - asym
      outstate(?SAVE?(Q__S, iElem, QQ, neigh)) = pdfTmp(Q__S ) - sym + asym

      ui   = u_x
      sym  = omega_h*(pdfTmp(Q__E ) + pdfTmp(Q__W) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__E ) - pdfTmp(Q__W ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__E, iElem, QQ, neigh)) = pdfTmp(Q__E ) - sym - asym
      outstate(?SAVE?(Q__W, iElem, QQ, neigh)) = pdfTmp(Q__W ) - sym + asym

      ui   = u_z
      sym  = omega_h*(pdfTmp(Q__T ) + pdfTmp(Q__B) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__T ) - pdfTmp(Q__B ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__T, iElem, QQ, neigh)) = pdfTmp(Q__T ) - sym - asym
      outstate(?SAVE?(Q__B, iElem, QQ, neigh)) = pdfTmp(Q__B ) - sym + asym
    enddo nodeloop

  end subroutine trt_advRel_d3q19_fixedVel
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with TRT collision
  !! operator
  !!
?? copy :: compute_routineHeader( trt_advRel_d3q19_les )
    !---------------------------------------------------------------------------
    integer :: iElem, iDir !< element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) !< temporary local pdf values
    real(kind=rk) :: rho     !< local density
    real(kind=rk) :: inv_rho !< inverse local density
    real(kind=rk) :: u_x     !< local x-velocity
    real(kind=rk) :: u_y     !< local y-velocity
    real(kind=rk) :: u_z     !< local z-velocity
    real(kind=rk) :: usq     !< square velocity
    ! derived constants
    real(kind=rk) :: ucx
    real(kind=rk) :: usqn
    real(kind=rk) :: omega, om, omega_h, asym_omega_h
    real(kind=rk) :: ui
    real(kind=rk) :: asym, sym, feq_common
    ! in case of the default model, the actual values depend on density
    real(kind=rk) ::  t0,t1x2,t2x2,fac1,fac2

    ! common prefactors for calculating the equilibrium parts (D3Q19 model)
    real(kind=rk), parameter :: t0_0 = 1.d0/3.d0
    ! twice the usual value due
    real(kind=rk), parameter :: t1x2_0 = 1.d0/18.d0 * 2.d0
    ! to the way of calculations
    real(kind=rk), parameter :: t2x2_0 = 1.d0/36.d0 * 2.d0
    ! Variables required for LES
    real(kind=rk) :: fEq(19) !< equilibrium distribution
    real(kind=rk) :: nEq(19) !< non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3)
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const
    !---------------------------------------------------------------------------
    ! read the relaxation parameter omega for the current level
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s
    tau0 = 1._rk/omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      !\todo MH: Replace this with Kannans equilibrium calculation derive%do
      ! This must be done outside the loop to allow for vectorized calculation 
      ! KM: We should compute equilibrium explicitly in compute kernel
      ! to improve performance
      ! Calculate the equilibrium distribution function
      ! square of velocity
      usq = u_x*u_x + u_y*u_y + u_z*u_z
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx =   real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk) * u_x       &
            & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk) * u_y       &
            & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk) * u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * rho * ( 1.d0 + ucx*cs2inv        &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv )

      enddo

      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )

      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)
      om_turb = 1._rk/(tau0+tau_turb)
      om = om_turb

      ! some global variables
      omega_h      = 0.5d0 * om
      asym_omega_h = 0.5d0 * 8.d0 * (2.d0-om)/(8.d0-om)

      ! square velocity and derived constants
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      feq_common = 1.d0  - 1.5d0 * usq
      t0 = t0_0 * rho

      ! let's start the relaxation process
      outstate(?SAVE?(Q__0, iElem, QQ, neigh)) =                               &
        &                            pdfTmp(Q__0 )*(1.d0-om) + om*t0*feq_common

      t2x2 = t2x2_0 * rho
      fac2 = t2x2 * t2cs4inv !inv2csq2

      ui   = u_x + u_y
      sym  = omega_h*(pdfTmp(Q_NE) + pdfTmp(Q_SW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_NE) - pdfTmp(Q_SW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_NE, iElem, QQ, neigh)) = pdfTmp(Q_NE) - sym - asym
      outstate(?SAVE?(Q_SW, iElem, QQ, neigh)) = pdfTmp(Q_SW) - sym + asym

      ui   = u_x - u_y
      sym  = omega_h*(pdfTmp(Q_SE) + pdfTmp(Q_NW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_SE) - pdfTmp(Q_NW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_SE, iElem, QQ, neigh)) = pdfTmp(Q_SE) - sym - asym
      outstate(?SAVE?(Q_NW, iElem, QQ, neigh)) = pdfTmp(Q_NW) - sym + asym

      ui   = u_x + u_z
      sym  = omega_h*(pdfTmp(Q_TE) + pdfTmp(Q_BW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TE) - pdfTmp(Q_BW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TE, iElem, QQ, neigh)) = pdfTmp(Q_TE) - sym - asym
      outstate(?SAVE?(Q_BW, iElem, QQ, neigh)) = pdfTmp(Q_BW) - sym + asym

      ui   = u_x - u_z
      sym  = omega_h*(pdfTmp(Q_BE) + pdfTmp(Q_TW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BE) - pdfTmp(Q_TW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BE, iElem, QQ, neigh)) = pdfTmp(Q_BE) - sym - asym
      outstate(?SAVE?(Q_TW, iElem, QQ, neigh)) = pdfTmp(Q_TW) - sym + asym

      ui   = u_y + u_z
      sym  = omega_h*(pdfTmp(Q_TN) + pdfTmp(Q_BS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TN) - pdfTmp(Q_BS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TN, iElem, QQ, neigh)) = pdfTmp(Q_TN) - sym - asym
      outstate(?SAVE?(Q_BS, iElem, QQ, neigh)) = pdfTmp(Q_BS) - sym + asym

      ui   = u_y - u_z
      sym  = omega_h*(pdfTmp(Q_BN) + pdfTmp(Q_TS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BN) - pdfTmp(Q_TS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BN, iElem, QQ, neigh)) = pdfTmp(Q_BN) - sym - asym
      outstate(?SAVE?(Q_TS, iElem, QQ, neigh)) = pdfTmp(Q_TS) - sym + asym

      t1x2 = t1x2_0 * rho
      fac1 = t1x2 * t2cs4inv !inv2csq2

      ui   = u_y
      sym  = omega_h*(pdfTmp(Q__N ) + pdfTmp(Q__S) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__N ) - pdfTmp(Q__S ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__N, iElem, QQ, neigh)) = pdfTmp(Q__N ) - sym - asym
      outstate(?SAVE?(Q__S, iElem, QQ, neigh)) = pdfTmp(Q__S ) - sym + asym

      ui   = u_x
      sym  = omega_h*(pdfTmp(Q__E ) + pdfTmp(Q__W) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__E ) - pdfTmp(Q__W ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__E, iElem, QQ, neigh)) = pdfTmp(Q__E ) - sym - asym
      outstate(?SAVE?(Q__W, iElem, QQ, neigh)) = pdfTmp(Q__W ) - sym + asym

      ui   = u_z
      sym  = omega_h*(pdfTmp(Q__T ) + pdfTmp(Q__B) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__T ) - pdfTmp(Q__B ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__T, iElem, QQ, neigh)) = pdfTmp(Q__T ) - sym - asym
      outstate(?SAVE?(Q__B, iElem, QQ, neigh)) = pdfTmp(Q__B ) - sym + asym
    enddo nodeloop

  end subroutine trt_advRel_d3q19_les
!******************************************************************************!



!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with TRT collision
  !! operator
  !!
?? copy :: compute_routineHeader( trt_advRel_d3q19_les_omegaSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem, iDir !< element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) !< temporary local pdf values
    real(kind=rk) :: rho     !< local density
    real(kind=rk) :: inv_rho !< inverse local density
    real(kind=rk) :: u_x     !< local x-velocity
    real(kind=rk) :: u_y     !< local y-velocity
    real(kind=rk) :: u_z     !< local z-velocity
    real(kind=rk) :: usq     !< square velocity
    ! derived constants
    real(kind=rk) :: ucx
    real(kind=rk) :: usqn
    real(kind=rk) :: omega, om, omega_h, asym_omega_h
    real(kind=rk) :: ui
    real(kind=rk) :: asym, sym, feq_common
    ! in case of the default model, the actual values depend on density
    real(kind=rk) ::  t0,t1x2,t2x2,fac1,fac2

    ! common prefactors for calculating the equilibrium parts (D3Q19 model)
    real(kind=rk), parameter :: t0_0 = 1.d0/3.d0
    ! twice the usual value due
    real(kind=rk), parameter :: t1x2_0 = 1.d0/18.d0 * 2.d0
    ! to the way of calculations
    real(kind=rk), parameter :: t2x2_0 = 1.d0/36.d0 * 2.d0
    ! Variables required for LES
    real(kind=rk) :: fEq(19) !< equilibrium distribution
    real(kind=rk) :: nEq(19) !< non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3)
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const, omSrc
    !---------------------------------------------------------------------------
    ! read the relaxation parameter omega for the current level
    omSrc   = fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s
    tau0 = 1._rk/omSrc

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      omega   = fieldProp(1)%fluid%viscSpatial%stored( level )%                &
        &                                                    val( iElem )*omSrc
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      !\todo MH: Replace this with Kannans equilibrium calculation derive%do
      ! This must be done outside the loop to allow for vectorized calculation 
      ! Calculate the equilibrium distribution function
      ! square of velocity
      usq = u_x*u_x + u_y*u_y + u_z*u_z
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx =   real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk) * u_x       &
            & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk) * u_y       &
            & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk) * u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * rho * ( 1.d0 + ucx*cs2inv        &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv )

      enddo

      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )

      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)
      om_turb = 1._rk/(tau0+tau_turb)
      om = om_turb

      ! some global variables
      omega_h      = 0.5d0 * om
      asym_omega_h = 0.5d0 * 8.d0 * (2.d0-om)/(8.d0-om)

      ! square velocity and derived constants
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      feq_common = 1.d0  - 1.5d0 * usq
      t0 = t0_0 * rho

      ! let's start the relaxation process
      outstate(?SAVE?(Q__0, iElem, QQ, neigh)) =                               &
        &                            pdfTmp(Q__0 )*(1.d0-om) + om*t0*feq_common

      t2x2 = t2x2_0 * rho
      fac2 = t2x2 * t2cs4inv !inv2csq2

      ui   = u_x + u_y
      sym  = omega_h*(pdfTmp(Q_NE) + pdfTmp(Q_SW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_NE) - pdfTmp(Q_SW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_NE, iElem, QQ, neigh)) = pdfTmp(Q_NE) - sym - asym
      outstate(?SAVE?(Q_SW, iElem, QQ, neigh)) = pdfTmp(Q_SW) - sym + asym

      ui   = u_x - u_y
      sym  = omega_h*(pdfTmp(Q_SE) + pdfTmp(Q_NW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_SE) - pdfTmp(Q_NW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_SE, iElem, QQ, neigh)) = pdfTmp(Q_SE) - sym - asym
      outstate(?SAVE?(Q_NW, iElem, QQ, neigh)) = pdfTmp(Q_NW) - sym + asym

      ui   = u_x + u_z
      sym  = omega_h*(pdfTmp(Q_TE) + pdfTmp(Q_BW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TE) - pdfTmp(Q_BW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TE, iElem, QQ, neigh)) = pdfTmp(Q_TE) - sym - asym
      outstate(?SAVE?(Q_BW, iElem, QQ, neigh)) = pdfTmp(Q_BW) - sym + asym

      ui   = u_x - u_z
      sym  = omega_h*(pdfTmp(Q_BE) + pdfTmp(Q_TW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BE) - pdfTmp(Q_TW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BE, iElem, QQ, neigh)) = pdfTmp(Q_BE) - sym - asym
      outstate(?SAVE?(Q_TW, iElem, QQ, neigh)) = pdfTmp(Q_TW) - sym + asym

      ui   = u_y + u_z
      sym  = omega_h*(pdfTmp(Q_TN) + pdfTmp(Q_BS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TN) - pdfTmp(Q_BS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TN, iElem, QQ, neigh)) = pdfTmp(Q_TN) - sym - asym
      outstate(?SAVE?(Q_BS, iElem, QQ, neigh)) = pdfTmp(Q_BS) - sym + asym

      ui   = u_y - u_z
      sym  = omega_h*(pdfTmp(Q_BN) + pdfTmp(Q_TS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BN) - pdfTmp(Q_TS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BN, iElem, QQ, neigh)) = pdfTmp(Q_BN) - sym - asym
      outstate(?SAVE?(Q_TS, iElem, QQ, neigh)) = pdfTmp(Q_TS) - sym + asym

      t1x2 = t1x2_0 * rho
      fac1 = t1x2 * t2cs4inv !inv2csq2

      ui   = u_y
      sym  = omega_h*(pdfTmp(Q__N ) + pdfTmp(Q__S) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__N ) - pdfTmp(Q__S ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__N, iElem, QQ, neigh)) = pdfTmp(Q__N ) - sym - asym
      outstate(?SAVE?(Q__S, iElem, QQ, neigh)) = pdfTmp(Q__S ) - sym + asym

      ui   = u_x
      sym  = omega_h*(pdfTmp(Q__E ) + pdfTmp(Q__W) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__E ) - pdfTmp(Q__W ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__E, iElem, QQ, neigh)) = pdfTmp(Q__E ) - sym - asym
      outstate(?SAVE?(Q__W, iElem, QQ, neigh)) = pdfTmp(Q__W ) - sym + asym

      ui   = u_z
      sym  = omega_h*(pdfTmp(Q__T ) + pdfTmp(Q__B) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__T ) - pdfTmp(Q__B ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__T, iElem, QQ, neigh)) = pdfTmp(Q__T ) - sym - asym
      outstate(?SAVE?(Q__B, iElem, QQ, neigh)) = pdfTmp(Q__B ) - sym + asym
    enddo nodeloop

  end subroutine trt_advRel_d3q19_les_omegaSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with TRT collision
  !! operator.
  !!
  !! Incompressible model
  !!
?? copy :: compute_routineHeader( trt_advRel_d3q19_incomp )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! derived constants
    real(kind=rk) usqn
    real(kind=rk) omega_2, cmpl_o, omega, om, omega_h, asym_omega_h
    real(kind=rk) ui
    real(kind=rk) :: asym, sym, feq_common

    ! common prefactors for calculating the equilibrium parts (D3Q19 model)
    real(kind=rk), parameter :: t0_0 = 1.d0/3.d0
    ! twice the usual value due
    real(kind=rk), parameter :: t1x2_0 = 1.d0/18.d0 * 2.d0
    ! to the way of calculations
    real(kind=rk), parameter :: t2x2_0 = 1.d0/36.d0 * 2.d0

    ! in case of the default model, the actual values depend on density
    real(kind=rk), parameter :: t0 = t0_0
    real(kind=rk), parameter :: t1x2 = t1x2_0
    real(kind=rk), parameter :: t2x2 = t2x2_0
    real(kind=rk), parameter :: fac1 = t1x2*t2cs4inv !inv2csq2
    real(kind=rk), parameter :: fac2 = t2x2*t2cs4inv !inv2csq2
    !---------------------------------------------------------------------------

    ! read the relaxation parameter omega for the current level
    omega   = fieldProp(1)%fluid%rampOmegas( level )
    ! pre-calculate partial collision constants
    omega_2 = 2._rk * omega
    cmpl_o  = 1._rk - omega

    ! some global variables
    omega_h      = 0.5d0 * omega
!#ifdef TRT_ASYM_1
!    asym_omega_h = 0.5d0 * asym_omega
!#else
!#ifdef TRT_BGK
!    ! check that BGK is recovered; o.k. works fine
!    asym_omega_h = 0.5d0 * omega
!#else
!    ! magic Irina for fixed walls
    asym_omega_h = 0.5d0 * 8.d0 * (2.d0-omega)/(8.d0-omega)
!#endif   /* not TRT_BGK */
!#endif   /* not TRT_ASYM_1 */
    om = omega

!$omp do schedule(static)

    nodeloop: do iElem=nStart,nEnd       
      ! First load all local values into temp array
      ! Generic! PUSH+PULL is possible
      ! Flexible array access
      ! pdfPos = scheme%pdfPos 
      ! pdfTmp( Q__W ) = inState(?FETCH?( sys%Variable%val(pdfPos)%varPos(Q__W), iElem, layout%stencil(1)%QQ, neigh ))
      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, QQ, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, QQ, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, QQ, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, QQ, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, QQ, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, QQ, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, QQ, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, QQ, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, QQ, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, QQ, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, QQ, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, QQ, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, QQ, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, QQ, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, QQ, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, QQ, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, QQ, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, QQ, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, QQ, neigh ))
 
 
      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)
                         
      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)
                         
      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      feq_common = rho - 1.5d0 * usq

      ! let's start the relaxation process
      outstate(?SAVE?(Q__0, iElem, QQ, neigh)) =                               &
        &                            pdfTmp(Q__0 )*(1.d0-om) + om*t0*feq_common

      ui   = u_x + u_y
      sym  = omega_h*(pdfTmp(Q_NE) + pdfTmp(Q_SW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_NE) - pdfTmp(Q_SW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_NE, iElem, QQ, neigh)) = pdfTmp(Q_NE) - sym - asym
      outstate(?SAVE?(Q_SW, iElem, QQ, neigh)) = pdfTmp(Q_SW) - sym + asym

      ui   = u_x - u_y
      sym  = omega_h*(pdfTmp(Q_SE) + pdfTmp(Q_NW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_SE) - pdfTmp(Q_NW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_SE, iElem, QQ, neigh)) = pdfTmp(Q_SE) - sym - asym
      outstate(?SAVE?(Q_NW, iElem, QQ, neigh)) = pdfTmp(Q_NW) - sym + asym

      ui   = u_x + u_z
      sym  = omega_h*(pdfTmp(Q_TE) + pdfTmp(Q_BW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TE) - pdfTmp(Q_BW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TE, iElem, QQ, neigh)) = pdfTmp(Q_TE) - sym - asym
      outstate(?SAVE?(Q_BW, iElem, QQ, neigh)) = pdfTmp(Q_BW) - sym + asym

      ui   = u_x - u_z
      sym  = omega_h*(pdfTmp(Q_BE) + pdfTmp(Q_TW) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BE) - pdfTmp(Q_TW) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BE, iElem, QQ, neigh)) = pdfTmp(Q_BE) - sym - asym
      outstate(?SAVE?(Q_TW, iElem, QQ, neigh)) = pdfTmp(Q_TW) - sym + asym

      ui   = u_y + u_z
      sym  = omega_h*(pdfTmp(Q_TN) + pdfTmp(Q_BS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_TN) - pdfTmp(Q_BS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_TN, iElem, QQ, neigh)) = pdfTmp(Q_TN) - sym - asym
      outstate(?SAVE?(Q_BS, iElem, QQ, neigh)) = pdfTmp(Q_BS) - sym + asym

      ui   = u_y - u_z
      sym  = omega_h*(pdfTmp(Q_BN) + pdfTmp(Q_TS) - fac2*ui*ui-t2x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q_BN) - pdfTmp(Q_TS) - 3.d0*t2x2*ui )
      outstate(?SAVE?(Q_BN, iElem, QQ, neigh)) = pdfTmp(Q_BN) - sym - asym
      outstate(?SAVE?(Q_TS, iElem, QQ, neigh)) = pdfTmp(Q_TS) - sym + asym

      ui   = u_y
      sym  = omega_h*(pdfTmp(Q__N ) + pdfTmp(Q__S) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__N ) - pdfTmp(Q__S ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__N, iElem, QQ, neigh)) = pdfTmp(Q__N ) - sym - asym
      outstate(?SAVE?(Q__S, iElem, QQ, neigh)) = pdfTmp(Q__S ) - sym + asym

      ui   = u_x
      sym  = omega_h*(pdfTmp(Q__E ) + pdfTmp(Q__W) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__E ) - pdfTmp(Q__W ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__E, iElem, QQ, neigh)) = pdfTmp(Q__E ) - sym - asym
      outstate(?SAVE?(Q__W, iElem, QQ, neigh)) = pdfTmp(Q__W ) - sym + asym

      ui   = u_z
      sym  = omega_h*(pdfTmp(Q__T ) + pdfTmp(Q__B) - fac1*ui*ui-t1x2*feq_common)
      asym = asym_omega_h*( pdfTmp(Q__T ) - pdfTmp(Q__B ) - 3.d0*t1x2*ui )
      outstate(?SAVE?(Q__T, iElem, QQ, neigh)) = pdfTmp(Q__T ) - sym - asym
      outstate(?SAVE?(Q__B, iElem, QQ, neigh)) = pdfTmp(Q__B ) - sym + asym
    enddo nodeloop

  end subroutine trt_advRel_d3q19_incomp
!******************************************************************************!

!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK and nonNewtonian
  !! power-law (PL) model. \n
  !! After BGK scheme, strain rate is calculated from nonEquilibilirium.
  !! Then shear rate is computed by strain rate. Next, viscosity is updated
  !! according to nonNewtonian model. Finally viscosity is converted into omega
  !! and updated.
  !!
  !! \author: Jiaxing Qi
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_nNwtn_PL )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho ! inverse local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    real(kind=rk) :: fEq__0
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2
    real(kind=rk) :: omega_2, cmpl_o, omega, inv_omg_2, coeffSR
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &              sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1,        &
      &              sum11_2, sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------
    ! parameters for non-Newtonian
    integer :: omegaPos, nScalars
    real(kind=rk) :: fNeq_NE_SW, fNeq_NW_SE, fNeq_BW_TE, &
      &              fNeq_BE_TW, fNeq_BS_TN, fNeq_BN_TS
    real(kind=rk) :: strainRate(6)
    real(kind=rk) :: shearRate ! second invariant of strainRate
    real(kind=rk) :: nuLB
    real(kind=rk) :: convSR(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convVisc(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: k, n_1 ! parameters for power-law
    !---------------------------------------------------------------------------

    omegaPos = layout%stencil(1)%QQ + 1
    nScalars = varSys%nScalars

    ! conversion factor for Shear Rate (or Strain Rate)
    convSR(:) = getConversionFac( &
      &           conf = varSys%solSpec%conf, &
      &           facName = 'strainRate', &
      &           nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! conversion factor for kinematic viscosity
    convVisc(:) = getConversionFac( &
      &             conf = varSys%solSpec%conf, &
      &             facName = 'visc', &
      &             nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! get nonNewtonian power law parameter
    k   = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%PL%k &
      &   / params%physics%rho0 / convVisc(level)
    n_1 = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%PL%n_1

!$omp do schedule(static)

    nodeloop: do iElem = nStart, nEnd

      ! read the relaxation parameter omega for the current element
      omega   = inState( ?IDX?(omegaPos, iElem, nScalars) )
      inv_omg_2 = 2._rk / omega

      ! pre-calculate partial collision constants
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega

      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, nScalars, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, nScalars, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, nScalars, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, nScalars, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, nScalars, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, nScalars, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, nScalars, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, nScalars, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, nScalars, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, nScalars, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, nScalars, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, nScalars, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, nScalars, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, nScalars, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, nScalars, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, nScalars, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, nScalars, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, nScalars, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, nScalars, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho
      coeffSR = -1.5_rk * omega * convSR(level) / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      fEq__0 = rho*(div1_3-0.5_rk*usq)

      ! f = (1-omega) * f + omega * fEq
      outState(?SAVE?(Q__0,iElem,nScalars,neigh)) =                       &
        &                                pdfTmp(Q__0)*cmpl_o + omega*fEq__0

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     = u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_NE)*cmpl_o +sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_SW)*cmpl_o -sum1_1 +sum1_2

      ! the term (omg*fEq) is already embedded in (sum1_1,sum1_2)
      ! so we devide (omg*fEq) by omg to get fNeq.
      ! This also applied to all other links.
      ! fNeq_NE + fNeq_SW
      fNeq_NE_SW = pdfTmp(Q_NE) + pdfTmp(Q_SW) - sum1_2*inv_omg_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      fNeq_NW_SE = pdfTmp(Q_NW) + pdfTmp(Q_SE) - sum3_2*inv_omg_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      fNeq_BW_TE = pdfTmp(Q_BW) + pdfTmp(Q_TE) - sum10_2*inv_omg_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      fNeq_BE_TW = pdfTmp(Q_BE) + pdfTmp(Q_TW) - sum12_2*inv_omg_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      fNeq_BS_TN = pdfTmp(Q_BS) + pdfTmp(Q_TN) - sum11_2*inv_omg_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      fNeq_BN_TS = pdfTmp(Q_BN) + pdfTmp(Q_TS) - sum13_2*inv_omg_2

      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      strainRate(2) = pdfTmp(Q__N) + pdfTmp(Q__S) - sum2_2*inv_omg_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      strainRate(1) = pdfTmp(Q__W) + pdfTmp(Q__E) - sum4_2*inv_omg_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__T)*cmpl_o +sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__B)*cmpl_o -sum9_1 +sum9_2

      strainRate(3) = pdfTmp(Q__T) + pdfTmp(Q__B) - sum9_2*inv_omg_2

      ! Calculate strain rate and convert physics unit
      ! s1 = (W+E) + (SW+NE) + (SE+NW) + (BW+TE) + (BE+TW)
      strainRate(1) =   strainRate(1) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BW_TE + fNeq_BE_TW

      ! s2 = (S+N) + (NE+SW) + (NW+SE) + (BS+TN) + (BN+TS)
      strainRate(2) =   strainRate(2) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BS_TN + fNeq_BN_TS

      ! s3 = (B+T) + (BS+TN) + (BN+TS) + (BW+TE) + (BE+TW)
      strainRate(3) =   strainRate(3) &
        &             + fNeq_BS_TN + fNeq_BN_TS + fNeq_BW_TE + fNeq_BE_TW

      ! s4 = (SW+NE) - (NW+SE)
      strainRate(4) = fNeq_NE_SW - fNeq_NW_SE

      ! s5 = (BS+TN) - (TS+BN)
      strainRate(5) = fNeq_BS_TN - fNeq_BN_TS

      ! s6 = (BW+TE) - (BE+TW)
      strainRate(6) = fNeq_BW_TE - fNeq_BE_TW

      ! coeffSR can NOT be moved into the following SQRT
      strainRate = strainRate * coeffSR

      !! Calculate physical shear rate
      shearRate = 2._rk * sqrt( sum( strainRate * strainRate ) )

      ! the exponential operation is quite expensive.
      ! it decreases performance about 3-4 MLUPS on SHU
      ! calculate kinematic visc by POWER-LAW model
      ! from paper: LBMs for non-newtonian flows, page 803
      ! k is actually k / rho0 / convVisc
      nuLB = ( shearRate ** n_1 ) * k

      ! calc omage and save state vector
outState( ?IDX?(omegaPos, iElem, nScalars) ) = 1._rk / ( 0.5_rk + 3._rk*nuLB )

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_nNwtn_PL
!******************************************************************************!

!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK and nonNewtonian
  !! Carreau-Yasuda (CY) model. \n
  !! After BGK scheme, strain rate is calculated from nonEquilibilirium.
  !! Then shear rate is computed by strain rate. Next, viscosity is updated
  !! according to nonNewtonian model. Finally viscosity is converted into omega
  !! and updated.
  !!
  !! \author: Jiaxing Qi
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_nNwtn_CY )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho ! inverse local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    real(kind=rk) :: fEq__0
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2
    real(kind=rk) :: omega_2, cmpl_o, omega, inv_omg_2, coeffSR
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &              sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1,        &
      &              sum11_2, sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------
    ! parameters for non-Newtonian
    integer :: omegaPos, nScalars
    real(kind=rk) :: fNeq_NE_SW, fNeq_NW_SE, fNeq_BW_TE, &
      &              fNeq_BE_TW, fNeq_BS_TN, fNeq_BN_TS
    real(kind=rk) :: strainRate(6)
    real(kind=rk) :: shearRate ! second invariant of strainRate
    real(kind=rk) :: nuLB, t
    real(kind=rk) :: convSR(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convVisc(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: a, n_1_a, lambda, v0_vInf, vInf ! parameters for CY model
    !---------------------------------------------------------------------------

    omegaPos = layout%stencil(1)%QQ + 1
    nScalars = varSys%nScalars

    ! conversion factor for Shear Rate (or Strain Rate)
    convSR(:) = getConversionFac( &
      &           conf = varSys%solSpec%conf, &
      &           facName = 'strainRate', &
      &           nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! conversion factor for kinematic viscosity
    convVisc(:) = getConversionFac( &
      &             conf = varSys%solSpec%conf, &
      &             facName = 'visc', &
      &             nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! get nonNewtonian power law parameter
    a       = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%a
    n_1_a   = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%n_1_a
    lambda  = 2._rk * scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%lambda
    vInf    = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%viscInf
    v0_vInf = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%visc0 - vInf

    vInf = vInf / params%physics%rho0 / convVisc(level)
    v0_vInf = v0_vInf / params%physics%rho0 / convVisc(level)

!$omp do schedule(static)

    nodeloop: do iElem = nStart, nEnd

      ! read the relaxation parameter omega for the current element
      omega   = inState( ?IDX?(omegaPos, iElem, nScalars) )
      inv_omg_2 = 2._rk / omega

      ! pre-calculate partial collision constants
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega

      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, nScalars, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, nScalars, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, nScalars, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, nScalars, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, nScalars, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, nScalars, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, nScalars, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, nScalars, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, nScalars, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, nScalars, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, nScalars, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, nScalars, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, nScalars, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, nScalars, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, nScalars, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, nScalars, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, nScalars, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, nScalars, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, nScalars, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho
      coeffSR = -1.5_rk * omega * convSR(level) / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      fEq__0 = rho*(div1_3-0.5_rk*usq)

      ! f = (1-omega) * f + omega * fEq
      outState(?SAVE?(Q__0,iElem,nScalars,neigh)) =                       &
        &                                pdfTmp(Q__0)*cmpl_o + omega*fEq__0

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     = u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      fNeq_NE_SW = pdfTmp(Q_NE) + pdfTmp(Q_SW) - sum1_2*inv_omg_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      fNeq_NW_SE = pdfTmp(Q_NW) + pdfTmp(Q_SE) - sum3_2*inv_omg_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      fNeq_BW_TE = pdfTmp(Q_BW) + pdfTmp(Q_TE) - sum10_2*inv_omg_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      fNeq_BE_TW = pdfTmp(Q_BE) + pdfTmp(Q_TW) - sum12_2*inv_omg_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_TN)*cmpl_o +sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_BS)*cmpl_o -sum11_1+sum11_2

      fNeq_BS_TN = pdfTmp(Q_BS) + pdfTmp(Q_TN) - sum11_2*inv_omg_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_TS)*cmpl_o +sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_BN)*cmpl_o -sum13_1+sum13_2

      fNeq_BN_TS = pdfTmp(Q_BN) + pdfTmp(Q_TS) - sum13_2*inv_omg_2

      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__N)*cmpl_o +sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__S)*cmpl_o -sum2_1 +sum2_2

      strainRate(2) = pdfTmp(Q__N) + pdfTmp(Q__S) - sum2_2*inv_omg_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__W)*cmpl_o -sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__E)*cmpl_o +sum4_1 +sum4_2

      strainRate(1) = pdfTmp(Q__W) + pdfTmp(Q__E) - sum4_2*inv_omg_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__T)*cmpl_o +sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__B)*cmpl_o -sum9_1 +sum9_2

      strainRate(3) = pdfTmp(Q__T) + pdfTmp(Q__B) - sum9_2*inv_omg_2

      ! s1 = (W+E) + (SW+NE) + (SE+NW) + (BW+TE) + (BE+TW)
      strainRate(1) =   strainRate(1) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BW_TE + fNeq_BE_TW

      ! s2 = (S+N) + (NE+SW) + (NW+SE) + (BS+TN) + (BN+TS)
      strainRate(2) =   strainRate(2) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BS_TN + fNeq_BN_TS

      ! s3 = (B+T) + (BS+TN) + (BN+TS) + (BW+TE) + (BE+TW)
      strainRate(3) =   strainRate(3) &
        &             + fNeq_BS_TN + fNeq_BN_TS + fNeq_BW_TE + fNeq_BE_TW

      ! s4 = (SW+NE) - (NW+SE)
      strainRate(4) = fNeq_NE_SW - fNeq_NW_SE

      ! s5 = (BS+TN) - (TS+BN)
      strainRate(5) = fNeq_BS_TN - fNeq_BN_TS

      ! s6 = (BW+TE) - (BE+TW)
      strainRate(6) = fNeq_BW_TE - fNeq_BE_TW

      ! convert physics unit
      strainRate = strainRate * coeffSR

      !! Calculate physical shear rate
      shearRate = sqrt( sum( strainRate * strainRate ) )

      ! calculate dynamic visc by Carreau-Yasuda model
      ! convert to kinematic visc
      t = ( 1._rk + (lambda*shearRate) ** a ) ** n_1_a
      nuLB = vInf + v0_vInf * t

      ! calc omage and save state vector
  outState( ?IDX?(omegaPos, iElem, nScalars) ) = 1._rk / ( 0.5_rk + 3._rk*nuLB )

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_nNwtn_CY
!******************************************************************************!

!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK and nonNewtonian
  !! Casson (CS) model. \n
  !! After BGK scheme, strain rate is calculated from nonEquilibilirium.
  !! Then shear rate is computed by strain rate. Next, viscosity is updated
  !! according to nonNewtonian model. Finally viscosity is converted into omega
  !! and updated.
  !!
  !! \author: Jiaxing Qi
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_nNwtn_CS )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho ! inverse local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    real(kind=rk) :: fEq__0
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2
    real(kind=rk) :: omega_2, cmpl_o, omega, inv_omg_2, coeffSR
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &              sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1,        &
      &              sum11_2, sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------
    ! parameters for non-Newtonian
    integer :: omegaPos, nScalars
    real(kind=rk) :: fNeq_NE_SW, fNeq_NW_SE, fNeq_BW_TE, &
      &              fNeq_BE_TW, fNeq_BS_TN, fNeq_BN_TS
    real(kind=rk) :: strainRate(6)
    real(kind=rk) :: shearRate ! second invariant of strainRate
    real(kind=rk) :: nuLB, t
    real(kind=rk) :: convSR(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convVisc(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: k0, k1 ! parameters for Casson model
    !---------------------------------------------------------------------------

    omegaPos = layout%stencil(1)%QQ + 1
    nScalars = varSys%nScalars

    ! conversion factor for Shear Rate (or Strain Rate)
    convSR(:) = getConversionFac( &
      &           conf = varSys%solSpec%conf, &
      &           facName = 'strainRate', &
      &           nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! conversion factor for kinematic viscosity
    convVisc(:) = getConversionFac( &
      &             conf = varSys%solSpec%conf, &
      &             facName = 'visc', &
      &             nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! get nonNewtonian Casson model parameter
    k0 = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CS%k0
    k1 = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CS%k1

!$omp do schedule(static)

    nodeloop: do iElem = nStart, nEnd

      ! read the relaxation parameter omega for the current element
      omega   = inState( ?IDX?(omegaPos, iElem, nScalars) )
      inv_omg_2 = 2._rk / omega

      ! pre-calculate partial collision constants
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega

      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, nScalars, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, nScalars, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, nScalars, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, nScalars, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, nScalars, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, nScalars, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, nScalars, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, nScalars, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, nScalars, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, nScalars, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, nScalars, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, nScalars, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, nScalars, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, nScalars, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, nScalars, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, nScalars, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, nScalars, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, nScalars, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, nScalars, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW) &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S) &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T) &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW) &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE) &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS) &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)  &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                 &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)  &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW) &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                 &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE) &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)  &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                 &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)  &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! inverse local density
      inv_rho = 1._rk / rho
      coeffSR = -1.5_rk * omega * convSR(level) / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (1._rk - 1.5_rk * usq) * rho
      fEq__0 = rho*(div1_3-0.5_rk*usq)

      ! f = (1-omega) * f + omega * fEq
      outState(?SAVE?(Q__0,iElem,nScalars,neigh)) =                       &
        &                                pdfTmp(Q__0)*cmpl_o + omega*fEq__0

      coeff_1 = div1_8 * omega * rho

      usqn_o1 = omega * usqn

      ui1     = u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      fNeq_NE_SW = pdfTmp(Q_NE) + pdfTmp(Q_SW) - sum1_2*inv_omg_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      fNeq_NW_SE = pdfTmp(Q_NW) + pdfTmp(Q_SE) - sum3_2*inv_omg_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      fNeq_BW_TE = pdfTmp(Q_BW) + pdfTmp(Q_TE) - sum10_2*inv_omg_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,nScalars,neigh)) =                           &
        &                                   pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      fNeq_BE_TW = pdfTmp(Q_BE) + pdfTmp(Q_TW) - sum12_2*inv_omg_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_TN)*cmpl_o +sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_BS)*cmpl_o -sum11_1+sum11_2

      fNeq_BS_TN = pdfTmp(Q_BS) + pdfTmp(Q_TN) - sum11_2*inv_omg_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_TS)*cmpl_o +sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,nScalars,neigh)) =                            &
        &                                   pdfTmp(Q_BN)*cmpl_o -sum13_1+sum13_2

      fNeq_BN_TS = pdfTmp(Q_BN) + pdfTmp(Q_TS) - sum13_2*inv_omg_2

      coeff_2 = div1_8 * omega_2 * rho
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__N)*cmpl_o +sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__S)*cmpl_o -sum2_1 +sum2_2

      strainRate(2) = pdfTmp(Q__N) + pdfTmp(Q__S) - sum2_2*inv_omg_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__W)*cmpl_o -sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__E)*cmpl_o +sum4_1 +sum4_2

      strainRate(1) = pdfTmp(Q__W) + pdfTmp(Q__E) - sum4_2*inv_omg_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__T)*cmpl_o +sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,nScalars,neigh )) =                          &
        &                                   pdfTmp(Q__B)*cmpl_o -sum9_1 +sum9_2

      strainRate(3) = pdfTmp(Q__T) + pdfTmp(Q__B) - sum9_2*inv_omg_2

      ! s1 = (W+E) + (SW+NE) + (SE+NW) + (BW+TE) + (BE+TW)
      strainRate(1) =   strainRate(1) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BW_TE + fNeq_BE_TW

      ! s2 = (S+N) + (NE+SW) + (NW+SE) + (BS+TN) + (BN+TS)
      strainRate(2) =   strainRate(2) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BS_TN + fNeq_BN_TS

      ! s3 = (B+T) + (BS+TN) + (BN+TS) + (BW+TE) + (BE+TW)
      strainRate(3) =   strainRate(3) &
        &             + fNeq_BS_TN + fNeq_BN_TS + fNeq_BW_TE + fNeq_BE_TW

      ! s4 = (SW+NE) - (NW+SE)
      strainRate(4) = fNeq_NE_SW - fNeq_NW_SE

      ! s5 = (BS+TN) - (TS+BN)
      strainRate(5) = fNeq_BS_TN - fNeq_BN_TS

      ! s6 = (BW+TE) - (BE+TW)
      strainRate(6) = fNeq_BW_TE - fNeq_BE_TW

      ! convert strain rate to physics unit
      strainRate = strainRate * coeffSR

      ! Calculate physical shear rate
      shearRate = 2._rk * sqrt( sum( strainRate * strainRate ) )

      ! calculate dynamic visc by Carreau-Yasuda model
      ! this equation could be simplified to k0^2/shear + k1^2 + 2k1/sqrt(shear)
      t = ( k0 + k1 * sqrt(shearRate) )
      nuLB = t * t / shearRate
      ! convert to kinematic visc
      nuLB = nuLB / params%physics%rho0 / convVisc(level)

      ! calc omage and save state vector
  outState( ?IDX?(omegaPos, iElem, nScalars) ) = 1._rk / ( 0.5_rk + 3._rk*nuLB )

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_nNwtn_CS
!******************************************************************************!

!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK for
  !! incompressible lbm model
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_incomp_nNwtn_CY )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho0 ! inverse local density
    real(kind=rk) :: rho0    ! reference density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    real(kind=rk) :: fEq__0
    ! derived constants
    real(kind=rk) usqn, usqn_o1, usqn_o2
    real(kind=rk) omega_2, cmpl_o, omega, inv_omg_2, coeffSR
    real(kind=rk) coeff_1, coeff_2
    real(kind=rk) ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,   &
      &           fac_13
    real(kind=rk) sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,      &
      &           sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1, sum11_2,  &
      &           sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------
    ! parameters for non-Newtonian
    integer :: omegaPos, nScalars
    real(kind=rk) :: fNeq_NE_SW, fNeq_NW_SE, fNeq_BW_TE, &
      &              fNeq_BE_TW, fNeq_BS_TN, fNeq_BN_TS
    real(kind=rk) :: strainRate(6)
    real(kind=rk) :: shearRate ! second invariant of strainRate
    real(kind=rk) :: nuLB, t
    real(kind=rk) :: convSR(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convVisc(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: a, n_1_a, lambda, v0_vInf, vInf ! parameters for CY model
    !---------------------------------------------------------------------------

    omegaPos = layout%stencil(1)%QQ + 1
    nScalars = varSys%nScalars

    ! conversion factor for Shear Rate (or Strain Rate)
    convSR(:) = getConversionFac( &
      &           conf = varSys%solSpec%conf, &
      &           facName = 'strainRate', &
      &           nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! conversion factor for kinematic viscosity
    convVisc(:) = getConversionFac( &
      &             conf = varSys%solSpec%conf, &
      &             facName = 'visc', &
      &             nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! get nonNewtonian power law parameter
    a       = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%a
    n_1_a   = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%n_1_a
    lambda  = 2._rk * scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%lambda
    vInf    = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%viscInf
    v0_vInf = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CY%visc0 - vInf

    vInf = vInf / params%physics%rho0 / convVisc(level)
    v0_vInf = v0_vInf / params%physics%rho0 / convVisc(level)

    rho0 = fieldProp(1)%fluid%rho0
    ! inverse local density
    inv_rho0 = 1._rk / rho0

!$omp do schedule(static)

    nodeloop: do iElem = nStart, nEnd
      ! read the relaxation parameter omega for the current element
      omega   = inState( ?IDX?(omegaPos, iElem, nScalars) )
      inv_omg_2 = 2._rk / omega
      coeffSR = -1.5_rk * omega * convSR(level) * inv_rho0

      ! pre-calculate partial collision constants
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega

      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, nScalars, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, nScalars, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, nScalars, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, nScalars, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, nScalars, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, nScalars, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, nScalars, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, nScalars, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, nScalars, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, nScalars, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, nScalars, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, nScalars, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, nScalars, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, nScalars, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, nScalars, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, nScalars, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, nScalars, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, nScalars, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, nScalars, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW)                         &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S)                         &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW)                         &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE)                         &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS)                         &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E)                         &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                                        &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW)                         &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)                        &
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                                        &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S ) - pdfTmp(Q_SE)                        &
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN)                         &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                                        &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN)                         &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! transfer moments to velocities
      u_x = u_x * inv_rho0
      u_y = u_y * inv_rho0
      u_z = u_z * inv_rho0

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (rho - 1.5_rk * usq * rho0)
      fEq__0 = rho*div1_3 - 0.5_rk*usq*rho0

      outState(?SAVE?(Q__0,iElem,nScalars,neigh)) = pdfTmp(Q__0)*cmpl_o + omega*fEq__0

      coeff_1 = div1_8 * omega * rho0

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,nScalars,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,nScalars,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      fNeq_NE_SW = pdfTmp(Q_NE) + pdfTmp(Q_SW) - sum1_2*inv_omg_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,nScalars,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,nScalars,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      fNeq_NW_SE = pdfTmp(Q_NW) + pdfTmp(Q_SE) - sum3_2*inv_omg_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,nScalars,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,nScalars,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      fNeq_BW_TE = pdfTmp(Q_BW) + pdfTmp(Q_TE) - sum10_2*inv_omg_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,nScalars,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,nScalars,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      fNeq_BE_TW = pdfTmp(Q_BE) + pdfTmp(Q_TW) - sum12_2*inv_omg_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,nScalars,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,nScalars,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      fNeq_BS_TN = pdfTmp(Q_BS) + pdfTmp(Q_TN) - sum11_2*inv_omg_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,nScalars,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,nScalars,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      fNeq_BN_TS = pdfTmp(Q_BN) + pdfTmp(Q_TS) - sum13_2*inv_omg_2

      coeff_2 = div1_8 * omega_2 * rho0
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,nScalars,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,nScalars,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      strainRate(2) = pdfTmp(Q__N) + pdfTmp(Q__S) - sum2_2*inv_omg_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,nScalars,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,nScalars,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      strainRate(1) = pdfTmp(Q__W) + pdfTmp(Q__E) - sum4_2*inv_omg_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,nScalars,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,nScalars,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

      strainRate(3) = pdfTmp(Q__T) + pdfTmp(Q__B) - sum9_2*inv_omg_2

      ! s1 = (W+E) + (SW+NE) + (SE+NW) + (BW+TE) + (BE+TW)
      strainRate(1) =   strainRate(1) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BW_TE + fNeq_BE_TW

      ! s2 = (S+N) + (NE+SW) + (NW+SE) + (BS+TN) + (BN+TS)
      strainRate(2) =   strainRate(2) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BS_TN + fNeq_BN_TS

      ! s3 = (B+T) + (BS+TN) + (BN+TS) + (BW+TE) + (BE+TW)
      strainRate(3) =   strainRate(3) &
        &             + fNeq_BS_TN + fNeq_BN_TS + fNeq_BW_TE + fNeq_BE_TW

      ! s4 = (SW+NE) - (NW+SE)
      strainRate(4) = fNeq_NE_SW - fNeq_NW_SE

      ! s5 = (BS+TN) - (TS+BN)
      strainRate(5) = fNeq_BS_TN - fNeq_BN_TS

      ! s6 = (BW+TE) - (BE+TW)
      strainRate(6) = fNeq_BW_TE - fNeq_BE_TW

      strainRate = strainRate * coeffSR

      ! Calculate physical shear rate
      shearRate = sqrt( sum( strainRate * strainRate ) )

      ! calculate dynamic visc by Carreau-Yasuda model
      ! convert to kinematic visc
      t = ( 1._rk + (lambda*shearRate) ** a ) ** n_1_a
      nuLB = vInf + v0_vInf * t

      ! calc omage and save state vector
outState( ?IDX?(omegaPos, iElem, nScalars) ) = 1._rk / ( 0.5_rk + 3._rk*nuLB )

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_incomp_nNwtn_CY
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK for
  !! incompressible lbm model
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_incomp_nNwtn_PL )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho0 ! inverse local density
    real(kind=rk) :: rho0    ! reference density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    real(kind=rk) :: fEq__0
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2, inv_omg_2, coeffSR
    real(kind=rk) :: omega_2, cmpl_o, omega
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &           sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1, sum11_2,  &
      &           sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------
    ! parameters for non-Newtonian
    integer :: omegaPos, nScalars
    real(kind=rk) :: fNeq_NE_SW, fNeq_NW_SE, fNeq_BW_TE, &
      &              fNeq_BE_TW, fNeq_BS_TN, fNeq_BN_TS
    real(kind=rk) :: strainRate(6)
    real(kind=rk) :: shearRate ! second invariant of strainRate
    real(kind=rk) :: nuLB
    real(kind=rk) :: convSR(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convVisc(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: k, n_1 ! parameters for power-law
    !---------------------------------------------------------------------------

    omegaPos = layout%stencil(1)%QQ + 1
    nScalars = varSys%nScalars

    ! conversion factor for Shear Rate (or Strain Rate)
    convSR(:) = getConversionFac( &
      &           conf = varSys%solSpec%conf, &
      &           facName = 'strainRate', &
      &           nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! conversion factor for kinematic viscosity
    convVisc(:) = getConversionFac( &
      &             conf = varSys%solSpec%conf, &
      &             facName = 'visc', &
      &             nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! get nonNewtonian power law parameter
    k   = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%PL%k &
      &   / params%physics%rho0 / convVisc(level)
    n_1 = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%PL%n_1

    rho0 = fieldProp(1)%fluid%rho0

    ! inverse local density
    inv_rho0 = 1._rk / rho0

!$omp do schedule(static)

    nodeloop: do iElem = nStart, nEnd
      ! read the relaxation parameter omega for the current element
      omega   = inState( ?IDX?(omegaPos, iElem, nScalars) )
      inv_omg_2 = 2._rk / omega
      coeffSR = -1.5_rk * omega * convSR(level) * inv_rho0

      ! pre-calculate partial collision constants
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega

      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, nScalars, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, nScalars, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, nScalars, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, nScalars, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, nScalars, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, nScalars, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, nScalars, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, nScalars, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, nScalars, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, nScalars, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, nScalars, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, nScalars, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, nScalars, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, nScalars, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, nScalars, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, nScalars, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, nScalars, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, nScalars, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, nScalars, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW) &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S) &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T) &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW) &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE) &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS) &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E) &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW) &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)&
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S)  - pdfTmp(Q_SE)&
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN) &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN) &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! transfer moments to velocities
      u_x = u_x * inv_rho0
      u_y = u_y * inv_rho0
      u_z = u_z * inv_rho0

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (rho - 1.5_rk * usq * rho0)
      fEq__0 = rho*div1_3 - 0.5_rk*usq*rho0

      outState(?SAVE?(Q__0,iElem,nScalars,neigh)) = pdfTmp(Q__0)*cmpl_o + omega*fEq__0

      coeff_1 = div1_8 * omega * rho0

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,nScalars,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,nScalars,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      fNeq_NE_SW = pdfTmp(Q_NE) + pdfTmp(Q_SW) - sum1_2*inv_omg_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,nScalars,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,nScalars,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      fNeq_NW_SE = pdfTmp(Q_NW) + pdfTmp(Q_SE) - sum3_2*inv_omg_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,nScalars,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,nScalars,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      fNeq_BW_TE = pdfTmp(Q_BW) + pdfTmp(Q_TE) - sum10_2*inv_omg_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,nScalars,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,nScalars,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      fNeq_BE_TW = pdfTmp(Q_BE) + pdfTmp(Q_TW) - sum12_2*inv_omg_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,nScalars,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,nScalars,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      fNeq_BS_TN = pdfTmp(Q_BS) + pdfTmp(Q_TN) - sum11_2*inv_omg_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,nScalars,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,nScalars,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      fNeq_BN_TS = pdfTmp(Q_BN) + pdfTmp(Q_TS) - sum13_2*inv_omg_2

      coeff_2 = div1_8 * omega_2 * rho0
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,nScalars,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,nScalars,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      strainRate(2) = pdfTmp(Q__N) + pdfTmp(Q__S) - sum2_2*inv_omg_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,nScalars,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,nScalars,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      strainRate(1) = pdfTmp(Q__W) + pdfTmp(Q__E) - sum4_2*inv_omg_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,nScalars,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,nScalars,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

      strainRate(3) = pdfTmp(Q__T) + pdfTmp(Q__B) - sum9_2*inv_omg_2

      ! s1 = (W+E) + (SW+NE) + (SE+NW) + (BW+TE) + (BE+TW)
      strainRate(1) =   strainRate(1) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BW_TE + fNeq_BE_TW

      ! s2 = (S+N) + (NE+SW) + (NW+SE) + (BS+TN) + (BN+TS)
      strainRate(2) =   strainRate(2) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BS_TN + fNeq_BN_TS

      ! s3 = (B+T) + (BS+TN) + (BN+TS) + (BW+TE) + (BE+TW)
      strainRate(3) =   strainRate(3) &
        &             + fNeq_BS_TN + fNeq_BN_TS + fNeq_BW_TE + fNeq_BE_TW

      ! s4 = (SW+NE) - (NW+SE)
      strainRate(4) = fNeq_NE_SW - fNeq_NW_SE

      ! s5 = (BS+TN) - (TS+BN)
      strainRate(5) = fNeq_BS_TN - fNeq_BN_TS

      ! s6 = (BW+TE) - (BE+TW)
      strainRate(6) = fNeq_BW_TE - fNeq_BE_TW

      strainRate = strainRate * coeffSR

      ! Calculate physical shear rate
      shearRate = 2._rk * sqrt( sum( strainRate * strainRate ) )

      ! the exponential operation is quite expensive.
      ! it decreases performance about 3-4 MLUPS on SHU
      ! calculate kinematic visc by POWER-LAW model
      ! from paper: LBMs for non-newtonian flows, page 803
      ! k is actually k / rho0 / convVisc
      nuLB = ( shearRate ** n_1 ) * k

      ! calc omage and save state vector
outState( ?IDX?(omegaPos, iElem, nScalars) ) = 1._rk / ( 0.5_rk + 3._rk*nuLB )

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_incomp_nNwtn_PL
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the D3Q19 model with BGK for
  !! incompressible lbm model
  !!
?? copy :: compute_routineHeader( bgk_advRel_d3q19_incomp_nNwtn_CS )
    !---------------------------------------------------------------------------
    integer :: iElem ! element counter
    real(kind=rk) :: pdfTmp(layout%stencil(1)%QQ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: inv_rho0 ! inverse local density
    real(kind=rk) :: rho0    ! reference density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    real(kind=rk) :: fEq__0
    ! derived constants
    real(kind=rk) :: usqn, usqn_o1, usqn_o2, inv_omg_2, coeffSR
    real(kind=rk) :: omega_2, cmpl_o, omega
    real(kind=rk) :: coeff_1, coeff_2
    real(kind=rk) :: ui1, ui3, ui10, ui11, ui12, ui13
    real(kind=rk) :: fac_1, fac_2, fac_3, fac_4, fac_9, fac_10, fac_11, fac_12,&
      &              fac_13
    real(kind=rk) :: sum1_1, sum1_2, sum2_1, sum2_2, sum3_1, sum3_2, sum4_1,   &
      &           sum4_2, sum9_1, sum9_2, sum10_1, sum10_2, sum11_1, sum11_2,  &
      &           sum12_1, sum12_2, sum13_1, sum13_2
    !---------------------------------------------------------------------------
    ! parameters for non-Newtonian
    integer :: omegaPos, nScalars
    real(kind=rk) :: fNeq_NE_SW, fNeq_NW_SE, fNeq_BW_TE, &
      &              fNeq_BE_TW, fNeq_BS_TN, fNeq_BN_TS
    real(kind=rk) :: strainRate(6)
    real(kind=rk) :: shearRate ! second invariant of strainRate
    real(kind=rk) :: nuLB, t
    real(kind=rk) :: convSR(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: convVisc(tree%global%minLevel:tree%global%maxLevel)
    real(kind=rk) :: k0, k1 ! parameters for Casson model
    !---------------------------------------------------------------------------

    omegaPos = layout%stencil(1)%QQ + 1
    nScalars = varSys%nScalars

    ! conversion factor for Shear Rate (or Strain Rate)
    convSR(:) = getConversionFac( &
      &           conf = varSys%solSpec%conf, &
      &           facName = 'strainRate', &
      &           nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! conversion factor for kinematic viscosity
    convVisc(:) = getConversionFac( &
      &             conf = varSys%solSpec%conf, &
      &             facName = 'visc', &
      &             nLevels = tree%global%maxLevel - tree%global%minLevel + 1 )

    ! get nonNewtonian Casson model parameter
    k0 = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CS%k0
    k1 = scheme( iScheme )%field(1)%fieldProp%fluid%nNwtn%CS%k1

    rho0 = fieldProp(1)%fluid%rho0

    ! inverse local density
    inv_rho0 = 1._rk / rho0

!$omp do schedule(static)

    nodeloop: do iElem = nStart, nEnd
      ! read the relaxation parameter omega for the current element
      omega   = inState( ?IDX?(omegaPos, iElem, nScalars) )
      inv_omg_2 = 2._rk / omega
      coeffSR = -1.5_rk * omega * convSR(level) * inv_rho0

      ! pre-calculate partial collision constants
      omega_2 = 2._rk * omega
      cmpl_o  = 1._rk - omega

      pdfTmp( Q__W ) = inState(?FETCH?( Q__W, iElem, nScalars, neigh ))
      pdfTmp( Q__S ) = inState(?FETCH?( Q__S, iElem, nScalars, neigh ))
      pdfTmp( Q__B ) = inState(?FETCH?( Q__B, iElem, nScalars, neigh ))
      pdfTmp( Q__E ) = inState(?FETCH?( Q__E, iElem, nScalars, neigh ))
      pdfTmp( Q__N ) = inState(?FETCH?( Q__N, iElem, nScalars, neigh ))
      pdfTmp( Q__T ) = inState(?FETCH?( Q__T, iElem, nScalars, neigh ))
      pdfTmp( Q_BS ) = inState(?FETCH?( Q_BS, iElem, nScalars, neigh ))
      pdfTmp( Q_TS ) = inState(?FETCH?( Q_TS, iElem, nScalars, neigh ))
      pdfTmp( Q_BN ) = inState(?FETCH?( Q_BN, iElem, nScalars, neigh ))
      pdfTmp( Q_TN ) = inState(?FETCH?( Q_TN, iElem, nScalars, neigh ))
      pdfTmp( Q_BW ) = inState(?FETCH?( Q_BW, iElem, nScalars, neigh ))
      pdfTmp( Q_BE ) = inState(?FETCH?( Q_BE, iElem, nScalars, neigh ))
      pdfTmp( Q_TW ) = inState(?FETCH?( Q_TW, iElem, nScalars, neigh ))
      pdfTmp( Q_TE ) = inState(?FETCH?( Q_TE, iElem, nScalars, neigh ))
      pdfTmp( Q_SW ) = inState(?FETCH?( Q_SW, iElem, nScalars, neigh ))
      pdfTmp( Q_NW ) = inState(?FETCH?( Q_NW, iElem, nScalars, neigh ))
      pdfTmp( Q_SE ) = inState(?FETCH?( Q_SE, iElem, nScalars, neigh ))
      pdfTmp( Q_NE ) = inState(?FETCH?( Q_NE, iElem, nScalars, neigh ))
      pdfTmp( Q__0 ) = inState(?FETCH?( Q__0, iElem, nScalars, neigh ))

      ! local density
      rho = pdfTmp(Q_NE) + pdfTmp(Q__N) + pdfTmp(Q_NW) &
        & + pdfTmp(Q__W) + pdfTmp(Q_SW) + pdfTmp(Q__S) &
        & + pdfTmp(Q_SE) + pdfTmp(Q__E) + pdfTmp(Q__T) &
        & + pdfTmp(Q_TE) + pdfTmp(Q_TN) + pdfTmp(Q_TW) &
        & + pdfTmp(Q_TS) + pdfTmp(Q__B) + pdfTmp(Q_BE) &
        & + pdfTmp(Q_BN) + pdfTmp(Q_BW) + pdfTmp(Q_BS) &
        & + pdfTmp(Q__0)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(Q_NE) + pdfTmp(Q_SE) + pdfTmp(Q__E) &
        & + pdfTmp(Q_TE) + pdfTmp(Q_BE)                &
        & - pdfTmp(Q_NW) - pdfTmp(Q__W) - pdfTmp(Q_SW) &
        & - pdfTmp(Q_TW) - pdfTmp(Q_BW)

      u_y = pdfTmp(Q_NE) + pdfTmp(Q__N)  + pdfTmp(Q_NW)&
        & + pdfTmp(Q_TN) + pdfTmp(Q_BN)                &
        & - pdfTmp(Q_SW) - pdfTmp(Q__S)  - pdfTmp(Q_SE)&
        & - pdfTmp(Q_TS) - pdfTmp(Q_BS)

      u_z = pdfTmp(Q__T) + pdfTmp(Q_TE) + pdfTmp(Q_TN) &
        & + pdfTmp(Q_TW) + pdfTmp(Q_TS)                &
        & - pdfTmp(Q__B) - pdfTmp(Q_BE) - pdfTmp(Q_BN) &
        & - pdfTmp(Q_BW) - pdfTmp(Q_BS)

      ! transfer moments to velocities
      u_x = u_x * inv_rho0
      u_y = u_y * inv_rho0
      u_z = u_z * inv_rho0

      ! square velocity and derived constants
      usq  = u_x*u_x + u_y*u_y + u_z*u_z
      usqn = div1_36 * (rho - 1.5_rk * usq * rho0)
      fEq__0 = rho*div1_3 - 0.5_rk*usq*rho0

      outState(?SAVE?(Q__0,iElem,nScalars,neigh)) = pdfTmp(Q__0)*cmpl_o + omega*fEq__0

      coeff_1 = div1_8 * omega * rho0

      usqn_o1 = omega * usqn

      ui1     =  u_x + u_y
      fac_1   = coeff_1 * ui1
      sum1_1  = fac_1 * div3_4h
      sum1_2  = fac_1 * ui1 + usqn_o1

      outState(?SAVE?(Q_NE,iElem,nScalars,neigh))=pdfTmp(Q_NE) *cmpl_o+sum1_1 +sum1_2
      outState(?SAVE?(Q_SW,iElem,nScalars,neigh))=pdfTmp(Q_SW) *cmpl_o-sum1_1 +sum1_2

      fNeq_NE_SW = pdfTmp(Q_NE) + pdfTmp(Q_SW) - sum1_2*inv_omg_2

      ui3     = -u_x + u_y
      fac_3   = coeff_1 * ui3
      sum3_1  = fac_3 * div3_4h
      sum3_2  = fac_3 * ui3 + usqn_o1

      outState(?SAVE?(Q_NW,iElem,nScalars,neigh))=pdfTmp(Q_NW) *cmpl_o+sum3_1 +sum3_2
      outState(?SAVE?(Q_SE,iElem,nScalars,neigh))=pdfTmp(Q_SE) *cmpl_o-sum3_1 +sum3_2

      fNeq_NW_SE = pdfTmp(Q_NW) + pdfTmp(Q_SE) - sum3_2*inv_omg_2

      ui10    =  u_x + u_z
      fac_10  = coeff_1 * ui10
      sum10_1 = fac_10 * div3_4h
      sum10_2 = fac_10 * ui10 + usqn_o1

      outState(?SAVE?(Q_TE,iElem,nScalars,neigh))=pdfTmp(Q_TE)*cmpl_o+sum10_1+sum10_2
      outState(?SAVE?(Q_BW,iElem,nScalars,neigh))=pdfTmp(Q_BW)*cmpl_o-sum10_1+sum10_2

      fNeq_BW_TE = pdfTmp(Q_BW) + pdfTmp(Q_TE) - sum10_2*inv_omg_2

      ui12    = -u_x + u_z
      fac_12  = coeff_1 * ui12
      sum12_1 = fac_12 * div3_4h
      sum12_2 = fac_12 * ui12 + usqn_o1

      outState(?SAVE?(Q_TW,iElem,nScalars,neigh))=pdfTmp(Q_TW)*cmpl_o+sum12_1+sum12_2
      outState(?SAVE?(Q_BE,iElem,nScalars,neigh))=pdfTmp(Q_BE)*cmpl_o-sum12_1+sum12_2

      fNeq_BE_TW = pdfTmp(Q_BE) + pdfTmp(Q_TW) - sum12_2*inv_omg_2

      ui11    =  u_y + u_z
      fac_11  = coeff_1 * ui11
      sum11_1 = fac_11 * div3_4h
      sum11_2 = fac_11 * ui11 + usqn_o1

      outState(?SAVE?(Q_TN,iElem,nScalars,neigh))=pdfTmp(Q_TN)*cmpl_o+sum11_1+sum11_2
      outState(?SAVE?(Q_BS,iElem,nScalars,neigh))=pdfTmp(Q_BS)*cmpl_o-sum11_1+sum11_2

      fNeq_BS_TN = pdfTmp(Q_BS) + pdfTmp(Q_TN) - sum11_2*inv_omg_2

      ui13    = -u_y + u_z
      fac_13  = coeff_1 * ui13
      sum13_1 = fac_13 * div3_4h
      sum13_2 = fac_13 * ui13 + usqn_o1

      outState(?SAVE?(Q_TS,iElem,nScalars,neigh))=pdfTmp(Q_TS)*cmpl_o+sum13_1+sum13_2
      outState(?SAVE?(Q_BN,iElem,nScalars,neigh))=pdfTmp(Q_BN)*cmpl_o-sum13_1+sum13_2

      fNeq_BN_TS = pdfTmp(Q_BN) + pdfTmp(Q_TS) - sum13_2*inv_omg_2

      coeff_2 = div1_8 * omega_2 * rho0
      usqn_o2 = omega_2 * usqn

      fac_2   = coeff_2 * u_y
      sum2_1  = fac_2 * div3_4h
      sum2_2  = fac_2 * u_y + usqn_o2

      outState(?SAVE?(Q__N,iElem,nScalars,neigh ))=pdfTmp(Q__N) *cmpl_o+sum2_1 +sum2_2
      outState(?SAVE?(Q__S,iElem,nScalars,neigh ))=pdfTmp(Q__S) *cmpl_o-sum2_1 +sum2_2

      strainRate(2) = pdfTmp(Q__N) + pdfTmp(Q__S) - sum2_2*inv_omg_2

      fac_4   = coeff_2 * u_x
      sum4_1  = fac_4 * div3_4h
      sum4_2  = fac_4 * u_x + usqn_o2

      outState(?SAVE?(Q__W,iElem,nScalars,neigh ))=pdfTmp(Q__W) *cmpl_o-sum4_1 +sum4_2
      outState(?SAVE?(Q__E,iElem,nScalars,neigh ))=pdfTmp(Q__E) *cmpl_o+sum4_1 +sum4_2

      strainRate(1) = pdfTmp(Q__W) + pdfTmp(Q__E) - sum4_2*inv_omg_2

      fac_9   = coeff_2 * u_z
      sum9_1  = fac_9 * div3_4h
      sum9_2  = fac_9 * u_z + usqn_o2

      outState(?SAVE?(Q__T,iElem,nScalars,neigh ))=pdfTmp(Q__T) *cmpl_o+sum9_1 +sum9_2
      outState(?SAVE?(Q__B,iElem,nScalars,neigh ))=pdfTmp(Q__B) *cmpl_o-sum9_1 +sum9_2

      strainRate(3) = pdfTmp(Q__T) + pdfTmp(Q__B) - sum9_2*inv_omg_2

      !--------------------------------------------------------------------
      ! s1 = (W+E) + (SW+NE) + (SE+NW) + (BW+TE) + (BE+TW)
      strainRate(1) =   strainRate(1) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BW_TE + fNeq_BE_TW

      ! s2 = (S+N) + (NE+SW) + (NW+SE) + (BS+TN) + (BN+TS)
      strainRate(2) =   strainRate(2) &
        &             + fNeq_NE_SW + fNeq_NW_SE + fNeq_BS_TN + fNeq_BN_TS

      ! s3 = (B+T) + (BS+TN) + (BN+TS) + (BW+TE) + (BE+TW)
      strainRate(3) =   strainRate(3) &
        &             + fNeq_BS_TN + fNeq_BN_TS + fNeq_BW_TE + fNeq_BE_TW

      ! s4 = (SW+NE) - (NW+SE)
      strainRate(4) = fNeq_NE_SW - fNeq_NW_SE

      ! s5 = (BS+TN) - (TS+BN)
      strainRate(5) = fNeq_BS_TN - fNeq_BN_TS

      ! s6 = (BW+TE) - (BE+TW)
      strainRate(6) = fNeq_BW_TE - fNeq_BE_TW

      strainRate = strainRate * coeffSR
      !--------------------------------------------------------------------

      ! Calculate physical shear rate
      shearRate = 2._rk * sqrt( sum( strainRate * strainRate ) )

      ! calculate dynamic visc by Carreau-Yasuda model
      t = ( k0 + k1 * sqrt(shearRate) )
      nuLB = t * t / shearRate
      ! convert to kinematic visc
      nuLB = nuLB / params%physics%rho0 / convVisc(level)

      ! calc omage and save state vector
outState( ?IDX?(omegaPos, iElem, nScalars) ) = 1._rk / ( 0.5_rk + 3._rk*nuLB )

    enddo nodeloop
!$omp end do nowait

  end subroutine bgk_advRel_d3q19_incomp_nNwtn_CS
!******************************************************************************!

end module mus_d3q19_module
!******************************************************************************!
