! See copyright notice in the COPYRIGHT file.
?? include 'header/lbm_macros.inc'
!******************************************************************************!
!> This module provides the definition and methods for
!! MRT advection relaxation scheme.
!! The LB equaton using MRT is
!!    f(t+dt,x+dx) = f - M^(-1) * S * ( (M*f) - m^(eq) )
!!
!! The moments m(1:19) = M * f(1:19) are labeled as
!!  m( 1) = rho
!!  m( 2) = e = rho * (ux^2 + uy^2 + uz^2)
!!  m( 3) = epsilon
!!  m( 4) = jx = rho * ux
!!  m( 5) = qx
!!  m( 6) = jy = rho * uy
!!  m( 7) = qy
!!  m( 8) = jz = rho * uz
!!  m( 9) = qz
!!  m(10) = 3 * pxx = rho * (2ux^2 - uy^2 - uz^2)
!!  m(11) = 3 * Pixx
!!  m(12) = pzz  = rho * (uy^2 - uz^2)
!!  m(13) = Piww
!!  m(14) = pxy  = rho * ux * uy
!!  m(15) = pyz  = rho * uy * uz
!!  m(16) = pzx  = rho * uz * ux
!!  m(17) = mx
!!  m(18) = my
!!  m(19) = mz
!!
!! The collision parameters S correspondes to the omega in BGK model.
!!
!! The MRT implementation here is taken from:\n
!! J. Toelke, S. Freudiger, and M. Krafczyk,
!! "An adaptive scheme using hierarchical grids for lattice Boltzmann
!! multi-phase flow simulations," Comput. Fluids, vol. 35, pp. 820–830,
!! 2006. \n
!! Notice that the collision matrix S used in this papar corresponds to
!! -omega in BGK model, because it express the LB equation is slightly
!! different way.
!! In this paper, the following notions are used:\n
!!  s(a) = s(2)
!!  s(b) = s(3)
!!  s(c) = s(5) = s(7) = s(9)
!!  s(d) = s(11) = s(13
!!  s(e) = s(17) = s(18) = s(19)
!!  s(w) = s(10) = s(12) = s(14) = s(15) = s(16)
!! It is suggested that, for D3Q19,
!!  s(a) = s(b) = s(c) = s(d) = s(e) = max( s(w), -1.0 )
!!
!! This MRT is based on the 3D initial paper
!! D. dHumieres, I. Ginzburg, M. Krafczyk, and P. Lallemand
!! "Multiple-relaxation-time lattice Boltzmann models in three dimensions,"
!! Philosophical Transactions: Mathematical, Jan. 2002
!! Attention: The collision matrix S used in the above two papers are
!! opposite with each other!
!!
module mus_mrt_module

  ! include treelm modules
  use env_module,               only: rk
  use tem_var_system_module,    only: tem_var_system_type
  use treelmesh_module,         only: treelmesh_type
  use tem_param_module,         only: div1_36, cs2inv, cs4inv, t2cs2inv, &
    &                                 t2cs4inv

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_pdf_module,           only: pdf_globInfo_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_param_module,         only: mus_param_type

  implicit none

  private

  public :: mrt_advRel_d3q19
  public :: mrt_advRel_d3q19_fixedVel
  public :: mrt_advRel_d3q19_omSpatial
  public :: mrt_advRel_d3q19_les
  public :: mrt_advRel_d3q19_les_omSpatial
  public :: mrt_advRel_d3q19_incomp
  public :: mrt_advRel_d3q19_incomp_omSpatial
  public :: mrt_advRel_d3q19_incomp_fixedVel
  public :: mrt_advRel_d3q19_incomp_les
  public :: mrt_advRel_d3q19_incomp_les_omSpatial

  !=============================================================================
  ! D3Q19 flow model
  !=============================================================================
  !> Definition of the discrete velocity set

  integer, parameter :: QQ  = 19   !> number of pdf directions
  integer, parameter :: QQN = 18   !> number of neighbors

  integer,parameter :: q__W     = 1     !< west             x-
  integer,parameter :: q__S     = 2     !< south            y-
  integer,parameter :: q__B     = 3     !< bottom           z-
  integer,parameter :: q__E     = 4     !< east             x+
  integer,parameter :: q__N     = 5     !< north            y+
  integer,parameter :: q__T     = 6     !< top              z+
  integer,parameter :: q_BS     = 7     !<                  z-,y-
  integer,parameter :: q_TS     = 8     !<                  z+,y-
  integer,parameter :: q_BN     = 9     !<                  z-,y+
  integer,parameter :: q_TN     = 10    !<                  z+,y+
  integer,parameter :: q_BW     = 11    !<                  x-,z-
  integer,parameter :: q_BE     = 12    !<                  x+,z-
  integer,parameter :: q_TW     = 13    !<                  x-,z+
  integer,parameter :: q_TE     = 14    !<                  x+,z+
  integer,parameter :: q_SW     = 15    !<                  y-,x-
  integer,parameter :: q_NW     = 16    !<                  y+,x-
  integer,parameter :: q_SE     = 17    !<                  y-,x+
  integer,parameter :: q_NE     = 18    !<                  y+,x+
  integer,parameter :: q__0     = 19    !< rest density is last

contains

!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
?? copy :: compute_routineHeader( mrt_advRel_d3q19 )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    !---------------------------------------------------------------------------

    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)

    nodeloop: do iElem = nStart, nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, QQ, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, QQ, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, QQ, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, QQ, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, QQ, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, QQ, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, QQ, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, QQ, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, QQ, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, QQ, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, QQ, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, QQ, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, QQ, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, QQ, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, QQ, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, QQ, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, QQ, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, QQ, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)

      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            &
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho
      meq( 2) = rho*usq
      meq( 4) = u_x*rho
      meq( 6) = u_y*rho
      meq( 8) = u_z*rho
      meq(10) = rho*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho*( u_y*u_y - u_z*u_z )
      meq(14) = rho*u_x*u_y
      meq(15) = rho*u_y*u_z
      meq(16) = rho*u_x*u_z

      ! Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) =  pdfTmp( 4)- pdfTmp( 1)+ pdfTmp(18)                            &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)

      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  pdfTmp( 5)- pdfTmp( 2)+ pdfTmp(18)                            &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  pdfTmp( 6)- pdfTmp( 3)                                        &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)

      ! MATRIX looks ok
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2) * (mneq( 2) - meq( 2))
      mout( 3) = s_mrt( 3) * (mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5) * (mneq( 5) - meq( 5))
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7) * (mneq( 7) - meq( 7))
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9) * (mneq( 9) - meq( 9))
      mout(10) = s_mrt(10) * (mneq(10) - meq(10))
      mout(11) = s_mrt(11) * (mneq(11) - meq(11))
      mout(12) = s_mrt(12) * (mneq(12) - meq(12))
      mout(13) = s_mrt(13) * (mneq(13) - meq(13))
      mout(14) = s_mrt(14) * (mneq(14) - meq(14))
      mout(15) = s_mrt(15) * (mneq(15) - meq(15))
      mout(16) = s_mrt(16) * (mneq(16) - meq(16))
      mout(17) = s_mrt(17) * mneq(17)
      mout(18) = s_mrt(18) * mneq(18)
      mout(19) = s_mrt(19) * mneq(19)

      ! Transformation back to 
      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7) + 1._rk/48._rk*mout(10)             &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) + 1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7) + 1._rk/48._rk*mout(10)             &
        &           + 1._rk/48._rk*mout(11) + 1._rk/16._rk*mout(12)            &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
?? copy :: compute_routineHeader( mrt_advRel_d3q19_les )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    integer :: iDir
    real(kind=rk) :: ucx
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ )  ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    ! Variables required for LES
    real(kind=rk) :: fEq(19) !< equilibrium distribution
    real(kind=rk) :: nEq(19) !< non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3)
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const
    !---------------------------------------------------------------------------
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)
    tau0 = 1._rk / fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s

    nodeloop: do iElem = nStart, nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, QQ, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, QQ, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, QQ, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, QQ, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, QQ, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, QQ, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, QQ, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, QQ, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, QQ, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, QQ, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, QQ, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, QQ, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, QQ, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, QQ, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, QQ, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, QQ, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, QQ, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, QQ, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)

      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            &
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho
      meq( 2) = rho*usq
      meq( 4) = u_x*rho
      meq( 6) = u_y*rho
      meq( 8) = u_z*rho
      meq(10) = rho*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho*( u_y*u_y - u_z*u_z )
      meq(14) = rho*u_x*u_y
      meq(15) = rho*u_y*u_z
      meq(16) = rho*u_x*u_z

      !---------------------- LES Part -----------------------------------------
      !\todo MH: Replace this with Kannans equilibrium calculation derive%do
      ! This must be done outside the loop to allow for vectorized calculation 
      ! Calculate the equilibrium distribution function
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx =   real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*u_x  &
          &   + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*u_y  &
          &   + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * rho * ( 1.d0 + ucx*cs2inv &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv )

      enddo

      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )

      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)
      om_turb = 1._rk/(tau0+tau_turb)
      s_mrt( 10 )     = om_turb
      s_mrt( 12 )     = om_turb
      s_mrt( 14:16 )  = om_turb
      ! Now the relaxation parameters are changed based on the strain rate
      !---------------------- LES Part -----------------------------------------

      !Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) =  pdfTmp( 4)- pdfTmp( 1)+ pdfTmp(18)                            &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      
      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  pdfTmp( 5)- pdfTmp( 2)+ pdfTmp(18)                            &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  pdfTmp( 6)- pdfTmp( 3)                                        &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)

      ! MATRIX looks ok
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 


      ! CHECK MINV ploed

      ! Transformation back to 
      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_les
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
?? copy :: compute_routineHeader( mrt_advRel_d3q19_les_omSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    integer :: iDir
    real(kind=rk) :: ucx
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) omega   ! relaxation parameter
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    ! Variables required for LES
    real(kind=rk) :: fEq(19) ! equilibrium distribution
    real(kind=rk) :: nEq(19) ! non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3)
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const
    !---------------------------------------------------------------------------
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)
    omega = fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s
    tau0 = 1._rk/omega

    nodeloop: do iElem=nStart,nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, varSys%nScalars, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, varSys%nScalars, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, varSys%nScalars, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, varSys%nScalars, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, varSys%nScalars, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, varSys%nScalars, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, varSys%nScalars, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, varSys%nScalars, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, varSys%nScalars, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, varSys%nScalars, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, varSys%nScalars, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, varSys%nScalars, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, varSys%nScalars, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, varSys%nScalars, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, varSys%nScalars, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, varSys%nScalars, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, varSys%nScalars, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, varSys%nScalars, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, varSys%nScalars, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)

      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            &
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho*usq
      meq( 4) = u_x*rho 
      meq( 6) = u_y*rho 
      meq( 8) = u_z*rho 
      meq(10) = rho*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho*( u_y*u_y - u_z*u_z )
      meq(14) = rho*u_x*u_y
      meq(15) = rho*u_y*u_z
      meq(16) = rho*u_x*u_z

      !---------------------- LES Part -----------------------------------------
      ! This must be done outside the loop to allow for vectorized calculation 
      ! Calculate the equilibrium distribution function
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*u_x           &
          & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*u_y           &
          & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * rho * ( 1.d0 + ucx*cs2inv        &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv )

      enddo

      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )
      ! Get the spatial omega definition directly into tau0
      tau0 = 1._rk/(fieldProp(1)%fluid%viscSpatial%stored( level )%            &
        &                                                   val( iElem )*omega)

      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)
      om_turb = 1._rk/(tau0+tau_turb)
      s_mrt( 10 )     = om_turb
      s_mrt( 12 )     = om_turb
      s_mrt( 14:16 )  = om_turb
      ! Now the relaxation parameters are changed based on the strain rate
      !---------------------- LES Part -----------------------------------------

      !Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) =  pdfTmp( 4)- pdfTmp( 1)+ pdfTmp(18)                            &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)

      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  pdfTmp( 5)- pdfTmp( 2)+ pdfTmp(18)                            &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  pdfTmp( 6)- pdfTmp( 3)                                        &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)

      ! MATRIX looks ok
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 


      ! CHECK MINV ploed

      ! Transformation back to 
      outstate( ?SAVE?( 19,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8) + 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_les_omSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
?? copy :: compute_routineHeader( mrt_advRel_d3q19_fixedVel )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    integer, parameter :: iField = 1
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    !---------------------------------------------------------------------------

    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)
    ! Get velocity which might be defined to vary in space and time
    nodeloop: do iElem=nStart,nEnd
      ! First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, QQ, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, QQ, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, QQ, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, QQ, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, QQ, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, QQ, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, QQ, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, QQ, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, QQ, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, QQ, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, QQ, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, QQ, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, QQ, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, QQ, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, QQ, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, QQ, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, QQ, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, QQ, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! use
      u_x = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(1)%       &
        &                                          stored( level )%val( iElem )
      u_y = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(2)%       &
        &                                          stored( level )%val( iElem )
      u_z = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(3)%       &
        &                                          stored( level )%val( iElem )

      ! inverse local density
      inv_rho = 1._rk / rho

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho*usq
      meq( 4) = u_x*rho 
      meq( 6) = u_y*rho 
      meq( 8) = u_z*rho 
      meq(10) = rho*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho*( u_y*u_y - u_z*u_z )
      meq(14) = rho*u_x*u_y
      meq(15) = rho*u_y*u_z
      meq(16) = rho*u_x*u_z

      ! Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) =  pdfTmp( 4)- pdfTmp( 1)+ pdfTmp(18)                            &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      
      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  pdfTmp( 5)- pdfTmp( 2)+ pdfTmp(18)                            &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  pdfTmp( 6)- pdfTmp( 3)                                        &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        & - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                                 &
        & + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                     &
        & -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)

      ! MATRIX looks ok
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = s_mrt( 4)*(mneq( 4) - meq( 4)) 
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = s_mrt( 6)*(mneq( 6) - meq( 6)) 
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = s_mrt( 8)*(mneq( 8) - meq( 8)) 
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 


      ! CHECK MINV ploed

      ! Transformation back to 
      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_fixedVel
!******************************************************************************!


!******************************************************************************
  !> Advection relaxation routine for the MRT model.
  !! 
  !! This implementation follows the MRT model as described by Tölke et al. in 
  !! J. Toelke, S. Freudiger, and M. Krafczyk,
  !! "An adaptive scheme using hierarchical grids for lattice Boltzmann
  !! multi-phase flow simulations," Comput. Fluids, vol. 35, pp. 820–830,
  !! 2006.
  !!
?? copy :: compute_routineHeader( mrt_advRel_d3q19_incomp )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0
    !---------------------------------------------------------------------------
    rho0  = fieldProp(1)%fluid%rho0
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)

    nodeloop: do iElem=nStart,nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, QQ, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, QQ, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, QQ, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, QQ, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, QQ, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, QQ, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, QQ, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, QQ, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, QQ, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, QQ, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, QQ, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, QQ, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, QQ, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, QQ, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, QQ, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, QQ, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, QQ, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, QQ, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)

      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            &
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho0*usq
      meq( 4) = u_x*rho0 
      meq( 6) = u_y*rho0 
      meq( 8) = u_z*rho0 
      meq(10) = rho0*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho0*( u_y*u_y - u_z*u_z )
      meq(14) = rho0*u_x*u_y
      meq(15) = rho0*u_y*u_z
      meq(16) = rho0*u_x*u_z

      ! Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) = u_x 

      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  u_y
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  u_z
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)

      ! MATRIX looks ok
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 


      ! CHECK MINV ploed

      ! Transformation back to 
      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_incomp
!******************************************************************************!


!******************************************************************************
  !> Advection relaxation routine for the MRT model.
  !! 
  !! This implementation follows the MRT model as described by Tölke et al. in 
  !! J. Tölke, S. Freudiger, and M. Krafczyk, 
  !! "An adaptive scheme using hierarchical grids for lattice Boltzmann 
  !! multi-phase flow simulations," Comput. Fluids, vol. 35, pp. 820–830,
  !! 2006.
  !!
?? copy :: compute_routineHeader( mrt_advRel_d3q19_omSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    real(kind=rk) :: omSrc
    !---------------------------------------------------------------------------
    omSrc = fieldProp(1)%fluid%rampOmegas( level )
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)

    nodeloop: do iElem=nStart,nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, QQ, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, QQ, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, QQ, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, QQ, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, QQ, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, QQ, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, QQ, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, QQ, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, QQ, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, QQ, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, QQ, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, QQ, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, QQ, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, QQ, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, QQ, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, QQ, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, QQ, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, QQ, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)
                         
      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            & 
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! inverse local density
      inv_rho = 1._rk / rho

      ! transfer moments to velocities
      u_x = u_x * inv_rho
      u_y = u_y * inv_rho
      u_z = u_z * inv_rho

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho*usq
      meq( 4) = u_x*rho 
      meq( 6) = u_y*rho 
      meq( 8) = u_z*rho 
      meq(10) = rho*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho*( u_y*u_y - u_z*u_z )
      meq(14) = rho*u_x*u_y
      meq(15) = rho*u_y*u_z
      meq(16) = rho*u_x*u_z

!Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) =  pdfTmp( 4)- pdfTmp( 1)+ pdfTmp(18)                            &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      
      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  pdfTmp( 5)- pdfTmp( 2)+ pdfTmp(18)                            &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  pdfTmp( 6)- pdfTmp( 3)                                        &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)

      ! Get the spatial omega from the defined array
      s_mrt(10) = fieldProp(1)%fluid%viscSpatial%stored( level )%              &
        &                                                    val( iElem )*omSrc
      s_mrt(12) = fieldProp(1)%fluid%viscSpatial%stored( level )%              &
        &                                                    val( iElem )*omSrc
      s_mrt(14:16) = fieldProp(1)%fluid%viscSpatial%stored( level )%           &
        &                                                    val( iElem )*omSrc

      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 
      
      ! Transformation back to f-space
      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_omSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
  !! This implementation follows the MRT model as described by Tölke et al. in 
  !! J. Tölke, S. Freudiger, and M. Krafczyk, 
  !! "An adaptive scheme using hierarchical grids for lattice Boltzmann 
  !! multi-phase flow simulations," Comput. Fluids, vol. 35, pp. 820–830,
  !! 2006.
  !!
?? copy :: compute_routineHeader( mrt_advRel_d3q19_incomp_omSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0, omSrc
    !---------------------------------------------------------------------------
    omSrc = fieldProp(1)%fluid%rampOmegas( level )
    rho0  = fieldProp(1)%fluid%rho0
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)

    nodeloop: do iElem=nStart,nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, varSys%nScalars, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, varSys%nScalars, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, varSys%nScalars, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, varSys%nScalars, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, varSys%nScalars, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, varSys%nScalars, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, varSys%nScalars, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, varSys%nScalars, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, varSys%nScalars, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, varSys%nScalars, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, varSys%nScalars, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, varSys%nScalars, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, varSys%nScalars, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, varSys%nScalars, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, varSys%nScalars, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, varSys%nScalars, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, varSys%nScalars, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, varSys%nScalars, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, varSys%nScalars, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)

      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            &
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho0*usq
      meq( 4) = u_x*rho0 
      meq( 6) = u_y*rho0 
      meq( 8) = u_z*rho0 
      meq(10) = rho0*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho0*( u_y*u_y - u_z*u_z )
      meq(14) = rho0*u_x*u_y
      meq(15) = rho0*u_y*u_z
      meq(16) = rho0*u_x*u_z

      ! Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) = u_x 
      
      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  u_y
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  u_z
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      
      ! MATRIX looks ok
      ! Get the spatial omega from the defined array
      s_mrt(10) = fieldProp(1)%fluid%viscSpatial%stored( level )%              &
        &                                                    val( iElem )*omSrc
      s_mrt(12) = fieldProp(1)%fluid%viscSpatial%stored( level )%              &
        &                                                    val( iElem )*omSrc
      s_mrt(14:16) = fieldProp(1)%fluid%viscSpatial%stored( level )%           &
        &                                                    val( iElem )*omSrc
      
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 
      
      ! Transformation back to 

      ! CHECK MINV ploed

      outstate( ?SAVE?( 19,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,varSys%nScalars,neigh )) =                    &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_incomp_omSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
  !! This implementation follows the MRT model as described by Tölke et al. in 
  !! J. Tölke, S. Freudiger, and M. Krafczyk, 
  !! "An adaptive scheme using hierarchical grids for lattice Boltzmann 
  !! multi-phase flow simulations," Comput. Fluids, vol. 35, pp. 820–830,
  !! 2006.
  !!
?? copy :: compute_routineHeader( mrt_advRel_d3q19_incomp_les )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    integer :: iDir
    real(kind=rk) :: ucx
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0
    ! Variables required for LES
    real(kind=rk) :: fEq(19) !< equilibrium distribution
    real(kind=rk) :: nEq(19) !< non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3)
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const
    !---------------------------------------------------------------------------

    rho0  = fieldProp(1)%fluid%rho0
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)
    tau0 = 1._rk / fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s

    nodeloop: do iElem=nStart,nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, QQ, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, QQ, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, QQ, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, QQ, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, QQ, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, QQ, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, QQ, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, QQ, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, QQ, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, QQ, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, QQ, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, QQ, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, QQ, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, QQ, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, QQ, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, QQ, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, QQ, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, QQ, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)
                         
      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            &
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho0*usq
      meq( 4) = u_x*rho0 
      meq( 6) = u_y*rho0 
      meq( 8) = u_z*rho0 
      meq(10) = rho0*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho0*( u_y*u_y - u_z*u_z )
      meq(14) = rho0*u_x*u_y
      meq(15) = rho0*u_y*u_z
      meq(16) = rho0*u_x*u_z

      !---------------------- LES Part -----------------------------------------
      !\todo MH: Replace this with Kannans equilibrium calculation derive%do
      ! This must be done outside the loop to allow for vectorized calculation 
      ! Calculate the equilibrium distribution function
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*u_x           &
          & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*u_y           &
          & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * ( rho + rho0 * ( ucx*cs2inv      &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv ) )

      enddo
!      fEq(:) = getEquilibrium( density = rho, velocity = (/u_x, u_y, u_z /), layout = layout)

      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )

      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)
      om_turb = 1._rk/(tau0+tau_turb)
      s_mrt( 10 )     = om_turb
      s_mrt( 12 )     = om_turb
      s_mrt( 14:16 )  = om_turb
      ! Now the relaxation parameters are changed based on the strain rate
      !---------------------- LES Part -----------------------------------------

      ! Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) = u_x 
      
      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  u_y
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  u_z
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      
      ! MATRIX looks ok
      
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 
      
      ! Transformation back to 

      ! CHECK MINV ploed

      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5)+ 1._rk/12._rk*mout(8)               &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5)- 1._rk/12._rk*mout(8)               &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5)- 1._rk/12._rk*mout(8)               &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5)+ 1._rk/12._rk*mout(8)               &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7)- 1._rk/12._rk*mout(8)               &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/12._rk*mout(8)               &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_incomp_les
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
  !! This implementation follows the MRT model as described by Tölke et al. in 
  !! J. Tölke, S. Freudiger, and M. Krafczyk, 
  !! "An adaptive scheme using hierarchical grids for lattice Boltzmann 
  !! multi-phase flow simulations," Comput. Fluids, vol. 35, pp. 820–830,
  !! 2006.
  !!
?? copy :: compute_routineHeader( mrt_advRel_d3q19_incomp_les_omSpatial )
    !---------------------------------------------------------------------------
    integer :: iElem                            ! element counter
    integer :: iDir
    real(kind=rk) :: ucx
    real(kind=rk) pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) rho     ! local density
!    real(kind=rk) inv_rho ! inverse local density
    real(kind=rk) omega   ! relaxation parameter
    real(kind=rk) u_x     ! local x-velocity
    real(kind=rk) u_y     ! local y-velocity
    real(kind=rk) u_z     ! local z-velocity
    real(kind=rk) usq     ! square velocity
    ! MRT Variables
    real(kind=rk) s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0
    ! Variables required for LES
    real(kind=rk) :: fEq(19) ! equilibrium distribution
    real(kind=rk) :: nEq(19) ! non-equilibrium part
    real(kind=rk) :: q_sum, q_bgk(3,3)
    real(kind=rk) :: tau0, tau_turb, om_turb, smag_const
    !---------------------------------------------------------------------------
    rho0  = fieldProp(1)%fluid%rho0
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)
    omega = fieldProp(1)%fluid%rampOmegas( level )
    smag_const = fieldProp(1)%fluid%les%c_s
    tau0 = 1._rk/omega

    nodeloop: do iElem=nStart,nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, varSys%nScalars, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, varSys%nScalars, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, varSys%nScalars, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, varSys%nScalars, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, varSys%nScalars, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, varSys%nScalars, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, varSys%nScalars, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, varSys%nScalars, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, varSys%nScalars, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, varSys%nScalars, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, varSys%nScalars, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, varSys%nScalars, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, varSys%nScalars, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, varSys%nScalars, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, varSys%nScalars, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, varSys%nScalars, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, varSys%nScalars, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, varSys%nScalars, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, varSys%nScalars, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! local x-, y- and z-velocity
      u_x = pdfTmp(18) + pdfTmp(17) + pdfTmp( 4)                               &
        & + pdfTmp(14) + pdfTmp(12)                                            &
        & - pdfTmp(16) - pdfTmp( 1) - pdfTmp(15)                               &
        & - pdfTmp(13) - pdfTmp(11)
                         
      u_y = pdfTmp(18) + pdfTmp( 5)  + pdfTmp(16)                              &
        & + pdfTmp(10) + pdfTmp( 9)                                            &
        & - pdfTmp(15) - pdfTmp( 2 ) - pdfTmp(17)                              &
        & - pdfTmp( 8) - pdfTmp( 7)

      u_z = pdfTmp( 6) + pdfTmp(14) + pdfTmp(10)                               &
        & + pdfTmp(13) + pdfTmp( 8)                                            &
        & - pdfTmp( 3) - pdfTmp(12) - pdfTmp( 9)                               &
        & - pdfTmp(11) - pdfTmp( 7)

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho0*usq
      meq( 4) = u_x*rho0 
      meq( 6) = u_y*rho0 
      meq( 8) = u_z*rho0 
      meq(10) = rho0*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho0*( u_y*u_y - u_z*u_z )
      meq(14) = rho0*u_x*u_y
      meq(15) = rho0*u_y*u_z
      meq(16) = rho0*u_x*u_z

      !---------------------- LES Part -----------------------------------------
      !\todo MH: Replace this with Kannans equilibrium calculation derive%do
      ! This must be done outside the loop to allow for vectorized calculation 
      ! Calculate the equilibrium distribution function
      do iDir = 1, layout%stencil(1)%QQ

        ! velocity times lattice unit velocity
        ucx = real( layout%stencil(1)%cxDir( 1, iDir ), kind=rk)*u_x           &
          & + real( layout%stencil(1)%cxDir( 2, iDir ), kind=rk)*u_y           &
          & + real( layout%stencil(1)%cxDir( 3, iDir ), kind=rk)*u_z

        ! calculate equilibrium density
        fEq( iDir ) = layout%weight( iDir ) * ( rho + rho0 * ( ucx*cs2inv      &
          &         + ucx*ucx*t2cs4inv - usq*t2cs2inv ) )

      enddo
!      fEq(:) = getEquilibrium( density = rho, velocity = (/u_x, u_y, u_z /), layout = layout)

      ! Calculate the non-equilibrium part
      nEq(:) = pdfTmp(:) - fEq(:)

      ! Now calculate the stress tensor part
      q_bgk      = 0._rk
      q_bgk(1,1) = sum( real(layout%stencil(1)%cxDir( 1, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(2,1) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,2) = q_bgk(2,1)
      q_bgk(2,2) = sum( real(layout%stencil(1)%cxDir( 2, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(3,3) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 3, :),kind=rk)*nEq(:) )
      q_bgk(3,1) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 1, :),kind=rk)*nEq(:) )
      q_bgk(1,3) = q_bgk(3,1)
      q_bgk(3,2) = sum( real(layout%stencil(1)%cxDir( 3, :)                    &
        &                   *layout%stencil(1)%cxDir( 2, :),kind=rk)*nEq(:) )
      q_bgk(2,3) = q_bgk(3,2)

      q_sum = sqrt(2._rk*(q_bgk(1,1)*q_bgk(1,1)                                &
        &               + q_bgk(2,1)*q_bgk(2,1)                                &
        &               + q_bgk(1,2)*q_bgk(1,2)                                &
        &               + q_bgk(2,2)*q_bgk(2,2)                                &
        &               + q_bgk(3,1)*q_bgk(3,1)                                &
        &               + q_bgk(3,2)*q_bgk(3,2)                                &
        &               + q_bgk(1,3)*q_bgk(1,3)                                &
        &               + q_bgk(2,3)*q_bgk(2,3)                                &
        &               + q_bgk(3,3)*q_bgk(3,3) ) )
      ! Get the spatial omega definition directly into tau0
      tau0 = 1._rk/(fieldProp(1)%fluid%viscSpatial%stored( level )%            &
        &                                                   val( iElem )*omega)

      ! and the turbulent part of the viscosity
      tau_turb = 0.5_rk *(sqrt(tau0*tau0 + 2._rk*smag_const                    &
        &              *smag_const/fieldProp(1)%fluid%rho0*cs4inv*q_sum)-tau0)
      om_turb = 1._rk/(tau0+tau_turb)
      s_mrt( 10 )     = om_turb
      s_mrt( 12 )     = om_turb
      s_mrt( 14:16 )  = om_turb
      ! Now the relaxation parameters are changed based on the strain rate
      !---------------------- LES Part -----------------------------------------

!Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !jx
      mneq(4) = u_x 
      
      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  u_y
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  u_z
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      
      ! MATRIX looks ok
      
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = 0.0_rk
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = 0.0_rk
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = 0.0_rk
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 

      ! Transformation back to 

      ! CHECK MINV ploed

      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_incomp_les_omSpatial
!******************************************************************************!


!******************************************************************************!
  !> Advection relaxation routine for the MRT model.
  !! 
?? copy :: compute_routineHeader( mrt_advRel_d3q19_incomp_fixedVel )
    !---------------------------------------------------------------------------
    integer :: iElem      ! element counter
    real(kind=rk) :: pdfTmp( layout%stencil(1)%QQ ) ! temporary local pdf values
    real(kind=rk) :: rho     ! local density
    real(kind=rk) :: u_x     ! local x-velocity
    real(kind=rk) :: u_y     ! local y-velocity
    real(kind=rk) :: u_z     ! local z-velocity
    real(kind=rk) :: usq     ! square velocity
    integer, parameter :: iField = 1
    ! MRT Variables
    real(kind=rk) :: s_mrt( layout%stencil(1)%QQ ) ,meq( layout%stencil(1)%QQ )
    real(kind=rk) :: mneq( layout%stencil(1)%QQ ) ,mout( layout%stencil(1)%QQ )
    real(kind=rk) :: rho0
    !---------------------------------------------------------------------------

    rho0  = fieldProp(1)%fluid%rho0
    s_mrt( :)= fieldProp(1)%fluid%mrt( level )%s_mrt(:)

    nodeloop: do iElem=nStart,nEnd
      !> First load all local values into temp array
      pdfTmp(  1 ) = inState( ?FETCH?(  1, iElem, QQ, neigh ))
      pdfTmp(  2 ) = inState( ?FETCH?(  2, iElem, QQ, neigh ))
      pdfTmp(  3 ) = inState( ?FETCH?(  3, iElem, QQ, neigh ))
      pdfTmp(  4 ) = inState( ?FETCH?(  4, iElem, QQ, neigh ))
      pdfTmp(  5 ) = inState( ?FETCH?(  5, iElem, QQ, neigh ))
      pdfTmp(  6 ) = inState( ?FETCH?(  6, iElem, QQ, neigh ))
      pdfTmp(  7 ) = inState( ?FETCH?(  7, iElem, QQ, neigh ))
      pdfTmp(  8 ) = inState( ?FETCH?(  8, iElem, QQ, neigh ))
      pdfTmp(  9 ) = inState( ?FETCH?(  9, iElem, QQ, neigh ))
      pdfTmp( 10 ) = inState( ?FETCH?( 10, iElem, QQ, neigh ))
      pdfTmp( 11 ) = inState( ?FETCH?( 11, iElem, QQ, neigh ))
      pdfTmp( 12 ) = inState( ?FETCH?( 12, iElem, QQ, neigh ))
      pdfTmp( 13 ) = inState( ?FETCH?( 13, iElem, QQ, neigh ))
      pdfTmp( 14 ) = inState( ?FETCH?( 14, iElem, QQ, neigh ))
      pdfTmp( 15 ) = inState( ?FETCH?( 15, iElem, QQ, neigh ))
      pdfTmp( 16 ) = inState( ?FETCH?( 16, iElem, QQ, neigh ))
      pdfTmp( 17 ) = inState( ?FETCH?( 17, iElem, QQ, neigh ))
      pdfTmp( 18 ) = inState( ?FETCH?( 18, iElem, QQ, neigh ))
      pdfTmp( 19 ) = inState( ?FETCH?( 19, iElem, QQ, neigh ))

      ! local density
      rho = pdfTmp(18) + pdfTmp( 5) + pdfTmp(16)                               &
        & + pdfTmp( 1) + pdfTmp(15) + pdfTmp( 2)                               &
        & + pdfTmp(17) + pdfTmp( 4) + pdfTmp( 6)                               &
        & + pdfTmp(14) + pdfTmp(10) + pdfTmp(13)                               &
        & + pdfTmp( 8) + pdfTmp( 3) + pdfTmp(12)                               &
        & + pdfTmp( 9) + pdfTmp(11) + pdfTmp( 7)                               &
        & + pdfTmp(19)

      ! use
      u_x = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(1)%       &
        &                                          stored( level )%val( iElem )
      u_y = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(2)%       &
        &                                          stored( level )%val( iElem )
      u_z = scheme(iScheme)%field(iField)%fieldprop%fluid%velSpatial(3)%       &
        &                                          stored( level )%val( iElem )

      ! square velocity and derived constants
      usq = u_x*u_x + u_y*u_y + u_z*u_z

      !-------------------------------------------------------------------------
      ! now calculate mrt against the bgk
      !
      ! Equilibrium moments
      meq(1:layout%stencil(1)%QQ) =  0.0_rk 

      meq( 1) = rho 
      meq( 2) = rho0*usq
      meq( 4) = u_x*rho0 
      meq( 6) = u_y*rho0 
      meq( 8) = u_z*rho0 
      meq(10) = rho0*(2.0_rk*u_x*u_x - ( u_y*u_y + u_z*u_z ))
      meq(12) = rho0*( u_y*u_y - u_z*u_z )
      meq(14) = rho0*u_x*u_y
      meq(15) = rho0*u_y*u_z
      meq(16) = rho0*u_x*u_z

      ! Non-equilibrium moments
      mneq(1) = rho
      mneq(2) = -pdfTmp(19) + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)               &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)
      !epsilon
      mneq(3) =  pdfTmp(19)- 2._rk*(pdfTmp( 4)+ pdfTmp( 1)+ pdfTmp( 5)         &
        &     + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))+ pdfTmp(18)                &
        &     + pdfTmp(15)+ pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)                 &
        &     + pdfTmp(13)+ pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)                 &
        &     + pdfTmp( 8)

      !jx
      mneq(4) =  pdfTmp( 4)- pdfTmp( 1)+ pdfTmp(18)                            &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      
      !qx
      mneq(5) =  -2.0_rk*( pdfTmp( 4)- pdfTmp( 1))+pdfTmp(18)                  &
        &     - pdfTmp(15)+ pdfTmp(17)                                         &
        &     - pdfTmp(16)+ pdfTmp(14)- pdfTmp(11)+ pdfTmp(12)                 &
        &     - pdfTmp(13)
      !jy
      mneq(6) =  pdfTmp( 5)- pdfTmp( 2)+ pdfTmp(18)                            &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !qy
      mneq(7) =  -2.0_rk*(pdfTmp( 5)- pdfTmp( 2))+ pdfTmp(18)                  &
        &     - pdfTmp(15)- pdfTmp(17)                                         &
        &     + pdfTmp(16)+ pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)                 &
        &     - pdfTmp( 8)
      !jz
      mneq(8) =  pdfTmp( 6)- pdfTmp( 3)                                        &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)      
      !qz
      mneq(9) =  -2.0_rk*(pdfTmp( 6)- pdfTmp( 3))                              &
        &     + pdfTmp(14)- pdfTmp(11)- pdfTmp(12)  + pdfTmp(13)               &
        &     + pdfTmp(10)- pdfTmp( 7)- pdfTmp( 9)  + pdfTmp( 8)
      !3pxx
      mneq(10) = 2.0_rk*(pdfTmp( 4)+ pdfTmp( 1))-(pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3))                           &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk* (pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !3pixx
      mneq(11) = -2._rk*(pdfTmp( 4)+ pdfTmp( 1))+ pdfTmp( 5)                   &
        &      + pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      + pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13)                &
        &      -2.0_rk*( pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)+ pdfTmp( 8))
      !pww
      mneq(12) = pdfTmp( 5)                                                    &
        &      + pdfTmp( 2)- pdfTmp( 6)- pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !piww
      mneq(13) = -pdfTmp( 5)                                                   &
        &      - pdfTmp( 2)+ pdfTmp( 6)+ pdfTmp( 3)                            &
        &      + pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)+ pdfTmp(16)                &
        &      -(pdfTmp(14)+ pdfTmp(11)+ pdfTmp(12)+ pdfTmp(13))
      !pxy
      mneq(14) = pdfTmp(18)+ pdfTmp(15)- pdfTmp(17)- pdfTmp(16)
      !pyz
      mneq(15) = pdfTmp(10)+ pdfTmp( 7)- pdfTmp( 9)- pdfTmp( 8)
      !pxz
      mneq(16) = pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)- pdfTmp(13)
      !mx
      mneq(17) = pdfTmp(18)- pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)                &
        &      - pdfTmp(14)+ pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)
      !my
      mneq(18) = -pdfTmp(18)+ pdfTmp(15)+ pdfTmp(17)- pdfTmp(16)               &
        &      + pdfTmp(10)- pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      !mz
      mneq(19) = +pdfTmp(14)- pdfTmp(11)- pdfTmp(12)+ pdfTmp(13)               &
        &      - pdfTmp(10)+ pdfTmp( 7)+ pdfTmp( 9)- pdfTmp( 8)
      
      mout( 1) = 0.0_rk
      mout( 2) = s_mrt( 2)*(mneq( 2) - meq( 2)) 
      mout( 3) = s_mrt( 3)*(mneq( 3) - meq( 3))
      mout( 4) = s_mrt( 4)*(mneq( 4) - meq( 4)) 
      mout( 5) = s_mrt( 5)*(mneq( 5) - meq( 5)) 
      mout( 6) = s_mrt( 6)*(mneq( 6) - meq( 6)) 
      mout( 7) = s_mrt( 7)*(mneq( 7) - meq( 7)) 
      mout( 8) = s_mrt( 8)*(mneq( 8) - meq( 8)) 
      mout( 9) = s_mrt( 9)*(mneq( 9) - meq( 9)) 
      mout(10) = s_mrt(10)*(mneq(10) - meq(10)) 
      mout(11) = s_mrt(11)*(mneq(11) - meq(11)) 
      mout(12) = s_mrt(12)*(mneq(12) - meq(12)) 
      mout(13) = s_mrt(13)*(mneq(13) - meq(13)) 
      mout(14) = s_mrt(14)*(mneq(14) - meq(14)) 
      mout(15) = s_mrt(15)*(mneq(15) - meq(15)) 
      mout(16) = s_mrt(16)*(mneq(16) - meq(16)) 
      mout(17) = s_mrt(17)*(mneq(17)) 
      mout(18) = s_mrt(18)*(mneq(18)) 
      mout(19) = s_mrt(19)*(mneq(19)) 

      ! Transformation back to 

      ! CHECK MINV ploed

      outstate( ?SAVE?( 19,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(19) -(1._rk/3._rk *mout(1) -1._rk/2._rk*mout(2)   &
        &           + 1._rk/6._rk*mout(3))

      outstate( ?SAVE?(  4,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 4) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           + 1._rk/6._rk *mout(4) - 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  1,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 1) -(1._rk/18._rk*mout(1) - 1._rk/18._rk*mout(3) &
        &           - 1._rk/6._rk *mout(4) + 1._rk/6._rk *mout(5)              &
        &           + 1._rk/12._rk*mout(10) - 1._rk/12._rk*mout(11))

      outstate( ?SAVE?(  5,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 5) -(1._rk/18._rk*mout(1) -1._rk/18._rk*mout(3)  &
        &           + 1._rk/6._rk*mout(6) - 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  2,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 2) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(6) + 1._rk/6._rk*mout(7)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           + 1._rk/8._rk*mout(12) -1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  6,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 6) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           + 1._rk/6._rk*mout(8) - 1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?(  3,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 3) -(1._rk/18._rk*mout(1)-1._rk/18._rk*mout(3)   &
        &           - 1._rk/6._rk*mout(8)+  1._rk/6._rk*mout(9)                &
        &           - 1._rk/24._rk*mout(10) + 1._rk/24._rk*mout(11)            &
        &           - 1._rk/8._rk*mout(12)+ 1._rk/8._rk*mout(13))

      outstate( ?SAVE?( 18,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(18) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 15,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(15) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 17,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(17) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(6)              &
        &           - 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+ 1._rk/16._rk*mout(12)             &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           + 1._rk/8._rk*mout(17)+  1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 16,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(16) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(6)              &
        &           + 1._rk/24._rk*mout(7)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)+1._rk/16._rk*mout(12)              &
        &           + 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(14)             &
        &           - 1._rk/8._rk*mout(17) -1._rk/8._rk*mout(18))

      outstate( ?SAVE?( 14,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(14) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)+ 1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 11,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(11) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) + 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 12,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(12) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(4)               &
        &           + 1._rk/24._rk*mout(5) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           - 1._rk/8._rk*mout(17)-1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 13,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(13) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(4)                &
        &           - 1._rk/24._rk*mout(5) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)+ 1._rk/48._rk*mout(10)              &
        &           + 1._rk/48._rk*mout(11)-1._rk/16._rk*mout(12)              &
        &           - 1._rk/16._rk*mout(13) - 1._rk/4._rk*mout(16)             &
        &           + 1._rk/8._rk*mout(17)+1._rk/8._rk*mout(19))

      outstate( ?SAVE?( 10,iElem,QQ,neigh )) =                                 &
        &             pdfTmp(10) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+ 1._rk/4._rk*mout(15)              &
        &           + 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  7,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 7) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)+1._rk/4._rk*mout(15)               &
        &           - 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  9,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 9) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)+ 1._rk/12._rk*mout(6)               &
        &           + 1._rk/24._rk*mout(7) - 1._rk/12._rk*mout(8)              &
        &           - 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11)-1._rk/4._rk*mout(15)               &
        &           + 1._rk/8._rk*mout(18) + 1._rk/8._rk*mout(19))

      outstate( ?SAVE?(  8,iElem,QQ,neigh )) =                                 &
        &             pdfTmp( 8) -(1._rk/36._rk*mout(1)+ 1._rk/24._rk*mout(2)  &
        &           + 1._rk/72._rk*mout(3)-1._rk/12._rk*mout(6)                &
        &           - 1._rk/24._rk*mout(7) + 1._rk/12._rk*mout(8)              &
        &           + 1._rk/24._rk*mout(9)-1._rk/24._rk*mout(10)               &
        &           - 1._rk/24._rk*mout(11) -1._rk/4._rk*mout(15)              &
        &           - 1._rk/8._rk*mout(18) - 1._rk/8._rk*mout(19))

    enddo nodeloop

  end subroutine mrt_advRel_d3q19_incomp_fixedVel
!******************************************************************************!


end module mus_mrt_module
!******************************************************************************!
