! See copyright notice in the COPYRIGHT file.
!******************************************************************************!
!> This module creates the required treeID lists, from which then the state
!! vector and the neighbor information can be constructed.
!!
!! Also, the structures for the ghost and halo elements are created.
!! For the ghost cells, this means establishing the dependencies. The source
!! element positions for each ghost is stored. For the halo elements, the
!! information from where to get the element information is stored in the mpi
!! buffers.
!!
!! \author{ Harald Klimach, Manuel Hasert, Jens Zudrop }
!!
module tem_construction_module

  ! include treelm modules
  use mpi
  use env_module,            only: long_k, rk, globalMaxLevels,                &
    &                              io_buffer_size, pathLen
  use tem_aux_module,        only: tem_abort
  use tem_dyn_array_module,  only: dyn_longArray_type, dyn_intArray_type,      &
    &                              init, append, expand, destroy,              &
    &                              PositionOfVal
  use tem_grow_array_module, only: grw_longArray_type, grw_intArray_type,      &
    &                              init, append, expand, destroy
  use treelmesh_module,      only: treelmesh_type
  use tem_geometry_module,   only: tem_tIdInfo, tem_PosOfPath, tem_baryOfID
  use tem_tools_module,      only: tem_PositionInSorted, tem_horizontalSpacer
  use tem_logging_module,    only: logUnit, tem_logging_isActive, tem_toStr
  use tem_debug_module,      only: tem_debug, main_debug, dbgUnit
  use tem_comm_env_module,   only: tem_comm_env_type
  use tem_comm_module,       only: tem_communication_type, tem_comm_dumpType,  &
    &                              tem_commPattern_type
  use tem_topology_module,   only: tem_LevelOf, tem_FirstIdAtLevel,            &
    &                              tem_PathComparison, tem_ParentOf,           &
    &                              tem_directChildren, tem_childNumber,        &
    &                              tem_PathOf, tem_path_type, tem_coordOfId,   &
    &                              tem_IDofCoord
  use tem_property_module,   only: prp_hasBnd, prp_missingNeigh
  use tem_bc_prop_module,    only: tem_bc_prop_type
  use tem_stencil_module,    only: tem_stencilHeader_type,                     &
    &                              tem_stencilElement_type,                    &
    &                              tem_stencil_map_toTreelmDef,                &
    &                              tem_stencil_getHeaderPos
  use tem_element_module,    only: tem_element_type, tem_element_dump,         &
    &                              init, append, PositionOfVal, truncate,      &
    &                              eT_fluidElem, eT_nonExistingElem,           &
    &                              eT_haloElem, destroy,                       &
    &                              eT_ghostFromCoarserElem, eT_minNumber,      &
    &                              eT_ghostFromFinerElem, eT_maxNumber,        &
    &                              eT_minNumRelevant, eT_undefinedElem,        &
    &                              eT_distributedGhostFromFinerElem
  use tem_halo_module,       only: dyn_haloArray_type, append, init, destroy,  &
    &                              tem_halo_type, PositionOfVal

  implicit none

  private

  public :: tem_levelDesc_type
  public :: tem_levelNeighbor_type
  public :: depSource_type
  public :: tem_updateTree_properties

  public :: prp_nLists

  public :: tem_find_allElements
  public :: tem_build_horizontalDependencies
  public :: tem_build_verticalDependencies
  public :: tem_cleanupDependencyArrays
  public :: tem_cleanup_arrays
  public :: tem_init_elemLevels
  public :: tem_eTypeOfId
  public :: tem_eTypeOfTotalPos
  public :: tem_treeIDinTotal
  public :: tem_killArrays
  public :: tem_find_depProc
  public :: tem_elemList_dump
  public :: tem_debug_HorizontalDependencies
  public :: tem_dumpPropBits
  public :: tem_dumpTreeIDlists
  public :: tem_debugDependencies
  public :: identify_stencilNeigh

  !> limit for searching neighbors
  !! * For acoustic scaling (2 coarser elements required) set to 1
  !! * For diffusive scaling (4 coarser elements required) set to 3
  integer, save :: nestingLimit

  !> Hash to quickly identify if an element was reconstructed before.
  !! If so, it is part of the hash
  integer(kind=long_k), allocatable, save :: hash(:)
  integer, allocatable, save :: hash_elemPos(:)
  !> Entries in the sh
  integer(kind=long_k) :: nHashes

  !> identification parameters for different lists
  integer, parameter :: prp_nLists    =  4

  !\todo replace by element%stencil%tID
  !> includes the direct neighbors of each tree ID
  type tem_levelNeighbor_type
    !> array of the neighbors in the resulting totalList.
    !! Use this one in the solver!
    !! size: stencil%QQN, nElems(to treat with stencil)
    integer,allocatable :: nghElems(:,:)
  end type tem_levelNeighbor_type

  !> Type to specify the dependencies of ghost and halo cells.
  !! E.g.: used to specify which cells have to be known to be able
  !! to interpolate a ghost/halo cell
  type depSource_type
    !> the source level, from where the current ghost element
    !! gets the source elements for the interpolation
    integer                  ::  dependencyLevel = -1
    !> position of the source elements in the totalList
    type( grw_intArray_type ) :: elem
    !> position of the source elements in the subbuffer
    type( grw_intArray_type ) :: elemBuffer
    !> Interpolation weight for each source element specified above
    real(kind=rk), allocatable :: weight(:)
    real(kind=rk) :: coord(3)
    !> What's this stencil?
    integer, allocatable :: stencil(:,:)
    integer :: childNum
  end type depSource_type

  !> detailed information of a complete level of elements
  !! including all treeIDs, properties and neighbors as well as informations
  !! about ghost/halo cells and its dependencies for
  !! interpolation/reconstruction
  type tem_levelDesc_type

    !---------new types for element description---------------------
    type(tem_element_type) :: elem
    !> This list includes treeIDs for which additionally
    !! neighbors have to be identified
    type(dyn_longArray_type) :: require
    type(dyn_haloArray_type) :: halosPrc
    !> points to the position of the element in the elem array
    !type(grw_intArray_type) :: fluid
    !type(grw_intArray_type) :: halo

    !> ghost treeIDs on current level, i.e. elements that are virtual, but local
    !type(grw_intArray_type) :: ghostFromFiner
    !type(grw_intArray_type) :: ghostFromCoarser
    !---------------------------------------------------------------

    !> list of treeIDs for this level. Consists of ordered treeIDs of
    !! first fluid, then ghost, then halo elements.
    !! total:
    !! \verbatim
    !! +---------------------+--------------------+------------------+---------+
    !! |........fluid........|..ghostFromCoarser..|..ghostFromFiner..|...halo..|
    !! +---------------------+--------------------+------------------+---------+
    !! \endverbatim
    integer(kind=long_k), allocatable :: total(:)
    !> pointer to elem%tID list
    !! Array size: nElems ( nElems_fluid+nElems_ghostFromCoarser
    !!                     +nElems_ghostFromFiner+nElems_halo )
    integer, allocatable :: totalPnt(:)
    !> list of property bits for this level. the same order as total list
    integer(kind=long_k), allocatable :: property(:)
    !> pointer from the levelDescriptor to the original treeID list
    !! ( only for fluids )
    integer, allocatable :: pntTID(:)

    !> neighbor relations for all fluid elements.
    !! Dimension: number of stencils
    !! We store the positions of the neighbor elements inside the total-list.
    !! If a fluid element does not have a neighbor in a direction
    !! (e.g. because of a boundary in that direction), we store the boundary ID
    !! as negative to indicate, that it is not a regular neighbor.
    type( tem_levelNeighbor_type ), allocatable :: neigh(:)

    !> Dependencies for ghost elements
    !! To reconstruct all the data you should
    !! iterate over this list and reconstruct the ghost elements
    !! with source element information from these data types
    !! data. Up = to coarser, down = to finer
    type( depSource_type ), allocatable :: depFromFiner(:)
    type( depSource_type ), allocatable :: depFromCoarser(:)
    !> Store the source elements for the ghostFromFiner
    type(dyn_intArray_type) :: sourceFromFiner
    !> Store the source elements for the ghostFromCoarser
    type(dyn_intArray_type) :: sourceFromCoarser
    !> Buffer storing intermediate values of the source elements for
    !! the interpolation
    real(kind=rk), allocatable :: intpBufFromFiner(:,:)
    real(kind=rk), allocatable :: intpBufFromCoarser(:,:)
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromCoarser_high
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromCoarser_low
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromFiner_high
    !> List to store high order interpolation ghost elements
    type(grw_intArray_type) :: intpFromFiner_low

    !> pointing to the position of boundary elements
    !! in the levelDescriptor total list
    type(dyn_intArray_type) :: bc_elemBuffer

    !> Offsets in the assembled lists for
    !! fluid (1), ghostFromCoarser(2), ghostFromFiner(3) and halo(4) elements
    !! for the assembled lists, i.e the totalList, invSorted, ...
    !! gets the values (0, nElems_fluid,
    !!   nElems_fluid+nElems_ghostCoarser,
    !!   nElems_fluid+nElems_ghostCoarser+nELems_ghostFiner)
    integer :: offset( 2, eT_minNumber:eT_maxNumber ) = 0

    !> send buffer
    type( tem_communication_type ) :: sendbuffer
    !> send buffer from coarser elements
    type( tem_communication_type ) :: sendbufferFromCoarser
    !> send buffer from finer elements
    type( tem_communication_type ) :: sendbufferFromFiner
    !> receive buffer
    type( tem_communication_type ) :: recvbuffer
    !> receive buffer from coarser elements
    type( tem_communication_type ) :: recvbufferFromCoarser
    !> receive buffer from finer elements
    type( tem_communication_type ) :: recvbufferFromFiner

    !> number of fluid elements
    integer :: nElems_fluid
    !> number of ghost elements from coarser
    integer :: nElems_ghostFromCoarser
    !> number of ghost elements from finer
    integer :: nElems_ghostFromFiner
    !> number of halo elements
    integer :: nElems_halo
    !> total number of elements
    integer :: nElems

  end type tem_levelDesc_type

  contains

!*******************************************************************************
  !> call this routine from your geometry initialization routine in the solver
  !! create all the necessary level-wise objects, such as element lists,
  !! dependencies
  !!
  !! 1.) build all dependencies for halos and ghost which
  !! are needed for interpolation/reconstruction (including MPI communications)
  !! 2.) build the pointers for each element to its neighbors/stencil elements.
  !! All this information is stored in tem_levelDesc_type
  !!
  subroutine tem_find_allElements( tree, levelDesc, levelPointer,              &
    &                              computeStencil, commPattern, cleanup,       &
    &                              reqNesting, proc )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(inout) :: tree
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> Pointer from treeIDlist entry to level-wise fluid part of total list
    integer, allocatable, intent(out) :: levelPointer(:)
    !> array of all stencils used in the simulation
    type(tem_stencilHeader_type) :: computeStencil(:)
    !> the communication pattern used
    type(tem_commPattern_type), intent(in)    :: commPattern
    !> nesting level
    integer, intent(in), optional :: reqNesting
    !> cleanup arrays afterwards?
    logical, intent(in), optional :: cleanup
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc 
    !---------------------------------------------------------------------------
    ! Geometry and Tree related variables
    integer,allocatable :: eligible_child(:) ! possible children (can be 1...8)
    integer :: iLevel
    type(tem_path_type), allocatable :: pathFirst(:), pathLast(:)
    logical :: doAdditional
    logical :: clean_constructionArrays
    !---------------------------------------------------------------------------

    if( present( cleanup )) then
      clean_constructionArrays = cleanup
    else
      clean_constructionArrays = .false.
    end if
    if( present( reqNesting )) then
      nestingLimit = reqNesting
    else
      nestingLimit = 1
    end if
    write(logUnit(3),*) 'Building up the element and total list.'

    ! first and last path in every process
    allocate(pathFirst( tree%global%nParts ))
    allocate(pathLast( tree%global%nParts ))

    ! Build the pathlist for improved performance when finding local element
    ! positions since the tree is the same for every scheme the allocation only
    ! has to be performed once
    if ( .not. allocated(tree%pathList))then
      allocate( tree%pathList(tree%nElems) )
    end if

    ! Fill the pathList for each element in the treeID list
    tree%pathList =  tem_PathOf( tree%treeID )
    pathFirst     =  tem_PathOf( tree%Part_First )
    pathLast      =  tem_PathOf( tree%Part_Last )

    ! Step 1: allocate and initialize levelPointer
    call init_arrays( levelPointer, tree, eligible_child )

    ! Step 2: build levelDesc element list including identification of neighbor 
    ! elements
    call build_levelElements( levelDesc      = levelDesc,                      &
      &                       levelPointer   = levelPointer,                   &
      &                       tree           = tree,                           &
      &                       proc           = proc,                           &
      &                       computeStencil = computeStencil,                 &
      &                       pathFirst      = pathFirst,                      &
      &                       pathLast       = pathLast )

    ! Step 3: assign totalPnt to elem%tID in sorted fashion and prepare 
    ! haloPrc list
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      call identify_lists( tree, levelDesc(iLevel) )
    end do

    ! Step 4: Communicate halo elements
    ! Here, first local halos are send to their corresponding process to 
    ! check for existence of halo elements.
    ! Each process checks whether requested element exist and returns
    ! list of available elements.
    ! With this new information halo list is redefined.
    call communicate_elements( tree, proc, levelDesc, commPattern,             &
      &               pathFirst, pathLast, computeStencil )

    ! Step 5: do additional communication if there are elements in require list
    ! which are neighbors of higher order boundaries.
    call check_additionalComm( levelDesc, proc, doAdditional,                  &
      &                        tree%global%minlevel ) 

    ! If doAdditional then identify neighbors of higher order boundary 
    ! neighbor elements.
    ! After this identification, new halo elements might have to be added so 
    ! we need to communicate all halo elements again.
    if( doAdditional ) then
      ! passing only first stencil as this is the required compute stencil
      call identify_additionalNeigh( tree, proc, levelDesc,                    &
        &        pathFirst, pathLast, computeStencil(1) )
      call communicate_elements( tree, proc, levelDesc, commPattern,           &
        &        pathFirst, pathLast, computeStencil )
    end if

    ! Step 6: assemble levelDesc total(treeID) list, property list and 
    ! pntTID list in sorted fashion (fluids+ghosts+halos) 
    ! which are pre-assembled in element type 
    call assemble_lists( levelDesc, tree%global%minLevel )

    ! Step 7:
    call update_elemPosToTotalPos( levelDesc, levelPointer, tree,              &
      &                            computeStencil )
    ! Warning: Truncation introduces a memory peak because of copy
    ! operations! Better not use...
    !call truncate_lists( levelDesc, tree%global%minLevel )

    if( clean_constructionArrays ) then
      call tem_cleanupDependencyArrays( levelDesc )
    endif

    write(logUnit(3),*) 'Finished building up element and total list. '
    deallocate(pathFirst)
    deallocate(pathLast)
    deallocate(hash)
    deallocate(hash_elempos)

  end subroutine tem_find_allElements
!******************************************************************************!


!******************************************************************************!
  !> subroutine to find neighbours of cells
  !!
  !! Typically every element requires information from its neighbors to perform
  !! an update of its state.
  !! All such required neighbors constitute a so called \ref tem_stencil_module
  !! "stencil".
  !! \ref tem_build_horizontaldependencies "horizontal relations", or the 
  !! connectivity of elements on the same refinement
  !! level in the tree, are therefore essential for the stencils in typical
  !! mesh-based numerical schemes.
  !! We now sketch the method to find the connectivity of elements in the mesh
  !! with the help of the described mesh layout.
  !! In a distributed mesh, the first distinction has to be made with respect to
  !! processor ownership.
  !! \todo: tem_find_depNeigh does not exist any more
  !! Is the element in question a \ref tem_find_depNeigh "local or remote
  !! element"?
  !!
  !! \todo: tem_identify_local_element does not exist any more
  !! For all \ref tem_identify_local_element "local elements", their actual
  !! position in the sparse mesh has to be identified for a given 
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID".
  !!
  !! A \ref tem_topology_module::tem_path_type "comparison of two nodes to
  !! decide their relative position" in the ordered list
  !! of elements has to take into account the hierarchy of the mesh.
  !!
  !! As in the actual simulation only leaves will be present, and no overlap of
  !! refinement levels is allowed, this is a sufficient determination.
  !! However, such overlapping regions might be created to enable the
  !! interpolation between different levels with virtual elements.
  !! These special elements are distinguished anyway, so this does not pose any
  !! problem in the search for neighborhood relations.
  !!
  !! An element can be either identified by its
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID"
  !! or by a tuple with four integer entries for the spatial coordinates and the
  !! refinement level: \f$(x, y, z, L)\f$.
  !! This coordinate fully describes the spatial shape and position of an
  !! element and is important to determine spatial relations between elements.
  !! Conversion between treeIDs and coordinates can be achieved by the routines
  !! - \ref tem_geometry_module::tem_coordofid "Coordinate from TreeID"
  !! - \ref tem_geometry_module::tem_idofcoord "TreeID from Coordinate"
  !!
  !! The spatial indices are limited by the refinement level:
  !! \f$ x, y, z \in \mathbf{Z}_{\ge 0}: x, y, z < 2^L \f$.
  !! To avoid undefined coordinates, movements through the mesh by additions to
  !! indices are done in a modulo(\f$2^L\f$) safeguard resulting in an periodic
  !! universe cube.
  !! This enables the movement in the mesh in horizontal direction by index
  !! alteration and translation into a
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID" again.
  !! The described procedure is completely reversible, enabling the construction
  !! of the \ref treelmesh_module::treelmesh_type::treeid "treeID" for any
  !! given coordinate.
  !! Thus, the conversion between this coordinate and the serialized 
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID"
  !! encoding is fully described by the topology of the octree and the chosen
  !! space-filling curve ordering of elements, as explained above.
  !! 
  !! With this method we can describe the neighborhood of any given element
  !! independently of the actual solver by a simple list of relative offsets.
  !! In contrast to the generic horizontal relation, the vertical relation
  !! between child and parent nodes in the tree requires an interpolation
  !! operator between different refinement levels.
  !! This interpolation usually has to take into account solver specific
  !! requirements, but is otherwise quite isolated from the numerical operation
  !! on each refinement level.
  !! The \a TreElM library offers the solver a level-wise view, as suggested
  !! by the properties described above.
  !! To find all required neighbors in the distributed octree, the solver merely
  !! has to provide its horizontal dependencies.
  !! These are described with the help of an element specific \ref
  !! tem_stencil_module "stencil".
  !! A stencil is basically a set of element-offsets \f$(s_x, s_y, s_z)\f$,
  !! describing the relative positions of all required elements for a given
  !! element.
  !!
  !! \ref tem_construction_example_stencil
  !!
  subroutine tem_init_elemLevels( me, boundary, tree, computeStencil )
    !---------------------------------------------------------------------------
    !> neighbor list containing all the neighbours for the
    !! cells given in treeidsubset. Result of this routine
    type(tem_levelDesc_type), allocatable, intent(out) :: me(:)
    !> boundaries for the elements with bnd property set
    type(tem_bc_prop_type), intent(in) :: boundary
    !> subset of tree ids for which the neighbours will be specified
    type(treelmesh_type), intent(in)   :: tree
    !> the given stencil
    type(tem_stencilHeader_type), intent(inout) :: computeStencil(:)
    !---------------------------------------------------------------------------
    integer :: addedPos
    type(tem_stencilElement_type) :: tStencil
    integer :: iElem, nElemsBnd, iQQN, iLevel, nElems, nProcs, hashpos
    integer :: x(4), nStencils, iStencil, elemPos, nStencilElems
    integer :: indElem
    logical :: wasAdded
    integer(kind=long_k) :: tOffset
    integer(kind=long_k), allocatable :: tStencilIDs(:)
    integer(kind=long_k) :: elemTID
    !---------------------------------------------------------------------------

    nElems = size(tree%treeID)
    nProcs = size(tree%Part_First)

    ! Find a suitable hash size.
    ! The hash lookup is O(1), while the identification grows with
    ! log(nElems) and log(nProcs), yet allocation, deallocation and
    ! filling will cause some overhead of O(nHashes), this is an attempt to
    ! find a compromise of these parameters.
    nHashes = int(max(nElems, nProcs, 64), kind=long_k)

    ! \todo Find the closest smaller prime for the hash size, to minimize the
    !       number of collisions?
    ! Limit the size of the hash:
    nHashes = min(nHashes, int(io_buffer_size, kind=long_k))

    allocate(hash(0:int(nHashes-1)))
    allocate(hash_elemPos(0:int(nHashes-1)))
    ! Reset the hash
    hash = -1_long_k
    hash_elemPos = -1

    write(logUnit(5),*) 'Allocating level Descriptor ...'
    allocate( me( tree%global%minLevel:tree%global%maxLevel ))

    ! initialize dynamic require array for requireNeighNeigh
    ! and element type
    do iLevel = tree%global%minLevel,tree%global%maxLevel
      call init(me = me(iLevel)%require, unique = .true.)
      write(logUnit(6),*) 'Initializing elemList on Level ', iLevel
      call init( me = me( iLevel )%elem )
    end do

    nElemsBnd = 0
    write(logUnit(4),*) 'Searching the neighbors for each element ...'
    nStencils = size( computeStencil )

    ! Loop over all given stencils
    stencilLoop: do iStencil = 1, nStencils
      write(logUnit(5), "('  Identifying element neighbors on stencil: ', I0)")&
        &                 iStencil

      ! map user stencil to treelm definition
      call tem_stencil_map_toTreelmDef( me = computeStencil( iStencil ))

      if ( computeStencil(iStencil)%useAll ) then
        ! This stencil belongs to all elements in the tree (fluid stencil)
        nStencilElems = tree%nElems
      else
        if( iStencil == 1) then
          write(logUnit(1),*)'tem_init_elemLevels:'
          write(logUnit(1),*)'Error: The first stencil does not have the'//    &
            &                'useAll flag set'
          write(logUnit(1),*)'       The first one is considered the compute ' &
            &            //'stencil and'
          write(logUnit(1),*)'       needs to be set for all elements.'
          call tem_abort()
        end if
        ! This stencil belongs to only a subset of the tree 
        ! (boundary and IBM stencils) 
        nStencilElems = computeStencil( iStencil )%nElems
      end if

      write(logUnit(6),*) '    nElems       ', nStencilElems
      write(logUnit(6),*) '    QQN          ', &
        &                 computeStencil(iStencil)%QQN
      if ( allocated(tStencilIDs) ) deallocate( tStencilIDs )
      allocate( tStencilIDs( computeStencil(iStencil)%QQN ) )

      ! Loop over all the elements connected to the current stencil 
      ! computeStencil( iStencil )
      elemLoop: do indElem = 1, nStencilElems
        ! Reset the empty stencil
        ! Create the temporary stencil for assigning the positions of the 
        ! neighbor treeIDs which are added in elem%neighID
        ! also store the original stencil it depends on (depends = iStencil )
        call init( me        = tStencil,                                       &
          &        QQN       = computeStencil( iStencil )%QQN,                 &
          &        headerPos = iStencil )

        if( computeStencil( iStencil )%useAll ) then
          ! if the stencil is defined to be connected to all elements, 
          ! then simply loop over all the fluid elements
          iElem = indElem
        else
          ! if the stencil is defined for a subset, loop over the
          ! defined subset in %elem
          iElem = computeStencil( iStencil )%elem( indElem )
        end if
        ! assign fluid element with info from the tree list and properties
        ! from disk
        elemTID = tree%treeID( iElem )
        ! Get topological info about coordinates+level
        x = tem_CoordOfId( elemTID )
        ! offset is needed to compute neighbor ID from tem_IdOfCoord 
        tOffset = tem_FirstIdAtLevel( x(4) )

        ! Find position of this element in the elemList as it might be
        ! there already from previous stencil iterations
        hashpos = int(mod( elemTID, nHashes))
        if (hash(hashpos) /= elemTID) then ! cache miss
          call append( me         = me( x(4) )%elem,                           &
            &          tID        = elemTID,                                   &
            &          pntTID     = iElem,                                     &
            &          eType      = eT_fluidElem,                              &
            &          nNeighIDs  = computeStencil( iStencil )%QQN,            &
            &          sourceProc = tree%global%myPart+1,                      &
            &          property   = tree%ElemPropertyBits( iElem ),            &
            &          pos        = elemPos,                                   &
            &          wasAdded   = wasAdded )
          hash(hashpos) = elemTID
          hash_elemPos( hashpos ) = elemPos
        else ! cache hit 
          elemPos = hash_elemPos( hashpos )
        end if

        ! Only perform the boundary check based on the treeID boundary
        ! information for stencils which work on all elements (so we can
        ! actually read out the boundary information from the tree )
        if( computeStencil( iStencil )%useAll .and. iStencil == 1) then
          ! First check, if I have prp_hasBnd
          ! If I don't take all the elements, then I can't analyze boundaries
          if( btest( me(x(4))%elem%property%val( elemPos ), prp_hasBnd )) then
          ! If yes, there might be no neighbors
            nElemsBnd = nElemsBnd +1
            me(x(4))%elem%boundaryPos%val( elemPos ) = nElemsBnd
          end if
        end if ! useAll elements? (boundary information)

        ! find nearest neighbors. If neighbor is boundary then
        ! identify the boundary ID from boundary%boundary_ID list
        ! if its fluid then identify its treeID from tem_IdOfCoord
        tStencilIDs = 0_long_k

        directionLoop: do iQQN = 1, computeStencil( iStencil )%QQN
          if ( me(x(4))%elem%boundaryPos%val( elemPos ) > 0                    &
            &  .and. computeStencil( iStencil )%map( iQQN ) > 0 ) then
            ! If there is a boundary in the current direction
            ! ... add it to the temporary stencil
            if ( boundary%boundary_ID( computeStencil(iStencil)%map(iQQN),     &
              &                        me(x(4))%elem%boundaryPos%val(elemPos)  &
              &                      )  > 0 ) then
              ! Set the negative boundary ID as neighbor, so we don't need
              !! more arrays for boundary information
              tStencilIDs( iQQN )                                              &
                &  =  -boundary                                                &
                &      %boundary_ID( computeStencil(iStencil)%map(iQQN),       &
                &                    me(x(4))%elem%boundaryPos%val(elemPos) )

            else if ( boundary                                                 &
              &       %boundary_ID( computeStencil(iStencil)%map(iQQN),        &
              &                     me(x(4))%elem%boundaryPos%val(elemPos)     &
              &                   )  < 0 ) then
              ! Is a periodic neighbor! value is -treeID_periodicNeighbor
              ! Set the id given as treeID and make positive
              tStencilIDs( iQQN )                                              &
                &  = abs( boundary                                             &
                &         %boundary_ID( computeStencil(iStencil)%map(iQQN),    &
                &                       me(x(4))%elem%boundaryPos%val(elemPos) &
                &                     )                                        &
                &       )
            else
              ! ... there is NO boundary condition, so store the required
              ! neighbor
              ! Just store the neighboring treeID which will have to be
              ! found or reconstructed (might be an actual fluid element
              ! or ghost / halo)
              tStencilIDs( iQQN ) = tem_IdOfCoord(                             &
                &     [ x(1) + computeStencil( iStencil )%cxDir( 1, iQQN ),    &
                &       x(2) + computeStencil( iStencil )%cxDir( 2, iQQN ),    &
                &       x(3) + computeStencil( iStencil )%cxDir( 3, iQQN ),    &
                &       x(4) ], tOffset)
            endif
          else
            ! ... there is NO boundary condition, so store the required neighbor
            ! Just store the neighboring treeID which will have to be
            !! found or reconstructed (might be an actual fluid element
            !! or ghost / halo)
            tStencilIDs( iQQN ) = tem_IdOfCoord(                               &
              &       [ x(1) + computeStencil( iStencil )%cxDir( 1, iQQN ),    &
              &         x(2) + computeStencil( iStencil )%cxDir( 2, iQQN ),    &
              &         x(3) + computeStencil( iStencil )%cxDir( 3, iQQN ),    &
              &         x(4) ], tOffset)
          endif

          if( computeStencil( iStencil )%requireNeighNeigh                     &
            & .and. iStencil > 1                                               &
            & .and. tStencilIDs(iQQN) > 0 ) then
            ! If the neighbors of this stencil neigbors are required, add this 
            ! to the require array. This includes treeIDs for which additional 
            ! neighbors have to be identified.
            call append( me  = me(x(4))%require,                               &
              &          val = tStencilIDs(iQQN),                              &
              &          pos = addedPos )
          end if
        end do directionLoop

        ! Append the stencil to the neighID array
        do iQQN = 1, computeStencil( iStencil )%QQN
          call append( me       = me(x(4))%elem%neighID%val( elemPos ),        &
            &          val      = tStencilIDs( iQQN ),                         &
            &          pos      = addedPos,                                    &
            &          wasAdded = wasAdded )
          tStencil%tIDpos( iQQN ) = addedPos
        end do

        if( computeStencil( iStencil )%QQN > 0 ) then
          ! Append the temporary stencil to the element
          call append( me  = me(x(4))%elem%stencil%val( elemPos ),             &
            &          val = tStencil )
        end if
      end do elemLoop
    end do stencilLoop
    write(logUnit(4),*) 'Finished searching the neighbors for each element.'

    if (tem_logging_isActive(main_debug%logger, 5)) then
      do iLevel = tree%global%minLevel, tree%global%maxLevel
        call tem_elemList_dump( me      = me( iLevel )%elem,                  &
          &                     compact = .true.,                             &
          &                     nUnit   = dbgUnit(5),                         &
          &                     stencil = .true.,                             &
          &                     string  = 'after initializing level elements' )
        if( me( iLevel )%require%nVals > 0 ) then
          call tem_require_dump( me      = me( iLevel )%require,               &
            &                    compact = .true.,                             &
            &                    nUnit   = dbgUnit(5),                         &
            &                    string  = 'after initializing level elements' )
        end if
      end do
    end if

  end subroutine tem_init_elemLevels
!******************************************************************************!


!******************************************************************************!
  !> Assemble the fluid list and identify neighbor relations
  !! start with building up the ghost and halo element collection as well
  !!
  subroutine build_levelElements( levelDesc, levelPointer, tree, proc,         &
    &                             computeStencil, pathFirst, pathLast )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(inout)     :: tree
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc( tree%global%minLevel:)
    !> Pointer from treeIDlist entry to level-wise fluid part of total list
    integer, intent(inout)                  :: levelPointer(:)
    !> Process description to use.
    type(tem_comm_env_type), intent(in)     :: proc
    !> array of all stencils used in the simulation
    type(tem_stencilHeader_type)            :: computeStencil(:)
    !> first and last treeID path in every process
    type(tem_path_type), intent(in)         :: pathFirst(:), pathLast(:)
    !---------------------------------------------------------------------------
    integer(kind=long_k) :: nTreeID   ! neighboring treeID
    integer(kind=long_k) :: elemTID   ! current element treeID
    integer :: nElems, nProcs, elemPos
    integer :: iElem, iNeighElem, iLevel, iStencil
    integer :: tLevel
    ! Nesting level to make sure all neighbors of first row of ghost elements
    ! are reconstructed
    integer :: nesting
    ! position of where to read the stencil neighbor treeid in the element
    integer :: neighPos
    !---------------------------------------------------------------------------
    nElems = size(tree%treeID)
    nProcs = size(tree%Part_First)

    ! Reset the has
    hash = -1_long_k
    hash_elemPos = -1

    write(logUnit(2),*) 'Building level-wise fluid list ...'

    ! Create levelpointer: create the pointer from the original
    ! treeID list to the level-wise lists
    ! KM: This steps is done in update_elemPosToTotalPos so commented here.
    ! \todo to be removed if no problem encountered. 
    !  Also remove LevelPointer from routine argument
    !KM!do iElem = 1, nElems
    !KM!  tLevel = tem_LevelOf( tree%treeID( iElem ) )
    !KM!  elemTID = tree%treeID( iElem )
    !KM!  levelPointer( iElem ) = PositionOfVal( me  = levelDesc(tLevel)%elem%tID, &
    !KM!    &                                    val = elemTID )
    !KM!end do

    ! number of fluid elements in each level
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      levelDesc( iLevel )%nElems_fluid = levelDesc( iLevel )%elem%tID%nVals
    end do

    ! step 1. Identify the neighbors of each elements including ghost from finer
    ! and ghost from coarser
    ! Now iterate over all the neighbor lists and add the required elements if
    ! necessary
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      do iElem = 1, levelDesc( iLevel )%nElems_fluid
        do iStencil = 1, levelDesc( iLevel )%elem%stencil%val(iElem )%nVals
          do iNeighElem = 1, levelDesc( iLevel )%elem                          &
            &                         %stencil%val( iElem )%val( iStencil )%QQN
            ! Same as in identify_additionalNeigh
            neighPos = levelDesc( iLevel )%elem%stencil%val( iElem )           &
              &                           %val( iStencil )%tIDpos( iNeighElem ) 
            if( neighPos > 0 ) then
              nTreeID = levelDesc( iLevel )%elem%neighID%val( iElem )%         &
                &                                               val( neighPos )
              nesting = 0
              ! identify the process in which requested nTreeID is located
              !   - if remote process add to elem list as halo element.
              !   - if different level add to elem list as ghostFromFiner or
              !     ghostFromCoarser.
              ! Also return position (elemPos) of nTreeID in elem%tID list
              call identify_elements( nTreeID        = nTreeID,                &
                &                     tree           = tree,                   &
                &                     pathFirst      = pathFirst,              &
                &                     pathLast       = pathLast,               &
                &                     levelDesc      = levelDesc,              &
                &                     elemPos        = elemPos,                &
                &                     proc           = proc,                   &
                &                     nesting        = nesting,                &
                &                     computeStencil = computeStencil(1) )
            else
              elemPos = 0 
            end if
            levelDesc( iLevel )%elem%stencil%val( iElem )%val( iStencil )      &
              &                               %totalPos( iNeighElem ) = elemPos
          end do ! iNeighElem!
        end do ! iStencil
      end do ! iElem
    end do ! iLevel

    ! Find neighbors of neighbors for elements in the require list.
    ! Same procedure as above but identify neighbors of elements in require 
    ! list along 1st stencil directions.
    ! What exactly is the require list for?
    ! - Used ONLY for boundary stencil with higher order neighbors i.e
    !   only when require nVals > 0
    call identify_additionalNeigh( tree           = tree,             &
      &                            proc           = proc,             &
      &                            levelDesc      = levelDesc,        &
      &                            pathFirst      = pathFirst,        &
      &                            pathLast       = pathLast,         &
      &                            computeStencil = computeStencil(1) )
 
    ! dump elemList into debug unit
    if (tem_logging_isActive(main_debug%logger, 5)) then
      do iLevel = tree%global%minLevel, tree%global%maxLevel
          call tem_elemList_dump( me      = levelDesc( iLevel )%elem, &
            &                     nUnit   = dbgUnit(5),               &
            &                     stencil = .true.,                   &
            &                     string  = 'after build level elm'   )
      end do ! iLevel
    end if

  end subroutine build_levelElements
!******************************************************************************!


!******************************************************************************!
  !> Allocate and initialise most important arrays
  !!
  subroutine init_arrays( levelPointer, tree, eligible_child )
    !---------------------------------------------------------------------------
    integer, allocatable, intent(out) :: levelPointer(:)
    type(treelmesh_type), intent(in) :: tree
    integer, allocatable, intent(out) :: eligible_child(:)
    !---------------------------------------------------------------------------

    ! allocate desciptors with minimal and maximal
    ! refinement level (of all MPI processs)
    allocate( levelPointer( tree%nElems ))
    levelPointer = -1 !< reset contents for easy debugging

    ! \todo This is LBM specific
    allocate( eligible_child(8) )
    eligible_child = (/1,2,3,4,5,6,7,8 /)

  end subroutine init_arrays
!******************************************************************************!


!******************************************************************************!
  !> Deallocate and de-initialise fluid, ghost and halo arrays
  !! todo: Give this subroutine s.th. to do
  !!
  subroutine tem_killArrays( levelDesc )
    !---------------------------------------------------------------------------
    type(tem_levelDesc_type), intent(inout) :: levelDesc(:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    ! Allocate fluidList for each level
    do iLevel = 1, globalMaxLevels
    end do

  end subroutine tem_killArrays
!******************************************************************************!


!******************************************************************************!
  !> Check, on which partition a given element is located add required elements 
  !! to corresponding lists:
  !! if remote, add to halo
  !! if ghost, add to resp. ghost list
  !!
  recursive subroutine identify_elements( nTreeID, tree, pathFirst, pathLast,  &
    &                                     levelDesc, elemPos, proc,            &
    &                                     computeStencil, nesting )
    !---------------------------------------------------------------------------
    !> neighboring treeID
    integer(kind=long_k), intent(in)          :: nTreeID
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !> first treeID path in every process
    type(tem_path_type), intent(in) :: pathFirst(:)
    !> last treeID path in every process
    type(tem_path_type), intent(in) :: pathLast(:)
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !> nTreeID element position in the levelDesc % elem list
    integer, intent(out) :: elemPos
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nesting level
    integer, intent(in) :: nesting
    !---------------------------------------------------------------------------
    integer(kind=long_k) :: children(8) ! child elements
    integer :: nDepProcs ! processes that this neigh depend on
    integer :: depProc
    integer :: iChild, neighLevel
    type(tem_path_type) :: neighPath
    type(tem_stencilElement_type) :: emptyStencil(1)
    integer :: relFirst, relLast
    integer :: childPos   ! position of child element
    integer :: p_lb, p_ub ! process lower and upper bound
    integer :: nNesting ! new nesting level
    ! Position of the current element in the hash
    integer :: hashpos, elemNesting
    logical :: cacheHit
    logical :: updated ! was the element updated during identify_local_element
    !--------------------------------------------------------------------------=
    ! Increase the nesting
    nNesting = nesting + 1
    cacheHit = .false.

    write(dbgUnit(6),*) 'identifying required TreeID for ', &
      &                 nTreeID,                            &
      &                 ' nesting ', nesting 
    ! if neighbor treeID > 0  then it is fluid else its boundary
    if( nTreeID > 0 ) then
      neighLevel = tem_LevelOf( nTreeID )
      neighPath  = tem_PathOf( nTreeID )
      hashpos = int(mod(nTreeID, nHashes))
      if( hash(hashpos) > -1_long_k ) then
        ! something is in the hash
        if( hash(hashpos) == nTreeID ) then
          if( nesting == levelDesc( neighlevel )%elem%haloNesting%val(       &
            &                                   hash_elemPos( hashPos )) ) then
            cacheHit = .true.
          end if
          if( levelDesc( neighlevel )%elem%needsUpdate%val(                    &
            &                                   hash_elemPos( hashPos )) ) then
            cacheHit = .false.
           ! Set the needs update to false as it was now updated
           levelDesc( neighLevel )%elem%needsUpdate%val(                       &
             &                               hash_elemPos( hashPos )) = .false.
          end if
        end if
      end if
      !\todo: Reactivate the cache! The problem is that the nesting of an
      !element is updated to 0 without actually looping over its neighbors!
      !cacheHit = .false.
      cachemiss: if( .not. cacheHit ) then
        ! Probably did not hit this element yet, put it in the hash now,
        ! and identify it.
        ! (If this treeID is already in the hash, we definitely have
        !  already processed this treeID, and need nothing to do anymore).
        ! This leaves elements to be identified multiple times only in
        ! the case of hash collisions, which should be pretty few, for
        ! a sufficiently large hash.
        !-----------------------------------------------
        ! Cache miss

        nDepProcs = 0
        ! First check if this neighbor is in my local tree range
        p_lb = 1
        ! rank starts from indice 0 where as fortran array pathFirst and
        ! pathLast starts from 1
        relFirst = tem_PathComparison(neighPath, PathFirst(proc%rank+1))
        relLast = tem_PathComparison(neighPath, PathLast(proc%rank+1))
        if (relFirst < 0) then
          ! The searched element is definitely left of myself          
          ! Do only search up to my own range
          p_ub = proc%rank
        else
          p_ub = int(min(int(proc%comm_size, kind=long_k),                      &
            &            tree%global%nElems))
          if (relLast > 0) then
            ! The searched element is definitely right of myself
            ! Do only search beyond my own range
            p_lb = proc%rank+2
          else

            ! The element might be (partly) on my own partition
            if (relLast < 0) then
              ! The element is at least left of my right border
              ! Set the upper bound to include myself
              p_ub = proc%rank + 1
            end if
            if (relFirst > 0) then
              ! The element is at least right of my left border
              ! Set the lower bound to include myself
              p_lb = proc%rank + 1
            end if
          end if ! relLast > 0
        end if ! relFirst < 0

        if ((p_lb == p_ub) .and. (p_lb == proc%rank+1)) then
          ! The element is local, do not go on to other processes
          nDepProcs = 1
          depProc = proc%rank+1
        else
          ! Possibly NON-LOCAL Element...
          ! Every process COULD hold a part of the current neighbor element
          call tem_find_depProc(depProc, nDepProcs, &
            &                    neighPath, p_lb, p_ub, PathFirst, PathLast)
        end if ! local or non-local
        ! How many processes possess a part of the requested treeID
        if ( nDepProcs == 1 ) then
          updated = .false.
          ! might be a local or halo of same level or a ghostFromCoarser
          ! or a ghostFromFiner. If halo and ghost it is distributed
          ! in single process.
          ! elemPos here is position of nTreeID in levelDesc elem list
          call single_process_neighbor( nTreeID        = nTreeID,              &
            &                           levelDesc      = levelDesc,            &
            &                           tree           = tree,                 &
            &                           proc           = proc,                 &
            &                           iProc          = depProc,              &
            &                           minLevel       = tree%global%minlevel, &
            &                           foundLevel     = neighLevel,           &
            &                           elemPos        = elemPos,              &
            &                           updated        = updated,              &
            &                           computeStencil = computeStencil,       &
            &                           nesting        = nesting )
          elemNesting = min( nesting,                                          &
                  levelDesc( neighLevel )%elem%haloNesting%val( elemPos ) )
          if ( nesting < levelDesc( neighLevel )%elem%haloNesting              &
            &                                            %val( elemPos )) then
            ! element needs updating as the current nesting is smaller than the
            ! one in the element property
            levelDesc( neighLevel )%elem%needsUpdate%val( elemPos ) = .true.
            levelDesc( neighLevel )%elem%haloNesting%val( elemPos )            &
              &  = elemNesting
            updated = .true.
          end if

          !updated = .true.
          ! it is ghost from coarser element in current process proc%rank
          if( updated ) then
            if ( levelDesc( neighLevel )%elem%eType%val( elemPos )             &
              &                               == eT_ghostFromCoarserElem ) then
              if( nesting < nestingLimit ) then
                ! Create the direct neighbors of the ghostFromCoarser

                ! identify all the compute neighbors of the current element
                call identify_stencilNeigh( iElem          = elemPos,          &
                  &                         iLevel         = neighLevel,       &
                  &                         tree           = tree,             &
                  &                         iStencil       = 1,                &
                  &                         pathFirst      = pathFirst,        &
                  &                         pathLast       = pathLast,         &
                  &                         levelDesc      = levelDesc,        &
                  &                         proc           = proc,             &
                  &                         computeStencil = computeStencil,   &
                  &                         nesting        = elemNesting + 1 )

              end if ! nesting < 1?
        
              write(dbgUnit(6),*) 'creating all parent neighbors for ', &
                &              levelDesc(neighLevel)%elem%tID%val(elemPos)
              ! Create all elements required up to the actual existing fluid
              ! element these include the neighbors of the parents. In a level
              ! jump >1, these intermediate levels have to provide valid
              ! quantities over two of their computation updates to account for
              ! the recursive algorithm.
              call create_allParentNeighbors( targetElem     = elemPos,        &
                &                             nesting        = nesting,        &
                &                             level          = neighLevel,     &
                &                             tree           = tree,           &
                &                             computeStencil = computeStencil, &
                &                             levelDesc      = levelDesc,      &
                &                             pathFirst      = pathFirst,      &
                &                             pathLast       = pathLast,       &
                &                             proc           = proc )
        
            endif ! ghostFromCoarser?
          endif ! updated?

        elseif ( nDepProcs > 1 ) then
          ! more than one depending processes
          ! create ghost and find children (can only have finer children)
          call init( me        = emptyStencil(1),                              &
            &        QQN       = computeStencil%QQN,                           &
            &        headerPos = 1 )
          ! Add to the level-wise ghost list
          ! append and store position of element for return
          call append( me      = levelDesc( neighLevel )%elem,                 &
               &       tID     = nTreeID, property = 0_long_k,                 &
               &       eType   = eT_distributedGhostFromFinerElem,             &
               &       stencil = emptyStencil,                                 &
               &       pos     = elemPos )

          !... and store the position in the ghost list !
          ! Now find all children of this ghost
          do iChild = 1,8
            children = tem_directChildren( nTreeID )
            call identify_elements( nTreeID        = children( iChild ),       &
              &                     tree           = tree,                     &
              &                     pathFirst      = pathFirst,                &
              &                     pathLast       = pathLast,                 &
              &                     levelDesc      = levelDesc,                &
              &                     proc           = proc,                     &
              &                     elemPos        = childPos,                 &
              &                     computeStencil = computeStencil,           &
              &                     nesting        = nNesting )
          end do
        else
          ! cell not existing. stop.
          write(logUnit(6),*) '---------- WARNING!!! -------------------------'
          write(logUnit(6),*) 'cell', nTreeID ,' not existing on proc ',       &
            &        proc%rank, '. Number of dependencies', nDepProcs
          call tem_tIDinfo( me = nTreeID, tree = tree, nUnit = logUnit(6) )
          write(logUnit(6),*) 'WARNING!!! This should never occur and '        &
            &        // 'points to a bug or a buggy mesh.'
        end if
        if( elemPos .gt. 0 ) then
          ! Set the encountered element position to the hash
          hash(hashpos) = nTreeID
          hash_elemPos( hashpos ) = elemPos
        end if
        ! end cache miss 
        !-----------------------------------------------
      else cachemiss
        !-----------------------------------------------
        ! Cache hit
        elemPos = hash_elemPos( hashpos )
        levelDesc( neighLevel )%elem%haloNesting%val( elemPos )                &
          &  = min( levelDesc( neighLevel )%elem%haloNesting%val(elemPos),     &
          &         nesting )
        !-----------------------------------------------
      end if cachemiss
    else
      elemPos = int( nTreeID )
    end if ! nTreeID .gt. 0


  end subroutine identify_elements
!******************************************************************************!


!******************************************************************************!
  !> Find partitions holding data on a given path
  !!
  !! Using a binary search over the processes first and last elements.
  !!
  !! Is the element in question a local or remote element?
  !! To look up a certain element by its
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID" in the distributed
  !! list of elements, it is sufficient to know the splitting positions of all
  !! chunks.
  !! That is, the first and last
  !! \ref treelmesh_module::treelmesh_type::treeid "treeID" of each partition.
  !! With a binary search over the splitting positions any requested element can
  !! then be identified to be either outside the computational domain at all, or
  !! inside of one or several known partitions.
  subroutine tem_find_depProc( depProc, nDepProcs, neighPath, p_lb, p_ub,      &
    &                          PathFirst, PathLast)
    !---------------------------------------------------------------------------
    !> List of partitions
    integer, intent(out) :: depProc
    !> Number of partitions
    integer, intent(out) :: nDepProcs
    !> Element to look up
    type(tem_path_type), intent(in) :: neighPath
    !> Left partition bounds
    type(tem_path_type), intent(in) :: PathFirst(:)
    !> Right partition bounds
    type(tem_path_type), intent(in) :: PathLast(:)
    !> Left interval bound to search in
    integer, intent(in) :: p_lb
    !> Right interval bound to search in
    integer, intent(in) :: p_ub
    !---------------------------------------------------------------------------
    integer :: lb, ub
    integer :: foundProc, lastProc
    integer :: curProc
    integer :: pComp
    !---------------------------------------------------------------------------

    lb = p_lb
    ub = p_ub


    ! Find the partition, to which neighPath is definitely right of
    ! the first element.
    do
      curProc = (lb + ub) / 2
      if( curProc <= 0 ) then
        ! Element not found on any proc
        foundProc = 0
        exit
      end if
      pComp = tem_PathComparison(neighPath, PathFirst(curProc))
      if (pComp > 0) then
        ! The first element of curProc is smaller than neighPath, therefore
        ! we do not have to search left of curProc. Can we even go to curProc+1?
        if (tem_PathComparison(neighPath, PathLast(curProc)) > 0) then
          ! The search element is not in the partition curProc at all, continue
          ! search with new lower bound of curProc+1.
          lb = min(curProc + 1, p_ub)
        else
          ! The left element is smaller than the search element and the right
          ! one is larger or equal to it, thus we found the first partition,
          ! which contains information on the element.
          foundProc = curProc
          exit
        end if
      else
        ! The first element of curProc is larger or equal to neighPath, thus
        ! we can restrict our search interval up to this partition.
        ub = curProc
        ! It is OK to not decrease the interval further here, as we always
        ! round down to find the middle, thus, if by this step we set
        ! ub = lb+1, the next iteration will test a new element, as we get
        ! curProc = (lb + lb + 1) / 2 = lb.
      end if
      if (ub <= lb) then
        ! Interval of size 0, found the first process to hold information on
        ! the element in neighpath.
        ! The lower bound was tested strictly, thus it is safe to set the first
        ! process to the lower bound here and exit.
        foundProc = lb
        exit
      end if
    end do

    ! Now linearly search for the last process spanned by the searched element.
    ! Usually this is just the single partition identified above or more
    lastProc = foundProc
    if( lastProc > 0 ) then
      nDepProcs = 1
      depProc = lastProc
      do
        if (lastProc == p_ub) exit
        if (tem_PathComparison(neighPath, PathLast(lastProc)) < 0) exit
        if (tem_PathComparison(neighPath, PathFirst(lastProc+1)) < 0) exit
        lastProc = lastProc + 1
        nDepProcs = nDepProcs + 1
      end do
    else
      nDepProcs = 0
    end if

  end subroutine tem_find_depProc
!******************************************************************************!


!******************************************************************************!
  !> create all the neighbors of an element's parent
  !!
  !! Create all elements required up to the actual existing fluid
  !! element these include the neighbors of the parents. In a level
  !! jump >1, these intermediate levels have to provide valid
  !! quantities over two of their computation updates to account for
  !! the recursive algorithm.                                         
  !! 
  !! Here the fromCoarser interpolation should be handed in.
  !!
  recursive subroutine create_allParentNeighbors(                              &
    &                    targetElem, level, computeStencil, tree, levelDesc,   &
    &                    pathFirst, pathLast, proc, nesting )
    !---------------------------------------------------------------------------
    !> requested element position (child element) in LevelDesc elem list
    integer, intent(in) :: targetElem
    !> requested element level
    integer, intent(in) :: level
    !> first treeID path in every process
    type(tem_path_type), intent(in) :: pathFirst(:)
    !> last treeID path in every process
    type(tem_path_type), intent(in) :: pathLast(:)
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> process
    type(tem_comm_env_type), intent(in) :: proc
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nesting level
    integer, intent(in) :: nesting
    !--------------------------------------------------------------------------=
    integer(kind=long_k) :: cTreeID  ! current tree ID
    integer :: coarserLevel, cPos, parentNesting
    !---------------------------------------------------------------------------
    !nNesting = nesting + 1
    coarserLevel = level
    ! The parent again must be computed, so the nesting has to be 0
    parentNesting = -1
    cPos = targetElem
    ! exit if we have reached the minimal level
    if( coarserLevel == tree%global%minlevel ) return

    ! current element treeID
    cTreeID = levelDesc( coarserLevel )%elem%tID%val( cPos )

    ! Get the parent of the current treeID
    coarserLevel = coarserLevel - 1
    cTreeID = tem_parentOf( cTreeID )

    ! ... and identify the parent (new current tree ID)
    call identify_elements( nTreeID        = cTreeID,                          &
      &                     tree           = tree,                             &
      &                     pathFirst      = pathFirst,                        &
      &                     pathLast       = pathLast,                         &
      &                     levelDesc      = levelDesc,                        &
      &                     elemPos        = cPos,                             &
      &                     proc           = proc,                             &
      &                     computeStencil = computeStencil,                   &
      &                     nesting        = parentNesting )

    if ( cPos <= 0 ) then
      write(dbgUnit(3),*) ' Element not found ', cTreeID
      write(dbgUnit(3),*) '   cPos         ', cPos
    end if

    ! identify the stencil neighbors of the parent.
    ! Here we should identify the fromCoarser interpolation stencil neighbors 
    ! instead of the compute stencil neighbors
    call identify_stencilNeigh( iElem          = cPos,                         &
      &                         iLevel         = coarserLevel,                 &
      &                         tree           = tree,                         &
      &                         iStencil       = 1,                            &
      &                         pathFirst      = pathFirst,                    &
      &                         pathLast       = pathLast,                     &
      &                         levelDesc      = levelDesc,                    &
      &                         proc           = proc,                         &
      &                         computeStencil = computeStencil,               &
      &                         nesting        = parentNesting )

  end subroutine create_allParentNeighbors
!******************************************************************************!


!******************************************************************************!
  !>  Invoke the identify_elements for each neighbor of the stencil 
  !! and store the positions of the encountered elements
  !!
  recursive subroutine identify_stencilNeigh( iElem, iLevel, iStencil, tree,   &
    &                                         pathFirst, pathLast, levelDesc,  &
    &                                         proc, computeStencil, nesting )
    !---------------------------------------------------------------------------
    !> element position in levelDesc to identify
    integer, intent(in)                      :: iElem
    !> element level
    integer, intent(in)                      :: iLevel
    !> stencil within the element to act on 
    integer, intent(in)                      :: iStencil
    !> tree information
    type(treelmesh_type), intent(in)         :: tree
    !> first treeID path in every process
    type(tem_path_type), intent(in)          :: pathFirst(:)
    !> last treeID path in every process
    type(tem_path_type), intent(in)          :: pathLast(:)
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout)  :: levelDesc(tree%global%minLevel:)
    !> process
    type(tem_comm_env_type), intent(in)      :: proc
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nesting level
    integer, intent(in)                      :: nesting
    !---------------------------------------------------------------------------
    integer :: iStencilElem, elemPos
    integer :: neighIDpos
    integer(kind=long_k) :: nTreeID
    !---------------------------------------------------------------------------
    ! identify all the compute neighbors of the current element
    do iStencilElem = 1, computeStencil%QQN
      neighIDpos = levelDesc( iLevel )%elem%stencil%val( iElem )               &
        &                               %val( iStencil )%tIDpos( iStencilElem )
      if( neighIDpos > 0 ) then
        nTreeID = levelDesc( iLevel )%elem%neighID%val( iElem )                &
          &                                                  %val( neighIDpos )
        ! This call might add new halo elements
        if ( nTreeID > 0_long_k ) then
          call identify_elements( nTreeID        = nTreeID,                    &
            &                     tree           = tree,                       &
            &                     pathFirst      = pathFirst,                  &
            &                     pathLast       = pathLast,                   &
            &                     levelDesc      = levelDesc,                  &
            &                     elemPos        = elemPos,                    &
            &                     proc           = proc,                       &
            &                     computeStencil = computeStencil,             &
            &                     nesting        = nesting )
        else
          elemPos = 0
        end if
      else
        elemPos = 0
      end if
      ! And add the encountered neighbor elements to the current element's
      ! stencil neighbors
      levelDesc( iLevel )%elem%stencil%val( iElem )                            &
        &           %val( iStencil )%totalPos( iStencilElem ) = elemPos
    end do
 
  end subroutine identify_stencilNeigh
!******************************************************************************!


!******************************************************************************!
  !> Determine the location (which process) of a requested element,
  !! which was identified to be located on one single process
  !! (can be local or remote)
  !! If it is located on a remote process: add to halo list
  !!                     local process: identify if ghost or fluid
  !!
  subroutine single_process_neighbor( nTreeID, levelDesc, tree, proc, iProc,   &
    &                                 minLevel, elemPos, foundLevel,           &
    &                                 computeStencil, nesting, updated )
    !---------------------------------------------------------------------------
    !> neighboring treeID
    integer(kind=long_k), intent(in)         :: nTreeID
    !> minimum level fluid element in the tree
    integer, intent(in)                      :: minLevel
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout)  :: levelDesc(minLevel:)
    !> Process description to use.
    type(tem_comm_env_type), intent(in)      :: proc
    !> Process on which nTreeID is located
    integer, intent(in)                      :: iProc
    !> tree information
    type(treelmesh_type), intent(in)         :: tree
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nTreeID element position in the levelDesc % elem list
    integer, intent(out)                     :: elemPos
    !> level of neighbor treeID
    integer, intent(out)                     :: foundLevel
    !> nesting level
    integer, intent(in)                      :: nesting
    !> was the element updated in this call?
    logical, intent(out)                     :: updated
    !---------------------------------------------------------------------------
    type(tem_stencilElement_type) :: emptyStencil(1)
    integer :: neighLevel
    logical :: wasAdded
    !---------------------------------------------------------------------------
    !KM: \todo this neighLevel is same as foundLevel
    neighLevel = tem_LevelOf( nTreeID ) ! Has to be same as tLevel!?
    if(( neighLevel < minLevel )  &
      &                      .or. (neighLevel > uBound(levelDesc,1)) ) then
      write(logUnit(1),*) ' ERROR: level which is not included in the fluid'// &
        &             ' tree was demanded in singleProcNeigh' 
      write(logUnit(1),*)'          treeID', nTreeID 
      call tem_abort()
    end if
    ! Set the element updated flag as a default to false
    updated = .false.

    ! If it is a remote cell on only one process -> regular halo
    if( iProc /= proc%rank + 1 )  then
      ! REMOTE
      call init( me = emptyStencil(1), QQN=computeStencil%QQN )
      ! append this nTreeID as halo element to levelDesc elem list
      call append( me          = levelDesc( neighLevel )%elem,                 &
        &          tID         = nTreeID,                                      &
        &          eType       = eT_haloElem,                                  &
        &          property    = 0_long_k,                                     &
        &          sourceProc  = iProc,                                        &
        &          haloNesting = nesting,                                      &
        &          stencil     = emptyStencil,                                 &
        &          pos         = elemPos,                                      &
        &          wasAdded    = wasAdded )

      if( .not. wasAdded ) then
        ! update the nesting to the current one
        if( nesting <  levelDesc( neighLevel )%elem%haloNesting%               &
          &                                                val( elemPos )) then
          levelDesc( neighLevel )%elem%needsUpdate%val( elemPos ) = .true.
          levelDesc( neighLevel )%elem%haloNesting%val( elemPos )              &
            &  = min( levelDesc( neighLevel )%elem%haloNesting%val( elemPos ), &
            &         nesting )
          updated = .true.
        end if
      else
        updated = .true.
      end if
    else
      ! LOCAL
      ! Either a ghost or fluid cell
      call identify_local_element( nTreeID        = nTreeID,                   &
        &                          levelDesc      = levelDesc,                 &
        &                          tree           = tree,                      &
        &                          elemPos        = elemPos,                   &
        &                          level          = foundLevel,                &
        &                          minLevel       = minLevel,                  &
        &                          nesting        = nesting,                   &
        &                          updated        = updated,                   &
        &                          computeStencil = computeStencil )
    end if

  end subroutine single_process_neighbor
!******************************************************************************!


!******************************************************************************!
  !> Determine if the requested (local) nTreeID is fluid or ghost in the
  !! local process
  !! If fluid: do nothing, as it will be added later on anyway (or already is)
  !!    ghostFromFiner (coarser than requested):
  !!        add all virtual children, i.e. all levels between requested treeID
  !!        and found one.
  !!    ghostFromCoarser (finer than requested):
  !!    not existing( localPos=0): add to halo (MH: why to %halo and not
  !!    halosPrc(iProc)%mine ?)
  !!
  subroutine identify_local_element( nTreeID, levelDesc, tree, minLevel, level,&
    &                                elemPos, nesting, updated, computeStencil )
    !---------------------------------------------------------------------------
    !> neighboring treeID
    integer(kind=long_k), intent(in) :: nTreeID
    !> minimum level fluid element in the tree
    integer, intent(in) :: minLevel
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minLevel:)
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !> nesting level
    integer, intent(in) :: nesting
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> nTreeID element position in the levelDesc % elem list
    integer, intent(out) :: elemPos
    !> level of neighbor treeID
    integer, intent(out) :: level
    !> was the element updated in this call?
    logical, intent(out) :: updated
    !---------------------------------------------------------------------------
    integer :: localPos, neighLevel, dPos, foundLevel
    integer(kind=long_k) :: foundTID
    type(tem_path_type) :: neighPath
    !---------------------------------------------------------------------------
    ! Set the element updated flag as a default to false
    updated = .false.

    ! Position of neighbor treeID
    neighLevel = tem_LevelOf( nTreeID )
    neighPath  = tem_PathOf( nTreeID )

    ! Return position of nTreeID in the treeIDlist
    localPos = tem_PosOfPath( neighPath, tree%pathList )
    ! By localPos we can determine how the element (might) exist locally:
    ! - fluid
    ! - ghostFromCoarser
    ! - ghostFromFiner
    if( localPos > 0) then
      ! ghostFromCoarser or fluid
      foundTID = tree%treeID( localPos )

      foundLevel = tem_LevelOf( foundTID )
      ! we can also between nTreeID and foundTID
      if( foundLevel == neighLevel ) then
        ! Fluid is already part of the elem list. Adding not necessary.
        updated = .false.
      elseif( foundLevel < neighLevel ) then
        ! ghostFromCoarser
        ! existing fluid is coarser than current element
        ! Add all the children of found treeID down to me (intermediate levels)
        ! LBM specific. \todo hand over eligible child
        ! is added ghostFromCoarser
        call add_all_virtual_children(                                         &
          &                sourceID       = foundTID,                          &
          &                foundPos       = localPos,                          &
          &                elemPath       = neighPath,                         &
          &                sourceProperty = tree%ElemPropertyBits( localPos ), &
          &                sourceLevel    = foundLevel,                        &
          &                targetLevel    = neighLevel,                        &
          &                levelDesc      = levelDesc,                         &
          &                minlevel       = minLevel,                          &
          &                nesting        = nesting,                           &
          &                updated        = updated,                           &
          &                tree           = tree,                              &
          &                computeStencil = computeStencil )
      end if ! on same level?
    elseif( localPos < 0) then
      ! ghostFromFiner
      ! Find all existing fluid cells within requested nTreeID position
      ! Add all the parent between requested nTreeID and available child ID in
      ! treeID list
      call fill_to_real_children( reqTreeID      = nTreeID,                    &
        &                         levelDesc      = levelDesc,                  &
        &                         minLevel       = minlevel,                   &
        &                         tree           = tree,                       &
        &                         foundPos       = dPos,                       &
        &                         updated        = updated,                    &
        &                         computeStencil = computeStencil              )
    else
      write(dbgUnit(6),*) 'Warning: element not existing',nTreeID,             &
        &        'adding to nonexisting ...'
      call tem_tIDinfo( me = nTreeID, tree = tree, nUnit = dbgUnit(6) )
      ! This case occurs, when a remote halo was added, which was not existing
      ! before.
      ! Halos are added in the transfer_treeIDs routine
      call append( me         = levelDesc( neighLevel )%elem,                  &
        &          tID        = nTreeID,                                       &
        &          eType      = eT_nonExistingElem,                            &
        &          sourceProc = tree%global%myPart+1,                          &
        &          pos        = dPos )

    endif ! localPos gt 0? coarser ghost or fluid?

    ! position of added nTreeID in the levelDesc elem list
    elemPos = PositionOfVal( me  = levelDesc( neighLevel )%elem%tID,           &
      &                      val = nTreeID )
    level = neighLevel

  end subroutine identify_local_element
!******************************************************************************!


!******************************************************************************!
  !> Find all the virtual children of the sourceTreeID down to the targetLevel
  !! and add to the level-wise ghostFromCoarser list in the level descriptor
  !!
  recursive subroutine add_all_virtual_children(                               &
    &       sourceID, sourceProperty, foundPos, elemPath, sourceLevel,         &
    &       targetLevel, levelDesc, minLevel, tree, computeStencil,            &
    &       nesting, updated )
    !---------------------------------------------------------------------------
    !> source treeID (existing founded ID in tree%treeID list or children ID
    !! from recursion)
    integer(kind=long_k), intent(in) :: sourceID
    !> property of source element
    integer(kind=long_k), intent(in) :: sourceProperty
    !> element path
    type(tem_path_type), intent(in)  :: elemPath
    !> nesting level
    integer, intent(in) ::  nesting
    !> position of this sourceID in elem%tID list
    integer, intent(in) :: foundPos
    !> level of source element
    integer,intent(in) ::  sourceLevel
    !> level upto which virtual children must be created
    integer,intent(in) ::  targetLevel
    !> minimum level in the tree
    integer, intent(in) :: minLevel
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minLevel:)
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> was the element updated in this call?
    logical, intent(out) :: updated
    !---------------------------------------------------------------------------
    integer :: targetPos, iChild, iDir, nVals
    ! position of the existing (source) tID in the elem list
    integer :: sourcePos, levelPos 
    logical :: wasAdded
    logical :: childUpdated
    integer(kind=long_k) :: cTreeID      ! current treeID to identify
    integer(kind=long_k) :: ichildID     ! child treeID
    integer(kind=long_k) :: curNeighborID, coarserNeighborID
    integer(kind=long_k), allocatable :: tNeighID(:)
    type(tem_stencilElement_type) :: tStencil(1)
    integer :: iChildCoord(4), curLevel, offset(4), xc(4), childCoord(4)
    integer :: addedPos
    !---------------------------------------------------------------------------
    !Position of the coarser source element
    sourcePos = PositionOfVal( me  = levelDesc( sourceLevel )%elem%tID,        &
      &                        val = sourceID )
    allocate( tNeighID( computeStencil%QQN ))
    offset(4)   = 0
    ! By default, set that no element was updated
    updated = .false.
    childUpdated = .false.

    ! create virual children until target level is reached
    if( sourceLevel < targetLevel ) then
      !children = tem_directChildren( sourceID )
      curLevel = sourceLevel + 1
      call init( me = tStencil(1), QQN = computeStencil%QQN, headerPos = 1 )
      !do iChild = 1, 8
        ! Add to the level-wise ghost list
        levelPos = targetLevel - sourceLevel
        cTreeID = elemPath%node( levelPos )
        ! determine which child it is
        iChild = tem_childNumber( cTreeID )
        ! coordinate of children treeID
        childCoord = tem_coordOfId( cTreeID ) !children( iChild ))
        call append( me          = levelDesc( curLevel )%elem,                 &
          &          tID         = cTreeID,                                    &
          &          property    = sourceProperty,                             &
          &          eType       = eT_ghostFromCoarserElem,                    &
          &          sourceProc  = tree%global%myPart+1,                       &
          &          stencil     = tStencil,                                   &
          &          pos         = targetPos,                                  &
          &          haloNesting = nesting,                                    &
          &          wasAdded    = wasAdded )

        if( wasAdded ) then
          updated = .true.
          ! inherit boundaries and neighbors from parent
          iChildCoord(:) = tem_coordOfId( int(iChild, kind=long_k) )
          ! inherit the boundary infos from the parent
          tNeighID = 0_long_k
          ! Get neighIds of (source) coarse element
          do iDir = 1, 3
            call tem_find_BCs_fromCoarser( dir            = iDir,              &
              &                            childCoord     = iChildCoord,       &
              &                            sourceLevel    = sourceLevel,       &
              &                            sourcePos      = sourcePos,         &
              &                            neighID        = tNeighID,          &
              &                            minLevel       = minLevel,          &
              &                            levelDesc      = levelDesc,         &
              &                            computeStencil = computeStencil )
          enddo
          ! loop over all directions to determine neighIDs for ghost child (targetPos)
          do iDir = 1, computeStencil%QQN
            ! offset direction to compute neighbor coord
            offset(1:3) = computeStencil%cxDir(1:3, iDir )
            !neighIDpos = levelDesc( curLevel )%elem%val( targetPos )%stencil%val(1)%tIDpos( iDir )
            !coarserNeighborID = levelDesc( curLevel )%elem%val( targetPos )%neighID%val( neighIDpos )
            coarserNeighborID = tNeighID( iDir )
            ! compute virtual neighbor child from coarser neighbor 
            ! in iDir direction
            if( coarserNeighborID > 0_long_k ) then
              ! find the corresponding children of the neighbor defined for my
              ! parent. Find the child in the corresponding direction
              xc(4) = 1
              ! get the corresponding neighbor in the offset direction, which
              ! is a child of the coarser neighbor element. Use +2 offset to get
              ! only positive numbers
              ! (child coordinates always range between {0..1})
              xc(1:3) = mod( iChildCoord(1:3) + 2 + offset(1:3), 2 )
              ! get the child number of the corresponding neighbor within the
              ! coarser element. (is {1...8})
              ichildID  = tem_IdOfCoord( xc )
              ! calculate the child treeID from the coarser neighbor and the
              ! childID
              curNeighborID = coarserNeighborID*8_long_k + ichildID
            elseif( coarserNeighborID == 0_long_k ) then
              ! virtual neighbor child exist in current parent.
              ! If the neighbor is still 0, it must be a direct sibling,
              ! so we can directly compute its tID.
              ! find the neighbor according to stencil offset from me
              curNeighborID = tem_IdOfCoord( childCoord + offset )
            else
              ! inherit the boundary ID
              curNeighborID = coarserNeighborID
            endif
            ! append the neighbor ID ...
            call append( me  = levelDesc( curLevel )%elem%neighID              &
              &                                    %val( targetPos ),          &
              &          val = curNeighborID,                                  &
              &          pos = addedPos )
            ! ... and store this position in the stencil
            levelDesc( curLevel )%elem%stencil%val( targetPos )                &
              &                               %val(1)%tIDpos( iDir )           &
              &  = addedPos
          enddo !iDir QQN
          ! Now all boundary conditions for the target elements are defined.
          nVals = levelDesc( curLevel )%elem%neighID%val( targetPos )%nVals
          if( minval( levelDesc( curLevel )%elem%neighID%val( targetPos )      &
            &          %val( 1: nVals )) < 0 ) then
            ! Update the boundary condition property bit if any boundary
            ! condition is present
            levelDesc( curLevel )%elem%property%val( targetPos )               &
              &  = ibset(levelDesc( curLevel )%elem%property%val(targetPos),   &
              &          prp_hasBnd)
          else
            ! Unset the property bit
            levelDesc( curLevel )%elem%property%val( targetPos )               &
              &  = ibclr(levelDesc( curLevel )%elem%property%val(targetPos),   &
              &          prp_hasBnd)
          end if
        else
          ! Existing element encountered.
          updated = .false.
        endif
        ! Overwrite the eventually existing nesting with the smallest value.
        ! The smallest nesting determines if further neighbors have to be
        ! retrieved in communicate_elements
        if( nesting .lt. levelDesc( curLevel )%elem%haloNesting%               &
          &                                              val( targetPos )) then
          ! needs update
          updated = .true.
          levelDesc( curLevel )%elem%needsUpdate%val( targetPos ) = .true.
          levelDesc( curLevel )%elem%haloNesting%val( targetPos )              &
            &       = min( nesting, levelDesc( curLevel )%elem%haloNesting%    &
            &                                                 val( targetPos ))
        end if
        ! In any case we have to recurse down to the target level
        ! lower levels might not yet exist.
        !MH: What about childUpdated?
        call add_all_virtual_children( sourceID       = cTreeID,               &
          &                            foundPos       = foundPos,              &
          &                            elemPath       = elemPath,              &
          &                            nesting        = nesting,               &
          &                            sourceLevel    = curLevel,              &
          &                            targetLevel    = targetLevel,           &
          &                            levelDesc      = levelDesc,             &
          &                            minLevel       = minLevel,              &
          &                            updated        = childUpdated,          &
          &                            tree           = tree,                  &
          &                            sourceProperty = sourceProperty,        &
          &                            computeStencil = computeStencil )
      !end do
    end if
    updated = ( updated .or. childUpdated )

  end subroutine add_all_virtual_children
!******************************************************************************!


!******************************************************************************!
  !> Inherit the neighborhood from the sourceELem to the targetElem
  !!
  subroutine tem_find_BCs_fromCoarser( dir, childCoord, sourceLevel,           &
    &                                  sourcePos, neighID, computeStencil,     &
    &                                  levelDesc, minLevel )
    !---------------------------------------------------------------------------
    !> coarse element level
    integer, intent(in) :: sourceLevel
    !> position of coarser element in original treeID list
    integer, intent(in) :: sourcePos
    !>
    integer, intent(in) :: dir
    !> minimum level in the tree
    integer, intent(in) :: minLevel
    !> neighbor treeIDs of child element
    integer(kind=long_k), intent(inout) :: neighID(:)
    !> coordinate of virtual(ghost) child
    integer, intent(in) :: childCoord(4)
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(in) :: levelDesc(minLevel:)
    !---------------------------------------------------------------------------
    ! Tangential direction iterators
    integer :: iDirX, iDirY
    integer :: dirX ! first tangential direction
    integer :: dirY ! second tangential direction
    integer :: curDir ! child face normal direction, always {-1,1}
    integer :: myLink(4) ! child link direction
    integer :: parentLink(3) ! each entry must be either {-1,0,1}
    integer :: iChildStencil ! child stencil element iterator
    integer :: iStencilElem  ! stencil element iterator
    integer :: iStencil
    !---------------------------------------------------------------------------

    ! curDir is -1 if childCoord(dir) is 0 
    ! curDir is 1 if childCoord(dir) is 1
    ! it is needed to find in which direction of child's neighIDs are to be found
    curDir = childToStencil( childCoord( dir ))
    myLink( dir ) = curDir
    myLink(4) = 0
    parentLink(dir) = curDir

    ! identify the tangential directions
    dirX = mod( dir, 3) + 1
    dirY = mod( dir+1, 3) + 1
    iStencil = 1

    ! Iterate over all nine child links of the surface with the normal pointing
    ! into dir
    do iDirY = -1, 1
      myLink( dirY ) = iDirY
      ! select the valid directions to look for the neighbors in the second
      ! tangential direction
      if( childCoord( dirY ) == 0) then
        parentLink( dirY ) = min( iDirY, 0 )
      else
        parentLink( dirY ) = max( iDirY, 0 )
      end if
      do iDirX = -1, 1
        myLink( dirX ) = iDirX
        if( childCoord( dirX ) == 0) then
          parentLink( dirX ) = min( iDirX, 0 )
        else
          parentLink( dirX ) = max( iDirX, 0 )
        end if
        ! matching direction of child to compute stencil
        iChildStencil = 0
        do iStencilElem = 1, computeStencil%QQN
          if  (      computeStencil%cxDir( 1, iStencilElem ) == myLink( 1 )  &
            &  .and. computeStencil%cxDir( 2, iStencilElem ) == myLink( 2 )  &
            &  .and. computeStencil%cxDir( 3, iStencilElem ) == myLink( 3 )  &
            & ) then
            ! Found matching stencil entry for current child direction
            iChildStencil = iStencilElem
          endif
        enddo
        if( iChildStencil .gt. 0 ) then
          ! matching direction of parent to compute stencil
          do iStencilElem = 1, computeStencil%QQN
            if (      computeStencil%cxDir(1, iStencilElem) == parentLink(1) &
              & .and. computeStencil%cxDir(2, iStencilElem) == parentLink(2) &
              & .and. computeStencil%cxDir(3, iStencilElem) == parentLink(3) &
              &  ) then
              ! Found matching stencil entry for current child direction
              ! Set the parent's neighbor here. Later we replace the
              ! parent's neighbor by the current level's neighbor
              !targetElem%neighID%val( targetElem%stencil%val(iStencil)%tIDpos( iChildStencil )) =  &
              neighID( iChildStencil )                                         &
                &  = levelDesc( sourceLevel )                                  &
                &    %elem%neighID%val(sourcePos)                              &
                &                 %val( levelDesc( sourceLevel )               &
                &                       %elem%stencil%val(sourcePos)           &
                &                                    %val(iStencil)            &
                &                                    %tIDpos(iStencilElem)&    
                &                     )
!                sourceElem%neighID%val( sourceElem%stencil%val(iStencil)%tIDpos( iStencilElem ))
            end if 
          end do ! iStencilElem
        end if ! iChildStencil > 0
      end do ! iDirX
    end do ! iDirY 

  end subroutine tem_find_BCs_fromCoarser
!******************************************************************************!


!******************************************************************************!
  !> Find all the children of the requested reqTreeID
  !! add to the level-wise ghostFromFine list in the level descriptor
  !!
  recursive subroutine fill_to_real_children( reqTreeID, levelDesc, minLevel,  &
    &                                         tree, updated, foundPos,         &
    &                                         computeStencil )
    !---------------------------------------------------------------------------
    !> requested treeID
    integer(kind=long_k), intent(in) :: reqTreeID
    !> minimum level fluid element in the tree
    integer, intent(in) :: minLevel
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minLevel:)
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    ! position of reqTreeID in elem%tID list
    integer, intent(out) :: foundPos
    !> was the current element updated in this call?
    logical, intent(out) :: updated
    !> current stencil definition
    type( tem_stencilHeader_type ), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    integer :: iChild, addedPos, reqLevel
    integer(kind=long_k) :: children(8), property
    logical :: wasAdded, childUpdated
    integer :: childPos(8), neighLevel
    type(tem_path_type) :: neighPath
    !---------------------------------------------------------------------------
    ! Set as not updated by default
    updated = .false.

    ! Create the ghostFromFiner
    reqLevel        = tem_LevelOf( reqTreeID )
    call append( me         = levelDesc( reqLevel )%elem,                      &
      &          tID        = reqTreeID,                                       &
      &          eType      = eT_ghostFromFinerElem,                           &
      &          property   = 0_long_k,                                        &
      &          sourceProc = tree%global%myPart+1,                            &
      &          pos        = addedPos,                                        &
      &          wasAdded   = wasAdded )

    ! Return the position of the appended element to the caller as foundpos.
    foundPos = addedPos

    if( wasAdded ) then

      updated = .true.
      children = tem_directChildren( reqTreeID )
      childPos = 0 ! reset child positions. non-existing children are 0
      ! reset property
      property = 0_long_k
      ! if added reqTreeID is more than level coarser than available child treeID in
      ! original treeID list then add all children between reqLevel and neighLevel
      do iChild = 1, 8
        ! What's the configuration of the requested element
        neighLevel = reqLevel + 1
        neighPath  = tem_PathOf( children( iChild ))

        ! Return position in the treeIDlist
        childPos( iChild ) = tem_PosOfPath( neighPath, tree%pathList )

        ! searched neighPath is parent of available treeIDs 
        if( childPos( iChild ) < 0 ) then
           ! finer elements! add them.
           call fill_to_real_children( reqTreeID      = children( iChild ),    &
             &                         levelDesc      = levelDesc,             &
             &                         minLevel       = minlevel,              &
             &                         tree           = tree,                  &
             &                         foundPos       = childPos( iChild ),    &
             &                         updated        = childUpdated,          &
             &                         computeStencil = computeStencil         )
           ! Unify all properties of the children
           property = ieor( property,                                          &
           &   levelDesc( reqLevel + 1)%elem%property%val( childPos( iChild )))
           updated = ( updated .or. childUpdated )
        else
          ! Fluid cells are already in the element list (Not existing cells do
          ! not have to be added). If the element is a fluid, we have to get its
          ! position in the levelwise element list of the level descriptor.
          childPos( iChild ) = PositionOfVal( levelDesc( reqLevel + 1 )%elem%  &
            &                                            tID, children(iChild))
        end if
      end do

      ! Now reconstruct current element's neighborhood
      ! based on the children's information
      call tem_find_BCs_fromFiner( childPos       = childPos,                  &
        &                          sourceLevel    = reqLevel + 1,              &
        &                          targetLevel    = reqLevel,                  &
        &                          targetPos      = addedPos,                  &
        &                          levelDesc      = levelDesc,                 &
        &                          minLevel       = tree%global%minLevel,      &
        &                          computeStencil = computeStencil )

    else
      ! ghostFromFiner element was not added
      ! Hence, we found available information which now has to be returned
      updated = .false.
    end if

  end subroutine fill_to_real_children
!******************************************************************************!


!******************************************************************************!
  !> Inherit the neighborhood from the sourceELem to the targetElem
  !! Note that targetElem is inout, as it might have already values assigned.
  !!
  subroutine tem_find_BCs_fromFiner( childPos, sourceLevel, targetLevel,       &
    &                                targetPos, levelDesc, minLevel,           &
    &                                computeStencil )
    !---------------------------------------------------------------------------
    !> position of all childs in the levelDesc elem tID list
    integer, intent(in) :: childPos(8)
    !> level of child
    integer, intent(in) :: sourceLevel
    !> level of parent
    integer, intent(in) :: targetLevel
    !> added position of parent in the levelDesc elem tID list
    integer, intent(in) :: targetPos
    !> minimum level in the tree
    integer, intent(in) :: minLevel
    !> the level descriptor to be filled
    type(tem_levelDesc_type ) :: levelDesc(minLevel:)
    !> current stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    integer :: dir
    ! Tangential direction iterators
    integer :: iDirX, iDirY, iDir
    integer :: dirX !< first tangential direction
    integer :: dirY !< second tangential direction
    integer :: iStencilElem  !< stencil element iterator
    integer :: childCoord(4)
    ! the type of the current stencil link (side, edge, corner)
    integer :: linktype
    integer :: iDirNormal
    integer :: iStencil, addedPos
    type(tem_stencilElement_type) :: tStencil
    integer(kind=long_k) :: tNeighID
    !---------------------------------------------------------------------------
    if  ( .not. allocated( levelDesc( targetLevel )%elem%stencil%              &
      &                                              val(targetPos)%val )) then
      call init( me     = levelDesc( targetLevel )%elem%stencil%val(targetPos),&
        &        length = 1 )
    endif
    call init( me = tStencil, QQN = computeStencil%QQN, headerPos = 1 )
    call append( me  = levelDesc( targetLevel )%elem%stencil%val(targetPos),   &
      &          val = tStencil )
    if( levelDesc( targetLevel )%elem%neighID%val(targetPos)%nVals .gt. 0 ) then
      write(dbgUnit(2),*) 'Warning: tem_find_BCs_fromFiner has already' &
        &                       //' encountered'
      write(dbgUnit(2),*) '         existing neighbor IDs for the stencil'
    end if
    iStencil = levelDesc( targetLevel )%elem%stencil%val( targetPos )%nVals

    childCoord(4) = 1
    do iStencilElem = 1, computeStencil%QQN
      ! Reset current direction
      ! only 0 can not be used for inheritance of BCs later because boundaries 
      ! are stored as negative treeIDs. Transformation of 0 into a HUGE 
      ! negative value enables easy inheritance.
      tNeighID = -huge( 0_long_k )
      ! Count the non-zero entries to decide the kind of the current
      ! link direction in the stencil. (only direct neighbors)
      ! Distinguishing three different kinds of neighbors:
      ! SIDE, EDGE and CORNER
      linktype = sum( abs(computeStencil%cxDir( :, iStencilElem )))
      select case(linktype)
      case(1)
        ! 1 non-zero
        ! SIDE, needs four children
        ! identify the normal direction of the side
        dir = maxloc( abs( computeStencil%cxDir(:, iStencilElem )), 1)
        ! Convert the stencil direction to child direction
        iDirNormal = stencilToChild(computeStencil%cxDir(dir, iStencilElem ))
        childCoord( dir ) = iDirNormal
        ! identify the tangential directions
        dirX = mod( dir, 3) + 1
        dirY = mod( dir+1, 3) + 1
        do iDirY = 0, 1
          childCoord( dirY ) = iDirY
          do iDirX = 0, 1
            childCoord( dirX ) = iDirX
            call update_childNeighborID( neighID      = tNeighID,              &
              &                          childCoord   = childCoord,            &
              &                          childPos     = childPos,              &
              &                          iStencil     = iStencil,              &
              &                          iStencilElem = iStencilElem,          &
              &                          levelDesc    = levelDesc(sourceLevel))
          end do
        end do

      case(2)
        ! 2 non-zeroes
        ! EDGE, needs two children
        ! Determine the 0-direction (dimension with the zero entry)
        dir = minloc( abs( computeStencil%cxDir(:,iStencilElem )), 1)
        ! Get the edge directions
        dirX = mod( dir, 3) + 1
        dirY = mod( dir+1, 3) + 1
        childCoord( dirX ) = stencilToChild(computeStencil                     &
          &                                 %cxDir(dirX,iStencilElem))
        childCoord( dirY ) = stencilToChild(computeStencil                     &
          &                                 %cxDir(dirY,iStencilElem))
        do iDir = 0, 1
          childCoord( dir ) = iDir
          call update_childNeighborID( neighID      = tNeighID,                &
            &                          childCoord   = childCoord,              &
            &                          childPos     = childPos,                &
            &                          iStencil     = iStencil,                &
            &                          iStencilElem = iStencilElem,            &
            &                          levelDesc    = levelDesc( sourceLevel ))
        enddo

      case(3)
        ! No zero at all, all three directions have a offset
        ! CORNER, just a single child is connected to this link
        childCoord( 1:3 ) = stencilToChild( computeStencil                     &
          &                                  %cxDir(:,iStencilElem) )
        call update_childNeighborID( neighID      = tNeighID,                  &
          &                          childCoord   = childCoord,                &
          &                          childPos     = childPos,                  &
          &                          iStencil     = iStencil,                  &
          &                          iStencilElem = iStencilElem,              &
          &                          levelDesc    = levelDesc( sourceLevel ))
      end select
      ! Append the neighID of virtual parent
      call append( me  = levelDesc( targetLevel )%elem%neighID%val(targetPos), &
        &          val = tNeighID,                                             &
        &          pos = addedPos )
      levelDesc( targetLevel )%elem%stencil%val(targetPos)                     &
        &                                  %val(iStencil)%tIDpos(iStencilElem) &
        &  = addedPos

    end do

  end subroutine tem_find_BCs_fromFiner
!******************************************************************************!


!******************************************************************************!
  !> Return the element type of a treeID by defining the position in the
  !! totalList
  !!
  function tem_eTypeOfTotalPos( pos, levelDesc ) result( eType )
    !---------------------------------------------------------------------------
    !> the element you are looking for
    integer, intent(in) :: pos
    !> the descriptor you use for searching
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !>
    integer :: eType
    !---------------------------------------------------------------------------
    integer :: levelPos
    !---------------------------------------------------------------------------
    eType   = eT_undefinedElem
    levelPos = PositionOfVal( me  = levelDesc%elem%tID,                        &
      &                       val = levelDesc%total( pos ))
    if( levelPos .gt. 0 ) then
      eType = levelDesc%elem%eType%val( levelPos )
    end if

  end function tem_eTypeOfTotalPos
!******************************************************************************!


!******************************************************************************!
  !> Return the element type of a treeID .
  !!
  function tem_eTypeOfId( tID, levelDesc ) result( eType )
    !---------------------------------------------------------------------------
    !> the element you are looking for
    integer(kind=long_k), intent(in) :: tID
    !> the descriptor you use for searching
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !>
    integer :: eType
    !---------------------------------------------------------------------------
    integer :: levelPos
    !---------------------------------------------------------------------------
    eType   = eT_undefinedElem
    levelPos = PositionOfVal( me  = levelDesc%elem%tID,                        &
      &                       val = tID )
    if( levelPos .gt. 0 ) then
      eType = levelDesc%elem%eType%val( levelPos )
    end if

  end function tem_eTypeOfId
!******************************************************************************!


!******************************************************************************!
  !> Returns the absolute position in the total list of a given treeID
  !! opposed to PosOfId, where the relative position in one of the separate
  !! lists is returned. Herefore, total list has to be created beforehand.
  !!
  function tem_treeIDinTotal( tID, levelDesc, eType ) result( elemPos )
    !---------------------------------------------------------------------------
    !> the element you are looking for
    integer(kind=long_k), intent(in) :: tID
    !> the descriptor you use for searching
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !> element type
    integer, intent(in), optional :: eType
    !> return position of tID in levelDesc%total list
    integer :: elemPos
    !---------------------------------------------------------------------------
    integer :: eType_loc, levelPos
    !---------------------------------------------------------------------------
    if( present( eType )) then
      eType_loc = eType
    else
      levelPos = PositionOfVal( me  = levelDesc%elem%tID,                      &
        &                       val = tID )
      if( levelPos > 0 ) then
        eType_loc = levelDesc%elem%eType%val( levelPos )
      else
        eType_loc = 0
      end if
    end if
    elemPos = 0
    if( eType_loc > 0 ) then
      elemPos = tem_PositionInSorted( me    = levelDesc%total,                 &
        &                             lower = levelDesc%offset(1, eType_loc)+1,&
        &                             upper = levelDesc%offset(2, eType_loc),  &
        &                             val   = tID )
    end if
    elemPos = max( elemPos, 0 )

  end function tem_treeIDinTotal
!******************************************************************************!


!******************************************************************************!
  !> create the intermediate, static list totalPnt, which holds pointers to the 
  !! elem%TID list, but in an ordered fashion. The order is the same as it will
  !! be in the total list later on, i.e.: fluid, ghostFC, ghostFF, halo.
  !! this four sub-lists are within sorted by their treeID.
  !! Additionally, the process-wise collections of halo elements are collected
  !! into halosPrc by grouping the treeIDs according to their belonging process
  !!
  subroutine identify_lists( tree, levelDesc )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(in) :: tree
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !---------------------------------------------------------------------------
    integer :: iElem, indElem
    integer :: dPos ! Position where the object was stored in the dynamic list
    integer :: procPos ! Position where the process is stored in dynamic list
    type( tem_halo_type ) :: tProc ! temporary halo type
    logical :: wasAdded ! was the element added to the list
    integer :: iPnt( eT_minNumber:eT_maxNumber ), eType, iVal
    integer(kind=long_k) :: tID
    !---------------------------------------------------------------------------
    ! Destroy lists
    call destroy( me = levelDesc%halosPrc  )
    ! init lists
    call init( me = levelDesc%halosPrc, unique = .true. )

    ! 1. count nelems
    levelDesc%nElems_fluid = 0
    levelDesc%nElems_ghostFromCoarser = 0
    levelDesc%nElems_ghostFromFiner = 0
    levelDesc%nElems_halo  = 0
    do indElem = 1, levelDesc%elem%tID%nVals
      iElem = levelDesc%elem%tID%sorted( indElem )
      select case( levelDesc%elem%eType%val( iElem ) )
      case( eT_fluidElem )
        levelDesc%nElems_fluid = levelDesc%nElems_fluid + 1
      case( eT_ghostFromCoarserElem )
        levelDesc%nElems_ghostFromCoarser =                          &
          &                     levelDesc%nElems_ghostFromCoarser + 1
      case( eT_ghostFromFinerElem, eT_distributedGhostFromFinerElem )
        levelDesc%nElems_ghostFromFiner =                            &
          &                       levelDesc%nElems_ghostFromFiner + 1
      case( eT_haloElem )
        levelDesc%nElems_halo = levelDesc%nElems_halo + 1
      end select     
    end do
    levelDesc%nElems =                                               &
      &                        + levelDesc%nElems_fluid              &
      &                        + levelDesc%nElems_ghostFromCoarser   &
      &                        + levelDesc%nElems_ghostFromFiner     &
      &                        + levelDesc%nElems_halo

    if( allocated( levelDesc%totalPnt ))                             &
      &     deallocate( levelDesc%totalPnt )
    allocate( levelDesc%totalPnt( levelDesc%nElems ))

    call set_offsets( me       = levelDesc%offset(:,:),              &
      &               nFluids  = levelDesc%nElems_fluid,             &
      &               nGhostFC = levelDesc%nElems_ghostFromCoarser,  &
      &               nGhostFF = levelDesc%nElems_ghostFromFiner,    &
      &               nHalos   = levelDesc%nElems_halo )

    ! Reset pointers to current eType element in the total list
    iPnt = 0
    levelDesc%totalPnt = -1
    do indElem = 1, levelDesc%elem%tID%nVals
      ! Access sorted list to maintain locality of space-filling curve order
      iElem = levelDesc%elem%tID%sorted( indElem )
      tID = levelDesc%elem%tID%val( iElem )
      eType = levelDesc%elem%eType%val( iElem ) 
      if ( eType == eT_distributedGhostFromFinerElem ) then
        eType = eT_ghostFromFinerElem
      end if
      ! increase counter for current element
      iPnt( eType ) = iPnt( eType ) + 1
      ! is the eType a required element (fluid, ghost or halo)?
      ! Get rid of nonExistent elements here.
      if( eType >= eT_minNumRelevant) then
        ! Add sorted position of tID to the pointer 
        levelDesc%totalPnt(                                          &
          & levelDesc%offset(1,eType) + iPnt( eType ) ) = iElem
        ! And add the halosPrc for halo elements
        if ( eType == eT_haloElem ) then
          ! get the process from where to get the element
          ! Create an entry in the process list for the current source process
          tProc%partnerProc = levelDesc%elem%sourceProc%val( iElem )
          call append( me  = levelDesc%halosPrc,                     &
            &          val = tProc, pos = procPos, wasAdded = wasAdded )
          ! If the source process was not in the list before, I need to
          ! initialize the element list inside the current source process entry
          if( wasAdded ) then
            call init( me  = levelDesc%halosPrc%val( procPos )%mine, &
              &        unique = .true. )
          end if
          ! ... and add the current halo element to the respective process entry
          call append( me  = levelDesc%halosPrc%val( procPos )%mine, &
            &          val = iElem,                                            &
            &          pos = dPos )
        end if
      end if
    end do
    
    ! Security check
    ! Check if there are no entries < 1
    do iElem = 1, levelDesc%nElems
      if( levelDesc%totalPnt( iElem ) < 1 ) then
        write(dbgUnit(1),*)"Error: Found index < 1 in the totalPnt array."
        write(dbgUnit(1),*)'Abort!'
        write(dbgUnit(1),*)'offset: ' // trim(tem_toStr(levelDesc      &
          &                                            %offset(1,:),           &
          &                                            '; '))            
        do iVal = 1, levelDesc%nElems
          write(dbgUnit(1),*)'totalPnt: ',levelDesc%totalPnt(iVal)
        end do
        call tem_abort
      end if
    enddo

  end subroutine identify_lists
!******************************************************************************!


!******************************************************************************!
  !> Create the level-wise total lists. they consist of
  !! fluid elements + ghost elements + halo elements
  !! and are sorted by the treeIDs
  !!
  subroutine truncate_lists( levelDesc, minlevel )
    !---------------------------------------------------------------------------
    !> minlevel in tree
    integer, intent(in) :: minlevel
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    do iLevel = minLevel, ubound( levelDesc, 1 )
      call truncate( me = levelDesc( iLevel )%elem )
    end do 
  end subroutine truncate_lists
!******************************************************************************!


!******************************************************************************!
  !>  Set the offsets
  !!
  subroutine set_offsets( me, nFluids, nGhostFC, nGhostFF, nHalos )
    !---------------------------------------------------------------------------
    !> element type offsets
    integer, intent(out) :: me( 2, eT_minNumber:eT_maxNumber)
    !>
    integer, intent(in) :: nFluids, nGhostFC, nGhostFF, nHalos
    !---------------------------------------------------------------------------
    integer ::  nTotal
    !---------------------------------------------------------------------------
    nTotal = nFluids + nGhostFC+ nGhostFF+ nHalos
    ! Offset for accessing totallist, invsorted etc. arrays for
    !! fluids, ghosts and halos
    ! store the offset for the beginning ...
    me( 1, eT_fluidElem ) = 0
    me( 1, eT_ghostFromCoarserElem )  = nFluids
    me( 1, eT_ghostFromFinerElem )  = nFluids + nGhostFC
    me( 1, eT_haloElem )  = nFluids + nGhostFC + nGhostFF
    ! ... and the end of the list
    me( 2, eT_fluidElem ) =  me( 1, eT_ghostFromCoarserElem )
    me( 2, eT_ghostFromCoarserElem ) = me( 1, eT_ghostFromFinerElem )
    me( 2, eT_ghostFromFinerElem )   = me( 1, eT_haloElem )
    me( 2, eT_haloElem )   =           nTotal
    me( :, eT_distributedGhostFromFinerElem ) = me( :, eT_GhostFromFinerElem )

  end subroutine set_offsets
!******************************************************************************!


!******************************************************************************!
  !> Create the level-wise total lists.
  !!
  !! They consist of 
  !! fluid elements + ghost elements + halo elements 
  !! and are sorted by the treeIDs
  !! Also create the property lists
  !!

!KJ: Assume there are two levels 1 and 2
!     level 1 will have a ghost from finer
!     level 2 will have ghosts from coarser
!     when refining, we mark a coarse cell and add its children to the LD
!     when coarsening, we mark a cell and turn it and its neighbors to coarser.
!         - the destroyed ones are just added to the list of ghosts as they
!         - might be neighbors of someone else
  subroutine assemble_lists( levelDesc, minlevel )
    !---------------------------------------------------------------------------
    !> Minimal level in the mesh
    integer, intent(in) :: minlevel
    !> Level descriptor to fill
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel, iIndex
    !---------------------------------------------------------------------------
    write(logUnit(2),*) 'Assembling total list ...'

    do iLevel = minlevel, ubound(levelDesc,1)
      allocate( levelDesc( iLevel)%total( levelDesc( iLevel )%nElems ))
      allocate( levelDesc( iLevel)%property(levelDesc( iLevel )%nElems ))
      allocate( levelDesc( iLevel)%pntTID( levelDesc( iLevel )%nElems_fluid ))

      ! compute offset for element types sorting in levelDesc total list
      call set_offsets( me       = levelDesc( iLevel )%offset(:,:),            &
        &               nFluids  = levelDesc( iLevel )%nElems_fluid,           &
        &               nGhostFC = levelDesc( iLevel )%nElems_ghostFromCoarser,&
        &               nGhostFF = levelDesc( iLevel )%nElems_ghostFromFiner,  &
        &               nHalos   = levelDesc( iLevel )%nElems_halo )

      write(logUnit(5),*) '  Level           ', iLevel
      write(logUnit(5),*) '    nElems_total  ', levelDesc(iLevel)%nElems
      write(logUnit(5),*) '    nElems_fluid  ', levelDesc(iLevel)%nElems_fluid
      write(logUnit(5),*) '    nElems_gfC    ', levelDesc(iLevel) &
        &                                       %nElems_ghostFromCoarser
      write(logUnit(5),*) '    nElems_gfF    ', levelDesc(iLevel) &
        &                                       %nElems_ghostFromFiner
      write(logUnit(5),*) '    nElems_halo   ', levelDesc(iLevel) &
        &                                       %nElems_halo

      ! Counter across all lists
      do iIndex = 1, levelDesc( iLevel )%nElems
        ! sorted treeID list
        levelDesc( iLevel)%total( iIndex ) = levelDesc( iLevel )               &
          &    %elem%tID%val( levelDesc( iLevel )%totalPnt(iIndex) )
        ! property  
        levelDesc( iLevel)%property( iIndex ) = levelDesc( iLevel )            &
          &    %elem%property%val( levelDesc( iLevel )%totalPnt(iIndex))

        if( iIndex <= levelDesc( iLevel )%nElems_fluid ) then
          ! pointer to original tree%treeID list
          levelDesc( iLevel)%pntTID( iIndex )  = levelDesc( iLevel )           &
            &    %elem%pntTID%val( levelDesc( iLevel )%totalPnt( iIndex ))
        end if
      end do

    end do ! iLevel

  end subroutine assemble_lists
!******************************************************************************!


!******************************************************************************!
  !> identify additionally required neighbor elements
  !! run over the 'require' list of elements, which was accumulated before in
  !! init_elemLevels. The list includes neighbor elements of stencil neighbors,
  !! for stencils with the requireNeighNeigh attribute set. 
  !! This is needed for example for LBM boundary stencil elements, which in turn
  !! require their compute stencil neighborhood to allow PULL operations from
  !! there
  !!
  !! What exactly is the require list for?
  !! - Used ONLY for boundary stencil with higher order neighbors i.e
  !!   only when require nVals > 0
  !!
  subroutine identify_additionalNeigh( tree, proc, levelDesc, pathFirst,        &
    &                                  pathLast, computeStencil )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(in) :: tree
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !> first treeID path in every process
    type(tem_path_type), intent(in) :: pathFirst(:)
    !> last treeID path in every process
    type(tem_path_type), intent(in) :: pathLast(:)
    !> the compute stencil, for which the additional neighbors are reconstructed
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    integer :: iLevel, iElem, neighPos, elemPos, iStencil, iNeighElem, indElem
    integer :: nesting
    integer(kind=long_k) :: nTreeID
    !---------------------------------------------------------------------------
    ! The position of the compute stencil
    iStencil = 1
    do iLevel = tree%global%minlevel, tree%global%maxLevel
      ! Run over the additionally required element list
      do indElem = 1, levelDesc( iLevel )%require%nVals 
        ! get the position of the treeID in the element list
        iElem = PositionOfVal( me  = levelDesc( iLevel )%elem%tID,             &
          &                    val = levelDesc( iLevel )%require%val( indElem ))
        ! get the element position 
        if( iElem > 0 ) then
          if ( levelDesc( iLevel )%elem%eType%val( iElem ) .gt. 0) then
            ! Run over all the neighbors of the compute stencil
            do iNeighElem = 1, levelDesc( iLevel )%elem%stencil%val( iElem )   &
                &                                          %val( iStencil )%QQN
              ! position of neighbor treeID in dynamic array of neighID
              neighPos = levelDesc( iLevel )%elem%stencil%val( iElem )         &
                &                         %val( iStencil )%tIDpos( iNeighElem ) 
              if( neighPos > 0 ) then
                nTreeID = levelDesc( iLevel )%elem%neighID%val( iElem )%       &
                  &                                             val( neighPos )
                nesting = 0
                call identify_elements( nTreeID        = nTreeID,              &
                  &                     tree           = tree,                 &
                  &                     pathFirst      = pathFirst,            &
                  &                     pathLast       = pathLast,             &
                  &                     levelDesc      = levelDesc,            &
                  &                     elemPos        = elemPos,              &
                  &                     proc           = proc,                 &
                  &                     nesting        = nesting,              &
                  &                     computeStencil = computeStencil )
              else
                elemPos = 0 
              end if
              levelDesc( iLevel )%elem%stencil%val( iElem )%val( iStencil )    &
                &                %totalPos( iNeighElem ) = elemPos
            end do ! neighPos > 0
          else
            write(logUnit(2),*) 'not found required element in mesh ', nTreeID
          endif  ! valid elemType
        end if ! iElem>0
      end do ! indElem
    end do ! iLevel

  end subroutine identify_additionalNeigh
!******************************************************************************!


!******************************************************************************!
  !> exchange the requested treeIDs between all MPI processs
  !!
  subroutine communicate_elements( tree, proc, levelDesc, commPattern,          &
    &                  pathFirst, pathLast, computeStencil )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(in) :: tree
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc 
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !> the communication pattern used
    type(tem_commPattern_type), intent(in)    :: commPattern
    !> first and last treeID path in every process
    type(tem_path_type), intent(in) :: pathFirst(:), pathLast(:)
    !> stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil(:)
    !---------------------------------------------------------------------------
    integer :: iLevel, iErr, nProcs, iProc, iHalo, elemPos
    integer,allocatable :: nHalos(:)
    integer :: nIterations
    logical :: redo ! locally indicate if another iteration has to be performed
    logical :: redo_global    ! global indicator for another iteration
    !---------------------------------------------------------------------------
    call tem_horizontalSpacer( fUnit = logUnit(3) )
    write(logUnit(3),*) 'Communicating elements ...'
    !----------- exchange number of requesting / requested treeIDs -------------
    nIterations = 0
    do  !iIter = 1, 1 ! exchange halos until no new elements are created
      nIterations = nIterations + 1
      write(logUnit(4),*) 'Halo count exchange iteration: ', nIterations
      redo = .false.
      ! communicate nHalos and allocate levelDesc%buffer
      call communicate_nElemsToTransfer( tree      = tree,                     &
        &                                proc      = proc,                     &
        &                                levelDesc = levelDesc )
      !
      !---done.--- exchange number of requesting / requested treeIDs ----------!
      write(logUnit(5),*) '  Done communicating nElems to Transfer'

      !-----------                                                   ----------!
      !
      ! now we request the halo cells from the mpi processes,
      ! so they can send us their dependencies for these halo cells.
      ! We allow only dependencies with a difference of one level
      ! from lower to higher refinement levels and an arbitrary difference
      ! of refinement level from higher to lower refinement level.
      ! Since we get only the leaves as a dependency from the other mpi process
      ! we have to figure out later which cells we have to add locally
      !
      ! 1)  get the number of cells we will receive from the active processes,
      !     since we have
      !     to communicate only to the mpi processs which have/need informations
      ! 2)  exchange the treeIDs of the halos
      !
      ! First collapse the levelDesc%halos.
      ! We remove the processes where we don't have any elements to exchange
      ! with.
      ! Inverse Communciation (send to sourceProc, recv from targetProc)
      do iLevel = tree%global%minLevel, tree%global%maxLevel
        nProcs = levelDesc( iLevel )%halosPrc%nVals
        if( allocated( nHalos )) deallocate( nHalos )
        allocate( nHalos( nProcs ))
        ! Send treeIDs
        write(logUnit(5),*) '  Requesting remote halos on level ', iLevel
        call request_remoteHalos( levelDesc      = levelDesc,                  &
          &                       tree           = tree,                       &
          &                       iLevel         = iLevel,                     &
          &                       pathFirst      = pathFirst,                  &
          &                       pathLast       = pathLast,                   &
          &                       computeStencil = computeStencil(1),          &
          &                       proc           = proc )
        if (tem_logging_isActive(main_debug%logger, 5)) then
          call tem_elemList_dump( me = levelDesc( iLevel )%elem,      &
            &                     nUnit = dbgUnit(5),                 &
            &                     stencil = .true.,                   &
            &                     string = 'after request remotHalos' )
        end if
        write(logUnit(5),*) '  Done requesting remote halos.'

        nHalos(:nProcs) = levelDesc( iLevel )%halosPrc%val(:nProcs )%mine%nVals
        write(logUnit(5),*) '  Identifying lists'
        call identify_lists( tree, levelDesc(iLevel) )
        if( any( levelDesc( iLevel )%halosPrc%val(1:nProcs)%mine%nVals         &
          &      /= nHalos(:) )                                                &
          &      .or. nProcs /= levelDesc( iLevel )%halosPrc%nVals)            &
          redo = .true.
      end do !iLevel
      write(logUnit(6),*) '  Allreduce to check if changes occurred on any' &
        &                 //' process'
      ! JUROPA work-around for crash in the mpi_allreduce
      call mpi_barrier( proc%comm, iErr )
      ! Determine among all neighbor processes, if further iterations required
      call mpi_allreduce( redo, redo_global, 1, mpi_logical, mpi_lor,          &
        &                 proc%comm, iErr )
      if( .not. redo_global ) exit
    end do !exchange halos
    write(logUnit(3),*) 'Done exchanging number of elements to communicate.'
     
    !!   Now each process knows, which halos are requested.
    !!   Continue with identifying the actual leaf elements, which
    !!   are then communicated

    write(logUnit(3),*) 'Return halo counts and Redefine halos ...'
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      write(logUnit(5),*) '  Returning halo counts on level ', iLevel
      ! Receive the number of really existing halo elements
      call return_haloCounts( sendbuffer = levelDesc( iLevel )%sendbuffer,     &
        &                     recvbuffer = levelDesc( iLevel )%recvbuffer,     &
        &                     proc       = proc )
      call return_haloCounts( sendbuffer = levelDesc( iLevel )                 &
        &                                  %sendbufferFromCoarser,             &
        &                     recvbuffer = levelDesc( iLevel )                 &
        &                                  %recvbufferFromCoarser,             &
        &                     proc       = proc )
      call return_haloCounts( sendbuffer = levelDesc( iLevel )                 &
        &                                  %sendbufferFromFiner,               &
        &                     recvbuffer = levelDesc( iLevel )                 &
        &                                  %recvbufferFromFiner,               &
        &                     proc       = proc )

      ! reset the halos in the elem list
      do iProc = 1, levelDesc( iLevel )%halosPrc%nVals
        ! First declare all local halos as non-existent, and set only those
        ! actually provided by the remote process.
        do iHalo = 1, levelDesc( iLevel )%halosPrc%val( iProc )%mine%nVals
          elemPos = levelDesc( iLevel )%halosPrc%val( iProc )%mine%val( iHalo )
          levelDesc( iLevel )%elem%eType%val( elemPos ) = eT_nonExistingElem
        end do
      end do

      write(logUnit(5),*) '  Redefining halos ... '
      ! Receive the number of really existing halo elements 
      call redefine_halos( levelDesc      = levelDesc( iLevel ),               &
        &                  sendbuffer     = levelDesc( iLevel )%sendbuffer,    &
        &                  recvbuffer     = levelDesc( iLevel )%recvbuffer,    &
        &                  commPattern    = commPattern,                       &
        &                  computeStencil = computeStencil,                    &
        &                  proc           = proc )
      
      call redefine_halos( levelDesc      = levelDesc( iLevel ),               &
        &                  sendbuffer     = levelDesc( iLevel )                &
        &                                             %sendbufferFromCoarser,  &
        &                  recvbuffer     = levelDesc( iLevel )                &
        &                                             %recvbufferFromCoarser,  &
        &                  commPattern    = commPattern,                       &
        &                  computeStencil = computeStencil,                    &
        &                  proc           = proc )
      
      call redefine_halos( levelDesc      = levelDesc( iLevel ),               &
        &                  sendbuffer     = levelDesc( iLevel )                &
        &                                              %sendbufferFromFiner,   &
        &                  recvbuffer     = levelDesc( iLevel )                &
        &                                              %recvbufferFromFiner,   &
        &                  commPattern    = commPattern,                       &
        &                  computeStencil = computeStencil,                    &
        &                  proc           = proc )
      call identify_lists( tree, levelDesc(iLevel) )
    end do
     
    ! dump debug output 
    if (tem_logging_isActive(main_debug%logger, 5)) then
      do iLevel = tree%global%minLevel, tree%global%maxLevel
        call tem_elemList_dump( me      = levelDesc( iLevel )%elem,   &
          &                     nUnit   = dbgUnit(5),                 &
          &                     stencil = .true.,                     &
          &                     string  = 'after redefine remotHalos' )
      end do
    end if
    write(logUnit(3),*) 'Done with communication of elements. '
    write(logUnit(3),*) '-------------------------------------------'

  end subroutine communicate_elements
!******************************************************************************!


!******************************************************************************!
  !> Map requested halo to a position in my local fluid list or
  !! add recursively ghosts until I reach valid fluid elements
  !! return type of added element in levelPos(2)
  !! Also, non-existing elements are reported as such (levelPos(2))
  !!
  subroutine identify_halo( haloTreeID, elemPos, haloLevel, levelDesc, tree,   &
    &                       updated,   nesting, minLevel, computeStencil )
    !---------------------------------------------------------------------------
    !>
    type(treelmesh_type), intent(in) :: tree
    !>
    integer, intent(in) :: minLevel
    !>
    integer, intent(in) :: nesting
    !> neighboring treeID
    integer(kind=long_k), intent(in) :: haloTreeID
    !> type and position in list of found treeID
    integer, intent(out) :: elemPos
    !>
    integer, intent(out) :: haloLevel
    !>
    logical, intent(out) :: updated  
    !>
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !>
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !---------------------------------------------------------------------------
    haloLevel = tem_levelOf( haloTreeID )
    elemPos = PositionOfVal( me  = levelDesc( halolevel )%elem%tID,            &
      &                      val = haloTreeID )
    
    if( elemPos == 0 ) then
      ! requested halo does not exist and has to be created somehow
      call identify_local_element( nTreeID        = haloTreeID,                &
        &                          levelDesc      = levelDesc,                 &
        &                          tree           = tree,                      &
        &                          elemPos        = elemPos,                   &
        &                          level          = haloLevel,                 &
        &                          nesting        = nesting,                   &
        &                          updated        = updated,                   &
        &                          minLevel       = minLevel,                  &
        &                          computeStencil = computeStencil )
    end if

  end subroutine identify_halo
!******************************************************************************!


!******************************************************************************!
  !> deallocate all dispensable dynamic data structures and arrays
  !!
  subroutine tem_cleanupDependencyArrays( levelDesc )
    !---------------------------------------------------------------------------
    !> the level descriptor
    type(tem_levelDesc_type) , intent(inout) :: levelDesc(:) 
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    do iLevel = 1,size(levelDesc)
      ! Kill empty fluid, ghost and halo lists
      call destroy( me = levelDesc( iLevel )%halosPrc )
      call destroy( me = levelDesc( iLevel )%require )
    end do

  end subroutine tem_cleanupDependencyArrays   
!******************************************************************************!


!******************************************************************************!
  !> deallocate the stencil treeID neighbor arrays for each element
  !! This routine can only be called after build_horizontalDependencies
  !! 
  subroutine tem_cleanup_arrays( levelDesc )
    !---------------------------------------------------------------------------
    !> the level descriptor
    type(tem_levelDesc_type) , intent(inout) :: levelDesc(:) 
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------

    ! loop over all levels and destroy the corresponding elem types
    do iLevel = 1,size(levelDesc)
      call destroy( me = levelDesc( iLevel )%elem )
    end do

  end subroutine tem_cleanup_arrays   
!******************************************************************************!


!******************************************************************************!
  !> Build the vertical dependencies of ghost elements
  !! 
  subroutine tem_build_verticalDependencies( levelDesc, minlevel ) 
    !---------------------------------------------------------------------------
    !> the level descriptor
    type(tem_levelDesc_type), intent(inout) :: levelDesc(minlevel:)
    !>
    integer, intent(in) :: minlevel
    !---------------------------------------------------------------------------
    integer(kind=long_k) :: ghostID, childTreeID(8)
    integer :: iLevel, iElem, iChild
    integer :: elemPos, offset
    !---------------------------------------------------------------------------
    write(logUnit(3),*) 'Build vertical dep ...'

    ! Start with GhostFromCoarser
    do iLevel = minLevel+1,ubound(levelDesc,1)
      allocate( levelDesc( iLevel )%depFromCoarser(                            &
        &       levelDesc( iLevel )%nElems_ghostFromCoarser))
      offset = levelDesc( iLevel )%offset(1,eT_ghostFromCoarserElem)
      do iElem = 1, levelDesc( iLevel )%nElems_ghostFromCoarser
        ! position of ghost finer element in elem%tID list
        elemPos = levelDesc( iLevel )%totalPnt( offset+iElem )      
        ! finer ghost ID
        ghostID = levelDesc( iLevel )%elem%tID%val( elemPos )
        ! position of ghost parent ID in total list in level-1
        elemPos = tem_treeIDinTotal(                                           &
          &       tem_parentOf( ghostID ), levelDesc( iLevel-1 ))
        ! Add the requested TreeID to the halos and add dependencies
        call appendGhostDependency(                                            &
          &           sourcePos   = elemPos,                                   &
          &           sourceLevel = iLevel-1,                                  &
          &           tgtDep      = levelDesc( iLevel )%depFromCoarser( iElem ))
      end do
    end do

    ! Continue with GhostFromFiner
    do iLevel = minLevel,ubound(levelDesc,1)
      allocate( levelDesc( iLevel )%depFromFiner(                              &
        &       levelDesc( iLevel )%nElems_ghostFromFiner ))
      offset = levelDesc( iLevel )%offset(1,eT_ghostFromFinerElem)
      do iElem = 1, levelDesc( iLevel )%nElems_ghostFromFiner
        ! position of ghost coarser element in elem%tID list
        elemPos = levelDesc( iLevel )%totalPnt( offset + iElem )
        ! coarser ghost ID
        ghostID = levelDesc( iLevel )%elem%tID%val( elemPos )
        ! children of coarser ghost
        childTreeID(:) = tem_directChildren( ghostID )
        ! append position of each children ID to depFromFiner list
        do iChild = 1, 8
          elemPos = tem_treeIDinTotal( childTreeID( iChild ),                  &
            &                          levelDesc( iLevel+1 ))
          if( elemPos .gt. 0 ) then
            ! if the child treeID exists, add dependency
            call appendGhostDependency(                                        &
              &         sourcePos   = elemPos,                                 &
              &         sourceLevel = iLevel+1,                                &
              &         tgtDep      = levelDesc( iLevel )%depFromFiner( iElem ))
          end if
        end do
      end do
    end do

  end subroutine tem_build_verticalDependencies
!******************************************************************************!


!******************************************************************************!
  !> Building neighor array
  !!
  subroutine tem_build_horizontalDependencies( iStencil, levelDesc, tree,      &
    &                                          computeStencil, check )
    !---------------------------------------------------------------------------
    !> Tree representation of your mesh.
    type(treelmesh_type), intent(in) :: tree
    !> Level descriptor for each level of your mesh (starting from minimum
    !! level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> Index of your neighbor list??
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> check dependencies
    logical, intent(in), optional :: check
    !---------------------------------------------------------------------------
    logical :: do_check   ! check dependencies
    integer :: iLevel, iIndex
    integer :: nElemsWithNeigh
    !---------------------------------------------------------------------------

    if( present( check )) then
      do_check  = check
    else
      do_check  = .false.
    end if

    write(logUnit(3),"(A,I0)") 'Build neighbor array for stencil: ', iStencil

    do iLevel = tree%global%minLevel,tree%global%maxLevel

      if( computeStencil%useAll .and. iStencil == 1 ) then
        ! all fluid and ghost and halo elements
        nElemsWithNeigh = levelDesc(iLevel)%nElems
      else if( computeStencil%useAll ) then
        ! all fluid and ghost elements no halos
        nElemsWithNeigh =   levelDesc(iLevel)%nElems     &
          &               - levelDesc(iLevel)%nElems_halo
      else
        ! only the fluids and ghosts??? that have the stencil iStencil
        nElemsWithNeigh = computeStencil%elemLvl( iLevel )%nVals
      end if

      write(logUnit(4),"(A,I2,A,I0)") ' level: ', iLevel, ', nElems to treat: ', nElemsWithNeigh

      ! allocate the nghElems array with the number of elements this stencil is
      ! used by
      allocate( levelDesc( iLevel )%neigh( iStencil )%            &
        &          nghElems( computeStencil%QQN, nElemsWithNeigh ) )

      iIndex = 0
      if( computeStencil%QQN > 0 ) then
        ! we init everything by default with 0 to find a bug as fast as possible
        levelDesc( iLevel )%neigh( iStencil )%nghElems(:,:) = 0
        if (computeStencil%useAll) then
!          write(logUnit(5),*) '     fluid elements  ', iIndex
          call tem_build_listHorizontalDep(                                    &
            &                  levelDesc      = levelDesc( iLevel ),           &
            &                  computeStencil = computeStencil,                &
            &                  iStencil       = iStencil,                      &
            &                  elem           = levelDesc( iLevel )%elem,      &
            &                  list           = levelDesc( iLevel )%totalPnt,  &
            &                  nElems         = nElemsWithNeigh,               &
            &                  iIndex         = iIndex )

        else ! stencil not using all elements
!          write(logUnit(5),*) '     not all elems   ', iIndex
          call tem_build_treeHorizontalDep(                                    &
            &         levelDesc      = levelDesc( iLevel ),                    &
            &         computeStencil = computeStencil,                         &
            &         iStencil       = iStencil,                               &
            &         list           = computeStencil%elemLvl( iLevel )%val,   &
            &         nElems         = nElemsWithNeigh,                        &
            &         tree           = tree,                                   &
            &         iIndex         = iIndex )
        end if
      end if

    end do ! iLevel

    if ( do_check .or. tem_logging_isActive(10) ) then
      call tem_debug_HorizontalDependencies( iStencil       = iStencil,        &
        &                                    levelDesc      = levelDesc,       &
        &                                    tree           = tree,            &
        &                                    computeStencil = computeStencil,  &
        &                                    check          = do_check )
    end if

  end subroutine tem_build_horizontalDependencies
!******************************************************************************!


!******************************************************************************!
  !> Building neighor array
  !!
  subroutine tem_debug_HorizontalDependencies( iStencil, levelDesc, tree,      &
    &                                          computeStencil, check )
    !---------------------------------------------------------------------------
    !> Tree representation of your mesh.
    type(treelmesh_type), intent(in) :: tree
    !> Level descriptor for each level of your mesh (starting from min level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minLevel:)
    !> Index of your neighbor list.
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type) :: computeStencil 
    !>
    logical, intent(in) :: check
    !---------------------------------------------------------------------------
    integer :: iElem, iLevel, nElems, nEntries, iVal, elemPos, tIDpos
    integer :: neighPos, stencilSize, stencilPos
    integer(kind=long_k) :: tID, bufID, refID
    logical :: abort_loc, warning, skipElem
    !---------------------------------------------------------------------------

    write(dbgUnit(1),*)'Check neighbor positions for stencil ',iStencil
    abort_loc = .false.
    warning = .false.
    write(logUnit(1),*)' Checking, if the horizontal dependencies were set'// &
      &                ' correctly'

    do iLevel = tree%global%minLevel,tree%global%maxLevel
      nEntries = size( levelDesc( iLevel )%neigh(iStencil)%nghElems, 1)
      nElems = size( levelDesc( iLevel )%neigh( iStencil )%nghElems, 2)
      if( nEntries .gt. 0 ) then
        do iElem = 1, nElems
          abort_loc    = .false.
          warning  = .false.
          skipElem = .false.
          ! First get the treeID
          if( computeStencil%useAll ) then
            tIDPos =iElem
            tID = levelDesc( iLevel )%total( tIDpos )
          else
            if(  computeStencil%elemLvl( iLevel )%nVals .le. 0 ) exit
            tIDPos = computeStencil%elemLvl( iLevel )%val( iElem )
            tID = tree%treeID( tIDpos )
          end if
          ! Get the position of the treeID in the element list
          elemPos = PositionOfVal(levelDesc( iLevel )%elem%tID, tID )
          ! \todo: SZ: maybe remove this if condition or extend it to treat
          !            ghosts as well
          if( elemPos .le. levelDesc( iLevel )%nElems_fluid)then
            ! find the right stencil
            stencilPos = tem_stencil_getHeaderPos(                               &
              &           me  = levelDesc( iLevel )%elem%stencil%val( elemPos ), &
              &           val = iStencil )
          else
            ! set the stencil position to be the fluid stencil
            stencilPos = 1
          end if
!write(*,*)'in debug_HorDeps, level ', iLevel,' elemPos: ', elemPos
!write(*,*)'stencilPos: ', stencilPos, ' iStencil: ', iStencil
!write(*,*)'nVals: ', levelDesc( iLevel )%elem%stencil%val( elemPos )%nVals
!write(*,*)'val%pos', levelDesc( iLevel )%elem%stencil%val( elemPos )%val(:)%headerPos
          stencilSize = size( levelDesc( iLevel )%elem%stencil%val( elemPos )  &
            &                                       %val( stencilPos )%tIDpos )
!write(*,*)'Stencil entries: ', nEntries, ' stencilSize: ', stencilSize
          if (nEntries .ne. stencilSize) then
            write(dbgUnit(1),*)'stencil entries do not match for tID ',tID,    &
              &                ': ', nEntries, ' vs. ', stencilSize
 
          end if
          write( dbgUnit(6), '(i9, a)', advance='no' ) tID,' ngh  '
          do iVal = 1, nEntries
            neighPos = levelDesc( iLevel )%neigh(iStencil)%                    &
              &                                         nghElems( iVal, iElem )
            tIDpos = levelDesc( iLevel )%elem%stencil%val( elemPos )%val(      &
              &                                     stencilPos )%tIDpos( iVal )
            if( tIDpos .gt. 0 ) then
              refID = levelDesc( iLevel )%elem%neighID%val( elemPos )%val(     &
                &                                                      tIDpos )
            else
              refID = 0
            end if
            if( neighPos .gt. 0 ) then
              bufID = levelDesc( iLevel )%total( neighPos )
            else
              bufID = neighPos
            endif
            write(dbgUnit(6),'(i9)',advance='no') bufID
            if( refID .ne. bufID ) then
              write(dbgUnit(6),'(a,i0,a)', advance='no') ' (',refID,')'
              warning = .true.
              if( bufID .gt. 0_long_k ) then
                abort_loc  = .true.
                exit  
              end if
            end if
          end do
          write(dbgUnit(6),*)''
          if (abort_loc .and. check) then
            write(dbgUnit(1),*)' Non matching stencil entries!!' 
            write(dbgUnit(1),*)'    original treeID:         ', tID
            write(dbgUnit(1),*)'    required stencil treeID: ', refID
            write(dbgUnit(6),*) '   encountered treeID:      ', bufID
            write(logUnit(1),*)' Found non-matching treeID in the neighborhood'
            write(logUnit(1),*)'    original treeID:         ', tID
            write(logUnit(1),*)'    required stencil treeID: ', refID
            write(logUnit(1),*)'    encountered treeID:      ', bufID
            call tem_abort()
          end if
        end do
      end if
    end do

  end subroutine tem_debug_HorizontalDependencies
!******************************************************************************!


!******************************************************************************!
  !> Update the neighor arrays depending on what is given in the element stencil
  !! 
  !! The array levelDesc( iLevel )%neigh( iStenci )%nghElems( 1:QQN, 1:nElems )
  !! is being filled up here
  !!
  subroutine tem_build_treeHorizontalDep( iStencil, levelDesc, computeStencil, &
    &                                     list, nElems, iIndex, tree )
    !---------------------------------------------------------------------------
    !> Level descriptor for each level of your mesh (starting from min level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !> tree information
    type(treelmesh_type), intent(in) :: tree
    !> Index of your neighbor list.
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> element index in nghElems(QQN, iIndex)
    integer, intent(inout) :: iIndex
    !> number of elements
    integer, intent(in) :: nElems
    !> stencil elemLvl points to sorted original treeID list
    integer, intent(in) :: list(:)
    !---------------------------------------------------------------------------
    integer :: iElem, iStencilElem
    integer :: elemPos, stencilPos, levelPos, totalPos
    integer(kind=long_k) :: tID
    logical :: missingNeigh
    real(kind=rk) :: bary(3) ! bary center of element
    !---------------------------------------------------------------------------

    ! run over the given element list to update
    do iElem = 1, nElems
      missingNeigh = .false.
      tID = tree%treeID( list( iElem ))
      ! Get the position of the treeID in the element list
      elemPos = PositionOfVal( me = levelDesc%elem%tID, val = tID )
      if( elemPos > 0 ) then
        totalPos = tem_treeIDinTotal(                                          &
          &                    tID       = tID,                                &
          &                    levelDesc = levelDesc,                          &
          &                    eType     = levelDesc%elem%eType%val( elemPos ))
        stencilPos = tem_stencil_getHeaderPos(                                 &
          &                       me  = levelDesc%elem%stencil%val( elemPos ), &
          &                       val = iStencil )
        if( stencilPos > 0 ) then
          do iStencilElem = 1, levelDesc%elem%stencil%val(elemPos)             &
            &                                        %val(stencilPos)%QQN
            ! Totalpos has before already been updated, so just read from
            ! totalPos
            levelPos = levelDesc%elem%stencil%val(elemPos)                     &
              &                              %val(stencilPos)                  &
              &                              %totalPos(iStencilElem)
            if( levelPos <= 0 .and. computeStencil%requireNeighNeigh ) then
              missingNeigh = .true.
            end if
            levelDesc%neigh( iStencil )%nghElems( iStencilElem, iElem )        &
              &                                                     = levelPos
          end do
        end if
        if( missingNeigh ) then
          bary = tem_BaryOfId(tree, tID)
          write(logUnit(10),"(A,I0,A,I0,A,3ES12.5)") &
            &                   'missing neighbor iElem: ', iElem,&
            &                   ', tID: ', tID, ', bary: ', bary
          ! Set the property of the current element to have missing neighbors
          levelDesc%property( totalPos )                                       &
            &  = ibset( levelDesc%property( totalPos ), prp_missingNeigh )
        end if
      else
        write(dbgUnit(1),*) 'Warning!!!'
        write(dbgUnit(1),*) 'element not found in build_treehorizontalDep'
        write(dbgUnit(1),*) 'That should not occur.'
      end if
    end do

  end subroutine tem_build_treeHorizontalDep
!******************************************************************************!
 

!******************************************************************************!
  !> Update the neighor arrays depending on what is given in the element stencil
  !! 
  !! The array levelDesc( iLevel )%neigh( iStenci )%nghElems( 1:QQN, 1:nElems )
  !! is being filled up here
  subroutine tem_build_listHorizontalDep( iStencil, levelDesc, computeStencil, &
    &                                     elem, list, nElems, iIndex )
    !---------------------------------------------------------------------------
    !> Level descriptor for each level of your mesh (starting from min level).
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !> Index of your neighbor list.
    integer, intent(in) :: iStencil
    !> The stencil you build the horizontal dependencies for.
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> element list
    type(tem_element_type), intent(in) :: elem          
    !>
    integer, intent(inout) :: iIndex
    !> number of elements 
    integer, intent(in) :: nElems
    !> totalPnt in levelDesc points to sorted elem%tID list
    integer, intent(in) :: list(:)
    !---------------------------------------------------------------------------
    integer :: iElem, iStencilElem
    integer :: elemPos, levelPos, neighPos, neighVal
    integer(kind=long_k) :: nTreeID ! neighbor treeID value to treat
    logical :: invalid
    !---------------------------------------------------------------------------
    ! run over the given element list to update
    do iElem = 1, nElems
      elemPos = list( iElem )
!write(*,*)'in build_listHorDep nVals: ', levelDesc%elem%stencil%val( elemPos )%nVals
!write(*,*)'in build_listHorDep iStencil: ', iStencil
      ! Check if there are enough stencils given in the element
      if( levelDesc%elem%stencil%val( elemPos )%nVals >= iStencil ) then
        ! Increase the index element counter, if a stencil was encountered
        iIndex = iIndex + 1
        ! Run over each entry of the stencil
        do iStencilElem = 1, levelDesc%elem%stencil%val( elemPos )%            &
          &                                                val( iStencil )%QQN
          neighPos = levelDesc%elem%stencil%val( elemPos )%val( iStencil )     &
            &                                           %tIDpos( iStencilElem )
          invalid = .false.
          if( neighPos > 0 ) then
            nTreeID = levelDesc%elem%neighID%val( elemPos )%val( neighPos )
            if( nTreeID < 0_long_k ) then
              ! If the entry in the stencil is < 0, it must be a boundary ID
              ! -> simply set
              neighVal = int(nTreeID)
            else
              ! If the entry is > 0, test if there is also an entry for the
              ! element position directly (%elem array)
              if( allocated( levelDesc%elem%stencil%val( elemPos )             &
                &                            %val( iStencil )%totalPos) ) then
                ! If there is an entry, test, if it is valid
                levelPos = levelDesc%elem%stencil%val( elemPos )               &
                &                    %val( iStencil )%totalPos( iStencilElem )
                if( levelPos > 0 ) then
                  ! If there is a valid entry, get the position of this element
                  ! in the total list
                  neighVal = levelPos 
                else 
                  ! If the entry in %elem was not valid, use the tID to find
                  invalid = .true.
                end if ! levelPos > 0
              else 
                ! if there was no %elem entry given at all, directly use the
                ! tID to find the element position in the total list
                invalid = .true.
              end if ! allocated totalPos
            end if ! nTreeID > 0
!            invalid = .true.
            ! If above anything went wrong, ...
            if( invalid ) then
              ! ... try to find it with the slowest method: just look up the
              ! tID in the total list
              neighVal = tem_treeIDinTotal( nTreeID, levelDesc )
            end if
          else ! neighPos < 0
            neighVal = neighPos
          end if ! neighPos > 0
          levelDesc%neigh(iStencil)%nghElems( iStencilElem, iIndex ) = neighVal
        end do ! stencil element
      end if ! nStencils > iStencil
    end do ! iElem
  end subroutine tem_build_listHorizontalDep
!******************************************************************************!


!******************************************************************************!
  !> Communicate with all existing process the number of requested halo elements
  !!
  !! After this routine, each process knows how many processes there are to
  !! communicate with and how many elements have to be transferred
  !!
  subroutine communicate_nElemsToTransfer( tree, proc, levelDesc )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(in) :: tree
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !---------------------------------------------------------------------------
    ! the number of halos which will be sent to this mpi process for each level
    ! and for each process:
    ! first dimension is the level, second one is the mpi process
    integer, allocatable :: nHalosHis(:,:)
    integer, allocatable :: nHalosMine(:,:)
    integer :: iProc, iLevel, ierr
    integer :: iIndexMine, iIndexHis
    integer :: nLevels, sourceProc
    !---------------------------------------------------------------------------

    ! allocate the send buffer. 
    ! Contains the number of elements for each level, which are requested 
    ! from the specific process
    allocate( nHalosMine( tree%global%minLevel:tree%global%maxLevel,           &
      &                   proc%comm_size ))
    ! allocate the recv buffer. 
    ! Will contain the number of elements for each level, which other processes 
    ! request from me. This one is filled in the mpi_alltoall call
    allocate( nHalosHis( tree%global%minLevel:tree%global%maxLevel,            &
      &                  proc%comm_size ))

    ! Reset the number of requested and requesting halos
    nHalosMine(:,:) = 0
    nHalosHis(:,:)  = 0

    ! Delete possible pre-allocated arrays from previous exchanges
    do iLevel = tree%global%minLevel,tree%global%maxLevel
      if( allocated( levelDesc( iLevel )%recvbuffer%proc ))                    &
        &                     deallocate( levelDesc( iLevel )%recvbuffer%proc )
      if( allocated( levelDesc( iLevel )%recvbuffer%nElemsProc ))              &
        &               deallocate( levelDesc( iLevel )%recvbuffer%nElemsProc ) 
      if( allocated( levelDesc( iLevel )%sendbuffer%proc ))                    &
        &                     deallocate( levelDesc( iLevel )%sendbuffer%proc )
      if( allocated( levelDesc( iLevel )%sendbuffer%nElemsProc ))              &
        &                deallocate(levelDesc( iLevel )%sendbuffer%nElemsProc ) 
    end do

    !---------------------------------------------------------------------------
    ! Fill the send buffer, i.e. number of elements, which I request from
    ! all other processes.
    do iLevel = tree%global%minLevel, tree%global%maxLevel
      do iProc = 1, levelDesc( iLevel )%halosPrc%nVals
        sourceProc = levelDesc( iLevel )%halosPrc%val( iProc )%partnerProc
        nHalosMine(iLevel, sourceProc) =                                       &
          &      levelDesc( iLevel )%halosPrc%val( iProc )%mine%nVals
      end do ! iProc
    end do   ! iLevel

    ! number of halos which are sent to a process 
    ! might be different from number of halos
    ! received from the same process for that we use mpi_alltoall

    ! JUROPA MPI WORKAROUND !
    ! This barrier is just here to workaround a deadlock problem in
    ! Parastation MPI on Juropa, which results in the consequent all
    ! to all to never complete on some processes, if PSP_ONDEMAND=1
    ! is set.
    ! \todo HK: remove this workaround for the buggy MPI implementation
    ! on Juropa again!
    call mpi_barrier(proc%comm, iErr) ! JUROPA
    write(logUnit(3),*) ''
    ! Exchange the number of requested treeIDs
    nLevels = tree%global%maxLevel - tree%global%minLevel + 1
    call mpi_alltoall( nHalosMine, nLevels, mpi_integer,                       &
      &                nHalosHis, nLevels, mpi_integer,                        &
      &                proc%comm, ierr )
    write(logUnit(3),*) '    Finished all-to-all communication for the'   &
      &                 //' number of elements to exchange.'

    ! JUROPA MPI WORKAROUND
    ! Security barrier for the mentioned Juropa shortcoming before the
    ! all to all.
    call mpi_barrier(proc%comm, iErr) ! JUROPA
    write(logUnit(5),*) '    Done.'

    do iLevel = tree%global%minLevel,tree%global%maxLevel
      ! now check which processs really ask for information and allocate buffer
      levelDesc( iLevel )%recvbuffer%nProcs                                    &
        &  = count( nHalosMine( iLevel, : ) > 0 )
      levelDesc( iLevel )%sendbuffer%nProcs                                    &
        &  = count( nHalosHis( iLevel, : ) > 0 )
      allocate( levelDesc( iLevel )%recvbuffer%proc(                           &
        &           levelDesc( iLevel )%recvbuffer%nProcs ))
      allocate( levelDesc( iLevel )%recvbuffer%nElemsProc(                     &
        &           levelDesc( iLevel )%recvbuffer%nProcs))
      allocate( levelDesc( iLevel )%sendbuffer%proc(                           &
        &           levelDesc( iLevel )%sendbuffer%nProcs ))
      allocate( levelDesc( iLevel )%sendbuffer%nElemsProc(                     &
        &           levelDesc( iLevel )%sendbuffer%nProcs ))
      iIndexMine = 0
      iIndexHis  = 0
      do iProc = 1, proc%comm_size
        if( nHalosMine( iLevel, iProc ) > 0 ) then 
          ! Store the processes numbers to receive from
          iIndexMine = iIndexMine + 1
          levelDesc( iLevel )%recvbuffer%nElemsProc( iIndexMine )              &
            & = nHalosMine( iLevel, iProc )
          levelDesc( iLevel )%recvbuffer%proc( iIndexMine ) = iProc - 1
        end if
        if( nHalosHis( iLevel, iProc ) > 0 ) then 
          ! Store the processes numbers to send to
          iIndexHis = iIndexHis + 1
          levelDesc( iLevel )%sendbuffer%nElemsProc( iIndexHis )               &
            & = nHalosHis( iLevel, iProc )
          levelDesc( iLevel )%sendbuffer%proc( iIndexHis ) = iProc - 1
        end if
      end do
      ! Copy all headers (information about target and source procs)
      ! to bufferFromCoarser and bufferFromFiner
      levelDesc( iLevel )%sendbufferFromCoarser = levelDesc( iLevel )%sendbuffer
      levelDesc( iLevel )%sendbufferFromFiner   = levelDesc( iLevel )%sendbuffer
      levelDesc( iLevel )%recvbufferFromCoarser = levelDesc( iLevel )%recvbuffer
      levelDesc( iLevel )%recvbufferFromFiner   = levelDesc( iLevel )%recvbuffer
    end do ! iLevel
    write(logUnit(3),*) '  Done. Finished comm nElems to transfer.'
 
  end subroutine communicate_nElemsToTransfer
!******************************************************************************!


!******************************************************************************!
  !> Inverse Communication: Communicate, which elements each process needs from
  !! me.
  !!
  !! In this routine, we send the treeIDs of the halo elements to the
  !! processes, where they are located. Later on, we fill these halos locally
  !! with information from these processes (sourceProcs). In this routine
  !! however, we now SEND information to these sourceProcs, so do not get
  !! confused here.
  !!
  subroutine request_remoteHalos( levelDesc, proc, tree, iLevel, computeStencil,&
    &                             pathFirst, pathLast )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(in) :: tree
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel: )
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !> current level
    integer, intent(in) :: iLevel
    !> stencil definition
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> first and last treeID path in every process
    type(tem_path_type), intent(in) :: pathFirst(:), pathLast(:)
    !---------------------------------------------------------------------------
    integer :: iProc, iErr, iElem, elemPos, procPos
    integer :: haloLevel
    integer, allocatable :: rq_handle(:)
    integer, allocatable :: status(:,: )
    integer :: nCommunications
    integer :: nesting
    type( grw_longArray_type ), allocatable :: sendBuf_long(:)
    type( grw_longArray_type ), allocatable :: recvBuf_long(:)
    type( grw_intArray_type ), allocatable :: sendBuf_int(:)
    type( grw_intArray_type ), allocatable :: recvBuf_int(:)
    type( tem_halo_type ) :: tProc
    logical :: wasAdded
    logical :: updated
    integer, parameter :: message_flag_long = 24
    integer, parameter :: message_flag_int  = 25
    !---------------------------------------------------------------------------
    ! two communications: treeID and nesting
    nCommunications = 2*(levelDesc( iLevel )%sendbuffer%nProcs                 &
      &             + levelDesc( iLevel )%recvbuffer%nProcs)
    allocate( rq_handle( nCommunications ) )
    allocate( status( mpi_status_size, nCommunications ) )
    allocate( recvBuf_long( levelDesc( iLevel )%sendbuffer%nProcs ))
    allocate( recvBuf_int( levelDesc( iLevel )%sendbuffer%nProcs ))
    !---------------------------------------------------------------------
    ! Warning: Inverse Communication ! (send to source, recv from target)
    !
    ! The following is CONFUSING, because I have to RECEIVE the SEND buffers.
    ! Allocate receive buffers, which are later on used as information, 
    ! which elements to SEND to which neighbors, that's why mustSend
    rq_handle(:) = MPI_REQUEST_NULL
    ! We have to use nHalosHis and the targetProc here!!
    do iProc = 1, levelDesc( iLevel )%sendbuffer%nProcs
      ! Try to add the target process to the process list
      ! It might not be exist there yet
      tProc%partnerProc = levelDesc( iLevel )%sendbuffer%proc( iProc )+1
      call append( me       = levelDesc( iLevel )%halosPrc,                    &
        &          val      = tProc,                                           &
        &          pos      = procPos,                                         &
        &          wasAdded = wasAdded )
      ! Allocate the buffers
      call init( me     = recvBuf_long( iProc ),                               &
        &        length = levelDesc( iLevel )%sendbuffer%nElemsProc( iProc ))
      call init( me     = recvBuf_int( iProc ),                                &
        &        length = levelDesc( iLevel )%sendbuffer%nElemsProc( iProc ))
      
      ! Receive the element tree IDs
      call mpi_irecv( recvBuf_long( iProc )%val,                               &
        &             recvBuf_long( iProc )%ContainerSize,                     &
        &             mpi_integer8,                                            &
        &             levelDesc( iLevel )%sendbuffer%proc(iProc),              &
        &             message_flag_long,                                       &
        &             proc%comm,                                                &
        &             rq_handle( iProc),                                       &
        &             iErr )
      ! Receive the element nestings
      call mpi_irecv( recvBuf_int( iProc )%val,                                &
        &             recvBuf_int( iProc )%ContainerSize,                      &
        &             mpi_integer,                                             &
        &             levelDesc( iLevel )%sendbuffer%proc(iProc),              &
        &             message_flag_int,                                        &
        &             proc%comm,                                                &
        &             rq_handle( iProc+nCommunications/2 ),                    &
        &             iErr  )                                              
      ! Update the number of elements inside the growing array of the recv 
      ! buffer
      recvBuf_long( iProc )%nVals = recvBuf_long( iProc )%ContainerSize
      recvBuf_int( iProc )%nVals  = recvBuf_long( iProc )%ContainerSize
    end do ! iProc
   
    ! I send the NEEDED elements to the neighbor processes.
    allocate( sendBuf_long( levelDesc( iLevel )%recvbuffer%nProcs ))
    allocate( sendBuf_int( levelDesc( iLevel )%recvbuffer%nProcs ))
    do iProc = 1, levelDesc( iLevel )%recvbuffer%nProcs
      tProc%partnerProc = levelDesc( iLevel )%recvbuffer%proc( iProc ) + 1
      ! Get the position of the process in the dynamic halos list (might be
      ! unordered)
      procPos = PositionOfVal( me = levelDesc( iLevel )%halosPrc, val = tProc )
      call init(  me     = sendBuf_long( iProc ),                              &
        &         length = levelDesc( iLevel )%recvbuffer%nElemsProc( iProc ))
      call init(  me     = sendBuf_int( iProc ),                               &
        &         length = levelDesc( iLevel )%recvbuffer%nElemsProc( iProc ))
      ! Collect the halo treeIDs into send buffers for all the processes to
      ! request from
      do iElem = 1, levelDesc( iLevel )%recvbuffer%nElemsProc( iProc )
        elemPos = levelDesc( iLevel )%halosPrc%val( procPos )%mine%val( iElem )
        call append( me  = sendBuf_long( iProc ),                              &
          &          val = levelDesc( iLevel )%elem%tID%val( elemPos ) )
        call append( me  = sendBuf_int( iProc ),                               &
          &          val = levelDesc( iLevel )%elem%haloNesting%val( elemPos ) )
      enddo
      ! Send treeIDs
      call mpi_isend(                                                          &
       &        sendBuf_long( iProc )%val,                                     &
       &        sendBuf_long( iProc )%nVals,                                   &
       &        mpi_integer8,                                                  &
       &        levelDesc( iLevel )%recvbuffer%proc( iProc ),                  &
       &        message_flag_long,                                             &
       &        proc%comm,                                                      &
       &        rq_handle( iProc + levelDesc( iLevel )%sendbuffer%nProcs),     &
       &        ierr  )
      ! Send nesting
      call mpi_isend(                                                          &
       &        sendBuf_int( iProc )%val,                                      &
       &        sendBuf_int( iProc )%nVals,                                    &
       &        mpi_integer,                                                   &
       &        levelDesc( iLevel )%recvbuffer%proc( iProc ),                  &
       &        message_flag_int,                                              &
       &        proc%comm,                                                      &
       &        rq_handle( iProc + levelDesc( iLevel )%sendbuffer%nProcs       &
       &                         + nCommunications/2),                         &
       &        ierr  )
    end do ! iProc

    call mpi_waitall( nCommunications, rq_handle, status, ierr)
    write(logUnit(5),*) '  Received requested halo elements successfully.'
    ! Requested halo elements were received. 
    !---------------------------------------------------------------------------

    !---------------------------------------------------------------------------
    ! Now identify the requested halos.
    ! Arrays might still be allocated from previous communication iteration
    if( allocated( levelDesc( iLevel )%sendbuffer%elemPos))                    &
      &                    deallocate( levelDesc( iLevel )%sendbuffer%elemPos )
    allocate( levelDesc( iLevel )%sendbuffer%elemPos(                          &
      &                                levelDesc( iLevel )%sendbuffer%nProcs ))
    if( allocated( levelDesc( iLevel )%sendbufferFromCoarser%elemPos))         &
      &         deallocate( levelDesc( iLevel )%sendbufferFromCoarser%elemPos )
    allocate( levelDesc( iLevel )%sendbufferFromCoarser%elemPos(               &
      &                                levelDesc( iLevel )%sendbuffer%nProcs ))
    if( allocated( levelDesc( iLevel )%sendbufferFromFiner%elemPos ))          &
      &           deallocate( levelDesc( iLevel )%sendbufferFromFiner%elemPos )
    allocate( levelDesc( iLevel )%sendbufferFromFiner%elemPos(                 &
      &                                levelDesc( iLevel )%sendbuffer%nProcs ))

    ! Add elements of received buffers to elem
    do iProc = 1, levelDesc( iLevel )%sendbuffer%nProcs
      ! Allocate the buffer for the element position indices
      call init( me = levelDesc( iLevel )%sendbuffer%elemPos(iProc) )
      call init( me = levelDesc( iLevel )%sendbufferFromCoarser%elemPos(iProc) )
      call init( me = levelDesc( iLevel )%sendbufferFromFiner%elemPos(iProc) )
      do iElem = 1, recvBuf_long( iProc )%nVals
        nesting = recvBuf_int( iProc )%val( iElem )
        ! identify requested halo treeID in local process
        call identify_halo( haloTreeID     = recvBuf_long( iProc )%val(iElem), &
          &                 elemPos        = elemPos,                          &
          &                 halolevel      = haloLevel,                        &
          &                 levelDesc      = levelDesc,                        &
          &                 nesting        = nesting,                          &
          &                 updated        = updated,                          &
          &                 tree           = tree,                             &
          &                 minLevel       = tree%global%minLevel,             &
          &                 computeStencil = computeStencil )                          
        if( elemPos > 0 ) then
          ! if requested halo is ghostFromCoarser then find stencil neighbors of 
          ! this halo element
          if( nesting < nestingLimit .and.                                  &
            &  levelDesc( iLevel )%elem%eType%val( elemPos )                   &
            &                                 == eT_ghostFromCoarserElem ) then
            ! identify all the compute neighbors of the current element
            call identify_stencilNeigh( iElem          = elemPos,              &
              &                         iLevel         = iLevel,               &
              &                         tree           = tree,                 &
              &                         iStencil       = 1,                    &
              &                         pathFirst      = pathFirst,            &
              &                         pathLast       = pathLast,             &
              &                         levelDesc      = levelDesc,            &
              &                         proc           = proc,                 &
              &                         computeStencil = computeStencil,       &
              &                         nesting        = nesting + 1 )
          end if
 
          ! if requested halo element haloNesting < found halo element (elemPos)
          ! haloNesting
          if( recvBuf_int( iProc )%val( iElem ) <                              &
            &   levelDesc( haloLevel )%elem%haloNesting%val( elemPos )) then
            levelDesc( haloLevel )%elem%needsUpdate%val( elemPos ) = .true.
            levelDesc( haloLevel )%elem%haloNesting%val( elemPos ) =           &
               min( recvBuf_int( iProc )%val( iElem ),                         &
                    levelDesc( haloLevel )%elem%haloNesting%val( elemPos ) )
          end if
          ! only add, if the element was added locally
          select case( levelDesc( iLevel )%elem%eType%val( elemPos ) ) 
          ! Depending on the type of the element, add to the 
          ! regular buffer, bufferFromCoarser, bufferFromFiner
          case( eT_fluidElem )
            call append( me  = levelDesc( iLevel )%sendbuffer%elemPos( iProc ),&
              &          val = elemPos )
          case( eT_ghostFromCoarserElem ) 
            call append( me  = levelDesc( iLevel )%sendbufferFromCoarser       &
              &                                             %elemPos( iProc ), &
              &          val = elemPos )

            ! for ghostFromCoarserElem determine neighbors of coarser element
            if( levelDesc( haloLevel )%elem%haloNesting%val( elemPos )         &
              &                                        < nestingLimit ) then
              call create_allParentNeighbors(                                  &
                &      targetElem     = elemPos,                               &
                &      nesting        = levelDesc( haloLevel )                 &
                &                            %elem%haloNesting%val( elemPos ), &
                &      level          = iLevel,                                &
                &      tree           = tree,                                  &
                &      computeStencil = computeStencil,                        &
                &      levelDesc      = levelDesc,                             &
                &      pathFirst      = pathFirst,                             &
                &      pathLast       = pathLast,                              &
                &      proc           = proc )
            end if

          case( eT_ghostFromFinerElem )
            call append( me  = levelDesc( iLevel )%sendbufferFromFiner         &
              &                %elemPos( iProc ),                              &
              &          val = elemPos )
          case( eT_distributedGhostFromFinerElem )
             write(logUnit(1),*)' Found distributed ghost From Finer in '//    &
               &                'request remote Halos'
             write(logUnit(1),*)' This case should not occur!'
             call tem_abort()
          end select
        end if ! elemPos > 0
      end do ! ielem recvbuffer
      levelDesc( iLevel )%sendbuffer%nElemsProc( iProc )                       &
        &           = levelDesc( iLevel )%sendbuffer%elemPos( iProc )%nVals 
      levelDesc( iLevel )%sendbufferFromCoarser%nElemsProc( iProc )            &
        &          = levelDesc( iLevel )%sendbufferFromCoarser                 &
        &                               %elemPos( iProc )%nVals 
      levelDesc( iLevel )%sendbufferFromFiner%nElemsProc( iProc )              &
        &          = levelDesc( iLevel )%sendbufferFromFiner                   &
        &                               %elemPos( iProc )%nVals 
    end do
    ! Now each Process knows, which elements to send to others
    !---------------------------------------------------------------------
    write(logUnit(5),*) 'Finished requesting remote halos'

  end subroutine request_remoteHalos
!******************************************************************************!


!******************************************************************************!
  !> Report the actually existing elements, which were requested as halos
  !! from remote
  !! 
  subroutine return_haloCounts( sendbuffer, recvbuffer, proc )
    !---------------------------------------------------------------------------
    !> send buffer
    type(tem_communication_type), intent(in) :: sendbuffer
    !> recv buffer
    type(tem_communication_type), intent(inout) :: recvbuffer
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !---------------------------------------------------------------------------
    integer :: iProc ! process iterator
    integer :: iErr  ! error flag
    integer, parameter :: nVals        = 1 ! number of values to comm. = 1
    integer, parameter :: message_tag = 23 ! arbitrary message flag
    integer, allocatable :: rq_handle(:)
    integer, allocatable :: status(:,:)
    integer :: nCommunications ! amount of communciations
    !---------------------------------------------------------------------------

    nCommunications = sendbuffer%nProcs + recvbuffer%nProcs
    allocate( rq_handle( nCommunications ) )
    allocate( status( mpi_status_size, nCommunications ) )
    !---------------------------------------------------------------------
    !! first report the number of actual halos
    rq_handle(:) = MPI_REQUEST_NULL

    do iProc = 1, recvbuffer%nProcs
      ! Receive the number of actual elements
      call mpi_irecv( recvbuffer%nElemsProc( iProc ),                          &
       &              nVals,                                                   &
       &              mpi_integer,                                             &
       &              recvbuffer%proc(iProc),                                  &
       &              message_tag,                                             &
       &              proc%comm,                                               &
       &              rq_handle(iProc),                                        &
       &              iErr  )
    end do ! iProc

    ! I send the number of actual existing halo elements to the requester.
    do iProc = 1, sendbuffer%nProcs
      call mpi_isend( sendbuffer%nElemsProc( iProc ),                          &
       &              nVals,                                                   &
       &              mpi_integer,                                             &
       &              sendbuffer%proc( iProc ),                                &
       &              message_tag,                                             &
       &              proc%comm,                                               &
       &              rq_handle(iProc + recvbuffer%nProcs),                    &
       &              iErr )
    end do ! iProc

    call mpi_waitall( nCommunications, rq_handle, status, iErr) 
    ! Now we know the number of how actually existent cells, to receive
    ! allocate the recv buffers and communicate
    !---------------------------------------------------------------------

  end subroutine return_haloCounts 
!******************************************************************************!


!******************************************************************************!
  !> Check if additional communications have to be performed
  !! 
  subroutine check_additionalComm( levelDesc, proc, doAdditional, minlevel )
    !---------------------------------------------------------------------------
    !> minlevel in tree
    integer, intent(in) :: minlevel
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc(minlevel:)
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !> do addtional steps to identify neighbors of elems in require list
    ! and do overall communication again, if require%nVals > 0
    logical, intent(out) :: doAdditional
    !---------------------------------------------------------------------------
    integer :: iError, iLevel
    logical :: do_local
    !---------------------------------------------------------------------------
    doAdditional = .false.
    do_local = .false.
    do iLevel = minlevel, ubound( levelDesc, 1 )
      do_local = ( do_local .or. (levelDesc( iLevel )%require%nVals > 0))
    end do
    ! JUROPA work-around for crash in the mpi_allreduce
    call mpi_barrier( proc%comm, iError )
    call mpi_allreduce( do_local, doAdditional, 1, mpi_logical,                &
      &                 mpi_LOR, proc%comm, iError )
    if( doAdditional ) then
      write(dbgUnit(1),*)'Perform additional exchanges for required neighbors' 
    end if

  end subroutine check_additionalComm
!******************************************************************************!
 

!******************************************************************************!
  !> Report the actually existing elements, which were requested as halos
  !! from remote
  !! 
  subroutine redefine_halos( levelDesc, sendbuffer, recvbuffer, proc,           &
    &                        commPattern, computeStencil )
    !---------------------------------------------------------------------------
    !> the level descriptor of specific level
    type(tem_levelDesc_type), intent(inout) :: levelDesc
    !> send and receive communication buffer type
    type(tem_communication_type), intent(inout) :: sendbuffer, recvbuffer
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !> communication pattern
    type(tem_commPattern_type)    :: commPattern
    !> array of all stencils used in the simulation
    type(tem_stencilHeader_type)  :: computeStencil(:)
    !---------------------------------------------------------------------------
    integer :: iProc, tPos, iElem, nElems
    integer :: elemCount
    integer(kind=long_k), allocatable :: recvBuf(:)
    integer :: message_flag
    logical :: wasAdded
    !---------------------------------------------------------------------------
    message_flag = 3

    if( allocated( sendbuffer%buf_long)) deallocate( sendbuffer%buf_long )
    allocate( sendbuffer%buf_long( sendbuffer%nProcs ))

    if( allocated( recvbuffer%buf_long)) deallocate( recvbuffer%buf_long )
    allocate( recvbuffer%buf_long( recvbuffer%nProcs ))
    if( allocated( recvBuffer%elemPos)) deallocate( recvBuffer%elemPos)
    allocate( recvbuffer%elemPos( recvbuffer%nProcs ))

    allocate( recvBuf(sum(recvbuffer%nElemsProc)) )

    elemCount = 0
    do iProc = 1, recvbuffer%nProcs
      nElems = recvbuffer%nElemsProc(iProc)
      call init(  me     = recvBuffer%elemPos( iProc ),                        &
        &         length = nElems )
      do iElem=1,nElems
        elemCount = elemCount + 1
        call append( me  = recvbuffer%elemPos( iProc ),                        &
          &          val = elemCount )
      end do
    end do

    if (tem_logging_isActive(main_debug%logger, 10)) then
      write(dbgUnit(0),*) '  SEND BUFFER '
      call tem_comm_dumpType( me = sendbuffer,    &
        &                     nUnit = dbgUnit(10) )
      write(dbgUnit(0),*) ' '
      write(dbgUnit(0),*) '  RECV BUFFER '
      call tem_comm_dumpType( me = recvbuffer,    &
        &                     nUnit = dbgUnit(10) )
    end if

    !---------------------------------------------------------------------
    ! initialize receive buffer
    do iProc = 1, recvbuffer%nProcs
      call commPattern%initBuf_long( me    = recvBuffer%buf_long( iProc ),     &
        &                            pos   = recvBuffer%elemPos( iProc )%val,  &
        &                            nVals = recvBuffer%nElemsProc( iProc ))
    end do

    ! initialize send buffer
    do iProc = 1, sendbuffer%nProcs
      call commPattern%initBuf_long( me    = sendBuffer%buf_long( iProc ),     &
        &                            pos   = sendBuffer%elemPos( iProc )%val,  &
        &                            nVals = sendBuffer%nElemsProc( iProc ))
    end do

    call commPattern%exchange_long( send         = sendbuffer,                 &
      &                             recv         = recvbuffer,                 &
      &                             state        = recvBuf,                    &
      &                             message_flag = message_flag,               &
      &                             proc         = proc,                       &
      &                             send_state   = levelDesc%elem%tID%val )

    ! Now I received all the halos, which are really existing
    ! and required on the remote processes
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! Add all received elements 
    ! Set the eType accordingly or leave as nonExisting if it was not received
    elemCount = 0
    do iProc = 1, recvbuffer%nProcs
      nElems = recvBuffer%nElemsProc( iProc )
      call commPattern%finBuf_long(me = recvbuffer%buf_long(iProc))
      call init(  me     = recvBuffer%elemPos( iProc ),                        &
        &         length = nElems )
      do iElem = 1, nElems
        elemCount = elemCount + 1
        call append( me         = levelDesc%elem,                              &
          &          tID        = recvBuf( elemCount ),                        &
          &          sourceProc = recvbuffer%proc( iProc )+1,                  &
          &          eType      = eT_haloElem,                                 &
          &          pos        = tPos,                                        &
          &          wasAdded   = wasAdded )
        if( wasAdded ) then
          ! New halo element received
        else
          ! If it was not added, it was there before.
          ! Then simply set the element type to halo (was initialized with
          ! eT_nonExistingElem above )
          levelDesc%elem%eType%val( tPos ) = eT_haloElem
        end if
        call append( me  = recvbuffer%elemPos( iProc ),                        &
          &          val = tPos )
      end do
      call commPattern%initBuf_long( me    = recvbuffer%buf_long( iProc ),     &
        &                            pos   = recvbuffer%elemPos( iProc )%val,  &
        &                            nVals = recvBuffer%nElemsProc( iProc ))
    end do

    deallocate(recvBuf)

    ! Allocate integer buffers
    if( allocated( recvbuffer%buf_int )) deallocate( recvbuffer%buf_int )
    allocate( recvbuffer%buf_int( recvbuffer%nProcs ))
    do iProc = 1, recvbuffer%nProcs
      call commPattern%initBuf_int( me    = recvbuffer%buf_int( iProc ),       &
        &                           pos   = recvbuffer%elemPos( iProc )%val,   &
        &                           nVals = recvBuffer%nElemsProc( iProc ))
    end do
    if( allocated( sendbuffer%buf_int )) deallocate( sendbuffer%buf_int )
    allocate( sendbuffer%buf_int( sendbuffer%nProcs ))
    do iProc = 1, sendbuffer%nProcs
      call commPattern%initBuf_int( me    = sendbuffer%buf_int( iProc ),       &
        &                           pos   = sendbuffer%elemPos( iProc )%val,   &
        &                           nVals = sendBuffer%nElemsProc( iProc ))
    end do
    ! added and initialized the receive buffers for extended properties of the
    ! halos
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! communicate the properties of the halos
    call commPattern%exchange_long(                                            &
      &                         send         = sendbuffer,                     &
      &                         recv         = recvbuffer,                     &
      &                         state        = levelDesc%elem%property%val(:), &
      &                         message_flag = 5,                              &
      &                         proc         = proc )

    ! communicate the stencil neighbors
    ! the stencil which includes the boundary information
    if ( recvbuffer%nProcs > 0 .or. sendbuffer%nProcs > 0 ) then
      ! communicate only fluid stencil which is defined for all
      ! elements in elem list including halo elements.
      ! In our case, fluid stencil is 1st Stencil so setting iStencil=1
      call tem_stencil_communicate( send           = sendbuffer,             &
        &                           recv           = recvbuffer,             &
        &                           elem           = levelDesc%elem,         &
        &                           proc           = proc,                   &
        &                           commPattern    = commPattern,            &
        &                           computeStencil = computeStencil(1),      &
        &                           iStencil       = 1 )
    end if

    write(logUnit(5),*) '  Done communicating stencil buffer'
    !---------------------------------------------------------------------

  end subroutine redefine_halos
!******************************************************************************!


!******************************************************************************!
  !> Communicate the complete stencil
  !!
  !! Currently, this assumes same stencils for all participating elements
  !!
  subroutine tem_stencil_communicate( send, recv, elem, computeStencil,        &
    &                                 proc, commPattern, iStencil )
    !---------------------------------------------------------------------------
    !> send and recv communication buffers
    type( tem_communication_type ), intent(inout) :: send, recv
    !> communication pattern
    type(tem_commPattern_type), intent(in)     :: commPattern
    !> levelDesc element list
    type(tem_element_type), intent(inout)  :: elem
    !> Process description to use.
    type(tem_comm_env_type), intent(in) :: proc
    !> array of all stencils used in the simulation
    type(tem_stencilHeader_type), intent(in) :: computeStencil
    !> amount of values to communicate
    integer, intent(in) :: iStencil
    !---------------------------------------------------------------------------
    integer :: iStencilElem, iElem, nElems
    integer(kind=long_k), allocatable :: buffer(:)
    type(tem_stencilElement_type ) :: tStencil
    integer :: addedPos, iProc, elemPos, neighPos, stencilPos
    logical :: wasAdded
    !---------------------------------------------------------------------------
    ! Number of local elements in element list
    nElems = elem%tID%nVals
    write(dbgUnit(10),*)'in stencil communicate, nElems: ', nElems

    ! buffer which sends stencil neighbors to remove process
    allocate( buffer( nElems ))

    ! Set the temporary (empty) stencil
    call init( me        = tStencil,                                           &
      &        QQN       = computeStencil%QQN ,                                &
      &        headerPos = iStencil ) 

    ! add stencil headerPos for all halo elements depends on requested stencil
    ! communication
    do iProc = 1, recv%nProcs
      do iElem = 1, recv%nElemsProc( iProc )
        elemPos = recv%elemPos( iProc )%val( iElem )
        if( elem%stencil%val( elemPos )%nVals > 0 ) then
          ! \todo: SZ: this does not make sense since all elements have a stencil
          !            elem%stencil%val( elemPos )%nVals .gt. 0 for all elements
          ! Find the stencil corresponding to the current one ( iStencil )
          stencilPos = tem_stencil_getHeaderPos(                               &
            &                               me  = elem%stencil%val( elemPos ), &
            &                               val = iStencil )
          ! if none was found, use the most appropriate one (headerPos = 0)
          ! this assigns the first stencil to the halo element
          ! why not assign it right away to be 1?
          if( stencilPos <= 0 )then
            stencilPos = tem_stencil_getHeaderPos(                             &
              &                             me  = elem%stencil%val( elemPos ), &
              &                             val = 0 )
            elem%stencil%val( elemPos )%val( stencilPos )%headerPos = iStencil
          end if
        else
          ! if stencil is not initialized for this element append tStencil
          call append( me = elem%stencil%val( elemPos ), val = tStencil )
        end if
      end do
    end do

    !--------------------------------------------------------------------------
    ! Do the communication for each stencil tID entry
    do iStencilElem = 1, computeStencil%QQN
      ! Fill the send buffer
      do iProc = 1, send%nProcs
        do iElem = 1, send%nElemsProc( iProc )
          elemPos = send%elemPos( iProc )%val( iElem )
          neighPos = elem%stencil%val( elemPos )%val( iStencil )%              &
            &                                            tIDpos( iStencilElem )
          buffer( elemPos ) = elem%neighID%val( elemPos )%val( neighPos )
        end do ! iElem send
      end do ! iProc send

      ! Do the exchange between all participating processes
      call commPattern%exchange_long( send         = send,                     &
        &                             recv         = recv,                     &
        &                             state        = buffer,                   &
        &                             message_flag = 0,                        &
        &                             proc         = proc )

      ! Read from the buffer
      do iProc = 1, recv%nProcs
        do iElem = 1, recv%nElemsProc( iProc )
          elemPos = recv%elemPos( iProc )%val( iElem )
          ! Find the stencil corresponding to the current one (iStencil )
          stencilPos = tem_stencil_getHeaderPos(                               &
            &                               me  = elem%stencil%val( elemPos ), &
            &                               val = iStencil )

          if( stencilPos > 0 ) then 
            ! Only append elements, if there is only one stencil for the element
            ! Append the treeID to the element's neighIDs
            call append( me       = elem%neighID%val( elemPos ),               &
              &          val      = buffer( elemPos),                          &
              &          pos      = addedPos,                                  &
              &          wasAdded = wasAdded )
            ! ... and store the appended position to the stencil
            elem%stencil%val( elemPos )%val( stencilPos )%                     &
              &                               tIDpos( iStencilElem ) = addedPos
          end if
        end do ! iElem recv
      end do ! iProc recv

    end do ! iStencilElem
    
  end subroutine tem_stencil_communicate
!******************************************************************************!


!******************************************************************************!
  !> Update the found dependencies, which were built for non-ordered lists
  !! Out of fluid, ghost and halo lists, the totalList is constructed in an
  !! ordered fashion. The element order as in the TotalList is later passed on 
  !! to the solver.
  !!
  subroutine update_elemPosToTotalPos( levelDesc, levelPointer, tree,          &
    &                                  computeStencil )
    !---------------------------------------------------------------------------
    !> the global tree
    type(treelmesh_type), intent(in) :: tree
    !> Pointer from original treeID list to level wise fluid list
    integer, intent(inout) :: levelPointer(:)
    !> the level descriptor to be filled
    type(tem_levelDesc_type), intent(inout) :: levelDesc(tree%global%minlevel:)
    !> array of all stencils used in the simulation
    type(tem_stencilHeader_type), intent(inout) :: computeStencil(:)
    !---------------------------------------------------------------------------
    integer :: iLevel, iElem, iStencil, iStencilElem, pos
    integer :: nUpdates, iError
    integer(kind=long_k) :: nTreeID
    logical :: needUpdate
    !---------------------------------------------------------------------------
    write(logUnit(3),*) 'Updating dependencies ...'

    ! Update the levelPointer
    ! KM: Do this step only since since levelPointer is not used within 
    ! tem_find_allElements. Commented out its dublication in build_levelElements
    do iElem = 1, tree%nElems
      iLevel = tem_levelOf( tree%treeID( iElem ))
      levelPointer( iElem ) = tem_PositionInSorted(                            &
        &                      me    = levelDesc(iLevel )%total,               &
        &                      val   = tree%treeID( iElem ),                   &
        &                      upper = levelDesc( iLevel )%nElems_fluid )
    end do

    ! Update the element positions from the position in the original 
    ! elem list to the total list 
    do iLevel = tree%global%minlevel, tree%global%maxlevel
      call update_buffer_elemPos( buffer    = levelDesc( iLevel )%sendbuffer,  &
        &                         levelDesc = levelDesc( iLevel ),             &
        &                         iError    = iError )
      if (iError > 0) &
        write(dbgUnit(1),*)  'error in sendbuffer ',iError,' l',iLevel

      call update_buffer_elemPos( buffer    = levelDesc( iLevel )%recvbuffer,  &
        &                         levelDesc = levelDesc( iLevel ),             &
        &                         iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in recvbuffer ',iError,' l',iLevel

      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%sendbufferFromCoarser,    &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in sendFCbuff ',iError,' l',iLevel

      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%recvbufferFromCoarser,    &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )      
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in recvFCbuff ',iError,' l',iLevel

      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%sendbufferFromFiner,      &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in sendFFbuff ',iError,' l',iLevel

      call update_buffer_elemPos(                                              &
        &            buffer    = levelDesc( iLevel )%recvbufferFromFiner,      &
        &            levelDesc = levelDesc( iLevel ),                          &
        &            iError    = iError )
      if( iError .gt. 0 ) &
        write(dbgUnit(1),*) 'error in recvFFbuff ',iError,' l',iLevel
    end do ! iLevel

    write(logUnit(4),*) 'Updating the stencil entries'
    ! update stencil elem entry from original treeID list to levelwise fluid 
    ! list
    do iStencil = 2, size( computeStencil )
      if( .not. computeStencil( iStencil )%useAll ) then
        do iElem = 1, computeStencil( iStencil )%nElems
          computeStencil( iStencil )%elem( iElem ) =                           &
            &   levelPointer( computeStencil( iStencil )%elem( iElem ))
        end do
      end if
    end do

    ! Update the totalPos in the element-stencils to the sorted total list
    write(logUnit(4),*) ' Updating entries in the element stencils...'
    nUpdates =  0
    do iLevel = tree%global%minlevel, tree%global%maxlevel
      do iElem = 1, levelDesc(iLevel)%elem%tID%nVals
        do iStencil = 1, levelDesc( iLevel )%elem%stencil%val( iElem )%nVals
          do iStencilElem = 1, levelDesc( iLevel )%elem%stencil%               &
            &                                  val( iElem )%val( iStencil )%QQN
            needUpdate = .false.
            pos = levelDesc( iLevel )%elem%stencil%val( iElem )%               &
              &                        val( iStencil )%totalPos( iStencilElem )
            if( pos > 0 ) then
              nTreeID = levelDesc( iLevel )%elem%tID%val( pos )
              if( pos > levelDesc( iLevel )%nElems ) needUpdate = .true.
              if( .not. needUpdate ) then
                if( levelDesc( iLevel )%total( pos ) /= nTreeID )            &
                  &                                         needUpdate = .true.
              end if
              if( needUpdate ) then
                ! Totalpos needs update!
                nUpdates = nUpdates + 1
                levelDesc( iLevel )%elem%stencil%val( iElem )%                 &
                  &                   val( iStencil )%totalPos( iStencilElem ) &
                  &  = tem_treeIDinTotal( nTreeID, levelDesc( iLevel ))
              end if
            end if
          end do
        end do
      end do
    end do
    write(logUnit(3),*) '     Updated nEntries = ', nUpdates
    write(logUnit(3),*) ' Finished updating dependencies'

   end subroutine update_elemPosToTotalPos
!******************************************************************************!


!******************************************************************************!
  !> Update the position of the elements inside the buffers from the original
  !! tID list to the later totalList
  !!
  subroutine update_buffer_elemPos( buffer, levelDesc, iError )
    !---------------------------------------------------------------------------
    !> communication buffer
    type(tem_communication_type), intent(inout) :: buffer
    !> levelDesc to be used
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !> return encountered error
    integer, intent(out) :: iError
    !---------------------------------------------------------------------------
    integer :: iElem, iProc, elemPos, iVal
    !---------------------------------------------------------------------------
    iError = 0
    do iProc = 1, buffer%nProcs
      do iElem = 1, buffer%elemPos( iProc )%nVals
        elemPos = buffer%elemPos( iProc )%val( iElem ) 
        ! replace buffer elemPos from levelDesc%elem%tID list to 
        ! levelDesc%total list
        buffer%elemPos( iProc )%val( iElem ) =                                 &
          &   tem_treeIDinTotal( tID       = levelDesc%elem%tID%val( elemPos ),&
          &                      eType     = levelDesc%elem%eType%val(elemPos),&
          &                      levelDesc = levelDesc )

        if( buffer%elemPos( iProc )%val( iElem ) < 1 ) then
          write(dbgUnit(2),*) ' tID ', levelDesc%elem%tID%val(elemPos),        &
            &                ' in buffer not found' 
          iError = iProc*10000000+iElem
          do iVal = 1, levelDesc%elem%tID%nVals
            if( levelDesc%elem%tID%val( iVal ) == levelDesc%elem%tID%          &
              &                                            val( elemPos )) then
              write(dbgUnit(2),*)'found at', iVal
            end if
          end do
          do iVal = 1, size(levelDesc%total)
            if( levelDesc%total( iVal ) .eq. levelDesc%elem%tID%               &
              &                                            val( elemPos )) then
              write(dbgUnit(2),*) 'found in total at', iVal
            end if
          end do
        end if ! if elemPos < 1

      end do !iElem
    end do !iProc
 
  end subroutine update_buffer_elemPos
!******************************************************************************!


!******************************************************************************!
  !> add here the dependency for interpolation between the levels
  !! For each target cell, there are one or more source cells.
  !! The source cell can be of type fluid, ghost or halo.
  !! We save the type to update the correct element position later on,
  !! when the lists have been assembled.
  !!
  subroutine appendGhostDependency( sourcePos, sourceLevel, tgtDep )
    !---------------------------------------------------------------------------
    !> position of the source cell within its list inside the
    integer, intent(in) :: sourcePos
    !> level of the source ghost cell
    integer, intent(in) :: sourceLevel
    !> array of level descriptors which holds the informations
    !! for all the levels. this methods adds new informations to
    !! levelDesc(:)%source -> inout
    type(depSource_type), intent(inout) :: tgtDep
    !---------------------------------------------------------------------------

    ! check if the current dependency list already contains information
    if( tgtDep%elem%nVals == 0 ) then
      call init( me=tgtDep%elem, length = 8 )
      tgtDep%dependencyLevel = sourceLevel
    end if

    ! append the new entry to the dependency list at the end+1
    call append( me  = tgtDep%elem,                                            &
         &       val = sourcePos )

  end subroutine appendGhostDependency
!******************************************************************************!


!******************************************************************************!
  !> Convert a non-zero stencil direction {-1,1} to child coordinate {0,1}
  !!
  elemental function stencilToChild( stencilCoord ) result( childCoord)
    !---------------------------------------------------------------------------
    integer, intent(in) :: stencilCoord
    integer :: childCoord
    !---------------------------------------------------------------------------

    childCoord = (stencilCoord + 1)/2

  end function stencilToChild
!******************************************************************************!


!******************************************************************************!
  !> Convert a child coordinate {0,1} to non-zero stencil direction {-1,1}
  !!
  function childToStencil( childCoord ) result( stencilCoord)
    !---------------------------------------------------------------------------
    integer, intent(in) :: childCoord
    integer :: stencilCoord
    !---------------------------------------------------------------------------

    stencilCoord = (childCoord*2) -1

  end function childToStencil
!******************************************************************************!


!******************************************************************************!
  !> Update the link into a given direction, based on the childs neighbor
  !! relations.
  !! Define here the trumping rule to decide, which of the neighbors or
  !! boundarie is taken for the ghostFromFiner element
  !!
  subroutine update_childNeighborID( neighID, childCoord, childPos,            &
    &                                iStencilElem, levelDesc, iStencil)
    !---------------------------------------------------------------------------
    !> neighID for coarser
    integer(kind=long_k),intent(inout) :: neighID
    !> child coordinates
    integer, intent(in) :: childCoord(4)
    !> position of childIds in levelDesc elem tID list
    integer, intent(in) :: childPos(8)
    !> current stencil direction
    integer, intent(in) :: iStencilElem
    !>
    integer, intent(in) :: iStencil
    !>
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !---------------------------------------------------------------------------
    integer :: childID, stencilTIDpos
    !---------------------------------------------------------------------------

    childID = int(tem_idOfCoord( childCoord ))
    ! childPos holds the positions of the child treeIDs in levelDesc%elem
    ! only if the current child element really exists, get information from its
    ! stencil
    if (childPos(childID) > 0) then
      ! if child at current position exists,
      ! take the max tID neighbor parent or the
      ! min boundaryID -> trumping rule (bIDs are stored as negative integers)
      stencilTIDpos = levelDesc%elem%stencil%val( childPos( childID ))         &
        &            %val(iStencil)%tIDpos( iStencilElem )
      neighID = max( neighID,                                                  &
        & tem_parentOf( levelDesc%elem%neighID%val( childPos( childID ))%val(  &
        &            stencilTIDpos )))
    else
    end if
  end subroutine update_childNeighborID
!******************************************************************************!


!******************************************************************************!
  !> write out the complete list of elements of a given level
  subroutine tem_elemList_dump( me, nUnit, string, stencil, stencilPos, &
    &                           compact )
    !---------------------------------------------------------------------------
    type( tem_element_type ), intent(in) :: me
    integer, intent(in) :: nUnit
    character(len=*), intent(in) :: string
    logical, intent(in), optional :: stencil
    logical, intent(in), optional :: stencilPos
    logical, intent(in), optional :: compact
    !---------------------------------------------------------------------------
    integer :: iElem
    !---------------------------------------------------------------------------
    write(nUnit, *) '========================================================='
    write(nUnit, *) '==               ', trim(string)
    write(nUnit, *) '========================================================='
    do iElem = 1, me%tID%nVals
      call tem_element_dump( me         = me,                                  &
        &                    elemPos    = iElem,                               &
        &                    nUnit      = nUnit,                               &
        &                    header     = (mod(iElem,32)==1),                  &
        &                    compact    = compact,                             &
        &                    stencil    = stencil,                             &
        &                    stencilPos = stencilPos)
    end do
    write(nUnit, *) '==  DONE         ', trim(string)
    write(nUnit, *) '========================================================='

  end subroutine tem_elemList_dump
!******************************************************************************!


!******************************************************************************!
  !> write out the complete list of elements of a given level
  !!
  subroutine tem_require_dump( me, nUnit, string, compact )
    !---------------------------------------------------------------------------
    type( dyn_longArray_type ), intent(in) :: me
    integer, intent(in) :: nUnit
    character(len=*), intent(in) :: string
    logical, intent(in), optional :: compact
    !---------------------------------------------------------------------------
    integer :: iElem
    integer :: nEntries, iLoop, iEntry
    character(len=pathLen) :: buffer
    !---------------------------------------------------------------------------
    nEntries = 15
    write(nUnit, *) '========================================================='
    write(nUnit, *) '==      REQUIRE  ', trim(string)
    write(nUnit, *) '========================================================='
    iElem = 0
    do iLoop = 1, ceiling( real(me%nVals)/real(nEntries) )
      buffer = ''
      do iEntry = 1, nEntries
        iElem = iElem + 1
        if( iElem .ge. me%nVals ) exit
        write(buffer, '(a, i10)') trim(buffer), me%val( iElem )
      end do
      write(nUnit,*) trim(buffer)
      if( iElem .ge. me%nVals ) exit
    end do
    write(nUnit, *) '==  DONE         ', trim(string)
    write(nUnit, *) '========================================================='

  end subroutine tem_require_dump
!******************************************************************************!


!******************************************************************************!
  !> output all level-wise treeIDs in a clean way
  !!
  subroutine tem_dumpTreeIDlists( levelDesc, levelPointer, minlevel )
    !---------------------------------------------------------------------------
    !> minimum level of the global tree
    integer, intent(in) :: minlevel
    !> level pointer from the original tree to the level wise total list
    integer, intent(in) :: levelPointer(:)
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    write(dbgUnit(2),*) '' 
    write(dbgUnit(2),*) '' 
    write(dbgUnit(2),*) '================================================'     &
      &                // '=================' 
    write(dbgUnit(2),*) '=======            A L L  TreeID  L I S T S     '     &
      &                // '      ==========' 
    write(dbgUnit(2),*) '================================================'     &
      &                // '=================' 
    do iLevel = minLevel, minLevel+size(levelDesc)-1

      if (size(levelDesc( iLevel )%total) .gt. 0) then
        write(dbgUnit(3),*) 'Level ', iLevel
        write(dbgUnit(3),*) '=============================================='   &  
          &                // '===================' 
        write(dbgUnit(3),*) 'nVals in total list ',                            &
          &                 size(levelDesc(iLevel)%total)
!      call tem_print_array( title = 'total',         &
!                        me = levelDesc(iLevel)%total,&
!                        itemLength=8,                &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%fluid%nvals .gt. 0 ) &
!      call tem_print_array( title = 'fluid',         &
!                        me = levelDesc( iLevel )%fluid%val(          &
!                                1:levelDesc( iLevel )%fluid%nVals ), &
!                        itemLength=8,                                &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%pntTID%nvals .gt. 0 ) &
!      call tem_print_array( title = 'pntTID',                                           &
!                        me = int( levelDesc( iLevel )%pntTID%val(                &
!                                     1:levelDesc( iLevel )%pntTID%nvals ), kind=long_k),         &
!                        itemLength=8,                                               &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%ghostFromCoarser%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromCoarser', &
!                        me = levelDesc( iLevel )%ghostFromCoarser%val(   &
!                                1:levelDesc( iLevel )%ghostFromCoarser%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%ghostFromFiner%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromFiner', &
!                        me = levelDesc( iLevel )%ghostFromFiner%val(   &
!                                1:levelDesc( iLevel )%ghostFromFiner%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%halo%nvals .gt. 0 ) &
!      call tem_print_array( title = 'halo',                                            &
!                          me = levelDesc( iLevel )%halo%val(                      &
!                                  1:levelDesc( iLevel )%halo%nVals ),                &
!                          itemLength=8,                                               &
!                          lineLength = 128 )
      endif
    enddo ! iLevel

  end subroutine tem_dumpTreeIDlists
!******************************************************************************!


!******************************************************************************!
  !> output the found dependencies
  !!
  !! \todo: DEB: Adapt this subroutine or kick it out
  !!
  subroutine tem_debugDependencies( levelDesc, levelPointer, minlevel )
    !---------------------------------------------------------------------------
    !> minimal level
    integer, intent(in) :: minlevel
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc(minlevel:)
    !> level pointer
    integer, intent(in) :: levelPointer(:)
    !---------------------------------------------------------------------------
!    integer :: iLevel
    !---------------------------------------------------------------------------
!  ! Debugging
!  write(dbgUnit,*)
!  write(dbgUnit,*)
!  write(dbgUnit,*) '================================================================='
!  write(dbgUnit,*) '=======            A L L   L I S T S                   =========='
!  write(dbgUnit,*) '================================================================='
!  do iLevel = minlevel, ubound( levelDesc, 1 )
!
!    if(size(levelDesc( iLevel )%total) .gt. 0)then
!    write(dbgUnit,'(a,i3)') 'Level',iLevel
!    write(dbgUnit,'(a)') &
!   '===================================================================='
!    write(dbgUnit,'(a,i8)') 'nvals',size(levelDesc( iLevel )%total)
!    write(dbgUnit,*) '---------------------------------------------------------------'
!    write(dbgUnit,'(a)') 'TOTAL:'
!
!    write(dbgUnit,*) '---------------------------------------------------------------'
!    !write(dbgUnit,'(a14,512i8)') 'pntTID:',&
!    !            levelDesc( iLevel )%pntTID%val(1:levelDesc( iLevel )%pntTID%nvals)
!    write(dbgUnit,'(a14,512i8)') 'fluid:',levelDesc( iLevel )%totalPnt(   &
!                1:levelDesc( iLevel )%nElems_fluid)
!    write(dbgUnit,*) '---------------------------------------------------------------'
!    if( levelDesc( iLevel )%ghostFromCoarser%nvals .gt. 0) then
!      write(dbgUnit,'(a14,512i8)') 'ghostFC:',levelDesc( iLevel )%totalPnt(   &
!                1:levelDesc( iLevel )%nElems_ghostFromCoarser)
!      write(dbgUnit,*) &
!          '------------------------------------------------------------------'
!    endif
!    if( levelDesc( iLevel )%ghostFromFiner%nvals .gt. 0) then
!      write(dbgUnit,'(a14,512i8)') 'ghostFF:',levelDesc( iLevel )%totalPnt(   &
!                1:levelDesc( iLevel )%nElems_ghostFromFiner)
!      write(dbgUnit,*) &
!          '------------------------------------------------------------------'
!    endif
!    if( levelDesc( iLevel )%halo%nvals .gt. 0) then
!      write(dbgUnit,'(a14,512i8)') 'halo:',levelDesc( iLevel )%halo%   &
!                val(1:levelDesc( iLevel )%halo%nvals)
!      write(dbgUnit,*) &
!          '------------------------------------------------------------------'
!    endif
! endif
! enddo
!

  end subroutine tem_debugDependencies
!******************************************************************************!


!******************************************************************************!
  !> output all level-wise PropertyBits in a clean way
  !!
  subroutine tem_dumpPropBits( levelDesc, minlevel )
    !---------------------------------------------------------------------------
    !> minimum level of the global tree
    integer, intent(in) :: minlevel
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc(minlevel:)
    !---------------------------------------------------------------------------
    integer :: iLevel
    !---------------------------------------------------------------------------
    write(dbgUnit(2),*)'' 
    write(dbgUnit(2),*)'' 
    write(dbgUnit(2),*) '================================================' &
      &                // '=================' 
    write(dbgUnit(2),*) '======        Property Bits Lists               ' &
      &                // '=================' 
    write(dbgUnit(2),*) '================================================' &
      &                // '=================' 
    do iLevel = minLevel, minLevel+size(levelDesc)-1

!      if(size(levelDesc( iLevel )%ttlProp) .gt. 0)then
!      write(dbgUnit,'(a,i3)') 'Level',iLevel
!      write(dbgUnit,'(a)') &
!     '===================================================================='
!      write(dbgUnit,'(a,i8)') 'nvals in total ',size(levelDesc( iLevel )%ttlProp)
!      call tem_print_array( title = 'total property',  &
!                        me = levelDesc(iLevel)%ttlProp,&
!                        itemLength=8,                &
!                        lineLength = 128 )
!
!      if( levelDesc( iLevel )%fldProp%nvals .gt. 0 ) &
!      call tem_print_array( title = 'fluid Property',         &
!                        me = levelDesc( iLevel )%fldProp%val(          &
!                                1:levelDesc( iLevel )%fluid%nVals ), &
!                        itemLength=8,                                &
!                        lineLength = 128 )

!    write(dbgUnit,'(a14,512i8)') 'sort fluid:',levelDesc( iLevel )%fluid%sorted(  &
!                1:levelDesc( iLevel )%fluid%nvals)
!    write(dbgUnit,'(a14,512i8)') 'invsort fluid:',levelDesc( iLevel )%invSorted(   &
!                1:levelDesc( iLevel )%fluid%nvals)
!      if( levelDesc( iLevel )%ghostFromCoarser%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromCoarser', &
!                        me = levelDesc( iLevel )%ghostFromCoarser%val(   &
!                                1:levelDesc( iLevel )%ghostFromCoarser%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%ghostFromFiner%nvals .gt. 0 ) &
!      call tem_print_array( title = 'ghostFromFiner', &
!                        me = levelDesc( iLevel )%ghostFromFiner%val(   &
!                                1:levelDesc( iLevel )%ghostFromFiner%nvals),&
!                        itemLength=8,     &
!                        lineLength = 128 )
!      if( levelDesc( iLevel )%halo%nvals .gt. 0 ) &
!      call tem_print_array( title = 'halo property',                  &
!                          me = levelDesc( iLevel )%haloProp%val(      &
!                                  1:levelDesc( iLevel )%halo%nVals ), &
!                          itemLength=8,                               &
!                          lineLength = 128 )
!      endif
    enddo ! iLevel

  end subroutine tem_dumpPropBits
!******************************************************************************!


!******************************************************************************!
  !> This routine updates the property bits in the tree with those of the
  !! level descriptor.
  !!
  subroutine tem_updateTree_properties( levelDesc, globTree )
    !---------------------------------------------------------------------------
    !> level descriptor
    type(tem_levelDesc_type), intent(in) :: levelDesc
    !> global tree
    type(treelmesh_type), intent(inout) :: globTree
    !---------------------------------------------------------------------------
    ! counter
    integer :: iElem
    !---------------------------------------------------------------------------

    do iElem = 1, levelDesc%nElems_fluid
      if( globTree%ElemPropertyBits( levelDesc%pntTID(iElem)) .ne.             &
        & levelDesc%property( iElem ))then
        globTree%ElemPropertyBits( levelDesc%pntTID( iElem )) =                &
          &                                         levelDesc%property( iElem )

        ! set the global mesh changed tag to be true such that the mesh is
        ! dumped
        globTree%global%meshChange = .true.
      end if
    end do

  end subroutine tem_updateTree_properties
!******************************************************************************!


end module tem_construction_module
!******************************************************************************!
!> \page tem_construction_example_stencil Example for the stencil construction
!!
!! To build the stencil for a given element, all required 
!! \ref treelmesh_module::treelmesh_type::treeid "treeIDs" can be obtained with
!! the following procedure:
!! First the integer coordinate tuple \f$(x,y,z,L)\f$ of the
!! \ref treelmesh_module::treelmesh_type::treeid "treeID", for which the stencil
!! has to be obtained, is computed.
!! Then for each stencil element the 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" is obtained by the 
!! following steps:
!! - Add the stencil offset to the coordinate of the central element to obtain
!!   the coordinate \f$(x + s_x, y + s_y, z + s_z, L)\f$ for the stencil element
!! - Convert the obtained coordinate back to a
!!   \ref treelmesh_module::treelmesh_type::treeid "treeID".
!!
!! To illustrate this procedure with a specific example, lets consider a two
!! dimensional mesh.
!! We would like to find the right neighbor of the element with 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" \f$ 40 \f$.
!! An illustration of the geometrical layout for this example is given in the 
!! following [Figure](@ref fig_rightNeighbor).
!! As we are looking at a two dimensional mesh, there are \f$ 4^L \f$ elements on
!! each refinement level \f$ L \f$.
!!
!! \anchor fig_rightNeighbor
!! ![Illustration of the right neighbor identification](find_rightNeighbor.png)
!! 
!! The first information, that is needed, is the level of the given 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID".
!! To find this, and the Morton index on that level, we subtract the elements
!! on coarse levels, until the reduced index is smaller than the next number of
!! elements to subtract:
!! 
!! - Level 0: \f$40 - 4^0 = 39\f$
!! - Level 1: \f$39 - 4^1 = 35\f$
!! - Level 2: \f$35 - 4^2 = 19\f$
!! - Level 3: \f$4^3 > 19\f$
!! - \ref treelmesh_module::treelmesh_type::treeid "treeID" \f$40\f$ is located
!!   on Level 3 and has a Morton index of \f$19\f$.
!! 
!! The binary representation of \f$19\f$ is \f$010011_b\f$.
!! Due to the bit-interleaving rule, odd bits correspond to the X coordinate and
!! even bits correspond to the Y coordinate.
!! Thus we obtain for X \f$101_b = 5\f$ and for Y \f$001_b = 1\f$.
!! Now we found the coordinate tuple \f$(x, y, L) = (5, 1, 3)\f$ of
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" \f$40\f$ and can
!! simply add the offset \f$(1, 0, 0)\f$, that represents the right neighbor.
!! By this we obtain the coordinate of this neighbor to be \f$(6, 1, 3)\f$,
!! resulting in a binary representation of \f$(110_b, 001_b)\f$.
!! After interleaving the bits, the Morton index of the right neighbor is found
!! to be \f$010110_b = 22\f$
!! Finally, by adding the offset, we obtain the 
!! \ref treelmesh_module::treelmesh_type::treeid "treeID" of the right neighbor
!! as \f$22 + 4^0 + 4^1 + 4^2 = 43\f$.
!! Note, that there is no restriction on the offsets, that are to be applied in
!! the neighborhood search, and due to the periodic universe all offset
!! locations are well defined.
